//
//
// Description: Class to set Boundary Conditions
//
//
// Author:  <Gauthier BECKER>, (C) 2010
//
// Copyright: See COPYING file that comes with this distribution
//
//
// Has to be grouped with BoundaryCondition defined in elasticitySolver.h These BC have to be defined in a separated file
// I add nonLinear above to avoid ambiguities
#ifndef NONLINEARBC_H_
#define NONLINEARBC_H_
#ifndef SWIG
#include "groupOfElements.h"
#include "simpleFunction.h"
#include "SVector3.h"
#include "quadratureRules.h"
#include "timeFunction.h"
#include "contactFunctionSpace.h"
#include "solverAlgorithms.h" // change this
#include "pbcVertex.h"
#include "nonLinearMicroBC.h"
#endif
class partDomain;

class nonLinearBoundaryCondition
{
 public:
  enum whichCondition{position=0, velocity=1, acceleration=2};
  enum location{UNDEF,ON_VERTEX,ON_EDGE,ON_FACE,ON_VOLUME,RIGIDCONTACT,ON_FACE_VERTEX};
  enum type{DIRICHLET, NEUMANN,INITIAL,PERIODIC,CONSTRAINT,RCONTACT,SAMEDISP,FIXONFACE,ALIGN};
  #ifndef SWIG
  location onWhat; // on vertices or elements
  int _tag; // tag for the dofManager or Physical number of BC
  groupOfElements *g; // support for this BC
  nonLinearBoundaryCondition() : g(NULL),_tag(0),onWhat(UNDEF) {};
  nonLinearBoundaryCondition(const nonLinearBoundaryCondition &source){
    this->onWhat = source.onWhat;
    this->_tag = source._tag;
    this->g = source.g;
  }
  virtual ~nonLinearBoundaryCondition()
  {
    if (g!=NULL) {delete g; g = NULL;};
  }
  virtual type getType()const=0;
#endif
};
#ifndef SWIG

class nonLinearDirichletBC : public nonLinearBoundaryCondition
{
 public:
  int _comp; // component
  simpleFunctionTime<double> *_f;
  /* BEAWARE the four following vectors have to be filled in the same time */
  std::vector<FunctionSpaceBase*> _vspace;
  std::vector<const groupOfElements*> _vgroup;
  std::vector<FilterDof*> _vfilter;
  std::vector<bool> _vdg; // true if discontinuous galerkin formulation (ie Dof entity = element number)
  nonLinearBoundaryCondition::whichCondition _mycondition;
  virtual nonLinearBoundaryCondition::type getType()const{return nonLinearBoundaryCondition::DIRICHLET;}
  nonLinearDirichletBC() : nonLinearBoundaryCondition(),_comp(0),_f(0), _mycondition(nonLinearBoundaryCondition::position){}
  nonLinearDirichletBC(const nonLinearDirichletBC &source) : nonLinearBoundaryCondition(source){
    this->_comp = source._comp;
    this->_f = source._f;
    this->_vspace = source._vspace;
    this->_vgroup = source._vgroup;
    this->_vfilter = source._vfilter;
    this->_vdg = source._vdg;
    this->_mycondition = source._mycondition;
  }
	virtual ~nonLinearDirichletBC(){
		this->clear();
	}
  void clear()
  {
    for(int i=0;i<_vspace.size();i++)
    {
      delete _vspace[i];
      delete _vfilter[i];
    }
    _vspace.clear();
    _vgroup.clear();
    _vfilter.clear();
    _vdg.clear();
  }
  bool isConstrained(MVertex* v) const {
    for (groupOfElements::vertexContainer::iterator it = g->vbegin(); it != g->vend(); it++){
      if (v== (*it)){
        return true;
      }
    }
    return false;
  };
};

class nonLinearDirichletBCAtCorner : public nonLinearBoundaryCondition{
  public:
    int _cornerNumber;
    int _comp;
    std::vector<Dof> constraintDofs;
    simpleFunctionTime<double> *_f;
    
  public:
    nonLinearDirichletBCAtCorner():nonLinearBoundaryCondition(){
      onWhat = ON_VERTEX;
      _f = NULL;
    }
    nonLinearDirichletBCAtCorner(const nonLinearDirichletBCAtCorner& src): nonLinearBoundaryCondition(src),_cornerNumber(src._cornerNumber),
          _comp(src._comp),constraintDofs(src.constraintDofs),_f(src._f){}
    virtual ~nonLinearDirichletBCAtCorner(){
      if (_f != NULL) delete _f;
    }
    virtual nonLinearBoundaryCondition::type getType()const{return nonLinearBoundaryCondition::DIRICHLET;}
};

class nonLinearNeumannBCAtCorner : public nonLinearBoundaryCondition{
  public:
    int _cornerNumber;
    int _comp;
    std::vector<Dof> constraintDofs;
    simpleFunctionTime<double> *_f;
    
  public:
    nonLinearNeumannBCAtCorner():nonLinearBoundaryCondition(){
      onWhat = ON_VERTEX;
      _f = NULL;
    }
    nonLinearNeumannBCAtCorner(const nonLinearNeumannBCAtCorner& src): nonLinearBoundaryCondition(src),_cornerNumber(src._cornerNumber),
          _comp(src._comp),constraintDofs(src.constraintDofs),_f(src._f){}
    virtual ~nonLinearNeumannBCAtCorner(){
      if (_f != NULL) delete _f;
    }
    virtual nonLinearBoundaryCondition::type getType()const{return nonLinearBoundaryCondition::NEUMANN;}
};

class nonLinearConstraintBC : public nonLinearBoundaryCondition
{
 public:
  int _comp; // component
  FilterDof *_filter;
  FunctionSpaceBase *_space;
  std::list<Dof> constraintDofs;
  virtual nonLinearBoundaryCondition::type getType() const{return nonLinearBoundaryCondition::CONSTRAINT;}
  nonLinearConstraintBC() : nonLinearBoundaryCondition(),_comp(0),_space(NULL), _filter(NULL){}
  virtual ~nonLinearConstraintBC(){
		if (_filter!=NULL){delete _filter; _filter = NULL;};
		if (_space != NULL){delete _space, _space = NULL;};
	}
  nonLinearConstraintBC(const nonLinearConstraintBC &source) : nonLinearBoundaryCondition(source){
    this->_comp = source._comp;
    this->_space = source._space;
    this->_filter = source._filter;
  }
};

class nonLinearFixOnFaceBC: public nonLinearBoundaryCondition{
  public:
    double A, B, C, D;

  public:
    virtual nonLinearBoundaryCondition::type getType() const {return nonLinearBoundaryCondition::FIXONFACE;}
    nonLinearFixOnFaceBC():nonLinearBoundaryCondition(),A(0.),B(0.),C(0.),D(0.){}
    nonLinearFixOnFaceBC(const double aa, const double bb, const double cc, const double dd):
            A(aa),B(bb),C(cc),D(dd),nonLinearBoundaryCondition(){}
    nonLinearFixOnFaceBC(const nonLinearFixOnFaceBC& src):nonLinearBoundaryCondition(src){
      this->A = src.A;
      this->B = src.B;
      this->C = src.C;
      this->D = src.D;
    }
    virtual ~nonLinearFixOnFaceBC(){};
};

class nonLinearSameDisplacementBC: public nonLinearBoundaryCondition{
  public:
    int comp; // component
    groupOfElements* gRoot;
    int tagRoot;
    double fac;

  public:
    virtual nonLinearBoundaryCondition::type getType() const {return nonLinearBoundaryCondition::SAMEDISP;}
    nonLinearSameDisplacementBC(): comp(-1),tagRoot(-1),gRoot(NULL),fac(1.),nonLinearBoundaryCondition(){}
    nonLinearSameDisplacementBC(const nonLinearSameDisplacementBC& src): nonLinearBoundaryCondition(src){
      this->comp = src.comp;
      this->gRoot = src.gRoot;
      this->tagRoot = src.tagRoot;
      this->fac = src.fac;
    }
    virtual ~nonLinearSameDisplacementBC(){
			if (gRoot!=NULL) {delete gRoot; gRoot = NULL;};
		}
};

class nonLinearSameDisplacementBCBetweenTwoGroups : public nonLinearBoundaryCondition{
public:
  int comp;
  groupOfElements* g1, *g2; // two pbc group

  virtual nonLinearBoundaryCondition::type getType() const {return nonLinearBoundaryCondition::PERIODIC;}
  nonLinearSameDisplacementBCBetweenTwoGroups():nonLinearBoundaryCondition(),comp(0),g1(NULL),g2(NULL){};
  nonLinearSameDisplacementBCBetweenTwoGroups(const nonLinearSameDisplacementBCBetweenTwoGroups& src) : nonLinearBoundaryCondition(src){
   this->comp = src.comp;
   this->g1 = src.g1;
   this->g2 = src.g2;
  }
  virtual ~nonLinearSameDisplacementBCBetweenTwoGroups(){
		if (g1!=NULL){delete g1; g1 = NULL;}
		if (g2!=NULL){delete g2; g2 = NULL;};
	}
};

class nonLinearPeriodicBCBetweenTwoGroups : public nonLinearBoundaryCondition{
 public:
  int comp;
	MVertex* v1, *v2; // two root vertices
  groupOfElements* g1, *g2; // two pbc group
  std::map<Dof,Dof> dofMap;
	std::pair<Dof,Dof>* rootDofs;

  virtual nonLinearBoundaryCondition::type getType() const {return nonLinearBoundaryCondition::PERIODIC;}
  nonLinearPeriodicBCBetweenTwoGroups():nonLinearBoundaryCondition(),comp(0),g1(NULL),g2(NULL),dofMap(),v1(NULL),v2(NULL),rootDofs(NULL){};
  nonLinearPeriodicBCBetweenTwoGroups(const nonLinearPeriodicBCBetweenTwoGroups& src) : nonLinearBoundaryCondition(src){
   this->comp = src.comp;
   this->g1 = src.g1;
   this->g2 = src.g2;
	 this->v1 = src.v1;
	 this->v2 = src.v2;
	 rootDofs = NULL;
   this->dofMap.clear();
  }
  virtual ~nonLinearPeriodicBCBetweenTwoGroups(){
		if (g1!=NULL){delete g1; g1 = NULL;}
		if (g2!=NULL){delete g2; g2 = NULL;};
	}
};

class nonLinearAveragePeriodicBCBetweenTwoGroups : public nonLinearBoundaryCondition{
 public:
  int comp;
	MVertex* v1, *v2; // two root vertices
  groupOfElements* g1, *g2; // two pbc group


  virtual nonLinearBoundaryCondition::type getType() const {return nonLinearBoundaryCondition::PERIODIC;}
  nonLinearAveragePeriodicBCBetweenTwoGroups():nonLinearBoundaryCondition(),comp(0),g1(NULL),g2(NULL),v1(NULL),v2(NULL){};
  nonLinearAveragePeriodicBCBetweenTwoGroups(const nonLinearAveragePeriodicBCBetweenTwoGroups& src) : nonLinearBoundaryCondition(src){
   this->comp = src.comp;
   this->g1 = src.g1;
   this->g2 = src.g2;
	 this->v1 = src.v1;
	 this->v2 = src.v2;
  }
  virtual ~nonLinearAveragePeriodicBCBetweenTwoGroups(){
		if (g1!=NULL){delete g1; g1 = NULL;}
		if (g2!=NULL){delete g2; g2 = NULL;};
	}
};

// group will be not used but allow to store in same vector
// tag == Master physical number
class rigidContactBC : public nonLinearBoundaryCondition
{
 public:
  rigidContactSpaceBase *space;
  int _comp; // component
  simpleFunctionTime<double> *_f;
  FilterDof *_filter;
  virtual nonLinearBoundaryCondition::type getType() const{return nonLinearBoundaryCondition::RCONTACT;}
  rigidContactBC(const int physMaster) : nonLinearBoundaryCondition() , _filter(NULL),_f(NULL){
    _tag = physMaster;
    onWhat = RIGIDCONTACT;
  }
  rigidContactBC(const rigidContactBC &source) : nonLinearBoundaryCondition(source)
  {
    _comp = source._comp;
    space = source.space;
    _f = source._f;
    _filter = source._filter;
  }
  virtual ~rigidContactBC(){
		if (_filter != NULL) {delete _filter; _filter = NULL;};
		if (_f != NULL){delete _f; _f = NULL;}
	}
  void setSpace(rigidContactSpaceBase *sp){space = sp;}
};
#endif // SWIG
class nonLinearNeumannBC  : public nonLinearBoundaryCondition
{
 public:
  enum NeumannBCType{NEUMANN_UNDEF=0,FORCE,PRESSURE,SCALAR_FLUX}; 
  // FORCE - traction applied in one direction in reference configure 
  // PRESSURE - 3 dimensional force appplied to the normal direction of the current surface
  // SCALAR_FLUX - scalar flux to current surface
  NeumannBCType _NeumannBCType;
  simpleFunctionTime<double> *_f;
  /* BE AWARE the four following vectors have to be filles in the same time */
  std::vector<FunctionSpaceBase*> _vspace;
  std::vector<const groupOfElements*> _vgroup;
  std::vector<QuadratureBase*> _vquadrature;
  std::vector<partDomain*> _vdom; // Used to create the term after
  // is fill after
  std::vector<LinearTermBase<double>*> _vterm; // given by partDomain
  std::vector<BilinearTermBase*> _vmatrix_term; // given by partDomain
  int _comp; // component x, y or z to create function space
  nonLinearNeumannBC () : nonLinearBoundaryCondition(),_f(NULL) {}
  nonLinearNeumannBC(const nonLinearNeumannBC &source) : _f(source._f), nonLinearBoundaryCondition(source),
                                                          _vspace(source._vspace), _vterm(source._vterm),
                                                          _comp(source._comp), _vquadrature(source._vquadrature),
                                                          _vgroup(source._vgroup), _vdom(source._vdom),
                                                          _NeumannBCType(NEUMANN_UNDEF){}
	virtual ~nonLinearNeumannBC(){
		this->clear();
	}
  virtual nonLinearBoundaryCondition::type getType() const {return nonLinearBoundaryCondition::NEUMANN;}
  void clear()
  {
    for(int i=0;i<_vspace.size();i++)
    {
      delete _vspace[i];
//      _vgroup[i]->clearAll(); // end problem ??
      delete _vquadrature[i];
      delete _vterm[i];
      delete _vmatrix_term[i];
    }
    _vspace.clear();
    _vgroup.clear();
    _vquadrature.clear();
    _vterm.clear();
    _vdom.clear();
  }
  virtual NeumannBCType getNeumannBCType() const {return _NeumannBCType;}
};
#ifndef SWIG
class initialCondition : public nonLinearDirichletBC {
 public:
  // for veloDCB the axis of beam is given by a simpleFunctionTime and is applied in the comp direction !!
  std::vector<groupOfElements*> _vDomGroup; // needed to prescribed normal on the boundary of domain
  /* COMP == -1 --> prescribed an initial value normal to the element */
  initialCondition(int comp, nonLinearBoundaryCondition::whichCondition wc) : nonLinearDirichletBC()
  {
    _comp=comp;
    _mycondition=wc;
  }
  virtual ~initialCondition(){}
  initialCondition(const initialCondition &source) : nonLinearDirichletBC(source){
    _vDomGroup = source._vDomGroup;
  }
  virtual nonLinearBoundaryCondition::type getType() const {return nonLinearBoundaryCondition::INITIAL;}
  void clear()
  {
    nonLinearDirichletBC::clear();
  }
};
#endif

#endif // non linear Boundary Conditions
