//
// C++ Interface: partDomain
//
// Description: Interface class to used solver. Your term ha to be store in a domain
//
// Author:  <Gauthier BECKER>, (C) 2010
//
// Copyright: See COPYING file that comes with this distribution
//
//
#include "partDomain.h"
#include "ipField.h"
#include "numericalMaterial.h"
#include "MInterfacePoint.h"
#include "ipFiniteStrain.h"
#include "extraDofIPVariable.h"
#include "nonLinearMechSolver.h"


#if defined(HAVE_MPI)
#include "mpi.h"
#endif //HAVE_MPI

partDomain::partDomain(const partDomain &source){
	// notice: only options are copied
	// this domain is created as usual domain in multiscale analyses,
	// all other members are created at the solver

	_tag = source._tag;
  _phys = source._phys;
  _fullDg = source._fullDg;
  _wsp = source._wsp;

	// all terms are not copied as they are created in the solver
  btermBulk = NULL; // source.btermBulk;
  btermElasticBulk = NULL; // not copy
  massterm = NULL; //source.massterm;
  ltermBulk = NULL; //source.ltermBulk;
  btermTangent = NULL; // btermTangent;
  btermElasticTangent = NULL;
  scalarTermPF = NULL;
  linearTermPF = NULL;

  // integBulk will be created in Solver
  integBulk = NULL; //source.integBulk;

  _numberNonLocalVariable = source._numberNonLocalVariable;
  _numberConstitutiveExtraDofDiffusionVariable = source._numberConstitutiveExtraDofDiffusionVariable;
  _bmbp = source._bmbp;
  _eps = source._eps;

  setmaterial = false; //source.setmaterial;

  #if defined(HAVE_MPI)
  _rootRank = source._rootRank;
  _otherRanks = source._otherRanks;
  // not copy
	//_domainIPBulk = source._domainIPBulk;
  //_mapIPBulk = source._mapIPBulk;
  #endif //HAVE_MPI

  _distributedOtherRanks = source._distributedOtherRanks;
  _distributedOnRoot = source._distributedOnRoot;

	_macroSolver = source._macroSolver;

	//
  g = NULL;
	if (source.g != NULL){
		g = new groupOfElements();
	};
	// group are added in solver
	_groupGhostMPI = NULL;
	if (source._groupGhostMPI != NULL){
		_groupGhostMPI = new groupOfElements();
	};
}

partDomain& partDomain::operator=(const partDomain &source)
{
  _tag = source._tag;
  _phys = source._phys;
  _fullDg = source._fullDg;
  _wsp = source._wsp;

	// all terms are not copied as they are created in the solver
  btermBulk = NULL; // source.btermBulk;
  btermElasticBulk = NULL;
  massterm = NULL; //source.massterm;
  ltermBulk = NULL; //source.ltermBulk;
  btermTangent = NULL; // btermTangent;
  btermElasticTangent = NULL;
	scalarTermPF = NULL;
	linearTermPF = NULL;

	integBulk = source.integBulk;

  _numberNonLocalVariable = source._numberNonLocalVariable;
  _numberConstitutiveExtraDofDiffusionVariable = source._numberConstitutiveExtraDofDiffusionVariable;
  _bmbp = source._bmbp;
  _eps = source._eps;

  setmaterial = false; //source.setmaterial;

  #if defined(HAVE_MPI)
  _rootRank = source._rootRank;
  _otherRanks = source._otherRanks;
  // not copy
	//_domainIPBulk = source._domainIPBulk;
  //_mapIPBulk = source._mapIPBulk;
  #endif //HAVE_MPI

  _distributedOtherRanks = source._distributedOtherRanks;
  _distributedOnRoot = source._distributedOnRoot;

  _macroSolver = source._macroSolver;

	//
  g = NULL;
  if (source.g != NULL){
    g = new groupOfElements();
  };
  // group are added in solver
  _groupGhostMPI = NULL;
  if (source._groupGhostMPI != NULL){
    _groupGhostMPI = new groupOfElements();
  };

  return *this;
}

bool partDomain::isDistributedRootRank() const {return _distributedOnRoot;};
void partDomain::distributeOnRootRank(const int flg){_distributedOnRoot = flg;};
bool partDomain::isDistributedOtherRanks() const {return _distributedOtherRanks;};
void partDomain::setDistributedOtherRanks(const bool flag) {_distributedOtherRanks = flag;};

void partDomain::setMacroSolver(const nonLinearMechSolver* solver){
	_macroSolver = solver;
}
const nonLinearMechSolver* partDomain::getMacroSolver() const {return _macroSolver;};

BilinearTermBase* partDomain::createNeumannMatrixTerm(FunctionSpace<double> *spneu, const groupOfElements* g,
                                                      const simpleFunctionTime<double>* f, const unknownField *uf,
                                                      const IPField * ip, 
                                                      const nonLinearBoundaryCondition::location onWhat,
                                                      const nonLinearNeumannBC::NeumannBCType neumann_type,
                                                      const int comp) const
{
  return new BiNonLinearTermVoid();
};

void partDomain::createIPState(AllIPState::ipstateContainer& map, const bool *state) const{
  IntPt* GP;
  const materialLaw *mlaw = this->getMaterialLaw();
  if (mlaw== NULL) return;
	#if defined(HAVE_MPI)
  if (_otherRanks.size() > 0 and mlaw->isNumeric()){
    if (Msg::GetCommRank() == _rootRank){
      // create IP State
      for (groupOfElements::elementContainer::const_iterator it = this->g->begin(); it != this->g->end(); ++it){
        MElement *ele = *it;
        int npts_bulk=this->getBulkGaussIntegrationRule()->getIntPoints(ele,&GP);
        AllIPState::ipstateElementContainer tp(npts_bulk);
        for(int i=0;i<npts_bulk;i++){
          int num = numericalMaterialBase::createTypeWithTwoInts(ele->getNum(),i);
          if (_domainIPBulk.find(num) != _domainIPBulk.end())
            mlaw->createIPState(tp[i],state,ele,ele->getNumVertices(),&(GP[i]),i);
          else{
            const numericalMaterialBase* nummat = dynamic_cast<const numericalMaterialBase*>(mlaw);
            if (nummat)
              nummat->createIPState(false,tp[i],state,ele,ele->getNumVertices(),&(GP[i]),i);
            else
              Msg::Error("Error error at creating the bulk IPState");
          }
          tp[i]->setLocation(IPVariable::BULK);
        }
        map.insert(AllIPState::ipstatePairType(ele->getNum(),tp));
      }
    }
    else if (_otherRanks.find(Msg::GetCommRank()) != _otherRanks.end()) {
      for (std::set<int>::iterator it = _domainIPBulk.begin(); it != _domainIPBulk.end(); it++){
        int num = *it;
        int elnum, gnum;
        IntPt *GPnum;
        numericalMaterialBase::getTwoIntsFromType(num,elnum,gnum);
        MElement* ele = new MPoint(NULL,elnum,0);
        int npts_bulk=this->getBulkGaussIntegrationRule()->getIntPoints(ele,&GPnum);
        AllIPState::ipstateElementContainer tp(1);
        mlaw->createIPState(tp[0],state,ele,ele->getNumVertices(),&(GPnum[0]),gnum);
        tp[0]->setLocation(IPVariable::BULK);
        delete ele;
        map.insert(AllIPState::ipstatePairType(num,tp));
      };
    }
  }
  else
	#endif //HAVE_MPI
	{
    for (groupOfElements::elementContainer::const_iterator it = this->g->begin(); it != this->g->end(); ++it){
      MElement *ele = *it;
      int npts_bulk=this->getBulkGaussIntegrationRule()->getIntPoints(ele,&GP);
      AllIPState::ipstateElementContainer tp(npts_bulk);
      for(int i=0;i<npts_bulk;i++){
        mlaw->createIPState(tp[i],state,ele,ele->getNumVertices(),&(GP[i]),i);
        tp[i]->setLocation(IPVariable::BULK);
      }
      map.insert(AllIPState::ipstatePairType(ele->getNum(),tp));
    }
  }
};

double partDomain::computeMeanValueDomain(const int num, const IPField* ipf, double& solid) const{
  IntPt* GP;
  double meanVal = 0;
  solid = 0;
  const ipFiniteStrain *vipv[256];
  for (groupOfElements::elementContainer::const_iterator it = this->g->begin(); it != this->g->end(); ++it){
    MElement *ele = *it;
    ipf->getIPv(ele,vipv);
    int npts_bulk=this->getBulkGaussIntegrationRule()->getIntPoints(ele,&GP);
    for (int i=0; i<npts_bulk; i++){
      double weight= GP[i].weight;
      double u = GP[i].pt[0];
      double v = GP[i].pt[1];
      double w = GP[i].pt[2];
      double detJ = ele->getJacobianDeterminant(u,v,w);
      meanVal += vipv[i]->get(num)*weight*detJ;
      solid += weight*detJ;
    };
  }
  if (solid > 0){
    meanVal /= solid;
  }
  return meanVal;
};

double partDomain::averagingOnActiveDissipationZone(const int num, const IPField* ipf, double& solid, const IPStateBase::whichState ws) const{
  IntPt* GP;
  double meanVal = 0;
  solid = 0.;
  const ipFiniteStrain *vipv[256];
  for (groupOfElements::elementContainer::const_iterator it = this->g->begin(); it != this->g->end(); ++it){
    MElement *ele = *it;
    ipf->getIPv(ele,vipv,ws);
    int npts_bulk=this->getBulkGaussIntegrationRule()->getIntPoints(ele,&GP);
    for (int i=0; i<npts_bulk; i++){
      // compute if damage is active and non unloading
      if (vipv[i]->dissipationIsActive()){
        double weight= GP[i].weight;
        double u = GP[i].pt[0];
        double v = GP[i].pt[1];
        double w = GP[i].pt[2];
        double detJ = ele->getJacobianDeterminant(u,v,w);
        double valnum = vipv[i]->get(num);
        if (num == IPField::F_XX or num == IPField::F_YY or num == IPField::F_ZZ){
          valnum -= 1.;
        }
        meanVal += valnum*weight*detJ;
        solid += weight*detJ;
      }
    };
  }
  if (solid > 0.)
    return meanVal/solid;
  else
    return 0.;
};

#if defined(HAVE_MPI)
void partDomain::setWorkingRanks(const int root, const std::set<int>& others) {
  _rootRank = root;
  _otherRanks = others;
};
int partDomain::getRootRank() const{
  return _rootRank;
};
void partDomain::getOtherRanks(std::set<int>& others) const{
  others = _otherRanks;
};

void partDomain::clearOtherRanks(){
  _otherRanks.clear();
};

void partDomain::createIPMap(){
  const materialLaw* mlaw = this->getMaterialLaw();
  if (mlaw == NULL) return;

  _mapIPBulk.clear();
  _domainIPBulk.clear();

  if (_otherRanks.size()>0 and mlaw->isNumeric()){
    if (Msg::GetCommRank() == _rootRank){
      // get all IP number
      std::vector<int> allRanks(_otherRanks.begin(),_otherRanks.end());
      if (_distributedOnRoot)
        allRanks.push_back(_rootRank);
      std::vector<int> allIPBulk;
      IntPt* GP;
      for (groupOfElements::elementContainer::iterator it = this->g->begin(); it != this->g->end(); ++it){
        MElement *ele = *it;
        int npts_bulk=this->getBulkGaussIntegrationRule()->getIntPoints(ele,&GP);
        for(int i=0;i<npts_bulk;i++){
          int num = numericalMaterialBase::createTypeWithTwoInts(ele->getNum(),i);
          allIPBulk.push_back(num);
        }
      }
      numericalMaterialBase::distributeMacroIP(allIPBulk,allRanks,_mapIPBulk);
      allIPBulk.clear();

      // Send data for bulk elements
      for (std::map<int,std::set<int> >::iterator it = _mapIPBulk.begin(); it!= _mapIPBulk.end(); it++){
        int rank = it->first;
        std::set<int>& numElOnRank = _mapIPBulk[rank];
        // at root rank
        if (rank == _rootRank){
          _domainIPBulk = numElOnRank;
        }
        // at other ranks
        else{
          int bufferSize = numElOnRank.size();
          MPI_Send(&bufferSize,1,MPI_INT,rank,numericalMaterialBase::createTypeWithTwoInts(getPhysical(),0),MPI_COMM_WORLD);
          if (bufferSize>0){
            int* buffer = new int[bufferSize];
            int idex = 0;
            for (std::set<int>::iterator its = numElOnRank.begin(); its != numElOnRank.end(); its++){
              buffer[idex] = *its;
              idex++;
            };
            MPI_Send(buffer,bufferSize,MPI_INT,rank,numericalMaterialBase::createTypeWithTwoInts(getPhysical(),1),MPI_COMM_WORLD);
            delete[] buffer;
          }
        }
      };
    }
    else if (_otherRanks.find(Msg::GetCommRank()) != _otherRanks.end()){
      // receive data from root for bulk elements
      MPI_Status status;
      int bufferSize;
      MPI_Recv(&bufferSize,1,MPI_INT,_rootRank,numericalMaterialBase::createTypeWithTwoInts(getPhysical(),0),MPI_COMM_WORLD,&status);
      if (bufferSize>0){
        int* buffer = new int[bufferSize];
        MPI_Recv(buffer,bufferSize,MPI_INT,_rootRank,numericalMaterialBase::createTypeWithTwoInts(getPhysical(),1),MPI_COMM_WORLD,&status);
        _domainIPBulk.clear();
        for (int i=0; i<bufferSize; i++)
          _domainIPBulk.insert(buffer[i]);
        delete[] buffer;
      }
    }
  }
};
#endif //HAVE_MPI

void partDomain::initMicroMeshId(){
  materialLaw* mlaw = this->getMaterialLaw();
  if (mlaw == NULL) return;
  if (mlaw->isNumeric() ){
    numericalMaterialBase* nummat = dynamic_cast<numericalMaterialBase*>(mlaw);
    if (nummat == NULL)
      Msg::Error("material law must be derived from numericalMaterialBase");
    IntPt* GP;
    for (groupOfElements::elementContainer::iterator it = g->begin(); it!= g->end(); it++){
      MElement* ele = *it;
      int npts = this->getBulkGaussIntegrationRule()->getIntPoints(ele,&GP);
      for (int i=0; i<npts; i++){
        nummat->assignMeshId(ele->getNum(),i);
      }
    }
  }
};

void partDomain::setBulkMatrixByPerturbation(const int i, const double eps)
{
  i == 0 ? _bmbp = false : _bmbp = true;
  _eps = eps;
}

int partDomain::getHomogenizationOrder() const {return _macroSolver->getHomogenizationOrder();}
bool partDomain::inMultiscaleSimulation()  const {return _macroSolver->inMultiscaleSimulation();};
bool partDomain::isExtractCohesiveLawFromMicroDamage() const {return _macroSolver->getDamageToCohesiveJumpFlag();};
bool partDomain::getExtractIrreversibleEnergyFlag() const {return _macroSolver->getExtractIrreversibleEnergyFlag();};
const elementErosionFilter& partDomain::getElementErosionFilter() const{return _macroSolver->getElementErosionFilter();}

void partDomain::computeAverageStress(const IPField* ipf, STensor3& sig) const{
	if (g->size() >0){
		IntPt* GP;
		const ipFiniteStrain *vipv[256];
		for (groupOfElements::elementContainer::iterator it= g->begin(); it!=g->end(); it++){
			MElement* ele= *it;
			ipf->getIPv(ele,vipv);
			int npts= integBulk->getIntPoints(ele,&GP);
			for (int i=0; i<npts; i++){
				double weight= GP[i].weight;
				double &detJ = vipv[i]->getJacobianDeterminant(ele,GP[i]);
				const STensor3& P = vipv[i]->getConstRefToFirstPiolaKirchhoffStress();
				sig+=P*detJ*weight;
			};
		};
	}
};


void partDomain::computeActiveDissipationAverageStressIncrement(const IPField* ipf, STensor3& stress) const{
  // compute average strain on active damaging part
  // damaging volume will besed on ws= current, previous, or initial
	if ((g->size()>0) and (ipf->getNumOfActiveDissipationIPs() > 0)){
		IntPt* GP;
	// compute from damaging zone
		for (groupOfElements::elementContainer::iterator it= g->begin(); it!=g->end(); it++){
			MElement* ele= *it;

			const AllIPState::ipstateElementContainer *vips = ipf->getAips()->getIPstate(ele->getNum());

			int npts= integBulk->getIntPoints(ele,&GP);
			for (int i=0; i<npts; i++){
				const IPStateBase *ips        = (*vips)[i];
				const ipFiniteStrain *ipv     = static_cast<const ipFiniteStrain*>(ips->getState(IPStateBase::current));
				const ipFiniteStrain *ipvprev = static_cast<const ipFiniteStrain*>(ips->getState(IPStateBase::activeDissipation));

				if (ipvprev->dissipationIsActive()){
					double weight= GP[i].weight;
					const double &detJ = ipv->getJacobianDeterminant(ele,GP[i]);
          double ratio = detJ*weight;
					const STensor3& P =ipv->getConstRefToFirstPiolaKirchhoffStress();
          const STensor3& Pprev = ipvprev->getConstRefToFirstPiolaKirchhoffStress();
					for (int j=0; j<3; j++){
						for (int k=0; k<3; k++){
							stress(j,k) += (P(j,k)-Pprev(j,k))*ratio;
						}
					}
				}
			};
		};
	}
};


void partDomain::computeAverageHighOrderStress(const IPField* ipf, STensor33& sig) const{
	if (g->size()>0){
		IntPt* GP;
		const ipFiniteStrain *vipv[256];
		const FunctionSpaceBase* space = this->getFunctionSpace();
		for (groupOfElements::elementContainer::iterator it= g->begin(); it!=g->end(); it++){
			MElement* ele= *it;
			if(ele->getParent()) ele = ele->getParent();
			ipf->getIPv(ele,vipv);
			int npts= integBulk->getIntPoints(ele,&GP);
			for (int i=0; i<npts; i++){
				double weight= GP[i].weight;
				double &detJ = vipv[i]->getJacobianDeterminant(ele,GP[i]);
				SPoint3 pt;
				std::vector<TensorialTraits<double>::ValType> &Vals = vipv[i]->f(space,ele,GP[i]);
				ele->pnt(Vals,pt);
				const STensor3& P = vipv[i]->getConstRefToFirstPiolaKirchhoffStress();
				double ratio = detJ*weight;
				for (int ii=0; ii<3; ++ii){
					for (int jj=0; jj<3; ++jj){
						for (int kk=0; kk<3; ++kk){
							sig(ii,jj,kk) += 0.5*(P(ii,jj)*pt[kk]+P(ii,kk)*pt[jj])*ratio;
						}
					}
				}
			};
		};
	}
};

double partDomain::computeVolumeDomain(const IPField* ipf) const{
  double volume = 0;
	if (g->size()>0){
		IntPt* GP;
		const ipFiniteStrain *vipv[256];

		for (groupOfElements::elementContainer::iterator it= g->begin(); it!=g->end(); it++){
			MElement* ele= *it;
			if(ele->getParent()) ele = ele->getParent();
			int npts= integBulk->getIntPoints(ele,&GP);
			ipf->getIPv(ele,vipv);
			for (int i=0; i<npts; i++){
				double weight= GP[i].weight;
				double& detJ = vipv[i]->getJacobianDeterminant(ele,GP[i]);
				double ratio = detJ*weight;
				volume+=ratio;
			};
		};
	}
  return volume;
};

void partDomain::computeActiveDissipationCenter(const IPField* ipf, SPoint3& pos) const{
	if ((g->size()>0) and (ipf->getNumOfActiveDissipationIPs() > 0)){
		IntPt* GP;
	// compute from damaging zone
		for (groupOfElements::elementContainer::iterator it= g->begin(); it!=g->end(); it++){
			MElement* ele= *it;
			const AllIPState::ipstateElementContainer *vips = ipf->getAips()->getIPstate(ele->getNum());
			int npts= integBulk->getIntPoints(ele,&GP);
			for (int i=0; i<npts; i++){
				const IPStateBase *ips        = (*vips)[i];
				const ipFiniteStrain *ipv     = static_cast<const ipFiniteStrain*>(ips->getState(IPStateBase::current));
				const ipFiniteStrain *ipvprev = static_cast<const ipFiniteStrain*>(ips->getState(IPStateBase::activeDissipation));
				if (ipvprev->dissipationIsActive()){
					double weight= GP[i].weight;
					const double &detJ = ipv->getJacobianDeterminant(ele,GP[i]);
          double ratio = detJ*weight;
					SPoint3 P;
					ele->pnt(GP[i].pt[0],GP[i].pt[1],GP[i].pt[2],P);
					for (int j=0; j<3; j++){
						pos[j] += P[j]*ratio;
					}
				}
			};
		};
	}
};

void partDomain::computeActiveDissipationAverageStrainIncrement(const IPField* ipf, STensor3& strain) const{
  // compute average strain on active damaging part
  // damaging volume will besed on ws= current, previous, or initial
	if ((g->size()>0) and (ipf->getNumOfActiveDissipationIPs() > 0)){
		IntPt* GP;
	// compute from damaging zone
		for (groupOfElements::elementContainer::iterator it= g->begin(); it!=g->end(); it++){
			MElement* ele= *it;
			const AllIPState::ipstateElementContainer *vips = ipf->getAips()->getIPstate(ele->getNum());
			int npts= integBulk->getIntPoints(ele,&GP);
			for (int i=0; i<npts; i++){
				const IPStateBase *ips        = (*vips)[i];
				const ipFiniteStrain *ipv     = static_cast<const ipFiniteStrain*>(ips->getState(IPStateBase::current));
				const ipFiniteStrain *ipvprev = static_cast<const ipFiniteStrain*>(ips->getState(IPStateBase::activeDissipation));

				if (ipvprev->dissipationIsActive()){
					double weight= GP[i].weight;
					const double &detJ = ipv->getJacobianDeterminant(ele,GP[i]);
          double ratio = detJ*weight;
					const STensor3& F =ipv->getConstRefToDeformationGradient();
					const STensor3& Fprev = ipvprev->getConstRefToDeformationGradient();
					for (int j=0; j<3; j++){
						for (int k=0; k<3; k++){
							strain(j,k) += (F(j,k) - Fprev(j,k))*ratio;
						}
					}
				}
			};
		};
	}
};

double partDomain::computeVolumeActiveDissipationDomain(const IPField* ipf) const{
  double volume = 0.;
	if (g->size()>0){
		IntPt* GP;
		const ipFiniteStrain *vipvRef[256];
	// compute from damaging zone

		for (groupOfElements::elementContainer::iterator it= g->begin(); it!=g->end(); it++){
			MElement* ele= *it;
			ipf->getIPv(ele,vipvRef,IPStateBase::activeDissipation);
			int npts= integBulk->getIntPoints(ele,&GP);
			for (int i=0; i<npts; i++){
				if (vipvRef[i]->dissipationIsActive()){
					double weight= GP[i].weight;
					double &detJ = vipvRef[i]->getJacobianDeterminant(ele,GP[i]);
					volume+=detJ*weight;
				}
			};
		};
	}

  return volume;
};

void partDomain::computeConstitutiveExtraDofAverageFlux(const int index, const IPField* ipf, SVector3& extraDofFluxpart) const{
	if (g->size() > 0){
		IntPt* GP;
		const IPVariable *vipv[256];

		for (groupOfElements::elementContainer::iterator it= g->begin(); it!=g->end(); it++){
			MElement* ele= *it;
			ipf->getIPv(ele,vipv);
			int npts= integBulk->getIntPoints(ele,&GP);
			for (int i=0; i<npts; i++){
				double weight= GP[i].weight;
				double detJ = 0;
				const ipFiniteStrain* ipfs = dynamic_cast<const ipFiniteStrain*>(vipv[i]);
				if (ipfs == NULL){
					detJ =  ipfs->getJacobianDeterminant(ele,GP[i]);
				}
				else{
					detJ = ele->getJacobianDeterminant(GP[i].pt[0],GP[i].pt[1],GP[i].pt[2]);
				}
				const extraDofIPVariableBase* extraipv = dynamic_cast<const extraDofIPVariableBase*>(vipv[i]);
				if (extraipv != NULL){
					SVector3 q(0.);
					extraipv->getConstitutiveExtraDofFlux(index,q);
					extraDofFluxpart+=q*(detJ*weight);
				}
				else{
					Msg::Error("extraDofIPVariableBase is not used");
				}

			};
		};
	}
};

void partDomain::computeConstitutiveExtraDofAverageInternalEnergy(const int index, const IPField* ipf, double &internalEnergyExtraDofpart) const{
  if (g->size() > 0){
		IntPt* GP;
		const IPVariable *vipv[256];
		for (groupOfElements::elementContainer::iterator it= g->begin(); it!=g->end(); it++){
			MElement* ele= *it;
			ipf->getIPv(ele,vipv);
			int npts= integBulk->getIntPoints(ele,&GP);
			for (int i=0; i<npts; i++){
				double weight= GP[i].weight;
				double detJ = 0;
				const ipFiniteStrain* ipfs = dynamic_cast<const ipFiniteStrain*>(vipv[i]);
				if (ipfs == NULL){
					detJ =  ipfs->getJacobianDeterminant(ele,GP[i]);
				}
				else{
					detJ = ele->getJacobianDeterminant(GP[i].pt[0],GP[i].pt[1],GP[i].pt[2]);
				}
				const extraDofIPVariableBase* extraipv = dynamic_cast<const extraDofIPVariableBase*>(vipv[i]);
				if (extraipv!=NULL){
					double internalEnergy = extraipv->getConstitutiveExtraDofInternalEnergy(index);
					internalEnergyExtraDofpart+=internalEnergy*(detJ*weight);
				}
				else{
					Msg::Error("extraDofIPVariableBase is not used");
				}
			};
		};
	}
};

void partDomain::computeConstitutiveExtraDofAverageFieldCapacityPerUnitField(const int index, const IPField* ipf, double& fieldCapacityPart) const{
	if (g->size() > 0){
		IntPt* GP;
		const IPVariable *vipv[256];

		for (groupOfElements::elementContainer::iterator it= g->begin(); it!=g->end(); it++){
			MElement* ele= *it;
			if (this->inMultiscaleSimulation()){
				ipf->getIPv(ele,vipv,IPStateBase::initial);
			}
			else{
				ipf->getIPv(ele,vipv,IPStateBase::previous);
			}

			int npts= integBulk->getIntPoints(ele,&GP);
			for (int i=0; i<npts; i++){
				double weight= GP[i].weight;
				double detJ = 0;
				const ipFiniteStrain* ipfs = dynamic_cast<const ipFiniteStrain*>(vipv[i]);
				if (ipfs == NULL){
					detJ =  ipfs->getJacobianDeterminant(ele,GP[i]);
				}
				else{
					detJ = ele->getJacobianDeterminant(GP[i].pt[0],GP[i].pt[1],GP[i].pt[2]);
				}
				const extraDofIPVariableBase* extraipv = dynamic_cast<const extraDofIPVariableBase*>(vipv[i]);
				if (extraipv!=NULL){
					double extraDofStoredEnergyPerUnitField = extraipv->getConstitutiveExtraDofFieldCapacityPerUnitField(index);
					fieldCapacityPart+=extraDofStoredEnergyPerUnitField*(detJ*weight);
				}
				else{
					Msg::Error("extraDofIPVariableBase is not used");
				}
			};
		};
	}
};

void partDomain::computeConstitutiveExtraDofAverageMechanicalSource(const int index, const IPField* ipf, double& mechanicalSource) const{
	if (g->size() > 0){
		IntPt* GP;
		const IPVariable *vipv[256];
		for (groupOfElements::elementContainer::iterator it= g->begin(); it!=g->end(); it++){
			MElement* ele= *it;
			ipf->getIPv(ele,vipv);
			int npts= integBulk->getIntPoints(ele,&GP);
			for (int i=0; i<npts; i++){
				double weight= GP[i].weight;
				double detJ = 0;
				const ipFiniteStrain* ipfs = dynamic_cast<const ipFiniteStrain*>(vipv[i]);
				if (ipfs == NULL){
					detJ =  ipfs->getJacobianDeterminant(ele,GP[i]);
				}
				else{
					detJ = ele->getJacobianDeterminant(GP[i].pt[0],GP[i].pt[1],GP[i].pt[2]);
				}
				const extraDofIPVariableBase* extraipv = dynamic_cast<const extraDofIPVariableBase*>(vipv[i]);
				if (extraipv!=NULL){
					double msource = extraipv->getConstitutiveExtraDofMechanicalSource(index);
					mechanicalSource+=msource*(detJ*weight);
				}
				else{
					Msg::Error("extraDofIPVariableBase is not used");
				}
			};
		};
	}
};

void partDomain::computeNonConstitutiveExtraDofAverageFlux(const int index, const IPField* ipf, SVector3& extraDofFluxpart) const{
	if (g->size() > 0){
		IntPt* GP;
		const IPVariable *vipv[256];

		for (groupOfElements::elementContainer::iterator it= g->begin(); it!=g->end(); it++){
			MElement* ele= *it;
			ipf->getIPv(ele,vipv);
			int npts= integBulk->getIntPoints(ele,&GP);
			for (int i=0; i<npts; i++){
				double weight= GP[i].weight;
				double detJ = 0;
				const ipFiniteStrain* ipfs = dynamic_cast<const ipFiniteStrain*>(vipv[i]);
				if (ipfs == NULL){
					detJ =  ipfs->getJacobianDeterminant(ele,GP[i]);
				}
				else{
					detJ = ele->getJacobianDeterminant(GP[i].pt[0],GP[i].pt[1],GP[i].pt[2]);
				}
				const extraDofIPVariableBase* extraipv = dynamic_cast<const extraDofIPVariableBase*>(vipv[i]);
				if (extraipv!=NULL){
					SVector3 q;
					extraipv->getNonConstitutiveExtraDofFlux(index,q);
					extraDofFluxpart+=q*(detJ*weight);
				}
				else{
					Msg::Error("extraDofIPVariableBase is not used");
				}
			};
		};
	}
};

double partDomain::computeDeformationEnergy(const IPField* ipf) const{
	double energy = 0.;
	if (g->size()>0){
		IntPt* GP;
		const ipFiniteStrain *vipv[256];
		for (groupOfElements::elementContainer::iterator it= g->begin(); it!=g->end(); it++){
			MElement* ele= *it;
			ipf->getIPv(ele,vipv);
			int npts= integBulk->getIntPoints(ele,&GP);
			for (int i=0; i<npts; i++){
				double weight= GP[i].weight;
				double &detJ = vipv[i]->getJacobianDeterminant(ele,GP[i]);
				energy += (vipv[i]->defoEnergy()*detJ*weight);
			};
		};
	}
	return energy;
};
double partDomain::computePlasticEnergy(const IPField* ipf) const{
	double energy = 0.;
	if (g->size()>0){
		IntPt* GP;
		const ipFiniteStrain *vipv[256];
		for (groupOfElements::elementContainer::iterator it= g->begin(); it!=g->end(); it++){
			MElement* ele= *it;
			ipf->getIPv(ele,vipv);
			int npts= integBulk->getIntPoints(ele,&GP);
			for (int i=0; i<npts; i++){
				double weight= GP[i].weight;
				double &detJ = vipv[i]->getJacobianDeterminant(ele,GP[i]);
				energy += (vipv[i]->plasticEnergy()*detJ*weight);
			};
		};
	}
	return energy;
};

double partDomain::computeIrreversibleEnergy(const IPField* ipf) const{
	double energy = 0.;
	if (g->size()>0){
		IntPt* GP;
		const ipFiniteStrain *vipv[256];
		for (groupOfElements::elementContainer::iterator it= g->begin(); it!=g->end(); it++){
			MElement* ele= *it;
			ipf->getIPv(ele,vipv);
			int npts= integBulk->getIntPoints(ele,&GP);
			for (int i=0; i<npts; i++){
				double weight= GP[i].weight;
				double &detJ = vipv[i]->getJacobianDeterminant(ele,GP[i]);
				energy += (vipv[i]->irreversibleEnergy()*detJ*weight);
			};
		};
	}
	return energy;
};



dgPartDomain::dgPartDomain(const dgPartDomain &source) : partDomain(source){

	btermBound = NULL;
  ltermBound = NULL;
  btermVirtBound = NULL;
  ltermVirtBound = NULL;

	scalarTermPFBound = NULL;
	linearTermPFBound = NULL;

  integBound = NULL; // source.integBound;
  _interByPert = source._interByPert;
  _virtByPert = source._virtByPert;

  _multipleFieldFormulation = source._multipleFieldFormulation;
	_multiFieldMethod = source._multiFieldMethod;
	_enrichedUnknownLocation = source._enrichedUnknownLocation;

	_averageStrainBased = source._averageStrainBased;
  _brokenIPFromBeginning = source._brokenIPFromBeginning;
  _fixInterfaceDofComp = source._fixInterfaceDofComp;

  #if defined(HAVE_MPI)
	// not copied
  //_domainIPInterfacePlus = source._domainIPInterfacePlus,
  //_domainIPInterfaceMinus = source._domainIPInterfaceMinus;
  // _mapIPInterfacePlus = source._mapIPInterfacePlus,
  // _mapIPInterfaceMinus = source._mapIPInterfaceMinus;
  #endif //HAVE_MPI

	gi = NULL;
	if (source.gi != NULL){
		gi = new groupOfElements();
	}
  gib = NULL;
	if (source.gib != NULL){
		gib  = new groupOfElements();
	}
  giv = NULL;
	if (source.giv != NULL){
		giv = new groupOfElements();
	}
}

dgPartDomain& dgPartDomain::operator=(const partDomain & dom)
{
  partDomain::operator=(dom);
	const dgPartDomain* source = dynamic_cast<const dgPartDomain*>(&dom);
	if (source!= NULL){
		btermBound = NULL;
		ltermBound = NULL;
		btermVirtBound = NULL;
		ltermVirtBound = NULL;

		scalarTermPFBound = NULL;
		linearTermPFBound = NULL;

		integBound = NULL;
		_interByPert = source->_interByPert;
		_virtByPert = source->_virtByPert;

		_multipleFieldFormulation = source->_multipleFieldFormulation;
		_multiFieldMethod = source->_multiFieldMethod;
		_enrichedUnknownLocation = source->_enrichedUnknownLocation;

		_averageStrainBased = source->_averageStrainBased;
		_brokenIPFromBeginning = source->_brokenIPFromBeginning;
		_fixInterfaceDofComp = source->_fixInterfaceDofComp;


		#if defined(HAVE_MPI)
		// not copied
		//_domainIPInterfacePlus = source._domainIPInterfacePlus,
		//_domainIPInterfaceMinus = source._domainIPInterfaceMinus;
		// _mapIPInterfacePlus = source._mapIPInterfacePlus,
		// _mapIPInterfaceMinus = source._mapIPInterfaceMinus;
		#endif //HAVE_MPI

		gi = NULL;
		if (source->gi != NULL){
			gi = new groupOfElements();
		}
		gib = NULL;
		if (source->gib != NULL){
			gib  = new groupOfElements();
		}
		giv = NULL;
		if (source->giv != NULL){
			giv = new groupOfElements();
		}
	}

  return *this;
}

void dgPartDomain::averageStrainBased(const bool flg){_averageStrainBased = flg;};
void dgPartDomain::setMultipleFieldFormulation(const bool flag, const int method, const int loc) {
	_multipleFieldFormulation = flag;
	_multiFieldMethod = (MultiFieldMethod)method;
	_enrichedUnknownLocation = (EnrichedUnknownLocation)loc;

	if (_multipleFieldFormulation){
    if (_multiFieldMethod == dgPartDomain::TWOFIELD){
      Msg::Info("Full twofield is used");
    }
    else{
      Msg::Error("Twofield method %d is not considered",_multiFieldMethod);
    }

		if (_enrichedUnknownLocation == partDomain::NODE){
			Msg::Info("enrich Dofs at nodes");
		}
		else if (_enrichedUnknownLocation == partDomain::GP){
			Msg::Info("enrich Dofs at GPs");
		}
		else{
			Msg::Error("enriched location %d is not considered",loc);
		}
	}
};
void dgPartDomain::fixInterfaceDofComponent(const int comp){_fixInterfaceDofComp.push_back(comp);};
void dgPartDomain::breakAllInterfaceIPsFromBeginning(const bool broken){ _brokenIPFromBeginning = broken;};

#if defined(HAVE_MPI)
void dgPartDomain::createIPMap(){
	if (_otherRanks.size() == 0) return; // if no other rank;

	const materialLaw* mlaw = this->getMaterialLaw();
	const materialLaw* mlawMinus = this->getMaterialLawMinus();
	if (mlawMinus== NULL) Msg::Error("getMaterialLawMinus is NULL");
	const materialLaw* mlawPlus = this->getMaterialLawPlus();
	if (mlawPlus== NULL) Msg::Error("getMaterialLawPlus is NULL");

	bool useNumMat = false;
	if (mlaw!=NULL){
		if (mlaw->isNumeric())
			useNumMat = true;
	}
	if (mlawMinus->isNumeric()){
		useNumMat = true;
	}
	if (mlawPlus->isNumeric()){
		useNumMat = true;
	}
	if (!useNumMat) return;
	// initMap is created when at leat one material law (among mlaw, mlawMinus, mlawPlus) is numeric

	if (Msg::GetCommRank() == _rootRank){
		IntPt* GP;

		std::set<int> allIPBulk; // for finding

		for (groupOfElements::elementContainer::iterator it = this->g->begin(); it != this->g->end(); ++it){
			MElement *ele = *it;
			int npts_bulk=this->getBulkGaussIntegrationRule()->getIntPoints(ele,&GP);
			for(int i=0;i<npts_bulk;i++){
				int num = numericalMaterialBase::createTypeWithTwoInts(ele->getNum(),i);
				allIPBulk.insert(num);
			}
		}

		// for minus part

		std::set<int> allIPMinus; // for finding
		for(groupOfElements::elementContainer::const_iterator it=this->gi->begin(); it!=this->gi->end();++it){
			MElement *ele = *it;
			int npts=this->getInterfaceGaussIntegrationRule()->getIntPoints(ele,&GP);
			for(int i=0;i<npts;i++){
				int num = numericalMaterialBase::createTypeWithTwoInts(ele->getNum(),i);
				allIPMinus.insert(num);
			}
		}
		// for plus part
		std::set<int> allIPlus; // for finding
		for(groupOfElements::elementContainer::const_iterator it=this->gi->begin(); it!=this->gi->end();++it){
			MElement *ele = *it;
			int npts=this->getInterfaceGaussIntegrationRule()->getIntPoints(ele,&GP);
			for(int i=0;i<npts;i++){
				int num = numericalMaterialBase::createTypeWithTwoInts(ele->getNum(),i+npts);
				allIPlus.insert(num);
			}
		}

		// distribute
		std::vector<int> allIP;
		if (mlaw!=NULL){
			if (mlaw->isNumeric()){
				for (std::set<int>::iterator it = allIPBulk.begin(); it!= allIPBulk.end(); it++){
					allIP.push_back(*it);
				}
			}
		}

		if (mlawMinus->isNumeric()){
			for (std::set<int>::iterator it = allIPMinus.begin(); it!= allIPMinus.end(); it++){
				allIP.push_back(*it);
			}
		}

		if (mlawPlus->isNumeric()){
			if (_averageStrainBased and  (mlawMinus->getNum() == mlawPlus->getNum())){
				// do nothing
			}
			else{
				for (std::set<int>::iterator it = allIPlus.begin(); it!= allIPlus.end(); it++){
					allIP.push_back(*it);
				}
			}
		}

		std::vector<int> allRanks(_otherRanks.begin(),_otherRanks.end());
		if (_distributedOnRoot)
			allRanks.push_back(_rootRank);
		// distribute multiscale IP
		std::map<int,std::set<int> > mapAllIP;
		numericalMaterialBase::distributeMacroIP(allIP,allRanks,mapAllIP);

		_mapIPBulk.clear();
		_mapIPInterfaceMinus.clear();
		_mapIPInterfacePlus.clear();
		for (int ip=0; ip < allRanks.size(); ip++){
			std::set<int> bulkPart;
			std::set<int> minusPart;
			std::set<int> plusPart;
			const int& rank = allRanks[ip];
			const std::set<int>& ipOnRank = mapAllIP[rank];
			for (std::set<int>::const_iterator itIP = ipOnRank.begin(); itIP!= ipOnRank.end(); itIP++){
				const int& num =  *itIP;
				if (allIPBulk.find(num) != allIPBulk.end()){
					bulkPart.insert(num);
				}
				else if (allIPMinus.find(num) != allIPMinus.end()){
					minusPart.insert(num);
				}
				else if (allIPlus.find(num) != allIPlus.end()){
					plusPart.insert(num);
				}
				else{
					Msg::Error("IP %d is not correctly defined");
				}
			}

			_mapIPBulk[rank] = bulkPart;
			_mapIPInterfaceMinus[rank] = minusPart;
			_mapIPInterfacePlus[rank] = plusPart;
		}

		// send data to other Ranks
	 // Send data for bulk elements
		for (std::map<int,std::set<int> >::iterator it = _mapIPBulk.begin(); it!= _mapIPBulk.end(); it++){
			int rank = it->first;
			std::set<int>& numElOnRank = _mapIPBulk[rank];
			// at root rank
			if (rank == _rootRank){
				_domainIPBulk = numElOnRank;
			}
			// at other ranks
			else{
				int bufferSize = numElOnRank.size();
				MPI_Send(&bufferSize,1,MPI_INT,rank,numericalMaterialBase::createTypeWithTwoInts(getPhysical(),0),MPI_COMM_WORLD);
				if (bufferSize>0){
					int* buffer = new int[bufferSize];
					int idex = 0;
					for (std::set<int>::iterator its = numElOnRank.begin(); its != numElOnRank.end(); its++){
						buffer[idex] = *its;
						idex++;
					};
					MPI_Send(buffer,bufferSize,MPI_INT,rank,numericalMaterialBase::createTypeWithTwoInts(getPhysical(),1),MPI_COMM_WORLD);
					delete[] buffer;
				}
			}
		};

		// send data for minus interface IP
		for (std::map<int,std::set<int> >::iterator it = _mapIPInterfaceMinus.begin(); it!= _mapIPInterfaceMinus.end(); it++){
			int rank = it->first;
			std::set<int>& numElOnRank = _mapIPInterfaceMinus[rank];
			// at root rank
			if (rank == _rootRank){
				_domainIPInterfaceMinus = numElOnRank;
			}
			// at other ranks
			else{
				int bufferSize = numElOnRank.size();
				MPI_Send(&bufferSize,1,MPI_INT,rank,numericalMaterialBase::createTypeWithTwoInts(getPhysical(),2),MPI_COMM_WORLD);
				if (bufferSize>0){
					int* buffer = new int[bufferSize];
					int idex = 0;
					for (std::set<int>::iterator its = numElOnRank.begin(); its != numElOnRank.end(); its++){
						buffer[idex] = *its;
						idex++;
					};
					MPI_Send(buffer,bufferSize,MPI_INT,rank,numericalMaterialBase::createTypeWithTwoInts(getPhysical(),3),MPI_COMM_WORLD);
					delete[] buffer;
				}
			}
		};

		// send data for plus interface IP
		for (std::map<int,std::set<int> >::iterator it = _mapIPInterfacePlus.begin(); it!= _mapIPInterfacePlus.end(); it++){
			int rank = it->first;
			std::set<int>& numElOnRank = _mapIPInterfacePlus[rank];
			// at root rank
			if (rank == _rootRank){
				_domainIPInterfacePlus = numElOnRank;
			}
			// at other ranks
			else{
				int bufferSize = numElOnRank.size();
				MPI_Send(&bufferSize,1,MPI_INT,rank,numericalMaterialBase::createTypeWithTwoInts(getPhysical(),4),MPI_COMM_WORLD);
				if (bufferSize>0){
					int* buffer = new int[bufferSize];
					int idex = 0;
					for (std::set<int>::iterator its = numElOnRank.begin(); its != numElOnRank.end(); its++){
						buffer[idex] = *its;
						idex++;
					};
					MPI_Send(buffer,bufferSize,MPI_INT,rank,numericalMaterialBase::createTypeWithTwoInts(getPhysical(),5),MPI_COMM_WORLD);
					delete[] buffer;
				}
			}
		};

		// add to root rank if material is not multiscale
		if (mlaw!=NULL){
			if (!mlaw->isNumeric()){
				for (std::set<int>::iterator it = allIPBulk.begin(); it!= allIPBulk.end(); it++){
					_domainIPBulk.insert(*it);
				}
			}
		}

		if (!mlawMinus->isNumeric()){
			for (std::set<int>::iterator it = allIPMinus.begin(); it!= allIPMinus.end(); it++){
				_domainIPInterfaceMinus.insert(*it);
			}
		}

		if (!mlawPlus->isNumeric()){
			if (_averageStrainBased and  (mlawMinus->getNum() == mlawPlus->getNum())){
				// do nothing
			}
			else{
				for (std::set<int>::iterator it = allIPlus.begin(); it!= allIPlus.end(); it++){
					_domainIPInterfacePlus.insert(*it);
				}
			}
		}

	}
	else if (_otherRanks.find(Msg::GetCommRank()) != _otherRanks.end()) {
		// receive bulk part
		// receive data from root for bulk elements
		MPI_Status status;
		int bufferSize;
		int* buffer = NULL;

		MPI_Recv(&bufferSize,1,MPI_INT,_rootRank,numericalMaterialBase::createTypeWithTwoInts(getPhysical(),0),MPI_COMM_WORLD,&status);
		if (bufferSize>0){
			buffer = new int[bufferSize];
			MPI_Recv(buffer,bufferSize,MPI_INT,_rootRank,numericalMaterialBase::createTypeWithTwoInts(getPhysical(),1),MPI_COMM_WORLD,&status);
			_domainIPBulk.clear();
			for (int i=0; i<bufferSize; i++)
				_domainIPBulk.insert(buffer[i]);
			delete[] buffer;
		}

		// receive minus part
		MPI_Recv(&bufferSize,1,MPI_INT,_rootRank,numericalMaterialBase::createTypeWithTwoInts(getPhysical(),2),MPI_COMM_WORLD,&status);
		if (bufferSize>0){
			buffer = new int[bufferSize];
			MPI_Recv(buffer,bufferSize,MPI_INT,_rootRank,numericalMaterialBase::createTypeWithTwoInts(getPhysical(),3),MPI_COMM_WORLD,&status);
			_domainIPInterfaceMinus.clear();
			for (int i=0; i<bufferSize; i++)
				_domainIPInterfaceMinus.insert(buffer[i]);
			delete[] buffer;
		}

		// receive plus part
		MPI_Recv(&bufferSize,1,MPI_INT,_rootRank,numericalMaterialBase::createTypeWithTwoInts(getPhysical(),4),MPI_COMM_WORLD,&status);
		if (bufferSize>0){
			buffer = new int[bufferSize];
			MPI_Recv(buffer,bufferSize,MPI_INT,_rootRank,numericalMaterialBase::createTypeWithTwoInts(getPhysical(),5),MPI_COMM_WORLD,&status);
			_domainIPInterfacePlus.clear();
			for (int i=0; i<bufferSize; i++)
				_domainIPInterfacePlus.insert(buffer[i]);
			delete[] buffer;
		}

	}
};
#endif // HAVE_MPI

void dgPartDomain::createIPState(AllIPState::ipstateContainer& map, const bool *state) const{
  partDomain::createIPState(map,state);
  IntPt* GP;
  const materialLaw *mlawMinus = this->getMaterialLawMinus();
  const materialLaw *mlawPlus = this->getMaterialLawPlus();

	#if defined(HAVE_MPI)
  if ((mlawMinus->isNumeric() or mlawPlus->isNumeric()) and _otherRanks.size()>0){
    if (Msg::GetCommRank() == _rootRank){
      // loop
      for(groupOfElements::elementContainer::const_iterator it=this->gi->begin(); it!=this->gi->end();++it){
        MElement *ele = *it;
        MInterfaceElement* iele = dynamic_cast<MInterfaceElement*>(ele);
      // 2* because IP is duplicated for DG formulation
        int npts = this->getInterfaceGaussIntegrationRule()->getIntPoints(ele,&GP);
        int npts_inter=2*npts;
        AllIPState::ipstateElementContainer tp(npts_inter);
        for(int i=0;i<npts;i++){
          if (mlawMinus->isNumeric()){
            int num = numericalMaterialBase::createTypeWithTwoInts(ele->getNum(),i);
            if (_domainIPInterfaceMinus.find(num) != _domainIPInterfaceMinus.end()){
              mlawMinus->createIPState(tp[i],state,ele,iele->getElem(0)->getNumVertices(),&(GP[i]),i);
            }
            else{
              const numericalMaterialBase* numatMinus = dynamic_cast<const numericalMaterialBase*>(mlawMinus);
              if (numatMinus)
                numatMinus->createIPState(false,tp[i],state,ele,iele->getElem(0)->getNumVertices(),&(GP[i]),i);
              else
                Msg::Error("Error error at creating the negative IPState");
            }
          }
          else{
            mlawMinus->createIPState(tp[i],state,ele,iele->getElem(0)->getNumVertices(),&(GP[i]),i);
          }
          tp[i]->setLocation(IPVariable::INTERFACE_MINUS);
        }
        for(int i=npts;i<npts_inter;i++){
          if (mlawPlus->isNumeric()){
            int num = numericalMaterialBase::createTypeWithTwoInts(ele->getNum(),i);
            if (_domainIPInterfacePlus.find(num) != _domainIPInterfacePlus.end()){
              mlawPlus->createIPState(tp[i],state,ele,iele->getElem(1)->getNumVertices(),&(GP[i-npts]),i);
            }
            else{
              const numericalMaterialBase* numatPlus = dynamic_cast<const numericalMaterialBase*>(mlawPlus);
              if (numatPlus)
                numatPlus->createIPState(false,tp[i],state,ele,iele->getElem(1)->getNumVertices(),&(GP[i-npts]),i);
              else
                Msg::Error("Error error at creating the positive IPState");
            }
          }
          else{
            mlawPlus->createIPState(tp[i],state,ele,iele->getElem(1)->getNumVertices(),&(GP[i-npts]),i);
          }

          tp[i]->setLocation(IPVariable::INTERFACE_PLUS);
        }
        map.insert(AllIPState::ipstatePairType(ele->getNum(),tp));
      }
    }
    else if (_otherRanks.find(Msg::GetCommRank()) != _otherRanks.end()) {
      for (std::set<int>::iterator it = _domainIPInterfaceMinus.begin(); it != _domainIPInterfaceMinus.end(); it++){
        int num = *it;
        int elnum, gnum;
        IntPt *GPnum;
        numericalMaterialBase::getTwoIntsFromType(num,elnum,gnum);
        MElement* ele = new MInterfacePoint(NULL,elnum,0);
        AllIPState::ipstateElementContainer tp(1);
        int npts_int=this->getInterfaceGaussIntegrationRule()->getIntPoints(ele,&GPnum);
        mlawMinus->createIPState(tp[0],state,ele,ele->getNumVertices(),&(GPnum[0]),gnum);
        tp[0]->setLocation(IPVariable::INTERFACE_MINUS);
        delete ele;
        map.insert(AllIPState::ipstatePairType(num,tp));
      };

      for (std::set<int>::iterator it = _domainIPInterfacePlus.begin(); it != _domainIPInterfacePlus.end(); it++){
        int num = *it;
        int elnum, gnum;
        IntPt *GPnum;
        numericalMaterialBase::getTwoIntsFromType(num,elnum,gnum);
        MElement* ele = new MInterfacePoint(NULL,elnum,0);
        AllIPState::ipstateElementContainer tp(1);
        int npts_int=this->getInterfaceGaussIntegrationRule()->getIntPoints(ele,&GPnum);
        mlawPlus->createIPState(tp[0],state,ele,ele->getNumVertices(),&(GPnum[0]),gnum);
        tp[0]->setLocation(IPVariable::INTERFACE_PLUS);
        delete ele;
        map.insert(AllIPState::ipstatePairType(num,tp));
      };
    }
  }
  else
	#endif //HAVE_MPI
	{
    // loop
    for(groupOfElements::elementContainer::const_iterator it=this->gi->begin(); it!=this->gi->end();++it){
      MElement *ele = *it;
      MInterfaceElement* iele = dynamic_cast<MInterfaceElement*>(ele);
    // 2* because IP is duplicated for DG formulation
      int npts = this->getInterfaceGaussIntegrationRule()->getIntPoints(ele,&GP);
      int npts_inter=2* npts;
      AllIPState::ipstateElementContainer tp(npts_inter);
      for(int i=0;i<npts;i++){
        mlawMinus->createIPState(tp[i],state,ele,iele->getElem(0)->getNumVertices(),&(GP[i]),i);
        tp[i]->setLocation(IPVariable::INTERFACE_MINUS);
      }
      for(int i=npts;i<npts_inter;i++){
        mlawPlus->createIPState(tp[i],state,ele,iele->getElem(1)->getNumVertices(),&(GP[i-npts]),i);
        tp[i]->setLocation(IPVariable::INTERFACE_PLUS);
      }
      map.insert(AllIPState::ipstatePairType(ele->getNum(),tp));
    }
  }


  // Virtual interface element (no duplication)
  const materialLaw *mlaw = this->getMaterialLaw();
  for(groupOfElements::elementContainer::const_iterator it=this->gib->begin(); it!=this->gib->end();++it){
    MElement *ele = *it;
    MInterfaceElement *iele = dynamic_cast<MInterfaceElement*>(ele);
    int npts_inter=this->getInterfaceGaussIntegrationRule()->getIntPoints(ele,&GP);
    AllIPState::ipstateElementContainer tp(npts_inter);
    for(int i=0;i<npts_inter;i++){
      mlaw->createIPState(tp[i],state,ele,iele->getElem(0)->getNumVertices(),&(GP[i]),i);
      tp[i]->setLocation(IPVariable::VIRTUAL_INTERFACE);
    }
    map.insert(AllIPState::ipstatePairType(ele->getNum(),tp));
  }
};

void dgPartDomain::initiallyBreakAllInterfaceIP(AllIPState *aips) const{
  const materialLaw* mlawMinus = this->getMaterialLawMinus();
  const materialLaw* mlawPlus = this->getMaterialLawPlus();
  const materialLaw2LawsInitializer* fracMinus = dynamic_cast<const materialLaw2LawsInitializer*>(mlawMinus);
  const materialLaw2LawsInitializer* fracPlus = dynamic_cast<const materialLaw2LawsInitializer*>(mlawPlus);
  #if defined(HAVE_MPI)
  if (Msg::GetCommRank() == _rootRank)
  #endif // HAVE_MPI
  {
    IntPt* GP;
    for (groupOfElements::elementContainer::iterator it = gi->begin(); it!= gi->end(); it++){
      MElement* ele = *it;

      AllIPState::ipstateElementContainer *vips = aips->getIPstate(ele->getNum());
      int npts=this->getInterfaceGaussIntegrationRule()->getIntPoints(ele,&GP);

      for (int j=0; j< npts; j++){
        IPStateBase* ipsm = (*vips)[j];
        IPStateBase* ipsp = (*vips)[j+npts];
        if (fracMinus!=NULL){
          fracMinus->initialBroken(ipsm);
        }
        if (fracPlus!=NULL){
          fracPlus->initialBroken(ipsp);
        }
      }
    }
  }
  #if defined(HAVE_MPI)
  else if (_otherRanks.find(Msg::GetCommRank()) != _otherRanks.end())
  {
    if ((mlawMinus->isNumeric()) and (fracMinus!=NULL)){
      for (std::set<int>::const_iterator it = _domainIPInterfaceMinus.begin(); it!= _domainIPInterfaceMinus.end(); it++){
        AllIPState::ipstateElementContainer *vips = aips->getIPstate(*it);
        fracMinus->initialBroken((*vips)[0]);
      }
    }

    if ((mlawPlus->isNumeric()) and (fracPlus!=NULL)){
      for (std::set<int>::const_iterator it = _domainIPInterfacePlus.begin(); it!= _domainIPInterfacePlus.end(); it++){
        AllIPState::ipstateElementContainer *vips = aips->getIPstate(*it);
        fracPlus->initialBroken((*vips)[0]);
      }
    }

  }
  #endif // HAVE_MPI
};


void dgPartDomain::numberValuesToTransfertMPI(AllIPState *aips,int* sizeeachrank)const
{
 #if defined(HAVE_MPI)
  IntPt *GP;
  for(groupOfElements::elementContainer::iterator ite=gi->begin(); ite!=gi->end(); ++ite)
  {
    MElement* ele = *ite;
    MInterfaceElement *iele = dynamic_cast<MInterfaceElement*>(ele);
    int rankminus = iele->getElem(0)->getPartition()-1;
    int rankplus = iele->getElem(1)->getPartition()-1;
    int rankcomm=-1;
    if(rankminus != Msg::GetCommRank())
    {
       rankcomm = rankminus;
    }
    else if(rankplus !=Msg::GetCommRank())
    {
      rankcomm = rankplus;
    }
    if(rankcomm!=-1)
    {
      int npts=integBound->getIntPoints(*ite,&GP);
      AllIPState::ipstateElementContainer *vips = aips->getIPstate(ele->getNum());
      int nvc=0;
      for(int j=0;j<npts;j++)
      {
         IPStateBase* ipsm = (*vips)[j];
         // take initial but supposed same value to transfert for all (so transfert only once)
         IPVariable *ipvm = static_cast<IPVariableMechanics*>(ipsm->getState(IPStateBase::initial));
         nvc += ipvm->numberValuesToCommunicateMPI();
      }
      if(nvc!=0) // otherwise no value to communicate
        sizeeachrank[rankcomm] += (2+nvc); // as the numbers of element minus and plus have to be communicate
    }
  }
 #endif // HAVE_MPI
}

void dgPartDomain::fillArrayIPvariableMPI(AllIPState *aips,const int rank,double* arrayMPI) const
{
 #if defined(HAVE_MPI)
  // We communicate also the number of each element (to identify later the interface)
  // so by element the number of value to tranfert is equal to 2 + npts
  // Only the value of minus element is transfert after in init term val + = val -
  IntPt *GP;
  int curpos=0;
  for(groupOfElements::elementContainer::iterator ite=gi->begin(); ite!=gi->end(); ++ite)
  {
    MElement* ele = *ite;
    MInterfaceElement *iele = dynamic_cast<MInterfaceElement*>(ele);
    int rankminus = iele->getElem(0)->getPartition()-1;
    int rankplus = iele->getElem(1)->getPartition()-1;
    if((rankminus == rank) or (rankplus == rank))
    {
      // write information about the interface element (the interface element number is local --> different on each rank)
      arrayMPI[curpos] = (double) iele->getElem(0)->getNum();
      arrayMPI[curpos+1] = (double) iele->getElem(1)->getNum();
      // get the value of the cohesive strength on minus element of each integration point
      // same for initial, previous and current.
      int npts=integBound->getIntPoints(*ite,&GP);
      AllIPState::ipstateElementContainer *vips = aips->getIPstate(ele->getNum());
      curpos +=2;
      for(int i=0;i<npts;i++)
      {
         IPStateBase* ipsm = (*vips)[i];
         IPVariable  *ipvm = ipsm->getState(IPStateBase::initial);
         ipvm->fillValuesToCommunicateMPI(&arrayMPI[curpos]);
         curpos+=ipvm->numberValuesToCommunicateMPI();
      }
    }
  }
 #endif // HAVE_MPI
}

void dgPartDomain::setIPVariableFromArrayMPI(AllIPState *aips,const int rank,const int arraysize,const double* arrayMPI)const
{
 #if defined(HAVE_MPI)
  int curpos=0;
  IntPt *GP;
  while(curpos<arraysize)
  {
     // get information to find the interelement (the number of the interface element is a local number which differ on each rank)
    int numminus = (int) arrayMPI[curpos];
    int numplus = (int) arrayMPI[curpos+1];
    MElement* elefound=NULL;
    for(groupOfElements::elementContainer::iterator it=gi->begin(); it!=gi->end();++it)
    {
      MInterfaceElement *ie = dynamic_cast<MInterfaceElement*>(*it);
      if((ie->getElem(0)->getNum()==numminus) and (ie->getElem(1)->getNum()==numplus))
      {
        elefound = *it;
        break;
      }
    }
    if(elefound!=NULL){
      int npts=integBound->getIntPoints(elefound,&GP);
      AllIPState::ipstateElementContainer *vips = aips->getIPstate(elefound->getNum());
      curpos+=2;
      for(int i=0;i<npts;i++)
      {
         IPStateBase* ipsm = (*vips)[i];
         // same values for the 3 states
         IPVariable *ipvm = ipsm->getState(IPStateBase::initial);
         ipvm->getValuesFromMPI(&arrayMPI[curpos]);
         ipvm = ipsm->getState(IPStateBase::previous);
         ipvm->getValuesFromMPI(&arrayMPI[curpos]);
         ipvm = ipsm->getState(IPStateBase::current);
         ipvm->getValuesFromMPI(&arrayMPI[curpos]);
         curpos += ipvm->numberValuesToCommunicateMPI();
      }
    }
    else
    {
      Msg::Error("Can't find an interelement where IPVariable have to be set. dom %d rank %d",_phys,Msg::GetCommRank());
    }
  }
 #endif // HAVE_MPI
}

void dgPartDomain::initMicroMeshId(){
  // for bulk element
  materialLaw* mlaw = this->getMaterialLaw();
  if(mlaw!=NULL)
  {
    if (mlaw->isNumeric()){
      partDomain::initMicroMeshId();
    }
  }

  if (IsInterfaceTerms()){
    // for interface elements
    materialLaw* mlawMinus = this->getMaterialLawMinus();
    materialLaw* mlawPlus = this->getMaterialLawPlus();

    if (mlawPlus->isNumeric() or mlawMinus->isNumeric()){
      numericalMaterialBase* nummatMinus = dynamic_cast<numericalMaterialBase*>(mlawMinus);
      numericalMaterialBase* nummatPlus = dynamic_cast<numericalMaterialBase*>(mlawPlus);
      if ((nummatMinus == NULL) and (nummatPlus == NULL))
        Msg::Error("material law must be derived from numericalMaterialBase");

      IntPt* GP;
      for (groupOfElements::elementContainer::iterator it = gi->begin(); it!= gi->end(); it++){
        MElement* ele = *it;
        int npts = this->getInterfaceGaussIntegrationRule()->getIntPoints(ele,&GP);
        for (int i=0; i<npts; i++){
          if (mlawMinus->isNumeric() == true && mlawPlus->isNumeric() == false ){
            nummatMinus->assignMeshId(ele->getNum(),i);
          }
          else if (mlawMinus->isNumeric() == false && mlawPlus->isNumeric() ==true){
            nummatPlus->assignMeshId(ele->getNum(),i+npts);
          }
          else {
            #if defined(HAVE_MPI)
            MInterfaceElement* ie =dynamic_cast<MInterfaceElement*>(ele);
            int posProc = ie->getElem(1)->getPartition()-1;
            int negProc = ie->getElem(0)->getPartition()-1;
            if (posProc == negProc){
            #endif // HAVE_MPI
              if (mlawMinus->getNum() == mlawPlus->getNum()){
                // same material law
                nummatMinus->assignMeshId(ele->getNum(),i);
                nummatPlus->setMeshId(ele->getNum(),i+npts,nummatMinus->getMeshId(ele->getNum(),i));
              }
              else{
                // different material law
                nummatMinus->assignMeshId(ele->getNum(),i);
                nummatPlus->assignMeshId(ele->getNum(),i+npts);
              }

            #if defined(HAVE_MPI)
            }
            else{
              // id initiate in lower proc and send to upper proc
              int smaller = std::min(posProc,negProc);
              int larger = std::max(posProc,negProc);

              if (Msg::GetCommRank() == smaller){
                if (mlawMinus->getNum() == mlawPlus->getNum()){
                  // same material law
                  nummatMinus->assignMeshId(ele->getNum(),i);
                  nummatPlus->setMeshId(ele->getNum(),i+npts,nummatMinus->getMeshId(ele->getNum(),i));
                }
                else{
                  // different material law
                  nummatMinus->assignMeshId(ele->getNum(),i);
                  nummatPlus->assignMeshId(ele->getNum(),i+npts);
                }
                // send data
                int bufferMinus = nummatMinus->getMeshId(ele->getNum(),i);
                int tagMinus = numericalMaterialBase::createTypeWithTwoInts(ie->getElem(0)->getNum(),i);

                int tagPlus = numericalMaterialBase::createTypeWithTwoInts(ie->getElem(1)->getNum(),i+npts);
                int bufferPlus  = nummatPlus->getMeshId(ele->getNum(),i+npts);

                MPI_Send(&bufferMinus,1,MPI_INT,larger,tagMinus,MPI_COMM_WORLD);
                MPI_Send(&bufferPlus,1,MPI_INT,larger,tagPlus,MPI_COMM_WORLD);
              }
              else if (Msg::GetCommRank() == larger){
                 // receive data
                int bufferMinus, bufferPlus;
                int tagMinus = numericalMaterialBase::createTypeWithTwoInts(ie->getElem(0)->getNum(),i);
                int tagPlus = numericalMaterialBase::createTypeWithTwoInts(ie->getElem(1)->getNum(),i+npts);

                MPI_Status status;
                MPI_Recv(&bufferMinus,1,MPI_INT,smaller,tagMinus,MPI_COMM_WORLD,&status);
                MPI_Recv(&bufferPlus,1,MPI_INT,smaller,tagPlus,MPI_COMM_WORLD,&status);

                nummatMinus->setMeshId(ele->getNum(),i,bufferMinus);
                nummatPlus->setMeshId(ele->getNum(),i+npts,bufferPlus);
              }
            }

            #endif // HAVE_MPI
          }
        }
      }
    };
  }
};
