//
// C++ Interface: partDomain
//
// Description: Interface class to used solver. Your term ha to be store in a domain
//
// Author:  <Gauthier BECKER>, (C) 2010
//
// Copyright: See COPYING file that comes with this distribution
//
//
#ifndef PARTDOMAIN_H_
#define PARTDOMAIN_H_
#ifndef SWIG
#include "mlaw.h"
#include "SVector3.h"
#include "MInterfaceElement.h"
#include "groupOfElements.h"
#include "timeFunction.h"
#include "functionSpaceType.h"
#include "nonLinearBC.h"
#include "nlTerms.h"
#include "unknownField.h"
#include "InterfaceBuilder.h"
#include "ElementErosionFilter.h"
#include "interfaceQuadrature.h"
// class for a domain (pure virtual class)
#endif
class dgPartDomain;
class partDomain{
	#ifndef SWIG
 public :
	enum EnrichedUnknownLocation{UNDEFINED=0,NODE=1,GP=2};

 protected :
  int _tag; // tag for the dofManager
  int _phys;
  bool _fullDg; // To know which formulation Cg/Dg or FullDg is used for this part
  functionSpaceType::whichSpace _wsp;
  BilinearTermBase* btermBulk;
  BilinearTermBase* btermElasticBulk;
  BilinearTermBase* massterm;
  LinearTermBase<double>* ltermBulk;
  QuadratureBase* integBulk;
  BilinearTermBase* btermTangent; // tangent estimation by homogenization
  BilinearTermBase* btermElasticTangent; // elastic tangent estimation by homogenization
  // for pathfollowing
  ScalarTermBase<double>* scalarTermPF; // scalar pathfollowing term
  LinearTermBase<double>* linearTermPF; // linear pathFollowing term

  // for matrix by perturbation
  int _numberNonLocalVariable; // number of nonlocal variables
  int _numberConstitutiveExtraDofDiffusionVariable; // number of nonlocal variables
  bool _bmbp;
  double _eps;
  virtual void setBulkMatrixByPerturbation(const int i, const double eps=1e-8);
  // To know if law is already set
  bool setmaterial;

  #if defined(HAVE_MPI)
  int _rootRank;
  std::set<int> _otherRanks;
  std::set<int> _domainIPBulk;  // bulk integration points lie on this domain
  std::map<int,std::set<int> > _mapIPBulk; // map IP for bulk elements
  #endif //HAVE_MPI

  bool _distributedOtherRanks;
  bool _distributedOnRoot;

  const nonLinearMechSolver* _macroSolver; // solver where this domain belongs to

  // Todo protect this variable

  groupOfElements *g; // support for this field
public:
  // add a group of element for prescribed Dirichlet BC of other partition in MPI
  groupOfElements* _groupGhostMPI;

  groupOfElements::elementContainer::const_iterator g_cbegin() const {return g->begin();}
  groupOfElements::elementContainer::const_iterator g_cend() const {return g->end();}
  groupOfElements::vertexContainer::const_iterator g_cvbegin() const {return g->vbegin();}
  groupOfElements::vertexContainer::const_iterator g_cvend() const {return g->vend();}
  size_t groupOfElementsSize() const {return g->size();}
  void addPhysicalGroup(int dim,int phys,elementFilter& elemfil)
  {
      if(g==NULL) g = new groupOfElements();
      g->addPhysical(dim,phys,elemfil);
  }
  virtual void clearGroupOfElements() {g->clearAll();}
  bool g_find(MElement* e) const {return g->find(e);}
  const groupOfElements* elements() const {return g;}
	#endif // SWIG
 public:
  // _distributedRootRanks is true if the microscopic solver are computed at
  // root ranks, false otherwise
  virtual bool isDistributedRootRank() const;
  virtual void distributeOnRootRank(const int flg);
  // _distributedOtherRanks is true if microscopic solver are computed at other
 // ranks, false otherwise
 // if _distributedOtherRanks = true and _distributedRootRanks = true mean the
 // microsopic BVPs are solved in both root and other ranks
 // while _distributedRootRanks = false means that the microscopic BVPs are
 // solved in other ranks only --> better for manage load balancing
 // Clearly, _distributedOtherRanks and _distributedRootRanks can not be
 // simultaneously false
  virtual bool isDistributedOtherRanks() const;
  virtual void setDistributedOtherRanks(const bool flag);
#ifndef SWIG
  // Constructors
  partDomain(const int tag, const int phys,
             const int ws = 0,const bool fdg=false, int numNL=0, int numExtraDof=0) : g(NULL), _tag(tag), _phys(phys),
             _bmbp(false), _eps(1e-8), _fullDg(fdg), setmaterial(false),
             _numberNonLocalVariable(numNL),_numberConstitutiveExtraDofDiffusionVariable(numExtraDof),
             btermBulk(NULL),btermElasticBulk(NULL),massterm(NULL),ltermBulk(NULL),
             integBulk(NULL),btermTangent(NULL), btermElasticTangent(NULL),_distributedOtherRanks(true),
             _distributedOnRoot(true),scalarTermPF(NULL),linearTermPF(NULL),
						 _macroSolver(NULL)
  {
    _groupGhostMPI = new groupOfElements();
    switch(ws){
     case 0:
      _wsp = functionSpaceType::Lagrange;
      break;
     case 10000: // Allow to set space later (used for interface domain)
      _wsp = functionSpaceType::Inter;
      break;
     default:
      Msg::Error("Function space type is unknown for partDomain %d. So Lagrange by default");
      _wsp = functionSpaceType::Lagrange;
    }
    #if defined(HAVE_MPI)
    _rootRank = Msg::GetCommRank();
    #endif //HAVE_MPI
  }
  partDomain(const partDomain &source);
  virtual partDomain& operator=(const partDomain &source);
  virtual ~partDomain(){
    if (btermBulk) {delete btermBulk; btermBulk = NULL;}
    if (ltermBulk) {delete ltermBulk; ltermBulk = NULL;}
    if (integBulk) {delete integBulk; integBulk = NULL;}
    if (btermElasticBulk) {delete btermElasticBulk; btermElasticBulk = NULL;};
    if (massterm) {delete massterm; massterm = NULL;}
    if (btermTangent) {delete btermTangent; btermTangent= NULL;};
    if (g) {delete g; g = NULL;}
    if (_groupGhostMPI) {delete _groupGhostMPI; _groupGhostMPI = NULL;};
    if (scalarTermPF) {delete scalarTermPF; scalarTermPF = NULL;}
    if (linearTermPF) {delete linearTermPF; linearTermPF = NULL;}
  }
  BilinearTermBase* getBilinearBulkTerm() const{return btermBulk;}
  BilinearTermBase* getBilinearElasticBulkTerm() const{return btermElasticBulk;}
  BilinearTermBase* getBilinearMassTerm() const{return massterm;}
  BilinearTermBase* getBilinearTangentTerm() const{return btermTangent;}
  BilinearTermBase* getBilinearElasticTangentTerm() const{return btermElasticTangent;}
  LinearTermBase<double>* getLinearBulkTerm() const{return ltermBulk;}

  ScalarTermBase<double>* getScalarTermPFBulk() const {return scalarTermPF;};
  LinearTermBase<double>* getLinearTermPFBulk() const {return linearTermPF;};

  void setMacroSolver(const nonLinearMechSolver* solver);
  const nonLinearMechSolver* getMacroSolver() const;

  QuadratureBase* getBulkGaussIntegrationRule() const{return integBulk;}
  int getNumNonLocalVariable() const {return _numberNonLocalVariable;};
  int getNumConstitutiveExtraDofDiffusionVariable() const {return _numberConstitutiveExtraDofDiffusionVariable;};
  virtual double getConstitutiveExtraDofDiffusionEqRatio(const int index) const {
    Msg::Error("partDomain::getConstitutiveExtraDofDiffusionEqRatio is not defined");
    return -1.;
  };
  virtual double getNonConstitutiveExtraDofEqRatio(const int index) const {
    Msg::Error("partDomain::getNonConstitutiveExtraDofEqRatio is not defined");
    return -1.;
  };

  // Dimension of domain
  virtual int getDim() const=0;
  int getTag()const{return _tag;}
  int getPhysical() const{return _phys;}

  virtual bool getAccountPathFollowingFlag() const {return true;};

  //check cohesive element insertion, failure
  virtual void checkFailure(IPField* ipf) const = 0;

  // check internal stata
  virtual void checkInternalState(IPField* ipf) const = 0;


  virtual void initializeTerms(unknownField *uf,IPField *ip)=0;
  // alocate the object (new) but is not responsiblefor the delete (the NeumannBC is in charged of the delete)
  virtual LinearTermBase<double>* createNeumannTerm(FunctionSpace<double> *spneu, const groupOfElements* g,
                                                    const simpleFunctionTime<double>* f, const unknownField *uf,
                                                    const IPField * ip, const nonLinearBoundaryCondition::location onWhat,
                                                    const nonLinearNeumannBC::NeumannBCType neumann_type,
                                                    const int comp) const=0;
  virtual BilinearTermBase* createNeumannMatrixTerm(FunctionSpace<double> *spneu, const groupOfElements* g,
                                                    const simpleFunctionTime<double>* f, const unknownField *uf,
                                                    const IPField * ip, const nonLinearBoundaryCondition::location onWhat,
                                                    const nonLinearNeumannBC::NeumannBCType neumann_type,
                                                    const int comp) const;
  // true is domain has interface terms
  virtual bool IsInterfaceTerms() const=0;
  functionSpaceType::whichSpace getFunctionSpaceType() const{return _wsp;}

  virtual FunctionSpaceBase* getFunctionSpace() =0;
  virtual const FunctionSpaceBase* getFunctionSpace() const=0;
  
  virtual void getNoMassDofs(std::set<Dof>& DofVector) const{};
  
// some data of BC have to be set by domain
  virtual FunctionSpaceBase* getSpaceForBC(const nonLinearBoundaryCondition::type bc_type, 
                                           const nonLinearBoundaryCondition::location bc_location,
                                           const nonLinearNeumannBC::NeumannBCType neumann_type,
                                           const int dof_comp,const groupOfElements *groupBC) const=0; // for dirichlet, neumann and initial
                                                                                    // groupBC = element of domain where BC is applied
  virtual QuadratureBase* getQuadratureRulesForNeumannBC(nonLinearNeumannBC &neu) const=0;

  virtual void createIPState(AllIPState::ipstateContainer& map, const bool *state) const;

  virtual void computeIPVariable(AllIPState *aips,const unknownField *ufield,const IPStateBase::whichState ws, bool stiff)=0;
  virtual void computeIpv(AllIPState *aips,MElement *e, IPStateBase::whichState ws,
                           materialLaw *mlaw,fullVector<double> &disp, bool stiff)=0;
  virtual void setGaussIntegrationRule()=0;
  virtual bool getFormulation() const{return _fullDg;}
  virtual void setMaterialLaw(const std::map<int,materialLaw*> &maplaw)=0;

  // No materialLaw store here but the function to give the material law exist
  // The law has to be store in the derivated class
  virtual materialLaw* getMaterialLaw()=0;
  virtual const materialLaw* getMaterialLaw() const=0;
  virtual int getLawNum() const=0;
  // For dg formulation the stability parameters decrease the critical time step size
  // in explicit. So this function allows to scale the time step
  virtual double scaleTimeStep() const {return 1.;}
  // creation of interface. At least boundary to create interface domain
  // can be empty be not interdomain creation in this case
  virtual void createInterface(manageInterface &maninter)=0;
  virtual MElement* createVirtualInterface(IElement *ie) const=0;
  virtual MElement* createInterface(IElement* ie1,IElement* ie2) const=0;
  // create an interface domain. Use only in mpi case (an interface domain can be given or a new one has to be
                                                       // created if dgdom==NULL)
  virtual void createInterfaceMPI(const int rankOtherPart, const groupOfElements &groupOtherPart,dgPartDomain* &dgdom){
    Msg::Error("To use parallel implementation you have to define createInterfaceMPI on your domain. Or your version is obsoleted to domain has to be given as last argument");
  }
  // Add for BC
  virtual FilterDof* createFilterDof(const int comp) const=0;

  // specify order of computational homogenization uses
  int getHomogenizationOrder() const;
  bool inMultiscaleSimulation()  const;
  bool isExtractCohesiveLawFromMicroDamage() const;
  bool getExtractIrreversibleEnergyFlag() const;
  const elementErosionFilter& getElementErosionFilter() const;

  virtual bool IsMultipleFieldFormulation() const {return false;};
  virtual EnrichedUnknownLocation getEnrichedUnknownLocation() const {return UNDEFINED;};

  virtual int getStressDimension() const{
    // this function is used to define the dimension of stress matrix in the
    // insystem nonliner system
    Msg::Error("This function must be defined if homogenized tangent is estimated by insystem method");
  }

  //compute mechanical average quantities
  virtual void computeAverageStress(const IPField* ipf, STensor3& stress) const;
  virtual void computeActiveDissipationAverageStressIncrement(const IPField* ipf, STensor3& stress) const;
  virtual void computeAverageHighOrderStress(const IPField* ipf, STensor33& stress) const;
  virtual double computeVolumeDomain(const IPField* ipf) const;

  virtual void computeActiveDissipationCenter(const IPField* ipf, SPoint3& pos) const;
  virtual void computeActiveDissipationAverageStrainIncrement(const IPField* _ipf, STensor3& strain) const;
  virtual double computeVolumeActiveDissipationDomain(const IPField* ipf) const;

  // compute constitutive extra-dof quantities
  virtual void computeConstitutiveExtraDofAverageFlux(const int index, const IPField* _ipf, SVector3& extraDofFluxpart) const;
  virtual void computeConstitutiveExtraDofAverageInternalEnergy(const int index, const IPField* _ipf, double &internalEnergyExtraDofpart) const;
  virtual void computeConstitutiveExtraDofAverageFieldCapacityPerUnitField(const int index, const IPField* _ipf, double& fieldCapacityPart) const;
  virtual void computeConstitutiveExtraDofAverageMechanicalSource(const int index, const IPField* _ipf, double& mechanicalSource) const;

  // compute non-constitutive extra-dof quantities
  virtual void computeNonConstitutiveExtraDofAverageFlux(const int index, const IPField* _ipf, SVector3& extraDofFluxpart) const;

  //
  virtual double computeDeformationEnergy(const IPField* ipf) const;
  virtual double computePlasticEnergy(const IPField* ipf) const;
  virtual double computeIrreversibleEnergy(const IPField* ipf) const;

  virtual partDomain* clone() const{
    #ifdef _DEBUG
    Msg::Warning("Define copyDomain on your domain for multiscale problem");
    #endif
    return NULL;
  }
  #if defined(HAVE_MPI)
  virtual void setWorkingRanks(const int root, const std::set<int>& others);
  virtual int getRootRank() const;
  virtual void getOtherRanks(std::set<int>& others) const;
  virtual void clearOtherRanks();
  virtual void createIPMap();
  virtual const std::map<int,std::set<int> >& getMapIPBulk() const {return _mapIPBulk;};
  virtual const std::set<int>& getDomainIPBulk() const {return _domainIPBulk;};
  #endif //HAVE_MPI
  virtual double computeMeanValueDomain(const int num, const IPField* ipf, double& volume) const;
  virtual double averagingOnActiveDissipationZone(const int num, const IPField* ipf, double & volume, const IPStateBase::whichState ws = IPStateBase::current) const;
  virtual void initMicroMeshId();
#endif // SWIG
};
// class for Dg part domain (pure virtual)
class dgPartDomain : public partDomain{
	#ifndef SWIG
 public:
  enum MultiFieldMethod{
            ONEFIELD=0,
            TWOFIELD=1, // full
                        }; //
 protected :
  BilinearTermBase* btermBound;
  LinearTermBase<double>* ltermBound;
  BilinearTermBase* btermVirtBound;
  LinearTermBase<double>* ltermVirtBound;
  QuadratureBase* integBound;

	// for pathfollowing
	ScalarTermBase<double>* scalarTermPFBound; // scalar pathfollowing term
	LinearTermBase<double>* linearTermPFBound; // linear pathFollowing term

  // For matrix by perturbation
  bool _interByPert;
  bool _virtByPert;

  bool _multipleFieldFormulation;
	MultiFieldMethod _multiFieldMethod; // 0  fullTwoField, 1- augemented twofield, 2-DG twofield
	EnrichedUnknownLocation _enrichedUnknownLocation;

  bool _averageStrainBased;
  bool _brokenIPFromBeginning;
  std::vector<int> _fixInterfaceDofComp; // fix components of interface Dofs

  #if defined(HAVE_MPI)
  std::set<int>  _domainIPInterfacePlus, // plus intergration points lie on this domain
                  _domainIPInterfaceMinus; //  minus integration points lie on this domain
  std::map<int,std::set<int> > _mapIPInterfacePlus, // map IP for positive part of interface elements
                              _mapIPInterfaceMinus; // map IP for negative part of interface elements
  #endif //HAVE_MPI


 public :
  // TODO protect these variables
  groupOfElements *gi; // support for the interfaceElement TODO cast to a groupOfElements
  groupOfElements *gib; // support for the interfaceElement TODO cast to a groupOfElements
  groupOfElements *giv; // support for the virtual interface element (used to set Neumann and Dirichlet BC)

  virtual void clearGroupOfElements()
  {
    partDomain::clearGroupOfElements();
    gi->clearAll();
  }
	#endif // SWIG
 public:

  virtual void averageStrainBased(const bool flg);
  virtual void setMultipleFieldFormulation(const bool flag, const int method, const int loc=2);
  virtual void fixInterfaceDofComponent(const int comp);
  virtual void breakAllInterfaceIPsFromBeginning(const bool broken);

#ifndef SWIG
  dgPartDomain(const int tag, const int phys, const int ws = 0,
               const bool fdg=false, int numNL=0, int numCED=0) : partDomain(tag,phys,ws,fdg,numNL,numCED), _interByPert(false), _virtByPert(false),
               btermBound(NULL),ltermBound(NULL),btermVirtBound(NULL),ltermVirtBound(NULL),integBound(NULL),
               _multipleFieldFormulation(false), _multiFieldMethod(dgPartDomain::ONEFIELD), _enrichedUnknownLocation(partDomain::UNDEFINED),
							 _averageStrainBased(false),_brokenIPFromBeginning(false),
							 scalarTermPFBound(NULL),linearTermPFBound(NULL)
  {
    gi = new groupOfElements();
    gib = new groupOfElements();
    giv = new groupOfElements();
  }
  dgPartDomain(const dgPartDomain &source);
  virtual dgPartDomain& operator=(const partDomain &source);
  virtual ~dgPartDomain(){
    if (btermBound) {delete btermBound; btermBound = NULL;};
    if (ltermBound) {delete ltermBound; ltermBound = NULL;};
    if (btermVirtBound) {delete btermVirtBound; btermVirtBound= NULL;};
    if (ltermVirtBound) {delete ltermVirtBound; ltermVirtBound=NULL;};
    if (integBound) {delete integBound; integBound = NULL;};
    if (gi) {delete gi; gi = NULL;};
    if (gib) {delete gib; gib = NULL;};
    if (giv) {delete giv; giv = NULL;};

		if (scalarTermPFBound) {delete scalarTermPFBound; scalarTermPFBound = NULL;}
		if (linearTermPFBound) {delete linearTermPFBound; linearTermPFBound = NULL;};
  }
//                  delete integBound; delete gib; delete giv}
  BilinearTermBase* getBilinearInterfaceTerm(){return btermBound;}
  LinearTermBase<double>* getLinearInterfaceTerm()const{return ltermBound;}
  BilinearTermBase* getBilinearVirtualInterfaceTerm(){return btermVirtBound;}
  LinearTermBase<double>* getLinearVirtualInterfaceTerm(){return ltermVirtBound;}
  QuadratureBase* getInterfaceGaussIntegrationRule() const {return integBound;}

	ScalarTermBase<double>* getScalarTermPFBound() const {return scalarTermPFBound;};
	LinearTermBase<double>* getLinearTermPFBound() const {return linearTermPFBound;};

  virtual void computeIPVariable(AllIPState *aips,const unknownField *ufield,const IPStateBase::whichState ws, bool stiff)=0;
  virtual void computeIpv(AllIPState *aips,MInterfaceElement *ie, IntPt *GP,const IPStateBase::whichState ws,
                            partDomain* efMinus, partDomain *efPlus,materialLaw *mlawminus,
                            materialLaw *mlawplus,fullVector<double> &dispm,
                            fullVector<double> &dispp, fullVector<double>& dispExtra,
                            const bool virt, bool stiff,const bool checkfrac=true)=0;
  virtual void computeIpv(AllIPState *aips,MElement *e, IPStateBase::whichState ws,
                            materialLaw *mlaw,fullVector<double> &disp, bool stiff)=0;

  virtual bool IsMultipleFieldFormulation() const {return _multipleFieldFormulation;};
	virtual EnrichedUnknownLocation getEnrichedUnknownLocation() const{ return _enrichedUnknownLocation;};
	virtual MultiFieldMethod getMultipleFieldMethod() const {return _multiFieldMethod;};
  virtual const std::vector<int>& getFixedInterfaceDofComponent() const {return _fixInterfaceDofComp;};
  virtual bool allInterfaceIPsAreBrokenFromBeginning() const {return _brokenIPFromBeginning;};
  virtual void initiallyBreakAllInterfaceIP(AllIPState *aips) const;

  #if defined(HAVE_MPI)
  virtual void createIPMap();
  virtual const std::map<int,std::set<int> >& getMapIPInterfacePlus() const {return _mapIPInterfacePlus;};
  virtual const std::set<int>& getDomainIPInterfacePlus() const {return _domainIPInterfacePlus;};
  virtual const std::map<int,std::set<int> >& getMapIPInterfaceMinus() const {return _mapIPInterfaceMinus;};
  virtual const std::set<int>& getDomainIPInterfaceMinus() const {return _domainIPInterfaceMinus;};
  #endif // HAVE_MPI

  virtual void setGaussIntegrationRule()=0;
  virtual void createIPState(AllIPState::ipstateContainer& map, const bool *state) const;
  virtual int getDim() const=0;
  virtual bool IsInterfaceTerms() const{return true;}
  virtual FunctionSpaceBase* getFunctionSpace() =0;
  virtual const FunctionSpaceBase* getFunctionSpace() const=0;

  virtual FunctionSpaceBase* getFunctionSpaceMinus()=0;
  virtual const FunctionSpaceBase* getFunctionSpaceMinus() const=0;

  virtual FunctionSpaceBase* getFunctionSpacePlus()=0;
  virtual const FunctionSpaceBase* getFunctionSpacePlus() const=0;

  virtual FunctionSpaceBase* getInterfaceFunctionSpace() =0;
  virtual const FunctionSpaceBase* getInterfaceFunctionSpace() const =0;

  virtual void matrixByPerturbation(const int ibulk, const int iinter, const int ivirt,const double eps=1e-8)=0;
  virtual void setMaterialLaw(const std::map<int,materialLaw*> &maplaw)=0;
  virtual materialLaw* getMaterialLaw(){Msg::Error("The law to retrieve is not given on a dgdom"); return NULL;}
  virtual const materialLaw* getMaterialLaw() const{Msg::Error("The law to retrieve is not given on a dgdom"); return NULL;}
  virtual materialLaw* getMaterialLawMinus()=0;
  virtual const materialLaw* getMaterialLawMinus() const=0;
  virtual materialLaw* getMaterialLawPlus()=0;
  virtual const materialLaw* getMaterialLawPlus() const=0;
  virtual int getLawNum() const=0;
  virtual int getMinusLawNum() const=0;
  virtual int getPlusLawNum() const=0;
  virtual const partDomain* getMinusDomain() const=0;
  virtual const partDomain* getPlusDomain() const=0;
  virtual partDomain* getMinusDomain()=0;
  virtual partDomain* getPlusDomain()=0;
  virtual void createInterface(manageInterface &maninter)=0;
  virtual MElement* createVirtualInterface(IElement *ie) const=0;
  virtual MElement* createInterface(IElement *ie1, IElement *ie2) const=0;
  // Add for BC
  virtual FilterDof* createFilterDof(const int comp) const=0;
	virtual interfaceQuadratureBase* getInterfaceQuadrature() const = 0;

  // as to fill the array (of size Msg::GetCommSize()) with at index i the number of values to communicate with rank i
  // Normally it is an array of zero in input
  virtual void numberValuesToTransfertMPI(AllIPState *aips,int* sizeeachrank)const;
  // fill the vector that will be communicate to rank rank
  virtual void fillArrayIPvariableMPI(AllIPState *aips,const int rank,double* arrayMPI) const;
  virtual void setIPVariableFromArrayMPI(AllIPState *aips,const int rank,const int arraysize,const double* arrayMPI)const;
  virtual void initMicroMeshId();
#endif
};
#endif
