//
// C++ Interface: ipGursonCoalescence
//
// Description: Storing class for coalescence law
//
//
// Author:  <J. Leclerc>, (C) 2017
//
// Copyright: See COPYING file that comes with this distribution
//
//

#ifndef IPCOALESCENCE_H_
#define IPCOALESCENCE_H_
#include <stdlib.h>
#include <stdio.h>
#include "GmshConfig.h"
#include "ipVoidState.h"
class IPNonLocalPorosity;

class IPCoalescence{
  // Base class for coalescence IPV
  protected:
    double _acceleratedRate;              // Rate acceleration - equal 1 by default
    double _porosityAtCoalescenceOnset;   // Porosity at coalescence onset
    // necking coalescence
    bool _coalescenceOnsetFlag; // =true if coalescence was active at least once
    // shear coalescence
    bool _shearCoalescenceOnsetFlag;
    
    IPNonLocalPorosity *_ipvAtCoalescenceOnset; // void state at coalescence

  public:
    // Constructors & destructor
    IPCoalescence();
    IPCoalescence(const IPCoalescence &source);
    virtual IPCoalescence &operator=(const IPCoalescence &source);
    virtual ~IPCoalescence();

    // Standard functions
    virtual void restart();
    virtual IPCoalescence * clone() const=0;

    // Access functions
    virtual bool getCoalescenceOnsetFlag() const {return _coalescenceOnsetFlag;};
    virtual bool& getRefToCoalescenceOnsetFlag() {return _coalescenceOnsetFlag;};
    
    virtual bool getShearCoalescenceOnsetFlag() const {return _shearCoalescenceOnsetFlag;};
    virtual bool& getRefToShearCoalescenceOnsetFlag() {return _shearCoalescenceOnsetFlag;};

    virtual double getPorosityAtCoalescenceOnset() const {return _porosityAtCoalescenceOnset;};
    virtual double& getRefToPorosityAtCoalescenceOnset() {return _porosityAtCoalescenceOnset;};

    virtual double getAccelerateRate() const {return _acceleratedRate;};
    virtual double& getRefToAccelerateRate() {return _acceleratedRate;};
    
    virtual const IPNonLocalPorosity* getIPvAtCoalescenceOnset() const;
    virtual void setIPvAtCoalescenceOnset(const IPNonLocalPorosity& ipvVoid);

    virtual double getCrackOffsetOnCft() const { return 1.;};
    virtual double getYieldOffset() const {return 0.;};
    virtual double getLodeParameterOffset() const {return 1.;};
};



class IPNoCoalescence : public IPCoalescence{
  // IPV for NoCoalescenceLaw
  public:
    // Constructors & destructor
    IPNoCoalescence();
    IPNoCoalescence(const IPNoCoalescence &source);
    virtual IPNoCoalescence &operator=(const IPCoalescence &source);
    virtual ~IPNoCoalescence(){};
    // Standard functions
    virtual IPCoalescence * clone() const;
};



class IPFstarCoalescence : public IPCoalescence{
  // IPV for FstarCoalescenceLaw
  public:
    // Constructors & destructor
    IPFstarCoalescence();
    IPFstarCoalescence(const IPFstarCoalescence &source);
    virtual IPFstarCoalescence &operator=(const IPCoalescence &source);
    virtual ~IPFstarCoalescence(){};
    // Standard functions
    virtual IPCoalescence * clone() const;
};


class IPThomasonCoalescence : public IPCoalescence{
    // IPV for ThomasonCoalescenceLaw
  protected:
    // Offset and coupling management (updated at the end of a converged step)
    double crackOffsetOnCft;  // Offset applied on Cft to obtain fT = 0
    double yieldOffset;       // Yield surface value for the unused plastic mode
    double LodeOffset;

  public:
    IPThomasonCoalescence();
    IPThomasonCoalescence(const IPThomasonCoalescence& src);
    virtual IPThomasonCoalescence& operator = (const IPCoalescence& src);
    virtual ~IPThomasonCoalescence(){};

    virtual void restart();

    virtual IPCoalescence * clone() const {return new IPThomasonCoalescence(*this);};

    virtual double getCrackOffsetOnCft() const {return crackOffsetOnCft;};
    virtual double& getRefToCrackOffsetOnCft() {return crackOffsetOnCft;};

    virtual double getYieldOffset() const {return yieldOffset;};
    virtual double& getRefToYieldOffset() {return yieldOffset;};
    
    virtual double getLodeParameterOffset() const {return LodeOffset;};
    virtual double& getRefToLodeParameterOffset() {return LodeOffset;};
};


#endif // IPCOALESCENCE_H_
