//
// Description: Define ip data for cohesive layer
// the parameters of coheisve law including Gc, Kp, sigmac, deltac are
// set using function  initializeCohesiveData
// to compute other paramters--> using mlawCohesive
//
// Author:  <V.D. Nguyen>, (C) 2015
//
// Copyright: See COPYING file that comes with this distribution
//
//

#ifndef IPCOHESIVE_H_
#define IPCOHESIVE_H_

#include "ipvariable.h"
#include "STensor3.h"

class IPCohesive : public IPVariableMechanics{
  protected:
    double _beta; // mixmode parameter
    double _Kp; // initial penalty
    double _sigmac; // critical stress
    double _deltac; // critical effective opening
    double _delta0; // damage onset--> clearly sigmac = sigmac/Kp;
    double _deltamax; // maximal effective crack opening
    double _sigmamax; // traction at _deltamax
    double _cohesiveDamage; // cohesive Damage
    double _DcohesiveDamageDdelta;
 		bool _tension; // true if tension state false if compression
    double _fracEnergy;
		
		double _effectivePotential;
		double _reversibleEnergy;
		double _irreversibleEnergy;
		SVector3 _DIrreversibleEnergyDJump;

  public:
    IPCohesive():IPVariableMechanics(),_beta(0.),_Kp(0.),_sigmac(0.),_deltac(0.),
                _delta0(0.),_deltamax(0.),_sigmamax(0.),_cohesiveDamage(0.),
                _DcohesiveDamageDdelta(0.),_tension(true),_fracEnergy(0.),
								_reversibleEnergy(0.),_irreversibleEnergy(0.),_effectivePotential(0.),
								_DIrreversibleEnergyDJump(0.){}
    IPCohesive(const IPCohesive& src):IPVariableMechanics(src),_beta(src._beta),_Kp(src._Kp),_sigmac(src._sigmac),_deltac(src._deltac),
                _delta0(src._delta0),_deltamax(src._deltamax),_sigmamax(src._sigmamax),_cohesiveDamage(src._cohesiveDamage),
                _DcohesiveDamageDdelta(src._DcohesiveDamageDdelta),_tension(src._tension),
                _fracEnergy(src._fracEnergy),_reversibleEnergy(src._reversibleEnergy),
								_irreversibleEnergy(src._irreversibleEnergy),_effectivePotential(src._effectivePotential),
								_DIrreversibleEnergyDJump(src._DIrreversibleEnergyDJump){}

    virtual IPCohesive& operator=(const IPVariable& src){
      IPVariableMechanics::operator=(src);
      const IPCohesive* psrc = dynamic_cast<const IPCohesive*>(&src);
      if (psrc != NULL){
        _beta = psrc->_beta;
        _Kp = psrc->_Kp;
        _sigmac =  psrc->_sigmac;
        _deltac =  psrc->_deltac;
        _delta0 =  psrc->_delta0;
        _deltamax =  psrc->_deltamax;
        _sigmamax =  psrc->_sigmamax;
        _cohesiveDamage =  psrc->_cohesiveDamage;
        _DcohesiveDamageDdelta =  psrc->_DcohesiveDamageDdelta;
				_tension = psrc->_tension;
        _fracEnergy = psrc->_fracEnergy;
				_reversibleEnergy = psrc->_reversibleEnergy;
				_irreversibleEnergy = psrc->_irreversibleEnergy;
				_effectivePotential = psrc->_effectivePotential;
				_DIrreversibleEnergyDJump = psrc->_DIrreversibleEnergyDJump;
      }
      return *this;

    }
    virtual ~IPCohesive(){}
    
    virtual double defoEnergy() const {return _reversibleEnergy;};
    virtual double plasticEnergy() const {return 0.;};
    virtual double damageEnergy() const {return _irreversibleEnergy;};

    virtual double& getRefToMixedModeParameter() {return _beta;}
    virtual const double& getConstRefToMixedModeParameter() const{return _beta;}

    virtual double& getRefToInitialPenality() {return _Kp;};
    virtual const double& getConstRefToInitialPenality() const{return _Kp;};

    virtual double& getRefToCriticalEffectiveStress() {return _sigmac;};
    virtual const double& getConstRefToCriticalEffectiveStress() const {return _sigmac;};

    virtual double& getRefToCriticalEffectiveJump() {return _deltac;};
    virtual const double& getConstRefToCriticalEffectiveJump() const {return _deltac;};

    virtual double& getRefToEffectiveJumpDamageOnset() {return _delta0;};
    virtual const double& getConstRefToEffectiveJumpDamageOnset() const {return _delta0;};

    virtual double& getRefToMaximalEffectiveJump() {return _deltamax;};
    virtual const double& getConstRefToMaximalEffectiveJump() const {return _deltamax;};

    virtual double& getRefToMaximalEffectiveStress() {return _sigmamax;};
    virtual const double& getConstRefToMaximalEffectiveStress() const {return _sigmamax;};

    virtual double& getRefToCohesiveDamage() {return _cohesiveDamage;};
    virtual const double& getConstRefToCohesiveDamage()const {return _cohesiveDamage;};

    virtual double& getRefToDCohesiveDamageDEffectiveJump() {return _DcohesiveDamageDdelta;};
    virtual const double& getConstRefToDCohesiveDamageDEffectiveJump() const {return _DcohesiveDamageDdelta;};

		virtual bool ifTension() const {return _tension;};
		virtual bool& getRefToTensionFlag() {return _tension;};

    virtual double& getRefToFractureEnergy() {return _fracEnergy;};
    virtual const double& getConstRefToFractureEnergy() const {return _fracEnergy;};
		
		virtual double& getRefToReversibleEnergy() {return _reversibleEnergy;};
		virtual const double& getConstRefToReversibleEnergy() const {return _reversibleEnergy;};
		
		virtual double& getRefToIrreversibleEnergy() {return _irreversibleEnergy;};
		virtual double irreversibleEnergy() const {return _irreversibleEnergy;};
		
		virtual double& getRefToEffectivePotential() {return _effectivePotential;};
		virtual const double& getConstRefToEffectivePotential() const {return _effectivePotential;};
		
		virtual SVector3& getRefToDIrreversibleEnergyDJump() {return _DIrreversibleEnergyDJump;};
		virtual const SVector3& getConstRefToDIrreversibleEnergyDJump() const {return _DIrreversibleEnergyDJump;};
		 
    virtual void initializeCohesiveData(const double beta, // mixmode parameter
                                        const double sigmac, // critical stress
                                        const double deltac, // critical effective opening
                                        const double delta0, // damage onset--> clearly delta0 = sigmac/Kp;
                                        const bool iftension)
    {
       _beta = beta;
       if(delta0<1.e-18) Msg::Error("IPCohesive::initializeCohesiveData with delta0 = 0");

       _Kp = sigmac/delta0;
       _sigmac = sigmac;
       _deltac = deltac;
       _delta0 = delta0;
       _tension = iftension;

       _deltamax = delta0;
       _sigmamax = sigmac;
       _cohesiveDamage = 0.;
       _DcohesiveDamageDdelta= 0.;
       _fracEnergy = 0.;
			 _reversibleEnergy = 0.;
			 _irreversibleEnergy = 0.;
			 _effectivePotential = 0.;
			 _DIrreversibleEnergyDJump *= 0.;

    }

    virtual void restart(){
      IPVariableMechanics::restart();
      restartManager::restart(_beta); // mixmode parameter
      restartManager::restart(_Kp); // initial penalty
      restartManager::restart(_sigmac); // critical stress
      restartManager::restart(_deltac); // critical effective opening
      restartManager::restart(_delta0); // damage onset--> clearly sigmac = sigmac/Kp;
      restartManager::restart(_deltamax); // maximal effective crack opening
      restartManager::restart(_sigmamax); // traction at _deltamax
      restartManager::restart(_cohesiveDamage); // cohesive Damage
      restartManager::restart(_DcohesiveDamageDdelta);
      restartManager::restart(_tension);
      restartManager::restart(_fracEnergy);
			restartManager::restart(_reversibleEnergy);
			restartManager::restart(_irreversibleEnergy);
			restartManager::restart(_effectivePotential);
			restartManager::restart(_DIrreversibleEnergyDJump);
    };
    virtual IPVariable* clone() const {return new IPCohesive(*this);};
};

class IPBiLinearCohesive : public IPCohesive{
  public:
    IPBiLinearCohesive():IPCohesive(){}
    IPBiLinearCohesive(const IPBiLinearCohesive& src): IPCohesive(src){}
    virtual IPBiLinearCohesive& operator=(const IPVariable& src){
      IPCohesive::operator=(src);
      return *this;
    }
    virtual ~IPBiLinearCohesive(){}
    virtual IPVariable* clone() const {return new IPBiLinearCohesive(*this);};
};

class IPTransverseIsotropicCohesive : public IPBiLinearCohesive{

  public:
    IPTransverseIsotropicCohesive():IPBiLinearCohesive(){}
    IPTransverseIsotropicCohesive(const IPTransverseIsotropicCohesive& src): IPBiLinearCohesive(src){}
    virtual IPTransverseIsotropicCohesive& operator=(const IPVariable& src){
      IPBiLinearCohesive::operator=(src);
      return *this;
    }
    virtual ~IPTransverseIsotropicCohesive(){}
    virtual IPVariable* clone() const {return new IPTransverseIsotropicCohesive(*this);};
};

class IPTransverseIsoCurvatureCohesive : public IPTransverseIsotropicCohesive{
  protected:
    SVector3 _A;
  public:
    IPTransverseIsoCurvatureCohesive(const SVector3 &GaussP, const SVector3 &Centroid, const SVector3 &PointStart1, const SVector3 &PointStart2, const double init_Angle):IPTransverseIsotropicCohesive()
    {
       SVector3 init_A = PointStart1 - Centroid;
       SVector3 Vec1 = PointStart2 - Centroid;
       SVector3 N_local = crossprod(init_A,Vec1);
       SVector3 Vec2 = GaussP - Centroid;
       SVector3 Proj_A = crossprod(N_local,Vec2);
       Proj_A.normalize(); 
       if(fabs(fabs(init_Angle)-90.)<1.e-12)
       {
          _A = N_local;
       }
       else
       {
          N_local.normalize();
          _A = Proj_A + tan(init_Angle*3.14159/180.)*N_local;
       }
       _A.normalize();
    }
    IPTransverseIsoCurvatureCohesive(const IPTransverseIsoCurvatureCohesive& src): IPTransverseIsotropicCohesive(src)
    {
       _A = src._A;
    }
    virtual IPTransverseIsoCurvatureCohesive& operator=(const IPVariable& src){
      IPTransverseIsotropicCohesive::operator=(src);
      const IPTransverseIsoCurvatureCohesive* psrc = dynamic_cast<const IPTransverseIsoCurvatureCohesive*>(&src);
      if (psrc != NULL){
        _A=psrc->_A;
      }

      return *this;
    }
    virtual ~IPTransverseIsoCurvatureCohesive(){}
    virtual void restart()
    {
      IPTransverseIsotropicCohesive::restart();
      restartManager::restart(_A); // direction
    }
    virtual IPVariable* clone() const {return new IPTransverseIsoCurvatureCohesive(*this);};
    virtual void getTransverseDirection(SVector3 &A) const { A =_A ;};

  private:
    IPTransverseIsoCurvatureCohesive():IPTransverseIsotropicCohesive()
    {
       Msg::Error("Cannot construct default IPTransverseIsoCurvatureCohesive");
    }
};

#endif // IPCOHESIVE_H_
