//
// C++ Interface: ipvariable
//
// Description: IP ViscoElastoPlastic
//
// Author:  V.D. Nguyen, (C) 2014
//
// Copyright: See COPYING file that comes with this distribution
//
//

#ifndef IPHYPERELASTIC_H_
#define IPHYPERELASTIC_H_

#include "j2IsotropicHardening.h"
#include "kinematicHardening.h"
#include "STensor43.h"

class IPHyperViscoElastic : public IPVariableMechanics{
  public:
    //viscoelastic history
    int _N; // number spring-Dashpot of element
    std::vector<STensor3> _A; // N elements
    std::vector<double> _B; // N elements
    double _elasticEnergy; // elastic energy stored
    STensor3 _Ee; // elastic strain
    STensor3 _kirchhoff; // corotational Kirchhoff stress
    
    double _irreversibleEnergy;
    STensor3 _DirreversibleEnergyDF;
    
  public:
    IPHyperViscoElastic(const int N);
    IPHyperViscoElastic(const IPHyperViscoElastic& src);
    virtual IPHyperViscoElastic& operator=(const IPVariable& src);
    virtual ~IPHyperViscoElastic(){};
    
    virtual IPVariable* clone() const{return new IPHyperViscoElastic(*this);};
    
    virtual STensor3& getRefToElasticStrain() {return _Ee;};
    virtual const STensor3& getConstRefToElasticStrain() const {return _Ee;};
    
    virtual STensor3& getRefToCorotationalKirchhoffStress() {return _kirchhoff;};
    virtual const STensor3& getConstRefToCorotationalKirchhoffStress() const {return _kirchhoff;};
    
    virtual double defoEnergy() const{return _elasticEnergy;}
    virtual double& getRefToElasticEnergy() {return _elasticEnergy;};
    virtual double plasticEnergy() const{return 0.;}
    
    virtual double irreversibleEnergy() const {return _irreversibleEnergy;};
    virtual double & getRefToIrreversibleEnergy() {return _irreversibleEnergy;};
    
    virtual STensor3& getRefToDIrreversibleEnergyDF() {return _DirreversibleEnergyDF;};
    virtual const STensor3& getConstRefToDIrreversibleEnergyDF() const{return _DirreversibleEnergyDF;};
    
    virtual void restart();
};

class IPHyperViscoElastoPlastic : public IPHyperViscoElastic{
  public: // defined with public to free access
    IPJ2IsotropicHardening* _ipCompression; // ip in compression
    IPJ2IsotropicHardening* _ipTraction; // ip in traction
    IPJ2IsotropicHardening* _ipShear; // ip in shear
    IPKinematicHardening* _ipKinematic;

   
    STensor3 _DgammaDF; // DequivalentPlasticStrainDF
    STensor3 _backsig; // backstress
    STensor3 _Fe; // elastic strain

    double _epspbarre; // equivalent plastic strain
    STensor3 _Fp;    // plastic part of the deformation gradient
    double _nup; // plastic poisson ration

    //plastic eq deformation in compression, traction and shear
    double _epspCompression;
    double _epspTraction;
    double _epspShear;

    double _DgammaDt; // plastic deformation rate

    // failure
		bool _inPostFailureStage;
    double _r;  // failure onset
    double _gF; //
    STensor3 _dgFdF;
    
    double _plasticEnergy;
    bool _dissipationBlocked;
    bool _dissipationActive;

  public:
    IPHyperViscoElastoPlastic(const J2IsotropicHardening* comp,
                    const J2IsotropicHardening* trac,const J2IsotropicHardening* shear,
                    const kinematicHardening* kin, const int N);
    IPHyperViscoElastoPlastic(const IPHyperViscoElastoPlastic& src);
    virtual IPHyperViscoElastoPlastic& operator =(const IPVariable &source);
    virtual ~IPHyperViscoElastoPlastic();

    virtual IPVariable* clone() const{return new IPHyperViscoElastoPlastic(*this);};
	
    virtual bool& getRefToDissipationActive(){return _dissipationActive;};
    virtual bool dissipationIsActive() const {return _dissipationActive;};

    virtual void blockDissipation(const bool fl){_dissipationBlocked = fl;};
    virtual bool dissipationIsBlocked() const{return _dissipationBlocked;};

    virtual double& getRefToPlasticEnergy() {return _plasticEnergy;};
    virtual double plasticEnergy() const{return _plasticEnergy;}

    virtual void restart();

		virtual bool inPostFailureStage() const {return _inPostFailureStage;};
		virtual bool& getRefToInPostFailureStage() {return _inPostFailureStage;};


    virtual double getFailureOnset() const {return _r;};
    virtual double& getRefToFailureOnset() {return _r;};

    virtual double getFailurePlasticity() const {return _gF;};
    virtual double& getRefToFailurePlasticity() {return _gF;};

    virtual const STensor3& getConstRefToDFailurePlasticityDF() const {return _dgFdF;};
    virtual STensor3& getRefToDFailurePlasticityDF() {return _dgFdF;};

    virtual const STensor3 & getConstRefToFp() const;
    virtual STensor3& getRefToFp();

    virtual const STensor3& getConstRefToBackStress() const;
    virtual STensor3& getRefToBackStress();

    virtual const STensor3& getConstRefToFe() const;
    virtual STensor3& getRefToFe();

    virtual const double& getConstRefToCompressionPlasticStrain() const;
    virtual double & getRefToCompressionPlasticStrain();

    virtual const double& getConstRefToTractionPlasticStrain() const;
    virtual double & getRefToTractionPlasticStrain();

    virtual const double& getConstRefToShearPlasticStrain() const;
    virtual double & getRefToShearPlasticStrain();

    virtual const double& getConstRefToEqPlasticStrain() const;
    virtual double& getRefToEqPlasticStrain();

    virtual const double getConstRefToPlasticPoissonRatio() const;
    virtual double& getRefToPlasticPoissonRatio();
    
    virtual const double& getConstRefToPlasticDeformationRate() const;
    virtual double& getRefToPlasticDeformationRate();

    virtual const IPJ2IsotropicHardening& getConstRefToIPCompressionHardening() const;
    virtual IPJ2IsotropicHardening& getRefToIPCompressionHardening();

    virtual const IPJ2IsotropicHardening& getConstRefToIPTractionHardening() const;
    virtual IPJ2IsotropicHardening& getRefToIPTractionHardening();

    virtual const IPJ2IsotropicHardening& getConstRefToIPShearHardening() const;
    virtual IPJ2IsotropicHardening& getRefToIPShearHardening();

    virtual const IPKinematicHardening& getConstRefToKinematicHardening() const;
    virtual IPKinematicHardening& getRefToKinematicHardening();
    
    virtual bool IPKinematicHardeningIsSaturated() const{
      if (_ipKinematic != NULL) return _ipKinematic->isSaturated();
      else return false;
    }
    virtual bool IPJ2IsotropicHardeningIsSaturated() const{
      if (_ipCompression!=NULL) return _ipCompression->isSaturated();
      else if (_ipTraction != NULL) return _ipTraction->isSaturated();
      else if (_ipShear != NULL) return _ipShear->isSaturated();
      else return false;
    }
    virtual bool isSaturated() const{
      if (IPKinematicHardeningIsSaturated()) return true;
      if (IPJ2IsotropicHardeningIsSaturated()) return true;
      return false;
    }
    
    virtual void saturate(const bool fl){
      if (_ipKinematic != NULL){
        _ipKinematic->getRefToSaturationState() = fl;
      }
      
      if (_ipCompression != NULL){
        _ipCompression->getRefToSaturationState() = fl;
      }
      
      if (_ipTraction != NULL){
        _ipTraction->getRefToSaturationState() = fl;
      }
      
      if (_ipShear != NULL){
        _ipShear->getRefToSaturationState() = fl;
      }
    
    };
};

#endif // IPHYPERELASTIC_H_
