//
//
// Description: Provide criterion class for nucleation 
//
//
// Author:  <J. Leclerc>, (C) 2019
//
// Copyright: See COPYING file that comes with this distribution
//
//
#include "NucleationCriterionLaw.h"
#include "mlawNonLocalPorous.h"

// NucleationCriterionLaw
// -------------------------------------------
NucleationCriterionLaw::NucleationCriterionLaw(): _num(0), _initialized(false)
{
  Msg::Error("NucleationCriterionLaw:: default constructor should be never called");
};

NucleationCriterionLaw::NucleationCriterionLaw(const int num, const bool init): _num(num), _initialized(init)
{

};

NucleationCriterionLaw::NucleationCriterionLaw(const NucleationCriterionLaw &source)
{
  _num = source._num;
  _initialized = source._initialized;
};

NucleationCriterionLaw& NucleationCriterionLaw::operator=(const NucleationCriterionLaw &source)
{
  _num = source._num;
  _initialized = source._initialized;
  return *this;
};




// TrivialNucleationCriterionLaw
// -------------------------------------------
TrivialNucleationCriterionLaw::TrivialNucleationCriterionLaw(const int num, const bool init):
    NucleationCriterionLaw(num,init)
{

}

TrivialNucleationCriterionLaw::TrivialNucleationCriterionLaw(const TrivialNucleationCriterionLaw &source) :
    NucleationCriterionLaw(source)
{

}

TrivialNucleationCriterionLaw& TrivialNucleationCriterionLaw::operator=(const NucleationCriterionLaw &source)
{
  NucleationCriterionLaw::operator=(source);
  const TrivialNucleationCriterionLaw* src = dynamic_cast<const TrivialNucleationCriterionLaw*>(&source);
  if(src != NULL)
  {

  }
  return *this;
}

NucleationCriterionLaw* TrivialNucleationCriterionLaw::clone() const
{
  return new TrivialNucleationCriterionLaw(*this);
}


void TrivialNucleationCriterionLaw::createIPVariable(IPNucleationCriterion* &ipv) const
{
  if(ipv != NULL) delete ipv;
  ipv = new IPTrivialNucleationCriterion();
}


void TrivialNucleationCriterionLaw::checkNucleationActivation(const mlawNonLocalPorosity* mlaw, IPNonLocalPorosity* q1Porous, const IPNonLocalPorosity* q0Porous) const
{
  // Nothing has to be done:
  // IPNucleationCriterion->getNucleationActivatedFlag() will always be true
};






// BereminNucleationCriterionLaw
// -------------------------------------------
BereminNucleationCriterionLaw::BereminNucleationCriterionLaw(const int num, 
                  const double sigma_cx, const double sigma_cy, const double sigma_cz, 
                  const double kx, const double ky, const double kz, 
                  const bool init):
    NucleationCriterionLaw(num,init),
    _criticalStress(0.),
    _stressConcentrationTensor(0.)
{
  // Set the stress concentration tensor to 0 and set the diagonal values.
  _criticalStress(0,0) = sigma_cx;
  _criticalStress(1,1) = sigma_cy;
  _criticalStress(2,2) = sigma_cz;
  
  _stressConcentrationTensor(0,0) = kx;
  _stressConcentrationTensor(1,1) = ky;
  _stressConcentrationTensor(2,2) = kz;
  
};


void BereminNucleationCriterionLaw::rotateMaterialAxes(const double alpha, const double beta, const double gamma)
{
  // Rotate the tensor of stress concentration factor following the rotation defined by the euler angles in arguments.
  // Angles are in degrees in input !
  
  // The rotation from the initial frame (need to be computed) to the body/material frame (given in input by constructor) 
  // is described as follows:
    // 1. a rotation along z0 of amplitude alpha = precession angle
    // 2. a rotation along x1 of amplitude beta = nutation angle
    // 3. a rotation along z2 of amplitude gamma = spin angle
  // R defines the rotation from the initial frame (need to be computed) to the body frame (given in input by constructor)
  // i.e.  X_body = R * X_initial =>    K_body = R * K_initital * R^T
  // and   X_initial = R^T * X_body =>  K_initital = R^T * K_body * R
  
  // Compute the rotation tensor R(alpha, beta, gamma)
    // Get values in rad
  double c1,c2,c3,s1,s2,s3;
  double s1c2, c1c2;
  double pi(3.14159265359);
  double fpi = pi/180.;

    // Compute components
  c1 = cos(alpha*fpi);
  s1 = sin(alpha*fpi);

  c2 = cos(beta*fpi);
  s2 = sin(beta*fpi);

  c3 = cos(gamma*fpi);
  s3 = sin(gamma*fpi);

  s1c2 = s1*c2;
  c1c2 = c1*c2;
  
  STensor3 R;
  R(0,0) = c3*c1 - s1c2*s3;
  R(0,1) = c3*s1 + c1c2*s3;
  R(0,2) = s2*s3;

  R(1,0) = -s3*c1 - s1c2*c3;
  R(1,1) = -s3*s1 + c1c2*c3;
  R(1,2) = s2*c3;

  R(2,0) = s1*s2;
  R(2,1) = -c1*s2;
  R(2,2) = c2;
  
  // Apply the rotation
  STensor3 criticalStressInOldAxes = _criticalStress; // This tensor is created for cache
  STensor3 stressFactorInOldAxes = _stressConcentrationTensor; // This tensor is created for cache
  for(int i=0;i<3;i++){
    for(int j=0;j<3;j++){
      _stressConcentrationTensor(i,j) = 0.;
      _criticalStress(i,j) = 0.;
      for(int k=0;k<3;k++){
        for(int l=0;l<3;l++){
          // K_(initial) = R^T * K_(material) * R
          _stressConcentrationTensor(i,j) += R(k,i)*stressFactorInOldAxes(k,l)*R(l,j);
          _criticalStress(i,j) += R(k,i)*criticalStressInOldAxes(k,l)*R(l,j);
        }
      }
    }
  }
};



BereminNucleationCriterionLaw::BereminNucleationCriterionLaw(const BereminNucleationCriterionLaw &source) :
    NucleationCriterionLaw(source),
    _criticalStress(source._criticalStress),
    _stressConcentrationTensor(source._stressConcentrationTensor)
{

}

BereminNucleationCriterionLaw& BereminNucleationCriterionLaw::operator=(const NucleationCriterionLaw &source)
{
  NucleationCriterionLaw::operator=(source);
  const BereminNucleationCriterionLaw* src = dynamic_cast<const BereminNucleationCriterionLaw*>(&source);
  if(src != NULL)
  {
    _criticalStress=src->_criticalStress;
    _stressConcentrationTensor=src->_stressConcentrationTensor;
  }
  return *this;
}

NucleationCriterionLaw* BereminNucleationCriterionLaw::clone() const
{
  return new BereminNucleationCriterionLaw(*this);
}


void BereminNucleationCriterionLaw::createIPVariable(IPNucleationCriterion* &ipv) const
{
  if(ipv != NULL) delete ipv;
  ipv = new IPBereminNucleationCriterion();
}


void BereminNucleationCriterionLaw::checkNucleationActivation(const mlawNonLocalPorosity* mlaw, IPNonLocalPorosity* q1Porous, const IPNonLocalPorosity* q0Porous) const
{
  // Update the activity of the nucleation
  // Get first the right ipv form
  const IPNucleationCriterion* q0Criterion = &(q0Porous->getConstRefToIPNucleationCriterion());
  IPNucleationCriterion* q1Criterion = &(q1Porous->getRefToIPNucleationCriterion());
  
  // Check if the criterion has been satisfied if it is not already the case
  if(q0Criterion->getNucleationActivatedFlag()){
    // If the criterion has been satified, keep its value as true
    q1Criterion->setNucleationActivatedFlag(true);
  }
  else{
    // If the criterion has not been satified yet,
    // Compute the maximal effective stress transmitted to the particles, which
    // corresponds to the eigenvalues of the criterion rewritten under a "tensorial" form:
    // criterion_tensor = stress + R*(k*(sigma_eq - sigma_y0) - sigma_nc)*R^T (in the current configuration)
    static STensor3 criterion_tensor;
    
    // Get 2nd and 3rd term := k*(sigma_eq - sigma_y0) - sigma_nc
    static STensor3 K_minus_sigma_nc;
    K_minus_sigma_nc = _stressConcentrationTensor;
    K_minus_sigma_nc *= (q1Porous->getCurrentViscoplasticYieldStress() - q1Porous->getConstRefToIPJ2IsotropicHardening().getR0());
    K_minus_sigma_nc -= _criticalStress;
    
    // Apply the rotation R from the initial to the current configuration:
    // i.e. k(current) = R * k(initial) * R^T
      // Get the rotation R from deformation gradient
    static STensor3 R, U; // cache for rotation and RU decomposition
    static STensor3 cache_R; // cache for rotation computation
    STensorOperation::RUDecomposition(q1Porous->getConstRefToDeformationGradient(), U, R);
      // apply the rotation k(current) = R * k(initial) * R^T
    STensorOperation::multSTensor3(R,K_minus_sigma_nc,cache_R);
    STensorOperation::multSTensor3SecondTranspose(cache_R,R,K_minus_sigma_nc);


    // Get stress tensor in current frame from corrotationnal frame
      // Compute Kirchhoff Stress from CorKir stress following: Stress = Fe^-T * StressCor * Fe^T
    static STensor3 invFp, Fe, invFe, invFe_x_CorStress, Stress; // cache
        // - get Fe = F * Fp^-1 and Fe^-1
    STensorOperation::inverseSTensor3(q1Porous->getConstRefToFp(), invFp);
    STensorOperation::multSTensor3(q1Porous->getConstRefToDeformationGradient(), invFp, Fe);
    STensorOperation::inverseSTensor3(Fe, invFe);
        // - apply the rotation Stress = Fe^-T * StressCor * Fe^T
    STensorOperation::multSTensor3FirstTranspose(invFe, q1Porous->getConstRefToCorotationalKirchhoffStress(), invFe_x_CorStress);
    STensorOperation::multSTensor3SecondTranspose(invFe_x_CorStress, Fe, Stress);
        // - transform, following the stress formulation, Kirchhoff stress to Cauchy stress = Kirchhoff / Jacobian
    if (mlaw->getStressFormulation() == mlawNonLocalPorosity::CORO_CAUCHY){
      double detF = STensorOperation::determinantSTensor3(q1Porous->getConstRefToDeformationGradient());
      Stress *= 1./detF;
    }


    // Get criterion_tensor
    criterion_tensor = Stress;
    criterion_tensor += K_minus_sigma_nc;
    
    // Get the max eigen values for the direction that maximises the criterion
    static fullMatrix<double> eigVec(3,3);
    static fullVector<double> eigVal(3);
    criterion_tensor.eig(eigVec,eigVal,true);
    
    double stress_max = std::max(eigVal(0),std::max(eigVal(1),eigVal(2)));
    
    // If criterion is satisfied, nucleation is activated, otherwise, it is false
    if (stress_max > 0.){
      q1Criterion->setNucleationActivatedFlag(true);
      Msg::Info("Nucleation has been activated on an IP in the direction [%e, %e ,%e]",eigVec(0,0),eigVec(0,1),eigVec(0,2));
    }
    else{
      q1Criterion->setNucleationActivatedFlag(false);
    }
  }
  
};




