//
// C++ Interface: Nucleation criterion
//
// Description: Provide criterion class for nucleation 
//
//
// Author:  <J. Leclerc>, (C) 2019
//
// Copyright: See COPYING file that comes with this distribution
//
//
//
#ifndef _NUCLEATIONCRITERIONLAW_H_
#define _NUCLEATIONCRITERIONLAW_H_
#ifndef SWIG
#include "ipstate.h"
#include "ipNucleationCriterion.h"
class IPNonLocalPorosity;
class mlawNonLocalPorosity;
#endif

class NucleationCriterionLaw{
  // Class for activation of nucleation. This law is used a the end of a time step.
  // It evaluates if nucleation should be activated at the next time step.
  // This information is stored in the corresponding IPv.
  
  // Basic variable
protected:
    // Name of the criterion
  enum nucleationCriterionName{trivialNucleationCriterion, bereminIsotropicNucleationCriterion};
  
protected:
  int _num; // number of law
  bool _initialized; // to initialise the law
  
#ifndef SWIG
  // Constructor-destructor
private:
  NucleationCriterionLaw();
public:
  NucleationCriterionLaw(const int num, const bool init=true);
  virtual ~NucleationCriterionLaw(){};
  NucleationCriterionLaw(const NucleationCriterionLaw &source);
  NucleationCriterionLaw& operator=(const NucleationCriterionLaw &source);
  virtual NucleationCriterionLaw * clone() const=0;
  
  // Function of materialLaw
  virtual int getNum() const{return _num;};
  virtual const bool isInitialized() {return _initialized;};
  virtual NucleationCriterionLaw::nucleationCriterionName getType() const=0;
  virtual void createIPVariable(IPNucleationCriterion* &ipv) const=0;
  
  // Specific functions
    // Determine at the end of a time step if the nucleation is activated or not
  virtual void checkNucleationActivation(const mlawNonLocalPorosity* mlaw, IPNonLocalPorosity* q1Porous, const IPNonLocalPorosity* q0Porous) const = 0;

#endif
};



class TrivialNucleationCriterionLaw : public NucleationCriterionLaw
{
  // Trivial criterion and used by default. 
  // With this class, nucleation is always activiated
protected:
  // Nothing is needed

public:
  // Constructor-destructor
  TrivialNucleationCriterionLaw(const int num, const bool init=true);
#ifndef SWIG
  virtual ~TrivialNucleationCriterionLaw(){};
  TrivialNucleationCriterionLaw(const TrivialNucleationCriterionLaw &source);
  TrivialNucleationCriterionLaw& operator=(const NucleationCriterionLaw &source);
  virtual NucleationCriterionLaw * clone() const;
  
  // Function of materialLaw
  virtual NucleationCriterionLaw::nucleationCriterionName getType() const{NucleationCriterionLaw::trivialNucleationCriterion;};
  virtual void createIPVariable(IPNucleationCriterion* &ipv) const;
  
  // Specific functions
  virtual void checkNucleationActivation(const mlawNonLocalPorosity* mlaw, IPNonLocalPorosity* q1Porous, const IPNonLocalPorosity* q0Porous) const;
#endif
};


class BereminNucleationCriterionLaw : public NucleationCriterionLaw
{
  // Beremin criterion based on Beremin1980 
  // Nucleation is allowed if n.(sigma + k(sigma_eq - sigma_y0)).n > n.sigma_c.n
  // (This criterion is identical either in Kirchhoff and in Sigma stress and is compatible with both formulations)
  
protected:
  // Stress threshold value at which nucleation occurs
  STensor3 _criticalStress;
  
  // Stress concentration factor represented under a tensor form
  // in initial material axis (an arbitrary rotation can be applied using the options)
  STensor3 _stressConcentrationTensor;
    
public :
  // Constructor-destructor
  BereminNucleationCriterionLaw(const int num, const double sigma_cx, const double sigma_cy, const double sigma_cz, const double kx, const double ky, const double kz, const bool init=true);
  // Applied a rotation on the material tensor using euler angles (in degrees)
  virtual void rotateMaterialAxes(const double alpha, const double beta, const double gamma);
#ifndef SWIG
  virtual ~BereminNucleationCriterionLaw(){};
  BereminNucleationCriterionLaw(const BereminNucleationCriterionLaw &source);
  BereminNucleationCriterionLaw& operator=(const NucleationCriterionLaw &source);
  virtual NucleationCriterionLaw* clone() const;
  
  // Standard functions
  virtual NucleationCriterionLaw::nucleationCriterionName getType() const{NucleationCriterionLaw::bereminIsotropicNucleationCriterion;};
  virtual void createIPVariable(IPNucleationCriterion* &ipv) const;
  
  
  // Specific functions
  virtual void checkNucleationActivation(const mlawNonLocalPorosity* mlaw, IPNonLocalPorosity* q1Porous, const IPNonLocalPorosity* q0Porous) const;
#endif
};


#endif //_NUCLEATIONCRITERIONLAW_H_
