//
//
//
// Description: Define gurson damake nucleation
//
//
// Author:  <L. Noels>, (C) 2013
//
// Copyright: See COPYING file that comes with this distribution
//
//
#include "NucleationLaw.h"

// NucleationLaw
// -------------------------------------------
NucleationLaw::NucleationLaw(const int num, const bool init): 
    _num(num), _initialized(init)
{

};

NucleationLaw::NucleationLaw(const NucleationLaw &source)
{
  _num = source._num;
  _initialized = source._initialized;
};

NucleationLaw& NucleationLaw::operator=(const NucleationLaw &source)
{
  _num = source._num;
  _initialized = source._initialized;
  return *this;
};



TrivialNucleationLaw::TrivialNucleationLaw(const int num): NucleationLaw(num){};
TrivialNucleationLaw& TrivialNucleationLaw::operator=(const NucleationLaw &source)
{
  NucleationLaw::operator =(source);
  return *this;
}

void TrivialNucleationLaw::createIPVariable(IPNucleation* &ipv) const
{
  if(ipv != NULL) delete ipv;
  ipv = new IPTrivialNucleationRate();
};


void TrivialNucleationLaw::nucleate(const double p, const double p0, const double f0, const  IPNucleation &ipvprev, IPNucleation &ipv) const{
  ipv.getRefToNucleatedPorosity() =  ipvprev.getNucleatedPorosity();
  ipv.getRefToNucleatedPorosityIncrement() = 0.;
  ipv.getRefToDNucleatedPorosityDMatrixPlasticDeformation() = 0.;
};


// LinearNucleationLaw
// -------------------------------------------
LinearNucleationLaw::LinearNucleationLaw(const int num, const double A0, const double fmin,
                                         const double fmax, const bool init) :
    NucleationLaw(num,init), _A0(A0), _fmin(fmin), _fmax(fmax)
{

};

LinearNucleationLaw::LinearNucleationLaw(const LinearNucleationLaw &source) :
    NucleationLaw(source)
{
  _A0 = source._A0;
  _fmin = source._fmin;
  _fmax = source._fmax;
};

LinearNucleationLaw& LinearNucleationLaw::operator=(const NucleationLaw &source)
{
  NucleationLaw::operator=(source);
  const LinearNucleationLaw* src = dynamic_cast<const LinearNucleationLaw*>(&source);
  if(src != NULL)
  {
    _A0 = src->_A0;
    _fmin = src->_fmin;
    _fmax = src->_fmax;
  }
  return *this;
};

NucleationLaw* LinearNucleationLaw::clone() const
{
  return new LinearNucleationLaw(*this);
};

void LinearNucleationLaw::createIPVariable(IPNucleation* &ipv) const
{
  if(ipv != NULL) delete ipv;
  ipv = new IPLinearNucleationRate();
};


void LinearNucleationLaw::nucleate(const double p, const double p0, const double f0, const  IPNucleation &ipvprev, IPNucleation &ipv) const{
  // Compute the nucleated porosity
    // If the plastic strain is not increasing, no porosity is nucleated
  if (p<p0){
    ipv.getRefToNucleatedPorosity() =  ipvprev.getNucleatedPorosity();
    ipv.getRefToNucleatedPorosityIncrement() = 0.;
    ipv.getRefToDNucleatedPorosityDMatrixPlasticDeformation() = 0.;
  }
  else{
    // If plastic strain is increasing, porosity is nucleated if the porosoty is in the allowed range
    if (_fmin<=f0 and f0<=_fmax)
    {
      // Porosity is outside the nucleation range : nothing is nucleated
      ipv.getRefToNucleatedPorosity() = ipvprev.getNucleatedPorosity()+ _A0*(p-p0);
      ipv.getRefToNucleatedPorosityIncrement() = _A0*(p-p0);
      ipv.getRefToDNucleatedPorosityDMatrixPlasticDeformation() = _A0;
    }
    else{
      // Porosity is inside the nucleation range
      ipv.getRefToNucleatedPorosity() =  ipvprev.getNucleatedPorosity();
      ipv.getRefToNucleatedPorosityIncrement() = 0.;
      ipv.getRefToDNucleatedPorosityDMatrixPlasticDeformation() = 0.;
    }
  }
};




// ExponentialNucleationLaw
// -------------------------------------------
ExponentialNucleationLaw::ExponentialNucleationLaw(const int num,  double fn, double sn, double epsilonn,
                                                   const bool init) :
      NucleationLaw(num,init), _fn(fn), _sn(sn), _epsilonn(epsilonn)
{

};

ExponentialNucleationLaw::ExponentialNucleationLaw(const ExponentialNucleationLaw &source) :
    NucleationLaw(source)
{
  _fn = source._fn;
  _sn = source._sn;
  _epsilonn = source._epsilonn;
};

ExponentialNucleationLaw& ExponentialNucleationLaw::operator=(const NucleationLaw &source)
{
  NucleationLaw::operator=(source);
  const ExponentialNucleationLaw* src = dynamic_cast<const ExponentialNucleationLaw*>(&source);
  if(src != NULL)
  {
    _fn = src->_fn;
    _sn = src->_sn;
    _epsilonn = src->_epsilonn;
  }
  return *this;
};

NucleationLaw* ExponentialNucleationLaw::clone() const
{
  return new ExponentialNucleationLaw(*this);
};



void ExponentialNucleationLaw::createIPVariable(IPNucleation* &ipv) const
{
  if(ipv != NULL) delete ipv;
  ipv = new IPExponentialNucleationRate();
};



void ExponentialNucleationLaw::nucleate(const double p, const double p0, const double f0, const  IPNucleation &ipvprev, IPNucleation &ipv) const{
  // Compute the nucleated porosity
    // If the plastic strain is not increasing, no porosity is nucleated
  if (p<p0){
    ipv.getRefToNucleatedPorosity() =  ipvprev.getNucleatedPorosity();
    ipv.getRefToNucleatedPorosityIncrement() = 0.;
    ipv.getRefToDNucleatedPorosityDMatrixPlasticDeformation() = 0.;
  }
  else{
    // Otherwise, if plastic strain is increasing, porosity is nucleated following the Gaussian law
    double F =  _fn/sqrt(2.*3.14159*_sn*_sn)*exp(-0.5*(p-_epsilonn)*(p-_epsilonn)/_sn/_sn);
    double DF = -F*(p-_epsilonn)/_sn/_sn;

    ipv.getRefToNucleatedPorosity() =  ipvprev.getNucleatedPorosity() + F*(p-p0);
    ipv.getRefToNucleatedPorosityIncrement() = F*(p-p0);
    ipv.getRefToDNucleatedPorosityDMatrixPlasticDeformation() = F+ DF*(p-p0);
  }
};