//
// C++ Interface: terms
//
// Description: Define Gurson nucleation
//
//
// Author:  <L. Noels>, (C) 2013
//
// Copyright: See COPYING file that comes with this distribution
//
//
#ifndef _NUCLEATIONLAW_H_
#define _NUCLEATIONLAW_H_
#ifndef SWIG
#include "ipstate.h"
#include "MElement.h"
#include "ipNucleation.h"
#include "STensorOperations.h"
#endif

class NucleationLaw{
  // Mother class for optional nucleation law used in porous law
  // wich is strain-controlled
public :
  enum nucleationname{none,linearDamageNucleationRate,exponentialDamageNucleationRate};
protected :
  int _num; // number of law (must be unique !)
  bool _initialized; // to initialize law
  
public:
  // Constructor-destructor
#ifndef SWIG
  NucleationLaw(const int num, const bool init=true);
  virtual ~NucleationLaw(){};
  NucleationLaw(const NucleationLaw &source);
  virtual NucleationLaw& operator=(const NucleationLaw &source);
  virtual NucleationLaw * clone() const=0;
  
  // Function of materialLaw
  virtual int getNum() const{return _num;};
  virtual nucleationname getType() const=0;
  virtual void createIPVariable(IPNucleation* &ipv) const=0;
  virtual const bool isInitialized() {return _initialized;};
  
  // Specific functions
    // compute the nucleated porosity from the plastic strain evolution
  virtual void nucleate(const double p, const double p0, const double f0, const  IPNucleation &ipvprev, IPNucleation &ipv) const = 0;
#endif
};

class TrivialNucleationLaw : public NucleationLaw
{
  public:
    TrivialNucleationLaw(const int num);
  #ifndef SWIG
    TrivialNucleationLaw(const TrivialNucleationLaw& src): NucleationLaw(src){}
    virtual TrivialNucleationLaw& operator=(const NucleationLaw &source);
    virtual ~TrivialNucleationLaw(){};
    virtual NucleationLaw * clone() const {return new TrivialNucleationLaw(*this);};
    virtual nucleationname getType() const {return NucleationLaw::none;};
    virtual void createIPVariable(IPNucleation* &ipv) const;
    virtual void nucleate(const double p, const double p0, const double f0, const  IPNucleation &ipvprev, IPNucleation &ipv) const;
  #endif //SWIG
};

class LinearNucleationLaw : public NucleationLaw
{
  // Linear nucleation law:
  // The nucleated porosity rate is proportional to the plastic strain rate 
  // as long it is included between fmin and fmax
protected:
  double _A0;   // Constant growth rate: dot_fV = A0 * dot_hatp
  double _fmin; // Lower bound for nucleation
  double _fmax; // Upper bound for nucleation
public :
  // Constructor-destructor
  LinearNucleationLaw(const int num, const double A0, const double fmin=0., const double fmax=1., const bool init=true);
  virtual ~LinearNucleationLaw(){}
#ifndef SWIG
  LinearNucleationLaw(const LinearNucleationLaw &source);
  LinearNucleationLaw& operator=(const NucleationLaw &source);
  // Function of materialLaw
  virtual nucleationname getType() const{return NucleationLaw::linearDamageNucleationRate; };
  virtual void createIPVariable(IPNucleation* &ipv) const;
  virtual void initLaws(const std::map<int,NucleationLaw*> &maplaw){};//nothing now, we will see if we use the mapping
  virtual NucleationLaw * clone() const;
  virtual void nucleate(const double p, const double p0, const double f0, const  IPNucleation &ipvprev, IPNucleation &ipv) const;
#endif
};

class ExponentialNucleationLaw : public NucleationLaw
{
  // Gaussian nucleation law:
  // The nucleated porosity is a gausssian law:
  // -with a mean plastic strain value epsilonn,
  // -with a normal dispersion sn,
  // -with a total porosity nucleated fn.
 protected:
  double _fn;
  double _sn;
  double _epsilonn;
 public :
  // Constructor-destructor
  ExponentialNucleationLaw(const int num, double fn, double sn, double epsilonn, const bool init=true);
  virtual ~ExponentialNucleationLaw(){};
#ifndef SWIG
  ExponentialNucleationLaw(const ExponentialNucleationLaw &source);
  ExponentialNucleationLaw& operator=(const NucleationLaw &source);
  virtual NucleationLaw * clone() const;
  
  // Function of materialLaw
  virtual nucleationname getType() const{return NucleationLaw::exponentialDamageNucleationRate; };
  virtual void createIPVariable(IPNucleation* &ipv) const;
  
  // Specific functions
  virtual void nucleate(const double p, const double p0, const double f0, const  IPNucleation &ipvprev, IPNucleation &ipv) const;
#endif
};
#endif //_NUCLEATIONLAW_H_
