#include "STensorOperations.h"


bool STensorOperation::polynomial(const STensor3& a, const int order, const fullVector<double>& coeffs, STensor3& F, STensor43* dF, STensor63* ddF){
  if (order+1 != coeffs.size()) {
    coeffs.print("coeffs");
    Msg::Error("approximation is wrongly defined STensorOperation::polynomialApproximation order = %d",order);
  };
  
    
  static const STensor3 I(1.);
  static const STensor43 I4(1.,1.);

  if (STensorOperation::isnan(a)){
    #ifdef _DEBUG
    Msg::Error("STensorOperation::polynomial of a NAN tensor");
    #endif // _DEBUG
    STensorOperation::zero(F);
    if (dF){
      (*dF) = I4;
      if (ddF){
        STensorOperation::zero(*ddF);
      }
    }
    return false;
  }

  static STensor3 Bn;
  static STensor43 dBnDa;
  static STensor63 ddBnDDa;
  
  for (int j=0; j<3; j++){
    for (int k=0; k<3; k++){
      Bn(j,k) = I(j,k)*coeffs(order);
    }
  }
  if (dF != NULL){
    STensorOperation::zero(dBnDa);
    if (ddF!= NULL){
      STensorOperation::zero(ddBnDDa);
    }
  }
  
  for (int i=0; i< order; i++){
    STensorOperation::zero(F);
    if (dF != NULL){
      STensorOperation::zero(*dF);
      if (ddF != NULL){
        STensorOperation::zero(*ddF);
      }
    }
    
    int n =order-i;
    for (int j=0; j<3; j++){
      for (int k=0; k<3; k++){
        F(j,k) = coeffs(n-1)*I(j,k);
        for (int l=0; l<3; l++){
          F(j,k) +=  Bn(j,l)*a(l,k);
          if (dF != NULL){
            for (int r=0; r<3; r++){
              for (int s=0; s<3; s++){
                (*dF)(j,k,r,s) +=  dBnDa(j,l,r,s)*a(l,k) + Bn(j,l)*I4(l,k,r,s);
                if (ddF != NULL){
                  for (int p=0; p<3; p++){
                    for (int q=0; q<3; q++){
                      (*ddF)(j,k,r,s,p,q) +=  ddBnDDa(j,l,r,s,p,q)*a(l,k)+dBnDa(j,l,r,s)*I4(l,k,p,q) + dBnDa(j,l,p,q)*I4(l,k,r,s);
                    }
                  }
                }
              }
            }              
          }
        }
      }
    }
    Bn = F;
    if (dF != NULL){
      dBnDa = (*dF);
      if (ddF != NULL){
        ddBnDDa = (*ddF);
      }
    }
  }
  return true;
};

void STensorOperation::logSTensor3(const STensor3& a, const int order, STensor3& loga, STensor43* dloga, STensor63* ddloga){
  // Tools
  static const STensor3 mI(-1.);
  static const STensor43 I4(1.,1.);
  static STensor3 ami;

  // Choose approximation order
  if (order == 1){
    // Order 1 : log(1+A) = A
    ami = mI;
    ami+= a;
    loga = ami;
    // First derivative
    if(dloga!=NULL)
    {
      (*dloga) = I4;
    }
    // Second derivative
    if(ddloga!=NULL)
    {
      (*ddloga) *= 0.;
    }
  }
  else if (order > 1){
    static STensor3 I(1.);
    ami = mI;
    ami+= a;
    
    fullVector<double> coeffs(order+1);
    
    for (int i=0; i< order+1; i++){
      if (i==0){
        coeffs(i) = 0.;
      }
      else if (i%2 == 0){
        coeffs(i) = -1./double(i);
      }
      else{
        coeffs(i) = 1./double(i);
      }
    }
    //coeffs.print("coeffs");
    
    bool ok =  STensorOperation::polynomial(ami,order,coeffs,loga,dloga,ddloga);
    if (!ok){
      #ifdef _DEBUG
      Msg::Error("error in computing logSTensor3");
      #endif // _DEBUG
    }
  }
  else if (order == -1)
  { 
    static STensor3 ainv;    
    static fullMatrix<double> m(3, 3);
    static fullVector<double> eigenValReal(3);
    static fullVector<double> eigenValImag(3);
    static fullMatrix<double> leftEigenVect(3,3);
    static fullMatrix<double> rightEigenVect(3,3);
    
    // Get eigen values and vectors
    a.getMat(m);
    m.eig(eigenValReal,eigenValImag,leftEigenVect,rightEigenVect,0);
    
    // Spectral sum for log function
    for (int i=0; i<3; i++){
      for (int j=0; j<3; j++){
        loga(i,j) =  log(eigenValReal(0)) *leftEigenVect(i,0)*rightEigenVect(j,0) 
                    +log(eigenValReal(1)) *leftEigenVect(i,1)*rightEigenVect(j,1)
                    +log(eigenValReal(2)) *leftEigenVect(i,2)*rightEigenVect(j,2);
      }
    }
    // First derivative
    inverseSTensor3(a,ainv);
    if(dloga!=NULL)
    {
			STensorOperation::zero(*dloga);
      for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
          for (int k=0; k<3; k++){
            for (int l=0; l<3; l++){
              for (int r=0; r<3; r++){
                dloga->operator()(i,j,k,l) += I4(i,r,k,l)*ainv(r,j);
              }
            }
          }
        }
      }
    }
    //Second derivative
    if(ddloga!=NULL)
    {
      STensorOperation::inverseSTensor3(a,ainv);
      for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
          for (int k=0; k<3; k++){
            for (int l=0; l<3; l++){
              for (int m=0; m<3; m++){
                for (int n=0; n<3; n++){
                  ddloga->operator()(i,j,k,l,m,n) = 0.;
                  for (int r=0; r<3; r++){
                    for (int s=0; s<3; s++){
                      for (int q=0; q<3; q++){
                        ddloga->operator()(i,j,k,l,m,n) -= I4(i,r,k,l)*I4(r,s,m,n)*ainv(s,q)*ainv(q,j);
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  else
  {
    Msg::Error("order = %d is not implemented",order);
  }
}


void STensorOperation::expSTensor3(const STensor3& a, const int order, STensor3& expa, STensor43* dexpa, STensor63* ddexpa)
{
  // Tools
  static const STensor3 I(1.);
  static const STensor43 I4(1.);
  
  // Choose approximation order 
  if (order == 1) 
  {
    // Order 1 : exp(A) = I + A
    expa = I;
    expa += a;
    // First derivative
    if(dexpa!=NULL)
    {
     (*dexpa) = I4;
    }
    // Second derivative
    if(ddexpa!=NULL)
    {
      (*ddexpa) *= 0.;
    }
  }
  else if (order > 1){
    fullVector<double> coeffs(order+1);
    coeffs(0) =1.;
    coeffs(1) =1;
    for (int i=2; i< order+1; i++){
      coeffs(i) = coeffs(i-1)/double(i);
    }
    //coeffs.print("coeffs");
    
    bool ok = STensorOperation::polynomial(a,order,coeffs,expa,dexpa,ddexpa);
    if (!ok){
      #ifdef _DEBUG
      Msg::Error("error in computing expSTensor3");
      #endif // _DEBUG
    }
  }
  else if (order == -1)
  {
    static fullMatrix<double> m(3, 3);
    static fullVector<double> eigenValReal(3);
    static fullVector<double> eigenValImag(3);
    static fullMatrix<double> leftEigenVect(3,3);
    static fullMatrix<double> rightEigenVect(3,3);
    
    // Get eigen values and vectors
    a.getMat(m);
    m.eig(eigenValReal,eigenValImag,leftEigenVect,rightEigenVect,0);
    
    // Spectral sum for exp function
    for (int i=0; i<3; i++){
      for (int j=0; j<3; j++){
        expa(i,j) =  exp(eigenValReal(0)) *leftEigenVect(i,0)*rightEigenVect(j,0) 
                    +exp(eigenValReal(1)) *leftEigenVect(i,1)*rightEigenVect(j,1)
                    +exp(eigenValReal(2)) *leftEigenVect(i,2)*rightEigenVect(j,2);
      }
    }
    // First derivative
    if(dexpa!=NULL)
    {
      for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
          for (int k=0; k<3; k++){
            for (int l=0; l<3; l++){
              dexpa->operator()(i,j,k,l) = 0.;
              for (int r=0; r<3; r++){
                dexpa->operator()(i,j,k,l) += I4(i,r,k,l)*expa(r,j);
              }
            }
          }
        }
      }
    }
    // Second derivative
    if(ddexpa!=NULL)
    {
      for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
          for (int k=0; k<3; k++){
            for (int l=0; l<3; l++){
              for (int m=0; m<3; m++){
                for (int n=0; n<3; n++){
                  ddexpa->operator()(i,j,k,l,m,n) = 0.;
                  for (int r=0; r<3; r++){
                    for (int s=0; s<3; s++){
                      ddexpa->operator()(i,j,k,l,m,n) += I4(i,r,k,l)*I4(r,s,m,n)*expa(s,j);
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  else
  {
    Msg::Error("order = %d is not implemented",order);
  }
}

void STensorOperation::VRDecomposition(const STensor3& a, STensor3& V, STensor3&R)
{
    static STensor3 biot,invV;
    static fullMatrix<double> m(3, 3);
    static fullVector<double> eigenValReal(3);
    static fullVector<double> eigenValImag(3);
    static fullMatrix<double> leftEigenVect(3,3);
    static fullMatrix<double> rightEigenVect(3,3);

    STensorOperation::multSTensor3SecondTranspose(a,a,biot);
    // Get eigen values and vectors
    biot.getMat(m);
    m.eig(eigenValReal,eigenValImag,leftEigenVect,rightEigenVect,0);
    
    // Spectral sum for exp function
    for (int i=0; i<3; i++){
      for (int j=0; j<3; j++){
        V(i,j) =  sqrt(eigenValReal(0)) *leftEigenVect(i,0)*rightEigenVect(j,0) 
                    +sqrt(eigenValReal(1)) *leftEigenVect(i,1)*rightEigenVect(j,1)
                    +sqrt(eigenValReal(2)) *leftEigenVect(i,2)*rightEigenVect(j,2);
        invV(i,j) =  1./sqrt(eigenValReal(0)) *leftEigenVect(i,0)*rightEigenVect(j,0) 
                    +1./sqrt(eigenValReal(1)) *leftEigenVect(i,1)*rightEigenVect(j,1)
                    +1./sqrt(eigenValReal(2)) *leftEigenVect(i,2)*rightEigenVect(j,2);
      }
    }
    STensorOperation::multSTensor3(invV, a, R);
    //check
    //static STensor3 abis;
    //abis=V;
    //abis*=R;
    //abis-=a;
    //double norm=abis.norm2();
    //Msg::Info("norm of VRDecomposition: %f",norm);
}


void STensorOperation::RUDecomposition(const STensor3& a, STensor3& U, STensor3&R)
{
    static STensor3 C,invU;
    static fullMatrix<double> m(3, 3);
    static fullVector<double> eigenValReal(3);
    static fullVector<double> eigenValImag(3);
    static fullMatrix<double> leftEigenVect(3,3);
    static fullMatrix<double> rightEigenVect(3,3);

    STensorOperation::multSTensor3FirstTranspose(a,a,C);
    // Get eigen values and vectors
    C.getMat(m);
    m.eig(eigenValReal,eigenValImag,leftEigenVect,rightEigenVect,0);
    
    // Spectral sum for exp function
    for (int i=0; i<3; i++){
      for (int j=0; j<3; j++){
        U(i,j) =  sqrt(eigenValReal(0)) *leftEigenVect(i,0)*rightEigenVect(j,0) 
                    +sqrt(eigenValReal(1)) *leftEigenVect(i,1)*rightEigenVect(j,1)
                    +sqrt(eigenValReal(2)) *leftEigenVect(i,2)*rightEigenVect(j,2);
        invU(i,j) =  1./sqrt(eigenValReal(0)) *leftEigenVect(i,0)*rightEigenVect(j,0) 
                    +1./sqrt(eigenValReal(1)) *leftEigenVect(i,1)*rightEigenVect(j,1)
                    +1./sqrt(eigenValReal(2)) *leftEigenVect(i,2)*rightEigenVect(j,2);
      }
    }
    STensorOperation::multSTensor3(a,invU, R);
    //check
    //static STensor3 abis;
    //abis=R;
    //abis*=U;
    //abis-=a;
    //double norm=abis.norm2();
    //Msg::Info("norm of RUDecomposition: %f ",norm);

}


