//
// C++ Interface: material law
//
// Author:  <Van Dung Nguyen>, (C) 2017
//
// Copyright: See COPYING file that comes with this distribution
//
//
#include "elasticPotential.h"

smallStrainLinearElasticPotential::smallStrainLinearElasticPotential(const int num, const double E, const double nu): elasticPotential(num), _E(E), _nu(nu){

  _K = _E/(3.*(1.-2.*_nu));
  _mu = _E/(2.*(1.+_nu));
  double lambda = _K- 2.*_mu/3.;
  static const STensor3 I(1.);
  for (int i=0; i<3; i++){
    for (int j=0; j<3; j++){
      for (int k=0; k<3; k++){
        for (int l=0; l<3; l++){
          _Cel(i,j,k,l) = lambda*I(i,j)*I(k,l)+_mu*(I(i,k)*I(j,l)+I(i,l)*I(j,k));
        }
      }
    }
	}
};
smallStrainLinearElasticPotential::smallStrainLinearElasticPotential(const smallStrainLinearElasticPotential& src) : elasticPotential(src),
	_E(src._E),_nu(src._nu),_K(src._K),_mu(src._mu), _Cel(src._Cel){};

double smallStrainLinearElasticPotential::get(const STensor3& F) const{
  // small strain value
  static STensor3 E;
  STensorOperation::diag(E,-1.);
	for (int i=0; i<3; i++){
		for (int j=0; j<3; j++){
			E(i,j) += (F(i,j) + F(j,i))*0.5;
		}
	}
	double val = 0.;
	for (int i=0; i<3;i++){
		for (int j=0; j<3; j++){
			for (int k=0; k<3; k++){
				for (int l=0; l<3; l++){
					val += E(i,j)*_Cel(i,j,k,l)*E(k,l)*0.5;
				}
			}
		}
	}
	return val;
};
void smallStrainLinearElasticPotential::constitutive(const STensor3& F, STensor3& P, const bool stiff, STensor43* L) const{
  static STensor3 E;
  STensorOperation::diag(E,-1.);
	for (int i=0; i<3; i++){
		for (int j=0; j<3; j++){
			E(i,j) += (F(i,j) + F(j,i))*0.5;
		}
	}
	
	for (int i=0; i<3;i++){
		for (int j=0; j<3; j++){
      P(i,j) = 0.;
			for (int k=0; k<3; k++){
				for (int l=0; l<3; l++){
					P(i,j) += _Cel(i,j,k,l)*E(k,l);
				}
			}
		}
	}
	if (stiff){
		(*L) = _Cel;
	}
};
void smallStrainLinearElasticPotential::getLocalValForNonLocalEquation(const STensor3& F, double& val, const bool stiff, STensor3* DvalDF) const{
	double Y = get(F); // potential
	double E = getYoungModulus(); // young modulus

	val = sqrt(2*Y/E); // isotropic nonlocal strain
	if (stiff){
		this->constitutive(F,*DvalDF,false,NULL);
		if (val > 1e-10){
			(*DvalDF)*= (1./(E*val));
		}
		else{
			(*DvalDF)*= (1./(E*1e-10));
		}

	}
};

// for anisotropic
double smallStrainLinearElasticPotential::getPositiveValue(const STensor3& F) const {
  return get(F);
};
double smallStrainLinearElasticPotential::getNegativeValue(const STensor3& F) const {
  return 0.;
};
void smallStrainLinearElasticPotential::constitutivePositive(const STensor3& F, STensor3& P, const bool stiff, STensor43* L) const {
  this->constitutive(F,P,stiff,L);
};
void smallStrainLinearElasticPotential::constitutiveNegative(const STensor3& F, STensor3& P, const bool stiff, STensor43* L) const {
  STensorOperation::zero(P);
  if (stiff){
    STensorOperation::zero(*L);
  }
};
void smallStrainLinearElasticPotential::getLocalPositiveValForNonLocalEquation(const STensor3& F, double& val, const bool stiff, STensor3* DvalDF) const {
  this->getLocalValForNonLocalEquation(F,val,stiff,DvalDF);
};

largeStrainAnisotropicPotential::largeStrainAnisotropicPotential(const int num, 
                                    const double Ex, const double Ey, const double Ez, 
                                    const double Vxy, const double Vxz, const double Vyz,
                                    const double MUxy, const double MUxz, const double MUyz,
                                    const double alpha, const double beta, const double gamma): elasticPotential(num){
  const double Vyx= Vxy*Ey/Ex  ; 
	const double Vzx= Vxz*Ez/Ex  ; 
	const double Vzy= Vyz*Ez/Ey  ;
	const double D=( 1-Vxy*Vyx-Vzy*Vyz-Vxz*Vzx-2*Vxy*Vyz*Vzx ) / ( Ex*Ey*Ez );

 	STensor43 ElasticityTensor(0.);
        
 	ElasticityTensor(0,0,0,0)=( 1-Vyz*Vzy ) / (Ey*Ez*D );
 	ElasticityTensor(1,1,1,1)=( 1-Vxz*Vzx ) / (Ex*Ez*D );
	ElasticityTensor(2,2,2,2)=( 1-Vyx*Vxy ) / (Ey*Ex*D );

	ElasticityTensor(0,0,1,1)=( Vyx+Vzx*Vyz ) / (Ey*Ez*D );
	ElasticityTensor(0,0,2,2)=( Vzx+Vyx*Vzy ) / (Ey*Ez*D );
	ElasticityTensor(1,1,2,2)=( Vzy+Vxy*Vzx ) / (Ex*Ez*D );

	ElasticityTensor(1,1,0,0)=( Vxy+Vzy*Vxz ) / (Ex*Ez*D );
	ElasticityTensor(2,2,0,0)=( Vxz+Vxy*Vyz ) / (Ey*Ex*D );
	ElasticityTensor(2,2,1,1)=( Vyz+Vxz*Vyx ) / (Ey*Ex*D );

 	ElasticityTensor(1,2,1,2)=MUyz;ElasticityTensor(1,2,2,1)=MUyz;
	ElasticityTensor(2,1,2,1)=MUyz;ElasticityTensor(2,1,1,2)=MUyz;

	ElasticityTensor(0,1,0,1)=MUxy;ElasticityTensor(0,1,1,0)=MUxy;
	ElasticityTensor(1,0,1,0)=MUxy;ElasticityTensor(1,0,0,1)=MUxy;

	ElasticityTensor(0,2,0,2)=MUxz;ElasticityTensor(0,2,2,0)=MUxz;
	ElasticityTensor(2,0,2,0)=MUxz;ElasticityTensor(2,0,0,2)=MUxz;

  
	double c1,c2,c3,s1,s2,s3;
	double s1c2, c1c2;
	double pi(3.14159265359);
  double fpi = pi/180.;

	c1 = cos(alpha*fpi);
	s1 = sin(alpha*fpi);

	c2 = cos(beta*fpi);
	s2 = sin(beta*fpi);

	c3 = cos(gamma*fpi);
	s3 = sin(gamma*fpi);

	s1c2 = s1*c2;
	c1c2 = c1*c2;
  
  STensor3 R; //3x3 rotation matrix
	R(0,0) = c3*c1 - s1c2*s3;
	R(0,1) = c3*s1 + c1c2*s3;
	R(0,2) = s2*s3;

	R(1,0) = -s3*c1 - s1c2*c3;
	R(1,1) = -s3*s1 + c1c2*c3;
	R(1,2) = s2*c3;

	R(2,0) = s1*s2;
	R(2,1) = -c1*s2;
	R(2,2) = c2;

  for(int i=0;i<3;i++){
    for(int j=0;j<3;j++){
      for(int k=0;k<3;k++){
        for(int l=0;l<3;l++){
          _Cel(i,j,k,l)=0.;
          for(int m=0;m<3;m++){
            for(int n=0;n<3;n++){
              for(int o=0;o<3;o++){
                for(int p=0;p<3;p++){
                  _Cel(i,j,k,l)+=R(m,i)*R(n,j)*R(o,k)*R(p,l)*ElasticityTensor(m,n,o,p);
                }
              }
            }
          }
        }
      }
    }
  } 
	
  _Emax = std::max(Ex,std::max(Ey,Ez));
  
	//Init du _poissonMax
	if (Vxy >=  Vxz && Vxy >= Vyz){
    _poissonMax=Vxy;
  }
	else if (Vxz >=  Vyz && Vxz >= Vxy){
    _poissonMax=Vxz;
  }
	else{
    _poissonMax=Vyz;
  }
};

largeStrainAnisotropicPotential::largeStrainAnisotropicPotential(const int num, const STensor43& C):elasticPotential(num),_Cel(C){
  _Emax = _Cel(0,0,0,0);
  _poissonMax = 0.;
};

double largeStrainAnisotropicPotential::get(const STensor3& F) const{
  static STensor3 E;
  STensorOperation::diag(E,-0.5);
  for (int i=0; i<3; i++){
    for (int j=0; j<3; j++){
      E(i,j) += 0.5*(F(0,i)*F(0,j)+F(1,i)*F(1,j)+F(2,i)*F(2,j));
    }
  }
  
  double val = 0.;
  for (int i=0; i<3; i++){
    for (int j=0; j<3; j++){
      for (int k=0; k<3; k++){
        for (int l=0; l<3; l++){
          val += 0.5*E(i,j)*_Cel(i,j,k,l)*E(k,l);
        }
      }
    }
  }
  return val;
}; // elastic potential
void largeStrainAnisotropicPotential::constitutive(const STensor3& F, STensor3& P, const bool stiff, STensor43* L) const{
  static STensor3 E;
  STensorOperation::diag(E,-0.5);
  for (int i=0; i<3; i++){
    for (int j=0; j<3; j++){
      E(i,j) += 0.5*(F(0,i)*F(0,j)+F(1,i)*F(1,j)+F(2,i)*F(2,j));
    }
  }
  
  static STensor3 secondPK;
  for (int i=0; i<3; i++){
    for (int j=0; j<3; j++){
      secondPK(i,j) = 0.;
      for (int k=0; k<3; k++){
        for (int l=0; l<3; l++){
          secondPK(i,j) += _Cel(i,j,k,l)*E(k,l);
        }
      }
    }
  }
  
  for (int i=0; i<3; i++){
    for (int j=0; j<3; j++){
      P(i,j) = F(i,0)*secondPK(0,j)+F(i,1)*secondPK(1,j)+F(i,2)*secondPK(2,j);
    }
  }

  if (stiff){
    STensorOperation::zero(*L);
    for (int i=0; i<3; i++){
      for (int j=0; j<3; j++){
        for (int p=0; p<3; p++){
          (*L)(i,j,i,p) += secondPK(p,j);
          for (int q=0; q<3; q++){
            for (int s=0; s<3; s++){
              for (int k=0; k<3; k++){
                (*L)(i,j,p,q) += F(i,k)*_Cel(k,j,q,s)*F(p,s);
              }
            }
          }
        }
      }
    }
  }  
};

void largeStrainAnisotropicPotential::getLocalValForNonLocalEquation(const STensor3& F, double& val, const bool stiff, STensor3* DvalDF) const{
  double Y = get(F);
	double E = getYoungModulus();

	val = sqrt(2*Y/E);
	if (stiff){
		(*DvalDF) *= 0.;
    this->constitutive(F,*DvalDF,false,NULL);
    if (val > 1e-10){
      (*DvalDF) *= (1./(E*val));
    }
    else{
      (*DvalDF) *= (1./(E*1e-10));
    }
	}
};

// for anisotropic
double largeStrainAnisotropicPotential::getPositiveValue(const STensor3& F) const{
  return get(F);
};
double largeStrainAnisotropicPotential::getNegativeValue(const STensor3& F) const{
  return 0.;
};
void largeStrainAnisotropicPotential::constitutivePositive(const STensor3& F, STensor3& P, const bool stiff, STensor43* L) const{
  this->constitutive(F,P,stiff,L);
};
void largeStrainAnisotropicPotential::constitutiveNegative(const STensor3& F, STensor3& P, const bool stiff, STensor43* L) const{
  STensorOperation::zero(P);
  if (stiff){
    STensorOperation::zero(*L);
  }
};
void largeStrainAnisotropicPotential::getLocalPositiveValForNonLocalEquation(const STensor3& F, double& val, const bool stiff, STensor3* DvalDF) const{
  getLocalValForNonLocalEquation(F,val,stiff,DvalDF);
};


largeStrainAnisotropicPotential::largeStrainAnisotropicPotential(const largeStrainAnisotropicPotential& src): elasticPotential(src),
          _Cel(src._Cel),_poissonMax(src._poissonMax){};
          
largeStrainTransverseIsotropicPotential::largeStrainTransverseIsotropicPotential(const int num, 
                                    const double ET, // inplane modulus
                                    const double EL,  // longitudinal modulus 
                                    const double nuTT, // inplane Poisson ratio
                                    const double nuLT,  // inplane-longitudinal Possion ratio
                                    const double MUTT,  // inplane shear
                                    const double MUTL, // out-plane shear
                                    const double alpha, const double beta, const double gamma)
                                    : elasticPotential(num){
  const double nuTL = ET*nuLT/EL;
  const double D=(1.+nuTT)*(1.-nuTT-2*nuTL*nuLT)/(ET*ET*EL);

 	STensor43 ElasticityTensor(0.);
        
 	ElasticityTensor(0,0,0,0)=( 1-nuTL*nuLT ) / (ET*EL*D );
 	ElasticityTensor(1,1,1,1)=( 1-nuTL*nuLT ) / (ET*EL*D );
	ElasticityTensor(2,2,2,2)=( 1-nuTT*nuTT ) / (ET*ET*D );

	ElasticityTensor(0,0,1,1)=( nuTT+nuLT*nuTL) / (ET*EL*D );
	ElasticityTensor(0,0,2,2)=( nuLT+nuTT*nuLT) / (ET*EL*D );
	ElasticityTensor(1,1,2,2)=( nuLT+nuTT*nuLT) / (ET*EL*D );

	ElasticityTensor(1,1,0,0)=ElasticityTensor(0,0,1,1); //( nuTT+nuLT*nuTL) / (ET*EL*D );
	ElasticityTensor(2,2,0,0)=ElasticityTensor(0,0,2,2); //( nuTL+nuTT*nuTL) / (ET*ET*D );
	ElasticityTensor(2,2,1,1)=ElasticityTensor(1,1,2,2); //( nuTL+nuTT*nuTL) / (ET*ET*D );

 	ElasticityTensor(1,2,1,2)=MUTL;
  ElasticityTensor(1,2,2,1)=MUTL;
	ElasticityTensor(2,1,2,1)=MUTL;
  ElasticityTensor(2,1,1,2)=MUTL;
  
  ElasticityTensor(0,2,0,2)=MUTL;
  ElasticityTensor(0,2,2,0)=MUTL;
	ElasticityTensor(2,0,2,0)=MUTL;
  ElasticityTensor(2,0,0,2)=MUTL;

	ElasticityTensor(0,1,0,1)=MUTT;
  ElasticityTensor(0,1,1,0)=MUTT;
	ElasticityTensor(1,0,1,0)=MUTT;
  ElasticityTensor(1,0,0,1)=MUTT;

	double c1,c2,c3,s1,s2,s3;
	double s1c2, c1c2;
	double pi(3.14159265359);
  double fpi = pi/180.;

	c1 = cos(alpha*fpi);
	s1 = sin(alpha*fpi);

	c2 = cos(beta*fpi);
	s2 = sin(beta*fpi);

	c3 = cos(gamma*fpi);
	s3 = sin(gamma*fpi);

	s1c2 = s1*c2;
	c1c2 = c1*c2;
  
  STensor3 R; //3x3 rotation matrix
	R(0,0) = c3*c1 - s1c2*s3;
	R(0,1) = c3*s1 + c1c2*s3;
	R(0,2) = s2*s3;

	R(1,0) = -s3*c1 - s1c2*c3;
	R(1,1) = -s3*s1 + c1c2*c3;
	R(1,2) = s2*c3;

	R(2,0) = s1*s2;
	R(2,1) = -c1*s2;
	R(2,2) = c2;

  for(int i=0;i<3;i++){
    for(int j=0;j<3;j++){
      for(int k=0;k<3;k++){
        for(int l=0;l<3;l++){
          _Cel(i,j,k,l)=0.;
          for(int m=0;m<3;m++){
            for(int n=0;n<3;n++){
              for(int o=0;o<3;o++){
                for(int p=0;p<3;p++){
                  _Cel(i,j,k,l)+=R(m,i)*R(n,j)*R(o,k)*R(p,l)*ElasticityTensor(m,n,o,p);
                }
              }
            }
          }
        }
      }
    }
  } 
	
  _Emax = std::max(ET,EL);
  _poissonMax = std::max(nuTT,std::max(nuTL,nuLT));
};

largeStrainTransverseIsotropicPotential::largeStrainTransverseIsotropicPotential(const largeStrainTransverseIsotropicPotential& src): elasticPotential(src),
          _Cel(src._Cel),_poissonMax(src._poissonMax){};


double largeStrainTransverseIsotropicPotential::get(const STensor3& F) const{
  static STensor3 E;
  STensorOperation::diag(E,-0.5);
  for (int i=0; i<3; i++){
    for (int j=0; j<3; j++){
      E(i,j) += 0.5*(F(0,i)*F(0,j)+F(1,i)*F(1,j)+F(2,i)*F(2,j));
    }
  }
  
  double val = 0.;
  for (int i=0; i<3; i++){
    for (int j=0; j<3; j++){
      for (int k=0; k<3; k++){
        for (int l=0; l<3; l++){
          val += 0.5*E(i,j)*_Cel(i,j,k,l)*E(k,l);
        }
      }
    }
  }
  return val;
}; // elastic potential
void largeStrainTransverseIsotropicPotential::constitutive(const STensor3& F, STensor3& P, const bool stiff, STensor43* L) const{
  static STensor3 E;
  STensorOperation::diag(E,-0.5);
  for (int i=0; i<3; i++){
    for (int j=0; j<3; j++){
      E(i,j) += 0.5*(F(0,i)*F(0,j)+F(1,i)*F(1,j)+F(2,i)*F(2,j));
    }
  }
  
  static STensor3 secondPK;
  for (int i=0; i<3; i++){
    for (int j=0; j<3; j++){
      secondPK(i,j) = 0.;
      for (int k=0; k<3; k++){
        for (int l=0; l<3; l++){
          secondPK(i,j) += _Cel(i,j,k,l)*E(k,l);
        }
      }
    }
  }
  
  for (int i=0; i<3; i++){
    for (int j=0; j<3; j++){
      P(i,j) = F(i,0)*secondPK(0,j)+F(i,1)*secondPK(1,j)+F(i,2)*secondPK(2,j);
    }
  }

  if (stiff){
    STensorOperation::zero(*L);
    for (int i=0; i<3; i++){
      for (int j=0; j<3; j++){
        for (int p=0; p<3; p++){
          (*L)(i,j,i,p) += secondPK(p,j);
          for (int q=0; q<3; q++){
            for (int s=0; s<3; s++){
              for (int k=0; k<3; k++){
                (*L)(i,j,p,q) += F(i,k)*_Cel(k,j,q,s)*F(p,s);
              }
            }
          }
        }
      }
    }
  }  
};

void largeStrainTransverseIsotropicPotential::getLocalValForNonLocalEquation(const STensor3& F, double& val, const bool stiff, STensor3* DvalDF) const{
  double Y = get(F);
	double E = getYoungModulus();

	val = sqrt(2*Y/E);
	if (stiff){
		(*DvalDF) *= 0.;
    this->constitutive(F,*DvalDF,false,NULL);
    if (val > 1e-10){
      (*DvalDF) *= (1./(E*val));
    }
    else{
      (*DvalDF) *= (1./(E*1e-10));
    }
	}
};

// for anisotropic
double largeStrainTransverseIsotropicPotential::getPositiveValue(const STensor3& F) const{
  return get(F);
};
double largeStrainTransverseIsotropicPotential::getNegativeValue(const STensor3& F) const{
  return 0.;
};
void largeStrainTransverseIsotropicPotential::constitutivePositive(const STensor3& F, STensor3& P, const bool stiff, STensor43* L) const{
  this->constitutive(F,P,stiff,L);
};
void largeStrainTransverseIsotropicPotential::constitutiveNegative(const STensor3& F, STensor3& P, const bool stiff, STensor43* L) const{
  STensorOperation::zero(P);
  if (stiff){
    STensorOperation::zero(*L);
  }
};
void largeStrainTransverseIsotropicPotential::getLocalPositiveValForNonLocalEquation(const STensor3& F, double& val, const bool stiff, STensor3* DvalDF) const{
  getLocalValForNonLocalEquation(F,val,stiff,DvalDF);
};



biLogarithmicElasticPotential::biLogarithmicElasticPotential(const int num, const double E, const double nu, const int order, const bool pert, const double tol) :
	elasticPotential(num), _E(E), _nu(nu), _order(order), _ByPerturbation(pert),_tol(tol){
  _K = _E/(3.*(1.-2.*_nu));
  _mu = _E/(2.*(1.+_nu));
  double lambda = _K- 2.*_mu/3.;
  const STensor3 I(1.);
  for (int i=0; i<3; i++){
    for (int j=0; j<3; j++){
      for (int k=0; k<3; k++){
        for (int l=0; l<3; l++){
          _Cel(i,j,k,l) = lambda*I(i,j)*I(k,l)+_mu*(I(i,k)*I(j,l)+I(i,l)*I(j,k));
        }
      }
    }
  }
};

biLogarithmicElasticPotential::biLogarithmicElasticPotential(const biLogarithmicElasticPotential& src) : elasticPotential(src),
	_E(src._E),_nu(src._nu),_K(src._K),_mu(src._mu), _Cel(src._Cel),_order(src._order),
	_ByPerturbation(src._ByPerturbation),_tol(src._tol){};

double biLogarithmicElasticPotential::get(const STensor3& F) const{
  double J = F.determinant();
  static STensor3 C, logC;
  
  for (int i=0; i<3; i++)
    for (int j=0; j<3; j++)
      C(i,j) = F(0,i)*F(0,j)+F(1,i)*F(1,j)+F(2,i)*F(2,j);

  STensorOperation::logSTensor3(C,_order,logC);
  static STensor3 logCdev;
  double trlogC;
  STensorOperation::decomposeDevTr(logC, logCdev, trlogC);
  
  double val = 0.5*_K*log(J)*log(J)+ 0.25*_mu*logCdev.dotprod();
  return val;
};
void biLogarithmicElasticPotential::constitutive(const STensor3& F, STensor3& P, const bool stiff, STensor43* L) const{
  if (_ByPerturbation){
     double potential = get(F);
     static STensor3 Fp;
     for (int i=0; i<3; i++){
       for (int j=0; j<3; j++){
         Fp=(F);
         Fp(i,j) += _tol;
         double potentialPlus = get(Fp);
         P(i,j) = (potentialPlus - potential)/_tol;
       }
     }

     if (stiff){
      static STensor3 Pp;
      for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
          Fp = F;
          Fp(i,j) += _tol;
          constitutive(Fp,Pp,false,NULL);
          for (int p=0; p<3; p++){
            for (int q=0; q<3; q++){
              (*L)(p,q,i,j) = (Pp(p,q) - P(p,q))/_tol;
            }
          }
        }
      }
    }
   }
   else{
     double J = F.determinant();
     static STensor3 C, logC;
     for (int i=0; i<3; i++)
       for (int j=0; j<3; j++)
         C(i,j) = F(0,i)*F(0,j)+F(1,i)*F(1,j)+F(2,i)*F(2,j);
    static STensor3 invC;
    STensorOperation::inverseSTensor3(C, invC);

    static STensor43 dlogCdC;
    STensorOperation::logSTensor3(C,_order,logC,&dlogCdC);
    static STensor3 logCdev;
    double trlogC;
    STensorOperation::decomposeDevTr(logC, logCdev, trlogC);

    static STensor3 Svol, Sdev;
    for (int i=0; i<3;i++){
      for (int j=0; j<3; j++){
        Svol(i,j) = _K*log(J)*invC(i,j);
        Sdev(i,j) = 0.;
        for (int k=0; k<3; k++){
          Sdev(i,j) += _mu*invC(i,k)*logCdev(k,j);
        }
      };
    }

    for (int i=0; i<3; i++){
      for (int j=0; j<3; j++){
        P(i,j) = F(i,0)*(Svol(0,j)+Sdev(0,j))+F(i,1)*(Svol(1,j)+Sdev(1,j))+F(i,2)*(Svol(2,j)+Sdev(2,j));
      }
    }

    if (stiff){
      static STensor43 dSvoldC, dSdevdC;
      for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
          for (int k=0; k<3; k++){
            for (int l=0; l<3; l++){
              dSvoldC(i,j,k,l) = 0.5*_K*invC(i,j)*invC(k,l) - 0.5*_K*log(J)*(invC(i,k)*invC(j,l)+invC(i,l)*invC(j,k));
              dSdevdC(i,j,k,l) = 0.;
              for (int p=0; p<3; p++){
                dSdevdC(i,j,k,l) += -_mu*0.5*(invC(i,k)*invC(p,l)+invC(i,l)*invC(p,k))*logCdev(p,j)
                                    +_mu*invC(i,p)*dlogCdC(p,j,k,l)
                                    -_mu*invC(i,j)*dlogCdC(p,p,k,l)/3.;
              }
            };
          };
        };
      };
      

      STensorOperation::zero(*L);
      for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
          for (int p=0; p<3; p++){
            (*L)(i,j,i,p) += Svol(p,j)+Sdev(p,j);
            for (int q=0; q<3; q++){
              for (int s=0; s<3; s++){
                for (int k=0; k<3; k++){
                  (*L)(i,j,p,q) += 2.*F(i,k)*(dSvoldC(k,j,q,s)+dSdevdC(k,j,q,s))*F(p,s);
                }
              }
            }
          }
        }
      }
		}

	}

};
void biLogarithmicElasticPotential::getLocalValForNonLocalEquation(const STensor3& F, double& val, const bool stiff, STensor3* DvalDF) const{
	double Y = get(F);
	double E = getYoungModulus();

	val = sqrt(2*Y/E);
	if (stiff){
		(*DvalDF) *= 0.;
		if (_ByPerturbation){
      static STensor3 Fp;
			for (int i=0; i<3; i++){
				for (int j=0; j<3; j++){
          Fp=(F);
					Fp(i,j) += _tol;
					double valp;
					getLocalValForNonLocalEquation(Fp,valp,false,NULL);
					(*DvalDF)(i,j) = (valp-val)/_tol;
				}
			}
		}
		else{
			this->constitutive(F,*DvalDF,false,NULL);
			if (val > 1e-10){
				(*DvalDF) *= (1./(E*val));
			}
			else{
				(*DvalDF) *= (1./(E*1e-10));
			}

		}
	}
};

// for anisotropic
double biLogarithmicElasticPotential::getPositiveValue(const STensor3& F) const {
  double J = F.determinant();
  static STensor3 C, logC;
  for (int i=0; i<3; i++)
    for (int j=0; j<3; j++)
      C(i,j) = F(0,i)*F(0,j)+F(1,i)*F(1,j)+F(2,i)*F(2,j);

  STensorOperation::logSTensor3(C,_order,logC);
  static STensor3 logCdev;
  double trlogC;
  STensorOperation::decomposeDevTr(logC, logCdev, trlogC);
  
  if (J >=1.){
    return 0.5*_K*log(J)*log(J)+ 0.25*_mu*logCdev.dotprod();
  }
  else{
    return 0.25*_mu*logCdev.dotprod();
  }
  
};
double biLogarithmicElasticPotential::getNegativeValue(const STensor3& F) const {
  double J = F.determinant();
  if (J>=1.) return 0.;
  else return 0.5*_K*log(J)*log(J);
};
void biLogarithmicElasticPotential::constitutivePositive(const STensor3& F, STensor3& P, const bool stiff, STensor43* L) const {
  double J = F.determinant();
  double fact = 1.;
  if (J < 1.) fact = 0.;
  
  STensor3 C, logC;
  for (int i=0; i<3; i++)
    for (int j=0; j<3; j++)
      C(i,j) = F(0,i)*F(0,j)+F(1,i)*F(1,j)+F(2,i)*F(2,j);
  static STensor3 invC;
  STensorOperation::inverseSTensor3(C, invC);

  static STensor43 dlogCdC;
  STensorOperation::logSTensor3(C,_order,logC,&dlogCdC);
  static STensor3 logCdev;
  double trlogC;
  STensorOperation::decomposeDevTr(logC, logCdev, trlogC);

  static STensor3 Svol, Sdev;
  
  for (int i=0; i<3;i++){
    for (int j=0; j<3; j++){
      Svol(i,j) = _K*log(J)*invC(i,j);
      Sdev(i,j) = 0.;
      for (int k=0; k<3; k++){
        Sdev(i,j) += _mu*invC(i,k)*logCdev(k,j);
      }
    };
  }

  for (int i=0; i<3; i++){
    for (int j=0; j<3; j++){
      P(i,j) = F(i,0)*(fact*Svol(0,j)+Sdev(0,j))+F(i,1)*(fact*Svol(1,j)+Sdev(1,j))+F(i,2)*(fact*Svol(2,j)+Sdev(2,j));
    }
  }

  if (stiff){
    static STensor43 dSvoldC, dSdevdC;
    for (int i=0; i<3; i++){
      for (int j=0; j<3; j++){
        for (int k=0; k<3; k++){
          for (int l=0; l<3; l++){
            dSvoldC(i,j,k,l) = 0.5*_K*invC(i,j)*invC(k,l) - 0.5*_K*log(J)*(invC(i,k)*invC(j,l)+invC(i,l)*invC(j,k));
            dSdevdC(i,j,k,l) = 0.;
            for (int p=0; p<3; p++){
              dSdevdC(i,j,k,l) += -_mu*0.5*(invC(i,k)*invC(p,l)+invC(i,l)*invC(p,k))*logCdev(p,j)
                                  +_mu*invC(i,p)*dlogCdC(p,j,k,l)
                                  -_mu*invC(i,j)*dlogCdC(p,p,k,l)/3.;
            }
          };
        };
      };
    };
    

    STensorOperation::zero(*L);
    for (int i=0; i<3; i++){
      for (int j=0; j<3; j++){
        for (int p=0; p<3; p++){
          (*L)(i,j,i,p) += fact*Svol(p,j)+Sdev(p,j);
          for (int q=0; q<3; q++){
            for (int s=0; s<3; s++){
              for (int k=0; k<3; k++){
                (*L)(i,j,p,q) += 2.*F(i,k)*(fact*dSvoldC(k,j,q,s)+dSdevdC(k,j,q,s))*F(p,s);
              }
            }
          }
        }
      }
    }
  }
};
void biLogarithmicElasticPotential::constitutiveNegative(const STensor3& F, STensor3& P, const bool stiff, STensor43* L) const {
  double J = F.determinant();
  if (J< 1.){
    static STensor3 C;
    for (int i=0; i<3; i++)
      for (int j=0; j<3; j++)
        C(i,j) = F(0,i)*F(0,j)+F(1,i)*F(1,j)+F(2,i)*F(2,j);
    static STensor3 invC;
    STensorOperation::inverseSTensor3(C, invC);
    
    static STensor3 Svol;
    Svol=(invC);
    Svol *= (_K*log(J));
    
		for (int i=0; i<3; i++){
			for (int j=0; j<3; j++){
				P(i,j) = F(i,0)*(Svol(0,j))+F(i,1)*(Svol(1,j))+F(i,2)*(Svol(2,j));
      }
    }

    if (stiff){
      static STensor43 dSvoldC;
      for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
          for (int k=0; k<3; k++){
            for (int l=0; l<3; l++){
              dSvoldC(i,j,k,l) = 0.5*_K*invC(i,j)*invC(k,l) - 0.5*_K*log(J)*(invC(i,k)*invC(j,l)+invC(i,l)*invC(j,k));
            };
          };
        };
      };
      

      STensorOperation::zero(*L);
      for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
          for (int p=0; p<3; p++){
            (*L)(i,j,i,p) += Svol(p,j);
            for (int q=0; q<3; q++){
              for (int s=0; s<3; s++){
                for (int k=0; k<3; k++){
                  (*L)(i,j,p,q) += 2.*F(i,k)*dSvoldC(k,j,q,s)*F(p,s);
                }
              }
            }
          }
        }
      }
		}
    
  }
  else{  
    STensorOperation::zero(P);
    if (stiff){
      STensorOperation::zero(*L);
    }
  }
};
void biLogarithmicElasticPotential::getLocalPositiveValForNonLocalEquation(const STensor3& F, double& val, const bool stiff, STensor3* DvalDF) const {
  double Y = getPositiveValue(F);
	double E = getYoungModulus();
  
	val = sqrt(2*Y/E);
	if (stiff){
		(*DvalDF) *= 0.;
		if (_ByPerturbation){
      static STensor3 Fp;
			for (int i=0; i<3; i++){
				for (int j=0; j<3; j++){
          Fp=(F);
					Fp(i,j) += _tol;
					double valp;
					getLocalPositiveValForNonLocalEquation(Fp,valp,false,NULL);
					(*DvalDF)(i,j) = (valp-val)/_tol;
				}
			}
		}
		else{
			this->constitutivePositive(F,*DvalDF,false,NULL);
			if (val > 1e-10){
				(*DvalDF) *= (1./(E*val));
			}
			else{
				(*DvalDF) *= (1./(E*1e-10));
			}

		}
	}
};

NeoHookeanElasticPotential::NeoHookeanElasticPotential(const int num, const double E, const double nu, const bool pert, const double tol) :
	elasticPotential(num), _E(E), _nu(nu), _ByPerturbation(pert),_tol(tol){
  _K = _E/(3.*(1.-2.*_nu));
  _mu = _E/(2.*(1.+_nu));
  double lambda = _K- 2.*_mu/3.;
  static const STensor3 I(1.);
  for (int i=0; i<3; i++){
    for (int j=0; j<3; j++){
      for (int k=0; k<3; k++){
        for (int l=0; l<3; l++){
          _Cel(i,j,k,l) = lambda*I(i,j)*I(k,l)+_mu*(I(i,k)*I(j,l)+I(i,l)*I(j,k));
        }
      }
    }
  }
};

NeoHookeanElasticPotential::NeoHookeanElasticPotential(const NeoHookeanElasticPotential& src) : elasticPotential(src),
	_E(src._E),_nu(src._nu),_K(src._K),_mu(src._mu), _Cel(src._Cel),
	_ByPerturbation(src._ByPerturbation),_tol(src._tol){};

double NeoHookeanElasticPotential::get(const STensor3& F) const{
	double J = F.determinant();
  double J2over3= pow(J,2./3.);
  static STensor3 C;
  for (int i=0; i<3; i++)
    for (int j=0; j<3; j++) {
      C(i,j) = 0.;
      for (int k=0; k<3; k++){
        C(i,j) += F(k,i)*F(k,j);
      }
    }
  C *= (1./J2over3);
  
  double val = 0.5*_K*log(J)*log(J) + 0.5*_mu*(C.trace()-3.);
  return val;
};
void NeoHookeanElasticPotential::constitutive(const STensor3& F, STensor3& P, const bool stiff, STensor43* L) const{
	if (_ByPerturbation){
		double potential = get(F);
    static STensor3 Fp;
		for (int i=0; i<3; i++){
			for (int j=0; j<3; j++){
				Fp=(F);
				Fp(i,j) += _tol;
				double potentialPlus = get(Fp);
				P(i,j) = (potentialPlus - potential)/_tol;
			}
		}

		if (stiff){
			
			static STensor3 Pp;
			for (int i=0; i<3; i++){
				for (int j=0; j<3; j++){
					Fp = F;
					Fp(i,j) += _tol;
					constitutive(Fp,Pp,false,NULL);
					for (int p=0; p<3; p++){
						for (int q=0; q<3; q++){
							(*L)(p,q,i,j) = (Pp(p,q) - P(p,q))/_tol;
						}
					}
				}
			}
		}
	}
  else{
    double J = F.determinant();
    double J2over3= pow(J,2./3.);
    double logJ = log(J);
    static STensor3 Finv;
    STensorOperation::inverseSTensor3(F, Finv);
    
    static STensor3 C;
    for (int i=0; i<3; i++)
      for (int j=0; j<3; j++)
        C(i,j) = (F(0,i)*F(0,j)+F(1,i)*F(1,j)+F(2,i)*F(2,j));
    
    double trC = C.trace();
    static STensor3 invC;
    STensorOperation::inverseSTensor3(C, invC);
        
    static const STensor3 I(1.);
    static STensor3 Svol, Sdev;
    for (int i=0; i<3; i++){
      for (int j=0; j<3; j++){
        Svol(i,j) = _K*logJ*invC(i,j);
        Sdev(i,j) = (_mu/J2over3)*(I(i,j) - trC*invC(i,j)/3.);
      }
    }
    
    
    for (int i=0; i<3; i++){
      for (int j=0; j<3; j++){
        P(i,j) = F(i,0)*(Svol(0,j)+Sdev(0,j))+ F(i,1)*(Svol(1,j)+Sdev(1,j))+F(i,2)*(Svol(2,j)+Sdev(2,j));
      }
    }
    
    if (stiff){
      
      static STensor43 dSvoldC, dSdevdC;
      for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
          for (int k=0; k<3; k++){
            for (int l=0; l<3; l++){
              dSvoldC(i,j,k,l) = 0.5*_K*invC(i,j)*invC(k,l) - 0.5*_K*log(J)*(invC(i,k)*invC(j,l)+invC(i,l)*invC(j,k));
              dSdevdC(i,j,k,l) = -Sdev(i,j)*invC(k,l)/3. - (_mu/J2over3)*invC(i,j)*I(k,l)/3. +
                               (_mu*trC/(6.*J2over3))*(invC(i,k)*invC(j,l)+invC(i,l)*invC(j,k));
              
            };
          };
        };
      };
      
      STensorOperation::zero(*L);
      for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
          for (int p=0; p<3; p++){
            (*L)(i,j,i,p) += (Svol(p,j)+Sdev(p,j));
            for (int q=0; q<3; q++){
              for (int s=0; s<3; s++){
                for (int k=0; k<3; k++){
                  (*L)(i,j,p,q) += 2.*F(i,k)*(dSvoldC(k,j,q,s)+dSdevdC(k,j,q,s))*F(p,s);
                }
              }
            }
          }
        }
      }
    }
  }
};
void NeoHookeanElasticPotential::getLocalValForNonLocalEquation(const STensor3& F, double& val, const bool stiff, STensor3* DvalDF) const{
	double Y = get(F);
	double E = getYoungModulus();

	val = sqrt(2*Y/E);
	if (stiff){
		(*DvalDF) *= 0.;
		if (_ByPerturbation){
      static STensor3 Fp;
			for (int i=0; i<3; i++){
				for (int j=0; j<3; j++){
					Fp=(F);
					Fp(i,j) += _tol;
					double valp;
					getLocalValForNonLocalEquation(Fp,valp,false,NULL);
					(*DvalDF)(i,j) = (valp-val)/_tol;
				}
			}
		}
		else{
			this->constitutive(F,*DvalDF,false,NULL);
			if (val > 1e-10){
				(*DvalDF) *= (1./(E*val));
			}
			else{
				(*DvalDF) *= (1./(E*1e-10));
			}

		}
	}
};


// for anisotropic
double NeoHookeanElasticPotential::getPositiveValue(const STensor3& F) const {
  double J = F.determinant();
  double J2over3= pow(J,2./3.);
  static STensor3 C;
  for (int i=0; i<3; i++)
    for (int j=0; j<3; j++)
      C(i,j) = (F(0,i)*F(0,j)+F(1,i)*F(1,j)+F(2,i)*F(2,j))/J2over3;
  
  double val = 0.;
  if (J >= 1.){
     val = 0.5*_K*log(J)*log(J) + 0.5*_mu*(C.trace()-3.);
  }
  else{
    val  = 0.5*_mu*(C.trace()-3.);
  }
  return val;
};
double NeoHookeanElasticPotential::getNegativeValue(const STensor3& F) const {
  double J = F.determinant();
  double val = 0.;
  if (J <1.){
    val =  0.5*_K*log(J)*log(J);
  }
  return val;
};
void NeoHookeanElasticPotential::constitutivePositive(const STensor3& F, STensor3& P, const bool stiff, STensor43* L) const {
  double J = F.determinant();
  double J2over3= pow(J,2./3.);
  double logJ = log(J);
  static STensor3 C;
  for (int i=0; i<3; i++)
    for (int j=0; j<3; j++)
      C(i,j) = (F(0,i)*F(0,j)+F(1,i)*F(1,j)+F(2,i)*F(2,j));
  
  double trC = C.trace();
  static STensor3 invC;
  STensorOperation::inverseSTensor3(C, invC);
  
  static const STensor3 I(1.);
  static STensor3 Spos, Sdev;
  
  for (int i=0; i<3; i++){
    for (int j=0; j<3; j++){
      Sdev(i,j) = (_mu/J2over3)*(I(i,j) - trC*invC(i,j)/3.);
      Spos(i,j) = Sdev(i,j);
      if (J>=1.){
        Spos(i,j) += _K*logJ*invC(i,j);
      }
    }
  }
  
  
  for (int i=0; i<3; i++){
    for (int j=0; j<3; j++){
      P(i,j) = F(i,0)*Spos(0,j)+ F(i,1)*Spos(1,j)+F(i,2)*Spos(2,j);
    }
  }
  
  if (stiff){
    static STensor43 dSposdC;
    for (int i=0; i<3; i++){
      for (int j=0; j<3; j++){
        for (int k=0; k<3; k++){
          for (int l=0; l<3; l++){
            dSposdC(i,j,k,l) = -Sdev(i,j)*invC(k,l)/3. - (_mu/J2over3)*invC(i,j)*I(k,l)/3. +
                               (_mu*trC/(6.*J2over3))*(invC(i,k)*invC(j,l)+invC(i,l)*invC(j,k));
            if (J >=1.){
              dSposdC(i,j,k,l) += 0.5*_K*invC(i,j)*invC(k,l) - 0.5*_K*log(J)*(invC(i,k)*invC(j,l)+invC(i,l)*invC(j,k));
            }
            
            
          };
        };
      };
    };
    
    STensorOperation::zero(*L);
    for (int i=0; i<3; i++){
      for (int j=0; j<3; j++){
        for (int p=0; p<3; p++){
          (*L)(i,j,i,p) += Spos(p,j);
          for (int q=0; q<3; q++){
            for (int s=0; s<3; s++){
              for (int k=0; k<3; k++){
                (*L)(i,j,p,q) += 2.*F(i,k)*dSposdC(k,j,q,s)*F(p,s);
              }
            }
          }
        }
      }
    }
  }
};
void NeoHookeanElasticPotential::constitutiveNegative(const STensor3& F, STensor3& P, const bool stiff, STensor43* L) const {
  double J = F.determinant();
  if (J < 1.){
    double logJ = log(J);
    static STensor3 C;
    
    for (int i=0; i<3; i++)
      for (int j=0; j<3; j++)
        C(i,j) = (F(0,i)*F(0,j)+F(1,i)*F(1,j)+F(2,i)*F(2,j));
    
    static STensor3 invC;
    STensorOperation::inverseSTensor3(C, invC);
    static STensor3 Svol;
    Svol=(invC);
    Svol*= (_K*logJ);
    
    
    for (int i=0; i<3; i++){
      for (int j=0; j<3; j++){
        P(i,j) = F(i,0)*Svol(0,j)+ F(i,1)*Svol(1,j)+F(i,2)*Svol(2,j);
      }
    }
    
    if (stiff){
      static STensor43 dSvoldC;
    
      for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
          for (int k=0; k<3; k++){
            for (int l=0; l<3; l++){
              dSvoldC(i,j,k,l) = 0.5*_K*invC(i,j)*invC(k,l)- 0.5*_K*log(J)*(invC(i,k)*invC(j,l)+invC(i,l)*invC(j,k));
            };
          };
        };
      };
      
      STensorOperation::zero(*L);
      for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
          for (int p=0; p<3; p++){
            (*L)(i,j,i,p) += (Svol(p,j));
            for (int q=0; q<3; q++){
              for (int s=0; s<3; s++){
                for (int k=0; k<3; k++){
                  (*L)(i,j,p,q) += 2.*F(i,k)*dSvoldC(k,j,q,s)*F(p,s);
                }
              }
            }
          }
        }
      }
		}
  }
  else{
    STensorOperation::zero(P);
    if (stiff){
      STensorOperation::zero(*L);
    }
  }
};
void NeoHookeanElasticPotential::getLocalPositiveValForNonLocalEquation(const STensor3& F, double& val, const bool stiff, STensor3* DvalDF) const {
  double Y = getPositiveValue(F);
	double E = getYoungModulus();
  
	val = sqrt(2*Y/E);
	if (stiff){
		(*DvalDF) *= 0.;
		if (_ByPerturbation){
      static STensor3 Fp;
			for (int i=0; i<3; i++){
				for (int j=0; j<3; j++){
					Fp=(F);
					Fp(i,j) += _tol;
					double valp;
					getLocalPositiveValForNonLocalEquation(Fp,valp,false,NULL);
					(*DvalDF)(i,j) = (valp-val)/_tol;
				}
			}
		}
		else{
			this->constitutivePositive(F,*DvalDF,false,NULL);
			if (val > 1e-10){
				(*DvalDF) *= (1./(E*val));
			}
			else{
				(*DvalDF) *= (1./(E*1e-10));
			}

		}
	}
};
