//
// C++ Interface: material law
//
// Author:  <Van Dung Nguyen>, (C) 2017
//
// Copyright: See COPYING file that comes with this distribution
//
//
#ifndef ELASTICPOTENTIAL_H_
#define ELASTICPOTENTIAL_H_

#include "mlaw.h"
#include "STensor3.h"
#include "STensor43.h"
#include "STensorOperations.h"

class elasticPotential{
	#ifndef SWIG
	protected:
		int _num;
		
	public:
		elasticPotential(const int num):_num(num){};
		elasticPotential(const elasticPotential& src): _num(src._num){}
		virtual ~elasticPotential(){};
    virtual const STensor43& getElasticTensor() const = 0;
		virtual double getYoungModulus() const = 0;
		virtual double getPoissonRatio() const = 0;
		virtual double get(const STensor3& F) const  = 0; // elastic potential
		virtual void constitutive(const STensor3& F, STensor3& P, const bool sitff, STensor43* L) const = 0; // constitutive law
    
		virtual void getLocalValForNonLocalEquation(const STensor3& F, double& val, const bool stiff, STensor3* DvalDF) const = 0;
		virtual elasticPotential* clone() const = 0;
    // for anisotropic
    virtual double getPositiveValue(const STensor3& F) const = 0;
    virtual double getNegativeValue(const STensor3& F) const = 0;
    virtual void constitutivePositive(const STensor3& F, STensor3& P, const bool sitff, STensor43* L) const = 0;
    virtual void constitutiveNegative(const STensor3& F, STensor3& P, const bool sitff, STensor43* L) const = 0;
    virtual void getLocalPositiveValForNonLocalEquation(const STensor3& F, double& val, const bool stiff, STensor3* DvalDF) const = 0;
	#endif //SWIG
};

class smallStrainLinearElasticPotential : public elasticPotential{
	protected:
		double _K, _mu, _E, _nu; // elastic constant
		STensor43 _Cel;
		
	public:
		smallStrainLinearElasticPotential(const int num, const double E, const double nu);
		#ifndef SWIG
		smallStrainLinearElasticPotential(const smallStrainLinearElasticPotential& src);
		virtual ~smallStrainLinearElasticPotential(){};
    virtual const STensor43& getElasticTensor() const {return _Cel;};
		virtual double getYoungModulus() const {return _E;};
		virtual double getPoissonRatio() const {return _nu;};
		virtual double get(const STensor3& F) const; // elastic potential
		virtual void constitutive(const STensor3& F, STensor3& P, const bool sitff, STensor43* L) const; // constitutive law
		virtual void getLocalValForNonLocalEquation(const STensor3& F, double& val, const bool stiff, STensor3* DvalDF) const;
		virtual elasticPotential* clone() const {
			return new smallStrainLinearElasticPotential(*this);
		};
    // for anisotropic
    virtual double getPositiveValue(const STensor3& F) const;
    virtual double getNegativeValue(const STensor3& F) const;
    virtual void constitutivePositive(const STensor3& F, STensor3& P, const bool sitff, STensor43* L) const;
    virtual void constitutiveNegative(const STensor3& F, STensor3& P, const bool sitff, STensor43* L) const;
    virtual void getLocalPositiveValForNonLocalEquation(const STensor3& F, double& val, const bool stiff, STensor3* DvalDF) const;
		#endif //SWIG
};

class largeStrainAnisotropicPotential : public elasticPotential{
   // eps*H*eps
  protected:
    STensor43 _Cel;
    double _Emax;
    double _poissonMax;
    
  public:
    largeStrainAnisotropicPotential(const int num, const double Ex, const double Ey, const double Ez, 
                                    const double Vxy, const double Vxz, const double Vyz,
                                    const double MUxy, const double MUxz, const double MUyz,
                                    const double alpha, const double beta, const double gamma);
    largeStrainAnisotropicPotential(const int num, const STensor43& C);
		#ifndef SWIG
		largeStrainAnisotropicPotential(const largeStrainAnisotropicPotential& src);
		virtual ~largeStrainAnisotropicPotential(){};
    virtual const STensor43& getElasticTensor() const {return _Cel;};
		virtual double getYoungModulus() const {return _Emax;};
		virtual double getPoissonRatio() const {return _poissonMax;};
		virtual double get(const STensor3& F) const; // elastic potential
		virtual void constitutive(const STensor3& F, STensor3& P, const bool sitff, STensor43* L) const; // constitutive law
		virtual void getLocalValForNonLocalEquation(const STensor3& F, double& val, const bool stiff, STensor3* DvalDF) const;
		virtual elasticPotential* clone() const {
			return new largeStrainAnisotropicPotential(*this);
		};
    // for anisotropic
    virtual double getPositiveValue(const STensor3& F) const;
    virtual double getNegativeValue(const STensor3& F) const;
    virtual void constitutivePositive(const STensor3& F, STensor3& P, const bool sitff, STensor43* L) const;
    virtual void constitutiveNegative(const STensor3& F, STensor3& P, const bool sitff, STensor43* L) const;
    virtual void getLocalPositiveValForNonLocalEquation(const STensor3& F, double& val, const bool stiff, STensor3* DvalDF) const;
		#endif //SWIG
};

class largeStrainTransverseIsotropicPotential : public elasticPotential{
   // eps*H*eps
  protected:
    STensor43 _Cel;
    double _Emax;
    double _poissonMax;
    
  public:
    largeStrainTransverseIsotropicPotential(const int num, 
                                    const double ETran, // inplane modulus
                                    const double ELong,  // longitudinal modulus 
                                    const double nuTT, // inplane Poisson ratio
                                    const double nuLT,  // inplane-longitudinal Possion ratio
                                    const double MUTT,  // inplane shear
                                    const double MUTL, // out-plane shear
                                    const double alpha, const double beta, const double gamma); 
		#ifndef SWIG
		largeStrainTransverseIsotropicPotential(const largeStrainTransverseIsotropicPotential& src);
		virtual ~largeStrainTransverseIsotropicPotential(){};
    virtual const STensor43& getElasticTensor() const {return _Cel;};
		virtual double getYoungModulus() const {return _Emax;};
		virtual double getPoissonRatio() const {return _poissonMax;};
		virtual double get(const STensor3& F) const; // elastic potential
		virtual void constitutive(const STensor3& F, STensor3& P, const bool sitff, STensor43* L) const; // constitutive law
		virtual void getLocalValForNonLocalEquation(const STensor3& F, double& val, const bool stiff, STensor3* DvalDF) const;
		virtual elasticPotential* clone() const {
			return new largeStrainTransverseIsotropicPotential(*this);
		};
    // for anisotropic
    virtual double getPositiveValue(const STensor3& F) const;
    virtual double getNegativeValue(const STensor3& F) const;
    virtual void constitutivePositive(const STensor3& F, STensor3& P, const bool sitff, STensor43* L) const;
    virtual void constitutiveNegative(const STensor3& F, STensor3& P, const bool sitff, STensor43* L) const;
    virtual void getLocalPositiveValForNonLocalEquation(const STensor3& F, double& val, const bool stiff, STensor3* DvalDF) const;
		#endif //SWIG
};


class biLogarithmicElasticPotential : public elasticPotential{
	protected:
		double _K, _mu, _E, _nu; // elastic constant
		int _order; // for log approx
		STensor43 _Cel;
		bool _ByPerturbation;
		double _tol;
	
	public:
		biLogarithmicElasticPotential(const int num, const double E, const double nu, const int order, const bool pert = false, const double tol =1e-8);
		#ifndef SWIG
		biLogarithmicElasticPotential(const biLogarithmicElasticPotential& src);
		virtual ~biLogarithmicElasticPotential(){}
    virtual const STensor43& getElasticTensor() const {return _Cel;};
		virtual double getYoungModulus() const {return _E;};
		virtual double getPoissonRatio() const {return _nu;};
		virtual double get(const STensor3& F) const; // elastic potential
		virtual void constitutive(const STensor3& F, STensor3& P, const bool sitff, STensor43* L) const; // constitutive law
		virtual void getLocalValForNonLocalEquation(const STensor3& F, double& val, const bool stiff, STensor3* DvalDF) const;
		virtual elasticPotential* clone() const {
			return new biLogarithmicElasticPotential(*this);
		}
    // for anisotropic
    virtual double getPositiveValue(const STensor3& F) const;
    virtual double getNegativeValue(const STensor3& F) const;
    virtual void constitutivePositive(const STensor3& F, STensor3& P, const bool sitff, STensor43* L) const;
    virtual void constitutiveNegative(const STensor3& F, STensor3& P, const bool sitff, STensor43* L) const;
    virtual void getLocalPositiveValForNonLocalEquation(const STensor3& F, double& val, const bool stiff, STensor3* DvalDF) const;
		#endif //SWIG
};

class NeoHookeanElasticPotential : public elasticPotential{
  protected:
    #ifndef SWIG
    double _K, _mu, _E, _nu;
    STensor43 _Cel;
		bool _ByPerturbation;
		double _tol;
    #endif //SWIG
  public:
    NeoHookeanElasticPotential(const int num, const double E, const double nu, const bool pert = false, const double tol =1e-8);
		#ifndef SWIG
		NeoHookeanElasticPotential(const NeoHookeanElasticPotential& src);
		virtual ~NeoHookeanElasticPotential(){}	
    virtual const STensor43& getElasticTensor() const {return _Cel;};
		virtual double getYoungModulus() const {return _E;};
		virtual double getPoissonRatio() const {return _nu;};
		virtual double get(const STensor3& F) const; // elastic potential
		virtual void constitutive(const STensor3& F, STensor3& P, const bool sitff, STensor43* L) const; // constitutive law
		virtual void getLocalValForNonLocalEquation(const STensor3& F, double& val, const bool stiff, STensor3* DvalDF) const;
		virtual elasticPotential* clone() const {
			return new NeoHookeanElasticPotential(*this);
		}
    // for anisotropic
    virtual double getPositiveValue(const STensor3& F) const;
    virtual double getNegativeValue(const STensor3& F) const;
    virtual void constitutivePositive(const STensor3& F, STensor3& P, const bool sitff, STensor43* L) const;
    virtual void constitutiveNegative(const STensor3& F, STensor3& P, const bool sitff, STensor43* L) const;
    virtual void getLocalPositiveValForNonLocalEquation(const STensor3& F, double& val, const bool stiff, STensor3* DvalDF) const;
    #endif //SWIG
};

#endif //ELASTICPOTENTIAL_H_