//
// C++ Interface: material law
//
// Description: clustering material law
//
//
// Author:  <Kevin SPILKER>, (C) 2019
//
// Copyright: See COPYING file that comes with this distribution
//
//
#include <math.h>
#include "MInterfaceElement.h"
#include <fstream>
#include "mlawClusterMaterialLaws.h"
#include "STensorOperations.h"
#include "FiniteStrain.h"
#include "matrix_operations.h"


mlawClusterMaterialLaws::mlawClusterMaterialLaws(const int num, const char *clusterName) : materialLaw(num,true)
{
  std::ifstream in(clusterName);
  if(!in) Msg::Error("Cannot open the file %s! Maybe is missing   ",clusterName);
  int elementnb, clusternb;
  //in >> elementnb; in>> clusternb; //to be corrected
  in >> elementnb >> clusternb;
  in.ignore(256,'\n');
  _clusternb.insert(std::pair<int,int>(elementnb,clusternb));

}
mlawClusterMaterialLaws::mlawClusterMaterialLaws(const mlawClusterMaterialLaws &source) :
                                        materialLaw(source)
{
  _clusternb=source._clusternb; //give the cluster number for element i
    
  _matlawnb=source._matlawnb; //give the material law number for cluster i
  _matlawptr=source._matlawptr; //give the material law pointer for cluster i
  //for(std::map<int,IP3State*>::const_iterator it=source._ipstates.begin(); it!=source._ipstates.end(); it++)
  //{
  //  int clusternb=it->first;
  //  IP3State *ips= new IP3State(*(it->second));
  //  _ipstates.insert(std::pair<int,IP3State*>(clusternb,ips));
  //}

  _eigen=source._eigen; //give the eigen for cluster i
}

mlawClusterMaterialLaws& mlawClusterMaterialLaws::operator=(const materialLaw &source)
{
  materialLaw::operator=(source);
  const mlawClusterMaterialLaws* src =static_cast<const mlawClusterMaterialLaws*>(&source);
  if(src!=NULL)
  {
    _clusternb=src->_clusternb; //give the cluster number for element i
    
    _matlawnb=src->_matlawnb; //give the material law number for cluster i
    _matlawptr=src->_matlawptr; //give the material law pointer for cluster i
    //for(std::map<int,IP3State*>::const_iterator it=src->_ipstates.begin(); it!=src->_ipstates.end(); it++)
    //{
    //  int clusternb=it->first;
    //  IP3State *ips= new IP3State(*(it->second));
    //  _ipstates.insert(std::pair<int,IP3State*>(clusternb,ips));
    //}
    _eigen=src->_eigen;
  }
  return *this;
}

mlawClusterMaterialLaws::~mlawClusterMaterialLaws()
{
  //for(std::map<int,IP3State*>::iterator it=_ipstates.begin(); it!=_ipstates.end(); it++)
  //{
  //  delete it->second;
  //}
  //_ipstates.clear();
}


void mlawClusterMaterialLaws::addClusterMaterialLaw(int clusterNb, int materialLawNb)
{
  _matlawnb.insert(std::pair<int,int>(clusterNb,materialLawNb));
}
void mlawClusterMaterialLaws::addClusterEigenStrain(int clusterNb, const STensor3 &Feig) //double F11,double F12,double F13,double F21,double F22,double F23,double F21,double F32,double F33)
{
  _eigen.insert(std::pair<int,STensor3>(clusterNb,Feig));

}

void mlawClusterMaterialLaws::initLaws(const std::map<int,materialLaw*> &maplaw)
{
  for(std::map<int,int>::const_iterator itcluster=_matlawnb.begin(); itcluster!=_matlawnb.end(); itcluster++)
  {
    int clusternb=itcluster->first;
    int matlawnb=itcluster->second;
    

    bool findlaw=false;
    for(std::map<int,materialLaw*>::const_iterator it = maplaw.begin(); it != maplaw.end(); ++it)
    {
      int num = it->first;
      if(num == matlawnb){
        findlaw=true;
        it->second->initLaws(maplaw);
        _matlawptr.insert(std::pair<int, const materialLaw*>(clusternb,it->second->getConstNonLinearSolverMaterialLaw()));
        break;
      }
    }
    if(!findlaw) Msg::Error("The law is not initialize for cluster number %d and material law number %d",clusternb,matlawnb);
  } 
} 

void mlawClusterMaterialLaws::constitutive(
            const STensor3& F0,         // initial deformation gradient (input @ time n)
            const STensor3& Fn,         // updated deformation gradient (input @ time n+1)
            STensor3 &P,                // updated 1st Piola-Kirchhoff stress tensor (output)
            const IPVariable *q0i,       // array of initial internal variable
            IPVariable *q1i,             // updated array of internal variable (in ipvcur on output),
            STensor43 &Tangent,         // constitutive tangents (output)
            const bool stiff,         // if true compute the tangents
           STensor43* elasticTangent) const
{


  const IPCluster *q0 = dynamic_cast<const IPCluster *> (q0i);
  IPCluster *q1 = dynamic_cast<IPCluster *> (q1i);
  int clusterNb=q1->getClusterNb();
  std::map<int,const materialLaw *>::const_iterator itmlaw=_matlawptr.find(clusterNb);
  
  const IPVariable *ipv0 = q0->getConstRefToIPv();
  IPVariable *ipv1 = q1->getRefToIPv();
  STensor3 Feig(Fn);
  //modify it to put eigen strain
  STensor3 eig;
  STensorOperation::unity(eig);
  bool find=false;
  for(std::map<int,STensor3>::const_iterator it = _eigen.begin(); it != _eigen.end(); ++it)
  {
    int num = it->first;
    if(num == clusterNb){
      find=true;
      eig=it->second;
      break;
    }
  }
  Feig*=eig;  
  itmlaw->second->constitutive(F0,Feig,P,ipv0,ipv1,Tangent,stiff,elasticTangent);   

}

void mlawClusterMaterialLaws::createIPState(IPCluster *ivi, IPCluster *iv1, IPCluster *iv2) const
{
  Msg::Error("mlawClusterMaterialLaws::createIPState(IPCluster *ivi, IPCluster *iv1, IPCluster *iv2 cannot be used without Ele");
}
void mlawClusterMaterialLaws::createIPVariable(IPCluster *&ipv,const MElement *ele,const int nbFF, const IntPt *GP, const int gpt) const
{

  //find the cluster nb
  int eleNb=ele->getNum();
  std::map<int,int>::const_iterator itcluster=_clusternb.find(eleNb);
  int clusterNb=itcluster->second;

  std::map<int,const materialLaw *>::const_iterator itmlaw=_matlawptr.find(clusterNb);
  //IPVariable* iplawi=NULL;
  //itmlaw->second->createIPVariable(iplawi, ele, nbFF, GP, gpt);
  // this would require to define   void createIPVariable(IPVariable *&ipv,const MElement *ele,const int nbFF, const IntPt *GP, const int gpt) const=0 in mlaw.
  // Here we do a workaround
  IPStateBase* ipslaw=NULL;
  const bool state_=true;
  itmlaw->second->createIPState(ipslaw, &state_,ele, nbFF, GP, gpt);
  std::vector<IPVariable*> allIP;
  ipslaw->getAllIPVariable(allIP);

  IPVariable* iplawi=allIP[0]->clone();
  ipv=new IPCluster(clusterNb,iplawi);
  delete ipslaw;
}

void mlawClusterMaterialLaws::createIPState(IPStateBase* &ips,const bool* state_,const MElement *ele, const int nbFF_, const IntPt *GP, const int gpt) const
{
  //find the cluster nb
  int eleNb=ele->getNum();
  std::map<int,int>::const_iterator itcluster=_clusternb.find(eleNb);
  int clusterNb=itcluster->second;
  std::map<int,const materialLaw *>::const_iterator itmlaw=_matlawptr.find(clusterNb);

  //IPVariable* iplawi=NULL;
  //itmlaw->second->createIPVariable(iplawi, ele, nbFF_, GP, gpt);
  // this would require to define   void createIPVariable(IPVariable *&ipv,const MElement *ele,const int nbFF, const IntPt *GP, const int gpt) const=0 in mlaw.
  // Here we do a workaround
  IPStateBase* ipslaw=NULL;
  itmlaw->second->createIPState(ipslaw, state_,ele, nbFF_, GP, gpt);
  std::vector<IPVariable*> allIP;
  ipslaw->getAllIPVariable(allIP);

  IPVariable* iplawi=allIP[0]->clone();
  IPVariable* ipvi = new IPCluster(clusterNb,iplawi);

  IPVariable* iplaw1=allIP[1]->clone();
  IPVariable* ipv1 = new IPCluster(clusterNb,iplaw1);

  IPVariable* iplaw2=allIP[2]->clone();
  IPVariable* ipv2 = new IPCluster(clusterNb,iplaw2);
  if(ips != NULL) delete ips;
  ips = new IP3State(state_,ipvi,ipv1,ipv2);
  delete ipslaw;
 }
