//
// C++ Interface: material law
//
// Author:  <Van Dung Nguyen>, (C) 2014
//
// Copyright: See COPYING file that comes with this distribution
//
//
#ifndef MLAWHYPERELASTIC_H_
#define MLAWHYPERELASTIC_H_

#include "mlaw.h"
#include "STensor3.h"
#include "STensor43.h"
#include "STensor63.h"
#include "ipHyperelastic.h"
#include "viscosityLaw.h"
#include "mlawHyperelasticFailureCriterion.h"


class mlawHyperViscoElastic : public materialLaw{
  #ifndef SWIG
  protected:
    double _rho; // density
    double _E; // young modulus
    double _nu; // Poisson ratio
    double _lambda; // 1st lame parameter
    double _mu; // 2nd lame parameter (=G)
    double _K; // bulk modulus

    STensor43 Cel; // elastic hook tensor
    int _order ; // to approximated log and exp of a tensor

    double _perturbationfactor; // perturbation factor
    bool _tangentByPerturbation; // flag for tangent by perturbation

     // viscoelastic data
    int _viscoMethod; // viscoelastic method 0- generalized Maxwell model, 1- generalized Kelvin-Voight model
    int _N; // number of elements
    std::vector<double> _Ki;
    std::vector<double> _ki;
    std::vector<double> _Gi;
    std::vector<double> _gi;

  protected:
    virtual double deformationEnergy(const STensor3 &C) const ;
    void updateViscoElasticFlow(const IPHyperViscoElastic *q0, IPHyperViscoElastic *q1, double& Ke, double& Ge) const;
    void viscoElasticPredictor(const STensor3& Ee, const STensor3& Ee0,
              const IPHyperViscoElastic *q0, IPHyperViscoElastic *q1,
              double& Ke, double& Ge) const;

    void isotropicHookTensor(const double K, const double G, STensor43& L) const;

    void predictorCorrector_ViscoElastic(const STensor3& F0, const STensor3& F, STensor3&P, const IPHyperViscoElastic *q0_, IPHyperViscoElastic *q1,
                                  const bool stiff, STensor43& Tangent) const;

  #endif // SWIG

  public:
    mlawHyperViscoElastic(const int num,const double E,const double nu, const double rho,
                        const bool matrixbyPerturbation = false, const double pert = 1e-8);
    virtual void setStrainOrder(const int order);
    virtual void setViscoelasticMethod(const int method);
    virtual void setViscoElasticNumberOfElement(const int N);
    virtual void setViscoElasticData(const int i, const double Ei, const double taui);
    virtual void setViscoElasticData_Bulk(const int i, const double Ki, const double ki);
    virtual void setViscoElasticData_Shear(const int i, const double Gi, const double gi);
    virtual void setViscoElasticData(const std::string filename);

    #ifndef SWIG
    mlawHyperViscoElastic(const mlawHyperViscoElastic& src);
    virtual ~mlawHyperViscoElastic(){};

    virtual matname getType() const{return materialLaw::hyperelastic;}
    virtual void createIPState(IPStateBase* &ips,const bool* state_=NULL,const MElement *ele=NULL, const int nbFF_=0, const IntPt *GP=NULL, const int gpt = 0) const{
        Msg::Error("mlawHyperViscoElastic::createIPState has not been defined");
    };
    virtual void initLaws(const std::map<int,materialLaw*> &maplaw){}; // this law is initialized so nothing to do

    virtual bool withEnergyDissipation() const {if (_N == 0) return false; else return true;};
    virtual int getViscoElasticNumberOfElement() const{return _N;};

    virtual double soundSpeed() const; // default but you can redefine it for your case
    virtual double density()const{return _rho;}
    virtual const double bulkModulus() const{return _K;}
    virtual const double shearModulus() const{return _mu;}
    virtual const double poissonRatio() const{return _nu;}

    virtual double scaleFactor() const { return _mu;};

    virtual void constitutive(
            const STensor3& F0,         // initial deformation gradient (input @ time n)
            const STensor3& Fn,         // updated deformation gradient (input @ time n+1)
            STensor3 &P,                // updated 1st Piola-Kirchhoff stress tensor (output)
            const IPVariable *q0,       // array of initial internal variable
            IPVariable *q1,             // updated array of internal variable (in ipvcur on output),
            STensor43 &Tangent,         // constitutive tangents (output)
            const bool stiff,          // if true compute the tangents
           STensor43* elasticTangent = NULL) const;


    virtual materialLaw* clone() const {return new mlawHyperViscoElastic(*this);};
    virtual void checkInternalState(IPVariable* ipv, const IPVariable* ipvprev) const{}; // do nothing
    #endif // SWIG
};

// general interface for yield surface f = a2 sigVM^n - (a1*p+a0)
// non-associated flow wit f = sigVM^2 + b* p^2
// associated flow
class mlawPowerYieldHyper : public mlawHyperViscoElastic{
  protected:
    double _tol; // for plasticity convergence
    STensor43 _I4; // symetric 4th order unit tensor
    STensor3  _I; // second order tensor
    STensor43 _Idev; // dev part of _I4

    // rate effects
    double _p; // exponent in rate
    viscosityLaw* _viscosity;

    // hardening data
    J2IsotropicHardening* _compression;
    J2IsotropicHardening* _traction;
    J2IsotropicHardening* _shear;
    kinematicHardening* _kinematic;

    // yield surface option
    double _n;
    double _b; // flow coefficient g = sigVM^2+ bp^2
    bool _nonAssociatedFlow;

  protected:

    virtual void hardening(const IPHyperViscoElastoPlastic* q0, IPHyperViscoElastoPlastic* q) const;

     virtual void updateEqPlasticDeformation(IPHyperViscoElastoPlastic *q1, const IPHyperViscoElastoPlastic *q0,
                                            const double& nup, const double& Dgamma) const;

    virtual void getYieldCoefficients(const IPHyperViscoElastoPlastic *q, fullVector<double>& coeffs) const;
    virtual void getYieldCoefficientDerivatives(const IPHyperViscoElastoPlastic *q, const double& nup, fullVector<double>& Dcoeffs) const;


  #ifndef SWIG
  protected:

    virtual void predictorCorrector_nonAssociatedFlow(const STensor3& F, const IPHyperViscoElastoPlastic *q0_, IPHyperViscoElastoPlastic *q1,
                            STensor3&P, const bool stiff, STensor43& Tangent,  STensor43& dFedF, STensor43& dFpdF) const;

    virtual void predictorCorrector_associatedFlow(const STensor3& F, const IPHyperViscoElastoPlastic *q0_, IPHyperViscoElastoPlastic *q1,
                            STensor3&P, const bool stiff, STensor43& Tangent, STensor43& dFedF, STensor43& dFpdF) const;

  protected:


    virtual void predictorCorrector(const STensor3& F, const IPHyperViscoElastoPlastic *q0_, IPHyperViscoElastoPlastic *q1,
                            STensor3&P, const bool stiff, STensor43& Tangent, STensor43& dFedF, STensor43& dFpdF,
                            STensor43* elasticTangent) const;
    virtual void tangent_full_perturbation( // compute tangent by perturbation
                             STensor43 &T, // current tangent
                             STensor43& dFedF,
                             STensor43& dFpdF,
                             const STensor3 &P, // current PK stress
                             const STensor3 &F, // current deformation gradient
                             const IPHyperViscoElastoPlastic* q0, // previous internal variables
                             IPHyperViscoElastoPlastic* q1 // current internal variables
                           ) const;

  #endif // SWIG
  public:
    mlawPowerYieldHyper(const int num,const double E,const double nu, const double rho,
                        const double tol = 1.e-6,
                        const bool matrixbyPerturbation = false, const double pert = 1e-8);

    void setPowerFactor(const double n);
    void nonAssociatedFlowRuleFactor(const double b);
    void setPlasticPoissonRatio(const double nup);
    void setNonAssociatedFlow(const bool flag);

    void setCompressionHardening(const J2IsotropicHardening& comp);
    void setTractionHardening(const J2IsotropicHardening& trac);
    void setShearHardening(const J2IsotropicHardening& shear);
    void setKinematicHardening(const kinematicHardening& kin);
    void setViscosityEffect(const viscosityLaw& v, const double p);

    #ifndef SWIG
    mlawPowerYieldHyper(const mlawPowerYieldHyper& src);
    virtual ~mlawPowerYieldHyper();

    virtual materialLaw* clone() const{return new mlawPowerYieldHyper(*this);};
    virtual bool withEnergyDissipation() const {return true;};

    virtual J2IsotropicHardening* getConstRefToCompressionHardening() const {return _compression;};
    virtual J2IsotropicHardening* getConstRefToTractionHardening() const{return _traction;};
    virtual J2IsotropicHardening* getConstRefToShearHardening() const {return _shear;};
    virtual kinematicHardening* getConstRefToKinematicHardening() const{return _kinematic;};


    virtual matname getType() const{return materialLaw::powerYieldLaw;}
    virtual void createIPState(IPStateBase* &ips,const bool* state_=NULL,const MElement *ele=NULL, const int nbFF_=0, const IntPt *GP=NULL, const int gpt = 0) const;
    virtual void initLaws(const std::map<int,materialLaw*> &maplaw){}; // this law is initialized so nothing to do

    virtual void constitutive(
            const STensor3& F0,         // initial deformation gradient (input @ time n)
            const STensor3& Fn,         // updated deformation gradient (input @ time n+1)
            STensor3 &P,                // updated 1st Piola-Kirchhoff stress tensor (output)
            const IPVariable *q0,       // array of initial internal variable
            IPVariable *q1,             // updated array of internal variable (in ipvcur on output),
            STensor43 &Tangent,         // constitutive tangents (output)
            const bool stiff,          // if true compute the tangents
           STensor43* elasticTangent = NULL) const;

    virtual void checkInternalState(IPVariable* ipv, const IPVariable* ipvprev) const{}; // do nothing
    #endif // SWIG
};

class mlawPowerYieldHyperWithFailure : public mlawPowerYieldHyper{
  protected:
    #ifndef SWIG
    mlawHyperelasticFailureCriterionBase* _failureCriterion;
  protected:
    virtual void predictorCorrector(const STensor3& F, const IPHyperViscoElastoPlastic *q0_, IPHyperViscoElastoPlastic *q1,
                            STensor3&P, const bool stiff, STensor43& Tangent, STensor43& dFedF, STensor43& dFpdF,
                            STensor43* elasticTangent) const;
    #endif // SWIG
  public:
    mlawPowerYieldHyperWithFailure(const int num,const double E,const double nu, const double rho,
                        const double tol = 1.e-6,
                        const bool matrixbyPerturbation = false, const double pert = 1e-8);


    void setFailureCriterion(const mlawHyperelasticFailureCriterionBase& fCr);

  #ifndef SWIG
    mlawPowerYieldHyperWithFailure(const mlawPowerYieldHyperWithFailure& src);
    virtual ~mlawPowerYieldHyperWithFailure();

    virtual materialLaw* clone() const{return new mlawPowerYieldHyperWithFailure(*this);};
    virtual void checkInternalState(IPVariable* ipv, const IPVariable* ipvprev) const;
    virtual matname getType() const{return materialLaw::powerYieldLawWithFailure;}
    virtual void initLaws(const std::map<int,materialLaw*> &maplaw){}; // this law is initialized so nothing to do

  #endif // SWIG

};
#endif // MLAWHYPERELASTIC_H_
