//
// C++ Interface: J2 material in small strains
//
//
// Author:  <Van Dung NGUYEN>, (C) 2016
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include "mlawJ2VMSmallStrain.h"


mlawJ2VMSmallStrain::mlawJ2VMSmallStrain(const int num, const double E, const double nu,const double rho,
                               const double sy0,const double h,const double tol,
                               const bool matrixbyPerturbation, const double pert, const bool init):
                               materialLaw(num,init),_tol(tol),_perturbationfactor(pert), _rho(rho),
                               _tangentByPerturbation(matrixbyPerturbation), _E(E),_nu(nu){
  _j2IH = new LinearExponentialJ2IsotropicHardening(num, sy0, h, 0., 10.);
  _K = E/(3.*(1.-2.*nu));
  _mu = E/(2.*(1.+nu));
  _lambda = (E*nu)/(1.+nu)/(1.-2.*nu);

  STensor3 I(1.);
  for (int i=0; i<3; i++){
    for (int j=0; j<3; j++){
      for (int k=0; k<3; k++){
        for (int l=0; l<3; l++){
          _I4(i,j,k,l) = 0.5*(I(i,k)*I(j,l)+I(i,l)*I(j,k));
          _I4dev(i,j,k,l) = 0.5*(I(i,k)*I(j,l)+I(i,l)*I(j,k)) - I(i,j)*I(k,l)/3.;
        }
      }
    }
  }


  _Cel*=0.;
  _Cel(0,0,0,0) = _lambda + 2.*_mu;
  _Cel(1,1,0,0) = _lambda;
  _Cel(2,2,0,0) = _lambda;
  _Cel(0,0,1,1) = _lambda;
  _Cel(1,1,1,1) = _lambda + 2.*_mu;
  _Cel(2,2,1,1) = _lambda;
  _Cel(0,0,2,2) = _lambda;
  _Cel(1,1,2,2) = _lambda;
  _Cel(2,2,2,2) = _lambda + 2.*_mu;

  _Cel(1,0,1,0) = _mu;
  _Cel(2,0,2,0) = _mu;
  _Cel(0,1,0,1) = _mu;
  _Cel(2,1,2,1) = _mu;
  _Cel(0,2,0,2) = _mu;
  _Cel(1,2,1,2) = _mu;

  _Cel(0,1,1,0) = _mu;
  _Cel(0,2,2,0) = _mu;
  _Cel(1,0,0,1) = _mu;
  _Cel(1,2,2,1) = _mu;
  _Cel(2,0,0,2) = _mu;
  _Cel(2,1,1,2) = _mu;
};

mlawJ2VMSmallStrain::mlawJ2VMSmallStrain(const int num, const double E, const double nu,const double rho,
                               const J2IsotropicHardening& j2IH, const double tol,
                               const bool matrixbyPerturbation, const double pert, const bool init):
                               materialLaw(num,init),_tol(tol),_perturbationfactor(pert), _rho(rho),
                               _tangentByPerturbation(matrixbyPerturbation), _E(E),_nu(nu){
  _j2IH = j2IH.clone();
  _K = E/(3.*(1.-2.*nu));
  _mu = E/(2.*(1.+nu));
  _lambda = (E*nu)/(1.+nu)/(1.-2.*nu);

  STensor3 I(1.);
  for (int i=0; i<3; i++){
    for (int j=0; j<3; j++){
      for (int k=0; k<3; k++){
        for (int l=0; l<3; l++){
          _I4dev(i,j,k,l) = 0.5*(I(i,k)*I(j,l)+I(i,l)*I(j,k)) - I(i,j)*I(k,l)/3.;
        }
      }
    }
  }

  _Cel*=0.;
  _Cel(0,0,0,0) = _lambda + 2.*_mu;
  _Cel(1,1,0,0) = _lambda;
  _Cel(2,2,0,0) = _lambda;
  _Cel(0,0,1,1) = _lambda;
  _Cel(1,1,1,1) = _lambda + 2.*_mu;
  _Cel(2,2,1,1) = _lambda;
  _Cel(0,0,2,2) = _lambda;
  _Cel(1,1,2,2) = _lambda;
  _Cel(2,2,2,2) = _lambda + 2.*_mu;

  _Cel(1,0,1,0) = _mu;
  _Cel(2,0,2,0) = _mu;
  _Cel(0,1,0,1) = _mu;
  _Cel(2,1,2,1) = _mu;
  _Cel(0,2,0,2) = _mu;
  _Cel(1,2,1,2) = _mu;

  _Cel(0,1,1,0) = _mu;
  _Cel(0,2,2,0) = _mu;
  _Cel(1,0,0,1) = _mu;
  _Cel(1,2,2,1) = _mu;
  _Cel(2,0,0,2) = _mu;
  _Cel(2,1,1,2) = _mu;
};



mlawJ2VMSmallStrain::mlawJ2VMSmallStrain(const mlawJ2VMSmallStrain& src):materialLaw(src){
  _j2IH = NULL;
  if (src._j2IH != NULL)
    _j2IH = src._j2IH->clone();
  _rho = src._rho;
  _K = src._K;
  _mu = src._mu;
  _E = src._E;
  _nu = src._nu;
  _lambda = src._lambda;
  _tol = src._tol; // tolerace of plastic corrector
  _perturbationfactor = src._perturbationfactor; // perturbation factor
  _tangentByPerturbation = src._tangentByPerturbation; // flag for tangent by perturbation
  _I4dev = src._I4dev;
  _Cel = src._Cel;
};

mlawJ2VMSmallStrain::~mlawJ2VMSmallStrain(){
  if (_j2IH) delete _j2IH; _j2IH = NULL;
};

void mlawJ2VMSmallStrain::createIPState(IPStateBase* &ips,const bool* state_,const MElement *ele, const int nbFF_, const IntPt *GP, const int gpt) const
{
  IPVariable* ipvi = new IPJ2linear(_j2IH);
  IPVariable* ipv1 = new IPJ2linear(_j2IH);
  IPVariable* ipv2 = new IPJ2linear(_j2IH);
  if(ips != NULL) delete ips;
  ips = new IP3State(state_,ipvi,ipv1,ipv2);
};

double mlawJ2VMSmallStrain::soundSpeed() const
{
  double factornu = (1.-_nu)/((1.+_nu)*(1.-2.*_nu));
  return sqrt(_E*factornu/_rho);
};

void mlawJ2VMSmallStrain::constitutive(const STensor3& F0,         // previous deformation gradient (input @ time n)
                            const STensor3& Fn,         // current deformation gradient (input @ time n+1)
                            STensor3 &P,                // current 1st Piola-Kirchhoff stress tensor (output)
                            const IPVariable *q0i,       // array of previous internal variables
                            IPVariable *q1i,             // current array of internal variable (in ipvcur on output),
                            STensor43 &Tangent,         // tangents (output)
                            const bool stiff,            // if true compute the tangents
                           STensor43* elasticTangent) const{

  static STensor43 depsEldF;
  static STensor3 dpdF;
  IPJ2linear* q1 = static_cast<IPJ2linear*>(q1i);
  const IPJ2linear* q0 = static_cast<const IPJ2linear*>(q0i);

  predictorCorector(F0,Fn,P,q0,q1,Tangent,depsEldF,dpdF,stiff,elasticTangent);
};


void mlawJ2VMSmallStrain::predictorCorector(const STensor3& F0,         // initial deformation gradient (input @ time n)
                            const STensor3& Fn,         // updated deformation gradient (input @ time n+1)
                            STensor3 &P,                // updated 1st Piola-Kirchhoff stress tensor (output)
                            const IPJ2linear *q0,       // array of initial internal variable
                            IPJ2linear *q1,             // updated array of internal variable (in ipvcur on output),
                            STensor43& Tangent,         // mechanical tangents (output)
                            STensor43& depsEldF,
                            STensor3& dpdF,
                            const bool stiff,
                            STensor43* elasticTangent
                            ) const {
   // same state
  // plastic strain
  STensor3& Ep = q1->_j2lepsp;
  double& gamma = q1->_j2lepspbarre;

  Ep = q0->_j2lepsp; // plastic strain tensor (equal I if no plastic occurs)
  gamma = q0->_j2lepspbarre; // equivalent plastic strain

  static const STensor3 I(1.); // identity tensor

  // small strain
  static STensor3 eps;
  for (int i=0; i<3; i++){
    for (int j=0; j<3; j++){
      eps(i,j) =  0.5*(Fn(i,j)+Fn(j,i)) - I(i,j);
    }
  }

  // trial elastic strain Eetr = eps-(Ep-I)
  static STensor3 Eetr;
  Eetr = (eps);
  Eetr -= Ep;
  Eetr += I;

  // trace and dev parts
  double traceEetr = Eetr.trace();
  static STensor3 devEetr;
  devEetr = Eetr.dev();

  // trial pressure and deviatoric stress
  double p = _K*traceEetr; // pressure
  static STensor3 devSigtr;
  devSigtr = (devEetr); // trial deviatoric stress
  devSigtr *= (2.*_mu);

  // stress
  P = devSigtr;
  P.daxpy(I,p);
  // elastic strain
  q1->_Ee = Eetr;
  q1->_plasticPower = 0.;
  q1->_j2ldsy = 0.;

  if (stiff and !_tangentByPerturbation){
    Tangent = _Cel;
    STensorOperation::zero(dpdF);
    STensorOperation::zero(depsEldF);
  }
  
  if (elasticTangent != NULL){
    (*elasticTangent) =  _Cel;
  }

  // VM stress
  double sigVMtr = sqrt(1.5*devSigtr.dotprod());

  // get current yield stress
  _j2IH->hardening(q0->_j2lepspbarre, q0->getConstRefToIPJ2IsotropicHardening(), gamma, q1->getRefToIPJ2IsotropicHardening());
  double  Sy0   = q1->getConstRefToIPJ2IsotropicHardening().getR();
  double H    = q1->getConstRefToIPJ2IsotropicHardening().getDR();

  double VMCriterion = sigVMtr -Sy0;
  if (VMCriterion >0.){
    // plastic occurs
    int ite =0, maxite = 1000;
    double deps = 0;
    double Sy = Sy0;
    while (fabs(VMCriterion/_j2IH->getYield0())> _tol){
      // plastic increment
      deps += VMCriterion/(3.*_mu+H);
      // recompute plastic state
      _j2IH->hardening(q0->_j2lepspbarre, q0->getConstRefToIPJ2IsotropicHardening(), gamma+deps, q1->getRefToIPJ2IsotropicHardening());
      Sy   = q1->getConstRefToIPJ2IsotropicHardening().getR();
      H    = q1->getConstRefToIPJ2IsotropicHardening().getDR();

      VMCriterion = sigVMtr - 3.*_mu*deps - Sy;
      ite ++;

      if(ite > maxite){
        Msg::Error("No convergence for plastic correction in j2linearVM small strain !!");
        break;
      }

    }

    gamma += deps;
    // update plastic strain
    static STensor3 N;
    N = (devSigtr);
    N*= (1.5/sigVMtr);

    Ep.daxpy(N,deps);
    double fact = -2.*_mu*deps;
    P.daxpy(N,fact);
    q1->_Ee.daxpy(N,-deps);
    q1->_plasticPower = (deps*Sy)/this->getTimeStep();
    q1->_j2ldsy = Sy - Sy0;

    if (stiff and !_tangentByPerturbation){
      STensorOperation::multSTensor3STensor43(N,_Cel,dpdF);
      dpdF *= (1./(3.*_mu+H));
      
      
      static STensor43 dNdSigpr;
      for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
          for (int k=0; k<3; k++){
            for (int l=0; l<3; l++){
              dNdSigpr(i,j,k,l) = (1.5*_I4dev(i,j,k,l) - N(i,j)*N(k,l))/sigVMtr;
            }
          }
        }
      }
      
      STensorOperation::multSTensor43(dNdSigpr,_Cel,depsEldF);
      depsEldF *= (-deps);
      STensorOperation::prodAdd(N,dpdF,-1.,depsEldF);
      depsEldF += _I4;
      
      
      for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
          for (int k=0; k<3; k++){
            for (int l=0; l<3; l++){
              for (int p=0; p<3; p++){
                for (int q=0; q<3; q++){
                  for (int r=0; r<3; r++){
                    for (int s=0; s<3; s++){
                      Tangent(i,j,k,l) -= _Cel(i,j,r,s)*N(r,s)*N(p,q)*_Cel(p,q,k,l)/(3.*_mu+H) + _Cel(i,j,r,s)*deps*dNdSigpr(r,s,p,q)*_Cel(p,q,k,l);
                    }
                  }

                }
              }
            }
          }
        }
      }
    }
  }
  
  q1->_elasticEnergy = deformationEnergy(q1->_Ee);

  if (stiff and _tangentByPerturbation){
    static STensor43 Temp;
    static STensor3 Fpp, Pp;
    static IPJ2linear q11(*q1);
    for (int i=0; i<3; i++){
      for (int j=0; j<3; j++){
        Fpp = (Fn);
        Fpp(i,j)+=_perturbationfactor;
        this->constitutive(F0,Fpp,Pp,q0,&q11,Temp,false);
        for (int k=0; k<3; k++){
          for (int l=0; l<3; l++){
            Tangent(k,l,i,j) = (Pp(k,l)-P(k,l))/_perturbationfactor;
            depsEldF(k,l,i,j) = (q11._Ee(k,l) - q1->_Ee(k,l))/_perturbationfactor;
          }
        }
        
        dpdF(i,j) = (q11._j2lepspbarre - q1->_j2lepspbarre)/_perturbationfactor;
      }
    }
  }  
};

double mlawJ2VMSmallStrain::deformationEnergy(const STensor3 &Ee) const
{
  double ener = 0.;
  for (int i=0; i< 3; i++){
    for (int j=0; j<3; j++){
      for (int k=0; k<3; k++){
        for (int l=0; l<3; l++){
          ener += 0.5*Ee(i,j)*_Cel(i,j,k,l)*Ee(k,l);
        }
      } 
    }
  }
  return ener;
}
