//
// C++ Interface: material law
//
// Description: Gurson elasto-plastic law with non local damage interface
//
// Author:  <L. Noels>, (C) 2013
//
// Copyright: See COPYING file that comes with this distribution
//
//
#ifndef MLAWNONLOCALDAMAGEGURSON_H_
#define MLAWNONLOCALDAMAGEGURSON_H_
#include "mlawNonLocalPorous.h"
#include "scalarFunction.h"
class mlawNonLocalDamageGurson : public mlawNonLocalPorosity
{
  #ifndef SWIG
  protected:
    // Material parameters - gurson yield surface
    double _q1;
    double _q2;
    double _q3;
    
     // Internal laws - void/damage nucleation
    NucleationLaw* _nucleationLaw;
    /// Nucleation criterion
    NucleationCriterionLaw* _nucleationCriterionLaw;
    /// Void state evolution law
    voidStateEvolutionLaw* _voidEvolutionLaw;
    /// Coalescense law
    CoalescenceLaw* _coalescenceLaw;

    // yield coefficients depend on stress triaxiality
    scalarFunction* _STfunc_q1;
    scalarFunction* _STfunc_q2;
    scalarFunction* _STfunc_q3;

    //adding temFunc for Gurson parameters----------------added
    scalarFunction* _temFunc_q1;
    scalarFunction* _temFunc_q2;
    scalarFunction* _temFunc_q3;
    //end of temFuncs---------------------------------------end

    #endif //SWIG

  public:
    /* Default constructor */
    mlawNonLocalDamageGurson(const int num,const double E,const double nu, const double rho,
                           const double q1, const double q2, const double q3, const double fVinitial,
                           const J2IsotropicHardening &j2IH, const CLengthLaw &cLLaw,
                           const double tol, const bool matrixbyPerturbation, const double pert);

    //set for temperature functions--------------------added
    void setTemperatureFunction_q1(const scalarFunction& Tfunc){
      if (_temFunc_q1 != NULL) delete _temFunc_q1;
      _temFunc_q1 = Tfunc.clone();
    };
    void setTemperatureFunction_q2(const scalarFunction& Tfunc){
      if (_temFunc_q2 != NULL) delete _temFunc_q2;
      _temFunc_q2 = Tfunc.clone();
    };
    void setTemperatureFunction_q3(const scalarFunction& Tfunc){
      if (_temFunc_q3 != NULL) delete _temFunc_q3;
      _temFunc_q3 = Tfunc.clone();
    };

    virtual double getq1T (double T) const{return _q1*_temFunc_q1->getVal(T);};
    virtual double getq2T (double T) const{return _q2*_temFunc_q2->getVal(T);};
    virtual double getq3T (double T) const{return _q3*_temFunc_q3->getVal(T);};
    virtual double getq1Der (double T)const{return _q1*_temFunc_q1->getDiff(T);};
    virtual double getq2Der (double T)const{return _q2*_temFunc_q2->getDiff(T);};
    virtual double getq3Der (double T)const{return _q3*_temFunc_q3->getDiff(T);};
    //end of get functions for parameters-----------------------end
    
    virtual void setNucleationLaw(const NucleationLaw& added_gdnLaw); 
    virtual void setNucleationCriterion(const NucleationCriterionLaw& nuclCriterion);
    virtual void setVoidEvolutionLaw(const voidStateEvolutionLaw& voidtateLaw);
    virtual void setCoalescenceLaw(const CoalescenceLaw& added_coalsLaw); 

    #ifndef SWIG
    // Constructors & destructor
    mlawNonLocalDamageGurson(const mlawNonLocalDamageGurson &source);
    virtual ~mlawNonLocalDamageGurson();
    virtual materialLaw* clone() const {return new mlawNonLocalDamageGurson(*this);};

    // Function of materialLaw
    virtual matname getType() const{return materialLaw::nonLocalDamageGurson;}
    virtual void createIPState(IPStateBase* &ips,const bool* state_=NULL,const MElement *ele=NULL, const int nbFF_=0, const IntPt *GP=NULL, const int gpt = 0) const;
    virtual void createIPVariable(const double fInit, IPNonLocalPorosity* &ipv) const;
    virtual void initLaws(const std::map<int,materialLaw*> &maplaw){}; // this law is initialized so nothing to do
  
    virtual const NucleationLaw* getNucleationLaw() const {return _nucleationLaw;};
    virtual const NucleationCriterionLaw* getNucleationCriterionLaw() const {return _nucleationCriterionLaw;};
    virtual const voidStateEvolutionLaw* getVoidEvolutionLaw() const {return _voidEvolutionLaw;};
    virtual const CoalescenceLaw* getCoalescenceLaw() const {return _coalescenceLaw;};
    
    // Specific functions
  public:
    virtual void voidStateGrowth(const double yieldfV0, const double yieldfV1, // yield porosity
                        const double DeltahatD, const double DeltahatQ, const double DeltahatP, // internal variables
                        const IPNonLocalPorosity *q0, IPNonLocalPorosity *q1) const;
                        
    virtual double yieldFunction(const double kcorEq, const double pcor, const double R, const double yieldfV,
                                const IPNonLocalPorosity* q0, const IPNonLocalPorosity* q1, const double* T = NULL,
                                bool diff=false, fullVector<double>* grad = NULL,
                                bool withthermic = false, double* dfdT = NULL) const;
                                
    virtual void plasticFlow(double & Ns, double& Nv,
                              const double kcorEq, const double pcor, const double R,
                              const IPNonLocalPorosity* q0, const IPNonLocalPorosity* q1, const double* T = NULL,
                              bool diff=false, fullVector<double>* gradNs=NULL, // grad of Ns with respect to [kcorEq pcor R fV Chi W]
                              fullVector<double>* gradNv = NULL, // grad of Nv with respect to  [kcorEq pcor R fV Chi W]
                              bool withTher = false, double* dNsDT = NULL, double* dNvDT=NULL) const;
                                  
                                  
    virtual double getOnsetCriterion(const IPNonLocalPorosity* q1) const;
    virtual void checkCoalescence(IPNonLocalPorosity* q1, const IPNonLocalPorosity *q0, const double* T) const;
    virtual void forceCoalescence(IPNonLocalPorosity* q1, const IPNonLocalPorosity *q0, const double* T) const;
    virtual void checkFailed(IPNonLocalPorosity* q1, const IPNonLocalPorosity *q0) const;

      
      // general interafce for I1-J2-J3 plasticity
  public:
    virtual double I1J2J3_yieldFunction(const STensor3& kcor, const double R, const IPNonLocalPorosity* q0, const IPNonLocalPorosity* q1, const double* T = NULL,
                                        bool diff=false, STensor3* DFDkcor=NULL, double* DFDR=NULL, std::vector<double>* DFDY=NULL, 
                                        bool withThermic=false, double *DFDT=NULL) const;
    virtual void I1J2J3_getYieldNormal(STensor3& Np, const STensor3& kcor, const double R, const IPNonLocalPorosity* q0, const IPNonLocalPorosity* q1, const double* T = NULL,
                                        bool diff=false, STensor43* DNpDkcor=NULL, STensor3* DNpDR=NULL, std::vector<STensor3>* DNpDY=NULL,
                                        bool withThermic=false, STensor3* DNpDT=NULL  
                                        ) const;
 #endif // SWIG
};

#endif // MLAWNONLOCALDAMAGEGURSON_H_
