// C++ Interface: material law
//
// Description: general class for nonlocal porous material law
//
// Author:  <V.D. Nguyen>, (C) 2018
//
// Copyright: See COPYING file that comes with this distribution
//
//
#ifndef MLAWNONLOCALPOROUS_H_
#define MLAWNONLOCALPOROUS_H_
#include "mlaw.h"
#include "ipNonLocalPorosity.h"
#include "j2IsotropicHardening.h"
#include "CoalescenceLaw.h"
#include "NucleationCriterionLaw.h"
#include "voidStateEvolutionLaw.h"
#include "NucleationLaw.h"
#include "CLengthLaw.h"
#include "viscosityLaw.h"
#include "STensor43.h"
#include "STensor63.h"
#include "scalarFunction.h"

class mlawNonLocalPorosity: public materialLaw
{
public:
    // Enumeration:
      // Non-local implementation (local / 1 non-local variable (porosity) / 3 NL var (plastic increment) / 1 NL var (ln(porosity)))
    enum nonLocalMethod{LOCAL=0, NONLOCAL_POROSITY=1,NONLOCAL_FULL_THREEVAR=2,NONLOCAL_LOG_POROSITY=3};
      // Stress tensor used in the yield surface and in the nucleation criterion
      // corotationnal Cauchy or Kirchhoff
    enum stressFormumation{CORO_CAUCHY=0,CORO_KIRCHHOFF=1};
  #ifndef SWIG
protected:
    // Internal laws
    /// Isotropic hardening
    J2IsotropicHardening *_j2IH;
    /// Viscoplastic law
    viscosityLaw *_viscoplastlaw;  // /!\ Viscosity is foressen to compute viscosity, not the overstress itself
    /// Lc law
    std::vector<CLengthLaw*> _cLLaw;

    // Material parameters
    /// basic parameters for hyperelasticity
    double _rho;  // density
    double _E;    // young modulus
    double _nu;   // Poisson ratio
    scalarFunction* _temFunc_rho; // temperature dependence of density;
    scalarFunction* _temFunc_E;   // temperature dependence of bulk modulus;
    scalarFunction* _temFunc_nu;  // temperature dependence of Poisson ratio;
    /// others
    double _lambda; // 1st lame parameter
    double _mu; // 2nd lame parameter (=G)
    double _K; // bulk modulus
    double _K3; // 3*bulk modulus
    double _mu3; // 3*_mu = 3*G
    double _mu2; // 2*_mu = 2*G

    // Material parameters - Thermal expansion parameters
    double _alp;
    STensor3 Stiff_alphaDilatation;
    scalarFunction* _temFunc_alp; // temperature dependent
    double _referenceT;
    bool _thermomechanicsCoupling;
    double _TaylorQuineyFactor; // for heat conversion
    double _ThermalConductivity;
    double _cp; // heat capacity per unit field
    scalarFunction* _temFunc_ThermalConductivity;
    scalarFunction* _temFunc_cp;
    bool _thermalEstimationPreviousConfig;
    STensor3 linearK;

    // Shear term
    double _kw; // shear term Nashon-Hutchinson
    scalarFunction* _triFunc_kw;
    scalarFunction* _LodeFunction; // func of xi = cos(3*theta)

    // Elastic tensor
    STensor43   Cel; // elastic Hook tensor

    // Initial porosity
    double _fVinitial; // initial porosity
    double _randMaxbound; // maximal incertainty bound
    double _randMinbound; // minimal incertainty bound

    // Failure management
    double _fV_failure;     // max porosity in yield surface (value for which fG=0)
    double _localfVFailure; // max local porosity at failure (almost 1)
    scalarFunction* _correctedRegularizedFunction;  // Regularisation function for yield/effective values
    double _failedTol;      // tolerance applied on max values at which the IP is considered as failed.

    // Transition management
    /// Behaviour after blocking
    int _postBlockedDissipationBehavior; // 1- purely elastic; 2-purely elastoplastic


    bool _nucleationVoidDuringCoalescence; // true if nucleation during coalesence, false if not, false by default

    // Non-local variable management
    int _numNonLocalVar;            // number of non-local variables
    nonLocalMethod _nonlocalMethod; // non-local method used

    stressFormumation _stressFormulation;

    // Numerical parameters
    /// Approx. order for tensorial log and exp
    int _orderlogexp;
    /// Newton Raphson parameters for plastic corrector
    double _tol; // tolerance for iterative process
    int _maxite; // maximal number of iteration
    /// Tangent by pertubation
    double _perturbationfactor; // perturbation factor
    bool _tangentByPerturbation; // flag for tangent by perturbation
    /// Substepping
    bool _withSubstepping;      // true if subbstepping is activated (false by default)
    int _maxAttemptSubstepping; // maximal attemp (times on which the defo is divided by 2)

    // Tools
    STensor43 _I4;    // 4th-order symmetric identity tensor
    STensor3  _I;     // Second order identity tensor
    STensor43 _I4dev;  // 4-th order deviatoric tensor
    STensor43 _I4vol;
    #endif //SWIG

  public:
    // Option settings
    virtual void setStressFormulation(const int fm);
    virtual void activateVoidNucleationDuringCoalescence(const bool flg);
    virtual void setNonLocalMethod(const int i);
    virtual void setOrderForLogExp(const int newOrder);
    virtual void setViscosityLaw(const viscosityLaw& visco);
    virtual void setScatterredInitialPorosity(double f0min, double f0max);
    virtual void setShearPorosityGrowthFactor(const double k);
    virtual void setStressTriaxialityFunction_kw(const scalarFunction& fct);
    virtual void setLodeFunction(const scalarFunction& fct);
    virtual void setFailureTolerance(const double NewFailureTol);
    virtual void setLocalFailurePorosity(const double newFvMax);
    virtual void setCorrectedRegularizedFunction(const scalarFunction& fct);
    virtual void setPostBlockedDissipationBehavior(const int method);
    virtual void clearCLengthLaw();
    virtual void setCLengthLaw(const CLengthLaw& clength);
    
    // depending on porous law
    virtual void setNucleationLaw(const NucleationLaw& added_gdnLaw) = 0; 
    virtual void setNucleationCriterion(const NucleationCriterionLaw& nuclCriterion) = 0;
    virtual void setVoidEvolutionLaw(const voidStateEvolutionLaw& voidtateLaw) = 0;
    virtual void setCoalescenceLaw(const CoalescenceLaw& added_coalsLaw) = 0; 
    
    // Options setting -- temperature part
    virtual void setThermomechanicsCoupling(const bool fl);
    virtual void setReferenceT(const double referenceT);
    virtual void setThermalExpansionCoefficient(const double alp);
    virtual void setReferenceThermalConductivity(const double KK);
    virtual void setTemperatureFunction_ThermalConductivity(const scalarFunction& Tfunc);
    virtual void setReferenceCp(const double Cp);
    virtual void setTemperatureFunction_Cp(const scalarFunction& Tfunc);
    virtual void setThermalEstimationPreviousConfig(const bool flag);
    virtual void setTaylorQuineyFactor(const double f);

    //
    virtual void setSubStepping(const bool fl, const int maxNumStep);

    virtual void setTemperatureFunction_rho(const scalarFunction& Tfunc);
    virtual void setTemperatureFunction_E(const scalarFunction& Tfunc);
    virtual void setTemperatureFunction_nu(const scalarFunction& Tfunc);
    virtual void setTemperatureFunction_alp(const scalarFunction& Tfunc);


  #ifndef SWIG
    virtual mlawNonLocalPorosity::nonLocalMethod getNonLocalMethod() const;
    static double frand(const double a,const double b);

    mlawNonLocalPorosity(const int num,const double E,const double nu, const double rho,
                           const double fVinitial, const J2IsotropicHardening &j2IH, const CLengthLaw &cLLaw,
                           const double tol=1.e-8, const bool matrixbyPerturbation = false, const double pert = 1e-8);
    mlawNonLocalPorosity(const mlawNonLocalPorosity &source);
    virtual ~mlawNonLocalPorosity();

    virtual materialLaw* clone() const  = 0;

    // Function of materialLaw
    virtual matname getType() const = 0;
    virtual void createIPState(IPStateBase* &ips,const bool* state_=NULL,const MElement *ele=NULL, const int nbFF_=0, const IntPt *GP=NULL, const int gpt = 0) const = 0;
    virtual void createIPVariable(const double fInit, IPNonLocalPorosity* &ipv) const = 0;
    virtual void initLaws(const std::map<int,materialLaw*> &maplaw) = 0;
//--------------------------------------------------added
    virtual double density(double T) const;
    virtual double density() const {return _rho;};
    virtual double young(double T) const;
    virtual double poisson(double T) const;
    virtual double youngDerivative(double T) const;
    virtual double poissonDerivative(double T) const;
    virtual double ThermalExpansion(double T) const;
    virtual double ThermalExpansionDerivative(double T) const;
//---------------------------------------------------end
    virtual double soundSpeed() const; // default but you can redefine it for your case
    virtual bool withEnergyDissipation() const {return true;};
    virtual double scaleFactor() const { return _mu;};

    // Access functions
    const std::vector<CLengthLaw*>* getCLengthLaw() const {return &_cLLaw; };
    const J2IsotropicHardening* getJ2IsotropicHardening() const {return _j2IH;};
    const viscosityLaw* getViscoplasticLaw() const {return _viscoplastlaw;};

    virtual const NucleationLaw* getNucleationLaw() const =0;
    virtual const NucleationCriterionLaw* getNucleationCriterionLaw() const =0;
    virtual const voidStateEvolutionLaw* getVoidEvolutionLaw() const =0;
    virtual const CoalescenceLaw* getCoalescenceLaw() const =0;

//----------------------------------------------------added
    virtual double bulkModulus(double T) const;
    virtual double bulkModulusDerivative (double T) const;
    virtual double shearModulus(double T) const;
    virtual double shearModulusDerivative (double T) const;
    virtual double lambda(double T) const;
    virtual double getShearFactor(const STensor3& kcor, bool stiff=false, STensor3* DkDkcor=NULL) const;
    //void elasticTensor(double T) const;
//-----------------------------------------------------end
    double poissonRatio() const{return _nu;}
    double getInitialPorosity() const {return _fVinitial*frand(_randMinbound,_randMaxbound);} // for Ipv creation only!
    virtual bool isThermomechanicallyCoupled() const {return _thermomechanicsCoupling;}
    virtual int getNumNonLocalVariables() const {return _numNonLocalVar;}
    virtual double getFailureYieldPorosityValue() const {return _fV_failure;};
    const STensor43& getCel() const {return Cel;};
    virtual const STensor3& getInitialConductivityTensor() const {return linearK;};
    virtual const STensor3& getStiff_alphaDilatation() const {return Stiff_alphaDilatation;};

    virtual const stressFormumation getStressFormulation() const {return _stressFormulation;};
  public:
    // Local case
    virtual void constitutive(
                              const STensor3& F0,         // initial deformation gradient (input @ time n)
                              const STensor3& Fn,         // updated deformation gradient (input @ time n+1)
                              STensor3 &P,                // updated 1st Piola-Kirchhoff stress tensor (output)
                              const IPVariable *q0,       // array of initial internal variable
                              IPVariable *q1,             // updated array of internal variable (in ipvcur on output),
                              STensor43 &Tangent,         // constitutive tangents (output)
                              const bool stiff,           // if true compute the tangents
                              STensor43* elasticTangent = NULL) const;

    virtual void constitutive(
                              const STensor3& F0,         // initial deformation gradient (input @ time n)
                              const STensor3& Fn,         // updated deformation gradient (input @ time n+1)
                              STensor3 &P,                // updated 1st Piola-Kirchhoff stress tensor (output)
                                                          // contains the initial values on input
                              const IPNonLocalPorosity *q0,       // array of initial internal variable
                              IPNonLocalPorosity *q1,             // updated array of internal variable (in ipvcur on output),
                              STensor43 &Tangent,         // constitutive tangents (output)
                              std::vector<STensor3>  &dLocalVarDStrain,
                              std::vector<STensor3>  &dStressDNonLocalVar,
                              fullMatrix<double>    &dLocalVarDNonLocalVar,
                              const bool stiff,
                              STensor43* elasticTangent = NULL) const;

    virtual void I1J2J3_constitutive(
                              const STensor3& F0,         // initial deformation gradient (input @ time n)
                              const STensor3& Fn,         // updated deformation gradient (input @ time n+1)
                              STensor3 &P,                // updated 1st Piola-Kirchhoff stress tensor (output)
                                                          // contains the initial values on input
                              const IPNonLocalPorosity *q0,       // array of initial internal variable
                              IPNonLocalPorosity *q1,             // updated array of internal variable (in ipvcur on output),
                              STensor43 &Tangent,         // constitutive tangents (output)
                              std::vector<STensor3>  &dLocalVarDStrain,
                              std::vector<STensor3>  &dStressDNonLocalVar,
                              fullMatrix<double>    &dLocalVarDNonLocalVar,
                              const bool stiff,
                              STensor43* elasticTangent = NULL) const;

    virtual void constitutive(const STensor3& F0,         // initial deformation gradient (input @ time n)
                              const STensor3& Fn,         // updated deformation gradient (input @ time n+1)
                              STensor3 &P,                // updated 1st Piola-Kirchhoff stress tensor (output)
                              const IPNonLocalPorosity *q0,       // array of initial internal variable
                              IPNonLocalPorosity *q1,             // updated array of internal variable (in ipvcur on output),
                              STensor43 &Tangent,         // constitutive tangents (output)
                              std::vector<STensor3>  &dLocalVarDStrain,
                              std::vector<STensor3>  &dStressDNonLocalVar,
                              fullMatrix<double>    &dLocalVarDNonLocalVar,
                              const bool stiff,
                              STensor3& dPdT,
                              STensor33 &dPdgradT,
                              fullMatrix<double>  &dLocalVarDT,
                              const double& T0, // previous temperature
                              const double& Tn, // current temperature
                              const SVector3 &gradT0, // previous temperature gradient
                              const SVector3 &gradT, // current temperature gradient
                              SVector3 &fluxT, // temperature flux
                              STensor33 &dfluxTdF, // thermal-mechanical coupling
                              std::vector<SVector3> &dfluxTdNonLocalVar,
                              SVector3 &dfluxTdT,
                              STensor3 &dfluxTdgradT, // thermal tangent
                              double &thermalSource,   // - cp*dTdt
                              STensor3 &dthermalSourcedF,
                              fullMatrix<double> &dthermalSourcedNonLocalVar,
                              double &dthermalSourcedT, // thermal source
                              SVector3 &dthermalSourcedgradT,
                              double& mechanicalSource, // mechanical source--> convert to heat
                              STensor3 & dmechanicalSourceF,
                              fullMatrix<double> & dmechanicalSourcedNonLocalVar,
                              double & dmechanicalSourcedT,
                              SVector3 &dmechanicalSourcedgradT,
                              STensor43* elasticTangent = NULL) const;


    virtual void predictorCorrector(const STensor3& F0,         // initial deformation gradient (input @ time n)
                              const STensor3& Fn,         // updated deformation gradient (input @ time n+1)
                              STensor3 &P,                // updated 1st Piola-Kirchhoff stress tensor (output)
                              const IPNonLocalPorosity *q0,       // array of initial internal variable
                              IPNonLocalPorosity *q1,             // updated array of internal variable (in ipvcur on output),
                              STensor43 &Tangent,         // constitutive tangents (output)
                              std::vector<STensor3>  &dLocalVarDStrain,
                              std::vector<STensor3>  &dStressDNonLocalVar,
                              fullMatrix<double>    &dLocalVarDNonLocalVar,
                              const bool stiff,
                              STensor43& dFedF, std::vector<STensor3>& dFeDNonlocalVar,
                              STensor3& dPdT,
                              STensor33 &dPdgradT,
                              fullMatrix<double>  &dLocalVarDT,
                              const double& T0, // previous temperature
                              const double& Tn, // current temperature
                              const SVector3 &gradT0, // previous temperature gradient
                              const SVector3 &gradT, // current temperature gradient
                              SVector3 &fluxT, // temperature flux
                              STensor33 &dfluxTdF, // thermal-mechanical coupling
                              std::vector<SVector3> &dfluxTdNonLocalVar,
                              SVector3 &dfluxTdT,
                              STensor3 &dfluxTdgradT, // thermal tangent
                              double &thermalSource,   // - cp*dTdt
                              STensor3 &dthermalSourcedF,
                              fullMatrix<double> &dthermalSourcedNonLocalVar,
                              double &dthermalSourcedT, // thermal source
                              SVector3 &dthermalSourcedgradT,
                              double& mechanicalSource, // mechanical source--> convert to heat
                              STensor3 & dmechanicalSourceF,
                              fullMatrix<double> & dmechanicalSourcedNonLocalVar,
                              double & dmechanicalSourcedT,
                              SVector3 &dmechanicalSourcedgradT
                              ) const;

    virtual void predictorCorrector(const STensor3& F0,         // initial deformation gradient (input @ time n)
                              const STensor3& Fn,         // updated deformation gradient (input @ time n+1)
                              STensor3 &P,                // updated 1st Piola-Kirchhoff stress tensor (output)
                                                          // contains the initial values on input
                              const IPNonLocalPorosity *q0,       // array of initial internal variable
                              IPNonLocalPorosity *q1,             // updated array of internal variable (in ipvcur on output),
                              STensor43 &Tangent,         // constitutive tangents (output)
                              std::vector<STensor3>  &dLocalVarDF,
                              std::vector<STensor3>  &dStressDNonLocalVar,
                              fullMatrix<double>    &dLocalVarDNonLocalVar,
                              const bool stiff,                   // if true compute the tangents
                              STensor43& dFedF, std::vector<STensor3>& dFeDNonlocalVar,
                              STensor3*  dStressDT = NULL,
                              fullMatrix<double>* dLocalVarDT = NULL,
                              const double T0 = 0.,const double T = 0.
                             ) const;

    // Case-dependent functions
    // void characteristic Y = [fV Chi W]
    virtual double yieldFunction(const double kcorEq, const double pcor, const double R, const double yieldfV,
                                const IPNonLocalPorosity* q0, const IPNonLocalPorosity* q1, const double* T = NULL,
                                bool diff=false, fullVector<double>* grad = NULL,
                                bool withthermic = false, double* dfdT = NULL) const = 0;

    virtual void plasticFlow(double & Ns, double& Nv,
                              const double kcorEq, const double pcor, const double R,
                              const IPNonLocalPorosity* q0, const IPNonLocalPorosity* q1, const double* T = NULL,
                              bool diff=false, fullVector<double>* gradNs=NULL, // grad of Ns with respect to [kcorEq pcor R fV Chi W]
                              fullVector<double>* gradNv = NULL, // grad of Nv with respect to  [kcorEq pcor R fV Chi W] )
                              bool withTher = false, double* dNsDT = NULL, double* dNvDT=NULL) const = 0;

    // Internal state and failure management
      // Function from mlaw that gathers checks for material status (i.e. for nucleation criterion, coalescence and failure)
    virtual void checkInternalState(IPVariable* ipv, const IPVariable* ipvprev) const;
      /// update nucleation criterion status
    virtual void checkNucleationCriterion(IPNonLocalPorosity* q1, const IPNonLocalPorosity *q0) const;
      /// update coalescence status
    virtual void checkCoalescence(IPNonLocalPorosity* q1, const IPNonLocalPorosity *q0, const double* T) const = 0;
    virtual void forceCoalescence(IPNonLocalPorosity* q1, const IPNonLocalPorosity *q0, const double* T) const = 0;
      /// update failure status
    virtual void checkFailed(IPNonLocalPorosity* q1, const IPNonLocalPorosity *q0) const = 0;

    // why ????
    virtual double getOnsetCriterion(const IPNonLocalPorosity* q1) const = 0;


    // Void nucleation
    // compute the increment of nucleated porosity
    virtual void nucleation(const double p1, const double p0, const double f0, IPNonLocalPorosity* q1, const IPNonLocalPorosity *q0) const;

    // Void expansion from growth, nucleation and shear
    // in the integrated form, it is a function of
    // plastic defos as DeltaHatD=deviatoric plastic defo, DeltaHatQ = volumetric plastic defor, DeltaHatP=matrixPlasticDefo
    // and kcor
    virtual void localPorosityGrowth(IPNonLocalPorosity *q1,
                      const double DeltaHatD, const double DeltaHatQ, const double DeltaHatP,
                      const STensor3& Kcor, const IPNonLocalPorosity *q0,
                      bool stiff = false,
                      double* DDeltafVDDeltaHatD=NULL, double* DDeltafVDDeltaHatQ=NULL, double* DDeltafVDDeltaHatP=NULL, STensor3* DDeltafVDKcor=NULL) const;
    // void state beside porosity
    virtual void voidStateGrowth(const double yieldfV0, const double yieldfV1, // yield porosity
                        const double DeltahatD, const double DeltahatQ, const double DeltahatP, // internal variables
                        const IPNonLocalPorosity *q0, IPNonLocalPorosity *q1) const = 0;
    
    // void characteristic evolutions
     virtual void voidCharacteristicsEvolution_NonLocalPorosity(const STensor3& kcor,
                                                const IPNonLocalPorosity *q0, IPNonLocalPorosity* q1,
                                                const double DeltaHatD, const double DeltaHatQ, const double DeltaHatP,
                                                double& DfVDDeltaHatD, double& DfVDDeltaHatQ, double& DfVDDeltaHatP,
                                                STensor3& DfVDkcorpr) const;
    
    virtual void voidCharacteristicsEvolutionLocal(const STensor3& kcor,
                                                const IPNonLocalPorosity *q0, IPNonLocalPorosity* q1,
                                                const double DeltaHatD, const double DeltaHatQ, const double DeltaHatP,
                                                const double DtildefVstarDDeltafV,
                                                double& DyieldfVDDeltafV,
                                                double& DDeltafVDDeltaHatD, double& DDeltafVDDeltaHatQ, double& DDeltafVDDeltaHatP,
                                                STensor3& DDeltafVDKcor) const;
    virtual void voidCharacteristicsEvolution_multipleNonLocalVariables(const STensor3& kcor,
                                                const IPNonLocalPorosity *q0, IPNonLocalPorosity* q1,
                                                std::vector<double>& DyieldfVDNonlocalVars,
                                                STensor3& DyieldfVDKcor, const bool blockVoidGrowth
                                                ) const;
     // Constitutive functions
  protected:
    // local and nonlocal vars depend on nonlocal method
    virtual double getLocalVariable(const IPNonLocalPorosity* q, const int idex) const;
    virtual double& getRefToLocalVariable(IPNonLocalPorosity* q, const int idex) const;
    virtual double getNonLocalVariable(const IPNonLocalPorosity* q, const int idex) const;
    virtual double& getRefToNonLocalVariable(IPNonLocalPorosity* q, const int idex) const;

    // Non-local case
    virtual void constitutive_NonLocalPorosity(
                              const STensor3& F0,         // initial deformation gradient (input @ time n)
                              const STensor3& Fn,         // updated deformation gradient (input @ time n+1)
                              STensor3 &P,                // updated 1st Piola-Kirchhoff stress tensor (output)
                                                          // contains the initial values on input
                              const IPNonLocalPorosity *q0,       // array of initial internal variable
                              IPNonLocalPorosity *q1,             // updated array of internal variable (in ipvcur on output),
                              STensor43 &Tangent,         // constitutive tangents (output)
                              STensor3  &dLocalPorosityDStrain,
                              STensor3  &dStressDNonLocalPorosity,
                              double    &dLocalPorosityDNonLocalPorosity,
                              const bool stiff,     // if true compute the tangents
                              STensor43* elasticTangent = NULL) const;

    // Non-local case
    virtual void constitutive_NonLocalLogarithmPorosity(
                              const STensor3& F0,         // initial deformation gradient (input @ time n)
                              const STensor3& Fn,         // updated deformation gradient (input @ time n+1)
                              STensor3 &P,                // updated 1st Piola-Kirchhoff stress tensor (output)
                                                          // contains the initial values on input
                              const IPNonLocalPorosity *q0,       // array of initial internal variable
                              IPNonLocalPorosity *q1,             // updated array of internal variable (in ipvcur on output),
                              STensor43 &Tangent,         // constitutive tangents (output)
                              STensor3  &dLocalPorosityDStrain,
                              STensor3  &dStressDNonLocalPorosity,
                              double    &dLocalPorosityDNonLocalPorosity,
                              const bool stiff,     // if true compute the tangents
                              STensor43* elasticTangent = NULL) const;

    // Multiple non-local var. cases
    virtual void constitutive_multipleNonLocalVariables(
                              const STensor3& F0,         // initial deformation gradient (input @ time n)
                              const STensor3& Fn,         // updated deformation gradient (input @ time n+1)
                              STensor3 &P,                // updated 1st Piola-Kirchhoff stress tensor (output)
                                                          // contains the initial values on input
                              const IPNonLocalPorosity *q0,       // array of initial internal variable
                              IPNonLocalPorosity *q1,             // updated array of internal variable (in ipvcur on output),
                              STensor43 &Tangent,         // constitutive tangents (output)
                              std::vector<STensor3>  &dLocalVarDStrain,
                              std::vector<STensor3>  &dStressDNonLocalVar,
                              fullMatrix<double>    &dLocalVarDNonLocalVar,
                              const bool stiff,
                              STensor43* elasticTangent = NULL) const;

    // Elasticity-related functions
    virtual double deformationEnergy(const STensor3 &C,const double* T = NULL) const;
    virtual void getDHookTensorDT(STensor43& DHDT, double T) const;
    virtual void elasticPredictor(const STensor3& F1, const STensor3& Fppr,
                                    STensor3& kCor,  STensor3& kcorDev, double& ppr,
                                    STensor43 &dKCordEe, STensor43& dDevKcordEe, STensor3& dpDEe,
                                    STensor3& Fepr, STensor3& Cepr, STensor3& Eepr,
                                    STensor43 &Le, STensor63 &dLe, const bool stiff, const double* T = NULL,
                                    STensor3 *dKCordT = NULL, STensor3* dDevKcordT = NULL, double* dpDT = NULL
                                    ) const;

    // Update internal variables during corrector iterations
    virtual void updateInternalVariables(IPNonLocalPorosity *q1, const IPNonLocalPorosity* q0,
                              const double DeltaHatD, const double DeltaHatQ, const double DeltaHatP,
                              const double* T = NULL) const;

    // Update plastic state at the end of a plastic corrector step (and check for coalescence)
    virtual void updatePlasticState(IPNonLocalPorosity *q1, const IPNonLocalPorosity* q0,  const STensor3& F1,
                              const double DeltaHatD, const double DeltaHatQ, const double DeltaHatP,
                              const STensor3& devKcorpr, const double kcorEqpr, const double pcorpr, const double yield,
                              STensor3& Fe1, STensor3& Ce, STensor3& Ee, STensor43& Le, STensor63& dLe,
                              STensor3& DGNp, STensor3& devDGNp, double& trDGNp, STensor43& Dexp,
                              const double* T=NULL) const;


    virtual void predictorCorrector_NonLocalLogarithmPorosity(const STensor3& F0,         // initial deformation gradient (input @ time n)
                              const STensor3& Fn,         // updated deformation gradient (input @ time n+1)
                              STensor3 &P,                // updated 1st Piola-Kirchhoff stress tensor (output)
                                                          // contains the initial values on input
                              const IPNonLocalPorosity *q0,       // array of initial internal variable
                              IPNonLocalPorosity *q1,             // updated array of internal variable (in ipvcur on output),
                              STensor43 &Tangent,         // constitutive tangents (output)
                              STensor3  &dLocalPorosityDStrain,
                              STensor3  &dStressDNonLocalPorosity,
                              double    &dLocalPorosityDNonLocalPorosity,
                              const bool stiff,                   // if true compute the tangents
                              STensor43& dFedF, STensor3& dFeDNonlocalVar,
                              STensor3*  dStressDT = NULL,
                              double* dLocalPorosityDT = NULL,
                              const double T0 = 0.,const double T = 0.
                             ) const;

     // when considereing nonlocal porosity,
    virtual void computeResidual_NonLocalPorosity(SVector3& res, STensor3& J, const STensor3& F,
                  const double kcorEqpr, const double pcorpr, const double R, const double H,
                  const double yieldfV,
                  const IPNonLocalPorosity *q0, const IPNonLocalPorosity* q1,
                  const double DeltaHatD, const double DeltaHatQ, const double DeltaHatP, // 3 unknonwn
                  const double* T = NULL
                ) const;

    virtual void computeDResidual_NonLocalPorosity(STensor3 &dres0dEpr, STensor3 &dres1dEpr, STensor3 &dres2dEpr,
                                    double &dres0dtildefV, double &dres1dtildefV, double &dres2dtildefV,
                                    const STensor3& F,
                                    const STensor3& devKcorpr, const double kcorEqpr, const double pcorpr, const double R,
                                    const double yieldfV, const STensor3& DyieldfVDKcorpr, const double DyieldfVDtildefV,
                                    const STensor43& DKcorprDEpr, const STensor43& DdevKcorprDEpr, const STensor3& DpcorprDEpr,
                                    const IPNonLocalPorosity *q0, const IPNonLocalPorosity* q1,
                                    const double DeltaHatD, const double DeltaHatQ, const double DeltaHatP,
                                    const double* T = NULL, const double* DRDT= NULL,
                                    const STensor3* DdevKcorprDT = NULL, const double* DpcorprDT = NULL, const STensor3* DKcorprDT = NULL,
                                    double *dres0dT= NULL, double *dres1dT= NULL, double *dres2dT = NULL
                                    ) const;


    // Predictor-corrector and plastic corrector for non-local case
    virtual void predictorCorrector_NonLocalPorosity(const STensor3& F0,         // initial deformation gradient (input @ time n)
                              const STensor3& Fn,         // updated deformation gradient (input @ time n+1)
                              STensor3 &P,                // updated 1st Piola-Kirchhoff stress tensor (output)
                                                          // contains the initial values on input
                              const IPNonLocalPorosity *q0,       // array of initial internal variable
                              IPNonLocalPorosity *q1,             // updated array of internal variable (in ipvcur on output),
                              STensor43 &Tangent,         // constitutive tangents (output)
                              STensor3  &dLocalPorosityDStrain,
                              STensor3  &dStressDNonLocalPorosity,
                              double    &dLocalPorosityDNonLocalPorosity,
                              const bool stiff,                   // if true compute the tangents
                              STensor43& dFedF, STensor3& dFeDtildefV,
                              STensor3*  dStressDT = NULL,
                              double* dLocalPorosityDT = NULL,
                              const double T0 = 0.,const double T = 0.
                             ) const;

   
    virtual bool plasticCorrector_NonLocalPorosity(const STensor3& F1,
                              const double& yieldfV,const double& DyieldfVDtildefV,
                              const STensor3& Kcorpr, const STensor3& devKcorpr,  const double& pcorpr,
                              const IPNonLocalPorosity *q0, IPNonLocalPorosity *q1,
                              STensor3& Fe, STensor3& Ce, STensor3& Ee, STensor43& Le, STensor63& dLe,
                              const bool stiff,
                              const STensor43& DKcorprDEpr, const STensor43& DdevKcorprDEpr, const STensor3& DpcorprDEpr, // predictor elastic tangent
                              STensor43& DKcorDEpr, STensor3& DKcorDtildefV,
                              STensor43& DFpDEpr, STensor3& DFpDtildefV,
                              STensor3& DfVDEpr, double& DfVDtildefV,
                              const double* T = NULL,
                              const STensor3* DKcorprDT = NULL, const STensor3* DdevKcorprDT = NULL, const double* DpcorprDT = NULL,
                              STensor3* DKcorDT= NULL, STensor3* DFpDT = NULL, double* DfVDT = NULL
                            ) const;
    virtual void computeResidualLocal(SVector3& res, STensor3& J, const STensor3& F,
                                    const double kcorEqpr, const double pcorpr, const double R, const double H,
                                    const double yieldfV,  const double DyieldfVDfV,
                                    const IPNonLocalPorosity *q0, const IPNonLocalPorosity* q1,
                                    const double DeltaHatD, const double DeltaHatQ, const double DeltaHatP, // 3 unknonwn
                                    const double DDeltafVDDeltaHatD, const double DDeltafVDDeltaHatQ, const double DDeltafVDDeltaHatP,
                                    const double* T = NULL
                                  ) const;

    virtual void computeDResidualLocal(STensor3 &dres0dEpr, STensor3 &dres1dEpr, STensor3 &dres2dEpr, const STensor3& F,
                                    const STensor3& devKcorpr, const double kcorEqpr, const double pcorpr, const double R,
                                    const double yieldfV, const double DyieldfVDfV,
                                    const STensor3& DDeltafVDEpr, const STensor43& DdevKcorprDEpr, const STensor3& DpcorprDEpr,
                                    const IPNonLocalPorosity *q0, const IPNonLocalPorosity* q1,
                                    const double DeltaHatD, const double DeltaHatQ, const double DeltaHatP,
                                    const double* T = NULL, const double* DRDT= NULL,
                                    const STensor3* DdevKcorprDT = NULL, const double* DpcorprDT = NULL, const double* DDeltafVDT = NULL,
                                    double *dres0dT= NULL, double *dres1dT= NULL, double *dres2dT = NULL
                                    ) const;

    // Predictor-corrector and plastic corrector for local case
    virtual void predictorCorrectorLocal(const STensor3& F0,         // initial deformation gradient (input @ time n)
                              const STensor3& Fn,         // updated deformation gradient (input @ time n+1)
                              STensor3 &P,                // updated 1st Piola-Kirchhoff stress tensor (output)
                                                          // contains the initial values on input
                              const IPNonLocalPorosity *q0,       // array of initial internal variable
                              IPNonLocalPorosity *q1,             // updated array of internal variable (in ipvcur on output),
                              STensor43 &Tangent,         // constitutive tangents (output)
                              const bool stiff           // if true compute the tangents
                             ) const;

    virtual bool plasticCorrectorLocal(const STensor3& F1, const double& tildefVstarPrev,const double& DtildefVstarDDeltafV,
                                    const STensor3& Kcorpr, const STensor3& devKcorpr,  const double& pcorpr,
                                    const IPNonLocalPorosity *q0, IPNonLocalPorosity *q1,
                                    STensor3& Fe, STensor3& Ce, STensor3& Ee, STensor43& Le, STensor63& dLe,
                                    const bool stiff,
                                    const STensor43& DKcorprDEpr, const STensor43& DdevKcorprDEpr, const STensor3& DpcorprDEpr, // predictor elastic tangent
                                    STensor43& DKcorDEpr,
                                    STensor43& DFpDEpr,
                                    const double* T = NULL,
                                    const STensor3* DKcorprDT = NULL, const STensor3* DdevKcorprDT = NULL, const double* DpcorprDT = NULL,
                                    STensor3* DKcorDT= NULL, STensor3* DFpDT = NULL
                                  ) const;


    // Tangent-related functions
    virtual void computePlasticTangentLocal(const IPNonLocalPorosity *q1, const IPNonLocalPorosity* q0,
                const STensor3& devKcorpr, const double kcorEqpr, const double pcorpr,
                const STensor3& DGNp, const STensor3& devDGNp, const double trDGNp, const STensor43& Dexp,
                const double DeltaHatD, const double DeltaHatQ, const double DeltaHatP,
                const STensor43& DKcorprDEpr, const STensor43& DdevKcorprDEpr, const STensor3& DpcorprDEpr,
                const STensor3& DDeltaHatDDEpr, const STensor3& DDeltaHatQDEpr, const STensor3& DDeltaHatPDEpr,
                STensor43& DKcorDEpr, STensor43& DFpDEpr,
                const double* T = NULL,
                const STensor3* DKcorprDT= NULL, const STensor3* DdevKcorprDT= NULL, const double* DpcorprDT= NULL,
                const double* DDeltaHatDDT= NULL, const double* DDeltaHatQDT= NULL, const double* DDeltaHatPDT= NULL,
                STensor3* DKcorDT= NULL, STensor3* DFpDT= NULL
              ) const;

    virtual void computePlasticTangentNonLocal(const IPNonLocalPorosity *q1, const IPNonLocalPorosity* q0,
                const STensor3& devKcorpr, const double kcorEqpr, const double pcorpr,
                const STensor3& DGNp, const STensor3& devDGNp, const double trDGNp, const STensor43& Dexp,
                const double DeltaHatD, const double DeltaHatQ, const double DeltaHatP,
                const STensor43& DKcorprDEpr, const STensor43& DdevKcorprDEpr, const STensor3& DpcorprDEpr,
                const STensor3& DDeltaHatDDEpr, const STensor3& DDeltaHatQDEpr, const STensor3& DDeltaHatPDEpr,
                STensor43& DKcorDEpr, STensor43& DFpDEpr,
                const double DDeltaHatDDtildefV, const double DDeltaHatQDtildefV, const double DDeltaHatPDtildefV,
                STensor3& DKcorDtildefV, STensor3& DFpDtildefV,
                const double* T= NULL, const STensor3* DKcorprDT= NULL, const STensor3* DdevKcorprDT= NULL, const double* DpcorprDT= NULL,
                const double* DDeltaHatDDT= NULL, const double* DDeltaHatQDT= NULL, const double* DDeltaHatPDT= NULL,
                STensor3* DKcorDT= NULL, STensor3* DFpDT= NULL
              ) const;


    virtual void tangentComputationLocal(STensor43& dStressDF,
                                const bool plastic,
                                const STensor3& F,  // current F
                                const STensor3& corKir,  // cor Kir
                                const STensor3& S, //  second PK
                                const STensor3& Fepr, const STensor3& Fppr, // predictor
                                const STensor43& Lpr,
                                const STensor3& Fe, const STensor3& Fp, // corrector
                                const STensor43& L, const STensor63& dL, // corrector value
                                const STensor43& DcorKirDEepr,
                                const STensor43& dFpDEepr,
                                const STensor43& EprToF, const STensor3& Fpinv
                                ) const;
      //----------------------added
    virtual void tangentComputationLocalWithTemperature(STensor43& dStressDF,STensor3& dStressDT,
                                const bool plastic,
                                const STensor3& F,  // current F
                                const STensor3& corKir,  // cor Kir
                                const STensor3& S, //  second PK
                                const STensor3& Fepr, const STensor3& Fppr, // predictor
                                const STensor43& Lpr,
                                const STensor3& Fe, const STensor3& Fp, // corrector
                                const STensor43& L, const STensor63& dL, // corrector value
                                const STensor43& DcorKirDEepr, const STensor3& DcorKirDT, // small strain tangents
                                const STensor43& dFpDEepr, const STensor3& DFpDT,
                                const STensor43& EprToF, const STensor3& Fpinv
                                ) const;

      //----------------------end


    virtual void tangentComputation(STensor43& dFedF, STensor3& dFeDtildefV, STensor43& dStressDF, STensor3& dStressDtildefV,
                                STensor3& DLocalPorositydF,
                                const bool plastic,
                                const STensor3& F,  // current F
                                const STensor3& corKir,  // cor Kir
                                const STensor3& S, //  second PK
                                const STensor3& Fepr, const STensor3& Fppr, // predictor
                                const STensor43& Lpr,
                                const STensor3& Fe, const STensor3& Fp, // corrector
                                const STensor43& L, const STensor63& dL, // corrector value
                                const STensor43& DcorKirDEepr, const STensor3& DcorKirDtildefV,  // small strain tangents
                                const STensor43& dFpDEepr, const STensor3& DFpDtildefV,
                                const STensor3& DfVDEpr,
                                const STensor43& EprToF, const STensor3& Fpinv
                                ) const;

    virtual void tangentComputation(STensor43& dFedF, std::vector<STensor3>& dFeDNonlocalVar, STensor43& dStressDF,
                                std::vector<STensor3>& dStressDNonlocalVar,
                                std::vector<STensor3>& DLocalVardF,
                                const bool plastic,
                                const STensor3& F,  // current F
                                const STensor3& corKir,  // cor Kir
                                const STensor3& S, //  second PK
                                const STensor3& Fepr, const STensor3& Fppr, // predictor
                                const STensor43& Lpr,
                                const STensor3& Fe, const STensor3& Fp, // corrector
                                const STensor43& L, const STensor63& dL, // corrector value
                                const STensor43& DcorKirDEepr, const std::vector<STensor3>& DcorKirDNonlocalVar,  // small strain tangents
                                const STensor43& dFpDEepr, const std::vector<STensor3>& DFpDNonlocalVar,
                                const std::vector<STensor3>& DLocalVarDEpr,
                                const STensor43& EprToF, const STensor3& Fpinv
                                ) const;
  //---------------------------------------------------------------------------------------------------------added
    virtual void tangentComputationWithTemperature(STensor43& dFedF, STensor3& dFeDtildefV, STensor43& dStressDF,
                              STensor3& dStressDtildefVstar, STensor3& dStressDT,
                                STensor3& DLocalPorositydF,
                                const bool plastic,
                                const STensor3& F,  // current F
                                const STensor3& corKir,  // cor Kir
                                const STensor3& S, //  second PK
                                const STensor3& Fepr, const STensor3& Fppr, // predictor
                                const STensor43& Lpr,
                                const STensor3& Fe, const STensor3& Fp, // corrector
                                const STensor43& L, const STensor63& dL, // corrector value
                                const STensor43& DcorKirDEepr, const STensor3& DcorKirDtildefVstar, const STensor3& DcorKirDT, // small strain tangents
                                const STensor43& dFpDEepr, const STensor3& DFpDtildefVstar, const STensor3& DFpDT,
                                const STensor3& DfVDEpr,
                                const STensor43& EprToF, const STensor3& Fpinv
                                ) const;


    virtual void tangentComputationWithTemperature(STensor43& dFedF, std::vector<STensor3>& dFeDNonlocalVar, STensor43& dStressDF,
                                std::vector<STensor3>& dStressDNonlocalVar, STensor3& dStressDT,
                                std::vector<STensor3>& DLocalVardF,
                                const bool plastic,
                                const STensor3& F,  // current F
                                const STensor3& corKir,  // cor Kir
                                const STensor3& S, //  second PK
                                const STensor3& Fepr, const STensor3& Fppr, // predictor
                                const STensor43& Lpr,
                                const STensor3& Fe, const STensor3& Fp, // corrector
                                const STensor43& L, const STensor63& dL, // corrector value
                                const STensor43& DcorKirDEepr, const std::vector<STensor3>& DcorKirDNonlocalVar, const STensor3& DcorKirDT, // small strain tangents
                                const STensor43& dFpDEepr, const std::vector<STensor3>& DFpDNonlocalVar, const STensor3& DFpDT,
                                const std::vector<STensor3>& DLocalVarDEpr,
                                const STensor43& EprToF, const STensor3& Fpinv
                                ) const;

    virtual void computePlasticTangentNonLocal_multipleNonLocalVariables(const IPNonLocalPorosity *q1, const IPNonLocalPorosity* q0,
                const STensor3& devKcorpr, const double kcorEqpr, const double pcorpr,
                const STensor3& DGNp, const STensor3& devDGNp, const double trDGNp, const STensor43& Dexp,
                const double DeltaHatD, const double DeltaHatQ, const double DeltaHatP,
                const STensor43& DKcorprDEpr, const STensor43& DdevKcorprDEpr, const STensor3& DpcorprDEpr,
                const STensor3& DDeltaHatDDEpr, const STensor3& DDeltaHatQDEpr, const STensor3& DDeltaHatPDEpr,
                STensor43& DKcorDEpr, STensor43& DFpDEpr,
                const std::vector<double>& DDeltaHatDDNonLocalVar, const std::vector<double>& DDeltaHatQDNonLocalVar, const std::vector<double>& DDeltaHatPDNonLocalVar,
                std::vector<STensor3>& DKcorDNonLocalVar, std::vector<STensor3>& DFpDNonLocalVar,
                const double* T= NULL, const STensor3* DKcorprDT= NULL, const STensor3* DdevKcorprDT= NULL, const double* DpcorprDT= NULL,
                const double* DDeltaHatDDT= NULL, const double* DDeltaHatQDT= NULL, const double* DDeltaHatPDT= NULL,
                STensor3* DKcorDT= NULL, STensor3* DFpDT= NULL
              ) const;

     // Multiple nonlocal version
    virtual void computeResidual_multipleNonLocalVariables(SVector3& res, STensor3& J, const STensor3& F,
                                  const double kcorEqpr, const double pcorpr, const double R, const double H,
                                  const double yieldfV,
                                  const IPNonLocalPorosity *q0, const IPNonLocalPorosity* q1,
                                  const double DeltaHatD, const double DeltaHatQ, const double DeltaHatP, // 3 unknonwn
                                  const double* T = NULL
                                ) const;


    virtual void computeDResidual_multipleNonLocalVariables(STensor3 &dres0dEpr, STensor3 &dres1dEpr, STensor3 &dres2dEpr,const STensor3& F,
                                  std::vector<double> &dres0dNonlocalVars, std::vector<double> &dres1dNonlocalVars, std::vector<double> &dres2dNonlocalVars,
                                  const STensor3& devKcorpr, const double kcorEqpr, const double pcorpr, const double R,
                                  const double yieldfV, const STensor3& DyieldfVDkcorpr, const std::vector<double>& DyieldfVDNonlocalVars,
                                  const STensor43& DkcorprDEpr, const STensor43& DdevKcorprDEpr, const STensor3& DpcorprDEpr,
                                  const IPNonLocalPorosity *q0, const IPNonLocalPorosity* q1,
                                  const double DeltaHatD, const double DeltaHatQ, const double DeltaHatP,
                                  const double* T = NULL, const double* DRDT= NULL,
                                  const STensor3* DdevKcorprDT = NULL, const double* DpcorprDT = NULL, const STensor3* DKcorprVDT = NULL,
                                  double *dres0dT= NULL, double *dres1dT= NULL, double *dres2dT = NULL
                                  ) const;

    virtual bool plasticCorrector_multipleNonLocalVariables(const STensor3& F1,
                                    const double& yieldfV,const std::vector<double>& DyieldfVDNonlocalVar, const STensor3& DyieldfVDKcorpr,
                                    const STensor3& Kcorpr, const STensor3& devKcorpr,  const double& pcorpr,
                                    const IPNonLocalPorosity *q0, IPNonLocalPorosity *q1,
                                    STensor3& Fe, STensor3& Ce, STensor3& Ee, STensor43& Le, STensor63& dLe,
                                    const bool stiff,
                                    const STensor43& DKcorprDEpr, const STensor43& DdevKcorprDEpr, const STensor3& DpcorprDEpr, // predictor elastic tangent
                                    STensor43& DKcorDEpr, std::vector<STensor3>& DKcorDnonlocalVar,
                                    STensor43& DFpDEpr, std::vector<STensor3>& DFpDnonlocalVar,
                                    std::vector<STensor3>& DlocalVarDEpr, fullMatrix<double>& DlocalVarDnonlocalVar,
                                    const double* T = NULL,
                                    const STensor3* DKcorprDT = NULL, const STensor3* DdevKcorprDT = NULL, const double* DpcorprDT = NULL,
                                    STensor3* DKcorDT= NULL, STensor3* DFpDT = NULL, fullMatrix<double>* DlocalVarDT = NULL
                                  ) const;


    virtual void predictorCorrector_multipleNonLocalVariables(const STensor3& F0,         // initial deformation gradient (input @ time n)
                              const STensor3& Fn,         // updated deformation gradient (input @ time n+1)
                              STensor3 &P,                // updated 1st Piola-Kirchhoff stress tensor (output)
                                                          // contains the initial values on input
                              const IPNonLocalPorosity *q0,       // array of initial internal variable
                              IPNonLocalPorosity *q1,             // updated array of internal variable (in ipvcur on output),
                              STensor43 &Tangent,         // constitutive tangents (output)
                              std::vector<STensor3>  &dLocalVarDF,
                              std::vector<STensor3>  &dStressDNonLocalVar,
                              fullMatrix<double>    &dLocalVarDNonLocalVar,
                              const bool stiff,                   // if true compute the tangents
                              STensor43& dFedF, std::vector<STensor3>& dFeDNonLocalVar,
                              STensor3*  dStressDT = NULL,
                              fullMatrix<double>* dLocalVarDT = NULL,
                              const double T0 = 0.,const double T = 0.
                             ) const;

  public:
    virtual int getNumOfYieldSurfaces() const;
    virtual int getNumOfVoidCharacteristics() const;
    // Y = [yieldfV Chi W gamma]
    virtual double I1J2J3_yieldFunction(const STensor3& kcor, const double R, const IPNonLocalPorosity* q0, const IPNonLocalPorosity* q1, const double* T = NULL,
                                        bool diff=false, STensor3* DFDkcor=NULL, double* DFDR=NULL, std::vector<double>* DFDY=NULL,
                                        bool withThermic=false, double *DFDT=NULL) const=0;
    virtual void I1J2J3_getYieldNormal(STensor3& Np, const STensor3& kcor, const double R, const IPNonLocalPorosity* q0, const IPNonLocalPorosity* q1, const double* T = NULL,
                                        bool diff=false, STensor43* DNpDkcor=NULL, STensor3* DNpDR=NULL, std::vector<STensor3>* DNpDY=NULL,
                                        bool withThermic=false, STensor3* DNpDT=NULL
                                        ) const=0;
  public:
    virtual bool I1J2J3_withPlastic(const STensor3& F, const STensor3& kcor, const double R, const IPNonLocalPorosity* q0, const IPNonLocalPorosity* q1, const double* T = NULL) const;

    virtual void I1J2J3_computeResidual(const STensor3& DeltaEp, const std::vector<double>& DeltaPlasticMult, const double DeltaHatP, // unknown value
                                        const STensor3& sig,  const double R, const IPNonLocalPorosity* q0, const IPNonLocalPorosity* q1,
                                        const std::vector<double>& Y, // void parameters
                                        std::vector<double>& res0, STensor3& res1, double& res2,
                                        bool stiff,
                                        const STensor43& DsigDDeltaEp, const double H,
                                        const std::vector<STensor3>& DYDsig, // stress -dependence
                                        std::vector<STensor3>& Dres0DDeltaEp, STensor43& Dres1DDeltaEp, STensor3& Dres2DDeltaEp,
                                        std::vector<double>& Dres0DDeltaPlasticMult,std::vector<STensor3>& Dres1DDeltaPlasticMult, std::vector<double>& Dres2DDeltaPlasticMult,
                                        std::vector<double>& Dres0DDeltaHatP, STensor3& Dres1DDeltaHatP, double& Dres2DDeltaHatP,
                                        const double* T=NULL) const;
    virtual void I1J2J3_computeDResidual(const STensor3& DeltaEp, const std::vector<double>& DeltaPlasticMult, const double DeltaHatP, // unknown value
                                const STensor3& sig, const double R,  const IPNonLocalPorosity* q0, const IPNonLocalPorosity* q1,
                                const STensor43& DsigDEepr, const double H,
                                const std::vector<double>& Y, // void parameters
                                const std::vector<STensor3>& DYDsig,
                                const std::vector<std::vector<double> >& DYDnonlocalVars,
                                std::vector<std::vector<double> >& Dres0DNonlocalVars, std::vector<STensor3>& Dres0DEepr,
                                std::vector<STensor3>& Dres1DNonlocalVars, STensor43& Dres1DEepr,
                                std::vector<double>& Dres2DNonlocalVars, STensor3& Dres2DEepr,
                                const double* T=NULL, const double* DRDT=NULL, const STensor3* DsigDT=NULL,
                                std::vector<double>* Dres0DT=NULL, STensor3* Dres1DT=NULL, double* Dres2DT=NULL) const;

    virtual bool I1J2J3_plasticCorrector_NonLocal(const STensor3& F1,
                                  const STensor3& Kcorpr,
                                  const IPNonLocalPorosity *q0, IPNonLocalPorosity *q1,
                                  STensor3& Fe, STensor3& Ce, STensor3& Ee, STensor43& Le, STensor63& dLe,
                                  const bool stiff,
                                  const STensor43& DKcorprDEepr, // predictor elastic tangent
                                  STensor43& DKcorDEepr, std::vector<STensor3>& DKcorDnonlocalVar,
                                  STensor43& DFpDEepr, std::vector<STensor3>& DFpDnonlocalVar,
                                  std::vector<STensor3>& DlocalVarDEepr, fullMatrix<double>& DlocalVarDnonlocalVar,
                                  const double* T,
                                  const STensor3* DKcorprDT,
                                  STensor3* DKcorDT, STensor3* DFpDT, fullMatrix<double>* DlocalVarDT
                                  ) const;
    virtual void I1J2J3_copyInternalVariableForElasticState(IPNonLocalPorosity* q1, const IPNonLocalPorosity* q0, const double T) const;
    virtual void I1J2J3_predictorCorrector_NonLocal(const STensor3& F0,         // initial deformation gradient (input @ time n)
                            const STensor3& F1,         // updated deformation gradient (input @ time n+1)
                            STensor3 &P,                // updated 1st Piola-Kirchhoff stress tensor (output)
                            const IPNonLocalPorosity *ipvprev,       // array of initial internal variable
                            IPNonLocalPorosity *ipvcur,             // updated array of internal variable (in ipvcur on output),
                            STensor43 &Tangent,         // constitutive tangents (output)
                            std::vector<STensor3>  &dLocalVarDF,
                            std::vector<STensor3>  &dStressDNonLocalVar,
                            fullMatrix<double>    &dLocalVarDNonLocalVar,
                            const bool stiff,                   // if true compute the tangents
                            STensor43& dFedF, std::vector<STensor3>& dFeDNonLocalVar,
                            STensor3*  dStressDT=NULL,
                            fullMatrix<double>* dLocalVarDT=NULL,
                            const double T0 =0,const double T=0
                           ) const;

    virtual void I1J2J3_constitutive_NonLocal(
                              const STensor3& F0,         // initial deformation gradient (input @ time n)
                              const STensor3& F1,         // updated deformation gradient (input @ time n+1)
                              STensor3 &P,                // updated 1st Piola-Kirchhoff stress tensor (output)
                              const IPNonLocalPorosity *ipvprev,       // array of initial internal variable
                              IPNonLocalPorosity *ipvcur,             // updated array of internal variable (in ipvcur on output),
                              STensor43 &Tangent,         // constitutive tangents (output)
                              std::vector<STensor3>  &dLocaldVarDStrain,
                              std::vector<STensor3>  &dStressDNonLocalVar,
                              fullMatrix<double>    &dLocalVarDNonLocalVar,
                              const bool stiff,            // if true compute the tangents
                             STensor43* elasticTangent = NULL) const;
    virtual void I1J2J3_voidCharacteristicsEvolution_NonLocal(const STensor3& sig,
                                                  const IPNonLocalPorosity *q0, IPNonLocalPorosity* q1,
                                                  bool stiff = false,
                                                  std::vector<double>* DyieldfVDNonlocalVars = NULL,
                                                  STensor3* DyieldfVDsig = NULL
                                                  ) const;

  protected:
    virtual void I1J2J3_updateInternalVariables(IPNonLocalPorosity *q1, const IPNonLocalPorosity* q0,
                            const STensor3& DeltaEp, const double DeltaHatP, const double* T=NULL) const;

    virtual void I1J2J3_getVoidCharacteristics_NonLocal(std::vector<double>& Y,
                                                std::vector<STensor3>& DYDsig,
                                                std::vector<std::vector<double> >& DYDnonlocalVars,
                                                const IPNonLocalPorosity *q0,  const IPNonLocalPorosity* q1,
                                                const std::vector<double>& DyieldfVDNonlocalVars,
                                                const STensor3& DyieldfVDsig) const;

    virtual void I1J2J3_getResAndJacobian(fullVector<double>& res, fullMatrix<double>& Jac,
                                const std::vector<double>& res0, const STensor3& res1, const double& res2,
                                const std::vector<STensor3>& Dres0DDeltaEp, const STensor43& Dres1DDeltaEp, const STensor3& Dres2DDeltaEp,
                                const std::vector<double>& Dres0DDeltaPlasticMult, const std::vector<STensor3>& Dres1DDeltaPlasticMult, const std::vector<double>& Dres2DDeltaPlasticMult,
                                const std::vector<double>& Dres0DDeltaHatP, const STensor3& Dres1DDeltaHatP,const double& Dres2DDeltaHatP) const;

    virtual void I1J2J3_updatePlasticState(IPNonLocalPorosity *q1, const IPNonLocalPorosity* q0,  const STensor3& F1, const STensor3& kcor,
                            const double R,  const STensor3& DeltaEp, const double DeltaHatP, const double H,
                            STensor3& Fe1, STensor3& Ce, STensor3& Ee, STensor43& Le, STensor63& dLe, STensor43& DFpDDeltaEp, const double* T=NULL) const;
    virtual void I1J2J3_getNonLocalVars(std::vector<double>& Vars,  const IPNonLocalPorosity* q) const;
 #endif // SWIG
};



class combinedPorousLaw : public mlawNonLocalPorosity
{
  #ifndef SWIG
  protected:
    std::vector<mlawNonLocalPorosity*> _multipleLaws; // cache for all relying laws


  public:    // Default constructor
    combinedPorousLaw(const int num,const double E,const double nu, const double rho, const double fVinitial,
                        const J2IsotropicHardening &j2IH, const CLengthLaw &cLLaw,
                        const double tol=1.e-8, const bool matrixbyPerturbation = false, const double pert = 1e-8);
    // Constructors & destructor
    combinedPorousLaw(const combinedPorousLaw &source);
    virtual ~combinedPorousLaw();
    virtual materialLaw* clone() const =0;
    virtual matname getType() const =0;

    // Function of materialLaw
    virtual void createIPState(IPStateBase* &ips,const bool* state_=NULL,const MElement *ele=NULL, const int nbFF_=0, const IntPt *GP=NULL, const int gpt = 0) const{};
    virtual void createIPState(IPNonLocalPorosity *ivi, IPNonLocalPorosity *iv1, IPNonLocalPorosity *iv2) const {};
    virtual void createIPVariable(IPNonLocalPorosity* &ipv, const MElement *ele, const int nbFF_, const IntPt *GP, const int gpt) const {};
    virtual void createIPVariable(IPNonLocalPorosity *&ipv) const {};
    virtual void initLaws(const std::map<int,materialLaw*> &maplaw){}; // this law is initialized so nothing to do
    virtual void setMacroSolver(const nonLinearMechSolver* sv);

    // Option settings
    virtual void setStressFormulation(const int fm);
    virtual void activateVoidNucleationDuringCoalescence(const bool flg);
    virtual void setNonLocalMethod(const int i);
    virtual void setOrderForLogExp(const int newOrder);
    virtual void setViscosityLaw(const viscosityLaw& vico);
    virtual void setScatterredInitialPorosity(double f0min, double f0max);
    virtual void clearCLengthLaw();
    virtual void setCLengthLaw(const CLengthLaw& clength);

    virtual void setShearPorosityGrowthFactor(const double k);
    virtual void setStressTriaxialityFunction_kw(const scalarFunction& fct);
    virtual void setLodeFunction(const scalarFunction& fct);
    virtual void setFailureTolerance(const double NewFailureTol);
    virtual void setCorrectedRegularizedFunction(const scalarFunction& fct);
    virtual void setPostBlockedDissipationBehavior(const int method);

    virtual void setSubStepping(const bool fl, const int maxNumStep);

    // Options setting -- temperature part
    virtual void setThermomechanicsCoupling(const bool fl);
    virtual void setReferenceT(const double referenceT);
    virtual void setThermalExpansionCoefficient(const double alp);
    virtual void setReferenceThermalConductivity(const double KK);
    virtual void setTemperatureFunction_ThermalConductivity(const scalarFunction& Tfunc);
    virtual void setReferenceCp(const double Cp);
    virtual void setTemperatureFunction_Cp(const scalarFunction& Tfunc);
    virtual void setThermalEstimationPreviousConfig(const bool flag);
    virtual void setTaylorQuineyFactor(const double f);
    //
    virtual void setTemperatureFunction_rho(const scalarFunction& Tfunc);
    virtual void setTemperatureFunction_E(const scalarFunction& Tfunc);
    virtual void setTemperatureFunction_nu(const scalarFunction& Tfunc);
    virtual void setTemperatureFunction_alp(const scalarFunction& Tfunc);

  public:
    static void LpNorm(const double power, const fullVector<double>& X, double& F,
                                     bool diff = false, fullVector<double>* DFDX = NULL,
                                     bool hess = false, fullMatrix<double>* DDFDXDX = NULL);

    void regulizationYieldFunctionBase(const double power, double p, const double s, double& f,
                          bool diff=false, double* dfdp=NULL , double* dfds=NULL,
                          bool hess=false, double* ddfdpdp=NULL, double* ddfdpds=NULL,double* ddfdsdp=NULL,double* ddfdsds=NULL) const;
    void regulizationYieldFunctionBase(const double power, double p, const double s, const double t, double& f,
                          bool diff=false, double* dfdp=NULL , double* dfds=NULL, double *dfdt = NULL,
                          bool hess=false, double* ddfdpdp=NULL, double* ddfdpds=NULL, double* ddfdpdt=NULL,
                          double* ddfdsdp=NULL,double* ddfdsds=NULL,double* ddfdsdt=NULL,
                          double* ddfdtdp=NULL,double* ddfdtds=NULL,double* ddfdtdt=NULL) const;
  #endif //SWIG
};


#endif //MLAWNONLOCALPOROUS_H_
