// C++ Interface: material law
// Description: SMP law
// Author:  <M. Pareja>, (C) 2017
// Copyright: See COPYING file that comes with this distribution

#include "mlawPhenomenologicalSMP.h"
#include <math.h>
#include <iostream>
#include "MInterfaceElement.h"
#include "ipPhenomenologicalSMP.h"
//#include "STensor63.h"
#include <vector>



mlawPhenomenologicalSMP::mlawPhenomenologicalSMP(const int num, const double rho, const double alpha, const double beta, const double gamma,
                                                 const double t0, const double Kxg,const double Kyg, const double Kzg, const double Kxr,
                                                 const double Kyr, const double Kzr,const double TaylorQuiney, const double c, const double cp,
                                                 const double cg, const double cr, const double Tg,const double Tc,const double xi, const double w1,
                                                 const double w2, const double Eg, const double Er, const double nug, const double nur,
                                                 const double alphag, const double alphar,const double alpha_trans, const double tau_y0)
    :materialLaw(num,true),_alpha(alpha),_beta(beta),_gamma(gamma),_Kxg(Kxg),_Kyg(Kyg),_Kzg(Kzg), _Kxr(Kxr),_Kyr(Kyr),_Kzr(Kzr),_t0(t0),
     _order(-1),_TaylorQuiney(TaylorQuiney),_nug(nug),_nur(nur), _c(c), _cp(cp),_cg(cg),_cr(cr),_Tg(Tg),_Tc(Tc),_xi(xi),_w1(w1),_w2(w2),
     _Gg(Eg/2./(1.+nug)),_Gr(Er/2./(1.+nur)),_Kg(Eg/3./(1.-2.*nug)),_Kr(Er/3./(1.-2.*nur)),_alphag(alphag),_alphar(alphar),
     _alpha_trans(alpha_trans),_Tonsetg((6.90675/(2*w1))+Tg),_Tonsetr((-6.90675/(2*w2))+Tc),_I2(1.),_I4(1.,1.)

{
    _j2IH = new PerfectlyPlasticJ2IsotropicHardening(num,tau_y0);
 
    // by default, no temperature dependence
    _temFunc_Sy0 =                      new constantScalarFunction(1.);
    _temFunc_H =                        new constantScalarFunction(1.);
    _temFunc_Kg =                       new constantScalarFunction(1.);
    _temFunc_Gg =                       new constantScalarFunction(1.);
    _temFunc_ThermalConductivity_g =    new constantScalarFunction(1.);
    _temFunc_cp_g =                     new constantScalarFunction(1.);
    _temFunc_alphag =                   new constantScalarFunction(1.);
    _temFunc_Kr =                       new constantScalarFunction(1.);
    _temFunc_Gr =                       new constantScalarFunction(1.);
    _temFunc_ThermalConductivity_r =    new constantScalarFunction(1.);
    _temFunc_cp_r =                     new constantScalarFunction(1.);
    _temFunc_alphar =                   new constantScalarFunction(1.);

    double _Gg2=_Gg+_Gg;
    double _lambdag=(Eg*nug)/(1.+nug)/(1.-2.*nug);
    double _lambdar=(Er*nur)/(1.+nur)/(1.-2.*nur);

    for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
            for (int k=0; k<3; k++){
                for (int l=0; l<3; l++){
                    _I4dev(i,j,k,l) = 0.5*(_I2(i,k)*_I2(j,l)+_I2(i,l)*_I2(j,k)) - _I2(i,j)*_I2(k,l)/3.;
                }
            }
        }
    }

    HookeTensor(_Kg,_Gg,_C0g);
    HookeTensor(_Kr,_Gr,_C0r);

    STensor3 R;		//3x3 rotation matrix
    double co1,co2,co3,si1,si2,si3;
    double s1c2, c1c2;
    double _pi(3.14159265359);
    double fpi = _pi/180.;
    co1 = cos(_alpha*fpi);
    si1 = sin(_alpha*fpi);
    co2 = cos(_beta*fpi);
    si2 = sin(_beta*fpi);
    co3 = cos(_gamma*fpi);
    si3 = sin(_gamma*fpi);
    s1c2 = si1*co2;
    c1c2 = co1*co2;

    R(0,0) = co3*co1 - s1c2*si3;
    R(0,1) = co3*si1 + c1c2*si3;
    R(0,2) = si2*si3;
    R(1,0) = -si3*co1 - s1c2*co3;
    R(1,1) = -si3*si1 + c1c2*co3;
    R(1,2) = si2*co3;
    R(2,0) = si1*si2;
    R(2,1) = -co1*si2;
    R(2,2) = co2;

    double a=25.e-5;
    STensor3 alphaDilatation;
    alphaDilatation(0,0)= a;
    alphaDilatation(1,1)= a;
    alphaDilatation(2,2)= a;
    for(int i=0; i<3; i++){
        for(int j=0; j<3; j++){
            _alphaDilatation(i,j)=0.;
            for(int m=0; m<3; m++){
                for(int n=0; n<3; n++){
                    _alphaDilatation(i,j)+=R(m,i)*R(n,j)*alphaDilatation(m,n);
                }
            }
        }
    }

    STensor43 K_;

    K_*=0.;
    K_(0,0,0,0) = _lambdag + _Gg2;
    K_(1,1,0,0) = _lambdag;
    K_(2,2,0,0) = _lambdag;
    K_(0,0,1,1) = _lambdag;
    K_(1,1,1,1) = _lambdag + _Gg2;
    K_(2,2,1,1) = _lambdag;
    K_(0,0,2,2) = _lambdag;
    K_(1,1,2,2) = _lambdag;
    K_(2,2,2,2) = _lambdag + _Gg2;
    K_(1,0,1,0) = _Gg;
    K_(2,0,2,0) = _Gg;
    K_(0,1,0,1) = _Gg;
    K_(2,1,2,1) = _Gg;
    K_(0,2,0,2) = _Gg;
    K_(1,2,1,2) = _Gg;
    K_(0,1,1,0) = _Gg;
    K_(0,2,2,0) = _Gg;
    K_(1,0,0,1) = _Gg;
    K_(1,2,2,1) = _Gg;
    K_(2,0,0,2) = _Gg;
    K_(2,1,1,2) = _Gg;


    for(int i=0; i<3; i++){
        for(int j=0; j<3; j++){
            _Stiff_alphaDilatation(i,j)=0.;
            for(int k=0; k<3; k++){
                for(int l=0; l<3; l++){
                    _Stiff_alphaDilatation(i,j)+=K_(i,j,k,l)*_alphaDilatation(k,l);
                }
            }
        }
    }
    STensor3 kg, kr;
    // to be unifom in DG3D q= k' gradT with k'=-k instead of q=-kgradT
    kg(0,0)  = -_Kxg;
    kg(1,1)  = -_Kyg;
    kg(2,2)  = -_Kzg;
    kr(0,0)  = -_Kxr;
    kr(1,1)  = -_Kyr;
    kr(2,2)  = -_Kzr;
    for(int i=0; i<3; i++){
        for(int j=0; j<3; j++){
            _kg(i,j)=0.;
            for(int m=0; m<3; m++){
                for(int n=0; n<3; n++){
                    _kg(i,j)+=R(m,i)*R(n,j)*kg(m,n);
                }
            }
        }
    }
    for(int i=0; i<3; i++){
        for(int j=0; j<3; j++){
            _kr(i,j)=0.;
            for(int m=0; m<3; m++){
                for(int n=0; n<3; n++){
                    _kr(i,j)+=R(m,i)*R(n,j)*kr(m,n);
                }
            }
        }
    }
}


mlawPhenomenologicalSMP::mlawPhenomenologicalSMP(const int num,const double rho,const J2IsotropicHardening &j2IH,const double alpha,
                                                 const double beta, const double gamma,const double t0, const double Kxg,const double Kyg,
                                                 const double Kzg, const double Kxr, const double Kyr, const double Kzr, const double TaylorQuiney,
                                                 const double c, const double cp, const double cg, const double cr, const double Tg,const double Tc,
                                                 const double xi, const double w1,const double w2,const double Eg, const double Er, const double nug,
                                                 const double nur, const double alphag, const double alphar,const double alpha_trans,const double tau_y0)
    :materialLaw(num,true),_alpha(alpha),_beta(beta),_gamma(gamma),_Kxg(Kxg),_Kyg(Kyg),_Kzg(Kzg), _Kxr(Kxr),_Kyr(Kyr),_Kzr(Kzr),_t0(t0),
     _order(-1),_TaylorQuiney(TaylorQuiney),_nug(nug),_nur(nur), _c(c), _cp(cp),_cg(cg),_cr(cr),_Tg(Tg),_Tc(Tc),_xi(xi),_w1(w1),_w2(w2),
    _Gg(Eg/2./(1.+nug)),_Gr(Er/2./(1.+nur)),_Kg(Eg/3./(1.-2.*nug)),_Kr(Er/3./(1.-2.*nur)),_alphag(alphag),_alphar(alphar),
    _alpha_trans(alpha_trans),_Tonsetg((6.90675/(2*_w1))+Tg),_Tonsetr((-6.90675/(2*_w2))+Tc),_I2(1.),_I4(1.,1.)

{
    _j2IH=j2IH.clone();

    double _Gg2=_Gg+_Gg;
    double _lambdag=(Eg*nug)/(1.+nug)/(1.-2.*nug);
    double _lambdar=(Er*nur)/(1.+nur)/(1.-2.*nur);

    for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
            for (int k=0; k<3; k++){
                for (int l=0; l<3; l++){
                    _I4dev(i,j,k,l) = 0.5*(_I2(i,k)*_I2(j,l)+_I2(i,l)*_I2(j,k)) - _I2(i,j)*_I2(k,l)/3.;
                }
            }
        }
    }

    HookeTensor(_Kg,_Gg,_C0g);
    HookeTensor(_Kr,_Gr,_C0r);

    STensor3 R; //3x3 rotation matrix
    double co1,co2,co3,si1,si2,si3;
    double s1c2, c1c2;
    double _pi(3.14159265359);
    double fpi = _pi/180.;
    co1 = cos(_alpha*fpi);
    si1 = sin(_alpha*fpi);
    co2 = cos(_beta*fpi);
    si2 = sin(_beta*fpi);
    co3 = cos(_gamma*fpi);
    si3 = sin(_gamma*fpi);
    s1c2 = si1*co2;
    c1c2 = co1*co2;
    R(0,0) = co3*co1 - s1c2*si3;
    R(0,1) = co3*si1 + c1c2*si3;
    R(0,2) = si2*si3;
    R(1,0) = -si3*co1 - s1c2*co3;
    R(1,1) = -si3*si1 + c1c2*co3;
    R(1,2) = si2*co3;
    R(2,0) = si1*si2;
    R(2,1) = -co1*si2;
    R(2,2) = co2;

    double a=25.e-5;
    STensor3 alphaDilatation;
    alphaDilatation(0,0)= a;
    alphaDilatation(1,1)= a;
    alphaDilatation(2,2)= a;
    for(int i=0; i<3; i++){
        for(int j=0; j<3; j++){
            _alphaDilatation(i,j)=0.;
            for(int m=0; m<3; m++){
                for(int n=0; n<3; n++){
                    _alphaDilatation(i,j)+=R(m,i)*R(n,j)*alphaDilatation(m,n);
                }
            }
        }
    }

    STensor43 K_;
    K_*=0.;
    K_(0,0,0,0) = _lambdag + _Gg2;
    K_(1,1,0,0) = _lambdag;
    K_(2,2,0,0) = _lambdag;
    K_(0,0,1,1) = _lambdag;
    K_(1,1,1,1) = _lambdag + _Gg2;
    K_(2,2,1,1) = _lambdag;
    K_(0,0,2,2) = _lambdag;
    K_(1,1,2,2) = _lambdag;
    K_(2,2,2,2) = _lambdag + _Gg2;
    K_(1,0,1,0) = _Gg;
    K_(2,0,2,0) = _Gg;
    K_(0,1,0,1) = _Gg;
    K_(2,1,2,1) = _Gg;
    K_(0,2,0,2) = _Gg;
    K_(1,2,1,2) = _Gg;
    K_(0,1,1,0) = _Gg;
    K_(0,2,2,0) = _Gg;
    K_(1,0,0,1) = _Gg;
    K_(1,2,2,1) = _Gg;
    K_(2,0,0,2) = _Gg;
    K_(2,1,1,2) = _Gg;

    for(int i=0; i<3; i++){
        for(int j=0; j<3; j++){
            _Stiff_alphaDilatation(i,j)=0.;
            for(int k=0; k<3; k++){
                for(int l=0; l<3; l++){
                    _Stiff_alphaDilatation(i,j)+=K_(i,j,k,l)*_alphaDilatation(k,l);
                }
            }
        }
    }
    STensor3 kg, kr;
    // to be unifom in DG3D q= k' gradT with k'=-k instead of q=-kgradT
    kg(0,0)  = -_Kxg;
    kg(1,1)  = -_Kyg;
    kg(2,2)  = -_Kzg;
    kr(0,0)  = -_Kxr;
    kr(1,1)  = -_Kyr;
    kr(2,2)  = -_Kzr;
    for(int i=0; i<3; i++){
        for(int j=0; j<3; j++){
            _kg(i,j)=0.;
            for(int m=0; m<3; m++){
                for(int n=0; n<3; n++){
                    _kg(i,j)+=R(m,i)*R(n,j)*kg(m,n);
                }
            }
        }
    }
    for(int i=0; i<3; i++){
        for(int j=0; j<3; j++){
            _kr(i,j)=0.;
            for(int m=0; m<3; m++){
                for(int n=0; n<3; n++){
                    _kr(i,j)+=R(m,i)*R(n,j)*kr(m,n);
                }
            }
        }
    }
}



mlawPhenomenologicalSMP::mlawPhenomenologicalSMP(const mlawPhenomenologicalSMP &source)
: materialLaw(source),_alpha(source._alpha),_beta(source._beta),_gamma(source._gamma) ,_Kxg(source. _Kxg),_Kyg(source._Kyg),
_Kzg(source._Kzg),_Kxr(source._Kxr),_Kyr(source._Kyr),_Kzr(source._Kzr), _kg(source._kg) , _kr(source._kr), _t0(source._t0),
_k0(source._k0), _Stiff_alphaDilatation(source._Stiff_alphaDilatation),_alphaDilatation(source._alphaDilatation), _order(source._order),
_w(source._w), _w1(source._w1),_w2(source._w2),_TaylorQuiney(source._TaylorQuiney), _c(source._c), _cp(source._cp),_cg(source._cg),
_cr(source._cr),_Tg(source._Tg),_Tc(source._Tc),_xi(source._xi),_Tonsetg(source._Tonsetg),_Tonsetr(source._Tonsetr),_Gg(source._Gg),
_Gr(source._Gr),_Kg(source._Kg),_Kr(source._Kr),_alphag(source._alphag),_alphar(source._alphar),_alpha_trans(source._alpha_trans),
_nur(source._nur),_nug(source._nug),_C0r(source._C0r),_C0g(source._C0g),_I2(1.),_I4(1.,1.)
{
    _j2IH = NULL;
    if (source._j2IH != NULL)
        _j2IH = source._j2IH->clone();

    _temFunc_Sy0 = NULL;
    if (source._temFunc_Sy0!= NULL)
        _temFunc_Sy0 = source._temFunc_Sy0->clone();

    _temFunc_H = NULL;
    if (source._temFunc_H != NULL)
    {_temFunc_H = source._temFunc_H->clone();}

    _temFunc_ThermalConductivity_g = NULL;
    if (source._temFunc_ThermalConductivity_g!= NULL)
    {_temFunc_ThermalConductivity_g = source._temFunc_ThermalConductivity_g->clone();}

    _temFunc_cp_g = NULL;
    if (source._temFunc_cp_g != NULL)
    {_temFunc_cp_g = source._temFunc_cp_g->clone();}

    _temFunc_alphag = NULL;
    if (source._temFunc_alphag != NULL)
    {_temFunc_alphag = source._temFunc_alphag->clone();}

    _temFunc_Kg = NULL;
    if (source._temFunc_Kg != NULL)
    {_temFunc_Kg = source._temFunc_Kg->clone();}

    _temFunc_Gg = NULL;
    if (source._temFunc_Gg!=NULL)
    {_temFunc_Gg = source._temFunc_Gg->clone();}

    _temFunc_ThermalConductivity_r = NULL;
    if (source._temFunc_ThermalConductivity_r!= NULL)
    {_temFunc_ThermalConductivity_r = source._temFunc_ThermalConductivity_r->clone();}

    _temFunc_cp_r = NULL;
    if (source._temFunc_cp_r != NULL)
    {_temFunc_cp_r = source._temFunc_cp_r->clone();}

    _temFunc_alphar = NULL;
    if (source._temFunc_alphar != NULL)
    {_temFunc_alphar = source._temFunc_alphar->clone();}

    _temFunc_Kr = NULL;
    if (source._temFunc_Kr != NULL)
    {_temFunc_Kr = source._temFunc_Kr->clone();}

    _temFunc_Gr = NULL;
    if (source._temFunc_Gr!=NULL)
    {_temFunc_Gr = source._temFunc_Gr->clone();}
}



mlawPhenomenologicalSMP& mlawPhenomenologicalSMP::operator=(const  materialLaw &source)
{
    materialLaw::operator=(source);
    const mlawPhenomenologicalSMP* src =static_cast<const mlawPhenomenologicalSMP*>(&source);

    _t0=                    src->_t0;
    _Stiff_alphaDilatation= src->_Stiff_alphaDilatation;
    _alphaDilatation=       src->_alphaDilatation;
    _order=                 src->_order;
    _w=                     src->_w;
    _w1=                    src->_w1;
    _w2=                    src->_w2;
    _TaylorQuiney=          src->_TaylorQuiney;
    _c=                     src->_c;
    _cp=                    src->_cp;
    _cg=                    src->_cg;
    _cr=                    src->_cr;
    _Tg=                    src->_Tg;
    _Tc=                    src->_Tc;
    _xi=                    src->_xi;
    _Tonsetg=               src->_Tonsetg;
    _Tonsetr=               src->_Tonsetr;
    _Kg=                    src->_Kg;
    _Kr=                    src->_Kr;
    _Gg=                    src->_Gg;
    _Gr=                    src->_Gr;
    _alphag=                src->_alphag;
    _alphar=                src->_alphar;
    _alpha_trans=           src->_alpha_trans;
    _alpha=                 src->_alpha;
    _beta=                  src->_beta;
    _gamma=                 src->_gamma;
    _Kxg=                   src->_Kxg;
    _Kyg=                   src->_Kyg;
    _Kzg=                   src->_Kzg;
    _Kxr=                   src->_Kxr;
    _Kyr=                   src->_Kyr;
    _Kzr=                   src->_Kzr;
    _kg=                    src->_kg;
    _kr=                    src->_kr;
    _k0=                    src->_k0;
    _C0r=                   src->_C0r;
    _C0g=                   src->_C0g;
    _nur=                   src->_nur;
    _nug=                   src->_nug;
}



double mlawPhenomenologicalSMP::soundSpeed() const
{
    //use glassy
    double K = _Kg;
    double mu = _Gg;
    double nu = (3.*K-2.*mu)/2./(3.*K+mu);
    double E = 2.*mu*(1.+nu);

    double factornu = (1.-nu)/((1.+nu)*(1.-2.*nu));
    return sqrt(E*factornu/_rho);
}



double mlawPhenomenologicalSMP::getExtraDofStoredEnergyPerUnitField(double zg) const
{
    double specificheat= zg*_cg+(1.-zg)*_cr;
    return specificheat;
}



void mlawPhenomenologicalSMP::createIPState(IPStateBase* &ips, const bool* state_, const MElement *ele, const int nbFF_,
                                            const IntPt *GP, const int gpt) const
{
    //bool inter=true;
    //const MInterfaceElement *iele = dynamic_cast<const MInterfaceElement*>(ele);
    //if(iele==NULL) inter=false;
    IPVariable* ipvi = new IPPhenomenologicalSMP(_t0,_Tg,_Tonsetg,_Tonsetr,_alpha_trans,_j2IH);
    IPVariable* ipv1 = new IPPhenomenologicalSMP(_t0,_Tg,_Tonsetg,_Tonsetr,_alpha_trans,_j2IH);
    IPVariable* ipv2 = new IPPhenomenologicalSMP(_t0,_Tg,_Tonsetg,_Tonsetr,_alpha_trans,_j2IH);
    if(ips != NULL) delete ips;
    ips = new IP3State(state_,ipvi,ipv1,ipv2);
}



void mlawPhenomenologicalSMP::HookeTensor(const double KT, const double GT, STensor43& C) const
{
    for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
            for (int k=0; k<3; k++){
                for (int l=0; l<3; l++){
                    C(i,j,k,l) = KT*_I2(i,j)*_I2(k,l) + 2.*GT*_I4dev(i,j,k,l);
                }
            }
        }
    }
}



void mlawPhenomenologicalSMP::invHookeTensor(const double KT, const double GT, STensor43& C) const
{
    double a = -KT/(2.*GT*(3.*KT+2.*GT));
    double b = 1./(2.*GT);

    for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
            for (int k=0; k<3; k++){
                for (int l=0; l<3; l++){
                    C(i,j,k,l) = a*_I2(i,j)*_I2(k,l) + b*_I4dev(i,j,k,l);
                }
            }
        }
    }
}



void mlawPhenomenologicalSMP::evaluate_reference_temperatures(const IPPhenomenologicalSMP *q0, IPPhenomenologicalSMP *q1, double zg,
                                                                double T_t, const double T0, double T, double Tdot, double& lambda,
                                                                double dzgdT, double *dlambdaTdT) const
{
    double &Tig=                q1->getRefToTig();
    double &Tir=                q1->getRefToTir();
    double &lambdaT=            q1->getRefToLambdaT();
    double &alphaig=            q1->getRefToAlphaig();
    double &Tmin=               q1->getRefToTmin();
    const double &Tig0=         q0->getConstRefToTig();
    const double &Tir0=         q0->getConstRefToTir();
    const double &lambdaT0=     q0->getConstRefToLambdaT();
    const double &alphaig0=     q0->getConstRefToAlphaig();
    double zg0=                 q0->getzg();
    const double &Tmin0=        q0->getConstRefToTmin();
    const double &T_t0=         q0->getConstRefToTt();

    double dalpharT, dalphagT;
    double alpharT = getalpha_rubbery(T, &dalpharT);
    double alphagT = getalpha_glassy(T, &dalphagT);

    if(T<Tmin0) {Tmin=T;}
    else        {Tmin=Tmin;}

    if(Tdot>0.001)  {zg0=1./(1.+exp(2.*_w1*(T0-_Tg)));}
    if(Tdot<0.001)  {zg0=1./(1.+exp(2.*_w2*(T0-_Tc)));}
    else{zg0=zg0;}

    double zg_inc;
    if(Tdot > 0.001)            {zg_inc= zg - zg0;}
    else if (Tdot < -0.001)     {zg_inc= zg - zg0;}
    else                        {zg_inc= 0.;}

    double zg_min=1./(1.+exp(2.*_w1*(Tmin-_Tc)));
    double _alpha_trans_t;
    if (Tdot < -0.001)  {_alpha_trans_t = _alpha_trans;}
    else                {_alpha_trans_t = zg_min*_alpha_trans;}

    double _alpha_exp;
    if(_Tg<Tmin)        {_alpha_exp=0.;}
    if(_Tc<=Tmin<=_Tg)  {_alpha_exp=(alpharT - alphagT)*(Tmin-_Tg);}
    if(Tmin<_Tc)        {_alpha_exp=(alpharT - alphagT)*(_Tc-_Tg);}

    if (Tdot > 0.001)   {alphaig= _alpha_trans_t + _alpha_exp;}
    else                {alphaig= _alpha_trans_t;}

    Tir= T;
    Tig= T;
    double zg_Av= (zg + zg0)/2.;
    lambda = lambdaT0;
    lambda += (1.-zg_Av)*(alpharT*(T-Tir0))   +   zg_Av*(alphagT*(T-Tig0))   +   zg_inc*alphaig;
    lambdaT= lambda;

    //std::cout<<"   _alpha_trans_t==="<<_alpha_trans_t<<"   _alpha_exp==="<<_alpha_exp<<"   alphaig==="<<alphaig
    //std::cout<<"   lambdaT==="<<lambdaT
    //<<"   zg_inc==="<<zg_inc<<"   zg0==="<<zg0<<"   zg==="<<zg
    //<<"   Tdot==="<<Tdot<<"   Tmin==="<<Tmin<<"   T0==="<<T0<<"   T==="<<T<<std::endl;

    if(*dlambdaTdT!=NULL)
    {
        if (Tdot > 0.001)
        {
        *dlambdaTdT = -dzgdT*alpharT*(T-Tir0) + (1.-zg_Av)*dalpharT*(T-Tir0) + (1.-zg_Av)*alpharT;
        *dlambdaTdT += dzgdT*alphagT*(T-Tig0) +      zg_Av*dalphagT*(T-Tig0) +      zg_Av*alphagT;
        *dlambdaTdT += dzgdT*alphaig + zg_inc*(dalpharT - dalphagT)*(_Tc-_Tg);
        }
        else
        {
        *dlambdaTdT = -dzgdT*alpharT*(T-Tir0) + (1.-zg_Av)*dalpharT*(T-Tir0) + (1.-zg_Av)*alpharT;
        *dlambdaTdT += dzgdT*alphagT*(T-Tig0) +      zg_Av*dalphagT*(T-Tig0) +      zg_Av*alphagT;
        *dlambdaTdT += dzgdT*alphaig;
        }
    }
}



void mlawPhenomenologicalSMP::getUf(const double T,STensor3& Uf,const STensor3& Fm, const STensor3& tauer) const
{
    double GrT = getGr(T);
    double KrT = getKr(T,GrT);

    static STensor3 C, U, E, Uinv, R;
    STensorOperation::multSTensor3FirstTranspose(Fm, Fm, C);
    STensorOperation::logSTensor3(C,_order,E);
    E*=0.5;
    STensorOperation::expSTensor3(E,_order,U);
    STensorOperation::inverseSTensor3(U,Uinv);
    STensorOperation::multSTensor3(Fm,Uinv,R);

    STensor43 C0inv;
    invHookeTensor(KrT,GrT,C0inv);

    static STensor3 aux, Uer;
    STensorOperation::multSTensor43STensor3(C0inv,tauer,aux);
    STensorOperation::expSTensor3(aux,_order,Uer);

    STensorOperation::multSTensor3(Uer,R,Uf);
}



void mlawPhenomenologicalSMP::constitutive(const STensor3& F0,const STensor3& Fn,STensor3 &P,
        const IPVariable *q0, IPVariable *q1,STensor43 &Tangent,const bool stiff, STensor43* elasticTangent) const
{
}



void mlawPhenomenologicalSMP::constitutive(
        const STensor3& F0,             // initial deformation gradient (input @ time n)
        const STensor3& Fn,             // updated deformation gradient (input @ time n+1)
                                        // updated 1st Piola-Kirchhoff stress tensor (output)
        STensor3 &P,                    // contains the initial values on input
        const IPVariable *q0i,          // array of initial internal variable
        IPVariable *q1i,                // updated array of internal variable (in ipvcur on output),
        STensor43 &Tangent,             // constitutive tangents (output)
        const double T0,
        double T,
        const SVector3 &gradT,
        SVector3 &fluxT,
        STensor3 &dPdT,
        STensor3 &dqdgradT,
        SVector3 &dqdT,
        STensor33 &dqdF,
        const bool stiff,               // if true compute the tangents
        double &wth,
        double &dwthdT,
        STensor3 &dwthdF,
        double&  mechSource,
        double& dmechSourcedT,
        STensor3 & dmechSourcedF,
        STensor43* elasticTangent) const
{
    const IPPhenomenologicalSMP *q0 = dynamic_cast<const IPPhenomenologicalSMP *>(q0i);
    IPPhenomenologicalSMP *q1       = dynamic_cast<IPPhenomenologicalSMP *>(q1i);
    if(elasticTangent!=NULL) Msg::Error("mlawPhenomenologicalSMP  elasticTangent not defined");

    //initilisation of variables to zero
    wth=0.;
    dwthdT=0.;
    mechSource=0.;
    dmechSourcedT=0.;
    STensorOperation::zero(dwthdF);
    STensorOperation::zero(dmechSourcedF);
    STensorOperation::zero(dqdF);
    STensorOperation::zero(dqdT);
    STensorOperation::zero(dqdgradT);
    STensorOperation::zero(dPdT);
    STensorOperation::zero(fluxT);
    static STensor3 Pg,Pr;
    STensorOperation::zero(Pg);
    STensorOperation::zero(Pr);
    static STensor43 Tangentr,Tangentg;
    static STensor3 dPrdT,dPgdT;
    STensorOperation::zero(Tangentg);
    STensorOperation::zero(dPgdT);
    STensorOperation::zero(Tangentr);
    STensorOperation::zero(dPrdT);
    double dgammataueq=0.;
    double dgammataueqdT=0.;
    static STensor3 dgammataueqdF;
    STensorOperation::zero(dgammataueqdF);
    static STensor3 Fm, Ginv;
    STensorOperation::zero(Ginv);
    STensorOperation::zero(Fm);

    double dh= getTimeStep();
    const double &Tdot0=    q0->getConstRefToTdot();
    double &Tdot=           q1->getRefToTdot();

    double &T_t =   q1->getRefToTt();
    double T_t0 =   q0->getConstRefToTt();
    double w  =     q1->getRefTow();
    double w0 =     q0->getConstRefTow();

    if (T == T0)
    {
        Tdot=0.;
    }
    if (T < T0)
    {
        Tdot= (T-T0)/dh;
        T_t0= _Tc;
        w0=   _w1;
    }
    if (T > T0)
    {
        Tdot= (T-T0)/dh;
        T_t0= _Tg;
        w0=   _w2;
    }

    STensor3 &G= q1->getRefToFth();
    STensorOperation::zero(G);
    static double lambda_th;
    static double dTtdT;

    get_Tt(Tdot0, T_t0, &T_t, &dTtdT, &w, w0, _w1, _w2);
    dTtdT*= 1./dh;

    double &zg= q1->getRefTozg();
    double dzgdT;
    zg= get_zg(T, w, T_t, &dzgdT, &dTtdT);

    double dlambdathdT;
    evaluate_reference_temperatures(q0, q1, zg, T_t, T0, T, Tdot, lambda_th, dzgdT, &dlambdathdT);

    G  = _I2;
    G *= lambda_th;
    STensorOperation::inverseSTensor3(G,Ginv);
    STensorOperation::multSTensor3(Fn,Ginv,Fm);


    if (zg<=0.00001)
    {constitutive1(F0,Fn,Fm,q0,q1,Tangentr,dPrdT, T0, T,stiff,dh,Pr,dgammataueq,dgammataueqdT,dgammataueqdF);}

    if ((zg>0.00001)and(zg<0.99999))
    {constitutive2(F0,Fn,Fm,q0,q1,Tangentr,Tangentg,dPrdT,dPgdT,T0,T,stiff,Pr,Pg,dh,lambda_th,dgammataueq,dgammataueqdT,dgammataueqdF);}

    if (zg>=0.99999)
    {constitutive3(F0,Fn,Fm,q0,q1,Tangentg,dPgdT, T0,T,stiff,Pg,dh,lambda_th,dgammataueq,dgammataueqdT,dgammataueqdF);}

    P =Pr;
    P+=Pg;

    //Wth & mecSource
    double dcgT, dcrT;
    double cgT = getcp_glassy(T,&dcgT);
    double crT = getcp_rubbery(T,&dcrT);

    double cp=zg*cgT+(1.-zg)*crT;
    static STensor3 dcpdF;
    STensorOperation::zero(dcpdF);
    double dcpdT  = dzgdT*(cgT-crT);
    dcpdT += (zg*dcgT + (1-zg)*dcrT);

    if(dh>0.)
    {
        wth=-cp*(T-T0)/dh;
        mechSource  = (dgammataueq)/dh;
    }

    if(dh>0.)
    {
        dwthdT=-dcpdT*(T-T0)/dh-cp/dh;
        dmechSourcedT+=(dgammataueqdT)/dh;
        for(int K = 0; K< 3; K++){
            for(int i = 0; i< 3; i++){
                dwthdF(K,i)=-dcpdF(K,i)*(T-T0)/dh;
                dmechSourcedF(K,i)+=(dgammataueqdF(K,i))/dh;
            }
        }
    }

    static STensor3 Fninv;
    STensorOperation::inverseSTensor3(Fn, Fninv);
    double J= Fn.determinant();
    static STensor3 _Kref,_k;
    STensorOperation::zero(_Kref);
    STensorOperation::zero(_k);

    STensor3 kgT, krT, dkgT, dkrT;

    getThermalConductivity_glassy(T,kgT,&dkgT);
    getThermalConductivity_rubbery(T,krT,&dkrT);

    _k=zg*kgT;
    _k+=(1-zg)*krT;

    for(int K = 0; K< 3; K++){
        for(int L = 0; L< 3; L++){
            for(int i = 0; i< 3; i++){
                for(int j = 0; j< 3; j++){
                    _Kref(K,L) += Fninv(K,i)*_k(i,j)*Fninv(L,j);
                }
            }
        }
    }

    STensorOperation::multSTensor3SVector3(_Kref, gradT, fluxT);
    fluxT*=J;

    static STensor3 dkdT;
    STensorOperation::zero(dkdT);
    if(stiff)
    {
        dPdT=dPgdT;
        dPdT+=dPrdT;
        Tangent=Tangentg;
        Tangent+=Tangentr;

        dqdgradT = _Kref;
        dqdgradT*=J;

        dkdT  = kgT;
        dkdT -= krT;
        dkdT *= dzgdT;
        dkdT += (zg*dkgT);
        dkdT += (1-zg)*dkrT;
        STensor3 dKrefdT;


        for(int K = 0; K< 3; K++){
            for(int m = 0; m< 3; m++){
                for(int N = 0; N< 3; N++){
                    for(int L = 0; L< 3; L++){
                        dKrefdT(K,m) += Fninv(K,N)*dkdT(N,L)*Fninv(L,m);
                        dqdF(K,m,N) -= Fninv(K,m)*_Kref(N,L)*gradT(L)*J;
                        dqdF(K,m,N) -= _Kref(K,N)*Fninv(L,m)*gradT(L)*J;
                        dqdF(K,m,N) += _Kref(K,L)*gradT(L)*Fninv(N,m)*J;
                    }
                }
            }
        }

        STensorOperation::multSTensor3SVector3(dKrefdT, gradT, dqdT);
        dqdT*=J;

    }
// Tangent.print("tangent=");
}



void mlawPhenomenologicalSMP::constitutive1(const STensor3& F0, const STensor3& Fn,const STensor3& Fm, const IPPhenomenologicalSMP *q0,
                                            IPPhenomenologicalSMP *q1,STensor43& Tangentr, STensor3& dPrdT, const double T0, double T, bool stiff, double dh,
                                            STensor3& Pr,double& dgammataueq, double& dgammataueqdT,STensor3& dgammataueqdF) const
{
    // Rubbery phase //
    // Initialisation of internal variables and material properties
    const STensor3 &Fp0=q0->getConstRefToFp();
    const STensor3 &Ff0=q0->getConstRefToFf();
    STensor3 &Fp=q1->getRefToFp();
    STensor3 &Ff=q1->getRefToFf();
    STensor3 &Fpg=q1->getRefToFpg();
    STensor3 &Uf=q1->getRefToUf();
    STensor3 &G=q1->getRefToFth();

    double &zg=q1->getRefTozg();
    double &Tig=q1->getRefToTig();
    double &Tir=q1->getRefToTir();
    double &alphaig=q1->getRefToAlphaig();
    double &Tdot=q1->getRefToTdot();
    const double &Tdot0=q0->getConstRefToTdot();

    double GrT = getGr(T);
    double KrT = getKr(T,GrT);
    double alpharT = getalpha_rubbery(T);

// Internal variables evolution (1)

    Fpg = _I2;
    Fp=Fp0;

// Resulting variables

    static STensor3 Fer, Eer,tauer;
    static STensor43 dlnCer;
    static STensor63 ddlnCer;
    STensorOperation::zero(dlnCer);
    STensorOperation::zero(ddlnCer);
    STensorOperation::zero(Eer);

    computePr(KrT,GrT,Fn,Fp,G,Fer,Eer,tauer,dlnCer,ddlnCer,zg,Pr);
    computeMechSource_rubbery(q0,q1,Fer,tauer,dlnCer,dgammataueq);

// Internal variables evolution: Fixity (2)

    static STensor3 R;

    getUf(T,Uf,Fer,tauer);

    static STensor3 subs1;
    STensorOperation::zero(subs1);
    subs1 = Fm;
    subs1-= _I2; //Uf
    Ff  = subs1;
    Ff *= _c;
    Ff += _I2; //Uf

// Tangent operators

    if(stiff)
    {
        double &Tig=q1->getRefToTig();
        double &Tir=q1->getRefToTir();
        double &alphaig=q1->getRefToAlphaig();
        tangentrubbery(Fn, Fer, q0, q1, tauer, dlnCer, ddlnCer, Eer, T0, T, Tdot, dh, Tir, Tig, alphaig, zg, G, dgammataueq, dgammataueqdT,
                       dgammataueqdF,dPrdT,Tangentr);
    }

}



void mlawPhenomenologicalSMP::constitutive2(const STensor3& F0, const STensor3& Fn,const STensor3& Fm, const IPPhenomenologicalSMP *q0,
                                            IPPhenomenologicalSMP *q1, STensor43& Tangentr, STensor43& Tangentg, STensor3& dPrdT,
                                            STensor3& dPgdT, const double T0, double T, bool stiff, STensor3& Pr, STensor3& Pg, const double dh,
                                            double lambda_th, double& dgammataueq,double& dgammataueqdT, STensor3& dgammataueqdF) const
{
// Mixed phase //
// Initialisation of internal variables and material properties
    const STensor3 &Fp0=q0->getConstRefToFp();
    STensor3 &Fp=q1->getRefToFp();
    const STensor3 &Ff0=q0->getConstRefToFf();
    STensor3 &Ff=q1->getRefToFf();
    const STensor3 &Fpg0=q0->getConstRefToFpg();
    STensor3 &Fpg=q1->getRefToFpg();
    STensor3 &G=q1->getRefToFth();

    double &zg=q1->getRefTozg();
    double &Tig=q1->getRefToTig();
    double &Tir=q1->getRefToTir();
    double &alphaig=q1->getRefToAlphaig();
    double &Tdot=q1->getRefToTdot();
    const double &Tdot0=q0->getConstRefToTdot();

    double GrT = getGr(T);
    double KrT = getKr(T,GrT);
    double GgT = getGg(T);
    double KgT = getKg(T,GgT);
    double alpharT = getalpha_rubbery(T);
    double alphagT = getalpha_glassy(T);

// Internal variables evolution

    Ff=Ff0;
    Fp=Fp0;

    if(T>T0)
    {
        Fpg=Fpg0;
        static STensor3 Feg, Eeg,taueg;
        static STensor43 dlnCeg;
        static STensor63 ddlnCeg;

        // Resulting Variables (Glassy)

        computePg(KgT,GgT,Fn,Ff,Fpg,G,Feg,Eeg,taueg,dlnCeg,ddlnCeg,zg,Pg);
        computeMechSource_glassy(q0,q1,Feg,taueg,dlnCeg,dgammataueq);

        // Tangent Operators Glassy

        double Eegpreq=0;
        static STensor3 N;
        N = _I2;
        double Delta_gamma=0.;
        double H_gamma=0.;
        double taueq=0.;
        static STensor43 dexpDgammaN;
        dexpDgammaN=_I4;

        if(stiff)
        {
            double &Tig=q1->getRefToTig();
            double &Tir=q1->getRefToTir();
            double &alphaig=q1->getRefToAlphaig();
            tangentglassy(Fn, Fm, q0, q1, dexpDgammaN, Feg, dlnCeg, Eeg, N, Eegpreq, Delta_gamma, zg, H_gamma, T0, T, Tdot, dh, Tir, Tig,
                          alphaig, taueq, Feg, Eeg, taueg, dlnCeg, ddlnCeg, G, dgammataueq, dgammataueqdT, dgammataueqdF, dPgdT, Tangentg);
        }
    }
    if(T<=T0)
    {
        predictorCorrector(F0,Fn,Fm,q0,q1,Pg,dh, T0,T,Tdot,stiff,lambda_th,dgammataueq,dgammataueqdT,dgammataueqdF,dPgdT,Tangentg);
    }

//Resulting variables (Rubbery)

    static STensor3 Fer, Eer,tauer;
    static STensor43 dlnCer;
    static STensor63 ddlnCer;
    STensorOperation::zero(dlnCer);
    STensorOperation::zero(ddlnCer);
    STensorOperation::zero(Eer);

    computePr(KrT,GrT,Fn,Fp,G,Fer,Eer,tauer,dlnCer,ddlnCer,zg,Pr);
    computeMechSource_rubbery(q0,q1,Fer,tauer,dlnCer,dgammataueq);

//Tangent operators (Rubbery)

    if(stiff)
    {
        double &Tig=q1->getRefToTig();
        double &Tir=q1->getRefToTir();
        double &alphaig=q1->getRefToAlphaig();
        tangentrubbery(Fn, Fer, q0, q1, tauer, dlnCer, ddlnCer, Eer, T0, T, Tdot, dh, Tir, Tig, alphaig, zg, G, dgammataueq, dgammataueqdT,
                       dgammataueqdF, dPrdT, Tangentr);
    }

}



void mlawPhenomenologicalSMP::constitutive3(const STensor3& F0, const STensor3& Fn,const STensor3& Fm, const IPPhenomenologicalSMP *q0,
                                            IPPhenomenologicalSMP *q1,STensor43& Tangentg, STensor3& dPgdT, const double T0, double T, bool stiff,
                                            STensor3& Pg, const double dh,double lambda_th, double& dgammataueq,double& dgammataueqdT,
                                            STensor3& dgammataueqdF) const
{
//Glassy phase//
// Initialisation of internal variables and material properties

    const STensor3 &Fp0=q0->getConstRefToFp();
    STensor3 &Fp=q1->getRefToFp();
    const STensor3 &Ff0=q0->getConstRefToFf();
    STensor3 &Ff=q1->getRefToFf();
    const STensor3 &Fpg0=q0->getConstRefToFpg();
    STensor3 &Fpg=q1->getRefToFpg();
    STensor3 &G=q1->getRefToFth();

    double &zg=q1->getRefTozg();
    double &Tig=q1->getRefToTig();
    double &Tir=q1->getRefToTir();
    double &alphaig=q1->getRefToAlphaig();
    double &Tdot=q1->getRefToTdot();

//Internal variables evolution

    Ff=Ff0;

    predictorCorrector(F0,Fn,Fm,q0,q1,Pg,dh, T0,T,Tdot,stiff,lambda_th,dgammataueq,dgammataueqdT,dgammataueqdF,dPgdT,Tangentg);

    static STensor3 subs3;
    STensorOperation::zero(subs3);
    subs3=Ff;
    subs3+=Fpg;
    subs3-=(2.*_I2);
    Fp  = subs3;
    Fp *= _cp;
    Fp +=_I2;
    if(!checkDeltaEp(Fp0,Fp))
    {
        Fp=Fp0;
    }
}



bool mlawPhenomenologicalSMP::checkDeltaEp(const STensor3& Fp0, const STensor3& Fp) const
{
    STensor3 Cp0,Cp1,Ep0,Ep1;
    STensor43 dlnCp0,dlnCp1;
    STensorOperation::zero(Cp0);
    STensorOperation::zero(Ep0);
    STensorOperation::zero(dlnCp0);
    STensorOperation::zero(Cp1);
    STensorOperation::zero(Ep1);
    STensorOperation::zero(dlnCp1);
    double Epnorm0;
    double Epnorm1;
    for(int i=0; i<2; i++)
    {
        if(i==0)
        {
            STensorOperation::multSTensor3FirstTranspose(Fp0,Fp0,Cp0);
            STensorOperation::logSTensor3(Cp0,_order,Ep0,&dlnCp0);
            Ep0*=0.5;
            Epnorm0=Ep0.norm2();
        }

        if(i==1)
        {
            STensorOperation::multSTensor3FirstTranspose(Fp,Fp,Cp1);
            STensorOperation::logSTensor3(Cp1,_order,Ep1,&dlnCp1);
            Ep1*=0.5;
            Epnorm1=Ep1.norm2();
        }
    }

    if(Epnorm1>Epnorm0)
    {
        return true;
    }

    else
    {
        return false;
    }
}



void mlawPhenomenologicalSMP::predictorCorrector(const STensor3& F0, const STensor3& Fn,const STensor3& Fm, const IPPhenomenologicalSMP *q0,
                                                 IPPhenomenologicalSMP *q1, STensor3& Pg, const double dh, const double T0,double T,double Tdot, bool stiff,
                                                 const double lambda_th,double& dgammataueq, double& dgammataueqdT,STensor3& dgammataueqdF,
                                                 STensor3& dPgdT, STensor43& Tangentg) const
{
    // Initialisation of variables
    const double &gamma0=q0->getConstRefToEquivalentPlasticDefo();
    double &gamma1=q1->getRefToEquivalentPlasticDefo();

    gamma1=gamma0;

    STensor3 &Ff=q1->getRefToFf();
    const STensor3 &Fpg0=q0->getConstRefToFpg();
    STensor3 &Fpg=q1->getRefToFpg();
    STensor3 &G=q1->getRefToFth();
    double &zg=q1->getRefTozg();

    double &Tig=q1->getRefToTig();
    double &Tir=q1->getRefToTir();
    double &alphaig=q1->getRefToAlphaig();

    double GgT = getGg(T);
    double KgT = getKg(T,GgT);
    double alphagT = getalpha_glassy(T);

    // Conpute predictor

    static STensor3 Fegpr, Eegpr,tauegpr;
    static STensor43 dlnCegpr;
    static STensor63 ddlnCegpr;

    computePg(KgT,GgT,Fn,Ff,Fpg0,G,Fegpr,Eegpr,tauegpr,dlnCegpr,ddlnCegpr,zg,Pg);
    computeMechSource_glassy(q0,q1,Fegpr,tauegpr,dlnCegpr,dgammataueq);

    double taueqpr=sqrt(1.5*tauegpr.dev().dotprod()); //predictor equivalent stress

    double Eegpreq;
    Eegpreq=sqrt((2./3.)*Eegpr.dev().dotprod());

    // Assume final values = predictor

    static STensor3 Feg, Eeg, devtaueg, taueg;
    static STensor43 dlnCeg;
    static STensor63 ddlnCeg;
    Feg=Fegpr;
    Eeg=Eegpr;
    taueg=tauegpr;
    dlnCeg=dlnCegpr;
    ddlnCeg=ddlnCegpr;

    static STensor3 N;
    static STensor43 dexpDgammaN;
    N = Eegpr.dev();
    N*= 1/Eegpreq;

    // Test plasticity //

    _j2IH->hardening(gamma0, q0->getConstRefToIPJ2IsotropicHardening(), gamma1, q1->getRefToIPJ2IsotropicHardening());
    double tau_y0T = _j2IH->getYield0()*_temFunc_Sy0->getVal(T);
    double tau_y   = tau_y0T + (q1->getConstRefToIPJ2IsotropicHardening().getR() - _j2IH->getYield0())*_temFunc_H->getVal(T);
    tau_y  *= zg;
    double H_gamma = q1->getConstRefToIPJ2IsotropicHardening().getDR()*_temFunc_H->getVal(T);

    double f=taueqpr-tau_y;

    double Delta_gamma=0.;
    double H_T=0.;
    double taueq=0.;

    if(f>0.)
    {

        //Newton-Raphson

        double tolNR=1.e-6;

        //Initialization of variables

        int ite = 0, maxite = 1000;
        double eps_f=fabs(f/taueqpr);

        //Iterative process

        while(eps_f>tolNR or ite<1)
        {
            //Compute Jacobian

            double j1=1./((3.*zg*GgT)+H_gamma);
            double deps=j1*f;
            gamma1+=deps;
            Eeg-=N*deps;
            devtaueg = Eeg.dev();
            devtaueg  *= (2.*GgT*zg);

            taueq=sqrt(1.5*devtaueg.dotprod());
            _j2IH->hardening(gamma0, q0->getConstRefToIPJ2IsotropicHardening(), gamma1, q1->getRefToIPJ2IsotropicHardening());
            tau_y  = tau_y0T + (q1->getConstRefToIPJ2IsotropicHardening().getR() - _j2IH->getYield0())*_temFunc_H->getVal(T);
            tau_y *= zg;
            H_gamma = q1->getConstRefToIPJ2IsotropicHardening().getDR()*_temFunc_H->getVal(T);

            f=taueq-tau_y;
            eps_f=fabs(f/taueqpr);
            if(ite > maxite)
            {
                Msg::Error("SMP itereration %d, residual %f,time step %f ", ite, eps_f,dh);
                break;
            }
            ite++;
        }
        Delta_gamma=gamma1-gamma0;
        static STensor3 DepsN;
        DepsN = N;
        DepsN *= Delta_gamma;
        static STensor3 expDgammaN;
        STensorOperation::expSTensor3(DepsN,_order,expDgammaN,&dexpDgammaN);

        //Update of variables

        STensorOperation::multSTensor3(expDgammaN,Fpg0,Fpg);
        computePg(KgT,GgT,Fn,Ff,Fpg,G,Feg,Eeg,taueg,dlnCeg,ddlnCeg,zg,Pg);
    }
    else
    {
        gamma1=gamma0;
    }

    computeMechSource_glassy(q0,q1,Feg,taueg,dlnCeg,dgammataueq);

    //Plastic dissipation term

    dgammataueq += taueq*Delta_gamma*_TaylorQuiney;

    //Tangent operators

    if(stiff)
    {
        tangentglassy(Fn, Fm, q0, q1, dexpDgammaN, Fegpr, dlnCegpr, Eegpr, N, Eegpreq, Delta_gamma, zg, H_gamma, T0, T, Tdot, dh, Tir, Tig,
                      alphaig, taueq, Feg, Eeg, taueg, dlnCeg, ddlnCeg, G, dgammataueq, dgammataueqdT, dgammataueqdF, dPgdT, Tangentg);
    }
}



void mlawPhenomenologicalSMP::computeMechSource_rubbery(const IPPhenomenologicalSMP *q0, IPPhenomenologicalSMP *q1,const STensor3& Fer,const STensor3& tauer,const STensor43& dlnCer, double& dgammataueq)  const
{
    const STensor3 &Fp0=q0->getConstRefToFp();
    STensor3 &Fp=q1->getRefToFp();
    static STensor3 Fpinv, DFp_int, DFp, Mer, Ser, Cer;
    STensorOperation::inverseSTensor3(Fp,Fpinv);

    DFp_int  = Fp;
    DFp_int -= Fp0;
    STensorOperation::multSTensor3(DFp_int,Fpinv,DFp);

    STensorOperation::multSTensor3FirstTranspose(Fer, Fer, Cer);
    STensorOperation::multSTensor3STensor43(tauer,dlnCer,Ser);
    STensorOperation::multSTensor3(Cer,Ser,Mer);

    static double Dr_recovery;
    Dr_recovery = STensorOperation::doubledot(Mer,DFp);

    dgammataueq += Dr_recovery;
}



void mlawPhenomenologicalSMP::computeMechSource_glassy(const IPPhenomenologicalSMP *q0, IPPhenomenologicalSMP *q1,const STensor3& Feg,const STensor3& taueg,const STensor43& dlnCeg, double& dgammataueq)  const
{
    const STensor3 &Ff0=q0->getConstRefToFf();
    STensor3 &Ff=q1->getRefToFf();
    STensor3 &Fpg=q1->getRefToFf();
    static STensor3 Ffinv, Fpginv, DFf_int, DFf, Meg, Seg, Ceg, aux1gf, aux2gf,aux1gth, aux2gth,aux3gth, aux4gth;
    STensorOperation::inverseSTensor3(Ff,Ffinv);
    STensorOperation::inverseSTensor3(Fpg,Fpginv);

    DFf_int  = Ff;
    DFf_int -= Ff0;
    STensorOperation::multSTensor3(DFf_int,Ffinv,DFf);

    STensorOperation::multSTensor3(Fpg,DFf,aux1gf);
    STensorOperation::multSTensor3(aux1gf,Fpginv,aux2gf);
    STensorOperation::multSTensor3FirstTranspose(Feg, Feg, Ceg);
    STensorOperation::multSTensor3STensor43(taueg,dlnCeg,Seg);
    STensorOperation::multSTensor3(Ceg,Seg,Meg);

    static double Dg_fixity;
    Dg_fixity = STensorOperation::doubledot(Meg,aux2gf);

    dgammataueq += Dg_fixity;
}



void mlawPhenomenologicalSMP::tangentrubbery(const STensor3& Fn,const STensor3& Fer,const IPPhenomenologicalSMP *q0,IPPhenomenologicalSMP *q1,
                                             const STensor3& tauer,const STensor43& dlnCer,const STensor63& ddlnCer,const STensor3& Eer, const double T0,
                                             const double T,const double Tdot,const double dh,const double Tir,const double Tig,
                                             const double alphaig,const double zg, const STensor3& G,double& dgammataueq, double& dgammataueqdT,
                                             STensor3& dgammataueqdF, STensor3& dPrdT, STensor43& Tangentr) const
{
    double dKrT,dGrT,dalpharT,dalphagT;
    double GrT = getGr(T,&dGrT);
    double KrT = getKr(T,GrT,&dKrT,&dGrT);
    double alpharT = getalpha_rubbery(T,&dalpharT);
    double alphagT = getalpha_glassy(T,&dalphagT);
    STensor43 C0r;
    HookeTensor(KrT,GrT,C0r);

    const STensor3 &Fp0=q0->getConstRefToFp();
    STensor3 &Fp=q1->getRefToFp();
    static STensor3 Fpinv, Ginv, Ser;
    static STensor43 dFerdF,dtauerdF;
    static STensor3 dtauerdT, dFerdT;
    STensorOperation::inverseSTensor3(Fp,Fpinv);
    STensorOperation::inverseSTensor3(G,Ginv);
    STensorOperation::multSTensor3STensor43(tauer,dlnCer,Ser);
    STensorOperation::zero(dFerdF);
    STensorOperation::zero(dtauerdF);
    STensorOperation::zero(dtauerdT);
    STensorOperation::zero(dFerdT);


    static double dTtdT;

    double T_t  = q1->getRefToTt();
    double T_t0 = q0->getConstRefToTt();
    double w  = q1->getRefTow();
    double w0 = q0->getConstRefTow();
    const double &Tdot0=q0->getConstRefToTdot();

    get_Tt(Tdot0,T_t0,&T_t, &dTtdT,&w,w0,_w1,_w2);
    dTtdT *= 1./dh;
    double dzgdT;
    get_zg(T,w,T_t,&dzgdT,&dTtdT);

    double lambda_th, dlambdathdT;
    evaluate_reference_temperatures(q0, q1, zg, T_t, T0, T, q1->getConstRefToTdot(), lambda_th, dzgdT, &dlambdathdT);

    static STensor3 dGdT;
    static STensor3 dGinvdT;
    dGdT  = _I2;
    dGdT *= dlambdathdT;
    STensorOperation::zero(dGinvdT);

    for (int R=0; R<3; R++){
        for (int O=0; O<3; O++){
            for (int L=0; L<3; L++){
                for (int A=0; A<3; A++){
                    dGinvdT(R,A) -=Ginv(R,O)*dGdT(O,L)*Ginv(L,A);
                }
            }
        }
    }

    dFerdF *= 0.;
    dFerdT *= 0.;
    for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
            for (int k=0; k<3; k++){
                for (int l=0; l<3; l++){
                    dFerdT(i,j) = Fn(i,l)*dGinvdT(l,k)*Fpinv(k,j);
                    dFerdF(i,j,i,k) += Ginv(k,l)*Fpinv(l,j);
                }
            }
        }
    }

    static STensor63 DLDF;
    if (_order != 1){
        for (int i=0; i<3; i++){
            for (int j=0; j<3; j++){
                for (int k=0; k<3; k++){
                    for (int l=0; l<3; l++){
                        for (int p=0; p<3; p++){
                            for (int q=0; q<3; q++){
                                DLDF(i,j,k,l,p,q) = 0.;
                                for (int r=0; r<3; r++){
                                    for (int s=0; s<3; s++){
                                        for (int a=0; a<3; a++){
                                            DLDF(i,j,k,l,p,q) += ddlnCer(i,j,k,l,r,s)*2.*Fer(a,r)*dFerdF(a,s,p,q);
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else
    {
        DLDF *= 0.;
    }

    static STensor43 EerToF, dtauerdEer;
    dtauerdEer=(C0r);
    dtauerdEer*=(1.-zg);
    for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
            for (int k=0; k<3; k++){
                for (int l=0; l<3; l++){
                    EerToF(i,j,k,l) = 0.;
                    for (int p=0; p<3; p++){
                        for (int q=0; q<3; q++){
                            for (int h=0; h<3; h++){
                                EerToF(i,j,k,l) += dlnCer(i,j,p,q)*Fer(k,p)*Ginv(l,h)*Fpinv(h,q);
                            }
                        }
                    }
                }
            }
        }
    }

    STensorOperation::multSTensor43(dtauerdEer,EerToF,dtauerdF);


    STensor3 devtauer;
    double p;
    STensorOperation::decomposeDevTr(tauer,devtauer,p);
    p=p/3.;

    dtauerdT *= 0.;
    for( int E=0; E<3; E++){
        dtauerdT(E,E) += p*dKrT/KrT;
        for( int Z=0; Z<3; Z++){
            dtauerdT(E,Z) += devtauer(E,Z)*dGrT/GrT;
            for( int N=0; N<3; N++){
                for( int O=0; O<3; O++){
                    dtauerdT(E,Z) -= dzgdT*C0r(E,Z,N,O)*Eer(N,O);
                }
            }
        }
    }

    static STensor43 DSDF; // S = corKir:L
    static STensor3 DSDT;
    STensorOperation::zero(DSDT);

    for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
            for (int k=0; k<3; k++){
                for (int l=0; l<3; l++){
                    DSDF(i,j,k,l) = 0.;
                    DSDT(i,j)+= dtauerdT(k,l)*dlnCer(k,l,i,j);
                    for (int r=0; r<3; r++){
                        for (int s=0; s<3; s++){
                            DSDF(i,j,k,l) += dtauerdF(r,s,k,l)*dlnCer(r,s,i,j) ;
                            if (_order != 1){
                                DSDF(i,j,k,l) += tauer(r,s)*DLDF(r,s,i,j,k,l);
                            }
                        }
                    }
                }
            }
        }
    }


// compute mechanical & thermal tangent

    static STensor43 dPrdFm;
    static STensor3 dauxrdT;
    STensorOperation::zero(dPrdFm);
    STensorOperation::zero(dauxrdT);

    Tangentr *= 0.;
    dPrdT    *= 0.;
    for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
            for (int k=0; k<3; k++){
                for (int l=0; l<3; l++){
                    for (int p=0; p<3; p++){
                        dPrdT(i,j) += (dFerdT(i,k)*Ser(k,p)*Ginv(l,p)*Fpinv(j,l));
                        dPrdT(i,j) += (Fer(i,k)*DSDT(k,p)*Ginv(l,p)*Fpinv(j,l));
                        dPrdT(i,j) += (Fer(i,k)*Ser(k,p)*dGinvdT(l,p)*Fpinv(j,l));
                        for (int q=0; q<3; q++){
                            for (int h=0; h<3; h++)
                            {
                                Tangentr(i,j,p,q) += (dFerdF(i,k,p,q)*Ser(k,h)*Ginv(l,h)*Fpinv(j,l));
                                Tangentr(i,j,p,q) += (Fer(i,k)*DSDF(k,h,p,q)*Ginv(l,h)*Fpinv(j,l));
                            }
                        }
                    }
                }
            }
        }
    }


}



void mlawPhenomenologicalSMP::tangentglassy(const STensor3& Fn,const STensor3& Fm,const IPPhenomenologicalSMP *q0, IPPhenomenologicalSMP *q1,
                                            const STensor43& dexpDelta_gammaNpr,const STensor3& Fegpr, STensor43& dlnCegpr,const STensor3& Eegpr,
                                            const STensor3& Npr,const double Eegpreq,const double Deltagamma,const double zg,const double H_gamma, const double T0,
                                            const double T,const double Tdot,const double dh,const double Tir,const double Tig,const double alphaig,
                                            const double taueq,const STensor3& Feg,const STensor3& Eeg,const STensor3& taueg,const STensor43& dlnCeg,
                                            const STensor63& ddlnCeg, const STensor3& G, double& dgammataueq, double& dgammataueqdT,STensor3& dgammataueqdF,
                                            STensor3& dPgdT, STensor43& Tangentg) const
{
    double dKgT,dGgT,dalphagT,dalpharT;
    double GgT = getGg(T,&dGgT);
    double KgT = getKg(T,GgT,&dKgT,&dGgT);
    double alpharT = getalpha_rubbery(T,&dalpharT);
    double alphagT = getalpha_glassy(T,&dalphagT);
    STensor43 C0g;
    HookeTensor(KgT,GgT,C0g);

    const STensor3& Fpg0 = q0->getConstRefToFpg();
    static STensor3 Fpg0inv, Ginv;
    STensor3 &Ff=q1->getRefToFf();
    STensor3 &Fpg=q1->getRefToFpg();
    STensor3 Ffinv, Fpginv;
    static STensor43 dtauegdF, dFpgdF;
    static STensor3 dtauegdT, dFpgdT;

    static STensor43 dFpgdEegpr, dtauegdEegpr;
    STensorOperation::inverseSTensor3(Fpg0,Fpg0inv);
    STensorOperation::inverseSTensor3(G,Ginv);
    STensorOperation::inverseSTensor3(Ff,Ffinv);
    STensorOperation::inverseSTensor3(Fpg,Fpginv);
    STensorOperation::zero(dtauegdF);
    STensorOperation::zero(dtauegdT);
    STensorOperation::zero(dFpgdF);
    STensorOperation::zero(dFpgdT);
    STensorOperation::zero(dFpgdEegpr);
    STensorOperation::zero(dtauegdEegpr);
    STensorOperation::zero(dgammataueqdF);
    dgammataueqdT=0.;

    dtauegdEegpr  = C0g;
    dtauegdEegpr *= zg;

    static double dTtdT;

    double T_t  = q1->getRefToTt();
    double T_t0 = q0->getConstRefToTt();
    double w  = q1->getRefTow();
    double w0 = q0->getConstRefTow();
    const double &Tdot0=q0->getConstRefToTdot();

    get_Tt(Tdot0,T_t0,&T_t, &dTtdT,&w,w0,_w1,_w2);
    dTtdT *= 1./dh;
    static double dzgdT;
    get_zg(T,w,T_t,&dzgdT,&dTtdT);

    double lambda_th, dlambdathdT;
    evaluate_reference_temperatures(q0, q1, zg, T_t, T0, T, q1->getConstRefToTdot(), lambda_th, dzgdT, &dlambdathdT);

    static STensor3 dGdT;
    dGdT  = _I2;
    dGdT *= dlambdathdT;

    static STensor3 dGinvdT;
    STensorOperation::zero(dGinvdT);

    for (int R=0; R<3; R++){
        for (int O=0; O<3; O++){
            for (int L=0; L<3; L++){
                for (int A=0; A<3; A++){
                    dGinvdT(R,A) -=Ginv(R,O)*dGdT(O,L)*Ginv(L,A);
                }
            }
        }
    }

    STensor3 devtaueg;
    double p;
    STensorOperation::decomposeDevTr(taueg,devtaueg,p);
    p=p/3.;

    dtauegdT *= 0.;
    for( int E=0; E<3; E++){
        dtauegdT(E,E) += p*dKgT/KgT;
        for( int Z=0; Z<3; Z++){
            dtauegdT(E,Z) += devtaueg(E,Z)*dGgT/GgT;
            for( int N=0; N<3; N++){
                for( int O=0; O<3; O++){
                    dtauegdT(E,Z) += dzgdT*C0g(E,Z,N,O)*Eegpr(N,O);
                }
            }
        }
    }


    static STensor43 EegprToF;
    for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
            for (int k=0; k<3; k++){
                for (int l=0; l<3; l++){
                    EegprToF(i,j,k,l) = 0.;
                    for (int p=0; p<3; p++){
                        for (int q=0; q<3; q++){
                            for (int r=0; r<3; r++){
                                for (int h=0; h<3; h++){
                                    EegprToF(i,j,k,l) += dlnCegpr(i,j,p,q)*Fegpr(k,p)*Ginv(l,h)*Ffinv(h,r)*Fpg0inv(r,q);
                                }
                            }
                        }
                    }
                }
            }
        }
    }


    if (Deltagamma > 0.)
    {
        double H_T = _j2IH->getYield0()*_temFunc_Sy0->getDiff(T) + (q1->getConstRefToIPJ2IsotropicHardening().getR() - _j2IH->getYield0())*_temFunc_H->getDiff(T);

        static STensor3 DgammaDEepr;
        DgammaDEepr = (Npr);
        DgammaDEepr *= ((2.*GgT*zg)/(3.*GgT*zg+H_gamma));
        double dDeltagammadT = -H_T;
        dDeltagammadT       -= 3.*Deltagamma*(GgT*dzgdT+zg*dGgT);
        dDeltagammadT       *= 1./(3.*zg*GgT+H_gamma);

        static STensor43 BE;
        tensprod(Npr,Npr,BE);
        BE *= (2.*GgT*zg/(3.*GgT*zg+H_gamma)- 2.*Deltagamma/3./Eegpreq);
        double val=Deltagamma/Eegpreq;
        BE.axpy(val,_I4dev);

        static STensor3  BT;
        BT = (Npr);
        BT*= dDeltagammadT;

        // update with plasticity corKir
        dtauegdEegpr.axpy(-2.*GgT*zg,BE);
        dtauegdT.daxpy(BT,-2.*GgT*zg);
        dtauegdT.daxpy(Npr,-2.*GgT*dzgdT*Deltagamma);
        dtauegdT.daxpy(Npr,-2.*dGgT*zg*Deltagamma);

        static STensor43 DexpABE;
        STensorOperation::multSTensor43(dexpDelta_gammaNpr,BE,DexpABE);
        // update with plasticity Fpg
        for (int i=0; i<3; i++){
            for (int j=0; j<3; j++){
                for (int k=0; k<3; k++){
                    for (int l=0; l<3; l++){
                        for (int p=0; p<3; p++){
                            dFpgdEegpr(i,j,k,l) += DexpABE(i,p,k,l)*Fpg0(p,j);
                        }
                    }
                }
            }
        }
        static STensor3 DexpABT;
        STensorOperation::multSTensor43STensor3(dexpDelta_gammaNpr,BT,DexpABT);
        STensorOperation::multSTensor3Add(DexpABT,Fpg0,1.,dFpgdT);

        static STensor3 dgammadF,dtaueqdF;
        STensorOperation::multSTensor3STensor43(DgammaDEepr,EegprToF,dgammadF);
        for (int j=0; j<3; j++){
            for (int X=0; X<3; X++){
                dtaueqdF(j,X) = 0.;
                for (int C=0; C<3; C++){
                    for (int D=0; D<3; D++){
                        for (int A=0; A<3; A++){
                            for (int B=0; B<3; B++){
                                dtaueqdF(j,X) += (3./2./taueq)*(devtaueg(C,D))*(dtauegdF(C,D,j,X)-((1/3.)*_I2(A,B)*dtauegdF(A,B,j,X)*_I2(C,D)));
                            }
                        }
                    }
                }
            }
        }

        dgammataueqdF  = (taueq*dgammadF);
        dgammataueqdF += (dtaueqdF*Deltagamma);

        double dtaueqdT=0.;

        for (int C=0; C<3; C++){
            for (int D=0; D<3; D++){
                for (int A=0; A<3; A++){
                    for (int B=0; B<3; B++){
                        dtaueqdT +=devtaueg(C,D)*(dtauegdT(C,D)-((1/3.)*_I2(A,B)*dtauegdT(A,B)*_I2(C,D)));
                    }
                }
            }
        }

        dtaueqdT *= (3./2./taueq);

        dgammataueqdT = (taueq*dDeltagammadT) + (Deltagamma*dtaueqdT);
    }

    STensorOperation::multSTensor43(dtauegdEegpr,EegprToF,dtauegdF);
    if (Deltagamma>0.)
    {
        STensorOperation::multSTensor43(dFpgdEegpr,EegprToF,dFpgdF);
    }

    else
    {
        STensorOperation::zero(dFpgdF);
    }


    // done DcorKirDF, DcorKirDT, DFpDF, DFpDT

    static STensor43 DinvFpgdF;
    static STensor3 DinvFpgdT;
    STensorOperation::zero(DinvFpgdF);
    STensorOperation::zero(DinvFpgdT);
    if (Deltagamma >0.){
        for (int i=0; i<3; i++){
            for (int j=0; j<3; j++){
                for (int p=0; p<3; p++){
                    for (int q=0; q<3; q++){
                        DinvFpgdT(i,j) -= Fpginv(i,p)*dFpgdT(p,q)*Fpginv(q,j);
                        for (int k=0; k<3; k++){
                            for (int l=0; l<3; l++){
                                DinvFpgdF(i,j,k,l) -= Fpginv(i,p)*dFpgdF(p,q,k,l)*Fpginv(q,j);
                            }
                        }
                    }
                }
            }
        }
    }

    static STensor43 dFegdF;
    static STensor3 dFegdT;
    STensorOperation::zero(dFegdF);
    STensorOperation::zero(dFegdT);
    for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
            for (int k=0; k<3; k++){
                for (int l=0; l<3; l++){
                    for (int r=0; r<3; r++){
                        dFegdT(i,j) += Fn(i,k)*dGinvdT(k,r)*Ffinv(r,l)*Fpginv(l,j);
                        dFegdT(i,j) += Fn(i,k)*Ginv(k,r)*Ffinv(r,l)*DinvFpgdT(l,j);
                        dFegdF(i,j,i,k) += Ginv(k,l)*Ffinv(l,r)*Fpginv(r,j);
                        if (Deltagamma> 0.){
                            for (int t=0; t<3; t++){
                                for (int h=0; h<3; h++){
                                    dFegdF(i,j,k,l) += Fn(i,h)*Ginv(h,r)*Ffinv(r,t)*DinvFpgdF(t,j,k,l);
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    static STensor63 DLDF;
    static STensor43 DLDT;
    STensorOperation::zero(DLDF);
    STensorOperation::zero(DLDT);
    for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
            for (int k=0; k<3; k++){
                for (int l=0; l<3; l++){
                    for (int r=0; r<3; r++){
                        for (int s=0; s<3; s++){
                            for (int a=0; a<3; a++){
                                DLDT(i,j,k,l) += ddlnCeg(i,j,k,l,r,s)*2.*Feg(a,r)*dFegdT(a,s);
                                for (int p=0; p<3; p++){
                                    for (int q=0; q<3; q++){
                                        DLDF(i,j,k,l,p,q) += ddlnCeg(i,j,k,l,r,s)*2.*Feg(a,r)*dFegdF(a,s,p,q);
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    static STensor43 DSDF; // S = corKir:L
    static STensor3 DSDT;
    STensorOperation::zero(DSDF);
    STensorOperation::zero(DSDT);
    for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
            for (int r=0; r<3; r++){
                for (int s=0; s<3; s++){
                    DSDT(i,j) += dtauegdT(r,s)*dlnCeg(r,s,i,j) + taueg(r,s)*DLDT(r,s,i,j);
                    for (int k=0; k<3; k++){
                        for (int l=0; l<3; l++){
                            DSDF(i,j,k,l) += dtauegdF(r,s,k,l)*dlnCeg(r,s,i,j) + taueg(r,s)*DLDF(r,s,i,j,k,l);
                        }
                    }
                }
            }
        }
    }

    static STensor3 Seg;
    STensorOperation::multSTensor3STensor43(taueg,dlnCeg,Seg);

    // compute mechanical tangent

    STensorOperation::zero(Tangentg);
    STensorOperation::zero(dPgdT);


    for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
            for (int k=0; k<3; k++){
                for (int l=0; l<3; l++){
                    for (int p=0; p<3; p++){
                        for (int q=0; q<3; q++){
                            dPgdT(i,j) += dFegdT(i,k)*Seg(k,l)*Fpginv(q,l)*Ffinv(p,q)*Ginv(j,p);
                            dPgdT(i,j) += Feg(i,k)*DSDT(k,l)*Fpginv(q,l)*Ffinv(p,q)*Ginv(j,p);
                            dPgdT(i,j) += Feg(i,k)*Seg(k,l)*DinvFpgdT(q,l)*Ffinv(p,q)*Ginv(j,p);
                            dPgdT(i,j) += Feg(i,k)*Seg(k,l)*Fpginv(q,l)*Ffinv(p,q)*dGinvdT(j,p);
                            for (int r=0; r<3; r++){
                                for (int h=0; h<3; h++){
                                    Tangentg(i,j,p,q) += (dFegdF(i,k,p,q)*Seg(k,l)*Fpginv(h,l)*Ffinv(r,h)*Ginv(j,r) + Feg(i,k)*DSDF(k,l,p,q)*Fpginv(h,l)*Ffinv(r,h)*Ginv(j,r)+Feg(i,k)*Seg(k,l)*DinvFpgdF(h,l,p,q)*Ffinv(r,h)*Ginv(j,r));
                                }
                            }
                        }
                    }
                }
            }
        }
    }


}



void mlawPhenomenologicalSMP::computePr(const double KrT, const double GrT,const STensor3& Fn,const STensor3& Fp, const STensor3& G, STensor3& Fer,STensor3& Eer,STensor3& tauer,STensor43& dlnCer,STensor63& ddlnCer,const double zg, STensor3& Pr)const
{
    static STensor3 Fpinv, Ginv;
    STensorOperation::inverseSTensor3(Fp,Fpinv);
    STensorOperation::inverseSTensor3(G,Ginv);
    static STensor3 Fer_int, Cer;
    STensorOperation::multSTensor3(Fn,Ginv,Fer_int);
    STensorOperation::multSTensor3(Fer_int,Fpinv,Fer);//Fer
    STensorOperation::multSTensor3FirstTranspose(Fer,Fer,Cer);//Cer

    STensorOperation::logSTensor3(Cer,_order,Eer,&dlnCer,&ddlnCer);
    Eer*=0.5; //(Elastic) Logarithmic strain of the rubbery phase

    static STensor3 per, devtauer,Ser;
    devtauer  = Eer.dev();
    devtauer*=((2.*GrT)*(1.-zg));
    double p_bracket = KrT*(1.-zg)*(Eer.trace());
    per=_I2;
    per*=p_bracket;
    tauer=per;
    tauer+=devtauer;//Corotational stress of the rubbery phase

    STensorOperation::multSTensor3STensor43(tauer,dlnCer,Ser);//Elastic 2nd Piola-Kirchhoff stress of the rubbery phase

    static STensor3 Pr_1, Pr_2, FpinvT, GinvT ;
    STensorOperation::transposeSTensor3(Fpinv,FpinvT);
    STensorOperation::transposeSTensor3(Ginv,GinvT);
    STensorOperation::multSTensor3(Fer,Ser,Pr_1);
    STensorOperation::multSTensor3(Pr_1,FpinvT,Pr_2);
    STensorOperation::multSTensor3(Pr_2,GinvT,Pr); //First Piola-Kirchhoff stres of the rubbery phase
}



void mlawPhenomenologicalSMP::computePg(const double KgT, const double GgT, const STensor3& Fn,const STensor3& Ff,const STensor3& Fpg,
                                        const STensor3& G,STensor3& Feg,STensor3& Eeg,STensor3& taueg,STensor43& dlnCeg,STensor63& ddlnCeg,
                                        const double zg,STensor3& Pg) const
{
    static STensor3 Ffinv, Fpginv, Ginv;
    STensorOperation::inverseSTensor3(Ff,Ffinv);
    STensorOperation::inverseSTensor3(Fpg,Fpginv);
    STensorOperation::inverseSTensor3(G,Ginv);
    static STensor3 Feg_int, Feg_int2;
    STensorOperation::multSTensor3(Fn, Ginv, Feg_int);
    STensorOperation::multSTensor3(Feg_int, Ffinv, Feg_int2);
    STensorOperation::multSTensor3(Feg_int2, Fpginv, Feg);

    static STensor3 Ceg;
    STensorOperation::multSTensor3FirstTranspose(Feg, Feg, Ceg);
    STensorOperation::logSTensor3(Ceg,_order,Eeg,&dlnCeg,&ddlnCeg);
    Eeg*=0.5;
    static STensor3 devtaueg,peg,Seg;
    devtaueg  = Eeg.dev();
    devtaueg*=(2.*GgT*zg);
    double peg_bracket=zg*KgT*(Eeg.trace());
    peg=_I2;
    peg*=peg_bracket;
    taueg=peg;
    taueg+=devtaueg;

    STensorOperation::multSTensor3STensor43(taueg,dlnCeg,Seg);

    static STensor3 Pg_1, Pg_2, Pg_3, FpginvT, FfinvT, GinvT;
    STensorOperation::transposeSTensor3(Fpginv,FpginvT);
    STensorOperation::transposeSTensor3(Ffinv,FfinvT);
    STensorOperation::transposeSTensor3(Ginv,GinvT);
    STensorOperation::multSTensor3(Feg,Seg,Pg_1);
    STensorOperation::multSTensor3(Pg_1,FpginvT,Pg_2);
    STensorOperation::multSTensor3(Pg_2,FfinvT,Pg_3);
    STensorOperation::multSTensor3(Pg_3,GinvT,Pg); //First Piola-Kirchhoff stres of the glassy phase

}
