//
// C++ Interface: numericalMaterial
//
// Description: Class creates microsolver from setting
//
//
// Author:  <Van Dung NGUYEN>, (C) 2012
//
// Copyright: See COPYING file that comes with this distribution
//
//
#include "numericalMaterial.h"
#include "nonLinearMechSolver.h"
#include "nonLinearBC.h"
#include "dirent.h"

void numericalMaterialBase::distributeMacroIP(const std::vector<int>& allIPNum, const std::vector<int>& allRanks,
                                    std::map<int,std::set<int> >& mapIP){
  int IPSize = allIPNum.size();
  mapIP.clear();

  int rankSize = allRanks.size();
  // init map
  /*
  for (int i=0; i<rankSize; i++){
    int rank = allRanks[i];
    std::set<int> numElOnRank;
    int startVal = IPSize*i/rankSize;
    int endVal = IPSize*(i+1)/rankSize;
    for (int j=startVal; j <endVal; j++){
      numElOnRank.insert(allIPNum[j]);
    }
    mapIP[rank] = numElOnRank;
  };
  */
  int numOnEachRank = IPSize/rankSize;
  for (int i=0; i<rankSize; i++){
    int rank = allRanks[i];
    std::set<int>& numElOnRank = mapIP[rank];
    for (int j=0; j<numOnEachRank; j++){
      int num = j+i*numOnEachRank;
      numElOnRank.insert(allIPNum[num]);
    };
  };
  int j=rankSize*numOnEachRank;
  while (j < IPSize){
    int i = j- rankSize*numOnEachRank;
    std::set<int>& numElOnRank = mapIP[allRanks[i]];
    numElOnRank.insert(allIPNum[j]);
    j++;
  };

  #ifdef _DEBUG
  for (std::map<int,std::set<int> >::iterator it = mapIP.begin(); it!= mapIP.end(); it++){
    int rank = it->first;
    std::set<int>& setIP = it->second;
    std::string fname = "IPonRank"+int2str(rank)+".csv";
    FILE* file = fopen(fname.c_str(),"a");
    for (std::set<int>::iterator it = setIP.begin(); it!= setIP.end(); it++){
      int num = *it;
      int e, g;
      numericalMaterialBase::getTwoIntsFromType(num,e,g);
      fprintf(file, "%d;%d;\n",e,g);
    }
    fclose(file);
  }
  #endif // _DEBUG
};

numericalMaterial::numericalMaterial(int tag)
                   :numericalMaterialBase(),_tag(tag),_numstep(1),_tol(1e-6),_absTol(1e-10),
                    _stressFlag(true),_stressIntegFlag(0), _sameStateCriterion(1e-10),
                    _tangentFlag(true),_tangentIntegFlag(0),_tangentPerturbation(1.e-7),
                    _isIterative(true),_stiffModification(true),
                    _stabilityAnalysis(false), _instabilitycriterion(1.e-8),
                    _perturbation(false),_perturbationfactor(0.01), _numeigenvalue(100), _eigenFollowing(false),
                    _isViewAll(false),_systemType(1), _rveVolume(0.),
                    _scheme(1),_solver(2), _pathFollowing(false),_controlType(1),
                    _microMeshFile(""),_microGeoFile(""),
                    _meshOrder(-1),_meshDim(-1),_meshIncomplete(false),
										_microBC(NULL),_extractIrreversibleEnergy(false){};

numericalMaterial::numericalMaterial(const numericalMaterial& src):numericalMaterialBase(src){
  _tag = src._tag; // for tag of microsolver
  _numstep = src._numstep; // number of time step at microsolve
  _tol = src._tol; // tolerance used in NR at microsolver
	_absTol = src._absTol;
  _systemType = src._systemType;
  _solver = src._solver;
  _isIterative = src._isIterative;
  _stiffModification = src._stiffModification;
  _sameStateCriterion = src._sameStateCriterion;
					// micro-data
  _microMeshFile = src._microMeshFile;	// micro meshfile
  _microGeoFile = src._microGeoFile;
  _meshDim = src._meshDim;
  _meshOrder = src._meshOrder;
  _meshIncomplete = src._meshIncomplete;

  _stressFlag = src._stressFlag;
  _stressIntegFlag = src._stressIntegFlag; // 0- volume, 1-surface
  _tangentFlag = src._tangentFlag;
  _tangentIntegFlag = src._tangentIntegFlag;
  _tangentPerturbation = src._tangentPerturbation;
  _allView = src._allView; // micro-solver to view
  _isViewAll = src._isViewAll; // true if view all
  _allDomain = src._allDomain;
  _allMaterialLaw = src._allMaterialLaw;
	_microBC = NULL;
	if (src._microBC != NULL){
		_microBC = src._microBC->clone();
	}

  _stabilityAnalysis = src._stabilityAnalysis;
  _modeview = src._modeview;
  _numeigenvalue = src._numeigenvalue;
  _perturbation = src._perturbation;
  _perturbationfactor = src._perturbationfactor;
  _instabilitycriterion = src._instabilitycriterion;
  _eigenFollowing = src._eigenFollowing;

  _allGP = src._allGP;
  _viewMicroField = src._viewMicroField;
  _viewMicroUnknownComponnent = src._viewMicroUnknownComponnent;

  _pathFollowing = src._pathFollowing;
  _controlType = src._controlType;

  _rveVolume = src._rveVolume;
  _allInitialBC = src._allInitialBC;
	_allDirichletBC = src._allDirichletBC;

	_extractIrreversibleEnergy = src._extractIrreversibleEnergy;
};

numericalMaterial::~numericalMaterial(){
	// delete cache
	for (int i=0; i< _allocatedDomains.size(); i++){
		delete _allocatedDomains[i];
	}
	_allocatedDomains.clear();
	for (int i=0; i< _allocatedLaws.size(); i++){
		delete _allocatedLaws[i];
	}
	_allocatedLaws.clear();
};

nonLinearMechSolver* numericalMaterial::createMicroSolver(const int ele, const int gpt) const {
  /**create serial micro-solver **/
  nonLinearMechSolver* solver = new nonLinearMechSolver(_tag,false);
  solver->setMicroSolverFlag(true);
  solver->setMultiscaleFlag(true);

  /** set microsolver identification **/
  solver->setMicroProblemIndentification(ele,gpt);
  int type = numericalMaterial::createTypeWithTwoInts(ele,gpt);

  std::map<int,int>::const_iterator itm = _meshIdMap.find(type);
  if (itm == _meshIdMap.end() or _allPertMesh.size() ==0){
    /**load model**/
		std::map<int,std::string>::const_iterator itMestConstraint =  _meshOnIP.find(type);
		if (itMestConstraint == _meshOnIP.end()){
			if (_microMeshFile.length() >0){
				Msg::Info("create model from mesh file %s",_microMeshFile.c_str());
				solver->loadModel(_microMeshFile);
			}
			else if (_microGeoFile.length()>0 ){
				Msg::Info("create model from geometry file %s",_microGeoFile.c_str());
				solver->createMicroModel(_microGeoFile,_meshDim,_meshOrder,_meshIncomplete);
			}
			else{
				Msg::Error("micro model file must be provided: proc %d, ele %d, GP %d",Msg::GetCommRank(),ele,gpt);
			}
		}
		else{
			std::string mshFile = itMestConstraint->second;
			printf("create model from mesh file %s  \n",mshFile.c_str());
			solver->loadModel(mshFile);
		}
  }
  else{
    std::map<int,std::string>::const_iterator itstr = _allPertMesh.find(itm->second);
    std::string mshFile = itstr->second;
    printf("perturbation mesh on ele %d gp %d \n",ele,gpt);
    printf("create model from mesh file %s  \n",mshFile.c_str());
    solver->loadModel(mshFile);
  }


  /** set domain for micro solver**/
  for (int i=0; i<_allDomain.size(); i++){
    partDomain* dom = _allDomain[i]->clone();
    if (dom) {
      solver->addDomain(dom);
			_allocatedDomains.push_back(dom);
    }
    else{
      Msg::Error("partDomain::clone has not been defined");
    }
  };
  /** set material law**/
  if (_pertMaterialLawIP.find(type) == _pertMaterialLawIP.end()){
    for (int i=0; i<_allMaterialLaw.size(); i++){
			materialLaw* law = _allMaterialLaw[i]->clone();
			if (law != NULL){
				solver->addMaterialLaw(law);
				_allocatedLaws.push_back(law);
			}
			else{
				Msg::Error("materialLaw::clone must be defined");
			}
    };
  }
  else{
    Msg::Warning("perturbation in ele %d GP %d",ele,gpt);
    for (int i=0; i<_allPertubationLaw.size(); i++){
			materialLaw* law = _allPertubationLaw[i]->clone();
			if (law != NULL){
				solver->addMaterialLaw(law);
				_allocatedLaws.push_back(law);
			}
			else{
				Msg::Error("materialLaw::clone must be defined");
			}
    };
  }

  if (_rveVolume >0){
    solver->setRVEVolume(_rveVolume);
  }

  /** set micro BC**/
  if (_microBC == NULL){
		Msg::Error("microscopic BC must be specified to launch multiscale analyses");
	}
	else{
		solver->addMicroBC(_microBC);
	}

  /**set control type --> always load control**/
  solver->setControlType(0);
  /** set system type**/
  solver->setSystemType(_systemType);
  /** set stress method**/
  solver->setStressAveragingMethod(_stressIntegFlag);
  /** set tangent method**/
  solver->setTangentAveragingMethod(_tangentIntegFlag,_tangentPerturbation);
  /** set scheme**/
  solver->Scheme(_scheme);
  /** set solver **/
  solver->Solver(_solver);

  /**set numstep and tolerance, compute time always from 0 to 1 **/
  solver->snlData(_numstep,1.,_tol,_absTol);
	solver->setSameStateCriterion(_sameStateCriterion);

	solver->setExtractIrreversibleEnergyFlag(_extractIrreversibleEnergy);

/** set eigensolver analysis**/
  if (_allStabilityView.find(type) != _allStabilityView.end() or _stabilityAnalysis){
    solver->eigenValueSolver(_numeigenvalue,true);
    for (int i=0; i<_modeview.size(); i++)
      solver->setModeView(_modeview[i]);
    solver->setEigenSolverFollowing(_eigenFollowing);
    solver->setInstabilityCriterion(_instabilitycriterion);
    solver->setPerturbationFactor(_perturbationfactor);
    solver->perturbateBucklingModeByMinimalEigenValueMode(_perturbation);
  }
  /** set stiffness midification **/
  solver->stiffnessModification(_stiffModification);
  /** set iterative solver**/
	solver->iterativeProcedure(_isIterative);

  /** for viewing**/
  std::set<int>::iterator it = _allView.find(type);
  if ( (ele>0) and (it!=_allView.end() or (_isViewAll and _allGP.find(gpt)!=_allGP.end()))){
    solver->setMessageView(1);
    solver->setStrainArchiveFlag(1);
    solver->setHomogenizationPropertyArchiveFlag(1);
    solver->setDisplacementAndIPArchiveFlag(1);

    for (std::map<int,std::string>::const_iterator itFieldView = _viewMicroField.begin(); itFieldView!= _viewMicroField.end(); itFieldView++){
      solver->internalPointBuildView(itFieldView->second,itFieldView->first,1,1);
    }

    for (std::map<int,std::string>::const_iterator itUView = _viewMicroUnknownComponnent.begin(); itUView!= _viewMicroUnknownComponnent.end(); itUView++){
      solver->OneUnknownBuildView(itUView->second,itUView->first,1);
    }

  }
  else{
    solver->setHomogenizationPropertyArchiveFlag(0);
    solver->setDisplacementAndIPArchiveFlag(0);
    solver->setMessageView(0);
    /** archive all strain to recompute if necessary**/
    solver->setStrainArchiveFlag(0);
  }
  /** initialize microsolver from setting data**/

  solver->pathFollowing(_pathFollowing,0);
  solver->setPathFollowingControlType(_controlType);

  // stress and tangent initial flag
  solver->stressAveragingFlag(_stressFlag);
  solver->tangentAveragingFlag(_tangentFlag);

  // initial boundayr condition
  for (int iInit=0; iInit< _allInitialBC.size(); iInit++){
    const initialBoundaryCondition& iBC =  _allInitialBC[iInit];
    solver->initialBC(iBC.onWhat,"Position",iBC.physical,iBC.comp,iBC.val);
  }

	for (int iD=0; iD < _allDirichletBC.size(); iD++){
		const DirichletBoundaryCondition& dirichLetBC= _allDirichletBC[iD];
		solver->displacementBC(dirichLetBC.onWhat,dirichLetBC.physical,dirichLetBC.comp,dirichLetBC.val);
	};

  return solver;
};
void numericalMaterial::loadModel(const std::string meshfile){
  _microMeshFile = meshfile;
};

void numericalMaterial::createMicroModel(const std::string geoFileName,const int dim,const int order,
                   const bool incomplete){
  _microGeoFile = geoFileName;
  _meshDim = dim;
  _meshOrder = order;
  _meshIncomplete = incomplete;
};

// set micro part domains
void numericalMaterial::addDomain(partDomain* domain){
  _allDomain.push_back(domain);
};
// set micro material law
void numericalMaterial::addMaterialLaw(materialLaw* mlaw){
  _allMaterialLaw.push_back(mlaw);
};

void numericalMaterial::addPertutationalMeshIP(const int ele, const int ip, const std::string meshfile){
	Msg::Info("pert mesh it element %d ip %d",ele,ip);
	int type = numericalMaterial::createTypeWithTwoInts(ele,ip);
	_meshOnIP[type] = meshfile;
};

void numericalMaterial::setPerturbationMaterialLawIP(const int ele, const int ip){
  int num = numericalMaterial::createTypeWithTwoInts(ele,ip);
  _pertMaterialLawIP.insert(num);
};

void numericalMaterial::addPerturbationMaterialLaw(materialLaw* mlaw){
  _allPertubationLaw.push_back(mlaw);
}

void numericalMaterial::addMicroBC(nonLinearMicroBC* bc){
  _microBC = bc;
}

// set tolerence for micro newton-raphson
void numericalMaterial::setNumStep(const int i){
  _numstep = i;
};
		//
void numericalMaterial::setTolerance(const double tol, const double abstol){
  _tol = tol;
	_absTol = abstol;
};

void numericalMaterial::setSameStateCriterion(const double cr){
	_sameStateCriterion = cr;
};
// set view to this element
void numericalMaterial::addViewMicroSolver(const int e, const int gp){
  int type = numericalMaterial::createTypeWithTwoInts(e,gp);
  _allView.insert(type);
};
// set view all elements
void numericalMaterial::setViewAllMicroProblems(const bool flag, const int a){
  _allGP.insert(a);
  _isViewAll= flag;
};

void numericalMaterial::tangentAveragingFlag(const bool fl){
  _tangentFlag = fl;

};
void numericalMaterial::stressAveragingFlag(const bool fl){
  _stressFlag = fl;
};

void numericalMaterial::setStressAveragingMethod(const int method){
  _stressIntegFlag = method;
};
void numericalMaterial::setTangentAveragingMethod(const int method, const double prec){
  _tangentIntegFlag = method;
  _tangentPerturbation = prec;
};

void numericalMaterial::Scheme(const int s){
  _scheme = s;
};
//
void numericalMaterial::Solver(const int s){
  _solver = s;
};
//
void numericalMaterial::setSystemType(const int s){
  _systemType = s;
};


void numericalMaterial::stiffnessModification(const bool flag){
  _stiffModification = flag;
};
		//
void numericalMaterial::iterativeProcedure(const bool flag){
  _isIterative = flag;
};

void numericalMaterial::pathFollowing(const bool p){
  _pathFollowing = p;
};

void numericalMaterial::setPathFollowingControlType(const int i){
  _controlType = i;
};


void numericalMaterial::setRVEVolume(const double val){
  _rveVolume = val;
};

void numericalMaterial::addStabilityMicroSolver(const int e, const int gp){
  int type = numericalMaterial::createTypeWithTwoInts(e,gp);
  _allStabilityView.insert(type);
};

void numericalMaterial::eigenValueSolver(const int num, const bool fl){
  _numeigenvalue = num;
  _stabilityAnalysis = fl;
}; // number of eigenvalue
void numericalMaterial::setModeView(const int view){
  _modeview.push_back(view);
}; // for view buckling mode
void numericalMaterial::setPerturbationFactor(const double val){
  _perturbationfactor = val;
};
void numericalMaterial::setInstabilityCriterion(const double val){
  _instabilitycriterion = val;
};
void numericalMaterial::perturbateBucklingModeByMinimalEigenValueMode(const bool flag){
  _perturbation = flag;
};
void numericalMaterial::setEigenSolverFollowing(const bool flag){
  _eigenFollowing = flag;
};

void numericalMaterial::internalPointBuildView(const std::string vname,const int comp){
  _viewMicroField[comp] = vname;
};

void numericalMaterial::OneUnknownBuildView(const std::string vname, const int comp){
  _viewMicroUnknownComponnent[comp] = vname;
};

void numericalMaterial::initialBC(std::string onwhat, const int numphys, const int comp, const double value){
  _allInitialBC.push_back(initialBoundaryCondition(onwhat,numphys,comp,value));
};

void numericalMaterial::dirichletBC(std::string onwhat, const int numphys, const int comp, const double value){
	_allDirichletBC.push_back(DirichletBoundaryCondition(onwhat,numphys,comp,value));
};

void numericalMaterial::setExtractIrreversibleEnergyFlag(const bool flg){
	_extractIrreversibleEnergy = flg;
};

void numericalMaterial::loadAllRVEMeshes(const std::string prefix){
  DIR *dir;
  struct dirent *ent;
  fileNameFilter* filter = new fileNameFilter(prefix,".msh");
  printf("read all file in current folder \n");
  if ((dir = opendir (".")) != NULL) {
    /* get all the files and directories within directory */
    while ((ent = readdir (dir)) != NULL) {
      std::string filename = ent->d_name;
      int num = filter->getNumber(filename);
      if (num>0){
        _allPertMesh[num] = filename;
        //printf("id: %d  -->filename: %s \n",num,filename.c_str());
      }
    }
    closedir (dir);
  }
  _idIterator = _allPertMesh.begin();
};

void numericalMaterial::printMeshIdMap() const{
  int rank = Msg::GetCommRank();
  std::string  filename = "idMap"+int2str(rank)+".txt";
  FILE* file =fopen(filename.c_str(),"a");
  for (std::map<int,int>::const_iterator it = _meshIdMap.begin(); it!= _meshIdMap.end(); it++){
    fprintf(file,"%d %d\n",it->first,it->second);
  }
  fclose(file);
};

void numericalMaterial::assignMeshId(const int e, const int gpt){
  if (_allPertMesh.size() == 0) return;
  int type = numericalMaterial::createTypeWithTwoInts(e, gpt);
  std::map<int,int>::iterator ittype = _meshIdMap.find(type);
  if (ittype!= _meshIdMap.end()) return;
  else{
    if (_idIterator == _allPertMesh.end())
      _idIterator = _allPertMesh.begin();


    _meshIdMap[type] = _idIterator->first;
    _idIterator++;
  }
};

void numericalMaterial::setMeshId(const int e, const int gpt, const int id) {
  if (_allPertMesh.size() == 0) return;
  int type = numericalMaterial::createTypeWithTwoInts(e, gpt);
  _meshIdMap[type] = id;
}

int numericalMaterial::getMeshId(const int e, const int gpt) const{
  int type = numericalMaterial::createTypeWithTwoInts(e,gpt);
  std::map<int,int>::const_iterator it = _meshIdMap.find(type);
  if (it !=_meshIdMap.end())
    return it->second;
  else{
    Msg::Warning("mesh id is not set: ELE %d GP %d",e,gpt);
    return 0;
  }
};

std::string numericalMaterial::getMeshFileName(const int e, const int gpt) const{
  if (_allPertMesh.size() ==0) return _microMeshFile;
  int id = this->getMeshId(e,gpt);
  std::map<int,std::string>::const_iterator it = _allPertMesh.find(id);
  if (it == _allPertMesh.end()){
    return _microMeshFile;
  }
  else{
    return it->second;
  }
};
#if defined(HAVE_MPI)
int numericalMaterial::getNumberValuesMPI() const {return _meshIdMap.size()*2;};
void numericalMaterial::fillValuesMPI(int* val) const{
  int iter=0;
  for (std::map<int,int>::const_iterator it = _meshIdMap.begin(); it!= _meshIdMap.end(); it++){
    val[iter] = it->first;
    val[iter+1] = it->second;
    iter+=2;
  }
};
void numericalMaterial::getValuesMPI(const int* val, const int size){
  for (int i=0; i<size; i+=2){
    _meshIdMap[val[i]] = val[i+1];
  }
};
#endif //HAVE_MPI



fileNameFilter::fileNameFilter(const std::string prefix, const std::string tag): _prefix(prefix),_fileTag(tag)  {}

int fileNameFilter::getNumber(const std::string& filename){
  if (filename.size() < _prefix.size() + _fileTag.size()) return 0;
  int count=0;
  for (int i=0; i<_prefix.size(); i++){
    if (_prefix[i]== filename[i])
      count++;
  }
  if (count != _prefix.size()) return 0;
  count=0;
  for (int i= _fileTag.size() -1; i>=0; i--){
    if (_fileTag[i] == filename[filename.size()-_fileTag.size()+i]){
      count++;
    }
  }
  if (count != _fileTag.size()) return 0;

  std::string temp;
  for (unsigned int i = 0; i<filename.size(); i++){
    if (isdigit(filename[i]))
      temp += filename[i];

  }

   std::istringstream stream(temp);
   int number;
   stream >> number;

   return number;
}


