//
// C++ Interface: numericalMaterial
//
// Description: Class creates microsolver from setting
//
//
// Author:  <Van Dung NGUYEN>, (C) 2012
//
// Copyright: See COPYING file that comes with this distribution
//
//
#ifndef NUMERICALMATERIAL_H_
#define NUMERICALMATERIAL_H_



#include "nonLinearMicroBC.h"
#include "mlaw.h"
#include "numericalFunctions.h"

class nonLinearMechSolver;
class partDomain;
class IPStateBase;

class numericalMaterialBase{
  #ifndef SWIG
  // two function allows creating a unique number from element number and gauss
  // number --> maximal 100 gausspoint per element (increasing if necessary)
  public:
    inline static int createTypeWithTwoInts(int i1, int i2)
    {
      return i2 + 100 * i1;
    }
    inline static void getTwoIntsFromType(int t, int &i1, int &i2)
    {
      i2 = t % 100;
      i1 = t / 100;
    }
    // function to distribut IP to different process
    static void distributeMacroIP(const std::vector<int>& allIPNum, const std::vector<int>& allRanks,
                                    std::map<int,std::set<int> >& mapIP);

  public:
    numericalMaterialBase(){}
    numericalMaterialBase(const numericalMaterialBase& src){}
    virtual numericalMaterialBase& operator = (const numericalMaterialBase& src){
    return *this;
    }
    virtual ~numericalMaterialBase(){}

    virtual nonLinearMechSolver* createMicroSolver(const int ele=0, const int gpt=0) const = 0;

    virtual void createIPState(const bool issolve,  /*create IPstate but create solver if true else false*/
                                IPStateBase* &ips, const bool* state_=NULL,
                                const MElement *ele=NULL, const int nbFF_=0, const IntPt *GP=NULL, const int gpt=0) const  = 0;

    virtual void printMeshIdMap() const = 0;
    //this function provides an id for a GP
		virtual void assignMeshId(const int e, const int gpt) = 0;
		virtual void setMeshId(const int e, const int gpt, const int id)= 0;
		// this function get mesh id
		virtual int getMeshId(const int e, const int gpt) const= 0;
		#if defined(HAVE_MPI)
		// these function used for comunicate MPI
		virtual int getNumberValuesMPI() const= 0;
		virtual void fillValuesMPI(int* val) const= 0;
		virtual void getValuesMPI(const int* val, const int size)= 0;
		#endif //HAVE_MPI
  #endif // SWIG
};

class numericalMaterial : public numericalMaterialBase{
  #ifndef SWIG
  public:
    class initialBoundaryCondition{
      public:
        std::string onWhat;
        int physical;
        int comp;
        double val;
      public:
        initialBoundaryCondition(const std::string locat, const int phys, const int c, const double v):
          onWhat(locat),physical(phys),comp(c),val(v){}
        initialBoundaryCondition(const initialBoundaryCondition& src): onWhat(src.onWhat),physical(src.physical),
                                comp(src.comp),val(src.val){}
        virtual initialBoundaryCondition& operator = (const initialBoundaryCondition& src){
          onWhat = src.onWhat;
          physical = src.physical;
          comp = src.comp;
          val = src.val;
          return *this;
        }
				virtual ~initialBoundaryCondition(){};
    };
		
		class DirichletBoundaryCondition : public initialBoundaryCondition{
			public:
				DirichletBoundaryCondition(const std::string locat, const int phys, const int c, const double v):
          initialBoundaryCondition(locat,phys,c,v){}
        DirichletBoundaryCondition(const DirichletBoundaryCondition& src): initialBoundaryCondition(src){}
        virtual DirichletBoundaryCondition& operator = (const initialBoundaryCondition& src){
					initialBoundaryCondition::operator =(src);
					return *this;
				};
				virtual ~DirichletBoundaryCondition(){};
		};


	protected:
		mutable std::vector<partDomain*> _allocatedDomains; // cache
		mutable std::vector<materialLaw*> _allocatedLaws;
		
    int _tag; // tag for solver
    int _numstep; // number of time step at microsolve
    double _tol; // tolerance used in NR at microsolver
    double _absTol; // absotlute tolerane
    int _systemType; // 0 - Disp elumination, 1 - mult elimination, 2 - disp-mult
    int _scheme; // 1- linear static, 2- nonlinear static
    int _solver; // 0 - gmm, 1 - tausc, 2 -petsc
    bool _isIterative; // 0 - non-iterative , 1 - iterative
    bool _stiffModification; // 0 - using one stiffness, 1- using multiple stiffness
    double _sameStateCriterion; // the microcompute will not be performed if strain different less than this value


    std::map<int,int> _meshIdMap; // mesh id map,
    std::map<int,std::string>::iterator _idIterator;// iterator to acces meshes
    std::map<int,std::string> _allPertMesh; // all perturbed meshes
		
		std::map<int,std::string> _meshOnIP; // constraint mesh on each ip

    std::string _microMeshFile;	// micro meshfile
    std::string _microGeoFile; // micro geometry
    int _meshOrder; // mesh ord
    int _meshDim; // mesh dimension
    bool _meshIncomplete; // mesh incomplete

    std::vector<partDomain*> _allDomain; // micro domain to define
    std::vector<materialLaw*> _allMaterialLaw; // micro material to define
    nonLinearMicroBC* _microBC; // microBC

    std::set<int> _pertMaterialLawIP; // all perturbed by material law
    std::vector<materialLaw*> _allPertubationLaw; // all perturbed law
		
    // initial BCs
    std::vector<initialBoundaryCondition> _allInitialBC;
		// dirichlet BCs
		std::vector<DirichletBoundaryCondition> _allDirichletBC;

		// homogenized flags
    bool _stressFlag; //  stress flag 0- no 1- yes
    int _stressIntegFlag; // stress method 0- volume, 1-surface
    bool _tangentFlag; // tangent flag 0- no 1- yes
    int _tangentIntegFlag; // tangent method 0- perturbation 1-condensation
    double _tangentPerturbation; // perturbation value to compute tangent operator
		
    std::set<int> _allStabilityView; // stability view
    bool _stabilityAnalysis; // true if using eigen analysis
		std::vector<int> _modeview; // view mode
    int _numeigenvalue; // number eigenvalue to computed
    double _perturbationfactor; // factor for adding buckling mode to solution
    double _instabilitycriterion; //  criterion for instability
    bool _perturbation; // perturbaton flag: 0- not perturbate buckling mode 1 - yes
		bool _eigenFollowing; // eigen following

    std::set<int> _allView; // micro-solver to view
    bool _isViewAll; // true if view all
    std::set<int> _allGP; // view all GP
    std::map<int,std::string> _viewMicroField;
    std::map<int,std::string> _viewMicroUnknownComponnent;

    // for pathFollowing analysis
    bool _pathFollowing; // true if using path following method
    int _controlType; // following strategies

    double _rveVolume; // volume of RVE
		bool _extractIrreversibleEnergy;
    #endif

	public:
    #ifndef SWIG
		numericalMaterial(int tag);
		numericalMaterial(const numericalMaterial& src);

		virtual ~numericalMaterial();
    // create micro solver
    virtual nonLinearMechSolver* createMicroSolver(const int ele, const int gpt) const;

		#endif
		// set micro meshes
		virtual void loadModel(const std::string meshfile);

		virtual void createMicroModel(const std::string geoFileName,const int dim,const int order,
                   const bool incomplete=false);

		// set micro part domains
		virtual void addDomain(partDomain* domain);
		// set micro material law
    virtual void addMaterialLaw(materialLaw* mlaw);
		
		virtual void addPertutationalMeshIP(const int ele, const int ip, const std::string meshfile);
    virtual void setPerturbationMaterialLawIP(const int ele, const int ip);
    virtual void addPerturbationMaterialLaw(materialLaw* mlaw);
		// add micro BC
    virtual void addMicroBC(nonLinearMicroBC* bc);
		// set tolerence for micro newton-raphson
		virtual void setNumStep(const int i);
		//
		virtual void setTolerance(const double tol, const double abstol = 1e-10);
		//
	  virtual void setSameStateCriterion(const double cr);

		// set view to this element
		virtual void addViewMicroSolver(const int e, const int gp = 0);
		// set view all elements
		virtual void setViewAllMicroProblems(const bool flag = true, const int algp = 0);

		virtual void tangentAveragingFlag(const bool fl = true);
		virtual void stressAveragingFlag(const bool fl = true);
		virtual void setStressAveragingMethod(const int method);
		virtual void setTangentAveragingMethod(const int method, const double prec = 1.e-8);
		//
		virtual void Scheme(const int s);
		virtual void Solver(const int s);
		virtual void setSystemType(const int s);

		virtual void stiffnessModification(const bool flag = true);
    virtual void iterativeProcedure(const bool flag = true);

    // path following flag
    virtual void pathFollowing(const bool p);
    virtual void setPathFollowingControlType(const int i);

    virtual void setRVEVolume(const double val);

    /** for stability analysis */
    virtual void addStabilityMicroSolver(const int e, const int gp);
    virtual void eigenValueSolver(const int num = 10, const bool fl = true); // number of eigenvalue
		virtual void setModeView(const int view); // for view buckling mode
		virtual void setPerturbationFactor(const double val = 1e-2);
		virtual void setInstabilityCriterion(const double val = 1e-6);
		virtual void perturbateBucklingModeByMinimalEigenValueMode(const bool flag = true);
		virtual void setEigenSolverFollowing(const bool flag);
		
		virtual void setExtractIrreversibleEnergyFlag(const bool flg);

    // for initial BC
    virtual void initialBC(std::string onwhat, const int numphys, const int comp, const double value);
		virtual void dirichletBC(std::string onwhat, const int numphys, const int comp, const double value);
		// for view
		virtual void internalPointBuildView(const std::string vname,const int comp);
		virtual void OneUnknownBuildView(const std::string vname, const int comp);

		// this function load all meshes in current folder
		virtual void loadAllRVEMeshes(const std::string prefix);
		virtual void printMeshIdMap() const;
		#ifndef SWIG
		//this function provides an id for a GP
		virtual void assignMeshId(const int e, const int gpt);
		virtual void setMeshId(const int e, const int gpt, const int id);
		// this function get mesh id
		virtual int getMeshId(const int e, const int gpt) const;
		// this function get mesh filename
		virtual std::string getMeshFileName(const int e, const int gpt) const;
		#if defined(HAVE_MPI)
		// these function used for comunicate MPI
		virtual int getNumberValuesMPI() const;
		virtual void fillValuesMPI(int* val) const;
		virtual void getValuesMPI(const int* val, const int size);
		#endif //HAVE_MPI
		#endif //SWIG
};

#ifndef SWIG
class fileNameFilter {
  protected:
    std::string _prefix;
    std::string _fileTag;
  public:
    fileNameFilter(const std::string prefix, const std::string tag);
    ~fileNameFilter(){}
    int getNumber(const std::string& filename);
};

#endif

#endif // NUMERICALMATERIAL_H_
