//
// Description: Define viscosity laws for  plasticity
//
//
// Author:  <V.D. Nguyen>, (C) 2014
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include "viscosityLaw.h"
#include "math.h"
#include "GmshMessage.h"

viscosityLaw::viscosityLaw(const int num, const bool init):_num(num),_initialized(init){};
viscosityLaw::viscosityLaw(const viscosityLaw& src):_num(src._num),_initialized(src._initialized){}
viscosityLaw& viscosityLaw::operator = (const viscosityLaw& src){
  _num = src._num;
  _initialized = src._initialized;
  return *this;
};




constantViscosityLaw::constantViscosityLaw(const int num, const double val, const bool init):viscosityLaw(num,init),_eta0(val){};
constantViscosityLaw::constantViscosityLaw(const constantViscosityLaw& src):viscosityLaw(src),_eta0(src._eta0){};
constantViscosityLaw& constantViscosityLaw::operator = (const viscosityLaw& src){
  viscosityLaw::operator=(src);
  const constantViscosityLaw* ps = dynamic_cast<const constantViscosityLaw*>(&src);
  if (ps != NULL){
    _eta0 = ps->_eta0;
  }
  return *this;
};




saturatePowerViscosityLaw::saturatePowerViscosityLaw(const int num, const double eta0, const double q, const bool init):
                viscosityLaw(num,init), _eta0(eta0), _q(q){};
saturatePowerViscosityLaw::saturatePowerViscosityLaw(const saturatePowerViscosityLaw& src):viscosityLaw(src),_eta0(src._eta0),_q(src._q){};
saturatePowerViscosityLaw& saturatePowerViscosityLaw::operator = (const viscosityLaw& src){
  viscosityLaw::operator=(src);
  const saturatePowerViscosityLaw* ps = dynamic_cast<const saturatePowerViscosityLaw*>(&src);
  if (ps != NULL){
    _eta0 = ps->_eta0;
    _q = ps->_q;
  }
  return *this;
};
void saturatePowerViscosityLaw::get(const double p, double& R, double& dR) const{
  R = _eta0 * pow(p,_q);
  if(p <1.e-6 and _q < 1.)
  {
    R =  _eta0 * pow(1.e-6,_q)*(p/1.e-6); 
    dR = _eta0 * pow(1.e-6,_q)   /1.e-6;
  }
  else
  {
    R = _eta0 * pow(p,_q);
    dR = _eta0 * pow(p,_q-1.) * _q;
  }
  
};




saturateSinhViscosityLaw::saturateSinhViscosityLaw(const int num, const double eta0, const double eta1, const double g0, const bool init):
                viscosityLaw(num,init),_eta0(eta0),_eta1(eta1),_g0(g0){};
saturateSinhViscosityLaw::saturateSinhViscosityLaw(const saturateSinhViscosityLaw& src):viscosityLaw(src),_eta0(src._eta0),_eta1(src._eta1),_g0(src._g0){};
saturateSinhViscosityLaw& saturateSinhViscosityLaw::operator = (const viscosityLaw& src){
  viscosityLaw::operator=(src);
  const saturateSinhViscosityLaw* ps = dynamic_cast<const saturateSinhViscosityLaw*>(&src);
  if (ps != NULL){
    _eta0 = ps->_eta0;
    _eta1 = ps->_eta1;
    _g0 = ps->_g0;
  }
  return *this;
};
void saturateSinhViscosityLaw::get(const double p, double& R, double& dR) const{
  double pa = p/_g0;
  if (pa< 1e-6){
    R = _eta0;
    dR = 0.;
  }
  else
  {
    R = _eta1 + (_eta0-_eta1)*pa/sinh(pa);
    dR = (_eta0-_eta1)*(1./sinh(pa) - (pa*cosh(pa))/(sinh(pa)*sinh(pa)))/_g0 ;
  }
};
