//
// C++ Interface: Void State
//
// Description: Storing class for computation of void state
//
//
// Author:  <V.-D. Nguyen>, (C) 2019
//
// Copyright: See COPYING file that comes with this distribution
//

#include "voidStateEvolutionLaw.h"
#include "ipNonLocalPorosity.h"

TrivialEvolutionLaw::TrivialEvolutionLaw(const int num): voidStateEvolutionLaw(num){

};

void TrivialEvolutionLaw::createIPVariable(const double f0, IPVoidState* &ipv) const {
  if(ipv != NULL) delete ipv;
  ipv = new IPVoidState(f0,1.,1,0.5);
};

void TrivialEvolutionLaw::evolve(const double yieldfV0, const double yieldfV1,
                        const double DeltahatD, const double DeltahatQ, const double DeltahatP,
                        const IPNonLocalPorosity *q0Porous, const IPNonLocalPorosity *q1Porous,
                        const IPVoidState& sPrev, IPVoidState& sCur) const
{

  sCur.getRefToRoughVoidLigamentRatio()  = sPrev.getRoughVoidLigamentRatio();

  sCur.getRefToVoidLigamentRatio() = sPrev.getVoidLigamentRatio();
  sCur.getRefToDVoidLigamentRatioDYieldPorosity() = 0.;
  sCur.getRefToDVoidLigamentRatioDVolumetricPlasticDeformation()  = 0.;
  sCur.getRefToDVoidLigamentRatioDDeviatoricPlasticDeformation() = 0.;
  sCur.getRefToDVoidLigamentRatioDMatrixPlasticDeformation() = 0.;

  sCur.getRefToVoidAspectRatio()  = sPrev.getVoidAspectRatio();
  sCur.getRefToDVoidAspectRatioDYieldPorosity() = 0.;
  sCur.getRefToDVoidAspectRatioDVolumetricPlasticDeformation() = 0.;
  sCur.getRefToDVoidAspectRatioDDeviatoricPlasticDeformation()  = 0.;
  sCur.getRefToDVoidAspectRatioDMatrixPlasticDeformation() = 0.;

  sCur.getRefToVoidSpacingRatio()  = sPrev.getVoidSpacingRatio();
  sCur.getRefToDVoidSpacingRatioDYieldPorosity() = 0.;
  sCur.getRefToDVoidSpacingRatioDVolumetricPlasticDeformation() = 0.;
  sCur.getRefToDVoidSpacingRatioDDeviatoricPlasticDeformation() = 0.;
  sCur.getRefToDVoidSpacingRatioDMatrixPlasticDeformation()  = 0.;

  sCur.getRefToVoidShapeFactor() = sPrev.getVoidShapeFactor();
  sCur.getRefToDVoidShapeFactorDYieldPorosity() = 0.;
  sCur.getRefToDVoidShapeFactorDVolumetricPlasticDeformation() = 0.;
  sCur.getRefToDVoidShapeFactorDDeviatoricPlasticDeformation() = 0.;
  sCur.getRefToDVoidShapeFactorDMatrixPlasticDeformation() = 0.;
};

SphericalVoidStateEvolutionLawWithMatrixPlasticDeformation::SphericalVoidStateEvolutionLawWithMatrixPlasticDeformation(const int num, const double l0, const double k):
voidStateEvolutionLaw(num),_lambda0(l0),_kappa(k){
  _ChiRegularizedFunction = NULL;
  //_ChiRegularizedFunction = new fourVariableExponentialSaturationScalarFunction(1.e-3, 1.e-6, 0.98, 0.995);
};
SphericalVoidStateEvolutionLawWithMatrixPlasticDeformation::SphericalVoidStateEvolutionLawWithMatrixPlasticDeformation(const SphericalVoidStateEvolutionLawWithMatrixPlasticDeformation& src):
  voidStateEvolutionLaw(src),_lambda0(src._lambda0),_kappa(src._kappa){
  _ChiRegularizedFunction = NULL;
  if (src._ChiRegularizedFunction != NULL){
    _ChiRegularizedFunction = src._ChiRegularizedFunction->clone();
  }
}
SphericalVoidStateEvolutionLawWithMatrixPlasticDeformation::~SphericalVoidStateEvolutionLawWithMatrixPlasticDeformation() {
  if (_ChiRegularizedFunction != NULL){
    delete _ChiRegularizedFunction;
     _ChiRegularizedFunction = NULL;
  }
};

void SphericalVoidStateEvolutionLawWithMatrixPlasticDeformation::setRegularizedFunction(const scalarFunction& f){
  if (_ChiRegularizedFunction != NULL){
    delete _ChiRegularizedFunction;
  }
  _ChiRegularizedFunction = f.clone();
};


void SphericalVoidStateEvolutionLawWithMatrixPlasticDeformation::createIPVariable(const double f0, IPVoidState* &ipv) const{
  if(ipv != NULL) delete ipv;
  ipv = new IPVoidState(f0,_lambda0,1.,0.5);
};
// in general, physics changes when coalesence occurs
void SphericalVoidStateEvolutionLawWithMatrixPlasticDeformation::evolve(const double yieldfV0, const double yieldfV1,
                        const double DeltahatD, const double DeltahatQ, const double DeltahatP,
                        const IPNonLocalPorosity *q0Porous, const IPNonLocalPorosity *q1Porous,
                        const IPVoidState& sPrev, IPVoidState& sCur) const
{
  // elastic predictor
  sCur.getRefToRoughVoidLigamentRatio()  = sPrev.getRoughVoidLigamentRatio();

  sCur.getRefToVoidLigamentRatio() = sPrev.getVoidLigamentRatio();
  sCur.getRefToDVoidLigamentRatioDYieldPorosity() = 0.;
  sCur.getRefToDVoidLigamentRatioDVolumetricPlasticDeformation()  = 0.;
  sCur.getRefToDVoidLigamentRatioDDeviatoricPlasticDeformation() = 0.;
  sCur.getRefToDVoidLigamentRatioDMatrixPlasticDeformation() = 0.;

  sCur.getRefToVoidAspectRatio()  = sPrev.getVoidAspectRatio();
  sCur.getRefToDVoidAspectRatioDYieldPorosity() = 0.;
  sCur.getRefToDVoidAspectRatioDVolumetricPlasticDeformation() = 0.;
  sCur.getRefToDVoidAspectRatioDDeviatoricPlasticDeformation()  = 0.;
  sCur.getRefToDVoidAspectRatioDMatrixPlasticDeformation() = 0.;

  sCur.getRefToVoidSpacingRatio()  = sPrev.getVoidSpacingRatio();
  sCur.getRefToDVoidSpacingRatioDYieldPorosity() = 0.;
  sCur.getRefToDVoidSpacingRatioDVolumetricPlasticDeformation() = 0.;
  sCur.getRefToDVoidSpacingRatioDDeviatoricPlasticDeformation() = 0.;
  sCur.getRefToDVoidSpacingRatioDMatrixPlasticDeformation()  = 0.;

  sCur.getRefToVoidShapeFactor() = sPrev.getVoidShapeFactor();
  sCur.getRefToDVoidShapeFactorDYieldPorosity() = 0.;
  sCur.getRefToDVoidShapeFactorDVolumetricPlasticDeformation() = 0.;
  sCur.getRefToDVoidShapeFactorDDeviatoricPlasticDeformation() = 0.;
  sCur.getRefToDVoidShapeFactorDMatrixPlasticDeformation() = 0.;

  // plastic corrector
  if (sPrev.isSaturated())
  {
    sCur.saturate(true);
    // copy only
  }
  else
  {
    // void shape factor and void aspect ratio do not change
    // update void spacing from previous state
    double lambdaPrev = sPrev.getVoidSpacingRatio();
    double& lambda = sCur.getRefToVoidSpacingRatio();
    double& DlambdaDHatP = sCur.getRefToDVoidSpacingRatioDMatrixPlasticDeformation();
    if (DeltahatP >= 0.)
    {
      lambda = lambdaPrev*exp(_kappa*DeltahatP);
      DlambdaDHatP =  _kappa*lambda;
    }
    else
    {
      lambda = lambdaPrev;
      DlambdaDHatP = 0.;
    }
    // compute void ligament ratio
    // we have X = [(3*gamma*f*lambda)/W]^(1/3)
    // we also Xprev = [(3*gamma*fPrev*lambdaPrev)/W]^(1/3)
    // leading (X/Xprev)^3 = f*lambda/(fPrev*lambdaPrev)  = fact
    //
    double fact = yieldfV1*lambda/(yieldfV0*lambdaPrev);
    // estimate rough value of void ligamant
    double roughChiPrev = sPrev.getRoughVoidLigamentRatio();
    double& roughChi = sCur.getRefToRoughVoidLigamentRatio();
    roughChi = roughChiPrev*pow(fact,1./3.);
    // void ligament ratio
    double& Chi = sCur.getRefToVoidLigamentRatio();
    Chi = roughChi;
    double DiffChiComputed = 1.;

    if (_ChiRegularizedFunction == NULL)
    {
      // estimate void ligament
      if (Chi >= _voidLigamentRatioAtFailure)
      {
        sCur.saturate(true);
        double chiPrev = sPrev.getVoidLigamentRatio();
        DiffChiComputed = (_voidLigamentRatioAtFailure - chiPrev)/(Chi - chiPrev);
        Chi = _voidLigamentRatioAtFailure;
      }
      else
      {
        sCur.saturate(false);
      }
    }
    else
    {
      // never saturated
      //
      Chi = _ChiRegularizedFunction->getVal(roughChi);
      DiffChiComputed = _ChiRegularizedFunction->getDiff(roughChi);
    }
    // derivatives
    double& dChidf = sCur.getRefToDVoidLigamentRatioDYieldPorosity();
    double& dChidhatP = sCur.getRefToDVoidLigamentRatioDMatrixPlasticDeformation();
    // value
    dChidf = DiffChiComputed*roughChi/(3.*yieldfV1);
    dChidhatP =  DiffChiComputed*DlambdaDHatP*roughChi/(3.*lambda);
  }
};


SphericalVoidStateEvolutionLawWithDeviatoricPlasticDeformation::SphericalVoidStateEvolutionLawWithDeviatoricPlasticDeformation(const int num, const double l0, const double k):
SphericalVoidStateEvolutionLawWithMatrixPlasticDeformation(num,l0,k){};
SphericalVoidStateEvolutionLawWithDeviatoricPlasticDeformation::SphericalVoidStateEvolutionLawWithDeviatoricPlasticDeformation(const SphericalVoidStateEvolutionLawWithDeviatoricPlasticDeformation& src):
SphericalVoidStateEvolutionLawWithMatrixPlasticDeformation(src){};
SphericalVoidStateEvolutionLawWithDeviatoricPlasticDeformation::~SphericalVoidStateEvolutionLawWithDeviatoricPlasticDeformation(){};

void SphericalVoidStateEvolutionLawWithDeviatoricPlasticDeformation::evolve(const double yieldfV0, const double yieldfV1,
                        const double DeltahatD, const double DeltahatQ, const double DeltahatP,
                        const IPNonLocalPorosity *q0Porous, const IPNonLocalPorosity *q1Porous,
                        const IPVoidState& sPrev, IPVoidState& sCur) const{

  // elastic predictor
  sCur.getRefToRoughVoidLigamentRatio()  = sPrev.getRoughVoidLigamentRatio();

  sCur.getRefToVoidLigamentRatio() = sPrev.getVoidLigamentRatio();
  sCur.getRefToDVoidLigamentRatioDYieldPorosity() = 0.;
  sCur.getRefToDVoidLigamentRatioDVolumetricPlasticDeformation()  = 0.;
  sCur.getRefToDVoidLigamentRatioDDeviatoricPlasticDeformation() = 0.;
  sCur.getRefToDVoidLigamentRatioDMatrixPlasticDeformation() = 0.;

  sCur.getRefToVoidAspectRatio()  = sPrev.getVoidAspectRatio();
  sCur.getRefToDVoidAspectRatioDYieldPorosity() = 0.;
  sCur.getRefToDVoidAspectRatioDVolumetricPlasticDeformation() = 0.;
  sCur.getRefToDVoidAspectRatioDDeviatoricPlasticDeformation()  = 0.;
  sCur.getRefToDVoidAspectRatioDMatrixPlasticDeformation() = 0.;

  sCur.getRefToVoidSpacingRatio()  = sPrev.getVoidSpacingRatio();
  sCur.getRefToDVoidSpacingRatioDYieldPorosity() = 0.;
  sCur.getRefToDVoidSpacingRatioDVolumetricPlasticDeformation() = 0.;
  sCur.getRefToDVoidSpacingRatioDDeviatoricPlasticDeformation() = 0.;
  sCur.getRefToDVoidSpacingRatioDMatrixPlasticDeformation()  = 0.;

  sCur.getRefToVoidShapeFactor() = sPrev.getVoidShapeFactor();
  sCur.getRefToDVoidShapeFactorDYieldPorosity() = 0.;
  sCur.getRefToDVoidShapeFactorDVolumetricPlasticDeformation() = 0.;
  sCur.getRefToDVoidShapeFactorDDeviatoricPlasticDeformation() = 0.;
  sCur.getRefToDVoidShapeFactorDMatrixPlasticDeformation() = 0.;

  // plastic corrector
  if (sPrev.isSaturated())
  {
    sCur.saturate(true);
    // copy only
  }
  else
  {
    // void shape factor and void aspect ratio do not change
    // update void spacing from previous state
    double lambdaPrev = sPrev.getVoidSpacingRatio();
    double& lambda = sCur.getRefToVoidSpacingRatio();
    double& DlambdaDHatD = sCur.getRefToDVoidSpacingRatioDDeviatoricPlasticDeformation();
    if (DeltahatD >= 0.)
    {
      lambda = lambdaPrev*exp(_kappa*DeltahatD);
      DlambdaDHatD =  _kappa*lambda;
    }
    else
    {
      lambda = lambdaPrev;
      DlambdaDHatD = 0.;
    }
    // compute void ligament ratio
    // we have X = [(3*gamma*f*lambda)/W]^(1/3)
    // we also Xprev = [(3*gamma*fPrev*lambdaPrev)/W]^(1/3)
    // leading (X/Xprev)^3 = f*lambda/(fPrev*lambdaPrev)  = fact
    //
    double fact = yieldfV1*lambda/(yieldfV0*lambdaPrev);
    // estimate rough value of void ligamant
    double roughChiPrev = sPrev.getRoughVoidLigamentRatio();
    double& roughChi = sCur.getRefToRoughVoidLigamentRatio();
    roughChi = roughChiPrev*pow(fact,1./3.);
    // void ligament ratio
    double& Chi = sCur.getRefToVoidLigamentRatio();
    Chi = roughChi;
    double DiffChiComputed = 1.;

    if (_ChiRegularizedFunction == NULL)
    {
      // estimate void ligament
      if (Chi >= _voidLigamentRatioAtFailure)
      {
        sCur.saturate(true);
        double chiPrev = sPrev.getVoidLigamentRatio();
        DiffChiComputed = (_voidLigamentRatioAtFailure - chiPrev)/(Chi - chiPrev);
        Chi = _voidLigamentRatioAtFailure;
      }
      else
      {
        sCur.saturate(false);
      }
    }
    else
    {
      // never saturated
      //
      Chi = _ChiRegularizedFunction->getVal(roughChi);
      DiffChiComputed = _ChiRegularizedFunction->getDiff(roughChi);
    }
    // derivatives
    double& dChidf = sCur.getRefToDVoidLigamentRatioDYieldPorosity();
    double& dChidhatD = sCur.getRefToDVoidLigamentRatioDDeviatoricPlasticDeformation();
    // value
    dChidf = DiffChiComputed*roughChi/(3.*yieldfV1);
    dChidhatD =  DiffChiComputed*DlambdaDHatD*roughChi/(3.*lambda);
  }
};

SpheroidalVoidStateEvolutionLawWithAspectRatioEvolution::SpheroidalVoidStateEvolutionLawWithAspectRatioEvolution(const int num, const double l0, const double W0, const double k):
voidStateEvolutionLaw(num),_lambda0(l0),_W0(W0),_kappa(k){
};

SpheroidalVoidStateEvolutionLawWithAspectRatioEvolution::SpheroidalVoidStateEvolutionLawWithAspectRatioEvolution(const SpheroidalVoidStateEvolutionLawWithAspectRatioEvolution& src):
  voidStateEvolutionLaw(src),_lambda0(src._lambda0),_W0(src._W0),_kappa(src._kappa){
}
SpheroidalVoidStateEvolutionLawWithAspectRatioEvolution::~SpheroidalVoidStateEvolutionLawWithAspectRatioEvolution() {

};

void SpheroidalVoidStateEvolutionLawWithAspectRatioEvolution::createIPVariable(const double f0, IPVoidState* &ipv) const{
  if(ipv != NULL) delete ipv;
  ipv = new IPVoidState(f0,_lambda0,_W0,0.5);
};
// in general, physics changes when coalesence occurs
void SpheroidalVoidStateEvolutionLawWithAspectRatioEvolution::evolve(const double yieldfV0, const double yieldfV1,
                        const double DeltahatD, const double DeltahatQ, const double DeltahatP,
                        const IPNonLocalPorosity *q0Porous, const IPNonLocalPorosity *q1Porous,
                        const IPVoidState& sPrev, IPVoidState& sCur) const{
  // elastic predictor
  sCur.getRefToVoidLigamentRatio() = sPrev.getVoidLigamentRatio();
  sCur.getRefToDVoidLigamentRatioDYieldPorosity() = 0.;
  sCur.getRefToDVoidLigamentRatioDVolumetricPlasticDeformation()  = 0.;
  sCur.getRefToDVoidLigamentRatioDDeviatoricPlasticDeformation() = 0.;
  sCur.getRefToDVoidLigamentRatioDMatrixPlasticDeformation() = 0.;

  sCur.getRefToVoidAspectRatio()  = sPrev.getVoidAspectRatio();
  sCur.getRefToDVoidAspectRatioDYieldPorosity() = 0.;
  sCur.getRefToDVoidAspectRatioDVolumetricPlasticDeformation() = 0.;
  sCur.getRefToDVoidAspectRatioDDeviatoricPlasticDeformation()  = 0.;
  sCur.getRefToDVoidAspectRatioDMatrixPlasticDeformation() = 0.;

  sCur.getRefToVoidSpacingRatio()  = sPrev.getVoidSpacingRatio();
  sCur.getRefToDVoidSpacingRatioDYieldPorosity() = 0.;
  sCur.getRefToDVoidSpacingRatioDVolumetricPlasticDeformation() = 0.;
  sCur.getRefToDVoidSpacingRatioDDeviatoricPlasticDeformation() = 0.;
  sCur.getRefToDVoidSpacingRatioDMatrixPlasticDeformation()  = 0.;

  sCur.getRefToVoidShapeFactor() = sPrev.getVoidShapeFactor();
  sCur.getRefToDVoidShapeFactorDYieldPorosity() = 0.;
  sCur.getRefToDVoidShapeFactorDVolumetricPlasticDeformation() = 0.;
  sCur.getRefToDVoidShapeFactorDDeviatoricPlasticDeformation() = 0.;
  sCur.getRefToDVoidShapeFactorDMatrixPlasticDeformation() = 0.;

  // evolutation with plastic deformation

  if (sPrev.isSaturated())
  { 
    // if saturated, do nothing
    sCur.saturate(true);
  }
  else
  {
    // previous values
    double lambdaPrev = sPrev.getVoidSpacingRatio();
    double XPrev = sPrev.getVoidLigamentRatio();
    double WPrev = sPrev.getVoidAspectRatio();

    // Get ipvs
    double& W = sCur.getRefToVoidAspectRatio();
    double& dWdHatD = sCur.getRefToDVoidAspectRatioDDeviatoricPlasticDeformation();

    double& X = sCur.getRefToVoidLigamentRatio();
    double& dXdHatD = sCur.getRefToDVoidLigamentRatioDDeviatoricPlasticDeformation();

    double& lambda = sCur.getRefToVoidSpacingRatio();
    double& DlambdaDhatD = sCur.getRefToDVoidSpacingRatioDDeviatoricPlasticDeformation();

  
     // void spacing
    if (DeltahatD > 0.)
    {
      lambda = lambdaPrev*exp(_kappa*DeltahatD);
      DlambdaDhatD = _kappa*lambda;
    }
    else
    {
      lambda = lambdaPrev;
      DlambdaDhatD = 0.;
    }

    double Zprev = XPrev*WPrev;
    double DeltaZ = 0.;
    if (_kappa > 0.)
    {      
      DeltaZ =  1.5*(lambda-lambdaPrev)/_kappa;
    }
    else
    {
      DeltaZ = 1.5*lambda*DeltahatD;
    }
    double Z = Zprev+DeltaZ;

    X = sqrt(1.5-(1.5-XPrev*XPrev)*Zprev/Z);
    dXdHatD = (3./4.)*(lambda*X/Z)*(1.5/X/X-1.);


    if (X >= _voidLigamentRatioAtFailure)
    {
      double ff = (_voidLigamentRatioAtFailure - XPrev)/(X - XPrev);
      dXdHatD *= ff;
      X = _voidLigamentRatioAtFailure;
      sCur.saturate(true);
    }
    else
    {
      sCur.saturate(false);
    }

    if (sCur.isSaturated())
    {
      W = WPrev;
      dWdHatD = 0.;
    }
    else
    {
      W = Z/X;
      dWdHatD = (9./4.)*(lambda/X)*(1.-0.5/X/X);
    }
  }
};

SpheroidalVoidStateEvolutionLawWithAspectRatioEvolutionAndShapeFactorEvolution::SpheroidalVoidStateEvolutionLawWithAspectRatioEvolutionAndShapeFactorEvolution(const int num, const double l0, const double W0, const double g0, const double gf, const double k):
voidStateEvolutionLaw(num),_lambda0(l0),_W0(W0),_kappa(k),_gamma0(g0),_gammaf(gf){
};
SpheroidalVoidStateEvolutionLawWithAspectRatioEvolutionAndShapeFactorEvolution::SpheroidalVoidStateEvolutionLawWithAspectRatioEvolutionAndShapeFactorEvolution(const SpheroidalVoidStateEvolutionLawWithAspectRatioEvolutionAndShapeFactorEvolution& src):
  voidStateEvolutionLaw(src),_lambda0(src._lambda0),_W0(src._W0),_kappa(src._kappa),_gamma0(src._gamma0),_gammaf(src._gammaf){
}
SpheroidalVoidStateEvolutionLawWithAspectRatioEvolutionAndShapeFactorEvolution::~SpheroidalVoidStateEvolutionLawWithAspectRatioEvolutionAndShapeFactorEvolution() {
};


void SpheroidalVoidStateEvolutionLawWithAspectRatioEvolutionAndShapeFactorEvolution::createIPVariable(const double f0, IPVoidState* &ipv) const{
  if(ipv != NULL) delete ipv;
  ipv = new IPVoidState(f0,_lambda0,_W0,_gamma0);
};
// in general, physics changes when coalesence occurs
void SpheroidalVoidStateEvolutionLawWithAspectRatioEvolutionAndShapeFactorEvolution::evolve(const double yieldfV0, const double yieldfV1,
                        const double DeltahatD, const double DeltahatQ, const double DeltahatP,
                        const IPNonLocalPorosity *q0Porous, const IPNonLocalPorosity *q1Porous,
                        const IPVoidState& sPrev, IPVoidState& sCur) const{
  // elastic predictor
  sCur.getRefToRoughVoidLigamentRatio()  = sPrev.getRoughVoidLigamentRatio();

  sCur.getRefToVoidLigamentRatio() = sPrev.getVoidLigamentRatio();
  sCur.getRefToDVoidLigamentRatioDYieldPorosity() = 0.;
  sCur.getRefToDVoidLigamentRatioDVolumetricPlasticDeformation()  = 0.;
  sCur.getRefToDVoidLigamentRatioDDeviatoricPlasticDeformation() = 0.;
  sCur.getRefToDVoidLigamentRatioDMatrixPlasticDeformation() = 0.;

  sCur.getRefToVoidAspectRatio()  = sPrev.getVoidAspectRatio();
  sCur.getRefToDVoidAspectRatioDYieldPorosity() = 0.;
  sCur.getRefToDVoidAspectRatioDVolumetricPlasticDeformation() = 0.;
  sCur.getRefToDVoidAspectRatioDDeviatoricPlasticDeformation()  = 0.;
  sCur.getRefToDVoidAspectRatioDMatrixPlasticDeformation() = 0.;

  sCur.getRefToVoidSpacingRatio()  = sPrev.getVoidSpacingRatio();
  sCur.getRefToDVoidSpacingRatioDYieldPorosity() = 0.;
  sCur.getRefToDVoidSpacingRatioDVolumetricPlasticDeformation() = 0.;
  sCur.getRefToDVoidSpacingRatioDDeviatoricPlasticDeformation() = 0.;
  sCur.getRefToDVoidSpacingRatioDMatrixPlasticDeformation()  = 0.;

  sCur.getRefToVoidShapeFactor() = sPrev.getVoidShapeFactor();
  sCur.getRefToDVoidShapeFactorDYieldPorosity() = 0.;
  sCur.getRefToDVoidShapeFactorDVolumetricPlasticDeformation() = 0.;
  sCur.getRefToDVoidShapeFactorDDeviatoricPlasticDeformation() = 0.;
  sCur.getRefToDVoidShapeFactorDMatrixPlasticDeformation() = 0.;
  
 
  if (sPrev.isSaturated())
  {
    sCur.saturate(true);
  }
  else
  {
     // Get ipvs
    double& W = sCur.getRefToVoidAspectRatio();
    double& dWdf = sCur.getRefToDVoidAspectRatioDYieldPorosity();
    double& dWdHatQ = sCur.getRefToDVoidAspectRatioDVolumetricPlasticDeformation();
    double& dWdHatD = sCur.getRefToDVoidAspectRatioDDeviatoricPlasticDeformation();
    double& dWdHatP = sCur.getRefToDVoidAspectRatioDMatrixPlasticDeformation();

    double& roughX = sCur.getRefToRoughVoidLigamentRatio();
    double& X = sCur.getRefToVoidLigamentRatio();
    double& dXdf = sCur.getRefToDVoidLigamentRatioDYieldPorosity();
    double& dXdHatQ = sCur.getRefToDVoidLigamentRatioDVolumetricPlasticDeformation();
    double& dXdHatD = sCur.getRefToDVoidLigamentRatioDDeviatoricPlasticDeformation();
    double& dXdHatP = sCur.getRefToDVoidLigamentRatioDMatrixPlasticDeformation();

    double& gamma = sCur.getRefToVoidShapeFactor();
    double& dgammadf = sCur.getRefToDVoidShapeFactorDYieldPorosity();
    double& dgammadHatQ = sCur.getRefToDVoidShapeFactorDVolumetricPlasticDeformation();
    double& dgammadHatD = sCur.getRefToDVoidShapeFactorDDeviatoricPlasticDeformation();
    double& dgammadHatP = sCur.getRefToDVoidShapeFactorDMatrixPlasticDeformation();

    double& lambda = sCur.getRefToVoidSpacingRatio();
    double& DlambdaDhatD = sCur.getRefToDVoidSpacingRatioDDeviatoricPlasticDeformation();
    
    // previpis value
    double roughXPrev = sPrev.getRoughVoidLigamentRatio();
    double lambdaPrev = sPrev.getVoidSpacingRatio();
    double XPrev = sPrev.getVoidLigamentRatio();
    double WPrev = sPrev.getVoidAspectRatio();
    double gammaPrev = sPrev.getVoidShapeFactor();
    
    const IPNonLocalPorosity* ipvAtCoalescence = q1Porous->getConstRefToIPCoalescence().getIPvAtCoalescenceOnset();
    const IPVoidState& voiStateAtCoales = ipvAtCoalescence->getConstRefToIPVoidState();
    double Xc = voiStateAtCoales.getVoidLigamentRatio();
    //Msg::Error("pass here Xc = %e",Xc);
    double agamma = (_gammaf - _gamma0)/(1.-Xc);

    // void spacing
    lambda = lambdaPrev*exp(_kappa*DeltahatD);
    DlambdaDhatD =_kappa*lambda;;

    double Cprev = 3.*gammaPrev/(XPrev*XPrev);
    double Zprev = roughXPrev*WPrev;
    double Z = Zprev+1.5*(lambda-lambdaPrev)/_kappa;

    double A = Cprev/(Cprev-1.)/Zprev;
    double C = (A*Z)/(A*Z-1.);
    if (C < 3.*_gammaf)
    {
      C = 3.*_gammaf;
      X = 1.;
      gamma = _gammaf;
    }
    else
    {
      // solve X*X - (3/C)*(gamma) = 0
      // aX*X + b*X + C = 0
      double a, b, c;
      a = 1.;
      b = -3.*agamma/C;
      c = -3.*(gammaPrev-agamma*XPrev)/C;
      // two solution
      double d = (b*b - 4.*a*c);
      if (d < 0) Msg::Error("No solution d = %e",d);
      else
      {
        double x1, x2;
        x1 = (-b - sqrt(d))/(2.*a);
        x2 = (-b + sqrt(d))/(2.*a);
        if (x1 > XPrev) X = x1;
        else X = x2;

        //Msg::Info("two solution x1 =%e x2 = %e X=%e",x1,x2,X);
      }

    }

    roughX = X;
    double DXDroughX = 1.;
    if (X > _voidLigamentRatioAtFailure)
    {
      DXDroughX = (_voidLigamentRatioAtFailure - XPrev)/(X - XPrev);
      X = _voidLigamentRatioAtFailure;
      sCur.saturate(true);
    }
    else
    {
      sCur.saturate(false);
    }
    gamma = gammaPrev+ agamma*(X-XPrev);
    double droughXdhatD = 0.75*(lambda*X/Z)*(3.*gamma/(X*X)-1.)*(2.*gamma)/(2.*gamma-X*agamma);

    dXdf =0.;
    dXdHatQ = 0.;
    dXdHatD = DXDroughX*droughXdhatD;
    dXdHatP = 0.;

    if (sCur.isSaturated())
    {
      W = WPrev;
      dWdf = 0.;
      dWdHatQ = 0.;
      dWdHatD = 0.;
      dWdHatP = 0.;
    }
    else
    {
      W = Z/X;
      dWdf = 0.;
      dWdHatQ = 0.;
      dWdHatD = (9./4.)*(lambda/X)*(1.-gamma/(X*X)) - W*agamma*dXdHatD/(2.*gamma);
      dWdHatP = 0.;
    }

    dgammadf = agamma* dXdf;
    dgammadHatQ = agamma* dXdHatQ;
    dgammadHatD = agamma* dXdHatD;
    dgammadHatP = agamma* dXdHatP;
  }
};

VoidStateLinearEvolutionLaw::VoidStateLinearEvolutionLaw(const int num, const double l0, const double kappa, const double kL):
  voidStateEvolutionLaw(num), _lambda0(l0), _kappa(kappa),_kLigament(kL){};

VoidStateLinearEvolutionLaw::VoidStateLinearEvolutionLaw(const VoidStateLinearEvolutionLaw& src):
    voidStateEvolutionLaw(src), _lambda0(src._lambda0), _kappa(src._kappa),_kLigament(src._kLigament){}

void VoidStateLinearEvolutionLaw::createIPVariable(const double f0, IPVoidState* &ipv) const
{
  if(ipv != NULL) delete ipv;
  ipv = new IPVoidState(f0,_lambda0,1.,0.5);
}

void VoidStateLinearEvolutionLaw::evolve(const double yieldfV0, const double yieldfV1,
                        const double DeltahatD, const double DeltahatQ, const double DeltahatP,
                        const IPNonLocalPorosity *q0Porous, const IPNonLocalPorosity *q1Porous,
                        const IPVoidState& sPrev, IPVoidState& sCur) const
{
  // elastic predictor
  sCur.getRefToVoidLigamentRatio() = sPrev.getVoidLigamentRatio();
  sCur.getRefToDVoidLigamentRatioDYieldPorosity() = 0.;
  sCur.getRefToDVoidLigamentRatioDVolumetricPlasticDeformation()  = 0.;
  sCur.getRefToDVoidLigamentRatioDDeviatoricPlasticDeformation() = 0.;
  sCur.getRefToDVoidLigamentRatioDMatrixPlasticDeformation() = 0.;

  sCur.getRefToVoidAspectRatio()  = sPrev.getVoidAspectRatio();
  sCur.getRefToDVoidAspectRatioDYieldPorosity() = 0.;
  sCur.getRefToDVoidAspectRatioDVolumetricPlasticDeformation() = 0.;
  sCur.getRefToDVoidAspectRatioDDeviatoricPlasticDeformation()  = 0.;
  sCur.getRefToDVoidAspectRatioDMatrixPlasticDeformation() = 0.;

  sCur.getRefToVoidSpacingRatio()  = sPrev.getVoidSpacingRatio();
  sCur.getRefToDVoidSpacingRatioDYieldPorosity() = 0.;
  sCur.getRefToDVoidSpacingRatioDVolumetricPlasticDeformation() = 0.;
  sCur.getRefToDVoidSpacingRatioDDeviatoricPlasticDeformation() = 0.;
  sCur.getRefToDVoidSpacingRatioDMatrixPlasticDeformation()  = 0.;

  sCur.getRefToVoidShapeFactor() = sPrev.getVoidShapeFactor();
  sCur.getRefToDVoidShapeFactorDYieldPorosity() = 0.;
  sCur.getRefToDVoidShapeFactorDVolumetricPlasticDeformation() = 0.;
  sCur.getRefToDVoidShapeFactorDDeviatoricPlasticDeformation() = 0.;
  sCur.getRefToDVoidShapeFactorDMatrixPlasticDeformation() = 0.;

  // evolutation with plastic deformation

  if (sPrev.isSaturated())
  { 
    // if saturated, do nothing
    sCur.saturate(true);
  }
  else
  {
    // previous values
    double XPrev = sPrev.getVoidLigamentRatio();
    double& X = sCur.getRefToVoidLigamentRatio();
    double& dXdHatD = sCur.getRefToDVoidLigamentRatioDDeviatoricPlasticDeformation();

     // void spacing
    if (DeltahatD >= 0.)
    {
      X = XPrev + _kLigament*DeltahatD;
      dXdHatD = _kLigament;
    }
    else
    {
      X = XPrev;
      dXdHatD = 0.;
    };
    
    
    if (X >= _voidLigamentRatioAtFailure)
    {
      double ff = (_voidLigamentRatioAtFailure - XPrev)/(X - XPrev);
      X = _voidLigamentRatioAtFailure;
      dXdHatD *= ff;
      sCur.saturate(true);
    }
    else
    {
      sCur.saturate(false);
    }
  }
};


VoidStateExponentialEvolutionLaw::VoidStateExponentialEvolutionLaw(const int num, const double l0, const double kappa, const double kL):
  VoidStateLinearEvolutionLaw(num,l0,kappa,kL){
  _voidLigamentRatioAtFailure = 0.999;  
};

VoidStateExponentialEvolutionLaw::VoidStateExponentialEvolutionLaw(const VoidStateExponentialEvolutionLaw& src):
    VoidStateLinearEvolutionLaw(src){}

void VoidStateExponentialEvolutionLaw::evolve(const double yieldfV0, const double yieldfV1,
                        const double DeltahatD, const double DeltahatQ, const double DeltahatP,
                        const IPNonLocalPorosity *q0Porous, const IPNonLocalPorosity *q1Porous,
                        const IPVoidState& sPrev, IPVoidState& sCur) const
{
  // elastic predictor
  sCur.getRefToVoidLigamentRatio() = sPrev.getVoidLigamentRatio();
  sCur.getRefToDVoidLigamentRatioDYieldPorosity() = 0.;
  sCur.getRefToDVoidLigamentRatioDVolumetricPlasticDeformation()  = 0.;
  sCur.getRefToDVoidLigamentRatioDDeviatoricPlasticDeformation() = 0.;
  sCur.getRefToDVoidLigamentRatioDMatrixPlasticDeformation() = 0.;

  sCur.getRefToVoidAspectRatio()  = sPrev.getVoidAspectRatio();
  sCur.getRefToDVoidAspectRatioDYieldPorosity() = 0.;
  sCur.getRefToDVoidAspectRatioDVolumetricPlasticDeformation() = 0.;
  sCur.getRefToDVoidAspectRatioDDeviatoricPlasticDeformation()  = 0.;
  sCur.getRefToDVoidAspectRatioDMatrixPlasticDeformation() = 0.;

  sCur.getRefToVoidSpacingRatio()  = sPrev.getVoidSpacingRatio();
  sCur.getRefToDVoidSpacingRatioDYieldPorosity() = 0.;
  sCur.getRefToDVoidSpacingRatioDVolumetricPlasticDeformation() = 0.;
  sCur.getRefToDVoidSpacingRatioDDeviatoricPlasticDeformation() = 0.;
  sCur.getRefToDVoidSpacingRatioDMatrixPlasticDeformation()  = 0.;

  sCur.getRefToVoidShapeFactor() = sPrev.getVoidShapeFactor();
  sCur.getRefToDVoidShapeFactorDYieldPorosity() = 0.;
  sCur.getRefToDVoidShapeFactorDVolumetricPlasticDeformation() = 0.;
  sCur.getRefToDVoidShapeFactorDDeviatoricPlasticDeformation() = 0.;
  sCur.getRefToDVoidShapeFactorDMatrixPlasticDeformation() = 0.;

  // evolutation with plastic deformation

  if (sPrev.isSaturated())
  { 
    // if saturated, do nothing
    sCur.saturate(true);
  }
  else
  {
    // previous values
    double XPrev = sPrev.getVoidLigamentRatio();
    double& X = sCur.getRefToVoidLigamentRatio();
    double& dXdHatD = sCur.getRefToDVoidLigamentRatioDDeviatoricPlasticDeformation();

     // void spacing
    if (DeltahatD >= 0.)
    {
      X = _voidLigamentRatioAtFailure- (_voidLigamentRatioAtFailure-XPrev)*exp(-_kLigament*DeltahatD);
      dXdHatD = (_voidLigamentRatioAtFailure-X)*_kLigament;
    }
    else
    {
      X = XPrev;
      dXdHatD = 0.;
    }
  };
};
