//
//
// Description: Non linear solver for mechanic problems
//              quasi-static implicit scheme & dynamic explicit Hulbert-Chung scheme
//
// Author:  <Gauthier BECKER>, (C) 2011
//
// Copyright: See COPYING file that comes with this distribution
//
//

#ifndef _NONLINEARMECHSOLVER_H_
#define _NONLINEARMECHSOLVER_H_

#ifndef SWIG
#include <map>
#include <string>
#include "SVector3.h"
#include "dofManager.h"
#include "simpleFunction.h"
#include "simpleFunctionTime.h"
#include "functionSpace.h"
#include "MInterfaceElement.h"
#include "groupOfElements.h"
#include "partDomain.h"
#include "explicitHulbertChung.h"
#include "contactDomain.h"
#include "NonLinearSolverConfig.h"
#include "eigenSolver.h"
#include "periodicBoundaryCondition.h"
#include "nonLinearMicroBC.h"
#include "ipField.h"
#include "unknownField.h"
#include "strainMapping.h"
#include "homogenizedData.h"
#include "ElementErosionFilter.h"
#include "failureCriterion.h"
#include "ipvariable.h"
#include "endSchemeMonitoring.h"
#include "selectiveUpdate.h"
#include "defoDefoContactDomain.h"

class PView;
class groupOfElements;
class energeticField;
class restartManager;
template<class T1,class T2> class DgC0BilinearTerm;
template<class T1> class DgC0LinearTerm;
template<class T> class staticDofManager;

class archiveForce{
 public:
  int numphys;
  int dim;
  int comp;
  int nstep;
  FILE *FP;
  int lastSaveStep;
  std::set<Dof> vdof; // With a set we are sur that a Dof will not be archived twice
  int contact_root; // rank where the archive is done for contact == -1 if not contact
  double fval; // tempory to get value from system
  archiveForce() : numphys(0), dim(0), comp(0), nstep(1), fval(0.), FP(NULL), contact_root(-1),lastSaveStep(0){}
  archiveForce(const archiveForce &source) : numphys(source.numphys), dim(source.dim),
                                             comp(source.comp), nstep(source.nstep),
                                             fval(source.fval),
                                             FP(source.FP),
                                             contact_root(source.contact_root),
                                             lastSaveStep(source.lastSaveStep)
  {
    for(std::set<Dof>::const_iterator itD = source.vdof.begin(); itD!= source.vdof.end(); ++itD)
    {
      vdof.insert(*itD);
    }
  }
  void openFile()
  {
    // create file only if there is dof to archive
    if(vdof.size()!=0){
      std::ostringstream oss;
      oss << numphys;
      std::string s = oss.str();
      oss.str("");
      oss << comp;
      std::string s2 = oss.str();
     #if defined(HAVE_MPI)
      if(Msg::GetCommSize() != 1){
        oss.str("");
        oss << Msg::GetCommRank();
        s2 += "_part"+oss.str();
      }
     #endif // HAVE_MPI
      std::string fname = "force"+s+"comp"+s2+".csv";
      FP = fopen(fname.c_str(),"w");
    }
    #if defined(HAVE_MPI)
    else if(Msg::GetCommSize() == 1)
    {
      Msg::Warning("Can't archive forces on physical %d because it contains no dofs!",numphys);
    }
    #endif
  }

  archiveForce(int np, int d, int c,int ns) : numphys(np), dim(d), comp(c), nstep(ns), fval(0.), FP(NULL), contact_root(-1),
                          lastSaveStep(0){ if (nstep <1) nstep=1;}
  ~archiveForce(){
    if(FP!=NULL) fclose(FP);
    FP=NULL;
  }
  void clear()
  {
    lastSaveStep = 0;
    if (FP!=NULL) 
    {
      fclose(FP); FP=NULL;
    }
    vdof.clear();
    contact_root = -1;
  }
  void resetArchiving()
  {
    lastSaveStep = 0;
    if(FP!=NULL) fclose(FP);
    openFile(); 
  }
};

class IPDataOnPhysical {
  public:
    std::vector<int> element; // ele and GP
    int physical; // physical
    int dim;
    int ipValue; // ipvalue IPField::Output
    int lastSaveStep;
    int nbArch;
    bool isInitialized;
    IPDataOnPhysical(const int ipval, const int phy, const int d, const int nstep): ipValue(ipval),physical(phy),dim(d),
                    nbArch(nstep),lastSaveStep(0),isInitialized(false){
    };
    ~IPDataOnPhysical(){
    };
    void resetArchiving()
    {
      lastSaveStep = 0;
    }
};

class IntegralVolume
{
  public:
    typedef enum {VolumeIntegral=1, VolumeAverage=2, ActiveDissipationAverage=3} IntegralVolumeType;
    IntegralVolumeType integType;
    int ipVal;
    int physical; // -1 if operation is performed in all domain, >0, operation is performed in this physical
    int lastSaveStep;
    int nbArch;
    FILE* fp;
    IntegralVolume(const int type, const int ip, const int p, const int nstep): 
          integType(IntegralVolumeType(type)),ipVal(ip), physical(p), nbArch(nstep),lastSaveStep(0),
          fp(NULL){}
    ~IntegralVolume(){if (fp!=NULL) fclose(fp);};
    std::string getString() const{
      if (integType == VolumeIntegral)
      {
        if (physical > 0) return "VolumeIntegral_"+int2str(physical);
        else return "VolumeIntegral";
      }
      else if (integType == VolumeAverage) return "Average";
      else if (integType == ActiveDissipationAverage) return "ActiveDissipationAverage";
      else
      {
        Msg::Error("integType %d has not been defined",integType);
      }
    }
    void openFile()
    {
      std::string fname = getString()+"_"+IPField::ToString(ipVal);
		  #if defined(HAVE_MPI)
		  if (Msg::GetCommSize() > 1){
		    fname += "_part"+int2str(Msg::GetCommRank());
		  }
		  #endif //HAVE_MPI
		  fname += ".csv";
      fp =  fopen(fname.c_str(),"w");
    }
    void resetArchiving()
    {
      lastSaveStep = 0;
      if (fp!=NULL) fclose(fp);
      openFile();
    }
};

#endif // SWIG


class nonLinearMechSolver
{
 public:
  typedef std::set<contactDomain*> contactContainer;
  typedef std::set<defoDefoContactDomain*> defoDefoContactContainer;
  enum unknownType{
    displacement = nonLinearBoundaryCondition::position,
    velocity = nonLinearBoundaryCondition::velocity,
    acceleration = nonLinearBoundaryCondition::acceleration
  };

 #ifndef SWIG
  enum solver{ Gmm=0,Taucs=1,Petsc=2, GMRESk=3};
  enum scheme{StaticLinear=0, StaticNonLinear=1, Explicit=2, Multi=3, Implicit=4, Eigen=5};
	enum pathFollowingMethod{GLOBAL_ARC_LENGTH_BASED=0, LOCAL_BASED=1, HYPERELLIPTIC_BASED=2};
	enum pathFollowingLocalIncrementType{DEFO_ENERGY=0, DISSIPATION_ENERGY=1,PLASTIC_ENERGY=2,DAMAGE_ENERGY=3};
	enum pathFollowingLocation{BULK_INTERFACE=0,BULK=1,INTERFACE=2};
	enum elementErosionType{FIRST_IP_FAILED=1, ALL_IP_FAILED=2};
 #endif //SWIG
 protected:
  GModel *pModel;
  int _dim, _tag;
  int _mpiUserDom;
	int _workingRank;
  bool _mpiDofManager; // true -> parallelization of dofManager. By default = true if Msg::GetCommSize() > 1
  dofManager<double> *pAssembler;
  // IPField declared here to allow to pass it from one scheme to an other scheme
  IPField* _ipf;
	// unknown field
  unknownField* _ufield;
  // energetic field
  energeticField* _energField;

//  dgGroupCollection _groups;
  // specific data
  int _numstepImpl, _numstepExpl; // Number of step not used for StaticLinearScheme and number of step between time step evaluation for Explicit Scheme
  unsigned long int _currentStep; // As it is used in different functions now (Onelab control)
  double endtime; // final time not used for StaticLinearScheme but it is not necesary to derive class because (small useless data)
  double _tol, _absTol; // relative and absolute tolerance for iteration not used for StaticLinearScheme but it is not necesary to derive class because (small useless data)
  // adaptative time step if no convergence for Static Non Linear Scheme
  int _maxNRite; // if this number of iteration is reached the time step is reduce
  double _timeStepFactorReduction; // the time step is divided by this value if no convergence (default 2)
  int _maxAttemptStepReduction; // number of attemp to reduce time step (default 6 which leads to a decrease by 64 (2^6) of time step)
  int _niteTimeStepIncrease; // if convergence with nite < _nite the time step is increase (default 5)(but can be greater than its max value given by numstep/endtime)
  double _maximalTimeStep; // maximal time step
  double _minimalTimeStep; // minimal time step
  int nsba; // number of step between two view
  // user solver option given as a string
  std::string _solver_options;

  /* data for restart */
  size_t _beginTime;

  /* specific data for explicit scheme */
  double _beta, _gamma, _alpham, _alphaf, _gammas, _rhoinfty;
  bool _timeStepByBenson;
  std::map<int,MElement*> _mapMElementTmp; // used to compute characteristique element size in the deformed configuration CLEAR THE MAP ???
  std::map<int,MElement*> _mapMElementFirstOrder; // idem

  /* crack tracking */
  FILE *_crackTrackingFile; // name of files for crack tracking no name == NULL and no track

  /* fragmentation */
  FILE *_fragmentationFile;

  std::vector<partDomain*> domainVector;
  // contact
  contactContainer _allContact;
  // defo defo contact BC
  defoDefoContactContainer _allDefoDefoContact;

  // neumann BC
  std::list<nonLinearNeumannBC> allNeumann;
  // dirichlet BC
  std::list<nonLinearDirichletBC> allDirichlet;

  // all periodic groups
  std::list<nonLinearPeriodicBCBetweenTwoGroups> allPeriodic;

  //all average PBC groups
  std::list<nonLinearAveragePeriodicBCBetweenTwoGroups> allAveragePeriodic;

  // all samedisp bc
  std::list<nonLinearSameDisplacementBC> allSameDisp;

  // all samedisp bc between two groups
  std::list<nonLinearSameDisplacementBCBetweenTwoGroups> allSameDispBetweenTwoGroups;
  // fix on face
  std::list<nonLinearFixOnFaceBC> allFixAllFace;

  // constraint BC
  std::list<nonLinearConstraintBC> allConstraint;

  // corner BC
  std::list<nonLinearDirichletBCAtCorner> allCornerConstraint;
  //
  std::list<nonLinearNeumannBCAtCorner> allCornerForce;

  // initial BC
  std::list<initialCondition> allinitial;

  // neumann BC theta (weak enforcement of rotation) group this with allNeumann ?
  std::list<nonLinearNeumannBC> allTheta;

  // dirichlet BC on rigid surface (prescribed the motion of gravity center)
  std::list<rigidContactBC> allContactBC;

  // set with virtual interfaceElements used to prescribed BC put this in an other place ??
  std::map<int,groupOfElements> mapvinter;

  // vector with material law
  std::map<int,materialLaw*> maplaw;
  // physical entities that are initialy broken
  std::vector<int> initbrokeninter;
  // std vector to archive a node displacement
  std::vector<unknownField::archiveNode> anoded;
  // std::vector to archive a force (info musy be store before creation because functionSpace are not initialize)
  std::vector<archiveForce> vaf;
  // std vector to archive ipvariable
  std::vector<IPField::ip2archive> vaip;
  // for view
  std::vector<nlsField::dataBuildView> unknownView;
  std::vector<nlsField::dataBuildView> ipView;
  std::vector<nlsField::dataBuildView> energyView;
  int _energyComputation, _fractureEnergyComputation; // step interval to compute total energy (never if 0)

  /* SWITCH scheme data */
  bool _previousInit; // To known if the initialization as already be made by an other scheme
  bool _notResetedBC; // To known if the BC are modified.
  bool _notResetedContact; // To known if the contact interaction are modified
  std::string _meshFileName; // To transfert the mesh file from one folder to an other one
  std::string _restartMshFileName;
  bool _resetRestart; //to avoid restart when shitfing schemes
  bool _disableResetRestart; //to allow restat in battery only
  bool _notResetUnknowns;
  bool _noMassPredicorWithoutVelocityAndAcceleration; // allows predicting nomass DOf differently with massDOf if implicicit CH is used

  /* Dynamic relaxation */
  bool _dynamicRelaxation; // performed explicit dynamic relaxtion

  /*Path following option with arc-length control*/
  bool _pathFollowing; // true if considering pathFollowing
  int _macroControlTypePathFollowing;
  int _correctionMethodPathFollowing;
  int _tranversalCriterionPathFollowing;
  int _solverTypePathFollowing;
  double _pathFollowingEqRatio;

  pathFollowingMethod _pathFollowingMethod; // 0- arclength based, 1 -
  double _loadStep; // for load control
  double _localStep; // to local cr control
  double _loadStepPrev; // for load control
  double _localStepPrev; // to local cr control
  double _localStepMinimal;
  double _localStepMaximal;


  double _arcLengthStep; // for arc-length control
  double _arcLengthStepPrev; // previous
  double _arcLengthStepMinimal;
  double _arcLengthStepMaximal;

  double _pathFollowingSwitchCriterion;
  bool _switchedControlType;
  pathFollowingLocalIncrementType _pathFollowingIncrementType;
  pathFollowingLocation _pathFollowingLocation;
  //
  std::vector<int> _hyperellipticControlComp; // comp used in hyperelliptic control


  int _numNROptimal; // num optimal
  int _numNROptimalLocal; // number optimal for local
  double _expNROptimal;
  double _timeStepAdaptationFactor; // for timestep adaptation = pow(_numNROptimal/previous number NR iterations,_expNROptimal)
  bool _timeStepAdaptation; // true if path following step is adapted with number NR


  #if defined(HAVE_MPI)
  std::map<int,std::vector<groupOfElements> >_mapOtherDomain; // map build by the creation of groupOfElement. Retrieve the element of other partition
  #endif // HAVE_MPI
  std::vector<partDomain*> _ghostDomainMPI; // to prescribed Dirichlet BC of other partition in MPI (avoid search in map in each iteration)

  /* MultiSystems solve */
  std::vector<int> _vcompBySys;
  std::vector<scheme> _vschemeBySys;

  bool _stiffEstimation, _stiffnessModification; // true if recalculate stiffness matrix
  bool _iterativeNR; // true if using iterative procedure
  bool _lineSearch;
  // for multiscale analysis
  std::map<int,std::set<int> > _mapRanks;
  bool _isPartitioned; // true if FE mesh is partitioned
  strainMapping* _strainMap;
  bool _strainMappingFollwing;
  // For Onelab display
  mutable int _GmshOneLabViewNum;

  // Failure management parameters
    // Crack insertion control
  int _maxAllowedNewlyBrokenElement;  // Max number of allowed newly cracked elements during one step (for implicit case)
  // Erosion control
  bool _bulkElementErosionFlag; // true if bulk element erosion
  bool _interfaceElementErosionFlag; // true if interface element erosion
  elementErosionType _erosionType; // all ip failed
  FailureCriterionBase* _erosionGlobalCriterion;
  bool _globalCheckFlag; // true if global verification based on _erosionGlobalCriterion
  elementErosionFilter _elementErosionFilter; // a set consists of all erosion elements

	//
  bool _GModelIsRotated; //
  std::map<int, STensor3> _allInterfaceLocalBasis; // ele+ gp, [n b t]
  std::map<TwoNum,int> _interfaceElements; // negative and positive element number versus interface
  std::map<int, TwoNum> _interfaceElementsInverseMap;

  bool _failureBasedOnPreviousState;

  std::list<IntegralVolume> _dataVolumeIntegral;
  std::list<IPDataOnPhysical> _dataOnPhysical;

  bool _endSchemeMonitoring;
  EndSchemeMonitoringBase* _endSchemeMonitoringObject;

  selectiveUpdateBase* _selectiveObject;


#ifndef SWIG
  void commonModel(); // common operation for loadModel and createModel
  void initMapRanks(const std::set<int>& parts);
  void init();
  void init2(); // Has to be regrouped with init() Van Dung ??
  void moveFiles(const std::string &dirName, const std::string &cm);
  void initAllBCsOnDofs();
  void initTerms();
  void endOfScheme(const double endtime, const int endstep); // common operation at the end of a scheme
  void oneStepPreSolve(const double curtime, const double timestep, const int numstep);
	void oneStepPreSolvePathFollowing(const double curtime, const double timestep, const int numstep);

  void fillMapOfInterfaceElementsInOneDomain(MElement *e, std::vector<MElement*> &eleFound,
                                const groupOfElements *g) const;

	void fillLocalReferenceBasisForAllInterfaceElement();


  void setPairSpaceElementForBoundaryConditions(); // new implementation

  // Function used by non linear solver (return the number of iteration to converge)
  int NewtonRaphson(const int numstep);

  /*
  NewtonRaphson function for arc-length path following
  */
  int NewtonRaphsonPathFollowing(const int numstep);

  bool localPathFollowingSwitching() const;
  double computeLocalPathFollowingStep() const;

  double computeRightHandSide();
  void computeLoadVector();
  void computePathFollowingConstraint();
  void computeDPathFollowingConstraintDUnknowns();

  void computeStiffMatrix();
  void computeElasticStiffMatrix();
  void computeMassMatrix();
  void computeExternalForces();
  void computeContactForces();
  void computeInternalForces();
  void fixNodalDofs();
  void numberDofs();
  void insertTheta(const int numphys, groupOfElements *goe);
  void initMapMElementToComputeTimeStep();
  bool considerForTimeStep(MElement* ele) const;
  double initialCriticalExplicitTimeStep();
  double criticalExplicitTimeStep();
  void setInitialCondition();
  void setTimeForBC(double time);
  void setTimeForLaw(const double t,const double dt, const int numstep);
  void initArchiveForce();
  void forceArchiving(const double curtime,const int numstep, const bool forceSave);

  void pathFollowingArchiving(const double curtime, const int numstep);
  void IPDataOnPhysicalArchiving(const double curtime, const int numstep, const bool forceSave);
  void IPVolumeIntegralArchiving(const double curtime, const int numstep, const bool forceSave);

  void initContactInteraction();
  void crackTracking(const double curtime); // 2D only for now !!!
  void postproFragment();

  void applyPBCBetweenTwoGroups();

  void applySameDispBC();
  void applyFixOnFace();

  void applySameDispBCBetweenTwoGroups();

  void fillConstraintDof();
  void applyConstraintBC();

  void initCornerBC();
  void fixDofCorner();

  // For Onelab
  void createOnelabViewOption(const std::string &fname,const int displayType) const;
  friend class restartManager;
  void restart();
  void restartTimeStepAdaptation();
  void restartPathFollowing();
  void restartErosion();
  void restartDefoDefoContact();

  void checkElementErosion(const IPStateBase::whichState ws);

#endif // SWIG
  // Unfortunately should be part of the interface to be able to control the loop from outside
 public:
  void oneStepPostSolve(const double curtime, const int numstep);

 protected:
  solver whatSolver; //  Solver used to solve
  scheme whatScheme; // scheme used to solve equation
  scheme _previousScheme; // to known the previous used scheme for switch
 public:
  nonLinearMechSolver(int tag,bool isParallel=true);
  virtual ~nonLinearMechSolver();
#ifndef SWIG
  int getNumRanks() const;
  int getNumRootRanks() const;
  bool rankOnSolver(const int rank) const;
  void getAllRootRanks(std::set<int>& rootRanks) const;
  void getAllOtherRanks(std::set<int>& otherRanks) const;
  void getRootRank(const int otherRank, int& rootRank) const;
  void getOtherRanks(const int rootRank, std::set<int>& otherRanks) const;

  bool isMultiscale() const;
  bool isDgDomain() const;
  bool withEnergyDissipation() const;
  bool withFractureLaw() const;
  void getLocalBasis(const MElement* ele, const int gpt, SVector3& n, SVector3& t, SVector3& b) const;

  int getDim() const {return _dim;};
  double getEndTime() const{return endtime;}
  scheme getScheme() const{return whatScheme;}
  virtual linearSystem<double>* createSNLSystem();
  virtual linearSystem<double>* createExplicitSystem();
  virtual linearSystem<double>* createImplicitSystem();
  virtual linearSystem<double>* createEigenSystem();
  virtual void createSystem();

  virtual double solveExplicit();
  virtual double solveMulti();
  virtual double solveEigen();
  virtual void createInterfaceElement();
  virtual materialLaw* getMaterialLaw(const int num);
  // create interfaceelement with dgGoupOfElement from dg project doesn't work (segmentation fault)
  virtual void createInterfaceElement_2();
  virtual bool mustRestart(const int numstep);
	pathFollowingLocalIncrementType getPathFollowingLocalIncrementType() const{return _pathFollowingIncrementType;};
	pathFollowingLocation getPathFollowingLocation() const {return _pathFollowingLocation;};

  bool withPathFollowing() const {return _pathFollowing;};
	void rotateModel(const SVector3& n1, const SVector3& n2, const SVector3& n3);
  bool GModelIsRotated() const {return _GModelIsRotated;}
  int getInterfaceElementNumber(const int em, const int ep) const;
  TwoNum getMinusAndPlusElementNumber(const int elnum) const;

	IPField* getIPField();
	const IPField* getIPField() const ;

	unknownField* getUnknownField();
	const unknownField* getUnknownField() const;

	dofManager<double>* getDofManager() ;
	const dofManager<double>* getDofManager() const;

	staticDofManager<double>* getStaticDofManager();
	const staticDofManager<double>* getStaticDofManager() const ;

	std::vector<partDomain*>* getDomainVector();
	const std::vector<partDomain*>* getDomainVector() const;

	std::vector<partDomain*>* getGhostDomainMPI();
	const std::vector<partDomain*>* getGhostDomainMPI() const;

	contactContainer* getAllContactDomain();
	const contactContainer* getAllContactDomain() const;

	defoDefoContactContainer* getAllDefoDefoContactDomain();
	const defoDefoContactContainer* getAllDefoDefoContactDomain() const;

	GModel* getGModel();
	const GModel* getGModel() const;

  std::map<int,materialLaw*> getMaplaw() {return maplaw;};
  const std::map<int,materialLaw*> getMaplaw() const {return maplaw;};

  bool withFailureBasedOnPreviousState() const  {return _failureBasedOnPreviousState;};
  void initArchiveReactionForceOnFixedPhysical(archiveForce& aforce);

#endif
  virtual double solveStaticLinear();
  virtual double solveSNL();
  virtual void perturbationProcedure(const double pre, double* mode);
  // functions for python (swig) interaction
  void loadModel(const std::string &meshFileName);
  void createModel(const std::string &geoFileName, const std::string &outFileName="",const int dim=3,const int order=1,
                   const bool incomplete=false);
  virtual double solve(); // return the time reaches at the end of the computation can be !=endtime if solving problem
  virtual void addDomain(partDomain *dom);
  virtual void addMaterialLaw(materialLaw *mlaw);
  void Solver(const int s){whatSolver= (solver)s;}
  void Scheme(const int s){whatScheme=(scheme)s;}
  void lineSearch(const bool line);

  // path following data
	void pathFollowing(const bool p, const int method);
  void setPathFollowingControlType(const int i);
	void setPathFollowingCorrectionMethod(const int i);
	void setPathFollowingTranversalCriterion(const int i);
	void setPathFollowingSolverType( const int i, const double eqRatio);
  void setBoundsOfPathFollowingArcLengthSteps(const double lowerBound, const double upperBound);

	//increment adaptation in path following based on number NR iterations
	void setPathFollowingIncrementAdaptation(const bool adap, const int numNROptimal, const double exp = 1.);
  void setPathFollowingIncrementAdaptationLocal(const bool adap, const int numNROptimalLoad, const int numNROptimalLocal,const double exp = 1.);
	// increment for local pathFollowing
	void setPathFollowingLocalSteps(const double loadStep, const double localCrStep); // if setPathFollowingIncrementAdaptation is active, localCrStep is not used
	void setPathFollowingLocalIncrementType(const int i);
  void setBoundsOfPathFollowingLocalSteps(const double lowerBound, const double upperBound);
	// for global pathFollowing
	void setPathFollowingArcLengthStep(const double arcStep);
	// set path folowing location
	void setPathFollowingLocation(const int loc);
  // set _pathFollowingSwichCriterion
  void setPathFollowingSwitchCriterion(const double cr);

  // set _hyperellipticControlComp
  void clearAllHyperellipticControlComp();
  void setHyperellipticControlComp(const int comp);

  void stiffnessModification(const bool flag = true);
  void iterativeProcedure(const bool flag = true);
  void snlData(const int ns, const double et, const double reltol, const double absTol=1.e-21);
  void snlManageTimeStep(const int miteNR,const int iteIncreaseTS, const double redfactor, const int maxAttemptRedFactor);
  void setTimeIncrementAdaptation(const bool adap, const int numNROptimal, const double exp, const double maximalStep, const double minTimeStep, const int maxFailedIter);
  void addSystem(const int ncomp,const int s);
  void switchSystem(const int sysindex,const int s); // first sysindex => sysnum =0
  void explicitData(const double ftime, const double gams=0.666667, const double beta=0, const double gamma=0.5, const double alpham=0.,const bool benson=false);
  void implicitData(const double beta, const double gamma, const double alpham , const double alphaf);
  void explicitSpectralRadius(const double ftime,const double gams=0.666667,const double rho=1.,const bool benson=false);
  void options(const std::string &options);
  void implicitSpectralRadius(const double rho);
  void explicitTimeStepEvaluation(const int nst);
  void stepBetweenArchiving(const int n);
  virtual void thetaBC(const int numphys);
  virtual void displacementBC(std::string onwhat, const int numphys, const int comp, const double value);
  virtual void displacementBC(std::string onwhat, const int numphys, const int comp, const double valueDiff, const double valueInit);
  virtual void displacementBC(std::string onwhat, const int numphys, const int comp, const double value,elementFilter *filter);
  virtual void displacementBC(std::string onwhat, const int numphys, const int comp, simpleFunction<double> *fct);
  virtual void displacementBC(std::string onwhat, const int numphys, const int comp, elementFilter *filter, simpleFunction<double> *fct);
  virtual void velocityBC(std::string onwhat, const int numphys, const int comp, const double value);
  virtual void velocityBC(std::string onwhat, const int numphys, const int comp, const double value,elementFilter *filter);
  virtual void velocityBC(std::string onwhat, const int numphys, const int comp, simpleFunction<double> *fct);
  virtual void velocityBC(std::string onwhat, const int numphys, const int comp, elementFilter *filter, simpleFunction<double> *fct);
  virtual void accelerationBC(std::string onwhat, const int numphys, const int comp, const double value);
  virtual void constraintBC(std::string onwhat, const int numphys, const int comp);

	virtual void periodicBC(std::string onwhat, const int phys1, const int phys2, const int v1, const int v2, const int comp);
  virtual void averagePeriodicBC(std::string onwhat, const int phys1, const int phys2, const int v1, const int v2, const int comp);
  virtual void sameDisplacementBC(std::string onwhat, const int phy, const int rootphy,
                                  const int comp, const double fact = 1.);
  virtual void sameDisplacementBCBetweenTwoGroups(std::string onwhat, const int phy1, const int phy2, const int comp);
  virtual void fixOnFace(std::string onwhat, const int phy, const double A, const double B, const double C, const double D);
  virtual void symetryBC(std::string onwhat, const int phy);
  virtual void displacementRigidContactBC(const int numphys, const int comp,const double value);
  virtual void displacementRigidContactBC(const int numphys, const int comp_, simpleFunction<double> *fct);
  virtual void initialBC(std::string onwhat, std::string whichC, const int numphys, const int comp, const double value);
  virtual void initialBC(std::string onwhat, std::string whichC, const int numphys, const int comp, const double value, elementFilter *filter);
  virtual void initialBC(std::string onwhat, std::string whichC, const int numphys, const int comp, simpleFunctionTime<double> *fct,elementFilter* filter=NULL);
  virtual void initialDCBVeloBC(std::string onwhat,const int numphys,const int axiscomp,const int deflcomp,const double length, const double value);
  
  virtual void forceBC(std::string onwhat, const int numphys, const int comp, const double val);
  virtual void forceBC(std::string onwhat, const int numphys, const int comp, simpleFunctionTime<double> *fct);
  
  virtual void scalarFluxBC(std::string onwhat, const int numphys, const int comp, const double val);
  virtual void scalarFluxBC(std::string onwhat, const int numphys, const int comp, simpleFunctionTime<double> *fct);
  
  virtual void independentDisplacementBC(std::string onwhat, const int numphys, const int comp, const double value);
  virtual void independentForceBC(std::string onwhat, const int numphys, const int comp, const double val);
  
  virtual void blastPressureBC(const int numphys,const double p0,const double p1, const double plexp, const double t0, const double t1);
  virtual void pressureOnPhysicalGroupBC(const int numphys, const double pressDiff, const double p0);
  virtual void pressureOnPhysicalGroupBC(const int numphys, simpleFunctionTime<double> *fct);
  
  virtual void blastPressureBC(std::string onwhat,const int numphys,const double p0,const double p1, const double plexp, const double t0, const double t1);
  virtual void pressureOnPhysicalGroupBC(std::string onwhat, const int numphys, const double pressDiff, const double p0);
  virtual void pressureOnPhysicalGroupBC(std::string onwhat, const int numphys, simpleFunctionTime<double> *fct);
  
  

  virtual void archivingForceOnPhysicalGroup(const std::string onwhat, const int numphys, const int comp,const int nstep=1);
  virtual void archivingRigidContactForce(const int numphys, const int comp,const int nstep=1);
  virtual void archivingNodeDisplacement(const int num, const int comp,const int nstep=1);
  virtual void archivingMeshVertexDisplacement(const int num, const int comp,const int nstep = 1);
  virtual void archivingNodeDisplacementOnPhysical(const int dim, const int physical, const int comp,const int nstep=1);
  virtual void archivingNodeDisplacementOnControlNode(const int pos, const int comp,const int nstep=1);

  virtual void archivingVolumeIntegralValue(const int num, const int nstep=1);
  virtual void archivingVolumeIntegralValueOnPhysical(const int phy, const int num, const int nstep=1);
  virtual void archivingAverageValue(const int num, const int nstep=1);
  virtual void archivingAverageValueActiveDissipation(const int num, const int nstep=1);

  virtual void archivingIPDataOnPhysical(const int ipval, const int phys, const int dim, const int nstep=1);

  virtual void archivingNodeVelocity(const int num, const int comp,const int nstep=1);
  virtual void archivingNodeAcceleration(const int num, const int comp,const int nstep=1);
  virtual void archivingNode(const int num, const int comp, nonLinearBoundaryCondition::whichCondition,const int nstep=1);
  virtual void archivingNode(const int numpphys, const int dim, const int comp, nonLinearBoundaryCondition::whichCondition,const int nstep);
  virtual void archivingNodeIP(const int numphys, const int ipval,const int elemval,const int nstep=1);
  virtual void archivingNodeIPOnPhysicalGroup(const int dim, const int numphys, const int ipval, const int nstep=1);
  virtual void archivingElementIP(const int elenum, const int ipval,const int elemval, const int numip=0, const int nstep=1);
  virtual void archivingInterfaceElementIP(const int elenumMinus,const int elenumPlus, const int ipval,const int elemval, const int numip=0, const int nstep=1);
  virtual void archivingIPOnPhysicalGroup(std::string onwhat, const int numphys, const int ipval, const int elemval,
                      const int nstep=1);
  virtual void physInitBroken(const int phys);
//  virtual void setInitOrRestartFileName(const std::string fname);
  virtual void contactInteraction(contactDomain *cdom);
  virtual void defoDefoContactInteraction(defoDefoContactDomain *cdom);
  virtual void archivingRigidContact(const int numphys, const int comp, const int wc=0,const int nstep=1);
  virtual void unknownBuildView(const int comp=0,const int nbstep=1);
  virtual void OneUnknownBuildView(const std::string vname, const int comp,const int nbstep=1);
  virtual void internalPointBuildView(const std::string vname,const int comp,const int nbstep=1,const int ev=-1);
  virtual void energyBuildView(const std::string vname,const int comp,const int nbstep=1);
  virtual void energyComputation(const int val=1);
  virtual void fractureEnergy(const int val=0);
  virtual void crackTracking(std::string fname);
  virtual void postproFragment(std::string fname);
  virtual void resetBoundaryConditions(const bool notResetUnknwown = true); // true if do not modify unknwons false if do
  virtual void resetContactInteraction();
  virtual void disableResetRestart();  //to avoid restart at a change of solver
  virtual void setFactorOnArchivingFiles(const int fact); // only for csv files (for view use nstepBetweenArchiving)
  virtual void dynamicRelaxation(const double gams,const double ftime,const double tol=1.e-6,const int wsolver=0,const bool benson=false);
  virtual void createRestartByTime(const int day,const int hour=0,const int minute=0,const int second=0);
  virtual void createRestartBySteps(const int numstep);

  virtual void createStrainMapping(const std::string  filename);
  virtual void setStrainMappingFollowing(const bool flag);
  // Get the value of a Dof contained in a txt file
  virtual double getArchivedNodalValue(const int numphys,const int comp,nonLinearBoundaryCondition::whichCondition) const;
  virtual double getArchivedForceOnPhysicalGroup(std::string onwhat, const int numphys,const int comp) const;
  virtual double getEnergy(int we) const;
  virtual double getHomogenizedStress(const int i, const int j) const;
  virtual double getHomogenizedTangent(const int i, const int j, const int k, const int l) const;
  virtual double getHomogenizedSecondStress(const int i, const int j, const int k) const;
  virtual double getHomogenizedSecondTangent(const int i, const int j, const int k, const int p, const int q, const int r) const;
  virtual double getHomogenizedFirstSecondTangent(const int i, const int j, const int p, const int q, const int r) const;
  virtual double getHomogenizedSecondFirstTangent(const int i, const int j, const int k, const int p, const int q) const;
  virtual double getHomogenizedCohesiveJump(const int i) const;
  virtual double getNodeIP(const int physical, const int ipVal) const;

  void setSelectiveUpdate(const selectiveUpdateBase& selectiveObj);
  void endSchemeMonitoring(const EndSchemeMonitoringBase& endObj);
  bool solverEndedByMornitoring() const;

  void setNoMassPredicorWithoutVelocityAndAcceleration(const bool fl);

  // Failure control settings
    // for crack insertion
  virtual void setMaxAllowedCrackInsertionAtOneStep(const int maxNum);
    // for element erosion
  virtual void setElementErosion(const bool bulkErosion, const bool interfaceErosion, const int method);
  virtual void setGlobalErosionCheck(const bool flag, const FailureCriterionBase* fcr);

  /** To control via Onelab **/
  // Explicit scheme control
  virtual void initializeExplicitScheme();
  virtual double oneExplicitStep(const double curtime, bool &forceTimeStepEvaluation);
  virtual double finalizeExplicitScheme(const double curtime);
  // Implicit scheme control
  virtual void initializeStaticScheme();
  virtual int oneStaticStep(const double curtime, const double dt, const int numstep);
  virtual double finalizeStaticScheme(const double curtime, const int numstep);

  // Display control
  virtual void createOnelabDisplacementView(const std::string &fname,const double time) const;
  virtual void createOnelabVelocityView(const std::string &fname,const double time) const;
  // Onelab can only access to dof that are archived in txt files
  // They are given in the order defined by the user in his/her python files
  virtual void getOnelabArchiveNodalUnknowns(fullVector<double> &nodalValues) const;

  virtual bool checkVertexBelongToDomains(const int phyVer) const;

// for periodic BC
 public:
  /**
   * 3 method for constraint treatment
   * DISP_ELIM = displacement elimination
   * MULT_ELIM = multiplier elimination
   * DISP_MULT = displacement + multiplier system, no elimination
   */
  enum SYSTEM_TYPE{DISP_ELIM=0, MULT_ELIM = 1, DISP_MULT=2, DISP_ELIM_UNIFIED=3};
   /**
     * homogenized stress method by volume or surface integral
     */
  enum STRESS_HOMO_TYPE{VOLUME=0, SURFACE=1};
  /**
    * homogenized tangent by perturbation of by condensation
    */
  enum TANGENT_HOMO_TYPE{PERTURB=0, CONDEN=1, INSYSTEMCONDEN=2, UNIFIED_CONDEN=3};
  /**
    * if  LOAD_CONTROL , solve until time = 1.
    * if ARC_CONTROL_EULER, solve by explicit Euler and only at microlevel
    */
  enum CONTROL_TYPE{LOAD_CONTROL=0, ARC_CONTROL_EULER=1};

  enum EIGENSOLVER_TYPE{STATIC = 0, DYNAMIC=1};

  /**
  ** microscopic tests
  **/
 protected:
 #ifndef SWIG
  // time and time step
  double _macroTimeStep; // for law which works on increment. (Use only in so no getTimeStep function)
  double _macroTime; // To save results vs time
  int _macroStep; // current macroscopic step

  //for micro flag
  SYSTEM_TYPE _systemType;
  CONTROL_TYPE _controlType;
  // micro flag--> true if microsolver is used
  bool _microFlag, _multiscaleFlag;
  // Element number and Integration point number
  int _enum, _gnum;
  //micro BC
  nonLinearMicroBC* _microBC;
	nonLinearMicroBC* _microBCOld;
	nonLinearMicroBC* _microFailureBC;
  //stress flag and tangent flag
  bool _stressflag, _tangentflag;
  // homogenized density
  double _rho;
  // homogenized stresses
  double _sameStateCriterion; // this parameter to check if same sate-> economise
  homogenizedData* _currentState, *_initialState, *_elasticState;
  // highorder flag : true if using second order multiscale analysis
  bool _highOrderFlag;
  // homogenized  method
  STRESS_HOMO_TYPE _homogenizeStressMethod;
  TANGENT_HOMO_TYPE _homogenizeTangentMethod;

  // for managing periodic boudary condition
  pbcAlgorithm* _pAl;
  // for tangent estimation
  stiffnessCondensation* _condensation;
  // eigensolver using for stability analysis
  bool _eigflag;
  eigenSolver* _eigenSolver;
  EIGENSOLVER_TYPE _eigenSolverType; // static - K only, and dynamic K and M
  bool _MKToFile;
  int _eigenSolverNumIteration;
  std::string _eigenSolverMethod; //
  double _eigenSolverConvergenCriterion;
  bool _hermitian; // if matrix is symmetrical or not
  std::vector<std::vector<nlsField::dataBuildView> > _eigview;
  std::vector<unknownField*> _eigfield;
  std::vector<int> _modeview;
  int _numeigenvalue;
  double _perturbationfactor, _instabilitycriterion;
  bool _eigenFollowing, _perturbate;
  // for archiving
  bool _archive;
  int _iterMax;
  // all homogenized filename
  bool _isHommProSaveToFile; // flag -->save homogenized properties to files
  bool _isHommStrainSaveToFile; //flag --> save homogenized strain to files
  bool _extractPerturbationToFile; //flag --> save perturbation on RVE boundary to file
  homogenizedDataFiles* _homogenizedFiles; // homogenized file data
  bool _messageView; // flag to archive message to file
  FILE* _outputFile; // output file --> if set _messagive
  FILE* _eigenValueFile; // eigenvalus file
  FILE* _criterionFile; // criterion to file
  FILE* _pbcNodeData; // periodic BC based on file
  double _tangentPerturbation; // tangent by perturbation
  double _rveVolume; // rve volume
  bool _isWriteDeformedMeshToFile; // write deformed mesh to file

  bool _stabilityCriterionFollowing;
  bool _isPerturbedEigenMode;
  int _numberPerturbedMode;
  double _valPerturbedMode;
  std::vector<FILE*> _allModeData;
  // for test
  bool _testFlag;
  pbcConstraintElementGroup* _pbcGroup;

  bool _damageIsBlocked;
  bool _solverIsBroken;
  bool _checkFailureOnset; //
  bool _checkWithNormal; //
  bool _damageToCohesiveJump; // true if extracting cohsive law

  STensor3 _FdamOnset; // average damage deformation atinitCornerBC failure onset
  SVector3 _lostSolutionUniquenssNormal; // normal to cohesive element
  double _RVELengthInCohesiveNormal; // length perpendicular to cohesive normal
	double _surfaceReductionRatio; // load carrying surface over nominal surface following cohesive normal

  double _maximalLostEllipticityCriterion; // maximal value
	double _lostSolutionUniquenssTolerance;
	double _voidPartInLocalizationBand; // void part in localization band

	bool _extractIrreversibleEnergy; // flag

	// last extract value
	STensor3 _homogenizedStressLastActiveDissipation;
  STensor3 _homogenizedStrainLastActiveDissipation;
	STensor3 _homogenizedDissipationStrainLastActiveDissipation;
	SVector3 _homogenizedCohesiveJumpLastActiveDissipation;

	double _homogenizedCrackSurface;
  bool _useMatMatSolveToInvertMatrix;

	//
	bool _failureBCIsSwitched; // true if FailureBC is switsched after failure,

  //
  bool _extractElasticTangentOperator; // true if elastic tangent operator is extracted at the same time as full tangent operator
  FILE*  _elasticDPDFTangentFile;

 #endif //SWIG
 protected:
 #ifndef SWIG
  void init2Micro();
  void switchMicroBC(const nonLinearMicroBC* mbc);
  void setTimeForMicroBC(const double time);
  void microNumberDof();

  void fillTangent(const fullMatrix<double>& L, homogenizedData* homoData);

  void computeSplittedStiffnessMatrix();
  stiffnessCondensation* createStiffnessCondensation(const bool sflag, const bool tflag);

  /** get homogzenized stress in case of no tangent computed **/
  void extractAverageStressByCondensation(homogenizedData* homoData);
  void extractAverageStressByVolumeIntegral(homogenizedData* homoData);
  void extractAverageStress(homogenizedData* homoData);


  /**get homogenized stress and tangent if stress and tangent are computed at the same time **/
  void extractAveragePropertiesByCondensation(homogenizedData* homoData);
  void extractAveragePropertiesPerturbation(homogenizedData* homoData);
  void extractAveragePropertiesByInSystemCondensation(homogenizedData* homoData);

  void extractAverageStressAndTangent(homogenizedData* homoData);

  void homogenizedDataToFile(const double time);
  void computeDensity();
  // for eigen value problem
  void createEigenvalueSolver();
  double critialPointTestFunction(double time, double* &mode);
  void modeViewToFile(std::vector<int>& num, double time, int step);
  // newton raphson
  int microNewtonRaphson(const int numstep);
  int microNewtonRaphsonPathFollowing(const int numstep);

  int pathFollowingPerturbation();

  linearSystem<double>* createMicroSNLSystem();
  void microPostSolveStep(const double curtime, const int iter);
  void createMicroSystem();

  double solveMicroSolverSNL();
  double solveMicroSolverStaticLinear();
  double solveMicroSolverForwardEuler();

  void OneStep(const int numstep);

  void writeDisturbedMeshByEigenVector(int step);
  void writeDeformedMesh(int step);
 #endif // SWIG
 public:
 #ifndef SWIG
  void setTime(const double ctime,const double dtime, const int curstep);
  // RVE analysis newton raphson
  double microSolveStaticLinear();
  double microSolveSNL();
  //rve volume
  double getRVEVolume() const;
  bool dissipationIsBlocked() const;
  void blockDissipation(const IPStateBase::whichState ws, const bool fl);
  int getElementSolverNumber() const;
  int getGaussPointSolverNumber() const;
  void extractAverageProperties(bool stiff);
  homogenizedData* getHomogenizationState(const IPStateBase::whichState state);
  const homogenizedData* getHomogenizationState(const IPStateBase::whichState state) const;

	nonLinearMicroBC* getMicroBC() {return _microBC;};
	const nonLinearMicroBC* getMicroBC() const {return _microBC;};

	nonLinearMicroBC* getMicroBCFailure() {return _microFailureBC;};
	const nonLinearMicroBC* getMicroBCFailure() const {return _microFailureBC;};


  void resetSolverToInitialStep();
  void closeAllFile();

	void checkFailureOnset();

	void computeHomogenizedCrackFace();

	double getConstitutiveExtraDofDiffusionEquationRatio(const int index) const;
	double getNonConstitutiveExtraDofEquationRatio(const int index) const;

	bool getExtractIrreversibleEnergyFlag() const;
	bool getDamageToCohesiveJumpFlag() const;
	int getHomogenizationOrder() const;
	bool inMultiscaleSimulation() const;
	const elementErosionFilter& getElementErosionFilter() const;
	bool solverIsBroken() const;
	void brokenSolver(const bool fl);

	const SVector3& getLostSolutionUniquenssNormal() const;

	double& getHomogenizedCrackFace();
	const double& getHomogenizedCrackFace() const;

	STensor3& getFdamOnset();
	const STensor3& getFdamOnset() const;

	FILE * getPBCFile() {return _pbcNodeData;};

 #endif // SWIG
  void createMicroModel(const std::string geoFile, const int dim, const int order, const bool incomplte  = false);
  void setMicroProblemIndentification(int ele, int gpt);
  void initMicroSolver();
  double microSolve();

  void nextStep();
  void archiveData(const double curtime, const int numstep);

  void setMessageView(const bool view);
  void setPeriodicity(const double x, const double y, const double z, const std::string direction);
  void addMicroBC(const nonLinearMicroBC* bc);
	void addMicroBCForFailure(const nonLinearMicroBC* bc);
  // for instability
  void eigenValueSolver(const int num = 10, const bool fl = true); // number of eigenvalue
  void setEigenSolverParamerters(const int type, const int numiter, const std::string method, const double tol, const bool mktofile = false, const bool hem = true);
  void setModeView(const int view); // for view buckling mode
  void setPerturbationFactor(const double val = 1e-2);
  void setInstabilityCriterion(const double val = 1e-6);
  void perturbateBucklingModeByMinimalEigenValueMode(const bool flag = true);
  void setEigenSolverFollowing(const bool flag);
  void setStabilityCriterionFollowing(const bool flag);

  // get macro properties
  void tangentAveragingFlag(const bool fl = true);
  void stressAveragingFlag(const bool fl = true);

  void setStressAveragingMethod(const int method);
  void setTangentAveragingMethod(const int method, const double prec = 1.e-8);

  void setDisplacementAndIPArchiveFlag(const bool flag = true);
  void setHomogenizationPropertyArchiveFlag(const bool flg = true);
  void setStrainArchiveFlag(const bool flg = true);

  double getHomogenizedDensity() const;

  // compute arbitrary quantitiy from microscopic problem defined in IPField::Output
  void getTotalIntegralByVolumeIntegralOnPhysical(const int phys, const int num, double& val) const;
  void getTotalIntegralByVolumeIntegral(const int num, double& val) const;
  void getHomogenizedPropertyByVolumeIntegral(const int num, double& val) const;
  void getHomogenizedPropertyByVolumeIntegralOnActiveDissipationDomain(const int num, double& val, const IPStateBase::whichState ws = IPStateBase::current) const;

  void setExtractPerturbationToFileFlag(const bool flag);
  void setSystemType(const int i);
  void setControlType(const int i);
  void setMicroSolverFlag(const bool flag);
  void setMultiscaleFlag( const bool flag);
  void setSameStateCriterion(const double cr);
  void saveStiffnessMatrixToFile(const int iter);
  void invertFlag(const bool fl);

  void readPBCDataFromFile(const std::string fn);
  void setRVEVolume(const double val);
  void setWriteDeformedMeshToFile(bool flag);
  void setDisturbedEigenMode(int numMode, double val, bool flag = true);
  void addModeData(std::string filename);
   void writeDisturbedMeshByEigenVectorFromFile();


	/** perform a particular test**/
	void activateTest(const bool flag);
	void initialize_test();
  void displacementBCOnControlNode(const int pos, const int comp, const double value);
  void forceBCOnControlNode(const int pos, const int comp, const double value);
  void getDofOnControlNode(const int pos, std::vector<Dof>& R);

	// for lost of solution uniqueness problem
	void setExtractCohesiveLawFromMicroDamage(const bool check);
	void setLostSolutionUniquenssTolerance(const double tol);
	void setExtractIrreversibleEnergyFlag(const bool ex);
	void setLocalizationNormal(const SVector3& normal);
	void setLocalizationNormal(const double n1, const double n2, const double n3);
	void setRVELengthInNormalDirection(const double l, const double surfaceReductionRatio=1.); // surface ratio
	void setVoidPartInLocalizationBand(const double vp);

	void setCheckFailureOnset(const bool fl, const bool withNormal);
	void setSwitchMicroBCFlag(const bool fl);

  void setExtractElasticTangentOperator(const bool fl);
};

#endif //_NONLINEARMECHSOLVER_H_

