//
//
// C++ Interface: dG3D domain with enhanced strain
//
// Description: Interface class for dg 3D
//
// Author:  <V.-D. NGUYEN>, (C) 2017
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include "dG3DEnhancedDomain.h"
#include "multipleFieldDG3DTerms.h"
#include "dG3DEnhancedStrainIPVariable.h"
#include "highOrderTensor.h"
#include "dG3DMultiscaleIPVariable.h"
#include "embeddedLocalizationBand.h"
#include "nonLinearMechSolver.h"
#include "pathFollowingTerms.h"

dG3DEnhancedDomain::dG3DEnhancedDomain(const int tag, const int phys, const int sp_, const int lnum, const int fdg, const int dim) :
dG3DDomain(tag,phys,sp_,lnum,0,dim){
	if (fdg){
		Msg::Warning("dG3DEnhancedDomain is implemented with CG only");
	}
	if(_space!=NULL) delete _space; 
	switch(_wsp)
  {
   case functionSpaceType::Lagrange:
		_space = new g3DLagrangeFunctionSpace(getTag(),12);
    _spaceInterUVW = new IsoparametricLagrangeFunctionSpace(getTag()); // 3 composantes
    break;
   case functionSpaceType::Inter :
    _spaceInterUVW = new IsoparametricLagrangeFunctionSpace(getTag()); // 3 composantes
    break; // InterfaceDomain set function space later
   default:
    Msg::Error("Function space type is unknown on domain %d",getTag());
  }
};

void dG3DEnhancedDomain::createTerms(unknownField *uf,IPField*ip)
{
  FunctionSpace<double>* dgspace = static_cast<FunctionSpace<double>*>(_space);
	ltermBulk = new twoFieldDG3DForceBulk(this,ip,_EqMultiFieldFactor);
	
	if(_bmbp)
    btermBulk = new BilinearTermPerturbation<double>(ltermBulk,*dgspace,*dgspace,uf,ip,this,_eps);
  else{
		btermBulk = new twoFieldDG3DStiffnessBulk(this,ip,_EqMultiFieldFactor);
	}
  
	ltermBound = new nonLinearTermVoid();
	btermBound = new BiNonLinearTermVoid();

  this->btermVirtBound = new BiNonLinearTermVoid();
  this->ltermVirtBound = new nonLinearTermVoid();

  massterm = new mass3D(*dgspace,_mlaw);
  
  if (this->getMacroSolver()->withPathFollowing() and _accountPathFollowing){
    scalarTermPF = new dG3DEnhancedStrainDissipationPathFollowingBulkScalarTerm(this,ip);
		linearTermPF = new dG3DEnhancedStrainDissipationPathFollowingBulkLinearTerm(this,ip);
  }
  else{
    scalarTermPF = new nonLinearScalarTermVoid();;
    linearTermPF = new nonLinearTermVoid();;
  }
  
  scalarTermPFBound = new nonLinearScalarTermVoid();;
  linearTermPFBound = new nonLinearTermVoid();
};

void dG3DEnhancedDomain::computeStrain(MElement *e, const int npts_bulk, IntPt* GP,
                                             AllIPState::ipstateElementContainer *vips,
                                             IPStateBase::whichState ws, fullVector<double> &disp, bool useBarF) const
{
	if (!getElementErosionFilter()(e)) return;
  if(useBarF)
    Msg::Error("dG3DEnhancedDomain::computeStrain bar F method not implemented");
  int nbFF = e->getNumShapeFunctions();
  for(int i = 0; i<npts_bulk; i++){
    IPStateBase* ips = (*vips)[i];
	
    dG3DEnhancedStrainIPVariable *ipv = static_cast<dG3DEnhancedStrainIPVariable*>(ips->getState(ws));
		const dG3DEnhancedStrainIPVariable *ipvprev = static_cast<const dG3DEnhancedStrainIPVariable*>(ips->getState(IPStateBase::previous));
		
		// bulk strain
		std::vector<TensorialTraits<double>::GradType> &Grads = ipv->gradf(_space,e,GP[i]);
		STensor3& Fbulk = ipv->getRefToBulkDeformationGradient();
    STensorOperation::unity(Fbulk);
    
    for (int j = 0; j < nbFF; j++){
			Fbulk(0,0) += Grads[j+0*nbFF][0]*disp(j);
			Fbulk(0,1) += Grads[j+0*nbFF][1]*disp(j);
			Fbulk(0,2) += Grads[j+0*nbFF][2]*disp(j);
			
			Fbulk(1,0) += Grads[j+1*nbFF][0]*disp(j+nbFF);
			Fbulk(1,1) += Grads[j+1*nbFF][1]*disp(j+nbFF);
			Fbulk(1,2) += Grads[j+1*nbFF][2]*disp(j+nbFF);
			
			Fbulk(2,0) += Grads[j+2*nbFF][0]*disp(j+2*nbFF);
			Fbulk(2,1) += Grads[j+2*nbFF][1]*disp(j+2*nbFF);
			Fbulk(2,2) += Grads[j+2*nbFF][2]*disp(j+2*nbFF);
    }
    if(STensorOperation::determinantSTensor3(Fbulk) < 1.e-15) Msg::Error("Negative Jacobian");
		
		ipv->getRefToDeformationGradient() = Fbulk;
		
		// enhanced strain
		std::vector<TensorialTraits<double>::ValType> &Vals = ipv->f(_space,e,GP[i]);
		STensor3& enF = ipv->getRefToEnhancedDeformationGradient();
    STensorOperation::zero(enF);
		
		for (int j=0; j<nbFF; j++){
			for (int k=0; k<3; k++){
				for (int l=0; l<3; l++){
					int row = Tensor23::getIndex(k,l);
					enF(k,l) += disp(j+row*nbFF+3*nbFF)*Vals[j+0*nbFF];
				}
			}
		}
		
		
		// compute betaE
		const double& bandWidth = ipv->getConstRefToBandWidth();
		const SVector3& normal = ipv->getConstRefToLocalizationNormal();
		ipv->getRefToBetaElement() = embeddedFiniteBand(e,GP[i],normal,bandWidth);
		//printf("rank %d bandWidth = %e normal = %e %e %e beta e = %e \n",Msg::GetCommRank(), bandWidth,normal[0],normal[1],normal[2],ipv->getConstRefToBetaElement());
  }
}

