//
// C++ Interface: homogenized tangent terms
//
//
//
// Author:  <V.-D. NGUYEN>, (C) 2017
//
// Copyright: See COPYING file that comes with this distribution
//
//
#include "dG3DHomogenizedTangentTerms.h"
#include "highOrderTensor.h"
#include "nonLinearMechSolver.h"
#include "nonLocalDamageDG3DIPVariable.h"

void dG3DHomogenizedTangent::get(MElement *ele,int npts,IntPt *GP,fullMatrix<double> &m) const{
  const FunctionSpaceBase* sp = _dom->getFunctionSpace();
	
  int nbDof = sp->getNumKeys(ele);
  int nbFF = ele->getNumShapeFunctions();
  int stressDim = _dom->getStressDimension();
  m.resize(stressDim,nbDof);
  m.setAll(0.);
	
  const AllIPState::ipstateElementContainer *vips = _ipf->getAips()->getIPstate(ele->getNum());

  for (int i = 0; i < npts; i++)
  {
    const IPStateBase *ips        = (*vips)[i];
    const dG3DIPVariableBase *ipv     = static_cast<const dG3DIPVariableBase*>(ips->getState(IPStateBase::current));
		
    const std::vector<TensorialTraits<double>::GradType>&  Grads  = ipv->gradf(sp,ele,GP[i]);
    const std::vector<TensorialTraits<double>::ValType>&  Vals  = ipv->f(sp,ele,GP[i]);
		
    const double weight = GP[i].weight;
    double &detJ = ipv->getJacobianDeterminant(ele,GP[i]);
		
    const double ratio = detJ*weight;
    const STensor43& L = ipv->getConstRefToTangentModuli();
		
		
    for (int row=firstRowP; row< lastRowP; row++){
      int p,q;
      Tensor23::getIntsFromIndex(row-firstRowP,p,q);
      for (int j=0; j<nbFF; j++)
        for (int k=0; k<3; k++)
          for (int r= 0; r<3; r++)
            m(row,j+k*nbFF)+= L(p,q,k,r)*Grads[j+k*nbFF](r)*ratio;
    }
    //nonlocal
    if(ipv->getNumberNonLocalVariable()>0)
    {
      const std::vector<STensor3>& dPdphiVec = ipv->getConstRefToDStressDNonLocalVariable();
      for (int row=firstRowP; row< lastRowP; row++){
        int p,q;
        Tensor23::getIntsFromIndex(row-firstRowP,p,q);
        for (int j=0; j<nbFF; j++)
        {
          for (int k=0; k<ipv->getNumberNonLocalVariable(); k++)
          {
            const STensor3& dPdphi = dPdphiVec[k];
            m(row,j+k*nbFF+3*nbFF)+= dPdphi(p,q)*Vals[j+k*nbFF+3*nbFF]*ratio;
          }
        }
      }
    } 
  
    if (_dom->getHomogenizationOrder() == 2)
    {
      SPoint3 pt;
      ele->pnt(Vals,pt);
      for (int row=firstRowQ; row< lastRowQ; row++){
        int ii,jj,kk;
        Tensor33::getIntsFromIndex(row-firstRowQ,ii,jj,kk);
        for (int j=0; j<nbFF; j++)
          for (int k=0; k<3; k++)
            for (int r= 0; r<3; r++)
              m(row,j+k*nbFF)+= 0.5*(L(ii,jj,k,r)*pt[kk]+L(ii,kk,k,r)*pt[jj])*Grads[j+k*nbFF](r)*ratio;
      }
      if(ipv->getNumberNonLocalVariable()>0)
      {
        Msg::Error("dG3DNonLocalHomogenizedTangent::get for second order has not been implemented");
      } 
    }
		
    if (_dom->getExtractIrreversibleEnergyFlag()){
      const STensor3& dIrrEnergdF = ipv->getConstRefToDIrreversibleEnergyDDeformationGradient();
      for (int row=firstRowIrEnergy; row < lastRowIrEnergy; row++){
        for (int j=0; j<nbFF; j++)
          for (int k=0; k<3; k++)
            for (int r= 0; r<3; r++)
              m(row,j+k*nbFF)+= dIrrEnergdF(k,r)*Grads[j+k*nbFF](r)*ratio;
      }
      //nonlocal
      if(ipv->getNumberNonLocalVariable()>0)
      {
       if( ipv->getNumberNonLocalVariable()>_dom->getNumNonLocalVariable() ) Msg::Error("Your material law use more non local variable than your domain dG3DHomogenizedTangent::get");
       for (int row=firstRowIrEnergy; row < lastRowIrEnergy; row++){
          for (int j=0; j<nbFF; j++){
            for (int k=0; k<ipv->getNumberNonLocalVariable(); k++){
              const double& dIrrEnergDphi = ipv->getConstRefToDIrreversibleEnergyDNonLocalVariable(k);
              m(row-firstRowIrEnergy,j+k*nbFF+3*nbFF)+= dIrrEnergDphi*Vals[j+k*nbFF+3*nbFF]*ratio;
            }
          }
        }
      }
    }
    if (_dom->isExtractCohesiveLawFromMicroDamage() and (_ipf->getNumOfActiveDissipationIPs() > 0)) {
      const dG3DIPVariableBase *ipvprev = static_cast<const dG3DIPVariableBase*>(ips->getState(IPStateBase::activeDissipation));
      static STensor43 I4(1.);
      if (ipvprev->dissipationIsActive()){
        for (int row=firstRowFD; row< lastRowFD; row++){
          int p,q;
          Tensor23::getIntsFromIndex(row-firstRowFD,p,q);
          for (int j=0; j<nbFF; j++){
            for (int k=0; k<3; k++){
              for (int r= 0; r<3; r++){
                m(row,j+k*nbFF)+= I4(p,q,k,r)*Grads[j+k*nbFF](r)*ratio;
              }
            }
          }
        }
      }
    }
    if (_dom->getNumConstitutiveExtraDofDiffusionVariable()>0  ){
      if( ipv->getNumConstitutiveExtraDofDiffusionVariable()>_dom->getNumConstitutiveExtraDofDiffusionVariable()) 
             Msg::Error("Your material law use more constitutive extra dof variable than your domain dG3DHomogenizedTangent::get");
      for (int extraDOFField = 0; extraDOFField< ipv->getNumConstitutiveExtraDofDiffusionVariable(); extraDOFField++){
        static bool msg=false;
        if(_dom->getNumConstitutiveExtraDofDiffusionVariable()>1)
        {
          if(!msg)
          {
            Msg::Info("we assume no cross dependency between the extra dof sources domain dG3DHomogenizedTangent::get");
            msg=true;
          }
        }
        int numMDof =nbFF*(3+_dom->getNumNonLocalVariable()+extraDOFField);
        int extraDOFField2=extraDOFField;
        const STensor3  &dPdField        = ipv->getConstRefTodPdField()[extraDOFField];
        const STensor3  &dFluxdGradField = ipv->getConstRefTodFluxdGradField()[extraDOFField][extraDOFField2]; //in reference conf
        const SVector3  &dFluxdField     = ipv->getConstRefTodFluxdField()[extraDOFField][extraDOFField2]; //in reference conf
        const STensor33 &dFluxdF         = ipv->getConstRefTodFluxdF()[extraDOFField]; //in reference conf
        // stress- extra dof
        for (int row=firstRowP; row< lastRowP; row++){
          int p,q;
          Tensor23::getIntsFromIndex(row-firstRowP,p,q);
          for (int j=0; j<nbFF; j++){
            m(row,numMDof+j)+= dPdField(p,q)*Vals[j+numMDof]*ratio;
          }
        }

        // flux
        for (int row=firstRowFlux; row< lastRowFlux; row++){
          // term dFluxdF
          int p;
          Tensor13::getIntsFromIndex(row-firstRowFlux,p);
          for (int j=0; j<nbFF; j++){
            for (int k=0; k<3; k++){
              for (int r= 0; r<3; r++){
                m(row,j+k*nbFF)+= dFluxdF(p,k,r)*Grads[j+k*nbFF](r)*ratio;
              }
            }
          }
          // term dFluxdGradField
          for (int j=0; j<nbFF; j++){
            for (int r= 0; r<3; r++){
              m(row,numMDof+j)+= dFluxdGradField(p,r)*Grads[numMDof+j](r)*ratio;
            }
          }
          // term dFluxdField
          for (int j=0; j<nbFF; j++){
            m(row,numMDof+j)+= dFluxdField(p)*Vals[numMDof+j]*ratio;
          }
        }

        // mechanical source converts to heat
        const STensor3& DmecaSourceDF = ipv->getConstRefTodMechanicalSourcedF()[extraDOFField];
        const SVector3& DmecaSourceDGradField = ipv->getConstRefTodMechanicalSourcedGradField()[extraDOFField][extraDOFField2];
        const double& DmecaSourceDField = ipv->getConstRefTodMechanicalSourcedField()(extraDOFField,extraDOFField2);

        // term DmecaSourceDF
        for (int row = firstRowMecaSource; row < lastRowMecaSource; row++){
          for (int j=0; j<nbFF; j++){
            for (int k=0; k<3; k++){
              for (int r= 0; r<3; r++){
                m(row,j+k*nbFF)+= DmecaSourceDF(k,r)*Grads[j+k*nbFF](r)*ratio;
              }
            }
          }
          // term DmecaSourceDGradField
          for (int j=0; j<nbFF; j++){
            for (int r= 0; r<3; r++){
              m(row,numMDof+j)+= DmecaSourceDGradField(r)*Grads[numMDof+j](r)*ratio;
            }
          }
          // term DmecaSourceDField
	  for (int j=0; j<nbFF; j++){
            m(row,numMDof+j)+= DmecaSourceDField*Vals[numMDof+j]*ratio;
          }
        }
      }
    }	
  }
};

void dG3DHomogenizedElasticTangent::get(MElement *ele,int npts,IntPt *GP,fullMatrix<double> &m) const{
  const FunctionSpaceBase* sp = _dom->getFunctionSpace();
	
  int nbDof = sp->getNumKeys(ele);
  int nbFF = ele->getNumShapeFunctions();
  int stressDim = _dom->getStressDimension();
  m.resize(stressDim,nbDof);
  m.setAll(0.);
	
  const AllIPState::ipstateElementContainer *vips = _ipf->getAips()->getIPstate(ele->getNum());

  for (int i = 0; i < npts; i++)
  {
    const IPStateBase *ips        = (*vips)[i];
    const dG3DIPVariableBase *ipv     = static_cast<const dG3DIPVariableBase*>(ips->getState(IPStateBase::current));
		
    const std::vector<TensorialTraits<double>::GradType>&  Grads  = ipv->gradf(sp,ele,GP[i]);
    const std::vector<TensorialTraits<double>::ValType>&  Vals  = ipv->f(sp,ele,GP[i]);
		
    const double weight = GP[i].weight;
    double &detJ = ipv->getJacobianDeterminant(ele,GP[i]);
		
    const double ratio = detJ*weight;
    const STensor43& L = ipv->getConstRefToElasticTangentModuli();
		
  
    for (int row=firstRowP; row< lastRowP; row++){
      int p,q;
      Tensor23::getIntsFromIndex(row-firstRowP,p,q);
      for (int j=0; j<nbFF; j++)
        for (int k=0; k<3; k++)
          for (int r= 0; r<3; r++)
            m(row,j+k*nbFF)+= L(p,q,k,r)*Grads[j+k*nbFF](r)*ratio;
    }
    //nonlocal
    /*if(ipv->getNumberNonLocalVariable()>0)
    {
      const std::vector<STensor3>& dPdphiVec = ipv->getConstRefToDStressDNonLocalVariable();
      for (int row=firstRowP; row< lastRowP; row++){
        int p,q;
        Tensor23::getIntsFromIndex(row-firstRowP,p,q);
        for (int j=0; j<nbFF; j++)
        {
          for (int k=0; k<ipv->getNumberNonLocalVariable(); k++)
          {
            const STensor3& dPdphi = dPdphiVec[k];
            m(row,j+k*nbFF+3*nbFF)+= dPdphi(p,q)*Vals[j+k*nbFF+3*nbFF]*ratio;
          }
        }
      }
    } 
      */
    
    if (_dom->getHomogenizationOrder() == 2)
    {
      SPoint3 pt;
      ele->pnt(Vals,pt);
      for (int row=firstRowQ; row< lastRowQ; row++){
        int ii,jj,kk;
        Tensor33::getIntsFromIndex(row-firstRowQ,ii,jj,kk);
        for (int j=0; j<nbFF; j++)
          for (int k=0; k<3; k++)
            for (int r= 0; r<3; r++)
              m(row,j+k*nbFF)+= 0.5*(L(ii,jj,k,r)*pt[kk]+L(ii,kk,k,r)*pt[jj])*Grads[j+k*nbFF](r)*ratio;
      }
      if(ipv->getNumberNonLocalVariable()>0)
      {
        Msg::Error("dG3DNonLocalHomogenizedTangent::get for second order has not been implemented");
      } 
    }
		
    if (_dom->getExtractIrreversibleEnergyFlag()){
      const STensor3& dIrrEnergdF = ipv->getConstRefToDIrreversibleEnergyDDeformationGradient();
      for (int row=firstRowIrEnergy; row < lastRowIrEnergy; row++){
        for (int j=0; j<nbFF; j++)
          for (int k=0; k<3; k++)
            for (int r= 0; r<3; r++)
              m(row,j+k*nbFF)+= dIrrEnergdF(k,r)*Grads[j+k*nbFF](r)*ratio;
      }
      //nonlocal
      if(ipv->getNumberNonLocalVariable()>0)
      {
       if( ipv->getNumberNonLocalVariable()>_dom->getNumNonLocalVariable() ) Msg::Error("Your material law use more non local variable than your domain dG3DHomogenizedTangent::get");
       for (int row=firstRowIrEnergy; row < lastRowIrEnergy; row++){
          for (int j=0; j<nbFF; j++){
            for (int k=0; k<ipv->getNumberNonLocalVariable(); k++){
              const double& dIrrEnergDphi = ipv->getConstRefToDIrreversibleEnergyDNonLocalVariable(k);
              m(row-firstRowIrEnergy,j+k*nbFF+3*nbFF)+= dIrrEnergDphi*Vals[j+k*nbFF+3*nbFF]*ratio;
            }
          }
        }
      }
    }
    if (_dom->isExtractCohesiveLawFromMicroDamage() and (_ipf->getNumOfActiveDissipationIPs() > 0)) {
      const dG3DIPVariableBase *ipvprev = static_cast<const dG3DIPVariableBase*>(ips->getState(IPStateBase::activeDissipation));
      static STensor43 I4(1.);
      if (ipvprev->dissipationIsActive()){
        for (int row=firstRowFD; row< lastRowFD; row++){
          int p,q;
          Tensor23::getIntsFromIndex(row-firstRowFD,p,q);
          for (int j=0; j<nbFF; j++){
            for (int k=0; k<3; k++){
              for (int r= 0; r<3; r++){
                m(row,j+k*nbFF)+= I4(p,q,k,r)*Grads[j+k*nbFF](r)*ratio;
              }
            }
          }
        }
      }
    }
    if (_dom->getNumConstitutiveExtraDofDiffusionVariable()>0  ){
      if( ipv->getNumConstitutiveExtraDofDiffusionVariable()>_dom->getNumConstitutiveExtraDofDiffusionVariable()) 
             Msg::Error("Your material law use more constitutive extra dof variable than your domain dG3DHomogenizedTangent::get");
      for (int extraDOFField = 0; extraDOFField< ipv->getNumConstitutiveExtraDofDiffusionVariable(); extraDOFField++){
        static bool msg=false;
        if(_dom->getNumConstitutiveExtraDofDiffusionVariable()>1)
        {
          if(!msg)
          {
            Msg::Info("we assume no cross dependency between the extra dof sources domain dG3DHomogenizedTangent::get");
            msg=true;
          }
        }
        int numMDof =nbFF*(3+_dom->getNumNonLocalVariable()+extraDOFField);
        int extraDOFField2=extraDOFField;
        const STensor3  &dPdField        = ipv->getConstRefTodPdField()[extraDOFField];
        const STensor3  &dFluxdGradField = ipv->getConstRefTodFluxdGradField()[extraDOFField][extraDOFField2]; //in reference conf
        const SVector3  &dFluxdField     = ipv->getConstRefTodFluxdField()[extraDOFField][extraDOFField2]; //in reference conf
        const STensor33 &dFluxdF         = ipv->getConstRefTodFluxdF()[extraDOFField]; //in reference conf
        // stress- extra dof
        for (int row=firstRowP; row< lastRowP; row++){
          int p,q;
          Tensor23::getIntsFromIndex(row-firstRowP,p,q);
          for (int j=0; j<nbFF; j++){
            m(row,numMDof+j)+= dPdField(p,q)*Vals[j+numMDof]*ratio;
          }
        }

        // flux
        for (int row=firstRowFlux; row< lastRowFlux; row++){
          // term dFluxdF
          int p;
          Tensor13::getIntsFromIndex(row-firstRowFlux,p);
          for (int j=0; j<nbFF; j++){
            for (int k=0; k<3; k++){
              for (int r= 0; r<3; r++){
                m(row,j+k*nbFF)+= dFluxdF(p,k,r)*Grads[j+k*nbFF](r)*ratio;
              }
            }
          }
          // term dFluxdGradField
          for (int j=0; j<nbFF; j++){
            for (int r= 0; r<3; r++){
              m(row,numMDof+j)+= dFluxdGradField(p,r)*Grads[numMDof+j](r)*ratio;
            }
          }
          // term dFluxdField
          for (int j=0; j<nbFF; j++){
            m(row,numMDof+j)+= dFluxdField(p)*Vals[numMDof+j]*ratio;
          }
        }

        // mechanical source converts to heat
        const STensor3& DmecaSourceDF = ipv->getConstRefTodMechanicalSourcedF()[extraDOFField];
        const SVector3& DmecaSourceDGradField = ipv->getConstRefTodMechanicalSourcedGradField()[extraDOFField][extraDOFField2];
        const double& DmecaSourceDField = ipv->getConstRefTodMechanicalSourcedField()(extraDOFField,extraDOFField2);

        // term DmecaSourceDF
        for (int row = firstRowMecaSource; row < lastRowMecaSource; row++){
          for (int j=0; j<nbFF; j++){
            for (int k=0; k<3; k++){
              for (int r= 0; r<3; r++){
                m(row,j+k*nbFF)+= DmecaSourceDF(k,r)*Grads[j+k*nbFF](r)*ratio;
              }
            }
          }
          // term DmecaSourceDGradField
          for (int j=0; j<nbFF; j++){
            for (int r= 0; r<3; r++){
              m(row,numMDof+j)+= DmecaSourceDGradField(r)*Grads[numMDof+j](r)*ratio;
            }
          }
          // term DmecaSourceDField
	  for (int j=0; j<nbFF; j++){
            m(row,numMDof+j)+= DmecaSourceDField*Vals[numMDof+j]*ratio;
          }
        }
      }
    }	
  }
};
 
 
/*
void dG3DExtraDofHomogenizedTangent::get(MElement *ele,int npts,IntPt *GP,fullMatrix<double> &m) const{
	dG3DHomogenizedTangent::get(ele,npts,GP,m);
	
	const FunctionSpaceBase* space = _dom->getFunctionSpace();
	
  int nbDof = space->getNumKeys(ele);
  int nbFF = ele->getNumShapeFunctions();

  int numMDof =nbFF*3;
	
	const AllIPState::ipstateElementContainer *vips = _ipf->getAips()->getIPstate(ele->getNum());
	
  for (int i = 0; i < npts; i++){
		const IPStateBase *ips        = (*vips)[i];
		const dG3DIPVariable *ipv     = static_cast<const dG3DIPVariable*>(ips->getState(IPStateBase::current));
		const dG3DIPVariable *ipvprev = static_cast<const dG3DIPVariable*>(ips->getState(IPStateBase::activeDissipation));		
		
    const double weight = GP[i].weight;
    const std::vector<TensorialTraits<double>::ValType> & Vals = ipv->f(space,ele,GP[i]);
    const std::vector<TensorialTraits<double>::GradType>&  Grads  = ipv->gradf(space,ele,GP[i]);
    double &detJ = ipv->getJacobianDeterminant(ele,GP[i]);
    const double ratio = detJ*weight;

    int extraDOFField=0;
    int extraDOFField2=0;
    const STensor3  &dPdField        = ipv->getConstRefTodPdField()[extraDOFField];
    const STensor3  &dFluxdGradField = ipv->getConstRefTodFluxdGradField()[extraDOFField][extraDOFField2]; //in reference conf
    const SVector3  &dFluxdField     = ipv->getConstRefTodFluxdField()[extraDOFField][extraDOFField2]; //in reference conf
    const STensor33 &dFluxdF         = ipv->getConstRefTodFluxdF()[extraDOFField]; //in reference conf


    // stress- extra dof
    for (int row=firstRowP; row< lastRowP; row++){
      int p,q;
      Tensor23::getIntsFromIndex(row-firstRowP,p,q);
      for (int j=0; j<nbFF; j++){
        m(row,numMDof+j)+= dPdField(p,q)*Vals[j+numMDof]*ratio;
      }
    }

    // flux
    for (int row=firstRowFlux; row< lastRowFlux; row++){
      // term dFluxdF
			int p;
			Tensor13::getIntsFromIndex(row-firstRowFlux,p);
      for (int j=0; j<nbFF; j++){
        for (int k=0; k<3; k++){
          for (int r= 0; r<3; r++){
            m(row,j+k*nbFF)+= dFluxdF(p,k,r)*Grads[j+k*nbFF](r)*ratio;
          }
        }
      }
      // term dFluxdGradField
      for (int j=0; j<nbFF; j++){
        for (int r= 0; r<3; r++){
          m(row,numMDof+j)+= dFluxdGradField(p,r)*Grads[numMDof+j](r)*ratio;
        }
      }
      // term dFluxdField
      for (int j=0; j<nbFF; j++){
          m(row,numMDof+j)+= dFluxdField(p)*Vals[numMDof+j]*ratio;
      }
    }

    // mechanical source converts to heat
    const STensor3& DmecaSourceDF = ipv->getConstRefTodMechanicalSourcedF()[extraDOFField];
    const SVector3& DmecaSourceDGradField = ipv->getConstRefTodMechanicalSourcedGradField()[extraDOFField][extraDOFField2];
    const double& DmecaSourceDField = ipv->getConstRefTodMechanicalSourcedField()(extraDOFField,extraDOFField2);

    // term DmecaSourceDF
		for (int row = firstRowMecaSource; row < lastRowMecaSource; row++){
			for (int j=0; j<nbFF; j++){
				for (int k=0; k<3; k++){
					for (int r= 0; r<3; r++){
						m(row,j+k*nbFF)+= DmecaSourceDF(k,r)*Grads[j+k*nbFF](r)*ratio;
					}
				}
			}
			// term DmecaSourceDGradField
			for (int j=0; j<nbFF; j++){
				for (int r= 0; r<3; r++){
					m(row,numMDof+j)+= DmecaSourceDGradField(r)*Grads[numMDof+j](r)*ratio;
				}
			}
			// term DmecaSourceDField
			for (int j=0; j<nbFF; j++){
					m(row,numMDof+j)+= DmecaSourceDField*Vals[numMDof+j]*ratio;
			}
		
		}

  };
};*/

