
//
//
// Description: Non linear solver for mechanic problems
//              quasi-static implicit scheme & dynamic explicit Hulbert-Chung scheme
//
// Author:  <Gauthier BECKER>, (C) 2011
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include <string.h>
#include <cstdlib>
#include <ctime>
#include <time.h>
#include <cstdio>
#include "GmshConfig.h"
#include "meshPartition.h"
#include "nonLinearMechSolver.h"
#include "linearSystemCSR.h"
#include "linearSystemPETSc.h"
#include "linearSystemGmm.h"
#include "Numeric.h"
#include "nlTerms.h"
#include "solverAlgorithms.h"
#include "quadratureRules.h"
#include "MPoint.h"
#include "ipstate.h"
#include "ipField.h"
#include "timeFunction.h"
#include "GModel.h"
#include "explicitHulbertChungPETSc.h"
#include "explicitHulbertChungBLAS.h"
#include "nlsolAlgorithms.h"
#include "nonLinearMechSolver.h"
#include "MTriangle.h"
#include "MQuadrangle.h"
#include "energyField.h"
#include "nlsolAlgorithms.h"
#include "solverAlgorithms.h"
#include "quadratureRules.h"
#include "MPoint.h"
#include "timeFunction.h"
#include "pbcDofManager.h"
#include "staticDofManager.h"
#include "dofManagerMultiSystems.h"
#include "contactTerms.h"
#include "NonLinearSolverConfig.h"
#include "currentConfig.h"
#include "MInterfaceLine.h"
#include "fragment.h"
#include "OS.h"
#include "pathFollowingSystem.h"
#include "StringUtils.h"
#include "Context.h"
#include "HighOrder.h"
#include "implicitHulbertChungPETSc.h"
#include "pathFollowingSystemPETSc.h"
#include "pbcPathFollowingSystemPETSc.h"
#include "generalLocalControlBasedPathFollowingSystemPETSc.h"
#include "eigenSolver.h"
#include "eigenModeView.h"
#include "eigenVectorData.h"
#include "failureDetection.h"
//#include "linearSystemGMRESk.h"
#include "restartManager.h"
#include "geometryRotation.h"
#include "interfaceReferenceLocalBasis.h"
#include "numericalMaterial.h"
#include "criticalTimeStepElements.h"
#include "generalHyperellipticPathFollowingSystemPETSc.h"

#if defined(HAVE_POST)
#include "PView.h"
#include "PViewData.h"
#endif

#if defined(HAVE_MPI)
#include "elementFilterMPI.h"
#include "nlmpiAlgorithms.h"
#endif // HAVE_MPI

 nonLinearMechSolver::nonLinearMechSolver(int tag, bool isParallel) : _tag(tag), _dim(0), pAssembler(NULL), _numstepExpl(1), _numstepImpl(1),
                                                                      endtime(1.), _tol(1.e-6), _absTol(1.e-12), whatSolver(nonLinearMechSolver::Gmm),
                                                                      whatScheme(nonLinearMechSolver::StaticLinear),
                                                                      _beta(0.25), _gamma(0.5), _alpham(0.), _alphaf(0.),_rhoinfty(1.), _gammas(0.6666),
                                                                      _mpiUserDom(0), _maxNRite(15),_pathFollowing(false),_pathFollowingMethod(GLOBAL_ARC_LENGTH_BASED),
																																			_loadStep(0.),_localStep(0.), _loadStepPrev(0.),_localStepPrev(0.),
                                                                      _localStepMinimal(0.),_localStepMaximal(1e100), nsba(0), 
                                                                      _arcLengthStepMinimal(0.),_arcLengthStepMaximal(1e100),
																																			_numNROptimal(5),_expNROptimal(1.),_numNROptimalLocal(5), _timeStepAdaptation(false),_switchedControlType(false),
																																			_pathFollowingIncrementType(DISSIPATION_ENERGY), _arcLengthStep(0.),_arcLengthStepPrev(0.),
																																			_timeStepAdaptationFactor(1.), _endSchemeMonitoring(false),_endSchemeMonitoringObject(NULL),
                                                                      _timeStepFactorReduction(2.),_maxAttemptStepReduction(6), _niteTimeStepIncrease(5), 
                                                                      _maximalTimeStep(1e10), _minimalTimeStep(0),
                                                                      _solver_options(""), _notResetUnknowns(true), 
                                                                      _crackTrackingFile(NULL), _mpiDofManager(false), _previousInit(false),
                                                                      _notResetedBC(false),_notResetedContact(false), _ipf(NULL), _meshFileName(""), _restartMshFileName("restart_"),
                                                                      _dynamicRelaxation(false),_previousScheme(StaticNonLinear),_criterionFile(NULL),
                                                                      _microFlag(false),_ufield(NULL),_pAl(NULL), _multiscaleFlag(false), _pbcNodeData(NULL),
                                                                       _extractPerturbationToFile(false), _sameStateCriterion(1e-10),
                                                                       _currentState(NULL),_initialState(NULL), _elasticState(NULL),
																																			_homogenizedFiles(NULL), _elasticDPDFTangentFile(NULL),
                                                                      _condensation(NULL),_eigenSolver(NULL),_eigflag(false),_energField(NULL),
                                                                      _stressflag(true),_tangentflag(false),_enum(0),_gnum(0),_rveVolume(0.),
                                                                      _systemType(nonLinearMechSolver::MULT_ELIM),_controlType(nonLinearMechSolver::LOAD_CONTROL),
                                                                      _highOrderFlag(false), _eigenValueFile(NULL), _perturbationfactor(0.1),
                                                                      _archive(true),_stiffEstimation(true),_stiffnessModification(true),
                                                                      _iterativeNR(true),_messageView(false), _outputFile(NULL), _eigenFollowing(false),
                                                                      _homogenizeStressMethod(nonLinearMechSolver::VOLUME), _stabilityCriterionFollowing(false),
                                                                      _homogenizeTangentMethod(nonLinearMechSolver::PERTURB), _isHommStrainSaveToFile(true),
                                                                      _isHommProSaveToFile(true), _tangentPerturbation(1.e-8), _perturbate(false),
                                                                      _fragmentationFile(NULL), _energyComputation(1), _fractureEnergyComputation(0),
                                                                      _timeStepByBenson(false),_lineSearch(false),_strainMap(NULL),_strainMappingFollwing(true),
                                                                      _GmshOneLabViewNum(0), _isWriteDeformedMeshToFile(false),_isPerturbedEigenMode(false),_numberPerturbedMode(1),_valPerturbedMode(1.),
                                                                      _pbcGroup(NULL),_testFlag(false), _eigenSolverType(DYNAMIC),_eigenSolverNumIteration(20),_eigenSolverMethod("krylovschur"), _eigenSolverConvergenCriterion(1e-6),
                                                                      _MKToFile(false),_lostSolutionUniquenssNormal(0.,0.,0.), _surfaceReductionRatio(1.),
                                                                      _damageToCohesiveJump(false), _RVELengthInCohesiveNormal(1.),_FdamOnset(0.),
                                                                      _useMatMatSolveToInvertMatrix(false), _maximalLostEllipticityCriterion(-1.),
																																			_microBC(NULL), _damageIsBlocked(false), _solverIsBroken(false),
																																			_bulkElementErosionFlag(false), _interfaceElementErosionFlag(false),_erosionType(ALL_IP_FAILED),
																																			_globalCheckFlag(false),_erosionGlobalCriterion(NULL),
																																			_macroControlTypePathFollowing(-1),_correctionMethodPathFollowing(-1),
																																			_tranversalCriterionPathFollowing(-1),_solverTypePathFollowing(-1),
																																			_pathFollowingEqRatio(1.),_extractIrreversibleEnergy(false),
																																			_pathFollowingLocation(BULK_INTERFACE),
																																			_homogenizedStressLastActiveDissipation(0.),_homogenizedStrainLastActiveDissipation(0.),
                                                                      _homogenizedDissipationStrainLastActiveDissipation(0.),
																																			_homogenizedCohesiveJumpLastActiveDissipation(0.),
																																			_lostSolutionUniquenssTolerance(0.), _voidPartInLocalizationBand(0.),
																																			_GModelIsRotated(false), _checkFailureOnset(false), _checkWithNormal(true), _homogenizedCrackSurface(0.),
																																			_microBCOld(NULL),_microFailureBC(NULL), _failureBCIsSwitched(false),
                                                                      _pathFollowingSwitchCriterion(0.), _failureBasedOnPreviousState(true), _selectiveObject(NULL),
                                                                      _macroTime(0.),_macroTimeStep(1.),_macroStep(0.),_hermitian(true),
                                                                      _extractElasticTangentOperator(false),
                                                                      _resetRestart(false),_disableResetRestart(false),
                                                                      _noMassPredicorWithoutVelocityAndAcceleration(false)
{
  // check parallelization of dofManager
 #if defined(HAVE_MPI)
 if(isParallel and Msg::GetCommSize() > 1)
   _mpiDofManager = true;
 #endif // HAVE_MPI

  _isPartitioned = false;
	#if defined(HAVE_MPI)
	_workingRank = Msg::GetCommRank();
	#else
	_workingRank = 0;
	#endif //HAVE_MPI

  // default view (displacement)
  unknownView.emplace_back("displacement",0,nlsField::crude,1);

  // initialize c++ function rand (already made in gmsh ?? put elsewhere ??)
  _beginTime = time(NULL);
  srand(_beginTime);

  restartManager::Init(2000000000);
  std::ostringstream oss;
  oss << Msg::GetCommRank();
  std::string srank = oss.str();
  _restartMshFileName += oss.str() + ".msh";



  _hyperellipticControlComp.resize(3);
  _hyperellipticControlComp[0] =0;
  _hyperellipticControlComp[1] =1;
  _hyperellipticControlComp[2] =2;
}

nonLinearMechSolver::~nonLinearMechSolver(){
  if (pModel) delete pModel;
  if (pAssembler) delete pAssembler;
  if (_crackTrackingFile != NULL) fclose(_crackTrackingFile);
  if (_fragmentationFile != NULL) fclose(_fragmentationFile);
  this->resetBoundaryConditions();
  this->resetContactInteraction();

  if (_strainMap) delete _strainMap;

  if (_pAl) delete _pAl;
  if (_condensation) delete _condensation;
  if (_eigenSolver) delete _eigenSolver;
  
  if(_ipf != NULL) delete _ipf;
  if (_ufield) delete _ufield;
  if (_energField) delete _energField;

  if (_pbcGroup != NULL) delete _pbcGroup;

  if (_microBC){
    delete _microBC;
    _microBC = NULL;
  }
  if (_microBCOld) delete _microBCOld;
  if (_microFailureBC) delete _microFailureBC;
  if (_erosionGlobalCriterion!=NULL){
    delete _erosionGlobalCriterion;
  }

  if (_endSchemeMonitoringObject != NULL) delete _endSchemeMonitoringObject;
  if (_selectiveObject) delete _selectiveObject;

  if(!_microFlag) // only on the master solver? -->ok, we need to pass once only to finish everything (PETSc, SLEPc, ...)
    Msg::Exit(0);
}

void nonLinearMechSolver::createRestartByTime(const int day,const int hour,const int minute,const int second)
{
  restartManager::setTimeBetweenCheckpoint(day,hour,minute,second);
}

void nonLinearMechSolver::createRestartBySteps(const int numstep)
{
  restartManager::setStepBetweenCheckpoint(numstep);
}

bool nonLinearMechSolver::mustRestart(const int numstep)
{
  return restartManager::checkpoint(numstep);
}

void nonLinearMechSolver::stepBetweenArchiving(const int n)
{  
  if (n > 0)
  {
    nsba = n;
    if (_ipf!=NULL) 
    {
      _ipf->stepBetweenArchiving(n);
    }
    if (_ufield != NULL)
    {
      _ufield->stepBetweenArchiving(n);
    }
    if (_energField != NULL)
    {
      _energField->stepBetweenArchiving(n);
    }
    
    // modify all existing view
    for(int i=0;i<ipView.size();i++)
    {
      ipView[i].nbstepArch = n;
    }
  
    for(int i=0;i<unknownView.size();i++)
    {
      unknownView[i].nbstepArch = n;
    }    
  
    for (int i=0; i< energyView.size(); i++)
    {
      energyView[i].nbstepArch = n;
    }    
  }
}

void nonLinearMechSolver::setFactorOnArchivingFiles(const int fact)
{
  if (fact > 0)
  {
    // chang factor in solver
    for (int i=0; i< vaip.size(); i++)
    {
      vaip[i].nstep*=fact;
    }
    
    for(int i=0;i<anoded.size();i++)
    {
      anoded[i].nstep*=fact;
    }
    // force
    for(int i=0;i<vaf.size();i++)
    {
      vaf[i].nstep*=fact;
    }
    
    _energyComputation *= fact;
    _fractureEnergyComputation *= fact;
    
    // change factor inside field if exist
    if (_ipf!=NULL)
    {
      _ipf->setFactorOnArchivingFiles(fact);
    }
    if (_ufield!=NULL)
    {
      _ufield->setFactorOnArchivingFiles(fact);
    
    }
    if (_energField !=NULL)
    {
      _energField->setFactorOnArchivingFiles(fact);
    }
  }
}

// nonLinearMechSolver
void nonLinearMechSolver::loadModel(const std::string &meshFileName)
{
  // handle restart case
/*  if(restartManager::available())
  {
    pModel = new GModel();
    pModel->setFileName(meshFileName);
    pModel->setAsCurrent();
    pModel->readMSH(_restartMshFileName.c_str());
    _dim = pModel->getNumRegions() ? 3 : 2;
    pModel->setFileName(meshFileName);
    pModel->setName(SplitFileName(meshFileName)[1]);
    return;
  }*/

  pModel = new GModel();
  pModel->setAsCurrent();
  pModel->readMSH(meshFileName.c_str());
  _meshFileName = meshFileName;
  pModel->setFileName(meshFileName);
  pModel->setName(SplitFileName(meshFileName)[1]);
  this->commonModel();
}

void nonLinearMechSolver::commonModel()
{
  _dim = pModel->getNumRegions() ? 3 : 2;
 #if defined(HAVE_MPI)
  int numPartitions = pModel->getNumPartitions();

  if (numPartitions >0) {
		_isPartitioned = true;
	}
  else
	{
		// only first
		_isPartitioned = false;
	}

  Msg::Info("Number of mesh partition %d for %d cpu(s)",numPartitions,Msg::GetCommSize());
  if(numPartitions > Msg::GetCommSize())
  {
    Msg::Error("Your mesh is wrong partitioned");
  }
  else if(numPartitions !=0 and Msg::GetCommSize() == 1)
  {
    Msg::Error("Cannot perform a serial computation with a partitionned mesh");
  }
  else if(numPartitions > 0 and numPartitions < Msg::GetCommSize())
  {
    Msg::Warning("You have more cpu than mesh partition so some cpus may have no work");
  }

  // partitions are numerated from 1 to numPartitions
  std::set<int> mshPart;
  for (int im=1; im < numPartitions+1; im++){
    mshPart.insert(im);
  }
  // int map of ranks
  if (!_microFlag)
    initMapRanks(mshPart);
 #endif // HAVE_MPI
}


void nonLinearMechSolver::createModel(const std::string &geoFileName, const std::string &outFileName,
                                      const int dim,const int order,const bool incomplete)
{
  // check reserved name
  //if(!strcmp(outFileName.c_str(),_restartMshFileName.c_str()))
  //{
  //  Msg::Error("The output file name you request to save your mesh is a reserved name! Change it to continue");
  //}

  // check restart file and if so use the saved msh file
  //if(restartManager::available())
  //{
  //  this->loadModel(_restartMshFileName);
  //  return;
  //}

  // Overwrite options
  CTX::instance()->mesh.order = order;
  CTX::instance()->mesh.secondOrderIncomplete = incomplete;
  // Use version 3 of mesh to avoid a renmbering of the elements
  if (Msg::GetCommSize()>1){
    CTX::instance()->mesh.mshFileVersion = 3.0;
    CTX::instance()->mesh.partitionCreatePhysicals = 0;
  }
  else{
    CTX::instance()->mesh.mshFileVersion = 3.0;
  }

  pModel = new GModel();
  pModel->setAsCurrent();
  pModel->setFileName(geoFileName);
  pModel->setName(SplitFileName(geoFileName)[1]);
  pModel->readGEO(geoFileName);
  pModel->mesh(dim);

  if((!_microFlag) && Msg::GetCommSize()>1) /// if _microFlag is true, mesh is not partitioned
  {
    CTX::instance()->mesh.numPartitions = Msg::GetCommSize();
    PartitionMesh(pModel);
  }

  if(!outFileName.empty() && Msg::GetCommRank() == 0){
    pModel->writeMSH(outFileName,
		     CTX::instance()->mesh.mshFileVersion,CTX::instance()->mesh.binary, CTX::instance()->mesh.saveAll,
                     CTX::instance()->mesh.saveParametric, CTX::instance()->mesh.scalingFactor);
  }

  _meshFileName = outFileName;
  commonModel();
}

void nonLinearMechSolver::createMicroModel(const std::string geoFile, const int dim, const int order,
                        const bool incomplte){
  std::string mshFile = "E_"+int2str(_enum)+"_GP_"+int2str(_gnum)+".msh";
  this->createModel(geoFile,mshFile,dim,order,incomplte);
};

void nonLinearMechSolver::initMapRanks(const std::set<int>& parts){
  _mapRanks.clear();
  std::vector<int> macroProcs;
  int partSize = parts.size();
  for (int i=0; i< partSize; i++){
    macroProcs.push_back(i);
	}
  if (macroProcs.size() == 0){
		// only rank 0 has a real solver
    macroProcs.push_back(0);
		_workingRank = 0;
	}
  int macroSize = macroProcs.size();
  int allSize = Msg::GetCommSize();
  int distSize = allSize - macroSize;
  int div = distSize/macroSize;
  for (int i=0; i<macroSize; i++){
    int root = macroProcs[i];
    std::set<int> others;
    for (int j=0; j<div; j++){
      int num = macroSize+i*div+j;
      others.insert(num);
    }
    _mapRanks[root] = others;
  }
  int j = macroSize+ div*macroSize;
  while (j< allSize){
    for (int i=0; i< macroSize; i++){
      int root = macroProcs[i];
      std::set<int>& others =_mapRanks[root];
      others.insert(j);
      j++;                                             // check here
      if (j>=allSize) break;
    }
  }

  #ifdef _DEBUG
  printf("print map ranks:\n");
  for (std::map<int,std::set<int> >::iterator it = _mapRanks.begin(); it!= _mapRanks.end(); it++){
    if (Msg::GetCommRank() == it->first){
      std::set<int>& others = it->second;
      printf("proc %d -- root rank = %d, other size = %d \n", it->first, it->first,others.size());
      for (std::set<int>::iterator its = others.begin(); its!= others.end(); its++){
        printf("proc %d -- other rank: %d \n",it->first,*its);
      }
    }
  }
  #endif // _DEBUG
};

int nonLinearMechSolver::getNumRanks() const{
  if (_mapRanks.size() > 0){
    int num=0;
    for (std::map<int,std::set<int> >::const_iterator it = _mapRanks.begin(); it != _mapRanks.end(); it++){
      const std::set<int>& others = it->second;
      num += (1+others.size());
    }
    return num;
	}
	else{
    return 1; // in serial case, we have at least 1 proc
	}
};

int nonLinearMechSolver::getNumRootRanks() const{
  if (_mapRanks.size() > 0){
    return _mapRanks.size(); //
	}
	else{
    return 1; // in serial case, we have at least 1 proc
	}
};

bool nonLinearMechSolver::rankOnSolver(const int rank) const{
	for (std::map<int,std::set<int> >::const_iterator it = _mapRanks.begin(); it != _mapRanks.end(); it++){
		if (rank == it->first) return true;
		else{
			const std::set<int>& others = it->second;
			if (others.find(rank) != others.end()){
				return true;
			}

		}
	}
	return false;
};

void nonLinearMechSolver::getAllRootRanks(std::set<int>& rootRanks) const{
	rootRanks.clear();
	for (std::map<int,std::set<int> >::const_iterator it = _mapRanks.begin(); it != _mapRanks.end(); it++){
		rootRanks.insert(it->first);
	}
};

void nonLinearMechSolver::getAllOtherRanks(std::set<int>& otherRanks) const{
  otherRanks.clear();
	for (std::map<int,std::set<int> >::const_iterator it = _mapRanks.begin(); it != _mapRanks.end(); it++){
		otherRanks.insert(it->second.begin(), it->second.end());
	}
};

void nonLinearMechSolver::getRootRank(const int otherRank, int& rootRank) const{
	bool found = false;
	for (std::map<int,std::set<int> >::const_iterator it = _mapRanks.begin(); it != _mapRanks.end(); it++){
		if (otherRank == it->first){
			Msg::Error("_mapRanks is wrongly initiated:  wrong rank %d is root in nonLinearMechSolver::getRootRank",otherRank);
		}

		const std::set<int>& others = it->second;
		if (others.find(otherRank) != others.end()){
			rootRank = it->first;
			found = true;
			break;
		}
	}
	if (!found){
		Msg::Error("_mapRanks is wrongly initiated: root of other rank %d can not be found nonLinearMechSolver::getRootRank",otherRank);
	}

};

void nonLinearMechSolver::getOtherRanks(const int rootRank, std::set<int>& otherRanks) const{
	otherRanks.clear();
	std::map<int,std::set<int> >::const_iterator it = _mapRanks.find(rootRank);
	if (it != _mapRanks.end()){
		otherRanks = it->second;
	}
};

bool nonLinearMechSolver::isMultiscale() const{
  bool isMult = false;
  for (std::map<int,materialLaw*>::const_iterator it = maplaw.begin(); it!= maplaw.end(); it++){
    materialLaw* mlaw = it->second;
    if (mlaw != NULL){
      if (mlaw->isNumeric()){
        isMult = true;
        break;
      }
    }
  }
  return isMult;
};

bool nonLinearMechSolver::isDgDomain() const{
  for (int i=0; i<domainVector.size(); i++){
    partDomain* dom = domainVector[i];
    if (dom->getFormulation()) return true;
  }
  return false;
};

IPField* nonLinearMechSolver::getIPField() {return _ipf;};
const IPField* nonLinearMechSolver::getIPField() const {return _ipf;};

unknownField* nonLinearMechSolver::getUnknownField() {return _ufield;};
const unknownField* nonLinearMechSolver::getUnknownField() const {return _ufield;};

dofManager<double>* nonLinearMechSolver::getDofManager() {return pAssembler;};
const dofManager<double>* nonLinearMechSolver::getDofManager() const {return pAssembler;};

staticDofManager<double>* nonLinearMechSolver::getStaticDofManager() {return dynamic_cast<staticDofManager<double>*>(pAssembler);};
const staticDofManager<double>* nonLinearMechSolver::getStaticDofManager() const {return dynamic_cast<const staticDofManager<double>*>(pAssembler);};

std::vector<partDomain*>* nonLinearMechSolver::getDomainVector() {return &domainVector;}
const std::vector<partDomain*>* nonLinearMechSolver::getDomainVector() const {return &domainVector;}

std::vector<partDomain*>* nonLinearMechSolver::getGhostDomainMPI(){return &_ghostDomainMPI;};
const std::vector<partDomain*>* nonLinearMechSolver::getGhostDomainMPI() const {return &_ghostDomainMPI;};

nonLinearMechSolver::contactContainer* nonLinearMechSolver::getAllContactDomain(){return &_allContact;};
const nonLinearMechSolver::contactContainer* nonLinearMechSolver::getAllContactDomain() const{return &_allContact;};
nonLinearMechSolver::defoDefoContactContainer* nonLinearMechSolver::getAllDefoDefoContactDomain(){return &_allDefoDefoContact;};
const nonLinearMechSolver::defoDefoContactContainer* nonLinearMechSolver::getAllDefoDefoContactDomain() const{return &_allDefoDefoContact;};

GModel* nonLinearMechSolver::getGModel(){return pModel;};
const GModel* nonLinearMechSolver::getGModel() const {return pModel;};

void nonLinearMechSolver::init(){

  if(!_previousInit)
  {
    // Set the material law for each domain and the gauss integration rule
    for(std::vector<partDomain*>::iterator it = domainVector.begin(); it!=domainVector.end(); ++it){
      partDomain *dom = *it;
      dom->setMaterialLaw(maplaw);
    }

		#if defined(HAVE_MPI)
		// remove all unecessary domains in other ranks
		if (this->getNumRanks() > 1){
      bool OneDomRemove = true;
      while (OneDomRemove){
        OneDomRemove = false;
        for(std::vector<partDomain*>::iterator it = domainVector.begin(); it!=domainVector.end(); ++it){
          partDomain *dom = *it;
          dgPartDomain* dgdom =  NULL;
          if (dom->IsInterfaceTerms()){
            dgdom = static_cast<dgPartDomain*>(dom);
          }
          std::set<int> otherRanks;
          int rootRank = dom->getRootRank();
          dom->getOtherRanks(otherRanks);

          if (otherRanks.size()> 0){
            bool isNumMat = false;
            if (dom->getMaterialLaw()){
              if (dom->getMaterialLaw()->isNumeric()) isNumMat = true;
            }
            if (!isNumMat){
              if (dom->IsInterfaceTerms()){
                if (dgdom->getMaterialLawMinus()->isNumeric() or (dgdom->getMaterialLawPlus()->isNumeric())) {
                  isNumMat = true;
                }
              }
            }
            if (!isNumMat){
              // if material is not numerical
              if (otherRanks.find(Msg::GetCommRank()) != otherRanks.end()){
                // in other rank
                if (dom->groupOfElementsSize() > 0)
                  Msg::Error("Remove wrong partDomain %d",dom->getPhysical());
                else if (dom->IsInterfaceTerms()){
                  if (dgdom->gi->size()> 0)
                    Msg::Error("Remove wrong dgPartDomain %d",dom->getPhysical());
                }
                domainVector.erase(it);
                OneDomRemove = true;
                break;
              }
              else if (rootRank == Msg::GetCommRank()){
                // in root rank, clear other ranks
                dom->clearOtherRanks();
              }
            }
          }
        }
      }
		}
		#endif // HAVE_MPI

    // create interfaceElement
    this->createInterfaceElement();

    for(std::vector<partDomain*>::iterator it = domainVector.begin(); it!=domainVector.end(); ++it){
      partDomain *dom = *it;
      dom->setGaussIntegrationRule();
			// set Solver
			dom->setMacroSolver(this); // some options in domaian come from solver
    }

		// add create new material cerate inside solver
		for(std::vector<partDomain*>::iterator it = domainVector.begin(); it!=domainVector.end(); ++it){
			dgPartDomain *dom = dynamic_cast<dgPartDomain*>(*it);
			if (dom != NULL){
				materialLaw* mlawMinus = dom->getMaterialLawMinus();
				if (maplaw.find(mlawMinus->getNum()) == maplaw.end()){
					maplaw[mlawMinus->getNum()] = mlawMinus;
				}

				materialLaw* mlawPlus = dom->getMaterialLawPlus();
				if (maplaw.find(mlawPlus->getNum()) == maplaw.end()){
					maplaw[mlawPlus->getNum()] = mlawPlus;
				}
			}
		};
		for (std::map<int,materialLaw*>::iterator it = maplaw.begin(); it!= maplaw.end(); it++){
			it->second->setMacroSolver(this);
		}

		// at root ranks
		#if defined(HAVE_MPI)
		if (this->getNumRanks() > 1){
      Msg::Barrier(); // wait to finish all previous mesages
		}
		#endif // HAVE_MPI

		if (isDgDomain()){
      _interfaceElements.clear();
      //printf("begin filling reference local basis\n");
      for (int idom= 0; idom < domainVector.size(); idom++){
        const partDomain* dom = domainVector[idom];
        if (dom->IsInterfaceTerms()){
          const dgPartDomain* dgdom = static_cast<const dgPartDomain*>(dom);
          for (groupOfElements::elementContainer::iterator ite = dgdom->gi->begin(); ite != dgdom->gi->end(); ite++){
            MElement* ele = *ite;
            MInterfaceElement* iele = dynamic_cast<MInterfaceElement*>(ele);
            TwoNum tn(iele->getElem(0)->getNum(),iele->getElem(1)->getNum());
            _interfaceElements[tn] = ele->getNum();
            _interfaceElementsInverseMap[ele->getNum()] = tn;
          }
        }
      }

      if (_interfaceElementsInverseMap.size() > 0){
        FILE* interFile = NULL;
        if (Msg::GetCommSize() > 1){
          std::ostringstream oss;
          oss<<Msg::GetCommRank();
          std::string partNum = oss.str();
          std::string filename = "InterfaceElement_part"+partNum+".csv";
          interFile = fopen(filename.c_str(),"w");
        }
        else{
          interFile = fopen("InterfaceElement.csv","w");
        }
        for (std::map<int,TwoNum>::const_iterator itm = _interfaceElementsInverseMap.begin(); itm != _interfaceElementsInverseMap.end(); itm++){
          fprintf(interFile,"%d %d %d\n",itm->first,itm->second.small,itm->second.large);
        }
        fclose(interFile);
      }

      #if defined(HAVE_MPI)
      if (this->getNumRanks() > 1){
        std::set<int> allRootRanks, allOtherRanks;
        this->getAllRootRanks(allRootRanks);
        this->getAllOtherRanks(allOtherRanks);

        if (allRootRanks.find(Msg::GetCommRank()) != allRootRanks.end()){
          std::set<int> otherRanks;
          this->getOtherRanks(Msg::GetCommRank(),otherRanks);
          if (otherRanks.size() > 0){
            // send to other rank
            int interfaceSize = _interfaceElements.size();
            for (std::set<int>::const_iterator itR = otherRanks.begin(); itR != otherRanks.end(); itR++){
              int oRank  = *itR;
              int tag = numericalMaterialBase::createTypeWithTwoInts(Msg::GetCommRank(),1);
              MPI_Send(&interfaceSize,1,MPI_INT,oRank,tag,MPI_COMM_WORLD);
              if (interfaceSize > 0){
                int bufferSize = interfaceSize*3;
                std::vector<int> buffer(bufferSize);
                int index = 0;
                for (std::map<TwoNum,int>::const_iterator itm = _interfaceElements.begin(); itm != _interfaceElements.end(); itm++){
                  buffer[index] = itm->first.small;
                  buffer[index+1] = itm->first.large;
                  buffer[index+2] = itm->second;
                  index += 3;
                }
                tag = numericalMaterialBase::createTypeWithTwoInts(Msg::GetCommRank(),2);
                MPI_Send(&buffer[0],bufferSize,MPI_INT,oRank,tag,MPI_COMM_WORLD);
                buffer.clear();
              }
            }
          }
        }
        else if (allOtherRanks.find(Msg::GetCommRank()) != allOtherRanks.end()){
          // receive data from root ranks
          int rootRank;
          this->getRootRank(Msg::GetCommRank(),rootRank);
          // receive message
          MPI_Status status;
          int tag = numericalMaterialBase::createTypeWithTwoInts(rootRank,1);
          int interfaceSize=0;
          MPI_Recv(&interfaceSize,1,MPI_INT,rootRank,tag,MPI_COMM_WORLD,&status);
          if (interfaceSize > 0){
            int bufferSize = 3*interfaceSize;
            tag = numericalMaterialBase::createTypeWithTwoInts(rootRank,2);
            std::vector<int> buffer(bufferSize);
            MPI_Recv(&buffer[0],bufferSize,MPI_INT,rootRank,tag,MPI_COMM_WORLD,&status);

            int idex = 0;
            for (int i=0; i< interfaceSize; i++){
              TwoNum tn(buffer[idex],buffer[idex+1]);
              _interfaceElements[tn] = buffer[idex+2];
              _interfaceElementsInverseMap[buffer[idex+2]] = tn;
              idex += 3;
            }
          }
        }
      }
      #endif // HAVE_MPI
    }

    // fill reference system axis of interface GPs
		if (isMultiscale()){
      this->fillLocalReferenceBasisForAllInterfaceElement();

      #if defined(HAVE_MPI)
      if (getNumRanks() > 1){
        std::set<int> allRootRanks, allOtherRanks;
        this->getAllRootRanks(allRootRanks);
        this->getAllOtherRanks(allOtherRanks);

        if (allRootRanks.find(Msg::GetCommRank()) != allRootRanks.end()){
          std::set<int> otherRanks;
          this->getOtherRanks(Msg::GetCommRank(),otherRanks);
          if (otherRanks.size() > 0){
            int sizeOfInterfaceLocalBasis = _allInterfaceLocalBasis.size();
            for (std::set<int>::const_iterator itR = otherRanks.begin(); itR != otherRanks.end(); itR++){
              int oRank  = *itR;
              int tag = numericalMaterialBase::createTypeWithTwoInts(Msg::GetCommRank(),3);
              MPI_Send(&sizeOfInterfaceLocalBasis,1,MPI_INT,oRank,tag,MPI_COMM_WORLD);
              if (sizeOfInterfaceLocalBasis > 0){
                int sizeBufferEleNum = sizeOfInterfaceLocalBasis;
                int sizeBufferLocalBasis = 9*sizeBufferEleNum;

                std::vector<int> bufferEleNum(sizeBufferEleNum);
                std::vector<double> bufferLocalBasis(sizeBufferLocalBasis);
                int rowEleNum =0;
                for (std::map<int,STensor3>::const_iterator it = _allInterfaceLocalBasis.begin(); it != _allInterfaceLocalBasis.end(); it++){
                  bufferEleNum[rowEleNum] = it->first;
                  const STensor3& mat = it->second;
                  for (int index = 0; index< 9; index++){
                    int i,j;
                    Tensor23::getIntsFromIndex(index,i,j);
                    bufferLocalBasis[9*rowEleNum + index] = mat(i,j);
                  }
                  rowEleNum++;
                }
                tag = numericalMaterialBase::createTypeWithTwoInts(Msg::GetCommRank(),4);
                MPI_Send(&bufferEleNum[0],sizeBufferEleNum,MPI_INT,oRank,tag,MPI_COMM_WORLD);
                tag = numericalMaterialBase::createTypeWithTwoInts(Msg::GetCommRank(),5);
                MPI_Send(&bufferLocalBasis[0],sizeBufferLocalBasis,MPI_DOUBLE,oRank,tag,MPI_COMM_WORLD);
              }
            };
          }
        }
        else if (allOtherRanks.find(Msg::GetCommRank()) != allOtherRanks.end()){
          // receive data from root ranks
          int rootRank;
          this->getRootRank(Msg::GetCommRank(),rootRank);
          int bufferSize=0;
          MPI_Status status;
          int tag = numericalMaterialBase::createTypeWithTwoInts(rootRank,3);
          MPI_Recv(&bufferSize,1,MPI_INT,rootRank,tag,MPI_COMM_WORLD,&status);
          if (bufferSize > 0){
            int sizeBufferEleNum = bufferSize;
            int sizeBufferLocalBasis = 9*bufferSize;
            std::vector<int> bufferEleNum(sizeBufferEleNum);
            std::vector<double> bufferLocalBasis(sizeBufferLocalBasis);

            tag = numericalMaterialBase::createTypeWithTwoInts(rootRank,4);
            MPI_Recv(&bufferEleNum[0],sizeBufferEleNum,MPI_INT,rootRank,tag,MPI_COMM_WORLD,&status);
            tag = numericalMaterialBase::createTypeWithTwoInts(rootRank,5);
            MPI_Recv(&bufferLocalBasis[0],sizeBufferLocalBasis,MPI_DOUBLE,rootRank,tag,MPI_COMM_WORLD,&status);
            // set data
            for (int id=0; id< bufferSize; id++){
              int type = bufferEleNum[id];
              STensor3 mat(0.);
              for (int index = 0; index< 9; index++){
                int i,j;
                Tensor23::getIntsFromIndex(index,i,j);
                mat(i,j) = bufferLocalBasis[9*id+index];
              }
              _allInterfaceLocalBasis[type] = mat;
            }
          }
        }
      }
      #endif // HAVE_MPI
		}

		#if defined(HAVE_MPI)
    if (this->getNumRanks() > 1){
      Msg::Barrier(); // wait to finish all previous mesages
		}
    #endif // HAVE_MPI
  }
  else  // move all archiving data in a folder to start a new analysis
        // keep only the previous one ( --> not valid if 2 switches)
        // Linux only ?? (implementation MAC and Windows)
  {
    moveFiles("previousScheme","mv");
    if(!_disableResetRestart) _resetRestart=true; //to avoid restart when shitfing schemes
  }


  if(whatScheme==Explicit or whatScheme==Multi)
    this->initMapMElementToComputeTimeStep();
  // find the domain associated to a physical slave number of a rigid contact interaction
  // WARNING a contact interaction cannot be defined on two domains in the same time
  // a physical group has to be define by domain
  if(!_notResetedContact){
    this->initContactInteraction();
    _notResetedContact =true;
  }

  if(!_notResetedBC){
    this->setPairSpaceElementForBoundaryConditions();
    _notResetedBC = true;
  }

  // initialisation of archiving force
  if (!_previousInit)
  {
    this->initArchiveForce();
  }
  else
  {
    if (_notResetUnknowns)
    {      
      for (std::vector<archiveForce>::iterator it = vaf.begin(); it!= vaf.end(); it++)
      {
        it->resetArchiving();
      }
    }    
    else
    {
      for (std::vector<archiveForce>::iterator it = vaf.begin(); it!= vaf.end(); it++)
      {
        it->clear();
      }
      this->initArchiveForce();
    }
    
    for (std::list<IntegralVolume>::iterator it = _dataVolumeIntegral.begin(); it!= _dataVolumeIntegral.end(); it++)
    {
      it->resetArchiving();
    }
    for (std::list<IPDataOnPhysical>::iterator it= _dataOnPhysical.begin(); it!= _dataOnPhysical.end(); it++)
    {
      it->resetArchiving();
    }
  }
}


void nonLinearMechSolver::moveFiles(const std::string &dirName, const std::string &cm)
{
   #if defined(HAVE_MPI)
    if(Msg::GetCommSize() > 1)
    {
       // get the host name of each rank
       const char* myhostname = GetEnvironmentVar("HOSTNAME");
       if(myhostname==NULL)
       {
         //Msg::Info("Hostname does not exist use rank");
         //std::ostringstream strm;
         //strm<<Msg::GetCommRank();
         //myhostname= strm.str().c_str();
         Msg::Info("Hostname does not exist use rank");
         myhostname= "DEFAULTNAME";
       }

       int* vsizename = new int[Msg::GetCommSize()];
       int* vallsizename = new int[Msg::GetCommSize()];
       for(int i=0;i<Msg::GetCommSize();i++)
       {
         vallsizename[i] = 0;
         vsizename[i] = 0;
       }
       vsizename[Msg::GetCommRank()] = (int) strlen(myhostname);
       MPI_Allreduce(vsizename,vallsizename,Msg::GetCommSize(),MPI_INT,MPI_SUM,MPI_COMM_WORLD);
       delete[] vsizename;

       char** vhostname = new char*[Msg::GetCommSize()];
       for(int i=0;i<Msg::GetCommSize();i++)
       {
           vhostname[i] = new char[vallsizename[i]];
           if(Msg::GetCommRank()==i)
           {
             vhostname[i] = (char*)myhostname;
           }
           MPI_Bcast(vhostname[i],vallsizename[i],MPI_CHAR,i,MPI_COMM_WORLD);
       }

       // At this point all processes have an array with the host name of all processes
       // if the host name of the rank doesn't match lower rank files are deleted
       bool firsthostname=true;
       for(int i=0;i<Msg::GetCommRank();i++){
         if(!strcmp(myhostname,vhostname[i]))
         {
            firsthostname = false;
            break;
         }
       }
       if(firsthostname)
       {
         Msg::Warning("File are moved on rank %d, hostname %s",Msg::GetCommRank(),vhostname[Msg::GetCommRank()]);
         std::string rmdir= "rm -rf " + dirName;
         system(rmdir.c_str());
         std::string mdir= "mkdir " + dirName;
         system(mdir.c_str());
         std::string mvcsv= cm + " *.csv " + dirName +"/";
         system(mvcsv.c_str());
         std::string mvdisp= cm + " disp*.msh " + dirName +"/";
         system(mvdisp.c_str());
         std::string mvstr= cm + " stress*.msh " + dirName +"/";
         system(mvstr.c_str());
	 if(!_disableResetRestart){
           struct stat st;
           if(stat("nonLinearMechSolver0.ckp",&st) == 0)
           {
             std::string mvstr= cm + " *.ckp " + dirName +"/";
             system(mvstr.c_str());
           }
	 }
         // get back the mesh file
         //if(!_meshFileName.empty())
         //{
         //  std::string backMesh = "cp previousScheme/" + _meshFileName + " .";
         //  system(backMesh.c_str());
         //}
       }
       else
       {
         Msg::Warning("Nothing moved on rank %d, hostname %s",Msg::GetCommRank(),vhostname[Msg::GetCommRank()]);
       }
       Msg::Barrier();
       // delete array
       delete[] vhostname;
       delete[] vallsizename;
    }
    else
   #endif // HAVE_MPI
    {
      std::string rmdir= "rm -rf " + dirName;
      system(rmdir.c_str());
      std::string mdir= "mkdir " + dirName;
      system(mdir.c_str());
      std::string mvcsv= cm + " *.csv " + dirName +"/";
      system(mvcsv.c_str());
      std::string mvdisp= cm + " disp*.msh " + dirName +"/";
      system(mvdisp.c_str());
      std::string mvstr= cm + " stress*.msh " + dirName +"/";
      system(mvstr.c_str());
      if(!_disableResetRestart){
        struct stat st;
        if(stat("nonLinearMechSolver0.ckp",&st) == 0)
        {
          std::string mvstr= cm + " *.ckp " + dirName +"/";
          system(mvstr.c_str());
        }
      }

      // get back the mesh file
      //if(!_meshFileName.empty())
      //{
      //  std::string backMesh = "cp previousScheme/" + _meshFileName + " .";
      //  system(backMesh.c_str());
      //}
    }
}



void nonLinearMechSolver::initAllBCsOnDofs(){
  if(whatScheme!=StaticLinear)
   this->setTimeForBC(0.);
  else
   this->setTimeForBC(1.);

  this->fixNodalDofs();

	#if defined(HAVE_MPI)
	if (Msg::GetCommRank() == _workingRank)
	#endif //HAVE_MPI
	{
		// Boundary Conditions and Dof creation
		this->fillConstraintDof();
		this->applyConstraintBC();

		// same disp BC
		this->applySameDispBC();
		//
		this->applySameDispBCBetweenTwoGroups();
		//
		this->applyFixOnFace();
    //periodic BC
		this->applyPBCBetweenTwoGroups();

		// for pbc
		if (_testFlag){
		  Msg::Info("initialize test with micro BC");
		  this->initialize_test();
		  this->initCornerBC();
		}
	}
};

void nonLinearMechSolver::init2()
{
  // initialize the system and dofManager
  this->createSystem();

  this->initAllBCsOnDofs();


  if(!_previousInit){
    Msg::Info("Fix and number Dofs");
    this->numberDofs();
    Msg::Info("Dofs are fixed and numbered");
  }

  if (!_notResetUnknowns and _previousInit){
    this->numberDofs();
    Msg::Info("renumbering unknowns because of modifying BCs");
    staticDofManager<double>* sAssembler = static_cast<staticDofManager<double>*>(pAssembler);
    sAssembler->updateFromPrevious();
  }
  /* MPI init */
 #if defined(HAVE_MPI)
  if(whatScheme!=StaticLinear and Msg::GetCommSize()>1)
  {
    staticDofManager<double>* sAssembler = static_cast<staticDofManager<double>*>(pAssembler);
    Msg::Info("MPI initialization begins");
    sAssembler->systemMPIComm(); // To initiate the MPI Communication
    Msg::Info("MPI initialization OK");
  }
 #endif // HAVE_MPI

 /* initialization of Dof value */
  // so now initial condition can be given to system
  this->setInitialCondition();

  #if defined(HAVE_MPI)
  if (Msg::GetCommSize()>1){
    bool isRootRank =false;
    for (int i=0; i<domainVector.size(); i++){
      partDomain* dom = domainVector[i];
      if (dom->getRootRank() == Msg::GetCommRank()){
        isRootRank = true;
        break;
      }
    }
    if (!isRootRank){
      // no view and energy computation
      unknownView.clear();
      ipView.clear();
      energyView.clear();
      _energyComputation = 0;
      _fractureEnergyComputation = 0;
    }
  }
  #endif // HAVE_MPI

  bool stiffcomputation=true;
  if(whatScheme==Explicit) stiffcomputation=false;
  if(!_previousInit){
    /* fields creation */
    if(_ufield) delete _ufield;
    _ufield = new unknownField(this,3,anoded,unknownView);

    Msg::Info("Creation of IPVariable");
    _ipf = new IPField(this,vaip,ipView); // Field for GaussPoint
    Msg::Info("IPVariable are created");

    _ipf->compute1state(IPStateBase::initial,stiffcomputation);
    _ipf->initialBroken(pModel, initbrokeninter);
    // starting state
    _ipf->copy(IPStateBase::initial,IPStateBase::previous);
    _ipf->copy(IPStateBase::initial,IPStateBase::current);

    // initially broken on whole domain
    for (int idom=0; idom< domainVector.size(); idom++){
      dgPartDomain* dgdom = dynamic_cast<dgPartDomain*>(domainVector[idom]);
      if (dgdom != NULL) {
        if (dgdom->allInterfaceIPsAreBrokenFromBeginning()){
          dgdom->initiallyBreakAllInterfaceIP(_ipf->getAips());
        }

      }
    }

		if(_energField) delete _energField;
		_energField = new energeticField(this, energyView,_energyComputation,_fractureEnergyComputation,getScheme());

    //
    #if defined(HAVE_MPI)
    if (Msg::GetCommSize() > 1){
      if (_dataVolumeIntegral.size() > 0){
        // because all average values are computed independently in each procs
        // the volume in chech procs must be estimated
        std::string fname = "VolumeOnProc";
        std::ostringstream oss;
        oss << Msg::GetCommRank();
        fname += oss.str();
        fname += ".csv";
        FILE* fileVolumeProc = fopen(fname.c_str(),"w");
        double volume = 0.;
        for (int i=0; i<domainVector.size(); i++){
          volume += domainVector[i]->computeVolumeDomain(_ipf);
        };
        fprintf(fileVolumeProc,"%e",volume);
        fclose(fileVolumeProc);
      }
    }
    #endif //HAVE_MPI
  }
  else
  {
    _ipf->resetArchiving();
    if (_notResetUnknowns)
    {
      _ufield->resetArchiving();
    }
    else
    {
      if (_ufield) delete _ufield;
      _ufield = new unknownField(this,3,anoded,unknownView);
    }
    
    if (_energField)
    {
      _energField->resetArchiving();
    }
  }
 
  /* terms initialization */
  this->initTerms();
  _previousInit = true;
  Msg::Info("End of initialization");
 #if defined(HAVE_MPI) && defined(_DEBUG)
  if(Msg::GetCommSize() > 1)
  {
    Msg::Error("End init OK from rank %d wait all in debug",Msg::GetCommRank());
    Msg::Barrier();
  }
 #endif // HAVE_MPI
}

void nonLinearMechSolver::endOfScheme(const double curtime, const int step)
{
  this->postproFragment();

  if (_microFlag){
    _homogenizedFiles->closeFiles();
  }

  if (_strainMap && !_strainMappingFollwing){
    Msg::Info("Create displacement and strain mapping");
    _strainMap->solve(pAssembler);
    std::string filename = "dispMap"+int2str(step)+".msh";
    _strainMap->buildDisplacementView(filename);
    filename = "strainMap"+int2str(step)+".msh";
    _strainMap->buildStrainView(filename);
  }
  
  _ufield->archive(curtime,step,true);
  _ipf->archive(curtime,step,true);
  
  if(_energField !=NULL)
  {
    _energField->archive(curtime,step,true);
  }
  forceArchiving(curtime,step,true);
  this->IPVolumeIntegralArchiving(curtime,step,true);
  this->IPDataOnPhysicalArchiving(curtime,step,true);

 #if defined(HAVE_MPI)
  Msg::Barrier(); // Wait all before new computation (MPI problem ??)
 #endif // HAVE_MPI
}

void nonLinearMechSolver::oneStepPreSolve(const double curtime, const double timestep, const int numstep)
{
  this->setTimeForBC(curtime);
  this->setTimeForLaw(curtime,timestep,numstep);
  //for periodic BC
  this->fixNodalDofs();
  if (_testFlag){
    if (_pbcGroup != NULL){
      this->fixDofCorner();
    }
  }
}

void nonLinearMechSolver::oneStepPreSolvePathFollowing(const double curtime, const double timestep, const int numstep){
	static std::string name = "A";
	linearSystem<double>* lsys =pAssembler->getLinearSystem(name);
	pathFollowingSystem<double>* pathsys = dynamic_cast<pathFollowingSystem<double>*>(lsys);
	// if
	double currentPathFollowingIncr = 0.;
	if (_pathFollowingMethod == LOCAL_BASED ){
		// local control always starts with load control
		if (!_switchedControlType){
			pathsys->setControlType(pathFollowingSystemBase::LOAD_CONTROL);
			if (localPathFollowingSwitching()){
				pathsys->setControlType(pathFollowingSystemBase::LOCAL_CONTROL);
				Msg::Info("local control is switched");
				_switchedControlType = true;
			}
		}

		if (pathsys->getControlType() == pathFollowingSystemBase::LOAD_CONTROL){
			_loadStep = _loadStepPrev*_timeStepAdaptationFactor;
			currentPathFollowingIncr = _loadStep;
		}
		else if (pathsys->getControlType() == pathFollowingSystemBase::LOCAL_CONTROL){
      double ff = 1.;
      if (_timeStepAdaptation){
        ff = pow(double(_numNROptimalLocal)/double(_numNROptimal),_expNROptimal);
      }
			_localStep = _localStepPrev*_timeStepAdaptationFactor*ff;
      // check if local step is out of range
      if (_localStep < _localStepMinimal) _localStep = _localStepMinimal;
      else if (_localStep > _localStepMaximal) _localStep = _localStepMaximal;

			currentPathFollowingIncr = _localStep;
		}
	}
	else if (_pathFollowingMethod == GLOBAL_ARC_LENGTH_BASED or _pathFollowingMethod == HYPERELLIPTIC_BASED){
		_arcLengthStep = _arcLengthStepPrev*_timeStepAdaptationFactor;
    if (_arcLengthStep < _arcLengthStepMinimal) _arcLengthStep = _arcLengthStepMinimal;
    else if (_arcLengthStep > _arcLengthStepMaximal) _arcLengthStep = _arcLengthStepMaximal;
    currentPathFollowingIncr = _arcLengthStep;
	}
	else
		Msg::Error("missing case nonLinearMechSolver::oneStepPreSolvePathFollowing");

	pathsys->setPseudoTimeIncrement(currentPathFollowingIncr);

   Msg::Info("path following step: %e  ",currentPathFollowingIncr);


	double loadParam = pathsys->getControlParameter();
	double dloadParam = fabs(pathsys->getControlParameterStep());
	this->oneStepPreSolve(loadParam,dloadParam, numstep);
}


void nonLinearMechSolver::oneStepPostSolve(const double curtime, const int numstep)
{
  // must be check first
  this->checkElementErosion(IPStateBase::current);
  if (_bulkElementErosionFlag or _interfaceElementErosionFlag)
  {
    _elementErosionFilter.print();
  }

  staticDofManager<double> *sAssembler = static_cast<staticDofManager<double>*>(pAssembler);

  /*writing deformed mesh */
  if (_isWriteDeformedMeshToFile){
  	this->writeDeformedMesh(numstep);
  }

  /* Archiving */
  _ufield->archive(curtime,numstep,false);
  _ipf->archive(curtime,numstep,false);
  _energField->archive(curtime,numstep,false);
  
  this->crackTracking(curtime);
  this->forceArchiving(curtime,numstep,false);   // Edge force value;
  this->pathFollowingArchiving(curtime,numstep);

  this->IPVolumeIntegralArchiving(curtime,numstep,false);
  this->IPDataOnPhysicalArchiving(curtime,numstep,false);

	if (_strainMap && _strainMappingFollwing){
		Msg::Info("Create displacement and strain mapping");
		_strainMap->solve(pAssembler);
		std::string filename = "dispMap"+int2str(numstep)+".msh";
		_strainMap->buildDisplacementView(filename);
		filename = "strainMap"+int2str(numstep)+".msh";
		_strainMap->buildStrainView(filename);
	}

  if (_pathFollowing){
    if ((_pathFollowingMethod == LOCAL_BASED) and (!_switchedControlType) and _timeStepAdaptation){
	_localStep = this->computeLocalPathFollowingStep();
    }
    _loadStepPrev = _loadStep;
    _localStepPrev = _localStep;
    _arcLengthStepPrev = _arcLengthStep;
  }

  // check element erosion

  /* field next step */
  if (_selectiveObject != NULL){
    _selectiveObject->nextStep();
  }

  _ipf->nextStep(curtime); //ipvariable
  sAssembler->nextStep();
  // contact
  for(contactContainer::iterator it = _allContact.begin(); it!=_allContact.end(); ++it){
    contactDomain *cdom = *it;
    cdom->nextStep();
  }
  for(defoDefoContactContainer::iterator it = _allDefoDefoContact.begin(); it!=_allDefoDefoContact.end(); ++it){
    defoDefoContactDomain *cdom = *it;
    cdom->nextStep();
  }
  // restart at end if not need to redo next step when reading restart
  if(this->mustRestart(numstep))
  {
    // special operation for the GModel to avoid libGmsh modifications
    pModel->writeMSH(_restartMshFileName, CTX::instance()->mesh.mshFileVersion,CTX::instance()->mesh.binary, CTX::instance()->mesh.saveAll,
                      CTX::instance()->mesh.saveParametric, CTX::instance()->mesh.scalingFactor);
    Msg::Barrier(); // To wait rank0 before end
    restartManager::createRestartFile("nonLinearMechSolver");
    restartManager::restart(this);
    // restart takes only nonm const argument! So convert numstep and curtime
    int restart_numstep = numstep;
    double restart_curtime = curtime;
    restartManager::restart(restart_numstep);
    restartManager::restart(restart_curtime);
    restartManager::closeRestartFile();
    //move output file for restart (if not the stress and disp files have duplicated steps)
    moveFiles("beforeRestart","cp");
    Msg::Barrier(); // To wait rank0 before end
  }

}

void nonLinearMechSolver::checkElementErosion(const IPStateBase::whichState ws){
  if (_bulkElementErosionFlag or _interfaceElementErosionFlag){
    groupOfElements* allg =  new groupOfElements();

    // add element to check
    for (int idom=0; idom< domainVector.size(); idom++){
			partDomain* dom = domainVector[idom];
      if (_bulkElementErosionFlag){
        for (groupOfElements::elementContainer::iterator it = dom->g_cbegin(); it!= dom->g_cend(); it++){
          allg->insert(*it);
        };
      }

      if (_interfaceElementErosionFlag){
        if (dom->getFormulation()){
          dgPartDomain* dgdom = static_cast<dgPartDomain*>(dom);
          for (groupOfElements::elementContainer::iterator it = dgdom->gi->begin(); it!= dgdom->gi->end(); it++){
            allg->insert(*it);
          };
        }
      }
    }
    //
    for (groupOfElements::elementContainer::iterator it = allg->begin(); it!= allg->end(); it++){
      MElement* ele = *it;
      AllIPState::ipstateElementContainer* ips = _ipf->getAips()->getIPstate(ele->getNum());

      if (_erosionType == FIRST_IP_FAILED){
        bool willBeEroded = false;

        for (int ip=0; ip < (*ips).size(); ip++){
          const IPVariable* ipvprev = (*ips)[ip]->getState(IPStateBase::previous);
          const IPVariable* ipv = (*ips)[ip]->getState(ws);
          // check element
          if (_globalCheckFlag){
            // global check
            if (_erosionGlobalCriterion->isFailed(ipvprev,ipv)){
              willBeEroded = true;
              break;
            }
          }
          else if (ipv->isDeleted()){
            // local criterion
            willBeEroded = true;
            break;
          }
        }

        if (willBeEroded){
          // add element
          _elementErosionFilter.addErodedElement(ele);
          // delete all remaining IP
          for (int ip=0; ip < (*ips).size(); ip++){
            IPVariable* ipv = (*ips)[ip]->getState(ws);
            if (!ipv->isDeleted()){
              ipv->Delete(true);
            }
          }
        }
      }
      else if (_erosionType == ALL_IP_FAILED){
        int numberNoFailedIP= 0;
        for (int ip=0; ip < (*ips).size(); ip++){
          const IPVariable* ipv = (*ips)[ip]->getState(ws);
          const IPVariable* ipvprev = (*ips)[ip]->getState(IPStateBase::previous);
          if (_globalCheckFlag){
            if (!_erosionGlobalCriterion->isFailed(ipvprev,ipv)){
              numberNoFailedIP++;
              break;
            }
          }
          else if (!ipv->isDeleted()){
            numberNoFailedIP++;
            break;
          }
        }
        if (numberNoFailedIP == 0){
          for (int ip=0; ip < (*ips).size(); ip++){
            IPVariable* ipv = (*ips)[ip]->getState(ws);
            if (!ipv->isDeleted()){
              ipv->Delete(true);
            }
          }
          _elementErosionFilter.addErodedElement(ele);
        }
      }
      else{
        Msg::Error("_erosionType = %d is not defined nonLinearMechSolver::checkElementErosion",_erosionType);
      }
    }

    delete allg;
  }
};

void nonLinearMechSolver::restart()
{
  int newtag = _tag;
  restartManager::restart(newtag);
  restartErosion();
  if(_tag != newtag) Msg::Error("The solver tag has change so restart cannot be done!");
  restartManager::restart(_ufield);
  restartManager::restart(_ipf);
  restartManager::restart(_energField);
  staticDofManager<double> *sAssembler = static_cast<staticDofManager<double>*>(pAssembler);
  sAssembler->restart();

  int cStep = (int) _currentStep;
  restartManager::restart(cStep);
  _currentStep = cStep;
  int newSolver = (int)whatSolver;
  restartManager::restart(newSolver);
  if(newSolver != (int)whatSolver) Msg::Error("You cannot change the solver at restart!");
  int newScheme = (int)whatScheme;
  restartManager::restart(newScheme);
  if(newScheme != (int)whatScheme) Msg::Error("You cannot change the scheme at restart!");

  restartTimeStepAdaptation();
  restartPathFollowing();
  restartDefoDefoContact();
  /// !!! ++ PeriodicBC solver Van Dung I do not know what to save !!!
  return;
}
void nonLinearMechSolver::restartTimeStepAdaptation()
{
  //restartManager::restart(_numNROptimal}; // num optimal
  //restartManager::restart(_numNROptimalLocal}; // number optimal for local
  //restartManager::restart(_expNROptimal};
  restartManager::restart(_timeStepAdaptationFactor); // for timestep adaptation = _numNROptimal/previous number NR iterations
  //restartManager::restart(_timeStepAdaptation}; // true if path following step is adapted with number NR
}
void nonLinearMechSolver::restartPathFollowing()
{
  //restartManager::restart(_pathFollowing); // true if considering pathFollowing
  //int _macroControlTypePathFollowing;
  //int _correctionMethodPathFollowing;
  //int _tranversalCriterionPathFollowing;
  //int _solverTypePathFollowing;
  //double _pathFollowingEqRatio;

  //pathFollowingMethod _pathFollowingMethod; // 0- arclength based, 1 -
  restartManager::restart(_loadStep); // for load control
  restartManager::restart(_localStep); // to local cr control
  restartManager::restart(_loadStepPrev); // for load control
  restartManager::restart(_localStepPrev); // to local cr control
  //double _localStepMinimal;
  //double _localStepMaximal;

  restartManager::restart(_arcLengthStep); // for arc-length control
  restartManager::restart(_arcLengthStepPrev); // previous

  //double _pathFollowingSwitchCriterion;
  restartManager::restart(_switchedControlType);
  //pathFollowingLocalIncrementType _pathFollowingIncrementType;
  //pathFollowingLocation _pathFollowingLocation;
}
void nonLinearMechSolver::restartErosion()
{
  //restartManager::restart(_maxAllowedNewlyBrokenElement);  // Max number of allowed newly cracked elements during one step (for implicit case)
  // Erosion control
  //restartManager::restart(_bulkElementErosionFlag); // true if bulk element erosion
  //restartManager::restart(_interfaceElementErosionFlag); // true if interface element erosion
  //restartManager::restart(_erosionType); // all ip failed
  //FailureCriterionBase* _erosionGlobalCriterion;
  //restartManager::restart(_globalCheckFlag); // true if global verification based on _erosionGlobalCriterion
  _elementErosionFilter.restart(); // a set consists of all erosion elements
  _elementErosionFilter.print();
}
void nonLinearMechSolver::restartDefoDefoContact()
{
  for(defoDefoContactContainer::iterator it = _allDefoDefoContact.begin(); it!=_allDefoDefoContact.end(); ++it){
    defoDefoContactDomain *cdom = *it;
    cdom->restart();
  }
}
void nonLinearMechSolver::initContactInteraction(){
  // find a common element between slave group of element and the group of element of partdomain
  bool flagfind;
  for(contactContainer::iterator it = _allContact.begin(); it!=_allContact.end(); ++it){
    contactDomain *cdom = *it;
    // get physical of domain
    int physSlave = cdom->getPhysSlave();
    flagfind=false;
    for(int kk=0;kk<_mpiUserDom; kk++){
      partDomain *dom = domainVector[kk];
      if(dom->getPhysical() == physSlave){
        flagfind =true;
        cdom->setDomainAndFunctionSpace(dom);
        break;
      }
    }
    if(!flagfind){
     #if defined(HAVE_MPI)
      // find the domain in the Ghosted domain (the search on an empty group OK)
      for(int kk=0; kk<_ghostDomainMPI.size(); kk++)
      {
        partDomain *dom = _ghostDomainMPI[kk];
        if(dom->getPhysical() == physSlave){
          flagfind =true;
          cdom->setDomainAndFunctionSpace(dom);
          break;
        }
      }
      if(!flagfind)
     #endif // HAVE_MPI
      {
        Msg::Error("The contact interaction %d can't be initiated because the domain is not find on rank %d",cdom->getPhys(),Msg::GetCommRank());
      }
    }
  }
  // now init the contact boundary conditions
  //for(rigidContactBC &diri : allContactBC){
  for(std::list<rigidContactBC>::iterator diri = allContactBC.begin(); diri!=allContactBC.end(); ++diri){
    if(diri->onWhat == nonLinearBoundaryCondition::RIGIDCONTACT){
      for(contactContainer::iterator itcdom = _allContact.begin(); itcdom!=_allContact.end(); ++itcdom){
        contactDomain *cdom = *itcdom;
        if(cdom->getPhys() == diri->_tag){ // we can have some slave domain with the same master --> don't put a break in the loop !
          rigidContactSpaceBase *sp = static_cast<rigidContactSpaceBase*>(cdom->getSpace());
          diri->setSpace(sp);
          diri->_filter = cdom->getDomain()->createFilterDof(diri->_comp);
        }
      }
    }
  }

  // Idem for defo defo contact. MPI not implemented 
  // for parallel we need to define the master faces as ghost faces in all the other partitions
  for(defoDefoContactContainer::iterator it = _allDefoDefoContact.begin(); it!=_allDefoDefoContact.end(); ++it){
    defoDefoContactDomain *cdom = (*it);

    bool domfindSlave=false;
    int physSlave = cdom->getPhysSlave();
    groupOfElements gSlave(cdom->getDimSlave(), physSlave); //will not be insterted, so delete it at the end
    // map
    std::map<partDomain*,std::vector<MElement*> > mapfindSlave;
    for(std::vector<partDomain*>::iterator itdom=domainVector.begin(); itdom!=domainVector.end(); ++itdom){
      partDomain *dom = *itdom;
      if(physSlave == dom->getPhysical()){
        domfindSlave = true;
        for(groupOfElements::elementContainer::iterator it2=dom->g_cbegin(); it2!=dom->g_cend(); ++it2)
          mapfindSlave[dom].push_back(*it2);
        break;
      }
    }
    #if defined(HAVE_MPI)
    // look if the BC is applied on a vertex
    if(!domfindSlave){
      if(cdom->getDimSlave() == 1){
        // loop on dof of other partition (if the dof is included only on this partition it will be find after)
        groupOfElements::vertexContainer::iterator itvBC = gSlave.vbegin();
        MVertex *verBC = *itvBC;
        for(std::map<int,std::vector<groupOfElements> >::iterator it=_mapOtherDomain.begin(); it!=_mapOtherDomain.end(); ++it){
          std::vector<groupOfElements> & groupOtherPart = it->second;
          for(int j=0;j<Msg::GetCommRank(); j++){ // look on domain with a lower partition number
            for(groupOfElements::vertexContainer::const_iterator itv = groupOtherPart[j].vbegin(); itv!=groupOtherPart[j].vend(); ++itv){
              MVertex *ver = *itv;
              if(verBC == ver){
                domfindSlave=true; // the BC will be applied on an other partition
                break;
              }
            }
          }
        }
      }
    }
    #endif //HAVE_MPI
    if(!domfindSlave){
      // loop on element
      std::map<partDomain*,MElement*> mapelemtype; // need for gauss quadrature rule TODO group with mapfind in a structure but no time benefit
      for(groupOfElements::elementContainer::iterator it=gSlave.begin(); it!=gSlave.end();++it){
        MElement *e = *it;
        bool flagfind = false;
        // Loop on all Domain to find on which domain the BC is applied
        for(int j=0; j<_mpiUserDom; j++){
          // if Dimensions are the same the element can be find with a comparison to element
          partDomain *dom = domainVector[j];
          // otherwise the vertex must be identified separately
          MElement *etest = *(dom->g_cbegin());
          mapelemtype[dom] = etest;
          if( (etest != NULL) and (etest->getDim() == e->getDim())){
            for(groupOfElements::elementContainer::iterator it2=dom->g_cbegin(); it2!=dom->g_cend(); ++it2){
              if(*it2 == e){ // The Element is contained in the domain
                mapfindSlave[dom].push_back(e);
                flagfind = true;
                break;
              }
            }
          }
          else{ // loop on Vertex identify the BC with first vertex of e
                 // an element is included in one domain --> one vertex of Element is OK
                 // in mpi the BC can be included in another partition
            #if defined(HAVE_MPI)
            for(int j=_mpiUserDom; j<domainVector.size(); j++){ // loop only on interface domain "added by mpi"
              dgPartDomain *dgdom = static_cast<dgPartDomain*>(domainVector[j]);
              // get the two partition number
              // Neumann BC cannot be applied on ghost element (will not be assembled in the end)
              // do not applied on ghost element AS THEY WILL NOT BE ABLE TO BE ASSEMBLED!
              // loop over ghost element and see if there is a match
              for(groupOfElements::elementContainer::const_iterator iteleDom = dgdom->gi->begin(); iteleDom != dgdom->gi->end();++iteleDom)
              {
                MInterfaceElement *ie = dynamic_cast<MInterfaceElement*>(*iteleDom);
                int numMinus = ie->getElem(0)->getPartition();
                int numPlus = ie->getElem(1)->getPartition();
                // identify the Ghost
                MElement* eleGhost = (numMinus == Msg::GetCommRank()+1) ? ie->getElem(1) : ie->getElem(0);
                // loop over nodes to see if they match
                int nmatch = 0;
                for(int i=0;i<eleGhost->getNumVertices();++i)
                {
                  MVertex* verGhost = eleGhost->getVertex(i);
                  for(int jj=0;jj<e->getNumVertices();jj++){
                    MVertex *verBC = e->getVertex(jj);
                    if(verBC == verGhost)
                    {
                      nmatch++;
                      if(nmatch == e->getNumVertices())
                      {
                        flagfind = true; // the BC will be applied on another partition as on ghost here
                        break;
                      }
                    }
                  }
                  if(flagfind) break;
                }
                if(flagfind) break;
              }
            }
            #endif // HAVE_MPI
            if(!flagfind){
              // check all vertex of the BC (node 0 could be on another partition)
              int ncount = 0;
              for(groupOfElements::vertexContainer::const_iterator itv=dom->g_cvbegin(); itv!=dom->g_cvend(); ++itv){
                for(int jj=0; jj<e->getNumVertices();++jj)
                {
                  MVertex *ver = e->getVertex(jj);
                  if((*itv) == ver){
                    ncount ++;
                    if(ncount == e->getNumVertices())
                    {
                      mapfindSlave[dom].push_back(e);
                      flagfind = true;
                      break;
                    }
                  }
                }
                if(flagfind) break;
              }
            }
          }
        if(flagfind)
          break;
        }
      }
    }



    bool domfindMaster=false;
    int physMaster = cdom->getPhysMaster();
    //groupOfElements *gMaster = new groupOfElements (cdom->getDimMaster(), physMaster);
    groupOfElements gMaster(cdom->getDimMaster(), physMaster); //will not be insterted, so delete it at the end
    // map
    std::map<partDomain*,std::vector<MElement*> > mapfindMaster;
    for(std::vector<partDomain*>::iterator itdom=domainVector.begin(); itdom!=domainVector.end(); ++itdom){
      partDomain *dom = *itdom;
      if(physMaster == dom->getPhysical()){
        domfindMaster = true;
        for(groupOfElements::elementContainer::iterator it2=dom->g_cbegin(); it2!=dom->g_cend(); ++it2)
          mapfindMaster[dom].push_back(*it2);
        break;
      }
    }
    #if defined(HAVE_MPI)
    // look if the BC is applied on a vertex
    if(!domfindMaster){
      if(cdom->getDimMaster() == 1){
        // loop on dof of other partition (if the dof is included only on this partition it will be find after)
        groupOfElements::vertexContainer::iterator itvBC = gMaster.vbegin();
        MVertex *verBC = *itvBC;
        for(std::map<int,std::vector<groupOfElements> >::iterator it=_mapOtherDomain.begin(); it!=_mapOtherDomain.end(); ++it){
          std::vector<groupOfElements> & groupOtherPart = it->second;
          for(int j=0;j<Msg::GetCommRank(); j++){ // look on domain with a lower partition number
            for(groupOfElements::vertexContainer::const_iterator itv = groupOtherPart[j].vbegin(); itv!=groupOtherPart[j].vend(); ++itv){
              MVertex *ver = *itv;
              if(verBC == ver){
                domfindMaster=true; // the BC will be applied on an other partition
                break;
              }
            }
          }
        }
      }
    }
    #endif //HAVE_MPI
    if(!domfindMaster){
      // loop on element
      std::map<partDomain*,MElement*> mapelemtype; // need for gauss quadrature rule TODO group with mapfind in a structure but no time benefit
      for(groupOfElements::elementContainer::iterator it=gMaster.begin(); it!=gMaster.end();++it){
        MElement *e = *it;
        bool flagfind = false;
        // Loop on all Domain to find on which domain the BC is applied
        for(int j=0; j<_mpiUserDom; j++){
          // if Dimensions are the same the element can be find with a comparison to element
          partDomain *dom = domainVector[j];
          // otherwise the vertex must be identified separately
          MElement *etest = *(dom->g_cbegin());
          mapelemtype[dom] = etest;
          if( (etest != NULL) and (etest->getDim() == e->getDim())){
            for(groupOfElements::elementContainer::iterator it2=dom->g_cbegin(); it2!=dom->g_cend(); ++it2){
              if(*it2 == e){ // The Element is contained in the domain
                mapfindMaster[dom].push_back(e);
                flagfind = true;
                break;
              }
            }
          }
          else{ // loop on Vertex identify the BC with first vertex of e
                 // an element is included in one domain --> one vertex of Element is OK
                 // in mpi the BC can be included in another partition
            #if defined(HAVE_MPI)
            for(int j=_mpiUserDom; j<domainVector.size(); j++){ // loop only on interface domain "added by mpi"
              dgPartDomain *dgdom = static_cast<dgPartDomain*>(domainVector[j]);
              // get the two partition number
              // Neumann BC cannot be applied on ghost element (will not be assembled in the end)
              // do not applied on ghost element AS THEY WILL NOT BE ABLE TO BE ASSEMBLED!
              // loop over ghost element and see if there is a match
              for(groupOfElements::elementContainer::const_iterator iteleDom = dgdom->gi->begin(); iteleDom != dgdom->gi->end();++iteleDom)
              {
                MInterfaceElement *ie = dynamic_cast<MInterfaceElement*>(*iteleDom);
                int numMinus = ie->getElem(0)->getPartition();
                int numPlus = ie->getElem(1)->getPartition();
                // identify the Ghost
                MElement* eleGhost = (numMinus == Msg::GetCommRank()+1) ? ie->getElem(1) : ie->getElem(0);
                // loop over nodes to see if they match
                int nmatch = 0;
                for(int i=0;i<eleGhost->getNumVertices();++i)
                {
                  MVertex* verGhost = eleGhost->getVertex(i);
                  for(int jj=0;jj<e->getNumVertices();jj++){
                    MVertex *verBC = e->getVertex(jj);
                    if(verBC == verGhost)
                    {
                      nmatch++;
                      if(nmatch == e->getNumVertices())
                      {
                        flagfind = true; // the BC will be applied on another partition as on ghost here
                        break;
                      }
                    }
                  }
                  if(flagfind) break;
                }
                if(flagfind) break;
              }
            }
            #endif // HAVE_MPI
            if(!flagfind){
              // check all vertex of the BC (node 0 could be on another partition)
              int ncount = 0;
              for(groupOfElements::vertexContainer::const_iterator itv=dom->g_cvbegin(); itv!=dom->g_cvend(); ++itv){
                for(int jj=0; jj<e->getNumVertices();++jj)
                {
                  MVertex *ver = e->getVertex(jj);
                  if((*itv) == ver){
                    ncount ++;
                    if(ncount == e->getNumVertices())
                    {
                      mapfindMaster[dom].push_back(e);
                      flagfind = true;
                      break;
                    }
                  }
                }
                if(flagfind) break;
              }
            }
          }
          if(flagfind)
            break;  
        }
      }
    }
    // now the map Space Element is filled for the BC
    for(std::map<partDomain*,std::vector<MElement*> >::iterator it=mapfindMaster.begin(); it!=mapfindMaster.end(); ++it)
    {
      groupOfElements mast(it->second); 
      for(std::map<partDomain*,std::vector<MElement*> >::iterator itslave=mapfindSlave.begin(); itslave!=mapfindSlave.end(); ++itslave)
      {
        groupOfElements slav(itslave->second);         
        cdom->insertSlaveMasterDomainsAndFunctionSpace(itslave->first,&slav,it->first,&mast);  
      }
    }
  }
}

void nonLinearMechSolver::setPairSpaceElementForBoundaryConditions()
{
  // The current GModel has to be selected to create the BC
  GModel* save_current_model = GModel::current();
  GModel::setCurrent(pModel);

  // loop on BC
  // BC which are prescribed weakly (the interfaceElement has to be finded in vinter)
  // to prescribed the BC
  // Theta (normally on Boundary only --> impossible that the condition be on 2 partitions)
  for(std::list<nonLinearNeumannBC>::iterator ittheta=allTheta.begin(); ittheta!=allTheta.end(); ++ittheta){
     for(groupOfElements::elementContainer::iterator ite=ittheta->g->begin(); ite!=ittheta->g->end(); ++ite){
       MVertex *vv = (*ite)->getVertex(2);
       for(int k=0;k<domainVector.size(); k++){
         partDomain *dom = domainVector[k];
         if(dom->IsInterfaceTerms()){ // change this (virtual interface has to be defined on domain)
           // Ok because 2nd degree min BoundaryInterfaceElement is created only for this vertex
           dgPartDomain *dgdom = static_cast<dgPartDomain*>(dom);
           groupOfElements *gvi = dgdom->giv;
           for(groupOfElements::elementContainer::const_iterator it_inter = gvi->begin(); it_inter!=gvi->end();++it_inter){
             MElement *minter = *it_inter;
             //loop on component of map_edge
             int numvertexminus1 = minter->getNumVertices()-1;
             if((vv == minter->getVertex(2)) or (vv == minter->getVertex(numvertexminus1))){
               dgdom->gib->insert(minter);
               break;
             }
           }
         }
       }
     }
   }

#if 1
  //Dirichlet (can be applied on each domain look on ghost element to prescribed BC on them avoid a search in unknown map at each iteration)
  for(std::list<nonLinearDirichletBC>::iterator itbc = allDirichlet.begin(); itbc != allDirichlet.end();++itbc){
    // first look if the BC is not applied on whole domain
    bool domfind=false;
    for(std::vector<partDomain*>::iterator itdom=domainVector.begin(); itdom!=domainVector.end(); ++itdom){
      partDomain *dom = *itdom;
      if(itbc->_tag == dom->getPhysical()){
        domfind = true;
        if(dom->_groupGhostMPI->size() > 0) // save BC to applied an other groupOfElement
        {
          FunctionSpaceBase *spdbc = dom->getSpaceForBC(nonLinearBoundaryCondition::DIRICHLET,itbc->onWhat, nonLinearNeumannBC::NEUMANN_UNDEF,
                                                        itbc->_comp,dom->_groupGhostMPI);
          FilterDof *fdof = dom->createFilterDof(itbc->_comp);
          itbc->_vspace.push_back(spdbc);
          itbc->_vgroup.push_back(dom->_groupGhostMPI);
          itbc->_vfilter.push_back(fdof);
          itbc->_vdg.push_back(dom->getFormulation());
        }
        FunctionSpaceBase *spdbc = dom->getSpaceForBC(nonLinearBoundaryCondition::DIRICHLET,itbc->onWhat, nonLinearNeumannBC::NEUMANN_UNDEF,
                                                        itbc->_comp,dom->elements());
        FilterDof *fdof = dom->createFilterDof(itbc->_comp);
        itbc->_vspace.push_back(spdbc);
        itbc->_vgroup.push_back(dom->elements());
        itbc->_vfilter.push_back(fdof);
        itbc->_vdg.push_back(dom->getFormulation());
        break;
      }
    }
    #if defined(HAVE_MPI)
    // same on ghost domain (be aware copy past of loop on domain !!)
    bool ghostdomfind=false;
    for(std::vector<partDomain*>::iterator itdom=_ghostDomainMPI.begin(); itdom!=_ghostDomainMPI.end(); ++itdom){
      partDomain *dom = *itdom;
      if(itbc->_tag == dom->getPhysical() and dom->_groupGhostMPI->size()>0){
        ghostdomfind = true;
	    FunctionSpaceBase *spdbc = dom->getSpaceForBC(nonLinearBoundaryCondition::DIRICHLET,itbc->onWhat, nonLinearNeumannBC::NEUMANN_UNDEF,
                                                        itbc->_comp,dom->_groupGhostMPI);
	    FilterDof* fdof = dom->createFilterDof(itbc->_comp);
        itbc->_vspace.push_back(spdbc);
        itbc->_vgroup.push_back(dom->_groupGhostMPI);
        itbc->_vfilter.push_back(fdof);
        itbc->_vdg.push_back(dom->getFormulation());
        break;
      }
    }
    #endif // HAVE_MPI
    if(!domfind){
      // loop on element
      // map
      std::map<partDomain*,std::vector<MElement*> > mapfind;
      std::map<partDomain*,std::vector<MElement*> > mapfindGhost;
      for(groupOfElements::elementContainer::iterator it=itbc->g->begin(); it!=itbc->g->end();++it){
        MElement *e = *it;
        // Loop on all Domain to find on which domain the BC is applied
        for(std::vector<partDomain*>::iterator itdom=domainVector.begin(); itdom!=domainVector.end(); ++itdom){
          // if Dimensions are the same the element can be find with a comparison to element
          partDomain *dom = *itdom;
          // otherwise the vertex must be identified separately
          const MElement *etest = *(dom->g_cbegin());
          if( (etest !=NULL) and (etest->getDim() == e->getDim())){
            for(groupOfElements::elementContainer::iterator it2=dom->g_cbegin(); it2!=dom->g_cend(); ++it2){
              if(*it2 == e){ // The Element is contained in the domain
                mapfind[dom].push_back(e);
                break;
              }
            }
            #if defined(HAVE_MPI)
            for(groupOfElements::elementContainer::iterator it2=dom->_groupGhostMPI->begin(); it2!=dom->_groupGhostMPI->end(); ++it2){
              if(*it2 == e){ // The Element is contained in the domain
                mapfind[dom].push_back(e);
                break;
              }
            }
            #endif // HAVE_MPI
          }
          else{
            // an element is included in one domain --> one vertex of Element is OK
            // On interface the BC is put on the first met domain
            fillMapOfInterfaceElementsInOneDomain(e, mapfind[(*itdom)], dom->elements());
            #if defined(HAVE_MPI)
            // change map in mapFindGhost???
            fillMapOfInterfaceElementsInOneDomain(e, mapfind[(*itdom)], dom->_groupGhostMPI);
            // check on element which are not included in _groupGhostMPI as there is no interface element with the element
            for(int j=0;j<Msg::GetCommSize();j++)
            {
              if(j!=Msg::GetCommRank() and !(*itdom)->getFormulation() ) // no ghost with itself
                fillMapOfInterfaceElementsInOneDomain(e, mapfindGhost[(*itdom)], &(_mapOtherDomain[(*itdom)->getPhysical()][j]));
            }

            #endif // HAVE_MPI
          }
        }
      }
      // now the space element map of the BC is filled
      for(std::map<partDomain*,std::vector<MElement*> >::iterator it=mapfind.begin(); it!=mapfind.end(); ++it){
        if(it->second.size() !=0){
          groupOfElements* dirig = new groupOfElements(it->second);
          FunctionSpaceBase *spdbc = it->first->getSpaceForBC(nonLinearBoundaryCondition::DIRICHLET,itbc->onWhat, nonLinearNeumannBC::NEUMANN_UNDEF,
                                                        itbc->_comp,dirig);
          FilterDof *fdof = it->first->createFilterDof(itbc->_comp);
          itbc->_vspace.push_back(spdbc);
          itbc->_vgroup.push_back(dirig);
          itbc->_vfilter.push_back(fdof);
          itbc->_vdg.push_back(it->first->getFormulation());
        }
      }
      // same with mapfindGhost
      #if defined(HAVE_MPI)
      for(std::map<partDomain*,std::vector<MElement*> >::iterator it=mapfindGhost.begin(); it!=mapfindGhost.end(); ++it){
        if(it->second.size() !=0){
          groupOfElements* dirig = new groupOfElements(it->second);
          FunctionSpaceBase *spdbc = it->first->getSpaceForBC(nonLinearBoundaryCondition::DIRICHLET,itbc->onWhat, nonLinearNeumannBC::NEUMANN_UNDEF,
                                                              itbc->_comp,dirig);
          FilterDof *fdof = it->first->createFilterDof(itbc->_comp);
          itbc->_vspace.push_back(spdbc);
          itbc->_vgroup.push_back(dirig);
          itbc->_vfilter.push_back(fdof);
          itbc->_vdg.push_back(it->first->getFormulation());
        }
      }
      #endif // HAVE_MPI
    }

    #if defined(HAVE_MPI)
    // same on ghostDomain (be aware copy past of loop on domain !!)
    if(!ghostdomfind){
      // loop on element
      // map
      std::map<partDomain*,std::vector<MElement*> > ghostmapfind;
      for(groupOfElements::elementContainer::iterator it=itbc->g->begin(); it!=itbc->g->end();++it){
        MElement *e = *it;
        // Loop on all Domain to find on which domain the BC is applied
        for(std::vector<partDomain*>::iterator itdom=_ghostDomainMPI.begin(); itdom!=_ghostDomainMPI.end(); ++itdom){
          // if Dimensions are the same the element can be find with a comparison to element
          partDomain *dom = *itdom;
          // otherwise the vertex must be identified separately
          MElement *etest = *(dom->g_cbegin());
          if( (etest !=NULL) and (etest->getDim() == e->getDim())){
            for(groupOfElements::elementContainer::iterator it2=dom->_groupGhostMPI->begin(); it2!=dom->_groupGhostMPI->end(); ++it2){
              if(*it2 == e){ // The Element is contained in the domain
                ghostmapfind[dom].push_back(e);
                break;
              }
            }
          }
          else
          {
            // an element is included in one domain --> one vertex of Element is OK
            // On interface the BC is put on the first met domain
            fillMapOfInterfaceElementsInOneDomain(e, ghostmapfind[(*itdom)], dom->_groupGhostMPI);
            // check on element which are not included in _groupGhostMPI as there is no interface element with the element
            for(int j=0;j<Msg::GetCommSize();j++)
            {
              if(j!=Msg::GetCommRank() and !(*itdom)->getFormulation() ) // no ghost with itself
                fillMapOfInterfaceElementsInOneDomain(e, ghostmapfind[(*itdom)], &(_mapOtherDomain[(*itdom)->getPhysical()][j]));
            }

          }
        }
      }
      // now the space element of the BC is filled
      for(std::map<partDomain*,std::vector<MElement*> >::iterator it=ghostmapfind.begin(); it!=ghostmapfind.end(); ++it){
        if(it->second.size() !=0){
          groupOfElements *dirig = new groupOfElements(it->second);
          FunctionSpaceBase *spdbc = it->first->getSpaceForBC(nonLinearBoundaryCondition::DIRICHLET,itbc->onWhat, nonLinearNeumannBC::NEUMANN_UNDEF,
                                                        itbc->_comp,dirig);
          FilterDof* fdof = it->first->createFilterDof(itbc->_comp);
          itbc->_vspace.push_back(spdbc);
          itbc->_vgroup.push_back(dirig);
          itbc->_vfilter.push_back(fdof);
          itbc->_vdg.push_back(it->first->getFormulation());
        }
      }
    }
    #endif // HAVE_MPI
  }

  // Idem for Neumann BC APPLY ON FIRST FOUNDED DOM. MPI problem the condition has be applied once (on the partition with the lowest number)
  for(std::list<nonLinearNeumannBC>::iterator itbc = allNeumann.begin(); itbc!=allNeumann.end();++itbc){
    // first look if the BC is not applied on whole domain
    bool domfind=false;
    for(std::vector<partDomain*>::iterator itdom=domainVector.begin(); itdom!=domainVector.end(); ++itdom){
      partDomain *dom = *itdom;
      if(itbc->_tag == dom->getPhysical()){
        domfind = true;
        FunctionSpaceBase *spnbc = dom->getSpaceForBC(nonLinearBoundaryCondition::NEUMANN,itbc->onWhat, itbc->_NeumannBCType,
                                                        itbc->_comp,dom->elements());
        QuadratureBase *integ = dom->getQuadratureRulesForNeumannBC(*itbc);
        itbc->_vspace.push_back(spnbc);
        itbc->_vgroup.push_back(dom->elements());
        itbc->_vquadrature.push_back(integ);
        itbc->_vdom.push_back(dom);
        break;
      }
    }
    // map
    std::map<partDomain*,std::vector<MElement*> > mapfind;
    #if defined(HAVE_MPI)
    // look if the BC is applied on a vertex
    if(!domfind){
      if(itbc->g->vsize() == 1){
        // loop on dof of other partition (if the dof is included only on this partition it will be find after)
        groupOfElements::vertexContainer::iterator itvBC = itbc->g->vbegin();
        MVertex *verBC = *itvBC;
        for(std::map<int,std::vector<groupOfElements> >::iterator it=_mapOtherDomain.begin(); it!=_mapOtherDomain.end(); ++it){
          std::vector<groupOfElements> & groupOtherPart = it->second;
          for(int j=0;j<Msg::GetCommRank(); j++){ // look on domain with a lower partition number
            for(groupOfElements::vertexContainer::const_iterator itv = groupOtherPart[j].vbegin(); itv!=groupOtherPart[j].vend(); ++itv){
              MVertex *ver = *itv;
              if(verBC == ver){
                domfind=true; // the BC will be applied on an other partition
                break;
              }
            }
          }
        }
      }
    }
    #endif //HAVE_MPI
    if(!domfind){
      // loop on element
      std::map<partDomain*,MElement*> mapelemtype; // need for gauss quadrature rule TODO group with mapfind in a structure but no time benefit
      for(groupOfElements::elementContainer::iterator it=itbc->g->begin(); it!=itbc->g->end();++it){
        MElement *e = *it;
        bool flagfind = false;
        // Loop on all Domain to find on which domain the BC is applied
        for(int j=0; j<_mpiUserDom; j++){
          // if Dimensions are the same the element can be find with a comparison to element
          partDomain *dom = domainVector[j];
          // otherwise the vertex must be identified separately
          MElement *etest = *(dom->g_cbegin());
          mapelemtype[dom] = etest;
          if( (etest != NULL) and (etest->getDim() == e->getDim())){
            for(groupOfElements::elementContainer::iterator it2=dom->g_cbegin(); it2!=dom->g_cend(); ++it2){
              if(*it2 == e){ // The Element is contained in the domain
                mapfind[dom].push_back(e);
                flagfind = true;
                break;
              }
            }
          }
          else{ // loop on Vertex identify the BC with first vertex of e
                 // an element is included in one domain --> one vertex of Element is OK
                 // in mpi the BC can be included in another partition
            #if defined(HAVE_MPI)
            for(int j=_mpiUserDom; j<domainVector.size(); j++){ // loop only on interface domain "added by mpi"
              dgPartDomain *dgdom = static_cast<dgPartDomain*>(domainVector[j]);
              // get the two partition number
              // Neumann BC cannot be applied on ghost element (will not be assembled in the end)
              // do not applied on ghost element AS THEY WILL NOT BE ABLE TO BE ASSEMBLED!
#if 0 // old code
              groupOfElements::elementContainer::const_iterator iteleDom = dgdom->gi->begin();
              MInterfaceElement *ie = dynamic_cast<MInterfaceElement*>(*iteleDom);
              int numPlus = ie->getElem(1)->getPartition();
              // if this partition is the one of numPlus it is the partition with the highest number and
              // it has been look  if the BC appear in the other partition
              if(Msg::GetCommRank()+1 == numPlus){
                // loop on element of the domain
                for(groupOfElements::vertexContainer::const_iterator itv=dgdom->gi->vbegin(); itv!=dgdom->gi->vend(); ++itv){
                  MVertex *verDom = *itv;
                  for(int jj=0;jj<e->getNumVertices();jj++){
                    MVertex *verBC = e->getVertex(jj);
                    if(verBC == verDom){
                      flagfind = true; // the BC will be applied on another partition
                      break;
                    }
                  }
                  if(flagfind) break;
                }
              }
#else
              // loop over ghost element and see if there is a match
              for(groupOfElements::elementContainer::const_iterator iteleDom = dgdom->gi->begin(); iteleDom != dgdom->gi->end();++iteleDom)
              {
                MInterfaceElement *ie = dynamic_cast<MInterfaceElement*>(*iteleDom);
                int numMinus = ie->getElem(0)->getPartition();
                int numPlus = ie->getElem(1)->getPartition();
                // identify the Ghost
                MElement* eleGhost = (numMinus == Msg::GetCommRank()+1) ? ie->getElem(1) : ie->getElem(0);
                // loop over nodes to see if they match
                int nmatch = 0;
                for(int i=0;i<eleGhost->getNumVertices();++i)
                {
                  MVertex* verGhost = eleGhost->getVertex(i);
                  for(int jj=0;jj<e->getNumVertices();jj++){
                    MVertex *verBC = e->getVertex(jj);
                    if(verBC == verGhost)
                    {
                      nmatch++;
                      if(nmatch == e->getNumVertices())
                      {
                        flagfind = true; // the BC will be applied on another partition as on ghost here
                        break;
                      }
                    }
                  }
                  if(flagfind) break;
                }
                if(flagfind) break;
              }
#endif //0
            }
            #endif // HAVE_MPI
            if(!flagfind){
              // check all vertex of the BC (node 0 could be on another partition)
              int ncount = 0;
              for(groupOfElements::vertexContainer::const_iterator itv=dom->g_cvbegin(); itv!=dom->g_cvend(); ++itv){
                for(int jj=0; jj<e->getNumVertices();++jj)
                {
                  MVertex *ver = e->getVertex(jj);
                  if((*itv) == ver){
                    ncount ++;
                    if(ncount == e->getNumVertices())
                    {
                      mapfind[dom].push_back(e);
                      flagfind = true;
                      break;
                    }
                  }
                }
                if(flagfind) break;
              }
            }
          }
        if(flagfind)
          break;
        }
      }
      // now the map Space Element is filled for the BC
      for(std::map<partDomain*,std::vector<MElement*> >::iterator it=mapfind.begin(); it!=mapfind.end(); ++it){
        //nonLinearNeumannBC *neu = new nonLinearNeumannBC(allNeumann[i]);
        groupOfElements *neug = new groupOfElements(it->second);
        FunctionSpaceBase *spnbc = it->first->getSpaceForBC(nonLinearBoundaryCondition::NEUMANN,itbc->onWhat, itbc->_NeumannBCType,
                                                            itbc->_comp,neug);
        QuadratureBase *integ = it->first->getQuadratureRulesForNeumannBC(*itbc);
        itbc->_vspace.push_back(spnbc);
        itbc->_vgroup.push_back(neug);
        itbc->_vquadrature.push_back(integ);
        itbc->_vdom.push_back(it->first);
      }
    }
  }

  for(std::list<initialCondition>::iterator itic = allinitial.begin(); itic != allinitial.end();++itic){
    // first look if the BC is not applied on whole domain
    bool domfind=false;
    for(std::vector<partDomain*>::iterator itdom=domainVector.begin(); itdom!=domainVector.end(); ++itdom){
      partDomain *dom = *itdom;
      if(itic->_tag == dom->getPhysical()){
        domfind = true;
        if(dom->_groupGhostMPI->size() > 0) // save BC to applied an other groupOfElement
        {
          FunctionSpaceBase *spdbc = dom->getSpaceForBC(nonLinearBoundaryCondition::INITIAL,itic->onWhat, nonLinearNeumannBC::NEUMANN_UNDEF,
                                                        itic->_comp,dom->_groupGhostMPI);
          FilterDof *fdof = dom->createFilterDof(itic->_comp);
          itic->_vspace.push_back(spdbc);
          itic->_vgroup.push_back(dom->_groupGhostMPI);
          itic->_vfilter.push_back(fdof);
          itic->_vdg.push_back(dom->getFormulation());
        }
        FunctionSpaceBase *spdbc = dom->getSpaceForBC(nonLinearBoundaryCondition::INITIAL,itic->onWhat, nonLinearNeumannBC::NEUMANN_UNDEF,
                                                      itic->_comp,dom->elements());
        FilterDof *fdof = dom->createFilterDof(itic->_comp);
        itic->_vspace.push_back(spdbc);
        itic->_vgroup.push_back(dom->elements());
        itic->_vfilter.push_back(fdof);
        itic->_vdg.push_back(dom->getFormulation());
        break;
      }
    }
    #if defined(HAVE_MPI)
    // same on ghost domain (be aware copy past of loop on domain !!)
    bool ghostdomfind=false;
    for(std::vector<partDomain*>::iterator itdom=_ghostDomainMPI.begin(); itdom!=_ghostDomainMPI.end(); ++itdom){
      partDomain *dom = *itdom;
      if(itic->_tag == dom->getPhysical() and dom->_groupGhostMPI->size()>0){
        ghostdomfind = true;
	    FunctionSpaceBase *spdbc = dom->getSpaceForBC(nonLinearBoundaryCondition::INITIAL,itic->onWhat, nonLinearNeumannBC::NEUMANN_UNDEF,
                                                        itic->_comp,dom->_groupGhostMPI);
	    FilterDof* fdof = dom->createFilterDof(itic->_comp);
        itic->_vspace.push_back(spdbc);
        itic->_vgroup.push_back(dom->_groupGhostMPI);
        itic->_vfilter.push_back(fdof);
        itic->_vdg.push_back(dom->getFormulation());
        break;
      }
    }
    #endif // HAVE_MPI
    if(!domfind){
      // loop on element
      // map
      std::map<partDomain*,std::vector<MElement*> > mapfind;
      std::map<partDomain*,std::vector<MElement*> > mapfindGhost;
      for(groupOfElements::elementContainer::iterator it=itic->g->begin(); it!=itic->g->end();++it){
        MElement *e = *it;
        // Loop on all Domain to find on which domain the BC is applied
        for(std::vector<partDomain*>::iterator itdom=domainVector.begin(); itdom!=domainVector.end(); ++itdom){
          // if Dimensions are the same the element can be find with a comparison to element
          partDomain *dom = *itdom;
          // otherwise the vertex must be identified separately
          MElement *etest = *(dom->g_cbegin());
          if( (etest !=NULL) and (etest->getDim() == e->getDim())){
            for(groupOfElements::elementContainer::const_iterator it2=dom->g_cbegin(); it2!=dom->g_cend(); ++it2){
              if(*it2 == e){ // The Element is contained in the domain
                mapfind[dom].push_back(e);
                break;
              }
            }
            #if defined(HAVE_MPI)
            for(groupOfElements::elementContainer::iterator it2=dom->_groupGhostMPI->begin(); it2!=dom->_groupGhostMPI->end(); ++it2){
              if(*it2 == e){ // The Element is contained in the domain
                mapfind[dom].push_back(e);
                break;
              }
            }
            #endif // HAVE_MPI
          }
          else{
                 // an element is included in one domain --> one vertex of Element is OK
                 // On interface the BC is put on the first met domain
            fillMapOfInterfaceElementsInOneDomain(e, mapfind[(*itdom)], dom->elements());
            #if defined(HAVE_MPI)
            // change map in mapFindGhost???
            fillMapOfInterfaceElementsInOneDomain(e, mapfind[(*itdom)], dom->_groupGhostMPI);
            // check on element which are not included in _groupGhostMPI as there is no interface element with the element
            for(int j=0;j<Msg::GetCommSize();j++)
            {
              if(j!=Msg::GetCommRank() and !(*itdom)->getFormulation() ) // no ghost with itself
                fillMapOfInterfaceElementsInOneDomain(e, mapfindGhost[(*itdom)], &(_mapOtherDomain[(*itdom)->getPhysical()][j]));
            }

            #endif // HAVE_MPI
          }
        }
      }
      // now the space element map of the BC is filled
      for(std::map<partDomain*,std::vector<MElement*> >::iterator it=mapfind.begin(); it!=mapfind.end(); ++it){
        if(it->second.size() !=0){
          groupOfElements* dirig = new groupOfElements(it->second);
          FunctionSpaceBase *spdbc = it->first->getSpaceForBC(nonLinearBoundaryCondition::INITIAL,itic->onWhat, nonLinearNeumannBC::NEUMANN_UNDEF,
                                                              itic->_comp,dirig);
          FilterDof *fdof = it->first->createFilterDof(itic->_comp);
          itic->_vspace.push_back(spdbc);
          itic->_vgroup.push_back(dirig);
          itic->_vfilter.push_back(fdof);
          itic->_vdg.push_back(it->first->getFormulation());
        }
      }
      // same with mapfindGhost
      #if defined(HAVE_MPI)
      for(std::map<partDomain*,std::vector<MElement*> >::iterator it=mapfindGhost.begin(); it!=mapfindGhost.end(); ++it){
        if(it->second.size() !=0){
          groupOfElements* dirig = new groupOfElements(it->second);
          FunctionSpaceBase *spdbc = it->first->getSpaceForBC(nonLinearBoundaryCondition::INITIAL,itic->onWhat, nonLinearNeumannBC::NEUMANN_UNDEF,
                                                              itic->_comp,dirig);
          FilterDof *fdof = it->first->createFilterDof(itic->_comp);
          itic->_vspace.push_back(spdbc);
          itic->_vgroup.push_back(dirig);
          itic->_vfilter.push_back(fdof);
          itic->_vdg.push_back(it->first->getFormulation());
        }
      }
      #endif // HAVE_MPI
    }

    #if defined(HAVE_MPI)
    // same on ghostDomain (be aware copy past of loop on domain !!)
    if(!ghostdomfind){
      // loop on element
      // map
      std::map<partDomain*,std::vector<MElement*> > ghostmapfind;
      for(groupOfElements::elementContainer::iterator it=itic->g->begin(); it!=itic->g->end();++it){
        MElement *e = *it;
        // Loop on all Domain to find on which domain the BC is applied
        for(std::vector<partDomain*>::iterator itdom=_ghostDomainMPI.begin(); itdom!=_ghostDomainMPI.end(); ++itdom){
          // if Dimensions are the same the element can be find with a comparison to element
          partDomain *dom = *itdom;
          // otherwise the vertex must be identified separately
          const MElement *etest = *(dom->g_cbegin());
          if( (etest !=NULL) and (etest->getDim() == e->getDim())){
            for(groupOfElements::elementContainer::iterator it2=dom->_groupGhostMPI->begin(); it2!=dom->_groupGhostMPI->end(); ++it2){
              if(*it2 == e){ // The Element is contained in the domain
                ghostmapfind[dom].push_back(e);
                break;
              }
            }
          }
          else
          {
                 // an element is included in one domain --> one vertex of Element is OK
                 // On interface the BC is put on the first met domain
            fillMapOfInterfaceElementsInOneDomain(e, ghostmapfind[(*itdom)], dom->_groupGhostMPI);
          }
        }
      }
      // now the space element of the BC is filled
      for(std::map<partDomain*,std::vector<MElement*> >::iterator it=ghostmapfind.begin(); it!=ghostmapfind.end(); ++it){
        if(it->second.size() !=0){
          groupOfElements *dirig = new groupOfElements(it->second);
          FunctionSpaceBase *spdbc = it->first->getSpaceForBC(nonLinearBoundaryCondition::INITIAL,itic->onWhat, nonLinearNeumannBC::NEUMANN_UNDEF,
                                                        itic->_comp,dirig);
          FilterDof* fdof = it->first->createFilterDof(itic->_comp);
          itic->_vspace.push_back(spdbc);
          itic->_vgroup.push_back(dirig);
          itic->_vfilter.push_back(fdof);
          itic->_vdg.push_back(it->first->getFormulation());
        }
      }
    }
    #endif // HAVE_MPI
  }
#endif // 0

  // put back the curent GModel
  GModel::setCurrent(save_current_model);

};


void nonLinearMechSolver::setTimeForBC(double time){
  for(std::list<nonLinearDirichletBC>::iterator it = allDirichlet.begin(); it != allDirichlet.end();++it)
  {
    it->_f->setTime(time);
  }

  for(std::list<nonLinearNeumannBC>::iterator it = allNeumann.begin(); it!= allNeumann.end();++it)
  {
    it->_f->setTime(time);
  }

  // set time for rigid contact BC
  for(std::list<rigidContactBC>::iterator it = allContactBC.begin(); it!=allContactBC.end(); ++it){
    it->_f->setTime(time);
  }

  if (_testFlag){
    for (std::list<nonLinearDirichletBCAtCorner>::iterator it = allCornerConstraint.begin(); it != allCornerConstraint.end(); it++){
      it->_f->setTime(time);
    }

    for (std::list<nonLinearNeumannBCAtCorner>::iterator it = allCornerForce.begin(); it != allCornerForce.end(); it++){
      it->_f->setTime(time);
    }
  }

	// for periodic BC
  this->setTimeForMicroBC(time);

}

void nonLinearMechSolver::setTimeForLaw(const double t,const double dt, const int numstep)
{
  for(std::map<int,materialLaw*>::iterator it=maplaw.begin(); it!=maplaw.end();++it)
  {
    if (_microFlag){
      double mtime = _macroTime- (1-t)*_macroTimeStep;
      double tstep = dt*_macroTimeStep; // because
      it->second->setTime(mtime,tstep);
    }
    else{
      it->second->setTime(t,dt);
    }
  }
  // if multiscale
  _ipf->getAips()->setTime(t,dt,numstep);
}

void nonLinearMechSolver::fixNodalDofs(){
  // Fixation (prescribed displacement)
  dofManager<double>* currentManager = NULL;
  for(std::list<nonLinearDirichletBC>::iterator it = allDirichlet.begin(); it!=allDirichlet.end();++it ){
    nonLinearDirichletBC &diri = *it;
    // select the system if required
    if(_vcompBySys.size()==0)
    {
      currentManager = pAssembler;
    }
    else
    {
      dofManagerMultiSystems<double>* multiManager = static_cast<dofManagerMultiSystems<double>*>(pAssembler);
      currentManager = multiManager->getDofManagerByComp(diri._comp);
    }
    if(diri._mycondition==nonLinearBoundaryCondition::position)
    {
      for(int j=0; j<diri._vspace.size();j++)
      {
        FunctionSpaceBase *spdiri = diri._vspace[j];
        const groupOfElements *gdiri = diri._vgroup[j];
        FilterDof* fdofdiri = diri._vfilter[j];
        bool fdg = diri._vdg[j];
        FixNodalDofs(spdiri,gdiri->begin(),gdiri->end(),*currentManager,*diri._f,*(fdofdiri),fdg);
      }
    }
    else
    {
      for(int j=0;j<diri._vspace.size();j++){
        // allow to prescribed a velocity or acceleration over time using a dirichlet BC
        SetInitialDofs(diri._vspace[j],diri._vgroup[j]->begin(),diri._vgroup[j]->end(),diri._mycondition,*currentManager,diri._f,*(diri._vfilter[j]),diri._vdg[j]);
      }
    }
  }

  // prescribed displacement of rigid bodies (CG)
  //for(rigidContactBC &rcbc : allContactBC){
  for(std::list<rigidContactBC>::iterator rcbc = allContactBC.begin(); rcbc!=allContactBC.end(); ++rcbc){
    // select the system if required
    if(_vcompBySys.size()==0)
    {
      currentManager = pAssembler;
    }
    else
    {
      dofManagerMultiSystems<double>* multiManager = static_cast<dofManagerMultiSystems<double>*>(pAssembler);
      currentManager = multiManager->getDofManagerByComp(rcbc->_comp);
    }
    FixNodalDofs(rcbc->space,*rcbc->_f,*(rcbc->_filter),*pAssembler);
  }


  for (int i =0; i< domainVector.size(); i++){
    partDomain* dom = domainVector[i];
    if (dom->IsMultipleFieldFormulation()){
      dgPartDomain* dgdom = dynamic_cast<dgPartDomain*>(dom);
      const std::vector<int>& fixComp = dgdom->getFixedInterfaceDofComponent();
      for (int j=0; j< fixComp.size(); j++){
        FilterDof* filter = dgdom->createFilterDof(fixComp[j]);
        FixInterfaceNodalDofs(dgdom->getInterfaceFunctionSpace(),dgdom->gi->begin(), dgdom->gi->end(),*pAssembler,*filter);
        delete filter;
      }

    }

  }
}

void nonLinearMechSolver::numberDofs(){
  // we number the dofs : when a dof is numbered, it cannot be numbered
  // again with another number.

  for(std::vector<partDomain*>::iterator itdom = domainVector.begin(); itdom!=domainVector.end(); ++itdom)
  {
    partDomain *dom = *itdom;
    if(dom->groupOfElementsSize() > 0){
      NumberDofsByVector(*(dom->getFunctionSpace()), dom->g_cbegin(), dom->g_cend(),*pAssembler);
    }
   #if defined(HAVE_MPI)
    // create the mpi communication on interface domain
    // no element in g --> interfaceDomain
    else if(Msg::GetCommSize()>1){
      dgPartDomain *dgdom = dynamic_cast<dgPartDomain*>(dom);
      if (dgdom){
        staticDofManager<double>* sAssembler = static_cast<staticDofManager<double>*>(pAssembler);
        NumberInterfaceDofsMPI(dom->getFunctionSpace(), dgdom->gi->begin(), dgdom->gi->end(),sAssembler);
      }
    }
   #endif // HAVE_MPI

    // number extra  DOfs when considering twofield formulation
   if (dom->IsMultipleFieldFormulation()){
    dgPartDomain *dgdom = dynamic_cast<dgPartDomain*>(dom);
    if(dgdom->gi->size() > 0){
      NumberDofsByVector(*(dgdom->getInterfaceFunctionSpace()), dgdom->gi->begin(), dgdom->gi->end(),*pAssembler);
    }
   }

  }
  if(whatScheme!=StaticLinear and whatScheme != Eigen)
  {
    staticDofManager<double>* sAssembler = static_cast<staticDofManager<double>*>(pAssembler);
    sAssembler->setFirstRigidContactUnknowns();
  }

  // NumberDofs of Rigid Contact Entities (Dofs for GC)
  // select the system if required
  dofManager<double>* currentManager = NULL;
  if(_vcompBySys.size()==0)
  {
    currentManager = pAssembler;
  }
  else
  {
    dofManagerMultiSystems<double>* multiManager = static_cast<dofManagerMultiSystems<double>*>(pAssembler);
    currentManager = multiManager->getManager(0);
    if(_allContact.size()>0)
      Msg::Warning("Contact with a multiSystemDofManager --> Do the hypothesis that displacement is the first system");
  }
  for(contactContainer::iterator it = _allContact.begin(); it!=_allContact.end(); ++it){
    contactDomain* cdom = *it;
    if(cdom->isRigid()){
      rigidContactSpaceBase *rcspace = static_cast<rigidContactSpaceBase*>(cdom->getSpace());
      NumberDofs(*rcspace,*currentManager);
    }
  }
  //HERE DO WE NEED SOMETHING FOR DEFODEFO? SHOULD NOT BE BECAUSE IT S FOR CG

  if (_testFlag and (_pbcGroup != NULL)){
    // number new dof coming from microBC if existing
    _pbcGroup->numberDof_virtualVertices(pAssembler);
  }

  if(_vcompBySys.size()==0)
  {
    // total number of unkowns to allocate the system
    int nunk = pAssembler->sizeOfR();

		if (_pathFollowing){
			// if path following and one rhs resolution
			if (_pathFollowingMethod == GLOBAL_ARC_LENGTH_BASED){

			}
			else if (_pathFollowingMethod == LOCAL_BASED){
				// add 1 DOF to total system as load factor unknown
				if (_solverTypePathFollowing == pathFollowingSystemBase::ONE_RHS){
					nunk++;
				}
			}
      else if (_pathFollowingMethod == HYPERELLIPTIC_BASED){

      }
		}
    // allocate system
    static std::string A("A");
    pAssembler->getLinearSystem(A)->allocate(nunk);

    if (whatScheme == Eigen and _eigenSolverType == DYNAMIC){
	    static std::string B("B");
    	pAssembler->getLinearSystem(B)->allocate(nunk);
    };
  }
  else
  {
    dofManagerMultiSystems<double>* multiManager = static_cast<dofManagerMultiSystems<double>*>(pAssembler);
    multiManager->allocateSystem();
  }


  // Try to fill the matrix sparsity if needed
  // this has to be done after the system allocation
  if(whatScheme == StaticNonLinear || whatScheme == Multi || whatScheme == Implicit)
  {
    staticDofManager<double>* sAssembler = static_cast<staticDofManager<double>*>(pAssembler);

    #if defined(HAVE_MPI)
    if (sAssembler->isMPIInit() == false)
      sAssembler->initMPI();
    #endif // HAVE_MPI

    for(int i=0; i<_mpiUserDom;++i)
    {
      partDomain *dom = domainVector[i];
      if(dom->groupOfElementsSize() > 0){
        SparsityDofs(*(dom->getFunctionSpace()), dom->g_cbegin(), dom->g_cend(),*sAssembler);
      }
      // account also interface
      dgPartDomain *dgdom = dynamic_cast<dgPartDomain*>(dom);
      if (dgdom){
        staticDofManager<double>* sAssembler = static_cast<staticDofManager<double>*>(pAssembler);
   #if defined(HAVE_MPI)
	if(Msg::GetCommSize()>1)
          SparsityInterfaceDofsMPI(dgdom->getFunctionSpace(), dgdom->gi->begin(), dgdom->gi->end(),sAssembler);
   #endif
        SparsityDofs(*(dgdom->getFunctionSpace()), dgdom->gi->begin(), dgdom->gi->end(),*sAssembler);
      }

      if (dom->IsMultipleFieldFormulation()){
        SparsityDofsInterface(*(dgdom->getInterfaceFunctionSpace()),*(dgdom->getFunctionSpaceMinus()),*(dgdom->getFunctionSpacePlus()),
          dgdom->gi->begin(), dgdom->gi->end(),*sAssembler);
      }

     }
   #if defined(HAVE_MPI)
     // MPI interface sparsity
    if(Msg::GetCommSize()>1){
      for(int i=_mpiUserDom; i<domainVector.size();++i)
      {
        dgPartDomain *dgdom = dynamic_cast<dgPartDomain*>(domainVector[i]);
        if (dgdom){
          staticDofManager<double>* sAssembler = static_cast<staticDofManager<double>*>(pAssembler);
          SparsityInterfaceDofsMPI(dgdom->getFunctionSpace(), dgdom->gi->begin(), dgdom->gi->end(),sAssembler);
        }
      }
    }
   #endif // HAVE_MPI
    if(_vcompBySys.size()==0)
    {
      // allocate system
      static std::string A("A");
      pAssembler->getLinearSystem(A)->preAllocateEntries();

      if (whatScheme == Eigen){
	      static std::string B("B");
    	  pAssembler->getLinearSystem(B)->preAllocateEntries();
      };
    }
    else
    {
      dofManagerMultiSystems<double>* multiManager = static_cast<dofManagerMultiSystems<double>*>(pAssembler);
      multiManager->preAllocateEntries();
    }
  }
  else if (whatScheme == StaticLinear){
    // sparsity when considering linear problem
    for(int i=0; i<domainVector.size();++i)
    {
      partDomain *dom = domainVector[i];
      if(dom->groupOfElementsSize() > 0){
        SparsityDofs(*(dom->getFunctionSpace()), dom->g_cbegin(), dom->g_cend(),*pAssembler);
      }
      // account also interface
      dgPartDomain *dgdom = dynamic_cast<dgPartDomain*>(dom);
      if (dgdom){
         SparsityDofs(*(dgdom->getFunctionSpace()), dgdom->gi->begin(), dgdom->gi->end(),*pAssembler);
        if (dom->IsMultipleFieldFormulation()){
          SparsityDofsInterface(*(dgdom->getInterfaceFunctionSpace()),*(dgdom->getFunctionSpaceMinus()),*(dgdom->getFunctionSpacePlus()),
            dgdom->gi->begin(), dgdom->gi->end(),*pAssembler);
        }

      }
    }
  }


  //
  if (_pathFollowing and (_pathFollowingMethod == HYPERELLIPTIC_BASED)){
    static std::string A("A");
    pathFollowingSystemBase* pfsysBase = dynamic_cast<pathFollowingSystemBase*>(pAssembler->getLinearSystem(A));
    if (pfsysBase != NULL){
      for (int i=0; i< _hyperellipticControlComp.size(); i++){

        int comp = _hyperellipticControlComp[i];
        for(std::vector<partDomain*>::iterator itdom = domainVector.begin(); itdom!=domainVector.end(); ++itdom)
        {
          partDomain *dom = *itdom;
          if (dom->getAccountPathFollowingFlag()){
            FunctionSpaceBase* space = dom->getFunctionSpace();
            FilterDof* filter = dom->createFilterDof(comp);
            for (groupOfElements::elementContainer::const_iterator ite = dom->g_cbegin(); ite != dom->g_cend(); ite++){
              MElement* ele = *ite;
              std::vector<Dof> keys;
              space->getKeys(ele,keys);
              for (int k=0; k< keys.size(); k++){
                if ((pAssembler->isAnUnknown(keys[k])) and (filter->operator ()(keys[k]))){
                  int num = pAssembler->getDofNumber(keys[k]);
                  if (num >=0){
                    pfsysBase->addToHyperellipticMatrix(num,num,1.);
                  }
                }
              }
            }
            delete filter;
          }
        }

      }
    }
    else{
      Msg::Error("path following system must be used");
    }
  };

  return;
}

void nonLinearMechSolver::addDomain(partDomain* dom){
  pModel->setAsCurrent();
  // initialization of GroupOfElements
  int dim = dom->getDim();
  int phys = dom->getPhysical();
#if defined(HAVE_MPI)
  std::vector<groupOfElements> groupOtherPart(Msg::GetCommSize());
  elementFilterMPI elemfil(&groupOtherPart) ;
  dom->addPhysicalGroup(dim,phys,elemfil);
  // add to map (to construct later interface domain)
  _mapOtherDomain.insert(std::pair<int,std::vector<groupOfElements> >(phys,groupOtherPart));
  if (_microFlag == false){
    if (dom->isDistributedOtherRanks()){
      // if multiscale material is used
      // at root ranks
      if (_mapRanks.find(Msg::GetCommRank()) != _mapRanks.end()){
        int addFlag = 0;
        if(dom->groupOfElementsSize()!= 0){ // otherwise the domain is include totally on an other partition
          dom->setWorkingRanks(Msg::GetCommRank(),_mapRanks[Msg::GetCommRank()]);
          domainVector.push_back(dom);
          addFlag= 1;
        }
        else if(dom->IsInterfaceTerms()){ // take into account for pure interface domain which have dom->g.size()=0
          dgPartDomain *dgdom = static_cast<dgPartDomain*>(dom);
          if(dgdom->getMinusDomain()->getPhysical() != dgdom->getPlusDomain()->getPhysical()){ // pure interface --> add
            dom->setWorkingRanks(Msg::GetCommRank(),_mapRanks[Msg::GetCommRank()]);
            domainVector.push_back(dom);
            addFlag = 2;
          }
          else{ // add the domain in ghostDomain vector to prescribed Dirichlet BC (applied on all cpu)
            _ghostDomainMPI.push_back(dom);
            addFlag = 3;
          }
        }

        std::set<int>& otherRanks = _mapRanks[Msg::GetCommRank()];
        for (std::set<int>::iterator its = otherRanks.begin(); its != otherRanks.end(); its++)
          MPI_Send(&addFlag,1,MPI_INT,*its,*its,MPI_COMM_WORLD);

      }
      // in other ranks
      else{
        for (std::map<int,std::set<int> >::iterator it = _mapRanks.begin(); it!= _mapRanks.end(); it++){
          std::set<int>& other = it->second;
          if (other.find(Msg::GetCommRank()) != other.end()){
            int rank = it->first;
            int addFlag = 0;
            MPI_Status status;
            MPI_Recv(&addFlag,1,MPI_INT,rank,Msg::GetCommRank(),MPI_COMM_WORLD,&status);
            if (addFlag == 1 or addFlag == 2){
              domainVector.push_back(dom);
              dom->setWorkingRanks(rank,other);
              dom->clearGroupOfElements();
            }
          }
        }
      }
    }
    else{
      // if normal Domain is used
      if (_mapRanks.find(Msg::GetCommRank()) != _mapRanks.end()){
        std::set<int> nullSet;
        if(dom->groupOfElementsSize() != 0){ // otherwise the domain is include totally on an other partition
          dom->setWorkingRanks(Msg::GetCommRank(),nullSet);
          domainVector.push_back(dom);
        }
        else if(dom->IsInterfaceTerms()){ // take into account for pure interface domain which have dom->g.size()=0
          dgPartDomain *dgdom = static_cast<dgPartDomain*>(dom);
          if(dgdom->getMinusDomain()->getPhysical() != dgdom->getPlusDomain()->getPhysical()){ // pure interface --> add
            dom->setWorkingRanks(Msg::GetCommRank(),nullSet);
            domainVector.push_back(dom);
          }
          else{ // add the domain in ghostDomain vector to prescribed Dirichlet BC (applied on all cpu)
            _ghostDomainMPI.push_back(dom);
          }
        }
      }

    }
  }
  else {
    if(dom->groupOfElementsSize() != 0){ // otherwise the domain is include totally on an other partition
      domainVector.push_back(dom);
    }
    else if(dom->IsInterfaceTerms()){ // take into account for pure interface domain which have dom->g.size()=0
      dgPartDomain *dgdom = static_cast<dgPartDomain*>(dom);
      if(dgdom->getMinusDomain()->getPhysical() != dgdom->getPlusDomain()->getPhysical()){ // pure interface --> add
        domainVector.push_back(dom);
      }
      else{ // add the domain in ghostDomain vector to prescribed Dirichlet BC (applied on all cpu)
        _ghostDomainMPI.push_back(dom);
      }
    }
  }


#else
  elementFilterTrivial elemfil;
  dom->addPhysicalGroup(dim,phys,elemfil);
  domainVector.push_back(dom);
#endif // HAVE_MPI
}

void nonLinearMechSolver::addMaterialLaw(materialLaw* mlaw){
  maplaw.insert(std::pair<int,materialLaw*>(mlaw->getNum(),mlaw));
}

materialLaw* nonLinearMechSolver::getMaterialLaw(const int num){
  return (maplaw.find(num)->second);
}

void nonLinearMechSolver::thetaBC(const int numphys){
  groupOfElements *goe = new groupOfElements(1,numphys);
  this->insertTheta(numphys,goe);
}
void nonLinearMechSolver::stiffnessModification(const bool flag){
  _stiffnessModification = flag;
};
void nonLinearMechSolver::iterativeProcedure(const bool flag){
  _iterativeNR = flag;
};
void nonLinearMechSolver::lineSearch(const bool lin){
  _lineSearch = lin;
}
void nonLinearMechSolver::pathFollowing(const bool p, const int method){
  _pathFollowing = p;
	_pathFollowingMethod = (pathFollowingMethod)method;
}

void nonLinearMechSolver::setPathFollowingIncrementAdaptation(const bool stepAdap, const int numNROptimal, const double exp){
  _numNROptimal = numNROptimal;
  _numNROptimalLocal = numNROptimal;
  _expNROptimal = exp;
  _timeStepAdaptation = stepAdap;
  if (_timeStepAdaptation) {
    Msg::Info("path following increment is adapted with number of NR iterations");
  }
};

void nonLinearMechSolver::setPathFollowingIncrementAdaptationLocal(const bool stepAdap, const int numNROptimalLoad, const int numNROptimalLocal, const double exp){
  _numNROptimal = numNROptimalLoad;
  _numNROptimalLocal = numNROptimalLocal;
  _expNROptimal = exp;
  _timeStepAdaptation = stepAdap;
  if (_timeStepAdaptation) {
     Msg::Info("path following increment is adapted with number of NR iterations");
  }
};

void nonLinearMechSolver::setPathFollowingControlType(const int i){
  _macroControlTypePathFollowing = i;
}
void nonLinearMechSolver::setPathFollowingCorrectionMethod(const int i){
	_correctionMethodPathFollowing = i;
};
void nonLinearMechSolver::setPathFollowingTranversalCriterion(const int i){
	_tranversalCriterionPathFollowing = i;
};
void nonLinearMechSolver::setPathFollowingSolverType( const int i, const double eqRatio){
	_solverTypePathFollowing = i;
	_pathFollowingEqRatio = eqRatio;
};

void nonLinearMechSolver::setPathFollowingLocalSteps(const double loadStep, const double localCrStep){
	_loadStep = loadStep;
	_loadStepPrev = loadStep;
	_localStep = localCrStep;
	_localStepPrev = localCrStep;
}

void nonLinearMechSolver::setPathFollowingArcLengthStep(const double arcStep){
	_arcLengthStep = arcStep;
	_arcLengthStepPrev = arcStep;
};

void nonLinearMechSolver::setBoundsOfPathFollowingLocalSteps(const double lowerBound, const double upperBound){
  _localStepMinimal = lowerBound;
  _localStepMaximal = upperBound;
  Msg::Info("path following local step is bounded by lowerbound %e and by upper bound %e",lowerBound,upperBound);
};

void nonLinearMechSolver::setBoundsOfPathFollowingArcLengthSteps(const double lowerBound, const double upperBound){
  _arcLengthStepMinimal = lowerBound;
  _arcLengthStepMaximal = upperBound;
  Msg::Info("path following local step is bounded by lowerbound %e and by upper bound %e",lowerBound,upperBound);
};

void nonLinearMechSolver::setPathFollowingLocalIncrementType(const int i){
  _pathFollowingIncrementType = (pathFollowingLocalIncrementType)i;
  if (i == 0) {
    Msg::Info("pf increment with defo energy");
  }
  else if (i == 1){
   Msg::Info("pf increment with dissipation energy");
  }
  else if (i == 2){
    Msg::Info("pf increment with plastic energy");
  }
  else if (i == 3){
    Msg::Info("pf increment with damage energy");
  }
  else{
    Msg::Error("missing case nonLinearMechSolver::setPathFollowingLocalIncrementType %d",i);
  }
};

void nonLinearMechSolver::setPathFollowingLocation(const int loc){
  _pathFollowingLocation = (pathFollowingLocation)loc;
  if (_pathFollowingLocation == BULK_INTERFACE){
    Msg::Info("path following constraint is contructed from bulk and interfaces elements");
  }
  else if (_pathFollowingLocation == BULK){
    Msg::Info("path following constraint is contructed from bulk elements");
  }
  else if (_pathFollowingLocation == INTERFACE){
    Msg::Info("path following constraint is contructed from interfaces elements");
  }
  else{
    Msg::Error("case missing %d nonLinearMechSolver::setPathFollowingLocation ",loc);
  }

};

void nonLinearMechSolver::setPathFollowingSwitchCriterion(const double cr){
  _pathFollowingSwitchCriterion = cr;
};

void nonLinearMechSolver::clearAllHyperellipticControlComp(){
  _hyperellipticControlComp.resize(0);
};

void nonLinearMechSolver::setHyperellipticControlComp(const int comp){
  _hyperellipticControlComp.push_back(comp);
  Msg::Info("path following with comp %d",comp);
};

/*
void nonLinearMechSolver::setInitOrRestartFileName(const std::string fname){
  initOrResartfname =fname;
  _restart = true;
}

void nonLinearMechSolver::restart(unknownField *ufield){
  if(!_restart) // no file to restart
    return;
  // create a map between elem number and adress
  std::map<long int,std::pair<partDomain*,MElement*> > allelem;
  for(std::vector<partDomain*>::iterator itdom = domainVector.begin(); itdom!=domainVector.end(); ++itdom){
    partDomain *dom = *itdom;
    for(groupOfElements::elementContainer::iterator ite = dom->g->begin(); ite!=dom->g->end(); ++ite){
      MElement *ele = *ite;
      std::pair<partDomain*,MElement*> pa(dom,ele);
      allelem.insert(std::pair<long int,std::pair<partDomain*,MElement*> >(ele->getNum(),pa));
    }
  }

  // read data in file
  FILE *f = fopen(initOrResartfname.c_str(), "r");
  char what[256];
  while(!feof(f)){
    fscanf(f, "%s", what);
    if(!strcmp(what,"$ElementNodeData")){
      int nstring;
      fscanf(f,"%d",&nstring);
      int i=0;
      char what2[256];
      while(i<nstring){
        fscanf(f,"%s",&what2);
        i++;
      }
      int nint, n;
      fscanf(f,"%d",&nint);
      i=0;
      while(i<nint){
        fscanf(f,"%d",&n);
        i++;
      }
      fscanf(f,"%d",&nint);
      if(nint != 4)
      {
        Msg::Error("Your restart file hasn't the good format since the number of int (!=4) is not good. The number of element is taken from the third one!");
      }
      int nelem;
      i=0;
      while(i<nint){
        fscanf(f,"%d",&n);
        if(i == 2)
          nelem = n;
        i++;
      }
      int elemnum,elemtype,ndofs,nvertex,ncomp;
      i = 0;
      int n2 = nelem; // because n is modified by fscanf(f,"%ld %d",&elemnum,&elemtype); WHY ??
      while(i<n2){
        fscanf(f,"%d %d",&elemnum,&elemtype);
        //MElement *ele = allelem[elemnum].second;
        //partDomain *dom = allelem[elem]
        std::map<long int,std::pair<partDomain*,MElement*> >::iterator it = allelem.find(elemnum);
        MElement *ele = (it->second).second;
        partDomain *dom = (it->second).first;
        FunctionSpaceBase *space = dom->getFunctionSpace();
        std::vector<Dof> R;
        space->getKeys(ele,R);
        ndofs = space->getNumKeys(ele);
        nvertex = ele->getNumVertices();
        ncomp = ndofs/nvertex;
        int j=0;
        std::vector<double> disp;
        disp.resize(ndofs);
        //double temp;
        while(j<nvertex){
          for(int k=0;k<ncomp;k++){
            fscanf(f,"%lf",&disp[j+k*nvertex]);
          }
          j++;
        }
        // set displacement in displacement field
        ufield->setInitial(R,disp);
        i++;
      }
    }
  }
  fclose(f);
  ufield->buildAllView(domainVector,0.,0);
}
*/
void nonLinearMechSolver::insertTheta(const int numphys, groupOfElements *goe){
  allTheta.emplace_back();
  nonLinearNeumannBC& neu = allTheta.back();
  neu.g = goe;
  neu.onWhat = nonLinearBoundaryCondition::UNDEF;
  neu._tag = numphys;
  neu._f = new simpleFunctionTimeWithConstant<double>(0.);
}

void nonLinearMechSolver::physInitBroken(const int numphys){
  initbrokeninter.push_back(numphys);
}

void nonLinearMechSolver::createInterfaceElement(){
  // Compute and add interface element to the model
  // Loop on mesh element and store each edge (which will be distributed between InterfaceElement, Boundary InterfaceElement and VirtualInterfaceElement)
  // A tag including the vertex number is used to identified the edge common to two elements. In this case a interface element is created
  // manage creation of interface between 2 domains
  pModel->setAsCurrent();
  manageInterface maninter(&domainVector);
  // loop on element
  for(std::vector<partDomain*>::iterator itdom=domainVector.begin(); itdom !=domainVector.end(); ++itdom)
  {
    partDomain *dom = *itdom;
    dom->createInterface(maninter);

   // remove interface element if domain is not DG ??
   // there is other way do that
    dgPartDomain* dgdom = dynamic_cast<dgPartDomain*>(dom);
    if (dgdom){
      if (!dgdom->IsInterfaceTerms()){
        dgdom->gi->clearAll();
      }
    }
  }


  // virtual interface
  for(std::vector<partDomain*>::iterator itdom = domainVector.begin(); itdom!=domainVector.end(); ++itdom){
    partDomain *dom = *itdom;
    int phys = dom->getPhysical();
    for(manageInterface::IelementContainer::iterator it=maninter.begin(); it!=maninter.end(); ++it){
      IElement *ie = it->second;
      if(ie->getPhys() == phys){
        dgPartDomain *dgdom = static_cast<dgPartDomain*>(dom);
        MElement* interel = dom->createVirtualInterface(ie);
        dgdom->giv->insert(interel);
      }
    }
  }
  _mpiUserDom = domainVector.size();


#if defined(HAVE_MPI)
  if(Msg::GetCommSize() > 1 and (_mapRanks.find(Msg::GetCommRank()) != _mapRanks.end())){
    // first add interface to the interdomain defined by the user
    for(std::vector<partDomain*>::iterator itdom = domainVector.begin(); itdom != domainVector.end(); ++itdom){
      // find interdomain (g->size==0)
      partDomain *dom = *itdom;
      if(dom->groupOfElementsSize()==0){
        int phys1,phys2;
        dgPartDomain *dgdom = static_cast<dgPartDomain*>(dom);
        manageInterface::getPhysDom(dom->getPhysical(),phys1,phys2);
        // find both domains
        partDomain *dom1=NULL;
        partDomain *dom2=NULL;
        for(std::vector<partDomain*>::iterator itdom = domainVector.begin(); itdom != domainVector.end(); ++itdom){
          partDomain* domtmp = *itdom;
          if(domtmp->getPhysical() == phys1) dom1=domtmp;
          else if(domtmp->getPhysical() == phys2) dom2=domtmp;
          if((dom1!=NULL) and(dom2!=NULL)) break;
        }
        if(dom1!=NULL){ // otherwise no element of dom1 on the partition
          // find the group of ghosted element of dom2
          std::map<int,std::vector<groupOfElements> >::iterator itg= _mapOtherDomain.find(phys2);
          if(itg!=_mapOtherDomain.end()){
            for(int i=0; i< itg->second.size(); i++){
              groupOfElements &group = itg->second[i];
              if(group.size()!=0)
                dom1->createInterfaceMPI(i,group,dgdom);
            }
          }
        }
        if(dom2!=NULL){ // otherwise no element of dom1 on partition
          // find the group of ghosted element of dom1
          std::map<int,std::vector<groupOfElements> >::iterator itg= _mapOtherDomain.find(phys1);
          if(itg!=_mapOtherDomain.end()){
            for(int i=0; i< itg->second.size(); i++){
              groupOfElements &group = itg->second[i];
              if(group.size()!=0)
                dom2->createInterfaceMPI(i,group,dgdom);
            }
          }
        }
      }
    }

    // create pure interface domain for MPI
    for(std::map<int,std::vector<groupOfElements> >::iterator it=_mapOtherDomain.begin(); it!=_mapOtherDomain.end(); ++it){
      int physdom = it->first;
      // get the domain
      partDomain *dom = NULL;
      for(std::vector<partDomain*>::iterator itdom = domainVector.begin(); itdom != domainVector.end(); ++itdom){
        if(physdom == (*itdom)->getPhysical()){
          dom = *itdom;
          break;
        }
      }
      if(dom != NULL){
        std::vector<groupOfElements> & groupOtherPart = it->second;
        dgPartDomain* newInterDom = NULL;
        int gsize = groupOtherPart.size();
        for(int i=0; i< gsize; i++){
          if(i!=Msg::GetCommRank()){
            groupOfElements &group = groupOtherPart[i];
            if(group.size() != 0){ // otherwise no domain to create
              dom->createInterfaceMPI(i,group,newInterDom);
            }
          }
        }
        if((newInterDom!=NULL) and (newInterDom->gi->size() > 0)){
          // law are set before (change this ??)
          newInterDom->setMaterialLaw(maplaw);
          domainVector.push_back(newInterDom); // otherwise no interface element
        }
        else{
         if(newInterDom!=NULL) delete newInterDom;
        }
      }
    }
  }
#endif // HAVE_MPI
}

void nonLinearMechSolver::createInterfaceElement_2()
{
  // The contructor of dgGroupCollection create automatically the GroupsOfElements
/*  _groups = dgGroupCollection(this->pModel,this->_dim,1); // TODO ?? Add parameter to model to store order
  _groups.buildGroupsOfInterfaces();
  // Affichage temporaire pour vérification
  int nn = _groups.getNbFaceGroups();
  printf("Number of group of faces : %d\n",nn);
  for(int i=0;i<nn;i++){
    printf("Group of face number %d\n",i);
    dgGroupOfFaces *inter = _groups.getFaceGroup(i);
    int nnn = inter->getNbGroupOfConnections();
    printf("Number of connection group %d \n",nnn);
    for(int j=0;j<nnn;j++){
      const dgGroupOfConnections connec = inter->getGroupOfConnections(j);
      printf("Connection's group number %d\n",j);
      int nnnn = connec.getNbClosures();
      printf("Number of closures %d\n",nnnn);
      for(int k=0;k<nnnn;k++){
        printf("Closure number %d\n",k);
        std::vector<int> vec = connec.getClosure(k);
        for(int kk=0;kk<vec.size();kk++){
          printf(" %d ",vec[kk]);
        }
        printf("\n");
      }
    }
  }*/
}

void nonLinearMechSolver::rotateModel(const SVector3& n1, const SVector3& n2, const SVector3& n3){
  if (!_GModelIsRotated){
    _GModelIsRotated = true;
    GeometryRotation::rotateGModel(_enum,_gnum,pModel,n1,n2,n3);
  }
};

int nonLinearMechSolver::getInterfaceElementNumber(const int em, const int ep) const{
  TwoNum tn(em,ep);
  std::map<TwoNum,int>::const_iterator itF = _interfaceElements.find(tn);
  if ( itF != _interfaceElements.end()){
    return itF->second;
  }
  else{
    Msg::Error("interface element with negative %d positive %d is not found nonLinearMechSolver::getInterfaceElementNumber!",em,ep);
  }
};

TwoNum nonLinearMechSolver::getMinusAndPlusElementNumber(const int elnum) const{
  std::map<int,TwoNum>::const_iterator itF = _interfaceElementsInverseMap.find(elnum);
  if (itF != _interfaceElementsInverseMap.end()){
    return itF->second;
  }
  else{
    #if defined (HAVE_MPI)
    if (Msg::GetCommSize() > 1)
      Msg::Error("interface element %d is not found nonLinearMechSolver::getMinusAndPlusElementNumber at rank %d",elnum,Msg::GetCommRank());
    else
    #endif // HAVE_MPI
      Msg::Error("interface element %d is not found nonLinearMechSolver::getMinusAndPlusElementNumber",elnum);
  }
};

void nonLinearMechSolver::initTerms(){
  // resize _vterm in NeumannBC
  if(_previousInit){ // if previous computation delete first the terms
    // BC
    for(std::list<nonLinearNeumannBC>::iterator itbc = allNeumann.begin(); itbc != allNeumann.end();++itbc)
    {
      for(int j=0;j<itbc->_vterm.size();j++)
      {
        delete itbc->_vterm[j];
      }
    }
    // domain
    for(std::vector<partDomain*>::iterator itdom = domainVector.begin(); itdom!=domainVector.end(); ++itdom){
      partDomain *dom = *itdom;
      if(dom->getBilinearBulkTerm() != NULL) delete dom->getBilinearBulkTerm();
      if(dom->getBilinearMassTerm() != NULL) delete dom->getBilinearMassTerm();
      if(dom->getLinearBulkTerm() != NULL) delete dom->getLinearBulkTerm();
      if(dom->IsInterfaceTerms()){
        dgPartDomain *dgdom = static_cast<dgPartDomain*>(dom);
        if(dgdom->getBilinearInterfaceTerm() != NULL) delete dgdom->getBilinearInterfaceTerm();
        if(dgdom->getBilinearVirtualInterfaceTerm() !=NULL) delete dgdom->getBilinearVirtualInterfaceTerm();
        if(dgdom->getLinearInterfaceTerm() != NULL) delete dgdom->getLinearInterfaceTerm();
        if(dgdom->getLinearVirtualInterfaceTerm() != NULL) delete dgdom->getLinearVirtualInterfaceTerm();
      }
    }
  }
  for(std::list<nonLinearNeumannBC>::iterator itbc = allNeumann.begin(); itbc != allNeumann.end();++itbc)
  {
    nonLinearNeumannBC &neu = *itbc;
    int nsize = neu._vspace.size();
    neu._vterm.resize(nsize);
    neu._vmatrix_term.resize(nsize);
    for(int j=0;j<nsize;++j)
    {
      neu._vterm[j] = neu._vdom[j]->createNeumannTerm(static_cast<FunctionSpace<double>*>(neu._vspace[j]),  neu._vgroup[j],
                                                      neu._f,_ufield, _ipf,neu.onWhat,neu._NeumannBCType, neu._comp);
      neu._vmatrix_term[j] = neu._vdom[j]->createNeumannMatrixTerm(static_cast<FunctionSpace<double>*>(neu._vspace[j]),  neu._vgroup[j],
                                                             neu._f,_ufield, _ipf,neu.onWhat,neu._NeumannBCType, neu._comp);
    }
  }

  for(std::vector<partDomain*>::iterator itdom = domainVector.begin(); itdom!=domainVector.end(); ++itdom){
    partDomain *dom = *itdom;
    dom->initializeTerms(_ufield,_ipf);
  }

  // contact domain
  for(contactContainer::iterator it = _allContact.begin(); it!=_allContact.end(); ++it){
    contactDomain *cdom = *it;
      cdom->initializeTerms(_ufield);
  }
  // contact domain
  for(defoDefoContactContainer::iterator it = _allDefoDefoContact.begin(); it!=_allDefoDefoContact.end(); ++it){
    defoDefoContactDomain *cdom = *it;
      cdom->initializeTerms(_ufield);
  }

}

double nonLinearMechSolver::solveStaticLinear()
{
  /* init data */
  #if defined(HAVE_MPI)
  if (Msg::GetCommSize() > 1)
  {
    Msg::Error("StaticLinear cannot run with multiple processors!!!");
  }
  #endif //HAVE_MPI
  
  this->init();
  this->init2();
  if (_energField != NULL){
    delete _energField; _energField=NULL; // energy cannot be computed with this scheme as the system does not computed the forces of fixed ddl!!
  }

  /* time initialization */
  double curtime = 1.;
  double timestep = 1.;
  int step = 1;

  /* solving */
  this->setTimeForLaw(curtime,timestep,step);
  std::cout <<  "Neumann BC"<< std::endl;
  double tsystresol = Cpu();
  this->computeExternalForces();
  this->computeStiffMatrix();
  printf("-- done assembling!\n");
  pAssembler->systemSolve();
  printf("-- done solving!\n");

  /* compute stress after solve */
  _ipf->compute1state(IPStateBase::current,true);
  /* end of scheme */
  tsystresol = Cpu() - tsystresol;
  Msg::Info("Time of resolution: %f",tsystresol);
  this->endOfScheme(curtime,step);
  Msg::Info("StaticLinear OK");
  return curtime;
}

double nonLinearMechSolver::solveEigen()
{
   /* init data */
   this->init();
   this->init2();
   if (_energField){
     delete _energField; _energField=NULL; // energy cannot be computed with this scheme as the system does not computed the forces of fixed ddl!!
   }

   /* time initialization */
   double curtime = 1.;
   double timestep = 1.;
   int step = 1;

   /* solving */
   this->setTimeForLaw(curtime,timestep,step);
   printf("--begin assembling \n");
   double tsystresol = Cpu();
   if (_eigenSolverType == DYNAMIC)
     this->computeMassMatrix();

   this->computeStiffMatrix();
   printf("-- done assembling!\n");

   #if defined (HAVE_SLEPC)
   this->createEigenvalueSolver();
   _eigenSolver->solve(_numeigenvalue,"smallestReal",_eigenSolverMethod,_eigenSolverConvergenCriterion,_eigenSolverNumIteration);
   tsystresol = Cpu() - tsystresol;
   Msg::Info("Time of Eigen Value resolution resolution: %f",tsystresol);


   if (domainVector.size() == 0)
    Msg::Error("no domain");
   const groupOfElements* g = domainVector[0]->elements();

   MElement* ele = *(g->begin());
   double scale = 1.;
   int d = ele->getDim();
   double vl = ele->getVolume();
   if (vl<=0 ){
     Msg::Error("element volume = 0");
   }
   else{
     if (d ==1) scale = vl;
     else if (d ==2) scale = sqrt(vl);
     else if (d==3) scale = pow(vl,1./3.);
     else Msg::Error("dim = %d not exist",d);
   };

  Msg::Info("scale eigenvector = %f",scale);
   _eigenSolver->normalize_mode(_modeview,scale);

  if (_MKToFile){
    #if defined(HAVE_PETSC)
     std::string Aname("A");
     linearSystem<double>* lsysA = pAssembler->getLinearSystem(Aname);
     std::string Bname("B");
     linearSystem<double>* lsysB = pAssembler->getLinearSystem(Bname);

     linearSystemPETSc<double>* Apet = dynamic_cast<linearSystemPETSc<double>*>(lsysA);
     if (Apet!= NULL) {
        functionPETSc::MatToFile(Apet->getMatrix(),"Mat_Stiffness.txt");

     linearSystemPETSc<double>* Bpet = dynamic_cast<linearSystemPETSc<double>*>(lsysB);
     if (Bpet!= NULL) {
        functionPETSc::MatToFile(Bpet->getMatrix(),"Mat_Mass.txt");
      };
    }

     #endif
  }

   for (int i=0; i<_numeigenvalue; i++){
     std::complex<double> val = _eigenSolver->getEigenValue(i);
     fprintf(_eigenValueFile,"%.16g  \t %.16g \n",val.real(), val.imag());
   }
   fprintf(_eigenValueFile,"\n");
   fflush(_eigenValueFile);

   this->modeViewToFile(_modeview,1.,1.);

   if (_isPerturbedEigenMode)
   		this->writeDisturbedMeshByEigenVector(0);
   #else
   Msg::Error("solveEigen::Slepc is not intalled");
   #endif // HAVE_SLEPC

   printf("-- done solving!\n");

   /* end of scheme */
   this->endOfScheme(curtime,step);
   Msg::Info("EigenSolve OK");
   return curtime;
}

void nonLinearMechSolver::computeLoadVector(){
  this->setTimeForBC(1.);
  staticDofManager<double> *sAssembler;
  for(std::list<nonLinearNeumannBC>::iterator itbc = allNeumann.begin(); itbc != allNeumann.end();++itbc)
  {
    nonLinearNeumannBC &neu = *itbc;
    // select the system if required
    if(_vcompBySys.size()==0)
    {
      sAssembler = static_cast<staticDofManager<double>*>(pAssembler);
    }
    else
    {
      dofManagerMultiSystems<double>* multiManager = static_cast<dofManagerMultiSystems<double>*>(pAssembler);
      sAssembler = static_cast<staticDofManager<double>*>(multiManager->getDofManagerByComp(neu._comp));
    }
    for(int j=0;j<neu._vspace.size();j++)
    {
      FunctionSpaceBase *spn = neu._vspace[j];
      const groupOfElements *gneu = neu._vgroup[j];
      QuadratureBase *qneu = neu._vquadrature[j];
      Assemble(neu._vterm[j],*spn,gneu->begin(),
             gneu->end(),*(qneu),_ufield,*sAssembler,nonLinearSystem<double>::LoadVector, _elementErosionFilter);
    }
  }

  //
  if (_testFlag){
    if (_pbcGroup != NULL){
      for (std::list<nonLinearNeumannBCAtCorner>::iterator itc = allCornerForce.begin(); itc != allCornerForce.end();++itc){
        nonLinearNeumannBCAtCorner& con = *itc;
        // select the system if required
        if(_vcompBySys.size()==0)
        {
          sAssembler = static_cast<staticDofManager<double>*>(pAssembler);
        }
        else
        {
          dofManagerMultiSystems<double>* multiManager = static_cast<dofManagerMultiSystems<double>*>(pAssembler);
          sAssembler = static_cast<staticDofManager<double>*>(multiManager->getDofManagerByComp(con._comp));
        }

        std::vector<Dof>& allDof = con.constraintDofs;
        int size = allDof.size();
        MVertex* v = _pbcGroup->getCornerVertex(con._cornerNumber);
        fullVector<double> F;
        F.resize(size,true);
        double val = con._f->operator ()(v->x(),v->y(),v->z());
        F.setAll(val);
        sAssembler->assemble(allDof,F,nonLinearSystem<double>::LoadVector);
      };

    };
  };
  this->setTimeForBC(0.);
}

void nonLinearMechSolver::computePathFollowingConstraint(){
	staticDofManager<double> *sAssembler = dynamic_cast<staticDofManager<double>*>(pAssembler);
	if (sAssembler == NULL) Msg::Error("staticDofManager must be used nonLinearMechSolver::computePathFollowingConstraint");

	for(std::vector<partDomain*>::iterator itdom=domainVector.begin(); itdom!=domainVector.end();++itdom){
    partDomain *dom = *itdom;
    if (dom->getMaterialLaw() != NULL){
      if (dom->getMaterialLaw()->withEnergyDissipation()){
        AssemblerPathFollowingConstraint(*(dom->getScalarTermPFBulk()),dom->g_cbegin(),dom->g_cend(),*(dom->getBulkGaussIntegrationRule()),*sAssembler,_elementErosionFilter);
      }
    }

		if (dom->IsInterfaceTerms()){
			dgPartDomain* dgDom = static_cast<dgPartDomain*>(dom);
      if (dgDom->getMaterialLawMinus()->withEnergyDissipation() or dgDom->getMaterialLawPlus()->withEnergyDissipation()){
        AssemblerPathFollowingConstraint(*(dgDom->getScalarTermPFBound()),dgDom->gi->begin(),dgDom->gi->end(),
									*(dgDom->getInterfaceGaussIntegrationRule()),*sAssembler,_elementErosionFilter);

      }
		}
	}
};

void nonLinearMechSolver::computeDPathFollowingConstraintDUnknowns(){
	staticDofManager<double> *sAssembler = dynamic_cast<staticDofManager<double>*>(pAssembler);
	if (sAssembler == NULL) Msg::Error("staticDofManager must be used nonLinearMechSolver::computeDPathFollowingConstraintDUnknowns");

	for(std::vector<partDomain*>::iterator itdom=domainVector.begin(); itdom!=domainVector.end();++itdom){
    partDomain *dom = *itdom;
    if (dom->getMaterialLaw() != NULL){
      if (dom->getMaterialLaw()->withEnergyDissipation()){
        AssemblerDPathFollowingConstraintDUnknowns(*(dom->getLinearTermPFBulk()),*(dom->getFunctionSpace()),
						 dom->g_cbegin(),dom->g_cend(),*(dom->getBulkGaussIntegrationRule()),*sAssembler,_elementErosionFilter);
      }
    }
		if (dom->IsInterfaceTerms()){
			dgPartDomain* dgDom = static_cast<dgPartDomain*>(dom);
      if (dgDom->getMaterialLawMinus()->withEnergyDissipation() or dgDom->getMaterialLawPlus()->withEnergyDissipation()){
        AssemblerDPathFollowingConstraintDUnknowns(*(dgDom->getLinearTermPFBound()),*(dgDom->getFunctionSpace()),
						 dgDom->gi->begin(),dgDom->gi->end(),*(dgDom->getInterfaceGaussIntegrationRule()),*sAssembler,_elementErosionFilter);
      }
		}
	}

};

void nonLinearMechSolver::computeExternalForces(){
  if(this->getScheme() != StaticLinear){
    staticDofManager<double> *sAssembler;
    for (std::list<nonLinearNeumannBC>::iterator itbc = allNeumann.begin(); itbc != allNeumann.end();++itbc)
    {
      nonLinearNeumannBC &neu = *itbc;
      // select the system if required
      if(_vcompBySys.size()==0)
      {
        sAssembler = static_cast<staticDofManager<double>*>(pAssembler);
      }
      else
      {
        dofManagerMultiSystems<double>* multiManager = static_cast<dofManagerMultiSystems<double>*>(pAssembler);
        sAssembler = static_cast<staticDofManager<double>*>(multiManager->getDofManagerByComp(neu._comp));
      }
      for(int j=0;j<neu._vspace.size();j++)
      {
        FunctionSpaceBase *spn = neu._vspace[j];
        const groupOfElements *gneu = neu._vgroup[j];
        QuadratureBase *qneu = neu._vquadrature[j];
        Assemble(neu._vterm[j],*spn,gneu->begin(),
               gneu->end(),*(qneu),_ufield,*sAssembler,nonLinearSystem<double>::Fext,_elementErosionFilter);
      }
    }

    //
    if (_testFlag){
      if (_pbcGroup != NULL){
        for (std::list<nonLinearNeumannBCAtCorner>::iterator itc = allCornerForce.begin(); itc != allCornerForce.end();++itc){
          nonLinearNeumannBCAtCorner& con = *itc;
          // select the system if required
          if(_vcompBySys.size()==0)
          {
            sAssembler = static_cast<staticDofManager<double>*>(pAssembler);
          }
          else
          {
            dofManagerMultiSystems<double>* multiManager = static_cast<dofManagerMultiSystems<double>*>(pAssembler);
            sAssembler = static_cast<staticDofManager<double>*>(multiManager->getDofManagerByComp(con._comp));
          }

          std::vector<Dof>& allDof = con.constraintDofs;
          int size = allDof.size();
          MVertex* v = _pbcGroup->getCornerVertex(con._cornerNumber);
          fullVector<double> F;
          F.resize(size,true);
          double val = con._f->operator ()(v->x(),v->y(),v->z());
          F.setAll(val);
          sAssembler->assemble(allDof,F,nonLinearSystem<double>::Fext);
        };

      };
    };
  }
  else{
    for(std::list<nonLinearNeumannBC>::iterator itbc = allNeumann.begin(); itbc != allNeumann.end();++itbc)
    {
      nonLinearNeumannBC &neu = *itbc;
      for(int j=0;j<neu._vspace.size();j++)
      {
        FunctionSpaceBase *spn = neu._vspace[j];
        const groupOfElements *gneu = neu._vgroup[j];
        QuadratureBase *qneu = neu._vquadrature[j];
        Assemble(*(neu._vterm[j]),*spn,gneu->begin(),gneu->end(),*(qneu),*pAssembler,_elementErosionFilter);
      }
    }

    if (_testFlag){
      if (_pbcGroup != NULL){
        for (std::list<nonLinearNeumannBCAtCorner>::iterator itc = allCornerForce.begin(); itc != allCornerForce.end();++itc){
          nonLinearNeumannBCAtCorner& con = *itc;
          std::vector<Dof>& allDof = con.constraintDofs;
          int size = allDof.size();
          MVertex* v = _pbcGroup->getCornerVertex(con._cornerNumber);
          fullVector<double> F;
          F.resize(size,true);
          double val = con._f->operator ()(v->x(),v->y(),v->z());
          F.setAll(val);
          pAssembler->assemble(allDof,F);
        };

      };
    };
  }
}

void nonLinearMechSolver::snlData(const int ns, const double et, const double reltol, const double absTol){
  if(whatScheme==StaticNonLinear or whatScheme==Multi or whatScheme== Implicit){
    _numstepImpl=ns;endtime=et;_tol=reltol;_absTol=absTol;
  }
  else{
    _numstepImpl = 1; // for static linear
    Msg::Error("Impossible to set data for Static Non Linear scheme because another is chosen to solve the problem");
  }
}

void nonLinearMechSolver::addSystem(const int comp,const int s)
{
  _vcompBySys.push_back(comp);
  _vschemeBySys.push_back((scheme)s);
}

void nonLinearMechSolver::switchSystem(const int sysindex,const int s)
{
  _vschemeBySys[sysindex] = (scheme)s;
}

void nonLinearMechSolver::setTimeIncrementAdaptation(const bool adap, const int numNROptimal, const double exp, const double maximalStep, const double minTimeStep, const int maxFailtIn)
{
  _numNROptimal = numNROptimal;
  _maximalTimeStep = maximalStep;
  _minimalTimeStep = minTimeStep;
  _expNROptimal = exp;
  _timeStepAdaptation = adap;
  _maxAttemptStepReduction = maxFailtIn;
  if (_timeStepAdaptation) {
    Msg::Info("time increment is adapted with number of NR iterations");
  }
}

void nonLinearMechSolver::snlManageTimeStep(const int miteNR,const int iteIncreaseTS, const double redfactor, const int maxAttemptRedFactor)
{
  _maxNRite = miteNR;
  _niteTimeStepIncrease = iteIncreaseTS;
  _timeStepFactorReduction = redfactor;
  _maxAttemptStepReduction = maxAttemptRedFactor;
}

void nonLinearMechSolver::explicitData(const double ftime, const double gams, const double beta, const double gamma, const double alpham,const bool benson){
  _timeStepByBenson = benson;
  endtime = ftime;
  _beta=beta;
  _gamma=gamma;
  _alpham = alpham;
  // depends on numerical damping
  double omegas;
  _rhoinfty = (alpham+1.)/(2.-alpham);
  omegas = sqrt((12.*(1.+_rhoinfty)*(1.+_rhoinfty)*(1.+_rhoinfty)*(2.-_rhoinfty))/(10.+15.*_rhoinfty-_rhoinfty*_rhoinfty+_rhoinfty*_rhoinfty*_rhoinfty-_rhoinfty*_rhoinfty*_rhoinfty*_rhoinfty));
    _gammas = gams*omegas;
}

void nonLinearMechSolver::implicitData(const double beta, const double gamma, const double alpham , const double alphaf){
  _beta = beta;
  _gamma = gamma;
  _alphaf = alphaf;
  _alpham = alpham;
};

void nonLinearMechSolver::explicitSpectralRadius(const double ftime,const double gams, const double rho,const bool benson){
  _timeStepByBenson = benson;
  endtime = ftime;
  _rhoinfty=rho;
  _beta = (5.-3.*_rhoinfty)/((1.+_rhoinfty)*(1.+_rhoinfty)*(2.-_rhoinfty));
  _alpham = (2.*_rhoinfty-1.)/(1.+_rhoinfty);
  _gamma = 1.5-_alpham;
  double omegas = sqrt((12.*(1.+_rhoinfty)*(1.+_rhoinfty)*(1.+_rhoinfty)*(2.-_rhoinfty))/(10.+15.*_rhoinfty-_rhoinfty*_rhoinfty+_rhoinfty*_rhoinfty*_rhoinfty-_rhoinfty*_rhoinfty*_rhoinfty*_rhoinfty));
    _gammas = gams*omegas;
}

void nonLinearMechSolver::options(const std::string &options)
{
  if(!strcmp("petsc_direct",options.c_str()))
  {
    _solver_options = std::string("-pc_type lu -ksp_type preonly");
  }
  else if(!strcmp("petsc_gmres",options.c_str()))
  {
    _solver_options = std::string("-pc_type jacobi -ksp_gmres_restart 200");
  }
  else if(!strcmp("petsc_mumps",options.c_str()))
  {
    _solver_options = std::string("-pc_mat_solver_package mumps");
  }
  else
  {
    _solver_options = std::string(options);
  }
  return;
}

void nonLinearMechSolver::implicitSpectralRadius(const double rho){
  _rhoinfty = rho;
  _alpham = (2.*_rhoinfty-1.)/(1.+_rhoinfty);
  _alphaf = _rhoinfty/(_rhoinfty +1.);
  _beta = 1./((_rhoinfty +1.)*(_rhoinfty +1.));
  _gamma = (3.-_rhoinfty)/(2.+2.*_rhoinfty);

};

void nonLinearMechSolver::explicitTimeStepEvaluation(const int nst){
  _numstepExpl = nst;
}

double nonLinearMechSolver::solve(){
  double reachtime=0.;
  if (_microFlag == false){
    switch(whatScheme){
      case StaticLinear :
          reachtime = this->solveStaticLinear();
        break;
      case StaticNonLinear :
          reachtime = this->solveSNL();
        break;
      case Implicit:
          reachtime = this->solveSNL();
        break;
      case Explicit:
        reachtime = this->solveExplicit();
        break;
      case Multi:
        reachtime = this->solveMulti();
		break;
	  case Eigen:
		reachtime = this->solveEigen();
		break;
    }
  }
  else{
    switch(whatScheme){
      case StaticLinear:
        reachtime = this->microSolveStaticLinear();
        break;
      case StaticNonLinear :
        reachtime = this->microSolveSNL();
		break;
    }
  };
  Msg::Barrier(); // wait all before end (for MPI)
  _previousScheme = whatScheme;
  _notResetedBC = true;
  _notResetedContact = true;
  _notResetUnknowns = true;
  _dynamicRelaxation = false;
  return reachtime;
}

void nonLinearMechSolver::resetBoundaryConditions(const bool notResetUnknwown)
{
  for(std::list<nonLinearDirichletBC>::iterator itbc = allDirichlet.begin(); itbc != allDirichlet.end();++itbc)
  {
    itbc->clear();
  }
  allDirichlet.clear();
  for(std::list<nonLinearNeumannBC>::iterator itbc = allNeumann.begin(); itbc != allNeumann.end();++itbc)
  {
    itbc->clear();
  }
  allNeumann.clear();
  for(std::list<initialCondition>::iterator itic = allinitial.begin(); itic !=allinitial.end();++itic)
  {
    itic->clear();
  }
  allinitial.clear();

  allCornerConstraint.clear();
  allCornerForce.clear();

  // clear all other BC
  // as objects created, destructor leading to clean the objects

  allPeriodic.clear();
  allAveragePeriodic.clear();
  allSameDisp.clear();
  allFixAllFace.clear();
  allConstraint.clear();

  _testFlag = false;
  if (_pbcGroup) {
    delete _pbcGroup;
    _pbcGroup = NULL;
  }

  _notResetedBC = false;
  _notResetUnknowns = notResetUnknwown;
}

void nonLinearMechSolver::resetContactInteraction()
{
 // for(contactContainer::iterator it = _allContact.begin(); it!=_allContact.end(); ++it){
 //   contactDomain *cdom = *it;
 //   delete cdom;
 // }
  _allContact.clear();
  //HERE NOT SURE WE NEED TO RESET THE DEFO DEFO CONTACT; DO ANOTHER FUNCTION?
  _notResetedContact = false;
}
void nonLinearMechSolver::disableResetRestart() //to avoid restart at a change of solver
{
  _disableResetRestart = true;
}

void nonLinearMechSolver::dynamicRelaxation(const double gams,const double ftime,const double tol,const int wsolver,const bool benson)
{
  _timeStepByBenson = true;
  _dynamicRelaxation = true;
  whatScheme = Explicit;
  if(wsolver == 2)
    whatSolver = Petsc;
  _tol = tol;
  this->explicitSpectralRadius(ftime,gams,0.99); //rho_infty == 0 for dynamic relaxtion (no influence ??)
}

void nonLinearMechSolver::displacementBC(std::string onwhat, const int numphys, const int comp, const double value)
{
  simpleFunction<double> *fct = new simpleFunctionTime<double>(value);
  elementFilterTrivial filter = elementFilterTrivial();
  this->displacementBC(onwhat,numphys,comp,&filter,fct);
}

void nonLinearMechSolver::displacementBC(std::string onwhat, const int numphys, const int comp, const double valueDiff, const double valueInit){
  simpleFunctionTime<double> *fct = new simpleFunctionTimeWithConstant<double>(valueDiff,true,1.,valueInit);
  elementFilterTrivial filter = elementFilterTrivial();
  this->displacementBC(onwhat,numphys,comp,&filter,fct);
};

void nonLinearMechSolver::velocityBC(std::string onwhat, const int numphys, const int comp, const double value)
{
  this->displacementBC(onwhat,numphys,comp,value);
  allDirichlet.back()._mycondition = nonLinearBoundaryCondition::velocity;
}

void nonLinearMechSolver::accelerationBC(std::string onwhat, const int numphys, const int comp, const double value)
{
  this->displacementBC(onwhat,numphys,comp,value);
  allDirichlet.back()._mycondition = nonLinearBoundaryCondition::acceleration;
}

void nonLinearMechSolver::displacementBC(std::string onwhat, const int numphys, const int comp, const double value, elementFilter *filter)
{
  simpleFunction<double> *fct = new simpleFunctionTime<double>(value);
  this->displacementBC(onwhat,numphys,comp,filter,fct);
}

void nonLinearMechSolver::velocityBC(std::string onwhat, const int numphys, const int comp, const double value, elementFilter *filter)
{
  this->displacementBC(onwhat,numphys,comp,value,filter);
  allDirichlet.back()._mycondition = nonLinearBoundaryCondition::velocity;
}

void nonLinearMechSolver::displacementBC(std::string onwhat, const int numphys, const int comp,simpleFunction<double> *fct)
{
  elementFilterTrivial filter = elementFilterTrivial();
  this->displacementBC(onwhat,numphys,comp,&filter,fct);
}

void nonLinearMechSolver::velocityBC(std::string onwhat, const int numphys, const int comp,simpleFunction<double> *fct)
{
  this->displacementBC(onwhat,numphys,comp,fct);
  allDirichlet.back()._mycondition = nonLinearBoundaryCondition::velocity;
}

void nonLinearMechSolver::displacementBC(std::string onwhat, const int numphys, const int comp, elementFilter *filter,
                                         simpleFunction<double> *fct)
{
  allDirichlet.emplace_back();
  nonLinearDirichletBC& diri = allDirichlet.back();
  const std::string node("Node");
  const std::string edge("Edge");
  const std::string face("Face");
  const std::string volume("Volume");
  diri.g = new groupOfElements();
  if(onwhat==node){
    diri.g->addPhysical(0, numphys,*filter);
    diri.onWhat=nonLinearBoundaryCondition::ON_VERTEX;
  }
  else if(onwhat==edge){
    diri.g->addPhysical(1, numphys,*filter);
    diri.onWhat=nonLinearBoundaryCondition::ON_EDGE;
  }
  else if(onwhat==face){
    diri.g->addPhysical(2, numphys,*filter);
    diri.onWhat=nonLinearBoundaryCondition::ON_FACE;
  }
  else if(onwhat==volume){
    diri.g->addPhysical(3, numphys,*filter);
    diri.onWhat=nonLinearBoundaryCondition::ON_VOLUME;
  }
  else{
    Msg::Error("Impossible to prescribe a displacement on a %s\n",onwhat.c_str());
    delete diri.g;
    return;
  }
  diri._f= static_cast<simpleFunctionTime<double>*>(fct);
  diri._comp=comp;
  diri._tag=numphys;
}

void nonLinearMechSolver::velocityBC(std::string onwhat, const int numphys, const int comp, elementFilter *filter,
                                         simpleFunction<double> *fct)
{
  this->displacementBC(onwhat,numphys,comp,filter,fct);
  allDirichlet.back()._mycondition = nonLinearBoundaryCondition::velocity;
}

void nonLinearMechSolver::displacementRigidContactBC(const int numphys, const int comp_, const double value)
{
  simpleFunction<double> *fct = new simpleFunctionTime<double>(value);
  this->displacementRigidContactBC(numphys,comp_,fct);
}
void nonLinearMechSolver::displacementRigidContactBC(const int numphys, const int comp_, simpleFunction<double> *fct)
{
  allContactBC.emplace_back(numphys);
  rigidContactBC& diri = allContactBC.back();
  diri.onWhat = nonLinearBoundaryCondition::RIGIDCONTACT;
  diri._comp = comp_;
  diri._f = static_cast<simpleFunctionTime<double>*>(fct);
}

void nonLinearMechSolver::initialBC(std::string onwhat, std::string whichC, const int numphys,
                                          const int comp, const double value)
{
  elementFilterTrivial filter = elementFilterTrivial();
  this->initialBC(onwhat,whichC,numphys,comp,value,&filter);
}

void nonLinearMechSolver::initialBC(std::string onwhat, std::string whichC, const int numphys,
                                          const int comp, const double value,
                                          elementFilter *filter)
{
  simpleFunctionTime<double>* sfunction = new simpleFunctionTime<double>(value,false,0.);
  this->initialBC(onwhat,whichC,numphys,comp,sfunction,filter);
}

void nonLinearMechSolver::initialBC(std::string onwhat, std::string whichC, const int numphys, const int comp,
                                    simpleFunctionTime<double> *fct,elementFilter* filter)
{
  bool nullfilter=false;
  if(filter==NULL)
  {
    filter = new elementFilterTrivial();
    nullfilter = true;
  }

  const std::string node("Node");
  const std::string edge("Edge");
  const std::string face("Face");
  const std::string volume("Volume");
  const std::string rigid("Rigid");
  const std::string position("Position");
  const std::string velocity("Velocity");
  const std::string acceleration("Acceleration");

  nonLinearBoundaryCondition::location ow;
  groupOfElements *gr = new groupOfElements();
  if(onwhat == node){
    ow = nonLinearBoundaryCondition::ON_VERTEX;
    gr->addPhysical(0,numphys,*filter);
  }
  else if(onwhat == edge){
    ow = nonLinearBoundaryCondition::ON_EDGE;
    gr->addPhysical(1,numphys,*filter);
  }
  else if(onwhat == face){
    ow = nonLinearBoundaryCondition::ON_FACE;
    gr->addPhysical(2,numphys,*filter);
  }
  else if(onwhat == volume){
    ow = nonLinearBoundaryCondition::ON_VOLUME;
    gr->addPhysical(3,numphys,*filter);
  }
  else if(onwhat == rigid)
  {
    ow = nonLinearBoundaryCondition::RIGIDCONTACT;
    // empty group insert gravity center later
  }
  else{
    ow = nonLinearBoundaryCondition::UNDEF;
    gr->addPhysical(2,numphys,*filter);
  }

  nonLinearBoundaryCondition::whichCondition wc;
  if(whichC == position)
    wc = nonLinearBoundaryCondition::position;
  else if(whichC == velocity)
    wc = nonLinearBoundaryCondition::velocity;
  else if(whichC == acceleration)
    wc = nonLinearBoundaryCondition::acceleration;
  else{
    Msg::Warning("Impossible to prescribed an initial condition %d",numphys);
    //return;
  }
  allinitial.emplace_back(comp,wc);
  initialCondition& initc = allinitial.back();
  initc.onWhat = ow;
  initc.g = gr;
  initc._tag = numphys;
  initc._f = fct;



  if(nullfilter)
  {
    delete filter;
  }
}

void nonLinearMechSolver::initialDCBVeloBC(std::string onwhat,const int numphys,const int axiscomp, const int deflcomp,const double length, const double value)
{
  const std::string edge("Edge");
  const std::string face("Face");
  const std::string volume("Volume");
  allinitial.emplace_back(deflcomp,nonLinearBoundaryCondition::velocity);
  initialCondition initc = allinitial.back();
  initc._f = new functionVelocityDCB(value,axiscomp,length);
  initc.g = new groupOfElements();
  initc._tag = numphys;
  if(onwhat == edge){
    initc.onWhat = nonLinearBoundaryCondition::ON_EDGE;
    initc.g->addPhysical(1,numphys);
  }
  else if(onwhat == face){
    initc.onWhat = nonLinearBoundaryCondition::ON_FACE;
    initc.g->addPhysical(2,numphys);
  }
  else if(onwhat == volume){
    initc.onWhat  = nonLinearBoundaryCondition::ON_VOLUME;
    initc.g->addPhysical(3,numphys);
  }
  else{
    Msg::Error("The initial velocity BC can be applied only on an edge or on a face");
  }
}

void nonLinearMechSolver::independentDisplacementBC(std::string onwhat, const int numphys, const int comp, const double value){
  allDirichlet.emplace_back();
  nonLinearDirichletBC& diri = allDirichlet.back();
  const std::string node("Node");
  const std::string edge("Edge");
  const std::string face("Face");
  const std::string volume("Volume");
  if(onwhat==node){
    diri.g = new groupOfElements (0, numphys);
    diri.onWhat=nonLinearBoundaryCondition::ON_VERTEX;
  }
  else if(onwhat==edge){
    diri.g = new groupOfElements (1, numphys);
    diri.onWhat=nonLinearBoundaryCondition::ON_EDGE;
  }
  else if(onwhat==face){
    diri.g = new groupOfElements (2, numphys);
    diri.onWhat=nonLinearBoundaryCondition::ON_FACE;
  }
  else if(onwhat==volume){
    diri.g = new groupOfElements (3, numphys);
    diri.onWhat=nonLinearBoundaryCondition::ON_VOLUME;
  }
  else Msg::Error("Impossible to prescribe a displacement on a %s\n",onwhat.c_str());
  diri._f= new simpleFunctionTime<double>(value,false);
  diri._comp=comp;
  diri._tag=numphys;
}

void nonLinearMechSolver::forceBC(std::string onwhat, const int numphys, const int comp, const double val)
{
  simpleFunctionTime<double> *fct = new simpleFunctionTime<double>(val,true,0.);
  this->forceBC(onwhat,numphys,comp,fct);
}
void nonLinearMechSolver::forceBC(std::string onwhat, const int numphys, const int comp, simpleFunctionTime<double> *fct)
{
   // in general, forceBC is applied for the flux of and arbitrary unknown field
  // (force in case of displacement, extradof field in case of extraDof, ...) in REFERENCE CONFIGURATION.
  // in case of using CURRENT CONFIGURATION, as pressure or scalar flux,
  // pressureOnPhysicalGroupBC for presurre or
  // scalarFluxBC for scalar flux  should be used

  const std::string node("Node");
  const std::string edge("Edge");
  const std::string face("Face");
  const std::string volume("Volume");
  const std::string facevertex("FaceVertex");

  allNeumann.emplace_back();
  nonLinearNeumannBC& neu = allNeumann.back();

  if(onwhat==node){
    neu.g = new groupOfElements (0, numphys);
    neu.onWhat=nonLinearBoundaryCondition::ON_VERTEX;
  }
  else if(onwhat==edge){
    neu.g = new groupOfElements (1, numphys);
    neu.onWhat=nonLinearBoundaryCondition::ON_EDGE;
  }
  else if(onwhat==face){
    neu.g = new groupOfElements (2, numphys);
    neu.onWhat=nonLinearBoundaryCondition::ON_FACE;
  }
  else if(onwhat==facevertex)
  {
    neu.g = new groupOfElements(2,numphys);
    neu.onWhat=nonLinearBoundaryCondition::ON_FACE_VERTEX;
  }
  else if(onwhat==volume){
    neu.g = new groupOfElements (3, numphys);
    neu.onWhat=nonLinearBoundaryCondition::ON_VOLUME;
  }
  else  {
    Msg::Error("Impossible to prescribe a force on a %s\n",onwhat.c_str());
    if (onwhat == "Pressure"){
      Msg::Error("pressureOnPhysicalGroupBC should be used instead of forceBC when applying pressure on a BC \n Applying pressure by this function has been removed");
    }
    else if (onwhat == "Flux"){
      Msg::Error("scalarFluxBC should be used instead of forceBC when applying scalar flux on a BC \n Applying scalar flux by this function has been removed");
    }
  };

  neu._NeumannBCType = nonLinearNeumannBC::FORCE;
  neu._f= fct;
  neu._tag=numphys;
  neu._comp=comp;
};

void nonLinearMechSolver::scalarFluxBC(std::string onwhat, const int numphys, const int comp, const double val){
  simpleFunctionTime<double> *fct = new simpleFunctionTime<double>(val,true,0.);
  this->scalarFluxBC(onwhat,numphys,comp,fct);
};
void nonLinearMechSolver::scalarFluxBC(std::string onwhat, const int numphys, const int comp, simpleFunctionTime<double> *fct){
  const std::string edge("Edge");
  const std::string face("Face");

  allNeumann.emplace_back();
  nonLinearNeumannBC& neu = allNeumann.back();
  if(onwhat==edge){
    neu.g = new groupOfElements (1, numphys);
    neu.onWhat=nonLinearBoundaryCondition::ON_EDGE;
  }
  else if(onwhat==face){
    neu.g = new groupOfElements (2, numphys);
    neu.onWhat=nonLinearBoundaryCondition::ON_FACE;
  }
  else  Msg::Error("Impossible to prescribe a scalar flux on a %s\n",onwhat.c_str());

  neu._NeumannBCType = nonLinearNeumannBC::SCALAR_FLUX;
  neu._f= fct;
  neu._tag=numphys;
  neu._comp=comp;
};

void nonLinearMechSolver::independentForceBC(std::string onwhat, const int numphys, const int comp, const double val){
  allNeumann.emplace_back();
  nonLinearNeumannBC& neu = allNeumann.back();
  const std::string node("Node");
  const std::string edge("Edge");
  const std::string face("Face");
  const std::string volume("Volume");
  if(onwhat==node){
    neu.g = new groupOfElements (0, numphys);
    neu.onWhat=nonLinearBoundaryCondition::ON_VERTEX;
  }
  else if(onwhat==edge){
    neu.g = new groupOfElements (1, numphys);
    neu.onWhat=nonLinearBoundaryCondition::ON_EDGE;
  }
  else if(onwhat==face){
    neu.g = new groupOfElements (2, numphys);
    neu.onWhat=nonLinearBoundaryCondition::ON_FACE;
  }
  else if(onwhat==volume){
    neu.g = new groupOfElements (3, numphys);
    neu.onWhat=nonLinearBoundaryCondition::ON_VOLUME;
  }
  else  Msg::Error("Impossible to prescribe a force on a %s\n",onwhat.c_str());
  neu._NeumannBCType = nonLinearNeumannBC::FORCE;
  neu._f= new simpleFunctionTimeWithConstant<double>(val,false);
  neu._tag=numphys;
  neu._comp = comp;
}

void nonLinearMechSolver::pressureOnPhysicalGroupBC(const int numphys, const double press, const double p0){
  Msg::Warning("applying pressure on surface");
  this->pressureOnPhysicalGroupBC("Face",numphys,press,p0);
}

void nonLinearMechSolver::pressureOnPhysicalGroupBC(const int numphys, simpleFunctionTime<double> *fct){
  Msg::Warning("applying pressure on surface");
  this->pressureOnPhysicalGroupBC("Face",numphys,fct);
};

void nonLinearMechSolver::pressureOnPhysicalGroupBC(std::string onwhat, const int numphys, const double pressDiff, const double p0){
  simpleFunctionTime<double> *fct = new simpleFunctionTimeWithConstant<double>(pressDiff,true,1.,p0);
  this->pressureOnPhysicalGroupBC(onwhat,numphys,fct);
};


void nonLinearMechSolver::pressureOnPhysicalGroupBC(std::string onwhat, const int numphys, simpleFunctionTime<double> *fct){
  allNeumann.emplace_back(nonLinearNeumannBC());
  nonLinearNeumannBC& neu = allNeumann.back();

  const std::string edge("Edge");
  const std::string face("Face");
  if (onwhat == face){
    neu.g = new groupOfElements(2,numphys);
    neu.onWhat = nonLinearBoundaryCondition::ON_FACE;
  }
  else if (onwhat == edge){
    neu.g = new groupOfElements(1,numphys);
    neu.onWhat = nonLinearBoundaryCondition::ON_EDGE;
  }
  else{
    Msg::Error("Impossible to prescribe a pressure on a %s\n",onwhat.c_str());
  }
  neu._NeumannBCType = nonLinearNeumannBC::PRESSURE;
  neu._f = fct;
  neu._tag=numphys;
};

void nonLinearMechSolver::blastPressureBC(std::string onwhat, const int numphys,const double p0,const double p1, const double plexp,
                                       const double t0, const double t1){
  simpleFunctionTime<double> *fct = new powerDecreasingFunctionTime(p0,p1,plexp,t0,t1,this->endtime);
  this->pressureOnPhysicalGroupBC(onwhat,numphys,fct);
};


void nonLinearMechSolver::blastPressureBC(const int numphys,const double p0,const double p1, const double plexp,
                                       const double t0, const double t1){
  Msg::Warning("applying pressure on surface");
  this->blastPressureBC("Face",numphys,p0,p1,plexp,t0,t1);
};

void nonLinearMechSolver::archivingForceOnPhysicalGroup(const std::string onwhat, const int numphys, const int comp,const int nstep){
  // get the node of the edge
  static std::string node("Node");
  static std::string edge("Edge");
  static std::string face("Face");
  static std::string volu("Volume");
  int dim=0;
  if(onwhat == node )
    dim = 0;
  else if(onwhat == edge)
    dim = 1;
  else if(onwhat == face)
    dim = 2;
  else if(onwhat == volu)
    dim = 3;
    
  // check if 
  bool found = false;
  for (int i=0; i< vaf.size(); i++)
  {
    if ((vaf[i].numphys == numphys) and (vaf[i].comp == comp))
    {
      found = true;
      break;
    }
  }
  if (!found)
  {
    vaf.emplace_back(numphys,dim,comp,nstep);
  }
}

void nonLinearMechSolver::archivingRigidContactForce(const int numphys, const int comp,const int nstep){
  // check if 
  bool found = false;
  for (int i=0; i< vaf.size(); i++)
  {
    if ((vaf[i].numphys == numphys) and (vaf[i].comp == comp))
    {
      found = true;
      break;
    }
  }
  
  if (!found)
  {
    vaf.emplace_back(numphys,0,comp,nstep);
  }
}

void nonLinearMechSolver::initArchiveReactionForceOnFixedPhysical(archiveForce& aforce){
  if(_previousInit) {
    aforce.vdof.clear();
  };
  std::vector<MVertex*> vv;
  pModel->getMeshVerticesForPhysicalGroup(aforce.dim,aforce.numphys,vv);

  if(vv.size() > 0) {
    for(int idom = 0; idom < _mpiUserDom;++idom){ // Do not consider Ghost domain when archiving forces
      partDomain *dom = domainVector[idom];
      FunctionSpaceBase *sp = dom->getFunctionSpace();
      std::vector<Dof> R;
      for(groupOfElements::elementContainer::iterator ite = dom->g_cbegin(); ite!=dom->g_cend(); ++ite){
        MElement *ele = *ite;
        int nbvertex = ele->getNumVertices();
        sp->getKeys(ele,R);
        for(std::vector<MVertex*>::iterator it=vv.begin(); it!=vv.end(); ++it){
          for(int j=0;j<nbvertex; j++){
            if(ele->getVertex(j) == *it){
              aforce.vdof.insert(R[j+nbvertex*aforce.comp]);
              break;
            }
          }
        }
        R.clear();
      }
    }
  }
};

void nonLinearMechSolver::initArchiveForce(){
 #if defined(HAVE_MPI)
  std::vector<int> has_this_contact(Msg::GetCommSize(),0);
 #endif // HAVE_MPI
  for(std::vector<archiveForce>::iterator itf=vaf.begin(); itf!=vaf.end(); ++itf){
    archiveForce &af = *itf;
    if(af.vdof.size()==0){// otherwise it has already be initialized before in a previous computation
      std::vector<MVertex*> vv;
      pModel->getMeshVerticesForPhysicalGroup(af.dim,af.numphys,vv);
      // loop on domain (to find the domain where the force is applied)
      if(vv.size() > 0) {// otherwise rigid contact force
        for(int idom = 0; idom < _mpiUserDom;++idom){ // Do not consider Ghost domain when archiving forces
          partDomain *dom = domainVector[idom];
          FunctionSpaceBase *sp = dom->getFunctionSpace();
          std::vector<Dof> R;
          for(groupOfElements::elementContainer::iterator ite = dom->g_cbegin(); ite!=dom->g_cend(); ++ite){
            MElement *ele = *ite;
            int nbvertex = ele->getNumVertices();
            sp->getKeys(ele,R);
            for(std::vector<MVertex*>::iterator it=vv.begin(); it!=vv.end(); ++it){
              for(int j=0;j<nbvertex; j++){
                if(ele->getVertex(j) == *it){
                  af.vdof.insert(R[j+nbvertex*af.comp]);
                  break;
                }
              }
            }
            R.clear();
          }
        }
      }
      else {
        for(nonLinearMechSolver::contactContainer::const_iterator itc = _allContact.begin(); itc!=_allContact.end(); ++itc){
          contactDomain *cdom = *itc;
          if(cdom->getPhys() != af.numphys)
            continue;
          partDomain* domf=NULL;
          int physDom = cdom->getPhysSlave();
          for(std::vector<partDomain*>::iterator itdom = domainVector.begin(); itdom!=domainVector.end(); ++itdom){
            partDomain *dom = *itdom;
            if(dom->getPhysical() == physDom){
              domf = dom;
              break;
            }
          }
        #if defined(HAVE_MPI)
          if(false) // domf == NULL)
          {
            for(std::vector<partDomain*>::iterator itdom = _ghostDomainMPI.begin(); itdom!=_ghostDomainMPI.end(); ++itdom)
            {
              partDomain *dom = *itdom;
              if(dom->getPhysical() == physDom)
              {
                domf = dom;
                break;
              }
            }
          }
        #endif // HAVE_MPI
          if(domf==NULL)
            continue;

          std::vector<Dof> R;
          (const_cast<rigidContactSpaceBase*>(static_cast<const rigidContactSpaceBase*>(cdom->getSpace())))->getKeysOfGravityCenter(R);

          FilterDof *fdof = domf->createFilterDof(af.comp);
          for(int j=0;j<R.size(); j++) {
            if(fdof->operator()(R[j])){
              af.vdof.insert(R[j]);
              break;   // only 1 dof allow...
            }
          }
          delete fdof;
        }
      #if defined(HAVE_MPI)
        if(!af.vdof.empty()){
           has_this_contact[Msg::GetCommRank()] = 1;
        }
        MPI_Allreduce(MPI_IN_PLACE,has_this_contact.data(),Msg::GetCommSize(),MPI_INT,MPI_SUM,MPI_COMM_WORLD);
        int root_rank = 0;
        for(; root_rank< Msg::GetCommSize();++root_rank)
        {
            if(has_this_contact[root_rank]==1)
                break;
        }
        if(root_rank == Msg::GetCommSize())
            Msg::Error("cannot find a contact on any of the partition...");
        af.contact_root = root_rank;
        std::fill(has_this_contact.begin(),has_this_contact.end(),0);
      #endif // HAVE_MPI
      }
    }
    if(!af.vdof.empty() && (af.contact_root == -1 || af.contact_root==Msg::GetCommRank()))
      af.openFile();
  }
}

void nonLinearMechSolver::pathFollowingArchiving(const double curtime, const int numstep){
  if (_pathFollowing){
    std::string name = "A";
    linearSystem<double>* lsys = pAssembler->getLinearSystem(name);
    nonLinearSystem<double>* nsys = dynamic_cast<nonLinearSystem<double>*>(lsys);
    pathFollowingSystem<double>* psys = dynamic_cast<pathFollowingSystem<double>*>(lsys);
    double control = psys->getControlParameter();
    double state = psys->getStateParameter();
    
    #if defined(HAVE_MPI)
    if (Msg::GetCommRank() == 0)
    #endif // HAVE_MPI
    {
      std::string filename = "pathResult_Time_ControlParameter_StateParameter.csv";
      FILE* file = fopen(filename.c_str(),"a");
      fprintf(file,"%e; %e; %e; \n",curtime,control,state);
      fclose(file);
    }
  }
};

void nonLinearMechSolver::archivingNodeDisplacement(const int numphys, const int comp,const int nstep){
  this->archivingNode(numphys,comp,nonLinearBoundaryCondition::position,nstep);
}

void nonLinearMechSolver::archivingNodeDisplacementOnControlNode(const int pos, const int comp,const int nstep){
  anoded.push_back(unknownField::archiveNode(pos,0,comp,nonLinearBoundaryCondition::position,nstep));
}

void nonLinearMechSolver::archivingMeshVertexDisplacement(const int num, const int comp,const int nstep){
  // no distinction between cG/dG and full Dg formulation. class Displacement Field manage it
  MVertex* v = NULL;
  v = pModel->getMeshVertexByTag(num);

  nonLinearBoundaryCondition::whichCondition wc = nonLinearBoundaryCondition::position;

  if(v != NULL){
    anoded.push_back(unknownField::archiveNode(0,v->getNum(),comp,wc,nstep));
  }
  else{
    Msg::Warning("No mesh vertex %d So it is impossible to archive nodal displacement",num);
  }

};

void nonLinearMechSolver::archivingVolumeIntegralValue(const int num, const int nstep)
{
  _dataVolumeIntegral.emplace_back(1,num,-1,nstep);
};

void nonLinearMechSolver::archivingVolumeIntegralValueOnPhysical(const int phy, const int num, const int nstep)
{
  _dataVolumeIntegral.emplace_back(1,num,phy,nstep);
};

void nonLinearMechSolver::archivingAverageValue(const int num, const int nstep){
  _dataVolumeIntegral.emplace_back(2,num,-1,nstep);
};
void nonLinearMechSolver::archivingAverageValueActiveDissipation(const int num, const int nstep){
  _dataVolumeIntegral.emplace_back(3,num,-1,nstep);
};

void nonLinearMechSolver::IPVolumeIntegralArchiving(const double curtime, const int numstep, const bool forceSave)
{
  for (std::list<IntegralVolume>::iterator it = _dataVolumeIntegral.begin(); it != _dataVolumeIntegral.end(); it++)
  {
    IntegralVolume& iv = *it;
    if (numstep > iv.lastSaveStep)
    {
      if (numstep%iv.nbArch==0 or forceSave)
      {
        iv.lastSaveStep = numstep;
        double val=0;
        if (iv.integType == IntegralVolume::VolumeIntegral)
        {
          if (iv.physical > 0)
          {
            this->getTotalIntegralByVolumeIntegralOnPhysical(iv.physical,iv.ipVal,val);
          }
          else
          {
            this->getTotalIntegralByVolumeIntegral(iv.ipVal,val);
          }
        }
        else if (iv.integType == IntegralVolume::VolumeAverage)
        {
          this->getHomogenizedPropertyByVolumeIntegral(iv.ipVal,val);
        }
        else if (iv.integType == IntegralVolume::ActiveDissipationAverage)
        {
          this->getHomogenizedPropertyByVolumeIntegralOnActiveDissipationDomain(iv.ipVal,val,IPStateBase::current);
        }
        else
        {
          Msg::Error("integType has not been defined");
        }
        if (iv.fp == NULL) iv.openFile();
        fprintf(iv.fp,"%e;%e\n",curtime,val);
        fflush(iv.fp);
      }
    }
  }
}

void nonLinearMechSolver::IPDataOnPhysicalArchiving(const double curtime, const int numstep, const bool forceSave)
{
  for (std::list<IPDataOnPhysical>::iterator idt = _dataOnPhysical.begin(); idt != _dataOnPhysical.end(); idt++)
  {
    IPDataOnPhysical& ipdata = *idt;
    int dim = ipdata.dim;
    int phys = ipdata.physical;
    int ipval = ipdata.ipValue;
    if (!ipdata.isInitialized)
    {
      ipdata.isInitialized = true;
      groupOfElements g;
      #if defined(HAVE_MPI)
      elementFilterMPI elemfil(NULL);
      g.addPhysical(dim,phys,elemfil);
      #else
      g.addPhysical(dim,phys);
      #endif // HAVE_MPI


      std::string fname = "coordinate_"+ IPField::ToString(ipval)+ "_OnPhysical_"+int2str(phys);
      #if defined(HAVE_MPI)
      if (this->getNumRanks() > 1){
        fname += "_part"+int2str(Msg::GetCommRank());
      }
      #endif //HAVE_MPIim
      fname += ".csv";


      FILE* coordinateFile = fopen(fname.c_str(),"w");
      ipdata.element.resize(g.size());
      int index= 0;
      SPoint3 p;
      for (groupOfElements::elementContainer::iterator it = g.begin(); it!= g.end(); it++)
      {
        MElement* ele = *it;
        ipdata.element[index] = ele->getNum(); index++;
        for (int idom = 0; idom < domainVector.size(); idom++){
          partDomain* dom = domainVector[idom];
          if (dom->g_find(ele)){
            IntPt* GP;
            int npts = dom->getBulkGaussIntegrationRule()->getIntPoints(ele,&GP);
            for (int j=0; j< npts; j++){
              double u = GP[j].pt[0]; double v = GP[j].pt[1]; double w = GP[j].pt[2];
              double weight = GP[j].weight;
              double detJ = ele->getJacobianDeterminant(u,v,w);
              ele->pnt(u,v,w,p);
              fprintf(coordinateFile,"%d; %e; %e; %e; %d; %e;\n",ele->getNum(),p.x(),p.y(),p.z(),j,weight*detJ);
            }
            break;
          }
        }
      }
      fclose(coordinateFile);
    }
    
    if (numstep > ipdata.lastSaveStep)
    {
      if (numstep%ipdata.nbArch==0 or forceSave)
      {
        ipdata.lastSaveStep = numstep;
        //
        std::string fname = "IPDdata_"+ IPField::ToString(ipval)+ "_OnPhysical_"+int2str(phys);
        #if defined(HAVE_MPI)
        if ( getNumRanks() > 1){
          fname += "_part"+int2str(Msg::GetCommRank());
        }
        #endif //HAVE_MPIim
        fname += "_step"+int2str(numstep)+".csv";

        FILE* file = fopen(fname.c_str(),"w");
        for (int i=0; i< ipdata.element.size(); i++){
          int elnum = ipdata.element[i];
          const std::vector<IPStateBase*>& ips = (*(_ipf->getAips()->getIPstate(elnum)));
          for (int j=0; j< ips.size(); j++)
          {
            double val = ips[j]->getState(IPStateBase::current)->get(IPField::Output(ipval));
            fprintf(file,"%e\n",val);
          }
        }
        fprintf(file,"\n");
        fclose(file);
      }
    }
  }
};

void nonLinearMechSolver::archivingIPDataOnPhysical(const int ipval, const int phys, const int dim, const int nstep){
  _dataOnPhysical.emplace_back(ipval,phys,dim,nstep);
};

void nonLinearMechSolver::archivingNodeDisplacementOnPhysical(const int dim, const int physical, const int comp,const int nstep){
  nonLinearBoundaryCondition::whichCondition wc = nonLinearBoundaryCondition::position;
  archivingNode(physical,dim,comp,wc,nstep);
};

void nonLinearMechSolver::archivingNodeVelocity(const int numphys, const int comp,const int nstep){
  this->archivingNode(numphys,comp,nonLinearBoundaryCondition::velocity,nstep);
}

void nonLinearMechSolver::archivingNodeAcceleration(const int numphys, const int comp, const int nstep){
  this->archivingNode(numphys,comp,nonLinearBoundaryCondition::acceleration,nstep);
}
void nonLinearMechSolver::constraintBC(std::string onwhat, const int numphys, const int comp)
{
  allConstraint.emplace_back();
  nonLinearConstraintBC& uc = allConstraint.back();
  const std::string node("Node");
  const std::string edge("Edge");
  const std::string face("Face");
  const std::string volume("Volume");
  if(onwhat==node){
    uc.g = new groupOfElements (0, numphys);
    uc.onWhat=nonLinearBoundaryCondition::ON_VERTEX;
  }
  else if(onwhat==edge){
    uc.g = new groupOfElements (1, numphys);
    uc.onWhat=nonLinearBoundaryCondition::ON_EDGE;
  }
  else if(onwhat==face){
    uc.g = new groupOfElements (2, numphys);
    uc.onWhat=nonLinearBoundaryCondition::ON_FACE;
  }
  else if(onwhat==volume){
    uc.g = new groupOfElements(3, numphys);
    uc.onWhat=nonLinearBoundaryCondition::ON_VOLUME;
  }
  else Msg::Error("Impossible to prescribe a constraint on a %s\n",onwhat.c_str());
  uc._comp=comp;
  uc._tag=numphys;

}

void nonLinearMechSolver::periodicBC(std::string onwhat, const int phys1, const int phys2, const int v1, const int v2, const int comp){
  Msg::Info("Periodic BC comp %d on physical %d physical %d  versur vertex %d %d",comp,phys1,phys2,v1,v2);

	allPeriodic.emplace_back();
	nonLinearPeriodicBCBetweenTwoGroups& pbc = allPeriodic.back();
  const std::string edge("Edge");
  const std::string face("Face");

 if(onwhat==edge){
    pbc.g1 = new groupOfElements (1, phys1);
    pbc.g2 = new groupOfElements (1, phys2);
    pbc.onWhat=nonLinearBoundaryCondition::ON_EDGE;
  }
  else if(onwhat==face){
    pbc.g1 = new groupOfElements (2, phys1);
    pbc.g2 = new groupOfElements (2, phys2);
    pbc.onWhat=nonLinearBoundaryCondition::ON_FACE;
  }
  else Msg::Error("Impossible to prescribe a constraint on a %s\n",onwhat.c_str());

	groupOfElements grv1(0,v1);
	groupOfElements grv2(0,v2);
	if (grv1.vsize() > 0 and grv2.vsize() > 0){
		pbc.v1 = *(grv1.vbegin());
		pbc.v2 = *(grv2.vbegin());
	}
  else{
    Msg::Error("periodic root vertices have not been correctly defined");
  }
  pbc.comp = comp;
};

void nonLinearMechSolver::averagePeriodicBC(std::string onwhat, const int phys1, const int phys2, const int v1, const int v2, const int comp){
  allAveragePeriodic.emplace_back();
	nonLinearAveragePeriodicBCBetweenTwoGroups& pbc = allAveragePeriodic.back();
  const std::string edge("Edge");
  const std::string face("Face");

 if(onwhat==edge){
    pbc.g1 = new groupOfElements (1, phys1);
    pbc.g2 = new groupOfElements (1, phys2);
    pbc.onWhat=nonLinearBoundaryCondition::ON_EDGE;
  }
  else if(onwhat==face){
    pbc.g1 = new groupOfElements (2, phys1);
    pbc.g2 = new groupOfElements (2, phys2);
    pbc.onWhat=nonLinearBoundaryCondition::ON_FACE;
  }
  else Msg::Error("Impossible to prescribe an average PBC constraint on a %s\n",onwhat.c_str());

	groupOfElements grv1(0,v1);
	groupOfElements grv2(0,v2);
	if (grv1.vsize() > 0 and grv2.vsize() > 0){
		pbc.v1 = *(grv1.vbegin());
		pbc.v2 = *(grv2.vbegin());
	}
  else{
    Msg::Error("periodic root vertices have not been correctly defined");
  }
  pbc.comp = comp;
};

void nonLinearMechSolver::sameDisplacementBC(std::string onwhat, const int phy, const int rootphy, const int comp, const double fac){
  Msg::Info("Displacement comp %d on physical %d is the same as on node %d",comp,phy,rootphy);
  allSameDisp.emplace_back();
  nonLinearSameDisplacementBC& bc = allSameDisp.back();

  bc.comp = comp;
  bc.tagRoot = rootphy;
  bc.gRoot = new groupOfElements(0,rootphy);
  bc._tag = phy;
  bc.fac = fac;

  const std::string node("Node");
  const std::string edge("Edge");
  const std::string face("Face");
  const std::string volume("Volume");

  if(onwhat==node){
    bc.onWhat=nonLinearBoundaryCondition::ON_VERTEX;
    bc.g = new groupOfElements(0,phy);
  }
  else if (onwhat == edge){
    bc.onWhat=nonLinearBoundaryCondition::ON_EDGE;
    bc.g = new groupOfElements(1,phy);
  }
  else if (onwhat == face){
    bc.onWhat=nonLinearBoundaryCondition::ON_FACE;
    bc.g = new groupOfElements(2,phy);
  }
  else if (onwhat == volume){
    bc.onWhat=nonLinearBoundaryCondition::ON_VOLUME;
    bc.g = new groupOfElements(3,phy);
  }
  else{
    Msg::Error("Impossible to prescribe a constraint on a %s\n",onwhat.c_str());
  }
}

void nonLinearMechSolver::sameDisplacementBCBetweenTwoGroups(std::string onwhat, const int phy1, const int phy2, const int comp){
  Msg::Info("Displacement comp %d on physical %d is the same as on physical %d",comp,phy1,phy2);
  allSameDispBetweenTwoGroups.emplace_back();
  nonLinearSameDisplacementBCBetweenTwoGroups& bc = allSameDispBetweenTwoGroups.back();

  bc.comp = comp;
  bc._tag = phy1;

  const std::string node("Node");
  const std::string edge("Edge");
  const std::string face("Face");

  if(onwhat==node){
    bc.onWhat=nonLinearBoundaryCondition::ON_VERTEX;
    bc.g1 = new groupOfElements(0,phy1);
    bc.g2 = new groupOfElements(0,phy2);
  }
  else if (onwhat == edge){
    bc.onWhat=nonLinearBoundaryCondition::ON_EDGE;
    bc.g1= new groupOfElements(1,phy1);
    bc.g2= new groupOfElements(1,phy2);
  }
  else if (onwhat == face){
    bc.onWhat=nonLinearBoundaryCondition::ON_FACE;
    bc.g1 = new groupOfElements(2,phy1);
    bc.g2 = new groupOfElements(2,phy2);
  }
  else{
    Msg::Error("Impossible to prescribe a constraint on a %s\n",onwhat.c_str());
  }
};

void nonLinearMechSolver::fixOnFace(std::string onwhat, const int phy, const double A, const double B, const double C, const double D){
  allFixAllFace.emplace_back(A,B,C,D);
  nonLinearFixOnFaceBC& bc = allFixAllFace.back();
  const std::string node("Node");
  const std::string edge("Edge");
  const std::string face("Face");
  const std::string volume("Volume");

  bc._tag = phy;

  if(onwhat==node){
    bc.onWhat=nonLinearBoundaryCondition::ON_VERTEX;
    bc.g = new groupOfElements(0,phy);
  }
  else if (onwhat == edge){
    bc.onWhat=nonLinearBoundaryCondition::ON_EDGE;
    bc.g = new groupOfElements(1,phy);
  }
  else if (onwhat == face){
    bc.onWhat=nonLinearBoundaryCondition::ON_FACE;
    bc.g = new groupOfElements(2,phy);
  }
  else if (onwhat == volume){
    bc.onWhat=nonLinearBoundaryCondition::ON_VOLUME;
    bc.g = new groupOfElements(3,phy);
  }
  else{
    Msg::Error("Impossible to prescribe a constraint on a %s\n",onwhat.c_str());
  }


};

void nonLinearMechSolver::symetryBC(std::string onwhat, const int phy){
  allFixAllFace.emplace_back();
  nonLinearFixOnFaceBC& bc = allFixAllFace.back();
  const std::string face("Face"); // for 3D

  bc._tag = phy;
  if (onwhat == face){
    bc.onWhat=nonLinearBoundaryCondition::ON_FACE;
    bc.g = new groupOfElements(2,phy);

    if (bc.g->size() == 0) Msg::Error("physical %d has not element",phy);
    MElement* ele = *(bc.g->begin());

    SPoint3 p1 = ele->getVertex(0)->point();
    SPoint3 p2 = ele->getVertex(1)->point();
    SVector3 u12(p2,p1);
    SPoint3 p3 = ele->getVertex(2)->point();
    SVector3 u13(p3,p1);

    SVector3 n = crossprod(u12,u13);
    if (n.norm() <= 0.) Msg::Info("zero normal to fixed face");
    n.normalize();

    bc.A = n(0);
    bc.B = n(1);
    bc.C = n(2);
    bc.D = -(n(0)*p1.x()+n(1)*p1.y()+n(2)*p1.z());
  }
  else{
    Msg::Error("Impossible to prescribe a constraint on a %s\n",onwhat.c_str());
  }
};

void nonLinearMechSolver::applySameDispBC(){
  for(std::list<nonLinearSameDisplacementBC>::const_iterator sdbc=allSameDisp.begin(); sdbc!=allSameDisp.end(); ++sdbc){
		if (isDgDomain()){
      MElement* e = *(sdbc->gRoot->begin());
      MVertex* vroot = e->getVertex(0);

      MElement* eleRoot = NULL;
      FunctionSpaceBase* spaceRoot = NULL;
      int vpos = 0.;

      for (int i=0; i< domainVector.size(); i++){
        partDomain* dom = domainVector[i];
        for (groupOfElements::elementContainer::iterator it = dom->g_cbegin(); it != dom->g_cend(); it++){
          MElement* ele = *it;
          for (int j=0; j< ele->getNumVertices(); j++){
            MVertex* vj = ele->getVertex(j);
            if (vj->getNum() == vroot->getNum()){
               eleRoot = ele;
               spaceRoot = dom->getFunctionSpace();
               vpos = j;
               Msg::Info("found root element %d for root node %d vpos = %d",eleRoot->getNum(),vroot->getNum(),vpos);
               break;
            }
          }
          if (eleRoot != NULL) break;
        }
        if (eleRoot != NULL) break;
      }

      if (eleRoot == NULL) Msg::Error("root vertex is wrongly defined in nonLinearMechSolver::applySameDispBC");
      std::vector<Dof> RRoot;
      std::vector<Dof> dofRoot;
      spaceRoot->getKeys(eleRoot,RRoot);
      dofRoot.push_back(RRoot[vpos+eleRoot->getNumVertices()*sdbc->comp]);

      // fill domain
      std::set<Dof> others;
      for (groupOfElements::elementContainer::iterator itg = sdbc->g->begin(); itg!=sdbc->g->end(); itg++){
        MElement* eleBound = *itg;
        std::vector<int> vvBound(eleBound->getNumPrimaryVertices());
        for (int iv=0; iv < eleBound->getNumPrimaryVertices(); iv++){
          vvBound[iv] = eleBound->getVertex(iv)->getNum();
        }

        // find bulk element

        MElement* eleBulk = NULL;
        FunctionSpaceBase* spaceBulk = NULL;

        for (int i=0; i< domainVector.size(); i++){
          partDomain* dom = domainVector[i];
          for (groupOfElements::elementContainer::iterator it = dom->g_cbegin(); it != dom->g_cend(); it++){
            MElement* eleb = *it;
            std::set<int> vvb;
            for (int iv=0; iv < eleb->getNumPrimaryVertices(); iv++){
              vvb.insert(eleb->getVertex(iv)->getNum());
            }

            bool isFound = true;
            for (int iv=0; iv< vvBound.size(); iv++){
              if (vvb.find(vvBound[iv]) == vvb.end()){
                isFound = false;
                break;
              }
            }

            if (isFound){
              eleBulk = eleb;
              spaceBulk = dom->getFunctionSpace();
              break;
            }

          }
          if (eleBulk != NULL) break;
        }

        if (eleBulk == NULL) Msg::Error("nonLinearSameDisplacementBC is wrongly defined nonLinearMechSolver::applySameDispBC");
        std::vector<Dof> R;
        spaceBulk->getKeys(eleBulk,R);
        int nbFFb = eleBulk->getNumShapeFunctions();

        std::vector<int> positionBulk;
        for (int iv=0; iv< eleBound->getNumVertices(); iv++){
          MVertex* vi = eleBound->getVertex(iv);
          //
          if (vi->getNum() != vroot->getNum()){
            for (int jv=0; jv < eleBulk->getNumVertices(); jv++){
              MVertex* vj = eleBulk->getVertex(jv);
              if (vi->getNum() == vj->getNum()){
                positionBulk.push_back(jv);
                break;
              }
            }
          }
        }

        for (int k=0; k<positionBulk.size(); k++){
          Dof*  D = &R[positionBulk[k]+nbFFb*sdbc->comp];
          if (others.find(*D) == others.end()){
            others.insert(*D);
          }
        }
      }

      DofAffineConstraint<double> dof;
      dof.linear.push_back(std::pair<Dof,double>(dofRoot[0],sdbc->fac));
      dof.shift = 0.;


      for (std::set<Dof>::iterator itd = others.begin(); itd != others.end(); itd++){
				if (!pAssembler->isConstrained(*itd) and !pAssembler->isFixed(*itd)){
        	pAssembler->setLinearConstraint(*itd,dof);
				}
      };

		}
    else{
      // because of CG
      FunctionSpaceBase* spb = domainVector[0]->getFunctionSpace();
      FilterDof* fiterDof = domainVector[0]->createFilterDof(sdbc->comp);

      // get Root Dof
      MElement* e = *(sdbc->gRoot->begin());
      MVertex* vroot = e->getVertex(0);
      std::vector<Dof> R, dofRoot;
      spb->getKeys(e,R);
      for (int idof=0; idof<R.size(); idof++){
        if (fiterDof->operator()(R[idof])){
          dofRoot.push_back(R[idof]);
          break;
        }
      }

      std::vector<Dof> others;
      // get constraint Dof
      for (groupOfElements::vertexContainer::iterator it = sdbc->g->vbegin(); it!=sdbc->g->vend(); it++){
        MVertex* vv = *it;
        if (vv->getNum() != vroot->getNum()){
          MPoint point(*it);
          R.clear();
          spb->getKeys(&point,R);
          for (int idof=0; idof<R.size(); idof++){
            if (fiterDof->operator()(R[idof])){
              others.push_back(R[idof]);
              break;
            }
          }
        }
      }

      DofAffineConstraint<double> dof;
      dof.linear.push_back(std::pair<Dof,double>(dofRoot[0],sdbc->fac));
      dof.shift = 0.;

      for (int j =0; j<others.size(); j++){
				if (!pAssembler->isConstrained(others[j]) and !pAssembler->isFixed(others[j])){
        	pAssembler->setLinearConstraint(others[j],dof);
				}
      };
    };
  };

};

void nonLinearMechSolver::applyFixOnFace(){
  if (allFixAllFace.size() > 0){
    Msg::Info("begin applying fix  on face BC");
  }
  if (isDgDomain()){
    //for (const nonLinearFixOnFaceBC & fofbc : allFixAllFace){
    for(std::list<nonLinearFixOnFaceBC>::const_iterator fofbc=allFixAllFace.begin(); fofbc!=allFixAllFace.end(); fofbc++){
      double A = fofbc->A;
      double B = fofbc->B;
      double C = fofbc->C;
      double D = fofbc->D;

      for (groupOfElements::elementContainer::iterator it = fofbc->g->begin();
                it!= fofbc->g->end(); it++){
        MElement* ele = *it;
        MElement* eleFound = NULL;
        partDomain* domFound = NULL;
        bool found = false;

        std::vector<MVertex*> vvMaster;
        ele->getVertices(vvMaster);

        for (int idom =0; idom< domainVector.size(); idom ++){
          partDomain* aldo = domainVector[idom];
          for (groupOfElements::elementContainer::iterator itedom = aldo->g_cbegin(); itedom != aldo->g_cend(); itedom++){
            MElement* receiveEle = *itedom;
            int numFace = receiveEle->getNumFaces();
            for (int kk=0; kk<numFace; kk++){
              std::vector<MVertex*> faceVer;
              receiveEle->getFaceVertices(kk,faceVer);
              if (ele->getType() == TYPE_TRI){
                if ((vvMaster[0] == faceVer[0] and vvMaster[1] == faceVer[1] and vvMaster[2] == faceVer[2]) or
                    (vvMaster[0] == faceVer[0] and vvMaster[1] == faceVer[2] and vvMaster[2] == faceVer[1]) or
                    (vvMaster[0] == faceVer[1] and vvMaster[1] == faceVer[0] and vvMaster[2] == faceVer[2]) or
                    (vvMaster[0] == faceVer[1] and vvMaster[1] == faceVer[2] and vvMaster[2] == faceVer[0]) or
                    (vvMaster[0] == faceVer[2] and vvMaster[1] == faceVer[1] and vvMaster[2] == faceVer[0]) or
                    (vvMaster[0] == faceVer[2] and vvMaster[1] == faceVer[0] and vvMaster[2] == faceVer[1])){
                  found = true;
                  eleFound = receiveEle;
                  domFound = aldo;
                  break;
                }
              }
              else if (ele->getType() == TYPE_QUA){
                if ((vvMaster[0] == faceVer[0] and vvMaster[1] == faceVer[1] and vvMaster[2] == faceVer[2]) or
                        (vvMaster[0] == faceVer[1] and vvMaster[1] == faceVer[2] and vvMaster[2] == faceVer[3]) or
                        (vvMaster[0] == faceVer[2] and vvMaster[1] == faceVer[3] and vvMaster[2] == faceVer[0]) or
                        (vvMaster[0] == faceVer[3] and vvMaster[1] == faceVer[0] and vvMaster[2] == faceVer[1]) or
                        (vvMaster[0] == faceVer[0] and vvMaster[1] == faceVer[3] and vvMaster[2] == faceVer[2]) or
                        (vvMaster[0] == faceVer[3] and vvMaster[1] == faceVer[2] and vvMaster[2] == faceVer[1]) or
                        (vvMaster[0] == faceVer[2] and vvMaster[1] == faceVer[1] and vvMaster[2] == faceVer[0]) or
                        (vvMaster[0] == faceVer[1] and vvMaster[1] == faceVer[0] and vvMaster[2] == faceVer[3])){

                  found = true;
                  eleFound = receiveEle;
                  domFound = aldo;
                  break;
                };
              }
              else{
                Msg::Warning("Element type %d is not implemented",ele->getType());
              }
            }

            if (found)  break;
          }
          if (found) break;
        };

        if (!found) Msg::Error("ele %d is not found",ele->getNum());

        int nbFF = eleFound->getNumShapeFunctions();
        std::vector<Dof> keys;
        domFound->getFunctionSpace()->getKeys(eleFound,keys);

        for (int iver = 0; iver < vvMaster.size(); iver++){
          MVertex* v = vvMaster[iver];
          int idex = -1;
          for (int ii=0; ii < eleFound->getNumVertices(); ii++){
            if (eleFound->getVertex(ii)->getNum() == v->getNum()){
              idex = ii;
              break;
            }
          }
          std::vector<Dof>  R;
          R.push_back(keys[idex]);
          R.push_back(keys[idex+nbFF]);
          R.push_back(keys[idex+2*nbFF]);

          DofAffineConstraint<double> dof;
          double val = -1.*(A*v->x()+B*v->y()+C*v->z()+D);

          if (fabs(A) > 1e-10 and !pAssembler->isFixed(R[0])){
            if (!pAssembler->isFixed(R[1]) or !pAssembler->isFixed(R[2])){
              dof.shift = val/A;
              dof.linear.push_back(std::pair<Dof,double>(R[1],-B/A));
              dof.linear.push_back(std::pair<Dof,double>(R[2],-C/A));
              pAssembler->setLinearConstraint(R[0],dof);
            }

          }
          else if (fabs(B) > 1e-10 and !pAssembler->isFixed(R[1])){
            if (!pAssembler->isFixed(R[0]) or !pAssembler->isFixed(R[2])){
              dof.shift = val/B;
              dof.linear.push_back(std::pair<Dof,double>(R[0],-A/B));
              dof.linear.push_back(std::pair<Dof,double>(R[2],-C/B));
              pAssembler->setLinearConstraint(R[1],dof);
            }

          }
          else if (fabs(C) > 1e-10 and !pAssembler->isFixed(R[2])){
            if (!pAssembler->isFixed(R[0]) or !pAssembler->isFixed(R[1])){
              dof.shift = val/C;
              dof.linear.push_back(std::pair<Dof,double>(R[0],-A/C));
              dof.linear.push_back(std::pair<Dof,double>(R[1],-B/C));
              pAssembler->setLinearConstraint(R[2],dof);
            }
          }
          //else
          //  Msg::Error("error in nonLinearMechSolver::applyFixOnFace");
        }
      }
    }
  }
  else{
    for(std::list<nonLinearFixOnFaceBC>::const_iterator fofBC=allFixAllFace.begin(); fofBC!=allFixAllFace.end(); fofBC++){
      double A = fofBC->A;
      double B = fofBC->B;
      double C = fofBC->C;
      double D = fofBC->D;

      // because of CG
      for (groupOfElements::vertexContainer::iterator it = fofBC->g->vbegin();
                it!= fofBC->g->vend(); it++){
        MVertex* v = *it;
        int idex = -1;
        MElement* ele = NULL;
        partDomain* dom = NULL;
        bool found = false;

        for (int idom =0; idom< domainVector.size(); idom ++){
          partDomain* aldo = domainVector[idom];
          for (groupOfElements::elementContainer::iterator itedom = aldo->g_cbegin(); itedom != aldo->g_cend(); itedom++){
            MElement* etemp = *itedom;
            for (int iv = 0; iv < etemp->getNumVertices(); iv++){
              if (etemp->getVertex(iv)->getNum() == v->getNum()){
                found = true;
                ele = etemp;
                dom = aldo;
                idex = iv;
                break;
              }
            }
            if (found)  break;
          }
          if (found) break;
        };

        if (!found) Msg::Error("vertex %d is not found",v->getNum());

        int nbFF = ele->getNumShapeFunctions();
        std::vector<Dof> keys;
        dom->getFunctionSpace()->getKeys(ele,keys);
        std::vector<Dof>  R;
        R.push_back(keys[idex]);
        R.push_back(keys[idex+nbFF]);
        R.push_back(keys[idex+2*nbFF]);

        DofAffineConstraint<double> dof;
        double val = -1.*(A*v->x()+B*v->y()+C*v->z()+D);

        if (fabs(A) > 1e-10 and !pAssembler->isFixed(R[0])){
          if (!pAssembler->isFixed(R[1]) or !pAssembler->isFixed(R[2])){
            dof.shift = val/A;
            dof.linear.push_back(std::pair<Dof,double>(R[1],-B/A));
            dof.linear.push_back(std::pair<Dof,double>(R[2],-C/A));
            pAssembler->setLinearConstraint(R[0],dof);
          }

        }
        else if (fabs(B) > 1e-10 and !pAssembler->isFixed(R[1])){
          if (!pAssembler->isFixed(R[0]) or !pAssembler->isFixed(R[2])){
            dof.shift = val/B;
            dof.linear.push_back(std::pair<Dof,double>(R[0],-A/B));
            dof.linear.push_back(std::pair<Dof,double>(R[2],-C/B));
            pAssembler->setLinearConstraint(R[1],dof);
          }

        }
        else if (fabs(C) > 1e-10 and !pAssembler->isFixed(R[2])){
          if (!pAssembler->isFixed(R[0]) or !pAssembler->isFixed(R[1])){
            dof.shift = val/C;
            dof.linear.push_back(std::pair<Dof,double>(R[0],-A/C));
            dof.linear.push_back(std::pair<Dof,double>(R[1],-B/C));
            pAssembler->setLinearConstraint(R[2],dof);
          }
        }

      }

    };
  };

  if (allFixAllFace.size() > 0){
    Msg::Info("end applying fix  on face BC");
  }
};

void nonLinearMechSolver::archivingNode(const int numphys, const int comp, nonLinearBoundaryCondition::whichCondition wc,const int nstep){
  // no distinction between cG/dG and full Dg formulation. class Displacement Field manage it
  
  groupOfElements g(0,numphys);

  if(g.vsize() >0){
    for (groupOfElements::vertexContainer::iterator it =g.vbegin(); it!= g.vend(); it++){
      MVertex* v = *it;
      bool found = false;
      for (int i=0; i< anoded.size(); i++)
      {
        if (anoded[i].physnum == numphys and anoded[i].nodenum == v->getNum() and anoded[i]._comp == comp and anoded[i].wc == wc)
        {
          found = true;
          break;
        }
      }
      
      if (!found)
      {
        anoded.emplace_back(numphys,v->getNum(),comp,wc,nstep);
      }
    }

  }
  else{
    Msg::Warning("No physical group %d So it is impossible to archive nodal displacement",numphys);
  }
}

void nonLinearMechSolver::archivingNode(const int numpphys, const int dim, const int comp, nonLinearBoundaryCondition::whichCondition wc,const int nstep){
  groupOfElements g(dim,numpphys);
  FILE* file = NULL;
  #if defined(HAVE_MPI)
  if (Msg::GetCommRank() == 0)
  #endif
  {
    std::ostringstream ossnumver;
    ossnumver << numpphys;
    std::string ss = ossnumver.str();
    std::string fnam = "numver" + ss+ ".csv";
    file = fopen(fnam.c_str(),"w");
  }

  for (groupOfElements::vertexContainer::iterator it =g.vbegin(); it!= g.vend(); it++){
    MVertex* vv = *it;
    bool found = false;
    for (int i=0; i< anoded.size(); i++)
    {
      if (anoded[i].physnum == numpphys and anoded[i].nodenum == vv->getNum() and anoded[i]._comp == comp and anoded[i].wc == wc)
      {
        found = true;
        break;
      }
    }
    
    if (!found)
    {
      anoded.emplace_back(numpphys,vv->getNum(),comp,wc,nstep);
    }
    #if defined(HAVE_MPI)
    if (Msg::GetCommRank() == 0)
    #endif
    {
      fprintf(file,"%d ;%f;%f;%f\n",vv->getNum(),vv->x(),vv->y(),vv->z());
    }
  }
  #if defined(HAVE_MPI)
  if (Msg::GetCommRank() == 0)
  #endif
  {
    fclose(file);
  }
};


void nonLinearMechSolver::archivingRigidContact(const int numphys, const int comp, const int wc,const int nstep){
  // For rigid contact the value are disponible on all rank --> archiving only on last rank
  #if defined(HAVE_MPI)
  if(Msg::GetCommRank() == Msg::GetCommSize() -1)
  #endif// HAVE_MPI
  {
    nonLinearBoundaryCondition::whichCondition whc;
    switch(wc){
     case 0:
      whc = nonLinearBoundaryCondition::position;
      break;
     case 1:
      whc = nonLinearBoundaryCondition::velocity;
      break;
     case 2:
      whc = nonLinearBoundaryCondition::acceleration;
      break;
     default:
      whc = nonLinearBoundaryCondition::position;
    }
    
    bool found = false;
    for (int i=0; i< anoded.size(); i++)
    {
      if (anoded[i].physnum == numphys and anoded[i].nodenum == numphys and 
          anoded[i]._comp == comp and anoded[i].wc == whc and 
          anoded[i].iscontact)
      {
        found = true;
        break;
      }
    }
    if (!found)
    {
      anoded.emplace_back(numphys,numphys,comp,whc,nstep,true);
    }
  }
}

void nonLinearMechSolver::archivingNodeIP(const int numphys, const int ipval,const int elemval,const int nstep){
  vaip.push_back(IPField::ip2archive(0, numphys,ipval,elemval,nstep));
}

void nonLinearMechSolver::archivingNodeIPOnPhysicalGroup(const int dim, const int numphys, const int ipval, const int nstep){
  groupOfElements g(dim,numphys);
  if (g.vsize() > 0){
    FILE* ff = NULL;
    #if defined(HAVE_MPI)
    if (Msg::GetCommRank() ==0)
    #endif // HAVE_MPI
    {
      ff = fopen("archivingNodeIPOnPhysicalGroup_vertexList.csv","w");
      fprintf(ff,"Num;X;Y;Z\n");
    }
    for (groupOfElements::vertexContainer::iterator it = g.vbegin(); it!= g.vend(); it++){
      MVertex* v = *it;
      vaip.push_back(IPField::ip2archive(ipval,nstep,v));
      if (ff != NULL){
        fprintf(ff,"%d;%e;%e;%e\n",v->getNum(),v->x(),v->y(),v->z());
      }
    }
    if (ff != NULL){
      fclose(ff);
    }
  }
};

void nonLinearMechSolver::archivingInterfaceElementIP(const int elenumMinus,const int elenumPlus, const int ipval,const int elemval, const int numip, const int nstep){
  vaip.push_back(IPField::ip2archive(-1, elenumMinus, elenumPlus,ipval,elemval,nstep,numip));
};

void nonLinearMechSolver::archivingElementIP(const int elenum,const int ipval,const int elemval, const int numip, const int nstep){
  // save data on numip on elemement elenum
  vaip.push_back(IPField::ip2archive(-1, elenum, elenum,ipval,elemval,nstep,numip));
};

void nonLinearMechSolver::archivingIPOnPhysicalGroup(std::string onwhat, const int numphys, const int ipval,const int elemval, const int nstep){

  const std::string node("Node");
  const std::string edge("Edge");
  const std::string face("Face");
  const std::string volume("Volume");
  int dim= 0;
  if(onwhat==node){
     dim =0;
  }
  else if(onwhat==edge){
     dim =1;
  }
  else if(onwhat==face){
     dim =2;
  }
  else if(onwhat==volume){
     dim =3;
  }
  else Msg::Error("Impossible to prescribe a constraint on a %s\n",onwhat.c_str());
  vaip.push_back(IPField::ip2archive(dim, numphys,ipval,elemval,nstep));
}

void nonLinearMechSolver::contactInteraction(contactDomain *cdom){
  _allContact.insert(cdom);

}
void nonLinearMechSolver::defoDefoContactInteraction(defoDefoContactDomain *cdom){
  _allDefoDefoContact.insert(cdom);

}


// create an element for each element type
// which will be used to compute the time step
void nonLinearMechSolver::initMapMElementToComputeTimeStep()
{
  MElementFactory factory;
  std::vector<MVertex*> vver;
  std::map<int,MElement*>::iterator ite;
  for(int i=0;i< _mpiUserDom;i++)
  {
    partDomain * dom = domainVector[i];
    for(groupOfElements::elementContainer::iterator it=dom->g_cbegin(); it!=dom->g_cend(); ++it){
      MElement *ele = *it;
      ite = _mapMElementTmp.find(ele->getTypeForMSH());
      if(ite==_mapMElementTmp.end()) // create one new element all vertex are in (0,0,0) They will be filled later
      {
        vver.clear();
        for(int j=0;j<ele->getNumVertices();j++)
        {
          vver.push_back(new MVertex(0.,0.,0.));
        }
        _mapMElementTmp.insert(std::pair<int,MElement*>(ele->getTypeForMSH(),factory.create(ele->getTypeForMSH(),vver)));
      }
      ite = _mapMElementFirstOrder.find(ele->getType());
      if(ite == _mapMElementFirstOrder.end())
      {
        vver.clear();
        for(int j=0;j<ele->getNumPrimaryVertices();j++)
        {
          vver.push_back(new MVertex(0.,0.,0.));
        }
        MElement *eletmp;
        switch(ele->getType())
        {
         case TYPE_PNT:
          eletmp = new MPoint(vver[0]); break;
         case TYPE_LIN:
          eletmp = factory.create(MSH_LIN_2,vver); break;
         case TYPE_TRI:
          eletmp = factory.create(MSH_TRI_3,vver); break;
         case TYPE_QUA:
          eletmp = factory.create(MSH_QUA_4,vver); break;
         case TYPE_TET:
          eletmp = factory.create(MSH_TET_4,vver); break;
         case TYPE_PYR:
          eletmp = factory.create(MSH_PYR_5,vver); break;
         case TYPE_PRI:
          eletmp = factory.create(MSH_PRI_6,vver); break;
         case TYPE_HEX:
          eletmp = factory.create(MSH_HEX_8,vver); break;
         case TYPE_POLYG:
          eletmp = factory.create(MSH_POLYG_,vver); break;
         case TYPE_POLYH:
          eletmp = factory.create(MSH_POLYH_,vver); break;
         default:
          Msg::Error("unknown element type %d",ele->getType());
        }
        _mapMElementFirstOrder.insert(std::pair<int,MElement*>(ele->getType(),eletmp));
      }
    }
  }
}

double nonLinearMechSolver::initialCriticalExplicitTimeStep(){
	// evaluate the critical time step carachteristic length = hs
  double hs;
  double dtmin=1.e100; // Initialization to infinity
  for(int i=0; i< _mpiUserDom; i++){
    partDomain *dom = domainVector[i];
    if(dom->groupOfElementsSize() > 0){ // otherwise no element (interface domain and no need to compute time step)
      double hsmin= 1.e100; // Initialization to infinity
      FunctionSpaceBase *spdom = dom->getFunctionSpace();
      std::vector<Dof> R;
      std::vector<double> disp;
      materialLaw *mlt = dom->getMaterialLaw();
      double celerity = mlt->soundSpeed();
      double scaleTimeStep = dom->scaleTimeStep();
      for(groupOfElements::elementContainer::iterator it=dom->g_cbegin(); it!=dom->g_cend(); ++it){
				MElement *e = *it;
       // nodal displacement 3 per nodes FIX THIS ASSUMPTION

				spdom->getKeys(e,R);
				int nkeys = spdom->getNumKeys(e);
				int nbvertex = e->getNumVertices();

				disp.resize(R.size());
				for (int iR = 0; iR < R.size(); iR++){
					disp[iR] = 0.;
				}
				MElement *etmp = NULL;
				if((e->getTypeForMSH()!=MSH_TET_10) and (e->getTypeForMSH()!=MSH_TET_20) and (e->getTypeForMSH()!=MSH_HEX_27)) // some functions are undefined in 3D so cannot use new method!
				{
					etmp = _mapMElementTmp[e->getTypeForMSH()];
          for(int j=0;j<nbvertex;j++)
					{
						//for(int jj=0;jj<nkeyspernodes;jj++){}
						MVertex *ver = e->getVertex(j);
						MVertex *vertmp = etmp->getVertex(j);
						vertmp->x() = ver->x();
						vertmp->y() = ver->y();
						vertmp->z() = ver->z();
					}
				}

        // Unfortunally, computation of characteristic length depends on the element type
        // ADD THIS FUNCTIONS directly in gmsh (method of MELEMENT)
        // Problem as inner radius is defined only for 1st order element
        switch(e->getTypeForMSH()){
         case MSH_QUA_9 :
          hs = quadrangle2orderCharacteristicSize(etmp,_mapMElementFirstOrder);
          break;
         case MSH_QUA_16:
          hs = quadrangle3orderCharacteristicSize(etmp,_mapMElementFirstOrder);
          break;
         case MSH_TRI_6 :
          hs = triangular2orderCharacteristicSize(etmp,_mapMElementFirstOrder);
          break;
         case MSH_TRI_10:
          hs = triangular3orderCharacteristicSize(etmp,_mapMElementFirstOrder);
          break;
         case MSH_TET_10:
          hs = tetrahedron2orderCharacteristicSize(e,disp);
          break;
         case MSH_TET_20:
          hs = tetrahedron3orderCharacteristicSize(e,disp);
          break;
         case MSH_HEX_27:
          hs = hexahedron2orderCharacteristicSize(e,disp);
          break;
         default :
          //default criteria innerRadius()/polynomialOrder() as innerRAdius is defined for first order only !!
          // keep divide by polynomialOrder to be general for undefined case and add 2 for security
          hs = etmp->getInnerRadius()/e->getPolynomialOrder()/2.;
        }
        if(hs<0)
        {
          Msg::Error("Negative characteristic size for element %d on rank %d",e->getNum(),Msg::GetCommRank());
        }
        if(hs<hsmin) hsmin = hs;
        R.clear(); disp.clear();
      }
      double dt = scaleTimeStep*hsmin/celerity;  
      if(dt < dtmin ) dtmin = dt;
    }
  }
  return dtmin;
};

bool nonLinearMechSolver::considerForTimeStep(MElement* ele) const{
  if (!getElementErosionFilter()(ele)) return false; // if element already eliminate--> do not account for
  const AllIPState::ipstateElementContainer *vips = _ipf->getAips()->getIPstate(ele->getNum()); // if one damage element is still work
  bool cons = false;
  for(int j=0;j<vips->size();j++){
    IPVariable *ipv = (*vips)[j]->getState(IPStateBase::current);
    if(ipv->get(IPField::DAMAGE)<0.9)
    {
      cons = true;
      break;
    }
  }
  return cons;
};

double nonLinearMechSolver::criticalExplicitTimeStep(){
  // evaluate the critical time step carachteristic length = hs
  double hs;
  double dtmin=1.e100; // Initialization to infinity
  for(int i=0; i< _mpiUserDom; i++){
    partDomain *dom = domainVector[i];
    if(dom->groupOfElementsSize() > 0){ // otherwise no element (interface domain and no need to compute time step)
      double hsmin= 1.e100; // Initialization to infinity
      FunctionSpaceBase *spdom = dom->getFunctionSpace();
      std::vector<Dof> R;
      std::vector<double> disp;
      materialLaw *mlt = dom->getMaterialLaw();
      double celerity = mlt->soundSpeed();
      double scaleTimeStep = dom->scaleTimeStep();
      for(groupOfElements::elementContainer::iterator it=dom->g_cbegin(); it!=dom->g_cend(); ++it){
       MElement *e = *it;
       // nodal displacement 3 per nodes FIX THIS ASSUMPTION
       if(this->considerForTimeStep(e))
       {
				spdom->getKeys(e,R);
				int nkeys = spdom->getNumKeys(e);
				int nbvertex = e->getNumVertices();

				_ufield->get(R,disp);

				MElement *etmp = NULL;
				if((e->getTypeForMSH()!=MSH_TET_10) and (e->getTypeForMSH()!=MSH_TET_20) and (e->getTypeForMSH()!=MSH_HEX_27)) // some functions are undefined in 3D so cannot use new method!
				{
					etmp = _mapMElementTmp[e->getTypeForMSH()];
					for(int j=0;j<nbvertex;j++)
					{
						//for(int jj=0;jj<nkeyspernodes;jj++){}
						MVertex *ver = e->getVertex(j);
						MVertex *vertmp = etmp->getVertex(j);
						vertmp->x() = ver->x() + disp[j];
						vertmp->y() = ver->y() + disp[j+nbvertex];
						vertmp->z() = ver->z() + disp[j+nbvertex+nbvertex];
					}
				}

        // Unfortunally, computation of characteristic length depends on the element type
        // ADD THIS FUNCTIONS directly in gmsh (method of MELEMENT)
        // Problem as inner radius is defined only for 1st order element
        switch(e->getTypeForMSH()){
         case MSH_QUA_9 :
          hs = quadrangle2orderCharacteristicSize(etmp,_mapMElementFirstOrder);
          break;
         case MSH_QUA_16:
          hs = quadrangle3orderCharacteristicSize(etmp,_mapMElementFirstOrder);
          break;
         case MSH_TRI_6 :
          hs = triangular2orderCharacteristicSize(etmp,_mapMElementFirstOrder);
          break;
         case MSH_TRI_10:
          hs = triangular3orderCharacteristicSize(etmp,_mapMElementFirstOrder);
          break;
         case MSH_TET_10:
          hs = tetrahedron2orderCharacteristicSize(e,disp);
          break;
         case MSH_TET_20:
          hs = tetrahedron3orderCharacteristicSize(e,disp);
          break;
         case MSH_HEX_27:
          hs = hexahedron2orderCharacteristicSize(e,disp);
          break;
         default :
          //default criteria innerRadius()/polynomialOrder() as innerRAdius is defined for first order only !!
          // keep divide by polynomialOrder to be general for undefined case and add 2 for security
          hs = etmp->getInnerRadius()/e->getPolynomialOrder()/2.;
        }
        if(hs<0)
        {
          Msg::Error("Negative characteristic size for element %d on rank %d",e->getNum(),Msg::GetCommRank());
        }
        if(hs<hsmin) hsmin = hs;
        R.clear(); disp.clear();
       }
      }
      double dt = scaleTimeStep*hsmin/celerity;  
      if(dt < dtmin ) dtmin = dt;
    }
  }

  // Delta t of contact
/*  explicitHCDofManager<double>* dynass = static_cast<explicitHCDofManager<double>*>(pAssembler);
  // filter because one component is needed (same mass in all direction)
  DgC0ShellFilterDofComponent filDof(0);
  for(nonLinearMechSolver::contactContainer::iterator itc = _allContact.begin(); itc!=_allContact.end(); ++itc){
    contactDomain *cdom = *itc;
    DgC0rigidContactSpace *csp = static_cast<DgC0rigidContactSpace*>(spaceManager->get(cdom));
    double penalty = cdom->getPenalty();
    for(groupOfElements::elementContainer::iterator it = cdom->gSlave->begin(); it!=cdom->gSlave->end(); ++it){
      MElement *ele = *it;
      // get mass
//      int nbdofs = csp->getNumKeys(ele);
      int nbvertex = ele->getNumVertices();
//      int nbdofsGC = csp->getNumKeysOfGravityCenter();
//      int nbcomp = (nbdofs-nbdofsGC)/nbvertex;
      std::vector<Dof> R;
      std::vector<double> vmass;
      csp->getKeys(ele,R);
      std::vector<Dof> Rfilter;
      filDof.filter(R,Rfilter);
      dynass->getVertexMass(Rfilter,vmass);
      double Mmaster = vmass[nbvertex];
      for(int i=0; i<nbvertex; i++){
        double dt = 10000.*sqrt(0.5*Mmaster*vmass[i]/penalty/(Mmaster+vmass[i]));
        if(dt < dtmin ){
          dtmin = dt;
          Msg::Info("Dtc is fixed by contact");
        }
      }
    }
  }*/
  return dtmin;
}

void nonLinearMechSolver::setInitialCondition(){
  dofManager<double>* currentManager = NULL;
  for(std::list<initialCondition>::iterator itic = allinitial.begin() ; itic != allinitial.end(); ++itic){
    initialCondition &initC = *itic;
    if(initC._comp != -1){
      // select the system if required
      if(_vcompBySys.size()==0)
      {
        currentManager = pAssembler;
      }
      else
      {
        dofManagerMultiSystems<double>* multiManager = static_cast<dofManagerMultiSystems<double>*>(pAssembler);
        currentManager = multiManager->getDofManagerByComp(initC._comp);
      }
      for(int j=0;j<initC._vspace.size();j++){
        SetInitialDofs(initC._vspace[j],initC._vgroup[j]->begin(),initC._vgroup[j]->end(),initC._mycondition,*currentManager,initC._f,*(initC._vfilter[j]),initC._vdg[j]);
      }
    }
    else{
      // select the system if required
      if(_vcompBySys.size()==0)
      {
        currentManager = pAssembler;
      }
      else
      {
        Msg::Warning("setNormalInitialDofs with a multisystem --> Do the hypothesis that displacement is the first system !");
        dofManagerMultiSystems<double>* multiManager = static_cast<dofManagerMultiSystems<double>*>(pAssembler);
        currentManager = multiManager->getManager(0);
      }
      for(int j=0;j<initC._vspace.size();j++){
        groupOfElements *groupDom = NULL;
        if(initC._vgroup[j] != initC._vDomGroup[j])
        {
          groupDom = initC._vDomGroup[j];
        }
        SetNormalInitialDofs(initC._vspace[j],initC._vgroup[j]->begin(),initC._vgroup[j]->end(),initC._mycondition,*currentManager,initC._f,*(initC._vfilter[j]),initC._vdg[j],groupDom);
      }
    }
  }
  // Eventually read file with initial position
/*  if(initOrResartfname != ""){
    // loop on all elements on all domain
    FunctionSpaceBase *sp; // to create the dofs
    std::vector<Dof> R;
    std::vector<double> val;
    for(std::vector<partDomain*>::iterator itdom = domainVector.begin(); itdom != domainVector.end(); ++itdom){
      partDomain *dom = *itdom;
      sp = spaceManager->get(dom);
      for(groupOfElements::elementContainer::iterator it = dom->g->begin(); it != dom->g->end(); ++it){
        MElement *ele = *it;
        // create dof of element
        sp->getKeys(ele,R);

        // clear
        R.clear();
      }

    }

  }*/
}

double nonLinearMechSolver::solveExplicit(){

  this->initializeExplicitScheme();
  double curtime = 0.;
  int ii=0; //current step already read in restart(this)
  bool forceTimeStepEvaluation=false;
  if(restartManager::available() and !_resetRestart)
  {
    Msg::Barrier(); // To wait rank0 before end
    restartManager::openRestartFile("nonLinearMechSolver");
    restartManager::restart(this);
    // restart takes only nonm const argument! So convert numstep and curtime
    restartManager::restart(ii);
    restartManager::restart(curtime);
    restartManager::closeRestartFile();
    Msg::Barrier(); // To wait rank0 before end
    _currentStep=ii;

    restartManager::InitFromRestartFile();
    Msg::Barrier(); // To wait rank0 before end
    struct stat st;
    if(stat("beforeRestart",&st) == 0)
    {
      std::ostringstream oss;
      oss << _currentStep;
      std::string cpDir="mv beforeRestart restart"+oss.str();
      system(cpDir.c_str());
    }
    forceTimeStepEvaluation=true;
  }
  _resetRestart=false;
  Msg::Barrier(); // To wait rank0 before end

  /* time loop */
  while(curtime<endtime){ // end via break for dynamic relaxation
    curtime = this->oneExplicitStep(curtime,forceTimeStepEvaluation);
  }

  /* end of scheme */
  return this->finalizeExplicitScheme(curtime);
}

void nonLinearMechSolver::initializeExplicitScheme()
{
  Msg::Info("Explicit Data: endtime = %e beta=%f gamma=%f alpha_m=%f",this->getEndTime(), this->_beta, this->_gamma, this->_alpham);

  /* init data */
  this->init();
  this->init2();
  staticDofManager<double> *sAssembler = static_cast<staticDofManager<double>*>(pAssembler);

  /* mass matrix computation outside the loop */
  for(std::vector<partDomain*>::iterator it = domainVector.begin(); it!=domainVector.end(); ++it){
    partDomain *dom = *it;
    AssembleMass(dom->getBilinearMassTerm(),*(dom->getFunctionSpace()),dom->g_cbegin(),dom->g_cend(),
               *(dom->getBulkGaussIntegrationRule()),*pAssembler);
  }
  if(_notResetedContact){
    for(nonLinearMechSolver::contactContainer::iterator itc = _allContact.begin(); itc!= _allContact.end(); ++itc){
      contactDomain *cdom = *itc;
      if(cdom->isRigid()){
        rigidContactSpaceBase *rcsp = static_cast<rigidContactSpaceBase*>(cdom->getSpace());
        AssembleMass(cdom->getMassTerm(),rcsp,pAssembler);
      }
    }
  }

  /* time initialization */
  sAssembler->nextStep();
  _currentStep = 0;
}

double nonLinearMechSolver::oneExplicitStep(const double curtime, bool &forceTimeStepEvaluation)
{
  // Make these to variables static as they have to be kept
  // betwwen 2 calls to this function
  static double timeStep=0.;
  static staticDofManager<double> *sAssembler = static_cast<staticDofManager<double>*>(pAssembler);

  // compute time step and current time
  if(_currentStep%_numstepExpl ==0 || forceTimeStepEvaluation){
    timeStep = _gammas * this->criticalExplicitTimeStep();
   #if defined(HAVE_MPI)
    if(Msg::GetCommSize()>1){
      // in case of mpi the time step is choosen as the minimal value compute on each partition
      double mpiTimeStep;
      MPI_Allreduce(&timeStep,&mpiTimeStep,1,MPI_DOUBLE,MPI_MIN,MPI_COMM_WORLD);
      timeStep = mpiTimeStep;
    }
   #endif // HAVE_MPI
    Msg::Info("time step value: %e at time %e",timeStep,curtime);
    sAssembler->setTimeStep(timeStep);
    forceTimeStepEvaluation=false;
  }
  double mytime= curtime + timeStep;

  if(mytime>endtime)  
  {
     mytime = endtime;
     timeStep = mytime - curtime;
  }

  this->oneStepPreSolve(mytime,timeStep,_currentStep+1);

  // solve via assembler to perform mpi operation (communication before and after systemSolve
  pAssembler->systemSolve();

  // compute the new forces
  _ipf->compute1state(IPStateBase::current,false);
  double normRHS = this->computeRightHandSide();
  if(std::isnan(normRHS))
  {
    Msg::Error("Nan force value --> end");
    return endtime; // To finish the loop
  }
  if(_dynamicRelaxation){
    double norm0 = sAssembler->norm0Inf();
    if(_currentStep%_numstepExpl ==0 ) Msg::Info("Rel norm for dynamic relaxation %e tol = %e",normRHS/norm0,_tol);
    if(normRHS/norm0 < _tol){
      Msg::Info("Dynamic relaxation procedure is converged. Return endtime to finish the loop");
      return endtime;
    }
    if(mytime == endtime)
      Msg::Info("Dynamic relaxation ended as the final time is reached. But convergence is not achieved!");
  }
  
  this->oneStepPostSolve(mytime,_currentStep+1);
  // current step
  _currentStep++;
  return mytime;
}

double nonLinearMechSolver::finalizeExplicitScheme(const double curtime)
{
  staticDofManager<double> *sAssembler = static_cast<staticDofManager<double>*>(pAssembler);
  sAssembler->nextStep(); // to make last current state at last explicit step
  this->endOfScheme(curtime,_currentStep);
  Msg::Info("Explicit OK");
  return curtime;
}

void nonLinearMechSolver::initializeStaticScheme()
{
  Msg::Info("SNL Data : nstep =%d endtime=%f",_numstepImpl,this->getEndTime());
  /* init data */
  this->init();
  this->init2();
  // iterative scheme (loop on timestep)
  staticDofManager<double> *sAssembler = static_cast<staticDofManager<double>*>(pAssembler);
  std::string name = "A";
  linearSystem<double>* lsys =pAssembler->getLinearSystem(name);
  nonLinearSystem<double>* nlsys = dynamic_cast<nonLinearSystem<double>*>(lsys);
  
  if (whatScheme == Implicit){
     Msg::Info("begin assembling mass matrix");
    /* mass matrix computation outside the loop */
    // zero matrix first
    nlsys->zeroMatrix(nonLinearSystem<double>::mass);
    
    for(std::vector<partDomain*>::iterator it = domainVector.begin(); it!=domainVector.end(); ++it){
      partDomain *dom = *it;
      AssembleMass(dom->getBilinearMassTerm(),*(dom->getFunctionSpace()),dom->g_cbegin(),dom->g_cend(),
                 *(dom->getBulkGaussIntegrationRule()),*pAssembler);
    }
    if(_notResetedContact){
      for(nonLinearMechSolver::contactContainer::iterator itc = _allContact.begin(); itc!= _allContact.end(); ++itc){
        contactDomain *cdom = *itc;
        if(cdom->isRigid()){
          rigidContactSpaceBase *rcsp = static_cast<rigidContactSpaceBase*>(cdom->getSpace());
          AssembleMass(cdom->getMassTerm(),rcsp,pAssembler);
        }
      }
    }
    Msg::Info("done assembling mass matrix");
  }

   

  if (whatScheme == Implicit){ //now only for implicit and dgdomain
    if (_noMassPredicorWithoutVelocityAndAcceleration)
    {
      Msg::Info("start no mass DOFs collecting");
      // collect all no mass DOFs
      std::set<Dof> nomassDOf;
      /// Loop on all domains
      for(std::vector<partDomain*>::iterator it=domainVector.begin(); it!=domainVector.end(); ++it){
        partDomain *dom = *it;
        // take back the keys of dofs
        dom->getNoMassDofs(nomassDOf);
      }
      Msg::Info("done collecting no mass DOF, total %d DOFs ",nomassDOf.size());
      std::vector<int>  noMasPosInGlobalVec(nomassDOf.size(),0);
      int idex = 0;
      for (std::set<Dof>::const_iterator itd = nomassDOf.begin(); itd != nomassDOf.end(); itd++){
        noMasPosInGlobalVec[idex] = pAssembler->getDofNumber(*itd);
        idex ++;
      }
      // transfer dof to non-linear system
      nlsys->setNoMassDofKeys(noMasPosInGlobalVec);
    }
  }

  if (_pathFollowing){
    this->computeLoadVector();
  }

  /* time initialization */
  // save solution of previous (initially last step == initial step BEAWARE FOR RESTART)
  sAssembler->nextStep();

  /* time loop */
  sAssembler->setTimeStep(double(endtime)/double(_numstepImpl)); // time step value prescribed by user (time step can not be greater than this value));
}

bool nonLinearMechSolver::localPathFollowingSwitching() const{
	// for test
	double pfCr = _ipf->computePathFollowingLocalValue(IPStateBase::previous);
  double Defor = _ipf->computeDeformationEnergy(IPStateBase::previous);
	#if defined(HAVE_MPI)
	if (Msg::GetCommSize()> 1){
    double Cr[2];
    Cr[0] = pfCr;
    Cr[1] = Defor;
		double totalCr[2];
		MPI_Allreduce(&Cr[0],&totalCr[0],2,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
    if (totalCr[1] > 0.)
      printf("rank %d pf = %e, deforEnerg = %e cr = %e impose value %e \n",Msg::GetCommRank(),totalCr[0],totalCr[1],totalCr[0]/totalCr[1],_pathFollowingSwitchCriterion);
		if (totalCr[0] > _pathFollowingSwitchCriterion*totalCr[1]) return true;
		else return false;
	}
	else
	#endif //HAVE_MPI
	{
    if (fabs(Defor) > 0.)
      printf("rank %d pf = %e, deforEnerg = %e cr = %e impose value %e \n",Msg::GetCommRank(),pfCr,fabs(Defor),pfCr/fabs(Defor),_pathFollowingSwitchCriterion);
		if (pfCr > _pathFollowingSwitchCriterion*fabs(Defor)) return true;
		else return false;
	}
};

double nonLinearMechSolver::computeLocalPathFollowingStep() const{
	// for adaptive purpose
	double localStep = _ipf->computePathFollowingLocalValue(IPStateBase::current) - _ipf->computePathFollowingLocalValue(IPStateBase::previous);
	// accumulate from all procs
	#if defined(HAVE_MPI)
	if (Msg::GetCommSize() > 1){
		double totalStep = 0;
		MPI_Allreduce(&localStep,&totalStep,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
		return totalStep;
	}
	else
	#endif //HAVE_MPI
	{
		return localStep;
	}

};

int nonLinearMechSolver::oneStaticStep(const double curtime,const double dt, const int numstep)
{
	double tsystresol = Cpu();
  int niteNR = 0;
  if (_pathFollowing){
    this->oneStepPreSolvePathFollowing(curtime,dt,numstep);
    niteNR = NewtonRaphsonPathFollowing(numstep);
  }
  else{
    //std::string name = "A";
    //linearSystem<double>* lsys =pAssembler->getLinearSystem(name);
    //lsys->zeroRightHandSide();
    //nonLinearSystemPETSc<double>* nonsys = dynamic_cast< nonLinearSystemPETSc<double>*>(lsys);
    //nonsys->zeroRightHandSide();
    //nonsys->printAll();    // Solve one step by NR scheme
    this->oneStepPreSolve(curtime,dt,numstep);
    niteNR = NewtonRaphson(numstep);
  }
	tsystresol = Cpu() - tsystresol;
  Msg::Info("Time of Newton-Raphson resolution %f",tsystresol);
  return niteNR;
}

double nonLinearMechSolver::finalizeStaticScheme(const double curtime, const int step)
{
  this->endOfScheme(curtime,step);
  Msg::Info("NonLinearStatic OK");
  return curtime;
}

void nonLinearMechSolver::createOnelabViewOption(const std::string &fname,const int displayType) const
{
  std::string optfname(fname);
  optfname += ".opt";
  FILE * fopt = fopen(optfname.c_str(),"w");

  fprintf(fopt,"n = %d;\n",_GmshOneLabViewNum);
  fprintf(fopt,"View[n].VectorType = %d;\n",displayType);

  fclose(fopt);

  _GmshOneLabViewNum++;
  return;
}

void nonLinearMechSolver::createOnelabDisplacementView(const std::string &fname, const double time) const
{
  static bool initialized = false;
  // ask the unknown field to create the view
  _ufield->onelabView(_meshFileName,fname,this,std::string("displacement"),0,time,_currentStep);

  if(!initialized)
  {
    this->createOnelabViewOption(fname,5); // 5 for displacement view type
    initialized = true;
  }
}

void nonLinearMechSolver::createOnelabVelocityView(const std::string &fname, const double time) const
{
  static bool initialized = false;
  // ask the unknown field to create the view
  _ufield->onelabView(_meshFileName,fname,this,std::string("velocity"),1,time,_currentStep);

  if(!initialized)
  {
    this->createOnelabViewOption(fname,4); // 4 for 3D arrow view type
    initialized = true;
  }
}

void nonLinearMechSolver::getOnelabArchiveNodalUnknowns(fullVector<double> &nodalValues) const
{
  _ufield->valuesForOnelab(nodalValues);
  return;
}


bool nonLinearMechSolver::checkVertexBelongToDomains(const int phyVer) const{
  groupOfElements gV(0,phyVer);
  MVertex* v = *(gV.vbegin());
  if (v == NULL) {
    Msg::Error("vertex does not exist");
    return true;
  }
  for (int i=0; i< domainVector.size(); i++){
    if (domainVector[i]->groupOfElementsSize() == 0){
      Msg::Error("domain vector must be initialized firts");
    }
    for (groupOfElements::vertexContainer::const_iterator it = domainVector[i]->g_cvbegin(); it!= domainVector[i]->g_cvend(); it++){
      MVertex* vv = *it;
      if (vv->getNum() == v->getNum()) return true;
    }
  }

  return false;
};

double nonLinearMechSolver::getArchivedNodalValue(const int numphys,const int comp,nonLinearBoundaryCondition::whichCondition wc) const
{
  bool found = false;
  double dofvalue = 0;
  for(std::vector<unknownField::archiveNode>::const_iterator itn = _ufield->getArchiveNodeContainer().begin(); itn!=_ufield->getArchiveNodeContainer().end();++itn)
  {
    if (itn->physnum == numphys && itn->_comp == comp &&itn->wc == wc)
    {
      found = true;
      if (whatScheme == StaticLinear)
      {
        pAssembler->getDofValue(itn->D,dofvalue);
      }
      else
      {
        static_cast<const staticDofManager<double>*>(pAssembler)->getDofValue(itn->D,dofvalue,itn->wc);
      }
      break;
    }
  }
  
  #if defined(HAVE_MPI)
  if (getNumRanks() > 1)
  {
    double sumValue;
    MPI_Allreduce(&dofvalue,&sumValue, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    return sumValue;
  }
  else
  #endif // HAVE_MPI
  {
    if (!found){
      Msg::Error("Cannot get a value in getArchivedNodalValue.\nTo get a value, the disp archiving on node with physical %d must be defined with archivingNodeDisplacement",numphys);
      return 0.;
    }
    else
      return dofvalue;
  }
};

double nonLinearMechSolver::getArchivedForceOnPhysicalGroup(std::string onwhat, const int numphys,const int comp) const
{
  static std::string node("Node");
  static std::string edge("Edge");
  static std::string face("Face");
  static std::string volu("Volume");
  int dim=0;
  if(onwhat == node )
    dim = 0;
  else if(onwhat == edge)
    dim = 1;
  else if(onwhat == face)
    dim = 2;
  else if(onwhat == volu)
    dim = 3;

  double force =0.;
  bool found = false;
  for(std::vector<archiveForce>::const_iterator itf = vaf.begin(); itf != vaf.end();++itf)
  {
    if(itf->numphys != numphys) continue;
    if(itf->comp != comp) continue;
    if(itf->dim != dim) continue;
    force = itf->fval; found = true; break;
  }

  #if defined(HAVE_MPI)
  if (getNumRanks() > 1){
    double sumValue;
    MPI_Allreduce(&force,&sumValue, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    return sumValue;
  }
  else
  #endif // HAVE_MPI
  {
    if (!found){
      Msg::Error("Cannot found a force value in getarchivedForceOnPhysicalGroup");
      return 0.;
    }
    else
      return force;
  }
}

double nonLinearMechSolver::getEnergy(int we) const
{
  double eneryPart =_energField->get(energeticField::whichEnergy(we));
  #if defined(HAVE_MPI)
  if(Msg::GetCommSize() > 1){
    double totalEnerg= 0.;;
    MPI_Allreduce(&eneryPart,&totalEnerg,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
    return totalEnerg;
  }
  #endif // HAVE_MPI
  return eneryPart;
}
void nonLinearMechSolver::forceArchiving(const double curtime,const int numstep, const bool forceSave)
{
  const staticDofManager<double> *sAssembler = dynamic_cast<const staticDofManager<double>*>(pAssembler);
  if (sAssembler == NULL)
  {
    #ifdef _DEBUG
    Msg::Warning("force archiving work only with staticDofManager\n");
    #endif //_DEBUG
    return;
  }
  sAssembler->getForces(vaf); // get force value from system
  for(int i=0;i<vaf.size();i++){
    if (vaf[i].vdof.size()!=0 && (vaf[i].contact_root == -1 || vaf[i].contact_root == Msg::GetCommRank()))
    {
      if (numstep > vaf[i].lastSaveStep){
        if((numstep%(vaf[i].nstep)==0 or forceSave))
        {
          vaf[i].lastSaveStep = numstep;
          fprintf(vaf[i].FP,"%e;%e\n",curtime,vaf[i].fval);
          fflush(vaf[i].FP);
        }
        
      }
    }
  }
}

double nonLinearMechSolver::getHomogenizedStress(const int i, const int j) const{
	return _currentState->getHomogenizedStress()(i,j);
};
double nonLinearMechSolver::getHomogenizedTangent(const int i, const int j, const int k, const int l) const{
	return _currentState->getHomogenizedTangentOperator_F_F()(i,j,k,l);
};

double nonLinearMechSolver::getHomogenizedSecondTangent(const int i, const int j, const int k, const int p, const int q, const int r) const{
  return _currentState->getHomogenizedTangentOperator_G_G()(i,j,k,p,q,r);
};

double nonLinearMechSolver::getHomogenizedSecondStress(const int i, const int j, const int k) const{
  return _currentState->getHomogenizedSecondOrderStress()(i,j,k);
};

double nonLinearMechSolver::getHomogenizedFirstSecondTangent(const int i, const int j, const int p, const int q, const int r) const{
  return _currentState->getHomogenizedTangentOperator_F_G()(i,j,p,q,r);
};

double nonLinearMechSolver::getHomogenizedSecondFirstTangent(const int i, const int j, const int k, const int p, const int q) const{
  return _currentState->getHomogenizedTangentOperator_G_F()(i,j,k,p,q);
};

double nonLinearMechSolver::getHomogenizedCohesiveJump(const int i) const{
  return _currentState->getHomogenizedCohesiveJump()(i);
};

double nonLinearMechSolver::getNodeIP(const int physical, const int ipVal) const{
  return _ipf->getIPValueOnPhysical(0,physical,ipVal);
};

void nonLinearMechSolver::setSelectiveUpdate(const selectiveUpdateBase& selectiveObj){
  _failureBasedOnPreviousState = false;
  if (_selectiveObject) delete _selectiveObject;
  _selectiveObject = selectiveObj.clone();
};

void nonLinearMechSolver::endSchemeMonitoring(const EndSchemeMonitoringBase& endObj){
  _endSchemeMonitoring = true;
  if (_endSchemeMonitoringObject != NULL) delete _endSchemeMonitoringObject;
  _endSchemeMonitoringObject = endObj.clone();
};
bool nonLinearMechSolver::solverEndedByMornitoring() const{
  if (_endSchemeMonitoring){
    return _endSchemeMonitoringObject->solverEndedByMornitoring();
  }
  else{
    return false;
  }
};

void nonLinearMechSolver::setNoMassPredicorWithoutVelocityAndAcceleration(const bool fl){
  _noMassPredicorWithoutVelocityAndAcceleration = fl;
  if (fl)
  {
    Msg::Info("no velocity in predictor of nomass dof with Implicit CH scheme");
  }
  
};

// for crack insertion control
void nonLinearMechSolver::setMaxAllowedCrackInsertionAtOneStep(const int maxNum)
{
  if(maxNum < 0) Msg::Error("broken number must be positive nonLinearMechSolver::setMaxAllowedCrackInsertionAtOneStep");
  Msg::Info("Only %d interface elements are allowed to be cracked during one step", maxNum);

  ReduceTimeStepWithMaximalNumberOfBrokenElements  endObj(maxNum);
  this->endSchemeMonitoring(endObj);
}


// for element erosion
void nonLinearMechSolver::setElementErosion(const bool bulkErosion, const bool interfaceErosion, const int method){
	_bulkElementErosionFlag  = bulkErosion;
	_interfaceElementErosionFlag = interfaceErosion;
	if (_bulkElementErosionFlag or _interfaceElementErosionFlag){
		_erosionType = (elementErosionType)method;
		if (_erosionType == nonLinearMechSolver::FIRST_IP_FAILED){
			Msg::Info("An element is removed when first IP reaches failure state");
		}
		else if (_erosionType == nonLinearMechSolver::ALL_IP_FAILED){
			Msg::Info("An element is removed when all IPs reach failure state");
		}
		else{
			Msg::Error("_erosionType is not defined nonLinearMechSolver::setElementErosion");
		}
	}
};

void nonLinearMechSolver::setGlobalErosionCheck(const bool flag, const FailureCriterionBase* fcr){
	_globalCheckFlag = flag;
	if (_globalCheckFlag){
		Msg::Info("erosion is checked globally based on a global criterion");
		_erosionGlobalCriterion = fcr->clone();
	}
};

//
// C++ Interface: terms Quasi Static
//
// Description: Files with definition of function : nonLinearMechSolver::solveSNL()
//
//
// Author:  <Gauthier BECKER>, (C) 2010
//
// Copyright: See COPYING file that comes with this distribution
//
//

// compute Fext-Fint
double nonLinearMechSolver::computeRightHandSide(){
  //lsys->zeroRightHandSide();
  staticDofManager<double> *pA2 = static_cast<staticDofManager<double>*>(pAssembler);
  pA2->clearRHSfixed();
  this->computeInternalForces();

  // compute ext forces
  this->computeExternalForces();

  // compute contact forces
  this->computeContactForces();

  // for periodic BC
  if (_microFlag){
    if (_systemType == nonLinearMechSolver::MULT_ELIM){
      _pAl->assembleConstraintResidualToSystem();
      this->computeStiffMatrix();
    }
    else if (_systemType == nonLinearMechSolver::DISP_MULT) {
      _pAl->computeLinearConstraints();
    }
  }

	if (_pathFollowing and _pathFollowingMethod == LOCAL_BASED){
		this->computePathFollowingConstraint();
	}

  // save Fint component to archive ( write in file when Fint = Fext)
//  return lsys->normInfRightHandSide();
  return pA2->normInfRightHandSide();
}

void nonLinearMechSolver::computeInternalForces(){
  staticDofManager<double>* nlAssembler = static_cast<staticDofManager<double>*>(pAssembler);
  for(std::vector<partDomain*>::iterator itdom=domainVector.begin(); itdom!=domainVector.end();++itdom)
  {
    partDomain *dom = *itdom;
    Assemble(dom->getLinearBulkTerm(),*(dom->getFunctionSpace()),dom->g_cbegin(),dom->g_cend(),
             *(dom->getBulkGaussIntegrationRule()),_ufield,*nlAssembler,
             nonLinearSystem<double>::Fint,_elementErosionFilter);

    if(dom->IsInterfaceTerms()){
      dgPartDomain *dgdom = static_cast<dgPartDomain*>(dom);
      if(whatScheme!=Multi)
      {
        // Assembling loop on elementary interface terms
        if (dom->IsMultipleFieldFormulation()){
          Assemble(*(dgdom->getInterfaceFunctionSpace()),dgdom->getLinearInterfaceTerm(),*(dgdom->getFunctionSpace()),dgdom->gi->begin(),dgdom->gi->end(),*(dgdom->getInterfaceGaussIntegrationRule()),
                 _ufield,*nlAssembler,nonLinearSystem<double>::Fint,_elementErosionFilter); // Use the same GaussQuadrature rule than on the boundary
        }
        else{

          Assemble(dgdom->getLinearInterfaceTerm(),*(dgdom->getFunctionSpace()),dgdom->gi->begin(),dgdom->gi->end(),*(dgdom->getInterfaceGaussIntegrationRule()),
                 _ufield,*nlAssembler,nonLinearSystem<double>::Fint,_elementErosionFilter); // Use the same GaussQuadrature rule than on the boundary
        }

      }
      else
      {

        dofManagerMultiSystems<double>* multiManager = static_cast<dofManagerMultiSystems<double>*>(pAssembler);
        // Assembling loop on elementary interface terms
        AssembleInterVectorMulti(dgdom->getLinearInterfaceTerm(),*(dgdom->getFunctionSpace()),dgdom->gi->begin(),dgdom->gi->end(),*(dgdom->getInterfaceGaussIntegrationRule()),
                        _ufield,*multiManager,_elementErosionFilter); // Use the same GaussQuadrature rule than on the boundary
      }
      // Assembling loop on elementary boundary interface terms

      Assemble(dgdom->getLinearVirtualInterfaceTerm(),*(dgdom->getFunctionSpace()),dgdom->gib->begin(),dgdom->gib->end(),
               *(dgdom->getInterfaceGaussIntegrationRule()),_ufield,*nlAssembler,nonLinearSystem<double>::Fint,_elementErosionFilter); // Use the same GaussQuadrature rule than on the boundary

    }
  }
}

void nonLinearMechSolver::computeContactForces(){
  staticDofManager<double>* nlAssembler = static_cast<staticDofManager<double>*>(pAssembler);
  for(contactContainer::iterator it = _allContact.begin(); it!=_allContact.end(); ++it){
    contactDomain *cdom = *it;
    contactLinearTermBase<double> *cterm = static_cast<contactLinearTermBase<double>*>(cdom->getForceTerm());
    cterm->clearContactNodes();
    Assemble(*(cterm),*(cdom->getSpace()),cdom->gSlave->begin(),
            cdom->gSlave->end(),*(cdom->getGaussIntegration()),*nlAssembler,nonLinearSystem<double>::Fext,_elementErosionFilter);
  }
  for(defoDefoContactContainer::iterator it = _allDefoDefoContact.begin(); it!=_allDefoDefoContact.end(); ++it){
    defoDefoContactDomain *cdom = *it;
    std::vector< LinearTermBase<double> *> &term = cdom->getRefToForceTerm(); //we might need to change the class
    for(int i=0;i<term.size(); i++)
    {
      contactLinearTermBase<double>* cterm=static_cast< contactLinearTermBase<double>* > (term[i]);
      cterm->clearContactNodes();
      AssembleDefoDefoContact(*(cterm),*((cdom->getRefToVSpace())[i]),((cdom->getRefToGroup())[i]).begin(),
            ((cdom->getRefToGroup())[i]).end(),
            *((cdom->getRefToVQuadratureMaster())[i]),*nlAssembler,nonLinearSystem<double>::Fext,_elementErosionFilter);
    }
  }
}

void  nonLinearMechSolver::computeMassMatrix(){
   static std::string Bname("B");
   pAssembler->getLinearSystem(Bname)->zeroMatrix();

   pAssembler->setCurrentMatrix(Bname);
   for(std::vector<partDomain*>::iterator itdom=domainVector.begin(); itdom!=domainVector.end(); ++itdom)
   {
     partDomain* dom = *itdom;
     // Assembling loop on Elementary terms
     Assemble(*(dom->getBilinearMassTerm()),*(dom->getFunctionSpace()),dom->g_cbegin(),dom->g_cend(),
               *(dom->getBulkGaussIntegrationRule()),*pAssembler,_elementErosionFilter);
    };
   static std::string Aname("A");
   pAssembler->setCurrentMatrix(Aname);;
};

void nonLinearMechSolver::computeStiffMatrix(){
  if (_stiffEstimation)
  {
    if (_stiffnessModification ==false){
      _stiffEstimation = false;
    }
    else{
      _stiffEstimation = true;
    };

    if(_vcompBySys.size()==0) // otherwise problem with the staticLinear scheme that uses a classical dofManager !
    {
      static std::string Aname("A");
      pAssembler->getLinearSystem(Aname)->zeroMatrix();
    }
    else
    {
      staticDofManager<double>* smanager = static_cast<staticDofManager<double>*>(pAssembler);
      smanager->zeroMatrix();
    }
    for(std::vector<partDomain*>::iterator itdom=domainVector.begin(); itdom!=domainVector.end(); ++itdom)
    {
      partDomain* dom = *itdom;
      // Assembling loop on Elementary terms
      Assemble(*(dom->getBilinearBulkTerm()),*(dom->getFunctionSpace()),dom->g_cbegin(),dom->g_cend(),
               *(dom->getBulkGaussIntegrationRule()),*pAssembler,_elementErosionFilter);

      if(dom->IsInterfaceTerms()){
        dgPartDomain *dgdom = static_cast<dgPartDomain*>(dom);
        if(whatScheme!=Multi)
        {
          if (dgdom->IsMultipleFieldFormulation()){
            Assemble(*(dgdom->getInterfaceFunctionSpace()),*(dgdom->getBilinearInterfaceTerm()),*(dom->getFunctionSpace()),dgdom->gi->begin(),dgdom->gi->end(),
                     *(dgdom->getInterfaceGaussIntegrationRule()),*pAssembler,_elementErosionFilter);

          }
          else{
            // Assembling loop on elementary interface terms
            Assemble(*(dgdom->getBilinearInterfaceTerm()),*(dom->getFunctionSpace()),dgdom->gi->begin(),dgdom->gi->end(),
                     *(dgdom->getInterfaceGaussIntegrationRule()),*pAssembler,_elementErosionFilter);
          }
        }
        else
        {
          dofManagerMultiSystems<double>* multiAssembler = static_cast<dofManagerMultiSystems<double>*>(pAssembler);
          // Assembling loop on elementary interface terms
          AssembleInterMatrixMulti(*(dgdom->getBilinearInterfaceTerm()),*(dom->getFunctionSpace()),dgdom->gi->begin(),dgdom->gi->end(),
                                   *(dgdom->getInterfaceGaussIntegrationRule()),*multiAssembler,_elementErosionFilter);
        }
        // Assembling loop on elementary boundary interface terms
        Assemble(*(dgdom->getBilinearVirtualInterfaceTerm()),*(dom->getFunctionSpace()),dgdom->gib->begin(),dgdom->gib->end(),
                 *(dgdom->getInterfaceGaussIntegrationRule()),*pAssembler,_elementErosionFilter);
      }
    }
    // NeumannBC (non linear flux)
    if(this->getScheme() != StaticLinear){
      staticDofManager<double> *sAssembler;
      for(std::list<nonLinearNeumannBC>::iterator itbc = allNeumann.begin(); itbc != allNeumann.end();++itbc){
        nonLinearNeumannBC &neu = *itbc;
        // select the system if required
        if(_vcompBySys.size()==0)
        {
          sAssembler = static_cast<staticDofManager<double>*>(pAssembler);
        }
        else
        {
          dofManagerMultiSystems<double>* multiManager = static_cast<dofManagerMultiSystems<double>*>(pAssembler);
          sAssembler = static_cast<staticDofManager<double>*>(multiManager->getDofManagerByComp(neu._comp));
        }
        for(int j=0;j<neu._vspace.size();j++)
        {
          FunctionSpaceBase *spn = neu._vspace[j];
          const groupOfElements *gneu = neu._vgroup[j];
          QuadratureBase *qneu = neu._vquadrature[j];
          AssembleSkipEmpty(*(neu._vmatrix_term[j]),*spn,gneu->begin(),gneu->end(),*(qneu),*pAssembler,_elementErosionFilter);
        }
      }
    }
    else{
      for(std::list<nonLinearNeumannBC>::iterator itbc = allNeumann.begin(); itbc != allNeumann.end();++itbc){
        nonLinearNeumannBC &neu = *itbc;
        for(int j=0;j<neu._vspace.size();j++)
        {
          FunctionSpaceBase *spn = neu._vspace[j];
          const groupOfElements *gneu = neu._vgroup[j];
          QuadratureBase *qneu = neu._vquadrature[j];
          AssembleSkipEmpty(*(neu._vmatrix_term[j]),*spn,gneu->begin(),gneu->end(),*(qneu),*pAssembler,_elementErosionFilter);
        }
      }
    }

    // Contact
    // little particular because it the term itself which know (via its linked linear term) the element
    // where there is contact and so a bilinearterm != 0 as to be computed
    for(contactContainer::iterator it = _allContact.begin(); it!=_allContact.end(); ++it){
      contactDomain *cdom = *it;
      contactBilinearTermBase<double> *sterm = static_cast<contactBilinearTermBase<double>*>(cdom->getStiffnessTerm());
      Assemble(*sterm,*(cdom->getSpace()),sterm->getContactNodes()->elemBegin(),
                      sterm->getContactNodes()->elemEnd(),*(cdom->getGaussIntegration()),*pAssembler,_elementErosionFilter);
  //    sterm->clearContactNodes();
    }
    for(defoDefoContactContainer::iterator it = _allDefoDefoContact.begin(); it!=_allDefoDefoContact.end(); ++it){
      defoDefoContactDomain *cdom = *it;
      const std::vector<BilinearTermBase*> &sterm = static_cast< const std::vector<BilinearTermBase* > >(cdom->getConstRefToStiffnessTerm()); //we might need to change the class
      for(int i=0; i<sterm.size(); i++)
      {
        AssembleSkipEmptyDefoDefoContact(*(sterm[i]),*(cdom->getRefToVSpace()[i]),((cdom->getRefToGroup())[i]).begin(),
            ((cdom->getRefToGroup())[i]).end(),
            *(cdom->getRefToVQuadratureMaster()[i]),*pAssembler,_elementErosionFilter);
      }
    }

    // for periodic BC
    if (_microFlag)
      if (_systemType == nonLinearMechSolver::DISP_MULT)
        _pAl->assembleConstraintMatrix();
  }

	if (_pathFollowing and _pathFollowingMethod == LOCAL_BASED){
		this->computeDPathFollowingConstraintDUnknowns();
	}
}

void nonLinearMechSolver::computeElasticStiffMatrix(){
  if (_stiffEstimation)
  {
    if (_stiffnessModification ==false){
      _stiffEstimation = false;
    }
    else{
      _stiffEstimation = true;
    };

    if(_vcompBySys.size()==0) // otherwise problem with the staticLinear scheme that uses a classical dofManager !
    {
      static std::string Aname("A");
      pAssembler->getLinearSystem(Aname)->zeroMatrix();
    }
    else
    {
      staticDofManager<double>* smanager = static_cast<staticDofManager<double>*>(pAssembler);
      smanager->zeroMatrix();
    }
    for(std::vector<partDomain*>::iterator itdom=domainVector.begin(); itdom!=domainVector.end(); ++itdom)
    {
      partDomain* dom = *itdom;
      // Assembling loop on Elementary terms
      Assemble(*(dom->getBilinearElasticBulkTerm()),*(dom->getFunctionSpace()),dom->g_cbegin(),dom->g_cend(),
               *(dom->getBulkGaussIntegrationRule()),*pAssembler,_elementErosionFilter);

      /*if(dom->IsInterfaceTerms()){
        dgPartDomain *dgdom = static_cast<dgPartDomain*>(dom);
        if(whatScheme!=Multi)
        {
          if (dgdom->IsMultipleFieldFormulation()){
            Assemble(*(dgdom->getInterfaceFunctionSpace()),*(dgdom->getBilinearInterfaceTerm()),*(dom->getFunctionSpace()),dgdom->gi->begin(),dgdom->gi->end(),
                     *(dgdom->getInterfaceGaussIntegrationRule()),*pAssembler,_elementErosionFilter);

          }
          else{
            // Assembling loop on elementary interface terms
            Assemble(*(dgdom->getBilinearInterfaceTerm()),*(dom->getFunctionSpace()),dgdom->gi->begin(),dgdom->gi->end(),
                     *(dgdom->getInterfaceGaussIntegrationRule()),*pAssembler,_elementErosionFilter);
          }
        }
        else
        {
          dofManagerMultiSystems<double>* multiAssembler = static_cast<dofManagerMultiSystems<double>*>(pAssembler);
          // Assembling loop on elementary interface terms
          AssembleInterMatrixMulti(*(dgdom->getBilinearInterfaceTerm()),*(dom->getFunctionSpace()),dgdom->gi->begin(),dgdom->gi->end(),
                                   *(dgdom->getInterfaceGaussIntegrationRule()),*multiAssembler,_elementErosionFilter);
        }
        // Assembling loop on elementary boundary interface terms
        Assemble(*(dgdom->getBilinearVirtualInterfaceTerm()),*(dom->getFunctionSpace()),dgdom->gib->begin(),dgdom->gib->end(),
                 *(dgdom->getInterfaceGaussIntegrationRule()),*pAssembler,_elementErosionFilter);
      }
      */
    }
    /*
    // NeumannBC (non linear flux)
    if(this->getScheme() != StaticLinear){
      staticDofManager<double> *sAssembler;
      for(std::list<nonLinearNeumannBC>::iterator itbc = allNeumann.begin(); itbc != allNeumann.end();++itbc){
        nonLinearNeumannBC &neu = *itbc;
        // select the system if required
        if(_vcompBySys.size()==0)
        {
          sAssembler = static_cast<staticDofManager<double>*>(pAssembler);
        }
        else
        {
          dofManagerMultiSystems<double>* multiManager = static_cast<dofManagerMultiSystems<double>*>(pAssembler);
          sAssembler = static_cast<staticDofManager<double>*>(multiManager->getDofManagerByComp(neu._comp));
        }
        for(int j=0;j<neu._vspace.size();j++)
        {
          FunctionSpaceBase *spn = neu._vspace[j];
          const groupOfElements *gneu = neu._vgroup[j];
          QuadratureBase *qneu = neu._vquadrature[j];
          AssembleSkipEmpty(*(neu._vmatrix_term[j]),*spn,gneu->begin(),gneu->end(),*(qneu),*pAssembler,_elementErosionFilter);
        }
      }
    }
    else{
      for(std::list<nonLinearNeumannBC>::iterator itbc = allNeumann.begin(); itbc != allNeumann.end();++itbc){
        nonLinearNeumannBC &neu = *itbc;
        for(int j=0;j<neu._vspace.size();j++)
        {
          FunctionSpaceBase *spn = neu._vspace[j];
          const groupOfElements *gneu = neu._vgroup[j];
          QuadratureBase *qneu = neu._vquadrature[j];
          AssembleSkipEmpty(*(neu._vmatrix_term[j]),*spn,gneu->begin(),gneu->end(),*(qneu),*pAssembler,_elementErosionFilter);
        }
      }
    }
     */

    // Contact
    // little particular because it the term itself which know (via its linked linear term) the element
    // where there is contact and so a bilinearterm != 0 as to be computed
    for(contactContainer::iterator it = _allContact.begin(); it!=_allContact.end(); ++it){
      contactDomain *cdom = *it;
      contactBilinearTermBase<double> *sterm = static_cast<contactBilinearTermBase<double>*>(cdom->getStiffnessTerm());
      Assemble(*sterm,*(cdom->getSpace()),sterm->getContactNodes()->elemBegin(),
                      sterm->getContactNodes()->elemEnd(),*(cdom->getGaussIntegration()),*pAssembler,_elementErosionFilter);
  //    sterm->clearContactNodes();
    }
    for(defoDefoContactContainer::iterator it = _allDefoDefoContact.begin(); it!=_allDefoDefoContact.end(); ++it){
      defoDefoContactDomain *cdom = *it;
      const std::vector<BilinearTermBase*> &sterm = static_cast< const std::vector<BilinearTermBase* > >(cdom->getConstRefToStiffnessTerm()); //we might need to change the class
      for(int i=0; i<sterm.size(); i++)
      {
        AssembleSkipEmptyDefoDefoContact(*(sterm[i]),*(cdom->getRefToVSpace()[i]),((cdom->getRefToGroup())[i]).begin(),
            ((cdom->getRefToGroup())[i]).end(),
            *(cdom->getRefToVQuadratureMaster()[i]),*pAssembler,_elementErosionFilter);
      }
    }
    // for periodic BC
    if (_microFlag)
      if (_systemType == nonLinearMechSolver::DISP_MULT)
        _pAl->assembleConstraintMatrix();
    }

    if (_pathFollowing and _pathFollowingMethod == LOCAL_BASED){
      this->computeDPathFollowingConstraintDUnknowns();
    }
}

// Newton Raphson scheme to solve one step
int nonLinearMechSolver::NewtonRaphson(const int numstep){
  staticDofManager<double>* staticAssembler = static_cast<staticDofManager<double>*>(pAssembler);
  if (whatScheme == Implicit){
    // dynamic predictor
    staticAssembler->systemSolveIntReturn();
  }

  // compute ipvariable
  _ipf->compute1state(IPStateBase::current, true);

  // Compute Right Hand Side (Fext-Fint)
  double normFinf = this->computeRightHandSide();
  // norm0 = norm(Fext) + norm(Fint) to have a relative convergence
  double norm0 = staticAssembler->norm0Inf();

  // loop until convergence
  int iter=0;
  if(norm0 == 0){ // no force (can append if contact )
    Msg::Info("no force go next step");
    if (_microFlag) this->computeStiffMatrix();
    return 0;
  }
  double relnorm = normFinf/norm0;
  Msg::Info("iteration n %d : residu : %e, absolute residu : %e",iter,relnorm,normFinf);
  while ((relnorm>_tol and normFinf>_absTol) or iter == 0){ // ensure 1 iteration min

    iter++;
    // Compute Stiffness Matrix
    this->computeStiffMatrix();

    // Solve KDu = Fext-Fint
    double t = Cpu();
    int succeed = staticAssembler->systemSolveIntReturn();
    t = Cpu() - t;
    if(!succeed)
    {
      iter = _maxNRite;
      break;
    }
    // update ipvariable
    _ipf->compute1state(IPStateBase::current,true);

    if (!_iterativeNR) break;
    // new forces
    normFinf = this->computeRightHandSide();

    // Check of convergence
    relnorm = normFinf/norm0;
    Msg::Info("iteration n %d : residu : %e, absolute residu : %e, solving time: %f",iter,relnorm, normFinf,t);

    // perform line search if activate (default)
    while((relnorm>_tol and normFinf>_absTol) and (iter > 1))    // avoid line search for linear case
    {
      int ls = staticAssembler->lineSearch();
      if(ls == 1) break; // processus is converged or failed
      // update ipvariable
      _ipf->compute1state(IPStateBase::current,true);
      // new RightHandSide
      this->computeRightHandSide();
    }
    if((iter == _maxNRite) or std::isnan(relnorm))
    {
      // reset system value
      iter = _maxNRite; // for isnan case
      break;
    }
  }
  //this->fixNodalDofs(); // reset values of converged state in accordance to BC
  return iter;
}


linearSystem<double>* nonLinearMechSolver::createSNLSystem()
{
  // Select the solver for linear system Ax=b (KDeltau = Fext-Fint)
  linearSystem<double> *lsys = NULL;
  if(nonLinearMechSolver::whatSolver == Taucs){
    #if defined(HAVE_TAUCS)
    lsys = new nonLinearSystemCSRTaucs<double>(_lineSearch);
    Msg::Info("Taucs is chosen to solve");
    #else
    lsys = new nonLinearSystemGmm<double>(_lineSearch);
    dynamic_cast<nonLinearSystemGmm<double>*>(lsys)->setNoisy(2);
    Msg::Error("Taucs is not installed\n Gmm is chosen to solve");
    #endif
  }
  else if(nonLinearMechSolver::whatSolver == Petsc){
    #if defined(HAVE_PETSC)
    if (_pathFollowing){
      if ((!_isPartitioned) and (Msg::GetCommSize()>1)){
				if (_pathFollowingMethod == GLOBAL_ARC_LENGTH_BASED){
					lsys = new pathFollowingSystemPETSC<double>(PETSC_COMM_SELF);
				}
				else if (_pathFollowingMethod == LOCAL_BASED){
					lsys = new generalLocalControlBasedPathFollowingSystemPETSc<double>(PETSC_COMM_SELF);
				}
        else if (_pathFollowingMethod == HYPERELLIPTIC_BASED){
          lsys = new generalHyperellipticPathFollowingSystemPETSC<double>(PETSC_COMM_SELF);
        }
				else
					Msg::Error("path following method %d is not defined",_pathFollowingMethod);
      }
      else{
				if (_pathFollowingMethod == GLOBAL_ARC_LENGTH_BASED){
					lsys = new pathFollowingSystemPETSC<double>(PETSC_COMM_WORLD);
				}
				else if (_pathFollowingMethod == LOCAL_BASED){
					lsys = new generalLocalControlBasedPathFollowingSystemPETSc<double>(PETSC_COMM_WORLD);
				}
        else if (_pathFollowingMethod == HYPERELLIPTIC_BASED){
          lsys = new generalHyperellipticPathFollowingSystemPETSC<double>(PETSC_COMM_WORLD);
        }
				else
					Msg::Error("path following method %d is not defined",_pathFollowingMethod);
      }
			// set path following options
			// paramater remains default if it is negative
			pathFollowingSystemBase* pfsysBase = dynamic_cast<pathFollowingSystemBase*>(lsys);
			if (_macroControlTypePathFollowing >=0){
				pfsysBase->setControlType(_macroControlTypePathFollowing);
			}
			if (_correctionMethodPathFollowing>=0){
				pfsysBase->setCorrectionMethod(_correctionMethodPathFollowing);
			}
			if (_tranversalCriterionPathFollowing>=0){
				pfsysBase->setTranversalCriterion(_tranversalCriterionPathFollowing);
			}
			if (_solverTypePathFollowing >=0){
				pfsysBase->setSolverType(_solverTypePathFollowing,_pathFollowingEqRatio);
			}

    }
    else{
      if ((!_isPartitioned) and (Msg::GetCommSize()>1)){
        lsys = new nonLinearSystemPETSc<double>(PETSC_COMM_SELF,_lineSearch);
      }
      else
      {
        lsys = new nonLinearSystemPETSc<double>(PETSC_COMM_WORLD,_lineSearch);
      }

    }
    if(_solver_options.size())
      lsys->setParameter(std::string("petsc_solver_options"),_solver_options);
    Msg::Info("PETSc is chosen to solve");
    #else
    Msg::Error("PETSc is not installed\n Gmm is chosen to solve");
    if (_pathFollowing){
      Msg::Error("path following has not been implemented with Gmm, load control is considered");
    }

    lsys = new nonLinearSystemGmm<double>(_lineSearch);
    dynamic_cast<nonLinearSystemGmm<double>*>(lsys)->setNoisy(2);

    #endif
  }
  else if(nonLinearMechSolver::whatSolver == GMRESk){
    lsys = new nonLinearSystemGMRESk<double>(_lineSearch);
    Msg::Info("GMRESk is chosen to solve");
  }
  else{
    lsys = new nonLinearSystemGmm<double>(_lineSearch);
    static_cast<nonLinearSystemGmm<double>*>(lsys)->setNoisy(2);
    Msg::Info("Gmm is chosen to solve");
  }

  nonLinearSystem<double>* nlsys = dynamic_cast<nonLinearSystem<double>*>(lsys);
  if (nlsys != NULL){
    // initialization data
    #if defined(HAVE_MPI)
    int rootRank = 0;
    std::set<int> OtherRanks;
    if ((!_isPartitioned) and (Msg::GetCommSize() >1) ){
      // if sequential mesh in mpi run
      if (!_microFlag){
        // if macro problem
        rootRank = 0;
        for (int i=1; i< Msg::GetCommSize(); i++){
          OtherRanks.insert(i);
        }
      }
      else{
        // if micro problem
        rootRank = Msg::GetCommRank();
      }
    }
    else{
      rootRank = Msg::GetCommRank();
    }
    nlsys->setWorkingRanks(rootRank,OtherRanks);
    #endif // HAVE_MPI
  }

  return lsys;
}

linearSystem<double>* nonLinearMechSolver::createExplicitSystem()
{
  linearSystem<double> *esys;
  if(_previousScheme != Explicit){
    if(nonLinearMechSolver::whatSolver == Petsc){
     #if defined(HAVE_PETSC)
       Msg::Info("Explicit scheme based on PETSc");
       esys = new explicitHulbertChungPetsc<double>(_alpham,_beta,_gamma,_timeStepByBenson,_dynamicRelaxation);
     #else
       Msg::Warning("petsc is not install on this computer. Use blas explicit system");
      #if !defined(HAVE_BLAS)
       Msg::Warning("blas is not installed. Use not optimized default implementation");
      #endif // HAVE_BLAS
       esys = new explicitHulbertChungBlas<double>(_alpham,_beta,_gamma,_timeStepByBenson,_dynamicRelaxation);
     #endif // HAVE_PETSC
    }
    else{
       #if !defined(HAVE_BLAS)
        Msg::Warning("blas is not installed. Use not optimized default implementation. Maybe you can use PETSc ?");
       #else
        Msg::Info("Explicit scheme based on BLAS");
       #endif // HAVE_BLAS
        esys = new explicitHulbertChungBlas<double>(_alpham,_beta,_gamma,_timeStepByBenson,_dynamicRelaxation);
    }
  }
   else{
     std::string sysname("A");
     esys = pAssembler->getLinearSystem(sysname);
     explicitHulbertChung<double>* expsys = static_cast<explicitHulbertChung<double>*>(esys);
     expsys->setDynamicRelaxation(_dynamicRelaxation);
   }
   return esys;
}

linearSystem<double>* nonLinearMechSolver::createImplicitSystem(){
  linearSystem<double> *esys;
  if(nonLinearMechSolver::whatSolver == Petsc){
   #if defined(HAVE_PETSC)
     Msg::Info("Implicit scheme based on PETSc");
     esys = new implicitHulbertChungPetsc<double>(_alpham,_alphaf,_beta,_gamma);
   #else
     Msg::Warning("petsc is not install on this computer. Use blas explicit system");
   #endif // HAVE_PETSC
  }
  else{
    Msg::Error("Petsc is not installed");
  }

   return esys;

}

linearSystem<double>* nonLinearMechSolver::createEigenSystem(){
  linearSystem<double> *sys = NULL;
  #if defined(HAVE_PETSC)
  sys = new linearSystemPETSc<double>();
  if(_solver_options.size())
    sys->setParameter(std::string("petsc_solver_options"),_solver_options);
  #else
  Msg::Error("Petsc is not intalled");
  #endif // HAVE_PETSC
  return sys;
}

void nonLinearMechSolver::createSystem()
{
  std::string Aname("A");
  if(_vcompBySys.size()==0)
  {
    /* System creation */
    linearSystem<double> *lsys=NULL;
    linearSystem<double> *lsys2 = NULL;
    if((!_previousInit) or (_previousScheme != whatScheme) or (!_notResetUnknowns))
    {
      if(whatScheme==Explicit)
      {
        lsys = this->createExplicitSystem();
        _mpiDofManager = false; // in explicit we need to useGlobalSystemPosition  = false to not use a global system position
      }
      else if (whatScheme == Implicit){
        lsys = this->createImplicitSystem();
      }
      else if (whatScheme == Eigen){
        lsys = this->createEigenSystem();
        if (_eigenSolverType == DYNAMIC)
          lsys2 = this->createEigenSystem();
      }
      else
        lsys = this->createSNLSystem();
    }
    else
    {
      lsys = pAssembler->getLinearSystem(Aname);
      if(whatScheme==Explicit)
      {
        explicitHulbertChung<double>* esys = static_cast<explicitHulbertChung<double>*>(lsys);
        esys->setDynamicRelaxation(_dynamicRelaxation);
      }
      if(whatScheme==Eigen) Msg::Error("Eigen solver in createSystem");
    }

    /* Create a dofManager for the system */
    if(!_previousInit or !_notResetUnknowns){
      dofManager<double>* oldAssembler = pAssembler;
      if(whatScheme!=StaticLinear and whatScheme != Eigen){
        pAssembler = new staticDofManager<double>(lsys,whatScheme,_mpiDofManager);
        if (oldAssembler){
          staticDofManager<double>* oldStaticDof = dynamic_cast<staticDofManager<double>*>(oldAssembler);
          if (oldStaticDof)
            dynamic_cast<staticDofManager<double>*>(pAssembler)->collecDataFromPrevious(oldStaticDof);
        }

      }
      else if(whatScheme == Eigen){
        if (lsys2 != NULL and lsys != NULL)
          pAssembler = new dofManager<double>(lsys,lsys2);
        else
          pAssembler = new dofManager<double>(lsys);

      }
      else
        pAssembler = new dofManager<double>(lsys);

      if (oldAssembler) {
        delete oldAssembler->getLinearSystem(Aname);
        delete oldAssembler;
      }
    }
    else{
      if(_previousScheme != whatScheme)
      {
        staticDofManager<double>* oldass = static_cast<staticDofManager<double>*>(pAssembler);
        pAssembler = new staticDofManager<double>(lsys,oldass,whatScheme);
        delete oldass->getLinearSystem(Aname);
        delete oldass;
        if(whatScheme==Eigen) Msg::Error("Eigen solver in createSystem");
      }
    }
  }
  else // Switch allowed only from a dofManagerMultiSystem
  {
    if(!_previousInit or (!_notResetUnknowns)){
      dofManager<double>* oldAssembler = pAssembler;

      pAssembler = new dofManagerMultiSystems<double>(StaticNonLinear);
      dofManagerMultiSystems<double>* allManager = static_cast<dofManagerMultiSystems<double>*>(pAssembler);
      for(int i=0;i<_vcompBySys.size();i++){
        linearSystem<double> *lsys;
        bool globalPos = true;
        if(_vschemeBySys[i] == Explicit)
        {
          lsys = this->createExplicitSystem();
          globalPos = false;
        }
        else
          lsys = this->createSNLSystem();
        dofManager<double>* currentAssembler = new staticDofManager<double>(lsys,_vschemeBySys[i],globalPos);
        allManager->add(currentAssembler,_vcompBySys[i]);
        if (oldAssembler){
          staticDofManager<double>* oldStaticDof = dynamic_cast<staticDofManager<double>*>(oldAssembler);
          if (oldStaticDof)
            dynamic_cast<staticDofManager<double>*>(currentAssembler)->collecDataFromPrevious(oldStaticDof);
        }
       }
       if (oldAssembler) {
          delete oldAssembler->getLinearSystem(Aname);
          delete oldAssembler;
        }
    }
    else{
      dofManagerMultiSystems<double> *oldass = dynamic_cast<dofManagerMultiSystems<double>*>(pAssembler);
      if(oldass!=NULL) // init from another multi system
      {
        for(int i=0;i<_vschemeBySys.size();i++)
        {
          // get the previous scheme
          staticDofManager<double> *oldmsys = dynamic_cast<staticDofManager<double>*>(oldass->getManager(i));
          scheme oldscheme = StaticLinear;
          if(oldmsys!=NULL)
          {
            oldscheme = oldmsys->getScheme();
          }

          // change the dofManger only if the scheme is different from the previous one
          if(_vschemeBySys[i]!=oldscheme)
          {
            // create a new system
            linearSystem<double> *lsys;
            if(_vschemeBySys[i]==Explicit)
            {
              lsys = this->createExplicitSystem();
            }
            else
            {
              lsys = this->createSNLSystem();
            }
            // create new manager
            dofManager<double> *newmanager = new staticDofManager<double>(lsys,oldmsys,_vschemeBySys[i]);
            oldass->reset(newmanager,i);
            // remove the old manager
            delete oldmsys->getLinearSystem(Aname);
            delete oldmsys;
          }
        }
      }
      else // init from a simple system
      {
        Msg::Info("A multi-system is initialized from a previous single system");
        // create the new systems
        std::vector<linearSystem<double>*> vsys(_vschemeBySys.size(),NULL);
        for(int i=0;i<_vschemeBySys.size();i++)
        {
          _vschemeBySys[i]==Explicit ? vsys[i] = this->createExplicitSystem() : vsys[i] = this->createSNLSystem();
        }
        // new dofManager
        staticDofManager<double> *oldstat = dynamic_cast<staticDofManager<double>*>(pAssembler);
        pAssembler = new dofManagerMultiSystems<double>(StaticNonLinear,oldstat,vsys,_vcompBySys,_vschemeBySys);
        // remove the old system and dofManager
        delete oldstat->getLinearSystem(Aname);
        delete oldstat;
        Msg::Info("Multi-system is ready");
      }
    }
  }

	// initial time step
	if(whatScheme==Explicit or whatScheme==Multi){
		staticDofManager<double>* sAssembler = dynamic_cast<staticDofManager<double>*>(pAssembler);
		if (sAssembler != NULL){
			double timeStep = _gammas * this->initialCriticalExplicitTimeStep();
			#if defined(HAVE_MPI)
			if(Msg::GetCommSize()>1){
				// in case of mpi the time step is choosen as the minimal value compute on each partition
				double mpiTimeStep;
				MPI_Allreduce(&timeStep,&mpiTimeStep,1,MPI_DOUBLE,MPI_MIN,MPI_COMM_WORLD);
				timeStep = mpiTimeStep;
			}
			#endif // HAVE_MPI
			Msg::Info("initial time step: %e",timeStep);
			sAssembler->setTimeStep(timeStep);
		}

	}
}

int nonLinearMechSolver::NewtonRaphsonPathFollowing(const int numstep){
   std::string name = "A";
  linearSystem<double>* lsys =pAssembler->getLinearSystem(name);
  nonLinearSystem<double>* nonsys = dynamic_cast<nonLinearSystem<double>*>(lsys);
	pathFollowingSystem<double>* pathsys = dynamic_cast<pathFollowingSystem<double>*>(lsys);
  staticDofManager<double>* staticAssembler = static_cast<staticDofManager<double>*>(pAssembler);

  // copy state from previous
	//_ipf->copy(IPStateBase::previous,IPStateBase::current);
	double normFinf= this->computeRightHandSide();

	// loop until convergence
  int iter=0;
  double relnorm = 0.;

	double norm0 = staticAssembler->norm0Inf();
	if (norm0 > 0) relnorm = normFinf/norm0;

	double pfRef = pathsys->getPathFollowingConstraintRelativeRedidual();

  Msg::Info("iteration n %d : residu : %e, absolute residu : %e norm0: %e pf residual: %e",iter,relnorm,normFinf,norm0,pfRef);

	while (((relnorm>_tol) and (normFinf>_absTol)) or ((pfRef> _tol) and (pfRef> _absTol)) or (iter == 0)){
    iter++;
    // Compute Stiffness Matrix
    this->computeStiffMatrix();

    // Solve KDu = Fext-Fint
    int succeed = staticAssembler->systemSolveIntReturn();
    if(!succeed)
    {
      iter = _maxNRite;
      break;
    }
    // update ipvariable
    _ipf->compute1state(IPStateBase::current,true);
		// update load
		double loadParam = pathsys->getControlParameter();
		double dloadParam = fabs(pathsys->getControlParameterStep());
		this->oneStepPreSolve(loadParam,dloadParam,numstep);

    if (!_iterativeNR) break;
    // new forces
    normFinf = this->computeRightHandSide();

    if (norm0 == 0.) {
      norm0 = staticAssembler->norm0Inf();
    }

    // Check of convergence
    relnorm = normFinf/norm0;
		pfRef = pathsys->getPathFollowingConstraintRelativeRedidual();
    Msg::Info("iteration n %d : residu : %e, absolute residu : %e pf residual: %e",  iter,relnorm,normFinf,pfRef);

    if((iter == _maxNRite) or std::isnan(relnorm))
    {
      iter = _maxNRite; // for isnan case
      break;
    }
  }

  return iter;
};

void nonLinearMechSolver::perturbationProcedure(const double pre, double* mode){
   /*
    Perturbation is case of simple buckling mode
   */
  if (pre>this->_instabilitycriterion) return;
  Msg::Warning("buckling occurs, criterion = %f < %f",pre,_instabilitycriterion);
  int sizeR = pAssembler->sizeOfR();
  /* Get infinity norm of mode to give the unity vector*/
  double norm = 0;
  for (int i=0; i<sizeR; i++){
   if (fabs(mode[i])> norm) norm = fabs(mode[i]);
  }

  std::string name = "A";
  linearSystem<double>* lsys = pAssembler->getLinearSystem(name);
  double normSol = lsys->normInfSolution();

  Msg::Warning("Perturbating to system: ");
  double factor = (_perturbationfactor*normSol)/norm;
  /* Perturbation to solution*/
  for (int i=0; i<sizeR; i++){
   double val = mode[i]*factor;
   lsys->addToSolution(i,val);
  }
}

double nonLinearMechSolver::solveSNL()
{
  this->initializeStaticScheme();
  staticDofManager<double> *sAssembler = static_cast<staticDofManager<double>*>(pAssembler);
  // change to introduce a dynamic modification of time step in case of convergence problem
  double curtime = 0.;
  double lasttime = 0.; // to avoid precision problem curtime = lasttime in place of curtime -= dt if no convergence
  double dendtime = double(endtime);
  double dtmax = dendtime/double(_numstepImpl); // time step value prescribed by user (time step can not be greater than this value)
  double dt = dtmax; // initial time step value
  int ii = 0; // iteration counter
  double tolend = 0.000001*dendtime; // OTHERWISE LAST STEP CAN BE DONE TWICE ??
  double toldtmax = 0.000001*dtmax; // IDEM (PRECISION PROBLEM)
  _timeStepAdaptationFactor = 1.;
  int numberTimeStepReduction=0;
  if(restartManager::available() and !_resetRestart)
  {
    Msg::Barrier(); // To wait rank0 before end
    restartManager::openRestartFile("nonLinearMechSolver");
    restartManager::restart(this);
    // restart takes only nonm const argument! So convert numstep and curtime
    restartManager::restart(ii);
    restartManager::restart(curtime);
    lasttime=curtime;
    restartManager::closeRestartFile();

    restartManager::InitFromRestartFile();
    Msg::Barrier(); // To wait rank0 before end

    struct stat st;
    if(stat("beforeRestart",&st) == 0)
    {
      std::ostringstream oss;
      oss << ii;
      std::string cpDir="mv beforeRestart restart"+oss.str();
      system(cpDir.c_str());
    }
    #if defined(HAVE_MPI)
    if(Msg::GetCommSize()>1)
    { //we need to reinitializae after the restart of the nonLinearSolver
      staticDofManager<double>* sAssembler = static_cast<staticDofManager<double>*>(pAssembler);
      Msg::Info("MPI restart initialization begins");
      sAssembler->systemMPIComm(); // To initiate the MPI Communication
      Msg::Info("MPI restart initialization OK");
    }
    #endif // HAVE_MPI
  }
  _resetRestart=false;
  Msg::Barrier(); // To wait rank0 before end

  while(fabs(curtime-dendtime) > tolend ){
    curtime+=dt;
    if(curtime > dendtime)
    {
      dt = dt -(curtime - dendtime);
      curtime = dendtime;
    }
    Msg::Info("t= %e on %e",curtime,dendtime);

    int niteNR  = this->oneStaticStep(curtime,dt,ii+1);

    bool redoStepWithSelectiveUpdate = false;
    if (niteNR < _maxNRite){
      // check for monitoring or selective update
      if (!_ipf->isAllChecked() or _endSchemeMonitoring or (_selectiveObject != NULL)){
        _ipf->checkAll();
      }

      if (_selectiveObject != NULL){
        _selectiveObject->checkSelectiveUpdate(this);
        if (_selectiveObject->solverIterateBySelectiveUpdate()){
          redoStepWithSelectiveUpdate = true;
        }
      }

      if (_endSchemeMonitoring ){
        // check of end cr is satisfy
        _endSchemeMonitoringObject->checkEnd(this);
        if (!_endSchemeMonitoringObject->solverEndedByMornitoring()){
          if (_endSchemeMonitoringObject->reduceTimeStepByMornitoring()){
            Msg::Warning("reduce by monitoring");
            niteNR = _maxNRite;
          }
        }
      }
    }

    if(niteNR == _maxNRite) // time step reduction
    {
       numberTimeStepReduction ++;
       if(numberTimeStepReduction > _maxAttemptStepReduction) // end of simulation
       {
         Msg::Error("Simulation end due to convergence problem! Values are set to the last converged time step in case you want to switch to another scheme");
				 // back to previous step if sovler ended by reaching maximal number of step reduction
				 sAssembler->resetUnknownsToPreviousTimeStep();
		     _ipf->copy(IPStateBase::previous,IPStateBase::current);
		     if (_selectiveObject != NULL){
		      _selectiveObject->resetToPreviousStep();
		     }
		     if (_endSchemeMonitoring){
            _endSchemeMonitoringObject->resetToPreviousStep();
		     }
         break;
       }
       Msg::Warning("Convergence of Newton-Raphson failed --> reduced time step");


       curtime = lasttime;
       dt /= _timeStepFactorReduction;
       _timeStepAdaptationFactor /= _timeStepFactorReduction;
       sAssembler->setTimeStep(dt);
       sAssembler->resetUnknownsToPreviousTimeStep();
       _ipf->copy(IPStateBase::previous,IPStateBase::current);
       if (_selectiveObject != NULL){
         _selectiveObject->resetToPreviousStep();
       }
       if (_endSchemeMonitoring){
         _endSchemeMonitoringObject->resetToPreviousStep();
       }
    }
    else if (redoStepWithSelectiveUpdate){
      curtime = lasttime;
      Msg::Warning("redo solving with selective update");
      _selectiveObject->selectiveUpdate(this);
    }
    else
    {
      lasttime = curtime;
      this->oneStepPostSolve(curtime,ii+1);
      ii ++; 
      // save for next step
      if (_timeStepAdaptation)
      {
        _timeStepAdaptationFactor = pow((double)_numNROptimal/(double)niteNR,_expNROptimal);
      }
      else{
        _timeStepAdaptationFactor = 1.0;
      }
      
			//
      if (_timeStepAdaptation && !_pathFollowing)
      {
        double dtPrev = dt;
        dt *= _timeStepAdaptationFactor;
        if (dt > _maximalTimeStep)
        {
          dt = _maximalTimeStep;
          _timeStepAdaptationFactor = _maximalTimeStep/dtPrev;
        }
        if (dt < _minimalTimeStep)
        {
          dt = _minimalTimeStep;
          _timeStepAdaptationFactor = _minimalTimeStep/dtPrev;
        }
        if (dtmax < dt){
         dtmax = dt;
       }
        sAssembler->setTimeStep(dt);
      }
      else if(niteNR <= _niteTimeStepIncrease && fabs(dtmax-dt) > toldtmax) // increase time step if convergence in a few iteration
      {
         numberTimeStepReduction --;
         dt *= _timeStepFactorReduction;
         if (_timeStepFactorReduction > 1.){
           Msg::Info("Time step is increased due to convergence of Newton-Raphson in less than %d",_niteTimeStepIncrease);
         }
         if (dtmax < dt){
           dtmax = dt;
         }
         _timeStepAdaptationFactor *= _timeStepFactorReduction;
         sAssembler->setTimeStep(dt);
      }

      
      if (_endSchemeMonitoring) {
        if (_endSchemeMonitoringObject->solverEndedByMornitoring()){
          Msg::Info("Computation is finished following _endSchemeMonitoring");
          _endSchemeMonitoringObject->nextStep();
          break;
        }
        // normal nextstep
        _endSchemeMonitoringObject->nextStep();
      }
    }
  }

  /* end of scheme */
  return this->finalizeStaticScheme(curtime,ii);
}

double nonLinearMechSolver::solveMulti()
{
  Msg::Info("Mixed static and dynamic solving: endtime=%e",this->getEndTime());
  Msg::Info("Static Data: tolerance: %lf",_tol);
  Msg::Info("Dynamic Data: beta=%f gamma=%f alpha_m=%f",this->_beta, this->_gamma, this->_alpham);
  /* init data */
  this->init();

  // set the first time step to 1 if previous computation to avoid initialization problem
  int step = 0; // iteration counter
  if(_previousInit) step = 1;
  this->init2();
  // iterative scheme (loop on timestep)

  staticDofManager<double> *sAssembler = static_cast<staticDofManager<double>*>(pAssembler);
  dofManagerMultiSystems<double> *multiAssembler = static_cast<dofManagerMultiSystems<double>*>(pAssembler);

  /* mass matrix computation outside the loop */
  for(std::vector<partDomain*>::iterator it = domainVector.begin(); it!=domainVector.end(); ++it){
    partDomain *dom = *it;
    AssembleMass(dom->getBilinearMassTerm(),*(dom->getFunctionSpace()),dom->g_cbegin(),dom->g_cend(),
               *(dom->getBulkGaussIntegrationRule()),*pAssembler);
  }
  if(_notResetedContact){
    for(nonLinearMechSolver::contactContainer::iterator itc = _allContact.begin(); itc!= _allContact.end(); ++itc){
      contactDomain *cdom = *itc;
      if(cdom->isRigid()){
        rigidContactSpaceBase *rcsp = static_cast<rigidContactSpaceBase*>(cdom->getSpace());
        AssembleMass(cdom->getMassTerm(),rcsp,pAssembler);
      }
    }
  }

  _currentStep = 0;
  sAssembler->nextStep();   // save solution of previous (initially last step == initial step BEAWARE FOR RESTART)
  /* time initialization */
  double curtime = 0.;
  double lasttime = 0.; // to avoid precision problem curtime = lasttime in place of curtime -= dt if no convergence
  double dendtime = double(endtime);
  double dtmax; // time step value prescribed by user (time step can not be greater than this value)
  double dt; // initial time step value
  double tolend = 0.000001*dendtime; // OTHERWISE LAST STEP CAN BE DONE TWICE ??
  double toldtmax; // IDEM (PRECISION PROBLEM)
  int numberTimeStepReduction=0;
  // if one system is explicit --> avoid the reduction of time step
  bool oneExplicitScheme=false;

  bool forceTimeStepEvaluation=false;
  if(restartManager::available() and !_resetRestart)
  {
    Msg::Barrier(); // To wait rank0 before end
    restartManager::openRestartFile("nonLinearMechSolver");
    restartManager::restart(this);
    // restart takes only nonm const argument! So convert numstep and curtime
    restartManager::restart(step);
    restartManager::restart(curtime);
    lasttime=curtime;
    restartManager::closeRestartFile();

    restartManager::InitFromRestartFile();
    Msg::Barrier(); // To wait rank0 before end

    struct stat st;
    if(stat("beforeRestart",&st) == 0)
    {
      std::ostringstream oss;
      oss << step;
      std::string cpDir="mv beforeRestart restart"+oss.str();
      system(cpDir.c_str());
    }
 #if defined(HAVE_MPI)
   if(Msg::GetCommSize()>1)
   { //we need to reinitializae after the restart of the nonLinearSolver
     //staticDofManager<double>* sAssembler = static_cast<staticDofManager<double>*>(pAssembler);
     Msg::Info("MPI restart initialization begins");
     sAssembler->systemMPIComm(); // To initiate the MPI Communication
     Msg::Info("MPI restart initialization OK");
   }
 #endif // HAVE_MPI
   forceTimeStepEvaluation=true;
   this->setTimeForBC(curtime); //for critical time step?
   //sAssembler->nextStep();
  }
/*  else
  {
    if(_previousInit)
      sAssembler->nextStep();   // save solution of previous (initially last step == initial step BEAWARE FOR RESTART)
  }*/
  _resetRestart=false;
  Msg::Barrier(); // To wait rank0 before end




  for(int i=0;i<_vschemeBySys.size();i++)
  {
    if(_vschemeBySys[i]==Explicit)
    {
      _maxAttemptStepReduction = 0;
      oneExplicitScheme = true;
    }
  }
  if(!oneExplicitScheme)
  {
    dt=dendtime/_numstepImpl;
    dtmax = dt;
  }
  else if (step==1 or forceTimeStepEvaluation) // compute the explicit time step in case of a previousScheme
  {
    dt = _gammas * this->criticalExplicitTimeStep();
    #if defined(HAVE_MPI)
     if(Msg::GetCommSize()>1){
       // in case of mpi the time step is choosen as the minimal value compute on each partition
       double mpiTimeStep;
       MPI_Allreduce(&dt,&mpiTimeStep,1,MPI_DOUBLE,MPI_MIN,MPI_COMM_WORLD);
       dt = mpiTimeStep;
      }
    #endif // HAVE_MPI
    Msg::Info("time step value: %e at time %e",dt,curtime);
    sAssembler->setTimeStep(dt);
    dtmax = dt;
    toldtmax = 1e-6*dtmax;
    forceTimeStepEvaluation=false;
  }

  /* time loop */
  while(fabs(curtime-dendtime) > tolend ){
    // compute the time step (explicit limitation)
    if(step%_numstepExpl ==0 and oneExplicitScheme){
      dt = _gammas * this->criticalExplicitTimeStep();
      #if defined(HAVE_MPI)
       if(Msg::GetCommSize()>1){
          // in case of mpi the time step is choosen as the minimal value compute on each partition
          double mpiTimeStep;
          MPI_Allreduce(&dt,&mpiTimeStep,1,MPI_DOUBLE,MPI_MIN,MPI_COMM_WORLD);
          dt = mpiTimeStep;
       }
      #endif // HAVE_MPI
      Msg::Info("time step value: %e at time %e",dt,curtime);
      sAssembler->setTimeStep(dt);
      dtmax = dt;
      toldtmax = 1e-6*dtmax;
    }
    else if(!oneExplicitScheme)
      Msg::Info("time step value: %e at time %e",dt,curtime);
    curtime+=dt;
    if(curtime > dendtime){curtime = dendtime;}

    this->oneStepPreSolve(curtime,dt,step+1);
    // Solve one step by NR scheme
    int niteNR=0;
    if(step%_numstepImpl==0 or !oneExplicitScheme)
    {
      Msg::Info("Solve implicit system(s)");
      double tsystresol = Cpu();
      tsystresol = Cpu() - tsystresol;
      niteNR = NewtonRaphson(step+1);
      Msg::Info("Time of Newton-Raphson resolution n %f: ",tsystresol);

      if(niteNR == _maxNRite) // time step reduction
      {
        numberTimeStepReduction ++;
        if(numberTimeStepReduction > _maxAttemptStepReduction) // end of simulation
        {
          Msg::Error("Simulation end due to convergence problem! Values are set to the last converged time step in case you want to switch to another scheme");
          // reset value to the previous converged time step and break (to switch from implicit to explicit in the pyfile for exemple)
          break;
        }
        Msg::Warning("Convergence of Newton-Raphson failed --> reduced time step");
        curtime = lasttime;
        dt /= _timeStepFactorReduction;
      }
      else if(niteNR!=0)
      {
        this->oneStepPostSolve(curtime,step+1);
        
        if(fabs(dtmax-dt) > toldtmax ) // increase time step if convergence in a few iteration
        {
          if(niteNR <= _niteTimeStepIncrease)
          {
            Msg::Warning("Time step is increased due to convergence of Newton-Raphson in less than %d",_niteTimeStepIncrease);
            numberTimeStepReduction --;
            dt *= _timeStepFactorReduction;
          }
        }
        lasttime = curtime;
        
        
      }
    }
    if(niteNR==0) // no Implicit computation or no force on the implicit systems
    {
      // solve only the Explicit scheme
      int globalsuc = 0;
      for(int i=0;i<_vschemeBySys.size();i++)
      {
        staticDofManager<double> *nlass = static_cast<staticDofManager<double>*>(multiAssembler->getManager(i));
        if(nlass->getScheme()==Explicit)
          globalsuc = nlass->systemSolveIntReturn();
      }
      if(globalsuc){ // otherwise no resolution (QS scheme only with no forces) !!
        // compute the new forces
        _ipf->compute1state(IPStateBase::current,false);
        double normRHS = this->computeRightHandSide();
        if(std::isnan(normRHS))
        {
          Msg::Error("Nan force value --> end");
          break;
        }
        if(_dynamicRelaxation)
        {
          double norm0 = sAssembler->norm0Inf();
          if(step%_numstepExpl ==0 ) Msg::Info("Rel norm for dynamic relaxation %e tol = %e",normRHS/norm0,_tol);
          if(normRHS/norm0 < _tol){
            Msg::Info("Dynamic relaxation procedure is achieved");
            break;
          }
          if(curtime == endtime)
            Msg::Info("Dynamic relaxation ended as the final time is reached. But convergence is not achieved!");
        }
      }
      lasttime = curtime;
      this->oneStepPostSolve(curtime,step+1);
    }
    step ++; // increase counter of iteration even if no convergence --> more archivage if no convergence (can be usefull)
  }
  /* end of scheme */
  this->endOfScheme(curtime,step);
  Msg::Info("Multi OK");
  return curtime;
}
void nonLinearMechSolver::unknownBuildView(const int comp,const int nbstep)
{
  int nstep_;
  nsba==0 ? nstep_ = nbstep : nstep_=nsba;
  
  switch(comp){
   case 1:
    unknownView.emplace_back("velocity",1,nlsField::crude,nstep_);
    break;
   case 2:
    unknownView.emplace_back("acceleration",2,nlsField::crude,nstep_);
    break;
   default:
    unknownView.emplace_back("displacement",0,nlsField::crude,nstep_);
  }
}

void nonLinearMechSolver::OneUnknownBuildView(const std::string vname, const int comp,const int nbstep)
{
  int nstep_;
  nsba==0 ? nstep_ = nbstep : nstep_=nsba;
  
  int unknownComp = -1 - comp;
  unknownView.emplace_back(vname,unknownComp,nlsField::crude,nstep_);
};

void nonLinearMechSolver::internalPointBuildView(const std::string vname,const int comp,
                                                 const int nbstep,const int ev)
{
  nlsField::ElemValue ev_;
  switch(ev){
   case 0:
    ev_ = nlsField::crude;
    break;
   case 1:
    ev_ = nlsField::mean;
    break;
   case 2 :
    ev_ = nlsField::min;
    break;
   case 3 :
    ev_ = nlsField::max;
    break;
   default:
    ev_ = nlsField::mean;
  }
  int nstep_;
  nsba==0 ? nstep_ = nbstep : nstep_=nsba;
  ipView.emplace_back(vname,comp,ev_,nstep_);
}
void nonLinearMechSolver::energyBuildView(const std::string vname,const int comp,const int nbstep)
{
  int nstep_;
  nsba==0 ? nstep_ = nbstep : nstep_=nsba;
  switch(comp){
   case 1:
    energyView.emplace_back("kinetic",1,nlsField::crude,nstep_);
    break;
   case 2:
    energyView.emplace_back("deformation",2,nlsField::crude,nstep_);
    break;
   case 3:
    energyView.emplace_back("external work",3,nlsField::crude,nstep_);
    break;
   default:
    energyView.emplace_back("total",4,nlsField::crude,nstep_);
  }
}

void nonLinearMechSolver::energyComputation(const int val)
{
  _energyComputation=val;
}

void nonLinearMechSolver::fractureEnergy(const int val)
{
  _fractureEnergyComputation=val;
}

void nonLinearMechSolver::fillConstraintDof()
{

  //for(nonLinearConstraintBC& cbc : allConstraint)
  for(std::list<nonLinearConstraintBC>::iterator cbc=allConstraint.begin(); cbc!=allConstraint.end(); cbc++)
  {
    cbc->constraintDofs.clear();
    std::vector<MVertex*> vv;
    int dim=0;
    switch(cbc->onWhat){
     case nonLinearBoundaryCondition::ON_VERTEX:
      dim = 0;
      break;
     case nonLinearBoundaryCondition::ON_EDGE:
      dim = 1;
      break;
     case nonLinearBoundaryCondition::ON_FACE:
      dim = 2;
      break;
     case nonLinearBoundaryCondition::ON_VOLUME:
      dim = 3;
      break;
    }
    pModel->getMeshVerticesForPhysicalGroup(dim,cbc->_tag,vv);

    // loop on domain (to find the domain where the force is applied)
    for(std::vector<partDomain*>::iterator itdom=domainVector.begin(); itdom!=domainVector.end(); ++itdom){
      partDomain *dom = *itdom;
      FunctionSpaceBase *sp = dom->getFunctionSpace();
      std::vector<Dof> R;
      if(!(dom->getFormulation())){  // Cg/Dg
        FilterDof* filter = dom->createFilterDof(cbc->_comp);
        for(groupOfElements::vertexContainer::const_iterator itv=dom->g_cvbegin(); itv!=dom->g_cvend(); ++itv){
          MVertex *ver = *itv;
          for(std::vector<MVertex*>::iterator it=vv.begin(); it!=vv.end(); ++it){
            MVertex *vver = *it;
            if(ver == vver){
              MPoint ele(ver);
              // build the dof
              R.clear();
              sp->getKeys(&ele,R);
              for(int j=0;j<R.size(); j++)
                if(filter->operator()(R[j]))
                  cbc->constraintDofs.push_back(R[j]);
              break;
            }
          }
        }
        delete filter;
      }
      else{ // full Dg
        for(groupOfElements::elementContainer::const_iterator ite = dom->g_cbegin(); ite!=dom->g_cend(); ++ite){
          MElement *ele = *ite;
          int nbvertex = ele->getNumVertices();
          sp->getKeys(ele,R);
          for(std::vector<MVertex*>::iterator it=vv.begin(); it!=vv.end(); ++it){
            for(int j=0;j<nbvertex; j++){
              if(ele->getVertex(j) == *it){
                cbc->constraintDofs.push_back(R[j+nbvertex*cbc->_comp]);
                break;
              }
            }
          }
          R.clear();
        }
      }
    }
    cbc->constraintDofs.unique();
  }
};

void nonLinearMechSolver::applyConstraintBC()
{
  //for(const nonLinearConstraintBC& cbc: allConstraint){
  for(std::list<nonLinearConstraintBC>::const_iterator cbc=allConstraint.begin(); cbc!=allConstraint.end(); cbc++){
    for (std::list<Dof>::const_iterator it =cbc->constraintDofs.begin(); it != cbc->constraintDofs.end(); it++)
    {
      if (it!=cbc->constraintDofs.begin())
      {
        DofAffineConstraint<double> cons;
        cons.linear.push_back(std::pair<Dof,double>(*(cbc->constraintDofs.begin()),1));
        cons.shift=0;
        pAssembler->setLinearConstraint(*it,cons);
        cons.linear.clear();
      }
    };
  }
};

void nonLinearMechSolver::initCornerBC(){
  if (!_testFlag) {
    Msg::Warning("fix corner Dof is only used for microscopic test only");
    return;
  }
  if (_pbcGroup != NULL){
    for(std::list<nonLinearDirichletBCAtCorner>::iterator cbc=allCornerConstraint.begin(); cbc!=allCornerConstraint.end(); cbc++){
      std::vector<Dof>& allDof = cbc->constraintDofs;
      std::vector<Dof> R;
      MVertex* v = _pbcGroup->getCornerVertex(cbc->_cornerNumber);
      _pbcGroup->getDofsOfCornerVertex(cbc->_cornerNumber,R);
      allDof.push_back(R[cbc->_comp]);
      double val = cbc->_f->operator()(v->x(),v->y(),v->z());
      pAssembler->fixDof(R[cbc->_comp],val);
    }

    for(std::list<nonLinearNeumannBCAtCorner>::iterator cbc=allCornerForce.begin(); cbc!=allCornerForce.end(); cbc++){
      std::vector<Dof>& allDof = cbc->constraintDofs;
      std::vector<Dof> R;
      MVertex* v = _pbcGroup->getCornerVertex(cbc->_cornerNumber);
      _pbcGroup->getDofsOfCornerVertex(cbc->_cornerNumber,R);
      allDof.push_back(R[cbc->_comp]);
    }
  }
};

void nonLinearMechSolver::fixDofCorner(){
  if (!_testFlag) {
    Msg::Warning("fix corner Dof is only used for microscopic test only");
    return;
  }
  if (_pbcGroup != NULL){
    for(std::list<nonLinearDirichletBCAtCorner>::const_iterator cbc=allCornerConstraint.begin(); cbc!=allCornerConstraint.end(); cbc++){
      const std::vector<Dof>& allDof = cbc->constraintDofs;
      MVertex* v = _pbcGroup->getCornerVertex(cbc->_cornerNumber);
      for (int i=0; i< allDof.size(); i++){
        double val = cbc->_f->operator()(v->x(),v->y(),v->z());
        pAssembler->fixDof(allDof[i],val);
      }
    }
  }
};

void nonLinearMechSolver::applyPBCBetweenTwoGroups(){
	// create pbcmap
	if (isDgDomain()){
		for (std::list<nonLinearPeriodicBCBetweenTwoGroups>::iterator pbctg=allPeriodic.begin(); pbctg!=allPeriodic.end(); pbctg++){
			// create element map
			std::set<MElement*> gSlave(pbctg->g2->begin(),pbctg->g2->end());
			for (groupOfElements::elementContainer::iterator itMaster = pbctg->g1->begin(); itMaster!= pbctg->g1->end(); itMaster++){
				MElement* master = *itMaster;
				SPoint3 masterBaryCenter = master->barycenter();
				std::set<MElement*>::iterator foundSlaveIt = gSlave.begin();
				MElement* slave = *(foundSlaveIt);

				double length = masterBaryCenter.distance(slave->barycenter());
				// find closest bary center as slave element
				std::set<MElement*>::iterator slaveIt = foundSlaveIt;
				slaveIt++;
				for (slaveIt; slaveIt != gSlave.end(); slaveIt++){
					MElement* ele = *slaveIt;
					double lengthtemp = masterBaryCenter.distance(ele->barycenter());
					if (lengthtemp < length){
						slave = ele;
						length = lengthtemp;
						foundSlaveIt = slaveIt;
					};
				}
				// remove slave in queue
				gSlave.erase(foundSlaveIt);

				std::vector<MVertex*> vvMaster, vvSlave;
				master->getVertices(vvMaster);
				slave->getVertices(vvSlave);

			 // Msg::Error("found master = %d, slave = %d comp = %d",master->getNum(), slave->getNum(),allPeriodic[i].comp);

				partDomain* domMaster = NULL;
				partDomain* domSlave = NULL;
				MElement* eleMaster = NULL;
				MElement* eleSlave = NULL;

				bool masterFound = false;
				bool slaveFound = false;

				for (int idom =0; idom < domainVector.size(); idom++){
					partDomain* dom = domainVector[idom];
					for (groupOfElements::elementContainer::const_iterator itg = dom->g_cbegin(); itg != dom->g_cend(); itg++){
						MElement* ele = *itg;
						// find element associated with master part
            if (!masterFound){
              if (checkInterfaceElementBelongToBulkElement(master,ele)){
                masterFound = true;
                eleMaster = ele;
                domMaster = dom;
              }
            }

            if (!slaveFound){
              if (checkInterfaceElementBelongToBulkElement(slave,ele)){
                slaveFound = true;
                eleSlave = ele;
                domSlave = dom;
              }
            }
						if (masterFound and slaveFound) break;
					}
					if (masterFound and slaveFound) break;
				}

				if (masterFound and slaveFound){
          // local periodic city
					for (int iv=0; iv< vvMaster.size(); iv++){
						pbcVertex pvMaster(domMaster,eleMaster,vvMaster[iv]);
						// find the closest vertex in slave element
						MVertex* vn = vvSlave[0];
						double dist = vvMaster[iv]->distance(vn);
						for (int in = 1; in< vvSlave.size(); in++){
							double disttemp = vvMaster[iv]->distance(vvSlave[in]);;
							if (disttemp< dist){
								vn = vvSlave[in];
								dist = disttemp;
							}
						}

						pbcVertex pvSlave(domSlave,eleSlave,vn);

						std::vector<Dof> masterKeys, slaveKeys;
						std::vector<int> pbccomp;
						pbccomp.push_back(pbctg->comp);

						// get Dofs in vertex
						pvMaster.getKeys(pbccomp,masterKeys);
						pvSlave.getKeys(pbccomp,slaveKeys);

						// add to dofMap in PBC
						for (int ikey = 0; ikey< masterKeys.size(); ikey++){
							bool noConstraint = true;
							for (std::list<nonLinearPeriodicBCBetweenTwoGroups>::const_iterator pbc2=allPeriodic.begin(); pbc2!=allPeriodic.end(); pbc2++){
								if (pbc2->dofMap.find(masterKeys[ikey]) != pbc2->dofMap.end()){
									noConstraint = false;
									break;
								}
							}

							if (noConstraint){
								// find root
								bool createRootDofs = false;
								if (pbctg->rootDofs == NULL){
									if (pbctg->v1 !=NULL and  pbctg->v2!=NULL){
										if ((pbctg->v1->getNum() == pvMaster.getVertex()->getNum()) and (pbctg->v2->getNum() == pvSlave.getVertex()->getNum())){
											createRootDofs = true;
											pbctg->rootDofs = new std::pair<Dof,Dof>(masterKeys[ikey],slaveKeys[ikey]);
										}
									}
									else{
										createRootDofs = true;
										pbctg->rootDofs = new std::pair<Dof,Dof>(masterKeys[ikey],slaveKeys[ikey]);
									}
								}

								if (!createRootDofs){
									pbctg->dofMap.insert(std::pair<Dof,Dof>(masterKeys[ikey],slaveKeys[ikey]));
								}
							}
						}
					}

				}
			};
		}

    // apply constraint
    if (allPeriodic.size() > 0)
      Msg::Info("begin accounting PBC rank %d",Msg::GetCommRank());

    for (std::list<nonLinearPeriodicBCBetweenTwoGroups>::iterator pbctg=allPeriodic.begin(); pbctg!=allPeriodic.end(); pbctg++){
      std::map<Dof,Dof>& dofMap = pbctg->dofMap;
      if (dofMap.size() > 0){
        if (pbctg->rootDofs == NULL){
          Msg::Error("Dof has not been created nonLinearMechSolver::applyPBCBetweenTwoGroups");
        }
        std::pair<Dof,Dof>& rootDof = *(pbctg->rootDofs);

        DofAffineConstraint<double> cons;
        cons.shift=0;
        // other constraints are expressed in terms of root dofs
        for (std::map<Dof,Dof>::iterator itm = dofMap.begin(); itm != dofMap.end(); itm++){
          if ((!pAssembler->isConstrained(itm->first)) and (!pAssembler->isFixed(itm->first))){
            cons.linear.push_back(std::pair<Dof,double>(rootDof.first,1.));
            cons.linear.push_back(std::pair<Dof,double>(rootDof.second,-1.));
            cons.linear.push_back(std::pair<Dof,double>(itm->second,1.));
            pAssembler->setLinearConstraint(itm->first,cons);
            cons.linear.clear();
          }
        }
      }
    };
    if (allPeriodic.size() > 0)
      Msg::Info("end accounting PBC");

	}
	else{
	// CG case
    for (std::list<nonLinearPeriodicBCBetweenTwoGroups>::iterator pbctg=allPeriodic.begin(); pbctg!=allPeriodic.end(); pbctg++){
      nonLinearPeriodicBCBetweenTwoGroups& bc = *pbctg;
      groupOfElements* gMaster = bc.g1;
      groupOfElements* gSlave = bc.g2;
      MVertex* vRootMaster = bc.v1;
      MVertex* vRootSlave = bc.v2;

      // find space for master and slave root Vertex
      partDomain* vRootMasterDom = NULL;
      partDomain* vRootSlaveDom = NULL;

      for (int i=0; i< domainVector.size(); i++){
        partDomain* dom = domainVector[i];
        for (groupOfElements::vertexContainer::iterator itvdom = dom->g_cvbegin(); itvdom != dom->g_cvend(); itvdom++){
          if ((*itvdom)->getNum() == vRootMaster->getNum() and vRootMasterDom == NULL){
            vRootMasterDom = dom;
          }
          if ((*itvdom)->getNum() == vRootSlave->getNum() and vRootSlaveDom == NULL){
            vRootSlaveDom = dom;
          }

          if (vRootMasterDom != NULL and vRootSlaveDom != NULL)
            break;
        }
        if (vRootMasterDom != NULL and vRootSlaveDom != NULL)
          break;
      }

      for (groupOfElements::vertexContainer::iterator itvm = gMaster->vbegin(); itvm != gMaster->vend(); itvm++){
        MVertex* vMaster = *itvm;
        if (vMaster->getNum() != vRootMaster->getNum()){
          MElement* eleSlave = NULL;
          if (bc.onWhat == nonLinearBoundaryCondition::ON_VERTEX){
            eleSlave = *(gSlave->begin());
          }
          else{
            for (groupOfElements::elementContainer::iterator ite = gSlave->begin(); ite != gSlave->end(); ite++){
              MElement* es = *ite;
              if (InterpolationOperations::isInside(vMaster,es)){
                eleSlave = es;
                break;
              }
            }
          }


          if (eleSlave == NULL){
            Msg::Error("slave element is not found nonLinearMechSolver::nonLinearPeriodicBCBetweenTwoGroups");
          }
          else{
            SPoint3 point;
            std::vector<MVertex*> vv;
            int size = eleSlave->getNumVertices();
            eleSlave->getVertices(vv);

            SPoint3 pp = vMaster->point();
            if (eleSlave->getDim() == 0){
              point = vv[0]->point();
            }
            else if (eleSlave->getDim() == 1){
              point = project(pp,vv[0]->point(),vv[1]->point());
            }
            else if (eleSlave->getDim() == 2){
              point = project(pp,vv[0]->point(),vv[1]->point(),vv[2]->point());
            };

            double xyz[3]={point[0],point[1],point[2]};
            double uvw[3];
            eleSlave->xyz2uvw(xyz,uvw);

            double fVal[size];
            eleSlave->getShapeFunctions(uvw[0],uvw[1],uvw[2],fVal);

            // find space for master and slave domain
            partDomain* masterDom = NULL;
            partDomain* slaveDom = NULL;

            for (int i=0; i< domainVector.size(); i++){
              partDomain* dom = domainVector[i];
              for (groupOfElements::vertexContainer::iterator itvdom = dom->g_cvbegin(); itvdom != dom->g_cvend(); itvdom++){
                if ((*itvdom)->getNum() == vMaster->getNum()){
                  masterDom = dom;
                  break;
                }
              }
              if (masterDom != NULL) break;
            }

            for (int i=0; i< domainVector.size(); i++){
              partDomain* dom = domainVector[i];
              for (groupOfElements::elementContainer::iterator itedom = dom->g_cbegin(); itedom != dom->g_cend(); itedom++){
                if (checkInterfaceElementBelongToBulkElement(eleSlave,*itedom)){
                  slaveDom = dom;
                  break;
                }
              }
              if (slaveDom != NULL) break;
            }

            std::vector<int> comp;
            comp.push_back(bc.comp);

            std::vector<Dof> keyRootMaster, keyRootSlave;
            getKeysFromVertex(vRootMasterDom->getFunctionSpace(),vRootMaster,comp,keyRootMaster);
            getKeysFromVertex(vRootSlaveDom->getFunctionSpace(),vRootSlave,comp,keyRootSlave);

            std::vector<Dof> RMaster, RSlave;
            getKeysFromVertex(masterDom->getFunctionSpace(),vMaster,comp,RMaster);
            for (int iv=0; iv < size; iv++){
              getKeysFromVertex(slaveDom->getFunctionSpace(),vv[iv],comp,RSlave);
            }

            if (!pAssembler->isConstrained(RMaster[0]) and !pAssembler->isFixed(RMaster[0])){
              DofAffineConstraint<double> cons;
              cons.shift = 0.;
              for (int ivf=0; ivf< size; ivf++){
                cons.linear.push_back(std::pair<Dof,double>(RSlave[ivf],fVal[ivf]));
              }
              cons.linear.push_back(std::pair<Dof,double>(keyRootMaster[0],1.));
              cons.linear.push_back(std::pair<Dof,double>(keyRootSlave[0],-1.));
              pAssembler->setLinearConstraint(RMaster[0],cons);
            }
          }
				}
			};
		};
	};



  // for all average PBC
  if (allAveragePeriodic.size() > 0)
    Msg::Info("begin accounting average PBC rank %d",Msg::GetCommRank());
  for (std::list<nonLinearAveragePeriodicBCBetweenTwoGroups>::iterator itpbc = allAveragePeriodic.begin(); itpbc != allAveragePeriodic.end(); itpbc++){
    nonLinearAveragePeriodicBCBetweenTwoGroups& pbc = *itpbc;
    if (isDgDomain()){
      Msg::Error("nonLinearAveragePeriodicBCBetweenTwoGroups has not been implemented for fullDG case");
    }
    else{
      FunctionSpaceBase* space = domainVector[0]->getFunctionSpace(); // with CG, all functions generated dof based on vertex

      std::vector<MVertex*> vecMaster, vecSlave;
      fullVector<double> weightMaster, weightSlave;

      supplementConstraint::computeWeight(pbc.g1,vecMaster,weightMaster);
      supplementConstraint::computeWeight(pbc.g2,vecSlave,weightSlave);

      int sizeDofMaster = vecMaster.size();
      int sizeDofSlave = vecSlave.size();

      std::vector<Dof> allDof;
      fullVector<double> weight;
      weight.resize(sizeDofMaster+sizeDofSlave+2,true);
      std::vector<int> comp;
      comp.push_back(pbc.comp);

      for (int i=0; i< sizeDofMaster; i++){
        getKeysFromVertex(space,vecMaster[i],comp,allDof);
        weight(i) = weightMaster(i);
      }
      for (int i=0; i< sizeDofSlave; i++){
        getKeysFromVertex(space,vecSlave[i],comp,allDof);
        weight(i+sizeDofMaster) = -weightSlave(i);
      }

      getKeysFromVertex(space,pbc.v1,comp,allDof);
      weight(sizeDofMaster+sizeDofSlave) = -1.;
      getKeysFromVertex(space,pbc.v2,comp,allDof);
      weight(sizeDofMaster+sizeDofSlave+1) = 1.;

      // find positive dof
      int positive = -1;
      double maxWeight = 0.;
      for (int i=0; i< sizeDofMaster+sizeDofSlave; i++){
        if ((!pAssembler->isConstrained(allDof[i])) and (!pAssembler->isFixed(allDof[i]))
            and (!(allDof[i] == allDof[sizeDofMaster+sizeDofSlave]))
            and (!(allDof[i] == allDof[sizeDofMaster+sizeDofSlave+1]))){
          if (fabs(weight(i)) > fabs(maxWeight)){
            positive = i;
            maxWeight = weight(i);
          }
        }
      }

      Msg::Info("positive = %d weigth = %e v1 = %d v2 = %d comp = %d",positive,maxWeight,pbc.v1->getNum(),pbc.v2->getNum(),pbc.comp);

      if (positive == -1){
        Msg::Error("nonLinearAveragePeriodicBCBetweenTwoGroups can not be used since all Dofs are constrained or fixed in other BCs");
      }
      else{
        double fact = 1./maxWeight;
        weight.scale(fact);
        DofAffineConstraint<double> cons;
        cons.shift = 0.;
        for (int i=0; i< sizeDofMaster+sizeDofSlave+2; i++){
          if (i != positive){
            cons.linear.push_back(std::pair<Dof,double>(allDof[i],-weight(i)));
          }
        }
        pAssembler->setLinearConstraint(allDof[positive],cons);
      }
    }
  }
  if (allAveragePeriodic.size() > 0)
    Msg::Info("end accounting average PBC");
};

void nonLinearMechSolver::applySameDispBCBetweenTwoGroups(){
  for (std::list<nonLinearSameDisplacementBCBetweenTwoGroups>::iterator itbc =allSameDispBetweenTwoGroups.begin(); itbc !=allSameDispBetweenTwoGroups.end(); itbc++){
    nonLinearSameDisplacementBCBetweenTwoGroups& bc = *itbc;
    if (isDgDomain()) {
      Msg::Error("nonLinearSameDisplacementBCBetweenTwoGroups for DG has not been implemented nonLinearMechSolver::applySameDispBCBetweenTwoGroups");
    }
    else{
      groupOfElements* gMaster = bc.g1;
      groupOfElements* gSlave = bc.g2;

      for (groupOfElements::vertexContainer::iterator itv = gMaster->vbegin(); itv != gMaster->vend(); itv++){
        MVertex* vm = *itv;
        MElement* eSlave = NULL;

        if (bc.onWhat == nonLinearBoundaryCondition::ON_VERTEX){
          eSlave = *(gSlave->begin());
        }
        else{
          for (groupOfElements::elementContainer::iterator ite = gSlave->begin(); ite != gSlave->end(); ite++){
            MElement* es = *ite;
            if (InterpolationOperations::isInside(vm,es)){
              eSlave = es;
              break;
            }
          }
        }
        if (eSlave == NULL){
          Msg::Error("slave element is not found nonLinearMechSolver::applySameDispBCBetweenTwoGroups");
        }
        else{
          SPoint3 point;
          std::vector<MVertex*> vv;
          int size = eSlave->getNumVertices();
          eSlave->getVertices(vv);
          SPoint3 pp = vm->point();
          if (eSlave->getDim() == 1){
            point = project(pp,vv[0]->point(),vv[1]->point());
          }
          else if (eSlave->getDim() == 2){
            point = project(pp,vv[0]->point(),vv[1]->point(),vv[2]->point());
          };

          double xyz[3]={point[0],point[1],point[2]};
          double uvw[3];
          eSlave->xyz2uvw(xyz,uvw);

          double fVal[size];
          eSlave->getShapeFunctions(uvw[0],uvw[1],uvw[2],fVal);

          // find space for master and slave
          std::vector<partDomain*> domMaster;
          std::vector<partDomain*> domSlave;

          for (int i=0; i< domainVector.size(); i++){
            partDomain* dom = domainVector[i];
            for (groupOfElements::vertexContainer::iterator itvdom = dom->g_cvbegin(); itvdom != dom->g_cvend(); itvdom++){
              if ((*itvdom)->getNum() == vm->getNum()){
                domMaster.push_back(dom);
                break;
              }
            }
            //
            for (groupOfElements::elementContainer::iterator itedom = dom->g_cbegin(); itedom != dom->g_cend(); itedom++){
              if (checkInterfaceElementBelongToBulkElement(eSlave,*itedom)){
                domSlave.push_back(dom);
                break;
              }
            }
          }

          std::vector<int> comp;
          comp.push_back(bc.comp);

          for (int i=0; i< domMaster.size(); i++){
            FunctionSpaceBase* spMaster = domMaster[i]->getFunctionSpace();
            std::vector<Dof> RMaster;
            getKeysFromVertex(spMaster,vm,comp,RMaster);
            if (!pAssembler->isConstrained(RMaster[0]) and !pAssembler->isFixed(RMaster[0])){
              for (int j=0; j< domSlave.size(); j++){
                FunctionSpaceBase* spSlave = domSlave[j]->getFunctionSpace();
                std::vector<Dof> RSlave;
                 for (int iv=0; iv< size; iv++){
                  getKeysFromVertex(spSlave,vv[iv],comp,RSlave);
                }

                DofAffineConstraint<double> cons;
                cons.shift = 0.;
                for (int ivf=0; ivf< size; ivf++){
                  cons.linear.push_back(std::pair<Dof,double>(RSlave[ivf],fVal[ivf]));
                }
                pAssembler->setLinearConstraint(RMaster[0],cons);
              }
            }
          }
        }
      }
    }
  };
};

void nonLinearMechSolver::crackTracking(std::string fname)
{
 #if defined(HAVE_MPI)
  if(Msg::GetCommSize() > 1){
    std::ostringstream oss;
    oss << Msg::GetCommRank();
    fname += oss.str();
  }
 #endif // HAVE_MPI
  _crackTrackingFile = fopen(fname.c_str(),"w");
  fprintf(_crackTrackingFile,"time0,time1;inter num;num minus;num plus;x0;y0;z0;x1;y1;z1\n");
}

// In MPI only 1 final file on rank 0
void nonLinearMechSolver::postproFragment(std::string fname)
{
 #if defined(HAVE_MPI)
  if(Msg::GetCommSize()>1)
  {
     // change file name (include the part number in it
    std::ostringstream oss;
    oss << Msg::GetCommRank();
    std::string partnum = oss.str();
    // retrieve file extension
    size_t ext_pos;
    ext_pos = fname.find_last_of('.');
    std::string ext(fname,ext_pos+1,fname.size());
    // set file name
    std::string newname(fname,0,ext_pos);
    std::string fileName = newname + "_part" + partnum +"." + ext;
    // open file
    _fragmentationFile = fopen(fileName.c_str(),"w");
  }
  else
 #endif //HAVE_MPI
  {
    _fragmentationFile = fopen(fname.c_str(),"w");
  }
}

void nonLinearMechSolver::crackTracking(const double curtime)
{
  if(_crackTrackingFile != NULL) // do nothing otherwise
  {
     // loop on domain
     for(std::vector<partDomain*>::const_iterator itdom = domainVector.begin(); itdom != domainVector.end(); ++itdom)
     {
        partDomain *dom = *itdom;
        if(dom->IsInterfaceTerms())
        {
          dgPartDomain *dgdom = static_cast<dgPartDomain*>(dom);
          if((dgdom->getMaterialLawMinus()->getType() == materialLaw::fracture)
             and (dgdom->getMaterialLawPlus()->getType() == materialLaw::fracture)
             and (dgdom->getDim() == 2))
          {
            // loop on interface element
            QuadratureBase *integBound = dgdom->getInterfaceGaussIntegrationRule();
            IntPt *GP;
            const materialLaw2LawsInitializer *mfminus = dynamic_cast<const materialLaw2LawsInitializer*>(dgdom->getMaterialLawMinus());
            const materialLaw2LawsInitializer *mfplus = dynamic_cast<const materialLaw2LawsInitializer*>(dgdom->getMaterialLawPlus());
            for(groupOfElements::elementContainer::const_iterator itele = dgdom->gi->begin(); itele != dgdom->gi->end(); ++itele)
            {
              MElement *ele = *itele;
              AllIPState::ipstateElementContainer *vips = _ipf->getAips()->getIPstate(ele->getNum());
              int npts_inter = integBound->getIntPoints(ele,&GP);
              int jj;
              /* Assume same value on minus and plus element !!*/
              bool prevOther = false; // other previous Gauss point already broken
              for(jj=0;jj<npts_inter;jj++)
              {
                IPStateBase *ipsm = (*vips)[jj];
                IPVariable *ipvmcur = ipsm->getState(IPStateBase::current);
                IPVariable *ipvmprev= ipsm->getState(IPStateBase::previous);
                bool fbmp = mfminus->fullBroken(ipvmprev);
                bool fbmc = mfminus->fullBroken(ipvmcur);
                if((fbmc==false))
                  break;
                else if(fbmp==false){
                  if(prevOther==false)
                    prevOther = true;
                  else
                    break;
                }
              }
              if( (jj==npts_inter) and (prevOther==true))// element is broken when all Gauss points are broken
              {
                /* get the time of crack initialization at extremities !! */
                IPStateBase *ipsm = (*vips)[0];
                IPVariable *ipvmcur = ipsm->getState(IPStateBase::current);
                double time00 = mfminus->timeInitBroken(ipvmcur);
                ipsm = (*vips)[npts_inter-1];
                ipvmcur = ipsm->getState(IPStateBase::current);
                double time01 = mfminus->timeInitBroken(ipvmcur);
                double time0;
                int index1,index0; // 0 or 1
                if(time00 < time01){
                  time0 = time00;
                  index0 = 0;
                  index1 = 1;
                }
                else{
                  time0 = time01;
                  index0 = 1;
                  index1 = 0;
                }
                /* find current position of element */
                FunctionSpaceBase *sp = dgdom->getFunctionSpace();
                std::vector<Dof> R;
                MInterfaceLine *iele = dynamic_cast<MInterfaceLine*>(ele);  // track a 2D crack--> interface == line
                MElement *minusele = iele->getElem(0);
                MElement *plusele = iele->getElem(1);
                int totalnodes = minusele->getNumVertices() + plusele->getNumVertices();
                // number of dimension
                sp->getKeys(iele,R); // can't only get keys of elem minus on interDomain !!
                int dimMinus = R.size()/totalnodes;
                int numdofsMinus = dimMinus*minusele->getNumVertices();
                int numdofsPlus = dimMinus*plusele->getNumVertices();
                // limit R to the dofs of minus elem (can't resize as Dof() is not defined)
                for(int kk=0;kk<numdofsPlus;kk++)
                {
                  R.pop_back();
                }
                fullVector<double> disp(R.size());
                _ufield->get(R,disp);
                std::vector<double> elempos;
                currentConfig::elementPositionXYZ(minusele,disp,elempos);
                std::vector<int> vn(2);
                iele->getLocalVertexNum(0,vn);
                int numvertex = minusele->getNumVertices();
                fprintf(_crackTrackingFile,"%e;%e;%d;%d;%d;%e;%e;%e;%e;%e;%e\n",time0,curtime,ele->getNum(),iele->getElem(index0)->getNum(),iele->getElem(index1)->getNum(),elempos[vn[index0]],elempos[vn[index0]+numvertex],elempos[vn[index0]+2*numvertex],elempos[vn[index1]],elempos[vn[index1]+numvertex],elempos[vn[index1]+2*numvertex]);
            }
          }
        }
      }
    }
    fflush(_crackTrackingFile);
  }
}

void nonLinearMechSolver::postproFragment()
{
  if(_fragmentationFile!=NULL)
  {
    std::set<interfaceFragment> sfrag; // contains interface element not broken
    std::set<pairMassElem> isolatedfrag; // contains elements of broken interface
    // loop on domain
    for(std::vector<partDomain*>::const_iterator itdom = domainVector.begin(); itdom != domainVector.end(); ++itdom)
    {
      partDomain *dom = *itdom;
      if(dom->IsInterfaceTerms())
      {
        dgPartDomain *dgdom = static_cast<dgPartDomain*>(dom);
        if((dgdom->getMaterialLawMinus()->getType() == materialLaw::fracture)
           and (dgdom->getMaterialLawPlus()->getType() == materialLaw::fracture)) // same for dim 3 normally. But not tested
        {
          // loop on interface element
          QuadratureBase *integBound = dgdom->getInterfaceGaussIntegrationRule();
          IntPt *GP;
          const materialLaw2LawsInitializer *mfminus = dynamic_cast<const materialLaw2LawsInitializer*>(dgdom->getMaterialLawMinus());
          const materialLaw2LawsInitializer *mfplus = dynamic_cast<const materialLaw2LawsInitializer*>(dgdom->getMaterialLawPlus());
          for(groupOfElements::elementContainer::const_iterator itele = dgdom->gi->begin(); itele != dgdom->gi->end(); ++itele)
          {
            MElement *ele = *itele;
            AllIPState::ipstateElementContainer *vips = _ipf->getAips()->getIPstate(ele->getNum());
            int npts_inter = integBound->getIntPoints(ele,&GP);
            int jj;
            /* Assume same value on minus and plus element !!*/
            for(jj=0;jj<npts_inter;jj++)
            {
              IPStateBase *ipsm = (*vips)[jj];
              IPVariable *ipvmcur = ipsm->getState(IPStateBase::current);
              if(!(mfminus->fullBroken(ipvmcur)))
                break;
            }
            MInterfaceElement *miel = dynamic_cast<MInterfaceElement*>(ele);
            if(!(jj==npts_inter))// element is not broken
            {
              sfrag.insert(interfaceFragment(miel,dgdom->getMaterialLawMinus()->density(),dgdom->getMaterialLawPlus()->density()));
            }
            else
            {
             #if defined(HAVE_MPI)
              if((miel->getElem(0)->getPartition() == Msg::GetCommRank()+1) or (Msg::GetCommSize()==1)) // insert only the element that is on the partition
             #endif // HAVE_MPI
              {
                isolatedfrag.insert(pairMassElem(miel->getElem(0),dgdom->getMaterialLawMinus()->density()));
              }
             #if defined(HAVE_MPI)
              if((miel->getElem(1)->getPartition() == Msg::GetCommRank()+1) or (Msg::GetCommSize()==1)) // insert only the element that is on the partition
             #endif // HAVE_MPI
              {
                isolatedfrag.insert(pairMassElem(miel->getElem(1),dgdom->getMaterialLawPlus()->density()));
              }
            }
          }
        }
      }
    }
    // loop on interelement
    std::set<fragment> fragmentlist;
    std::set<fragment>::iterator it1=fragmentlist.end();
    std::set<fragment>::iterator it2=fragmentlist.end();
    for(std::set<interfaceFragment>::iterator it=sfrag.begin(); it!=sfrag.end();++it)
    {
       it1=fragmentlist.end();
       it2=fragmentlist.end();
       for(std::set<fragment>::iterator itf=fragmentlist.begin(); itf!=fragmentlist.end(); ++itf)
       {
          if((*itf).match(*it))
          {
            if(it1==fragmentlist.end())
            {
              it1 = itf;
            }
            else
            {
              it2 = itf;
              break; // can only found 2 elements;
            }
          }
       }
       // insert fragment
       if((it1!=fragmentlist.end()) and(it2==fragmentlist.end())) // match with one --> insert interfacefragment to the fragment
       {
          (*it1).add(*it);
       }
       else if(it2!=fragmentlist.end()) // match twice --> regroup both fragment
       {
          (*it1).add(*it2);
          fragmentlist.erase(it2);
       }
       else // no match --> create a new one
       {
          fragmentlist.insert(fragment(*it));
       }
    }
    // Loop on isolated fragment. If not founded the fragment contains only 1 element which has to be inserted in fragmentlist
    for(std::set<pairMassElem>::iterator it=isolatedfrag.begin(); it!=isolatedfrag.end();++it)
    {
      std::set<fragment>::iterator itfound=fragmentlist.end();
      for(std::set<fragment>::iterator itf=fragmentlist.begin(); itf!=fragmentlist.end();++itf)
      {
        if((*itf).match(*it))
        {
          itfound = itf;
          break;
        }
      }
      if(itfound==fragmentlist.end())
      {
        fragmentlist.insert(fragment(*it));
      }
    }

    // at this point fragmentlist contains all fragment of the partition
    // the file can be written except for fragment defined on more than 1
    // partition which have to be regrouped.
    double massallfrag=0.;
    int fragnum=1;
    fprintf(_fragmentationFile,"number;number of element;mass\n");
    #if defined(HAVE_MPI)
    std::set<fragmentMPI> fragmentlistMPI;
    int mysizeMPI = 0;  // compute the size of data to transfert
    #endif // HAVE_MPI
    std::vector<std::set<fragment>::iterator> fragmypart;
    for(std::set<fragment>::iterator it=fragmentlist.begin(); it!=fragmentlist.end();++it)
    {
      #if defined(HAVE_MPI)
      if((Msg::GetCommSize()>1) and ((*it).definedOnMoreThanOneRank())){
        fragmentlistMPI.insert(fragmentMPI(*it));
        mysizeMPI +=(*it).numberOfValueToCommunicateMPI();
      }
      else // fragment is totally included on this rank so can be written into the files
      #endif // HAVE_MPI
      {
        fprintf(_fragmentationFile,"%d;%d;%e\n",fragnum,(*it).numberOfElements(),(*it).mass());
        fragnum++;
        massallfrag+= (*it).mass();
        fragmypart.push_back(it);
      }
    }
    #if defined(HAVE_MPI)
    // regroup inter rank fragment (the data is send to rank 0 which do the job (change this ??))
    // size to transfert
    int *alldatasize;
    double* allfragMPI;
    double** allfragMPIrank0;
    MPI_Status mpistatus;
    if(Msg::GetCommRank()==0){
      // get fragment of the other ranks
      alldatasize = new int[Msg::GetCommSize()];
      allfragMPIrank0 = new double*[Msg::GetCommSize()];
      alldatasize[0] = mysizeMPI;
      for(int i=1;i<Msg::GetCommSize();i++)
      {
        MPI_Recv(&alldatasize[i],1,MPI_INT,i,i,MPI_COMM_WORLD,&mpistatus);
        allfragMPIrank0[i] = new double[alldatasize[i]];
        MPI_Recv(allfragMPIrank0[i],alldatasize[i],MPI_DOUBLE,i,i,MPI_COMM_WORLD,&mpistatus);
      }
      // double loop to avoid wait of other rank for MPI_comm ??
      // insert fragment of each rank
      std::set<fragmentMPI>::iterator it1;
      std::set<fragmentMPI>::iterator it2;
      for(int i=1;i<Msg::GetCommSize();i++)
      {
        int curpos=0;
        while(curpos<alldatasize[i])
        {
          fragmentMPI newfrag = fragmentMPI(allfragMPIrank0[i],curpos);
          curpos += newfrag.memorySize();
          // add fragment in the list
          std::vector<std::set<fragmentMPI>::iterator> vit;
          for(std::set<fragmentMPI>::iterator itf=fragmentlistMPI.begin(); itf!=fragmentlistMPI.end(); ++itf)
          {
            if((*itf).match(newfrag))
            {
              newfrag.add(*itf);
              vit.push_back(itf);
            }
          }
          // delete matched fragment
          for(int i=vit.size()-1; i>=0;i--)
          {
            fragmentlistMPI.erase(vit[i]);
          }
          // insert the new one
          fragmentlistMPI.insert(newfrag);
        }
      }
      // write to files
      for(std::set<fragmentMPI>::iterator it=fragmentlistMPI.begin(); it!=fragmentlistMPI.end();++it)
      {
        fprintf(_fragmentationFile,"%d;%d;%e\n",fragnum,(*it).numberOfElements(),(*it).mass());
        fragnum++;
        massallfrag+= (*it).mass();
      }


      // free memory
      for(int i=1;i<Msg::GetCommSize();i++)
        delete[] allfragMPIrank0[i];
      delete[] allfragMPIrank0;
    }
    else
    {
      // build a double vector to transfert all data to node 0
      allfragMPI = new double[mysizeMPI];
      int vecpos = 0;
      for(std::set<fragmentMPI>::iterator it=fragmentlistMPI.begin(); it!=fragmentlistMPI.end(); ++it)
      {
        vecpos +=(*it).fillMPIvector(&allfragMPI[vecpos]);
      }
      MPI_Send(&mysizeMPI,1,MPI_INT,0,Msg::GetCommRank(),MPI_COMM_WORLD);
      // send data
      MPI_Send(allfragMPI,mysizeMPI,MPI_DOUBLE,0,Msg::GetCommRank(),MPI_COMM_WORLD);

      delete[] allfragMPI;
    }
    Msg::Barrier(); // To wait rank0 before end
    #endif // HAVE_MPI
    // total mass at the end
    fprintf(_fragmentationFile,"total mass; %e\n number of fragment %d\n",massallfrag,fragnum-1);
    fflush(_fragmentationFile);
    /* create a view of fragment ( ie a mesh with a different physical number for each fragment) */
    //filename
    std::string fragmshname;
   #if defined(HAVE_MPI)
    if(Msg::GetCommSize()>1)
    {
      std::ostringstream oss;
      oss << Msg::GetCommRank();
      std::string partnum = oss.str();
      fragmshname = "fragview_part"+partnum+".msh";
    }
    else
   #endif //HAVE_MPI
    {
      fragmshname = "fragview.msh";
    }
    FILE *fpfragmsh = fopen(fragmshname.c_str(),"w");
    FILE *fpmsh = fopen(_meshFileName.c_str(),"r");
    fprintf(fpfragmsh,"$MeshFormat\n2.2 0 8\n$EndMeshFormat\n");
    // The node are copy only on rank 0
    char what[256];
//    if(Msg::GetCommRank()==0)
    {
      while(1)
      {
        fscanf(fpmsh,"%s",what);
        if(!strcmp(what,"$Nodes"))
        {
          fprintf(fpfragmsh,"$Nodes\n");
          fscanf(fpmsh,"%s",what);
          fprintf(fpfragmsh,"%s\n",what);
          break;
        }
      }
      int j=1;
      while(!feof(fpmsh))
      {
        fscanf(fpmsh,"%s",what);
        fprintf(fpfragmsh,"%s",what);
        if(!strcmp(what,"$EndNodes"))
        {
          fprintf(fpfragmsh,"\n");
          break;
        }
        if(j%4==0)
        {
          fprintf(fpfragmsh,"\n");
        }
        else
        {
          fprintf(fpfragmsh," ");
        }
        j++;
      }
    }
/*    else{
      while(!feof(fpmsh))
      {
        fscanf(fpmsh,"%s",what);
        if(!strcmp(what,"$EndNodes"))
        {
          break;
        }
      }
    }
*/
    fscanf(fpmsh,"%s",what);
    fprintf(fpfragmsh,"%s\n",what);
    int allelem; // total number of element in the mesh file
    fscanf(fpmsh,"%d",&allelem);
    // compute the total number of element included on this partition
    int totalelem=0;
    for(int i=0;i<fragmypart.size();i++)
    {
      totalelem +=fragmypart[i]->numberOfElements();
    }
   #if defined(HAVE_MPI)
    if(Msg::GetCommRank()==0)
    {
      for(std::set<fragmentMPI>::iterator it = fragmentlistMPI.begin(); it!=fragmentlistMPI.end();++it)
      {
        totalelem +=it->numberOfElements();
      }
    }
   #endif // HAVE_MPI
    fprintf(fpfragmsh,"%d\n",totalelem);
    int num1,num2;
    int physfrag;
    int typemshint;
    for(int k=0;k<allelem;k++)
    {
      int elenum;
      fscanf(fpmsh,"%d",&elenum);
      // check if the element is in the fragment of this partition
      bool writeelem=false;
      physfrag=0;
      for(int i=0;i<fragmypart.size();i++)
      {
        physfrag++;
        if(fragmypart[i]->isElem(elenum))
        {
          writeelem=true;
          break;
        }
      }
     #if defined(HAVE_MPI)
      if(Msg::GetCommRank()==0 and !writeelem)
      {
        for(std::set<fragmentMPI>::iterator it = fragmentlistMPI.begin(); it!=fragmentlistMPI.end();++it)
        {
          physfrag++;
          if(it->isElem(elenum))
          {
            writeelem = true;
            break;
          }
        }
      }
     #endif // HAVE_MPI
      fscanf(fpmsh,"%d %d",&typemshint,&num2);
      if(writeelem) fprintf(fpfragmsh,"%d %d %d ",elenum,typemshint,num2);
      fscanf(fpmsh,"%d",&num1); // read the physical number which has to be replaced
      if(writeelem) fprintf(fpfragmsh,"%d ",physfrag+1000000*Msg::GetCommRank());
      int nver = MElement::getInfoMSH(typemshint);       // get the number of vertices of the element
      // the gentity has to be different for each fragment ie same as physical
      fscanf(fpmsh,"%d",&num1);
      if(writeelem) fprintf(fpfragmsh,"%d ",physfrag+1000000*Msg::GetCommRank());
      for(int i=2;i<num2;i++){
        fscanf(fpmsh,"%d",&num1);
        if(writeelem) fprintf(fpfragmsh,"%d ",num1);
      }
      for(int i=0;i<nver;i++)
      {
        fscanf(fpmsh,"%d",&num1);
        if(writeelem){
          if(i!=nver-1)
            fprintf(fpfragmsh,"%d ",num1);
          else
            fprintf(fpfragmsh,"%d\n",num1);
        }
      }
      //if(writeelem) fprintf(fpfragmsh,"\n");
    }
    fprintf(fpfragmsh,"$EndElements");
    fclose(fpmsh);
    fclose(fpfragmsh);
    Msg::Barrier(); // wait all otherwise archiving problem ??
  }
}

void nonLinearMechSolver::fillMapOfInterfaceElementsInOneDomain(MElement *e, std::vector<MElement*> &eleFound,
                                const groupOfElements *g) const
{
    // take only the primary vertex (avoid unnecessery check for element with order > 1
   int nbVertex = e->getNumPrimaryVertices();

  std::vector<MVertex*> vv;
  for(int i=0;i<nbVertex;i++)
    vv.push_back(e->getVertex(i));

  // In 3D, in //, it is possible that all the nodes are in group g, but not from the same element
  // We need to identify by face.
  for(groupOfElements::elementContainer::const_iterator ite=g->begin(); ite !=g->end(); ite++)
  {
    MElement *gele = *ite;
    int nbVertex2 = gele->getNumPrimaryVertices();
    bool oneNotFound = false;
    for(std::vector<MVertex*>::const_iterator itvv = vv.begin(); itvv!= vv.end(); itvv++)
    {
      oneNotFound = true;
      for(int j=0;j<nbVertex2;j++)
      {
        if((gele->getVertex(j)) == (*itvv))
        {
          oneNotFound = false;
          break;
        }
      }
      if(oneNotFound) break;
    }
    if(!oneNotFound)
    {
      eleFound.push_back(e);
      break;
    }
  }
}

void nonLinearMechSolver::fillLocalReferenceBasisForAllInterfaceElement(){
  //printf("begin filling reference local basis\n");
	for (int idom= 0; idom < domainVector.size(); idom++){
		const dgPartDomain* dgdom = dynamic_cast<const dgPartDomain*>(domainVector[idom]);
		if (dgdom != NULL){
			IntPt* GP;
			for (groupOfElements::elementContainer::iterator ite = dgdom->gi->begin(); ite != dgdom->gi->end(); ite++){
				MElement* ele = *ite;
				int npts  = dgdom->getInterfaceGaussIntegrationRule()->getIntPoints(ele,&GP);
				for (int i=0; i< npts; i++){
					int type = numericalMaterialBase::createTypeWithTwoInts(ele->getNum(), i);
					SVector3 n, b, t;
					computeInterfaceReferenceLocalBasis(ele,&GP[i],n,t,b);

					STensor3 mat;
					mat(0,0) = n(0); mat(1,0) = n(1);  mat(2,0) = n(2);
					mat(0,1) = t(0); mat(1,1) = t(1);  mat(2,1) = t(2);
					mat(0,2) = b(0); mat(1,2) = b(1);  mat(2,2) = b(2);

					_allInterfaceLocalBasis[type] = mat; // negative
					type = numericalMaterialBase::createTypeWithTwoInts(ele->getNum(), i+npts);
					_allInterfaceLocalBasis[type] = mat; // positive
				}
			}
		}
	}
	//printf("done filling reference local basis size = %d\n",allInterfaceLocalBasis.size());
};

void nonLinearMechSolver::getLocalBasis(const MElement* ele, const int gpt, SVector3& n, SVector3& t, SVector3& b) const {
	int type = numericalMaterialBase::createTypeWithTwoInts(ele->getNum(), gpt);
	std::map<int,STensor3>::const_iterator itfind = _allInterfaceLocalBasis.find(type);
	if ( itfind == _allInterfaceLocalBasis.end()){
		Msg::Error(" rank %d local basis at interface element %d gpt %d does not exist",Msg::GetCommRank(), ele->getNum(),gpt);
	}
	else{
		const STensor3& mat = itfind->second;
		n(0) = mat(0,0); n(1) = mat(1,0);  n(2) = mat(2,0);
		t(0) = mat(0,1); t(1) = mat(1,1);  t(2) = mat(2,1);
		b(0) = mat(0,2); b(1) = mat(1,2);  b(2) = mat(2,2);

		//mat.print("found basis");
	}
};



void nonLinearMechSolver::createStrainMapping(const std::string  filename){
  if (_strainMap) delete _strainMap;
  _strainMap = new strainMapping(domainVector);
  _strainMap->readMesh(filename);

  GModel::setCurrent(pModel);
};

void nonLinearMechSolver::setStrainMappingFollowing(const bool flag){
  _strainMappingFollwing = flag;
};

void nonLinearMechSolver::blockDissipation(const IPStateBase::whichState ws, const bool fl){
  if (_ipf){
		if (fl and !_damageIsBlocked){
      printf("rank %d damage is blocked ele = %d, gp = %d  \n",Msg::GetCommRank(),_enum,_gnum);
    }
		else if (fl and _damageIsBlocked){
			printf("rank %d damage is reblocked ele = %d, gp = %d  \n",Msg::GetCommRank(),_enum,_gnum);
		}
    else if (!fl and _damageIsBlocked){
      printf("rank %d damage is unblocked ele = %d, gp = %d  \n",Msg::GetCommRank(),_enum,_gnum);
    }
		_damageIsBlocked = fl;

  }
  else
    Msg::Error("IPField has not been initialized");
}

bool nonLinearMechSolver::dissipationIsBlocked() const{
	return _damageIsBlocked;
};

void nonLinearMechSolver::setTimeForMicroBC(const double time){
	if (_microBC!=NULL){
    _microBC->setTime(time);
  }
};

void nonLinearMechSolver::microNumberDof(){
  if (_pathFollowing){
    _systemType = MULT_ELIM;
    _pAl->numberDof(MULT_ELIM);
  }
  else{
    if (_controlType == LOAD_CONTROL){
    // number normally dof
			_pAl->numberDof(_systemType);
    // allocate micro system
    }
    else if (_controlType == ARC_CONTROL_EULER){
      _systemType = DISP_MULT;
      _pAl->numberDof(DISP_MULT);
    }
    else{
      Msg::Error("control Type must be correct defined");
    }
  }
	std::string A("A");
  if (_systemType == DISP_ELIM_UNIFIED){
    // PAL has to be set before allocating linear system
    pbcSystemConden<double>* pbc = dynamic_cast<pbcSystemConden<double>*>(pAssembler->getLinearSystem(A));
    if (pbc){
      pbc->setPBCAlgorithm(_pAl);
    }
  }
  // allocate system
  int systemSize = pAssembler->sizeOfR();
	if (_pathFollowing){
		// if path following and one rhs resolution
		if (_pathFollowingMethod == GLOBAL_ARC_LENGTH_BASED){

		}
		else if (_pathFollowingMethod == LOCAL_BASED){
			// add 1 DOF to total system as load factor unknown
			if (_solverTypePathFollowing == pathFollowingSystemBase::ONE_RHS){
				systemSize++;
			}
		}
		else{
			Msg::Error("_pathFollowingMethod = %d has not been implemented",_pathFollowingMethod);
		}
	}

  pAssembler->getLinearSystem(A)->allocate(systemSize);

  if (_systemType == DISP_ELIM_UNIFIED){
    // initial kinematic variable for this method
    pbcSystemConden<double>* pbc = dynamic_cast<pbcSystemConden<double>*>(pAssembler->getLinearSystem(A));
    if (pbc){
      fullVector<double> kinVec;
      _pAl->getMBC()->getKinematicalVector(kinVec);
      for (int i=0; i< kinVec.size(); i++){
        double val = kinVec(i);
        pbc->initialKinematicVector(i,val);
      }

    }
  }

  // preallocated system
  for(int i=0; i<_mpiUserDom;++i)
  {
    staticDofManager<double>* sAssembler = static_cast<staticDofManager<double>*>(pAssembler);
    partDomain *dom = domainVector[i];
    if(dom->groupOfElementsSize() > 0){
        SparsityDofs(*(dom->getFunctionSpace()), dom->g_cbegin(), dom->g_cend(),*sAssembler);
    }
    // account also interface
    dgPartDomain *dgdom = dynamic_cast<dgPartDomain*>(dom);
    if (dgdom){
      staticDofManager<double>* sAssembler = static_cast<staticDofManager<double>*>(pAssembler);
      #if defined(HAVE_MPI)
      if(Msg::GetCommSize()>1)
          SparsityInterfaceDofsMPI(dgdom->getFunctionSpace(), dgdom->gi->begin(), dgdom->gi->end(),sAssembler);
      #endif
      SparsityDofs(*(dgdom->getFunctionSpace()), dgdom->gi->begin(), dgdom->gi->end(),*sAssembler);
    }
  }
  #if defined(HAVE_MPI)
   // MPI interface sparsity
  if(Msg::GetCommSize()>1){
    for(int i=_mpiUserDom; i<domainVector.size();++i)
    {
      dgPartDomain *dgdom = dynamic_cast<dgPartDomain*>(domainVector[i]);
      if (dgdom){
        staticDofManager<double>* sAssembler = static_cast<staticDofManager<double>*>(pAssembler);
        SparsityInterfaceDofsMPI(dgdom->getFunctionSpace(), dgdom->gi->begin(), dgdom->gi->end(),sAssembler);
      }
    }
  }
  #endif // HAVE_MPI

  if (_systemType == nonLinearMechSolver::DISP_MULT){
    _pAl->sparsityLagMultiplierDofs();
  }

	pAssembler->getLinearSystem(A)->preAllocateEntries();


  // assemble all linear constraint matrix to system in case of using mult -elim method
  int dimKin = _pAl->getNumberOfMacroToMicroKinematicVariables();
  int dimStress = _pAl->getNumberMicroToMacroHomogenizedVariables();

  bool insystemCondenFlag = false;

  if (_tangentflag and _homogenizeTangentMethod == INSYSTEMCONDEN){
    insystemCondenFlag = true;
  };

  if (_systemType == MULT_ELIM){
    _pAl->allocateConstraintMatrixToSystem(dimKin,dimStress,insystemCondenFlag);
    // assembleLinearConstraintMatrixToSystem must be called at least once
    _pAl->assembleLinearConstraintMatrixToSystem();
    if (insystemCondenFlag){
      _pAl->assembleKinematicMatrix();
    }
  }
  else if (_systemType == DISP_ELIM_UNIFIED ){
    std::string A("A");
    pbcSystemConden<double>* pbc = dynamic_cast<pbcSystemConden<double>*>(pAssembler->getLinearSystem(A));
    if (pbc == NULL) Msg::Error("pbcSystemConden has to be used");
    // computeConstraintProjectMatrices must be called at least once
    pbc->computeConstraintProjectMatrices();
    if (insystemCondenFlag){
      // only stress matrix is necessary in this case
      pbc->allocateStressMatrix(dimStress,pAssembler->sizeOfR());
    }
  }
};

void nonLinearMechSolver::setPeriodicity(const double x, const double y, const double z, const std::string direction){
  Msg::Warning(" nonLinearMechSolver::setPeriodicity is no longer necessary!");
};

void nonLinearMechSolver::addMicroBC(const nonLinearMicroBC* bc){
  _microFlag = true;
	if (_microBCOld) delete _microBCOld;
	_microBCOld = _microBC;
  _microBC = bc->clone();
	if (_microBCOld != NULL){
		_microBC->collecDataFromPrevious(*_microBCOld);
	}

	// set highorder flag
	if (_microBC->getOrder() >1){
		_highOrderFlag = true;
	}
	else{
		_highOrderFlag = false;
	}

	// add BC suppport
	std::vector<groupOfElements*>& bgroup  = _microBC->getBoundaryGroupOfElements();
	const std::vector<int>& bphysical = _microBC->getBoundaryPhysicals();
	bgroup.clear();
	for (int i=0; i< bphysical.size(); i++){
		groupOfElements* gr = new groupOfElements(_dim-1,bphysical[i]);
		bgroup.push_back(gr);
	}
};
void nonLinearMechSolver::addMicroBCForFailure(const nonLinearMicroBC* mbc){
	// this BC sleeps util failure
	// once failure occurs, _microBC is replaced by _microFailureBC using addMicroBC
	if (_microFailureBC!=NULL) delete _microFailureBC;
	_microFailureBC = mbc->clone();
};

double nonLinearMechSolver::getRVEVolume() const{
	return _rveVolume;
};

void nonLinearMechSolver::setRVEVolume(const double val){
  _rveVolume = val;
}

double nonLinearMechSolver::getHomogenizedDensity() const{
  return _rho;
};

double nonLinearMechSolver::getConstitutiveExtraDofDiffusionEquationRatio(const int index) const
{
  double eqRatio=1.;
  bool found=false;

	for (int i=0; i<domainVector.size(); i++)
	{
		const partDomain* dom = domainVector[i];
		double eqR=dom->getConstitutiveExtraDofDiffusionEqRatio(index);
		if(eqR>0.)
		{
			 if(!found)
			 {
				 eqRatio = eqR;
				 found=true;
			 }
			 else if(eqR!=eqRatio)
				 Msg::Error("not all the domains have the same EqRatio field %d",index);
		}
	}
  return eqRatio;
};

double nonLinearMechSolver::getNonConstitutiveExtraDofEquationRatio(const int index) const
{
  double eqRatio=1.;
  bool found=false;

	for (int i=0; i<domainVector.size(); i++)
	{
		const partDomain* dom = domainVector[i];
		double eqR=dom->getNonConstitutiveExtraDofEqRatio(index);
		if(eqR>0.)
		{
			 if(!found)
			 {
				 eqRatio = eqR;
				 found=true;
			 }
			 else if(eqR!=eqRatio)
				 Msg::Error("not all the domains have the same EqRatio field %d",index);
		}
	}
  return eqRatio;
};

bool nonLinearMechSolver::getExtractIrreversibleEnergyFlag() const{return _extractIrreversibleEnergy;};
bool nonLinearMechSolver::getDamageToCohesiveJumpFlag() const {return _damageToCohesiveJump;};
int nonLinearMechSolver::getHomogenizationOrder() const {
  if (_microBC != NULL){
    return _microBC->getOrder();
  }
  else{
    return 0;
  }
};
bool nonLinearMechSolver::inMultiscaleSimulation() const{return _multiscaleFlag;};

const SVector3& nonLinearMechSolver::getLostSolutionUniquenssNormal() const{
	return _lostSolutionUniquenssNormal;
};

double& nonLinearMechSolver::getHomogenizedCrackFace(){
	return _homogenizedCrackSurface;
};
const double& nonLinearMechSolver::getHomogenizedCrackFace() const{
	return _homogenizedCrackSurface;
};

STensor3& nonLinearMechSolver::getFdamOnset() {return _FdamOnset;};
const STensor3& nonLinearMechSolver::getFdamOnset() const {return _FdamOnset;};

bool nonLinearMechSolver::withEnergyDissipation() const{
	for (std::map<int, materialLaw*>::const_iterator it = maplaw.begin(); it!= maplaw.end(); it++){
		if (it->second->withEnergyDissipation()){
			return true;
		}
	}
	return false;
};

bool nonLinearMechSolver::withFractureLaw() const{
  for (std::map<int, materialLaw*>::const_iterator it = maplaw.begin(); it!= maplaw.end(); it++){
		if (it->second->getType() == materialLaw::fracture){
			return true;
		}
	}
	return false;
};

const elementErosionFilter& nonLinearMechSolver::getElementErosionFilter() const{
	return _elementErosionFilter;
};

bool nonLinearMechSolver::solverIsBroken() const{
  return _solverIsBroken;
}

void nonLinearMechSolver::brokenSolver(const bool fl){
  _solverIsBroken = fl;
};

void nonLinearMechSolver::getTotalIntegralByVolumeIntegralOnPhysical(const int phys, const int num, double& val) const{
  val = 0.;
  for (int i=0; i< domainVector.size(); i++){
    partDomain* dom = domainVector[i];
    if (dom->getPhysical() == phys){
      double volumePart = 0;
      double valPartAverage = dom->computeMeanValueDomain(num,_ipf,volumePart);
      val += (valPartAverage*volumePart);
    }
  }
  #if defined(HAVE_MPI)
  if (_isPartitioned and (Msg::GetCommSize() > 1)){
    double valTmp = val;
    MPI_Allreduce(&valTmp,&val,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
  }
  #endif // HAVE_MPI
};

void nonLinearMechSolver::getTotalIntegralByVolumeIntegral(const int num, double& val) const{
  val = 0.;
  for (int i=0; i< domainVector.size(); i++){
    partDomain* dom = domainVector[i];
    double volumePart = 0;
    double valPartAverage = dom->computeMeanValueDomain(num,_ipf,volumePart);
    val += (valPartAverage*volumePart);
  }
  #if defined(HAVE_MPI)
  if (_isPartitioned and (Msg::GetCommSize() > 1)){
    double valTmp = val;
    MPI_Allreduce(&valTmp,&val,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
  }
  #endif // HAVE_MPI
};

void nonLinearMechSolver::getHomogenizedPropertyByVolumeIntegral(const int num, double & val) const{
  val = 0.;
  double volume = 0.;
  for (int i=0; i< domainVector.size(); i++){
    partDomain* dom = domainVector[i];
    double volumePart = 0;
    double valPart = dom->computeMeanValueDomain(num,_ipf,volumePart);
    val += (valPart*volumePart);
    volume += volumePart;
  }

  #if defined(HAVE_MPI)
  if (_isPartitioned and (Msg::GetCommSize() > 1)){
    double send[2] = {volume, val };
    double receive[2];
    MPI_Allreduce(send,receive,2,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
    volume = receive[0];
    val = receive[1];
  }
  #endif // HAVE_MPI
  double invV = (1./volume);
  val *= invV;
};

void nonLinearMechSolver::getHomogenizedPropertyByVolumeIntegralOnActiveDissipationDomain(const int num, double& val, const IPStateBase::whichState ws) const{
  val = 0.;
  double dissipationVolume = 0.;
  for (int i=0; i< domainVector.size(); i++){
    partDomain* dom = domainVector[i];
    double dissipationVolumePart = 0;
    double valPart = dom->averagingOnActiveDissipationZone(num,_ipf,dissipationVolumePart,ws);
    val += (valPart*dissipationVolumePart);
    dissipationVolume += dissipationVolumePart;
  }
  if (dissipationVolume >0){
    val /= dissipationVolume;
  }
};


void nonLinearMechSolver::extractAverageStressByVolumeIntegral(homogenizedData* homoData){
	// in case of multiscale analyses, previous macro-state is initial micro-state
	double volumeRVEInv = 1./this->getRVEVolume();
  // basic mechanics quantities
	if (_microBC->getTotalNumberOfMechanicalDofs()> 0){
		double& damageVolume = homoData->getActiveDissipationVolume();
		damageVolume = 0.;
		double& clength = homoData->getAverageLocalizationBandWidth();
		clength = 0.;
		if (withEnergyDissipation()){
      // compute current damage volume
			for (int i=0; i<domainVector.size(); i++){
				damageVolume += domainVector[i]->computeVolumeActiveDissipationDomain(_ipf);
			}
			// active damage center
			if (damageVolume > 0.){
				// compute homogenized crack length
				this->computeHomogenizedCrackFace();
        // true damage surface = _surfaceReductionRatio*_homogenizedCrackSurface
        clength = damageVolume/(_surfaceReductionRatio*_homogenizedCrackSurface);
			}
		}

		STensor3& P = homoData->getHomogenizedStress();
    STensorOperation::zero(P);
    for (int i=0; i<domainVector.size(); i++){
      domainVector[i]->computeAverageStress(_ipf,P);
    };
    P *= (volumeRVEInv);

		if (_microBC->getOrder() == 2){
			STensor33& Q = homoData->getHomogenizedSecondOrderStress();
			STensorOperation::zero(Q);
			for (int i=0; i<domainVector.size(); i++){
				domainVector[i]->computeAverageHighOrderStress(_ipf,Q);
			};
			Q *= (volumeRVEInv);
		}

    // extract cohesive law
    if (this->withEnergyDissipation() and this->getDamageToCohesiveJumpFlag()){
      // cohesive traction
      SVector3& T = homoData->getHomogenizedCohesiveTraction();
      STensorOperation::zero(T);
      // cohesive traction
      for (int i=0; i<3; i++){
        for (int j=0; j<3; j++){
          T(i) += P(i,j)*_lostSolutionUniquenssNormal(j);
        }
      }

      // compute cohesive jump
      STensor3& Fdam = homoData->getHomogenizedActiveDissipationDeformationGradient();
      Fdam = _homogenizedDissipationStrainLastActiveDissipation;

      STensor3 dFdam(0.);
      if (damageVolume > 0.){
        for (int i=0; i<domainVector.size(); i++){
          domainVector[i]->computeActiveDissipationAverageStrainIncrement(_ipf,dFdam);
        }
        dFdam *= (1./damageVolume);
        Fdam += dFdam;
      }

      // compute cohesive jump
      SVector3& cjump = homoData->getHomogenizedCohesiveJump();
      cjump  = _homogenizedCohesiveJumpLastActiveDissipation;

      if (_solverIsBroken){
        STensor3 dFinterf = this->getMicroBC()->getFirstOrderKinematicalVariable();
        dFinterf -= _homogenizedStrainLastActiveDissipation;
        double beta = damageVolume*volumeRVEInv/_surfaceReductionRatio;

        for (int i=0; i<3; i++){
          for (int j=0; j<3; j++){
            cjump(i) += (clength/(1.-beta))*(dFdam(i,j) - dFinterf(i,j))*_lostSolutionUniquenssNormal(j);
          }
        };
      }
    };

		double& defoEnergy = homoData->getDeformationEnergy();
		double& plasEnergy = homoData->getPlasticEnergy();
		double& irrEnergy = homoData->getIrreversibleEnergy();

		defoEnergy = 0.;
		plasEnergy = 0.;
		irrEnergy = 0.;
		for (int i=0; i< domainVector.size(); i++){
			defoEnergy += domainVector[i]->computeDeformationEnergy(_ipf);
			plasEnergy += domainVector[i]->computePlasticEnergy(_ipf);
			irrEnergy += domainVector[i]->computeIrreversibleEnergy(_ipf);
		}
		defoEnergy *= volumeRVEInv;
		plasEnergy *= volumeRVEInv;
		irrEnergy *= volumeRVEInv;
	};

	if (_microBC->getTotalNumberOfConDofs() > 0){
		// for constitutive-extra Dof
		for (int index=0; index< _microBC->getTotalNumberOfConDofs(); index++ ){
			SVector3& fluxT = homoData->getHomogenizedConstitutiveExtraDofFlux(index);
			double &internalEnergyExtraDof = homoData->getHomogenizedConstitutiveExtraDofInternalEnergy(index);
			double &mecaSource = homoData->getHomogenizedConstitutiveExtraDofMechanicalSource(index);
			double &fieldCapacityPerUnitField = homoData->getHomogenizedConstitutiveExtraDofFieldCapacityPerUnitField(index);

      STensorOperation::zero(fluxT);
			internalEnergyExtraDof = 0.;
			mecaSource = 0.;
			fieldCapacityPerUnitField = 0.;
			for (int i=0; i<domainVector.size(); i++){
				domainVector[i]->computeConstitutiveExtraDofAverageFlux(index,_ipf,fluxT);
				domainVector[i]->computeConstitutiveExtraDofAverageInternalEnergy(index,_ipf,internalEnergyExtraDof);
				domainVector[i]->computeConstitutiveExtraDofAverageMechanicalSource(index,_ipf,mecaSource);
				domainVector[i]->computeConstitutiveExtraDofAverageFieldCapacityPerUnitField(index,_ipf,fieldCapacityPerUnitField);
			};

			fluxT *= volumeRVEInv;
			internalEnergyExtraDof *= volumeRVEInv;
			mecaSource *= volumeRVEInv;
			fieldCapacityPerUnitField *= volumeRVEInv;
		}
	}

	// non-constitutive extraDof
	if (_microBC->getTotalNumberOfNonConDofs()> 0){
		for (int index=0; index<_microBC->getTotalNumberOfNonConDofs(); index++){
			SVector3& fluxV = homoData->getHomogenizedNonConstitutiveExtraDofFlux(index);
			STensorOperation::zero(fluxV);
			for (int i=0; i<domainVector.size(); i++){
				domainVector[i]->computeNonConstitutiveExtraDofAverageFlux(index,_ipf,fluxV);
			};
			fluxV *= volumeRVEInv;
		}
	}
};

void nonLinearMechSolver::homogenizedDataToFile(const double time){
  homogenizedData* data = this->getHomogenizationState(IPStateBase::current);

  _homogenizedFiles->dataToFile(time, this, data);
  if (_extractElasticTangentOperator){
    Tensor43::writeData(_elasticDPDFTangentFile,_elasticState->getHomogenizedTangentOperator_F_F(),time);
  }
};

void nonLinearMechSolver::extractAverageStressByCondensation(homogenizedData* homoData){
  // condensation can compute conly only stresses and extraDof flux
  // other homogenized properties can be directly computed from
  // volumetric integrals

  /**create object in case of NULL **/
  if (_condensation == NULL){
    _condensation = this->createStiffnessCondensation(true,false);
  }

  /** solve**/
  fullVector<double> vecStress;
  _condensation->stressSolve(vecStress,getRVEVolume());

	int atRow = 0;
	// mechanical parts
	if (_microBC->getTotalNumberOfMechanicalDofs()> 0){
		STensor3& P = homoData->getHomogenizedStress();
		P*= (0.);
		// mechanics stress
		for (int row = atRow; row<atRow+9; row++){
			int i,j;
			Tensor23::getIntsFromIndex(row-atRow,i,j);
			P(i,j) =vecStress(row);
		};
		atRow+= 9;

		if (_microBC->getOrder() == 2){
			STensor33& Q = homoData->getHomogenizedSecondOrderStress();
			Q*= (0.);
			// mechanics second order stress
			for (int row = atRow; row<atRow+27; row++){
				int i,j,k;
				Tensor33::getIntsFromIndex(row-atRow,i,j,k);
				Q(i,j,k) = vecStress(row);
			};
			atRow += 27;
		}

		// compute energy by volume integral
		double& defoEnergy = homoData->getDeformationEnergy();
		double& plasEnergy = homoData->getPlasticEnergy();
		double& irrEnergy = homoData->getIrreversibleEnergy();

		defoEnergy = 0.;
		plasEnergy = 0.;
		irrEnergy = 0.;
		for (int i=0; i< domainVector.size(); i++){
			defoEnergy += domainVector[i]->computeDeformationEnergy(_ipf);
			plasEnergy += domainVector[i]->computePlasticEnergy(_ipf);
			irrEnergy += domainVector[i]->computeIrreversibleEnergy(_ipf);
		}

		defoEnergy /= (this->getRVEVolume());
		plasEnergy /= (this->getRVEVolume());
		irrEnergy /= (this->getRVEVolume());

	}

  if (_microBC->getTotalNumberOfConDofs() > 0){
		for (int index = 0; index < _microBC->getTotalNumberOfConDofs(); index++){
			SVector3& extraDofFlux = homoData->getHomogenizedConstitutiveExtraDofFlux(index);
			extraDofFlux*=(0.);
			// extradofs
			for (int row = atRow; row<atRow+3; row++){
				extraDofFlux(row-atRow) = vecStress(row)/getConstitutiveExtraDofDiffusionEquationRatio(index);
			};
			atRow+=3;

			//compute by volume integrals
			double &internalEnergyExtraDof = homoData->getHomogenizedConstitutiveExtraDofInternalEnergy(index);
			double &mecaSource = homoData->getHomogenizedConstitutiveExtraDofMechanicalSource(index);
			double &fieldCapacityPerUnitField = homoData->getHomogenizedConstitutiveExtraDofFieldCapacityPerUnitField(index);

			internalEnergyExtraDof =0.;
			mecaSource = 0;
			fieldCapacityPerUnitField = 0.;

			for (int i=0; i<domainVector.size(); i++){
				domainVector[i]->computeConstitutiveExtraDofAverageInternalEnergy(index,_ipf,internalEnergyExtraDof);
				domainVector[i]->computeConstitutiveExtraDofAverageMechanicalSource(index,_ipf,mecaSource);
				domainVector[i]->computeConstitutiveExtraDofAverageFieldCapacityPerUnitField(index,_ipf,fieldCapacityPerUnitField);
			}
			double vtotal = this->getRVEVolume();
			double vin = 1./vtotal;

			internalEnergyExtraDof*= vin;
			mecaSource *= vin;
			fieldCapacityPerUnitField *= vin;
		}
  }

	if (_microBC->getTotalNumberOfNonConDofs()){
		for (int index = 0; index < _microBC->getTotalNumberOfNonConDofs(); index++){
			SVector3& extraDofFlux = homoData->getHomogenizedNonConstitutiveExtraDofFlux(index);
			extraDofFlux*=(0.);
			// extradofs
			for (int row = atRow; row<atRow+3; row++){
				extraDofFlux(row-atRow) = vecStress(row)/getNonConstitutiveExtraDofEquationRatio(index);
			};
			atRow+=3;
		}
	}
};

void nonLinearMechSolver::extractAveragePropertiesByCondensation(homogenizedData* homoData){
  // condensation can compute conly only tangent operators of stresses and extraDof flux
  // as surface integrals can be used
  // other homogenized properties cannot be computed
  // using other method to estimated tangent
  // if considering mechanical source

  /**create object in case of NULL **/
  if (_condensation == NULL){
    if (_homogenizeStressMethod == VOLUME)
      _condensation = this->createStiffnessCondensation(false, true);
    else
      _condensation = this->createStiffnessCondensation(true, true);
  }

  if (_pAl->isModifiedConstraintMatrix()){
    // recomputing all constraint matrix if necessary
    _condensation->clear();
    _pAl->setIsModifiedConstraintMatrixFlag(false);
  }

  if (_homogenizeStressMethod == VOLUME){
    extractAverageStressByVolumeIntegral(homoData);
  }
  else{
    extractAverageStressByCondensation(homoData);
  }

  /** compute splitted stiffness matrix**/
  this->computeSplittedStiffnessMatrix();

  /** solve**/
  fullMatrix<double> L;
  double t=Cpu();
  if (_multiscaleFlag == false)
    Msg::Info("Begin performing static condensation");
  _condensation->tangentCondensationSolve(L,getRVEVolume());
  t = Cpu() -t;
  if (_multiscaleFlag == false)
    Msg::Info("Done performing static condensation (%f s)",t);
  fillTangent(L,homoData);

  if (_microBC->getTotalNumberOfConDofs() > 0) {
    // re-estimate tangent values due to use eqRatio
		for (int index=0; index < _microBC->getTotalNumberOfConDofs(); index++){
			double eqRatio=1./getConstitutiveExtraDofDiffusionEquationRatio(index);
			if (_microBC->getTotalNumberOfMechanicalDofs()>0){
				homoData->getHomogenizedTangentOperator_gradT_F(index) *= eqRatio;
				if (_microBC->getOrder() == 2){
					homoData->getHomogenizedTangentOperator_gradT_G(index) *= eqRatio;
				}
			}
			for (int j=0; j< _microBC->getTotalNumberOfConDofs(); j++){
				homoData->getHomogenizedTangentOperator_gradT_gradT(index,j) *= eqRatio;
				homoData->getHomogenizedTangentOperator_gradT_T(index,j) *= eqRatio;
			}

			for (int j=0; j< _microBC->getTotalNumberOfNonConDofs(); j++){
				homoData->getHomogenizedTangentOperator_gradT_gradV(index,j) *= eqRatio;
			}

		}

  }

	if (_microBC->getTotalNumberOfNonConDofs() > 0) {
		for (int index=0; index< _microBC->getTotalNumberOfNonConDofs(); index++){
			double eqRatio=1./getNonConstitutiveExtraDofEquationRatio(index);
			if (_microBC->getTotalNumberOfMechanicalDofs() >0){
				homoData->getHomogenizedTangentOperator_gradV_F(index) *= eqRatio;
				if (_microBC->getOrder() == 2){
					homoData->getHomogenizedTangentOperator_gradV_G(index) *= eqRatio;
				}
			}

			for (int j=0; j< _microBC->getTotalNumberOfNonConDofs(); j++){
				homoData->getHomogenizedTangentOperator_gradV_gradV(index,j) *= eqRatio;
			}

			for (int j=0; j< _microBC->getTotalNumberOfConDofs(); j++){
				homoData->getHomogenizedTangentOperator_gradV_T(index,j) *= eqRatio;
				homoData->getHomogenizedTangentOperator_gradV_gradT(index,j) *= eqRatio;
			}
		}
	};
};

void nonLinearMechSolver::extractAveragePropertiesByInSystemCondensation(homogenizedData* homoData){
  // extract stress
  this->extractAverageStress(homoData);
  // extract tangent

  // estimation of stress matrix
  _pAl->zeroStressMatrix();
  staticDofManager<double>* sAssembler = dynamic_cast<staticDofManager<double>*>(pAssembler);
  if (sAssembler == NULL){
    Msg::Error("staticDofManager<double> must be used nonLinearMechSolver::extractAveragePropertiesByInSystemCondensation");
  }
  for (int idom=0; idom<domainVector.size(); idom++){
    partDomain* dom = domainVector[idom];
		if (dom->groupOfElementsSize()>0){
			AssembleStressMatrix(*dom->getBilinearTangentTerm(),*dom->getFunctionSpace(),dom->g_cbegin(),
                        dom->g_cend(),*dom->getBulkGaussIntegrationRule(),*sAssembler,_elementErosionFilter);
		}

  }

	double vRVE = this->getRVEVolume();

  fullMatrix<double> L;
  double t= Cpu();
  if (_multiscaleFlag == false)
    Msg::Info("Begin performing in-system condensation");
  int goodSolve = _pAl->condensationSolve(L,vRVE);
  t = Cpu() -t;
  if (_multiscaleFlag == false)
    Msg::Info("Done performing in-system condensation (%f s)",t);

  fillTangent(L,homoData);


  // extract cohesive law
  if (withEnergyDissipation() and getDamageToCohesiveJumpFlag()){
		STensor33& DcjumpDF = homoData->getHomogenizedTangentOperator_CohesiveJump_F();
    STensorOperation::zero(DcjumpDF);

		const double& clength = homoData->getAverageLocalizationBandWidth();
    double beta = homoData->getActiveDissipationVolume()/(_surfaceReductionRatio*vRVE);
    double fact = (clength/(1.-beta));

    static STensor43 Ifourth(1.,1.);

		if (_solverIsBroken){
			const STensor43&  dFdamDF = homoData->getHomogenizedTangentOperator_ActiveDissipationDeformationGradient_F();
			for (int i=0; i<3; i++){
				for (int j=0; j<3; j++){
					for (int k=0; k<3; k++){
						for (int l=0; l<3; l++){
							DcjumpDF(i,j,k) += fact*(dFdamDF(i,l,j,k) - Ifourth(i,l,j,k))*_lostSolutionUniquenssNormal(l);
						}
					}
				}
			}
		}

		if (_microBC->getOrder() == 2){
			STensor43& dcjumpdG = homoData->getHomogenizedTangentOperator_CohesiveJump_G();
      STensorOperation::zero(dcjumpdG);
			if (_solverIsBroken){
				const STensor53& dFdamDG = homoData->getHomogenizedTangentOperator_ActiveDissipationDeformationGradient_G();
				for (int i=0; i<3; i++){
					for (int j=0; j<3; j++){
						for (int k=0; k<3; k++){
							for (int l=0; l<3; l++){
								for (int r=0; r<3; r++){
									dcjumpdG(i,j,k,r) += fact*dFdamDG(i,l,j,k,r)*_lostSolutionUniquenssNormal(l);
								}
							}
						}
					}
				}
			}
		}

		if (_microBC->getTotalNumberOfConDofs()> 0){
			for (int index=0; index < _microBC->getTotalNumberOfConDofs(); index++){
				STensor3& dcjumpdgradT = homoData->getHomogenizedTangentOperator_CohesiveJump_gradT(index);
        STensorOperation::zero(dcjumpdgradT);

				SVector3& dcjumpdT = homoData->getHomogenizedTangentOperator_CohesiveJump_T(index);
        STensorOperation::zero(dcjumpdT);

				if (_solverIsBroken){
					const STensor3& dFdamDT = homoData->getHomogenizedTangentOperator_ActiveDissipationDeformationGradient_T(index);
					const STensor33& dFdamDgradT = homoData->getHomogenizedTangentOperator_ActiveDissipationDeformationGradient_gradT(index);
					for (int i=0; i<3; i++){
						for (int j=0; j<3; j++){
							dcjumpdT(i) += fact*dFdamDT(i,j)*_lostSolutionUniquenssNormal(j);
							for (int k=0; k<3; k++){
								dcjumpdgradT(i,k) += fact*dFdamDgradT(i,j,k)*_lostSolutionUniquenssNormal(j);
							};
						}
					}
				};
			}
		}

		if (_microBC->getTotalNumberOfNonConDofs()> 0){
			for (int index=0; index < _microBC->getTotalNumberOfNonConDofs(); index++){
				STensor3& dcjumpdgradV = homoData->getHomogenizedTangentOperator_CohesiveJump_gradV(index);
        STensorOperation::zero(dcjumpdgradV);

				if (_solverIsBroken){
					const STensor33& dFdamDgradV = homoData->getHomogenizedTangentOperator_ActiveDissipationDeformationGradient_gradV(index);
					for (int i=0; i<3; i++){
						for (int j=0; j<3; j++){
							for (int k=0; k<3; k++){
								dcjumpdgradV(i,k) += fact*dFdamDgradV(i,j,k)*_lostSolutionUniquenssNormal(j);
							};
						}
					}
				};
			}
		}
  }

  if (_extractElasticTangentOperator){
    if (_multiscaleFlag == false)
      Msg::Info("extract elastic tengent operator");
    //estimating elastic stiffness matrix
    this->computeElasticStiffMatrix();

    // estimation of stress matrix
    _pAl->zeroStressMatrix();
    staticDofManager<double>* sAssembler = dynamic_cast<staticDofManager<double>*>(pAssembler);
    if (sAssembler == NULL){
      Msg::Error("staticDofManager<double> must be used nonLinearMechSolver::extractAveragePropertiesByInSystemCondensation");
    }
    for (int idom=0; idom<domainVector.size(); idom++){
      partDomain* dom = domainVector[idom];
      if (dom->groupOfElementsSize()>0){
        AssembleStressMatrix(*dom->getBilinearElasticTangentTerm(),*dom->getFunctionSpace(),dom->g_cbegin(),
                          dom->g_cend(),*dom->getBulkGaussIntegrationRule(),*sAssembler,_elementErosionFilter);
      }

    }

    //
    t= Cpu();
    if (_multiscaleFlag == false)
      Msg::Info("Begin performing in-system condensation");
    int goodSolve = _pAl->condensationSolve(L,vRVE);
    t = Cpu() -t;
    if (_multiscaleFlag == false)
      Msg::Info("Done performing in-system condensation (%f s)",t);

    fillTangent(L,_elasticState);

    if (_multiscaleFlag == false){
      _elasticState->getHomogenizedTangentOperator_F_F().print("Elastic tangent operator");
      Msg::Info("done extracting elastic tengent operator");
    }
  };

};

void nonLinearMechSolver::extractAverageStress(homogenizedData* homoData){
  if (_homogenizeStressMethod == VOLUME)
    this->extractAverageStressByVolumeIntegral(homoData);
  else if (_homogenizeStressMethod == SURFACE)
    this->extractAverageStressByCondensation(homoData);
  else {
    Msg::Error("This method is not implemented");
  }
};

void nonLinearMechSolver::extractAverageStressAndTangent(homogenizedData* homoData){
  if (_homogenizeTangentMethod == CONDEN or _homogenizeTangentMethod==UNIFIED_CONDEN){
    this->extractAveragePropertiesByCondensation(homoData);
  }
  else if (_homogenizeTangentMethod == PERTURB){
    this->extractAveragePropertiesPerturbation(homoData);
  }
  else if (_homogenizeTangentMethod == INSYSTEMCONDEN){
    this->extractAveragePropertiesByInSystemCondensation(homoData);
  }
  else{
    Msg::Error("Homogenization tangent method must be correctly defined");
  }

  // estimate loss of ellipticity
  if (_checkFailureOnset){
    const STensor43& L = homoData->getHomogenizedTangentOperator_F_F();
    double& cr = homoData->getLostOfSolutionUniquenessCriterion();
    cr = _lostSolutionUniquenssTolerance - 1.;

    double detNLN = 0;
    if (_checkWithNormal){
      // check lost of solution uniqueness
      if (_lostSolutionUniquenssNormal.norm() <=0.){
        Msg::Error("can not check failure with zero normal");
      }
      //
      FailureDetection::getLostSolutionUniquenessCriterionFollowingDirection(_dim,L,_lostSolutionUniquenssNormal,detNLN);
    }
    else{
      // the material softening is not allowed in bulk IP
      FailureDetection::getLostSolutionUniquenessCriterion(_dim,L,_lostSolutionUniquenssNormal,detNLN);
    }

    // maximal Cr
		if (_maximalLostEllipticityCriterion < detNLN and detNLN > 0.){
			_maximalLostEllipticityCriterion = detNLN;
		}
		if (_maximalLostEllipticityCriterion > 0){
      cr = _lostSolutionUniquenssTolerance -  detNLN/_maximalLostEllipticityCriterion;
		}
  }
};

void nonLinearMechSolver::extractAveragePropertiesPerturbation(homogenizedData* homoData){
  // extract stress
  this->extractAverageStress(homoData);

	// mesh  chracteristic size for perturbation
	double elength = 0.;
	for (int i=0; i< domainVector.size(); i++){
		if (domainVector[i]->groupOfElementsSize() > 0){
			const MElement* ele = *(domainVector[0]->g_cbegin());
			double volume = const_cast<MElement*>(ele)->getVolume();

			if (_dim == 2){
				elength = sqrt(volume);
			}
			else if (_dim ==3){
				elength = pow(volume,1.0/3.0);
			}
			break;
		}
	}

	//
	/** function for compute tangent by perturbation on current state**/
    /*
    save all computed current data to tmp step
  */

  _ipf->copy(IPStateBase::current,IPStateBase::temp);
  std::string Aname ="A";
  linearSystem<double>* lsys = pAssembler->getLinearSystem(Aname);
  nonLinearSystem<double>* nonlsys = dynamic_cast<nonLinearSystem<double>*>(lsys);
  nonlsys->copy(IPStateBase::current,IPStateBase::temp);

	homogenizedData homoDataPlus(*homoData);

	double t= Cpu();
  if (_multiscaleFlag == false)
    Msg::Info("Begin performing tangent by perturbation");

	if (_microBC->getTotalNumberOfMechanicalDofs()> 0){
		STensor3 F = _microBC->getFirstOrderKinematicalVariable();
		for (int i=0; i<_dim; i++){
			for (int j=0; j<_dim; j++){
				nonlsys->copy(IPStateBase::previous,IPStateBase::current);
				/** perturbation in deformation gradient **/
				STensor3 Fplus(F);
				Fplus(i,j) += _tangentPerturbation;
				_microBC->setFirstOrderKinematicalVariable(Fplus);

				/** solve perturbated system**/
				this->OneStep(0);

				/** get homogenized stress in perturbed system **/
				this->extractAverageStress(&homoDataPlus);

				const STensor3& P = homoData->getHomogenizedStress();
				const STensor3& Pplus = homoDataPlus.getHomogenizedStress();
				for (int k=0; k<3; k++){
					for (int l=0; l<3; l++){
						homoData->getHomogenizedTangentOperator_F_F()(k,l,i,j) = (Pplus(k,l) - P(k,l))/(_tangentPerturbation);
					}
				}

				if (_microBC->getOrder() == 2){
					const STensor33& Q = homoData->getHomogenizedSecondOrderStress();
					const STensor33& Qplus = homoDataPlus.getHomogenizedSecondOrderStress();
					for (int k=0; k<3; k++)
						for (int l=0; l<3; l++)
							for (int m=0; m<3; m++)
								homoData->getHomogenizedTangentOperator_G_F()(k,l,m,i,j) = (Qplus(k,l,m) - Q(k,l,m))/(_tangentPerturbation);
				}

				if (_damageToCohesiveJump){
					const SVector3& cjump = homoData->getHomogenizedCohesiveJump();
					const SVector3& cjumpplus = homoDataPlus.getHomogenizedCohesiveJump();
					for (int k=0; k<3; k++){
						homoData->getHomogenizedTangentOperator_CohesiveJump_F()(k,i,j) = (cjumpplus(k) - cjump(k))/(_tangentPerturbation);
					}
				}

				if (_extractIrreversibleEnergy){
					const double& irrEnerg = homoData->getIrreversibleEnergy();
					const double& irrEnergPlus = homoDataPlus.getIrreversibleEnergy();
					homoData->getHomogenizedTangentOperator_IrreversibleEnergy_F()(i,j) = (irrEnergPlus - irrEnerg)/_tangentPerturbation;
				}


				if (_microBC->getTotalNumberOfConDofs() > 0){
					for (int index=0; index < _microBC->getTotalNumberOfConDofs(); index++){
						const SVector3& flux = homoData->getHomogenizedConstitutiveExtraDofFlux(index);
						const SVector3& fluxPlus = homoDataPlus.getHomogenizedConstitutiveExtraDofFlux(index);
						for (int k=0; k<3; k++){
							homoData->getHomogenizedTangentOperator_gradT_F(index)(k,i,j) = (fluxPlus(k)-flux(k))/(_tangentPerturbation);

						}
						const double& mecasrc = homoData->getHomogenizedConstitutiveExtraDofMechanicalSource(index);
						const double& mecasrcPlus = homoDataPlus.getHomogenizedConstitutiveExtraDofMechanicalSource(index);
						homoData->getHomogenizedTangentOperator_ConstitutiveExtraDofMechanicalSource_F(index)(i,j) = (mecasrcPlus-mecasrc)/(_tangentPerturbation);
					}
				}

				if (_microBC->getTotalNumberOfNonConDofs() > 0){
					for (int index=0; index < _microBC->getTotalNumberOfNonConDofs(); index++){
						const SVector3& flux = homoData->getHomogenizedNonConstitutiveExtraDofFlux(index);
						const SVector3& fluxPlus = homoDataPlus.getHomogenizedNonConstitutiveExtraDofFlux(index);
						for (int k=0; k<3; k++){
							homoData->getHomogenizedTangentOperator_gradV_F(index)(k,i,j) = (fluxPlus(k)-flux(k))/(_tangentPerturbation);

						}
					}
				}
			}
		}
		/** back micro BC as before perturbation**/
		_microBC->setFirstOrderKinematicalVariable(F);



		if (_microBC->getOrder() == 2){
			/**perturbation in gradient of deformation gradient **/
			STensor33 G = _microBC->getSecondOrderKinematicalVariable();
			double secondpert = _tangentPerturbation/elength;

			for (int i=0; i<_dim; i++){
				for (int j=0; j<_dim; j++){
					for (int s=0; s<=j; s++){
						nonlsys->copy(IPStateBase::previous,IPStateBase::current);
						STensor33 Gplus(G);
						Gplus(i,j,s) += secondpert;
						Gplus(i,s,j) += secondpert;

						_microBC->setSecondOrderKinematicalVariable(Gplus);
						/** solve perturbed system**/
						this->OneStep(0);

						/** get homogenized stress in perturbed system **/
						this->extractAverageStress(&homoDataPlus);

						//
						const STensor3& P = homoData->getHomogenizedStress();
						const STensor3& Pplus = homoDataPlus.getHomogenizedStress();
						for (int k=0; k<3; k++){
							for (int l=0; l<3; l++){
								homoData->getHomogenizedTangentOperator_F_G()(k,l,i,j,s) = (Pplus(k,l) - P(k,l))/(2.*secondpert);
								homoData->getHomogenizedTangentOperator_F_G()(k,l,i,s,j) = homoData->getHomogenizedTangentOperator_F_G()(k,l,i,j,s);
							}
						}

						const STensor33& Q = homoData->getHomogenizedSecondOrderStress();
						const STensor33& Qplus = homoDataPlus.getHomogenizedSecondOrderStress();
						for (int k=0; k<3; k++){
							for (int l=0; l<3; l++){
								for (int m=0; m<3; m++){
									homoData->getHomogenizedTangentOperator_G_G()(k,l,m,i,j,s) = (Qplus(k,l,m) - Q(k,l,m))/(2.*secondpert);
									homoData->getHomogenizedTangentOperator_G_G()(k,l,m,i,s,j)  = homoData->getHomogenizedTangentOperator_G_G()(k,l,m,i,j,s);
								}
							}
						}

						if (_damageToCohesiveJump){
							const SVector3& cjump = homoData->getHomogenizedCohesiveJump();
							const SVector3& cjumpplus = homoDataPlus.getHomogenizedCohesiveJump();
							for (int k=0; k<3; k++){
								homoData->getHomogenizedTangentOperator_CohesiveJump_G()(k,i,j,s) = (cjumpplus(k) - cjump(k))/(2.*secondpert);
								homoData->getHomogenizedTangentOperator_CohesiveJump_G()(k,i,s,j) = homoData->getHomogenizedTangentOperator_CohesiveJump_G()(k,i,j,s);
							}
						}

						if (_extractIrreversibleEnergy){
							const double& irrEnerg = homoData->getIrreversibleEnergy();
							const double& irrEnergPlus = homoDataPlus.getIrreversibleEnergy();
							homoData->getHomogenizedTangentOperator_IrreversibleEnergy_G()(i,j,s) = (irrEnergPlus - irrEnerg)/(2.*secondpert);
							homoData->getHomogenizedTangentOperator_IrreversibleEnergy_G()(i,s,j) = homoData->getHomogenizedTangentOperator_IrreversibleEnergy_G()(i,j,s);
						}

						if (_microBC->getTotalNumberOfConDofs() > 0){
							for (int index=0; index < _microBC->getTotalNumberOfConDofs(); index++){
								const SVector3& flux = homoData->getHomogenizedConstitutiveExtraDofFlux(index);
								const SVector3& fluxPlus = homoDataPlus.getHomogenizedConstitutiveExtraDofFlux(index);
								for (int k=0; k<3; k++){
									homoData->getHomogenizedTangentOperator_gradT_G(index)(k,i,j,s) = (fluxPlus(k)-flux(k))/(2.*secondpert);
									homoData->getHomogenizedTangentOperator_gradT_G(index)(k,i,s,j) = homoData->getHomogenizedTangentOperator_gradT_G(index)(k,i,j,s);
								}
								const double& mecasrc = homoData->getHomogenizedConstitutiveExtraDofMechanicalSource(index);
								const double& mecasrcPlus = homoDataPlus.getHomogenizedConstitutiveExtraDofMechanicalSource(index);
								homoData->getHomogenizedTangentOperator_ConstitutiveExtraDofMechanicalSource_G(index)(i,j,s) = (mecasrcPlus-mecasrc)/(2.*secondpert);
								homoData->getHomogenizedTangentOperator_ConstitutiveExtraDofMechanicalSource_G(index)(i,s,j) = homoData->getHomogenizedTangentOperator_ConstitutiveExtraDofMechanicalSource_G(index)(i,j,s);
							}
						}

						if (_microBC->getTotalNumberOfNonConDofs() > 0){
							for (int index=0; index < _microBC->getTotalNumberOfNonConDofs(); index++){
								const SVector3& flux = homoData->getHomogenizedNonConstitutiveExtraDofFlux(index);
								const SVector3& fluxPlus = homoDataPlus.getHomogenizedNonConstitutiveExtraDofFlux(index);
								for (int k=0; k<3; k++){
									homoData->getHomogenizedTangentOperator_gradV_G(index)(k,i,j,s) = (fluxPlus(k)-flux(k))/(2.*secondpert);
									homoData->getHomogenizedTangentOperator_gradV_G(index)(k,i,s,j) = homoData->getHomogenizedTangentOperator_gradV_G(index)(k,i,j,s);
								}
							}
						}

					}
				}
			}
			_microBC->setSecondOrderKinematicalVariable(G);
		}
	}

	if (_microBC->getTotalNumberOfConDofs() > 0){
		for (int index = 0; index < _microBC->getTotalNumberOfConDofs(); index ++){
			SVector3 gradT = _microBC->getConstitutiveExtraDofDiffusionKinematicalVariable(index);
			double gradTpert = _tangentPerturbation*(_microBC->getInitialConstitutiveExtraDofDiffusionValue(index))/elength;

			for (int i=0; i<_dim; i++){
				nonlsys->copy(IPStateBase::previous,IPStateBase::current);
				SVector3 gradTplus(gradT);

				gradTplus(i) += gradTpert;
				_microBC->setConstitutiveExtraDofDiffusionKinematicalVariable(index,gradTplus);
				/** solve perturbed system**/
				this->OneStep(0);

				/** get homogenized stress in perturbed system **/
				this->extractAverageStress(&homoDataPlus);

				if (_microBC->getTotalNumberOfMechanicalDofs() > 0){
					const STensor3& P = homoData->getHomogenizedStress();
					const STensor3& Pplus = homoDataPlus.getHomogenizedStress();
					for (int p=0; p<3; p++){
						for (int q=0; q<3; q++){
							homoData->getHomogenizedTangentOperator_F_gradT(index)(p,q,i) = (Pplus(p,q)- P(p,q))/(gradTpert);
						}
					}

					if (_microBC->getOrder() == 2){
						const STensor33& Q = homoData->getHomogenizedSecondOrderStress();
						const STensor33& Qplus = homoDataPlus.getHomogenizedSecondOrderStress();
						for (int p=0; p<3; p++)
							for (int q=0; q<3; q++)
								for (int r=0; r<3; r++)
									homoData->getHomogenizedTangentOperator_G_gradT(index)(p,q,r,i) = (Qplus(p,q,r) - Q(p,q,r))/(gradTpert);
					}

					if (_damageToCohesiveJump){
						const SVector3& cjump = homoData->getHomogenizedCohesiveJump();
						const SVector3& cjumpplus = homoDataPlus.getHomogenizedCohesiveJump();
						for (int p=0; p<3; p++){
							homoData->getHomogenizedTangentOperator_CohesiveJump_gradT(index)(p,i) = (cjumpplus(p) - cjump(p))/(gradTpert);
						}
					}

					if (_extractIrreversibleEnergy){
						const double& irrEnerg = homoData->getIrreversibleEnergy();
						const double& irrEnergPlus = homoDataPlus.getIrreversibleEnergy();
						homoData->getHomogenizedTangentOperator_IrreversibleEnergy_gradT(index)(i) = (irrEnergPlus-irrEnerg)/gradTpert;
					}
				}

				for (int fluxIndex = 0; fluxIndex < _microBC->getTotalNumberOfConDofs(); fluxIndex ++){
					const SVector3& flux = homoData->getHomogenizedConstitutiveExtraDofFlux(fluxIndex);
					const SVector3& fluxPlus = homoDataPlus.getHomogenizedConstitutiveExtraDofFlux(fluxIndex);
					for (int p=0; p<3; p++){
						homoData->getHomogenizedTangentOperator_gradT_gradT(fluxIndex,index)(p,i) = (fluxPlus(p) - flux(p))/(gradTpert);
					}

					const double& source = homoData->getHomogenizedConstitutiveExtraDofMechanicalSource(fluxIndex);
					const double& sourcePlus = homoDataPlus.getHomogenizedConstitutiveExtraDofMechanicalSource(fluxIndex);
					homoData->getHomogenizedTangentOperator_ConstitutiveExtraDofMechanicalSource_gradT(fluxIndex,index)(i) = (sourcePlus-source)/gradTpert;
				};

				for (int fluxIndex = 0; fluxIndex < _microBC->getTotalNumberOfNonConDofs(); fluxIndex ++){
					const SVector3& flux = homoData->getHomogenizedNonConstitutiveExtraDofFlux(fluxIndex);
					const SVector3& fluxPlus = homoDataPlus.getHomogenizedNonConstitutiveExtraDofFlux(fluxIndex);
					for (int p=0; p<3; p++){
						homoData->getHomogenizedTangentOperator_gradV_gradT(fluxIndex,index)(p,i) = (fluxPlus(p) - flux(p))/(gradTpert);
					}
				}
				_microBC->setConstitutiveExtraDofDiffusionKinematicalVariable(index,gradT);
			}

			// perturtabtion on T
			const double T = _microBC->getConstitutiveExtraDofDiffusionConstantVariable(index);
			double Tpert = _tangentPerturbation*(_microBC->getInitialConstitutiveExtraDofDiffusionValue(index));

			nonlsys->copy(IPStateBase::previous,IPStateBase::current);
			double Tplus = T+Tpert;

			_microBC->setConstitutiveExtraDofDiffusionConstantVariable(index,Tplus);
			/** solve perturbed system**/
			this->OneStep(0);

			/** get homogenized stress in perturbed system **/
			this->extractAverageStress(&homoDataPlus);

			if (_microBC->getTotalNumberOfMechanicalDofs() > 0){
				const STensor3& P = homoData->getHomogenizedStress();
				const STensor3& Pplus = homoDataPlus.getHomogenizedStress();
				for (int p=0; p<3; p++){
					for (int q=0; q<3; q++){
						homoData->getHomogenizedTangentOperator_F_T(index)(p,q) = (Pplus(p,q)- P(p,q))/(Tpert);
					}
				}

				if (_microBC->getOrder() == 2){
					const STensor33& Q = homoData->getHomogenizedSecondOrderStress();
					const STensor33& Qplus = homoDataPlus.getHomogenizedSecondOrderStress();
					for (int p=0; p<3; p++)
						for (int q=0; q<3; q++)
							for (int r=0; r<3; r++)
								homoData->getHomogenizedTangentOperator_G_T(index)(p,q,r) = (Qplus(p,q,r) - Q(p,q,r))/(Tpert);
				}

				if (_damageToCohesiveJump){
					const SVector3& cjump = homoData->getHomogenizedCohesiveJump();
					const SVector3& cjumpplus = homoDataPlus.getHomogenizedCohesiveJump();
					for (int p=0; p<3; p++){
						homoData->getHomogenizedTangentOperator_CohesiveJump_T(index)(p) = (cjumpplus(p) - cjump(p))/(Tpert);
					}
				}
				if (_extractIrreversibleEnergy){
					const double& irrEnerg = homoData->getIrreversibleEnergy();
					const double& irrEnergPlus = homoDataPlus.getIrreversibleEnergy();
					homoData->getHomogenizedTangentOperator_IrreversibleEnergy_T(index) = (irrEnergPlus-irrEnerg)/Tpert;
				}
			}

			for (int fluxIndex = 0; fluxIndex < _microBC->getTotalNumberOfConDofs(); fluxIndex ++){
				const SVector3& flux = homoData->getHomogenizedConstitutiveExtraDofFlux(fluxIndex);
				const SVector3& fluxPlus = homoDataPlus.getHomogenizedConstitutiveExtraDofFlux(fluxIndex);
				for (int p=0; p<3; p++){
					homoData->getHomogenizedTangentOperator_gradT_T(fluxIndex,index)(p) = (fluxPlus(p) - flux(p))/(Tpert);
				}

				const double& source = homoData->getHomogenizedConstitutiveExtraDofMechanicalSource(fluxIndex);
				const double& sourcePlus = homoDataPlus.getHomogenizedConstitutiveExtraDofMechanicalSource(fluxIndex);
				homoData->getHomogenizedTangentOperator_ConstitutiveExtraDofMechanicalSource_T(fluxIndex,index) = (sourcePlus-source)/Tpert;
			};

			for (int fluxIndex = 0; fluxIndex < _microBC->getTotalNumberOfNonConDofs(); fluxIndex ++){
				const SVector3& flux = homoData->getHomogenizedNonConstitutiveExtraDofFlux(fluxIndex);
				const SVector3& fluxPlus = homoDataPlus.getHomogenizedNonConstitutiveExtraDofFlux(fluxIndex);
				for (int p=0; p<3; p++){
					homoData->getHomogenizedTangentOperator_gradV_T(fluxIndex,index)(p) = (fluxPlus(p) - flux(p))/(Tpert);
				}
			}
			_microBC->setConstitutiveExtraDofDiffusionConstantVariable(index,T);
		}
	}

	if (_microBC->getTotalNumberOfNonConDofs() > 0){
		for (int index = 0; index < _microBC->getTotalNumberOfNonConDofs(); index ++){
			SVector3 gradV = _microBC->getNonConstitutiveExtraDofDiffusionKinematicalVariable(index);
			double gradVpert = (1+gradV.norm())*_tangentPerturbation;

			for (int i=0; i<_dim; i++){
				nonlsys->copy(IPStateBase::previous,IPStateBase::current);
				SVector3 gradVplus(gradV);
				gradVplus(i) += gradVpert;
				_microBC->setNonConstitutiveExtraDofDiffusionKinematicalVariable(index,gradVplus);
				/** solve perturbed system**/
				this->OneStep(0);

				/** get homogenized stress in perturbed system **/
				this->extractAverageStress(&homoDataPlus);

				if (_microBC->getTotalNumberOfMechanicalDofs() > 0){
					const STensor3& P = homoData->getHomogenizedStress();
					const STensor3& Pplus = homoDataPlus.getHomogenizedStress();
					for (int p=0; p<3; p++){
						for (int q=0; q<3; q++){
							homoData->getHomogenizedTangentOperator_F_gradV(index)(p,q,i) = (Pplus(p,q)- P(p,q))/(gradVpert);
						}
					}

					if (_microBC->getOrder() == 2){
						const STensor33& Q = homoData->getHomogenizedSecondOrderStress();
						const STensor33& Qplus = homoDataPlus.getHomogenizedSecondOrderStress();
						for (int p=0; p<3; p++)
							for (int q=0; q<3; q++)
								for (int r=0; r<3; r++)
									homoData->getHomogenizedTangentOperator_G_gradV(index)(p,q,r,i) = (Qplus(p,q,r) - Q(p,q,r))/(gradVpert);
					}

					if (_damageToCohesiveJump){
						const SVector3& cjump = homoData->getHomogenizedCohesiveJump();
						const SVector3& cjumpplus = homoDataPlus.getHomogenizedCohesiveJump();
						for (int p=0; p<3; p++){
							homoData->getHomogenizedTangentOperator_CohesiveJump_gradV(index)(p,i) = (cjumpplus(p) - cjump(p))/(gradVpert);
						}
					}
					if (_extractIrreversibleEnergy){
						const double& irrEnerg = homoData->getIrreversibleEnergy();
						const double& irrEnergPlus = homoDataPlus.getIrreversibleEnergy();
						homoData->getHomogenizedTangentOperator_IrreversibleEnergy_gradV(index)(i) = (irrEnergPlus-irrEnerg)/gradVpert;
					}
				}

				for (int fluxIndex = 0; fluxIndex < _microBC->getTotalNumberOfConDofs(); fluxIndex ++){
					const SVector3& flux = homoData->getHomogenizedConstitutiveExtraDofFlux(fluxIndex);
					const SVector3& fluxPlus = homoDataPlus.getHomogenizedConstitutiveExtraDofFlux(fluxIndex);
					for (int p=0; p<3; p++){
						homoData->getHomogenizedTangentOperator_gradT_gradV(fluxIndex,index)(p,i) = (fluxPlus(p) - flux(p))/(gradVpert);
					}

					const double& source = homoData->getHomogenizedConstitutiveExtraDofMechanicalSource(fluxIndex);
					const double& sourcePlus = homoDataPlus.getHomogenizedConstitutiveExtraDofMechanicalSource(fluxIndex);
					homoData->getHomogenizedTangentOperator_ConstitutiveExtraDofMechanicalSource_gradV(fluxIndex,index)(i) = (sourcePlus-source)/gradVpert;
				};

				for (int fluxIndex = 0; fluxIndex < _microBC->getTotalNumberOfNonConDofs(); fluxIndex ++){
					const SVector3& flux = homoData->getHomogenizedNonConstitutiveExtraDofFlux(fluxIndex);
					const SVector3& fluxPlus = homoDataPlus.getHomogenizedNonConstitutiveExtraDofFlux(fluxIndex);
					for (int p=0; p<3; p++){
						homoData->getHomogenizedTangentOperator_gradV_gradV(fluxIndex,index)(p,i) = (fluxPlus(p) - flux(p))/(gradVpert);
					}
				}
				_microBC->setNonConstitutiveExtraDofDiffusionKinematicalVariable(index,gradV);
			}
		}
	}

  t = Cpu() -t;
  if (_multiscaleFlag == false)
    Msg::Info("Done performing tangent by perturbation (%f s)",t);;

  /**back to current step from temp state **/
  _ipf->copy(IPStateBase::temp,IPStateBase::current);
  nonlsys->copy(IPStateBase::temp,IPStateBase::current);
}

void nonLinearMechSolver::computeDensity(){
  _rho = 0;
  for (int i=0; i<domainVector.size(); i++){
    partDomain* dom = domainVector[i];
		// only on volumic domain
		if (dom->groupOfElementsSize()>0){
			double rhodom = dom->getMaterialLaw()->density();
			double voldom =dom->computeVolumeDomain(_ipf);
			_rho += (voldom*rhodom);
		}
  };
  double vtotal = this->getRVEVolume();
  _rho /= vtotal;
};

stiffnessCondensation* nonLinearMechSolver::createStiffnessCondensation(const bool sflag, const bool tflag){
  stiffnessCondensation* stiffCon = NULL;

  /**condensation require **/
  if (_pAl == NULL){
    _pAl = new pbcAlgorithm(this);
  };
  if (!_pAl->isSplittedDof()) _pAl->splitDofs();

  if (_homogenizeTangentMethod == UNIFIED_CONDEN){
    #if defined(HAVE_PETSC)
    Msg::Info("stiffnessCondensationPETSc is used for nonLinearMixedBC");
    stiffCon = new stiffnessCondensationPETSc(pAssembler,_pAl,sflag,tflag);
    #else
    Msg::Error("Petsc must be used");
    #endif // HAVE_PETSC
  }
  else if (_homogenizeTangentMethod == CONDEN){
    if (dynamic_cast<nonLinearMinimalKinematicBC*>(_microBC)){
      #if defined(HAVE_PETSC)
      Msg::Info("stiffnessCondensationPETScWithoutDirectConstraints is used for nonLinearMinimalKinematicBC");
      stiffCon = new stiffnessCondensationPETScWithoutDirectConstraints(pAssembler,_pAl,sflag,tflag);
      #else
      Msg::Error("Petsc must be used");
      #endif // HAVE_PETSC
    }
    else if (dynamic_cast<nonLinearDisplacementBC*>(_microBC)){
      #if defined(HAVE_PETSC)
      Msg::Info("stiffnessCondensationLDBCPETSc is used for nonLinearDisplacementBC");
      stiffCon = new stiffnessCondensationLDBCPETSc(pAssembler,_pAl,sflag,tflag);
      #else
      Msg::Error("Petsc must be used");
      #endif // HAVE_PETSC
    }
    else if (dynamic_cast<nonLinearPeriodicBC*>(_microBC)){
      nonLinearPeriodicBC* pbc = dynamic_cast<nonLinearPeriodicBC*>(_microBC);
      if (_pAl->getSplittedDof()->sizeOfVirtualDof() >0 and pbc->getOrder()==1){
        #if defined(HAVE_PETSC)
        Msg::Info("interpolationStiffnessCondensationPETSC is used for nonLinearPeriodicBC");
        stiffCon = new interpolationStiffnessCondensationPETSC(pAssembler,_pAl,sflag,tflag);
        #else
        Msg::Error("Petsc must be used");
        #endif
      }
      else if ((pbc->getPBCMethod() == nonLinearPeriodicBC::LIM) and (pbc->getMaxDegree() == 1)){
        #if defined(HAVE_PETSC)
        Msg::Info("stiffnessCondensationLDBCPETSc is used for nonLinearPeriodicBC as order = 1");
        stiffCon = new stiffnessCondensationLDBCPETSc(pAssembler,_pAl,sflag,tflag);
        #else
        Msg::Error("Petsc must be used");
        #endif
      }
      else{
        #if defined(HAVE_PETSC)
        if (_pAl->getNumberDirectConstraints()>0){
          Msg::Info("stiffnessCondensationPETScWithDirectConstraints is used for nonLinearPeriodicBC");
          stiffCon = new stiffnessCondensationPETScWithDirectConstraints(pAssembler,_pAl,sflag,tflag);
        }
        else{
          Msg::Info("stiffnessCondensationPETScWithoutDirectConstraints is used for nonLinearPeriodicBC");
          stiffCon = new stiffnessCondensationPETScWithoutDirectConstraints(pAssembler,_pAl,sflag,tflag);
        }
        #else
        if (_pAl->getNumberDirectConstraints()>0){
          Msg::Info("stiffnessCondensationFullMatrixWithDirectConstraints is used for nonLinearPeriodicBC");
          stiffCon = new stiffnessCondensationFullMatrixWithDirectConstraints(pAssembler,_pAl,sflag,tflag);
        }
        else{
          Msg::Error("Petsc must be used");
        }
        #endif // HAVE_PETSC
      }

    }
    else if (dynamic_cast<nonLinearMixedBC*>(_microBC)){
      #if defined(HAVE_PETSC)
      if (_pAl->getNumberDirectConstraints()>0){
        Msg::Info("stiffnessCondensationPETScWithDirectConstraints is used for nonLinearMixedBC");
        stiffCon = new stiffnessCondensationPETScWithDirectConstraints(pAssembler,_pAl,sflag,tflag);
      }
      else{
        Msg::Info("stiffnessCondensationPETScWithoutDirectConstraints is used for nonLinearMixedBC");
        stiffCon = new stiffnessCondensationPETScWithoutDirectConstraints(pAssembler,_pAl,sflag,tflag);
      }
      #else
      if (_pAl->getNumberDirectConstraints()>0){
        Msg::Info("stiffnessCondensationFullMatrixWithDirectConstraints is used for nonLinearMixedBC");
        stiffCon = new stiffnessCondensationFullMatrixWithDirectConstraints(pAssembler,_pAl,sflag,tflag);
      }
      else{
        Msg::Error("Petsc must be used");
      }
      #endif // HAVE_PETSC
    }
    else{
      Msg::Error("Micro BC type must be defined");
    }

  }

  //
  if (stiffCon == NULL){
    Msg::Warning("only homogenized stress is computed by surface integral");
    #if defined(HAVE_PETSC)
    Msg::Info("stiffnessCondensationPETSc is used");
    stiffCon = new stiffnessCondensationPETSc(pAssembler,_pAl,sflag,tflag);
    #else
    Msg::Error("Petsc must be used");
    #endif // HAVE_PETSC
  }

  stiffCon->setSplittedDofSystem(_pAl);

  if (tflag) {
    _pAl->getSplittedDof()->allocateSplittedMatrix();
    // preallocate
    for(std::vector<partDomain*>::iterator itdom=domainVector.begin(); itdom!=domainVector.end(); ++itdom){
      partDomain* dom = *itdom;
      SparsityDofsSplittedStiffness(*(dom->getFunctionSpace()), dom->g_cbegin(), dom->g_cend(),*_pAl->getSplittedDof());
			dgPartDomain* dgdom = dynamic_cast<dgPartDomain*>(dom);
			if (dgdom!=NULL){
				SparsityDofsSplittedStiffness(*(dgdom->getFunctionSpace()), dgdom->gi->begin(), dgdom->gi->end(),*_pAl->getSplittedDof());
			}
    }
  };

  return stiffCon;
};

void nonLinearMechSolver::fillTangent(const fullMatrix<double>& L, homogenizedData* homoData){
	// stress vector is arranged by this order
	// S = [P Q FD fluxT1 mecaSrc1 -- fluxTn mecaSrcn fluxV1 --- fluxVm]
	// K = [F G gradT1 T1 --- gradTn Tn gradV1 --- gradVm] see nonLinearMicroBC::getKinematicVector
	// from dS/dK --> all tangents are explicitly given

  // mechanics part
  int atRow = 0;
	if (_microBC->getTotalNumberOfMechanicalDofs() > 0){
		for (int row = atRow; row<atRow+9; row++){
			int i,j;
			Tensor23::getIntsFromIndex(row-atRow,i,j);
			// tangent P-F
			int atCol = 0;

			for (int col = atCol; col< atCol+9; col++){
				int k,l;
				Tensor23::getIntsFromIndex(col-atCol,k,l);
				homoData->getHomogenizedTangentOperator_F_F()(i,j,k,l) = L(row,col);
			}
			atCol+= 9;

			// tangent P-secondorder
			if (_microBC->getOrder() == 2){
				for (int col =atCol; col< atCol+27; col++){
					int p,q,r;
					Tensor33::getIntsFromIndex(col-atCol,p,q,r);
					homoData->getHomogenizedTangentOperator_F_G()(i,j,p,q,r) = L(row,col);
				};
				atCol += 27;
			}

			if (_microBC->getTotalNumberOfConDofs()>0){
				for (int index=0; index < _microBC->getTotalNumberOfConDofs(); index++){
					// tangent P-GradextraDof
					for (int col = atCol; col< atCol+3; col++){
						int p;
						Tensor13::getIntsFromIndex(col-atCol,p);
						homoData->getHomogenizedTangentOperator_F_gradT(index)(i,j,p) = L(row,col);
					}
					atCol += 3;
					// tangent P-ExtraDof
					homoData->getHomogenizedTangentOperator_F_T(index)(i,j) =L(row,atCol);
					atCol += 1;

				}
			};

			if (_microBC->getTotalNumberOfNonConDofs() > 0){
				for (int index=0; index < _microBC->getTotalNumberOfNonConDofs(); index++){
					// tangent P-GradextraDof
					for (int col = atCol; col< atCol+3; col++){
						int p;
						Tensor13::getIntsFromIndex(col-atCol,p);
						homoData->getHomogenizedTangentOperator_F_gradV(index)(i,j,p) = L(row,col);
					}
					atCol += 3;
				}
			}
		};
		atRow += 9;

		if (_microBC->getOrder() == 2){
			// second order stress
			for (int row = atRow; row<atRow+27; row++){
				int i,j,k;
				Tensor33::getIntsFromIndex(row-atRow,i,j,k);

				// tangent Q-F
				int atCol = 0;
				for (int col = atCol; col<atCol+9; col++){
					int p,q;
					Tensor23::getIntsFromIndex(col-atCol,p,q);
					homoData->getHomogenizedTangentOperator_G_F()(i,j,k,p,q) = L(row,col);
				}
				atCol+= 9;

				// tangent Q-Q;
				for (int col=atCol; col < atCol+27; col++){
					int p,q,r;
					Tensor33::getIntsFromIndex(col-atCol,p,q,r);
					homoData->getHomogenizedTangentOperator_G_G()(i,j,k,p,q,r) = L(row,col);
				}
				atCol += 27;

				if (_microBC->getTotalNumberOfConDofs()>0){
					for (int index=0; index < _microBC->getTotalNumberOfConDofs(); index++){
						// tangent P-GradextraDof
						for (int col = atCol; col< atCol+3; col++){
							int p;
							Tensor13::getIntsFromIndex(col-atCol,p);
							homoData->getHomogenizedTangentOperator_G_gradT(index)(i,j,k,p) = L(row,col);
						}
						atCol += 3;
						// tangent P-ExtraDof
						homoData->getHomogenizedTangentOperator_G_T(index)(i,j,k) =L(row,atCol);
						atCol += 1;

					}
				};

				if (_microBC->getTotalNumberOfNonConDofs() > 0){
					for (int index=0; index < _microBC->getTotalNumberOfNonConDofs(); index++){
						// tangent P-GradextraDof
						for (int col = atCol; col< atCol+3; col++){
							int p;
							Tensor13::getIntsFromIndex(col-atCol,p);
							homoData->getHomogenizedTangentOperator_G_gradV(index)(i,j,k,p) = L(row,col);
						}
						atCol += 3;
					}
				}

			};
			atRow += 27;
		}

		if (_extractIrreversibleEnergy and (_homogenizeTangentMethod == INSYSTEMCONDEN)){
			for (int row = atRow; row<atRow+1; row++){
				// tangent FD-F
				int atCol = 0;
				for (int col = atCol; col< atCol+9; col++){
					int k,l;
					Tensor23::getIntsFromIndex(col-atCol,k,l);
					homoData->getHomogenizedTangentOperator_IrreversibleEnergy_F()(k,l) = L(row,col);
				}
				atCol+= 9;

				// tangent FD-G
				if (_microBC->getOrder() == 2){
					for (int col = atCol; col< atCol+27; col++){
						int p,q,r;
						Tensor33::getIntsFromIndex(col-atCol,p,q,r);
						homoData->getHomogenizedTangentOperator_IrreversibleEnergy_G()(p,q,r) = L(row,col);
					}
					atCol+= 27;
				}

				if (_microBC->getTotalNumberOfConDofs()>0){
					for (int index=0; index < _microBC->getTotalNumberOfConDofs(); index++){
						// tangent P-GradextraDof
						for (int col = atCol; col< atCol+3; col++){
							int p;
							Tensor13::getIntsFromIndex(col-atCol,p);
							homoData->getHomogenizedTangentOperator_IrreversibleEnergy_gradT(index)(p) = L(row,col);
						}
						atCol += 3;
						// tangent P-ExtraDof
						homoData->getHomogenizedTangentOperator_IrreversibleEnergy_T(index) =L(row,atCol);
						atCol += 1;

					}
				};

				if (_microBC->getTotalNumberOfNonConDofs() > 0){
					for (int index=0; index < _microBC->getTotalNumberOfNonConDofs(); index++){
						// tangent P-GradextraDof
						for (int col = atCol; col< atCol+3; col++){
							int p;
							Tensor13::getIntsFromIndex(col-atCol,p);
							homoData->getHomogenizedTangentOperator_IrreversibleEnergy_gradV(index)(p) = L(row,col);
						}
						atCol += 3;
					}
				}
			}
			atRow += 1;
		}

		// this tangent is only computed by using option INSYSTEMCONDEN
		if (_damageToCohesiveJump and (_homogenizeTangentMethod == INSYSTEMCONDEN)){
			// first order damage only
			double Vd = homoData->getActiveDissipationVolume();
			double vRve = this->getRVEVolume();
			double fact = vRve/Vd;

			for (int row = atRow; row<atRow+9; row++){
				int i,j;
				Tensor23::getIntsFromIndex(row-atRow,i,j);

				// tangent FD-F
				int atCol = 0;
				for (int col = atCol; col< atCol+9; col++){
					int k,l;
					Tensor23::getIntsFromIndex(col-atCol,k,l);
					homoData->getHomogenizedTangentOperator_ActiveDissipationDeformationGradient_F()(i,j,k,l) = fact*L(row,col);
				}
				atCol+= 9;

				// tangent FD-G
				if (_microBC->getOrder() == 2){
					for (int col = atCol; col< atCol+27; col++){
						int p,q,r;
						Tensor33::getIntsFromIndex(col-atCol,p,q,r);
						homoData->getHomogenizedTangentOperator_ActiveDissipationDeformationGradient_G()(i,j,p,q,r) = fact*L(row,col);
					}
					atCol+= 27;
				}

				if (_microBC->getTotalNumberOfConDofs()>0){
					for (int index=0; index < _microBC->getTotalNumberOfConDofs(); index++){
						// tangent P-GradextraDof
						for (int col = atCol; col< atCol+3; col++){
							int p;
							Tensor13::getIntsFromIndex(col-atCol,p);
							homoData->getHomogenizedTangentOperator_ActiveDissipationDeformationGradient_gradT(index)(i,j,p) = fact*L(row,col);
						}
						atCol += 3;
						// tangent P-ExtraDof
						homoData->getHomogenizedTangentOperator_ActiveDissipationDeformationGradient_T(index)(i,j) = fact*L(row,atCol);
						atCol += 1;

					}
				};

				if (_microBC->getTotalNumberOfNonConDofs() > 0){
					for (int index=0; index < _microBC->getTotalNumberOfNonConDofs(); index++){
						// tangent P-GradextraDof
						for (int col = atCol; col< atCol+3; col++){
							int p;
							Tensor13::getIntsFromIndex(col-atCol,p);
							homoData->getHomogenizedTangentOperator_ActiveDissipationDeformationGradient_gradV(index)(i,j,p) = fact*L(row,col);
						}
						atCol += 3;
					}
				}
			}
			atRow += 9;
		}
	}

  if (_microBC->getTotalNumberOfConDofs() >0){
		for (int index=0; index < _microBC->getTotalNumberOfConDofs(); index++){
			for (int row = atRow; row<atRow+3; row++){
				int i;
				Tensor13::getIntsFromIndex(row-atRow, i);
				int atCol = 0;
				if (_microBC->getTotalNumberOfMechanicalDofs() > 0){
					// // gradExtraDof - P
					for (int col = atCol; col<atCol+9; col++){
						int p,q;
						Tensor23::getIntsFromIndex(col-atCol,p,q);
						homoData->getHomogenizedTangentOperator_gradT_F(index)(i,p,q) = L(row,col);
					}
					atCol += 9;
					// grad ExtraDof- Q --> need to complete
					if (_microBC->getOrder() == 2){
						for (int col = atCol; col<atCol+27; col++){
							int p,q,r;
							Tensor33::getIntsFromIndex(col-atCol,p,q,r);
							homoData->getHomogenizedTangentOperator_gradT_G(index)(i,p,q,r) = L(row,col);
						}
						atCol += 27;
					}

				}

				for (int fieldInx = 0; fieldInx < _microBC->getTotalNumberOfConDofs(); fieldInx++){
				// gradExtraDOf- grad ExtraDof
					for (int col = atCol; col < atCol+3; col++){
						int p;
						Tensor13::getIntsFromIndex(col-atCol,p);
						homoData->getHomogenizedTangentOperator_gradT_gradT(index,fieldInx)(i,p) = L(row,col);
					}
					atCol += 3;
					// gradExtraDof- ExtraDof
					homoData->getHomogenizedTangentOperator_gradT_T(index,fieldInx)(i) = L(row,atCol);
					atCol += 1;

				}
				if (_microBC->getTotalNumberOfNonConDofs() > 0){
					for (int fieldInx = 0; fieldInx < _microBC->getTotalNumberOfNonConDofs(); fieldInx++){
						for (int col = atCol; col < atCol+3; col++){
							int p;
							Tensor13::getIntsFromIndex(col-atCol,p);
							homoData->getHomogenizedTangentOperator_gradT_gradV(index,fieldInx)(i,p) = L(row,col);
						}
						atCol += 3;
					}
				}
			};
			atRow+=3;

			if (_homogenizeTangentMethod == INSYSTEMCONDEN){
				// for mechanical source
				for (int row = atRow; row<atRow+1; row++){
					// mecasource - P
					int atCol = 0;
					if (_microBC->getTotalNumberOfMechanicalDofs() > 0){
						for (int col = atCol; col<atCol+9; col++){
							int p,q;
							Tensor23::getIntsFromIndex(col-atCol,p,q);
							homoData->getHomogenizedTangentOperator_ConstitutiveExtraDofMechanicalSource_F(index)(p,q) = L(row,col);
						}
						atCol += 9;
						// grad ExtraDof- Q --> need to complete
						if (_microBC->getOrder() == 2){
							for (int col = atCol; col<atCol+27; col++){
								int p,q,r;
								Tensor33::getIntsFromIndex(col-atCol,p,q,r);
								homoData->getHomogenizedTangentOperator_ConstitutiveExtraDofMechanicalSource_G(index)(p,q,r) = L(row,col);
							}
							atCol += 9;
						}

					}
					// mecasource- grad ExtraDof
					for (int fieldInx = 0; fieldInx < _microBC->getTotalNumberOfConDofs(); fieldInx++){
					// grad ExtraDof
						for (int col = atCol; col < atCol+3; col++){
							int p;
							Tensor13::getIntsFromIndex(col-atCol,p);
							homoData->getHomogenizedTangentOperator_ConstitutiveExtraDofMechanicalSource_gradT(index,fieldInx)(p) = L(row,col);
						}
						atCol += 3;

						// - ExtraDof
						homoData->getHomogenizedTangentOperator_ConstitutiveExtraDofMechanicalSource_T(index,fieldInx) = L(row,atCol);
						atCol += 1;

					}
					if (_microBC->getTotalNumberOfNonConDofs() > 0){
						for (int fieldInx = 0; fieldInx < _microBC->getTotalNumberOfNonConDofs(); fieldInx++){
						//  grad nonConExtraDof
							for (int col = atCol; col < atCol+3; col++){
								int p;
								Tensor13::getIntsFromIndex(col-atCol,p);
								homoData->getHomogenizedTangentOperator_ConstitutiveExtraDofMechanicalSource_gradV(index,fieldInx)(p) = L(row,col);
							}
							atCol += 3;
						}
					}
				};
				atRow+=1;
			};
		}
  }

	if (_microBC->getTotalNumberOfNonConDofs() > 0){
		for (int index=0; index < _microBC->getTotalNumberOfNonConDofs(); index++){
			for (int row = atRow; row<atRow+3; row++){
				int i;
				Tensor13::getIntsFromIndex(row-atRow, i);
				int atCol = 0;
				if (_microBC->getTotalNumberOfMechanicalDofs() > 0){
					// // gradExtraDof - P
					for (int col = atCol; col<atCol+9; col++){
						int p,q;
						Tensor23::getIntsFromIndex(col-atCol,p,q);
						homoData->getHomogenizedTangentOperator_gradV_F(index)(i,p,q) = L(row,col);
					}
					atCol += 9;
					// grad ExtraDof- Q --> need to complete
					if (_microBC->getOrder() == 2){
						for (int col = atCol; col<atCol+27; col++){
							int p,q,r;
							Tensor33::getIntsFromIndex(col-atCol,p,q,r);
							homoData->getHomogenizedTangentOperator_gradV_G(index)(i,p,q,r) = L(row,col);
						}
						atCol += 27;
					}

				}
				if (_microBC->getTotalNumberOfConDofs() > 0){
					for (int fieldInx = 0; fieldInx < _microBC->getTotalNumberOfConDofs(); fieldInx++){
					// gradExtraDOf- grad ExtraDof
						for (int col = atCol; col < atCol+3; col++){
							int p;
							Tensor13::getIntsFromIndex(col-atCol,p);
							homoData->getHomogenizedTangentOperator_gradV_gradT(index,fieldInx)(i,p) = L(row,col);
						}
						atCol += 3;

						// gradExtraDof- ExtraDof
						homoData->getHomogenizedTangentOperator_gradV_T(index,fieldInx)(i) = L(row,atCol);
						atCol += 1;

					}
				}

				for (int fieldInx = 0; fieldInx < _microBC->getTotalNumberOfNonConDofs(); fieldInx++){
				// gradExtraDOf- grad ExtraDof
					for (int col = atCol; col < atCol+3; col++){
						int p;
						Tensor13::getIntsFromIndex(col-atCol,p);
						homoData->getHomogenizedTangentOperator_gradV_gradV(index,fieldInx)(i,p) = L(row,col);
					}
					atCol += 3;
				}
			};
			atRow+=3;
		}
	}

};

void nonLinearMechSolver::checkFailureOnset(){
  if (_checkFailureOnset){
    const homogenizedData* homoDataPrev = this->getHomogenizationState(IPStateBase::previous);
    homogenizedData* homoData = this->getHomogenizationState(IPStateBase::current);

    if (!_solverIsBroken){
      const double cr = homoData->getLostOfSolutionUniquenessCriterion();
      if (cr > 0.){
        _solverIsBroken = true;
        _FdamOnset = homoData->getHomogenizedActiveDissipationDeformationGradient();
        if (_checkWithNormal){
          printf("solver is broken at ele %d gpt %d at rank %d  real tolearance = %e, following prescribed normal direction [%e %e %e] \n",
                _enum,_gnum,Msg::GetCommRank(),cr,_lostSolutionUniquenssNormal[0],_lostSolutionUniquenssNormal[1],_lostSolutionUniquenssNormal[2]);
        }
        else{
          printf("solver is broken at ele %d gpt %d at rank %d  real tolearance = %e, with minimal normal direction [%e %e %e] \n",
                _enum,_gnum,Msg::GetCommRank(),cr, _lostSolutionUniquenssNormal[0],_lostSolutionUniquenssNormal[1],_lostSolutionUniquenssNormal[2]);
        }
      }
      else {
        _solverIsBroken = false;
      }
    }
	}
};

void nonLinearMechSolver::computeHomogenizedCrackFace(){
	if (_pAl == NULL){
		Msg::Error("pbcAlgorithm object must be created");
	}

  SVector3 Lx, Ly, Lz;
  _pAl->getPBCConstraintGroup()->getPeriodicity(Lx, Ly, Lz);

	if (_lostSolutionUniquenssNormal.norm() < 1e-6){
		// normally, _lostSolutionUniquenssNormal.norm() == 1
		_homogenizedCrackSurface = std::max(Lx.norm(),Ly.norm()); // nothing to do
		return;
	}

	SPoint3 v1 = _pAl->getPBCConstraintGroup()->getRootPoint();
	Msg::Info("rootPoint %f %f %f ",v1[0],v1[1],v1[2]);

  SPoint3 damageVolumeCenter(0.,0.,0.);
  double damageVolume = 0.;
  for (int i=0; i<domainVector.size(); i++){
    damageVolume += domainVector[i]->computeVolumeActiveDissipationDomain(_ipf);
  }
  if (damageVolume > 0.){
    for (int i=0; i<domainVector.size(); i++){
      domainVector[i]->computeActiveDissipationCenter(_ipf,damageVolumeCenter);
    }
    damageVolumeCenter *= (1./damageVolume);
    Msg::Info("center of active damage zone : %f %f %f \n",damageVolumeCenter[0],damageVolumeCenter[1],damageVolumeCenter[2]);
  }


	_homogenizedCrackSurface = 0.;

	if (getDim() == 2){
		SPoint3 v2(v1);
		v2 += (Lx.point());
		SPoint3 v4(v1);
		v4 += (Ly.point());
		SPoint3 v3(v1);
		v3 += (Lx.point());
		v3 += (Ly.point());

		if (dot(_lostSolutionUniquenssNormal,Lx)/(Lx.norm()) < 1e-8){
			_homogenizedCrackSurface = Lx.norm();
			Msg::Info("_homogenizedCrackSurface following Lx = %f",_homogenizedCrackSurface);
		}
		else if (dot(_lostSolutionUniquenssNormal,Ly)/(Ly.norm()) < 1e-8){
			_homogenizedCrackSurface = Ly.norm();
			Msg::Info("_homogenizedCrackSurface following Ly = %f",_homogenizedCrackSurface);
		}
		else{
			std::vector<SPoint3> twoPoint;
			SPoint3 prCenterTo12 = planeDirProject(damageVolumeCenter,v1,v2,_lostSolutionUniquenssNormal);
			if (inside(prCenterTo12,v1,v2)){
				twoPoint.push_back(prCenterTo12);
			}
			SPoint3 prCenterTo41 = planeDirProject(damageVolumeCenter,v1,v4,_lostSolutionUniquenssNormal);
			if (inside(prCenterTo41,v1,v4)){
				twoPoint.push_back(prCenterTo41);
			}
			if (twoPoint.size() < 2){
			SPoint3 prCenterTo23 = planeDirProject(damageVolumeCenter,v2,v3,_lostSolutionUniquenssNormal);
				if (inside(prCenterTo23,v2,v3)){
					twoPoint.push_back(prCenterTo23);
				}
			}
			if (twoPoint.size() < 2){
				SPoint3 prCenterTo34 = planeDirProject(damageVolumeCenter,v3,v4,_lostSolutionUniquenssNormal);
				if (inside(prCenterTo34,v3,v4)){
					twoPoint.push_back(prCenterTo34);
				}
			}

			if (twoPoint.size() == 2){
				_homogenizedCrackSurface = twoPoint[0].distance(twoPoint[1]);
				Msg::Info("_homogenizedCrackSurface = %f ",_homogenizedCrackSurface);
			}
			else{
				Msg::Error("computeHomogenizedCrackLength is wrong ele %d, gp %d _lostSolutionUniquenssNormal = [%f %f %f]",_enum,_gnum,
								_lostSolutionUniquenssNormal[0],_lostSolutionUniquenssNormal[1],_lostSolutionUniquenssNormal[2]);
			}

		}
		_homogenizedCrackSurface *= (Lz.norm()); // generally, _LzNorm = 1

		#ifdef _DEBUG
		printf("on ele %d gp %d _homogenizedCrackSurface = %e \n",_enum,_gnum,_homogenizedCrackSurface);
		#endif //_DEBUG

	}
	else if (getDim() == 3){
		Msg::Info("nonLinearMechSolver::computeHomogenizedCrackFace for 3D has not been implemented");
	}
	else{
		Msg::Info("computeHomogenizedCrackLength has not been implemented for %d D problems");

	}
};


void nonLinearMechSolver::setMessageView(const bool view){
  _messageView = view;
}

void nonLinearMechSolver::extractAverageProperties(bool stiff){
  homogenizedData* homoData = this->getHomogenizationState(IPStateBase::current);
  /* get homogenization value */
  if (stiff){
    this->extractAverageStressAndTangent(homoData);
  }
  else{
    this->extractAverageStress(homoData);
  }

};

void nonLinearMechSolver::createEigenvalueSolver(){
	if (_eigflag){
	  if (_eigenSolver) delete _eigenSolver;
	  std::string strA = "A";
	  linearSystem<double>* lsysA = pAssembler->getLinearSystem(strA);
	  linearSystemPETSc<double>* lpetA = dynamic_cast<linearSystemPETSc<double>*>(lsysA);

      std::string strB = "B";
	  linearSystem<double>* lsysB = pAssembler->getLinearSystem(strB);
	  linearSystemPETSc<double>* lpetB = dynamic_cast<linearSystemPETSc<double>*>(lsysB);

		if (lpetA != NULL ){
			_eigenSolver = new eigenSolver(lpetA,lpetB,_hermitian);
      // for eigen mode view
			if (_eigfield.size() > 0){
				for (int i= 0; i< _eigfield.size(); i++){
					delete _eigfield[i];
				}
				_eigfield.clear();
			}
      for (int i=0; i<_modeview.size(); i++){
        std::string modefilename = "E_"+int2str(_enum)+"_GP_"+int2str(_gnum)+"_mode";
        unknownField* field = new eigenVectorField(this,3,anoded,_eigview[i],_eigenSolver,_modeview[i],modefilename);
        _eigfield.push_back(field);
      };
      std::string eigfilename = "E_"+int2str(_enum)+"_GP_"+int2str(_gnum)+"_eigenValue.csv";
        _eigenValueFile = fopen(eigfilename.c_str(),"w");
	  }
		else{
				Msg::Error("eigensolver requires PETSc");
		}
  };
};

void nonLinearMechSolver::computeSplittedStiffnessMatrix(){
	_pAl->getSplitStiffness()->zeroSubMatrix();
	for(std::vector<partDomain*>::iterator itdom=domainVector.begin(); itdom!=domainVector.end(); ++itdom){
    partDomain* dom = *itdom;
    AssembleSplittedStiffness(*(dom->getBilinearBulkTerm()),*(dom->getFunctionSpace()),dom->g_cbegin(),dom->g_cend(),
             *(dom->getBulkGaussIntegrationRule()),*_pAl->getSplittedDof());

    if(dom->IsInterfaceTerms()){
      dgPartDomain *dgdom = static_cast<dgPartDomain*>(dom);
      // Assembling loop on elementary interface terms
      AssembleSplittedStiffness(*(dgdom->getBilinearInterfaceTerm()),*(dom->getFunctionSpace()),dgdom->gi->begin(),dgdom->gi->end(),
               *(dgdom->getInterfaceGaussIntegrationRule()),*_pAl->getSplittedDof());
      // Assembling loop on elementary boundary interface terms
      AssembleSplittedStiffness(*(dgdom->getBilinearVirtualInterfaceTerm()),*(dom->getFunctionSpace()),dgdom->gib->begin(),dgdom->gib->end(),
               *(dgdom->getInterfaceGaussIntegrationRule()),*_pAl->getSplittedDof());
    }
  }

  for(contactContainer::iterator it = _allContact.begin(); it!=_allContact.end(); ++it){
    contactDomain *cdom = *it;
    contactBilinearTermBase<double> *sterm = static_cast<contactBilinearTermBase<double>*>(cdom->getStiffnessTerm());
    AssembleSplittedStiffness(*sterm,*(cdom->getSpace()),sterm->getContactNodes()->elemBegin(),
                    sterm->getContactNodes()->elemEnd(),*(cdom->getGaussIntegration()),*_pAl->getSplittedDof());
  }
  for(defoDefoContactContainer::iterator it = _allDefoDefoContact.begin(); it!=_allDefoDefoContact.end(); ++it){
      defoDefoContactDomain *cdom = *it;
      const std::vector<BilinearTermBase*> &sterm = static_cast< const std::vector<BilinearTermBase* > >(cdom->getConstRefToStiffnessTerm()); //we might need to change the class
      for(int i=0; i<sterm.size(); i++)
      {
        AssembleSplittedStiffnessDefoDefoContact(*(sterm[i]),*(cdom->getRefToVSpace()[i]),((cdom->getRefToGroup())[i]).begin(),
            ((cdom->getRefToGroup())[i]).end(),
            *(cdom->getRefToVQuadratureMaster()[i]),*_pAl->getSplittedDof());
      }
  }
};

homogenizedData* nonLinearMechSolver::getHomogenizationState(const IPStateBase::whichState state){
  if (state == IPStateBase::initial or state == IPStateBase::previous){
    return _initialState;
  }
  else if (state == IPStateBase::current){
    return _currentState;
  }
  else{
    Msg::Error("This state does not exist");
    return NULL;
  }
}

const homogenizedData* nonLinearMechSolver::getHomogenizationState(const IPStateBase::whichState state) const{
  if (state == IPStateBase::initial or state == IPStateBase::previous){
    return _initialState;
  }
  else if (state == IPStateBase::current){
    return _currentState;
  }
  else{
    Msg::Error("This state does not exist");
    return NULL;
  }
}


void nonLinearMechSolver::resetSolverToInitialStep(){
  // system
  std::string name = "A";
  linearSystem<double>* lsys = pAssembler->getLinearSystem(name);
  nonLinearSystem<double>* nonlsys = dynamic_cast<nonLinearSystem<double>*>(lsys);
  if (nonlsys){
    nonlsys->copy(IPStateBase::initial,IPStateBase::current);
    nonlsys->copy(IPStateBase::initial,IPStateBase::previous);
  }


  pathFollowingSystem<double>* pathSys = dynamic_cast<pathFollowingSystem<double>*>(nonlsys);
  if (pathSys){
    pathSys->resetControlParameter();
  }

  // ipf field
  _ipf->copy(IPStateBase::initial,IPStateBase::current);
  _ipf->copy(IPStateBase::initial,IPStateBase::previous);

  // homogenized data
  (*_currentState) = (*_initialState);

  if (_damageIsBlocked){
    _ipf->blockDissipation(IPStateBase::current,true);
  }
  else{
    _ipf->blockDissipation(IPStateBase::current,false);
  }

};

void nonLinearMechSolver::closeAllFile(){
  if (_outputFile != NULL) fclose(_outputFile);

  if (_eigenValueFile != NULL) fclose(_eigenValueFile);

	if (_homogenizedFiles!=NULL){
		_homogenizedFiles->closeFiles();
		delete _homogenizedFiles;
		_homogenizedFiles= NULL;
	}

  if (_elasticDPDFTangentFile) {fclose(_elasticDPDFTangentFile); _elasticDPDFTangentFile = NULL;};
  if (_initialState) {delete _initialState;}
  if (_currentState) {delete _currentState;}
  if (_elasticState) {delete _elasticState;};

  if (_criterionFile != NULL) fclose(_criterionFile);
};

void nonLinearMechSolver::nextStep(){
  // in microBC
	bool BCIsChanged = false;
	if (_failureBCIsSwitched){
		if (_microFailureBC == NULL) {
			Msg::Error("_microFailureBC is null in nonLinearMechSolver::nextStep");
		}
		else{
			const homogenizedData* homodataPrev = this->getHomogenizationState(IPStateBase::initial);
			const homogenizedData* homodata = this->getHomogenizationState(IPStateBase::current);
			if (_solverIsBroken){
				this->switchMicroBC(_microFailureBC);
				printf("ranks %d microBC has just been changed \n",Msg::GetCommRank());
				BCIsChanged = true;
        _failureBCIsSwitched = false; // pass one time
			}
		}
	}

	if (!BCIsChanged){
		_microBC->nextStep();
	}

  // archive data
  this->archiveData(_macroTime,_macroStep);

  // next step for solver
  std::string name = "A";
  linearSystem<double>* lsys = pAssembler->getLinearSystem(name);
  nonLinearSystem<double>* nonlsys = dynamic_cast<nonLinearSystem<double>*>(lsys);
  pathFollowingSystem<double>* pathSys = dynamic_cast<pathFollowingSystem<double>*>(nonlsys);
  if (nonlsys){
    nonlsys->copy(IPStateBase::current,IPStateBase::initial);
  }

  if (pathSys){
    pathSys->resetControlParameter();
  }

  if (_damageIsBlocked){
    _ipf->blockDissipation(IPStateBase::current,true);
  }
  else{
    _ipf->blockDissipation(IPStateBase::current,false);
  }


	if (withEnergyDissipation()){
    _ipf->checkActiveDissipation();
		if (_ipf->getNumOfActiveDissipationIPsCurrent() > 0){
			// store historical value
			_homogenizedStressLastActiveDissipation = this->getHomogenizationState(IPStateBase::current)->getHomogenizedStress();
			_homogenizedStrainLastActiveDissipation = this->getMicroBC()->getFirstOrderKinematicalVariable();
      _homogenizedDissipationStrainLastActiveDissipation = this->getHomogenizationState(IPStateBase::current)->getHomogenizedActiveDissipationDeformationGradient();
			_homogenizedCohesiveJumpLastActiveDissipation = this->getHomogenizationState(IPStateBase::current)->getHomogenizedCohesiveJump();
		}
	}

	 // save data for nex macroscopic step
  _ipf->copy(IPStateBase::current,IPStateBase::initial);

  // for homogenized data
  (*_initialState)=(*_currentState);
};

double nonLinearMechSolver::critialPointTestFunction(double time, double* &mode){
  if (_eigflag){
    if (_eigenSolver == NULL) createEigenvalueSolver();
    _eigenSolver->clear();
    int nsize = pAssembler->sizeOfR();
    if (_numeigenvalue >= nsize) {
      _numeigenvalue = nsize;
      if (_outputFile)
      fprintf(_outputFile,"Modify number of eigenvalues computed: %d \n",nsize);
    }
    // solver eigen solver
    //_eigenSolver->solve(_numeigenvalue,"smallestReal");
    _eigenSolver->solve(_numeigenvalue,"smallestReal",_eigenSolverMethod,_eigenSolverConvergenCriterion,_eigenSolverNumIteration);

    std::complex<double> ls = _eigenSolver->getEigenValue(0);
    std::complex<double> lm = _eigenSolver->getEigenValue(_numeigenvalue-1);

    fprintf(_eigenValueFile,"%.16g\t",time);
    for (int i=0; i<_numeigenvalue; i++){
      std::complex<double> val = _eigenSolver->getEigenValue(i);
        fprintf(_eigenValueFile,"%.16g  \t %.16g \t",val.real(), val.imag());
    }
    fprintf(_eigenValueFile,"\n");
    fflush(_eigenValueFile);

    mode = new double[nsize];
    for (int i=0; i<nsize; i++){
      std::complex<double>  s = _eigenSolver->getEigenVectorComp(0,i);
        mode[i] = s.real();
    }

    return ls.real();
  }
  else
  {
    return 1.e10;
  }
};

linearSystem<double>* nonLinearMechSolver::createMicroSNLSystem(){
  linearSystem<double>* lsys = NULL;
  if (_pathFollowing){
    _controlType = nonLinearMechSolver::LOAD_CONTROL; // this means compute until time = 1.
    _systemType == nonLinearMechSolver::MULT_ELIM; // always multiplier elimination
    #if defined(HAVE_PETSC)
    Msg::Info("Path following is used");
    lsys = new pbcPathFollowingSystemPETSC<double>(PETSC_COMM_SELF);

		// set path following options
		// paramater remains default if it is negative
		pathFollowingSystemBase* pfsysBase = dynamic_cast<pathFollowingSystemBase*>(lsys);
		if (_macroControlTypePathFollowing >=0){
			pfsysBase->setControlType(_macroControlTypePathFollowing);
		}
		if (_correctionMethodPathFollowing>=0){
			pfsysBase->setCorrectionMethod(_correctionMethodPathFollowing);
		}
		if (_tranversalCriterionPathFollowing>=0){
			pfsysBase->setTranversalCriterion(_tranversalCriterionPathFollowing);
		}
		if (_solverTypePathFollowing >=0){
			pfsysBase->setSolverType(_solverTypePathFollowing,_pathFollowingEqRatio);
		}

    #else
    Msg::Error("This option used with petsc only");
    #endif
  }
  else{
    if (_controlType == nonLinearMechSolver::LOAD_CONTROL){
      if (_systemType == nonLinearMechSolver::MULT_ELIM){
        if(nonLinearMechSolver::whatSolver == Petsc){
          #if defined(HAVE_PETSC)
          lsys = new pbcNonLinearSystemPETSc<double>(PETSC_COMM_SELF);
          (dynamic_cast<pbcNonLinearSystemPETSc<double>*>(lsys))->invertFlag(_useMatMatSolveToInvertMatrix);
          if(_solver_options.size())
            lsys->setParameter(std::string("petsc_solver_options"),_solver_options);
          if (_outputFile)
            fprintf(_outputFile,"PETSc is chosen to solve \n");
          #endif
        }
        else{
          Msg::Error("Petsc is required");
        }
      }
      else if (_systemType == nonLinearMechSolver::DISP_ELIM_UNIFIED){
        if(nonLinearMechSolver::whatSolver == Petsc){
          #if defined(HAVE_PETSC)
          lsys = new pbcConstraintEliminationNonLinearSystemPETSc<double>(PETSC_COMM_SELF);
          (dynamic_cast<pbcConstraintEliminationNonLinearSystemPETSc<double>*>(lsys))->invertFlag(_useMatMatSolveToInvertMatrix);
          if(_solver_options.size())
            lsys->setParameter(std::string("petsc_solver_options"),_solver_options);
          if (_outputFile)
            fprintf(_outputFile,"PETSc is chosen to solve \n");
          #endif
        }
        else{
          Msg::Error("Petsc is required");
        }

      }
      else{
        if(nonLinearMechSolver::whatSolver == Petsc){
          #if defined(HAVE_PETSC)
          lsys = new nonLinearSystemPETSc<double>(PETSC_COMM_SELF,_lineSearch);
          if(_solver_options.size())
            lsys->setParameter(std::string("petsc_solver_options"),_solver_options);
          if (_outputFile)
            fprintf(_outputFile,"PETSc is chosen to solve \n");
          #else
          lsys = new nonLinearSystemGmm<double>(_lineSearch);
          dynamic_cast<nonLinearSystemGmm<double>*>(lsys)->setNoisy(2);
          if (_outputFile)
            fprintf(_outputFile,"PETSc is not installed\n Gmm is chosen to solve \n");
          #endif
        }
        else
          lsys= nonLinearMechSolver::createSNLSystem();
      }

    }
    else if (_controlType == nonLinearMechSolver::ARC_CONTROL_EULER){
      if (nonLinearMechSolver::whatSolver == Petsc){
        #if defined(HAVE_PETSC)
        lsys = new forwardEulerLinearSystemPETSc<double>();
        if(_solver_options.size())
          lsys->setParameter(std::string("petsc_solver_options"),_solver_options);
        if (_outputFile)
          fprintf(_outputFile,"PETSc is chosen to solve \n");
        #else
        Msg::Error("Petsc is not available");
        #endif
      }
      else{
        Msg::Error("This is not implemented without PETSc");
      };
    };
  }

  if (_outputFile)
    fflush(_outputFile);

  return lsys;
};

void nonLinearMechSolver::createMicroSystem(){
	linearSystem<double>* lsys = createMicroSNLSystem();
	dofManager<double>* oldManager = pAssembler; // store old manager

	// create a new one
	if (_controlType == ARC_CONTROL_EULER or whatScheme == StaticLinear){
		pAssembler = new dofManager<double>(lsys,false);
	}
	else{
		pAssembler = new staticDofManager<double>(lsys,whatScheme,false,false);
	}

	// transfer data if needed
	if (oldManager != NULL){
		staticDofManager<double>* oldSmanager = dynamic_cast<staticDofManager<double>*>(oldManager);
		if (oldSmanager != NULL){
			staticDofManager<double>* sAssembler = dynamic_cast<staticDofManager<double>*>(pAssembler);
			if (sAssembler != NULL){
				sAssembler->collecDataFromPrevious(oldSmanager);
			}
			delete oldSmanager;
		}
	}
};

void nonLinearMechSolver::init2Micro(){
	// get quantities for micro BC
	if (_microBC == NULL){
		Msg::Error("Microscopic BC must be specified nonLinearMechSolver::init2Micro");
	}

  // create system, dofManager
  this->createMicroSystem();

	// init all other BC
  this->initAllBCsOnDofs();

  // create periodic algorithm
  if (_pAl != NULL){
		delete _pAl; Msg::Info("pbcAlgorithm is deleted");
	}
	_pAl = new pbcAlgorithm(this); // create all constraints, etc

  // number Dof and allocate the system
  this->microNumberDof();

	if (_previousInit){
    staticDofManager<double>* sAssembler = static_cast<staticDofManager<double>*>(pAssembler);
    sAssembler->updateFromPrevious();
  }

  /* initialization of Dof value */
  this->setInitialCondition();

	if (!_archive){
		unknownView.clear();
		ipView.clear();
		energyView.clear();
	}

	if (_ufield) delete _ufield;
	std::string dispfile ="E_"+int2str(_enum)+"_GP_"+int2str(_gnum)+"_disp";
	_ufield = new unknownField(this,3,anoded,unknownView,dispfile); // 3 components by nodes User choice ??

	if (!_previousInit){
		if (_ipf) delete _ipf;
		std::string ipfile = "E_"+int2str(_enum)+"_GP_"+int2str(_gnum)+"_stress";
		_ipf = new IPField(this,vaip,ipView,ipfile); // Field for GaussPoint

		_ipf->compute1state(IPStateBase::current,true);
		_ipf->copy(IPStateBase::current,IPStateBase::previous);
		_ipf->copy(IPStateBase::current,IPStateBase::initial);

		_ipf->initialBroken(pModel, initbrokeninter);
		// initially broken on whole domain
		for (int idom=0; idom< domainVector.size(); idom++){
			dgPartDomain* dgdom = dynamic_cast<dgPartDomain*>(domainVector[idom]);
			if (dgdom != NULL) {
				if (dgdom->allInterfaceIPsAreBrokenFromBeginning()){
					dgdom->initiallyBreakAllInterfaceIP(_ipf->getAips());
				}

			}
		}
	}
	else{
    _ipf->copy(IPStateBase::current,IPStateBase::initial);
    _ipf->copy(IPStateBase::current,IPStateBase::previous);
	}

	if (_energField) delete _energField;
	_energyComputation = 0;
	_fractureEnergyComputation = 0;
	_energField = new energeticField(this,energyView,_energyComputation,_fractureEnergyComputation,getScheme());

	if (_eigflag){
		if (_outputFile)
			fprintf(_outputFile,"Create EigenSolver \n");
		createEigenvalueSolver();
		if (_outputFile)
			fprintf(_outputFile,"EigenSolver created \n");
	};

  // init terms
  this->initTerms();

  if (_tangentflag and _homogenizeTangentMethod == INSYSTEMCONDEN){
    this->computeStiffMatrix();
  }

  if (_stabilityCriterionFollowing == true){
    std::string filename = "E_"+int2str(_enum)+"_GP_"+int2str(_gnum)+"_stabilityCriterion.csv";
		if (_criterionFile != NULL) fclose(_criterionFile);
    _criterionFile = fopen(filename.c_str(),"w");
  }
  // rve_volume from periodicity
	if (!_previousInit){
    if (_rveVolume <=0.){
      _rveVolume = _pAl->getRVEVolume();
    }
    Msg::Info("RVE volume used = %e",_rveVolume);
		// compute density
		this->computeDensity();
		if (_isHommProSaveToFile){
			std::string denfile = "E_"+int2str(_enum)+"_GP_"+int2str(_gnum)+"_density.csv";
			FILE* fileden = fopen(denfile.c_str(),"w");
			fprintf(fileden,"%e",this->_rho);
			fclose(fileden);
		}


		_initialState = new homogenizedData(this,_damageToCohesiveJump,_extractIrreversibleEnergy);
		_currentState = new homogenizedData(this,_damageToCohesiveJump,_extractIrreversibleEnergy);

    if (_extractElasticTangentOperator){
      _elasticState = new homogenizedData(this,_damageToCohesiveJump,_extractIrreversibleEnergy);
      std::string elasticFileName = "E_"+int2str(_enum)+"_GP_"+int2str(_gnum)+"_elasticTangent.csv";
      _elasticDPDFTangentFile = Tensor43::createFile(elasticFileName);
    }

    /** initial tangent**/

		this->extractAverageProperties(_tangentflag);


		(*_initialState) = (*_currentState);

		if (_multiscaleFlag == false){
			if (_microBC->getTotalNumberOfMechanicalDofs() > 0) {
				_currentState->getHomogenizedStress().print("stress");
				if (_tangentflag){
					_currentState->getHomogenizedTangentOperator_F_F().print("tangent");
				}
				if (_microBC->getOrder() == 2)
					_currentState->getHomogenizedSecondOrderStress().print("ho stress");
			}

			for (int index=0; index< _microBC->getTotalNumberOfConDofs(); index++){
				_currentState->getHomogenizedConstitutiveExtraDofFlux(index).print("ConExtraDof flux");
				printf("_currentState constitutive extra dof %d internal energy: %f \n", index, _currentState->getHomogenizedConstitutiveExtraDofInternalEnergy(index));
			}

			for (int index=0; index< _microBC->getTotalNumberOfNonConDofs(); index++){
				_currentState->getHomogenizedNonConstitutiveExtraDofFlux(index).print("nonConExtraDof flux");
			}
		}

		_homogenizedFiles = new homogenizedDataFiles(this);
		bool saveCohesiveLawFile = (_isHommProSaveToFile and _damageToCohesiveJump);
		_homogenizedFiles->openFiles(_enum,_gnum,this,_isHommStrainSaveToFile,_isHommProSaveToFile,saveCohesiveLawFile);
		this->homogenizedDataToFile(0.);

	}
	else{
		// extract with new BC
		this->extractAverageProperties(_tangentflag);
		if (_multiscaleFlag == false){
			if (_microBC->getTotalNumberOfMechanicalDofs() > 0) {
				_currentState->getHomogenizedStress().print("stress");
				if (_tangentflag){
					_currentState->getHomogenizedTangentOperator_F_F().print("tangent");
				}
				if (_microBC->getOrder() == 2)
					_currentState->getHomogenizedSecondOrderStress().print("ho stress");
			}

			for (int index=0; index< _microBC->getTotalNumberOfConDofs(); index++){
				_currentState->getHomogenizedConstitutiveExtraDofFlux(index).print("ConExtraDof flux");
				printf("_currentState constitutive extra dof %d internal energy: %f \n", index, _currentState->getHomogenizedConstitutiveExtraDofInternalEnergy(index));
			}

			for (int index=0; index< _microBC->getTotalNumberOfNonConDofs(); index++){
				_currentState->getHomogenizedNonConstitutiveExtraDofFlux(index).print("nonConExtraDof flux");
			}
		}

		bool saveCohesiveLawFile = (_isHommProSaveToFile and _damageToCohesiveJump);
		_homogenizedFiles->openFiles(_enum,_gnum,this,_isHommStrainSaveToFile,_isHommProSaveToFile,saveCohesiveLawFile);
		this->homogenizedDataToFile(0.);
	}

	_previousInit = true;

};

void nonLinearMechSolver::initMicroSolver(){
  if (_messageView){
    std::string filename = "E_"+int2str(_enum)+"_GP_"+int2str(_gnum)+"_out.txt";
		if (_outputFile != NULL) fclose(_outputFile);
    _outputFile = fopen(filename.c_str(),"w");

    fprintf(_outputFile,"Initializing on processor %d of %d \nMicro-problem: element %d, gauss point %d \nMicro SNL Data : nstep =%d  endtime = %f\n",
        Msg::GetCommRank(),Msg::GetCommSize(),this->getElementSolverNumber(),this->getGaussPointSolverNumber(),
        _numstepImpl,endtime);
  }
  // read mesh file
  this->init();
	this->init2Micro();

	if (_outputFile) {
		fprintf(_outputFile,"---------End of initialization-------- \n");
		fflush(_outputFile);
	}
};

double nonLinearMechSolver::microSolve(){
  double time = 0;
  this->resetSolverToInitialStep();
  if (_microBC->notSameState(_sameStateCriterion,_absTol)){
    if (_outputFile)
      fprintf(_outputFile,"Solving in procs %d \n",Msg::GetCommRank());
    // solve system
    if (whatScheme ==StaticLinear){
      time = solveMicroSolverStaticLinear();
    }
    else if (whatScheme == StaticNonLinear){
      time = solveMicroSolverSNL();
    }
    else
      Msg::Error("This scheme is not implemented");

    this->extractAverageProperties(_tangentflag);

    if (_outputFile)
      fflush(_outputFile);
  }
  return time;
};

void nonLinearMechSolver::OneStep(const int numstep){
  int niteNR = 0;
  if (_pathFollowing){
    niteNR = pathFollowingPerturbation();
  }
  else{
    if (_systemType == MULT_ELIM)
      niteNR = microNewtonRaphson(numstep);
    else {
      if (_systemType == DISP_ELIM)
        _pAl->applyPBCByConstraintElimination();
       else if (_systemType == DISP_ELIM_UNIFIED){
         _pAl->updateSystemUnknown();
       }
      niteNR = NewtonRaphson(numstep);
    }
    if(niteNR == _maxNRite) // time step reduction
    {
      Msg::Error("convergent fails in tangent perturbation ele %d GP %d",_enum,_gnum);
    }
  }

};

double nonLinearMechSolver::solveMicroSolverSNL(){
  std::string name = "A";
  linearSystem<double>* lsys =pAssembler->getLinearSystem(name);
  nonLinearSystem<double>* nonsys = dynamic_cast<nonLinearSystem<double>*>(lsys);
  pathFollowingSystem<double>* pathSys = dynamic_cast<pathFollowingSystem<double>*>(lsys);
  staticDofManager<double> *sAssembler = dynamic_cast<staticDofManager<double>*>(pAssembler);

  // change to introduce a dynamic modification of time step in case of convergence problem
  double curtime = 0.;
  double lasttime = 0.; // to avoid precision problem curtime = lasttime in place of curtime -= dt if no convergence
  double dendtime = double(endtime);
  double dtmax = dendtime/double(_numstepImpl); // time step value prescribed by user (time step can not be greater than this value)
  double dt = dtmax; // initial time step value
  int ii =0.; // iteration counter
  double tolend = 0.000001*dendtime; // OTHERWISE LAST STEP CAN BE DONE TWICE ??
  double toldtmax = 0.000001*dtmax; // IDEM (PRECISION PROBLEM)
  int numberTimeStepReduction=0;
  bool isSuccess = true;

  if (_pathFollowing){
    _pAl->computeLoadVector();
  }

  bool willFinish = false;

  while(fabs(curtime-dendtime) > tolend ){
    curtime+=dt;
    if(curtime > dendtime)
    {
      dt -= (curtime -dendtime);
      curtime = dendtime;
    }
    if (_outputFile)
      fprintf(_outputFile,"MICRO t= %e on %e \n",curtime,dendtime);

    // switch next step
    if (isSuccess){
      if (sAssembler)
        sAssembler->nextStep();
      _ipf->nextStep();
    }

    if (sAssembler)
      sAssembler->setTimeStep(dt);
    // Solve one step by NR scheme
    int niteNR = 0;
    if (_pathFollowing){
      pathSys->setPseudoTimeIncrement(dt);
      _pAl->updateConstraint(_ipf);
      niteNR = microNewtonRaphsonPathFollowing(ii+1);

      double control = pathSys->getControlParameter();
      if (_outputFile)
        fprintf(_outputFile,"control variable %e \n",control);
      if (control>1. and fabs(control-1) >tolend){
        if (_outputFile)
          fprintf(_outputFile,"Load correction step\n");
        willFinish = true;

        nonsys->resetUnknownsToPreviousTimeStep();
        _ipf->copy(IPStateBase::previous,IPStateBase::current);

        pathSys->setControlType(2);
        curtime = lasttime;
        dt = 1-pathSys->getControlParameter();
        dtmax = dt;
        isSuccess = false;
        continue;
      }
      else if (fabs(control-1) <=tolend) {
        willFinish = true;
      }
      else if (control<0){
        Msg::Error("path following is not converge e= %d, gp = %d!",_enum,_gnum);
      }
    }
    else{
      this->oneStepPreSolve(curtime,dt,ii+1);
      _pAl->updateConstraint(_ipf);
      if (_systemType == MULT_ELIM)
        niteNR = microNewtonRaphson(ii+1);
      else {
        if (_systemType == DISP_ELIM){
          _pAl->applyPBCByConstraintElimination();
        }
        else if (_systemType == DISP_ELIM_UNIFIED){
          _pAl->updateSystemUnknown();
        }

        niteNR = NewtonRaphson(ii+1);
      }
    }

    if(niteNR == _maxNRite) // time step reduction
    {
       isSuccess = false;
       numberTimeStepReduction ++;
       if(numberTimeStepReduction > _maxAttemptStepReduction) // end of simulation
       {
         if (_outputFile)
          fprintf(_outputFile,"Simulation end due to convergence problem  element %d, IP %d !!!\n",
                  getElementSolverNumber(),getGaussPointSolverNumber());
         Msg::Error("Simulation end due to convergence problem  element %d, IP %d !!!",
                  getElementSolverNumber(),getGaussPointSolverNumber());
         break;
       }
       if (_outputFile)
        fprintf(_outputFile,"Convergence of Newton-Raphson failed at element %d, IP %d --> reduced time step",
                    getElementSolverNumber(),getGaussPointSolverNumber());
       Msg::Warning("Convergence of Newton-Raphson failed at element %d, IP %d --> reduced time step",
                    getElementSolverNumber(),getGaussPointSolverNumber());
       curtime = lasttime;
       dt /= _timeStepFactorReduction;

       nonsys->resetUnknownsToPreviousTimeStep();
       _ipf->copy(IPStateBase::previous,IPStateBase::current);
    }
    else
    {
      isSuccess = true;
      // save solution of previous as sucessfull convergence of NR process
      if(fabs(dtmax-dt) > toldtmax ) // increase time step if convergence in a few iteration
      {
        if(niteNR <= _niteTimeStepIncrease)
        {
           if (_outputFile)
            fprintf(_outputFile,"Time step is increased due to convergence of Newton-Raphson in less than %d at element %d, IP %d ",_niteTimeStepIncrease,
                        getElementSolverNumber(),getGaussPointSolverNumber());
           Msg::Warning("Time step is increased due to convergence of Newton-Raphson in less than %d at element %d, IP %d ",_niteTimeStepIncrease,
                        getElementSolverNumber(),getGaussPointSolverNumber());
           numberTimeStepReduction --;
           dt *= _timeStepFactorReduction;
        }
      }
      // next step

      if (_eigenFollowing){
        double* mode = NULL;
        double pre = this->critialPointTestFunction(curtime,mode);
        printf("Buckling criterion = %e\n",pre);
        if (_outputFile)
          fprintf(_outputFile,"Critial Point test function = %f \n",pre);
        this->modeViewToFile(_modeview,pre,ii+1);
        if ((pre < _instabilitycriterion) and _perturbate){
          this->perturbationProcedure(pre,mode);
        }
        if (mode) delete mode;
      }

      lasttime = curtime;
      if (!_multiscaleFlag){
				if (_outputFile)
					fflush(_outputFile);
        this->microPostSolveStep(curtime,ii+1);
      }
    }
    ii ++; // increase counter of iteration even if no convergence --> more archivage if no convergence (can be usefull)
    if(_pathFollowing){
      if (fabs(curtime-dendtime) < tolend){
        double control = pathSys->getControlParameter();
        if (control<1.){
          dendtime *=2.;
        }
      }
      if (willFinish) {
        dendtime = 1.;
        pathSys->setControlType(_macroControlTypePathFollowing);
        break;
      }
    }
  }
  _iterMax = ii;
  return curtime;
};

void nonLinearMechSolver::microPostSolveStep(const double curtime, const int numstep){
	Msg::Info("Extract macro properties");
  // echeck erosion
  this->checkElementErosion(IPStateBase::current);
  if (_bulkElementErosionFlag or _interfaceElementErosionFlag)
  {
    _elementErosionFilter.print();
  }

	// copy current to initial state
	(*_initialState) = (*_currentState);

	this->extractAverageProperties(_tangentflag);
  
  this->IPVolumeIntegralArchiving(curtime,numstep,false);
  this->IPDataOnPhysicalArchiving(curtime,numstep,false);
  

	if (getDamageToCohesiveJumpFlag() and !_multiscaleFlag){
		this->checkFailureOnset();
	}

	this->pathFollowingArchiving(curtime,numstep);

	if (_microBC->getTotalNumberOfMechanicalDofs() > 0){
		if (_stressflag){
			_currentState->getHomogenizedStress().print("Stress ");
			if (_microBC->getOrder() == 2)
				_currentState->getHomogenizedSecondOrderStress().print("hostress ");
		}
	}

	for (int index=0; index < _microBC->getTotalNumberOfConDofs(); index++){
		_currentState->getHomogenizedConstitutiveExtraDofFlux(index).print("con extra dof flux");
		 printf("double con extra dof %d internal energy %f\n",index,_currentState->getHomogenizedConstitutiveExtraDofInternalEnergy(index));
	}

	if (_tangentflag){
		 _currentState->getHomogenizedTangentOperator_F_F().print("tangent");
	}

	if (_extractPerturbationToFile){
		_pAl->getPBCConstraintGroup()->writePertBoundaryToFile(pAssembler,_enum,_gnum,numstep);
	}

	if (_isWriteDeformedMeshToFile){
		this->writeDeformedMesh(numstep);
	}
	this->archiveData(curtime,numstep);

	if (withEnergyDissipation()){
    _ipf->checkActiveDissipation();
		if (_ipf->getNumOfActiveDissipationIPsCurrent() > 0){
			// store historical value
			_homogenizedStressLastActiveDissipation = this->getHomogenizationState(IPStateBase::current)->getHomogenizedStress();
			_homogenizedStrainLastActiveDissipation = this->getMicroBC()->getFirstOrderKinematicalVariable();
      _homogenizedDissipationStrainLastActiveDissipation = this->getHomogenizationState(IPStateBase::current)->getHomogenizedActiveDissipationDeformationGradient();
			_homogenizedCohesiveJumpLastActiveDissipation = this->getHomogenizationState(IPStateBase::current)->getHomogenizedCohesiveJump();

		}
	}

};


void nonLinearMechSolver::archiveData(const double curtime, const int numstep){
  // Archiving
  if (_outputFile)
    fprintf(_outputFile,"Archiving on element %d, IP point %d \n",getElementSolverNumber(),getGaussPointSolverNumber());

	if (_criterionFile != NULL){
    std::string name = "A";
    linearSystem<double>* lsys = pAssembler->getLinearSystem(name);
    nonLinearSystem<double>* nonsys = dynamic_cast<nonLinearSystem<double>*>(lsys);
    double val = nonsys->getInstabilityCriterion();
    fprintf(_criterionFile,"%e; %e; \n",curtime,val);
  }

  if (_archive){
    _energField->archive(curtime,numstep,true);
    _ufield->archive(curtime,numstep,true);
    _ipf->archive(curtime,numstep,true);
  }

  this->homogenizedDataToFile(curtime);

  this->forceArchiving(curtime,numstep,true);
};


double nonLinearMechSolver::solveMicroSolverStaticLinear(){
  /* time initialization */
  double curtime = 1.;
  double timestep = 1.;
  int step = 1;
  /* solving */
  this->setTimeForBC(curtime);
  this->setTimeForLaw(curtime,timestep,step);

  if (_systemType == DISP_ELIM)
    _pAl->applyPBCByConstraintElimination();
  else if (_systemType == DISP_MULT)
    _pAl->assembleRightHandSide();
  else if (_systemType == MULT_ELIM)
    _pAl->assembleConstraintResidualToSystem();

  this->computeExternalForces();
  this->computeStiffMatrix();
  if (_outputFile)
    fprintf(_outputFile,"-- done assembling!\n");
  pAssembler->systemSolve();
  if (_outputFile)
    fprintf(_outputFile,"-- done solving!\n");
  _ipf->compute1state(IPStateBase::current,true);

   /* end of scheme */
  _iterMax = step;
  return curtime;
};

double nonLinearMechSolver::solveMicroSolverForwardEuler(){
  // external force vector // not change in function of time
	this->setTimeForBC(1.);
	_pAl->assembleRightHandSide();

  double dt = endtime/_numstepImpl;
  std::string Aname = "A";
  linearSystem<double>* lsys = pAssembler->getLinearSystem(Aname);

	double time = 0;
	for (int i=0; i<_numstepImpl; i++){
    _ipf->nextStep();
		double arclength = dt;
		time += dt;
		if (_outputFile)
      fprintf(_outputFile," iter = %d, t = %f\n",i+1,time);

		this->computeStiffMatrix();
    _pAl->assembleConstraintMatrix();
    forwardEulerArcLengthBase<double>* elsys = dynamic_cast<forwardEulerArcLengthBase<double>*>(lsys);
    elsys->setArcLengthControl(arclength);
    pAssembler->systemSolve();
    double control;
    elsys->getControlParameter(control);
    this->setTimeForBC(control);
    // update ipvariable
    _ipf->compute1state(IPStateBase::current,true);
    this->microPostSolveStep(time,i+1);
		lsys->zeroMatrix();
	};
};

void nonLinearMechSolver::setTime(const double ctime,const double dtime, const int curstep){
  _macroTimeStep = dtime;
  _macroTime = ctime;
  _macroStep = curstep;
}

double nonLinearMechSolver::microSolveStaticLinear(){
  double t= Cpu();
  initMicroSolver();
  double time = 0;
  if (_controlType == LOAD_CONTROL){
    time= solveMicroSolverStaticLinear();

    extractAverageProperties(_tangentflag);

		if (_microBC->getTotalNumberOfMechanicalDofs() > 0){
			if (_stressflag)
				_currentState->getHomogenizedStress().print("stress");
			if (_tangentflag)
				_currentState->getHomogenizedTangentOperator_F_F().print("tangent");

		}

		for (int index =0; index < _microBC->getTotalNumberOfConDofs(); index++){
      _currentState->getHomogenizedConstitutiveExtraDofFlux(index).print("con extra dof flux");
      printf("double con extra dof %d internal energy %f\n",index,_currentState->getHomogenizedConstitutiveExtraDofInternalEnergy(index));
		}

    this->endOfScheme(time,_iterMax);
  }
  else if (_controlType == ARC_CONTROL_EULER)
    Msg::Error("This is not exist");


  Msg::Info("StaticLinear OK, total time = %e seconds",t);
  return time;
};

double nonLinearMechSolver::microSolveSNL(){
  double t= Cpu();
  initMicroSolver();
  double time = 0;
  if (_controlType == LOAD_CONTROL or _pathFollowing)
    time = solveMicroSolverSNL();
  else if (_controlType == ARC_CONTROL_EULER)
    time = solveMicroSolverForwardEuler();

  this->endOfScheme(time,_iterMax);
  t = Cpu() -t;
  Msg::Info("StaticNonLinear OK, total time = %e seconds",t);

  return time;
};

int nonLinearMechSolver::pathFollowingPerturbation(){
  std::string name = "A";
  linearSystem<double>* lsys = pAssembler->getLinearSystem(name);
  pbcSystem<double> * pbcSys = dynamic_cast<pbcSystem<double>*>(lsys);
  nonLinearSystem<double>* nlsys = dynamic_cast<nonLinearSystem<double>*>(lsys);
	staticDofManager<double>* staticAssembler = static_cast<staticDofManager<double>*>(pAssembler);
	pathFollowingSystem<double>* pathsys = dynamic_cast<pathFollowingSystem<double>*>(lsys);

  // compute ipvariable
  _ipf->compute1state(IPStateBase::current, true);
  double normFinfInit = this->computeRightHandSide();

  // loop until convergence
  int iter=0;
  if(normFinfInit <_tol){ // no force (can append if contact )
    if (_outputFile)
      fprintf(_outputFile,"NO FORCE GO TO NEXT STEP\n");
      Msg::Error("perturbation is too small!");
    return 0;
  }
  double relnorm = 1;

  if (_outputFile)
    fprintf(_outputFile,"MICROITERATION n %d : RESIDU : %e \n",iter,relnorm);

  while (relnorm>_tol){
    iter++;
    // Solve KDu = Fext-Fint
    int succeed = pathsys->solvePerturbedSystem();
    if(!succeed)
    {
      Msg::Error("solution is not converged in perturbed system");
      iter = _maxNRite;
      break;
    }

    // update ipvariable
    _ipf->compute1state(IPStateBase::current,true);

    // break in case of non-iterative procedure
    if (!_iterativeNR) break;

    // check convergence criterion
    double normFinf = this->computeRightHandSide();
    relnorm = normFinf/normFinfInit;
    if (_outputFile)
      fprintf(_outputFile,"MICROITERATION n %d : RESIDU : %e \n",iter,relnorm);
    if((iter == _maxNRite) or std::isnan(relnorm))
    {
      // reset system value
      Msg::Error("solution is not converged in perturbed system");
      iter = _maxNRite; // for isnan case
      break;
    }
  }
  return iter;
};

int nonLinearMechSolver::microNewtonRaphson(const int numstep){
  std::string name = "A";
  linearSystem<double>* lsys = pAssembler->getLinearSystem(name);
  pbcSystem<double> * pbcSys = dynamic_cast<pbcSystem<double>*>(lsys);
  nonLinearSystem<double>* nlsys = dynamic_cast<nonLinearSystem<double>*>(lsys);
	staticDofManager<double>* staticAssembler = static_cast<staticDofManager<double>*>(pAssembler);

  // compute ipvariable
  _ipf->compute1state(IPStateBase::current, true);
  double normFinfInit = this->computeRightHandSide();

  // loop until convergence
  int iter=0;
  if(normFinfInit <_tol){ // no force (can append if contact )
    if (_outputFile)
      fprintf(_outputFile,"NO FORCE GO TO NEXT STEP\n");
    return 0;
  }
  double relnorm = 1;
  double normFinf =1;

  if (_outputFile)
    fprintf(_outputFile,"MICROITERATION n %d : RESIDU : %e \n",iter,relnorm);

  while ((relnorm>_tol and normFinf>_absTol)){
    iter++;
    // Solve KDu = Fext-Fint
    double t = Cpu();
    int succeed = staticAssembler->systemSolveIntReturn();
    t = Cpu() -t;

    if(!succeed)
    {
      iter = _maxNRite;
      break;
    }

    // update ipvariable
    _ipf->compute1state(IPStateBase::current,true);

    // break in case of non-iterative procedure
    if (!_iterativeNR) break;

    // check convergence criterion
    normFinf = this->computeRightHandSide();
    relnorm = normFinf/normFinfInit;
    if (_outputFile)
      fprintf(_outputFile,"MICROITERATION n %d : RESIDU : %e COMPUTING TIME: %f\n",iter,relnorm,t);
    if((iter == _maxNRite) or std::isnan(relnorm))
    {
      // reset system value
      iter = _maxNRite; // for isnan case
      break;
    }
  }
  return iter;
}

int nonLinearMechSolver::microNewtonRaphsonPathFollowing(const int numstep){
  std::string name = "A";
  linearSystem<double>* lsys =pAssembler->getLinearSystem(name);
  nonLinearSystem<double>* nonsys = dynamic_cast<nonLinearSystem<double>*>(lsys);
  pathFollowingSystem<double>* pathSys = dynamic_cast<pathFollowingSystem<double>*>(lsys);
  staticDofManager<double>* staticAssembler = static_cast<staticDofManager<double>*>(pAssembler);

  // copy state from previous
  _ipf->compute1state(IPStateBase::current,true);
  double normFint = this->computeRightHandSide();
  double norm0 = 0.;

  // loop until convergence
  int iter=0;
  double relnorm = 1;
  if (_outputFile)
    fprintf(_outputFile,"iteration n %d : residu : %e, absolute residu : %e \n",iter,relnorm,normFint);
  else if (!_multiscaleFlag)
    Msg::Info("iteration n %d : residu : %e, absolute residu : %e ",iter,relnorm,normFint);
  while ((relnorm>_tol)){

    iter++;
    int succeed = staticAssembler->systemSolveIntReturn();
    if(!succeed)
    {
      iter = _maxNRite;
      break;
    }
    // update ipvariable
    this->setTimeForBC(pathSys->getControlParameter());
    _ipf->compute1state(IPStateBase::current,true);

    if (!_iterativeNR) break;
    normFint = this->computeRightHandSide();
    if (norm0 == 0.) norm0 = nonsys->norm0Inf();
    relnorm = normFint/norm0;

    if (_outputFile)
      fprintf(_outputFile,"iteration n %d : residu : %e, absolute residu : %e \n",iter,relnorm,normFint);
    else if (!_multiscaleFlag)
      Msg::Info("iteration n %d : residu : %e, absolute residu : %e",iter,relnorm,normFint);

    if((iter == _maxNRite) or std::isnan(relnorm))
    {
      // reset system value
      iter = _maxNRite; // for isnan case
      break;
    }
  }
  return iter;
};


void nonLinearMechSolver::setPerturbationFactor(const double fact){
  _perturbationfactor = fact;
};

void nonLinearMechSolver::setInstabilityCriterion(const double val){
  _instabilitycriterion = val;
}
void nonLinearMechSolver::perturbateBucklingModeByMinimalEigenValueMode(const bool flag){
  _perturbate = flag;
};

void nonLinearMechSolver::setModeView(const int view){
	_modeview.push_back(view);
	std::string str = "ModeShape" + int2str(view);
	std::vector<nlsField::dataBuildView> tempview;
	tempview.push_back(nlsField::dataBuildView(str,-1,nlsField::crude,1));
	_eigview.push_back(tempview);
};

void nonLinearMechSolver::modeViewToFile(std::vector<int>& num, double time, int step){
	if (_eigflag){
		for (int i=0; i<num.size(); i++){
			printf("printting mode %d to file \n", num[i]);
			_eigfield[i]->archive(time,step,true);
		};
	}
	else{
		Msg::Info("Eigen solver is not activated");
	};
};

void nonLinearMechSolver::tangentAveragingFlag(const bool fl){
	_tangentflag = fl;
	#ifdef _DEBUG
	if (_tangentflag) Msg::Info("Tangent averaging is activated");
	else Msg::Info("Tangent averaging is desactivated");
	#endif
};

void nonLinearMechSolver::stressAveragingFlag(const bool fl){
	_stressflag = fl;
	#ifdef _DEBUG
	if (_stressflag) Msg::Info("Stress averaging is activated");
	else Msg::Info("Stress averaging is desactivated");
	#endif
};

void nonLinearMechSolver::setStressAveragingMethod(const int method){
  _homogenizeStressMethod = STRESS_HOMO_TYPE(method);
};
void nonLinearMechSolver::setTangentAveragingMethod(const int method, const double prec){
  _homogenizeTangentMethod = (TANGENT_HOMO_TYPE)method;
  _tangentPerturbation = prec;
};

void nonLinearMechSolver::eigenValueSolver(const int num , const bool fl){
	_numeigenvalue = num;
  _eigflag = fl;
  #ifdef _DEBUG
  if (_eigflag) Msg::Info("Eigen solver is activated");
  else Msg::Info("Eigen solver is desactivated");
  #endif
};

void nonLinearMechSolver::setEigenSolverParamerters(const int type, const int numiter, const std::string method, const double tol, const bool mktofile, const bool hem){
  if (_eigflag){
    if (type == 0){
      _eigenSolverType = STATIC;
      Msg::Info("static eigen solver is used");
    }
    else if (type ==1){
      _eigenSolverType = DYNAMIC;
      Msg::Info("dynamic eigen solver is used");
    }
    else
      Msg::Error("eigensolver type %d is not correctly defined",type);

    _eigenSolverNumIteration= numiter;
    _eigenSolverMethod = method;
    _eigenSolverConvergenCriterion = tol;
    _MKToFile = mktofile;
    _hermitian = hem;
  }
};


void nonLinearMechSolver::setMicroProblemIndentification(int ele, int gpt){
	_enum = ele;
	_gnum = gpt;
};
int nonLinearMechSolver::getElementSolverNumber() const{
	return _enum;
};
int nonLinearMechSolver::getGaussPointSolverNumber() const{
	return _gnum;
};

void nonLinearMechSolver::setHomogenizationPropertyArchiveFlag(const bool flg){
  _isHommProSaveToFile = flg;
};
void nonLinearMechSolver::setDisplacementAndIPArchiveFlag(const bool flg){
  _archive = flg;
}
void nonLinearMechSolver::setStrainArchiveFlag(const bool flg){
  _isHommStrainSaveToFile = flg;
};


void nonLinearMechSolver::setExtractPerturbationToFileFlag(const bool flag){
  _extractPerturbationToFile = flag;
};

void nonLinearMechSolver::setSystemType(const int i){
  _systemType = (SYSTEM_TYPE)i;
};
void nonLinearMechSolver::setControlType(const int i){
  _controlType = (CONTROL_TYPE)i;
};
void nonLinearMechSolver::setMicroSolverFlag(const bool flag){
  _microFlag = flag;
};

void nonLinearMechSolver::setEigenSolverFollowing(const bool f){
  _eigenFollowing = f;
}

void nonLinearMechSolver::setStabilityCriterionFollowing(const bool flag){
  _stabilityCriterionFollowing = flag;
}

void nonLinearMechSolver::setMultiscaleFlag( const bool flag){
  _multiscaleFlag = true;
};

void nonLinearMechSolver::setSameStateCriterion(const double cr){
 Msg::Info("microscopic computation does not occurs if the strain diffrence is less than %e",cr);
 _sameStateCriterion = cr;
};

void nonLinearMechSolver::readPBCDataFromFile(const std::string fn){
  if (_pbcNodeData != NULL) fclose(_pbcNodeData);
  _pbcNodeData = NULL;

  _pbcNodeData = fopen(fn.c_str(),"r");
  if (_pbcNodeData == NULL){
    Msg::Error("Unable to open file %s",fn.c_str());
  }
};


void nonLinearMechSolver::saveStiffnessMatrixToFile(const int iter){
  linearSystem<double>* lsys=NULL;
  std::string name="A";
  lsys=pAssembler->getLinearSystem(name);
#if defined(HAVE_PETSC)
  linearSystemPETSc<double>* lpet=dynamic_cast<linearSystemPETSc<double>*>(lsys);
  if(lpet==NULL)return;
  else{
    functionPETSc::MatToFile(lpet->getMatrix(),"data/Mat_Stiffness.txt");
  }
  implicitHulbertChungPetsc<double>* hcsys=dynamic_cast<implicitHulbertChungPetsc<double>*>(lsys);
  if(hcsys==NULL)return;
  else{
    functionPETSc::MatToFile(hcsys->getMassMatrix(),"data/Mat_Mass.txt");
  }
#endif
};

void nonLinearMechSolver::invertFlag(const bool fl){
  _useMatMatSolveToInvertMatrix = fl;
}

void nonLinearMechSolver::setDisturbedEigenMode(int num, double val, bool flag){
	_isPerturbedEigenMode = flag;
	_numberPerturbedMode = num;
	_valPerturbedMode = val;
	if (flag)
		Msg::Info("Set disturbing by eigenmodes: num = %d, val = %f ",num,val);
};

void nonLinearMechSolver::writeDisturbedMeshByEigenVector(int step){
	Msg::Info("Begin writing deformed mesh");
  GModel* dispgmodel = new GModel();
  dispgmodel->readMSH(_meshFileName.c_str());
  std::set<MVertex*> computedVertex;

  int numMode = _eigenSolver->getNumberEigenvectors();
  if (_numberPerturbedMode >numMode)
    _numberPerturbedMode = numMode;

  eigenVectorData* eigData = new eigenVectorData(_eigenSolver,pAssembler,_numberPerturbedMode,_valPerturbedMode);

  for (int i=0; i<domainVector.size(); i++){
  	partDomain* dom = domainVector[i];
    FunctionSpaceBase* sp = dom->getFunctionSpace();

    for (groupOfElements::elementContainer::const_iterator it = dom->g_cbegin(); it!= dom->g_cend(); it++){
      MElement* e = *it;
      std::vector<Dof> keys;
      sp->getKeys(e,keys);
      // get displacement results
			std::vector<double> vals;
			eigData->get(keys,vals);

			for (int iver =0; iver< e->getNumVertices(); iver++){
        MVertex* v = e->getVertex(iver);
				MVertex* vdisp = dispgmodel->getMeshVertexByTag(v->getNum());
				// modify directly vertex coordinates
        if (computedVertex.find(vdisp) == computedVertex.end()){
          computedVertex.insert(vdisp);
          double x = vals[iver];
          double y = vals[iver+1*e->getNumVertices()];
          double z = 0.;
          if (_dim == 3)
            z = vals[iver+2*e->getNumVertices()];
          vdisp->x() += x;
          vdisp->y() += y;
          vdisp->z() += z;
    		}
    	}
 	 	}
  }

  // write deformed mesh to file
  std::string filename = "Disturb_mesh_"+int2str(step)+ ".msh";
  dispgmodel->writeMSH(filename, CTX::instance()->mesh.mshFileVersion,CTX::instance()->mesh.binary, CTX::instance()->mesh.saveAll,CTX::instance()->mesh.saveParametric, CTX::instance()->mesh.scalingFactor);
  computedVertex.clear();
  delete dispgmodel;
	delete eigData;
	Msg::Info("End writing deformed mesh");

};

void nonLinearMechSolver::setWriteDeformedMeshToFile(bool flag){
  _isWriteDeformedMeshToFile = flag;
};

void nonLinearMechSolver::writeDeformedMesh(int step){
  Msg::Info("Begin writing deformed mesh");
  GModel* dispgmodel = new GModel();
  dispgmodel->readMSH(_meshFileName.c_str());
  std::set<MVertex*> computedVertex;

  for (int i=0; i<domainVector.size(); i++){
  	partDomain* dom = domainVector[i];
    FunctionSpaceBase* sp = dom->getFunctionSpace();
    for (groupOfElements::elementContainer::iterator it = dom->g_cbegin(); it!= dom->g_cend(); it++){
      MElement* e = *it;
      std::vector<Dof> keys;
      sp->getKeys(e,keys);
      // get displacement results
      std::vector<double> vals;
      pAssembler->getDofValue(keys,vals);
      for (int iver =0; iver< e->getNumVertices(); iver++){
        MVertex* v = e->getVertex(iver);
				MVertex* vdisp = dispgmodel->getMeshVertexByTag(v->getNum());
				// modify directly vertex coordinates
        if (computedVertex.find(vdisp) == computedVertex.end()){
          computedVertex.insert(vdisp);
          double x = vals[iver];
          double y = vals[iver+1*e->getNumVertices()];
          double z = 0.;
          if (_dim == 3)
            z = vals[iver+2*e->getNumVertices()];
          vdisp->x() += x;
          vdisp->y() += y;
          vdisp->z() += z;
    		}
    	}
 	 	}
  }
  // write deformed mesh to file
  std::string filename = "deformed_mesh_"+int2str(step)+ ".msh";
  dispgmodel->writeMSH(filename, CTX::instance()->mesh.mshFileVersion,CTX::instance()->mesh.binary, CTX::instance()->mesh.saveAll,CTX::instance()->mesh.saveParametric, CTX::instance()->mesh.scalingFactor);
  computedVertex.clear();
  delete dispgmodel;
	Msg::Info("End writing deformed mesh");
};

void nonLinearMechSolver::addModeData(std::string filename){
	FILE* file = fopen(filename.c_str(),"r");
	if (file == NULL){
		Msg::Error("File is not exist");
	}
	else
		_allModeData.push_back(file);
};

void nonLinearMechSolver::writeDisturbedMeshByEigenVectorFromFile(){
	Msg::Info("Begin writing deformed mesh");

  groupOfElements* g = new groupOfElements();

  std::map<int, std::vector<GEntity*> > groups[4];
	pModel->getPhysicalGroups(groups);
  std::map<int, std::vector<GEntity*> > &entmap = groups[_dim];
  for (std::map<int, std::vector<GEntity*> >::iterator it = entmap.begin(); it!= entmap.end(); it++){
    std::vector<GEntity*> &ent = it->second;
    for (unsigned int i = 0; i < ent.size(); i++){
      for (unsigned int j = 0; j < ent[i]->getNumMeshElements(); j++){
        MElement *e = ent[i]->getMeshElement(j);
        g->insert(e);
      }
    }
  }

  int numMode = _allModeData.size();

	if (numMode == 0){
		Msg::Info("no mode is used");
		return;
	};

	std::map<int, SVector3> disp;
	std::set<int> computedNode;

	for (int i=0; i<numMode; i++){
		// read data from file
		FILE* f = _allModeData[i];
		Msg::Info("begin reading in file %d",i);
		while (!feof(f)){
			char what[256];
			fscanf(f,"%s",what);
			if (!strcmp(what,"$ElementNodeData")){
				int nstr;
				fscanf(f,"%d",&nstr);
				for (int i=0; i<nstr; i++){
					char str[256];
					fscanf(f,"%s",str);
					printf("%s \n",str);
				};

				int nreal;
				fscanf(f,"%d",&nreal);
				for (int i=0; i<nreal; i++){
					double val;
					fscanf(f,"%lf",&val);
					printf("%f \n",val);
				};

				int ninteg;
				fscanf(f,"%d",&ninteg);

				int timestep;
				fscanf(f,"%d",&timestep);
				printf("time step = %d \n",timestep);

				if (timestep ==0) continue;

				int numcomp;
				fscanf(f,"%d",&numcomp);
				printf("num comp = %d \n", numcomp);

				int nele;
				fscanf(f,"%d",&nele);
				printf("num ele = %d \n", nele);

				int part;
				fscanf(f,"%d",&part);

				for (int i=0; i<nele; i++){
					int ele;
					fscanf(f,"%d",&ele);
					//printf("read ele = %d \n",ele);
					int numnode;
					std::vector<double> eledisp;
					fscanf(f,"%d",&numnode);
					for (int j=0; j<numnode*numcomp; j++) {
						double d;
						fscanf(f,"%lf",&d);
						eledisp.push_back(d);
					};
					if (timestep == 1){
						MElement* e = pModel->getMeshElementByTag(ele);
						std::vector<MVertex*> vv;
						e->getVertices(vv);
						for (int k=0; k<vv.size(); k++){
							MVertex* ver = vv[k];
							//Msg::Info("node =¨%d",ver->getNum());
							if (computedNode.find(ver->getNum()) == computedNode.end()){
								computedNode.insert(ver->getNum());
								SVector3 vec(0.,0.,0.);
								vec[0] =  _valPerturbedMode*eledisp[k*numcomp+0];
								vec[1] =  _valPerturbedMode*eledisp[k*numcomp+1];
								vec[2] =  _valPerturbedMode*eledisp[k*numcomp+2];
								//vec.print("vec  = ");
								disp[ver->getNum()] += vec;
							};
						};
					}
				};
			};

		}
		Msg::Info("End reading");
		fclose(f);
		computedNode.clear();
	};

	for (groupOfElements::vertexContainer::iterator it = g->vbegin(); it!= g->vend(); it++){
   	MVertex* v = *it;
    SVector3& pert = disp[v->getNum()];
    v->x() += pert[0];
    v->y() += pert[1];
    v->z() += pert[2];

	}

	std::string filename = "deformed_mesh.msh";
  pModel->writeMSH(filename, CTX::instance()->mesh.mshFileVersion,CTX::instance()->mesh.binary, CTX::instance()->mesh.saveAll,CTX::instance()->mesh.saveParametric, CTX::instance()->mesh.scalingFactor);
	Msg::Info("End writing deformed mesh");

};


/** perform a particular test**/
void nonLinearMechSolver::activateTest(const bool fl){
  _testFlag = fl;
  _microFlag = false; // desactive the microFlag
  Msg::Info("activation test");
};



void nonLinearMechSolver::initialize_test(){
  if (_microBC == NULL) {
    Msg::Error("test with micro BC must be coupled with one micro BC");
  };
  if (_pbcGroup != NULL) delete _pbcGroup;

  std::vector<groupOfElements*>& bgroup  = _microBC->getBoundaryGroupOfElements();
  const std::vector<int>& bphysical = _microBC->getBoundaryPhysicals();
  bgroup.clear();
  for (int i=0; i< bphysical.size(); i++){
    groupOfElements* gr = new groupOfElements(_dim-1,bphysical[i]);
    bgroup.push_back(gr);
  }

  FunctionSpaceBase* lagspace = domainVector[0]->getFunctionSpace();
  _pbcGroup = new pbcConstraintElementGroup(this,lagspace,NULL);
  /*create all linear constraints*/
  _pbcGroup->createConstraintForTest();
  /*enforce constraints to system*/
  _pbcGroup->applyLinearConstraintsToSystem(pAssembler);
  /* init homogenized stress file*/
  _rveVolume = _pbcGroup->getRVEVolume();

};

void nonLinearMechSolver::displacementBCOnControlNode(const int pos, const int comp, const double value){
  allCornerConstraint.emplace_back();
  nonLinearDirichletBCAtCorner& cornerConstraint = allCornerConstraint.back();
  cornerConstraint._tag = pos;
  cornerConstraint._comp = comp;
  cornerConstraint._cornerNumber = pos;
  cornerConstraint._f = new simpleFunctionTime<double>(value);
};

void nonLinearMechSolver::forceBCOnControlNode(const int pos, const int comp, const double value){
  allCornerForce.emplace_back();
  nonLinearNeumannBCAtCorner& cornerConstraint = allCornerForce.back();
  cornerConstraint._tag = pos;
  cornerConstraint._comp = comp;
  cornerConstraint._cornerNumber = pos;
  cornerConstraint._f = new simpleFunctionTime<double>(value);
};

void nonLinearMechSolver::getDofOnControlNode(const int pos, std::vector<Dof>& R){
  if (_testFlag){
    if (_pbcGroup != NULL){
      _pbcGroup->getDofsOfCornerVertex(pos,R);
    }
  }
};

void nonLinearMechSolver::setExtractCohesiveLawFromMicroDamage(const bool check){
  _damageToCohesiveJump = check;
  if (check){
    _tangentflag = true;
    Msg::Info("extract cohesive jump from microdamage, tangent computation has to be activated");
    _checkFailureOnset = true;
  }
};

void nonLinearMechSolver::setLostSolutionUniquenssTolerance(const double tol){
	_lostSolutionUniquenssTolerance = tol;
}

void nonLinearMechSolver::setExtractIrreversibleEnergyFlag(const bool ex){
	_extractIrreversibleEnergy =ex;
}

void nonLinearMechSolver::setLocalizationNormal(const SVector3& normal){
  _lostSolutionUniquenssNormal = normal;

  if (_lostSolutionUniquenssNormal.norm() == 0.){
    Msg::Info("zero normal is set");
  }
  else{
    _lostSolutionUniquenssNormal.normalize();
  }
};
void nonLinearMechSolver::setLocalizationNormal(const double n1, const double n2, const double n3){
  _lostSolutionUniquenssNormal(0) = n1;
  _lostSolutionUniquenssNormal(1) = n2;
  _lostSolutionUniquenssNormal(2) = n3;

  if (_lostSolutionUniquenssNormal.norm() == 0.){
    Msg::Info("zero normal is set");
  }
  else{
    _lostSolutionUniquenssNormal.normalize();
  }
};

void nonLinearMechSolver::setRVELengthInNormalDirection(const double l, const double sfR){
  _RVELengthInCohesiveNormal = l;
	_surfaceReductionRatio = sfR;
};

void nonLinearMechSolver::setVoidPartInLocalizationBand(const double vp){
	_voidPartInLocalizationBand = vp;
};

void nonLinearMechSolver::setCheckFailureOnset(const bool fl, const bool withNormal){
	_checkFailureOnset = fl;
	_checkWithNormal = withNormal;
	if (_checkFailureOnset){
    if (_checkWithNormal){
      printf("failure will check with prescribed normal\n");
    }
    else{
      printf("failure will check without normal\n");
    }
	}
};

void nonLinearMechSolver::switchMicroBC(const nonLinearMicroBC* mbc){
	this->addMicroBC(mbc);

	// update data with solver
	nonLinearMicroBC*  microBC = this->getMicroBC();
	microBC->updateWithSolver(this);

	// recreate system,
  this->createMicroSystem();

	// init all other BC, fixed DOf, etc
  this->initAllBCsOnDofs();

  // create new pbcAlgorithm
  if (_pAl != NULL){
		delete _pAl;
	}
	_pAl = new pbcAlgorithm(this); // create all constraints, etc

  // renumber Dof and reallocate the system
  this->microNumberDof();

	// update unknown from previous system
	staticDofManager<double>* sAssembler = this->getStaticDofManager();
	if (sAssembler != NULL){
		sAssembler->updateFromPrevious();
	}
	_ipf->copy(IPStateBase::current,IPStateBase::initial);
	_ipf->copy(IPStateBase::current,IPStateBase::previous);
};

void nonLinearMechSolver::setSwitchMicroBCFlag(const bool fl){
	_failureBCIsSwitched = fl;
};

void nonLinearMechSolver::setExtractElasticTangentOperator(const bool fl){
  _extractElasticTangentOperator = fl;
  if (_extractElasticTangentOperator){
    Msg::Info("extract elastic tangent works if setTangentAveragingMethod(2) is used ");
  }
};
