// C++ Interface: partDomain
//
// Description: Interface class for dg 3D
//
// Author:  <L. Noels>, (C) 2011
//
// Copyright: See COPYING file that comes with this distribution
//
//
#include "nonLinearMechSolver.h"
#include "dG3DTerms.h"
#include "dG3DDomain.h"
#include "dG3DFilterDofComponent.h"
#include "IFace.h"
#include "GaussLobatto2DQuadratureRules.h"
#include "GmshMessage.h"
#include "numericalMaterial.h"
#include "IEdge.h"
#include "MInterfaceLine.h"
#include "MInterfacePoint.h"
#include "dG3DMultiscaleMaterialLaw.h"
#include "extraFieldFunctionSpace.h"
#include "multipleFieldDG3DTerms.h"
#include "terms.h"
#include "pathFollowingTerms.h"
#include "dG3DHomogenizedTangentTerms.h"
#include "FractureCohesiveDG3DMaterialLaw.h"
#include "nonLocalDamageDG3DIPVariable.h"



dG3DDomain::dG3DDomain (const int tag,const int phys, const int ws, const int lnum,
                                          const int fdg, const int dim, int nonLocalVar, int nonConstitutiveExtraDOFVar) : dgPartDomain(tag,phys,ws,fdg,nonLocalVar, nonConstitutiveExtraDOFVar),
                                                            _lnum(lnum),
                                                            _beta1(10.),_sts(1.), _isHighOrder(false), _interQuad(NULL),
                                                            _space(NULL), _dim(dim),
                                                            _gqt(Gauss), _gaussorderbulk(-1), _gaussorderbound(-1), _evalStiff(false),
                                                            _bulkDamageBlockMethod(-1),_subSteppingMethod(0),
                                                            _forceInterfaceElementBroken(false),_failureIPRatio(1.), _EqMultiFieldFactor(1.),
                                                            _nonLocalBeta(100.), _nonLocalContinuity(true), _nonLocalEqRatio(100.),
                                                            _constitutiveExtraDofDiffusionBeta(100.), _constitutiveExtraDofDiffusionContinuity(true), _constitutiveExtraDofDiffusionEqRatio(100.),
                                                            _constitutiveExtraDofDiffusionUseEnergyConjugatedField(false), _constitutiveExtraDofDiffusionAccountFieldSource(true),
                                                            _constitutiveExtraDofDiffusionAccountMecaSource(false), _imposeCrackFlag(false),_accountPathFollowing(true),
                                                            _incrementNonlocalBased(false), _planeStressState(false)
{
  switch(_wsp)
  {
   case functionSpaceType::Lagrange:
    if(_fullDg)
      _space = new dG3DLagrangeFunctionSpace(getTag(),3+getNumNonLocalVariable()+getNumConstitutiveExtraDofDiffusionVariable());
    else
      _space = new g3DLagrangeFunctionSpace(getTag(),3+getNumNonLocalVariable()+getNumConstitutiveExtraDofDiffusionVariable());
    _spaceInterUVW = new IsoparametricLagrangeFunctionSpace(getTag()); // 3 composantes
    break;
   case functionSpaceType::Inter :
    _spaceInterUVW = new IsoparametricLagrangeFunctionSpace(getTag()); // 3 composantes
    break; // InterfaceDomain set function space later
   default:
    Msg::Error("Function space type is unknown on domain %d",getTag());
  }

  _extraSpace = new extraFieldFunctionSpace(this,tag,3); // is getNumNonLocalVariable() missing

}

dG3DDomain::dG3DDomain(const dG3DDomain &source) : dgPartDomain(source), _lnum(source._lnum){
	_mlaw = NULL; // law is not copied
  _dim  = source._dim;
  _beta1 = source._beta1;
  //
  _sts = source._sts;
  _evalStiff = source._evalStiff;
  //
  _isHighOrder = source._isHighOrder;

  _extraSpace = NULL;
  if (source._extraSpace != NULL)
    _extraSpace = source._extraSpace->clone(source._extraSpace->getId());
  //
  _space = NULL;
  if (source._space != NULL)
    _space = source._space->clone(source._space->getId());
  //
  _spaceInterUVW = NULL;
  if (source._spaceInterUVW != NULL)
    _spaceInterUVW = source._spaceInterUVW->clone(source._spaceInterUVW->getId());

  _gaussorderbulk = source._gaussorderbulk;
  _gaussorderbound = source._gaussorderbound;
  //
	// _interQuad will be set in solver
  _interQuad = NULL; // source._interQuad;
  _gqt = source._gqt;

  _subSteppingMethod = source._subSteppingMethod;
  _subStepNumber = source._subStepNumber;

  _bulkDamageBlockMethod = source._bulkDamageBlockMethod;
  _forceInterfaceElementBroken = source._forceInterfaceElementBroken;
  _failureIPRatio = source._failureIPRatio;

  _EqMultiFieldFactor = source._EqMultiFieldFactor;
  _imposeCrackFlag = source._imposeCrackFlag;

  _accountPathFollowing = source._accountPathFollowing;

  _nonLocalBeta = source._nonLocalBeta;
  _nonLocalContinuity = source._nonLocalContinuity;
  _nonLocalEqRatio = source._nonLocalEqRatio;
  //
  _constitutiveExtraDofDiffusionBeta = source._constitutiveExtraDofDiffusionBeta;
  _constitutiveExtraDofDiffusionContinuity = source._constitutiveExtraDofDiffusionContinuity;
  _constitutiveExtraDofDiffusionEqRatio = source._constitutiveExtraDofDiffusionEqRatio;
  _constitutiveExtraDofDiffusionUseEnergyConjugatedField = source._constitutiveExtraDofDiffusionUseEnergyConjugatedField;
  _constitutiveExtraDofDiffusionAccountFieldSource = source._constitutiveExtraDofDiffusionAccountFieldSource;
  _constitutiveExtraDofDiffusionAccountMecaSource = source._constitutiveExtraDofDiffusionAccountMecaSource;

  _incrementNonlocalBased = source._incrementNonlocalBased;
	_planeStressState = source._planeStressState;
}

dG3DDomain::~dG3DDomain(){
  if (_interQuad) {delete _interQuad; _interQuad = NULL;};
  if (_space) {delete _space; _space = NULL;};
  if (_spaceInterUVW) {delete _spaceInterUVW; _spaceInterUVW = NULL;};
  if (_extraSpace)  {delete _extraSpace; _extraSpace = NULL;}
};

void dG3DDomain::setMaterialLaw(const std::map<int,materialLaw*> &maplaw)
{
  if(!setmaterial){
    for(std::map<int,materialLaw*>::const_iterator it=maplaw.begin(); it!=maplaw.end(); ++it){
      if(it->first == _lnum){
        _mlaw = it->second;
        _mlaw->initLaws(maplaw);
        _isHighOrder = _mlaw->isHighOrder();
      }
    }
    setmaterial = true;
  }
}
void dG3DDomain::accountPathFollowing(const bool fl){
  _accountPathFollowing = fl;
  if (fl){
    Msg::Info("Domain %d is considered in path following",getPhysical());
  }
  else{
    Msg::Info("Domain %d is not considered in path following",getPhysical());
  }
};

void dG3DDomain::setIncrementNonlocalBased(const bool fl){
  _incrementNonlocalBased = fl;
  if (_incrementNonlocalBased){
    Msg::Info("nonlocal based on increment is considered in dom %d",getPhysical());
  }
}

void dG3DDomain::setPlaneStressState(const bool fl){
	_planeStressState = fl;
	if (_planeStressState){
		Msg::Info("Plane stress problem is considered");
	}
};

void dG3DDomain::matrixByPerturbation(const int iinter, const double eps){
	_interByPert = iinter;
	_eps = eps;
}

void dG3DDomain::stabilityParameters(const double b1)
{
  _beta1 = b1;
  _sts = 1./sqrt(_beta1);
}

void dG3DDomain::setBulkDamageBlockedMethod(const int method){_bulkDamageBlockMethod = method;};
void dG3DDomain::forceCohesiveInsertionAllIPs(const bool flg, const double fratio){_forceInterfaceElementBroken = flg; _failureIPRatio = fratio;};
void dG3DDomain::strainSubstep(const int method, const int numstep){ _subSteppingMethod = method; _subStepNumber = numstep;};

void dG3DDomain::setMultipleFieldEqRatio(const double val){
 _EqMultiFieldFactor = val;
};

void dG3DDomain::setNewIdForComp(const int comp, const int type){
  ThreeDLagrangeFunctionSpace* gspace = dynamic_cast<ThreeDLagrangeFunctionSpace*>(this->getFunctionSpace());
  if (gspace != NULL){
    gspace->setNewIdForComp(comp,type);
  }
};

void dG3DDomain::setImposedCrack(const bool flag){
	_imposeCrackFlag = flag;
};

void dG3DDomain::imposeCrack(const int em, const int ep){
	printf("add crack between %d and %d \n",em,ep);
	_imposedInterfaceCrack.insert(TwoNum(em,ep));
};


const double dG3DDomain::stabilityParameter(const int i) const
{
  switch(i)
  {
   case 1:
    return _beta1;
   default:
    Msg::Error("This domain has only 1 stability parameters !");
    return 0.;
  }
}

void dG3DDomain::matrixByPerturbation(const int ibulk, const int iinter, const int ivirt, const double eps)
{
  this->setBulkMatrixByPerturbation(ibulk,eps);
  iinter == 0 ? _interByPert = false : _interByPert = true;
  ivirt == 0 ? _virtByPert = false : _virtByPert = true;
}

FunctionSpaceBase* dG3DDomain::getSpaceForBC(const nonLinearBoundaryCondition::type bc_type,
                                             const nonLinearBoundaryCondition::location bc_location,
                                             const nonLinearNeumannBC::NeumannBCType neumann_type, 
                                             const int dof_comp,const groupOfElements *groupBC,const int domcomp) const
{
  // when DG is used
  // if BC is applied on a dimension smaller than one of current domain dG3DBoundaryConditionLagrangeFunctionSpace must be use
  // if BC is applied on the same dimension as one of the current domain dG3DLagrangeFunctionSpace must be used
  //
  // when CG is used
  // g3DDirichletBoundaryConditionLagrangeFunctionSpace, g3DNeumannBoundaryConditionLagrangeFunctionSpace should be used
  // as keys are generated based on vertex
  //

  // with PRESSURE, SCALAR_FLUX boundary conditions
  // dof_comp is not used as all Dof must be present, only domcomp is used
  // with others,
  // dof_comp is used, domcomp is not used

  FunctionSpaceBase* spacebc;
  if(bc_type == nonLinearBoundaryCondition::DIRICHLET)
  {
    switch(_wsp){
     case functionSpaceType::Lagrange:
      if(_fullDg){
        if (getDim() == 3){
          if(bc_location !=nonLinearBoundaryCondition::ON_VOLUME)
            spacebc= new dG3DBoundaryConditionLagrangeFunctionSpace(getTag(),1,dof_comp,groupBC,g,_groupGhostMPI);
          else
            spacebc = new dG3DLagrangeFunctionSpace(getTag(),1,dof_comp);
        }
        else if (getDim() == 2){
          if(bc_location !=nonLinearBoundaryCondition::ON_FACE)
            spacebc= new dG3DBoundaryConditionLagrangeFunctionSpace(getTag(),1,dof_comp,groupBC,g,_groupGhostMPI);
          else
            spacebc = new dG3DLagrangeFunctionSpace(getTag(),1,dof_comp);
        }
      }
      else{
        spacebc = new g3DDirichletBoundaryConditionLagrangeFunctionSpace(getTag(),1,dof_comp);

		}
		break;
		default :
       Msg::Error("Unknown function space type for a dirichlet BC on domain %d",_phys);
    }
  }
  else if(bc_type == nonLinearBoundaryCondition::NEUMANN)
  {
    switch(_wsp){
     case functionSpaceType::Lagrange :
      if(_fullDg){
        if (neumann_type == nonLinearNeumannBC::FORCE){
          // force in reference configuration following dof_comp is applied 
          if ((this->getDim() == 3 and bc_location == nonLinearBoundaryCondition::ON_VOLUME) or
              (this->getDim() == 2 and bc_location == nonLinearBoundaryCondition::ON_FACE) or 
              (this->getDim() == 1 and bc_location == nonLinearBoundaryCondition::ON_EDGE)){
            // applying BC in an entity having same dimension as current domain
            spacebc = new dG3DLagrangeFunctionSpace(getTag(),1,dof_comp);
          }
          else{
            // applying BC in an entity having a smaller dimension as current domain
            spacebc = new dG3DBoundaryConditionLagrangeFunctionSpace(getTag(),1,dof_comp,groupBC,gi,giv,NULL); // no ghostMPI for Neumann BC
          }
        }
        else if (neumann_type == nonLinearNeumannBC::PRESSURE or neumann_type == nonLinearNeumannBC::SCALAR_FLUX){
          // allways applying BC in an entity having a smaller dimension as current domain
          spacebc = new dG3DBoundaryConditionLagrangeFunctionSpace(getTag(),domcomp,groupBC,gi,giv,NULL); // no ghostMPI for Neumann BC
        }
        else{
          Msg::Error("nonLinearNeumannBC::NeumannBCType %s is not defined",neumann_type);
        }
      }
      else
        if (neumann_type == nonLinearNeumannBC::FORCE){
          spacebc = new g3DNeumannBoundaryConditionLagrangeFunctionSpace(getTag(),1,dof_comp);
        }
        else if (neumann_type == nonLinearNeumannBC::PRESSURE or neumann_type == nonLinearNeumannBC::SCALAR_FLUX){
          spacebc = new g3DNeumannBoundaryConditionLagrangeFunctionSpace(getTag(),domcomp);
        }
        else{
          Msg::Error("nonLinearNeumannBC::NeumannBCType %s is not defined",neumann_type);
        }

     break;
      default :
      Msg::Error("Unknown function space type on a neumman BC on domain %d",_phys);
    }
  }
  else if(bc_type == nonLinearBoundaryCondition::INITIAL)
  {
    switch(_wsp){
     case functionSpaceType::Lagrange:
      if(_fullDg){
        if (getDim()==3){
          if(bc_location !=nonLinearBoundaryCondition::ON_VOLUME)
            spacebc = new dG3DBoundaryConditionLagrangeFunctionSpace(getTag(),1,dof_comp,groupBC,g,_groupGhostMPI);
          else
            spacebc = new dG3DLagrangeFunctionSpace(getTag(),1,dof_comp);
        }
        else if (getDim()==2){
          if(bc_location !=nonLinearBoundaryCondition::ON_FACE)
            spacebc = new dG3DBoundaryConditionLagrangeFunctionSpace(getTag(),1,dof_comp,groupBC,g,_groupGhostMPI);
          else
            spacebc = new dG3DLagrangeFunctionSpace(getTag(),1,dof_comp);
        }
      }
      else{
        spacebc = new g3DDirichletBoundaryConditionLagrangeFunctionSpace(getTag(),1,dof_comp);
      }
      break;
     default :
       Msg::Error("Unknown function space type for an initial BC on domain %d",_phys);
    }
  }
  return spacebc;
}

QuadratureBase* dG3DDomain::getQuadratureRulesForNeumannBC(nonLinearNeumannBC &neu) const
{
  QuadratureBase *integ = NULL;
  if((neu.onWhat == nonLinearBoundaryCondition::ON_VOLUME) or
     (neu.onWhat == nonLinearBoundaryCondition::ON_FACE and this->getDim() == 2) or
     (neu.onWhat == nonLinearBoundaryCondition::ON_EDGE and this->getDim() == 1) or
     (neu.onWhat == nonLinearBoundaryCondition::ON_VERTEX and this->getDim() == 0)){
    // when applying BC on domain
    GaussQuadrature *integgauss = static_cast<GaussQuadrature*>(integBulk);
    integ = new GaussQuadrature(*integgauss);
  }
  else{
    if (neu.g->size() > 0){
      groupOfElements::elementContainer::iterator it = neu.g->begin();
      MElement *ie = *it;

      if (ie->getTypeForMSH() == MSH_TRI_3)
      {
        if( _gqt == Gauss){
          integ = new GaussQuadrature(2); // 3 Gauss points
        }
        else if (_gqt == Lobatto)
        {
          integ = new GaussLobatto2DQuadrature(2); //order 2
        }
      }
      else if (ie->getTypeForMSH() == MSH_TRI_6)
      {
        if( _gqt == Gauss){
          integ = new GaussQuadrature(4);
        }
        else if (_gqt == Lobatto)
        {
          integ = new GaussLobatto2DQuadrature(2); //order 2
        }
      }
      else if (ie->getTypeForMSH() == MSH_QUA_9)
      {
        if( _gqt == Gauss){
          integ = new GaussQuadrature(4);
        }
        else if (_gqt == Lobatto)
        {
          integ = new GaussLobatto2DQuadrature(2); //order 2
        }
      }
      else if (ie->getTypeForMSH() == MSH_QUA_4)
      {
        if( _gqt == Gauss)
        {
          integ = new GaussQuadrature(2);
        }
        else if (_gqt == Lobatto)
        {
          integ = new GaussLobatto2DQuadrature(2); //order 2
        }
      }
      else if (ie->getTypeForMSH() == MSH_LIN_4)
      {
        if( _gqt == Gauss)
        {
          integ = new GaussQuadrature(7); // 4 pts
        }
      }
      else if (ie->getTypeForMSH() == MSH_LIN_3)
      {
        if( _gqt == Gauss)
        {
          integ = new GaussQuadrature(5); // 3 pts
        }
      }
      //else if (e->getTypeForMSH() == MSH_QUA_4)
      else if (ie->getTypeForMSH() == MSH_LIN_2)
      {
        if( _gqt == Gauss)
        {
          integ = new GaussQuadrature(3); // 2 pts
        }
      }
      else if (ie->getTypeForMSH() == MSH_TRI_10)
      {
        if( _gqt == Gauss){
          integ = new GaussQuadrature(7);
        }
        else if (_gqt == Lobatto)
        {
          integ = new GaussLobatto2DQuadrature(3); //order 2
        }
      }
      else if (ie->getTypeForMSH() == MSH_PNT){
        if( _gqt == Gauss)
        {
          integ = new GaussQuadrature(1);
        }
      }
      else{
        Msg::Error("gauss quadrature for NeumannBC is not defined by default");
      }

      IntPt* GP;
      int npts = integ->getIntPoints(ie,&GP);
      Msg::Info("NeumannBC: use %d integration points for element type %d",npts,ie->getTypeForMSH());

    }
    else{
      integ = new QuadratureVoid();
      Msg::Error("Neumann BC has no element");
    }
  }
  if (integ == NULL){
    Msg::Error("missing case in dG3DDomain::getQuadratureRulesForNeumannBC");
  }
  return integ;
}



void dG3DDomain::createTerms(unknownField *uf,IPField*ip)
{
  FunctionSpace<double>* dgspace = static_cast<FunctionSpace<double>*>(_space);
  massterm = new mass3D(*dgspace,_mlaw);
  ltermBulk = new dG3DForceBulk(*dgspace,_mlaw,_fullDg,ip,_incrementNonlocalBased);
  if(_bmbp)
    btermBulk = new BilinearTermPerturbation<double>(ltermBulk,*dgspace,*dgspace,uf,ip,this,_eps);
  else
    btermBulk = new dG3DStiffnessBulk(*dgspace,_mlaw,_fullDg,ip);

  btermElasticBulk = new dG3DElasticStiffnessBulk(*dgspace,_mlaw,_fullDg,ip);

  if(_fullDg){
    if (IsMultipleFieldFormulation()){
      if (getMultipleFieldMethod() == dgPartDomain::TWOFIELD){
        ltermBound = new twoFieldDG3DForceInter(this,ip,_EqMultiFieldFactor,_beta1);
      }
      else{
        Msg::Error("two field method %d has not been correctly defined",getMultipleFieldMethod());
      }
    }
    else{
      ltermBound = new dG3DForceInter(*dgspace,dgspace,_mlaw,_mlaw,_interQuad,_beta1,_fullDg,ip,_incrementNonlocalBased);
    }
    if(_interByPert)
      btermBound = new BilinearTermPerturbation<double>(ltermBound,*dgspace,*dgspace,uf,ip,this,_eps);
    else{
      if (IsMultipleFieldFormulation()){
        if (getMultipleFieldMethod() == dgPartDomain::TWOFIELD){
          btermBound = new twoFieldDG3DStiffnessInter(this,ip,_EqMultiFieldFactor,_beta1);
        }
        else
        {
          Msg::Error("two field method %d has not been correctly defined",getMultipleFieldMethod());
        }
      }
      else
      {
        btermBound = new dG3DStiffnessInter(*dgspace,*dgspace,_mlaw,_mlaw,_interQuad,_beta1,ip,uf,_fullDg,_incrementNonlocalBased);
      }
    }
  }
  else
  {
     ltermBound = new nonLinearTermVoid();
     btermBound = new BiNonLinearTermVoid();
  }


  if( getNumNonLocalVariable() >0)
  {
    (static_cast<dG3DForceBulk*>(ltermBulk))->setNonLocalEqRatio(getNonLocalEqRatio());
    (static_cast<dG3DForceBulk*>(ltermBulk))->setNumNonLocalVariable(getNumNonLocalVariable());
    if(!_bmbp)
    {
      (static_cast<dG3DStiffnessBulk*>(btermBulk))->setNonLocalEqRatio(getNonLocalEqRatio());
      (static_cast<dG3DStiffnessBulk*>(btermBulk))->setNumNonLocalVariable(getNumNonLocalVariable());
    }

    (static_cast<dG3DElasticStiffnessBulk*>(btermElasticBulk))->setNonLocalEqRatio(getNonLocalEqRatio());
    (static_cast<dG3DElasticStiffnessBulk*>(btermElasticBulk))->setNumNonLocalVariable(getNumNonLocalVariable());

    if(_fullDg)
    {
      if(!IsMultipleFieldFormulation())
      {
        (static_cast<dG3DForceInter*>(ltermBound))->setNonLocalEqRatio(getNonLocalEqRatio());
        (static_cast<dG3DForceInter*>(ltermBound))->setNumNonLocalVariable(getNumNonLocalVariable());
        (static_cast<dG3DForceInter*>(ltermBound))->setNonLocalStabilityParameter(getNonLocalStabilityParameter());
        (static_cast<dG3DForceInter*>(ltermBound))->setNonLocalContinuity(getNonLocalContinuity());
        if (!_interByPert){
          (static_cast<dG3DStiffnessInter*>(btermBound))->setNonLocalEqRatio(getNonLocalEqRatio());
          (static_cast<dG3DStiffnessInter*>(btermBound))->setNumNonLocalVariable(getNumNonLocalVariable());
          (static_cast<dG3DStiffnessInter*>(btermBound))->setNonLocalStabilityParameter(getNonLocalStabilityParameter());
          (static_cast<dG3DStiffnessInter*>(btermBound))->setNonLocalContinuity(getNonLocalContinuity());
        }
      }
    }
  }
  if( getNumConstitutiveExtraDofDiffusionVariable() >0)
  {
    (static_cast<dG3DForceBulk*>(ltermBulk))->setNumConstitutiveExtraDofDiffusionVariable(getNumConstitutiveExtraDofDiffusionVariable());
    (static_cast<dG3DForceBulk*>(ltermBulk))->setConstitutiveExtraDofDiffusionEqRatio(getConstitutiveExtraDofDiffusionEqRatio());
    (static_cast<dG3DForceBulk*>(ltermBulk))->setConstitutiveExtraDofDiffusionUseEnergyConjugatedField(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField());
    (static_cast<dG3DForceBulk*>(ltermBulk))->setConstitutiveExtraDofDiffusionAccountFieldSource(getConstitutiveExtraDofDiffusionAccountFieldSource());
    (static_cast<dG3DForceBulk*>(ltermBulk))->setConstitutiveExtraDofDiffusionAccountMecaSource(getConstitutiveExtraDofDiffusionAccountMecaSource());
    if(!_bmbp)
    {
      (static_cast<dG3DStiffnessBulk*>(btermBulk))->setNumConstitutiveExtraDofDiffusionVariable(getNumConstitutiveExtraDofDiffusionVariable());
      (static_cast<dG3DStiffnessBulk*>(btermBulk))->setConstitutiveExtraDofDiffusionEqRatio(getConstitutiveExtraDofDiffusionEqRatio());
      (static_cast<dG3DStiffnessBulk*>(btermBulk))->setConstitutiveExtraDofDiffusionUseEnergyConjugatedField(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField());
      (static_cast<dG3DStiffnessBulk*>(btermBulk))->setConstitutiveExtraDofDiffusionAccountFieldSource(getConstitutiveExtraDofDiffusionAccountFieldSource());
      (static_cast<dG3DStiffnessBulk*>(btermBulk))->setConstitutiveExtraDofDiffusionAccountMecaSource(getConstitutiveExtraDofDiffusionAccountMecaSource());
    }
    if(_fullDg)
    {
      if(!IsMultipleFieldFormulation())
      {
        (static_cast<dG3DForceInter*>(ltermBound))->setNumConstitutiveExtraDofDiffusionVariable(getNumConstitutiveExtraDofDiffusionVariable());
        (static_cast<dG3DForceInter*>(ltermBound))->setConstitutiveExtraDofDiffusionEqRatio(getConstitutiveExtraDofDiffusionEqRatio());
        (static_cast<dG3DForceInter*>(ltermBound))->setConstitutiveExtraDofDiffusionUseEnergyConjugatedField(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField());
        (static_cast<dG3DForceInter*>(ltermBound))->setConstitutiveExtraDofDiffusionAccountFieldSource(getConstitutiveExtraDofDiffusionAccountFieldSource());
        (static_cast<dG3DForceInter*>(ltermBound))->setConstitutiveExtraDofDiffusionAccountMecaSource(getConstitutiveExtraDofDiffusionAccountMecaSource());
        (static_cast<dG3DForceInter*>(ltermBound))->setConstitutiveExtraDofDiffusionStabilityParameter(getConstitutiveExtraDofDiffusionStabilityParameter());
        (static_cast<dG3DForceInter*>(ltermBound))->setConstitutiveExtraDofDiffusionContinuity(getConstitutiveExtraDofDiffusionContinuity());
        if(!_interByPert)
        {
          (static_cast<dG3DStiffnessInter*>(btermBound))->setNumConstitutiveExtraDofDiffusionVariable(getNumConstitutiveExtraDofDiffusionVariable());
          (static_cast<dG3DStiffnessInter*>(btermBound))->setConstitutiveExtraDofDiffusionEqRatio(getConstitutiveExtraDofDiffusionEqRatio());
          (static_cast<dG3DStiffnessInter*>(btermBound))->setConstitutiveExtraDofDiffusionUseEnergyConjugatedField(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField());
          (static_cast<dG3DStiffnessInter*>(btermBound))->setConstitutiveExtraDofDiffusionAccountFieldSource(getConstitutiveExtraDofDiffusionAccountFieldSource());
          (static_cast<dG3DStiffnessInter*>(btermBound))->setConstitutiveExtraDofDiffusionAccountMecaSource(getConstitutiveExtraDofDiffusionAccountMecaSource());
          (static_cast<dG3DStiffnessInter*>(btermBound))->setConstitutiveExtraDofDiffusionStabilityParameter(getConstitutiveExtraDofDiffusionStabilityParameter());
          (static_cast<dG3DStiffnessInter*>(btermBound))->setConstitutiveExtraDofDiffusionContinuity(getConstitutiveExtraDofDiffusionContinuity());
        }
      }
    }
  }

  this->btermVirtBound = new BiNonLinearTermVoid();
  this->ltermVirtBound = new nonLinearTermVoid();

  //path following
  if (this->getMacroSolver()->withPathFollowing() and _accountPathFollowing){
    if (getMacroSolver()->getPathFollowingLocation() == nonLinearMechSolver::INTERFACE){
      scalarTermPF = new nonLinearScalarTermVoid();
      linearTermPF = new nonLinearTermVoid();
    }
    else{
      scalarTermPF = new dG3DDissipationPathFollowingBulkScalarTerm(this,ip);
      linearTermPF = new dG3DDissipationPathFollowingBulkLinearTerm(this,ip);
    }

    if (getMacroSolver()->getPathFollowingLocation() == nonLinearMechSolver::BULK){
      scalarTermPFBound = new nonLinearScalarTermVoid();
      linearTermPFBound = new nonLinearTermVoid();
    }
    else{
      if (this->IsMultipleFieldFormulation()){
        scalarTermPFBound = new twoFieldDG3DDissipationPathFollowingBoundScalarTerm(this,ip);
        linearTermPFBound = new twoFieldDG3DDissipationPathFollowingBoundLinearTerm(this,ip);
      }
      else{
        scalarTermPFBound = new dG3DDissipationPathFollowingBoundScalarTerm(this,ip);
        linearTermPFBound = new dG3DDissipationPathFollowingBoundLinearTerm(this,ip);
      }
    }
  }
  else{
    scalarTermPF = new nonLinearScalarTermVoid();;
    linearTermPF = new nonLinearTermVoid();;
    scalarTermPFBound = new nonLinearScalarTermVoid();;
    linearTermPFBound = new nonLinearTermVoid();
  }
  btermTangent = new dG3DHomogenizedTangent(this,ip);
  btermElasticTangent = new dG3DHomogenizedElasticTangent(this,ip);
}

void dG3DDomain::initializeTerms(unknownField *uf,IPField*ip)
{
  createTerms(uf,ip);

   /* same value for fracture strength on both side */
   // not very elegant put very efficient to put here
   if(_mlaw->getType() == materialLaw::fracture)
   {
     // loop on ipvariable
     IntPt *GP;
     for(groupOfElements::elementContainer::iterator it=gi->begin(); it!=gi->end(); ++it)
     {
       MElement *ele = *it;
       int npts_inter=integBound->getIntPoints(ele,&GP);
       AllIPState::ipstateElementContainer *vips = ip->getAips()->getIPstate(ele->getNum());
       for(int j=0;j<npts_inter;j++){ // npts_inter points on both sides
         IPStateBase* ipsm = (*vips)[j];
         IPStateBase* ipsp = (*vips)[j+npts_inter];
         FractureCohesive3DIPVariable *ipvm = static_cast<FractureCohesive3DIPVariable*>(ipsm->getState(IPStateBase::initial));
         FractureCohesive3DIPVariable *ipvp = static_cast<FractureCohesive3DIPVariable*>(ipsp->getState(IPStateBase::initial));
         ipvp->setFractureStrengthFactor(ipvm->fractureStrengthFactor());
         ipvm = static_cast<FractureCohesive3DIPVariable*>(ipsm->getState(IPStateBase::previous));
         ipvp = static_cast<FractureCohesive3DIPVariable*>(ipsp->getState(IPStateBase::previous));
         ipvp->setFractureStrengthFactor(ipvm->fractureStrengthFactor());
         ipvm = static_cast<FractureCohesive3DIPVariable*>(ipsm->getState(IPStateBase::current));
         ipvp = static_cast<FractureCohesive3DIPVariable*>(ipsp->getState(IPStateBase::current));
         ipvp->setFractureStrengthFactor(ipvm->fractureStrengthFactor());
       }
     }
   }
}

LinearTermBase<double>*
dG3DDomain::createNeumannTerm(FunctionSpace<double> *spneu,  const groupOfElements* g,const simpleFunctionTime<double>* f,
                                   const unknownField *uf,const IPField * ip, const nonLinearBoundaryCondition::location onWhat,
                                   const nonLinearNeumannBC::NeumannBCType neumann_type, const int comp) const
{
  LinearTermBase<double>*term = NULL;
  if(neumann_type == nonLinearNeumannBC::PRESSURE)
  {
    term = new g3DFiniteStrainsPressureTerm(this,*spneu,uf,f);
  }
  else if(neumann_type == nonLinearNeumannBC::SCALAR_FLUX)
  {
    term = new g3DFiniteStrainsScalarFluxLinearTerm(this,*spneu,uf,std::vector<int>(1,comp),f);
  }
  else if (neumann_type == nonLinearNeumannBC::FORCE)
  {
    term= new g3DLoadTerm(*spneu,f,comp);
  }
  else{
    Msg::Error("nonLinearNeumannBC::NeumannBCType %s is not defined dG3DDomain::createNeumannTerm",neumann_type);
  }
  
  g3DLinearTerm<double>* gterm = static_cast<g3DLinearTerm<double>*>(term);
  if(getNumNonLocalVariable()>0)
  {
    gterm->setNonLocalEqRatio(getNonLocalEqRatio());
    gterm->setNumNonLocalVariable(getNumNonLocalVariable());
  }
  if( getNumConstitutiveExtraDofDiffusionVariable() >0)
  {
    gterm->setNumConstitutiveExtraDofDiffusionVariable(getNumConstitutiveExtraDofDiffusionVariable());
    gterm->setConstitutiveExtraDofDiffusionEqRatio(getConstitutiveExtraDofDiffusionEqRatio());
    gterm->setConstitutiveExtraDofDiffusionUseEnergyConjugatedField(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField());
    gterm->setConstitutiveExtraDofDiffusionAccountFieldSource(getConstitutiveExtraDofDiffusionAccountFieldSource());
    gterm->setConstitutiveExtraDofDiffusionAccountMecaSource(getConstitutiveExtraDofDiffusionAccountMecaSource());
  }

  return term;
}
BilinearTermBase* dG3DDomain::createNeumannMatrixTerm(FunctionSpace<double> *spneu, const groupOfElements* g,
                                                                     const simpleFunctionTime<double>* f, const unknownField *uf,
                                                                     const IPField * ip, const nonLinearBoundaryCondition::location onWhat,
                                                                     const nonLinearNeumannBC::NeumannBCType neumann_type,
                                                                     const int comp) const
{
  BilinearTermBase* bneuterm = NULL;
  if (neumann_type == nonLinearNeumannBC::PRESSURE){
    bneuterm = new g3DFiniteStrainsPressureBilinearTerm(this,*spneu,uf,f);
  }
  else if (neumann_type == nonLinearNeumannBC::SCALAR_FLUX){
    bneuterm =  new g3DFiniteStrainsScalarFluxBiLinearTerm(this,*spneu,*spneu,uf,std::vector<int>(1,comp),f);
  }
  else if (neumann_type == nonLinearNeumannBC::FORCE){
    bneuterm = new BiNonLinearTermVoid();
  }
  else{
    Msg::Error("nonLinearNeumannBC::NeumannBCType %s is not defined dG3DDomain::createNeumannMatrixTerm",neumann_type);
  }
  
  g3DBilinearTerm<double,double>* gterm = dynamic_cast<g3DBilinearTerm<double,double>*>(bneuterm);
  if (gterm != NULL){
    if(getNumNonLocalVariable()>0)
    {
      gterm->setNonLocalEqRatio(getNonLocalEqRatio());
      gterm->setNumNonLocalVariable(getNumNonLocalVariable());
    }
    if( getNumConstitutiveExtraDofDiffusionVariable() >0)
    {
      gterm->setNumConstitutiveExtraDofDiffusionVariable(getNumConstitutiveExtraDofDiffusionVariable());
      gterm->setConstitutiveExtraDofDiffusionEqRatio(getConstitutiveExtraDofDiffusionEqRatio());
      gterm->setConstitutiveExtraDofDiffusionUseEnergyConjugatedField(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField());
      gterm->setConstitutiveExtraDofDiffusionAccountFieldSource(getConstitutiveExtraDofDiffusionAccountFieldSource());
      gterm->setConstitutiveExtraDofDiffusionAccountMecaSource(getConstitutiveExtraDofDiffusionAccountMecaSource());
    }
  }
  return bneuterm;
}



void dG3DDomain::getNoMassDofs(std::set<Dof>& DofVector) const
{
  // Loop on all elements
  int numExtraDof = getNumNonLocalVariable()+getNumConstitutiveExtraDofDiffusionVariable();
  std::vector<Dof> R;
  const partDomain *dom = static_cast<const partDomain*>(this);
  for(groupOfElements::elementContainer::const_iterator it=g->begin(); it!=g->end();++it){
    // get dofs of an element
    MElement *ele = *it;
    R.clear();
    _space->getKeys(ele,R);
    
    int nbFF = ele->getNumShapeFunctions();
    // Loop on all non-local var and shape functions
    for(int nl = 0; nl < numExtraDof; ++nl){
      for(int i = 0; i < nbFF; ++i){
        DofVector.insert(R[i+(3+nl)*nbFF]);
      }
    }
  }
}


void dG3DDomain::computeAllIPStrain(AllIPState *aips,const unknownField *ufield,const IPStateBase::whichState ws, const bool virt, bool useBarF){
  IntPt *GP;
  fullVector<double> dispm;
  fullVector<double> dispp;
  fullVector<double> dispExtra;
  std::vector<Dof> R;
  // interface elements
  partDomain *dom = static_cast<partDomain*>(this);
  for(groupOfElements::elementContainer::const_iterator it=gi->begin(); it!=gi->end();++it){
    MInterfaceElement *ie = dynamic_cast<MInterfaceElement*>(*it);
    R.clear();
    this->getFunctionSpaceMinus()->getKeys(ie->getElem(0),R);
    dispm.resize(R.size());
    ufield->get(R,dispm);
    R.clear();
    this->getFunctionSpacePlus()->getKeys(ie->getElem(1),R);
    dispp.resize(R.size());
    ufield->get(R,dispp);
    int npts = integBound->getIntPoints(*it,&GP);

    if (this->IsMultipleFieldFormulation()){
      FunctionSpaceBase* extraSp = this->getInterfaceFunctionSpace();
      R.clear();
      extraSp->getKeys(*it,R);
      if (R.size()>0){
        dispExtra.resize(R.size());
        ufield->get(R,dispExtra);
      }
    }

    this->computeStrain(aips,ie, GP,ws,dom,dom,dispm,dispp,dispExtra,virt, useBarF);
  }
  // bulk
  for (groupOfElements::elementContainer::const_iterator it = g->begin(); it != g->end(); ++it){
    MElement *e = *it;
    R.clear();
    _space->getKeys(e,R);
    dispm.resize(R.size());
    ufield->get(R,dispm);
    IntPt *GP;
    int npts_bulk=integBulk->getIntPoints(e,&GP);
    this->computeStrain(e,npts_bulk,GP,aips->getIPstate(e->getNum()),ws,dispm, useBarF);
  }
};


void dG3DDomain::computeIPVariable(AllIPState *aips,const unknownField *ufield, const IPStateBase::whichState ws, bool stiff)
{
  if (ws == IPStateBase::initial){
    this->initialIPVariable(aips,ufield,ws,stiff);
    return;
  }

  _evalStiff = false; // directly access to computeIPv is for matrix by perturbation otherwise compute ipvarible thanks to this function
  #if defined(HAVE_MPI)
  if (_mlaw->isNumeric() and _otherRanks.size()>0){
    dG3DMaterialLaw *mlaw;
    if(_mlaw->getType() == materialLaw::fracture)
      mlaw = static_cast< dG3DMaterialLaw* >((static_cast< FractureByCohesive3DLaw* > (_mlaw) )->getBulkLaw());
    else{
      mlaw= static_cast<dG3DMaterialLaw*>(_mlaw);
    }
    bool useBarF=mlaw->getUseBarF();    
    this->computeIPVariableMPI(aips,ufield,ws,stiff,useBarF);
  }
  else
  #endif //HAVE_MPI
  {
    dG3DMaterialLaw *mlaw;
    if(_mlaw->getType() == materialLaw::fracture)
      mlaw = static_cast< dG3DMaterialLaw* >((static_cast< FractureByCohesive3DLaw* > (_mlaw) )->getBulkLaw());
    else{
      mlaw= static_cast<dG3DMaterialLaw*>(_mlaw);
    }
    bool useBarF=mlaw->getUseBarF();
    this->computeAllIPStrain(aips,ufield,ws,false, useBarF);
    IntPt *GP;
    double uem,vem,uep,vep;
    fullVector<double> dispm;
    fullVector<double> dispp;
    fullVector<double> dispExtra;
    std::vector<Dof> R;
    // interface elements
    partDomain *dom = static_cast<partDomain*>(this);
    for(groupOfElements::elementContainer::const_iterator it=gi->begin(); it!=gi->end();++it){
      MInterfaceElement *ie = dynamic_cast<MInterfaceElement*>(*it);
      R.clear();
      this->getFunctionSpaceMinus()->getKeys(ie->getElem(0),R);
      dispm.resize(R.size());
      ufield->get(R,dispm);
      R.clear();
      this->getFunctionSpacePlus()->getKeys(ie->getElem(1),R);
      dispp.resize(R.size());
      ufield->get(R,dispp);

      if (this->IsMultipleFieldFormulation()){
        FunctionSpaceBase* extraSp = this->getInterfaceFunctionSpace();
        R.clear();
        extraSp->getKeys(*it,R);
        if (R.size()>0){
          dispExtra.resize(R.size());
          ufield->get(R,dispExtra);
        }
      }

      int npts = integBound->getIntPoints(*it,&GP);
      this->computeIpv(aips,ie, GP,ws,dom,dom,_mlaw,_mlaw,dispm,dispp,dispExtra,false,stiff);
    }
    // bulk
    for (groupOfElements::elementContainer::const_iterator it = g->begin(); it != g->end(); ++it){
      MElement *e = *it;
      R.clear();
      _space->getKeys(e,R);
      dispm.resize(R.size());
      ufield->get(R,dispm);
      this->computeIpv(aips,e,ws,_mlaw,dispm,stiff);
    }

  }
  _evalStiff = true;
}

void dG3DDomain::computeStrain(MElement *e, const int npts_bulk, IntPt *GP,
                               AllIPState::ipstateElementContainer *vips,
                               IPStateBase::whichState ws,
                               fullVector<double> &disp, bool useBarF) const
{
  if (!getElementErosionFilter()(e)) return;

  int nbFF = e->getNumShapeFunctions();
  //Msg::Info("nb bulk gp %d",npts_bulk);
  for(int j=0;j<npts_bulk;j++){
    IPStateBase* ips = (*vips)[j];

    dG3DIPVariableBase *ipv = static_cast<dG3DIPVariableBase*>(ips->getState(ws));
    // get grad of shape function from Gauss point
    const std::vector<TensorialTraits<double>::ValType> &Vals = ipv->f(_space,e,GP[j]);
    const std::vector<TensorialTraits<double>::GradType> &Grads = ipv->gradf(_space,e,GP[j]);
    //x y z
    STensor3 &F = ipv->getRefToDeformationGradient();
    STensorOperation::unity(F);
    for (int i = 0; i < nbFF; i++)
    {
        //x y z: take compoenent x arbitrarily
        // xx yy zz xy xz yz
        F(0,0) += Grads[i+0*nbFF][0]*disp(i);
        F(0,1) += Grads[i+0*nbFF][1]*disp(i);
        F(0,2) += Grads[i+0*nbFF][2]*disp(i);
        F(1,0) += Grads[i+0*nbFF][0]*disp(i+nbFF);
        F(1,1) += Grads[i+0*nbFF][1]*disp(i+nbFF);
        F(1,2) += Grads[i+0*nbFF][2]*disp(i+nbFF);
        F(2,0) += Grads[i+0*nbFF][0]*disp(i+2*nbFF);
        F(2,1) += Grads[i+0*nbFF][1]*disp(i+2*nbFF);
        F(2,2) += Grads[i+0*nbFF][2]*disp(i+2*nbFF);
    }
    if(STensorOperation::determinantSTensor3(F) < 1.e-15) Msg::Error("Negative Jacobian in ele %d",e->getNum());

    if( ipv->getNumberNonLocalVariable()>getNumNonLocalVariable() )
      Msg::Error("Your material law uses more non local variables than your domain dG3DDomain::computeStrain bulk");
    for (int nl = 0; nl < ipv->getNumberNonLocalVariable(); nl++){
      double& nonlocalVar = ipv->getRefToNonLocalVariable(nl);
      SVector3& gradNonlocalVar = ipv->getRefToGradNonLocalVariable()[nl];

      nonlocalVar = 0.;
      STensorOperation::zero(gradNonlocalVar);
      for (int i = 0; i < nbFF; i++)
      {
        nonlocalVar += Vals[i+3*nbFF+nl*nbFF]*disp(i+3*nbFF+nl*nbFF);
        gradNonlocalVar(0) += Grads[i+3*nbFF+nl*nbFF][0]*disp(i+3*nbFF+nl*nbFF);
        gradNonlocalVar(1) += Grads[i+3*nbFF+nl*nbFF][1]*disp(i+3*nbFF+nl*nbFF);
        gradNonlocalVar(2) += Grads[i+3*nbFF+nl*nbFF][2]*disp(i+3*nbFF+nl*nbFF);
      }
    }
    if( ipv->getNumConstitutiveExtraDofDiffusionVariable()>getNumConstitutiveExtraDofDiffusionVariable() ) 
      Msg::Error("Your material law uses more constitutive extra dof diffusion variables than your domain dG3DDomain::computeStrain bulk");
    for (int extraDOFField = 0; extraDOFField < ipv->getNumConstitutiveExtraDofDiffusionVariable(); extraDOFField++){
        
      	double& T = ipv->getRefToField(extraDOFField);
      	SVector3& gradT = ipv->getRefToGradField()[extraDOFField]; 	
      	T = 0.;
      	STensorOperation::zero(gradT);

	      for (int i = 0; i < nbFF; i++)
      		{
        	  int indexField=3+getNumNonLocalVariable()+extraDOFField;
  	           T += Vals[i+indexField*nbFF]*disp(i+indexField*nbFF);
	           gradT(0) += Grads[i+indexField*nbFF][0]*disp(i+indexField*nbFF);
                   gradT(1) += Grads[i+indexField*nbFF][1]*disp(i+indexField*nbFF);
                   gradT(2) += Grads[i+indexField*nbFF][2]*disp(i+indexField*nbFF);
    	        }
        
        if((getNumConstitutiveExtraDofDiffusionVariable() == 2)&&(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField()))
        {
         double& energyConjugatedField = ipv->getRefToEnergyConjugatedField(extraDOFField);
         SVector3& gradEnergyConjugatedField = ipv->getRefToGradEnergyConjugatedField()[extraDOFField];
         energyConjugatedField = 0.;
         STensorOperation::zero(gradEnergyConjugatedField);

         int extraDOFField_aux=0;
             if((extraDOFField+1)< ipv->getNumConstitutiveExtraDofDiffusionVariable()) extraDOFField_aux=extraDOFField+1;
             else extraDOFField_aux=extraDOFField-1;  

         double& dFielddEnergyConjugatedField = ipv->getRefTodFielddEnergyConjugatedField()[extraDOFField][extraDOFField];
         double& dFielddEnergyConjugatedField2 = ipv->getRefTodFielddEnergyConjugatedField()[extraDOFField][extraDOFField_aux];
         SVector3& dGradFielddEnergyConjugatedField = ipv->getRefTodGradFielddEnergyConjugatedField()[extraDOFField][extraDOFField];
         SVector3& dGradFielddEnergyConjugatedField2 = ipv->getRefTodGradFielddEnergyConjugatedField()[extraDOFField][extraDOFField_aux];
         STensor3& dGradFielddGradEnergyConjugatedField = ipv->getRefTodGradFielddGradEnergyConjugatedField()[extraDOFField][extraDOFField];
         STensor3& dGradFielddGradEnergyConjugatedField2 = ipv->getRefTodGradFielddGradEnergyConjugatedField()[extraDOFField][extraDOFField_aux];
 
         STensorOperation::zero(dGradFielddGradEnergyConjugatedField); STensorOperation::zero(dGradFielddGradEnergyConjugatedField2);

         int indexFieldfT=3+getNumNonLocalVariable()+extraDOFField;

          for (int i = 0; i < nbFF; i++)
           {
             const double& ftemp = disp(i+indexFieldfT*nbFF);
              energyConjugatedField += Vals[i+indexFieldfT*nbFF]*ftemp;
              gradEnergyConjugatedField(0) += Grads[i+indexFieldfT*nbFF][0]*ftemp;
              gradEnergyConjugatedField(1) += Grads[i+indexFieldfT*nbFF][1]*ftemp;
              gradEnergyConjugatedField(2) += Grads[i+indexFieldfT*nbFF][2]*ftemp;
            }
         
           if(extraDOFField == 0)
            {
             T = 1/(energyConjugatedField);

             gradT(0)=-1/(pow(energyConjugatedField,2))*(gradEnergyConjugatedField(0));
             gradT(1)=-1/(pow(energyConjugatedField,2))*(gradEnergyConjugatedField(1));
             gradT(2)=-1/(pow(energyConjugatedField,2))*(gradEnergyConjugatedField(2));
  
             dFielddEnergyConjugatedField  = -1./(energyConjugatedField)/(energyConjugatedField);
             dFielddEnergyConjugatedField2 = 0.;

             dGradFielddEnergyConjugatedField(0)=2/(energyConjugatedField)/(energyConjugatedField)/(energyConjugatedField)*(gradEnergyConjugatedField(0));
             dGradFielddEnergyConjugatedField(1)=2/(energyConjugatedField)/(energyConjugatedField)/(energyConjugatedField)*(gradEnergyConjugatedField(1));
             dGradFielddEnergyConjugatedField(2)=2/(energyConjugatedField)/(energyConjugatedField)/(energyConjugatedField)*(gradEnergyConjugatedField(2));

             dGradFielddEnergyConjugatedField2(0)=0.;
             dGradFielddEnergyConjugatedField2(1)=0.;
             dGradFielddEnergyConjugatedField2(2)=0.;
             
             STensorOperation::diag(dGradFielddGradEnergyConjugatedField,-1./(energyConjugatedField)/(energyConjugatedField));

            }
           else
            {
            
             double fT = ipv->getConstRefToEnergyConjugatedField(0);
             SVector3 gradfT = ipv->getConstRefToGradEnergyConjugatedField()[0];
             
             T = -1*(energyConjugatedField)/(fT);

             gradT(0)= -1/(fT)*(gradEnergyConjugatedField(0))+(energyConjugatedField)/(pow(fT,2))*(gradfT(0));
             gradT(1)= -1/(fT)*(gradEnergyConjugatedField(1))+(energyConjugatedField)/(pow(fT,2))*(gradfT(1));
             gradT(2)= -1/(fT)*(gradEnergyConjugatedField(2))+(energyConjugatedField)/(pow(fT,2))*(gradfT(2));

             dFielddEnergyConjugatedField=-1./(fT);
             dFielddEnergyConjugatedField2=(energyConjugatedField)/(fT)/(fT);

             dGradFielddEnergyConjugatedField(0)=1./(fT)/(fT)*(gradfT(0));
             dGradFielddEnergyConjugatedField(1)=1./(fT)/(fT)*(gradfT(1));
             dGradFielddEnergyConjugatedField(2)=1./(fT)/(fT)*(gradfT(2));

             dGradFielddEnergyConjugatedField2(0)=1./(fT)/(fT)*(gradEnergyConjugatedField(0))-2*(energyConjugatedField)/(fT)/(fT)/(fT)*(gradfT(0));
             dGradFielddEnergyConjugatedField2(1)=1./(fT)/(fT)*(gradEnergyConjugatedField(1))-2*(energyConjugatedField)/(fT)/(fT)/(fT)*(gradfT(1));
             dGradFielddEnergyConjugatedField2(2)=1./(fT)/(fT)*(gradEnergyConjugatedField(2))-2*(energyConjugatedField)/(fT)/(fT)/(fT)*(gradfT(2));
    
             STensorOperation::diag(dGradFielddGradEnergyConjugatedField,-1./(fT));
             STensorOperation::diag(dGradFielddGradEnergyConjugatedField2,(energyConjugatedField)/(fT)/(fT));


            }
        }
    }
  }

  if(useBarF)
  {
    double barJ=0.;
    double totalWeight=0.;
    for(int j=0;j<npts_bulk;j++){
      IPStateBase* ips = (*vips)[j];
      dG3DIPVariableBase *ipv = static_cast<dG3DIPVariableBase*>(ips->getState(ws));
      double J=ipv->getJ();
      ipv->getRefToLocalJ()=J;
      double weight = GP[j].weight;
      barJ+=J*weight;
      totalWeight+=weight;
    }
    barJ/=totalWeight;
    for(int j=0;j<npts_bulk;j++){
      IPStateBase* ips = (*vips)[j];
      dG3DIPVariableBase *ipv = static_cast<dG3DIPVariableBase*>(ips->getState(ws));
      double localJ=ipv->getLocalJ();
      ipv->getRefToBarJ()=barJ;
      STensor3 &barF = ipv->getRefToDeformationGradient();
      barF*=pow(barJ/localJ,1./3.);
    }
  }
}
void dG3DDomain::computeIpv(AllIPState *aips,MElement *e, IPStateBase::whichState ws,
                                        materialLaw *mlaw__,fullVector<double> &disp, bool stiff)
{
  if (!getElementErosionFilter()(e)) return;

  dG3DMaterialLaw *mlaw;
  AllIPState::ipstateElementContainer *vips = aips->getIPstate(e->getNum());
  IntPt *GP;
  int npts_bulk=integBulk->getIntPoints(e,&GP);

  if(mlaw__->getType() == materialLaw::fracture)
    mlaw = static_cast< dG3DMaterialLaw* >((static_cast< FractureByCohesive3DLaw* > (mlaw__) )->getBulkLaw());
  else{
    mlaw= static_cast<dG3DMaterialLaw*>(mlaw__);
  }
  bool useBarF=mlaw->getUseBarF();
  if(_evalStiff) { this->computeStrain(e,npts_bulk,GP,vips,ws,disp,useBarF);}

  for(int j=0;j<npts_bulk;j++)
  {
    // get dgipv from ipstate
    IPStateBase* ips = (*vips)[j];
    dG3DIPVariableBase *ipv = static_cast<dG3DIPVariableBase*>(ips->getState(ws));
    const dG3DIPVariableBase *ipvprev = static_cast<const dG3DIPVariableBase*>(ips->getState(IPStateBase::previous));

    if ((_subSteppingMethod == 0) or (_subSteppingMethod == 1))
    {
      if (_planeStressState)
        mlaw->stress3DTo2D(ipv,ipvprev,stiff,false);
      else
      	mlaw->stress(ipv,ipvprev, stiff);
    }
    else if (_subSteppingMethod == 2)
    {
        // get
        dG3DIPVariableBase* dgipv = static_cast<dG3DIPVariableBase*>(ipv);
        const dG3DIPVariableBase* dgipvprev = static_cast<const dG3DIPVariableBase*>(ipvprev);

        // create a temporary ipv
        IPVariable* ipvTemp =  dgipvprev->clone();
        dG3DIPVariableBase* dgipvTemp = dynamic_cast<dG3DIPVariableBase*>(ipvTemp);
        if (dgipvTemp == NULL) Msg::Error("dG3DIPVariableBase::clone is not correctly defined");

        // compute dF
        static STensor3 dF;
        dF  = dgipv->getConstRefToDeformationGradient();
        dF -= dgipvprev->getConstRefToDeformationGradient();
        // compute dNLV if nonlocal damage is used
        fullVector<double> dnonlocalVar;
        int numNL = dgipv->getNumberNonLocalVariable();
        if (numNL > 0){
          dnonlocalVar.resize(numNL);
          for (int iv = 0; iv < numNL; iv++){
            dnonlocalVar(iv) = (dgipv->getConstRefToNonLocalVariable(iv) - dgipvprev->getConstRefToNonLocalVariable(iv));
          }
        }

        int iter = 1;
        while (1){
          // update kinematics
          double time = double(iter)/(double(_subStepNumber));
          //Msg::Info("step %d, time = %f",iter,time);
          STensor3& F = dgipv->getRefToDeformationGradient();
          F = dgipvprev->getConstRefToDeformationGradient();
          F.daxpy(dF,time);

          for (int iv=0; iv < numNL; iv++){
            dgipv->getRefToNonLocalVariable(iv) = dgipvprev->getConstRefToNonLocalVariable(iv);
            dgipv->getRefToNonLocalVariable(iv) += (time*dnonlocalVar(iv));
          }

          if (iter == _subStepNumber){
            // in last substep
            if (_planeStressState)
              mlaw->stress3DTo2D(ipv,ipvTemp,stiff,false);
            else
              mlaw->stress(ipv,ipvTemp,stiff);
            break;
          }
          else
          {
            if (_planeStressState)
              mlaw->stress3DTo2D(ipv,ipvTemp,false,false);
            else
              mlaw->stress(ipv,ipvTemp,false);
            dgipvTemp->operator=(*(dynamic_cast<const IPVariable*>(ipv)));
          }
          // next step
          iter++;
        }
        if (ipvTemp!=NULL) delete ipvTemp;
    }
    else
    {
      Msg::Error("This substepping method %d has not been implemented",_subSteppingMethod);
    }
  }
}

void dG3DDomain::computeStrain(AllIPState *aips,MInterfaceElement *ie, IntPt *GP,const IPStateBase::whichState ws,
                                        partDomain* efMinus, partDomain *efPlus, fullVector<double> &dispm,
                                        fullVector<double> &dispp, fullVector<double>& dispExtra, const bool virt,
                                        bool useBarF)
{

  MElement *ele = dynamic_cast<MElement*>(ie);
  if (!getElementErosionFilter()(ele)) return;

  if(!virt){
    FunctionSpaceBase* _spaceminus = efMinus->getFunctionSpace();
    FunctionSpaceBase* _spaceplus = efPlus->getFunctionSpace();

    int npts=integBound->getIntPoints(ele,&GP);

    MElement *em = ie->getElem(0);
    MElement *ep = ie->getElem(1);

     // gauss point
    int nbFFm = em->getNumVertices();
    int nbFFp = ep->getNumVertices();
    // Get Gauss points values on minus and plus elements
    IntPt *GPm; IntPt *GPp;
    _interQuad->getIntPoints(ie,GP,&GPm,&GPp);

    // the current outward normal is computed at the mean surface
    // which defined as the reference coordinates plus mean displacement of positive and negative part
    // displacement at the mean face
    fullVector<double> dispinter, jumpinter;
    int nbdofinter = _space->getNumKeys(ele);
    int nbvertexInter = ele->getNumVertices();
    std::vector<int> vn;

    vn.resize(nbvertexInter);
    ie->getLocalVertexNum(0,vn);
    dispinter.resize(nbdofinter);
    jumpinter.resize(nbdofinter);
    for(int i=0;i<nbvertexInter;i++){
      dispinter(i) = 0.5*dispm(vn[i]);
      dispinter(i+nbvertexInter) = 0.5*dispm(vn[i]+nbFFm);
      dispinter(i+2*nbvertexInter) = 0.5*dispm(vn[i]+2*nbFFm);
      jumpinter(i) = -dispm(vn[i]);
      jumpinter(i+nbvertexInter) = -dispm(vn[i]+nbFFm);
      jumpinter(i+2*nbvertexInter) = -dispm(vn[i]+2*nbFFm);
    }

    std::vector<int> vp;
    vp.resize(nbvertexInter);
    ie->getLocalVertexNum(1,vp);
    for(int i=0;i<nbvertexInter;i++){
      dispinter(i) += 0.5*dispp(vp[i]);
      dispinter(i+nbvertexInter) += 0.5*dispp(vp[i]+nbFFp);
      dispinter(i+2*nbvertexInter) += 0.5*dispp(vp[i]+2*nbFFp);
      jumpinter(i) += dispp(vp[i]);
      jumpinter(i+nbvertexInter) += dispp(vp[i]+nbFFp);
      jumpinter(i+2*nbvertexInter) += dispp(vp[i]+2*nbFFp);
    }

    nlsFunctionSpaceUVW<double> *spinter = static_cast<nlsFunctionSpaceUVW<double>*>(_spaceInterUVW);
    // Get grad shape functions values at all Gauss points in order to compute normal to interfaces
    std::vector<std::vector<TensorialTraits<double>::GradType> > vgrads;
    spinter->getgradf(ele,npts,GP,vgrads);


    AllIPState::ipstateElementContainer *vips = aips->getIPstate(ele->getNum());
    for(int j=0;j<npts;j++){

      IPStateBase* ipsm = (*vips)[j];
      IPStateBase* ipsp = (*vips)[j+npts];

      dG3DIPVariableBase* ipvm = static_cast<dG3DIPVariableBase*>(ipsm->getState(ws));
      dG3DIPVariableBase* ipvp = static_cast<dG3DIPVariableBase*>(ipsp->getState(ws));

      // get shape function gradients
      const std::vector<TensorialTraits<double>::GradType> &Gradm = ipvm->gradf(_spaceminus,em,GPm[j]);
      const std::vector<TensorialTraits<double>::GradType> &Gradp = ipvp->gradf(_spaceplus,ep,GPp[j]);
      const std::vector<TensorialTraits<double>::ValType> &Valm = ipvm->f(_spaceminus,em,GPm[j]);
      const std::vector<TensorialTraits<double>::ValType> &Valp = ipvp->f(_spaceplus,ep,GPp[j]);

      STensor3& Fm = ipvm->getRefToDeformationGradient();
      STensorOperation::unity(Fm);
      for (int i = 0; i < nbFFm; i++)
      {
        //x y z: take compoenent x arbitrarily
        // xx yy zz xy xz yz
        Fm(0,0) += Gradm[i+0*nbFFm][0]*dispm(i);
        Fm(0,1) += Gradm[i+0*nbFFm][1]*dispm(i);
        Fm(0,2) += Gradm[i+0*nbFFm][2]*dispm(i);
        Fm(1,0) += Gradm[i+0*nbFFm][0]*dispm(i+nbFFm);
        Fm(1,1) += Gradm[i+0*nbFFm][1]*dispm(i+nbFFm);
        Fm(1,2) += Gradm[i+0*nbFFm][2]*dispm(i+nbFFm);
        Fm(2,0) += Gradm[i+0*nbFFm][0]*dispm(i+2*nbFFm);
        Fm(2,1) += Gradm[i+0*nbFFm][1]*dispm(i+2*nbFFm);
        Fm(2,2) += Gradm[i+0*nbFFm][2]*dispm(i+2*nbFFm);
     }
     if(STensorOperation::determinantSTensor3(Fm) < 1.e-15) Msg::Error("Negative Jacobian Fm, em = %d, ele=%d",em->getNum(),ele->getNum());

     STensor3& Fp = ipvp->getRefToDeformationGradient();
     STensorOperation::unity(Fp);
     for (int i = 0; i < nbFFp; i++)
     {
        //x y z: take compoenent x arbitrarily
        // xx yy zz xy xz yz
        Fp(0,0) += Gradp[i+0*nbFFp][0]*dispp(i);
        Fp(0,1) += Gradp[i+0*nbFFp][1]*dispp(i);
        Fp(0,2) += Gradp[i+0*nbFFp][2]*dispp(i);
        Fp(1,0) += Gradp[i+0*nbFFp][0]*dispp(i+nbFFp);
        Fp(1,1) += Gradp[i+0*nbFFp][1]*dispp(i+nbFFp);
        Fp(1,2) += Gradp[i+0*nbFFp][2]*dispp(i+nbFFp);
        Fp(2,0) += Gradp[i+0*nbFFp][0]*dispp(i+2*nbFFp);
        Fp(2,1) += Gradp[i+0*nbFFp][1]*dispp(i+2*nbFFp);
        Fp(2,2) += Gradp[i+0*nbFFp][2]*dispp(i+2*nbFFp);
     }
     if(STensorOperation::determinantSTensor3(Fp) < 1.e-15) Msg::Error("Negative Jacobian Fp, ep =%d ele = %d",ep->getNum(),ele->getNum());
    }
    if(useBarF)
    {
      double barJm=0.;
      double barJp=0.;
      double totalWeight=0.;
      for(int j=0;j<npts;j++){
        IPStateBase* ipsm = (*vips)[j];
        IPStateBase* ipsp = (*vips)[j+npts];
        double weight = GP[j].weight;
        totalWeight+=weight;
        dG3DIPVariableBase* ipvm = static_cast<dG3DIPVariableBase*>(ipsm->getState(ws));
        dG3DIPVariableBase* ipvp = static_cast<dG3DIPVariableBase*>(ipsp->getState(ws));
 
        double Jm=ipvm->getJ();
        ipvm->getRefToLocalJ()=Jm;
        barJm+=Jm*weight;
        double Jp=ipvp->getJ();
        ipvp->getRefToLocalJ()=Jp;
        barJp+=Jp*weight;
      }
      barJm/=totalWeight;
      barJp/=totalWeight;
      for(int j=0;j<npts;j++){
        IPStateBase* ipsm = (*vips)[j];
        IPStateBase* ipsp = (*vips)[j+npts];

        dG3DIPVariableBase* ipvm = static_cast<dG3DIPVariableBase*>(ipsm->getState(ws));
        dG3DIPVariableBase* ipvp = static_cast<dG3DIPVariableBase*>(ipsp->getState(ws));

        double localJm=ipvm->getLocalJ();
        ipvm->getRefToBarJ()=barJm;
        STensor3 &barFm = ipvm->getRefToDeformationGradient();
        barFm*=pow(barJm/localJm,1./3.);

        double localJp=ipvp->getLocalJ();
        ipvp->getRefToBarJ()=barJp;
        STensor3 &barFp = ipvp->getRefToDeformationGradient();
        barFp*=pow(barJp/localJp,1./3.);
      }
    }
    for(int j=0;j<npts;j++){

      IPStateBase* ipsm = (*vips)[j];
      IPStateBase* ipsp = (*vips)[j+npts];

      dG3DIPVariableBase* ipvm = static_cast<dG3DIPVariableBase*>(ipsm->getState(ws));
      dG3DIPVariableBase* ipvp = static_cast<dG3DIPVariableBase*>(ipsp->getState(ws));

      // get shape function gradients
      const std::vector<TensorialTraits<double>::GradType> &Gradm = ipvm->gradf(_spaceminus,em,GPm[j]);
      const std::vector<TensorialTraits<double>::GradType> &Gradp = ipvp->gradf(_spaceplus,ep,GPp[j]);
      const std::vector<TensorialTraits<double>::ValType> &Valm = ipvm->f(_spaceminus,em,GPm[j]);
      const std::vector<TensorialTraits<double>::ValType> &Valp = ipvp->f(_spaceplus,ep,GPp[j]);

      STensor3& Fm = ipvm->getRefToDeformationGradient();
      STensor3& Fp = ipvp->getRefToDeformationGradient();

      if (_averageStrainBased and (efMinus->getMaterialLaw()->getNum() == efPlus->getMaterialLaw()->getNum())){
       // compute Fm as average 0.5*(Fm+Fp)
       // assign Fp = Fm
       Fm += Fp;
       Fm *= (0.5);
       Fp = Fm;
       if(useBarF)
         Msg::Error("The setBulkDeformationGradient does not allow to recover the local one for the derivatives");
      }

      const std::vector<TensorialTraits<double>::GradType> &Grads = vgrads[j];
      // minus outward normal
      static SVector3 referencePhi0[2];
      static SVector3 currentPhi0[2];

      // zero vector first
      STensorOperation::zero(referencePhi0[0]);
      STensorOperation::zero(referencePhi0[1]);
      STensorOperation::zero(currentPhi0[0]);
      STensorOperation::zero(currentPhi0[1]);

      if (efMinus->getDim() == 3){
        for(int k=0;k<nbvertexInter;k++)
        {
          double x = ele->getVertex(k)->x();
          double y = ele->getVertex(k)->y();
          double z = ele->getVertex(k)->z();
          //evaulate displacements at interface
          for(int i=0;i<2;i++)
          {
           referencePhi0[i](0) += Grads[k](i)*x;
           referencePhi0[i](1) += Grads[k+nbvertexInter](i)*y;
           referencePhi0[i](2) += Grads[k+nbvertexInter+nbvertexInter](i)*z;
           currentPhi0[i](0) += Grads[k](i)*(x+dispinter(k));
           currentPhi0[i](1) += Grads[k+nbvertexInter](i)*(y+dispinter(k+nbvertexInter));
           currentPhi0[i](2) += Grads[k+nbvertexInter+nbvertexInter](i)*(z+dispinter(k+2*nbvertexInter));
          }
        }
      }
      else if (efMinus->getDim() == 2){
        for(int k=0;k<nbvertexInter;k++)
        {
          double x = ele->getVertex(k)->x();
          double y = ele->getVertex(k)->y();
          double z = ele->getVertex(k)->z();
           referencePhi0[0](0) += Grads[k](0)*x;
           referencePhi0[0](1) += Grads[k+nbvertexInter](0)*y;
           referencePhi0[0](2) += Grads[k+nbvertexInter+nbvertexInter](0)*z;
           currentPhi0[0](0) += Grads[k](0)*(x+dispinter(k));
           currentPhi0[0](1) += Grads[k+nbvertexInter](0)*(y+dispinter(k+nbvertexInter));
           currentPhi0[0](2) += Grads[k+nbvertexInter+nbvertexInter](0)*(z+dispinter(k+2*nbvertexInter));
        }

        static SVector3 phi0[2];
        static SVector3 phi0Cur[2];

        STensorOperation::zero(phi0[0]);
        STensorOperation::zero(phi0[1]);

        STensorOperation::zero(phi0Cur[0]);
        STensorOperation::zero(phi0Cur[1]);

        FunctionSpace<double>* spaceMinus = dynamic_cast<FunctionSpace<double>*> (_spaceminus);

        std::vector<TensorialTraits<double>::GradType> gradmUVW;
        spaceMinus->gradfuvw(em,GPm[j].pt[0],GPm[j].pt[1],GPm[j].pt[2],gradmUVW);
        for(int k=0;k<nbFFm;k++)
        {
          double x = em->getVertex(k)->x();
          double y = em->getVertex(k)->y();
          double z = em->getVertex(k)->z();
          //evaulate displacements at interface
          for(int i=0;i<2;i++)
          {
           phi0[i](0) += gradmUVW[k](i)*x;
           phi0[i](1) += gradmUVW[k+nbFFm](i)*y;
           phi0[i](2) += gradmUVW[k+nbFFm+nbFFm](i)*z;
           phi0Cur[i](0) += gradmUVW[k](i)*(x+dispm(k));
           phi0Cur[i](1) += gradmUVW[k+nbFFm](i)*(y+dispm(k+nbFFm));
           phi0Cur[i](2) += gradmUVW[k+nbFFm+nbFFm](i)*(z+dispm(k+2*nbFFm));
          }
        }

        currentPhi0[1] = crossprod(phi0Cur[0],phi0Cur[1]);
        currentPhi0[1].normalize();
        referencePhi0[1] = crossprod(phi0[0],phi0[1]);
        referencePhi0[1].normalize();
      }


      (ipvm->getRefToCurrentOutwardNormal())   = crossprod(currentPhi0[0],currentPhi0[1]);
      (ipvm->getRefToReferenceOutwardNormal()) = crossprod(referencePhi0[0],referencePhi0[1]);

      (ipvp->getRefToCurrentOutwardNormal())   = crossprod(currentPhi0[0],currentPhi0[1]);
      (ipvp->getRefToReferenceOutwardNormal()) = crossprod(referencePhi0[0],referencePhi0[1]);


      SVector3& ujump = ipvm->getRefToJump();
      computeJump(Valm,nbFFm,Valp,nbFFp,dispm,dispp,ujump);
      ipvp->getRefToJump() = ujump;

      FractureCohesive3DIPVariable *ipvfm = dynamic_cast<FractureCohesive3DIPVariable*>(ipvm);
      FractureCohesive3DIPVariable *ipvfp = dynamic_cast<FractureCohesive3DIPVariable*>(ipvp);

      if((ipvfm!=NULL) and (ipvfp!= NULL) ){

        Cohesive3DIPVariableBase* cohipvfm = ipvfm->getIPvFrac();
        Cohesive3DIPVariableBase* cohipvfp = ipvfp->getIPvFrac();

        // store bulk deformation gradient at interface IP
        if (cohipvfm->withDeformationGradient() and cohipvfp->withDeformationGradient()){
          cohipvfm->setBulkDeformationGradient(Fm);
          cohipvfp->setBulkDeformationGradient(Fp);
        }

        // Compute spatial gradient of jump on interface at interface IP
        if (cohipvfm->withJumpGradient() and cohipvfp->withJumpGradient()){
          const std::vector<TensorialTraits<double>::GradType>& gradVal = ipvm->gradf(_extraSpace,ie,GP[j]);
          STensor3& Gradjumpm =  cohipvfm->getRefToJumpGradient();
          Gradjumpm *= 0.;
          for(int k=0; k<3; k++)
          {
            for(int l=0; l<nbvertexInter; l++)
            {
              Gradjumpm(0,k) += jumpinter(l)*gradVal[l](k);
              Gradjumpm(1,k) += jumpinter(l+nbvertexInter)*gradVal[l+nbvertexInter](k);
              Gradjumpm(2,k) += jumpinter(l+2*nbvertexInter)*gradVal[l+2*nbvertexInter](k);
            }
          }
          cohipvfp->getRefToJumpGradient() = Gradjumpm;
        };


        if (cohipvfm->useLocalBasis()){
          // for cohesiev band model
          cohipvfm->getRefToCohesiveNormal() =ipvm->getConstRefToCurrentOutwardNormal();
          cohipvfm->getRefToCohesiveTangent() = currentPhi0[0];
          cohipvfm->getRefToCohesiveBiTangent() = crossprod(cohipvfm->getConstRefToCohesiveNormal(),currentPhi0[0]);

          cohipvfm->getRefToCohesiveNormal().normalize();
          cohipvfm->getRefToCohesiveTangent().normalize();
          cohipvfm->getRefToCohesiveBiTangent().normalize();

          cohipvfm->getRefToCohesiveReferenceNormal() =ipvm->getConstRefToReferenceOutwardNormal();
          cohipvfm->getRefToCohesiveReferenceTangent() = referencePhi0[0];
          cohipvfm->getRefToCohesiveReferenceBiTangent() = crossprod(cohipvfm->getConstRefToCohesiveReferenceNormal(),referencePhi0[0]);

          cohipvfm->getRefToCohesiveReferenceNormal().normalize();
          cohipvfm->getRefToCohesiveReferenceTangent().normalize();
          cohipvfm->getRefToCohesiveReferenceBiTangent().normalize();
        }

        if (cohipvfp->useLocalBasis()){
          cohipvfp->getRefToCohesiveNormal() =ipvp->getConstRefToCurrentOutwardNormal();
          cohipvfp->getRefToCohesiveTangent() = currentPhi0[0];
          cohipvfp->getRefToCohesiveBiTangent() = crossprod(cohipvfp->getConstRefToCohesiveNormal(),currentPhi0[0]);

          cohipvfp->getRefToCohesiveNormal().normalize();
          cohipvfp->getRefToCohesiveTangent().normalize();
          cohipvfp->getRefToCohesiveBiTangent().normalize();

          cohipvfp->getRefToCohesiveReferenceNormal() =ipvp->getConstRefToReferenceOutwardNormal();
          cohipvfp->getRefToCohesiveReferenceTangent() = referencePhi0[0];
          cohipvfp->getRefToCohesiveReferenceBiTangent() = crossprod(cohipvfp->getConstRefToCohesiveReferenceNormal(),referencePhi0[0]);

          cohipvfp->getRefToCohesiveReferenceNormal().normalize();
          cohipvfp->getRefToCohesiveReferenceTangent().normalize();
          cohipvfp->getRefToCohesiveReferenceBiTangent().normalize();
        }
      }

      if (IsMultipleFieldFormulation()){
        // compute extra field at gauss points
        const std::vector<TensorialTraits<double>::ValType>& Vals = ipvm->f(_extraSpace,ie,GP[j]);
        extraFieldFunctionSpace* extraStp = static_cast<extraFieldFunctionSpace*>(_extraSpace);
        int numExtraKey = extraStp->getNumExtraKeysPerVertex();
        fullVector<double> extraVal;
        extraVal.resize(numExtraKey,true);
        for (int id = 0; id < numExtraKey; id++){
          if (this->getEnrichedUnknownLocation() == partDomain::NODE){
            for (int i=0; i< nbvertexInter; i++){
							extraVal(id) += Vals[i]*dispExtra(i+id*nbvertexInter);
						}
          }
          else if (this->getEnrichedUnknownLocation() == partDomain::GP){
            extraVal(id) = dispExtra(j+id*npts);
          }
        }

        double hs = ie->getCharacteristicSize();
        if (this->getMultipleFieldMethod() == dgPartDomain::TWOFIELD){
          SVector3& iJump = ipvm->getRefToIncompatibleJump();
          iJump(0) = extraVal(0)/(_beta1*hs);
          iJump(1) = extraVal(1)/(_beta1*hs);
          iJump(2) = extraVal(2)/(_beta1*hs);
          ipvp->getRefToIncompatibleJump() = iJump;

          // full two field
          static SVector3 refN;
          refN = ipvm->getConstRefToReferenceOutwardNormal();
          refN.normalize();
          // update interface deformation gradients with incompatible strain field
          for (int p=0; p<3; p++){
            for (int q=0; q<3; q++){
              Fm(p,q) += (iJump(p)*refN(q));
              Fp(p,q) += (iJump(p)*refN(q));
            }
          }
        }
	else{
          Msg::Error("two field method %d is not correctly defined",this->getMultipleFieldMethod());
	}
      }
      // non-local for minus part
      if( ipvm->getNumberNonLocalVariable()>getNumNonLocalVariable() )
        Msg::Error("Your material law uses more non local variables than your domain dG3DDomain::computeStrain negative interface");
      for (int nlm = 0; nlm < ipvm->getNumberNonLocalVariable(); nlm++)
      {
        double& nonlocalVarm = ipvm->getRefToNonLocalVariable(nlm);
        nonlocalVarm = 0.;

        SVector3& gradNonlocalVarm = ipvm->getRefToGradNonLocalVariable()[nlm];
        STensorOperation::zero(gradNonlocalVarm);


        for (int i = 0; i < nbFFm; i++)
        {
           //epsbar: take compoenent epsbar
          nonlocalVarm += Valm[i+3*nbFFm+nlm*nbFFm]*dispm(i+3*nbFFm+nlm*nbFFm);
          gradNonlocalVarm(0) += Gradm[i+3*nbFFm+nlm*nbFFm][0]*dispm(i+3*nbFFm+nlm*nbFFm);
          gradNonlocalVarm(1) += Gradm[i+3*nbFFm+nlm*nbFFm][1]*dispm(i+3*nbFFm+nlm*nbFFm);
          gradNonlocalVarm(2) += Gradm[i+3*nbFFm+nlm*nbFFm][2]*dispm(i+3*nbFFm+nlm*nbFFm);
        }

      }

      // for plus part
      if( ipvp->getNumberNonLocalVariable()>getNumNonLocalVariable() ) Msg::Error("Your material law uses more non local variables than your domain dG3DDomain::computeStrain positive interface");
      for (int nlp = 0; nlp < ipvp->getNumberNonLocalVariable(); nlp++){
        double& nonlocalVarp = ipvp->getRefToNonLocalVariable(nlp);
        nonlocalVarp = 0.;
        SVector3& gradNonlocalVarp = ipvp->getRefToGradNonLocalVariable()[nlp];
        STensorOperation::zero(gradNonlocalVarp);

        for (int i = 0; i < nbFFp; i++)
        {
           //epsbar: take compoenent epsbar
          nonlocalVarp += Valp[i+3*nbFFp+nlp*nbFFp]*dispp(i+3*nbFFp+nlp*nbFFp);
          gradNonlocalVarp(0) += Gradp[i+3*nbFFp+nlp*nbFFp][0]*dispp(i+3*nbFFp+nlp*nbFFp);
          gradNonlocalVarp(1) += Gradp[i+3*nbFFp+nlp*nbFFp][1]*dispp(i+3*nbFFp+nlp*nbFFp);
          gradNonlocalVarp(2) += Gradp[i+3*nbFFp+nlp*nbFFp][2]*dispp(i+3*nbFFp+nlp*nbFFp);
        }
      }

      // compute the jump
      for (int nl = 0; nl < ipvp->getNumberNonLocalVariable(); nl++){
        double& epsjump= ipvm->getRefToNonLocalJump()(nl);
        computeNonLocalJump(Valm,nbFFm,Valp,nbFFp,dispm,dispp,epsjump,nl);
        ipvp->getRefToNonLocalJump()(nl) = epsjump;
      }

      // mean part
      if (_averageStrainBased and (efMinus->getMaterialLaw()->getNum() == efPlus->getMaterialLaw()->getNum())){
        // average nonlocal values
        for (int nlv = 0; nlv < ipvm->getNumberNonLocalVariable(); nlv++){

          // value averge
          double& valminus = ipvm->getRefToNonLocalVariable(nlv);
          valminus += ipvp->getConstRefToNonLocalVariable(nlv);
          valminus *= 0.5;
          ipvp->getRefToNonLocalVariable(nlv) = valminus;

          // grad average
          SVector3& gradValminus = ipvm->getRefToGradNonLocalVariable()[nlv];
          gradValminus += ipvp->getConstRefToGradNonLocalVariable()[nlv];
          gradValminus *= 0.5;
          ipvp->getRefToGradNonLocalVariable()[nlv] = gradValminus;
        }
      }
      // constitutive extra dofs
      if( ipvm->getNumConstitutiveExtraDofDiffusionVariable()>getNumConstitutiveExtraDofDiffusionVariable() )
        Msg::Error("Your material law uses more constitutive extra dof diffusion variables than your domain dG3DDomain::computeStrain bulk");
      for (int extraDOFFieldm = 0; extraDOFFieldm < ipvm->getNumConstitutiveExtraDofDiffusionVariable(); extraDOFFieldm++)
      {
        double& Tm = ipvm->getRefToField(extraDOFFieldm);
        Tm = 0.;
        SVector3& gradTm = ipvm->getRefToGradField()[extraDOFFieldm];
        STensorOperation::zero(gradTm);
        for (int i = 0; i < nbFFm; i++)
        {
          int indexField=3+getNumNonLocalVariable()+extraDOFFieldm;
          Tm += Valm[i+indexField*nbFFm]*dispm(i+indexField*nbFFm);
          gradTm(0) += Gradm[i+indexField*nbFFm][0]*dispm(i+indexField*nbFFm);
          gradTm(1) += Gradm[i+indexField*nbFFm][1]*dispm(i+indexField*nbFFm);
          gradTm(2) += Gradm[i+indexField*nbFFm][2]*dispm(i+indexField*nbFFm);
        }

        if((ipvm->getNumConstitutiveExtraDofDiffusionVariable() == 2)&&(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField()))
        {
         double& energyConjugatedFieldm = ipvm->getRefToEnergyConjugatedField(extraDOFFieldm);
         SVector3& gradEnergyConjugatedFieldm = ipvm->getRefToGradEnergyConjugatedField()[extraDOFFieldm];
         energyConjugatedFieldm = 0.;
         STensorOperation::zero(gradEnergyConjugatedFieldm);

         int extraDOFField_aux=0;
             if((extraDOFFieldm+1)< ipvm->getNumConstitutiveExtraDofDiffusionVariable()) extraDOFField_aux=extraDOFFieldm+1;
             else extraDOFField_aux=extraDOFFieldm-1;  

         double& dFielddEnergyConjugatedFieldm = ipvm->getRefTodFielddEnergyConjugatedField()[extraDOFFieldm][extraDOFFieldm];
         double& dFielddEnergyConjugatedField2m = ipvm->getRefTodFielddEnergyConjugatedField()[extraDOFFieldm][extraDOFField_aux];
         SVector3& dGradFielddEnergyConjugatedFieldm = ipvm->getRefTodGradFielddEnergyConjugatedField()[extraDOFFieldm][extraDOFFieldm];
         SVector3& dGradFielddEnergyConjugatedField2m = ipvm->getRefTodGradFielddEnergyConjugatedField()[extraDOFFieldm][extraDOFField_aux];
         STensor3& dGradFielddGradEnergyConjugatedFieldm = ipvm->getRefTodGradFielddGradEnergyConjugatedField()[extraDOFFieldm][extraDOFFieldm];
         STensor3& dGradFielddGradEnergyConjugatedField2m = ipvm->getRefTodGradFielddGradEnergyConjugatedField()[extraDOFFieldm][extraDOFField_aux];
 
         STensorOperation::zero(dGradFielddGradEnergyConjugatedFieldm); STensorOperation::zero(dGradFielddGradEnergyConjugatedField2m);

         int indexFieldfT=3+getNumNonLocalVariable()+extraDOFFieldm;

          for (int i = 0; i < nbFFm; i++)
           {
             const double& ftemp = dispm(i+indexFieldfT*nbFFm);
              energyConjugatedFieldm += Valm[i+indexFieldfT*nbFFm]*ftemp;
              gradEnergyConjugatedFieldm(0) += Gradm[i+indexFieldfT*nbFFm][0]*ftemp;
              gradEnergyConjugatedFieldm(1) += Gradm[i+indexFieldfT*nbFFm][1]*ftemp;
              gradEnergyConjugatedFieldm(2) += Gradm[i+indexFieldfT*nbFFm][2]*ftemp;
            }
         
           if(extraDOFFieldm == 0)
            {
             Tm = 1/(energyConjugatedFieldm);

             gradTm(0)=-1/(pow(energyConjugatedFieldm,2))*(gradEnergyConjugatedFieldm(0));
             gradTm(1)=-1/(pow(energyConjugatedFieldm,2))*(gradEnergyConjugatedFieldm(1));
             gradTm(2)=-1/(pow(energyConjugatedFieldm,2))*(gradEnergyConjugatedFieldm(2));
  
             dFielddEnergyConjugatedFieldm  = -1./(energyConjugatedFieldm)/(energyConjugatedFieldm);
             dFielddEnergyConjugatedField2m = 0.;

             dGradFielddEnergyConjugatedFieldm(0)=2/(energyConjugatedFieldm)/(energyConjugatedFieldm)/(energyConjugatedFieldm)*(gradEnergyConjugatedFieldm(0));
             dGradFielddEnergyConjugatedFieldm(1)=2/(energyConjugatedFieldm)/(energyConjugatedFieldm)/(energyConjugatedFieldm)*(gradEnergyConjugatedFieldm(1));
             dGradFielddEnergyConjugatedFieldm(2)=2/(energyConjugatedFieldm)/(energyConjugatedFieldm)/(energyConjugatedFieldm)*(gradEnergyConjugatedFieldm(2));

             dGradFielddEnergyConjugatedField2m(0)=0.;
             dGradFielddEnergyConjugatedField2m(1)=0.;
             dGradFielddEnergyConjugatedField2m(2)=0.;
             
             STensorOperation::diag(dGradFielddGradEnergyConjugatedFieldm,-1./(energyConjugatedFieldm)/(energyConjugatedFieldm));
             

            }
           else
            {
            
             double fTm = ipvm->getConstRefToEnergyConjugatedField(0);
             SVector3 gradfTm = ipvm->getConstRefToGradEnergyConjugatedField()[0];
             
             Tm = -1*(energyConjugatedFieldm)/(fTm);

             gradTm(0)= -1/(fTm)*(gradEnergyConjugatedFieldm(0))+(energyConjugatedFieldm)/(pow(fTm,2))*(gradfTm(0));
             gradTm(1)= -1/(fTm)*(gradEnergyConjugatedFieldm(1))+(energyConjugatedFieldm)/(pow(fTm,2))*(gradfTm(1));
             gradTm(2)= -1/(fTm)*(gradEnergyConjugatedFieldm(2))+(energyConjugatedFieldm)/(pow(fTm,2))*(gradfTm(2));

             dFielddEnergyConjugatedFieldm  = -1./(fTm);
             dFielddEnergyConjugatedField2m = (energyConjugatedFieldm)/(fTm)/(fTm);

             dGradFielddEnergyConjugatedFieldm(0)=1./(fTm)/(fTm)*(gradfTm(0));
             dGradFielddEnergyConjugatedFieldm(1)=1./(fTm)/(fTm)*(gradfTm(1));
             dGradFielddEnergyConjugatedFieldm(2)=1./(fTm)/(fTm)*(gradfTm(2));

             dGradFielddEnergyConjugatedField2m(0)=1./(fTm)/(fTm)*(gradEnergyConjugatedFieldm(0))-2*(energyConjugatedFieldm)/(fTm)/(fTm)/(fTm)*(gradfTm(0));
             dGradFielddEnergyConjugatedField2m(1)=1./(fTm)/(fTm)*(gradEnergyConjugatedFieldm(1))-2*(energyConjugatedFieldm)/(fTm)/(fTm)/(fTm)*(gradfTm(1));
             dGradFielddEnergyConjugatedField2m(2)=1./(fTm)/(fTm)*(gradEnergyConjugatedFieldm(2))-2*(energyConjugatedFieldm)/(fTm)/(fTm)/(fTm)*(gradfTm(2));
    
             STensorOperation::diag(dGradFielddGradEnergyConjugatedFieldm,-1./(fTm));
             STensorOperation::diag(dGradFielddGradEnergyConjugatedField2m,(energyConjugatedFieldm)/(fTm)/(fTm));

            }
        }
      }
      if( ipvp->getNumConstitutiveExtraDofDiffusionVariable()>getNumConstitutiveExtraDofDiffusionVariable() )
        Msg::Error("Your material law uses more constitutive extra dof diffusion variables than your domain dG3DDomain::computeStrain bulk");
      for (int extraDOFFieldp = 0; extraDOFFieldp < ipvp->getNumConstitutiveExtraDofDiffusionVariable(); extraDOFFieldp++)
      {
        double&  Tp = ipvp->getRefToField(extraDOFFieldp);
        Tp = 0.;
        SVector3& gradTp = ipvp->getRefToGradField()[extraDOFFieldp];
        STensorOperation::zero(gradTp);

        for (int i = 0; i < nbFFp; i++)
        {
          int indexField=3+getNumNonLocalVariable()+extraDOFFieldp;
          Tp += Valp[i+indexField*nbFFp]*dispp(i+indexField*nbFFp);
          gradTp(0) += Gradp[i+indexField*nbFFp][0]*dispp(i+indexField*nbFFp);
          gradTp(1) += Gradp[i+indexField*nbFFp][1]*dispp(i+indexField*nbFFp);
          gradTp(2) += Gradp[i+indexField*nbFFp][2]*dispp(i+indexField*nbFFp);
        }

        if((ipvp->getNumConstitutiveExtraDofDiffusionVariable() == 2)&&(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField()))
        {
         double& energyConjugatedFieldp = ipvp->getRefToEnergyConjugatedField(extraDOFFieldp);
         SVector3& gradEnergyConjugatedFieldp = ipvp->getRefToGradEnergyConjugatedField()[extraDOFFieldp];
         energyConjugatedFieldp = 0.;
         STensorOperation::zero(gradEnergyConjugatedFieldp);

         int extraDOFField_aux=0;
             if((extraDOFFieldp+1)< ipvp->getNumConstitutiveExtraDofDiffusionVariable()) extraDOFField_aux=extraDOFFieldp+1;
             else extraDOFField_aux=extraDOFFieldp-1;  

         double& dFielddEnergyConjugatedFieldp = ipvp->getRefTodFielddEnergyConjugatedField()[extraDOFFieldp][extraDOFFieldp];
         double& dFielddEnergyConjugatedField2p = ipvp->getRefTodFielddEnergyConjugatedField()[extraDOFFieldp][extraDOFField_aux];
         SVector3& dGradFielddEnergyConjugatedFieldp = ipvp->getRefTodGradFielddEnergyConjugatedField()[extraDOFFieldp][extraDOFFieldp];
         SVector3& dGradFielddEnergyConjugatedField2p = ipvp->getRefTodGradFielddEnergyConjugatedField()[extraDOFFieldp][extraDOFField_aux];
         STensor3& dGradFielddGradEnergyConjugatedFieldp = ipvp->getRefTodGradFielddGradEnergyConjugatedField()[extraDOFFieldp][extraDOFFieldp];
         STensor3& dGradFielddGradEnergyConjugatedField2p = ipvp->getRefTodGradFielddGradEnergyConjugatedField()[extraDOFFieldp][extraDOFField_aux];
 
         STensorOperation::zero(dGradFielddGradEnergyConjugatedFieldp); STensorOperation::zero(dGradFielddGradEnergyConjugatedField2p);

         int indexFieldfT=3+getNumNonLocalVariable()+extraDOFFieldp;

          for (int i = 0; i < nbFFp; i++)
           {
             const double& ftemp = dispp(i+indexFieldfT*nbFFp);
              energyConjugatedFieldp += Valp[i+indexFieldfT*nbFFp]*ftemp;
              gradEnergyConjugatedFieldp(0) += Gradp[i+indexFieldfT*nbFFp][0]*ftemp;
              gradEnergyConjugatedFieldp(1) += Gradp[i+indexFieldfT*nbFFp][1]*ftemp;
              gradEnergyConjugatedFieldp(2) += Gradp[i+indexFieldfT*nbFFp][2]*ftemp;
            }
         
           if(extraDOFFieldp == 0)
            {
             Tp = 1/(energyConjugatedFieldp);

             gradTp(0)=-1/(pow(energyConjugatedFieldp,2))*(gradEnergyConjugatedFieldp(0));
             gradTp(1)=-1/(pow(energyConjugatedFieldp,2))*(gradEnergyConjugatedFieldp(1));
             gradTp(2)=-1/(pow(energyConjugatedFieldp,2))*(gradEnergyConjugatedFieldp(2));
  
             dFielddEnergyConjugatedFieldp  = -1./(energyConjugatedFieldp)/(energyConjugatedFieldp);
             dFielddEnergyConjugatedField2p = 0.;

             dGradFielddEnergyConjugatedFieldp(0)=2/(energyConjugatedFieldp)/(energyConjugatedFieldp)/(energyConjugatedFieldp)*(gradEnergyConjugatedFieldp(0));
             dGradFielddEnergyConjugatedFieldp(1)=2/(energyConjugatedFieldp)/(energyConjugatedFieldp)/(energyConjugatedFieldp)*(gradEnergyConjugatedFieldp(1));
             dGradFielddEnergyConjugatedFieldp(2)=2/(energyConjugatedFieldp)/(energyConjugatedFieldp)/(energyConjugatedFieldp)*(gradEnergyConjugatedFieldp(2));

             dGradFielddEnergyConjugatedField2p(0)=0.;
             dGradFielddEnergyConjugatedField2p(1)=0.;
             dGradFielddEnergyConjugatedField2p(2)=0.;
             
             STensorOperation::diag(dGradFielddGradEnergyConjugatedFieldp,-1./(energyConjugatedFieldp)/(energyConjugatedFieldp));

            }
           else
            {
            
             double fTp = ipvp->getConstRefToEnergyConjugatedField(0);
             SVector3 gradfTp = ipvp->getConstRefToGradEnergyConjugatedField()[0];
             
             Tp = -1*(energyConjugatedFieldp)/(fTp);

             gradTp(0)= -1/(fTp)*(gradEnergyConjugatedFieldp(0))+(energyConjugatedFieldp)/(pow(fTp,2))*(gradfTp(0));
             gradTp(1)= -1/(fTp)*(gradEnergyConjugatedFieldp(1))+(energyConjugatedFieldp)/(pow(fTp,2))*(gradfTp(1));
             gradTp(2)= -1/(fTp)*(gradEnergyConjugatedFieldp(2))+(energyConjugatedFieldp)/(pow(fTp,2))*(gradfTp(2));

             dFielddEnergyConjugatedFieldp  = -1./(fTp);
             dFielddEnergyConjugatedField2p = (energyConjugatedFieldp)/(fTp)/(fTp);

             dGradFielddEnergyConjugatedFieldp(0)=1./(fTp)/(fTp)*(gradfTp(0));
             dGradFielddEnergyConjugatedFieldp(1)=1./(fTp)/(fTp)*(gradfTp(1));
             dGradFielddEnergyConjugatedFieldp(2)=1./(fTp)/(fTp)*(gradfTp(2));

             dGradFielddEnergyConjugatedField2p(0)=1./(fTp)/(fTp)*(gradEnergyConjugatedFieldp(0))-2*(energyConjugatedFieldp)/(fTp)/(fTp)/(fTp)*(gradfTp(0));
             dGradFielddEnergyConjugatedField2p(1)=1./(fTp)/(fTp)*(gradEnergyConjugatedFieldp(1))-2*(energyConjugatedFieldp)/(fTp)/(fTp)/(fTp)*(gradfTp(1));
             dGradFielddEnergyConjugatedField2p(2)=1./(fTp)/(fTp)*(gradEnergyConjugatedFieldp(2))-2*(energyConjugatedFieldp)/(fTp)/(fTp)/(fTp)*(gradfTp(2));
    
             STensorOperation::diag(dGradFielddGradEnergyConjugatedFieldp,-1./(fTp));
             STensorOperation::diag(dGradFielddGradEnergyConjugatedField2p,(energyConjugatedFieldp)/(fTp)/(fTp));

            }
        }
      }
      // compute the jump
      for (int extraDOFField = 0; extraDOFField < ipvm->getNumConstitutiveExtraDofDiffusionVariable(); extraDOFField++)
      {
        int indexField=3+getNumNonLocalVariable()+extraDOFField;

        if(ipvm->getNumConstitutiveExtraDofDiffusionVariable()==1) //thermomec
        {
          double& tempjump= ipvm->getRefToFieldJump()(extraDOFField);
          computeExtraFieldJump(indexField,Valm,nbFFm,Valp,nbFFp,dispm,dispp,tempjump);
          double tpj=tempjump;
          ipvp->getRefToFieldJump()(extraDOFField) = tpj;
          // compute 1/T jump to be energetically consistent
          if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())
          {         
            double& ftjump = ipvm->getRefToEnergyConjugatedFieldJump()(extraDOFField);
            double& dftjumpdTp = ipvp->getRefTodEnergyConjugatedFieldJumpdFieldp()[extraDOFField][extraDOFField];
            double& dftjumpdTm = ipvm->getRefTodEnergyConjugatedFieldJumpdFieldm()[extraDOFField][extraDOFField];
            computeftjump(3+getNumNonLocalVariable(),Valm,nbFFm,Valp,nbFFp,dispm,dispp,ftjump,dftjumpdTp,dftjumpdTm);
            ipvp->getRefToEnergyConjugatedFieldJump()(extraDOFField) = ftjump;
            ipvm->getRefTodEnergyConjugatedFieldJumpdFieldm()[extraDOFField][extraDOFField] = dftjumpdTm;
            ipvp->getRefTodEnergyConjugatedFieldJumpdFieldp()[extraDOFField][extraDOFField] = dftjumpdTp;
          }
        }
        else if(ipvm->getNumConstitutiveExtraDofDiffusionVariable()==2) //elecThermech
        { 
          if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())
          {  
           double& ftjump = ipvm->getRefToEnergyConjugatedFieldJump()(extraDOFField);
           computeExtraFieldJump(indexField,Valm,nbFFm,Valp,nbFFp,dispm,dispp,ftjump);
           double tpj=ftjump;
           ipvp->getRefToEnergyConjugatedFieldJump()(extraDOFField) = tpj;
          }
          else
          {
            double& tempjump= ipvm->getRefToFieldJump()(extraDOFField);
            computeExtraFieldJump(indexField,Valm,nbFFm,Valp,nbFFp,dispm,dispp,tempjump);
            double tpj =tempjump;
            ipvp->getRefToFieldJump()(extraDOFField) = tpj;
             
            double& ftjump = ipvm->getRefToEnergyConjugatedFieldJump()(extraDOFField);
            double& dftjumpdTp = ipvp->getRefTodEnergyConjugatedFieldJumpdFieldp()[extraDOFField][extraDOFField];
            double& dftjumpdTm = ipvm->getRefTodEnergyConjugatedFieldJumpdFieldm()[extraDOFField][extraDOFField];
            
            if(extraDOFField==0)
            {
              computeftjump(indexField,Valm,nbFFm,Valp,nbFFp,dispm,dispp,ftjump,dftjumpdTp,dftjumpdTm);
              ipvp->getRefToEnergyConjugatedFieldJump()(extraDOFField) = ftjump;
              ipvm->getRefTodEnergyConjugatedFieldJumpdFieldm()[extraDOFField][extraDOFField] = dftjumpdTm;
              ipvp->getRefTodEnergyConjugatedFieldJumpdFieldp()[extraDOFField][extraDOFField] = dftjumpdTp;
            }
            else
            {
             double& dftjumpdT2p = ipvp->getRefTodEnergyConjugatedFieldJumpdFieldp()[extraDOFField][0];
             double& dftjumpdT2m = ipvm->getRefTodEnergyConjugatedFieldJumpdFieldm()[extraDOFField][0];
             computefvjump(indexField-1,indexField,Valm,nbFFm,Valp,nbFFp,dispm,dispp,ftjump,dftjumpdT2p,dftjumpdT2m,dftjumpdTp,dftjumpdTm); 
             ipvp->getRefToEnergyConjugatedFieldJump()(extraDOFField) = ftjump;
             ipvm->getRefTodEnergyConjugatedFieldJumpdFieldm()[extraDOFField][extraDOFField] = dftjumpdTm;
             ipvp->getRefTodEnergyConjugatedFieldJumpdFieldp()[extraDOFField][extraDOFField] = dftjumpdTp;
             ipvm->getRefTodEnergyConjugatedFieldJumpdFieldm()[extraDOFField][0] = dftjumpdT2m;
             ipvp->getRefTodEnergyConjugatedFieldJumpdFieldp()[extraDOFField][0] = dftjumpdT2p;
            }
          }
        }
        else Msg::Error("UseEnergyConjugatedField with more than 2 extra dofs not implemented");
      }
      //
    }

  }
  else
  {  // Virtual interface element
     Msg::Error("No virtual interface for 3D");
  }
}
void dG3DDomain::computeIpv(AllIPState *aips,MInterfaceElement *ie, IntPt *GP,const IPStateBase::whichState ws,
                                        partDomain* efMinus, partDomain *efPlus,materialLaw *mlawmi,
                                        materialLaw *mlawp,fullVector<double> &dispm,
                                        fullVector<double> &dispp, fullVector<double>& dispExtra,
                                        const bool virt, bool stiff,const bool checkfrac)
{
	MElement *ele = dynamic_cast<MElement*>(ie);
	if (!getElementErosionFilter()(ele)) return;

  if(!virt){

    dG3DMaterialLaw *mlawminus = static_cast<dG3DMaterialLaw*>(mlawmi);
    dG3DMaterialLaw *mlawplus = static_cast<dG3DMaterialLaw*>(mlawp);

    dG3DMaterialLaw *mlawbulkm;
    dG3DMaterialLaw *mlawbulkp;
    if(mlawminus->getType() == materialLaw::fracture)
      mlawbulkm = static_cast< dG3DMaterialLaw* >((static_cast< FractureByCohesive3DLaw* > (mlawminus) )->getBulkLaw());
    else
      mlawbulkm= static_cast<dG3DMaterialLaw*>(mlawminus);
    if(mlawplus->getType() == materialLaw::fracture)
      mlawbulkp = static_cast< dG3DMaterialLaw* >((static_cast< FractureByCohesive3DLaw* > (mlawplus) )->getBulkLaw());
    else
      mlawbulkp= static_cast<dG3DMaterialLaw*>(mlawplus);
    if (mlawbulkm->getUseBarF()!=mlawbulkp->getUseBarF()) //check with the bulk law
      Msg::Error("dG3DDomain::computeIpv All the constitutive laws need to use Fbar or not");
    bool useBarF=mlawbulkm->getUseBarF();

    if(_evalStiff){ this->computeStrain(aips,ie, GP,ws,efMinus,efPlus,dispm,dispp,dispExtra,virt, useBarF);}

    int npts=integBound->getIntPoints(ele,&GP);


    AllIPState::ipstateElementContainer *vips = aips->getIPstate(ele->getNum());
    //Msg::Info("nb interf gp %d",npts);
    for(int j=0;j<npts;j++){

      IPStateBase* ipsm = (*vips)[j];
      IPStateBase* ipsp = (*vips)[j+npts];

      IPVariable* ipvm = ipsm->getState(ws);
      IPVariable* ipvp = ipsp->getState(ws);
      const IPVariable* ipvmprev = ipsm->getState(IPStateBase::previous);
      const IPVariable* ipvpprev = ipsp->getState(IPStateBase::previous);

      if (_subSteppingMethod == 0){
        // no substepting
        if (mlawminus->getNum() != mlawplus->getNum()){
        // if minus and plus material are different
					if (_planeStressState){
						mlawminus->stress3DTo2D(ipvm,ipvmprev, stiff, checkfrac);
          	mlawplus->stress3DTo2D(ipvp,ipvpprev, stiff, checkfrac);
					}
					else{
          	mlawminus->stress(ipvm,ipvmprev, stiff, checkfrac);
          	mlawplus->stress(ipvp,ipvpprev, stiff, checkfrac);
					}
        }
        else{
          // if both part are the same
					if (_planeStressState)
						mlawminus->stress3DTo2D(ipvm,ipvmprev,stiff,checkfrac);
					else
          	mlawminus->stress(ipvm,ipvmprev,stiff,checkfrac);
          if (_averageStrainBased){
           // if the computation is performed based on the average strain
          // positive IP is equal to minus IP
            ipvp->operator=(*(dynamic_cast<const IPVariable*>(ipvm)));

          }
          else{
						if (_planeStressState)
							mlawplus->stress3DTo2D(ipvp,ipvpprev,stiff,checkfrac);
						else
            	mlawplus->stress(ipvp,ipvpprev,stiff,checkfrac);
          }
        }
      }
      else if ((_subSteppingMethod == 1) or (_subSteppingMethod == 2))  {
        // compute negative part based on the substepting
        dG3DIPVariableBase* dgipvm = static_cast<dG3DIPVariableBase*>(ipvm);
        const dG3DIPVariableBase* dgipvmprev = static_cast<const dG3DIPVariableBase*>(ipvmprev);

        // create a temporary ipv
        // it cannot be static because different Ipv can be use with different domains
        IPVariable* ipvTemp =  dgipvmprev->clone();
        dG3DIPVariableBase* dgipvTemp = dynamic_cast<dG3DIPVariableBase*>(ipvTemp);
        if (dgipvTemp == NULL) Msg::Error("dG3DIPVariableBase::clone is not correctly defined");

        static STensor3 dF;
        dF = dgipvm->getConstRefToDeformationGradient();
        dF -= dgipvmprev->getConstRefToDeformationGradient();

        // if nonlocal damage is used
        fullVector<double> dnonlocalVarm;
        int numNL = dgipvm->getNumberNonLocalVariable();
        if (numNL > 0){
          dnonlocalVarm.resize(numNL);
          for (int iv = 0; iv < numNL; iv++){
            dnonlocalVarm(iv) = (dgipvm->getConstRefToNonLocalVariable(iv) - dgipvmprev->getConstRefToNonLocalVariable(iv));
          }
        }

        int iter = 1;
        while (1){
          // update kinematics
          double time = double(iter)/(double(_subStepNumber));
          //Msg::Info("step %d, time = %f",iter,time);
          STensor3& F = dgipvm->getRefToDeformationGradient();
          F = dgipvmprev->getConstRefToDeformationGradient();
          F.daxpy(dF,time);

          for (int iv=0; iv < numNL; iv++){
            dgipvm->getRefToNonLocalVariable(iv) = dgipvmprev->getConstRefToNonLocalVariable(iv);
            dgipvm->getRefToNonLocalVariable(iv) += (time*dnonlocalVarm(iv));
          }

          if (iter == _subStepNumber){
            // in last substep
						if (_planeStressState)
            	mlawminus->stress3DTo2D(ipvm,ipvTemp,stiff,checkfrac);
						else
							mlawminus->stress(ipvm,ipvTemp,stiff,checkfrac);
            break;
          }
          else{
						if (_planeStressState)
            	mlawminus->stress3DTo2D(ipvm,ipvTemp,false,checkfrac);
						else
							mlawminus->stress(ipvm,ipvTemp,false,checkfrac);
            dgipvTemp->operator=(*(dynamic_cast<const IPVariable*>(ipvm)));
          }
          // next step
          iter++;
        }
        if (ipvTemp!=NULL) delete ipvTemp;

        if (_averageStrainBased and (mlawminus->getNum() == mlawplus->getNum())){
           // if the computation is performed based on the average strain
          // positive IP is equal to minus IP
          ipvp->operator=(*(dynamic_cast<const IPVariable*>(ipvm)));

        }
        else{
          dG3DIPVariableBase* dgipvp = static_cast<dG3DIPVariableBase*>(ipvp);
          const dG3DIPVariableBase* dgipvpprev = static_cast<const dG3DIPVariableBase*>(ipvpprev);

          // create a temporary ipv
          ipvTemp =  dgipvpprev->clone();
          dgipvTemp = dynamic_cast<dG3DIPVariableBase*>(ipvTemp);
          if (dgipvTemp == NULL) Msg::Error("dG3DIPVariableBase::clone is not correctly defined");

          dF = dgipvp->getConstRefToDeformationGradient();
          dF -= dgipvpprev->getConstRefToDeformationGradient();

          fullVector<double> dnonlocalVarp;
          numNL = dgipvp->getNumberNonLocalVariable();
          if (numNL > 0){
            dnonlocalVarp.resize(numNL);
            for (int iv = 0; iv < numNL; iv++){
              dnonlocalVarp(iv) = (dgipvp->getConstRefToNonLocalVariable(iv) - dgipvpprev->getConstRefToNonLocalVariable(iv));
            }
          }

          int iter = 1;
          while (1){
            // update kinematics
            double time = double(iter)/(double(_subStepNumber));
            //Msg::Info("step %d, time = %f",iter,time);
            STensor3& F = dgipvp->getRefToDeformationGradient();
            F = dgipvpprev->getConstRefToDeformationGradient();
            F.daxpy(dF,time);

            for (int iv=0; iv < numNL; iv++){
              dgipvp->getRefToNonLocalVariable(iv) = dgipvpprev->getConstRefToNonLocalVariable(iv);
              dgipvp->getRefToNonLocalVariable(iv) += (time*dnonlocalVarp(iv));
            }

            // stress  computation at substep

            if (iter == _subStepNumber){
            // in last substep
							if (_planeStressState)
              	mlawplus->stress3DTo2D(ipvp,ipvTemp,stiff,checkfrac);
							else
								mlawplus->stress(ipvp,ipvTemp,stiff,checkfrac);
              break;
            }
            else{
							if (_planeStressState)
              	mlawplus->stress3DTo2D(ipvp,ipvTemp,false,checkfrac);
							else
								mlawplus->stress(ipvp,ipvTemp,false,checkfrac);
              dgipvTemp->operator=(*(dynamic_cast<const IPVariable*>(ipvp)));
            }
            // next step
            iter++;
          }
          if (ipvTemp!=NULL) delete ipvTemp;

        }


      }
      else{
        Msg::Error("This substepping method %d has not been implemented",_subSteppingMethod);
      }

    }
  }
  else
  {  // Virtual interface element
     Msg::Error("No virtual interface for 3D");

  }
}

int dG3DDomain::getStressDimension() const{
	int stresDim = 0;
  if (getHomogenizationOrder() == 1){
    stresDim += 9; // stress
  }
  else if (getHomogenizationOrder() == 2){
    stresDim +=  (9+27); // stress and higher order stress
  }
  else{
    Msg::Error("Homogenized order is not well defined %d",getHomogenizationOrder());
  }

  if (getExtractIrreversibleEnergyFlag()){
   stresDim += 1;
  }

  if (this->isExtractCohesiveLawFromMicroDamage()){
  // all stress
  // and FD
    stresDim+=9;
  }

  if(getNumConstitutiveExtraDofDiffusionVariable()>0)
  {
    // all mechanical stress + extra flux + source
    stresDim+=4*getNumConstitutiveExtraDofDiffusionVariable();
  }
  return stresDim;
}

bool dG3DDomain::getAccountPathFollowingFlag() const{
  return _accountPathFollowing;
};

void dG3DDomain::checkInternalState(IPField* ipf) const{
  #if defined(HAVE_MPI)
  if (_otherRanks.size()>0){
    // accumulation data from other ranks and send to rootRanks
    for (std::set<int>::iterator it = _domainIPBulk.begin(); it != _domainIPBulk.end(); it++){
      int num = *it;
      IPStateBase* ips = NULL;
      if (Msg::GetCommRank() == _rootRank){
        int elnum, gnum;
        numericalMaterialBase::getTwoIntsFromType(num,elnum,gnum);
        AllIPState::ipstateElementContainer *vips = ipf->getAips()->getIPstate(elnum);
        ips = (*vips)[gnum];
      }
      else if (_otherRanks.find(Msg::GetCommRank()) != _otherRanks.end()){
        AllIPState::ipstateElementContainer *vips = ipf->getAips()->getIPstate(num);
        ips = (*vips)[0];
      }
      IPVariable *ipv = ips->getState(IPStateBase::current);
      const IPVariable *ipvprev = ips->getState(IPStateBase::previous);
      this->getMaterialLaw()->checkInternalState(ipv,ipvprev);
      // check to send
      if (_otherRanks.find(Msg::GetCommRank()) != _otherRanks.end()){
        const dG3DMultiscaleIPVariable* ipvMult = static_cast<const dG3DMultiscaleIPVariable*>(ipv);
        const dG3DMultiscaleIPVariable* ipvMultPrev = static_cast<const dG3DMultiscaleIPVariable*>(ipvprev);
        if (ipvMult->solverIsBroken() and !ipvMultPrev->solverIsBroken()){
          ipf->solverBroken(num);
        }
      }
    };

    if (_fullDg){
      for (std::set<int>::iterator it = _domainIPInterfaceMinus.begin(); it != _domainIPInterfaceMinus.end(); it++){
        int num = *it;
        IPStateBase* ips = NULL;
        if (Msg::GetCommRank() == _rootRank){
          int elnum, gnum;
          numericalMaterialBase::getTwoIntsFromType(num,elnum,gnum);
          AllIPState::ipstateElementContainer *vips = ipf->getAips()->getIPstate(elnum);
          ips = (*vips)[gnum];
        }
        else if (_otherRanks.find(Msg::GetCommRank()) != _otherRanks.end()){
          AllIPState::ipstateElementContainer *vips = ipf->getAips()->getIPstate(num);
          ips = (*vips)[0];
        }
        IPVariable *ipv = ips->getState(IPStateBase::current);
        const IPVariable *ipvprev = ips->getState(IPStateBase::previous);
        this->getMaterialLawMinus()->checkInternalState(ipv,ipvprev);

        // check to send
        if (_otherRanks.find(Msg::GetCommRank()) != _otherRanks.end()){
          const dG3DMultiscaleIPVariable* ipvMult = NULL;
          const dG3DMultiscaleIPVariable* ipvMultPrev = NULL;
          const FractureCohesive3DIPVariable *fipv = dynamic_cast<const FractureCohesive3DIPVariable*>(ipv);
          if (fipv != NULL){
            const FractureCohesive3DIPVariable *fipvprev = static_cast<const FractureCohesive3DIPVariable*>(ipvprev);
            ipvMult = static_cast<const dG3DMultiscaleIPVariable*>(fipv->getIPvBulk());
            ipvMultPrev = static_cast<const dG3DMultiscaleIPVariable*>(fipvprev->getIPvBulk());
          }
          else{
            ipvMult = static_cast<const dG3DMultiscaleIPVariable*>(ipv);
            ipvMultPrev = static_cast<const dG3DMultiscaleIPVariable*>(ipvprev);
          }

          if (ipvMult->solverIsBroken() and !ipvMultPrev->solverIsBroken()){
            ipf->solverBroken(num);
          }
        }
      };


      for (std::set<int>::iterator it = _domainIPInterfacePlus.begin(); it != _domainIPInterfacePlus.end(); it++){
        int num = *it;
        IPStateBase* ips = NULL;
        if (Msg::GetCommRank() == _rootRank){
          int elnum,gnum;
          numericalMaterialBase::getTwoIntsFromType(num,elnum,gnum);
          AllIPState::ipstateElementContainer *vips = ipf->getAips()->getIPstate(elnum);
          ips = (*vips)[gnum];
        }
        else if (_otherRanks.find(Msg::GetCommRank()) != _otherRanks.end()){
          AllIPState::ipstateElementContainer *vips = ipf->getAips()->getIPstate(num);
          ips = (*vips)[0];
        }
        IPVariable *ipv = ips->getState(IPStateBase::current);
        const IPVariable *ipvprev = ips->getState(IPStateBase::previous);
        this->getMaterialLawPlus()->checkInternalState(ipv,ipvprev);

        // check to send
        if (_otherRanks.find(Msg::GetCommRank()) != _otherRanks.end()){
          const dG3DMultiscaleIPVariable* ipvMult = NULL;
          const dG3DMultiscaleIPVariable* ipvMultPrev = NULL;
          const FractureCohesive3DIPVariable *fipv = dynamic_cast<const FractureCohesive3DIPVariable*>(ipv);
          if (fipv != NULL){
            const FractureCohesive3DIPVariable *fipvprev = static_cast<const FractureCohesive3DIPVariable*>(ipvprev);
            ipvMult = static_cast<const dG3DMultiscaleIPVariable*>(fipv->getIPvBulk());
            ipvMultPrev = static_cast<const dG3DMultiscaleIPVariable*>(fipvprev->getIPvBulk());
          }
          else{
            ipvMult = static_cast<const dG3DMultiscaleIPVariable*>(ipv);
            ipvMultPrev = static_cast<const dG3DMultiscaleIPVariable*>(ipvprev);
          }

          if (ipvMult->solverIsBroken() and !ipvMultPrev->solverIsBroken()){
            ipf->solverBroken(num);
          }
        }
      }
    }
  }
  else
  #endif // HAVE_MPI
  {
    for (groupOfElements::elementContainer::const_iterator it = this->g_cbegin(); it!= this->g_cend(); it++){
      MElement* ele = *it;
      IntPt* GP;
      int npts= integBulk->getIntPoints(ele,&GP);
      AllIPState::ipstateElementContainer *vips = ipf->getAips()->getIPstate(ele->getNum());
      for (int i=0; i< npts; i++){
        IPStateBase* ips = (*vips)[i];
        IPVariable *ipv = ips->getState(IPStateBase::current);
        const IPVariable *ipvprev = ips->getState(IPStateBase::previous);
        this->getMaterialLaw()->checkInternalState(ipv,ipvprev);
      }
    }
    if (_fullDg){
      for (groupOfElements::elementContainer::const_iterator it = this->gi->begin(); it != this->gi->end(); it++){
        MElement* ele = *it;
        IntPt* GP;
        int npts= integBound->getIntPoints(ele,&GP);
        AllIPState::ipstateElementContainer *vips = ipf->getAips()->getIPstate(ele->getNum());

        for (int i=0; i< npts; i++){
          IPStateBase* ipsm = (*vips)[i];
          // minus
          IPVariable* ipvm = ipsm->getState(IPStateBase::current);
          const IPVariable* ipvprevm =ipsm->getState(IPStateBase::previous);
          this->getMaterialLawMinus()->checkInternalState(ipvm,ipvprevm);

          IPStateBase* ipsp = (*vips)[i+npts];
          // plus
          IPVariable* ipvp = ipsp->getState(IPStateBase::current);
          const IPVariable* ipvprevp = ipsp->getState(IPStateBase::previous);
          this->getMaterialLawPlus()->checkInternalState(ipvp,ipvprevp);
        }
      }
    }
  }
};

void dG3DDomain::checkFailure(IPField* ipf) const{
  // Check and manage insertion of cohesive element

  const dG3DMaterialLaw *mlawminus = static_cast<const dG3DMaterialLaw*>(this->getMaterialLawMinus());
  const dG3DMaterialLaw *mlawplus = static_cast<const dG3DMaterialLaw*>(this->getMaterialLawPlus());

  // Check failure only if needed (if DG and fracture law are used)
  if (_fullDg and (mlawminus->getType() == materialLaw::fracture and mlawplus->getType() == materialLaw::fracture))
  {
    // Get mlaw
      // in minus part
    const FractureByCohesive3DLaw * flawMinus = static_cast<const FractureByCohesive3DLaw*>(mlawminus);
    const Cohesive3DLaw *mclaw = static_cast<const Cohesive3DLaw*>(flawMinus->getFractureLaw());
    const dG3DMaterialLaw* bulkLawMinus =static_cast<const dG3DMaterialLaw*>(flawMinus->getBulkLaw());
      // in plus part
    const FractureByCohesive3DLaw* flawPlus = static_cast<const FractureByCohesive3DLaw*>(mlawplus);
    const dG3DMaterialLaw* bulkLawPlus = static_cast<const dG3DMaterialLaw*>(flawPlus->getBulkLaw());

    // if only domain work only on rootranks
    // Loop on interface elements to check failure
    for(groupOfElements::elementContainer::iterator it=gi->begin(); it!=gi->end(); ++it){
      MElement *ele = *it;
      MInterfaceElement *ie = dynamic_cast<MInterfaceElement*>(ele);

      int numElMinus = ie->getElem(0)->getNum();
      int numElPlus = ie->getElem(1)->getNum();


      // Get ipstate container of the given interface element
      IntPt* GP;
      int npts=integBound->getIntPoints(ele,&GP);
      AllIPState::ipstateElementContainer *vips = ipf->getAips()->getIPstate(ele->getNum());


      // Synchronise ipv deletion on both sides of an interface element
        // loop on ips
      for(int j=0;j<npts;j++){
        // get ipv_frac
        IPStateBase* ipsm = (*vips)[j];
        IPStateBase* ipsp = (*vips)[j+npts];
        FractureCohesive3DIPVariable *ipvfm = static_cast<FractureCohesive3DIPVariable*>(ipsm->getState(IPStateBase::current));
        FractureCohesive3DIPVariable *ipvfp = static_cast<FractureCohesive3DIPVariable*>(ipsp->getState(IPStateBase::current));
        const FractureCohesive3DIPVariable *ipvfprevm = static_cast<const FractureCohesive3DIPVariable*>(ipsm->getState(IPStateBase::previous));
        const FractureCohesive3DIPVariable *ipvfprevp = static_cast<const FractureCohesive3DIPVariable*>(ipsp->getState(IPStateBase::previous));

        // force ipv of each side to be deleted if one of them is already deleted
        if (!ipvfprevm->isDeleted())
        {
          if (ipvfm->isDeleted()){
            Msg::Info("dG3DDomain::checkFailure: a couple of IP was deleted at one interface element");
            ipvfp->Delete(true);

          }
          else if (ipvfp->isDeleted()){
            ipvfm->Delete(true);
            Msg::Info("dG3DDomain::checkFailure: a couple of IP was deleted at one interface element");
          }
        }
      }




      // check broken state when element is not broken yet from previous state
      if (!(ipf->elementIsBroken(ele))){
        // if element is not yet broken
        bool willCheckFailure = true;
        if (_imposeCrackFlag){
          // if crack path is forced following some interfaces
          willCheckFailure =false;
          if (_imposedInterfaceCrack.size() > 0){
            TwoNum key(numElMinus, numElPlus);
            if (_imposedInterfaceCrack.find(key) != _imposedInterfaceCrack.end()){
              willCheckFailure = true;
              printf("checking failure at interface between ele %d  and ele %d \n",numElMinus, numElPlus);
            }
          }
        }

        if (willCheckFailure){
          bool interfaceElementBroken = false; // element broken flag (said element is broken if all Ip are already broken)

          std::set<int> brokenIPs; // all broken IPs

          /* Check for cohesive insertion on ip's of the interface element */
          for(int j=0;j<npts;j++){
            IPStateBase* ipsm = (*vips)[j];
            IPStateBase* ipsp = (*vips)[j+npts];

            const FractureCohesive3DIPVariable *ipvfprevm = static_cast<const FractureCohesive3DIPVariable*>(ipsm->getState(IPStateBase::previous));
            const FractureCohesive3DIPVariable *ipvfprevp = static_cast<const FractureCohesive3DIPVariable*>(ipsp->getState(IPStateBase::previous));
            FractureCohesive3DIPVariable *ipvfm = static_cast<FractureCohesive3DIPVariable*>(ipsm->getState(IPStateBase::current));
            FractureCohesive3DIPVariable *ipvfp = static_cast<FractureCohesive3DIPVariable*>(ipsp->getState(IPStateBase::current));

            // if previously broken
            if (ipvfprevm->isbroken()){
              ipvfm->broken();
              ipvfp->broken();
            }
            else{
              // check for current state
              if (!ipvfm->isbroken()){
                mclaw->checkCohesiveInsertion(ipsm,ipsp);
                if(ipvfm->isbroken()){
                  // synchonize data when one ipv is broken
                  ipf->brokenIP(numericalMaterialBase::createTypeWithTwoInts(ele->getNum(),j));
                  ipf->brokenIP(numericalMaterialBase::createTypeWithTwoInts(ele->getNum(),j+npts));
                  printf("Checked: failed IPs, ipminus %d ip plus %d of interface %d between elements %d %d \n",j,j+npts,ele->getNum(),numElMinus,numElPlus);
                };
              }
            };

            // Count the number of currently broken ipv
            if(ipvfm->isbroken()){
              brokenIPs.insert(j);
            };
          }
          int nbIpVBroken = brokenIPs.size();       // Number of IPv broken at the interface element


          if (nbIpVBroken ==npts) // all GP are broken
          {
            interfaceElementBroken = true; // interfaceElement is naturally broken
          }

          // Check if forcing is alowed and not already done
          if (_forceInterfaceElementBroken and !interfaceElementBroken)
          {
            // Determine the needed number of broken ipv before forcing
            int nptsToBreakInterfaceElement(0);
            if (_failureIPRatio <= 0.)
            {
              // break interface element when cohesive IP is inserted in first IP
              nptsToBreakInterfaceElement = 1;
            }
            else if (_failureIPRatio >= 1.)
            {
              // break interface element when cohesive IPs are inserted in all IP
              nptsToBreakInterfaceElement = npts;
            }
            else
            {
              // break interface element when cohesive IPs are inserted in agiven number of IPs
              nptsToBreakInterfaceElement = npts*_failureIPRatio;
              if (nptsToBreakInterfaceElement < 1) nptsToBreakInterfaceElement = 1;
            }

            // Check if a sufficient number of IP's are broken
            if(nbIpVBroken >= nptsToBreakInterfaceElement)
            {
              interfaceElementBroken = true; // element is force to break
              // Force cohesive insertion at all cohesive ipv if not already broken
              for(int j=0;j<npts;j++)
              {
                if (brokenIPs.find(j) == brokenIPs.end()){
                  // Get Ipstate
                  IPStateBase* ipsm = (*vips)[j];
                  IPStateBase* ipsp = (*vips)[j+npts];
                  // Check if failure already detected and force insertion if needed
                  FractureCohesive3DIPVariable *ipvfm = static_cast<FractureCohesive3DIPVariable*>(ipsm->getState(IPStateBase::current));
                    // force broken all remaining IPVs
                  mclaw->checkCohesiveInsertion(ipsm,ipsp,true);
                  ipf->brokenIP(numericalMaterialBase::createTypeWithTwoInts(ele->getNum(),j));
                  ipf->brokenIP(numericalMaterialBase::createTypeWithTwoInts(ele->getNum(),j+npts));

                  printf("Forced: failed IPs, ipminus %d ip plus %d of interface %d between elements %d %d \n",j,j+npts,ele->getNum(),numElMinus,numElPlus);
                }
              }
            }
          }

          // Control of crack insertion
            // If an interface element is totally broken at this step
          if (interfaceElementBroken){
            // add to the list of newly cracked elements
            ipf->addOneBrokenElement(ele);
          }

          // Control damage blocking inside bulk elements in terms of chosen method
          bool blockBulkDamage = false; // true if damage is blocked in bulk elements
          if (_bulkDamageBlockMethod == 0)
          {
           // block damage when cohesive crack is inserted in first IP
           if (nbIpVBroken > 0) blockBulkDamage = true;
          }
          else if (_bulkDamageBlockMethod == 1)
          {
            // block damage when cohesive crack is interface element is broken
            if (interfaceElementBroken) blockBulkDamage = true;
          }
          else
          {
            // no block damage in bulk ips
            blockBulkDamage = false;
          }


          // Block damage in bulk in Plus and Minus bulk elements when using
          // bulk material with damage if needed
          if (blockBulkDamage){
            if (bulkLawMinus->withEnergyDissipation()){
              // Get bulk element and check if it exists
              const partDomain* dom = this->getMinusDomain();
              MElement* em = ie->getElem(0);
              if(dom->g_find(em) and (!ipf->dissipationIsBlockedOnElement(em))){
                ipf->addOneBlockedDissipationElement(em);
                // Loop on bulk ipv of the element
                int npts_bulk = dom->getBulkGaussIntegrationRule()->getIntPoints(em,&GP);
                AllIPState::ipstateElementContainer *vips_bulk = ipf->getAips()->getIPstate(em->getNum());
                for(int k=0;k<npts_bulk;k++)
                {
                  // Update critical damage and flag in bulk ipv's
                  IPStateBase* ips_bulk = (*vips_bulk)[k];
                  dG3DIPVariableBase *ipv_bulk = static_cast<dG3DIPVariableBase*>(ips_bulk->getState(IPStateBase::current));
                  const dG3DIPVariableBase* ipv_bulkPrev = static_cast<const dG3DIPVariableBase*>(ips_bulk->getState(IPStateBase::previous));

                  if (ipv_bulkPrev->dissipationIsBlocked()){
                    ipv_bulk->blockDissipation(true);
                    ipv_bulk->setBulkCriticalDamage(ipv_bulkPrev->getBulkCriticalDamage());
                  }
                  else{
                    ipv_bulk->blockDissipation(true);
                    ipv_bulk->setBulkCriticalDamage(ipv_bulk->get(IPField::DAMAGE));
                    ipf->blockDissipationIP(numericalMaterialBase::createTypeWithTwoInts(em->getNum(),k));
                  }
                }
              }
            }

            if (bulkLawPlus->withEnergyDissipation())
            {
              const partDomain* dom = this->getPlusDomain();
              MElement* ep = ie->getElem(1);
              if(dom->g_find(ep) and (!ipf->dissipationIsBlockedOnElement(ep)))
              {
                ipf->addOneBlockedDissipationElement(ep);
                // Loop on bulk ipv of the element
                int npts_bulk = dom->getBulkGaussIntegrationRule()->getIntPoints(ep,&GP);
                AllIPState::ipstateElementContainer *vips_bulk = ipf->getAips()->getIPstate(ep->getNum());
                for(int k=0;k<npts_bulk;k++)
                {
                  // Update critical damage and flag in bulk ipv's
                  IPStateBase* ips_bulk = (*vips_bulk)[k];
                  dG3DIPVariableBase *ipv_bulk = static_cast<dG3DIPVariableBase*>(ips_bulk->getState(IPStateBase::current));
                  const dG3DIPVariableBase* ipv_bulkPrev = static_cast<const dG3DIPVariableBase*>(ips_bulk->getState(IPStateBase::previous));

                  if (ipv_bulkPrev->dissipationIsBlocked()){
                    ipv_bulk->blockDissipation(true);
                    ipv_bulk->setBulkCriticalDamage(ipv_bulkPrev->getBulkCriticalDamage());
                  }
                  else{
                    ipv_bulk->blockDissipation(true);
                    ipv_bulk->setBulkCriticalDamage(ipv_bulk->get(IPField::DAMAGE));
                    ipf->blockDissipationIP(numericalMaterialBase::createTypeWithTwoInts(ep->getNum(),k));
                  }
                }
              }
            }
          }
        }
      }
      #ifdef _DEBUG
      else{
        printf("element %d with negative %d, positive %d is fully broken\n", ele->getNum(),numElMinus,numElPlus);
      }
      #endif //_DEBUG
    }
  }


  // block damage in elements which are not related to crack mouth
  if (this->groupOfElementsSize() > 0 and ((_bulkDamageBlockMethod ==0) or (_bulkDamageBlockMethod ==1))){
    const materialLaw* mlaw = this->getMaterialLaw();
    if (mlaw->withEnergyDissipation()){
      IntPt* GP;
      for (groupOfElements::elementContainer::iterator ite = this->g_cbegin(); ite!= this->g_cend(); ite++){
        MElement* ele = *ite;

        if (!ipf->dissipationIsBlockedOnElement(ele)){
          int npts_bulk = this->getBulkGaussIntegrationRule()->getIntPoints(ele,&GP);
          AllIPState::ipstateElementContainer *vips_bulk = ipf->getAips()->getIPstate(ele->getNum());

          // if one IP is broken, block all element
          bool willBlock = false;
          for(int k=0;k<npts_bulk;k++){
            IPStateBase* ips_bulk = (*vips_bulk)[k];
            dG3DIPVariableBase *ipv_bulk = static_cast<dG3DIPVariableBase*>(ips_bulk->getState(IPStateBase::current));
            const dG3DIPVariableBase *ipvprev_bulk = static_cast<const dG3DIPVariableBase*>(ips_bulk->getState(IPStateBase::previous));

            if (ipvprev_bulk->dissipationIsBlocked() or mlaw->brokenCheck(ipv_bulk)){
              willBlock = true;
              break;
            }
          }

          if (willBlock){
            ipf->addOneBlockedDissipationElement(ele);
            // block all IP
            for(int k=0;k<npts_bulk;k++){
              IPStateBase* ips_bulk = (*vips_bulk)[k];
              dG3DIPVariableBase *ipv_bulk = static_cast<dG3DIPVariableBase*>(ips_bulk->getState(IPStateBase::current));
              const dG3DIPVariableBase *ipvprev_bulk = static_cast<const dG3DIPVariableBase*>(ips_bulk->getState(IPStateBase::previous));
              if (ipvprev_bulk->dissipationIsBlocked()){
                ipv_bulk->blockDissipation(true);
                ipv_bulk->setBulkCriticalDamage(ipvprev_bulk->getBulkCriticalDamage());
              }
              else{
                ipv_bulk->blockDissipation(true);
                ipv_bulk->setBulkCriticalDamage(ipv_bulk->get(IPField::DAMAGE));
                ipf->blockDissipationIP(numericalMaterialBase::createTypeWithTwoInts(ele->getNum(),k));
              }
            }
          }
        }
      }
    }
  }

};



void dG3DDomain::initialIPVariable(AllIPState *aips,const unknownField *ufield, const IPStateBase::whichState ws, bool stiff)
{
  dG3DMaterialLaw* mlawMinus = dynamic_cast<dG3DMaterialLaw*>(this->getMaterialLawMinus());
  dG3DMaterialLaw* mlawPlus = dynamic_cast<dG3DMaterialLaw*>(this->getMaterialLawPlus());

    dG3DMaterialLaw *mlawbulkm;
    dG3DMaterialLaw *mlawbulkp;
    if(mlawMinus->getType() == materialLaw::fracture)
      mlawbulkm = static_cast< dG3DMaterialLaw* >((static_cast< FractureByCohesive3DLaw* > (mlawMinus) )->getBulkLaw());
    else
      mlawbulkm= static_cast<dG3DMaterialLaw*>(mlawMinus);
    if(mlawPlus->getType() == materialLaw::fracture)
      mlawbulkp = static_cast< dG3DMaterialLaw* >((static_cast< FractureByCohesive3DLaw* > (mlawPlus) )->getBulkLaw());
    else
      mlawbulkp= static_cast<dG3DMaterialLaw*>(mlawPlus);
    if (mlawbulkm->getUseBarF()!=mlawbulkp->getUseBarF()) //check with the bulk law
      Msg::Error("dG3DDomain::computeIpv All the constitutive laws need to use Fbar or not");
    bool useBarF=mlawbulkm->getUseBarF();

  // compute strain
  this->computeAllIPStrain(aips,ufield,ws,false,useBarF);
  //


  fullVector<double> unknowns_m;
  fullVector<double> unknowns_p;
  fullVector<double> unknowns_inter;
  std::vector<Dof> R;
  R.clear();
  nlsFunctionSpaceUVW<double> *spinter = static_cast<nlsFunctionSpaceUVW<double>*>(_spaceInterUVW);
  IntPt *GP;
  IntPt *GPm; IntPt *GPp;
  // loop on interface
  for(groupOfElements::elementContainer::iterator it=gi->begin(); it!=gi->end(); ++it){
    MInterfaceElement *ie = dynamic_cast<MInterfaceElement*>(*it);
    MElement *ele = *it; //dynamic_cast<MElement*>(ie);
    MElement *em = ie->getElem(0);
    MElement *ep = ie->getElem(1);
    // get displacement
    R.clear();
    this->getFunctionSpaceMinus()->getKeys(em,R);
    unknowns_m.resize(R.size());
    ufield->get(R,unknowns_m);
    R.clear();
    this->getFunctionSpacePlus()->getKeys(ep,R);
    unknowns_p.resize(R.size());
    ufield->get(R,unknowns_p);

    int nbdofinter = _space->getNumKeys(ele);
    int nbvertexInter = ele->getNumVertices();
    int nbFFm = ie->getElem(0)->getNumVertices();
    int nbFFp = ie->getElem(1)->getNumVertices();

    unknowns_inter.resize(nbdofinter);
    unknowns_inter.setAll(0.);

    std::vector<int> vn;
    vn.resize(nbvertexInter);
    ie->getLocalVertexNum(0,vn);
    for(int i=0;i<nbvertexInter;i++){
      unknowns_inter(i) += 0.5*unknowns_m(vn[i]);
      unknowns_inter(i+nbvertexInter) += 0.5*unknowns_m(vn[i]+nbFFm);
      unknowns_inter(i+2*nbvertexInter) += 0.5*unknowns_m(vn[i]+2*nbFFm);
    }
    std::vector<int> vp;
    vp.resize(nbvertexInter);
    ie->getLocalVertexNum(1,vp);
    for(int i=0;i<nbvertexInter;i++){
      unknowns_inter(i) += 0.5*unknowns_p(vp[i]);
      unknowns_inter(i+nbvertexInter) += 0.5*unknowns_p(vp[i]+nbFFp);
      unknowns_inter(i+2*nbvertexInter) +=0.5* unknowns_p(vp[i]+2*nbFFp);
    }

    AllIPState::ipstateElementContainer *vips = aips->getIPstate(ele->getNum());
    int npts=integBound->getIntPoints(ele,&GP);
    _interQuad->getIntPoints(ie,GP,&GPm,&GPp);
    // Grads values on all Gauss points;
    std::vector<std::vector<TensorialTraits<double>::GradType> > vgrads;
    spinter->getgradf(ele,npts,GP,vgrads);

    for(int j=0;j<npts;j++){

      std::vector<TensorialTraits<double>::GradType> &Grads = vgrads[j];

      IPStateBase* ipsm = (*vips)[j];
      IPStateBase* ipsp = (*vips)[j+npts];

      dG3DIPVariableBase* ipvm = static_cast<dG3DIPVariableBase*>(ipsm->getState(ws));
      dG3DIPVariableBase* ipvp = static_cast<dG3DIPVariableBase*>(ipsp->getState(ws));

      // minus outward normal
      static SVector3 referencePhi0[2];
      static SVector3 currentPhi0[2];

      STensorOperation::zero(referencePhi0[0]);
      STensorOperation::zero(referencePhi0[1]);

      STensorOperation::zero(currentPhi0[0]);
      STensorOperation::zero(currentPhi0[1]);

      if (this->getMinusDomain()->getDim() == 3){
        for(int k=0;k<nbvertexInter;k++)
        {
          double x = ele->getVertex(k)->x();
          double y = ele->getVertex(k)->y();
          double z = ele->getVertex(k)->z();
          //evaulate displacements at interface
          for(int i=0;i<2;i++)
          {
            referencePhi0[i](0) += Grads[k](i)*x;
            referencePhi0[i](1) += Grads[k+nbvertexInter](i)*y;
            referencePhi0[i](2) += Grads[k+nbvertexInter+nbvertexInter](i)*z;
            currentPhi0[i](0) += Grads[k](i)*(x+unknowns_inter(k));
            currentPhi0[i](1) += Grads[k+nbvertexInter](i)*(y+unknowns_inter(k+nbvertexInter));
            currentPhi0[i](2) += Grads[k+nbvertexInter+nbvertexInter](i)*(z+unknowns_inter(k+2*nbvertexInter));
          }
        }
      }
      else if (this->getMinusDomain()->getDim() == 2){
        for(int k=0;k<nbvertexInter;k++)
        {
          double x = ele->getVertex(k)->x();
          double y = ele->getVertex(k)->y();
          double z = ele->getVertex(k)->z();
           referencePhi0[0](0) += Grads[k](0)*x;
           referencePhi0[0](1) += Grads[k+nbvertexInter](0)*y;
           referencePhi0[0](2) += Grads[k+nbvertexInter+nbvertexInter](0)*z;
           currentPhi0[0](0) += Grads[k](0)*(x+unknowns_inter(k));
           currentPhi0[0](1) += Grads[k+nbvertexInter](0)*(y+unknowns_inter(k+nbvertexInter));
           currentPhi0[0](2) += Grads[k+nbvertexInter+nbvertexInter](0)*(z+unknowns_inter(k+2*nbvertexInter));
        }

        static SVector3 phi0[2];
        static SVector3 phi0Cur[2];

        STensorOperation::zero(phi0[0]);
        STensorOperation::zero(phi0[1]);

        STensorOperation::zero(phi0Cur[0]);
        STensorOperation::zero(phi0Cur[1]);

        FunctionSpace<double>* spaceMinus = dynamic_cast<FunctionSpace<double>*> (this->getMinusDomain()->getFunctionSpace());

        std::vector<TensorialTraits<double>::GradType> gradmUVW;
        spaceMinus->gradfuvw(em,GPm[j].pt[0],GPm[j].pt[1],GPm[j].pt[2],gradmUVW);
        for(int k=0;k<nbFFm;k++)
        {
          double x = em->getVertex(k)->x();
          double y = em->getVertex(k)->y();
          double z = em->getVertex(k)->z();
          //evaulate displacements at interface
          for(int i=0;i<2;i++)
          {
           phi0[i](0) += gradmUVW[k](i)*x;
           phi0[i](1) += gradmUVW[k+nbFFm](i)*y;
           phi0[i](2) += gradmUVW[k+nbFFm+nbFFm](i)*z;
           phi0Cur[i](0) += gradmUVW[k](i)*(x+unknowns_m(k));
           phi0Cur[i](1) += gradmUVW[k+nbFFm](i)*(y+unknowns_m(k+nbFFm));
           phi0Cur[i](2) += gradmUVW[k+nbFFm+nbFFm](i)*(z+unknowns_m(k+2*nbFFm));
          }
        }
        currentPhi0[1] = crossprod(phi0Cur[0],phi0Cur[1]);
        currentPhi0[1].normalize();
        referencePhi0[1] = crossprod(phi0[0],phi0[1]);
        referencePhi0[1].normalize();
      }

      (ipvm->getRefToCurrentOutwardNormal())       = crossprod(currentPhi0[0],currentPhi0[1]);
      (ipvm->getRefToReferenceOutwardNormal())     = crossprod(referencePhi0[0],referencePhi0[1]);
      (ipvp->getRefToCurrentOutwardNormal())       = crossprod(currentPhi0[0],currentPhi0[1]);
      (ipvp->getRefToReferenceOutwardNormal())     = crossprod(referencePhi0[0],referencePhi0[1]);

			if (mlawMinus!=NULL){
				mlawMinus->initialIPVariable(ipvm,stiff);
			}
			else{
				Msg::Error("Minus material is NULL dG3DDomain::initialIPVariable");
			}

			if (mlawPlus!=NULL){
				mlawPlus->initialIPVariable(ipvp,stiff);
			}
			else{
				Msg::Error("Plus material is NULL dG3DDomain::initialIPVariable");
			}
    }
  }
  // loop on element
  fullVector<double> disp;
  dG3DMaterialLaw* mlaw = dynamic_cast<dG3DMaterialLaw*>(this->getMaterialLaw());
  for(groupOfElements::elementContainer::iterator it=g->begin(); it!=g->end(); ++it)
  {
    MElement *ele = *it;
    // get displacement of element to known the current element configuration
    R.clear();
    this->getFunctionSpace()->getKeys(ele,R);
    disp.resize(R.size());
    ufield->get(R,disp);
    // loop on Gauss's point
    int npts = integBulk->getIntPoints(ele,&GP);
    AllIPState::ipstateElementContainer *vips = aips->getIPstate(ele->getNum());

    for(int j=0;j<npts;j++){
      dG3DIPVariableBase *ipv = static_cast<dG3DIPVariableBase*>((*vips)[j]->getState(ws));
      if (mlaw!=NULL){
        mlaw->initialIPVariable(ipv,stiff);
      };
    }
  }
}
void dG3DDomain::setGaussIntegrationRule(){
  if (this->g->size()>0){
    groupOfElements::elementContainer::iterator it = g->begin();
    MElement *e = *it;
    // default case bulk
    if(_gaussorderbulk == -1)
    {
      if(e->getTypeForMSH() == MSH_TET_4)
      {
        integBulk = new GaussQuadrature(2); // full integration
      }
      else if(e->getTypeForMSH() == MSH_TET_10)
      {
        integBulk = new GaussQuadrature(2); // under integration
      }
      else if(e->getTypeForMSH() == MSH_HEX_27)
      {
        integBulk = new GaussQuadrature(4); // full integration
      }
      else if(e->getTypeForMSH() == MSH_HEX_8)
      {
        integBulk = new GaussQuadrature(2); // full integration
      }
      else if (e->getTypeForMSH() == MSH_QUA_9){
        integBulk = new GaussQuadrature(4);
      }
      else if (e->getTypeForMSH() == MSH_QUA_8){
        integBulk = new GaussQuadrature(4);
      }
      else if (e->getTypeForMSH() == MSH_QUA_12){
        integBulk = new GaussQuadrature(4);
      }
      else if (e->getTypeForMSH() == MSH_QUA_4){
        integBulk = new GaussQuadrature(2);
      }
      else if (e->getTypeForMSH() == MSH_TRI_6){
        integBulk = new GaussQuadrature(2); // 3GPs
      }
      else if (e->getTypeForMSH() == MSH_TRI_3){
        if (this->getNumNonLocalVariable() == 0)
          integBulk = new GaussQuadrature(1);
        else{
          integBulk = new GaussQuadrature(2);
        }
      }
      else if(e->getTypeForMSH() == MSH_TET_20)
      {
        integBulk = new GaussQuadrature(4);
      }
      else if (e->getTypeForMSH() == MSH_PRI_6){
        integBulk = new GaussQuadrature(2);
      }
      else if (e->getTypeForMSH() == MSH_PRI_18){
        integBulk = new GaussQuadrature(4);
      }
      else if (e->getTypeForMSH() == MSH_LIN_2){
        integBulk = new GaussQuadrature(2);
      }
      else if (e->getTypeForMSH() == MSH_LIN_3){
        integBulk = new GaussQuadrature(5);
      }
      else{
        Msg::Error("Bulk gauss quadrature is not defined for this element type by default");
      }
    }
    else{
      integBulk = new GaussQuadrature(_gaussorderbulk);
    }

    IntPt* GP;
    int npts = integBulk->getIntPoints(e,&GP);
    Msg::Info("bulk: use %d integration points for element type %d",npts,e->getType());
  }
  else{
    integBulk = new QuadratureVoid();
  }
    // default case bound

  if (gi->size() > 0){
    groupOfElements::elementContainer::iterator it = gi->begin();
    MElement *ie = *it;
    if(_gaussorderbound == -1)
    {
      if (ie->getTypeForMSH() == MSH_TRI_3)
      {
        if( _gqt == Gauss){
          integBound = new GaussQuadrature(2); // 3 Gauss points
          _interQuad = new interface2DQuadrature(2,false);  // same number of Gauss point
        }
        else if (_gqt == Lobatto)
        {
          integBound = new GaussLobatto2DQuadrature(2); //order 2
          _interQuad = new interface2DQuadrature(2,true);
        }
        else{
          Msg::Error("interface quadarature method %d has not been implemented",_gqt);
        }
      }
      else if (ie->getTypeForMSH() == MSH_TRI_6)
      {
        if( _gqt == Gauss){
          integBound = new GaussQuadrature(4);
          _interQuad = new interface2DQuadrature(4,false);  // same number of Gauss point Gauss & Lobato = 2 ??
        }
        else if (_gqt == Lobatto)
        {
          integBound = new GaussLobatto2DQuadrature(2); //order 2
          _interQuad = new interface2DQuadrature(2,true);  // same number of Gauss point Gauss & Lobato = 2 ??
        }
        else{
          Msg::Error("interface quadarature method %d has not been implemented",_gqt);
        }
      }
      else if (ie->getTypeForMSH() == MSH_QUA_9)
      {
        if( _gqt == Gauss){
          integBound = new GaussQuadrature(4);
         _interQuad = new interface2DQuadrature(4,false);  // same number of Gauss point Gauss
        }
        else if (_gqt == Lobatto)
        {
          integBound = new GaussLobatto2DQuadrature(2); //order 2
          _interQuad = new interface2DQuadrature(2,true);  // same number of Gauss point Gauss
        }
        else{
          Msg::Error("interface quadarature method %d has not been implemented",_gqt);
        }
      }
      else if (ie->getTypeForMSH() == MSH_QUA_4)
      {
        if( _gqt == Gauss)
        {
          integBound = new GaussQuadrature(2);
         _interQuad = new interface2DQuadrature(2,false);  // same number of Gauss point Gauss
        }
        else if (_gqt == Lobatto)
        {
          integBound = new GaussLobatto2DQuadrature(2); //order 2
          _interQuad = new interface2DQuadrature(2,true);
        }
        else{
          Msg::Error("interface quadarature method %d has not been implemented",_gqt);
        }
      }
      else if (ie->getTypeForMSH() == MSH_LIN_4)
      {
        if( _gqt == Gauss){
          integBound = new GaussQuadrature(7); // 4 pts
          _interQuad = new interface1DQuadrature(7);
        }
        else{
          Msg::Error("interface quadarature method %d has not been implemented",_gqt);
        }
      }
      else if (ie->getTypeForMSH() == MSH_LIN_3)
      {
        if( _gqt == Gauss){
          integBound = new GaussQuadrature(5); // 3 pts
          _interQuad = new interface1DQuadrature(5);
        }
        else{
          Msg::Error("interface quadarature method %d has not been implemented",_gqt);
        }
      }
      else if (ie->getTypeForMSH() == MSH_LIN_2)
      {
        if( _gqt == Gauss){
          integBound = new GaussQuadrature(3); // 2 pts
          _interQuad = new interface1DQuadrature(3);
        }
        else{
          Msg::Error("interface quadarature method %d has not been implemented",_gqt);
        }
      }
      else if (ie->getTypeForMSH() == MSH_TRI_10)
      {
        if( _gqt == Gauss){
          integBound = new GaussQuadrature(7);
          _interQuad = new interface2DQuadrature(7,false);
        }
        else if (_gqt == Lobatto)
        {
          integBound = new GaussLobatto2DQuadrature(3); //order 2
          _interQuad = new interface2DQuadrature(3,true);
        }
        else{
          Msg::Error("interface quadarature method %d has not been implemented",_gqt);
        }
      }
      else if (ie->getTypeForMSH() == MSH_PNT){
        integBound = new GaussQuadrature(1);
        _interQuad = new interface1DQuadrature(1);
      }
      else {
        Msg::Error("Interface gauss quadrature is not defined for this element type by default");
      }
    }
    else{
      if(getDim()==3){
        _gqt == Gauss ? integBound = new GaussQuadrature(_gaussorderbound) : integBound = new GaussLobatto2DQuadrature(_gaussorderbound);
        _gqt == Gauss ? _interQuad = new interface2DQuadrature(_gaussorderbound,false) : _interQuad = new interface2DQuadrature(_gaussorderbound,true);
      }
      else{
        integBound = new GaussQuadrature(_gaussorderbound);
        _interQuad = new interface1DQuadrature(_gaussorderbound);
      }
    }

    IntPt* GP;
    int npts = integBound->getIntPoints(ie,&GP);
    Msg::Info("interface: use %d integration points for element type %d",npts,ie->getType());

  }
  else{

    integBound = new QuadratureVoid();
    _interQuad = new interfaceQuadratureVoid();
  }
}

void dG3DDomain::allocateInterfaceMPI(const int rankOtherPart, const groupOfElements &groupOtherPart,dgPartDomain** dgdom)
{
  // domain creation has it function create a domain with a part of the same domain contained on an other part,
  // the type of domain is known
  if(*dgdom==NULL){ // domain doesn't exist --> creation
    (*dgdom) = new interDomainBetween3D(this->getTag(),this,this,0);
    // set the stability parameters
    interDomainBetween3D* newInterDom3D = static_cast< interDomainBetween3D* >(*dgdom);
    newInterDom3D->stabilityParameters(this->stabilityParameter(1));
    // set other options from domain
    newInterDom3D->setBulkDamageBlockedMethod(_bulkDamageBlockMethod);
    newInterDom3D->strainSubstep(_subSteppingMethod,_subStepNumber);
    newInterDom3D->forceCohesiveInsertionAllIPs(_forceInterfaceElementBroken,_failureIPRatio);
    newInterDom3D->setNonLocalStabilityParameters(this->getNonLocalStabilityParameter(),this->getNonLocalContinuity());
    newInterDom3D->setNonLocalEqRatio(this->getNonLocalEqRatio());

    newInterDom3D->setConstitutiveExtraDofDiffusionStabilityParameters(this->getConstitutiveExtraDofDiffusionStabilityParameter(), this->getConstitutiveExtraDofDiffusionContinuity());
    newInterDom3D->setConstitutiveExtraDofDiffusionEqRatio(this->getConstitutiveExtraDofDiffusionEqRatio());
    newInterDom3D->setConstitutiveExtraDofDiffusionUseEnergyConjugatedField(this->getConstitutiveExtraDofDiffusionUseEnergyConjugatedField());
    newInterDom3D->setConstitutiveExtraDofDiffusionAccountSource(this->getConstitutiveExtraDofDiffusionAccountFieldSource(),this->getConstitutiveExtraDofDiffusionAccountMecaSource());

    newInterDom3D->setIncrementNonlocalBased(_incrementNonlocalBased);
		newInterDom3D->setPlaneStressState(_planeStressState);
  }
}


// copy paste of linear change this !!
void dG3DDomain::createInterfaceMPI(const int rankOtherPart, const groupOfElements &groupOtherPart,dgPartDomain* &dgdom)
{
  // domain creation has it function create a domain with a part of the same domain contained on an other part,
  // the type of domain is known
  allocateInterfaceMPI(rankOtherPart, groupOtherPart,&dgdom);
  // create interface (no need to use an interface builder ??)

  if (getDim() ==2){
    std::set<IEdge> map_edge;// the type Iedge is known
    // create interface with groupOtherPart (no match possible match has to be done with dom->gib)
    for(groupOfElements::elementContainer::const_iterator it = groupOtherPart.begin(); it != groupOtherPart.end(); ++it)
    {
      MElement *e=*it;
      int nedge = e->getNumEdges();
      for(int j=0;j<nedge;j++){
        std::vector<MVertex*> vv;
        e->getEdgeVertices(j,vv);
        //map_edge.emplace(vv,e,this);
        IEdge *ie = new IEdge(vv,e,this);
        std::set<IEdge>::const_iterator itie=map_edge.find(*ie);
        if(itie==map_edge.end())
          map_edge.insert(*ie);
        delete ie;
      }
    }
    // now use interface created on boundary to identify the interface element
    for(groupOfElements::elementContainer::iterator it=giv->begin(); it!=giv->end(); ++it) // loop on gib too (normally no need)
    {
      MInterfaceElement *interele = dynamic_cast<MInterfaceElement*>(*it);
      MElement *e = interele->getElem(0); // on boundary interface geetElem(0) == getElem(1)
      int nedge = e->getNumEdges();
      for(int j=0;j<nedge;j++){
        std::vector<MVertex*> vv;
        e->getEdgeVertices(j,vv);
        IEdge ie(vv,e,this);
        std::set<IEdge>::iterator it_edge = map_edge.find(ie);
        if(it_edge != map_edge.end()){
          MElement *interel;
          // the partition with the lowest number is choosen by convention like the one corresponding to minus element
          if(dgdom->getPhysical() == _phys){
            if(Msg::GetCommRank() < rankOtherPart){
              interel =  new MInterfaceLine(vv, 0, Msg::GetCommRank()+1, e, it_edge->getElement());
            }
            else{
              // vv has to be oriented has the minus element --> get the
              std::vector<MVertex*> vvother  = it_edge->getVertices();
              interel = new MInterfaceLine(vvother,0,Msg::GetCommRank()+1,it_edge->getElement(),e);
            }
            _groupGhostMPI->insert(it_edge->getElement());
          }
          else{ // interDomain defined between 2 partitions. The minus element has to be included in the minus domain
            groupOfElements::elementContainer::const_iterator itfound = dgdom->getMinusDomain()->g_cend();
            for(groupOfElements::elementContainer::const_iterator ite=dgdom->getMinusDomain()->g_cbegin(); ite!=dgdom->getMinusDomain()->g_cend();++ite)
            {
              if(e->getNum()==(*ite)->getNum())
              {
                itfound = ite;
                break;
              }
            }
            if(itfound!=dgdom->getMinusDomain()->g_cend())
          {
              interel =  new MInterfaceLine(vv, 0, Msg::GetCommRank()+1, e, it_edge->getElement());
              // add element of other partition to prescribed Dirichlet BC later
              if(e->getPartition() != Msg::GetCommRank()+1)
                 dgdom->getMinusDomain()->_groupGhostMPI->insert(e);
              else
                 dgdom->getPlusDomain()->_groupGhostMPI->insert(it_edge->getElement());

            }
            else{ // e is included in the plus Domain
              // vv has to be oriented has the minus element --> get the
              std::vector<MVertex*> vvother  = it_edge->getVertices();
              interel = new MInterfaceLine(vvother,0,Msg::GetCommRank()+1,it_edge->getElement(),e);
              // add element of other partition to prescribed Dirichlet BC later
              if(e->getPartition() != Msg::GetCommRank()+1)
                 dgdom->getPlusDomain()->_groupGhostMPI->insert(e);
              else
                 dgdom->getMinusDomain()->_groupGhostMPI->insert(it_edge->getElement());
            }
          }
          dgdom->gi->insert(interel);
          // add element of other partition to prescribed Dirichlet BC later
          map_edge.erase(it_edge);
        }
      }
    }
  }
  else if (getDim() == 3) {
    std::set<IFace> map_face;// the type Iedge is known
    // create interface with groupOtherPart (no match possible match has to be done with dom->gib)
    for(groupOfElements::elementContainer::const_iterator it = groupOtherPart.begin(); it != groupOtherPart.end(); ++it)
    {
      MElement *e=*it;
      int nface = e->getNumFaces();
      for(int j=0;j<nface;j++){
        std::vector<MVertex*> vv;
        e->getFaceVertices(j,vv);
        //map_face.emplace(vv,e,this);
        IFace *ie = new IFace(vv,e,this);
        std::set<IFace>::const_iterator itie=map_face.find(*ie);
        if(itie==map_face.end())
          map_face.insert(*ie);
        delete ie;
       
      }
    }
    // now use interface created on boundary to identify the interface element
    for(groupOfElements::elementContainer::iterator it=giv->begin(); it!=giv->end(); ++it) // loop on gib too (normally no need)
    {
      MInterfaceElement *interele = dynamic_cast<MInterfaceElement*>(*it);
      MElement *e = interele->getElem(0); // on boundary interface geetElem(0) == getElem(1)
      int nface = e->getNumFaces();
      for(int j=0;j<nface;j++)
      {
        std::vector<MVertex*> vv;
        e->getFaceVertices(j,vv);
        IFace ie(vv,e,this);
        std::set<IFace>::iterator it_face = map_face.find(ie);
        if(it_face != map_face.end()){
          MElement *interel;
          // the partition with the lowest number is choosen by convention like the one corresponding to minus element
          if(_phys == dgdom->getPhysical())
          {
            if(Msg::GetCommRank() < rankOtherPart)
            {
              if(e->getTypeForMSH() == MSH_TET_4)
              {
                interel =  new MInterfaceTriangleN(vv, 1, 0, Msg::GetCommRank()+1, e, it_face->getElement());
              }
              else if(e->getTypeForMSH() == MSH_TET_10)
              {
                interel =  new MInterfaceTriangleN(vv, 2, 0, Msg::GetCommRank()+1, e, it_face->getElement());
              }
              else if(e->getTypeForMSH() == MSH_HEX_8)
              {
                interel =  new MInterfaceQuadrangleN(vv, 1, 0, Msg::GetCommRank()+1, e, it_face->getElement());
              }
              else if(e->getTypeForMSH() == MSH_HEX_27)
              {
                interel =  new MInterfaceQuadrangleN(vv, 2, 0, Msg::GetCommRank()+1, e, it_face->getElement());
              }
              else if(e->getTypeForMSH() == MSH_TET_20)
              {
                interel =  new MInterfaceTriangleN(vv, 3, 0, Msg::GetCommRank()+1, e, it_face->getElement());
              }
              else if (e->getTypeForMSH() == MSH_PRI_6)
              {
                if (vv.size() == 3){
                  interel =  new MInterfaceTriangleN(vv, 1, 0, Msg::GetCommRank()+1, e, it_face->getElement());
                }
                else if (vv.size() == 4){
                  interel =  new MInterfaceQuadrangleN(vv, 1, 0, Msg::GetCommRank()+1, e, it_face->getElement());
                }
              }
              else if (e->getTypeForMSH() == MSH_PRI_18){
                if (vv.size() == 6){
                  interel =  new MInterfaceTriangleN(vv, 2, 0, Msg::GetCommRank()+1, e, it_face->getElement());
                }
                else if (vv.size() == 9){
                  interel =  new MInterfaceQuadrangleN(vv, 2, 0, Msg::GetCommRank()+1, e, it_face->getElement());
                }
              }
              else
              {
                Msg::Error("dG3DDomain::createInterfaceMPI");
              }
            }
            else
            {
              // vv has to be oriented has the minus element --> get the
              std::vector<MVertex*> vvother  = it_face->getVertices();
              if(it_face->getElement()->getTypeForMSH() == MSH_TET_4)
              {
                interel =  new MInterfaceTriangleN(vvother,1, 0,Msg::GetCommRank()+1,it_face->getElement(),e);
              }
              else if(it_face->getElement()->getTypeForMSH() == MSH_TET_10)
              {
                interel =  new MInterfaceTriangleN(vvother,2, 0,Msg::GetCommRank()+1,it_face->getElement(),e);
              }
              else if(it_face->getElement()->getTypeForMSH() == MSH_HEX_8)
              {
                interel =  new MInterfaceQuadrangleN(vvother,1,0,Msg::GetCommRank()+1,it_face->getElement(),e);
              }
              else if(it_face->getElement()->getTypeForMSH() == MSH_HEX_27)
              {
                interel =  new MInterfaceQuadrangleN(vvother,2,0,Msg::GetCommRank()+1,it_face->getElement(),e);
              }
              else if(it_face->getElement()->getTypeForMSH() == MSH_TET_20)
              {
                interel =  new MInterfaceTriangleN(vvother,3, 0,Msg::GetCommRank()+1,it_face->getElement(),e);
              }
              else if (it_face->getElement()->getTypeForMSH() == MSH_PRI_6){
                if (vvother.size() == 3)
                  interel =  new MInterfaceTriangleN(vvother,1, 0,Msg::GetCommRank()+1,it_face->getElement(),e);
                else if (vvother.size() == 4)
                  interel =  new MInterfaceQuadrangleN(vvother,1,0,Msg::GetCommRank()+1,it_face->getElement(),e);
              }
              else if (it_face->getElement()->getTypeForMSH() == MSH_PRI_18){
                if (vvother.size() == 6)
                  interel =  new MInterfaceTriangleN(vvother,2, 0,Msg::GetCommRank()+1,it_face->getElement(),e);
                else if (vvother.size() == 9)
                  interel =  new MInterfaceQuadrangleN(vvother,2,0,Msg::GetCommRank()+1,it_face->getElement(),e);
              }

              else
              {
                Msg::Error("dG3DDomain::createInterfaceMPI");
              }
            }
            // add element of other partition to prescribed Dirichlet BC later
            _groupGhostMPI->insert(it_face->getElement());
          }
          else{ // interDomain defined between 2 partitions the minus element has to be included in the minus domain
                 // idem when filling _groupGhsotMPI
            groupOfElements::elementContainer::const_iterator itfound = dgdom->getMinusDomain()->g_cend();
            for(groupOfElements::elementContainer::const_iterator ite=dgdom->getMinusDomain()->g_cbegin(); ite!=dgdom->getMinusDomain()->g_cend();++ite)
            {
              if(e->getNum()==(*ite)->getNum())
              {
                itfound = ite;
                break;
              }
            }
            if(itfound!=dgdom->getMinusDomain()->g_cend())
            {
              if(e->getTypeForMSH() == MSH_TET_4)
              {
                interel =  new MInterfaceTriangleN(vv, 1, 0, Msg::GetCommRank()+1, e, it_face->getElement());
              }
              else if(e->getTypeForMSH() == MSH_TET_10)
              {
                interel =  new MInterfaceTriangleN(vv, 2, 0, Msg::GetCommRank()+1, e, it_face->getElement());
              }
              else if(e->getTypeForMSH() == MSH_HEX_8)
              {
                interel =  new MInterfaceQuadrangleN(vv, 1, 0, Msg::GetCommRank()+1, e, it_face->getElement());
              }
              else if(e->getTypeForMSH() == MSH_HEX_27)
              {
                interel =  new MInterfaceQuadrangleN(vv, 2, 0, Msg::GetCommRank()+1, e, it_face->getElement());
              }
              else if(e->getTypeForMSH() == MSH_TET_20)
              {
                interel =  new MInterfaceTriangleN(vv, 3, 0, Msg::GetCommRank()+1, e, it_face->getElement());
              }
              else if(e->getTypeForMSH() == MSH_PRI_6)
              {
                if (vv.size() == 3)
                  interel =  new MInterfaceTriangleN(vv, 1, 0, Msg::GetCommRank()+1, e, it_face->getElement());
                else if (vv.size()== 4)
                  interel =  new MInterfaceQuadrangleN(vv, 1, 0, Msg::GetCommRank()+1, e, it_face->getElement());
              }
              else if(e->getTypeForMSH() == MSH_PRI_18)
              {
                if (vv.size() == 6)
                  interel =  new MInterfaceTriangleN(vv, 2, 0, Msg::GetCommRank()+1, e, it_face->getElement());
                else if (vv.size()== 9)
                  interel =  new MInterfaceQuadrangleN(vv, 2, 0, Msg::GetCommRank()+1, e, it_face->getElement());
              }
              else
              {
                Msg::Error("dG3DDomain::createInterfaceMPI");
              }
              // add element of other partition to prescribed Dirichlet BC later
              if(e->getPartition() != Msg::GetCommRank()+1)
                 dgdom->getMinusDomain()->_groupGhostMPI->insert(e);
              else
                 dgdom->getPlusDomain()->_groupGhostMPI->insert(it_face->getElement());

            }
            else
            {
               // vv has to be oriented has the minus element --> get the
              std::vector<MVertex*> vvother  = it_face->getVertices();
              if(it_face->getElement()->getTypeForMSH() == MSH_TET_4)
              {
                interel =  new MInterfaceTriangleN(vvother,1, 0,Msg::GetCommRank()+1,it_face->getElement(),e);
              }
              else if(it_face->getElement()->getTypeForMSH() == MSH_TET_10)
              {
                interel =  new MInterfaceTriangleN(vvother,2, 0,Msg::GetCommRank()+1,it_face->getElement(),e);
              }
              else if(it_face->getElement()->getTypeForMSH() == MSH_HEX_8)
              {
                interel =  new MInterfaceQuadrangleN(vvother,1,0,Msg::GetCommRank()+1,it_face->getElement(),e);
              }
              else if(it_face->getElement()->getTypeForMSH() == MSH_HEX_27)
              {
                interel =  new MInterfaceQuadrangleN(vvother,2,0,Msg::GetCommRank()+1,it_face->getElement(),e);
              }
              else if(it_face->getElement()->getTypeForMSH() == MSH_TET_20)
              {
                interel =  new MInterfaceTriangleN(vvother,3, 0,Msg::GetCommRank()+1,it_face->getElement(),e);
              }
              else if(it_face->getElement()->getTypeForMSH() == MSH_PRI_6)
              {
                if (vvother.size() == 3)
                  interel =  new MInterfaceTriangleN(vvother,1, 0,Msg::GetCommRank()+1,it_face->getElement(),e);
                else if (vv.size() == 4)
                  interel =  new MInterfaceQuadrangleN(vvother,1,0,Msg::GetCommRank()+1,it_face->getElement(),e);
              }
              else if(it_face->getElement()->getTypeForMSH() == MSH_PRI_18)
              {
                if (vvother.size() == 6)
                  interel =  new MInterfaceTriangleN(vvother,2, 0,Msg::GetCommRank()+1,it_face->getElement(),e);
                else if (vv.size() == 9)
                  interel =  new MInterfaceQuadrangleN(vvother,2,0,Msg::GetCommRank()+1,it_face->getElement(),e);
              }
              else
              {
                Msg::Error("dG3DDomain::createInterfaceMPI");
              }
              // add element of other partition to prescribed Dirichlet BC later
              if(e->getPartition() != Msg::GetCommRank()+1)
                 dgdom->getPlusDomain()->_groupGhostMPI->insert(e);
              else
                 dgdom->getMinusDomain()->_groupGhostMPI->insert(it_face->getElement());
            }
          }
          dgdom->gi->insert(interel);
          map_face.erase(it_face);
        }
      }
    }
  }
}



FilterDof* dG3DDomain::createFilterDof(const int comp) const
{
  FilterDof* fdof = new DG3DFilterDofComponent(comp);
  return fdof;

}

void dG3DDomain::gaussIntegration(const GaussType gqt,const int orderbulk_, const int orderbound_){
  _gqt = gqt;
  _gaussorderbulk = orderbulk_;
  _gaussorderbound= orderbound_;
}


void dG3DDomain::createInterface(manageInterface &maninter)
{
// create all interface even if CG to allow CG/DG mixed formulation
  for (groupOfElements::elementContainer::const_iterator it = g->begin(); it != g->end(); ++it)
  {
    MElement *e=*it;
    if (e->getDim() == 3){
      int nface = e->getNumFaces();
      for(int j=0;j<nface;j++){
          std::vector<MVertex*> vv;
          e->getFaceVertices(j,vv);
          IElement *ie = new IFace(vv,e,this);
          maninter.insert(ie,this);
        }
      }
      else if (e->getDim() == 2){
        int nedge = e->getNumEdges();
        for(int j=0;j<nedge;j++){
          std::vector<MVertex*> vv;
          e->getEdgeVertices(j,vv);
          IElement *ie = new IEdge(vv,e,this);
          maninter.insert(ie,this);
        }
      }
    }
//  }
}

MElement* dG3DDomain::createVirtualInterface(IElement *ie) const
{
  std::vector<MVertex*> Mv  = ie->getVertices();
  MElement *ele;
  if(ie->getElement()->getTypeForMSH() == MSH_TET_4)
  {
    ele = new MInterfaceTriangleN(Mv,1,0,0,ie->getElement(),ie->getElement());
  }
  else if(ie->getElement()->getTypeForMSH() == MSH_TET_10)
  {
    ele = new MInterfaceTriangleN(Mv,2,0,0,ie->getElement(),ie->getElement());
  }
  else if(ie->getElement()->getTypeForMSH() == MSH_HEX_8)
  {
    ele = new MInterfaceQuadrangleN(Mv,1,0,0,ie->getElement(),ie->getElement());
  }
  else if(ie->getElement()->getTypeForMSH() == MSH_HEX_27)
  {
    ele = new MInterfaceQuadrangleN(Mv,2,0,0,ie->getElement(),ie->getElement());
  }
  else if (ie->getElement()->getDim() == 2){
    ele = new MInterfaceLine(Mv,0,0,ie->getElement(),ie->getElement());
  }
  else if(ie->getElement()->getTypeForMSH() == MSH_TET_20)
  {
    ele = new MInterfaceTriangleN(Mv,3,0,0,ie->getElement(),ie->getElement());
  }
  else if (ie->getElement()->getTypeForMSH() == MSH_PRI_6)
  {
    if (Mv.size() == 3){
      ele = new MInterfaceTriangleN(Mv,1,0,0,ie->getElement(),ie->getElement());
    }
    else if (Mv.size() == 4){
      ele = new MInterfaceQuadrangleN(Mv,1,0,0,ie->getElement(),ie->getElement());
    }
  }
  else if (ie->getElement()->getTypeForMSH() == MSH_PRI_18)
  {
    if (Mv.size() == 6){
      ele = new MInterfaceTriangleN(Mv,2,0,0,ie->getElement(),ie->getElement());
    }
    else if (Mv.size() == 9){
      ele = new MInterfaceQuadrangleN(Mv,2,0,0,ie->getElement(),ie->getElement());
    }
  }

  else
  {
    Msg::Error("missing case in dG3DDomain::createVirtualInterface");
  }
  return ele;
}
MElement* dG3DDomain::createInterface(IElement *ie1, IElement *ie2) const
{
  #if defined(HAVE_MPI)
  if(Msg::GetCommSize() != 1){ // with mpi special attention is needed for creation of
                               // interface on interface domain
                               // minus element == element with the lowest part number (if they are different)
    int part1 = ie1->getElement()->getPartition();
    int part2 = ie2->getElement()->getPartition();
    MElement *nelem;
    // HERE WE USE DG FOR
    //         1° 2 domains -> if phys1 != phys2
    //         2° Parallel -> if part1 != part2
    //         3° if choice -> _full dg

    if(ie2->getPhys() > ie1->getPhys() )
    {
      std::vector<MVertex*> vv  = ie1->getVertices();
      if(ie1->getElement()->getTypeForMSH() == MSH_TET_4)
      {
        nelem = new MInterfaceTriangleN(vv, 1, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
      }
      else if(ie1->getElement()->getTypeForMSH() == MSH_TET_10)
      {
        nelem = new MInterfaceTriangleN(vv, 2, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
      }
      else if(ie1->getElement()->getTypeForMSH() == MSH_HEX_8)
      {
        nelem = new MInterfaceQuadrangleN(vv, 1, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
      }
      else if(ie1->getElement()->getTypeForMSH() == MSH_HEX_27)
      {
        nelem = new MInterfaceQuadrangleN(vv, 2, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
      }
      else if (ie1->getElement()->getDim()==2){
        nelem = new MInterfaceLine(vv,0,Msg::GetCommRank()+1,ie1->getElement(),ie2->getElement());
      }
      else if(ie1->getElement()->getTypeForMSH() == MSH_TET_20)
      {
        nelem = new MInterfaceTriangleN(vv, 3, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
      }
      else if (ie1->getElement()->getTypeForMSH() == MSH_PRI_6)
      {
        if (vv.size() == 3){
          nelem = new MInterfaceTriangleN(vv, 1, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
        }
        else if (vv.size() == 4){
          nelem = new MInterfaceQuadrangleN(vv, 1, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
        }
      }
      else if (ie1->getElement()->getTypeForMSH() == MSH_PRI_18)
      {
        if (vv.size() == 6){
          nelem = new MInterfaceTriangleN(vv,2, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
        }
        else if (vv.size() == 9){
          nelem = new MInterfaceQuadrangleN(vv, 2, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
        }
      }
      else
      {
        Msg::Error("missing case in dG3DDomain::createInterface");
        nelem = NULL;
      }
    }
    else if(ie2->getPhys() < ie1->getPhys() )
    {
      std::vector<MVertex*> vv  = ie2->getVertices();
      if(ie2->getElement()->getTypeForMSH() == MSH_TET_4)
      {
        nelem = new MInterfaceTriangleN(vv, 1, 0,Msg::GetCommRank()+1,ie2->getElement(),ie1->getElement());
      }
      else if(ie2->getElement()->getTypeForMSH() == MSH_TET_10)
      {
        nelem = new MInterfaceTriangleN(vv, 2, 0,Msg::GetCommRank()+1,ie2->getElement(),ie1->getElement());
      }
      else if(ie2->getElement()->getTypeForMSH() == MSH_HEX_8)
      {
        nelem = new MInterfaceQuadrangleN(vv, 1, 0,Msg::GetCommRank()+1,ie2->getElement(),ie1->getElement());
      }
      else if(ie2->getElement()->getTypeForMSH() == MSH_HEX_27)
      {
        nelem = new MInterfaceQuadrangleN(vv, 2, 0,Msg::GetCommRank()+1,ie2->getElement(),ie1->getElement());
      }
      else if (ie2->getElement()->getDim()==2){
        nelem  = new MInterfaceLine(vv,0,Msg::GetCommRank()+1,ie2->getElement(),ie1->getElement());
      }
      else if(ie2->getElement()->getTypeForMSH() == MSH_TET_20)
      {
        nelem = new MInterfaceTriangleN(vv, 3, 0,Msg::GetCommRank()+1,ie2->getElement(),ie1->getElement());
      }
      else if (ie2->getElement()->getTypeForMSH() == MSH_PRI_6)
      {
        if (vv.size() == 3){
          nelem = new MInterfaceTriangleN(vv, 1, 0,Msg::GetCommRank()+1,ie2->getElement(),ie1->getElement());
        }
        else if (vv.size() == 4){
          nelem = new MInterfaceQuadrangleN(vv, 1, 0,Msg::GetCommRank()+1,ie2->getElement(),ie1->getElement());
        }
      }
      else if (ie2->getElement()->getTypeForMSH() == MSH_PRI_18)
      {
        if (vv.size() == 6){
          nelem = new MInterfaceTriangleN(vv, 2, 0,Msg::GetCommRank()+1,ie2->getElement(),ie1->getElement());
        }
        else if (vv.size() == 9){
          nelem = new MInterfaceQuadrangleN(vv, 2, 0,Msg::GetCommRank()+1,ie2->getElement(),ie1->getElement());
        }
      }
      else
      {
        Msg::Error("missing case in dG3DDomain::createInterface");
        nelem=NULL;
      }
    }
    else if(part2 > part1){
      std::vector<MVertex*> vv = ie1->getVertices();
      if(ie1->getElement()->getTypeForMSH() == MSH_TET_4)
      {
        nelem = new MInterfaceTriangleN(vv, 1, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
      }
      else if(ie1->getElement()->getTypeForMSH() == MSH_TET_10)
      {
        nelem = new MInterfaceTriangleN(vv, 2, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
      }
      else if(ie1->getElement()->getTypeForMSH() == MSH_HEX_8)
      {
        nelem = new MInterfaceQuadrangleN(vv, 1, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
      }
      else if(ie1->getElement()->getTypeForMSH() == MSH_HEX_27)
      {
        nelem = new MInterfaceQuadrangleN(vv, 2, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
      }
      else if (ie1->getElement()->getDim()==2){
        nelem = new MInterfaceLine(vv,0,Msg::GetCommRank()+1,ie1->getElement(),ie2->getElement());
      }
      else if(ie1->getElement()->getTypeForMSH() == MSH_TET_20)
      {
        nelem = new MInterfaceTriangleN(vv, 3, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
      }
      else if(ie1->getElement()->getTypeForMSH() == MSH_PRI_6)
      {
        if (vv.size() == 3){
          nelem = new MInterfaceTriangleN(vv, 1, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
        }
        else if (vv.size() == 4){
          nelem = new MInterfaceQuadrangleN(vv, 1, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
        }
      }
      else if(ie1->getElement()->getTypeForMSH() == MSH_PRI_18)
      {
        if (vv.size() == 6){
          nelem = new MInterfaceTriangleN(vv, 2, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
        }
        else if (vv.size() == 9){
          nelem = new MInterfaceQuadrangleN(vv, 2, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
        }
      }

      else
      {
        Msg::Error("missing case in dG3DDomain::createInterface");
      }
    }
    else if(part1 > part2){
      std::vector<MVertex*> vv  = ie2->getVertices();
      if(ie2->getElement()->getTypeForMSH() == MSH_TET_4)
      {
        nelem = new MInterfaceTriangleN(vv, 1, 0,Msg::GetCommRank()+1,ie2->getElement(),ie1->getElement());
      }
      else if(ie2->getElement()->getTypeForMSH() == MSH_TET_10)
      {
        nelem = new MInterfaceTriangleN(vv, 2, 0,Msg::GetCommRank()+1,ie2->getElement(),ie1->getElement());
      }
      else if(ie2->getElement()->getTypeForMSH() == MSH_HEX_8)
      {
        nelem = new MInterfaceQuadrangleN(vv, 1, 0,Msg::GetCommRank()+1,ie2->getElement(),ie1->getElement());
      }
      else if(ie2->getElement()->getTypeForMSH() == MSH_HEX_27)
      {
        nelem = new MInterfaceQuadrangleN(vv, 2, 0,Msg::GetCommRank()+1,ie2->getElement(),ie1->getElement());
      }
      else if (ie2->getElement()->getDim()==2){
        nelem = new MInterfaceLine(vv,0,Msg::GetCommRank()+1,ie2->getElement(),ie1->getElement());
      }
      else if(ie2->getElement()->getTypeForMSH() == MSH_TET_20)
      {
        nelem = new MInterfaceTriangleN(vv, 3, 0,Msg::GetCommRank()+1,ie2->getElement(),ie1->getElement());
      }
      else if(ie2->getElement()->getTypeForMSH() == MSH_PRI_6)
      {
        if (vv.size() == 3){
          nelem = new MInterfaceTriangleN(vv, 1, 0,Msg::GetCommRank()+1,ie2->getElement(),ie1->getElement());
        }
        else if (vv.size() == 4){
          nelem = new MInterfaceQuadrangleN(vv, 1, 0,Msg::GetCommRank()+1,ie2->getElement(),ie1->getElement());
        }
      }
      else if(ie2->getElement()->getTypeForMSH() == MSH_PRI_18)
      {
        if (vv.size() == 6){
          nelem = new MInterfaceTriangleN(vv, 2, 0,Msg::GetCommRank()+1,ie2->getElement(),ie1->getElement());
        }
        else if (vv.size() == 9){
          nelem = new MInterfaceQuadrangleN(vv, 2, 0,Msg::GetCommRank()+1,ie2->getElement(),ie1->getElement());
        }
      }
      else
      {
        Msg::Error("missing case in dG3DDomain::createInterface");
        nelem = NULL;
      }
    }
    else if(_fullDg or _isHighOrder) //arbitrary choice
    {
      std::vector<MVertex*> vv  = ie1->getVertices();
      if(ie1->getElement()->getTypeForMSH() == MSH_TET_4)
      {
        nelem = new MInterfaceTriangleN(vv, 1, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
      }
      else if(ie1->getElement()->getTypeForMSH() == MSH_TET_10)
      {
        nelem = new MInterfaceTriangleN(vv, 2, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
      }
      else if(ie1->getElement()->getTypeForMSH() == MSH_HEX_8)
      {
        nelem = new MInterfaceQuadrangleN(vv, 1, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
      }
      else if(ie1->getElement()->getTypeForMSH() == MSH_HEX_27)
      {
        nelem = new MInterfaceQuadrangleN(vv, 2, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
      }
      else if (ie1->getElement()->getDim()==2){
        nelem = new MInterfaceLine(vv,0,Msg::GetCommRank()+1,ie1->getElement(),ie2->getElement());
      }
      else if(ie1->getElement()->getTypeForMSH() == MSH_TET_20)
      {
        nelem = new MInterfaceTriangleN(vv, 3, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
      }
      else if(ie1->getElement()->getTypeForMSH() == MSH_PRI_6)
      {
        if (vv.size() == 3) {
          nelem = new MInterfaceTriangleN(vv, 1, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
        }
        else if (vv.size() == 4){
          nelem = new MInterfaceQuadrangleN(vv, 1, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
        }
      }
      else if(ie1->getElement()->getTypeForMSH() == MSH_PRI_18)
      {
        if (vv.size() == 6) {
          nelem = new MInterfaceTriangleN(vv, 2, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
        }
        else if (vv.size() == 9){
          nelem = new MInterfaceQuadrangleN(vv, 2, 0, Msg::GetCommRank()+1, ie1->getElement(), ie2->getElement());
        }
      }

      else
      {
        Msg::Error("missing case in dG3DDomain::createInterface");
      }
    }
    else
    {
      nelem = NULL;
    }
    if((part1 != Msg::GetCommRank() + 1) and (part1 != 0)) // add ghost interface for Dirichlet BC
    {
      ie1->getDomain()->_groupGhostMPI->insert(ie1->getElement());
    }
    if((part2 != Msg::GetCommRank() + 1) and (part2 != 0)) // add ghost interface for Dirichlet BC
    {
      ie2->getDomain()->_groupGhostMPI->insert(ie2->getElement());
    }
    return nelem;
  }
  else
 #endif // HAVE_MPI
  {
    if(_isHighOrder or _fullDg or (ie1->getPhys() != ie2->getPhys())){
      if(ie1->getPhys() <= ie2->getPhys())
      {
        std::vector<MVertex*> vv = ie1->getVertices();
        MElement *nelem;
        if(ie1->getElement()->getTypeForMSH() == MSH_TET_4)
        {
          nelem = new MInterfaceTriangleN(vv, 1, 0, 0, ie1->getElement(), ie2->getElement());
        }
        else if(ie1->getElement()->getTypeForMSH() == MSH_TET_10)
        {
          nelem = new MInterfaceTriangleN(vv, 2, 0, 0, ie1->getElement(), ie2->getElement());
        }
        else if(ie1->getElement()->getTypeForMSH() == MSH_HEX_8)
        {
          nelem = new MInterfaceQuadrangleN(vv, 1, 0, 0, ie1->getElement(), ie2->getElement());
        }
        else if(ie1->getElement()->getTypeForMSH() == MSH_HEX_27)
        {
          nelem = new MInterfaceQuadrangleN(vv, 2, 0, 0, ie1->getElement(), ie2->getElement());
        }
        else if (ie1->getElement()->getDim()==2){
          nelem = new MInterfaceLine(vv,0,0,ie1->getElement(),ie2->getElement());
        }
        else if(ie1->getElement()->getTypeForMSH() == MSH_TET_20)
        {
          nelem = new MInterfaceTriangleN(vv, 3, 0, 0, ie1->getElement(), ie2->getElement());
        }
        else if (ie1->getElement()->getTypeForMSH() == MSH_PRI_6)
        {
          if (vv.size() == 3){
            nelem = new MInterfaceTriangleN(vv, 1, 0, 0, ie1->getElement(), ie2->getElement());
          }
          else if (vv.size() == 4){
            nelem = new MInterfaceQuadrangleN(vv, 1, 0, 0, ie1->getElement(), ie2->getElement());
          }
        }
        else if (ie1->getElement()->getTypeForMSH() == MSH_PRI_18)
        {
          if (vv.size() == 6){
            nelem = new MInterfaceTriangleN(vv, 2, 0, 0, ie1->getElement(), ie2->getElement());
          }
          else if (vv.size() == 9){
            nelem = new MInterfaceQuadrangleN(vv, 2, 0, 0, ie1->getElement(), ie2->getElement());
          }
        }
        else
        {
          Msg::Error("missing case in dG3DDomain::createInterface");
        }
        return nelem;
      }
      else
      {
        std::vector<MVertex*> vv  = ie2->getVertices();
        MElement *nelem;
        if(ie2->getElement()->getTypeForMSH() == MSH_TET_4)
        {
          nelem = new MInterfaceTriangleN(vv, 1, 0, 0, ie2->getElement(), ie1->getElement());
        }
        else if(ie2->getElement()->getTypeForMSH() == MSH_TET_10)
        {
          nelem = new MInterfaceTriangleN(vv, 2, 0, 0, ie2->getElement(), ie1->getElement());
        }
        else if(ie2->getElement()->getTypeForMSH() == MSH_HEX_8)
        {
          nelem = new MInterfaceQuadrangleN(vv, 1, 0, 0, ie2->getElement(), ie1->getElement());
        }
        else if(ie2->getElement()->getTypeForMSH() == MSH_HEX_27)
        {
          nelem = new MInterfaceQuadrangleN(vv, 2, 0, 0, ie2->getElement(), ie1->getElement());
        }
        else if (ie2->getElement()->getDim()==2){
          nelem = new MInterfaceLine(vv,0,0,ie2->getElement(),ie1->getElement());
        }
        else if(ie2->getElement()->getTypeForMSH() == MSH_TET_20)
        {
          nelem = new MInterfaceTriangleN(vv, 3, 0, 0, ie2->getElement(), ie1->getElement());
        }
        else if(ie2->getElement()->getTypeForMSH() == MSH_PRI_6)
        {
          if (vv.size() == 3){
            nelem = new MInterfaceTriangleN(vv, 1, 0, 0, ie2->getElement(), ie1->getElement());
          }
          else if (vv.size() == 4){
            nelem = new MInterfaceQuadrangleN(vv, 1, 0, 0, ie2->getElement(), ie1->getElement());
          }
        }
        else if(ie2->getElement()->getTypeForMSH() == MSH_PRI_18)
        {
          if (vv.size() == 6){
            nelem = new MInterfaceTriangleN(vv, 2, 0, 0, ie2->getElement(), ie1->getElement());
          }
          else if (vv.size() == 9){
            nelem = new MInterfaceQuadrangleN(vv, 2, 0, 0, ie2->getElement(), ie1->getElement());
          }
        }
        else
        {
          Msg::Error("missing case in dG3DDomain::createInterface");
        }
        return nelem;
      }
    }
    else{
      return NULL;
    }
  }
}


#if defined(HAVE_MPI)
void dG3DDomain::computeIPVariableMPI(AllIPState *aips,const unknownField *ufield,const IPStateBase::whichState ws, bool stiff, bool useBarF){
  IntPt *GP;
  MPI_Status status;
  if (Msg::GetCommRank() == _rootRank){
    // compute all IP strain on root rank
    this->computeAllIPStrain(aips,ufield,ws,false, useBarF);
    // send strain to other procs
    // for bulk elements
    for (std::set<int>::iterator it = _otherRanks.begin(); it!= _otherRanks.end(); it++){
      int otherRank = *it;
      std::set<int>& IPBulkonRank = _mapIPBulk[otherRank];
      for (std::set<int>::iterator its = IPBulkonRank.begin(); its!= IPBulkonRank.end(); its++){
        int num = *its;
        int elnum,gnum;
        numericalMaterialBase::getTwoIntsFromType(num,elnum,gnum);
        AllIPState::ipstateElementContainer *vips = aips->getIPstate(elnum);
        IPStateBase* ips = (*vips)[gnum];
        IPVariable* ipv = ips->getState(ws);
        int bufferSize = ipv->getMacroNumberElementDataSendToMicroProblem(); // already known
        double *buffer = new double[bufferSize];
        ipv->getMacroDataSendToMicroProblem(buffer);
        MPI_Send(buffer,bufferSize,MPI_DOUBLE,otherRank,num,MPI_COMM_WORLD);
        delete[] buffer;
      };

      std::set<int>& IPInterfaceMinusonRank = _mapIPInterfaceMinus[otherRank];
      for (std::set<int>::iterator its = IPInterfaceMinusonRank.begin(); its!= IPInterfaceMinusonRank.end(); its++){
        int num = *its;
        int elnum,gnum;
        numericalMaterialBase::getTwoIntsFromType(num,elnum,gnum);
        AllIPState::ipstateElementContainer *vips = aips->getIPstate(elnum);
        IPStateBase* ips = (*vips)[gnum];
        IPVariable* ipv = ips->getState(ws);
        int bufferSize = ipv->getMacroNumberElementDataSendToMicroProblem(); // already known
        double *buffer = new double[bufferSize];
        ipv->getMacroDataSendToMicroProblem(buffer);
        MPI_Send(buffer,bufferSize,MPI_DOUBLE,otherRank,num,MPI_COMM_WORLD);
        delete[] buffer;
      };

      std::set<int>& IPInterfacePlusonRank = _mapIPInterfacePlus[otherRank];
      for (std::set<int>::iterator its = IPInterfacePlusonRank.begin(); its!= IPInterfacePlusonRank.end(); its++){
        int num = *its;
        int elnum,gnum;
        numericalMaterialBase::getTwoIntsFromType(num,elnum,gnum);
        AllIPState::ipstateElementContainer *vips = aips->getIPstate(elnum);
        IPStateBase* ips = (*vips)[gnum];
        IPVariable* ipv = ips->getState(ws);
        int bufferSize = ipv->getMacroNumberElementDataSendToMicroProblem(); // already known
        double *buffer = new double[bufferSize];
        ipv->getMacroDataSendToMicroProblem(buffer);
        MPI_Send(buffer,bufferSize,MPI_DOUBLE,otherRank,num,MPI_COMM_WORLD);
        delete[] buffer;
      };
    }
  }
  else if (_otherRanks.find(Msg::GetCommRank()) != _otherRanks.end()){
    for (std::set<int>::iterator it = _domainIPBulk.begin(); it != _domainIPBulk.end(); it++){
      int num = *it;
      int elnum,gnum;
      numericalMaterialBase::getTwoIntsFromType(num,elnum,gnum);
      AllIPState::ipstateElementContainer *vips = aips->getIPstate(num);
      IPStateBase* ips = (*vips)[0];
      IPVariable* ipv = ips->getState(ws);
      int bufferSize = ipv->getMacroNumberElementDataSendToMicroProblem();
      double *buffer = new double[bufferSize];
      MPI_Recv(buffer,bufferSize,MPI_DOUBLE,_rootRank,num,MPI_COMM_WORLD,&status);
      ipv->setReceivedMacroDataToMicroProblem(buffer);
      delete [] buffer;
    };

    for (std::set<int>::iterator it = _domainIPInterfaceMinus.begin(); it != _domainIPInterfaceMinus.end(); it++){
      int num = *it;
      int elnum, phys, gnum;
      numericalMaterialBase::getTwoIntsFromType(num,elnum,gnum);
      AllIPState::ipstateElementContainer *vips = aips->getIPstate(num);
      IPStateBase* ips = (*vips)[0];
      IPVariable* ipv = ips->getState(ws);
      int bufferSize = ipv->getMacroNumberElementDataSendToMicroProblem();
      double *buffer = new double[bufferSize];
      MPI_Recv(buffer,bufferSize,MPI_DOUBLE,_rootRank,num,MPI_COMM_WORLD,&status);
      ipv->setReceivedMacroDataToMicroProblem(buffer);
      delete [] buffer;
    };

    for (std::set<int>::iterator it = _domainIPInterfacePlus.begin(); it != _domainIPInterfacePlus.end(); it++){
      int num = *it;
      int elnum,gnum;
      numericalMaterialBase::getTwoIntsFromType(num,elnum,gnum);
      AllIPState::ipstateElementContainer *vips = aips->getIPstate(num);
      IPStateBase* ips = (*vips)[0];
      IPVariable* ipv =  ips->getState(ws);
      int bufferSize = ipv->getMacroNumberElementDataSendToMicroProblem();
      double *buffer = new double[bufferSize];
      MPI_Recv(buffer,bufferSize,MPI_DOUBLE,_rootRank,num,MPI_COMM_WORLD,&status);
      ipv->setReceivedMacroDataToMicroProblem(buffer);
      delete [] buffer;
    };
  };

  for (std::set<int>::iterator it = _domainIPBulk.begin(); it != _domainIPBulk.end(); it++){
    int num = *it;
    int elnum, gnum;
    numericalMaterialBase::getTwoIntsFromType(num,elnum,gnum);
    IPStateBase* ips = NULL;
    if (Msg::GetCommRank() == _rootRank){
      AllIPState::ipstateElementContainer *vips = aips->getIPstate(elnum);
      ips = (*vips)[gnum];
    }
    else if (_otherRanks.find(Msg::GetCommRank()) != _otherRanks.end()){
      AllIPState::ipstateElementContainer *vips = aips->getIPstate(num);
      ips = (*vips)[0];
    }
    IPVariable *ipv = ips->getState(ws);
    const IPVariable *ipvprev = ips->getState(IPStateBase::previous);
    dG3DMaterialLaw *mlaw = dynamic_cast<dG3DMaterialLaw*>(this->getMaterialLaw());
    if (_planeStressState)
      mlaw->stress3DTo2D(ipv,ipvprev,stiff,false);
    else
      mlaw->stress(ipv,ipvprev,stiff,false);
  };

  for (std::set<int>::iterator it = _domainIPInterfaceMinus.begin(); it != _domainIPInterfaceMinus.end(); it++){
    int num = *it;
    int elnum, gnum;
    numericalMaterialBase::getTwoIntsFromType(num,elnum,gnum);
    IPStateBase* ips = NULL;
    if (Msg::GetCommRank() == _rootRank){
      AllIPState::ipstateElementContainer *vips = aips->getIPstate(elnum);
      ips = (*vips)[gnum];
    }
    else if (_otherRanks.find(Msg::GetCommRank()) != _otherRanks.end()){
      AllIPState::ipstateElementContainer *vips = aips->getIPstate(num);
      ips = (*vips)[0];
    }
    IPVariable *ipv = ips->getState(ws);
    const IPVariable *ipvprev = ips->getState(IPStateBase::previous);
    dG3DMaterialLaw *mlawMinus = dynamic_cast<dG3DMaterialLaw*>(this->getMaterialLawMinus());
		if (_planeStressState)
			mlawMinus->stress3DTo2D(ipv,ipvprev,stiff,true);
		else
    	mlawMinus->stress(ipv,ipvprev,stiff,true);
  };


  for (std::set<int>::iterator it = _domainIPInterfacePlus.begin(); it != _domainIPInterfacePlus.end(); it++){
    int num = *it;
    int elnum,gnum;
    numericalMaterialBase::getTwoIntsFromType(num,elnum,gnum);
    IPStateBase* ips = NULL;
    if (Msg::GetCommRank() == _rootRank){
      AllIPState::ipstateElementContainer *vips = aips->getIPstate(elnum);
      ips = (*vips)[gnum];
    }
    else if (_otherRanks.find(Msg::GetCommRank()) != _otherRanks.end()){
      AllIPState::ipstateElementContainer *vips = aips->getIPstate(num);
      ips = (*vips)[0];
    }
    IPVariable *ipv = ips->getState(ws);
    const IPVariable *ipvprev = ips->getState(IPStateBase::previous);
    dG3DMaterialLaw *mlawPlus= dynamic_cast<dG3DMaterialLaw*>(this->getMaterialLawPlus());
		if (_planeStressState)
			mlawPlus->stress3DTo2D(ipv,ipvprev,stiff,true);
		else
    	mlawPlus->stress(ipv,ipvprev,stiff,true);
  }


  if (_otherRanks.find(Msg::GetCommRank()) != _otherRanks.end()){
    for (std::set<int>::iterator it = _domainIPBulk.begin(); it != _domainIPBulk.end(); it++){
      int num = *it;
      int elnum,gnum;
      numericalMaterialBase::getTwoIntsFromType(num,elnum,gnum);
      AllIPState::ipstateElementContainer *vips = aips->getIPstate(num);
      IPStateBase* ips = (*vips)[0];
      IPVariable* ipv = ips->getState(ws);
      int bufferSize = ipv->getMicroNumberElementDataSendToMacroProblem();
      double *buffer = new double[bufferSize];
      ipv->getMicroDataToMacroProblem(buffer);
      MPI_Send(buffer,bufferSize,MPI_DOUBLE,_rootRank,num,MPI_COMM_WORLD);
      delete [] buffer;
    };
    for (std::set<int>::iterator it = _domainIPInterfaceMinus.begin(); it != _domainIPInterfaceMinus.end(); it++){
      int num = *it;
      int elnum,gnum;
      numericalMaterialBase::getTwoIntsFromType(num,elnum,gnum);
      AllIPState::ipstateElementContainer *vips = aips->getIPstate(num);
      IPStateBase* ips = (*vips)[0];
      IPVariable* ipv = ips->getState(ws);
      int bufferSize = ipv->getMicroNumberElementDataSendToMacroProblem();
      double *buffer = new double[bufferSize];
      ipv->getMicroDataToMacroProblem(buffer);
      MPI_Send(buffer,bufferSize,MPI_DOUBLE,_rootRank,num,MPI_COMM_WORLD);
      delete [] buffer;
    };
    for (std::set<int>::iterator it = _domainIPInterfacePlus.begin(); it != _domainIPInterfacePlus.end(); it++){
      int num = *it;
      int elnum,gnum;
      numericalMaterialBase::getTwoIntsFromType(num,elnum,gnum);
      AllIPState::ipstateElementContainer *vips = aips->getIPstate(num);
      IPStateBase* ips = (*vips)[0];
      IPVariable* ipv = ips->getState(ws);
      int bufferSize = ipv->getMicroNumberElementDataSendToMacroProblem();
      double *buffer = new double[bufferSize];
      ipv->getMicroDataToMacroProblem(buffer);
      MPI_Send(buffer,bufferSize,MPI_DOUBLE,_rootRank,num,MPI_COMM_WORLD);
      delete [] buffer;
    };
  }
  else if (Msg::GetCommRank() == _rootRank){

    for (std::set<int>::iterator it = _otherRanks.begin(); it!= _otherRanks.end(); it++){
      int otherRank = *it;
      std::set<int>& IPBulkonRank = _mapIPBulk[otherRank];
      for (std::set<int>::iterator its = IPBulkonRank.begin(); its!= IPBulkonRank.end(); its++){
        int num = *its;
        int elnum,gnum;
        numericalMaterialBase::getTwoIntsFromType(num,elnum,gnum);
        AllIPState::ipstateElementContainer *vips = aips->getIPstate(elnum);
        IPStateBase* ips = (*vips)[gnum];
        IPVariable* ipv = ips->getState(ws);
        int bufferSize = ipv->getMicroNumberElementDataSendToMacroProblem();
        double *buffer = new double[bufferSize];
        MPI_Recv(buffer,bufferSize,MPI_DOUBLE,otherRank,num,MPI_COMM_WORLD,&status);
        ipv->setReceivedMicroDataToMacroProblem(buffer);
        dG3DMaterialLaw* mlaw = dynamic_cast<dG3DMaterialLaw*>(this->getMaterialLaw());
        mlaw->setElasticStiffness(ipv);
        delete[] buffer;
      };
      std::set<int>& IPInterfaceMinusonRank = _mapIPInterfaceMinus[otherRank];
      for (std::set<int>::iterator its = IPInterfaceMinusonRank.begin(); its!= IPInterfaceMinusonRank.end(); its++){
        int num = *its;
        int elnum,gnum;
        numericalMaterialBase::getTwoIntsFromType(num,elnum,gnum);
        AllIPState::ipstateElementContainer *vips = aips->getIPstate(elnum);
        IPStateBase* ips = (*vips)[gnum];
        IPVariable* ipv = ips->getState(ws);
        int bufferSize = ipv->getMicroNumberElementDataSendToMacroProblem();
        double *buffer = new double[bufferSize];
        MPI_Recv(buffer,bufferSize,MPI_DOUBLE,otherRank,num,MPI_COMM_WORLD,&status);
        ipv->setReceivedMicroDataToMacroProblem(buffer);
        dG3DMaterialLaw* mlawMinus = dynamic_cast<dG3DMaterialLaw*>(this->getMaterialLawMinus());
        mlawMinus->setElasticStiffness(ipv);
        delete[] buffer;
      };

      std::set<int>& IPInterfacePlusonRank = _mapIPInterfacePlus[otherRank];
      for (std::set<int>::iterator its = IPInterfacePlusonRank.begin(); its!= IPInterfacePlusonRank.end(); its++){
        int num = *its;
        int elnum,gnum;
        numericalMaterialBase::getTwoIntsFromType(num,elnum,gnum);
        AllIPState::ipstateElementContainer *vips = aips->getIPstate(elnum);
        IPStateBase* ips = (*vips)[gnum];
        IPVariable* ipv = ips->getState(ws);
        int bufferSize = ipv->getMicroNumberElementDataSendToMacroProblem();
        double *buffer = new double[bufferSize];
        MPI_Recv(buffer,bufferSize,MPI_DOUBLE,otherRank,num,MPI_COMM_WORLD,&status);
        ipv->setReceivedMicroDataToMacroProblem(buffer);
        dG3DMaterialLaw* mlawPlus = dynamic_cast<dG3DMaterialLaw*>(this->getMaterialLawPlus());
        mlawPlus->setElasticStiffness(ipv);
        delete[] buffer;
      };
    }

    if (_averageStrainBased and (this->getMaterialLawMinus()->getNum() == this->getMaterialLawPlus()->getNum())){
      for (groupOfElements::elementContainer::iterator ite = gi->begin(); ite!= gi->end(); ite++){
        MElement* ele = *ite;
        int npts = this->getInterfaceGaussIntegrationRule()->getIntPoints(ele,&GP);
        AllIPState::ipstateElementContainer *vips = aips->getIPstate(ele->getNum());
        for(int j=0;j<npts;j++){
          IPStateBase* ipsm = (*vips)[j];
          IPStateBase* ipsp = (*vips)[j+npts];

          IPVariable* ipvm = ipsm->getState(IPStateBase::current);
          IPVariable* ipvp = ipsp->getState(IPStateBase::current);

          ipvp->operator=(*ipvm);
        }

      }
    }
  };
};
#endif

interDomainBase::interDomainBase(partDomain *dom1,partDomain *dom2, const int bnum) :  _mlawMinus(NULL), _mlawPlus(NULL), _bulkNum(bnum),
    _lawFromBulkPart(false)
{
  // By convention the domain with the small number is the domMinus
  // For MPI in domMinus->getPhysical() = domPlus->getPhysical() --> domMinus = domain on the partion with the lowest rank

  // look for element (INTERDOMAIN HAVE TO BE CREATED AFTER addDomain to solver)
  if(dom2->getPhysical() > dom1->getPhysical())
  {
    _domMinus = dom1;
    _domPlus = dom2;
  }
  else if(dom1->getPhysical() > dom2->getPhysical())
  {
    _domMinus = dom2;
    _domPlus = dom1;
  }
#if defined(HAVE_MPI)
  else if((Msg::GetCommSize() > 1) and (dom1->getPhysical() == dom2->getPhysical()))
  {
    _domMinus = dom1; // same --> no matter
    _domPlus = dom2;
  }
  else if ( (Msg::GetCommSize() > 1) and ((dom1->groupOfElementsSize() !=0) and (dom2->groupOfElementsSize() !=0))){
    // Check if the two domains are added to solver
    groupOfElements::elementContainer::const_iterator it1 = dom1->g_cbegin();
    const MElement *ele1 = *it1;
    groupOfElements::elementContainer::const_iterator it2 = dom2->g_cbegin();
    const MElement *ele2 = *it2;
    if(ele1->getPartition() < ele2->getPartition())
    {
      _domMinus = dom1;
      _domPlus = dom2;
    }
    else if(ele2->getPartition() < ele1->getPartition())
    {
      _domMinus = dom2;
      _domPlus = dom1;
    }
  }
 #endif // HAVE_MPI
  else  // same partition choose in function of physical number
  {
    Msg::Error("Impossible to known the partion of domain --> no inter domain created between %d and %d",dom1->getPhysical(),dom2->getPhysical());
  }

  _spaceMinus = _domMinus->getFunctionSpace();
  _spacePlus = _domPlus->getFunctionSpace();

}

interDomainBase::interDomainBase(const interDomainBase &source) : _domMinus(source._domMinus),
                                                                  _domPlus(source._domPlus),
                                                                  _mlawMinus(source._mlawMinus),
                                                                  _mlawPlus(source._mlawPlus),
                                                                  _spaceMinus(source._spaceMinus),
                                                                  _spacePlus(source._spacePlus),
                                                                  _bulkNum(source._bulkNum),
                                                                  _lawFromBulkPart(source._lawFromBulkPart)
{

}

void interDomainBase::setLawFromBulkPartFlag(const bool fl){_lawFromBulkPart = fl;};

void interDomainBase::initializeFractureBase(unknownField *uf,IPField*ip,
                                         groupOfElements *gi, QuadratureBase* integBound)
{

  // same strength fracture value on both side (to avoid discontinuities)
  // no very elegant to put here but very efficient
  // take the value of minus element by choice
     if(_mlawMinus->getType() == materialLaw::fracture and _mlawPlus->getType() == materialLaw::fracture)
     {
       // loop on ipvariable
       IntPt *GP;
       for(groupOfElements::elementContainer::iterator it=gi->begin(); it!=gi->end(); ++it)
       {
         MElement *ele = *it;
         int npts_inter=integBound->getIntPoints(ele,&GP);
         AllIPState::ipstateElementContainer *vips = ip->getAips()->getIPstate(ele->getNum());
         for(int j=0;j<npts_inter;j++){ // npts_inter points on both sides
           IPStateBase* ipsm = (*vips)[j];
           IPStateBase* ipsp = (*vips)[j+npts_inter];
           FractureCohesive3DIPVariable *ipvm = static_cast<FractureCohesive3DIPVariable*>(ipsm->getState(IPStateBase::initial));
           FractureCohesive3DIPVariable *ipvp = static_cast<FractureCohesive3DIPVariable*>(ipsp->getState(IPStateBase::initial));
           ipvp->setFractureStrengthFactor(ipvm->fractureStrengthFactor());
           ipvm = static_cast<FractureCohesive3DIPVariable*>(ipsm->getState(IPStateBase::previous));
           ipvp = static_cast<FractureCohesive3DIPVariable*>(ipsp->getState(IPStateBase::previous));
           ipvp->setFractureStrengthFactor(ipvm->fractureStrengthFactor());
           ipvm = static_cast<FractureCohesive3DIPVariable*>(ipsm->getState(IPStateBase::current));
           ipvp = static_cast<FractureCohesive3DIPVariable*>(ipsp->getState(IPStateBase::current));
           ipvp->setFractureStrengthFactor(ipvm->fractureStrengthFactor());
         }
       }
     }
}

void interDomainBase::setGaussIntegrationRuleBase(groupOfElements *gi, QuadratureBase** integBound, QuadratureBase** integBulk,
                                                  interfaceQuadratureBase** interQuad, int _gaussorderbound, dG3DDomain::GaussType _gqt)
{
 #if defined(HAVE_MPI)
   // InterDomain defined by user can be empty on a partition !!
  if(gi->size() == 0)
  {
    *integBound = new QuadratureVoid();
    *interQuad = new interface2DQuadrature(0,false);
  }
  else
 #endif //HAVE_MPI
  {
    groupOfElements::elementContainer::const_iterator it = gi->begin();
    MElement *e = *it;
    if(_gaussorderbound == -1)
    {
      if(e->getTypeForMSH() == MSH_TET_4 or e->getTypeForMSH() == MSH_TRI_3)
      {
        if( _gqt == dG3DDomain::Gauss)
        {
              *integBound = new GaussQuadrature(2);
              *interQuad = new interface2DQuadrature(2,false);  // same number of Gauss point Gauss
        }
        else
        {
          *integBound = new GaussLobatto2DQuadrature(2); //order 2
          *interQuad = new interface2DQuadrature(2,true);
        }
      }
      else if(e->getTypeForMSH() == MSH_TET_10 or e->getTypeForMSH() == MSH_TRI_6)
      {
        if( _gqt == dG3DDomain::Gauss)
        {
              *integBound = new GaussQuadrature(4);
              *interQuad = new interface2DQuadrature(4,false);  // same number of Gauss points, why was it 2?
        }
        else
        {
              *integBound = new GaussLobatto2DQuadrature(2); //order 2
              *interQuad = new interface2DQuadrature(2,true);  // same number of Gauss point Gauss & Lobato = 2 ??
        }
      }
      else if(e->getTypeForMSH() == MSH_HEX_27 or e->getTypeForMSH() == MSH_QUA_9)
      {
        if( _gqt == dG3DDomain::Gauss)
        {
              *integBound = new GaussQuadrature(4);
              *interQuad = new interface2DQuadrature(4,false);  // same number of Gauss point Gauss
        }
        else
        {
              *integBound = new GaussLobatto2DQuadrature(2); //order 2
              *interQuad = new interface2DQuadrature(2,true);
        }
      }
      else if(e->getTypeForMSH() == MSH_HEX_8 or e->getTypeForMSH() == MSH_QUA_4)
      {
        if( _gqt == dG3DDomain::Gauss)
        {
              *integBound = new GaussQuadrature(3);
              *interQuad = new interface2DQuadrature(3,false);  // same number of Gauss point
        }
        else
        {
              *integBound = new GaussLobatto2DQuadrature(2); //order 2
              *interQuad = new interface2DQuadrature(2,true);
        }
      }
      else if (e->getTypeForMSH() == MSH_LIN_2){
        *integBound = new GaussQuadrature(2); // 2 Gauss points
        *interQuad = new interface1DQuadrature(2);
      }
      else if (e->getTypeForMSH() == MSH_LIN_3){
        *integBound = new GaussQuadrature(5); // 3 Gauss  point
        *interQuad = new interface1DQuadrature(5);
      }
      else if(e->getTypeForMSH() == MSH_TET_20 or e->getTypeForMSH() == MSH_TRI_10)
      {
        if( _gqt == dG3DDomain::Gauss)
        {
              *integBound = new GaussQuadrature(7);
              *interQuad = new interface2DQuadrature(7,false);  // same number of Gauss point Gauss & Lobato = 2 ??
        }
        else
        {
              *integBound = new GaussLobatto2DQuadrature(3); //order 3
              *interQuad = new interface2DQuadrature(3,true);  // same number of Gauss point Gauss & Lobato = 2 ??
        }
      }
      else{
        Msg::Error("Interface gauss quadrature is not defined for this element type by default");
      }
    }
    else{
      if (_domMinus->getDim()==3){
        _gqt == dG3DDomain::Gauss ? *integBound = new GaussQuadrature(_gaussorderbound) : *integBound = new GaussLobatto2DQuadrature(_gaussorderbound);
        _gqt == dG3DDomain::Gauss ? *interQuad = new interface2DQuadrature(_gaussorderbound,false) : *interQuad = new interface2DQuadrature(_gaussorderbound,true);
      }
      else if (_domMinus->getDim()==2){
        *integBound = new GaussQuadrature(_gaussorderbound);
        *interQuad = new interface1DQuadrature(_gaussorderbound);
      }

    }
   }
}


interDomainBetween3D::interDomainBetween3D(const int tag, partDomain *dom1,partDomain *dom2,const int lnum, const int bnum) :
                                  interDomainBase(dom1,dom2,bnum),
                                  dG3DDomain(tag,manageInterface::getKey(dom1->getPhysical(),
                                                                    dom2->getPhysical()),10000,lnum,true,dom1->getDim(),
                                                                    dom1->getNumNonLocalVariable(),
                                                                    dom1->getNumConstitutiveExtraDofDiffusionVariable())
{
  if(dom1->getNumNonLocalVariable()<dom2->getNumNonLocalVariable() or dom1->getNumConstitutiveExtraDofDiffusionVariable()<dom2->getNumConstitutiveExtraDofDiffusionVariable())
    Msg::Error("interDomainBetween3D::interDomainBetween3D: domain 2 has less variable than domain 1");
  // create empty groupOfElements
  g = new groupOfElements();
  // set functionalSpace

  if((_domMinus->getFunctionSpaceType() == functionSpaceType::Lagrange) and
     (_domPlus->getFunctionSpaceType() == functionSpaceType::Lagrange)){
    if(_space != NULL) delete _space;
    _space = new dG3DLagrangeBetween2DomainsFunctionSpace(3+getNumNonLocalVariable()+getNumConstitutiveExtraDofDiffusionVariable(),_spaceMinus,_spacePlus);
    _wsp = functionSpaceType::Lagrange;
  }
  else{
    Msg::Error("Impossible to create functional space on interDomainBetween3D %d",_phys);
  }

  _nonLocalBeta = (static_cast<dG3DDomain *> (dom1))-> getNonLocalStabilityParameter();
  _nonLocalContinuity = (static_cast<dG3DDomain *> (dom1))->getNonLocalContinuity();
  _nonLocalEqRatio =  (static_cast<dG3DDomain *> (dom1))->getNonLocalEqRatio();

  _constitutiveExtraDofDiffusionBeta = (static_cast<dG3DDomain *> (dom1))->getConstitutiveExtraDofDiffusionStabilityParameter();
  _constitutiveExtraDofDiffusionContinuity = (static_cast<dG3DDomain *> (dom1))->getConstitutiveExtraDofDiffusionContinuity();
  _constitutiveExtraDofDiffusionEqRatio = (static_cast<dG3DDomain *> (dom1))->getConstitutiveExtraDofDiffusionEqRatio();
  _constitutiveExtraDofDiffusionUseEnergyConjugatedField = (static_cast<dG3DDomain *> (dom1))->getConstitutiveExtraDofDiffusionUseEnergyConjugatedField();
  _constitutiveExtraDofDiffusionAccountFieldSource = (static_cast<dG3DDomain *> (dom1))->getConstitutiveExtraDofDiffusionAccountFieldSource();
  _constitutiveExtraDofDiffusionAccountMecaSource = (static_cast<dG3DDomain *> (dom1))->getConstitutiveExtraDofDiffusionAccountMecaSource();

}


interDomainBetween3D::interDomainBetween3D(const interDomainBetween3D &source) : interDomainBase(source), dG3DDomain(source)
{}

void interDomainBetween3D::createTerms(unknownField *uf,IPField*ip)
{
  // void term (as they have to exist)
  this->ltermBulk = new nonLinearTermVoid();
  this->btermBulk = new BiNonLinearTermVoid();

  this->btermVirtBound = new BiNonLinearTermVoid();
  this->ltermVirtBound = new nonLinearTermVoid();
  this->massterm = new BiNonLinearTermVoid();

  FunctionSpace<double>* dgspace = static_cast<FunctionSpace<double>*>(_space);
  FunctionSpace<double>* dgspace1 = static_cast<FunctionSpace<double>*>(_spaceMinus);
  FunctionSpace<double>* dgspace2 = static_cast<FunctionSpace<double>*>(_spacePlus);
  dG3DDomain* dgldomMinus = static_cast<dG3DDomain*>(_domMinus);
  dG3DDomain* dgldomPlus = static_cast<dG3DDomain*>(_domPlus);
  if (IsMultipleFieldFormulation()){
    if (this->getMultipleFieldMethod() == dgPartDomain::TWOFIELD){
      ltermBound = new twoFieldDG3DForceInter(this,ip,this->_EqMultiFieldFactor,this->_beta1);
    }
    else{
      Msg::Error("two field method %d has not been correctly defined",getMultipleFieldMethod());
    }
  }
  else
    ltermBound = new dG3DForceInter(*dgspace1,dgspace2,_mlawMinus,_mlawPlus,_interQuad,_beta1,true,ip,_incrementNonlocalBased);
  if(_interByPert)
    btermBound = new BilinearTermPerturbation<double>(ltermBound,*dgspace1,*dgspace2,uf,ip,this,_eps);
  else{
    if (IsMultipleFieldFormulation()){
      if (this->getMultipleFieldMethod() == dgPartDomain::TWOFIELD){
        btermBound = new twoFieldDG3DStiffnessInter(this,ip,this->_EqMultiFieldFactor,this->_beta1);
      }
      else
      {
        Msg::Error("two field method %d has not been correctly defined",getMultipleFieldMethod());
      }
    }
    else
      btermBound = new dG3DStiffnessInter(*dgspace1,*dgspace2,_mlawMinus,_mlawPlus,_interQuad,_beta1,ip,uf,true,_incrementNonlocalBased);
  }

  scalarTermPF = new nonLinearScalarTermVoid();
  linearTermPF = new nonLinearTermVoid();

  if( getNumNonLocalVariable() >0)
  {
    if(!IsMultipleFieldFormulation())
    {
        (static_cast<dG3DForceInter*>(ltermBound))->setNonLocalEqRatio(getNonLocalEqRatio());
        (static_cast<dG3DForceInter*>(ltermBound))->setNumNonLocalVariable(getNumNonLocalVariable());
        (static_cast<dG3DForceInter*>(ltermBound))->setNonLocalStabilityParameter(getNonLocalStabilityParameter());
        (static_cast<dG3DForceInter*>(ltermBound))->setNonLocalContinuity(getNonLocalContinuity());
        if(!_interByPert)
        {
          (static_cast<dG3DStiffnessInter*>(btermBound))->setNonLocalEqRatio(getNonLocalEqRatio());
          (static_cast<dG3DStiffnessInter*>(btermBound))->setNumNonLocalVariable(getNumNonLocalVariable());
          (static_cast<dG3DStiffnessInter*>(btermBound))->setNonLocalStabilityParameter(getNonLocalStabilityParameter());
          (static_cast<dG3DStiffnessInter*>(btermBound))->setNonLocalContinuity(getNonLocalContinuity());
        }
    }
  }
  if (getNumConstitutiveExtraDofDiffusionVariable()>0)
  {
    if(!IsMultipleFieldFormulation())
    {
      (static_cast<dG3DForceInter*>(ltermBound))->setNumConstitutiveExtraDofDiffusionVariable(getNumConstitutiveExtraDofDiffusionVariable());
      (static_cast<dG3DForceInter*>(ltermBound))->setConstitutiveExtraDofDiffusionEqRatio(getConstitutiveExtraDofDiffusionEqRatio());
      (static_cast<dG3DForceInter*>(ltermBound))->setConstitutiveExtraDofDiffusionUseEnergyConjugatedField(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField());
      (static_cast<dG3DForceInter*>(ltermBound))->setConstitutiveExtraDofDiffusionAccountFieldSource(getConstitutiveExtraDofDiffusionAccountFieldSource());
      (static_cast<dG3DForceInter*>(ltermBound))->setConstitutiveExtraDofDiffusionAccountMecaSource(getConstitutiveExtraDofDiffusionAccountMecaSource());
      (static_cast<dG3DForceInter*>(ltermBound))->setConstitutiveExtraDofDiffusionStabilityParameter(getConstitutiveExtraDofDiffusionStabilityParameter());
      (static_cast<dG3DForceInter*>(ltermBound))->setConstitutiveExtraDofDiffusionContinuity(getConstitutiveExtraDofDiffusionContinuity());
      if(!_interByPert)
      {
        (static_cast<dG3DStiffnessInter*>(btermBound))->setNumConstitutiveExtraDofDiffusionVariable(getNumConstitutiveExtraDofDiffusionVariable());
        (static_cast<dG3DStiffnessInter*>(btermBound))->setConstitutiveExtraDofDiffusionEqRatio(getConstitutiveExtraDofDiffusionEqRatio());
        (static_cast<dG3DStiffnessInter*>(btermBound))->setConstitutiveExtraDofDiffusionUseEnergyConjugatedField(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField());
        (static_cast<dG3DStiffnessInter*>(btermBound))->setConstitutiveExtraDofDiffusionAccountFieldSource(getConstitutiveExtraDofDiffusionAccountFieldSource());
        (static_cast<dG3DStiffnessInter*>(btermBound))->setConstitutiveExtraDofDiffusionAccountMecaSource(getConstitutiveExtraDofDiffusionAccountMecaSource());
        (static_cast<dG3DStiffnessInter*>(btermBound))->setConstitutiveExtraDofDiffusionStabilityParameter(getConstitutiveExtraDofDiffusionStabilityParameter());
        (static_cast<dG3DStiffnessInter*>(btermBound))->setConstitutiveExtraDofDiffusionContinuity(getConstitutiveExtraDofDiffusionContinuity());
      }
    }
  }
  if (this->getMacroSolver()->withPathFollowing() and _accountPathFollowing){
    if (this->getMacroSolver()->getPathFollowingLocation() == nonLinearMechSolver::BULK){
      scalarTermPFBound = new nonLinearScalarTermVoid();
      linearTermPFBound = new nonLinearTermVoid();
    }
    else{
      if (this->IsMultipleFieldFormulation()){
        scalarTermPFBound = new twoFieldDG3DDissipationPathFollowingBoundScalarTerm(this,ip);
        linearTermPFBound = new twoFieldDG3DDissipationPathFollowingBoundLinearTerm(this,ip);
      }
      else{
        scalarTermPFBound = new dG3DDissipationPathFollowingBoundScalarTerm(this,ip);
        linearTermPFBound = new dG3DDissipationPathFollowingBoundLinearTerm(this,ip);
      }
    }
  }
  else{
    scalarTermPFBound = new nonLinearScalarTermVoid();
    linearTermPFBound = new nonLinearTermVoid();
  }
}
void interDomainBetween3D::initializeTerms(unknownField *uf,IPField*ip)
{

  createTerms(uf,ip);
  initializeFractureBase(uf,ip, gi, integBound);
}


void interDomainBetween3D::computeAllIPStrain(AllIPState *aips,const unknownField *ufield,const IPStateBase::whichState ws, const bool virt, bool useBarF){
  IntPt *GP;
  fullVector<double> dispm;
  fullVector<double> dispp;
  fullVector<double> dispExtra;
  std::vector<Dof> R;

  for(groupOfElements::elementContainer::iterator it=gi->begin(); it!=gi->end(); ++it){
    MInterfaceElement *iele = dynamic_cast<MInterfaceElement*>(*it);
    R.clear();
    this->getFunctionSpaceMinus()->getKeys(iele->getElem(0),R);
    dispm.resize(R.size());
    ufield->get(R,dispm);
    R.clear();
    this->getFunctionSpacePlus()->getKeys(iele->getElem(1),R);
    dispp.resize(R.size());
    ufield->get(R,dispp);

    if (this->IsMultipleFieldFormulation()){
      FunctionSpaceBase* extraSp = this->getInterfaceFunctionSpace();
      R.clear();
      extraSp->getKeys(*it,R);
      if (R.size()>0){
        dispExtra.resize(R.size());
        ufield->get(R,dispExtra);
      }
    }

    this->computeStrain(aips,iele,GP,ws,_domMinus,_domPlus,dispm,dispp,dispExtra,virt, useBarF);
  }
};

void interDomainBetween3D::computeIPVariable(AllIPState *aips,const unknownField *ufield,
                                                const IPStateBase::whichState ws, bool stiff)
{
  if (ws == IPStateBase::initial){
    this->initialIPVariable(aips,ufield,ws,stiff);
    return;
  }
 _evalStiff = false;
    dG3DMaterialLaw *mlawMinus = static_cast<dG3DMaterialLaw*>(this->getMaterialLawMinus());
    dG3DMaterialLaw *mlawPlus = static_cast<dG3DMaterialLaw*>(this->getMaterialLawPlus());
    dG3DMaterialLaw *mlawbulkm;
    dG3DMaterialLaw *mlawbulkp;
    if(mlawMinus->getType() == materialLaw::fracture)
      mlawbulkm = static_cast< dG3DMaterialLaw* >((static_cast< FractureByCohesive3DLaw* > (mlawMinus) )->getBulkLaw());
    else
      mlawbulkm= static_cast<dG3DMaterialLaw*>(mlawMinus);
    if(mlawPlus->getType() == materialLaw::fracture)
      mlawbulkp = static_cast< dG3DMaterialLaw* >((static_cast< FractureByCohesive3DLaw* > (mlawPlus) )->getBulkLaw());
    else
      mlawbulkp= static_cast<dG3DMaterialLaw*>(mlawPlus);
    if (mlawbulkm->getUseBarF()!=mlawbulkp->getUseBarF()) //check with the bulk law
      Msg::Error("dG3DDomain::computeIpv All the constitutive laws need to use Fbar or not");
    bool useBarF=mlawbulkm->getUseBarF();

  #if defined(HAVE_MPI)
  if ((this->getMaterialLawMinus()->isNumeric() or this->getMaterialLawPlus()->isNumeric()) and this->_otherRanks.size()>0){
    this->computeIPVariableMPI(aips,ufield,ws,stiff,useBarF);
  }
  else
  #endif //HAVE_MPI
  {

    this->computeAllIPStrain(aips,ufield,ws,false, useBarF);
    IntPt *GP;
    fullVector<double> dispm;
    fullVector<double> dispp;
    fullVector<double> dispExtra;
    std::vector<Dof> R;
    for(groupOfElements::elementContainer::iterator it=gi->begin(); it!=gi->end(); ++it){
      MInterfaceElement *iele = dynamic_cast<MInterfaceElement*>(*it);
      int npts_inter= integBound->getIntPoints(*it,&GP);
      R.clear();
      this->getFunctionSpaceMinus()->getKeys(iele->getElem(0),R);
      dispm.resize(R.size());
      ufield->get(R,dispm);
      R.clear();
      this->getFunctionSpacePlus()->getKeys(iele->getElem(1),R);
      dispp.resize(R.size());
      ufield->get(R,dispp);

      if (this->IsMultipleFieldFormulation()){
        FunctionSpaceBase* extraSp = this->getInterfaceFunctionSpace();
        R.clear();
        extraSp->getKeys(*it,R);
        if (R.size()>0){
          dispExtra.resize(R.size());
          ufield->get(R,dispExtra);
        }
      }

      this->computeIpv(aips,iele,GP,ws,_domMinus,_domPlus,mlawMinus,mlawPlus,dispm,dispp,dispExtra,false,stiff);
    }
  }
  _evalStiff = true;
}

void interDomainBetween3D::setMaterialLaw(const std::map<int,materialLaw*> &maplaw)
{
  // oterwise the law is already set
  if(!setmaterial){
    _domMinus->setMaterialLaw(maplaw);
    _domPlus->setMaterialLaw(maplaw);
    _mlaw = NULL;
    if(_lnum == 0){
      // no fracture law prescribed --> materialLawMinus and Plus are given via domain
      if (_bulkNum != 0){
        for(std::map<int,materialLaw*>::const_iterator it=maplaw.begin(); it!=maplaw.end(); ++it){
          if(it->first == _bulkNum){
            _mlawMinus = it->second;
            _mlawPlus = it->second;
            break;
          }
        }
        if (_mlawMinus == NULL){
          Msg::Error("The material law %d is not found",_bulkNum);
        }
      }
      else{
        if ((_domMinus->getMaterialLaw()->getType() == materialLaw::fracture) and _lawFromBulkPart){
          materialLaw2LawsInitializer *m2law = dynamic_cast<materialLaw2LawsInitializer*>(_domMinus->getMaterialLaw());
          if (m2law){
            int lawMinusNum = m2law->bulkLawNumber();
            for(std::map<int,materialLaw*>::const_iterator it=maplaw.begin(); it!=maplaw.end(); ++it){
              if (it->first == lawMinusNum){
                _mlawMinus = it->second;
                break;
              }
            }
            if (_mlawMinus == NULL){
              Msg::Error("The material law %d is not found",lawMinusNum);
            }
          }
          else{
            Msg::Error("fractureBy2Laws must be used interDomainBetween3D::setMaterialLaw");
          }
        }
        else{
          _mlawMinus = _domMinus->getMaterialLaw();
        }

        if ((_domPlus->getMaterialLaw()->getType() == materialLaw::fracture) and _lawFromBulkPart){
          materialLaw2LawsInitializer *m2law = dynamic_cast<materialLaw2LawsInitializer*>(_domPlus->getMaterialLaw());
          if (m2law){
            int lawPlusNum = m2law->bulkLawNumber();
            for(std::map<int,materialLaw*>::const_iterator it=maplaw.begin(); it!=maplaw.end(); ++it){
              if (it->first == lawPlusNum){
                _mlawPlus = it->second;
                break;
              }
            }
            if (_mlawPlus == NULL){
              Msg::Error("The material law %d is not found",lawPlusNum);
            }

          }
          else{
            Msg::Error("fractureBy2Laws must be used interDomainBetween3D::setMaterialLaw");
          }
        }
        else{
          _mlawPlus = _domPlus->getMaterialLaw();
        }
      }
    }
    else{
      // create law (use lnum with negative number same number for the two but it doesn't matter)
      // these law don't appear in map with all laws
      // check if minus and plus laws are already fracture by 2 laws;
      if (_bulkNum != 0){
        // if _bulkNum is set, this used for bulk law
        materialLaw* mlawInt = NULL;
        for(std::map<int,materialLaw*>::const_iterator it=maplaw.begin(); it!=maplaw.end(); ++it){
          if(it->first == _bulkNum){
            mlawInt = it->second;
            break;
          }
        }
        if (mlawInt == NULL){
          Msg::Error("The material law %d is not found",_bulkNum);
        }

        if (mlawInt->isNumeric()){
          _mlawMinus = new MultiscaleFractureByCohesive3DLaw(-_bulkNum,_bulkNum,_lnum);
        }
        else{
          _mlawMinus = new FractureByCohesive3DLaw(-_bulkNum,_bulkNum,_lnum);
        }

        _mlawPlus = _mlawMinus;
      }
      else{
        // if _bulkNum is not set, the material laws are constructed from positive and negative parts
        int numlawBulkMinus(0), numlawBulkPlus(0);
        if(_domMinus->getMaterialLaw()->getType() != materialLaw::fracture){
          numlawBulkMinus = _domMinus->getLawNum();
        }
        else{
          materialLaw2LawsInitializer *m2law = dynamic_cast<materialLaw2LawsInitializer*>(_domMinus->getMaterialLaw());
          numlawBulkMinus = m2law->bulkLawNumber();
        }
        if(_domPlus->getMaterialLaw()->getType() != materialLaw::fracture){
          numlawBulkPlus = _domPlus->getLawNum();
        }
        else{
          materialLaw2LawsInitializer *m2law = dynamic_cast<materialLaw2LawsInitializer*>(_domPlus->getMaterialLaw());
          numlawBulkPlus = m2law->bulkLawNumber();
        }

        if (_domMinus->getMaterialLaw()->isNumeric()){
          _mlawMinus = new MultiscaleFractureByCohesive3DLaw(-numlawBulkMinus,numlawBulkMinus,_lnum);
        }
        else
          _mlawMinus = new FractureByCohesive3DLaw(-numlawBulkMinus,numlawBulkMinus,_lnum);

        if (_domPlus->getMaterialLaw()->isNumeric()){
          _mlawPlus = new MultiscaleFractureByCohesive3DLaw(-numlawBulkPlus,numlawBulkPlus,_lnum);
        }
        else
          _mlawPlus = new FractureByCohesive3DLaw(-numlawBulkPlus,numlawBulkPlus,_lnum);
      }
    }
    _mlawMinus->initLaws(maplaw);
    _mlawPlus->initLaws(maplaw);
    setmaterial=true;
  }
}


void computeJump(const std::vector<TensorialTraits<double>::ValType> &Val_m,const int n_m,
                      const std::vector<TensorialTraits<double>::ValType> &Val_p,
                      const int n_p,const fullVector<double> & dispm,const fullVector<double> & dispp,SVector3 &ujump)
{
  for(int i=0;i<3;i++){
    ujump(i) =0.;
    for(int j=0;j<n_m;j++)
      ujump(i)-=Val_m[j]*dispm(j+i*n_m);
    for(int j=0;j<n_p;j++)
      ujump(i)+=Val_p[j]*dispp(j+i*n_p);
  }
}

void computeNonLocalJump(const std::vector<TensorialTraits<double>::ValType> &Val_m,const int n_m,
                      const std::vector<TensorialTraits<double>::ValType> &Val_p,
                      const int n_p,const fullVector<double> & dispm,const fullVector<double> & dispp, double &epsjump,
                      const int idex)
{
  epsjump =0.;
  for(int j=0;j<n_m;j++)
    epsjump-=Val_m[j]*dispm(j+3*n_m+idex*n_m);
  for(int j=0;j<n_p;j++)
    epsjump+=Val_p[j]*dispp(j+3*n_p+idex*n_p);
}

void computeExtraFieldJump(const int index_field,const std::vector<TensorialTraits<double>::ValType> &Val_m,const int n_m,
                      const std::vector<TensorialTraits<double>::ValType> &Val_p,
                      const int n_p,const fullVector<double> & dispm,const fullVector<double> & dispp, double &temperaturejump)
{
  temperaturejump =0.;
  for(int j=0;j<n_m;j++)
    temperaturejump-=Val_m[j]*dispm(j+index_field*n_m); //depending on index_field could be temperature jump or voltage jump
  for(int j=0;j<n_p;j++)
    temperaturejump+=Val_p[j]*dispp(j+index_field*n_p);
}


void computeftjump(const int index_field,const std::vector<TensorialTraits<double>::ValType> &Val_m,const int n_m,
                      const std::vector<TensorialTraits<double>::ValType> &Val_p,
                      const int n_p,const fullVector<double> & dispm,const fullVector<double> & dispp, double &ftjump, double &dftjumpdTp, double &dftjumpdTm)
{
  ftjump =0.;dftjumpdTp=0.;dftjumpdTm=0.;
  double tp =0.,tm =0.;
  for(int j=0;j<n_m;j++)
    tm+=Val_m[j]*dispm(j+index_field*n_m); //depending on index_field could be temperature jump or voltage jump
  for(int j=0;j<n_p;j++)
    tp+=Val_p[j]*dispp(j+index_field*n_p);
  if (tp>0. &&tm>0.)
  {
    ftjump=1./tp -1/tm;

     dftjumpdTp=-1./pow(tp,2);
     dftjumpdTm= 1./pow(tm,2);
  }
 else
  {  // Virtual interface element
     Msg::Error("No in computeftjump in Domain for 3D");

  }
}

void computefvjump(const int index_fieldT,const int index_fieldv,const std::vector<TensorialTraits<double>::ValType> &Val_m,const int n_m,
                  const std::vector<TensorialTraits<double>::ValType> &Val_p,const int n_p,const fullVector<double> & dispm,const fullVector<double> & dispp,
		  double &fvjump, double &dfvjumpdTp, double &dfvjumpdTm, double &dfvjumpdvp, double &dfvjumpdvm)
{
  double ftjump=0.;
  fvjump =0.;dfvjumpdTp=0.;dfvjumpdTm=0.;dfvjumpdvp=0.;dfvjumpdvm=0.;
  double tp =0.,tm =0.;
  for(int j=0;j<n_m;j++)
    tm+=Val_m[j]*dispm(j+index_fieldT*n_m); //depending on index_field could be temperature jump
  for(int j=0;j<n_p;j++)
    tp+=Val_p[j]*dispp(j+index_fieldT*n_p);



  double vp =0.,vm =0.;
  for(int j=0;j<n_m;j++)
    vm+=Val_m[j]*dispm(j+index_fieldv*n_m); //
  for(int j=0;j<n_p;j++)
    vp+=Val_p[j]*dispp(j+index_fieldv*n_p);


  if (tp>0.&&tm>0.)
  {
   ftjump=1./tp -1./tm;

   fvjump=vm/tm -vp/tp;

   dfvjumpdvp=-1./tp;
   dfvjumpdvm= 1./tm;

   dfvjumpdTp= vp/pow(tp,2);
   dfvjumpdTm=-vm/pow(tm,2);
  }
   else
  {  // Virtual interface element
     Msg::Error("No in computefvjump in Domain for 3D");

  }

}



// New thermomechanical

ThermoMechanicsDG3DDomain::ThermoMechanicsDG3DDomain (const int tag,const int phys, const int ws, const int lnum,
                                          const int fdg, const double ThermoMechanicsEqRatio, const int dim) :
                                          dG3DDomain(tag,phys,ws,lnum,fdg, dim,0,1)
{
  setConstitutiveExtraDofDiffusionEqRatio(ThermoMechanicsEqRatio);

}

ThermoMechanicsDG3DDomain::ThermoMechanicsDG3DDomain(const ThermoMechanicsDG3DDomain &source) : dG3DDomain(source)
{

}

ThermoMechanicsInterDomainBetween3D::ThermoMechanicsInterDomainBetween3D(const int tag, partDomain *dom1,partDomain *dom2,const double ThermoMechanicsEqRatio, const int lnum, const int bnum) :
                                 interDomainBetween3D(tag, dom1, dom2, lnum, bnum)
{
  setConstitutiveExtraDofDiffusionEqRatio(ThermoMechanicsEqRatio);
}


ThermoMechanicsInterDomainBetween3D::ThermoMechanicsInterDomainBetween3D(const ThermoMechanicsInterDomainBetween3D &source) :
				interDomainBetween3D(source)
{

}


ElecTherMechDG3DDomain::ElecTherMechDG3DDomain (const int tag,const int phys, const int ws, const int lnum,
                                          const int fdg, const double ElecTherMechEqRatio, const int dim) : dG3DDomain(tag,phys,ws,lnum,fdg,dim,0,2)
{
/*  if(_space!=NULL) delete _space;   // A space is created in dG3DDomain remove it!
  switch(_wsp)
  {
   case functionSpaceType::Lagrange:
    if(_fullDg)
      _space = new dG3DLagrangeFunctionSpace(getTag(),3+getNumNonLocalVariable()+2);
    else
      _space = new g3DLagrangeFunctionSpace(getTag(),3+getNumNonLocalVariable()+2);
    break;
   case functionSpaceType::Inter :
    break; // InterfaceDomain set function space later
   default:
    Msg::Error("Function space type is unknown on domain %d",_phys);
  }*/
  setConstitutiveExtraDofDiffusionEqRatio(ElecTherMechEqRatio);
/*  _continuity=true;
  _ElecTherMechBeta=100.;*/
}

ElecTherMechDG3DDomain::ElecTherMechDG3DDomain(const ElecTherMechDG3DDomain &source) : dG3DDomain(source)
{
  /*   _ElecTherMechBeta = source._ElecTherMechBeta;
     _continuity = source._continuity;
     _ElecTherMechEqRatio = source._ElecTherMechEqRatio;*/
}


/* void ElecTherMechDG3DDomain::computeStrain(MElement *e, const int npts_bulk, IntPt* GP,
                                           AllIPState::ipstateElementContainer *vips,
                                           IPStateBase::whichState ws, fullVector<double> &disp) const
{
  if (!getElementErosionFilter()(e)) return;

  dG3DDomain::computeStrain(e, npts_bulk,GP,vips, ws,disp);
  int nbFF = e->getNumShapeFunctions();
  for(int j=0;j<npts_bulk;j++){

    IPStateBase* ips = (*vips)[j];
    ElecTherMechDG3DIPVariableBase *ipv = static_cast<ElecTherMechDG3DIPVariableBase*>(ips->getState(ws));

    const std::vector<TensorialTraits<double>::ValType> &Vals = ipv->f(_space,e,GP[j]);
    const std::vector<TensorialTraits<double>::GradType> &Grads = ipv->gradf(_space,e,GP[j]);

    if(getUseFtFv())
    {
      double *fT = &(ipv->getRefTofT());
      SVector3 *gradfT = &(ipv->getRefToGradfT());
      *fT= 0.;
      STensorOperation::zero(*gradfT);

      for (int i = 0; i < nbFF; i++)
      {
        int indexFieldfT=3+getNumNonLocalVariable();
        const double& ftemp = disp(i+indexFieldfT*nbFF);
        *fT += Vals[i+indexFieldfT*nbFF]*ftemp;
        gradfT->operator()(0) += Grads[i+indexFieldfT*nbFF][0]*ftemp;
        gradfT->operator()(1) += Grads[i+indexFieldfT*nbFF][1]*ftemp;
        gradfT->operator()(2) += Grads[i+indexFieldfT*nbFF][2]*ftemp;
      }


      double *fV = &(ipv->getRefTofV());
      SVector3 *gradfV = &(ipv->getRefToGradfV());
      *fV= 0.;
      STensorOperation::zero(*gradfV);
      for (int i = 0; i < nbFF; i++)
      {
        int indexFieldfV=3+getNumNonLocalVariable()+1;
        const double& fvolt = disp(i+indexFieldfV*nbFF);
        *fV += Vals[i+indexFieldfV*nbFF]*fvolt;
        gradfV->operator()(0) += Grads[i+indexFieldfV*nbFF][0]*fvolt;
        gradfV->operator()(1) += Grads[i+indexFieldfV*nbFF][1]*fvolt;
        gradfV->operator()(2) += Grads[i+indexFieldfV*nbFF][2]*fvolt;
      }
      SVector3 *gradT = &(ipv->getRefToGradT());
      SVector3 *gradV = &(ipv->getRefToGradV());

      double *T = &(ipv->getRefToTemperature());
      double *V = &(ipv->getRefToVoltage());
      *V = -1*(*fV)/(*fT);
      *T = 1/(*fT);


      gradV->operator()(0)= -1/(*fT)*(gradfV->operator()(0))+(*fV)/(pow(*fT,2))*(gradfT->operator()(0));
      gradV->operator()(1)= -1/(*fT)*(gradfV->operator()(1))+(*fV)/(pow(*fT,2))*(gradfT->operator()(1));
      gradV->operator()(2)= -1/(*fT)*(gradfV->operator()(2))+(*fV)/(pow(*fT,2))*(gradfT->operator()(2));


      gradT->operator()(0)=-1/(pow(*fT,2))*(gradfT->operator()(0));
      gradT->operator()(1)=-1/(pow(*fT,2))*(gradfT->operator()(1));
      gradT->operator()(2)=-1/(pow(*fT,2))*(gradfT->operator()(2));

      STensor3 *dgardVdgradfV=&(ipv->getRefTodgradVdgradfV());
      STensor3 *dgardVdgradfT=&(ipv->getRefTodgradVdgradfT());
      STensor3 *dgardTdgradfT=&(ipv->getRefTodgradTdgradfT());
      STensor3 *dgardTdgradfV=&(ipv->getRefTodgradTdgradfV());


      SVector3 *dgradVdfV=&(ipv->getRefTodgradVdfV());
      SVector3 *dgradVdfT=&(ipv->getRefTodgradVdfT());
      SVector3 *dgradTdfT=&(ipv->getRefTodgradTdfT());
      SVector3 *dgradTdfV=&(ipv->getRefTodgradTdfV());

      double *dVdfV=&(ipv->getRefTodVdfV());
      double *dVdfT=&(ipv->getRefTodVdfT());
      double *dTdfT=&(ipv->getRefTodTdfT());
      double *dTdfV=&(ipv->getRefTodTdfV());


      STensorOperation::zero(*dgardVdgradfV);
      dgardVdgradfV->operator()(0,0) =-1./(*fT);
      dgardVdgradfV->operator()(1,1) =-1./(*fT);
      dgardVdgradfV->operator()(2,2) =-1./(*fT);

      STensorOperation::zero(*dgardVdgradfT);
      dgardVdgradfT->operator()(0,0) =(*fV)/(*fT)/(*fT);
      dgardVdgradfT->operator()(1,1) =(*fV)/(*fT)/(*fT);
      dgardVdgradfT->operator()(2,2) =(*fV)/(*fT)/(*fT);

      STensorOperation::zero(*dgardTdgradfT);
      dgardTdgradfT->operator()(0,0) =-1./(*fT)/(*fT);
      dgardTdgradfT->operator()(1,1) =-1./(*fT)/(*fT);
      dgardTdgradfT->operator()(2,2) =-1./(*fT)/(*fT);

      STensorOperation::zero(*dgardTdgradfV);
      *dVdfV=-1./(*fT);
      *dVdfT=(*fV)/(*fT)/(*fT);
      *dTdfT=-1./(*fT)/(*fT);
      *dTdfV=0.;

      dgradVdfV->operator()(0)=1./(*fT)/(*fT)*(gradfT->operator()(0));
      dgradVdfV->operator()(1)=1./(*fT)/(*fT)*(gradfT->operator()(1));
      dgradVdfV->operator()(2)=1./(*fT)/(*fT)*(gradfT->operator()(2));

      dgradVdfT->operator()(0)=1./(*fT)/(*fT)*(gradfV->operator()(0))-2*(*fV)/(*fT)/(*fT)/(*fT)*(gradfT->operator()(0));
      dgradVdfT->operator()(1)=1./(*fT)/(*fT)*(gradfV->operator()(1))-2*(*fV)/(*fT)/(*fT)/(*fT)*(gradfT->operator()(1));
      dgradVdfT->operator()(2)=1./(*fT)/(*fT)*(gradfV->operator()(2))-2*(*fV)/(*fT)/(*fT)/(*fT)*(gradfT->operator()(2));

      dgradTdfT->operator()(0)=2/(*fT)/(*fT)/(*fT)*(gradfT->operator()(0));
      dgradTdfT->operator()(1)=2/(*fT)/(*fT)/(*fT)*(gradfT->operator()(1));
      dgradTdfT->operator()(2)=2/(*fT)/(*fT)/(*fT)*(gradfT->operator()(2));

      dgradTdfV->operator()(0)=0.;
      dgradTdfV->operator()(1)=0.;
      dgradTdfV->operator()(2)=0.;
    }
    else
    {
      double *temperature = &(ipv->getRefToTemperature());
      SVector3 *gradT = &(ipv->getRefToGradT());
      *temperature= 0.;
      STensorOperation::zero(*gradT);

      for (int i = 0; i < nbFF; i++)
      {
        int indexFieldT=3+getNumNonLocalVariable();
        const double& temp = disp(i+indexFieldT*nbFF);
        *temperature += Vals[i+indexFieldT*nbFF]*temp;
        gradT->operator()(0) += Grads[i+indexFieldT*nbFF][0]*temp;
        gradT->operator()(1) += Grads[i+indexFieldT*nbFF][1]*temp;
        gradT->operator()(2) += Grads[i+indexFieldT*nbFF][2]*temp;
      }


      double *Voltage = &(ipv->getRefToVoltage());
      SVector3 *gradV = &(ipv->getRefToGradV());
      *Voltage= 0.;
      STensorOperation::zero(*gradV);

      for (int i = 0; i < nbFF; i++)
      {
        int indexFieldV=3+getNumNonLocalVariable()+1;
        const double& volt = disp(i+indexFieldV*nbFF);
        *Voltage += Vals[i+indexFieldV*nbFF]*volt;
        gradV->operator()(0) += Grads[i+indexFieldV*nbFF][0]*volt;
        gradV->operator()(1) += Grads[i+indexFieldV*nbFF][1]*volt;
        gradV->operator()(2) += Grads[i+indexFieldV*nbFF][2]*volt;
      }
    }
  }
};*/


/*void ElecTherMechDG3DDomain::computeStrain(AllIPState *aips,MInterfaceElement *ie, IntPt *GP,const IPStateBase::whichState ws,
                                        partDomain* efMinus, partDomain *efPlus, fullVector<double> &dispm,
                                        fullVector<double> &dispp, fullVector<double>& dispExtra, const bool virt)
{
  MElement *ele = dynamic_cast<MElement*>(ie);
  if (!getElementErosionFilter()(ele)) return;

  if(!virt){
    dG3DDomain::computeStrain(aips,ie, GP,ws,efMinus,efPlus, dispm,dispp,dispExtra,virt);
    int npts=integBound->getIntPoints(ele,&GP);
    // Get Gauss points values on minus and plus elements
    IntPt *GPm; IntPt *GPp;
    _interQuad->getIntPoints(ie,GP,&GPm,&GPp);

    FunctionSpaceBase* _spaceminus = this->getFunctionSpaceMinus();
    FunctionSpaceBase* _spaceplus = this->getFunctionSpacePlus();


    MElement *em = ie->getElem(0);
    MElement *ep = ie->getElem(1);

    int nbFFm = em->getNumVertices();
    int nbFFp = ep->getNumVertices();

    AllIPState::ipstateElementContainer *vips = aips->getIPstate(ie->getNum());
    for(int j=0;j<npts;j++){
      IPStateBase* ipsm = (*vips)[j];
      IPStateBase* ipsp = (*vips)[j+npts];

      ElecTherMechDG3DIPVariableBase* ipvm = NULL;
      ElecTherMechDG3DIPVariableBase* ipvp = NULL;
      // check if fracture or not via minus if true for minus --> true for plus too
      FractureCohesive3DIPVariable *ipvtmp = dynamic_cast<FractureCohesive3DIPVariable*>(ipsm->getState(ws));
      if(ipvtmp!=NULL)
      {
        ipvm = static_cast<ElecTherMechDG3DIPVariableBase*>(ipvtmp->getIPvBulk());
        ipvtmp = static_cast<FractureCohesive3DIPVariable*>(ipsp->getState(ws)); // here sure cast OK
        ipvp = static_cast<ElecTherMechDG3DIPVariableBase*>(ipvtmp->getIPvBulk());
      }
      else
      {
        ipvm = static_cast<ElecTherMechDG3DIPVariableBase*>(ipsm->getState(ws));
        ipvp = static_cast<ElecTherMechDG3DIPVariableBase*>(ipsp->getState(ws));
      }

      const std::vector<TensorialTraits<double>::ValType> &Valsm = ipvm->f(_spaceminus,em,GPm[j]);
      const std::vector<TensorialTraits<double>::ValType> &Valsp = ipvp->f(_spaceplus,ep,GPp[j]);

      const std::vector<TensorialTraits<double>::GradType> &Gradsm = ipvm->gradf(_spaceminus,em,GPm[j]);
      const std::vector<TensorialTraits<double>::GradType> &Gradsp = ipvp->gradf(_spaceplus,ep,GPp[j]);

      if(getUseFtFv())
      {
        double *fTm = &(ipvm->getRefTofT());
        *fTm = 0.;

        SVector3 *gradfTm = &(ipvm->getRefToGradfT());
        STensorOperation::zero(*gradfTm);

         int indexFieldfT=3+getNumNonLocalVariable();
         for (int i = 0; i < nbFFm; i++)
         {
            const double& ftemp = dispm(i+indexFieldfT*nbFFm);
            *fTm += Valsm[i+indexFieldfT*nbFFm]*ftemp;
            gradfTm->operator()(0) += Gradsm[i+indexFieldfT*nbFFm][0]*ftemp;
            gradfTm->operator()(1) += Gradsm[i+indexFieldfT*nbFFm][1]*ftemp;
            gradfTm->operator()(2) += Gradsm[i+indexFieldfT*nbFFm][2]*ftemp;
          }

        double *fTp = &(ipvp->getRefTofT());
        *fTp = 0.;
        SVector3 *gradfTp = &(ipvp->getRefToGradfT());
        STensorOperation::zero(*gradfTp);

        for (int i = 0; i < nbFFp; i++)
        {
          const double& ftemp = dispp(i+indexFieldfT*nbFFp);
          *fTp += Valsp[i+indexFieldfT*nbFFp]*ftemp;
          gradfTp->operator()(0) += Gradsp[i+indexFieldfT*nbFFp][0]*ftemp;
          gradfTp->operator()(1) += Gradsp[i+indexFieldfT*nbFFp][1]*ftemp;
          gradfTp->operator()(2) += Gradsp[i+indexFieldfT*nbFFp][2]*ftemp;
        }

        double& fTjump=ipvm->getRefTofTJump();

        // compute the jump
        computeExtraFieldJump(indexFieldfT,Valsm,nbFFm,Valsp,nbFFp,dispm,dispp,fTjump);
        //Msg::Error("Temperature jump: %e",tempjump);
        ipvp->getRefToOneOverTemperatureJump() = fTjump;


        double *fVm = &(ipvm->getRefTofV());
        *fVm = 0.;

        SVector3 *gradfVm = &(ipvm->getRefToGradfV());
        STensorOperation::zero(*gradfVm);
        int indexFieldfV=3+getNumNonLocalVariable()+1;

        for (int i = 0; i < nbFFm; i++)
        {
          const double& fvolt = dispm(i+indexFieldfV*nbFFm);
          *fVm += Valsm[i+indexFieldfV*nbFFm]*fvolt;
          gradfVm->operator()(0) += Gradsm[i+indexFieldfV*nbFFm][0]*fvolt;
          gradfVm->operator()(1) += Gradsm[i+indexFieldfV*nbFFm][1]*fvolt;
          gradfVm->operator()(2) += Gradsm[i+indexFieldfV*nbFFm][2]*fvolt;
        }

        double *fVp = &(ipvp->getRefTofV());
        *fVp = 0.;
        SVector3 *gradfVp = &(ipvp->getRefToGradfV());
        STensorOperation::zero(*gradfVp);

        for (int i = 0; i < nbFFp; i++)
        {
          const double& fvolt = dispp(i+indexFieldfV*nbFFp);
          *fVp += Valsp[i+indexFieldfV*nbFFp]*fvolt;
          gradfVp->operator()(0) += Gradsp[i+indexFieldfV*nbFFp][0]*fvolt;
          gradfVp->operator()(1) += Gradsp[i+indexFieldfV*nbFFp][1]*fvolt;
          gradfVp->operator()(2) += Gradsp[i+indexFieldfV*nbFFp][2]*fvolt;
        }

        double *Tm = &(ipvm->getRefToTemperature());
        double *Vm = &(ipvm->getRefToVoltage());
        *Vm = -1*(*fVm)/(*fTm);
        *Tm = 1/(*fTm);

        double *Tp = &(ipvp->getRefToTemperature());
        double *Vp = &(ipvp->getRefToVoltage());
        *Vp = -1*(*fVp)/(*fTp);
        *Tp = 1/(*fTp);

        double& fVjump= ipvm->getRefTofvJump();
        // compute the jump
        computeExtraFieldJump(indexFieldfV,Valsm,nbFFm,Valsp,nbFFp,dispm,dispp,fVjump);
        ipvp->getRefTofvJump() = fVjump;

        SVector3 &gradTm = ipvm->getRefToGradT();
        SVector3 &gradVm = ipvm->getRefToGradV();
        SVector3 &gradTp = ipvp->getRefToGradT();
        SVector3 &gradVp = ipvp->getRefToGradV();

        gradVp(0)= -1/(*fTp)*(gradfVp->operator()(0))+(*fVp)/(pow(*fTp,2))*gradfTp->operator()(0);
        gradVp(1)= -1/(*fTp)*(gradfVp->operator()(1))+(*fVp)/(pow(*fTp,2))*gradfTp->operator()(1);
        gradVp(2)= -1/(*fTp)*(gradfVp->operator()(2))+(*fVp)/(pow(*fTp,2))*gradfTp->operator()(2);

        gradVm(0)= -1/(*fTm)*(gradfVm->operator()(0))+(*fVm)/(pow(*fTm,2))*gradfTm->operator()(0);
        gradVm(1)= -1/(*fTm)*(gradfVm->operator()(1))+(*fVm)/(pow(*fTm,2))*gradfTm->operator()(1);
        gradVm(2)= -1/(*fTm)*(gradfVm->operator()(2))+(*fVm)/(pow(*fTm,2))*gradfTm->operator()(2);

        gradTp(0)=-1/(pow(*fTp,2))*gradfTp->operator()(0);
        gradTp(1)=-1/(pow(*fTp,2))*gradfTp->operator()(1);
        gradTp(2)=-1/(pow(*fTp,2))*gradfTp->operator()(2);

        gradTm(0)=-1/(pow(*fTm,2))*gradfTm->operator()(0);
        gradTm(1)=-1/(pow(*fTm,2))*gradfTm->operator()(1);
        gradTm(2)=-1/(pow(*fTm,2))*gradfTm->operator()(2);

        STensor3 &dgardVdgradfVm=ipvm->getRefTodgradVdgradfV();
        STensor3 &dgardVdgradfTm=ipvm->getRefTodgradVdgradfT();
        STensor3 &dgardTdgradfTm=ipvm->getRefTodgradTdgradfT();
        STensor3 &dgardTdgradfVm=ipvm->getRefTodgradTdgradfV();

        SVector3 &dgradVdfVm=ipvm->getRefTodgradVdfV();
        SVector3 &dgradVdfTm=ipvm->getRefTodgradVdfT();
        SVector3 &dgradTdfTm=ipvm->getRefTodgradTdfT();
        SVector3 &dgradTdfVm=ipvm->getRefTodgradTdfV();

        double &dVdfVm=ipvm->getRefTodVdfV();
        double &dVdfTm=ipvm->getRefTodVdfT();
        double &dTdfTm=ipvm->getRefTodTdfT();
        double &dTdfVm=ipvm->getRefTodTdfV();

        STensor3 &dgardVdgradfVp=ipvp->getRefTodgradVdgradfV();
        STensor3 &dgardVdgradfTp=ipvp->getRefTodgradVdgradfT();
        STensor3 &dgardTdgradfTp=ipvp->getRefTodgradTdgradfT();
        STensor3 &dgardTdgradfVp=ipvp->getRefTodgradTdgradfV();

        SVector3 &dgradVdfVp=ipvp->getRefTodgradVdfV();
        SVector3 &dgradVdfTp=ipvp->getRefTodgradVdfT();
        SVector3 &dgradTdfTp=ipvp->getRefTodgradTdfT();
        SVector3 &dgradTdfVp=ipvp->getRefTodgradTdfV();

        double &dVdfVp=ipvp->getRefTodVdfV();
        double &dVdfTp=ipvp->getRefTodVdfT();
        double &dTdfTp=ipvp->getRefTodTdfT();
        double &dTdfVp=ipvp->getRefTodTdfV();


        STensorOperation::diag(dgardVdgradfVm,-1./(*fTm));
        STensorOperation::diag(dgardVdgradfTm,*fVm/(*fTm)/(*fTm));
        STensorOperation::diag(dgardTdgradfTm,-1./(*fTm)/(*fTm));
        STensorOperation::zero(dgardTdgradfVm);

        STensorOperation::diag(dgardVdgradfVp,-1./(*fTp));
        STensorOperation::diag(dgardVdgradfTp,*fVp/(*fTp)/(*fTp));
        STensorOperation::diag(dgardTdgradfTp,-1./(*fTp)/(*fTp));
        STensorOperation::zero(dgardTdgradfVp);

        dVdfVm=-1./(*fTm);
        dVdfTm=(*fVm)/(*fTm)/(*fTm);
        dTdfTm=-1./(*fTm)/(*fTm);
        dTdfVm=0.;

        dVdfVp=-1./(*fTp);
        dVdfTp=(*fVp)/(*fTp)/(*fTp);
        dTdfTp=-1./(*fTp)/(*fTp);
        dTdfVp=0.;

        dgradVdfVm(0)=1./(*fTm)/(*fTm)*(gradfTm->operator()(0));
        dgradVdfVm(1)=1./(*fTm)/(*fTm)*(gradfTm->operator()(1));
        dgradVdfVm(2)=1./(*fTm)/(*fTm)*(gradfTm->operator()(2));

        dgradVdfVp(0)=1./(*fTp)/(*fTp)*(gradfTp->operator()(0));
        dgradVdfVp(1)=1./(*fTp)/(*fTp)*(gradfTp->operator()(1));
        dgradVdfVp(2)=1./(*fTp)/(*fTp)*(gradfTp->operator()(2));

        dgradVdfTm(0)=1./(*fTm)/(*fTm)*(gradfVm->operator()(0))-2*(*fVm)/(*fTm)/(*fTm)/(*fTm)*(gradfTm->operator()(0));
        dgradVdfTm(1)=1./(*fTm)/(*fTm)*(gradfVm->operator()(1))-2*(*fVm)/(*fTm)/(*fTm)/(*fTm)*(gradfTm->operator()(1));
        dgradVdfTm(2)=1./(*fTm)/(*fTm)*(gradfVm->operator()(2))-2*(*fVm)/(*fTm)/(*fTm)/(*fTm)*(gradfTm->operator()(2));

        dgradVdfTp(0)=1./(*fTp)/(*fTp)*(gradfVp->operator()(0))-2*(*fVp)/(*fTp)/(*fTp)/(*fTp)*(gradfTp->operator()(0));
        dgradVdfTp(1)=1./(*fTp)/(*fTp)*(gradfVp->operator()(1))-2*(*fVp)/(*fTp)/(*fTp)/(*fTp)*(gradfTp->operator()(1));
        dgradVdfTp(2)=1./(*fTp)/(*fTp)*(gradfVp->operator()(2))-2*(*fVp)/(*fTp)/(*fTp)/(*fTp)*(gradfTp->operator()(2));

        dgradTdfTm(0)=2/(*fTm)/(*fTm)/(*fTm)*(gradfTm->operator()(0));
        dgradTdfTm(1)=2/(*fTm)/(*fTm)/(*fTm)*(gradfTm->operator()(1));
        dgradTdfTm(2)=2/(*fTm)/(*fTm)/(*fTm)*(gradfTm->operator()(2));

        dgradTdfTp(0)=2/(*fTp)/(*fTp)/(*fTp)*(gradfTp->operator()(0));
        dgradTdfTp(1)=2/(*fTp)/(*fTp)/(*fTp)*(gradfTp->operator()(1));
        dgradTdfTp(2)=2/(*fTp)/(*fTp)/(*fTp)*(gradfTp->operator()(2));

        dgradTdfVm(0)=0.;
        dgradTdfVm(1)=0.;
        dgradTdfVm(2)=0.;

        dgradTdfVp(0)=0.;
        dgradTdfVp(1)=0.;
        dgradTdfVp(2)=0.;
      }
      else
      {
        int indexFieldT=3+getNumNonLocalVariable();
        double *temperaturem = &(ipvm->getRefToTemperature());
        *temperaturem = 0.;

        SVector3 *gradTm = &(ipvm->getRefToGradT());
        STensorOperation::zero(*gradTm);

        for (int i = 0; i < nbFFm; i++)
        {
          const double& temp = dispm(i+indexFieldT*nbFFm);
          *temperaturem += Valsm[i+indexFieldT*nbFFm]*temp;
          gradTm->operator()(0) += Gradsm[i+indexFieldT*nbFFm][0]*temp;
          gradTm->operator()(1) += Gradsm[i+indexFieldT*nbFFm][1]*temp;
          gradTm->operator()(2) += Gradsm[i+indexFieldT*nbFFm][2]*temp;
        }

        double *temperaturep = &(ipvp->getRefToTemperature());
        *temperaturep = 0.;
        SVector3 *gradTp = &(ipvp->getRefToGradT());
        STensorOperation::zero(*gradTp);

        for (int i = 0; i < nbFFp; i++)
        {
          const double& temp = dispp(i+indexFieldT*nbFFp);
          *temperaturep += Valsp[i+indexFieldT*nbFFp]*temp;
          gradTp->operator()(0) += Gradsp[i+indexFieldT*nbFFp][0]*temp;
          gradTp->operator()(1) += Gradsp[i+indexFieldT*nbFFp][1]*temp;
          gradTp->operator()(2) += Gradsp[i+indexFieldT*nbFFp][2]*temp;
        }

        double& tempjump= ipvm->getRefToTemperatureJump();
        // compute the jump
        computeExtraFieldJump(indexFieldT,Valsm,nbFFm,Valsp,nbFFp,dispm,dispp,tempjump);
        //Msg::Error("Temperature jump: %e",tempjump);
        ipvp->getRefToTemperatureJump() = tempjump;


        double *voltagem = &(ipvm->getRefToVoltage());
        *voltagem = 0.;
        SVector3 *gradVm = &(ipvm->getRefToGradV());
        STensorOperation::zero(*gradVm);
        int indexFieldV=3+getNumNonLocalVariable()+1;

        for (int i = 0; i < nbFFm; i++)
        {
          const double& volt = dispm(i+indexFieldV*nbFFm);
          *voltagem += Valsm[i+indexFieldV*nbFFm]*volt;
          gradVm->operator()(0) += Gradsm[i+indexFieldV*nbFFm][0]*volt;
          gradVm->operator()(1) += Gradsm[i+indexFieldV*nbFFm][1]*volt;
          gradVm->operator()(2) += Gradsm[i+indexFieldV*nbFFm][2]*volt;
        }
        double *voltagep = &(ipvp->getRefToVoltage());
        *voltagep = 0.;
        SVector3 *gradVp = &(ipvp->getRefToGradV());
        STensorOperation::zero(*gradVp);

        for (int i = 0; i < nbFFp; i++)
        {
          const double& volt = dispp(i+indexFieldV*nbFFp);
          *voltagep += Valsp[i+indexFieldV*nbFFp]*volt;
          gradVp->operator()(0) += Gradsp[i+indexFieldV*nbFFp][0]*volt;
          gradVp->operator()(1) += Gradsp[i+indexFieldV*nbFFp][1]*volt;
          gradVp->operator()(2) += Gradsp[i+indexFieldV*nbFFp][2]*volt;
        }

        double& voltjump= ipvm->getRefToVoltageJump();
        // compute the jump
        computeExtraFieldJump(indexFieldV,Valsm,nbFFm,Valsp,nbFFp,dispm,dispp,voltjump);
        ipvp->getRefToVoltageJump() = voltjump;


        double fvjump=0.;double ftjump=0.;
        double dftjumpdTp=0.;double dftjumpdTm=0.;
        double dfvjumpdvp, dfvjumpdvm,dfvjumpdTm,dfvjumpdTp;
        dfvjumpdvp=0.; dfvjumpdvm=0.;dfvjumpdTm=0.;dfvjumpdTp=0.;

        computeftjump(indexFieldT,Valsm,nbFFm,Valsp,nbFFp,dispm,dispp,ftjump,dftjumpdTp,dftjumpdTm);
        computefvjump(indexFieldT,indexFieldV,Valsm,nbFFm,Valsp,nbFFp,dispm,dispp,fvjump,dfvjumpdTp,dfvjumpdTm,dfvjumpdvp,dfvjumpdvm);

        ipvm->getRefTofvJump() = fvjump;
        ipvp->getRefTofvJump() = fvjump;

        ipvm->getRefToOneOverTemperatureJump() = ftjump;
        ipvp->getRefToOneOverTemperatureJump() = ftjump;

        ipvm->getRefTodfvjumpdvm() = dfvjumpdvm;
        ipvp->getRefTodfvjumpdvp() = dfvjumpdvp;

        ipvm->getRefTodfvjumpdTm() = dfvjumpdTm;
        ipvp->getRefTodfvjumpdTp() = dfvjumpdTp;

        ipvm->getRefTodOneOverTemperatureJumpdTm() = dftjumpdTm;
        ipvp->getRefTodOneOverTemperatureJumpdTp() = dftjumpdTp;
      }
    }
  }
  else
  {  // Virtual interface element
     Msg::Error("No virtual interface for 3D");
  }
}*/

/*void ElecTherMechDG3DDomain::createTerms(unknownField *uf,IPField*ip)
{
  FunctionSpace<double>* dgspace = static_cast<FunctionSpace<double>*>(_space);
  ltermBulk = new ElecTherMechDG3DForceBulk(*dgspace,_mlaw,_fullDg,ip,_incrementNonlocalBased,getElecTherMechEqRatio(), getUseFtFv());
  if(_fullDg)
    ltermBound = new ElecTherMechDG3DForceInter(*dgspace,dgspace,_mlaw,_mlaw,_interQuad,_beta1,_fullDg,ip,_incrementNonlocalBased,getElecTherMechStabilityParameter(),
                  getElecTherMechContinuity(), getElecTherMechEqRatio(), getUseFtFv());
  else
     ltermBound = new nonLinearTermVoid();
  if(_bmbp)
    btermBulk = new BilinearTermPerturbation<double>(ltermBulk,*dgspace,*dgspace,uf,ip,this,_eps);
  else
   btermBulk = new ElecTherMechDG3DStiffnessBulk(*dgspace,_mlaw,_fullDg,ip,getElecTherMechEqRatio(), getUseFtFv());
  if(_fullDg)
  {
    if(_interByPert)
      btermBound = new BilinearTermPerturbation<double>(ltermBound,*dgspace,*dgspace,uf,ip,this,_eps);
    else
      btermBound = new ElecTherMechDG3DStiffnessInter(*dgspace,*dgspace,_mlaw,_mlaw,_interQuad,_beta1,ip,uf,_fullDg, _incrementNonlocalBased,
                        getElecTherMechStabilityParameter() ,getElecTherMechContinuity(),getElecTherMechEqRatio(), getUseFtFv());
  }
  else
    btermBound = new BiNonLinearTermVoid();

  this->btermVirtBound = new BiNonLinearTermVoid();
  this->ltermVirtBound = new nonLinearTermVoid();


  massterm = new ElecTherMechMassG3D(*dgspace,_mlaw);

}*/

/*LinearTermBase<double>* ElecTherMechDG3DDomain::createNeumannTerm(FunctionSpace<double> *spneu,  const groupOfElements* g,
                                                                  const simpleFunctionTime<double>* f,const unknownField *uf,
                                                                  const IPField * ip,nonLinearBoundaryCondition::location onWhat) const
{
  //  return  new ElecTherMechG3DLoadTerm<double>(*spneu,f);

}*/

/* void ElecTherMechDG3DDomain::allocateInterfaceMPI(const int rankOtherPart, const groupOfElements &groupOtherPart,dgPartDomain**  dgdom)
{
 // domain creation has it function create a domain with a part of the same domain contained on an other part,
  // the type of domain is known
  if(*dgdom==NULL){ // domain doesn't exist --> creation
    *dgdom = new ElecTherMechInterDomainBetween3D(this->getTag(),this,this,this->getElecTherMechEqRatio(),0); // TO CHANGE HERE
    // set the stability parameters
    ElecTherMechInterDomainBetween3D* newInterDom3D = static_cast< ElecTherMechInterDomainBetween3D* >(*dgdom);
    newInterDom3D->stabilityParameters(this->stabilityParameter(1));
    newInterDom3D->ElecTherMechStabilityParameters(this->getElecTherMechStabilityParameter(),this->getElecTherMechContinuity());
    newInterDom3D->useFtFv(this->getUseFtFv());
  }
}*/


ElecTherMechInterDomainBetween3D::ElecTherMechInterDomainBetween3D(const int tag, partDomain *dom1,partDomain *dom2,const double ElecTherMechEqRatio, const int lnum, const int bnum) :
                                  interDomainBase(dom1,dom2,bnum),
                                  ElecTherMechDG3DDomain(tag,manageInterface::getKey(dom1->getPhysical(),
                                                                    dom2->getPhysical()),10000,lnum,true,ElecTherMechEqRatio,dom1->getDim())
{
/*
  // create empty groupOfElements
  g = new groupOfElements();
  // set functionalSpace
  if((_domMinus->getFunctionSpaceType() == functionSpaceType::Lagrange) and
     (_domPlus->getFunctionSpaceType() == functionSpaceType::Lagrange)){
    if(_space != NULL) delete _space;
    _space = new dG3DLagrangeBetween2DomainsFunctionSpace(5,_spaceMinus,_spacePlus);
    _wsp = functionSpaceType::Lagrange;
  }
  else{
    Msg::Error("Impossible to create functional space on ElecTherMechInterDomain3D %d",_phys);
  }
  _continuity=(static_cast<ElecTherMechDG3DDomain *> (dom1))->getElecTherMechContinuity();
  _ElecTherMechBeta=(static_cast<ElecTherMechDG3DDomain *> (dom1))->getElecTherMechStabilityParameter();*/
  setConstitutiveExtraDofDiffusionEqRatio(ElecTherMechEqRatio);
}


ElecTherMechInterDomainBetween3D::ElecTherMechInterDomainBetween3D(const ElecTherMechInterDomainBetween3D &source) :
interDomainBase(source), ElecTherMechDG3DDomain(source)
{

}

/*void ElecTherMechInterDomainBetween3D::initializeTerms(unknownField *uf,IPField*ip)
{

  createTerms(uf,ip);
  initializeFractureBase(uf,ip, gi, integBound);
}*/

/*void ElecTherMechInterDomainBetween3D::setGaussIntegrationRule()
{
  setGaussIntegrationRuleBase(gi, &integBound, &integBulk,&_interQuad, _gaussorderbound, _gqt);
}*/

/*void ElecTherMechInterDomainBetween3D::computeAllIPStrain(AllIPState *aips,const unknownField *ufield,const IPStateBase::whichState ws, const bool virt){
  IntPt *GP;
  fullVector<double> dispm;
  fullVector<double> dispp;
  fullVector<double> dispExtra;
  std::vector<Dof> R;
  for(groupOfElements::elementContainer::iterator it=gi->begin(); it!=gi->end(); ++it){
    MInterfaceElement *iele = dynamic_cast<MInterfaceElement*>(*it);
    R.clear();
    this->getFunctionSpaceMinus()->getKeys(iele->getElem(0),R);
    dispm.resize(R.size());
    ufield->get(R,dispm);
    R.clear();
    this->getFunctionSpacePlus()->getKeys(iele->getElem(1),R);
    dispp.resize(R.size());
    ufield->get(R,dispp);

    if (this->IsMultipleFieldFormulation()){
      FunctionSpaceBase* extraSp = this->getInterfaceFunctionSpace();
      R.clear();
      extraSp->getKeys(*it,R);
      if (R.size()>0){
        dispExtra.resize(R.size());
        ufield->get(R,dispExtra);
      }
    }

    this->computeStrain(aips,iele,GP,ws,_domMinus,_domPlus,dispm,dispp,dispExtra,virt);
  }
};*/


/*void ElecTherMechInterDomainBetween3D::computeIPVariable(AllIPState *aips,const unknownField *ufield,
                                                const IPStateBase::whichState ws, bool stiff)
{

 if (ws == IPStateBase::initial){
    this->initialIPVariable(aips,ufield,ws,stiff);
    return;
  }


  _evalStiff = false;
  #if defined(HAVE_MPI)
  if ((_mlawMinus->isNumeric() or _mlawPlus->isNumeric()) and this->_otherRanks.size()>0){
    this->computeIPVariableMPI(aips,ufield,ws,stiff);
  }
  else
  #endif //HAVE_MPI
  {
    this->computeAllIPStrain(aips,ufield,ws,false);
    IntPt *GP;
    fullVector<double> dispm;
    fullVector<double> dispp;
    fullVector<double> dispExtra;
    std::vector<Dof> R;
    dG3DMaterialLaw *mlawMinus = static_cast<dG3DMaterialLaw*>(_mlawMinus);
    dG3DMaterialLaw *mlawPlus = static_cast<dG3DMaterialLaw*>(_mlawPlus);
    for(groupOfElements::elementContainer::iterator it=gi->begin(); it!=gi->end(); ++it){
      MInterfaceElement *iele = dynamic_cast<MInterfaceElement*>(*it);
      int npts_inter= integBound->getIntPoints(*it,&GP);
      R.clear();
      this->getFunctionSpaceMinus()->getKeys(iele->getElem(0),R);
      dispm.resize(R.size());
      ufield->get(R,dispm);
      R.clear();
      this->getFunctionSpacePlus()->getKeys(iele->getElem(1),R);
      dispp.resize(R.size());
      ufield->get(R,dispp);

      if (this->IsMultipleFieldFormulation()){
        FunctionSpaceBase* extraSp = this->getInterfaceFunctionSpace();
        R.clear();
        extraSp->getKeys(*it,R);
        if (R.size()>0){
          dispExtra.resize(R.size());
          ufield->get(R,dispExtra);
        }
      }

      this->computeIpv(aips,iele,GP,ws,_domMinus,_domPlus,mlawMinus,mlawPlus,dispm,dispp,dispExtra,false,stiff);
    }
  }
  _evalStiff = true;
}*/

/*void ElecTherMechInterDomainBetween3D::setMaterialLaw(const std::map<int,materialLaw*> &maplaw)
{
  // oterwise the law is already set
  if(!setmaterial){
    _domMinus->setMaterialLaw(maplaw);
    _domPlus->setMaterialLaw(maplaw);
    _mlaw = NULL;
    if(_lnum == 0){ // no fracture law prescribed --> materialLawMinus and Plus are given via domain
      _mlawMinus = _domMinus->getMaterialLaw();
      _mlawPlus = _domPlus->getMaterialLaw();
      if(_domMinus->getPhysical() == _domPlus->getPhysical())
      {
         _mlaw = _mlawMinus; // same law on both
      }
      else if(_mlawMinus->getNum() == _mlawPlus->getNum())
      {
        _mlaw = _mlawMinus; // same law on both
      }
      else
      {
       //
	Msg::Warning("Cannot set the interface law on interDomain3D");
      }
    }
    else{
      // create law (use lnum with negative number same number for the two but it doesn't matter)
      // these law don't appear in map with all laws
      // check if minus and plus laws are already fracture by 2 laws;
      int numlawBulkMinus, numlawBulkPlus;
      if(_domMinus->getMaterialLaw()->getType() != materialLaw::fracture){
        numlawBulkMinus = _domMinus->getLawNum();
      }
      else{
        materialLaw2LawsInitializer *m2law = dynamic_cast<materialLaw2LawsInitializer*>(_domMinus->getMaterialLaw());
        numlawBulkMinus = m2law->bulkLawNumber();
      }
      if(_domPlus->getMaterialLaw()->getType() != materialLaw::fracture){
        numlawBulkPlus = _domPlus->getLawNum();
      }
      else{
        materialLaw2LawsInitializer *m2law = dynamic_cast<materialLaw2LawsInitializer*>(_domPlus->getMaterialLaw());
        numlawBulkPlus = m2law->bulkLawNumber();
      }
      _mlawMinus = new FractureByCohesive3DLaw(-numlawBulkMinus,numlawBulkMinus,_lnum);
      _mlawPlus = new FractureByCohesive3DLaw(-numlawBulkPlus,numlawBulkPlus,_lnum);
    }
    _mlawMinus->initLaws(maplaw);
    _mlawPlus->initLaws(maplaw);
    setmaterial=true;
  }
}*/
/*
void ElecTherMechInterDomainBetween3D::createTerms(unknownField *uf,IPField*ip)
{
  // void term (as they have to exist)
  this->ltermBulk = new nonLinearTermVoid();
  this->btermBulk = new BiNonLinearTermVoid();

  this->btermVirtBound = new BiNonLinearTermVoid();
  this->ltermVirtBound = new nonLinearTermVoid();
  this->massterm = new BiNonLinearTermVoid();

  FunctionSpace<double>* dgspace = static_cast<FunctionSpace<double>*>(_space);
  FunctionSpace<double>* dgspace1 = static_cast<FunctionSpace<double>*>(_spaceMinus);
  FunctionSpace<double>* dgspace2 = static_cast<FunctionSpace<double>*>(_spacePlus);
  dG3DDomain* dgldomMinus = static_cast<dG3DDomain*>(_domMinus);
  dG3DDomain* dgldomPlus = static_cast<dG3DDomain*>(_domPlus);
  ltermBound = new ElecTherMechDG3DForceInter(*dgspace1,dgspace2,_mlawMinus,_mlawPlus,_interQuad,_beta1,true,ip, _incrementNonlocalBased,
                    getElecTherMechStabilityParameter() ,getElecTherMechContinuity(),getElecTherMechEqRatio(), getUseFtFv());
  if(_interByPert)
    btermBound = new BilinearTermPerturbation<double>(ltermBound,*dgspace1,*dgspace2,uf,ip,this,_eps);
  else
    btermBound = new ElecTherMechDG3DStiffnessInter(*dgspace1,*dgspace2,_mlawMinus,_mlawPlus,_interQuad,_beta1,ip,uf,true, _incrementNonlocalBased,
                           getElecTherMechStabilityParameter() ,getElecTherMechContinuity(),getElecTherMechEqRatio(), getUseFtFv());
}*/
