//
// C++ Interface: terms
//
// Description: Class of terms for dg non linear shell
//
//
// Author:  <Gauthier BECKER>, (C) 2011
//
// Copyright: See COPYING file that comes with this distribution
//
//
#include "dG3DTerms.h"
#include "highOrderTensor.h"
#include "FractureCohesiveDG3DIPVariable.h"
#include "nonLocalDamageDG3DIPVariable.h"
#include "FractureCohesiveDG3DMaterialLaw.h"

void g3DLoadTerm::get(MElement *ele,int npts,IntPt *GP,fullVector<double> &m) const
{
  const int nbdof=this->space1.getNumKeys(ele);
  m.resize(nbdof);
  m.setAll(0.);
  // Get shape functions values and gradients at all Gauss points
  nlsFunctionSpaceXYZ<double>* sp1 = static_cast<nlsFunctionSpaceXYZ<double>*>(&(this->space1));
  std::vector<std::vector<TensorialTraits<double>::ValType> >vvals;
  std::vector<std::vector<TensorialTraits<double>::GradType> >vgradsuvw;
  sp1->get(ele,npts,GP,vvals);
  sp1->getgradfuvw(ele,npts,GP,vgradsuvw);
  double jac[3][3];
  SPoint3 p;
  double scaleEq = 1.;
  if (comp < 3){
    scaleEq = 1.; // disp
  }
  else if (comp > 2 and comp < 3+getNumNonLocalVariable()){
    scaleEq = getNonLocalEqRatio(); // nonlocal Dof
  }
  else if (comp > 2+getNumNonLocalVariable() and comp < 3+getNumNonLocalVariable()+getNumConstitutiveExtraDofDiffusionVariable()){
    scaleEq = getConstitutiveExtraDofDiffusionEqRatio(); // extraDof
  }
  else{
    Msg::Error("comp = %d has not been wrongly introduced in g3DLoadTerm::get",comp);
  }
  for (int i = 0; i < npts; i++)
  {
    const double weight = GP[i].weight;
    std::vector<double> &Vals = vvals[i];
    // not optimal as evertything is computed (especially if hessian and vthird are true but it should be false for Neumann space) but better than before...
    const double detJ = ele->getJacobian(vgradsuvw[i],jac);
    ele->pnt(Vals,p);
    double load=Load->operator()(p.x(),p.y(),p.z());
    double loadvalue = load*weight*detJ;
    for (int j = 0; j < nbdof ; ++j){
      m(j)+=Vals[j]*loadvalue*scaleEq;
    }
  }
}

void g3DFiniteStrainsPressureTerm::get(MElement *ele,int npts,IntPt *GP,fullVector<double> &m) const
{
  // resize the force vector
  const int nbdof=this->space1.getNumKeys(ele);
  const int nbFF = ele->getNumShapeFunctions();
  m.resize(nbdof);
  m.setAll(0.);

  // get the displacements of the element
  std::vector<Dof> keys;
  this->space1.getKeys(ele,keys);
  fullVector<double> disp(keys.size());
  _ufield->get(keys,disp);
  
  // Get shape functions values and gradients at all Gauss points
  nlsFunctionSpaceXYZ<double>* sp1 = static_cast<nlsFunctionSpaceXYZ<double>*>(&(this->space1));
  std::vector<std::vector<TensorialTraits<double>::ValType> >vvals;
  std::vector<std::vector<TensorialTraits<double>::GradType> >vgrads;
  std::vector<std::vector<TensorialTraits<double>::GradType> >vgradsuvw;
  sp1->get(ele,npts,GP,vvals);
  sp1->getgradf(ele,npts,GP,vgrads);
  sp1->getgradfuvw(ele,npts,GP,vgradsuvw);
  
  SPoint3 p;

  for (int i = 0; i < npts; i++)
  {
    const double weight = GP[i].weight;
    std::vector<TensorialTraits<double>::ValType> &Vals = vvals[i];
    std::vector<TensorialTraits<double>::GradType>& Grads = vgrads[i];

    static STensor3 defo_grad;
    STensorOperation::unity(defo_grad);
    for (int j = 0; j < nbFF; j++)
    {
      //x y z: take compoenent x arbitrarily
      // xx yy zz xy xz yz
      defo_grad(0,0) += Grads[j+0*nbFF][0]*disp(j);
      defo_grad(0,1) += Grads[j+0*nbFF][1]*disp(j);
      defo_grad(0,2) += Grads[j+0*nbFF][2]*disp(j);
      defo_grad(1,0) += Grads[j+nbFF][0]*disp(j+nbFF);
      defo_grad(1,1) += Grads[j+nbFF][1]*disp(j+nbFF);
      defo_grad(1,2) += Grads[j+nbFF][2]*disp(j+nbFF);
      defo_grad(2,0) += Grads[j+2*nbFF][0]*disp(j+2*nbFF);
      defo_grad(2,1) += Grads[j+2*nbFF][1]*disp(j+2*nbFF);
      defo_grad(2,2) += Grads[j+2*nbFF][2]*disp(j+2*nbFF);
    }
    if(STensorOperation::determinantSTensor3(defo_grad) < 1.e-15) Msg::Error("Negative Jacobian g3DFiniteStrainsPressureTerm::get");
    

    // defo gradient and jac
    double jdefo = STensorOperation::determinantSTensor3(defo_grad);
    static STensor3 invF;
    STensorOperation::inverseSTensor3(defo_grad,invF);
    
    std::vector<TensorialTraits<double>::GradType> &graduv = vgradsuvw[i];
    SVector3 ref_normal(0.,0.,0.);
    if (dom->getDim() == 3){
      // initial normal de l'élément (compute and store it once at the beginning)
      SVector3 phi1(0.), phi2(0.);
      for(int j=0;j<ele->getNumVertices();++j)
      {
          phi1[0] += graduv[j](0) * ele->getVertex(j)->x(); phi2[0] += graduv[j](1) * ele->getVertex(j)->x();
          phi1[1] += graduv[j](0) * ele->getVertex(j)->y(); phi2[1] += graduv[j](1) * ele->getVertex(j)->y();
          phi1[2] += graduv[j](0) * ele->getVertex(j)->z(); phi2[2] += graduv[j](1) * ele->getVertex(j)->z();
      }
      ref_normal = crossprod(phi1,phi2);        
    }
    else if (dom->getDim() == 2){
      // initial normal de l'élément (compute and store it once at the beginning)
      SVector3 phi1(0.);
      for(int j=0;j<ele->getNumVertices();++j)
      {
          phi1[0] += graduv[j](0) * ele->getVertex(j)->x(); 
          phi1[1] += graduv[j](0) * ele->getVertex(j)->y(); 
          phi1[2] += graduv[j](0) * ele->getVertex(j)->z(); 
      }
      ref_normal = crossprod(phi1,normalToDomain);
    }
    
    double detJ = ref_normal.normalize(); //  norm of ref_normal equal to detJ (Gauss integration) and  ref_normal become unit vector

    // current normal using Nanson formula
    static SVector3 cur_normal;
    STensorOperation::multSVector3STensor3(ref_normal,invF,cur_normal); // cur_normal  is not unit vector

    // value of the force at the given Gauss points. 
    ele->pnt(Vals,p);
    double load=Load->operator()(p.x(),p.y(),p.z());

    double loadvalue = -load*weight*detJ; /// minus as cur_normal == outward normal
    for (int j = 0; j < nbFF ; j++){
      for (int k=0; k< 3; k++){
        m(j+k*nbFF)+=Vals[j+k*nbFF]*loadvalue*jdefo*cur_normal[k];
      }
    }
  }
  //m.print("pressure term");
}

void g3DFiniteStrainsPressureBilinearTerm::get(MElement *ele,int npts,IntPt *GP,fullMatrix<double> &m) const
{
  // resize the force vector
  const int nbdof=this->space1.getNumKeys(ele);
  const int nbFF = ele->getNumShapeFunctions();
  m.resize(nbdof,nbdof);
  m.setAll(0.);

  // get the displacements of the element
  std::vector<Dof> keys;
  this->space1.getKeys(ele,keys);
  fullVector<double> disp(keys.size());
  _ufield->get(keys,disp);
  
  // Get shape functions values and gradients at all Gauss points
  nlsFunctionSpaceXYZ<double>* sp1 = static_cast<nlsFunctionSpaceXYZ<double>*>(&(this->space1));
  std::vector<std::vector<TensorialTraits<double>::ValType> >vvals;
  std::vector<std::vector<TensorialTraits<double>::GradType> >vgrads;
  std::vector<std::vector<TensorialTraits<double>::GradType> >vgradsuvw;
  sp1->get(ele,npts,GP,vvals);
  sp1->getgradf(ele,npts,GP,vgrads);
  sp1->getgradfuvw(ele,npts,GP,vgradsuvw);
  
  SPoint3 p;

  for (int i = 0; i < npts; i++)
  {
    const double weight = GP[i].weight;
    std::vector<TensorialTraits<double>::ValType> &Vals = vvals[i];
    std::vector<TensorialTraits<double>::GradType>& Grads = vgrads[i];

    static STensor3 defo_grad;
    STensorOperation::unity(defo_grad);
    for (int j = 0; j < nbFF; j++)
    {
      //x y z: take compoenent x arbitrarily
      // xx yy zz xy xz yz
      defo_grad(0,0) += Grads[j+0*nbFF][0]*disp(j);
      defo_grad(0,1) += Grads[j+0*nbFF][1]*disp(j);
      defo_grad(0,2) += Grads[j+0*nbFF][2]*disp(j);
      defo_grad(1,0) += Grads[j+nbFF][0]*disp(j+nbFF);
      defo_grad(1,1) += Grads[j+nbFF][1]*disp(j+nbFF);
      defo_grad(1,2) += Grads[j+nbFF][2]*disp(j+nbFF);
      defo_grad(2,0) += Grads[j+2*nbFF][0]*disp(j+2*nbFF);
      defo_grad(2,1) += Grads[j+2*nbFF][1]*disp(j+2*nbFF);
      defo_grad(2,2) += Grads[j+2*nbFF][2]*disp(j+2*nbFF);
    }
    if(STensorOperation::determinantSTensor3(defo_grad) < 1.e-15) Msg::Error("Negative Jacobian g3DFiniteStrainsPressureBilinearTerm::");
    

    // defo gradient and jac
    double jdefo = STensorOperation::determinantSTensor3(defo_grad);
    static STensor3 invF;
    STensorOperation::inverseSTensor3(defo_grad,invF);

    SVector3 ref_normal(0.,0.,0.);
    std::vector<TensorialTraits<double>::GradType> &graduv = vgradsuvw[i];
    if (dom->getDim() == 3){
      // initial normal de l'élément (compute and store it once at the beginning)
      SVector3 phi1(0.), phi2(0.);
      for(int j=0;j<ele->getNumVertices();++j)
      {
          phi1[0] += graduv[j](0) * ele->getVertex(j)->x(); phi2[0] += graduv[j](1) * ele->getVertex(j)->x();
          phi1[1] += graduv[j](0) * ele->getVertex(j)->y(); phi2[1] += graduv[j](1) * ele->getVertex(j)->y();
          phi1[2] += graduv[j](0) * ele->getVertex(j)->z(); phi2[2] += graduv[j](1) * ele->getVertex(j)->z();
      }
      ref_normal = crossprod(phi1,phi2);        
    }
    else if (dom->getDim() == 2){
      // initial normal de l'élément (compute and store it once at the beginning)
      SVector3 phi1(0.);
      for(int j=0;j<ele->getNumVertices();++j)
      {
          phi1[0] += graduv[j](0) * ele->getVertex(j)->x(); 
          phi1[1] += graduv[j](0) * ele->getVertex(j)->y(); 
          phi1[2] += graduv[j](0) * ele->getVertex(j)->z(); 
      }
      ref_normal = crossprod(phi1,normalToDomain);
    }
    const double detJ = ref_normal.normalize();  //  norm of ref_normal equal to detJ (Gauss integration) and  ref_normal become unit vector
    
    // define cur_normal = ref_normal*invF
    // current normal using Nanson formula
    // Jcur_normal = jdefo*ref_normal*invF
    static STensor33 DJcur_normalDF;
    STensorOperation::zero(DJcur_normalDF);
    for (int k=0; k<3; k++){
      for (int l=0; l<3; l++){
        for (int r=0; r<3; r++){
          for (int s=0; s<3; s++){
            DJcur_normalDF(l,r,s) +=  jdefo*invF(s,r)*ref_normal(k)*invF(k,l) -jdefo*0.5*ref_normal(k)*(invF(k,r)*invF(l,s)+ invF(k,s)*invF(l,r));
          }
        }
      }
    }

    // value of the force at the given Gauss points.
    ele->pnt(Vals,p);
    double load=Load->operator()(p.x(),p.y(),p.z());

    double loadvalue = -load*weight*detJ; /// minus as cur_normal == outward normal
    for (int j = 0; j < nbFF; j++){
      for (int k=0; k<3; k++){
        for (int p=0; p< nbFF; p++){
          for (int q=0; q<3; q++){
            for (int r=0; r<3; r++){
              m(j+k*nbFF,p+q*nbFF) -=Vals[j+k*nbFF]*loadvalue*DJcur_normalDF(k,q,r)*Grads[p+q*nbFF](r); // negative as it is the derivative of Fext
            }
          }
        }
      }
    }
  }
}

void g3DFiniteStrainsScalarFluxLinearTerm::get(MElement *ele,int npts,IntPt *GP,fullVector<double> &m) const
{
  // resize the force vector
  const int nbdof=this->space1.getNumKeys(ele);
  const int nbFF = ele->getNumShapeFunctions();
  m.resize(nbdof);
  m.setAll(0.);

  // get the displacements of the element
  std::vector<Dof> keys;
  this->space1.getKeys(ele,keys);
  fullVector<double> disp(keys.size());
  _ufield->get(keys,disp);

  // Get shape functions values and gradients at all Gauss points
  nlsFunctionSpaceXYZ<double>* sp1 = static_cast<nlsFunctionSpaceXYZ<double>*>(&(this->space1));
  std::vector<std::vector<TensorialTraits<double>::ValType> >vvals;
  std::vector<std::vector<TensorialTraits<double>::GradType> >vgrads;
  std::vector<std::vector<TensorialTraits<double>::GradType> >vgradsuvw;
  sp1->get(ele,npts,GP,vvals);
  sp1->getgradf(ele,npts,GP,vgrads);
  sp1->getgradfuvw(ele,npts,GP,vgradsuvw);
  
  
  SPoint3 p;

  for (int i = 0; i < npts; i++)
  {
    const double weight = GP[i].weight;
    std::vector<TensorialTraits<double>::ValType> &Vals = vvals[i];
    std::vector<TensorialTraits<double>::GradType>& Grads = vgrads[i];

    static STensor3 defo_grad;
    STensorOperation::unity(defo_grad);
    for (int j = 0; j < nbFF; j++)
    {
      //x y z: take compoenent x arbitrarily
      // xx yy zz xy xz yz
      defo_grad(0,0) += Grads[j+0*nbFF][0]*disp(j);
      defo_grad(0,1) += Grads[j+0*nbFF][1]*disp(j);
      defo_grad(0,2) += Grads[j+0*nbFF][2]*disp(j);
      defo_grad(1,0) += Grads[j+nbFF][0]*disp(j+nbFF);
      defo_grad(1,1) += Grads[j+nbFF][1]*disp(j+nbFF);
      defo_grad(1,2) += Grads[j+nbFF][2]*disp(j+nbFF);
      defo_grad(2,0) += Grads[j+2*nbFF][0]*disp(j+2*nbFF);
      defo_grad(2,1) += Grads[j+2*nbFF][1]*disp(j+2*nbFF);
      defo_grad(2,2) += Grads[j+2*nbFF][2]*disp(j+2*nbFF);
    }
    if(STensorOperation::determinantSTensor3(defo_grad) < 1.e-15) Msg::Error("Negative Jacobian g3DFiniteStrainsScalarFluxLinearTerm::get");
    
    // defo gradient and jac
    double jdefo = STensorOperation::determinantSTensor3(defo_grad);
    static STensor3 invF;
    STensorOperation::inverseSTensor3(defo_grad,invF);
    
    SVector3 ref_normal(0.,0.,0.);
    std::vector<TensorialTraits<double>::GradType> &graduv = vgradsuvw[i];
    if (_dom->getDim() == 3){
      // initial normal de l'élément (compute and store it once at the beginning)
      SVector3 phi1(0.), phi2(0.);
      for(int j=0;j<ele->getNumVertices();++j)
      {
          phi1[0] += graduv[j](0) * ele->getVertex(j)->x(); phi2[0] += graduv[j](1) * ele->getVertex(j)->x();
          phi1[1] += graduv[j](0) * ele->getVertex(j)->y(); phi2[1] += graduv[j](1) * ele->getVertex(j)->y();
          phi1[2] += graduv[j](0) * ele->getVertex(j)->z(); phi2[2] += graduv[j](1) * ele->getVertex(j)->z();
      }
      ref_normal = crossprod(phi1,phi2);        
    }
    else if (_dom->getDim() == 2){
      // initial normal de l'élément (compute and store it once at the beginning)
      SVector3 phi1(0.);
      for(int j=0;j<ele->getNumVertices();++j)
      {
          phi1[0] += graduv[j](0) * ele->getVertex(j)->x(); 
          phi1[1] += graduv[j](0) * ele->getVertex(j)->y(); 
          phi1[2] += graduv[j](0) * ele->getVertex(j)->z(); 
      }
      ref_normal = crossprod(phi1,normalToDomain);
    }
    const double detJ = ref_normal.normalize();  //  norm of ref_normal equal to detJ (Gauss integration) and  ref_normal become unit vector
    
    
    // current normal using Nanson formula
    static SVector3 cur_normal;
    STensorOperation::multSVector3STensor3(ref_normal,invF,cur_normal); // cur_normal  is not unit vector
    double normJcur_normal = jdefo*cur_normal.norm();
    
    // value of the force at the given Gauss points
    ele->pnt(Vals,p);
    // loop on the different components
    for(size_t k=0;k<vload.size();++k)
    {
      double load=vload[k]->operator()(p.x(),p.y(),p.z());
      double loadvalue = - load*weight*detJ;
      double depenValue = 1.;
      // get field value
      if (vload[k]->withUnknownDependency()){
        const std::vector<int>& depenComps = vload[k]->getDependenceComponents();
        const scalarFunction* depenFunc = vload[k]->getDependenceFunction();
        std::vector<double> fieldValue(depenComps.size());
        for (int ic = 0; ic < depenComps.size(); ic++){
          fieldValue[ic] = 0.;
          for (int jc = 0; jc < nbFF; jc++){
            fieldValue[ic] += Vals[jc+depenComps[ic]*nbFF]*disp(jc+depenComps[ic]*nbFF); 
          }
        }
        depenValue = depenFunc->getVal(fieldValue); 
        //Msg::Info("depenValue = %e fieldValue = %e",depenValue,fieldValue[0]);
      }
      double scaleEq = 1.;
      if (_vcomp[k] < 3){
        scaleEq = 1.; // disp
      }
      else if (_vcomp[k] > 2 and _vcomp[k] < 3+getNumNonLocalVariable()){
        scaleEq = getNonLocalEqRatio(); // nonlocal Dof
      }
      else if (_vcomp[k] > 2+getNumNonLocalVariable() and _vcomp[k] < 3+getNumNonLocalVariable()+getNumConstitutiveExtraDofDiffusionVariable()){
        scaleEq = getConstitutiveExtraDofDiffusionEqRatio(); // extraDof
      }
      else{
        Msg::Error("comp = %d has not been wrongly introduced in g3DFiniteStrainsScalarFluxLinearTerm::get",_vcomp[k]);
      }
      for (int j = 0; j < nbFF ; ++j){
        m(j+_vcomp[k]*nbFF)+=Vals[j]*loadvalue*depenValue*normJcur_normal*scaleEq;
      }
    }
  }
};

void g3DFiniteStrainsScalarFluxBiLinearTerm::get(MElement *ele, int npts, IntPt *GP, fullMatrix<double> &m) const
{
  // resize the force vector
  const int nbdof=this->space2.getNumKeys(ele);
  const int nbFF = ele->getNumShapeFunctions();
  m.resize(nbdof,nbdof);
  m.setAll(0.);

  // get the displacements of the element
  std::vector<Dof> keys;
  this->space2.getKeys(ele,keys);
  fullVector<double> disp(keys.size());
  _ufield->get(keys,disp);


  // Get shape functions values and gradients at all Gauss points
  nlsFunctionSpaceXYZ<double>* sp1 = static_cast<nlsFunctionSpaceXYZ<double>*>(&(this->space1));
  std::vector<std::vector<TensorialTraits<double>::ValType> >vvals;
  std::vector<std::vector<TensorialTraits<double>::GradType> >vgradsuvw;
  std::vector<std::vector<TensorialTraits<double>::GradType> >vgradsxyz;
  sp1->get(ele,npts,GP,vvals);
  sp1->getgradfuvw(ele,npts,GP,vgradsuvw);
  sp1->getgradf(ele,npts,GP,vgradsxyz);
  
  SPoint3 p;

  for (int i = 0; i < npts; i++)
  {
    const double weight = GP[i].weight;
    std::vector<TensorialTraits<double>::ValType> &Vals = vvals[i];
    std::vector<TensorialTraits<double>::GradType>& Grads = vgradsxyz[i];

    static STensor3 defo_grad;
    STensorOperation::unity(defo_grad);
    for (int j = 0; j < nbFF; j++)
    {
      //x y z: take compoenent x arbitrarily
      // xx yy zz xy xz yz
      defo_grad(0,0) += Grads[j+0*nbFF][0]*disp(j);
      defo_grad(0,1) += Grads[j+0*nbFF][1]*disp(j);
      defo_grad(0,2) += Grads[j+0*nbFF][2]*disp(j);
      defo_grad(1,0) += Grads[j+nbFF][0]*disp(j+nbFF);
      defo_grad(1,1) += Grads[j+nbFF][1]*disp(j+nbFF);
      defo_grad(1,2) += Grads[j+nbFF][2]*disp(j+nbFF);
      defo_grad(2,0) += Grads[j+2*nbFF][0]*disp(j+2*nbFF);
      defo_grad(2,1) += Grads[j+2*nbFF][1]*disp(j+2*nbFF);
      defo_grad(2,2) += Grads[j+2*nbFF][2]*disp(j+2*nbFF);
    }
    if(STensorOperation::determinantSTensor3(defo_grad) < 1.e-15) Msg::Error("Negative Jacobian");
    
    // defo gradient and jac
    double jdefo = STensorOperation::determinantSTensor3(defo_grad);
    static STensor3 invF;
    STensorOperation::inverseSTensor3(defo_grad,invF);
    
    SVector3 ref_normal(0.,0.,0.);
    std::vector<TensorialTraits<double>::GradType> &graduv = vgradsuvw[i];
    if (_dom->getDim() == 3){
      // initial normal de l'élément (compute and store it once at the beginning)
      SVector3 phi1(0.), phi2(0.);
      for(int j=0;j<ele->getNumVertices();++j)
      {
          phi1[0] += graduv[j](0) * ele->getVertex(j)->x(); phi2[0] += graduv[j](1) * ele->getVertex(j)->x();
          phi1[1] += graduv[j](0) * ele->getVertex(j)->y(); phi2[1] += graduv[j](1) * ele->getVertex(j)->y();
          phi1[2] += graduv[j](0) * ele->getVertex(j)->z(); phi2[2] += graduv[j](1) * ele->getVertex(j)->z();
      }
      ref_normal = crossprod(phi1,phi2);        
    }
    else if (_dom->getDim() == 2){
      // initial normal de l'élément (compute and store it once at the beginning)
      SVector3 phi1(0.);
      for(int j=0;j<ele->getNumVertices();++j)
      {
          phi1[0] += graduv[j](0) * ele->getVertex(j)->x(); 
          phi1[1] += graduv[j](0) * ele->getVertex(j)->y(); 
          phi1[2] += graduv[j](0) * ele->getVertex(j)->z(); 
      }
      ref_normal = crossprod(phi1,normalToDomain);
    }
    const double detJ = ref_normal.normalize();  //  norm of ref_normal equal to detJ (Gauss integration) and  ref_normal become unit vector
    
    // current normal using Nanson formula
    static SVector3 cur_normal;
    STensorOperation::multSVector3STensor3(ref_normal,invF,cur_normal); // cur_normal  is not unit vector
    double normJcur_normal = jdefo*cur_normal.norm();
    
    // define cur_normal = ref_normal*invF
    // current normal using Nanson formula
    // Jcur_normal = jdefo*ref_normal*invF
    static STensor33 DJcur_normalDF;
    STensorOperation::zero(DJcur_normalDF);
    for (int k=0; k<3; k++){
      for (int l=0; l<3; l++){
        for (int r=0; r<3; r++){
          for (int s=0; s<3; s++){
            DJcur_normalDF(l,r,s) +=  jdefo*invF(s,r)*ref_normal(k)*invF(k,l) -jdefo*0.5*ref_normal(k)*(invF(k,r)*invF(l,s)+ invF(k,s)*invF(l,r));
          }
        }
      }
    }
    static STensor3 DnormJcur_normalDF;
    STensorOperation::zero(DnormJcur_normalDF);
    for (int k=0; k<3; k++){
      for (int l=0; l<3; l++){
        for (int r=0; r<3; r++){
          DnormJcur_normalDF(k,l) += jdefo*cur_normal(r)*DJcur_normalDF(r,k,l)/normJcur_normal;
        }
      }
    }

    // value of the force at the given Gauss points. Change this by a evaluation only at the barycenter?
    ele->pnt(Vals,p);
    for(size_t k=0;k<vload.size();++k)
    {
      double loadvalue= - vload[k]->operator()(p.x(),p.y(),p.z())*weight*detJ;
      double depenValue = 1.;
      std::vector<double> DdepenValue;
      std::vector<int> depenComps;
      const scalarFunction* depenFunc = NULL;
      // get field value
      if (vload[k]->withUnknownDependency()){
        depenComps = vload[k]->getDependenceComponents();
        depenFunc =  vload[k]->getDependenceFunction();
        std::vector<double> fieldValue(depenComps.size());
        for (int ic = 0; ic < depenComps.size(); ic++){
          fieldValue[ic] = 0.;
          for (int jc = 0; jc < nbFF; jc++){
            fieldValue[ic] += Vals[jc+depenComps[ic]*nbFF]*disp(jc+depenComps[ic]*nbFF); 
          }
        }
        depenValue *= depenFunc->getVal(fieldValue); 
        depenFunc->getDiff(fieldValue,DdepenValue);
      }
      
      double scaleEq = 1.;
      if (_vcomp[k] < 3){
        scaleEq = 1.; // disp
      }
      else if (_vcomp[k] > 2 and _vcomp[k] < 3+getNumNonLocalVariable()){
        scaleEq = getNonLocalEqRatio(); // nonlocal Dof
      }
      else if (_vcomp[k] > 2+getNumNonLocalVariable() and _vcomp[k] < 3+getNumNonLocalVariable()+getNumConstitutiveExtraDofDiffusionVariable()){
        scaleEq = getConstitutiveExtraDofDiffusionEqRatio(); // extraDof
      }
      else{
        Msg::Error("comp = %d has not been wrongly introduced in g3DFiniteStrainsScalarFluxLinearTerm::get",_vcomp[k]);
      }
      
      // loop on the different components
      for (int j = 0; j < nbFF ; ++j){
        for (int p=0; p< nbFF; p++){
          for (int q=0; q<3; q++){
            for (int r=0; r<3; r++){
              m(j+_vcomp[k]*nbFF,p+q*nbFF) -=Vals[j]*loadvalue*depenValue*DnormJcur_normalDF(q,r)*Grads[p](r)*scaleEq; // negative as it is the derivative of Fext
            }
          }
          
          if (vload[k]->withUnknownDependency()){
            for (int ic = 0; ic < depenComps.size(); ic++){
              m(j+_vcomp[k]*nbFF,p+depenComps[ic]*nbFF) -= Vals[j]* loadvalue*normJcur_normal*DdepenValue[ic]*Vals[p]*scaleEq; // negative as it is the derivative of Fext
            }
          }
        }
        
      }
    }
    
  }
}

void dG3DForceBulk::get(MElement *ele,int npts,IntPt *GP,fullVector<double> &vFor) const
{

  // Initialization of some data
  int nbdof = this->space1.getNumKeys(ele);
  int nbFF = ele->getNumShapeFunctions();
  vFor.resize(nbdof);
  vFor.setAll(0.);
  static SVector3 BT;

  // get gauss' point data
  _ipf->getIPv(ele,vipv,vipvprev);

  /*double averagePressure = 0.;
  if((ele->getTypeForMSH() == MSH_HEX_8 or ele->getTypeForMSH() == MSH_QUA_4 or 
      ele->getTypeForMSH() == MSH_TET_4 or ele->getTypeForMSH() == MSH_TRI_3 or ele->getTypeForMSH() == MSH_PRI_6) and !_fullDg) //not correct in large deformations
  {
    for (int i = 0; i < npts; i++)
    {
      const STensor3 *PK1 = &(vipv[i]->getConstRefToFirstPiolaKirchhoffStress());
      double pres = (PK1->operator()(0,0)+PK1->operator()(1,1)+PK1->operator()(2,2))/3.;
      averagePressure += pres/npts;
    }
  }*/
  for (int i = 0; i < npts; i++)
  {
    double weight = GP[i].weight;
    double & detJ = vipv[i]->getJacobianDeterminant(ele,GP[i]);

    const STensor3 *PK1 = &(vipv[i]->getConstRefToFirstPiolaKirchhoffStress());

    double ratio = detJ*weight;
    double pres = (PK1->operator()(0,0)+PK1->operator()(1,1)+PK1->operator()(2,2))/3.;

    const std::vector<TensorialTraits<double>::GradType> &Grads = vipv[i]->gradf(&space1,ele,GP[i]);
    const std::vector<TensorialTraits<double>::ValType> &Vals = vipv[i]->f(&space1,ele,GP[i]);
    for(int j=0;j<nbFF; j++)
    {
      BT(0) = Grads[j+0*nbFF][0];
      BT(1) = Grads[j+0*nbFF][1];
      BT(2) = Grads[j+0*nbFF][2];
      // x y z
      for(int kk=0;kk<3;kk++)
      {
        /*if((ele->getTypeForMSH() == MSH_HEX_8 or ele->getTypeForMSH() == MSH_QUA_4  or
      ele->getTypeForMSH() == MSH_TET_4 or ele->getTypeForMSH() == MSH_TRI_3 or ele->getTypeForMSH() == MSH_PRI_6) and !_fullDg)
        {
          for(int m=0; m<3; m++)
          {
            vFor(j+kk*nbFF) += ratio*(BT(m)*(PK1->operator()(kk,m)));
          }
          vFor(j+kk*nbFF) -= ratio*(BT(kk)*(pres-averagePressure));
        }
        else
        {*/
        for(int m=0; m<3; m++)
        {
          vFor(j+kk*nbFF) += ratio*(BT(m)*PK1->operator()(kk,m));
        }
        //}
      }
    }
    //non local
    if( vipv[i]->getNumberNonLocalVariable()>getNumNonLocalVariable() ) Msg::Error("Your material law uses more non local variables than your domain dG3DForceBulk::get");
    for (int nl = 0; nl< vipv[i]->getNumberNonLocalVariable(); nl++){
      const STensor3 *cg = &vipv[i]->getConstRefToCharacteristicLengthMatrix(nl);
      
      
      static double nonlocalVar;
      nonlocalVar = vipv[i]->getConstRefToNonLocalVariable(nl);
      if (_incrementNonlocalBased){
        nonlocalVar -= vipvprev[i]->getConstRefToNonLocalVariable(nl);
      }
      static SVector3 gradNonlocalVar;
      gradNonlocalVar = vipv[i]->getConstRefToGradNonLocalVariable()[nl];
      if (_incrementNonlocalBased){
        gradNonlocalVar -= vipvprev[i]->getConstRefToGradNonLocalVariable()[nl];
      }
      
      static double localVar;
      localVar = vipv[i]->getConstRefToLocalVariable(nl);
      if (_incrementNonlocalBased){
        localVar -= vipvprev[i]->getConstRefToLocalVariable(nl);
      }
      
      int nbdof = this->space1.getNumKeys(ele);
      const int threeTimesNbFF = 3*nbFF;

      for(int j=0;j<nbFF; j++)
      {
        vFor(j+nl*nbFF+threeTimesNbFF) +=  getNonLocalEqRatio()*ratio*(Vals[j+nl*nbFF+threeTimesNbFF]*nonlocalVar);
        for(int m=0; m<3; m++)
        {
          for(int n=0; n<3; n++)
          {
            vFor(j+nl*nbFF+threeTimesNbFF) +=  getNonLocalEqRatio()*ratio*(Grads[j+nl*nbFF+threeTimesNbFF](m)*cg->operator()(m,n)*gradNonlocalVar(n));
          }
        }
        vFor(j+nl*nbFF+threeTimesNbFF) -=  getNonLocalEqRatio()*ratio*(Vals[j+nl*nbFF+threeTimesNbFF]*localVar);
      }
    }
    //extra dof
    if( vipv[i]->getNumConstitutiveExtraDofDiffusionVariable()>getNumConstitutiveExtraDofDiffusionVariable() ){
         Msg::Error("Your material law uses more constitutive extra dof variables than your domain dG3DForceBulk::get");}
    for (int extraDOFField = 0; extraDOFField< vipv[i]->getNumConstitutiveExtraDofDiffusionVariable(); extraDOFField++){
      const SVector3 &flux =   vipv[i]-> getConstRefToFlux()[extraDOFField];
      double w = 0.;
      if (getConstitutiveExtraDofDiffusionAccountFieldSource()){
         w += vipv[i]->getConstRefToFieldSource()(extraDOFField);
        }
      if (getConstitutiveExtraDofDiffusionAccountMecaSource()){
         w += vipv[i]->getConstRefToMechanicalSource()(extraDOFField);
      }
      int indexField=3+getNumNonLocalVariable()+extraDOFField;

      for(int j=0;j<nbFF; j++)
      {
        for(int m=0; m<3; m++)
        {
          vFor(j+indexField*nbFF) += getConstitutiveExtraDofDiffusionEqRatio()*ratio*(Grads[j+indexField*nbFF][m]*flux(m));
        }
        if(extraDOFField==0){ // corresponds to temperature, but when considering electro-mechanics, ???
          vFor(j+indexField*nbFF) += getConstitutiveExtraDofDiffusionEqRatio()*ratio*(w*Vals[j+indexField*nbFF]);// related to cp
        }
      }
    }
  }
  //vFor.print("bulkF");
}

void dG3DStiffnessBulk::get(MElement *ele,int npts,IntPt *GP,fullMatrix<double> &mStiff) const
{
  dG3DStiffnessBulk::get(ele, npts, GP, mStiff, true);
}

void dG3DStiffnessBulk::get(MElement *ele,int npts,IntPt *GP,fullMatrix<double> &mStiff, bool considerNonElasticOnes) const
{
  if (sym)
  {
    // Initialization of some data
    int nbdof = g3DBilinearTerm<double,double>::space1.getNumKeys(ele);
    int nbFF = ele->getNumShapeFunctions();
    const dG3DMaterialLaw *mlawbulk;
    if(_mlaw->getType() == materialLaw::fracture)
      mlawbulk = static_cast<const dG3DMaterialLaw* >((static_cast<const FractureByCohesive3DLaw* > (_mlaw) )->getBulkLaw());
    else
      mlawbulk= static_cast<const dG3DMaterialLaw*>(_mlaw);
    bool useBarF=mlawbulk->getUseBarF();

    mStiff.resize(nbdof, nbdof,true); // true --> setAll(0.)

    // get ipvariables
    _ipf->getIPv(ele,vipv);

    // sum on Gauss' points
    static SVector3 B;
    static SVector3 BT;
    static SVector3 Bipg;

/*    std::vector<SVector3> HBaverage;
    std::vector< std::vector<double> > dPdFieldVaverage;
    std::vector< std::vector<double> > dPdNonLocalVaverage;

    if((ele->getTypeForMSH() == MSH_HEX_8 or ele->getTypeForMSH() == MSH_QUA_4  or
      ele->getTypeForMSH() == MSH_TET_4 or ele->getTypeForMSH() == MSH_TRI_3 or ele->getTypeForMSH() == MSH_PRI_6) and !_fullDg)
    {
      HBaverage.resize(nbFF);
      for (int ipg = 0; ipg < npts; ipg++)
      {
        const std::vector<TensorialTraits<double>::GradType>& Gradsipg = vipv[ipg]->gradf(&space1,ele,GP[ipg]);
        const STensor43 *Hipg =&vipv[ipg]->getConstRefToTangentModuli();
        for(int k=0;k<nbFF;k++){
          SVector3 HBaverageTmp;
          STensorOperation::zero(HBaverageTmp);
          Bipg(0) = Gradsipg[k][0];
          Bipg(1) = Gradsipg[k][1];
          Bipg(2) = Gradsipg[k][2];
          for(int ll=0; ll<3; ll++)
          {
            for(int n=0; n<3; n++)
            {
              HBaverageTmp(ll) += (Hipg->operator()(0,0,ll,n)+Hipg->operator()(1,1,ll,n)+Hipg->operator()(2,2,ll,n))*Bipg(n)/3./npts;
            }
          }
          HBaverage[k]+=HBaverageTmp; //.push_back(HBaverageTmp);
        }
      }
      dPdFieldVaverage.resize(getNumConstitutiveExtraDofDiffusionVariable());
      for (int extraDOFField = 0; extraDOFField < getNumConstitutiveExtraDofDiffusionVariable(); extraDOFField++){
        dPdFieldVaverage[extraDOFField].resize(nbFF);
        int indexField=3+getNumNonLocalVariable()+extraDOFField;
        for (int ipg = 0; ipg < npts; ipg++)
        {
          for(int k=0;k<nbFF;k++){
            if(vipv[ipg]->getNumConstitutiveExtraDofDiffusionVariable()>extraDOFField)
            {
              const std::vector<TensorialTraits<double>::ValType> &Valsipg = vipv[ipg]->f(&space1,ele,GP[ipg]);
              const STensor3  &dPdField        = vipv[ipg]->getConstRefTodPdField()[extraDOFField];
              dPdFieldVaverage[extraDOFField][k]+=(dPdField(0,0)+dPdField(1,1)+dPdField(2,2))/3./npts*Valsipg[k+indexField*nbFF];
            }
          }
        }
      }
      dPdNonLocalVaverage.resize(getNumNonLocalVariable());
      for (int nlk = 0; nlk < getNumNonLocalVariable(); nlk++){
        dPdNonLocalVaverage[nlk].resize(nbFF);
        int indexField=3+nlk;
        for (int ipg = 0; ipg < npts; ipg++)
        {
          for(int k=0;k<nbFF;k++){
            if(vipv[ipg]->getNumberNonLocalVariable()>nlk)
            {
              const std::vector<TensorialTraits<double>::ValType> &Valsipg = vipv[ipg]->f(&space1,ele,GP[ipg]);
              const STensor3  &dsdp = vipv[ipg]->getConstRefToDStressDNonLocalVariable()[nlk];
              dPdNonLocalVaverage[nlk][k]+=(dsdp(0,0)+dsdp(1,1)+dsdp(2,2))/3./npts*Valsipg[k+indexField*nbFF];
            }
          }
        }
      }
    }*/
    static std::vector<SVector3> dBarJdu;
    static std::vector< std::vector< STensor33 > > dBarFdu; //for each GP and for each node
    if(useBarF)
    {
      dBarJdu.resize(nbFF);
      dBarFdu.resize(npts);
      for (int k=0;k<nbFF;k++)
      {
        STensorOperation::zero(dBarJdu[k]); //check here if ok
      }
      for (int ipg = 0; ipg < npts; ipg++)
      {
        dBarFdu[ipg].resize(nbFF);
        for (int k=0;k<nbFF;k++)
        {
          STensorOperation::zero(dBarFdu[ipg][k]); //check here if ok
        }
      }
      double totalWeight=0.;
      for(int ipg = 0; ipg < npts; ipg++)
        totalWeight+=GP[ipg].weight;
      for(int ipg = 0; ipg < npts; ipg++)
      {
        const std::vector<TensorialTraits<double>::GradType>& Gradsipg = vipv[ipg]->gradf(&space1,ele,GP[ipg]);
        const STensor3 &barF = vipv[ipg]->getConstRefToDeformationGradient(); //this is actually barF
        STensor3 invBarF(barF.invert());
        double localJ=vipv[ipg]->getLocalJ();
        double barJ=vipv[ipg]->getBarJ();
        double weight=GP[ipg].weight*pow(localJ,2./3.)*pow(barJ,1./3.)/totalWeight;
        for(int k=0;k<nbFF;k++)
        {
          SVector3 dBarJduTmp;
          STensorOperation::zero(dBarJduTmp);
          Bipg(0) = Gradsipg[k][0];
          Bipg(1) = Gradsipg[k][1];
          Bipg(2) = Gradsipg[k][2];
          for(int ll=0; ll<3; ll++)
          {
            for(int n=0; n<3; n++)
            {
              dBarJduTmp(ll) += invBarF(n,ll)*Bipg(n)*weight;
            }
          }
          dBarJdu[k]+=dBarJduTmp; 
        }
      }
      
      static STensor3 I;
      STensorOperation::unity(I);
      for (int ipg = 0; ipg < npts; ipg++)
      {
        const std::vector<TensorialTraits<double>::GradType>& Gradsipg = vipv[ipg]->gradf(&space1,ele,GP[ipg]);
        const STensor3 &barF = vipv[ipg]->getConstRefToDeformationGradient(); //this is actually barF
        STensor3 invBarF(barF.invert());
        double localJ=vipv[ipg]->getLocalJ();
        double barJ=vipv[ipg]->getBarJ();
        for(int k=0;k<nbFF;k++)
        {
          static STensor33 dBarFduTmp;
          STensorOperation::zero(dBarFduTmp);
          for(int kk=0; kk<3; kk++)
          {
            for(int m=0; m<3; m++)
            {
              for(int jj=0; jj<3; jj++)
              {
                dBarFduTmp(kk,m,jj) += pow(barJ,1./3.)*pow(localJ,-1./3.)*I(kk,jj)*Gradsipg[k](m);
                for(int p=0; p<3; p++)
                {
                  dBarFduTmp(kk,m,jj) -= 1./3.*pow(barJ,1./3.)*pow(localJ,-1./3.)*
                                                        barF(kk,m)*invBarF(p,jj)*Gradsipg[k](p);             
                }
                dBarFduTmp(kk,m,jj) += 1./3./barJ*barF(kk,m)*dBarJdu[k](jj);
              }                                
            }
          }
          dBarFdu[ipg][k]=dBarFduTmp;
        }
      }
    }

    for (int i = 0; i < npts; i++)
    {
      double weight = GP[i].weight;
      double& detJ = vipv[i]->getJacobianDeterminant(ele,GP[i]);
      double ratio = detJ*weight;
      // x, y, z
      const STensor43 *H =&vipv[i]->getConstRefToTangentModuli();
      const std::vector<TensorialTraits<double>::GradType> &Grads = vipv[i]->gradf(&space1,ele,GP[i]);
      const std::vector<TensorialTraits<double>::ValType> &Vals = vipv[i]->f(&space1,ele,GP[i]);
      const int threeTimesNbFF = 3*nbFF;
      //for barF
      const STensor3 &barF = vipv[i]->getConstRefToDeformationGradient(); //this is actually barF
      STensor3 invBarF;
      if(useBarF)
        invBarF=barF.invert();
      double localJ=vipv[i]->getLocalJ();
      double barJ=vipv[i]->getBarJ();

      for(int j=0;j<nbFF; j++)
      {
        BT(0) = Grads[j][0];
        BT(1) = Grads[j][1];
        BT(2) = Grads[j][2];
        for(int k=0;k<nbFF;k++){
          // x y z
          B(0) = Grads[k][0];
          B(1) = Grads[k][1];
          B(2) = Grads[k][2];

          for(int kk=0;kk<3;kk++)
          {
            for(int ll=0; ll<3; ll++)
            {
              /*if((ele->getTypeForMSH() == MSH_HEX_8 or ele->getTypeForMSH() == MSH_QUA_4  or
      ele->getTypeForMSH() == MSH_TET_4 or ele->getTypeForMSH() == MSH_TRI_3 or ele->getTypeForMSH() == MSH_PRI_6) and !_fullDg)
              {
                for(int m=0; m<3; m++)
                {
                  for(int n=0; n<3; n++)
                  {
                    mStiff(j+kk*nbFF,k+ll*nbFF) += ratio*(BT(m)*H->operator()(kk,m,ll,n)*B(n));
                  }
                  mStiff(j+kk*nbFF,k+ll*nbFF) -= ratio*(BT(kk)*(H->operator()(0,0,ll,m)+H->operator()(1,1,ll,m)+H->operator()(2,2,ll,m))*B(m))/3.;
                }
                mStiff(j+kk*nbFF,k+ll*nbFF) += ratio*(BT(kk)*HBaverage[k](ll));
              }*/
              if(useBarF)
              {
                for(int m=0; m<3; m++)
                {
                  for(int ii=0; ii<3; ii++)
                  {
                    for(int p=0; p<3; p++)
                    {
                      mStiff(j+kk*nbFF,k+ll*nbFF) += ratio*(BT(m)*H->operator()(kk,m,ii,p)*dBarFdu[i][k](ii,p,ll));
                    }                                
                  }
                }
              }
              else
              {
                for(int m=0; m<3; m++)
                {
                  for(int n=0; n<3; n++)
                  {
                    mStiff(j+kk*nbFF,k+ll*nbFF) += ratio*(BT(m)*H->operator()(kk,m,ll,n)*B(n));
                  }
                }
              }
            }
          }
        }
      }
      if( vipv[i]->getNumberNonLocalVariable()>getNumNonLocalVariable() )  Msg::Error("Your material law uses more non local variables than your domain dG3DStiffnessBulk::get");
      if( vipv[i]->getNumConstitutiveExtraDofDiffusionVariable()>getNumConstitutiveExtraDofDiffusionVariable() ) 
            Msg::Error("Your material law uses more constitutive extra dof variables than your domain dG3DStiffnessBulk::get");
			
      for (int nlk = 0; nlk < vipv[i]->getNumberNonLocalVariable(); nlk++){
        const STensor3  *dpds = &vipv[i]->getConstRefToDLocalVariableDStrain()[nlk];
        const STensor3  *dsdp = &vipv[i]->getConstRefToDStressDNonLocalVariable()[nlk];
        const STensor3    *cg = &vipv[i]->getConstRefToCharacteristicLengthMatrix(nlk);           // by Wu Ling
        

        for(int j=0;j<nbFF; j++){
          for(int k=0;k<nbFF;k++){
            // principal term
            double BtB = 0.;
            for(int m=0; m<3; m++){
              for(int n=0; n<3; n++){
                BtB += Grads[j+nlk*nbFF+threeTimesNbFF](m)*(cg->operator()(m,n))*Grads[k+nlk*nbFF+threeTimesNbFF](n);
              }
            }
            mStiff(j+nlk*nbFF+threeTimesNbFF,k+nlk*nbFF+threeTimesNbFF)
                    += getNonLocalEqRatio()*ratio*(Vals[j+nlk*nbFF+threeTimesNbFF]*Vals[k+nlk*nbFF+threeTimesNbFF]+ BtB);

            // cross terms to check
            if(considerNonElasticOnes)
            {
              for(int kk=0;kk<3;kk++)
              {
              /*if((ele->getTypeForMSH() == MSH_HEX_8 or ele->getTypeForMSH() == MSH_QUA_4  or
      ele->getTypeForMSH() == MSH_TET_4 or ele->getTypeForMSH() == MSH_TRI_3 or ele->getTypeForMSH() == MSH_PRI_6) and !_fullDg)
              {
                for(int m=0; m<3; m++)
                {
                  mStiff(j+kk*nbFF,k+nlk*nbFF+threeTimesNbFF) += ratio*(Grads[j+kk*nbFF](m)*dsdp->operator()(kk,m)*Vals[k+nlk*nbFF+threeTimesNbFF]); 
                }
                mStiff(j+kk*nbFF,k+nlk*nbFF+threeTimesNbFF) -= ratio*(Grads[j+kk*nbFF][kk]*(dsdp->operator()(0,0)+dsdp->operator()(1,1)+dsdp->operator()(2,2))*Vals[k+nlk*nbFF+threeTimesNbFF])/3.;
                mStiff(j+kk*nbFF,k+nlk*nbFF+threeTimesNbFF) += ratio*(Grads[j+kk*nbFF][kk]*dPdNonLocalVaverage[nlk][k]);
              }
              else
              {*/
                for(int m=0; m<3; m++)
                  mStiff(j+kk*nbFF,k+nlk*nbFF+threeTimesNbFF) += ratio*(Grads[j+kk*nbFF](m)*dsdp->operator()(kk,m)*Vals[k+nlk*nbFF+threeTimesNbFF]);
              //}
                if(useBarF)
                {
                  for(int ii=0; ii<3; ii++)
                  {
                    for(int p=0; p<3; p++)
                    {
                      mStiff(j+nlk*nbFF+threeTimesNbFF,k+kk*nbFF) -= getNonLocalEqRatio()*ratio*(Vals[j+nlk*nbFF+threeTimesNbFF]*
                                             dpds->operator()(ii,p)*dBarFdu[i][k](ii,p,kk));
                    }                                
                  }
                }
                else
                {
                  for(int m=0; m<3; m++)
                  {
                    mStiff(j+nlk*nbFF+threeTimesNbFF,k+kk*nbFF) -= getNonLocalEqRatio()*ratio*(Vals[j+nlk*nbFF+threeTimesNbFF]*
                                             dpds->operator()(kk,m)*Grads[k+kk*nbFF](m));
                  }
                }
              }
            }
          }
        }
        
        for (int mlk = 0; mlk < vipv[i]->getNumberNonLocalVariable(); mlk++){
          double  SpBar = vipv[i]->getConstRefToDLocalVariableDNonLocalVariable()(nlk,mlk);
          for(int j=0;j<nbFF; j++){
            for(int k=0;k<nbFF;k++){
              mStiff(j+nlk*nbFF+threeTimesNbFF,k+mlk*nbFF+threeTimesNbFF) -= getNonLocalEqRatio()*ratio*(Vals[j+nlk*nbFF+threeTimesNbFF]*SpBar*Vals[k+mlk*nbFF+threeTimesNbFF]);
            }
          }
        }
        
        for (int extraDOFField = 0; extraDOFField < vipv[i]->getNumConstitutiveExtraDofDiffusionVariable(); extraDOFField++){
          //Msg::Error("Check non-local / extra dof coupling in dG3DStiffnessBulk::get");
          double  dLocalVariableDExtraDof = vipv[i]->getConstRefTodLocalVariableDExtraDofDiffusionField()(nlk,extraDOFField);
          for(int j=0;j<nbFF; j++){
            for(int k=0;k<nbFF;k++){
              int indexField=3+getNumNonLocalVariable()+extraDOFField;
              // principal term
              mStiff(j+nlk*nbFF+threeTimesNbFF,k+indexField*nbFF)
                    += getNonLocalEqRatio()*ratio*(Vals[j+nlk*nbFF+threeTimesNbFF]*Vals[k+indexField*nbFF]*dLocalVariableDExtraDof);
            }
          }
        }
      }
  if(!getConstitutiveExtraDofDiffusionUseEnergyConjugatedField()){
      for (int extraDOFField = 0; extraDOFField< vipv[i]->getNumConstitutiveExtraDofDiffusionVariable(); extraDOFField++){

        const STensor33 &dFluxdF         = vipv[i]->getConstRefTodFluxdF()[extraDOFField]; //in reference conf
        const STensor3  &dPdField        = vipv[i]->getConstRefTodPdField()[extraDOFField];
        static STensor3 dwdf;
        STensorOperation::zero(dwdf);

        if (getConstitutiveExtraDofDiffusionAccountFieldSource()){
          if(extraDOFField==0){
          dwdf += vipv[i]->getConstRefTodFieldSourcedF()[extraDOFField];
           }
        }
        if (getConstitutiveExtraDofDiffusionAccountMecaSource()){
          if(extraDOFField==0){
          dwdf += vipv[i]->getConstRefTodMechanicalSourcedF()[extraDOFField];
          }
        }
        for(int j=0;j<nbFF; j++){
          for(int k=0;k<nbFF;k++){
            int indexField=3+getNumNonLocalVariable()+extraDOFField;

 // F-related terms
            for(int kk=0;kk<3;kk++)
            {
              /*if((ele->getTypeForMSH() == MSH_HEX_8 or ele->getTypeForMSH() == MSH_QUA_4  or
      ele->getTypeForMSH() == MSH_TET_4 or ele->getTypeForMSH() == MSH_TRI_3 or ele->getTypeForMSH() == MSH_PRI_6) and !_fullDg)
              {
                for(int m=0; m<3; m++)
                {
                  mStiff(j+kk*nbFF,k+indexField*nbFF) += ratio*(Grads[j+kk*nbFF][m]*dPdField.operator()(kk,m)*Vals[k+indexField*nbFF]);

                }
                mStiff(j+kk*nbFF,k+indexField*nbFF) -= ratio*(Grads[j+kk*nbFF][kk]*(dPdField.operator()(0,0)+dPdField.operator()(1,1)+dPdField.operator()(2,2))*Vals[k+indexField*nbFF])/3.;
                mStiff(j+kk*nbFF,k+indexField*nbFF) += ratio*(Grads[j+kk*nbFF][kk]*dPdFieldVaverage[extraDOFField][k]);
              }
              else
              {*/
              for(int m=0; m<3; m++)
                  mStiff(j+kk*nbFF,k+indexField*nbFF) += ratio*(Grads[j+kk*nbFF][m]*dPdField.operator()(kk,m)*Vals[k+indexField*nbFF]);
              //}
              if(useBarF)
              {
                for(int m=0; m<3; m++)
                {
                  for(int ii=0; ii<3; ii++)
                  {
                    for(int p=0; p<3; p++)
                    {
                      mStiff(j+indexField*nbFF,k+kk*nbFF) += getConstitutiveExtraDofDiffusionEqRatio()*ratio*
                           (Grads[j+indexField*nbFF][m]*dFluxdF.operator()(m,ii,p)*dBarFdu[i][k](ii,p,kk));
                    }
                    if(extraDOFField==0) 
                      mStiff(j+indexField*nbFF,k+kk*nbFF)+=getConstitutiveExtraDofDiffusionEqRatio()*ratio*
                                            (Vals[j+indexField*nbFF]*dwdf.operator()(ii,m)*dBarFdu[i][k](ii,m,kk)); // related to cp
                  }
                }
              }
              else
              {
                for(int m=0; m<3; m++)
                {
                  for(int n=0; n<3; n++)
                  {
                    mStiff(j+indexField*nbFF,k+kk*nbFF) += getConstitutiveExtraDofDiffusionEqRatio()*ratio*
                                     (Grads[j+indexField*nbFF][m]*dFluxdF.operator()(m,kk,n)*Grads[k][n]);
                  }
                  if(extraDOFField==0) 
                     mStiff(j+indexField*nbFF,k+kk*nbFF)+=getConstitutiveExtraDofDiffusionEqRatio()*ratio*
                                            (Vals[j+indexField*nbFF]*dwdf.operator()(kk,m)*Grads[k][m]); // related to cp
                }
              }
            }
          }
        }
	for (int extraDOFField2 = 0; extraDOFField2< vipv[i]->getNumConstitutiveExtraDofDiffusionVariable(); extraDOFField2++){
        
        //int extraDOFField2=extraDOFField;
        const STensor3  &dFluxdGradField = vipv[i]->getConstRefTodFluxdGradField()[extraDOFField][extraDOFField2]; //in reference conf
        const SVector3  &dFluxdField     = vipv[i]->getConstRefTodFluxdField()[extraDOFField][extraDOFField2]; //in reference conf

        double dwdt = 0.;
        if (getConstitutiveExtraDofDiffusionAccountFieldSource()){
          if(extraDOFField==0 && extraDOFField2==0){
            dwdt += vipv[i]->getConstRefTodFieldSourcedField()(extraDOFField,extraDOFField2);
          }
        }
        if (getConstitutiveExtraDofDiffusionAccountMecaSource()){
          if(extraDOFField==0 && extraDOFField2==0){
            dwdt += vipv[i]->getConstRefTodMechanicalSourcedField()(extraDOFField,extraDOFField2);
          }
        }
        for(int j=0;j<nbFF; j++){
          for(int k=0;k<nbFF;k++){
            int indexField=3+getNumNonLocalVariable()+extraDOFField;
            int indexField2=3+getNumNonLocalVariable()+extraDOFField2;

            // Pure & Cross Terms
            for(int m=0; m<3; m++)
            {
              mStiff(j+indexField*nbFF,k+indexField2*nbFF)+=getConstitutiveExtraDofDiffusionEqRatio()*ratio*(Grads[j+indexField*nbFF][m]*dFluxdField.operator()(m)*Vals[k+indexField2*nbFF]);
              for(int n=0; n<3; n++)
              {
                mStiff(j+indexField*nbFF,k+indexField2*nbFF)+=getConstitutiveExtraDofDiffusionEqRatio()*ratio*(Grads[j+indexField*nbFF][m]*dFluxdGradField.operator()(m,n)*Grads[k+indexField2*nbFF][n]);
              }
            }
         
            if(extraDOFField==0 && extraDOFField2==0) mStiff(j+indexField*nbFF,k+indexField*nbFF)+=getConstitutiveExtraDofDiffusionEqRatio()*ratio*dwdt*Vals[j+indexField*nbFF]*Vals[k+indexField*nbFF];// related to cp
          }
        }
      }
     }
    }

    if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField()){
      for (int extraDOFField = 0; extraDOFField< vipv[i]->getNumConstitutiveExtraDofDiffusionVariable(); extraDOFField++){

        const STensor33 &dFluxdF         = vipv[i]->getConstRefTodFluxdF()[extraDOFField]; //in reference conf
        const STensor3  &dPdField1       = vipv[i]->getConstRefTodPdField()[extraDOFField];

        int extraDOFField_aux=0;
        if((extraDOFField+1)< vipv[i]->getNumConstitutiveExtraDofDiffusionVariable()) extraDOFField_aux=extraDOFField+1;
        else extraDOFField_aux=extraDOFField-1;        

        const STensor3  &dPdField2       = vipv[i]->getConstRefTodPdField()[extraDOFField_aux];

        double dFielddEnergyConjugatedField1 = vipv[i]->getConstRefTodFielddEnergyConjugatedField()[extraDOFField][extraDOFField];
        double dFielddEnergyConjugatedField2 = vipv[i]->getConstRefTodFielddEnergyConjugatedField()[extraDOFField_aux][extraDOFField];

        static STensor3 dwdf;
        STensorOperation::zero(dwdf);

        if (getConstitutiveExtraDofDiffusionAccountFieldSource()){
          if(extraDOFField==0){
            dwdf += vipv[i]->getConstRefTodFieldSourcedF()[extraDOFField];
          }
        }
        if (getConstitutiveExtraDofDiffusionAccountMecaSource()){
          if(extraDOFField==0){
            dwdf += vipv[i]->getConstRefTodMechanicalSourcedF()[extraDOFField];
          }
        }
        for(int j=0;j<nbFF; j++){
          for(int k=0;k<nbFF;k++){
            int indexField=3+getNumNonLocalVariable()+extraDOFField;

 // F-related terms
            for(int kk=0;kk<3;kk++)
            {
              for(int m=0; m<3; m++)
              {
                mStiff(j+kk*nbFF,k+indexField*nbFF) += ratio*(Grads[j+kk*nbFF][m]*(dPdField1.operator()(kk,m)*dFielddEnergyConjugatedField1+
                                             dPdField2.operator()(kk,m)*dFielddEnergyConjugatedField2)*Vals[k+indexField*nbFF]);
                if(useBarF)
                {
                  for(int ii=0; ii<3; ii++)
                  {
                    for(int p=0; p<3; p++)
                    {
                      mStiff(j+indexField*nbFF,k+kk*nbFF) += getConstitutiveExtraDofDiffusionEqRatio()*ratio*
                                              (Grads[j+indexField*nbFF][m]*dFluxdF.operator()(m,ii,p)*dBarFdu[i][k](ii,p,kk));
                    }
                    if(extraDOFField==0) 
                    {
                      mStiff(j+indexField*nbFF,k+kk*nbFF)+=getConstitutiveExtraDofDiffusionEqRatio()*ratio*
                                              (Vals[j+indexField*nbFF]*dwdf.operator()(ii,m)*dBarFdu[i][k](ii,m,kk)); // related to cp
                    }
                  }
                }
                else
                {
                  for(int n=0; n<3; n++)
                  {
                    mStiff(j+indexField*nbFF,k+kk*nbFF) += getConstitutiveExtraDofDiffusionEqRatio()*ratio*
                                              (Grads[j+indexField*nbFF][m]*dFluxdF.operator()(m,kk,n)*Grads[k][n]);
                  }
                  if(extraDOFField==0) 
                  {
                    mStiff(j+indexField*nbFF,k+kk*nbFF)+=getConstitutiveExtraDofDiffusionEqRatio()*ratio*
                                              (Vals[j+indexField*nbFF]*dwdf.operator()(kk,m)*Grads[k][m]); // related to cp
                  }
                }
              }
            }
	  }
        }
	for (int extraDOFField2 = 0; extraDOFField2< vipv[i]->getNumConstitutiveExtraDofDiffusionVariable(); extraDOFField2++)
        {
          //int extraDOFField2=extraDOFField;
          const STensor3  &dFluxdGradField1                         = vipv[i]->getConstRefTodFluxdGradField()[extraDOFField][extraDOFField2]; //in reference conf
          const SVector3  &dFluxdField1                             = vipv[i]->getConstRefTodFluxdField()[extraDOFField][extraDOFField2]; //in reference conf

          int extraDOFField_aux=0;
          if((extraDOFField2+1)< vipv[i]->getNumConstitutiveExtraDofDiffusionVariable()) extraDOFField_aux=extraDOFField2+1;
          else extraDOFField_aux=extraDOFField2-1;        

          const STensor3  &dFluxdGradField2 = vipv[i]->getConstRefTodFluxdGradField()[extraDOFField][extraDOFField_aux]; //in reference conf
          const SVector3  &dFluxdField2     = vipv[i]->getConstRefTodFluxdField()[extraDOFField][extraDOFField_aux]; //in reference conf

          const STensor3  &dGradFielddGradEnergyConjugatedField1 = vipv[i]->getConstRefTodGradFielddGradEnergyConjugatedField()[extraDOFField2][extraDOFField2];
          const SVector3  &dGradFielddEnergyConjugatedField1 = vipv[i]->getConstRefTodGradFielddEnergyConjugatedField()[extraDOFField2][extraDOFField2];
          double          dFielddEnergyConjugatedField1 = vipv[i]->getConstRefTodFielddEnergyConjugatedField()[extraDOFField2][extraDOFField2];
      
          const STensor3  &dGradFielddGradEnergyConjugatedField2 = vipv[i]->getConstRefTodGradFielddGradEnergyConjugatedField()[extraDOFField_aux][extraDOFField2];
          const SVector3  &dGradFielddEnergyConjugatedField2 = vipv[i]->getConstRefTodGradFielddEnergyConjugatedField()[extraDOFField_aux][extraDOFField2];
          double          dFielddEnergyConjugatedField2 = vipv[i]->getConstRefTodFielddEnergyConjugatedField()[extraDOFField_aux][extraDOFField2];

          double dwdt = 0.;
          if (getConstitutiveExtraDofDiffusionAccountFieldSource()){
            if(extraDOFField==0 && extraDOFField2==0){
              dwdt += vipv[i]->getConstRefTodFieldSourcedField()(extraDOFField,extraDOFField2);
            }
          }
          if (getConstitutiveExtraDofDiffusionAccountMecaSource()){
            if(extraDOFField==0 && extraDOFField2==0){
              dwdt += vipv[i]->getConstRefTodMechanicalSourcedField()(extraDOFField,extraDOFField2);
            }
          }
          for(int j=0;j<nbFF; j++){
            for(int k=0;k<nbFF;k++){
              int indexField=3+getNumNonLocalVariable()+extraDOFField;
              int indexField2=3+getNumNonLocalVariable()+extraDOFField2;

              // Pure & Cross Terms
              for(int m=0; m<3; m++)
              {
                mStiff(j+indexField*nbFF,k+indexField2*nbFF)+=getConstitutiveExtraDofDiffusionEqRatio()*ratio*
                              (Grads[j+indexField*nbFF][m]*(dFluxdField1.operator()(m)*dFielddEnergyConjugatedField1+
                               dFluxdField2.operator()(m)*dFielddEnergyConjugatedField2)*Vals[k+indexField2*nbFF]);

                 for(int n=0; n<3; n++)
                {
                  mStiff(j+indexField*nbFF,k+indexField2*nbFF)+=getConstitutiveExtraDofDiffusionEqRatio()*ratio*
                             (Grads[j+indexField*nbFF][m]*(dFluxdGradField1.operator()(m,n)*dGradFielddEnergyConjugatedField1.operator()(n)+
                             dFluxdGradField2.operator()(m,n)*dGradFielddEnergyConjugatedField2.operator()(n))*Vals[k+indexField2*nbFF]);
             
                  for(int p=0; p<3; p++)
              	  {
                    mStiff(j+indexField*nbFF,k+indexField2*nbFF)+=getConstitutiveExtraDofDiffusionEqRatio()*ratio*
                             (Grads[j+indexField*nbFF][m]*(dFluxdGradField1.operator()(m,p)*dGradFielddGradEnergyConjugatedField1.operator()(p,n)+
                             dFluxdGradField2.operator()(m,p)*dGradFielddGradEnergyConjugatedField2.operator()(p,n))*Grads[k+indexField2*nbFF][n]);
             	  }
                }
              }
              if(extraDOFField==0 && extraDOFField2==0) mStiff(j+indexField*nbFF,k+indexField*nbFF)+=
                            getConstitutiveExtraDofDiffusionEqRatio()*ratio*dwdt*dFielddEnergyConjugatedField1*
                            Vals[j+indexField*nbFF]*Vals[k+indexField*nbFF];// related to cp
            }
          }
        }
      }
    }
    for (int extraDOFField = 0; extraDOFField< vipv[i]->getNumConstitutiveExtraDofDiffusionVariable(); extraDOFField++){
      for (int nlk = 0; nlk < vipv[i]->getNumberNonLocalVariable(); nlk++)
      {
        //Msg::Error("Check extra dof / non-local coupling in dG3DStiffnessBulk::get");
        const SVector3 &dFluxdNonLocalVariable = vipv[i]->getConstRefTodFluxdNonLocalVariable()[extraDOFField][nlk];
        double dwdNonLocalVariable   = 0.;  
        if (getConstitutiveExtraDofDiffusionAccountFieldSource()){
          dwdNonLocalVariable += vipv[i]->getConstRefTodFieldSourcedNonLocalVariable()(extraDOFField,nlk);
        }
        if (getConstitutiveExtraDofDiffusionAccountMecaSource()){
          dwdNonLocalVariable += vipv[i]->getConstRefTodMechanicalSourcedNonLocalVariable()(extraDOFField,nlk);
        }
        if(considerNonElasticOnes)
        {
          for(int j=0;j<nbFF; j++){
            for(int k=0;k<nbFF;k++){
              int indexField=3+getNumNonLocalVariable()+extraDOFField;
              int indexNL=3+nlk;
              for(int m=0; m<3; m++)
              {
                mStiff(j+indexField*nbFF,k+indexNL*nbFF)+=getConstitutiveExtraDofDiffusionEqRatio()*ratio*
                                    (Grads[j+indexField*nbFF][m]*dFluxdNonLocalVariable.operator()(m)*Vals[k+indexNL*nbFF]);
              }
              mStiff(j+indexField*nbFF,k+indexNL*nbFF)+=getConstitutiveExtraDofDiffusionEqRatio()*ratio*
                                            dwdNonLocalVariable*Vals[j+indexField*nbFF]*Vals[k+indexNL*nbFF];
            }
          }
        }
      }
    }
         /*   // F-related terms
            for(int kk=0;kk<3;kk++)
            {
              if((ele->getTypeForMSH() == MSH_HEX_8 or ele->getTypeForMSH() == MSH_QUA_4  or
      ele->getTypeForMSH() == MSH_TET_4 or ele->getTypeForMSH() == MSH_TRI_3 or ele->getTypeForMSH() == MSH_PRI_6) and !_fullDg)
              {
                for(int m=0; m<3; m++)
                {
                  mStiff(j+kk*nbFF,k+indexField*nbFF) += ratio*(Grads[j+kk*nbFF][m]*dPdField.operator()(kk,m)*Vals[k+indexField*nbFF]);

                }
                mStiff(j+kk*nbFF,k+indexField*nbFF) -= ratio*(Grads[j+kk*nbFF][kk]*(dPdField.operator()(0,0)+dPdField.operator()(1,1)+dPdField.operator()(2,2))*Vals[k+indexField*nbFF])/3.;
                mStiff(j+kk*nbFF,k+indexField*nbFF) += ratio*(Grads[j+kk*nbFF][kk]*dPdFieldVaverage[extraDOFField][k]);
              }
              else
              {
                for(int m=0; m<3; m++)
                  mStiff(j+kk*nbFF,k+indexField*nbFF) += ratio*(Grads[j+kk*nbFF][m]*dPdField.operator()(kk,m)*Vals[k+indexField*nbFF]);
              }
              for(int m=0; m<3; m++)
              {
                for(int n=0; n<3; n++)
                {
                  mStiff(j+indexField*nbFF,k+kk*nbFF) += getConstitutiveExtraDofDiffusionEqRatio()*ratio*(Grads[j+indexField*nbFF][m]*dFluxdF.operator()(m,kk,n)*Grads[k][n]);
                }
                mStiff(j+indexField*nbFF,k+kk*nbFF)+=getConstitutiveExtraDofDiffusionEqRatio()*ratio*(Grads[k][m]*dwdf.operator()(kk,m)*Vals[j+indexField*nbFF]); // related to cp
              }
            }*/
       /* for (int nlk = 0; nlk < vipv[i]->getNumberNonLocalVariable(); nlk++){
          //Msg::Error("Check extra dof / non-local coupling in dG3DStiffnessBulk::get");
          const SVector3 &dFluxdNonLocalVariable = vipv[i]->getConstRefTodFluxdNonLocalVariable()[extraDOFField][nlk];
          double dwdNonLocalVariable   = 0.;  
          if (getConstitutiveExtraDofDiffusionAccountFieldSource()){
            dwdNonLocalVariable += vipv[i]->getConstRefTodFieldSourcedNonLocalVariable()(extraDOFField,nlk);
          }
          if (getConstitutiveExtraDofDiffusionAccountMecaSource()){
            dwdNonLocalVariable += vipv[i]->getConstRefTodMechanicalSourcedNonLocalVariable()(extraDOFField,nlk);
          }
          for(int j=0;j<nbFF; j++){
            for(int k=0;k<nbFF;k++){
              int indexField=3+getNumNonLocalVariable()+extraDOFField;
              int indexNL=3+nlk;
              for(int m=0; m<3; m++)
              {
                mStiff(j+indexField*nbFF,k+indexNL*nbFF)+=getConstitutiveExtraDofDiffusionEqRatio()*ratio*(Grads[j+indexField*nbFF][m]*dFluxdNonLocalVariable.operator()(m)*Vals[k+indexNL*nbFF]);
              }
              mStiff(j+indexField*nbFF,k+indexNL*nbFF)+=getConstitutiveExtraDofDiffusionEqRatio()*ratio*dwdNonLocalVariable*Vals[j+indexField*nbFF]*Vals[k+indexNL*nbFF];
            }
          }
        }*/
    }
  //mStiff.print("bulk");
  }
  else
    printf("not implemented\n");
}

void dG3DElasticStiffnessBulk::get(MElement *ele,int npts,IntPt *GP,fullMatrix<double> &mStiff) const
{
    dG3DStiffnessBulk::get(ele, npts, GP, mStiff, false);
  //mStiff.print("Elastic bulk");
}

void dG3DForceInter::get(MElement *ele, int npts, IntPt *GP, fullVector<double> &m)const
{
  MInterfaceElement *ie = dynamic_cast<MInterfaceElement*>(ele);
  const int nbFFm = ie->getElem(0)->getNumVertices();
  const int nbFFp = ie->getElem(1)->getNumVertices();
  const int nbdofm = _minusSpace->getNumKeys(ie->getElem(0));
  const int nbdofp = _plusSpace->getNumKeys(ie->getElem(1));
  m.resize(nbdofm+nbdofp);
  m.setAll(0.);
  if(_fullDg)
  {
    // characteristic size
    const double hs = ie->getCharacteristicSize();
    const double b1hs = getStabilityParameter()/hs;

    // get value at gauss's point
    const AllIPState::ipstateElementContainer *vips = _ipf->getAips()->getIPstate(ele->getNum());
    // Values on minus and plus elements
    IntPt *GPm; IntPt *GPp;
    _interQuad->getIntPoints(ie,GP,&GPm,&GPp);

    // loop on Gauss Point
    for(int i=0;i<npts; i++)
    {
      // IP
      const IPStateBase *ipsm        = (*vips)[i];
      const dG3DIPVariableBase *ipvm     = static_cast<const dG3DIPVariableBase*>(ipsm->getState(IPStateBase::current));
      const dG3DIPVariableBase *ipvmprev = static_cast<const dG3DIPVariableBase*>(ipsm->getState(IPStateBase::previous));

      const SVector3 &referenceOutwardNormalm = ipvm->getConstRefToReferenceOutwardNormal();

      const IPStateBase *ipsp        = (*vips)[i+npts];
      const dG3DIPVariableBase *ipvp     = static_cast<const dG3DIPVariableBase*>(ipsp->getState(IPStateBase::current));
      const dG3DIPVariableBase *ipvpprev    = static_cast<const dG3DIPVariableBase*>(ipsp->getState(IPStateBase::previous));

      const SVector3 &referenceOutwardNormalp = ipvp->getConstRefToReferenceOutwardNormal();
      const SVector3 &jump = ipvp->getConstRefToJump();

      // Weight of Gauss' point i and Jacobian's value at this point
      const double weight = GP[i].weight;
      const double& detJ = ipvm->getJacobianDeterminant(ie,GP[i]);
      const double wJ = detJ* weight;


      const std::vector<TensorialTraits<double>::GradType> &Gradsm = ipvm->gradf(_minusSpace,ie->getElem(0),GPm[i]);
      const std::vector<TensorialTraits<double>::GradType> &Gradsp = ipvp->gradf(_plusSpace,ie->getElem(1),GPp[i]);
      const std::vector<TensorialTraits<double>::ValType> &Valsm = ipvm->f(_minusSpace,ie->getElem(0),GPm[i]);
      const std::vector<TensorialTraits<double>::ValType> &Valsp = ipvp->f(_plusSpace,ie->getElem(1),GPp[i]);
      // possibility of fracture
      bool broken = false;
      if(_mlawMinus->getType() == materialLaw::fracture )
      {
        // when if failure is checked at the convergence state, previous or current is not important
        // but when failure is checked on the fly,  current must be used
        // this  corresponds to a more general case - Van Dung
        const FractureCohesive3DIPVariable *ipvwf = static_cast<const FractureCohesive3DIPVariable*>(ipvm); // broken via minus (OK broken on both sides)
        broken = ipvwf->isbroken();
      }

      if(!broken)
      {
        static SVector3 nu;
        nu = referenceOutwardNormalm;
        nu*=(1./referenceOutwardNormalm.norm());

        const std::vector<TensorialTraits<double>::ValType> &Valsm = ipvm->f(_minusSpace,ie->getElem(0),GPm[i]);
        const std::vector<TensorialTraits<double>::ValType> &Valsp = ipvp->f(_plusSpace,ie->getElem(1),GPp[i]);
        // consistency \mean{P}
        static SVector3 meanPN;
        for(int k =0; k<3; k++)
        {
          meanPN[k]  = 0.;
          for(int l =0; l<3; l++)
          {
            meanPN[k] += (ipvp->getConstRefToFirstPiolaKirchhoffStress()(k,l)+ipvm->getConstRefToFirstPiolaKirchhoffStress()(k,l))*(nu(l)/2.);
          }
        }
        // Stability  N \mean{H} (jump N)

        static STensor43 meanH;
        meanH = ipvp->getConstRefToDGElasticTangentModuli();
        meanH+= ipvm->getConstRefToDGElasticTangentModuli();
        meanH*=(0.5);

        static SVector3 NMeanHJumpNBetasc;
        for(int k = 0; k <3; k++)
        {
          NMeanHJumpNBetasc(k) = 0.;
          for(int l = 0; l <3; l++)
          {
            for(int m = 0; m <3; m++)
            {
              for(int n = 0; n <3; n++)
              {
                NMeanHJumpNBetasc(k) += meanH(k,l,m,n)*jump(m)*nu(n)*b1hs*nu(l);
              }
            }
          }
        }
        // Assembly consistency + stability (if not broken)
        for(int j=0;j<nbFFm;j++){
          for(int k=0;k<3;k++){
            m(j+k*nbFFm) -= (meanPN[k]+NMeanHJumpNBetasc[k])*(Valsm[j+0*nbFFm]*wJ);
          }
        }
        for(int j=0;j<nbFFp;j++){
          for(int k=0;k<3;k++){
            m(j+k*nbFFp+nbdofm) += (meanPN[k]+NMeanHJumpNBetasc[k])*(Valsp[j+0*nbFFp]*wJ);
          }
        }
        // compatibility membrane (4 terms to assembly)
        static STensor3 HmJumpN;
        static STensor3 HpJumpN;

        for(int k = 0; k <3; k++)
        {
          for(int l = 0; l <3; l++)
          {
            HmJumpN(k,l) = 0.;
            HpJumpN(k,l) = 0.;
            for(int m = 0; m <3; m++)
            {
              for(int n = 0; n <3; n++)
              {
                HmJumpN(k,l) += ipvm->getConstRefToDGElasticTangentModuli()(m,n,k,l)*jump(m)*nu(n);
                HpJumpN(k,l) += ipvp->getConstRefToDGElasticTangentModuli()(m,n,k,l)*jump(m)*nu(n);
              }
            }
          }
        }
        // Assembly (loop on shape function)
        for(int j=0;j<nbFFm;j++)
        {
          for(int k=0;k<3;k++)
          {
            for(int l=0;l<3;l++)
            {
              m(j+k*nbFFm) += (HmJumpN(k,l)*Gradsm[j+0*nbFFm][l])*(wJ/2.);
            }
          }
        }
        for(int j=0;j<nbFFm;j++)
        {
          for(int k=0;k<3;k++)
          {
            for(int l=0;l<3;l++)
            {
              m(j+k*nbFFp+nbdofm) += (HpJumpN(k,l)*Gradsp[j+0*nbFFp][l])*(wJ/2.);
            }
          }
        }
        //non local
        if( ipvm->getNumberNonLocalVariable()>getNumNonLocalVariable() ) Msg::Error("Your minus material law uses more non local variable than your domain dG3DForceInter::get");
        if( ipvp->getNumberNonLocalVariable()>getNumNonLocalVariable() ) Msg::Error("Your plus material law uses more non local variable than your domain dG3DForceInter::get");
        if( ipvm->getNumberNonLocalVariable()!= ipvm->getNumberNonLocalVariable()  ) Msg::Error("Your minus and plus material laws do not have the same number of non local variables dG3DForceInter::get");
        if( getNonLocalContinuity())
        {
          // characteristic size
          const double nbetahs = getNonLocalStabilityParameter()/hs;
 
          for (int nlk = 0; nlk< ipvp->getNumberNonLocalVariable(); nlk++){
            double epljump = ipvp->getConstRefToNonLocalJump()(nlk);
            if (_incrementNonlocalBased){
              epljump -= ipvpprev->getConstRefToNonLocalJump()(nlk);
            }
            const STensor3  *cgm = &(ipvm->getConstRefToCharacteristicLengthMatrix(nlk));           // by Wu Ling
            const STensor3  *cgp = &(ipvp->getConstRefToCharacteristicLengthMatrix(nlk));           // by Wu Ling
            
            static SVector3 gradm, gradp;
            gradm = ipvm->getConstRefToGradNonLocalVariable()[nlk];
            gradp = ipvp->getConstRefToGradNonLocalVariable()[nlk];
            if (_incrementNonlocalBased){
              gradm -= ipvmprev->getConstRefToGradNonLocalVariable()[nlk];
              gradp -= ipvpprev->getConstRefToGradNonLocalVariable()[nlk];
            };

            SVector3 cgGradm;
            SVector3 cgGradp;

            for(int k =0; k<3; k++)
            {
              cgGradm[k] = 0.;
              cgGradp[k] = 0.;
              for(int l =0; l<3; l++)
              {
                cgGradm[k] += cgm->operator()(k,l)*gradm(l);
                cgGradp[k] += cgp->operator()(k,l)*gradp(l);
              }
            }

            // consistency \mean{cg*grad eps}
            double meanCgGradN =0.;
            for(int k =0; k<3; k++)
            {
              meanCgGradN += (cgGradm(k)+cgGradp(k))*(nu(k)/2.);
            }
            // Stability  N \mean{H} (jump N)
            double NMeanCgJumpNBetasc = 0.;
            static STensor3 meanCg;
            meanCg = (*cgm);
            meanCg+= (*cgp);
            meanCg*=(1./2.);
            for(int l = 0; l <3; l++)
            {
              for(int n = 0; n <3; n++)
              {
                NMeanCgJumpNBetasc += meanCg(l,n)*epljump*nu(n)*nbetahs*nu(l);
              }
            }
            // Assembly consistency + stability (if not broken)
            for(int j=0;j<nbFFm;j++)
            {
              m(j+nlk*nbFFm+3*nbFFm) -= (meanCgGradN+NMeanCgJumpNBetasc)*(Valsm[j+nlk*nbFFm+3*nbFFm]*wJ*getNonLocalEqRatio());
            }
            for(int j=0;j<nbFFp;j++)
            {
              m(j+nlk*nbFFp+3*nbFFp+nbdofm) += (meanCgGradN+NMeanCgJumpNBetasc)*(Valsp[j+nlk*nbFFp+3*nbFFp]*wJ*getNonLocalEqRatio());
            }
            // compatibility membrane (4 terms to assembly)
            SVector3 CgmJumpN;
            SVector3 CgpJumpN;

            for(int k = 0; k <3; k++)
            {
              CgmJumpN(k) = 0.;
              CgpJumpN(k) = 0.;
              for(int m = 0; m <3; m++)
              {
                CgmJumpN(k) += cgm->operator()(k,m)*epljump*nu(m);
                CgpJumpN(k) += cgp->operator()(k,m)*epljump*nu(m);
              }
            }
            // Assembly (loop on shape function)
            for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<3;l++)
              {
                m(j+nlk*nbFFm+3*nbFFm) += (CgmJumpN(l)*Gradsm[j+nlk*nbFFm+3*nbFFm][l])*(wJ*getNonLocalEqRatio()/2.);
              }
            }
            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<3;l++)
              {
                m(j+nlk*nbFFp+3*nbFFp+nbdofm) += (CgpJumpN(l)*Gradsp[j+nlk*nbFFp+3*nbFFp][l])*(wJ*getNonLocalEqRatio()/2.);
              }
            }
          }
        }
        // constitutive extra dof
//if( ipvm->getNumConstitutiveExtraDofDiffusionVariable()==getNumConstitutiveExtraDofDiffusionVariable() ){
        if( ipvm->getNumConstitutiveExtraDofDiffusionVariable()>getNumConstitutiveExtraDofDiffusionVariable() ) 
            Msg::Error("Your material law uses more constitutive extra dof variables than your domain dG3DForceInter::get");
        if( ipvp->getNumConstitutiveExtraDofDiffusionVariable()>getNumConstitutiveExtraDofDiffusionVariable() ) 
            Msg::Error("Your material law uses more constitutive extra dof variables than your domain dG3DForceInter::get");
        if( ipvp->getNumConstitutiveExtraDofDiffusionVariable()!= ipvm->getNumConstitutiveExtraDofDiffusionVariable() ) 
            Msg::Error("Your plus and minus material laws do not use the same number of constitutive extra dof variables dG3DForceInter::get");
        for (int extraDOFField = 0; extraDOFField< ipvp->getNumConstitutiveExtraDofDiffusionVariable(); extraDOFField++){
         for (int extraDOFField2 = 0; extraDOFField2< ipvp->getNumConstitutiveExtraDofDiffusionVariable(); extraDOFField2++){
          if( getConstitutiveExtraDofDiffusionContinuity())
          {
            const double nbetahs = getConstitutiveExtraDofDiffusionStabilityParameter()/hs;
            double fieldJump;
            double fieldp=0.;double fieldm=0.;

            if(ipvp->getNumConstitutiveExtraDofDiffusionVariable()==1)
            {
             fieldJump = ipvp->getConstRefToFieldJump()(extraDOFField);
             fieldp = ipvp->getConstRefToField(extraDOFField);
             fieldm = ipvm->getConstRefToField(extraDOFField);
            }
            else
            {
             fieldJump = ipvp->getConstRefToEnergyConjugatedFieldJump()(extraDOFField2);
             fieldp = ipvp->getConstRefToEnergyConjugatedField(extraDOFField);
             fieldm = ipvm->getConstRefToEnergyConjugatedField(extraDOFField);
            }

            static STensor3  k0m;
            k0m = *(ipvm->getConstRefToLinearK()[extraDOFField][extraDOFField2]);
            static STensor3  k0p;
            k0p = *(ipvp->getConstRefToLinearK()[extraDOFField][extraDOFField2]);
            
            static STensor3  Stiff_alphadialitationm;
            static STensor3  Stiff_alphadialitationp;
            Stiff_alphadialitationm = *(ipvm->getConstRefToLinearSymmetrizationCoupling()[0]);
            Stiff_alphadialitationp = *(ipvp->getConstRefToLinearSymmetrizationCoupling()[0]);
            
            if(ipvp->getNumConstitutiveExtraDofDiffusionVariable()==1) //thermomec
            {
                if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())
               {
                k0m*=(-1.)*fieldm*fieldm;
                k0p*=(-1.)*fieldp*fieldp;
                Stiff_alphadialitationm*= (-1.)*fieldm*fieldm;
                Stiff_alphadialitationp*= (-1.)*fieldp*fieldp;
                }
             }
             else
             {
              if(extraDOFField==0)
                   {
                    Stiff_alphadialitationm*= (-1.)/(fieldm*fieldm);
                    Stiff_alphadialitationp*= (-1.)/(fieldp*fieldp); 
                   }
             }  
            const SVector3  &fluxm = (ipvm->getConstRefToFlux()[extraDOFField]);
            const SVector3  &fluxp = (ipvp->getConstRefToFlux()[extraDOFField]);
            // consistency \mean{cg*grad eps}
            double meantflux =0.;
            for(int k =0; k<3; k++)
            {
              meantflux += (fluxm(k)+fluxp(k))*(nu(k)/2.);
            }
            double NMeank0JumpNBetasc = 0.;
            static STensor3 meank;
            meank = (k0m);
            meank+= (k0p);
            meank*=(1./2.);

            for(int l = 0; l <3; l++)
            {
              for(int n = 0; n <3; n++)
              {
                NMeank0JumpNBetasc += meank(l,n)*fieldJump*nu(n)*nbetahs*nu(l);
              }
            }

            double Stiff_alphadialitationp_JumpN=0.;
            double Stiff_alphadialitationm_JumpN=0.;
            for(int k = 0; k <3; k++)
            {
              for(int m = 0; m <3; m++)
              {
                Stiff_alphadialitationp_JumpN += Stiff_alphadialitationp(k,m)*jump(k)*nu(m);
                Stiff_alphadialitationm_JumpN += Stiff_alphadialitationm(k,m)*jump(k)*nu(m);
              }
            }
            int indexField=3+getNumNonLocalVariable()+extraDOFField;
            
            if(extraDOFField==extraDOFField2)
            {
              for(int j=0;j<nbFFm;j++)
              {
                m(j+indexField*nbFFm) -= ( meantflux+NMeank0JumpNBetasc)*(Valsm[j+indexField*nbFFm]*wJ*getConstitutiveExtraDofDiffusionEqRatio());
              }
              for(int j=0;j<nbFFp;j++)
              {
                m(j+indexField*nbFFp+nbdofm) += ( meantflux+NMeank0JumpNBetasc)*(Valsp[j+indexField*nbFFp]*wJ*getConstitutiveExtraDofDiffusionEqRatio());
              }
            }
            else
            {
              for(int j=0;j<nbFFm;j++)
              {
                m(j+indexField*nbFFm) -= (NMeank0JumpNBetasc)*(Valsm[j+indexField*nbFFm]*wJ*getConstitutiveExtraDofDiffusionEqRatio());
              }
              for(int j=0;j<nbFFp;j++)
              {
                m(j+indexField*nbFFp+nbdofm) += (NMeank0JumpNBetasc)*(Valsp[j+indexField*nbFFp]*wJ*getConstitutiveExtraDofDiffusionEqRatio());
              }
            }
    
            SVector3 k0pJumpN;
            STensorOperation::zero(k0pJumpN);
            SVector3 k0mJumpN;
            STensorOperation::zero(k0mJumpN);
            for(int k = 0; k <3; k++)
            {
              for(int m = 0; m <3; m++)
              {
                k0mJumpN(k) += k0m(k,m)*fieldJump*nu(m);
                k0pJumpN(k) += k0p(k,m)*fieldJump*nu(m);
              }
            }
            // Assembly (loop on shape function)
            for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<3;l++)
              {
                m(j+indexField*nbFFm) += ( k0mJumpN(l)*Gradsm[j+indexField*nbFFm][l])*(wJ*getConstitutiveExtraDofDiffusionEqRatio()/2.);
              }
            }
            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<3;l++)
              {
                m(j+indexField*nbFFp+nbdofm) += ( k0pJumpN(l)*Gradsp[j+indexField*nbFFp][l])*(wJ*getConstitutiveExtraDofDiffusionEqRatio()/2.);
              }
            }
            //new
            if(extraDOFField==0){
              double gamma=0.;
              for(int j=0;j<nbFFm;j++){
                m(j+indexField*nbFFm) += gamma*( Stiff_alphadialitationm_JumpN*Valsm[j+indexField*nbFFm])*(wJ/2.)*(-1.);
              }
              for(int j=0;j<nbFFp;j++){
                m(j+indexField*nbFFp+nbdofm) += gamma*( Stiff_alphadialitationp_JumpN*Valsp[j+indexField*nbFFp])*(wJ/2.)*(-1.);
              }
            }
          }
         }
        }
      }
      else{ // broken case
        const FractureCohesive3DIPVariable *ipvmf = static_cast<const FractureCohesive3DIPVariable*>(ipvm);
        const FractureCohesive3DIPVariable *ipvpf = static_cast<const FractureCohesive3DIPVariable*>(ipvp);

        static SVector3 interfaceForce;
        interfaceForce = ipvmf->getConstRefToInterfaceForce();
        interfaceForce += ipvpf->getConstRefToInterfaceForce();
        interfaceForce*= 0.5;

        const std::vector<TensorialTraits<double>::ValType> &Valsm = ipvm->f(_minusSpace,ie->getElem(0),GPm[i]);
        const std::vector<TensorialTraits<double>::ValType> &Valsp = ipvp->f(_plusSpace,ie->getElem(1),GPp[i]);

        // Assembly consistency + stability (if not broken)
        for(int j=0;j<nbFFm;j++){
          for(int k=0;k<3;k++){
            m(j+k*nbFFm) -= interfaceForce[k]*(Valsm[j+0*nbFFm]*wJ);
          }
        }
        for(int j=0;j<nbFFp;j++){
          for(int k=0;k<3;k++){
            m(j+k*nbFFp+nbdofm) += interfaceForce[k]*(Valsp[j+0*nbFFp]*wJ);
          }
        }
      }
    }
  }
//m.print("Force inter");
}

void dG3DStiffnessInter::get(MElement *ele,int npts,IntPt *GP,fullMatrix<double> &stiff) const
{
  MInterfaceElement *ie = dynamic_cast<MInterfaceElement*>(ele);
  const int nbFFm = ie->getElem(0)->getNumVertices();
  const int nbFFp = ie->getElem(1)->getNumVertices();
  const int nbdofm = _minusSpace->getNumKeys(ie->getElem(0));
  const int nbdofp = _plusSpace->getNumKeys(ie->getElem(1));
  stiff.resize(nbdofm+nbdofp,nbdofm+nbdofp);
  stiff.setAll(0.);
  double jac[3][3];
  if(_fullDg)
  {
    // characteristic size
    const double hs = ie->getCharacteristicSize();
    const double b1hs = getStabilityParameter()/hs;

    // get value at gauss's point
    const AllIPState::ipstateElementContainer *vips = _ipf->getAips()->getIPstate(ele->getNum());
    // Values on minus and plus elements
    IntPt *GPm; IntPt *GPp;
    _interQuad->getIntPoints(ie,GP,&GPm,&GPp);
    // loop on Gauss Point


    static std::vector<SVector3> dBarJdum,dBarJdup;
    static std::vector< std::vector< STensor33 > > dBarFdum,dBarFdup; //for each GP and for each node

    const dG3DMaterialLaw *mlawbulkm;
    const dG3DMaterialLaw *mlawbulkp;
    if(_mlawMinus->getType() == materialLaw::fracture)
      mlawbulkm = static_cast<const dG3DMaterialLaw* >((static_cast<const FractureByCohesive3DLaw* > (_mlawMinus) )->getBulkLaw());
    else
      mlawbulkm= static_cast<const dG3DMaterialLaw*>(_mlawMinus);
    if(_mlawPlus->getType() == materialLaw::fracture)
      mlawbulkp = static_cast<const dG3DMaterialLaw* >((static_cast<const FractureByCohesive3DLaw* > (_mlawPlus) )->getBulkLaw());
    else
      mlawbulkp= static_cast<const dG3DMaterialLaw*>(_mlawPlus);
    if(mlawbulkm->getUseBarF()!=mlawbulkp->getUseBarF())
      Msg::Error("dG3DStiffnessInter::get both law should use barF or not");
    bool useBarF=mlawbulkm->getUseBarF();
    if(useBarF)
    {
      dBarJdum.resize(nbFFm);
      dBarFdum.resize(npts);
      dBarJdup.resize(nbFFp);
      dBarFdup.resize(npts);
      for (int k=0;k<nbFFm;k++)
      {
        STensorOperation::zero(dBarJdum[k]); //check here if ok
      }
      for (int k=0;k<nbFFp;k++)
      {
        STensorOperation::zero(dBarJdup[k]); //check here if ok
      }
      for (int ipg = 0; ipg < npts; ipg++)
      {
        dBarFdum[ipg].resize(nbFFm);
        dBarFdup[ipg].resize(nbFFp);
        for (int k=0;k<nbFFm;k++)
        {
          STensorOperation::zero(dBarFdum[ipg][k]); //check here if ok
        }
        for (int k=0;k<nbFFp;k++)
        {
          STensorOperation::zero(dBarFdup[ipg][k]); //check here if ok
        }
      }
      double totalWeight=0.;
      for(int ipg = 0; ipg < npts; ipg++)
        totalWeight+=GP[ipg].weight;
      for(int ipg = 0; ipg < npts; ipg++)
      {
        const IPStateBase *ipsm        = (*vips)[ipg];
        const dG3DIPVariableBase *ipvm     = static_cast<const dG3DIPVariableBase*>(ipsm->getState(IPStateBase::current));
        const IPStateBase *ipsp        = (*vips)[ipg+npts];
        const dG3DIPVariableBase *ipvp     = static_cast<const dG3DIPVariableBase*>(ipsp->getState(IPStateBase::current));
        const FractureCohesive3DIPVariable *ipvwfm;
        const FractureCohesive3DIPVariable *ipvwfp;
        const STensor3 *barFm;
        const STensor3 *barFp;
        if(_mlawMinus->getType() == materialLaw::fracture )
        {
          ipvwfm = static_cast<const FractureCohesive3DIPVariable*>(ipvm);
          barFm=&ipvwfm->getBulkDeformationGradient(); //this is actually barF
          ipvwfp = static_cast<const FractureCohesive3DIPVariable*>(ipvp);
          barFp=&ipvwfp->getBulkDeformationGradient(); //this is actually barF
        }
        else
        {
          barFm=&ipvm->getConstRefToDeformationGradient(); //this is actually barF
          barFp=&ipvp->getConstRefToDeformationGradient(); //this is actually barF
        }
        const std::vector<TensorialTraits<double>::GradType> &Gradsm = ipvm->gradf(_minusSpace,ie->getElem(0),GPm[ipg]);
        const std::vector<TensorialTraits<double>::GradType> &Gradsp = ipvp->gradf(_plusSpace,ie->getElem(1),GPp[ipg]);

        STensor3 invBarFm(barFm->invert());
        double localJm=ipvm->getLocalJ();
        double barJm=ipvm->getBarJ();
        double weightm=GP[ipg].weight*pow(localJm,2./3.)*pow(barJm,1./3.)/totalWeight;

        STensor3 invBarFp(barFp->invert());
        double localJp=ipvp->getLocalJ();
        double barJp=ipvp->getBarJ();
        double weightp=GP[ipg].weight*pow(localJp,2./3.)*pow(barJp,1./3.)/totalWeight;

        for(int k=0;k<nbFFm;k++)
        {
          SVector3 dBarJduTmp;
          STensorOperation::zero(dBarJduTmp);
          for(int ll=0; ll<3; ll++)
          {
            for(int n=0; n<3; n++)
            {
              dBarJduTmp(ll) += invBarFm(n,ll)*Gradsm[k][n]*weightm;
            }
          }
          dBarJdum[k]+=dBarJduTmp; 
        }
        for(int k=0;k<nbFFp;k++)
        {
          SVector3 dBarJduTmp;
          STensorOperation::zero(dBarJduTmp);
          for(int ll=0; ll<3; ll++)
          {
            for(int n=0; n<3; n++)
            {
              dBarJduTmp(ll) += invBarFp(n,ll)*Gradsp[k][n]*weightp;
            }
          }
          dBarJdup[k]+=dBarJduTmp; 
        }
      }
      
      static STensor3 I;
      STensorOperation::unity(I);
      for (int ipg = 0; ipg < npts; ipg++)
      {
        const IPStateBase *ipsm        = (*vips)[ipg];
        const dG3DIPVariableBase *ipvm     = static_cast<const dG3DIPVariableBase*>(ipsm->getState(IPStateBase::current));
        const IPStateBase *ipsp        = (*vips)[ipg+npts];
        const dG3DIPVariableBase *ipvp     = static_cast<const dG3DIPVariableBase*>(ipsp->getState(IPStateBase::current));
        const FractureCohesive3DIPVariable *ipvwfm;
        const FractureCohesive3DIPVariable *ipvwfp;
        const STensor3 *barFm;
        const STensor3 *barFp;
        if(_mlawMinus->getType() == materialLaw::fracture )
        {
          ipvwfm = static_cast<const FractureCohesive3DIPVariable*>(ipvm);
          barFm=&ipvwfm->getBulkDeformationGradient(); //this is actually barF
          ipvwfp = static_cast<const FractureCohesive3DIPVariable*>(ipvp);
          barFp=&ipvwfp->getBulkDeformationGradient(); //this is actually barF
        }
        else
        {
          barFm=&ipvm->getConstRefToDeformationGradient(); //this is actually barF
          barFp=&ipvp->getConstRefToDeformationGradient(); //this is actually barF
        }

        const std::vector<TensorialTraits<double>::GradType> &Gradsm = ipvm->gradf(_minusSpace,ie->getElem(0),GPm[ipg]);
        const std::vector<TensorialTraits<double>::GradType> &Gradsp = ipvp->gradf(_plusSpace,ie->getElem(1),GPp[ipg]);

        STensor3 invBarFm(barFm->invert());
        double localJm=ipvm->getLocalJ();
        double barJm=ipvm->getBarJ();
        double weightm=GP[ipg].weight*pow(localJm,2./3.)*pow(barJm,1./3.)/totalWeight;

        STensor3 invBarFp(barFp->invert());
        double localJp=ipvp->getLocalJ();
        double barJp=ipvp->getBarJ();
        double weightp=GP[ipg].weight*pow(localJp,2./3.)*pow(barJp,1./3.)/totalWeight;

        for(int k=0;k<nbFFm;k++)
        {
          static STensor33 dBarFduTmp;
          STensorOperation::zero(dBarFduTmp);
          for(int kk=0; kk<3; kk++)
          {
            for(int m=0; m<3; m++)
            {
              for(int jj=0; jj<3; jj++)
              {
                dBarFduTmp(kk,m,jj) += pow(barJm,1./3.)*pow(localJm,-1./3.)*I(kk,jj)*Gradsm[k](m);
                for(int p=0; p<3; p++)
                {
                  dBarFduTmp(kk,m,jj) -= 1./3.*pow(barJm,1./3.)*pow(localJm,-1./3.)*
                                                        barFm->operator()(kk,m)*invBarFm(p,jj)*Gradsm[k](p);             
                }
                dBarFduTmp(kk,m,jj) += 1./3./barJm*barFm->operator()(kk,m)*dBarJdum[k](jj);
              }                                
            }
          }
          dBarFdum[ipg][k]=dBarFduTmp;
        }
        for(int k=0;k<nbFFp;k++)
        {
          static STensor33 dBarFduTmp;
          STensorOperation::zero(dBarFduTmp);
          for(int kk=0; kk<3; kk++)
          {
            for(int m=0; m<3; m++)
            {
              for(int jj=0; jj<3; jj++)
              {
                dBarFduTmp(kk,m,jj) += pow(barJp,1./3.)*pow(localJp,-1./3.)*I(kk,jj)*Gradsp[k](m);
                for(int p=0; p<3; p++)
                {
                  dBarFduTmp(kk,m,jj) -= 1./3.*pow(barJp,1./3.)*pow(localJp,-1./3.)*
                                                        barFp->operator()(kk,m)*invBarFp(p,jj)*Gradsp[k](p);             
                }
                dBarFduTmp(kk,m,jj) += 1./3./barJp*barFp->operator()(kk,m)*dBarJdup[k](jj);
              }                                
            }
          }
          dBarFdup[ipg][k]=dBarFduTmp;
        }
      }
    }

    for(int i=0;i<npts; i++)
    {
      // IP
      const IPStateBase *ipsm        = (*vips)[i];
      const dG3DIPVariableBase *ipvm     = static_cast<const dG3DIPVariableBase*>(ipsm->getState(IPStateBase::current));
      const dG3DIPVariableBase *ipvmprev = static_cast<const dG3DIPVariableBase*>(ipsm->getState(IPStateBase::previous));
      const dG3DIPVariableBase *ipvm0    = static_cast<const dG3DIPVariableBase*>(ipsm->getState(IPStateBase::initial));

      const SVector3 &referenceOutwardNormalm = ipvm->getConstRefToReferenceOutwardNormal();

      const IPStateBase *ipsp        = (*vips)[i+npts];
      const dG3DIPVariableBase *ipvp     = static_cast<const dG3DIPVariableBase*>(ipsp->getState(IPStateBase::current));
      const dG3DIPVariableBase *ipvpprev = static_cast<const dG3DIPVariableBase*>(ipsp->getState(IPStateBase::previous));
      const dG3DIPVariableBase *ipvp0    = static_cast<const dG3DIPVariableBase*>(ipsp->getState(IPStateBase::initial));

      const SVector3 &referenceOutwardNormalp = ipvp->getConstRefToReferenceOutwardNormal();
      const SVector3 &jump = ipvp->getConstRefToJump();

      // Weight of Gauss' point i and Jacobian's value at this point
      const double weight = GP[i].weight;
      const double& detJ = ipvm->getJacobianDeterminant(ie,GP[i]);
      const double wJ = detJ* weight;

      const std::vector<TensorialTraits<double>::GradType> &Gradsm = ipvm->gradf(_minusSpace,ie->getElem(0),GPm[i]);
      const std::vector<TensorialTraits<double>::GradType> &Gradsp = ipvp->gradf(_plusSpace,ie->getElem(1),GPp[i]);


      // possibility of fracture
      bool broken = false;
//    bool prevbroken = false;
      if(_mlawMinus->getType() == materialLaw::fracture )
      {
        const FractureCohesive3DIPVariable *ipvwf = static_cast<const FractureCohesive3DIPVariable*>(ipvm); // broken via minus (OK broken on both sides)
        broken = ipvwf->isbroken();
      }

      static SVector3 nu;
      nu = referenceOutwardNormalm;
      nu.normalize();

      const std::vector<TensorialTraits<double>::ValType> &Valsm = ipvm->f(_minusSpace,ie->getElem(0),GPm[i]);
      const std::vector<TensorialTraits<double>::ValType> &Valsp = ipvp->f(_plusSpace,ie->getElem(1),GPp[i]);

      if(broken)
      {
        const FractureCohesive3DIPVariable *ipvmf = static_cast<const FractureCohesive3DIPVariable*>(ipvm);
        const FractureCohesive3DIPVariable *ipvpf = static_cast<const FractureCohesive3DIPVariable*>(ipvp);

        static STensor3 DInterfaceForceDjump;
        DInterfaceForceDjump = ipvmf->getConstRefToDInterfaceForceDjump();
        DInterfaceForceDjump += ipvpf->getConstRefToDInterfaceForceDjump();
        DInterfaceForceDjump *= 0.5;

        // Assembly (if broken)
        for(int j=0;j<nbFFm;j++)
        {
          for(int k=0;k<3;k++)
          {
            for(int l=0;l<nbFFm;l++)
            {
              for(int m=0;m<3;m++)
              {
                stiff(j+k*nbFFm,l+m*nbFFm) += DInterfaceForceDjump(k,m)*(Valsm[j+0*nbFFm]*Valsm[l+0*nbFFm]*wJ);
              }
            }
            for(int l=0;l<nbFFp;l++)
            {
              for(int m=0;m<3;m++)
              {
                stiff(j+k*nbFFm,l+m*nbFFp+nbdofm) -= DInterfaceForceDjump(k,m)*(Valsm[j+0*nbFFm]*Valsp[l+0*nbFFp]*wJ);
              }
            }
          }
        }
        for(int j=0;j<nbFFp;j++)
        {
          for(int k=0;k<3;k++)
          {
            for(int l=0;l<nbFFm;l++)
            {
              for(int m=0;m<3;m++)
              {
                stiff(j+k*nbFFp+nbdofm,l+m*nbFFm) -= DInterfaceForceDjump(k,m)*(Valsp[j+0*nbFFp]*Valsm[l+0*nbFFm]*wJ);
              }
            }
            for(int l=0;l<nbFFp;l++)
            {
                for(int m=0;m<3;m++)
              {
                stiff(j+k*nbFFp+nbdofm,l+m*nbFFp+nbdofm) += DInterfaceForceDjump(k,m)*(Valsp[j+0*nbFFp]*Valsp[l+0*nbFFp]*wJ);
              }
            }
          }
        }
        
        if (ipvmf->withDeformationGradient() and ipvpf->withDeformationGradient()){
          const STensor33 &DInterfaceForceDFm = ipvmf->getConstRefToDInterfaceForceDDeformationGradient();
          const STensor33 &DInterfaceForceDFp = ipvpf->getConstRefToDInterfaceForceDDeformationGradient();

          for(int j=0;j<nbFFm;j++){
            for(int k=0;k<3;k++){
              for(int l=0;l<nbFFm;l++){
                for(int m=0;m<3;m++)
                {
                  if(useBarF)
                  {
                    for(int ii=0; ii<3; ii++)
                    {
                      for(int p=0; p<3; p++)
                      {
                        stiff(j+k*nbFFm,l+m*nbFFm) -= DInterfaceForceDFm(k,ii,p)*dBarFdum[i][l](ii,p,m)*Valsm[j+0*nbFFm]*wJ*0.5;
                      }
                    }
                  }
                  else
                  {
                    for (int n=0; n<3; n++){
                      stiff(j+k*nbFFm,l+m*nbFFm) -= DInterfaceForceDFm(k,m,n)*Gradsm[l+0*nbFFm](n)*Valsm[j+0*nbFFm]*wJ*0.5;
                    }
                  }
                }
              }
              for(int l=0;l<nbFFp;l++){
                for(int m=0;m<3;m++){
                  if(useBarF)
                  {
                    for(int ii=0; ii<3; ii++)
                    {
                      for(int p=0; p<3; p++)
                      { 
                        stiff(j+k*nbFFm,l+m*nbFFp+nbdofm) -= DInterfaceForceDFp(k,ii,p)*dBarFdup[i][l](ii,p,m)*Valsm[j+0*nbFFm]*wJ*0.5;
                      }
                    }
                  }
                  else
                  {
                    for (int n=0; n<3; n++){
                      stiff(j+k*nbFFm,l+m*nbFFp+nbdofm) -= DInterfaceForceDFp(k,m,n)*Gradsp[l+0*nbFFp](n)*Valsm[j+0*nbFFm]*wJ*0.5;
                    }
                  }
                }
              }
            }
          }

          for(int j=0;j<nbFFp;j++){
            for(int k=0;k<3;k++){
              for(int l=0;l<nbFFm;l++){
                for(int m=0;m<3;m++){
                  if(useBarF)
                  {
                    for(int ii=0; ii<3; ii++)
                    {
                      for(int p=0; p<3; p++)
                      {
                        stiff(j+k*nbFFp+nbdofm,l+m*nbFFm) += DInterfaceForceDFm(k,ii,p)*dBarFdum[i][l](ii,p,m)*Valsp[j+0*nbFFp]*wJ*0.5;
                      }
                    }
                  }
                  else
                  {
                    for (int n=0; n<3; n++){
                      stiff(j+k*nbFFp+nbdofm,l+m*nbFFm) += DInterfaceForceDFm(k,m,n)*Gradsm[l+0*nbFFm](n)*Valsp[j+0*nbFFp]*wJ*0.5;
                    }
                  }
                }
              }
              for(int l=0;l<nbFFp;l++){
                for(int m=0;m<3;m++){
                  if(useBarF)
                  {
                    for(int ii=0; ii<3; ii++)
                    {
                      for(int p=0; p<3; p++)
                      { 
                        stiff(j+k*nbFFp+nbdofm,l+m*nbFFp+nbdofm) += DInterfaceForceDFp(k,ii,p)*
                                                                   dBarFdup[i][l](ii,p,m)*Valsp[j+0*nbFFp]*wJ*0.5;
                      }
                    }
                  }
                  else
                  {
                    for (int n=0; n<3; n++){
                      stiff(j+k*nbFFp+nbdofm,l+m*nbFFp+nbdofm) += DInterfaceForceDFp(k,m,n)*Gradsp[l+0*nbFFp](n)*Valsp[j+0*nbFFp]*wJ*0.5;
                    }
                  }
                }
              }
            }
          }
        }
        
        if (ipvmf->withJumpGradient() and ipvpf->withJumpGradient()){

          //std::vector<TensorialTraits<double>::GradType> gradValInt;
          //_minusSpace->gradf(ele,GP[i].pt[0],GP[i].pt[1],GP[i].pt[2],gradValInt);
          std::vector<TensorialTraits<double>::GradType>& gradValInt = ipvm->gradf(_minusSpace,ie,GP[i]);

          int nbvertexInter = ele->getNumVertices();
          std::vector<int> vn;
          vn.resize(nbvertexInter);
          ie->getLocalVertexNum(0,vn);
          std::vector<int> vp;
          vp.resize(nbvertexInter);
          ie->getLocalVertexNum(1,vp);

          const STensor33 &DInterfaceForceDjumpGradm = ipvmf->getConstRefToDInterfaceForceDjumpGradient();
          const STensor33 &DInterfaceForceDjumpGradp = ipvpf->getConstRefToDInterfaceForceDjumpGradient();

          for(int j=0;j<nbFFm;j++){
            for(int k=0;k<3;k++){
              for(int l=0;l<nbvertexInter;l++){
                for(int m=0;m<3;m++){
                  for (int n=0; n<3; n++){
                    stiff(j+k*nbFFm,vn[l]+m*nbFFm) += 0.5*(DInterfaceForceDjumpGradm(k,m,n)+DInterfaceForceDjumpGradp(k,m,n))*gradValInt[l+m*nbvertexInter](n)*Valsm[j+0*nbFFm]*wJ;
                  }
                }
              }
              for(int l=0;l<nbvertexInter;l++){
                for(int m=0;m<3;m++){
                  for (int n=0; n<3; n++){
                    stiff(j+k*nbFFm,vp[l]+m*nbFFp+nbdofm) -= 0.5*(DInterfaceForceDjumpGradm(k,m,n)+DInterfaceForceDjumpGradp(k,m,n))*gradValInt[l+m*nbvertexInter](n)*Valsm[j+0*nbFFm]*wJ;
                  }
                }
              }
            }
          }

          for(int j=0;j<nbFFp;j++){
            for(int k=0;k<3;k++){
              for(int l=0;l<nbvertexInter;l++){
                for(int m=0;m<3;m++){
                  for (int n=0; n<3; n++){
                    stiff(j+k*nbFFp+nbdofm,vn[l]+m*nbFFm) -= 0.5*(DInterfaceForceDjumpGradm(k,m,n)+DInterfaceForceDjumpGradp(k,m,n))*gradValInt[l+m*nbvertexInter](n)*Valsp[j+0*nbFFp]*wJ;
                  }
                }
              }
              for(int l=0;l<nbvertexInter;l++){
                for(int m=0;m<3;m++){
                  for (int n=0; n<3; n++){
                    stiff(j+k*nbFFp+nbdofm,vp[l]+m*nbFFp+nbdofm) += 0.5*(DInterfaceForceDjumpGradm(k,m,n)+DInterfaceForceDjumpGradp(k,m,n))*gradValInt[l+m*nbvertexInter](n)*Valsp[j+0*nbFFp]*wJ;
                  }
                }
              }
            }
          }
        }
        // Here we have extra dof part
        if( ipvm->getNumConstitutiveExtraDofDiffusionVariable()>getNumConstitutiveExtraDofDiffusionVariable() ) 
            Msg::Error("Your material law uses more constitutive extra dof variable than your domain dG3DStiffnessInter::get");
        if( ipvp->getNumConstitutiveExtraDofDiffusionVariable()>getNumConstitutiveExtraDofDiffusionVariable() ) 
            Msg::Error("Your material law uses more constitutive extra dof variable than your domain dG3DStiffnessInter::get");
        if( ipvp->getNumConstitutiveExtraDofDiffusionVariable()!= ipvm->getNumConstitutiveExtraDofDiffusionVariable() ) 
            Msg::Error("Your plus and minus material laws do not use the same number of constitutive extra dof variables dG3DStiffnessInter::get");
        for (int extraDOFField = 0; extraDOFField< ipvp->getNumConstitutiveExtraDofDiffusionVariable(); extraDOFField++){
            Msg::Error("Need to implement the stiffness matrix related to the CZM/CBM and constitutiveExtraDOFVariable dG3DStiffnessInter::get");
        }
      }
      else
      {
        // consistency D\mean{P}DF
        const STensor43 &Hp = ipvp->getConstRefToTangentModuli();
        const STensor43 &Hm = ipvm->getConstRefToTangentModuli();
        const STensor43 &eHp = ipvp->getConstRefToDGElasticTangentModuli();
        const STensor43 &eHm = ipvm->getConstRefToDGElasticTangentModuli();
        // Assembly consistency
        for(int j=0;j<nbFFm;j++)
        {
          for(int k=0;k<3;k++)
          {
            for(int l=0;l<nbFFm;l++)
            {
              for(int m=0;m<3;m++)
              {
                for(int q = 0; q< 3; q++)
                {
                  if(useBarF)
                  {
                    for(int ii=0; ii<3; ii++)
                    {
                      for(int p=0; p<3; p++)
                      {
                        stiff(j+k*nbFFm,l+m*nbFFm) -= Hm(k,q,ii,p)*(dBarFdum[i][l](ii,p,m)*Valsm[j+0*nbFFm]*wJ/2.*nu(q));
                      }
                    }
                  }
                  else
                  {
                    for(int p = 0; p< 3; p++)
                    {
                      stiff(j+k*nbFFm,l+m*nbFFm) -= Hm(k,q,m,p)*(Gradsm[l+0*nbFFm][p]*Valsm[j+0*nbFFm]*wJ/2.*nu(q));
                    }
                  }
                }
              }
            }
            for(int l=0;l<nbFFp;l++)
            {
              for(int m=0;m<3;m++)
              {
                for(int q = 0; q< 3; q++)
                {
                  if(useBarF)
                  {
                    for(int ii=0; ii<3; ii++)
                    {
                      for(int p=0; p<3; p++)
                      {
                        stiff(j+k*nbFFm,l+m*nbFFp+nbdofm) -= Hp(k,q,ii,p)*(dBarFdup[i][l](ii,p,m)*Valsm[j+0*nbFFm]*wJ/2.*nu(q));
                      }
                    }
                  }
                  else
                  {
                    for(int p = 0; p< 3; p++)
                    {
                      stiff(j+k*nbFFm,l+m*nbFFp+nbdofm) -= Hp(k,q,m,p)*(Gradsp[l+0*nbFFp][p]*Valsm[j+0*nbFFm]*wJ/2.*nu(q));
                    }
                  }
                }
              }
            }
          }
        }
        for(int j=0;j<nbFFp;j++)
        {
          for(int k=0;k<3;k++)
          {
            for(int l=0;l<nbFFm;l++)
            {
              for(int m=0;m<3;m++)
              {
                for(int q = 0; q< 3; q++)
                {
                  if(useBarF)
                  {
                    for(int ii=0; ii<3; ii++)
                    {
                      for(int p=0; p<3; p++)
                      {
                        stiff(j+k*nbFFp+nbdofm,l+m*nbFFm) += Hm(k,q,ii,p)*(dBarFdum[i][l](ii,p,m)*Valsp[j+0*nbFFp]*wJ/2.*nu(q));
                      }
                    }
                  }
                  else
                  {
                    for(int p = 0; p< 3; p++)
                    {
                      stiff(j+k*nbFFp+nbdofm,l+m*nbFFm) += Hm(k,q,m,p)*(Gradsm[l+0*nbFFm][p]*Valsp[j+0*nbFFp]*wJ/2.*nu(q));
                    }
                  }
                }
              }
            }
            for(int l=0;l<nbFFp;l++)
            {
              for(int m=0;m<3;m++)
              {
                for(int q = 0; q< 3; q++)
                {
                  if(useBarF)
                  {
                    for(int ii=0; ii<3; ii++)
                    {
                      for(int p=0; p<3; p++)
                      {
                        stiff(j+k*nbFFp+nbdofm,l+m*nbFFp+nbdofm) += Hp(k,q,ii,p)*(dBarFdup[i][l](ii,p,m)*Valsp[j+0*nbFFp]*wJ/2.*nu(q));
                      }
                    }
                  }
                  else
                  {
                    for(int p = 0; p< 3; p++)
                    {
                      stiff(j+k*nbFFp+nbdofm,l+m*nbFFp+nbdofm) += Hp(k,q,m,p)*(Gradsp[l+0*nbFFp][p]*Valsp[j+0*nbFFp]*wJ/2.*nu(q));
                    }
                  }
                }
              }
            }
          }
        }

        // Stability  N \mean{H} (N)
        static STensor3 NMeanHNBetasc;
        for(int k = 0; k <3; k++)
        {
          for(int m = 0; m <3; m++)
          {
            NMeanHNBetasc(k,m) = 0.;
            for(int l = 0; l <3; l++)
            {
              for(int n = 0; n <3; n++)
              {
                NMeanHNBetasc(k,m) += (eHp(k,l,m,n)+eHm(k,l,m,n))*nu(n)*b1hs*nu(l)/2.;
              }
            }
          }
        }
        // Assembly stability (if not broken)
        for(int j=0;j<nbFFm;j++)
        {
          for(int k=0;k<3;k++)
          {
            for(int l=0;l<nbFFm;l++)
            {
              for(int m=0;m<3;m++)
              {
                stiff(j+k*nbFFm,l+m*nbFFm) += NMeanHNBetasc(k,m)*(Valsm[j+0*nbFFm]*Valsm[l+0*nbFFm]*wJ);
              }
            }
            for(int l=0;l<nbFFp;l++)
            {
              for(int m=0;m<3;m++)
              {
                stiff(j+k*nbFFm,l+m*nbFFp+nbdofm) -= NMeanHNBetasc(k,m)*(Valsm[j+0*nbFFm]*Valsp[l+0*nbFFp]*wJ);
              }
            }
          }
        }
        for(int j=0;j<nbFFp;j++)
        {
          for(int k=0;k<3;k++)
          {
            for(int l=0;l<nbFFm;l++)
            {
              for(int m=0;m<3;m++)
              {
                stiff(j+k*nbFFp+nbdofm,l+m*nbFFm) -= NMeanHNBetasc(k,m)*(Valsp[j+0*nbFFp]*Valsm[l+0*nbFFm]*wJ);
              }
            }
            for(int l=0;l<nbFFp;l++)
            {
              for(int m=0;m<3;m++)
              {
                stiff(j+k*nbFFp+nbdofm,l+m*nbFFp+nbdofm) += NMeanHNBetasc(k,m)*(Valsp[j+0*nbFFp]*Valsp[l+0*nbFFp]*wJ);
              }
            }
          }
        }
        // compatibility membrane (4 terms to assembly)
        static STensor3 HmJumpN;
        static STensor3 HpJumpN;

        for(int k = 0; k <3; k++)
        {
          for(int l = 0; l <3; l++)
          {
            HmJumpN(k,l) = 0.;
            HpJumpN(k,l) = 0.;
            for(int m = 0; m <3; m++)
            {
              for(int n = 0; n <3; n++)
              {
                HmJumpN(k,l) += eHm(m,n,k,l)*jump(m)*nu(n);
                HpJumpN(k,l) += eHp(m,n,k,l)*jump(m)*nu(n);
              }
            }
          }
        }
        // Assembly (loop on shape function)
        for(int j=0;j<nbFFm;j++)
        {
          for(int k=0;k<3;k++)
          {
            for(int l=0;l<nbFFm;l++)
            {
              for(int m=0;m<3;m++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+k*nbFFm,l+m*nbFFm) -= (eHm(m,p,k,q)*Gradsm[j+0*nbFFm][q])*(nu(p)*wJ/2.*Valsm[l+0*nbFFm]);
                  }
                }
              }
            }
            for(int l=0;l<nbFFp;l++)
            {
              for(int m=0;m<3;m++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+k*nbFFm,l+m*nbFFp+nbdofm) += (eHm(m,p,k,q)*Gradsm[j+0*nbFFm][q])*(nu(p)*wJ/2.*Valsp[l+0*nbFFp]);
                  }
                }
              }
            }
          }
        }
        for(int j=0;j<nbFFp;j++)
        {
          for(int k=0;k<3;k++)
          {
            for(int l=0;l<nbFFm;l++)
            {
              for(int m=0;m<3;m++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+k*nbFFp+nbdofm,l+m*nbFFm) -= (eHp(m,p,k,q)*Gradsp[j+0*nbFFp][q])*(nu(p)*wJ/2.*Valsm[l+0*nbFFm]);
                  }
                }
              }
            }
            for(int l=0;l<nbFFp;l++)
            {
              for(int m=0;m<3;m++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+k*nbFFp+nbdofm,l+m*nbFFp+nbdofm) += (eHp(m,p,k,q)*Gradsp[j+0*nbFFp][q])*(nu(p)*wJ/2.*Valsp[l+0*nbFFp]);
                  }
                }
              }
            }
          }
        }
        //non local
        if( ipvm->getNumberNonLocalVariable()>getNumNonLocalVariable() ) Msg::Error("Your minus material law uses more non local variables than your domain dG3DStiffnessInter::get");
        if( ipvp->getNumberNonLocalVariable()>getNumNonLocalVariable() ) Msg::Error("Your plus material law uses more non local variables than your domain dG3DStiffnessInter::get");
        if( getNonLocalContinuity() && !broken)
        {
//
          const double nbetahs = getNonLocalStabilityParameter()/hs;
          for (int nlk = 0; nlk < ipvp->getNumberNonLocalVariable(); nlk++)
          {
            double epljump = ipvp->getConstRefToNonLocalJump()(nlk);
            static SVector3  gradm, gradp; 
            gradm = ipvm->getConstRefToGradNonLocalVariable()[nlk];
            gradp = ipvp->getConstRefToGradNonLocalVariable()[nlk];
            if (_incrementNonlocalBased){
              epljump -= ipvpprev->getConstRefToNonLocalJump()(nlk);
              gradm -= ipvmprev->getConstRefToGradNonLocalVariable()[nlk];
              gradp -= ipvpprev->getConstRefToGradNonLocalVariable()[nlk];
            }
            
            
            const STensor3  *cgm = &(ipvm->getConstRefToCharacteristicLengthMatrix(nlk));           // by Wu Ling
            const STensor3  *cgp = &(ipvp->getConstRefToCharacteristicLengthMatrix(nlk));           // by Wu Ling
            
            const STensor3  *dsdpp = &(ipvp->getConstRefToDStressDNonLocalVariable()[nlk]);
            const STensor3  *dsdpm = &(ipvm->getConstRefToDStressDNonLocalVariable()[nlk]);
            
            

            // Assembly consistency (from derivative of sigma with tilde{p}
            for(int j=0;j<nbFFm;j++)
            {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+k*nbFFm,l+nlk*nbFFm+3*nbFFm) -= dsdpm->operator()(k,q)*(Valsm[l+nlk*nbFFm+3*nbFFm]*Valsm[j+k*nbFFm]*wJ/2.*nu(q));
                  }
                }
                for(int l=0;l<nbFFp;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+k*nbFFm,l+nlk*nbFFp+3*nbFFp+nbdofm) -= dsdpp->operator()(k,q)*(Valsp[l+nlk*nbFFp+3*nbFFp]*Valsm[j+k*nbFFm]*wJ/2.*nu(q));
                  }
                }
              }
            }
            for(int j=0;j<nbFFp;j++)
            {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+k*nbFFp+nbdofm,l+nlk*nbFFm+3*nbFFm) += dsdpm->operator()(k,q)*(Valsm[l+nlk*nbFFm+3*nbFFm]*
                                                                           Valsp[j+k*nbFFp]*wJ/2.*nu(q));
                  }
                }
                for(int l=0;l<nbFFp;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+k*nbFFp+nbdofm,l+nlk*nbFFp+3*nbFFp+nbdofm) += dsdpp->operator()(k,q)*(Valsp[l+nlk*nbFFp+3*nbFFp]* 
                                                                            Valsp[j+k*nbFFp]*wJ/2.*nu(q));
                  }
                }
              }
            }
     
            // consistency D\mean{cG grad epl} Depl
            // Assembly consistency
            for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p = 0; p< 3; p++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+nlk*nbFFm+3*nbFFm,l+nlk*nbFFm+3*nbFFm) -= cgm->operator()(p,q)*
                                      (Gradsm[l+nlk*nbFFm+3*nbFFm][p]*Valsm[j+nlk*nbFFm+3*nbFFm]*wJ*getNonLocalEqRatio()/2.*nu(q));
                  }
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p = 0; p< 3; p++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+nlk*nbFFm+3*nbFFm,l+nlk*nbFFp+3*nbFFp+nbdofm) -= cgp->operator()(p,q)*
                                       (Gradsp[l+nlk*nbFFp+3*nbFFp][p]*Valsm[j+nlk*nbFFm+3*nbFFm]*wJ*getNonLocalEqRatio()/2.*nu(q));
                  }
                }
              }
            }
            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p = 0; p< 3; p++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+nlk*nbFFp+3*nbFFp+nbdofm,l+nlk*nbFFm+3*nbFFm) += cgm->operator()(p,q)*
                                       (Gradsm[l+nlk*nbFFm+3*nbFFm][p]*Valsp[j+nlk*nbFFp+3*nbFFp]*wJ*getNonLocalEqRatio()/2.*nu(q));
                  }
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p = 0; p< 3; p++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+nlk*nbFFp+3*nbFFp+nbdofm,l+nlk*nbFFp+3*nbFFp+nbdofm) += cgp->operator()(p,q)*
                                       (Gradsp[l+nlk*nbFFp+3*nbFFp][p]*Valsp[j+nlk*nbFFp+3*nbFFp]*wJ*getNonLocalEqRatio()/2.*nu(q));
                  }
                }
              }
            }

            // Stability  N \mean{cg} (N)
            double NMeanCgNBetasc=0.;
            for(int l = 0; l <3; l++)
            {
              for(int n = 0; n <3; n++)
              {
                NMeanCgNBetasc += (cgp->operator()(l,n)+cgm->operator()(l,n))*nu(n)*nbetahs*nu(l)/2.;
              }
            }
            // Assembly stability (if not broken)
            for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                stiff(j+nlk*nbFFm+3*nbFFm,l+nlk*nbFFm+3*nbFFm) += NMeanCgNBetasc*
                                     (Valsm[j+nlk*nbFFm+3*nbFFm]*Valsm[l+nlk*nbFFm+3*nbFFm]*wJ*getNonLocalEqRatio());
              }
              for(int l=0;l<nbFFp;l++)
              {
                stiff(j+nlk*nbFFm+3*nbFFm,l+nlk*nbFFp+3*nbFFp+nbdofm) -= NMeanCgNBetasc*
                                     (Valsm[j+nlk*nbFFm+3*nbFFm]*Valsp[l+nlk*nbFFp+3*nbFFp]*wJ*getNonLocalEqRatio());
              }
            }
            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                stiff(j+nlk*nbFFp+3*nbFFp+nbdofm,l+nlk*nbFFm+3*nbFFm) -= NMeanCgNBetasc*
                                     (Valsp[j+nlk*nbFFp+3*nbFFp]*Valsm[l+nlk*nbFFm+3*nbFFm]*wJ*getNonLocalEqRatio());
              }
              for(int l=0;l<nbFFp;l++)
              {
                stiff(j+nlk*nbFFp+3*nbFFp+nbdofm,l+nlk*nbFFp+3*nbFFp+nbdofm) += NMeanCgNBetasc*
                                    (Valsp[j+nlk*nbFFp+3*nbFFp]*Valsp[l+nlk*nbFFp+3*nbFFp]*wJ*getNonLocalEqRatio());
              }
            }
            // Assembly (loop on shape function)
            for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+nlk*nbFFm+3*nbFFm,l+nlk*nbFFm+3*nbFFm) -= (cgm->operator()(p,q)*
                                Gradsm[j+nlk*nbFFm+3*nbFFm][q])*(nu(p)*wJ*getNonLocalEqRatio()/2.*Valsm[l+nlk*nbFFm+3*nbFFm]);
                  }
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+nlk*nbFFm+3*nbFFm,l+nlk*nbFFp+3*nbFFp+nbdofm) += (cgm->operator()(p,q)*
                                Gradsm[j+nlk*nbFFm+3*nbFFm][q])*(nu(p)*wJ*getNonLocalEqRatio()/2.*Valsp[l+nlk*nbFFp+3*nbFFp]);
                  }
                }
              }
            }
            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+nlk*nbFFp+3*nbFFp+nbdofm,l+nlk*nbFFm+3*nbFFm) -= (cgp->operator()(p,q)*
                                Gradsp[j+nlk*nbFFp+3*nbFFp][q])*(nu(p)*wJ*getNonLocalEqRatio()/2.*Valsm[l+nlk*nbFFm+3*nbFFm]);
                  }
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+nlk*nbFFp+3*nbFFp+nbdofm,l+nlk*nbFFp+3*nbFFp+nbdofm) += (cgp->operator()(p,q)*
                                Gradsp[j+nlk*nbFFp+3*nbFFp][q])*(nu(p)*wJ*getNonLocalEqRatio()/2.*Valsp[l+nlk*nbFFp+3*nbFFp]);
                  }
                }
              }
            }
            for (int extraDOFField = 0; extraDOFField < ipvm->getNumConstitutiveExtraDofDiffusionVariable(); extraDOFField++){
              static bool mess=false;
              if(!mess)
                Msg::Info("Implement non-local / extra dof coupling in dG3DStiffnessBulk::get");
              mess=true;
              //double  dLocalVariableDNonLocalVariablem = ipvm->getConstRefToDLocalVariableDNonLocalVariable()(nlk,extraDOFField);
              //double  dLocalVariableDNonLocalVariablep = ipvp->getConstRefToDLocalVariableDNonLocalVariable()(nlk,extraDOFField);
            }
          }
        }

        // constitutive extra dof
        if( ipvm->getNumConstitutiveExtraDofDiffusionVariable()>getNumConstitutiveExtraDofDiffusionVariable() ) 
            Msg::Error("Your material law uses more constitutive extra dof variable than your domain dG3DStiffnessInter::get");
        if( ipvp->getNumConstitutiveExtraDofDiffusionVariable()>getNumConstitutiveExtraDofDiffusionVariable() ) 
            Msg::Error("Your material law uses more constitutive extra dof variable than your domain dG3DStiffnessInter::get");
        if( ipvp->getNumConstitutiveExtraDofDiffusionVariable()!= ipvm->getNumConstitutiveExtraDofDiffusionVariable() ) 
            Msg::Error("Your plus and minus material laws do not use the same number of constitutive extra dof variables dG3DStiffnessInter::get");
        for (int extraDOFField = 0; extraDOFField< ipvp->getNumConstitutiveExtraDofDiffusionVariable(); extraDOFField++){

            const double nbetahs = getConstitutiveExtraDofDiffusionStabilityParameter()/hs;

            const STensor33 *dqdFm = &(ipvm->getConstRefTodFluxdF()[extraDOFField]);
            const STensor33 *dqdFp = &(ipvp->getConstRefTodFluxdF()[extraDOFField]);
            const STensor3  *dPdFieldm = &(ipvm->getConstRefTodPdField()[extraDOFField]);
            const STensor3  *dPdFieldp = &(ipvp->getConstRefTodPdField()[extraDOFField]);

            double fieldJump=0.; double fieldJump2=0.;
            double fieldp=0.;double fieldm=0.; double dfieldJumpdFieldp=0.; double dfieldJumpdFieldm=0.;
            double dfieldJumpdField2p=0.; double dfieldJumpdField2m=0.; double oneOverFieldJump=0.;
            double dFielddEnergyConjugatedFieldp=0.; double dFielddEnergyConjugatedFieldm=0.; 
            double dFielddEnergyConjugatedField2p=0.; double dFielddEnergyConjugatedField2m=0.;
            static SVector3 dGradFielddEnergyConjugatedFieldp, dGradFielddEnergyConjugatedFieldm, dGradFielddEnergyConjugatedField2p, dGradFielddEnergyConjugatedField2m;
            static STensor3 dGradFielddGradEnergyConjugatedFieldp, dGradFielddGradEnergyConjugatedFieldm, dGradFielddGradEnergyConjugatedField2p, dGradFielddGradEnergyConjugatedField2m;

            static STensor3 Stiff_alphadialitationm, Stiff_alphadialitationp, dk0mdFieldm, dk0pdFieldp, dk0mdField2m, dk0pdField2p, dk0mdField_auxm, dk0pdField_auxp, dk0mdField_aux2m, dk0pdField_aux2p, dStiff_alphadialitationmdFieldm, dStiff_alphadialitationpdFieldp, dPdField2m, dPdField2p;

            if(extraDOFField==0)
             {
              Stiff_alphadialitationm = *(ipvm->getConstRefToLinearSymmetrizationCoupling()[0]);
              Stiff_alphadialitationp = *(ipvp->getConstRefToLinearSymmetrizationCoupling()[0]);
             }
            else
             {
              STensorOperation::zero(Stiff_alphadialitationm);
              STensorOperation::zero(Stiff_alphadialitationp);
             }

            if(ipvp->getNumConstitutiveExtraDofDiffusionVariable()==1)
            {
             fieldJump = ipvp->getConstRefToFieldJump()(extraDOFField);
             fieldp = ipvp->getConstRefToField(extraDOFField);
             fieldm = ipvm->getConstRefToField(extraDOFField);
             dfieldJumpdFieldp = 1.;
             dfieldJumpdFieldm = -1.;
             oneOverFieldJump = ipvp->getConstRefToOneOverFieldJump()(extraDOFField);
             STensorOperation::zero(dk0mdFieldm);
             STensorOperation::zero(dk0pdFieldp);
             STensorOperation::zero(dStiff_alphadialitationmdFieldm);
             STensorOperation::zero(dStiff_alphadialitationpdFieldp);
              if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())
               {
                 fieldJump=ipvp->getConstRefToOneOverFieldJump()(extraDOFField);
                 Stiff_alphadialitationm*=(-1.)*fieldm*fieldm;
                 Stiff_alphadialitationp*=(-1.)*fieldp*fieldp;
                 dStiff_alphadialitationmdFieldm=Stiff_alphadialitationm;//the - is removed
                 dStiff_alphadialitationmdFieldm*=(1.)*2./fieldm;
                 dStiff_alphadialitationpdFieldp*=Stiff_alphadialitationp;//the - is removed
                 dStiff_alphadialitationpdFieldp*=(1.)*2./fieldp;
               }
            }
            else
            {
             fieldp = ipvp->getConstRefToEnergyConjugatedField(extraDOFField);
             fieldm = ipvm->getConstRefToEnergyConjugatedField(extraDOFField);
             if(extraDOFField==0)
                   {
                    Stiff_alphadialitationm*= (-1.)/(fieldm*fieldm);
                    Stiff_alphadialitationp*= (-1.)/(fieldp*fieldp); 
                   }
             if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())
                   {
                    int extraDOFField_aux=0;
                      if((extraDOFField+1)< ipvp->getNumConstitutiveExtraDofDiffusionVariable()) extraDOFField_aux=extraDOFField+1;
                      else extraDOFField_aux=extraDOFField-1;  

                    dFielddEnergyConjugatedFieldp = ipvp->getConstRefTodFielddEnergyConjugatedField()[extraDOFField][extraDOFField];
                    dFielddEnergyConjugatedFieldm = ipvm->getConstRefTodFielddEnergyConjugatedField()[extraDOFField][extraDOFField]; 
                    dFielddEnergyConjugatedField2p = ipvp->getConstRefTodFielddEnergyConjugatedField()[extraDOFField_aux][extraDOFField];
                    dFielddEnergyConjugatedField2m = ipvm->getConstRefTodFielddEnergyConjugatedField()[extraDOFField_aux][extraDOFField];  
                    dPdField2p       = ipvp->getConstRefTodPdField()[extraDOFField_aux];
                    dPdField2m       = ipvm->getConstRefTodPdField()[extraDOFField_aux]; 
                   }
            }

         int indexField=3+getNumNonLocalVariable()+extraDOFField;

         // Assembly consistency (from derivative of sigma with extra field)
          for(int j=0;j<nbFFm;j++)
            {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    if((getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())&&(ipvp->getNumConstitutiveExtraDofDiffusionVariable()==2)) 
                      stiff(j+k*nbFFm,l+indexField*nbFFm) -= (dPdFieldm->operator()(k,q)*dFielddEnergyConjugatedFieldm+dPdField2m(k,q)*
                                                                 dFielddEnergyConjugatedField2m)*(Valsm[l+indexField*nbFFm]*Valsm[j+0*nbFFm]*wJ/2.*nu(q));
                    else 
                      stiff(j+k*nbFFm,l+indexField*nbFFm) -= dPdFieldm->operator()(k,q)*(Valsm[l+indexField*nbFFm]*Valsm[j+0*nbFFm]*wJ/2.*nu(q));
                  }
                }
                for(int l=0;l<nbFFp;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    if((getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())&&(ipvp->getNumConstitutiveExtraDofDiffusionVariable()==2)) 
                      stiff(j+k*nbFFm,l+indexField*nbFFp+nbdofm) -= (dPdFieldp->operator()(k,q)*dFielddEnergyConjugatedFieldp+
                                                  dPdField2p(k,q)*dFielddEnergyConjugatedField2p)*(Valsp[l+indexField*nbFFp]*Valsm[j+0*nbFFm]*wJ/2.*nu(q));
                    else 
                      stiff(j+k*nbFFm,l+indexField*nbFFp+nbdofm) -= dPdFieldp->operator()(k,q)*(Valsp[l+indexField*nbFFp]*Valsm[j+0*nbFFm]*wJ/2.*nu(q));
                  }
                }
              }
            }
            for(int j=0;j<nbFFp;j++)
            {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    if((getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())&&(ipvp->getNumConstitutiveExtraDofDiffusionVariable()==2)) 
                      stiff(j+k*nbFFp+nbdofm,l+indexField*nbFFm) += (dPdFieldm->operator()(k,q)*dFielddEnergyConjugatedFieldm+dPdField2m(k,q)*
                                                                dFielddEnergyConjugatedField2m)*(Valsm[l+indexField*nbFFm]*Valsp[j+0*nbFFp]*wJ/2.*nu(q));
                    else 
                      stiff(j+k*nbFFp+nbdofm,l+indexField*nbFFm) += dPdFieldm->operator()(k,q)*(Valsm[l+indexField*nbFFm]*Valsp[j+0*nbFFp]*wJ/2.*nu(q));
                  }
                }
                for(int l=0;l<nbFFp;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    if((getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())&&(ipvp->getNumConstitutiveExtraDofDiffusionVariable()==2)) 
                      stiff(j+k*nbFFp+nbdofm,l+indexField*nbFFp+nbdofm) += (dPdFieldp->operator()(k,q)*dFielddEnergyConjugatedFieldp+dPdField2p(k,q)*
                                                                   dFielddEnergyConjugatedField2p)*(Valsp[l+indexField*nbFFp]*Valsp[j+0*nbFFp]*wJ/2.*nu(q));
                    else 
                      stiff(j+k*nbFFp+nbdofm,l+indexField*nbFFp+nbdofm) += dPdFieldp->operator()(k,q)*(Valsp[l+indexField*nbFFp]*Valsp[j+0*nbFFp]*wJ/2.*nu(q));
                  }
                }
              }
            }

         //F-related terms
           for(int j=0;j<nbFFm;j++)
            {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int N = 0; N< 3; N++)
                  {
                    if(useBarF)
                    {
                      for(int ii=0; ii<3; ii++)
                      {
                        for(int q=0; q<3; q++)
                        {
                          stiff(j+indexField*nbFFm,l+k*nbFFm) -= dqdFm->operator()(N,ii,q)*(dBarFdum[i][l](ii,q,k)*Valsm[j+indexField*nbFFm]*
                                                               wJ*getConstitutiveExtraDofDiffusionEqRatio()/2.*nu(N));
                        }
                      }
                    }
                    else
                    {
                      for(int q = 0; q< 3; q++)
                      {
                        stiff(j+indexField*nbFFm,l+k*nbFFm) -= dqdFm->operator()(N,k,q)*(Gradsm[l+0*nbFFm][q]*Valsm[j+indexField*nbFFm]*
                                                               wJ*getConstitutiveExtraDofDiffusionEqRatio()/2.*nu(N));
                      }
                    }
                  }
                }
                for(int l=0;l<nbFFp;l++)
                {
                  for(int N = 0; N< 3; N++)
                  {
                    if(useBarF)
                    {
                      for(int ii=0; ii<3; ii++)
                      {
                        for(int q=0; q<3; q++)
                        {
                          stiff(j+indexField*nbFFm,l+k*nbFFp+nbdofm) -= dqdFp->operator()(N,ii,q)*(dBarFdup[i][l](ii,q,k)*Valsm[j+indexField*nbFFm]*
                                                               wJ*getConstitutiveExtraDofDiffusionEqRatio()/2.*nu(N));
                        }
                      }
                    }
                    else
                    {
                      for(int q = 0; q< 3; q++)
                      {
                        stiff(j+indexField*nbFFm,l+k*nbFFp+nbdofm) -= dqdFp->operator()(N,k,q)*(Gradsp[l+0*nbFFp][q]*Valsm[j+indexField*nbFFm]*
                                                               wJ*getConstitutiveExtraDofDiffusionEqRatio()/2.*nu(N));
                      }
                    }
                  }
                }
              }
            }
            for(int j=0;j<nbFFp;j++)
            {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int N = 0; N< 3; N++)
                  {
                    if(useBarF)
                    {
                      for(int ii=0; ii<3; ii++)
                      {
                        for(int q=0; q<3; q++)
                        {
                          stiff(j+indexField*nbFFp+nbdofm,l+k*nbFFm) += dqdFm->operator()(N,ii,q)*(dBarFdum[i][l](ii,q,k)*Valsp[j+indexField*nbFFp]*
                                                                 wJ*getConstitutiveExtraDofDiffusionEqRatio()/2.*nu(N));
                        }
                      }
                    }
                    else
                    {
                      for(int q = 0; q< 3; q++)
                      {
                        stiff(j+indexField*nbFFp+nbdofm,l+k*nbFFm) += dqdFm->operator()(N,k,q)*(Gradsm[l+0*nbFFm][q]*Valsp[j+indexField*nbFFp]*
                                                               wJ*getConstitutiveExtraDofDiffusionEqRatio()/2.*nu(N));
                      }
                    }
                  }
                }
                for(int l=0;l<nbFFp;l++)
                {
                  for(int N = 0; N< 3; N++)
                  {
                    if(useBarF)
                    {
                      for(int ii=0; ii<3; ii++)
                      {
                        for(int q=0; q<3; q++)
                        {
                          stiff(j+indexField*nbFFp+nbdofm,l+k*nbFFp+nbdofm) += dqdFp->operator()(N,ii,q)*(dBarFdup[i][l](ii,q,k)*Valsp[j+indexField*nbFFp]*
                                                              wJ*getConstitutiveExtraDofDiffusionEqRatio()/2.*nu(N));
                        }
                      }
                    }
                    else
                    {
                      for(int q = 0; q< 3; q++)
                      {
                        stiff(j+indexField*nbFFp+nbdofm,l+k*nbFFp+nbdofm) += dqdFp->operator()(N,k,q)*(Gradsp[l+0*nbFFp][q]*Valsp[j+indexField*nbFFp]*
                                                              wJ*getConstitutiveExtraDofDiffusionEqRatio()/2.*nu(N));
                      }
                    }
                  }
                }
              }
            }


           if(extraDOFField==0)
           {
            double gamma=0.;
            if(ipvp->getNumConstitutiveExtraDofDiffusionVariable()==1)
            {
             for(int j=0;j<nbFFm;j++)
             {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int p =0; p < 3; p++)
                  {
                    stiff(j+indexField*nbFFm,l+k*nbFFm) -= gamma*(Stiff_alphadialitationm(k,p)*Valsm[j+indexField*nbFFm])*(nu(p)*wJ/2.*Valsm[l+0*nbFFm])*(-1.);
                  }
                }
                for(int l=0;l<nbFFp;l++)
                {
                  for(int p =0; p < 3; p++)
                  {
                    stiff(j+indexField*nbFFm,l+k*nbFFp+nbdofm) += gamma*(Stiff_alphadialitationm(k,p)*Valsm[j+indexField*nbFFm])*(nu(p)*wJ/2.*Valsp[l+0*nbFFp])*(-1.);
                   }
                 }
               }
             }
             for(int j=0;j<nbFFp;j++)
             {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int p =0; p < 3; p++)
                  {
                    stiff(j+indexField*nbFFp+nbdofm,l+k*nbFFm) -= gamma*(Stiff_alphadialitationp(k,p)*Valsp[j+indexField*nbFFp])*(nu(p)*wJ/2.*Valsm[l+0*nbFFm])*(-1.);
                  }
                }
                for(int l=0;l<nbFFp;l++)
                {
                  for(int p =0; p < 3; p++)
                  {
                    stiff(j+indexField*nbFFp+nbdofm,l+k*nbFFp+nbdofm) += gamma*(Stiff_alphadialitationp(k,p)*Valsp[j+indexField*nbFFp])*(nu(p)*wJ/2.*Valsp[l+0*nbFFp])*(-1.);
                  }
                }
              }
             }
             if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())
             {
              for(int j=0;j<nbFFm;j++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int k =0; k < 3; k++)
                  {
                    for(int p =0; p < 3; p++)
                    {
                      stiff(j+indexField*nbFFm,l+indexField*nbFFm) += gamma*jump(k)*
                          (dStiff_alphadialitationmdFieldm(k,p)*Valsm[j+indexField*nbFFm])*nu(p)*Valsm[l+indexField*nbFFm]*wJ/2.*(-1.);
                    }
                  }
                }
              }
              for(int j=0;j<nbFFp;j++)
              {
                for(int l=0;l<nbFFp;l++)
                {
                  for(int k =0; k < 3; k++)
                  {
                    for(int p =0; p< 3; p++)
                    {
                      stiff(j+indexField*nbFFp+nbdofm,l+indexField*nbFFp+nbdofm) += gamma*jump(k)*
                          (dStiff_alphadialitationpdFieldp(k,p)*Valsp[j+indexField*nbFFp])*nu(p)*Valsp[l+indexField*nbFFp]*wJ/2.*(-1.);
                     }
                   }
                 }
               }
              }
          }
          if(ipvp->getNumConstitutiveExtraDofDiffusionVariable()==2)
          {
           double gamma=0.;
           for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int k =0; k < 3; k++)
                {
                  for(int p =0; p < 3; p++)
                  {
                    stiff(j+indexField*nbFFm,l+indexField*nbFFm) += gamma*jump(k)*(Stiff_alphadialitationm(k,p)*
                                                Valsm[j+indexField*nbFFm])*(2.*fieldm)*nu(p)*Valsm[l+indexField*nbFFm]*wJ/2.*(-1.);
                  }
                }
              }
            }

            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<nbFFp;l++)
              {
                for(int k =0; k < 3; k++)
                {
                  for(int p =0; p< 3; p++)
                  {

                    stiff(j+indexField*nbFFp+nbdofm,l+indexField*nbFFp+nbdofm) += gamma*jump(k)*(Stiff_alphadialitationp(k,p)*
                                                (2.*fieldp)*Valsp[j+indexField*nbFFp])*nu(p)*Valsp[l+indexField*nbFFp]*wJ/2.*(-1.);
                  }
                }
              }
            }
            for(int j=0;j<nbFFm;j++)
            {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int p =0; p < 3; p++)
                  {
                      stiff(j+indexField*nbFFm,l+k*nbFFm) -= gamma*(Stiff_alphadialitationm(k,p)*Valsm[j+indexField*nbFFm])*(nu(p)*wJ/2.*Valsm[l+0*nbFFm])*(-1.);
                  }
                }
                for(int l=0;l<nbFFp;l++)
                {
                  for(int p =0; p < 3; p++)
                  {
                      stiff(j+indexField*nbFFm,l+k*nbFFp+nbdofm) +=  gamma*(Stiff_alphadialitationm(k,p)*Valsm[j+indexField*nbFFm])*(nu(p)*wJ/2.*Valsp[l+0*nbFFp])*(-1.);
                  }
                }
              }
            }

            for(int j=0;j<nbFFp;j++)
            {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int p =0; p < 3; p++)
                  {
                    stiff(j+indexField*nbFFp+nbdofm,l+k*nbFFm) -=  gamma*(Stiff_alphadialitationp(k,p)*Valsp[j+indexField*nbFFp])*(nu(p)*wJ/2.*Valsm[l+0*nbFFm])*(-1.);
                  }
                }
                for(int l=0;l<nbFFp;l++)
                {
                  for(int p =0; p < 3; p++)
                  {
                      stiff(j+indexField*nbFFp+nbdofm,l+k*nbFFp+nbdofm) +=  gamma*(Stiff_alphadialitationp(k,p)*Valsp[j+indexField*nbFFp])*(nu(p)*wJ/2.*Valsp[l+0*nbFFp])*(-1.);
                  }
                }
              }
            }
          }
        }


         for (int extraDOFField2 = 0; extraDOFField2< ipvp->getNumConstitutiveExtraDofDiffusionVariable(); extraDOFField2++){
          if( getConstitutiveExtraDofDiffusionContinuity() && !broken)
          {
            const STensor3  *dqdGradFieldm = &(ipvm->getConstRefTodFluxdGradField()[extraDOFField][extraDOFField2]);
            const STensor3  *dqdGradFieldp = &(ipvp->getConstRefTodFluxdGradField()[extraDOFField][extraDOFField2]);
            const SVector3  *dqdFieldm = &(ipvm->getConstRefTodFluxdField()[extraDOFField][extraDOFField2]);
            const SVector3  *dqdFieldp = &(ipvp->getConstRefTodFluxdField()[extraDOFField][extraDOFField2]);

            static STensor3  k0m, k0p, k02m, k02p, dqdGradField2m, dqdGradField2p;
            static SVector3 dqdField2m, dqdField2p;
            static STensor43 dk0mdFm, dk0pdFp;
            STensorOperation::zero(dk0mdFm);STensorOperation::zero(dk0pdFp);

            k0m = *(ipvm->getConstRefToLinearK()[extraDOFField][extraDOFField2]);
            k0p = *(ipvp->getConstRefToLinearK()[extraDOFField][extraDOFField2]);

            if(ipvp->getNumConstitutiveExtraDofDiffusionVariable()==1) //thermomec
            {
              if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())
               {
                 k0m*=(-1.)*fieldm*fieldm;
                 k0p*=(-1.)*fieldp*fieldp;
                 dk0mdFieldm=k0m;
                 dk0mdFieldm*=(1.)*2./fieldm;//the - is removed
                 dk0pdFieldp=k0p;
                 dk0pdFieldp*=(1.)*2./fieldp;//the - is removed
               }
            }
             else
            {
              int extraDOFField_aux=0;
                      if((extraDOFField2+1)< ipvp->getNumConstitutiveExtraDofDiffusionVariable()) extraDOFField_aux=extraDOFField2+1;
                      else extraDOFField_aux=extraDOFField2-1; 

              fieldJump = ipvp->getConstRefToEnergyConjugatedFieldJump()(extraDOFField2);
              fieldJump2 = ipvp->getConstRefToEnergyConjugatedFieldJump()(extraDOFField_aux);

              dk0mdFieldm  = ipvm->getConstRefTodLinearKdField()[extraDOFField][extraDOFField2][extraDOFField2];
              dk0pdFieldp  = ipvp->getConstRefTodLinearKdField()[extraDOFField][extraDOFField2][extraDOFField2];
              dk0mdField2m = ipvm->getConstRefTodLinearKdField()[extraDOFField][extraDOFField_aux][extraDOFField2];
              dk0pdField2p = ipvp->getConstRefTodLinearKdField()[extraDOFField][extraDOFField_aux][extraDOFField2];

              dk0mdFm  = ipvm->getConstRefTodLinearKdF()[extraDOFField][extraDOFField2];
              dk0pdFp  = ipvp->getConstRefTodLinearKdF()[extraDOFField][extraDOFField2];

              if(!getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())
                {
                 if(extraDOFField==extraDOFField2)
                 {
                 dfieldJumpdFieldp=ipvp->getConstRefTodEnergyConjugatedFieldJumpdFieldp()[extraDOFField][extraDOFField2];
                 dfieldJumpdFieldm=ipvm->getConstRefTodEnergyConjugatedFieldJumpdFieldm()[extraDOFField][extraDOFField2];
                 dfieldJumpdField2p=ipvp->getConstRefTodEnergyConjugatedFieldJumpdFieldp()[extraDOFField_aux][extraDOFField];
                 dfieldJumpdField2m=ipvm->getConstRefTodEnergyConjugatedFieldJumpdFieldm()[extraDOFField_aux][extraDOFField];

                 k02m = *(ipvm->getConstRefToLinearK()[extraDOFField][extraDOFField_aux]);
                 k02p = *(ipvp->getConstRefToLinearK()[extraDOFField][extraDOFField_aux]);
                }
                else 
                 {
                 dfieldJumpdFieldp=ipvp->getConstRefTodEnergyConjugatedFieldJumpdFieldp()[extraDOFField][extraDOFField2];
                 dfieldJumpdFieldm=ipvm->getConstRefTodEnergyConjugatedFieldJumpdFieldm()[extraDOFField][extraDOFField2];
                 dfieldJumpdField2p=ipvp->getConstRefTodEnergyConjugatedFieldJumpdFieldp()[extraDOFField2][extraDOFField2];
                 dfieldJumpdField2m=ipvm->getConstRefTodEnergyConjugatedFieldJumpdFieldm()[extraDOFField2][extraDOFField2];

                 k0m  = *(ipvm->getConstRefToLinearK()[extraDOFField][extraDOFField]);
                 k0p  = *(ipvp->getConstRefToLinearK()[extraDOFField][extraDOFField]);
                 k02m = *(ipvm->getConstRefToLinearK()[extraDOFField][extraDOFField2]);
                 k02p = *(ipvp->getConstRefToLinearK()[extraDOFField][extraDOFField2]);
                }
              }
              if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())
                   {

                    dk0mdField_auxm = ipvm->getConstRefTodLinearKdField()[extraDOFField][extraDOFField2][extraDOFField_aux];
                    dk0pdField_auxp = ipvp->getConstRefTodLinearKdField()[extraDOFField][extraDOFField2][extraDOFField_aux];
                    dk0mdField_aux2m = ipvm->getConstRefTodLinearKdField()[extraDOFField][extraDOFField_aux][extraDOFField_aux];
                    dk0pdField_aux2p = ipvp->getConstRefTodLinearKdField()[extraDOFField][extraDOFField_aux][extraDOFField_aux];
                   
                    dFielddEnergyConjugatedFieldp = ipvp->getConstRefTodFielddEnergyConjugatedField()[extraDOFField2][extraDOFField2];
                    dFielddEnergyConjugatedFieldm = ipvm->getConstRefTodFielddEnergyConjugatedField()[extraDOFField2][extraDOFField2]; 
                    dFielddEnergyConjugatedField2p = ipvp->getConstRefTodFielddEnergyConjugatedField()[extraDOFField_aux][extraDOFField2];
                    dFielddEnergyConjugatedField2m = ipvm->getConstRefTodFielddEnergyConjugatedField()[extraDOFField_aux][extraDOFField2]; 

                    dGradFielddEnergyConjugatedFieldp = ipvp->getConstRefTodGradFielddEnergyConjugatedField()[extraDOFField2][extraDOFField2]; 
                    dGradFielddEnergyConjugatedFieldm = ipvm->getConstRefTodGradFielddEnergyConjugatedField()[extraDOFField2][extraDOFField2];
                    dGradFielddGradEnergyConjugatedFieldp = ipvp->getConstRefTodGradFielddGradEnergyConjugatedField()[extraDOFField2][extraDOFField2];
                    dGradFielddGradEnergyConjugatedFieldm = ipvm->getConstRefTodGradFielddGradEnergyConjugatedField()[extraDOFField2][extraDOFField2]; 

                    dGradFielddEnergyConjugatedField2p = ipvp->getConstRefTodGradFielddEnergyConjugatedField()[extraDOFField_aux][extraDOFField2]; 
                    dGradFielddEnergyConjugatedField2m = ipvm->getConstRefTodGradFielddEnergyConjugatedField()[extraDOFField_aux][extraDOFField2];
                    dGradFielddGradEnergyConjugatedField2p = ipvp->getConstRefTodGradFielddGradEnergyConjugatedField()[extraDOFField_aux][extraDOFField2];
                    dGradFielddGradEnergyConjugatedField2m = ipvm->getConstRefTodGradFielddGradEnergyConjugatedField()[extraDOFField_aux][extraDOFField2]; 

                    dqdField2m = ipvm->getConstRefTodFluxdField()[extraDOFField][extraDOFField_aux];
                    dqdField2p = ipvp->getConstRefTodFluxdField()[extraDOFField][extraDOFField_aux];
                    dqdGradField2m = ipvm->getConstRefTodFluxdGradField()[extraDOFField][extraDOFField_aux];
                    dqdGradField2p = ipvp->getConstRefTodFluxdGradField()[extraDOFField][extraDOFField_aux];
                   }
            }  

            int indexField2=3+getNumNonLocalVariable()+extraDOFField2;

            // consistency D\mean{q} DT
            // Assembly consistency

            for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p = 0; p< 3; p++)
                {

                  if((getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())&&(ipvp->getNumConstitutiveExtraDofDiffusionVariable()==2)) 
                    stiff(j+indexField*nbFFm,l+indexField2*nbFFm) -= (dqdFieldm->operator()(p)*dFielddEnergyConjugatedFieldm+dqdField2m(p)*
                                                       dFielddEnergyConjugatedField2m)*(Valsm[l+indexField2*nbFFm]*Valsm[j+indexField*nbFFm]*
                                                       getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.*nu(p));
                  else 
                    stiff(j+indexField*nbFFm,l+indexField2*nbFFm) -= dqdFieldm->operator()(p)*(Valsm[l+indexField2*nbFFm]*
                                                        Valsm[j+indexField*nbFFm]*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.*nu(p));
                  for(int q = 0; q< 3; q++)
                  {

                    if((getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())&&(ipvp->getNumConstitutiveExtraDofDiffusionVariable()==2)) 
                      stiff(j+indexField*nbFFm,l+indexField2*nbFFm) -=(dqdGradFieldm->operator()(p,q)*dGradFielddEnergyConjugatedFieldm(q)+dqdGradField2m(p,q)*
                                                    dGradFielddEnergyConjugatedField2m(q))*(Valsm[j+indexField*nbFFm]*
                                                    Valsm[l+indexField2*nbFFm]*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.*nu(p));
                    else 
                      stiff(j+indexField*nbFFm,l+indexField2*nbFFm) -= dqdGradFieldm->operator()(q,p)*
                                     (Gradsm[l+indexField2*nbFFm][p]*Valsm[j+indexField*nbFFm]*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.*nu(q));

                    if((getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())&&(ipvp->getNumConstitutiveExtraDofDiffusionVariable()==2))
                     {
                       for(int o = 0; o< 3; o++)
                        {

                          stiff(j+indexField*nbFFm,l+indexField2*nbFFm) -=(dqdGradFieldm->operator()(p,o)*dGradFielddGradEnergyConjugatedFieldm(o,q)+dqdGradField2m(p,o)*
                                                     dGradFielddGradEnergyConjugatedField2m(o,q))*Valsm[j+indexField*nbFFm]*
                                                     (Gradsm[l+indexField2*nbFFm][q]*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.*nu(p));
                        }
                     }
                  }
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p = 0; p< 3; p++)
                {
                  
                  if((getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())&&(ipvp->getNumConstitutiveExtraDofDiffusionVariable()==2)) 
                    stiff(j+indexField*nbFFm,l+indexField2*nbFFp+nbdofm) -= (dqdFieldp->operator()(p)*dFielddEnergyConjugatedFieldp+dqdField2p(p)*
                                dFielddEnergyConjugatedField2p)*(Valsp[l+indexField2*nbFFp]*
                                Valsm[j+indexField*nbFFm]*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.*nu(p));
                  else 
                    stiff(j+indexField*nbFFm,l+indexField2*nbFFp+nbdofm) -= dqdFieldp->operator()(p)*
                               (Valsp[l+indexField2*nbFFp]*Valsm[j+indexField*nbFFm]*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.*nu(p));

                  for(int q = 0; q< 3; q++)
                  {
                    if((getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())&&(ipvp->getNumConstitutiveExtraDofDiffusionVariable()==2)) 
                      stiff(j+indexField*nbFFm,l+indexField2*nbFFp+nbdofm) -= (dqdGradFieldp->operator()(p,q)*
                                                 dGradFielddEnergyConjugatedFieldp(q)+dqdGradField2p(p,q)*dGradFielddEnergyConjugatedField2p(q))*
                                                 Valsm[j+indexField*nbFFm]*(Valsp[l+indexField2*nbFFp]*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.*nu(p));
                    else 
                      stiff(j+indexField*nbFFm,l+indexField2*nbFFp+nbdofm) -= dqdGradFieldp->operator()(q,p)*
                                                 (Gradsp[l+indexField2*nbFFp][p]*Valsm[j+indexField*nbFFm]*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.*nu(q));

                    if((getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())&&(ipvp->getNumConstitutiveExtraDofDiffusionVariable()==2))
                    {
                      for(int o = 0; o< 3; o++)
                      {

                          stiff(j+indexField*nbFFm,l+indexField2*nbFFp+nbdofm) -=(dqdGradFieldp->operator()(p,o)*dGradFielddGradEnergyConjugatedFieldp(o,q)+
                                                    dqdGradField2p(p,o)*dGradFielddGradEnergyConjugatedField2p(o,q))*Valsm[j+indexField*nbFFm]*
                                                    (Gradsp[l+indexField2*nbFFp][q]*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.*nu(p));
                      }
                    }
                  }
                }
              }
            }
            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p = 0; p< 3; p++)
                {

                  if((getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())&&(ipvp->getNumConstitutiveExtraDofDiffusionVariable()==2)) 
                    stiff(j+indexField*nbFFp+nbdofm,l+indexField2*nbFFm) += (dqdFieldm->operator()(p)*dFielddEnergyConjugatedFieldm+dqdField2m(p)*
                               dFielddEnergyConjugatedField2m)*
                               (Valsm[l+indexField2*nbFFm]*Valsp[j+indexField*nbFFp]*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.*nu(p));
                  else 
                    stiff(j+indexField*nbFFp+nbdofm,l+indexField2*nbFFm) += dqdFieldm->operator()(p)*(Valsm[l+indexField2*nbFFm]*
                                Valsp[j+indexField*nbFFp]*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.*nu(p));
                  for(int q = 0; q< 3; q++)
                  {

                    if((getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())&&(ipvp->getNumConstitutiveExtraDofDiffusionVariable()==2)) 
                      stiff(j+indexField*nbFFp+nbdofm,l+indexField2*nbFFm) +=(dqdGradFieldm->operator()(p,q)*dGradFielddEnergyConjugatedFieldm(q)+
                                    dqdGradField2m(p,q)*dGradFielddEnergyConjugatedField2m(q))*Valsp[j+indexField*nbFFp]*
                                    (Valsm[l+indexField2*nbFFm]*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.*nu(p));
                    else 
                      stiff(j+indexField*nbFFp+nbdofm,l+indexField2*nbFFm) += dqdGradFieldm->operator()(q,p)*
                                    (Gradsm[l+indexField2*nbFFm][p]*Valsp[j+indexField*nbFFp]*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.*nu(q));
                    if((getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())&&(ipvp->getNumConstitutiveExtraDofDiffusionVariable()==2))
                    {
                      for(int o = 0; o< 3; o++)
                      {

                        stiff(j+indexField*nbFFp+nbdofm,l+indexField2*nbFFm) +=(dqdGradFieldm->operator()(p,o)*dGradFielddGradEnergyConjugatedFieldm(o,q)+
                                             dqdGradField2m(p,o)*dGradFielddGradEnergyConjugatedField2m(o,q))*Valsp[j+indexField*nbFFp]*
                                             (Gradsm[l+indexField2*nbFFm][q]*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.*nu(p));
                      }
                    }
                  }
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p = 0; p< 3; p++)
                {

                  if((getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())&&(ipvp->getNumConstitutiveExtraDofDiffusionVariable()==2)) 
                    stiff(j+indexField*nbFFp+nbdofm,l+indexField2*nbFFp+nbdofm) += (dqdFieldp->operator()(p)*dFielddEnergyConjugatedFieldp+
                                                          dqdField2p(p)*dFielddEnergyConjugatedField2p)*
                              (Valsp[l+indexField2*nbFFp]*Valsp[j+indexField*nbFFp]*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.*nu(p));
                  else 
                    stiff(j+indexField*nbFFp+nbdofm,l+indexField2*nbFFp+nbdofm) += dqdFieldp->operator()(p)*
                              (Valsp[l+indexField2*nbFFp]*Valsp[j+indexField*nbFFp]*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.*nu(p));
                  for(int q = 0; q< 3; q++)
                  {

                    if((getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())&&(ipvp->getNumConstitutiveExtraDofDiffusionVariable()==2))
                      stiff(j+indexField*nbFFp+nbdofm,l+indexField2*nbFFp+nbdofm) += (dqdGradFieldp->operator()(p,q)*dGradFielddEnergyConjugatedFieldp(q)+
                                dqdGradField2p(p,q)*dGradFielddEnergyConjugatedField2p(q))*(Valsp[j+indexField*nbFFp]*
                                Valsp[l+indexField2*nbFFp]*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.*nu(p));
                    else 
                      stiff(j+indexField*nbFFp+nbdofm,l+indexField2*nbFFp+nbdofm) += dqdGradFieldp->operator()(q,p)*(Gradsp[l+indexField2*nbFFp][p]*
                                Valsp[j+indexField*nbFFp]*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.*nu(q));

                    if((getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())&&(ipvp->getNumConstitutiveExtraDofDiffusionVariable()==2))
                    {
                      for(int o = 0; o< 3; o++)
                      {

                        stiff(j+indexField*nbFFp+nbdofm,l+indexField2*nbFFp+nbdofm) += (dqdGradFieldp->operator()(p,o)*dGradFielddGradEnergyConjugatedFieldp(o,q)+
                                 dqdGradField2p(p,o)*dGradFielddGradEnergyConjugatedField2p(o,q))*Valsp[j+indexField*nbFFp]*
                                 (Gradsp[l+indexField2*nbFFp][q]*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.*nu(p));
                      }
                    }
                  }
                }
              }
            }


          if(ipvp->getNumConstitutiveExtraDofDiffusionVariable()==1)
          {
            // Stability  N \mean{k0} (N)
            double NMeankNBetasc=0.;
            double dNMeankNBetascdFieldp=0.;
            double dNMeankNBetascdFieldm=0.;

            for(int l = 0; l <3; l++)
            {
              for(int n = 0; n <3; n++)
              {
                NMeankNBetasc += (k0p(l,n)+k0m(l,n))*nu(n)*nbetahs*nu(l)/2.;
                dNMeankNBetascdFieldp+=dk0pdFieldp(l,n)*nu(n)*nbetahs*nu(l)/2.;
                dNMeankNBetascdFieldm+=dk0mdFieldm(l,n)*nu(n)*nbetahs*nu(l)/2.;
              }
            }

            // Assembly stability (if not broken)
            for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                stiff(j+indexField*nbFFm,l+indexField*nbFFm) -= dfieldJumpdFieldm*NMeankNBetasc*
                                      (Valsm[j+indexField*nbFFm]*Valsm[l+indexField*nbFFm]*wJ*getConstitutiveExtraDofDiffusionEqRatio());
                if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())
                  stiff(j+indexField*nbFFm,l+indexField*nbFFm) -= oneOverFieldJump*dNMeankNBetascdFieldm*
                                      (Valsm[j+indexField*nbFFm]*Valsm[l+indexField*nbFFm]*wJ*getConstitutiveExtraDofDiffusionEqRatio());
              }
              for(int l=0;l<nbFFp;l++)
              {
                stiff(j+indexField*nbFFm,l+indexField*nbFFp+nbdofm) -= dfieldJumpdFieldp*NMeankNBetasc*
                                      (Valsm[j+indexField*nbFFm]*Valsp[l+indexField*nbFFp]*wJ*getConstitutiveExtraDofDiffusionEqRatio());
                if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())
                  stiff(j+indexField*nbFFm,l+indexField*nbFFp+nbdofm) -= oneOverFieldJump*dNMeankNBetascdFieldp*
                                                  (Valsm[j+indexField*nbFFm]*Valsp[l+indexField*nbFFp]*wJ*getConstitutiveExtraDofDiffusionEqRatio());
              }
            }
            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                stiff(j+indexField*nbFFp+nbdofm,l+indexField*nbFFm) += dfieldJumpdFieldm*NMeankNBetasc*
                                                  (Valsp[j+indexField*nbFFp]*Valsm[l+indexField*nbFFm]*wJ*getConstitutiveExtraDofDiffusionEqRatio());
                if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())
                  stiff(j+indexField*nbFFp+nbdofm,l+indexField*nbFFm) += oneOverFieldJump*dNMeankNBetascdFieldm*
                                                  (Valsp[j+indexField*nbFFp]*Valsm[l+indexField*nbFFm]*wJ*getConstitutiveExtraDofDiffusionEqRatio());
              }
              for(int l=0;l<nbFFp;l++)
              {
                stiff(j+indexField*nbFFp+nbdofm,l+indexField*nbFFp+nbdofm) += dfieldJumpdFieldp*NMeankNBetasc*
                                                  (Valsp[j+indexField*nbFFp]*Valsp[l+indexField*nbFFp]*wJ*getConstitutiveExtraDofDiffusionEqRatio());
                if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())
                  stiff(j+indexField*nbFFp+nbdofm,l+indexField*nbFFp+nbdofm) += oneOverFieldJump*dNMeankNBetascdFieldp*
                                                  (Valsp[j+indexField*nbFFp]*Valsp[l+indexField*nbFFp]*wJ*getConstitutiveExtraDofDiffusionEqRatio());
              }
            }
            // compatibility
            // Assembly (loop on shape function)
            for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+indexField*nbFFm,l+indexField*nbFFm) += dfieldJumpdFieldm*(k0m(p,q)*
                                                  Gradsm[j+indexField*nbFFm][q])*(nu(p)*wJ*getConstitutiveExtraDofDiffusionEqRatio()/2.*Valsm[l+indexField*nbFFm]);
                    if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())
                      stiff(j+indexField*nbFFm,l+indexField*nbFFm) += oneOverFieldJump*(dk0mdFieldm(p,q)*
                                                  Gradsm[j+indexField*nbFFm][q])*(nu(p)*wJ*getConstitutiveExtraDofDiffusionEqRatio()/2.*Valsm[l+indexField*nbFFm]);
                  }
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+indexField*nbFFm,l+indexField*nbFFp+nbdofm) += dfieldJumpdFieldp*(k0m(p,q)*
                                                Gradsm[j+indexField*nbFFm][q])*(nu(p)*wJ*getConstitutiveExtraDofDiffusionEqRatio()/2.*Valsp[l+indexField*nbFFp]);
                  }
                }
              }
            }
            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+indexField*nbFFp+nbdofm,l+indexField*nbFFm) += dfieldJumpdFieldm*(k0p(p,q)*Gradsp[j+indexField*nbFFp][q])*
                                               (nu(p)*wJ*getConstitutiveExtraDofDiffusionEqRatio()/2.*Valsm[l+indexField*nbFFm]);
                  }
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+indexField*nbFFp+nbdofm,l+indexField*nbFFp+nbdofm) += dfieldJumpdFieldp*(k0p(p,q)*
                                      Gradsp[j+indexField*nbFFp][q])*(nu(p)*wJ*getConstitutiveExtraDofDiffusionEqRatio()/2.*Valsp[l+indexField*nbFFp]);
                    if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())
                      stiff(j+indexField*nbFFp+nbdofm,l+indexField*nbFFp+nbdofm) += oneOverFieldJump*
                                 (dk0pdFieldp(p,q)*Gradsp[j+indexField*nbFFp][q])*(nu(p)*wJ*getConstitutiveExtraDofDiffusionEqRatio()/2.*Valsp[l+indexField*nbFFp]);
                  }
                }
              }
            }
          }
          else
          {
            // Stability  N \mean{k0} (N)
            double NMeankNBetasc=0.;
            double NMeank2NBetasc=0.;
            double dNMeankNBetascdFieldp=0.;
            double dNMeankNBetascdFieldm=0.;
            double dNMeankNBetascdField2p=0.;
            double dNMeankNBetascdField2m=0.;
            double dNMeankNBetascdField_auxp=0.;
            double dNMeankNBetascdField_auxm=0.;
            double dNMeankNBetascdField_aux2p=0.;
            double dNMeankNBetascdField_aux2m=0.;          
            static STensor3 dNMeankNBetascdFm, dNMeankNBetascdFp;

            STensorOperation::zero(dNMeankNBetascdFm); STensorOperation::zero(dNMeankNBetascdFp);

            for(int l = 0; l <3; l++)
            {
              for(int n = 0; n <3; n++)
              {
                NMeankNBetasc  += (k0p(l,n)+k0m(l,n))*nu(n)*nbetahs*nu(l)/2.;
                NMeank2NBetasc += (k02p(l,n)+k02m(l,n))*nu(n)*nbetahs*nu(l)/2.;

                dNMeankNBetascdFieldp  += dk0pdFieldp(l,n)*nu(n)*nbetahs*nu(l)/2.;
                dNMeankNBetascdFieldm  += dk0mdFieldm(l,n)*nu(n)*nbetahs*nu(l)/2.;
                dNMeankNBetascdField2p += dk0pdField2p(l,n)*nu(n)*nbetahs*nu(l)/2.;
                dNMeankNBetascdField2m += dk0mdField2m(l,n)*nu(n)*nbetahs*nu(l)/2.; 
                
                if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())
                {
                  dNMeankNBetascdField_auxp += dk0pdField_auxp(l,n)*nu(n)*nbetahs*nu(l)/2.;         
                  dNMeankNBetascdField_auxm += dk0mdField_auxm(l,n)*nu(n)*nbetahs*nu(l)/2.; 
                  dNMeankNBetascdField_aux2p += dk0pdField_aux2p(l,n)*nu(n)*nbetahs*nu(l)/2.;         
                  dNMeankNBetascdField_aux2m += dk0mdField_aux2m(l,n)*nu(n)*nbetahs*nu(l)/2.;                 
                }
                for(int j = 0; j<3; j++)
                  {
                   for(int p = 0; p<3; p++)
                    {
                      dNMeankNBetascdFm(j,p)+=dk0mdFm(l,n,j,p)*nu(j)*nbetahs*nu(p)/2.;
                      dNMeankNBetascdFp(j,p)+=dk0pdFp(l,n,j,p)*nu(j)*nbetahs*nu(p)/2.;
                   }                
                 }
              }
            }

            if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())
            {
                 dNMeankNBetascdFieldp *= dFielddEnergyConjugatedFieldp;
                 dNMeankNBetascdFieldp += (dNMeankNBetascdField_auxp*dFielddEnergyConjugatedField2p);
                 dNMeankNBetascdFieldm *= dFielddEnergyConjugatedFieldm;
                 dNMeankNBetascdFieldm += (dNMeankNBetascdField_auxp*dFielddEnergyConjugatedField2m);
                 dNMeankNBetascdField2p *= dFielddEnergyConjugatedFieldp;
                 dNMeankNBetascdField2p += (dNMeankNBetascdField_aux2p*dFielddEnergyConjugatedField2p);
                 dNMeankNBetascdField2m *= dFielddEnergyConjugatedFieldm;
                 dNMeankNBetascdField2m += (dNMeankNBetascdField_aux2p*dFielddEnergyConjugatedField2m);
            }

            // Assembly stability (if not broken)
            for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField()) 
                  stiff(j+indexField*nbFFm,l+indexField2*nbFFm) += NMeankNBetasc*(Valsm[j+indexField*nbFFm]*Valsm[l+indexField2*nbFFm]*
                                                                    wJ*getConstitutiveExtraDofDiffusionEqRatio());
                else 
                  stiff(j+indexField*nbFFm,l+indexField2*nbFFm) -= (dfieldJumpdFieldm*NMeankNBetasc+dfieldJumpdField2m*NMeank2NBetasc)*
                                      (Valsm[j+indexField*nbFFm]*Valsm[l+indexField2*nbFFm]*wJ*getConstitutiveExtraDofDiffusionEqRatio());
                stiff(j+indexField*nbFFm,l+indexField2*nbFFm) -= (fieldJump*dNMeankNBetascdFieldm+fieldJump2*dNMeankNBetascdField2m)*
                                      (Valsm[j+indexField*nbFFm]*Valsm[l+indexField2*nbFFm]*wJ*getConstitutiveExtraDofDiffusionEqRatio());
              }
              for(int l=0;l<nbFFp;l++)
              {
                if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField()) 
                   stiff(j+indexField*nbFFm,l+indexField2*nbFFp+nbdofm) -= NMeankNBetasc*(Valsm[j+indexField*nbFFm]*
                                      Valsp[l+indexField2*nbFFp]*wJ*getConstitutiveExtraDofDiffusionEqRatio());
                else 
                   stiff(j+indexField*nbFFm,l+indexField2*nbFFp+nbdofm) -= (dfieldJumpdFieldp*NMeankNBetasc+dfieldJumpdField2p*NMeank2NBetasc)*
                                      (Valsm[j+indexField*nbFFm]*Valsp[l+indexField2*nbFFp]*wJ*getConstitutiveExtraDofDiffusionEqRatio());
                stiff(j+indexField*nbFFm,l+indexField2*nbFFp+nbdofm) -= (fieldJump*dNMeankNBetascdFieldp+fieldJump2*dNMeankNBetascdField2p)*
                                      (Valsm[j+indexField*nbFFm]*Valsp[l+indexField2*nbFFp]*wJ*getConstitutiveExtraDofDiffusionEqRatio());
              }
            }
            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField()) 
                  stiff(j+indexField*nbFFp+nbdofm,l+indexField2*nbFFm) -= NMeankNBetasc*
                                                     (Valsp[j+indexField*nbFFp]*Valsm[l+indexField2*nbFFm]*wJ*getConstitutiveExtraDofDiffusionEqRatio());
                else 
                  stiff(j+indexField*nbFFp+nbdofm,l+indexField2*nbFFm) += (dfieldJumpdFieldm*NMeankNBetasc+dfieldJumpdField2m*NMeank2NBetasc)*
                                                      (Valsp[j+indexField*nbFFp]*Valsm[l+indexField2*nbFFm]*wJ*getConstitutiveExtraDofDiffusionEqRatio());
                stiff(j+indexField*nbFFp+nbdofm,l+indexField2*nbFFm) += (fieldJump*dNMeankNBetascdFieldm+fieldJump2*dNMeankNBetascdField2m)*
                                                      (Valsp[j+indexField*nbFFp]*Valsm[l+indexField2*nbFFm]*wJ*getConstitutiveExtraDofDiffusionEqRatio());
              }
              for(int l=0;l<nbFFp;l++)
              {
                if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField()) 
                  stiff(j+indexField*nbFFp+nbdofm,l+indexField2*nbFFp+nbdofm) += NMeankNBetasc*
                                         (Valsp[j+indexField*nbFFp]*Valsp[l+indexField2*nbFFp]*wJ*getConstitutiveExtraDofDiffusionEqRatio());
                else 
                  stiff(j+indexField*nbFFp+nbdofm,l+indexField2*nbFFp+nbdofm) += (dfieldJumpdFieldp*NMeankNBetasc+dfieldJumpdField2p*NMeank2NBetasc)*
                                         (Valsp[j+indexField*nbFFp]*Valsp[l+indexField2*nbFFp]*wJ*getConstitutiveExtraDofDiffusionEqRatio());
                stiff(j+indexField*nbFFp+nbdofm,l+indexField2*nbFFp+nbdofm) += (fieldJump*dNMeankNBetascdFieldp+fieldJump2*dNMeankNBetascdField2p)*
                                         (Valsp[j+indexField*nbFFp]*Valsp[l+indexField2*nbFFp]*wJ*getConstitutiveExtraDofDiffusionEqRatio());
              }
            }
 

           for(int j=0;j<nbFFm;j++)
           {
              for(int k=0;k<3;k++)
              {
                 for(int l=0;l<nbFFm;l++)
                 {
                    if(useBarF)
                    {
                      for(int ii=0; ii<3; ii++)
                      {
                        for(int p=0; p<3; p++)
                        {
                          stiff(j+indexField*nbFFm,l+k*nbFFm) -= fieldJump*dNMeankNBetascdFm(ii,p)*dBarFdum[i][l](ii,p,k)*Valsm[j+indexField*nbFFm]*
                                                             wJ*getConstitutiveExtraDofDiffusionEqRatio();
                        }
                      }
                    }
                    else
                    {
                      for(int p=0;p<3;p++)
                      {
                        stiff(j+indexField*nbFFm,l+k*nbFFm) -= fieldJump*dNMeankNBetascdFm(k,p)*Gradsm[l+0*nbFFm][p]*Valsm[j+indexField*nbFFm]*
                                                             wJ*getConstitutiveExtraDofDiffusionEqRatio();
                      }
                    }
                  }
                  for(int l=0;l<nbFFp;l++)
                  {
                    if(useBarF)
                    {
                      for(int ii=0; ii<3; ii++)
                      {
                        for(int p=0; p<3; p++)
                        {
                         stiff(j+indexField*nbFFm,l+k*nbFFp+nbdofm) -= fieldJump*dNMeankNBetascdFp(ii,p)*dBarFdup[i][l](ii,p,k)*Valsm[j+indexField*nbFFm]*
                                                                wJ*getConstitutiveExtraDofDiffusionEqRatio();
                        }
                      }
                    }
                    else
                    {
                      for(int p=0;p<3;p++)
                      {
                        stiff(j+indexField*nbFFm,l+k*nbFFp+nbdofm) -= fieldJump*dNMeankNBetascdFp(k,p)*Gradsp[l+0*nbFFp][p]*Valsm[j+indexField*nbFFm]*
                                                               wJ*getConstitutiveExtraDofDiffusionEqRatio();
                      }
                    }
                  }
                } 
              }

            for(int j=0;j<nbFFp;j++)
            {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  if(useBarF)
                  {
                    for(int ii=0; ii<3; ii++)
                    {
                      for(int p=0; p<3; p++)
                      {
                        stiff(j+indexField*nbFFp+nbdofm,l+k*nbFFm) += fieldJump*dNMeankNBetascdFm(ii,p)*dBarFdum[i][l](ii,p,k)*Valsp[j+indexField*nbFFp]*
                                                                    wJ*getConstitutiveExtraDofDiffusionEqRatio();
                      }
                    }
                  }
                  else
                  {
                    for(int p=0;p<3;p++)
                    {
                      stiff(j+indexField*nbFFp+nbdofm,l+k*nbFFm) += fieldJump*dNMeankNBetascdFm(k,p)*Gradsm[l+0*nbFFm][p]*Valsp[j+indexField*nbFFp]*
                                                                    wJ*getConstitutiveExtraDofDiffusionEqRatio();
                    }
                  }
                }

                for(int l=0;l<nbFFp;l++)
                {
                  if(useBarF)
                  {
                    for(int ii=0; ii<3; ii++)
                    {
                      for(int p=0; p<3; p++)
                      {
                        stiff(j+indexField*nbFFp+nbdofm,l+k*nbFFp+nbdofm) +=fieldJump*dNMeankNBetascdFp(ii,p)*dBarFdup[i][l](ii,p,k)*Valsp[j+indexField*nbFFp]*
                                                                     wJ*getConstitutiveExtraDofDiffusionEqRatio();
                      }
                    }
                  }
                  else
                  {
                    for(int p=0;p<3;p++)
                    {
                      stiff(j+indexField*nbFFp+nbdofm,l+k*nbFFp+nbdofm) +=fieldJump*dNMeankNBetascdFp(k,p)*Gradsp[l+0*nbFFp][p]*Valsp[j+indexField*nbFFp]*
                                                                     wJ*getConstitutiveExtraDofDiffusionEqRatio();
                    }
                  }
                }
              }
            }
            // compatibility
            // Assembly (loop on shape function)

            for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField()) 
                       stiff(j+indexField*nbFFm,l+indexField2*nbFFm) -= (k0m(p,q)*Gradsm[j+indexField*nbFFm][q])*nu(p)*
                                                   Valsm[l+indexField2*nbFFm]*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.;
                    else 
                       stiff(j+indexField*nbFFm,l+indexField2*nbFFm) += ((k0m(p,q)*Gradsm[j+indexField*nbFFm][q])*nu(p)*Valsm[l+indexField2*nbFFm]*
                                      dfieldJumpdFieldm*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.)+((k02m(p,q)*Gradsm[j+indexField*nbFFm][q])*
                                      nu(p)*Valsm[l+indexField2*nbFFm]*dfieldJumpdField2m*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.);

                    if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField()) 
                       stiff(j+indexField*nbFFm,l+indexField2*nbFFm) += (fieldJump*(dk0mdFieldm(p,q)*dFielddEnergyConjugatedFieldm+dk0mdField_auxm(p,q)*
                                      dFielddEnergyConjugatedField2m)+fieldJump2*(dk0mdField2m(p,q)*dFielddEnergyConjugatedFieldm+dk0mdField_aux2m(p,q)*
                                      dFielddEnergyConjugatedField2m))*Gradsm[j+indexField*nbFFm][q]*nu(p)*Valsm[l+indexField2*nbFFm]*
                                      getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.;
                    else 
                       stiff(j+indexField*nbFFm,l+indexField2*nbFFm) += (fieldJump*(dk0mdFieldm(p,q)*Gradsm[j+indexField*nbFFm][q])*nu(p)*
                                      Valsm[l+indexField2*nbFFm]*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.)+
                                      (fieldJump2*(dk0mdField2m(p,q)*Gradsm[j+indexField*nbFFm][q])*nu(p)*
                                      Valsm[l+indexField2*nbFFm]*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.);
                  }
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField()) 
                       stiff(j+indexField*nbFFm,l+indexField2*nbFFp+nbdofm) += (k0m(p,q)*Gradsm[j+indexField*nbFFm][q])*nu(p)*
                                          Valsp[l+indexField2*nbFFp]*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.;
                    else 
                       stiff(j+indexField*nbFFm,l+indexField2*nbFFp+nbdofm) += ((k0m(p,q)*Gradsm[j+indexField*nbFFm][q])*nu(p)*
                                          Valsp[l+indexField2*nbFFp]*dfieldJumpdFieldp*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.)+
                                          ((k02m(p,q)*Gradsm[j+indexField*nbFFm][q])*nu(p)*Valsp[l+indexField2*nbFFp]*
                                          dfieldJumpdField2p*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.);
                  }
                }
              }
            }

            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField()) 
                      stiff(j+indexField*nbFFp+nbdofm,l+indexField2*nbFFm) -= (k0p(p,q)*Gradsp[j+indexField*nbFFp][q])*
                                           (nu(p)*Valsm[l+indexField2*nbFFm])*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.;
                    else 
                      stiff(j+indexField*nbFFp+nbdofm,l+indexField2*nbFFm) += ((k0p(p,q)*Gradsp[j+indexField*nbFFp][q])*nu(p)*
                                           Valsm[l+indexField2*nbFFm]*dfieldJumpdFieldm*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.)+
                                           ((k02p(p,q)*Gradsp[j+indexField*nbFFp][q])*nu(p)*Valsm[l+indexField2*nbFFm]*
                                           dfieldJumpdField2m*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.);
                  }
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField()) 
                      stiff(j+indexField*nbFFp+nbdofm,l+indexField2*nbFFp+nbdofm) += (k0p(p,q)*Gradsp[j+indexField*nbFFp][q])*nu(p)*
                                                  Valsp[l+indexField2*nbFFp]*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.;
                    else 
                      stiff(j+indexField*nbFFp+nbdofm,l+indexField2*nbFFp+nbdofm) += ((k0p(p,q)*Gradsp[j+indexField*nbFFp][q])*nu(p)*
                                                   Valsp[l+indexField2*nbFFp]*dfieldJumpdFieldp*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.)+
                                                  ((k02p(p,q)*Gradsp[j+indexField*nbFFp][q])*nu(p)*Valsp[l+indexField2*nbFFp]*
                                                  dfieldJumpdField2p*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.);

                    if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField()) 
                       stiff(j+indexField*nbFFp+nbdofm,l+indexField2*nbFFp+nbdofm) += (fieldJump*(dk0pdFieldp(p,q)*
                                                  dFielddEnergyConjugatedFieldp+dk0pdField_auxp(p,q)*dFielddEnergyConjugatedField2p)+
                                                  fieldJump2*(dk0pdField2p(p,q)*dFielddEnergyConjugatedFieldp+
                                                  dk0pdField_aux2p(p,q)*dFielddEnergyConjugatedField2p))*Gradsp[j+indexField*nbFFp][q]*
                                                  (nu(p)*getConstitutiveExtraDofDiffusionEqRatio()*(wJ/2.)*Valsp[l+indexField2*nbFFp]);
                    else 
                      stiff(j+indexField*nbFFp+nbdofm,l+indexField2*nbFFp+nbdofm) += (fieldJump*(dk0pdFieldp(p,q)*Gradsp[j+indexField*nbFFp][q])*nu(p)*
                                                  Valsp[l+indexField2*nbFFp]*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.)+
                                                  (fieldJump2*(dk0pdField2p(p,q)*Gradsp[j+indexField*nbFFp][q])*nu(p)*
                                                  Valsp[l+indexField2*nbFFp]*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.);
                  }
                }
              }
            }
           
            for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int m =0; m < 3; m++)
                {
                  for(int p =0; p < 3; p++)
                  {
                    for(int q =0; q < 3; q++)
                    {
                      if(useBarF)
                      {
                        for(int ii=0; ii<3; ii++)
                        {
                          for(int n=0; n<3; n++)
                          {
                            stiff(j+indexField*nbFFm,l+m*nbFFm) += fieldJump*(dk0mdFm(p,q,ii,n)*
                                      Gradsm[j+indexField*nbFFm][q]*dBarFdum[i][l](ii,n,m))*(nu(p)*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.);
                          }
                        }
                      }
                      else
                      {
                        for(int n =0; n< 3; n++)
                        {
                          stiff(j+indexField*nbFFm,l+m*nbFFm) += fieldJump*(dk0mdFm(p,q,m,n)*
                                      Gradsm[j+indexField*nbFFm][q]*Gradsm[l+0*nbFFm][n])*(nu(p)*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.);
                        }
                      }
                    }
                  }
                }
              }
            }

            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<nbFFp;l++)
              {
                for(int m =0; m < 3; m++)
                {
                  for(int q =0; q< 3; q++)
                  {
                    for(int p =0; p < 3; p++)
                    {
                      if(useBarF)
                      {
                        for(int ii=0; ii<3; ii++)
                        {
                          for(int n=0; n<3; n++)
                          {
                            stiff(j+indexField*nbFFp+nbdofm,l+m*nbFFp+nbdofm) += fieldJump*(dk0pdFp(p,q,ii,n)*
                                      Gradsp[j+indexField*nbFFp][q]*dBarFdup[i][l](ii,n,m))*
                                      (nu(p)*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.);
                          }
                        }
                      }
                      else
                      {
                        for(int n =0; n < 3; n++)
                        {
                          stiff(j+indexField*nbFFp+nbdofm,l+m*nbFFp+nbdofm) += fieldJump*(dk0pdFp(p,q,m,n)*
                                      Gradsp[j+indexField*nbFFp][q]*Gradsp[l+0*nbFFp][n])*
                                      (nu(p)*getConstitutiveExtraDofDiffusionEqRatio()*wJ/2.);
                        }
                      }
                    }
                  }
                }
              }
            }

          }
            for (int nlk = 0; nlk < ipvp->getNumberNonLocalVariable(); nlk++){
              static bool mess=false;
              if(!mess)
                Msg::Info("Implement non-local / extra dof coupling in dG3DStiffnessInter::get");
              mess=true;
            }
           }
          }
         }
      }
    }
  }
   //stiff.print("dginter");
}


/*void ElecTherMechMassG3D::get(MElement *ele,int npts,IntPt *GP,fullMatrix<double> &m) const
{
    mass3D::get(ele,npts,GP,m);
    Msg::Error("We should fill the 5th degree of freedom for the mass matrix");
}*/


/*void ElecTherMechDG3DForceBulk::get(MElement *ele,int npts,IntPt *GP,fullVector<double> &vFor) const
{

  // xyz Part
  dG3DForceBulk::get(ele,npts,GP,vFor);

  int nbdof = this->space1.getNumKeys(ele);
  int nbFF = ele->getNumShapeFunctions();

//  double jac[3][3];
  static SVector3 ATepsbar3;
  static SVector3 ATepsbar4;

  // get gauss' point data
  _ipf->getIPv(ele,vipv,vipvprev);

  for (int i = 0; i < npts; i++)
  {
    // Weight of Gauss' point i
    double weight = GP[i].weight;
    double &detJ = vipv[i]->getJacobianDeterminant(ele,GP[i]);


    // CHANGE FROMM HERE

    const STensor3 *PK1 = &(vipv[i]->getConstRefToFirstPiolaKirchhoffStress());
    double ratio = detJ*weight;
    double eqRatio = _ElecTherMechEqRatio;

//    const SVector3 &fluxT = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefToThermalFlux());
    const SVector3 &fluxV = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefToFluxV());
    const SVector3 &gradV = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefToGradV());
    const double    w     = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefToThermalSource());
    const SVector3 &fluxjy = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefToFluxEnergy());

    const std::vector<TensorialTraits<double>::GradType> &Grads = vipv[i]->gradf(&space1,ele,GP[i]);
    const std::vector<TensorialTraits<double>::ValType> &Vals = vipv[i]->f(&space1,ele,GP[i]);

    int indexFieldT=3+getNumNonLocalVariable();
    int indexFieldV=3+getNumNonLocalVariable()+1;

    for(int j=0;j<nbFF; j++)
    {
      ATepsbar3(0) = Grads[j+indexFieldT*nbFF][0];
      ATepsbar3(1) = Grads[j+indexFieldT*nbFF][1];
      ATepsbar3(2) = Grads[j+indexFieldT*nbFF][2];

      ATepsbar4(0) = Grads[j+indexFieldV*nbFF][0];
      ATepsbar4(1) = Grads[j+indexFieldV*nbFF][1];
      ATepsbar4(2) = Grads[j+indexFieldV*nbFF][2];

      for(int m=0; m<3; m++)
      {
        vFor(j+indexFieldT*nbFF) += eqRatio*ratio*(ATepsbar3(m)*fluxjy(m));
      }

      vFor(j+indexFieldT*nbFF) += eqRatio*ratio*(w*Vals[j+indexFieldT*nbFF]);  // related to cp

      for(int m=0; m<3; m++)
      {
        vFor(j+indexFieldV*nbFF) += eqRatio*ratio*(ATepsbar4(m)*fluxV(m));
      }
    }
  }
  //vFor.print("interF");
};*/


/*void ElecTherMechDG3DStiffnessBulk::get(MElement *ele,int npts,IntPt *GP,fullMatrix<double> &mStiff) const
{
  dG3DStiffnessBulk::get(ele,npts,GP,mStiff);

  if (sym)
  {
    // Initialization of some data
    int nbdof = BilinearTerm<double,double>::space1.getNumKeys(ele);
    int nbFF = ele->getNumShapeFunctions();

    // get ipvariables
    _ipf->getIPv(ele,vipv);

    static SVector3 Bepsbar4;
    static SVector3 Bepsbar3;
    static SVector3 Aepsbar4;
    static SVector3 Aepsbar3;
    static SVector3 newB;
    static SVector3 newA;

    if(_useFtFv==true)
    {
      for (int i = 0; i < npts; i++)
      {
        double weight = GP[i].weight;
        double &detJ = vipv[i]->getJacobianDeterminant(ele,GP[i]);
        // x, y, z
        const STensor3  &dFluxVdGradV  = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodFluxVdGradV());
        const SVector3  &dFluxVdV      = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodFluxVdV());
        const STensor3  &dFluxVdGradT  = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodFluxVdGradT());
        const SVector3  &dFluxVdT      = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodFluxVdT());
    //  const SVector3  &dFluxTdT      = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodFluxTdT());
    //  const STensor3  &dFluxTdGradT  = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodThermalFluxdGradT());
    //  const SVector3  &dFluxTdV      = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodFluxTdV());
    //  const STensor3  &dFluxTdGradV  = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodFluxTdGradV());
        const STensor3  &dPdV          = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodPdV());
        const STensor3  &dPdT          = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodPdT());
        const SVector3  &gradV         = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefToGradV());
        const SVector3  &FluxV         = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefToFluxV());
        const STensor33 &dFluxTdF      = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodThermalFluxdF());
        const STensor33 &dFluxVdF      = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodFluxVdF());
        const double    w              = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefToThermalSource());
        const double    dwdt           = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodThermalSourcedT());
        const STensor3  &dwdF          = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodThermalSourcedF());
        const double    dwdV           = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodThermalSourcedV());
        const SVector3  &dwdGradT      = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodThermalSourcedGradT());
        const SVector3  &dwdGradV      = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodThermalSourcedGradV());

        const STensor33 &djydF		= ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodFluxEnergydF());
        const STensor3  &djydGradT	= ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodFluxEnergydGradT());
        const SVector3  &djydT		= ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodFluxEnergydT());
        const STensor3  &djydGradV	= ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodFluxEnergydGradV());
        const SVector3  &djydV		= ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodFluxEnergydV());
        const SVector3  &fluxjy		= ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefToFluxEnergy());

        const STensor3 &dgardVdgradfV = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodgradVdgradfV());
        const STensor3 &dgardVdgradfT = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodgradVdgradfT());
        const STensor3 &dgardTdgradfT = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodgradTdgradfT());
        const STensor3 &dgardTdgradfV = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodgradTdgradfV());

        const double dVdfV = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodVdfV());
        const double dVdfT = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodVdfT());
        const double dTdfT = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodTdfT());
        const double dTdfV = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodTdfV());

        const SVector3 &dgradVdfV=((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodgradVdfV());
        const SVector3 &dgradVdfT=((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodgradVdfT());
        const SVector3 &dgradTdfT=((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodgradTdfT());
        const SVector3 &dgradTdfV=((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodgradTdfV());

        double eqRatio = _ElecTherMechEqRatio;

        const std::vector<TensorialTraits<double>::GradType> &Grads = vipv[i]->gradf(&space1,ele,GP[i]);
        const std::vector<TensorialTraits<double>::ValType> &Vals = vipv[i]->f(&space1,ele,GP[i]);

        int indexFieldfT=3+getNumNonLocalVariable();
        int indexFieldfV=3+getNumNonLocalVariable()+1;
        for(int j=0;j<nbFF; j++)
        {
          newA(0) = Grads[j][0];
          newA(1) = Grads[j][1];
          newA(2) = Grads[j][2];

          Aepsbar4(0) = Grads[j+indexFieldfV*nbFF][0];
          Aepsbar4(1) = Grads[j+indexFieldfV*nbFF][1];
          Aepsbar4(2) = Grads[j+indexFieldfV*nbFF][2];

          Aepsbar3(0) = Grads[j+indexFieldfT*nbFF][0];
          Aepsbar3(1) = Grads[j+indexFieldfT*nbFF][1];
          Aepsbar3(2) = Grads[j+indexFieldfT*nbFF][2];

          for(int k=0;k<nbFF;k++){
            double ratio = detJ*weight;
              // x y z
            newB(0) = Grads[k][0];
            newB(1) = Grads[k][1];
            newB(2) = Grads[k][2];

            Bepsbar4(0) = Grads[k+indexFieldfV*nbFF][0];
            Bepsbar4(1) = Grads[k+indexFieldfV*nbFF][1];
            Bepsbar4(2) = Grads[k+indexFieldfV*nbFF][2];

            Bepsbar3(0) = Grads[k+indexFieldfT*nbFF][0];
            Bepsbar3(1) = Grads[k+indexFieldfT*nbFF][1];
            Bepsbar3(2) = Grads[k+indexFieldfT*nbFF][2];

            // pure thermal dfint dt
            for(int m=0; m<3; m++)
            {
              mStiff(j+indexFieldfT*nbFF,k+indexFieldfT*nbFF)+=eqRatio*ratio*Aepsbar3(m)*(djydT.operator()(m)*dTdfT+djydV.operator()(m)*dVdfT)*Vals[k+indexFieldfT*nbFF];
              for(int n=0; n<3; n++)
              {
                mStiff(j+indexFieldfT*nbFF,k+indexFieldfT*nbFF)+=eqRatio*ratio*Aepsbar3(m)*(djydGradT.operator()(m,n)*dgradTdfT.operator()(n)
                              +djydGradV.operator()(m,n)*dgradVdfT.operator()(n))*Vals[k+indexFieldfT*nbFF];
                for(int p=0; p<3; p++)
                {
                  mStiff(j+indexFieldfT*nbFF,k+indexFieldfT*nbFF)+=eqRatio*ratio*Aepsbar3(m)*(djydGradT.operator()(m,p)*dgardTdgradfT.operator()(p,n)+
                                      djydGradV.operator()(m,p)*dgardVdgradfT.operator()(p,n))*Bepsbar3(n);
                }
              }
            }
            mStiff(j+indexFieldfT*nbFF,k+indexFieldfT*nbFF)+=eqRatio*ratio*dwdt*dTdfT*Vals[j+indexFieldfT*nbFF]*Vals[k+indexFieldfT*nbFF];// related to cp


            //dfintdv!!
            for(int m=0; m<3; m++)
            {
              mStiff(j+indexFieldfT*nbFF,k+indexFieldfV*nbFF)+=eqRatio*ratio*Aepsbar3(m)*(djydT.operator()(m)*dTdfV+djydV.operator()(m)*dVdfV)*Vals[k+indexFieldfV*nbFF];
              for(int n=0; n<3; n++)
              {
                mStiff(j+indexFieldfT*nbFF,k+indexFieldfV*nbFF)+=eqRatio*ratio*Aepsbar3(m)*(djydGradT.operator()(m,n)*dgradTdfV.operator()(n)
                            +djydGradV.operator()(m,n)*dgradVdfV.operator()(n))*Vals[k+indexFieldfV*nbFF];
                for(int p=0; p<3; p++)
                {
                  mStiff(j+indexFieldfT*nbFF,k+indexFieldfV*nbFF)+=eqRatio*ratio*Aepsbar3(m)*(djydGradV.operator()(m,p)*dgardVdgradfV.operator()(p,n)
                                +djydGradT.operator()(m,p)*dgardTdgradfV.operator()(p,n))*Bepsbar4(n);
                }
              }
            }

            // pure electrical dfinvdv!!
            for(int m=0; m<3; m++)
            {
              mStiff(j+indexFieldfV*nbFF,k+indexFieldfV*nbFF)+=eqRatio*ratio*Aepsbar4(m)*(dFluxVdV.operator()(m)*dVdfV+
                            dFluxVdT.operator()(m)*dTdfV)*Vals[k+indexFieldfV*nbFF];
              for(int n=0; n<3; n++)
              {
                mStiff(j+indexFieldfV*nbFF,k+indexFieldfV*nbFF)+=eqRatio*ratio*Aepsbar4(m)*(dFluxVdGradV.operator()(m,n)*dgradVdfV.operator()(n)+
                                  dFluxVdGradT.operator()(m,n)*dgradTdfV.operator()(n))*Vals[k+indexFieldfV*nbFF];
                for(int p=0; p<3; p++)
                {
                  mStiff(j+indexFieldfV*nbFF,k+indexFieldfV*nbFF)+=eqRatio*ratio*Aepsbar4(m)*(dFluxVdGradV.operator()(m,p)*dgardVdgradfV.operator()(p,n)+
                                dFluxVdGradT.operator()(m,p)*dgardTdgradfV.operator()(p,n))*Bepsbar4(n);
                }
              }
            }
            //dfinvdt
            for(int m=0; m<3; m++)
            {
              mStiff(j+indexFieldfV*nbFF,k+indexFieldfT*nbFF)+=eqRatio*ratio*Aepsbar4(m)*(dFluxVdT.operator()(m)*dTdfT+
                             dFluxVdV.operator()(m)*dVdfT)*Vals[k+indexFieldfT*nbFF];
              for(int n=0; n<3; n++)
              {
                mStiff(j+indexFieldfV*nbFF,k+indexFieldfT*nbFF)+=eqRatio*ratio*Aepsbar4(m)*(dFluxVdGradT.operator()(m,n)*dgradTdfT.operator()(n)+
                                     dFluxVdGradV.operator()(m,n)*dgradVdfT.operator()(n))*Vals[k+indexFieldfT*nbFF];

                for(int p=0; p<3; p++)
                {
                  mStiff(j+indexFieldfV*nbFF,k+indexFieldfT*nbFF)+=eqRatio*ratio*Aepsbar4(m)*(dFluxVdGradT.operator()(m,p)*dgardTdgradfT.operator()(p,n)+
                                       dFluxVdGradV.operator()(m,p)*dgardVdgradfT.operator()(p,n))*Bepsbar3(n);
                }
              }
            }

            // cross terms
            for(int kk=0;kk<3;kk++)
            {
              for(int m=0; m<3; m++)
              {
                mStiff(j+kk*nbFF,k+indexFieldfV*nbFF)  += ratio*(newA(m)*(dPdV.operator()(kk,m)*dVdfV+dPdT.operator()(kk,m)*dTdfV)*Vals[k+indexFieldfV*nbFF]); //dpdv
                mStiff(j+kk*nbFF,k+indexFieldfT*nbFF) += ratio*(newA(m)*(dPdT.operator()(kk,m)*dTdfT+dPdV.operator()(kk,m)*dVdfT)*Vals[k+indexFieldfT*nbFF]); //dpdT

                for(int n=0; n<3; n++)
                {
                  mStiff(j+indexFieldfV*nbFF,k+kk*nbFF)  += eqRatio*ratio*(Aepsbar4(m)*dFluxVdF.operator()(m,kk,n)*newB(n));
                  mStiff(j+indexFieldfT*nbFF,k+kk*nbFF) += eqRatio*ratio*(Aepsbar3(m)*djydF.operator()(m,kk,n)*newB(n));
               //  mStiff(j+threeTimesNbFF,k+kk*nbFF) -= eqRatio*ratio*(gradV(m)*dFluxVdF.operator()(m,kk,n)*newB(n)*Vals[j+threeTimesNbFF]); //before combining with w
                }
                mStiff(j+indexFieldfT*nbFF,k+kk*nbFF) += eqRatio*ratio*(newB(m)*dwdF.operator()(kk,m)*Vals[j+indexFieldfT*nbFF]); // related to cp
              }
            }
          }
        }
      }
    }
    else
    {
      for (int i = 0; i < npts; i++)
      {
        double weight = GP[i].weight;
        double &detJ = vipv[i]->getJacobianDeterminant(ele,GP[i]);
        // x, y, z
        const STensor3  &dFluxVdGradV  = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodFluxVdGradV());
        const SVector3  &dFluxVdV      = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodFluxVdV());
        const STensor3  &dFluxVdGradT  = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodFluxVdGradT());
        const SVector3  &dFluxVdT      = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodFluxVdT());
        const STensor3  &dPdV          = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodPdV());
        const STensor3  &dPdT          = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodPdT());
        const SVector3  &gradV         = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefToGradV());
        const SVector3  &FluxV         = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefToFluxV());
        const STensor33 &dFluxVdF      = (*(static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodFluxVdF());
        const double    w              = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefToThermalSource());
        const double    dwdt           = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodThermalSourcedT());
        const STensor3  &dwdF          = ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodThermalSourcedF());
        const STensor33 &djydF		= ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodFluxEnergydF());
        const STensor3  &djydGradT	= ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodFluxEnergydGradT());
        const SVector3  &djydT		= ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodFluxEnergydT());
        const STensor3  &djydGradV	= ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodFluxEnergydGradV());
        const SVector3  &djydV		= ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefTodFluxEnergydV());
        const SVector3  &fluxjy		= ((static_cast <const ElecTherMechDG3DIPVariableBase * > (vipv[i]))->getConstRefToFluxEnergy());

        double eqRatio = _ElecTherMechEqRatio;

        const std::vector<TensorialTraits<double>::GradType> &Grads = vipv[i]->gradf(&space1,ele,GP[i]);
        const std::vector<TensorialTraits<double>::ValType> &Vals = vipv[i]->f(&space1,ele,GP[i]);

        int indexFieldfT=3+getNumNonLocalVariable();
        int indexFieldfV=3+getNumNonLocalVariable()+1;
        for(int j=0;j<nbFF; j++)
        {
          newA(0) = Grads[j][0];
          newA(1) = Grads[j][1];
          newA(2) = Grads[j][2];

          Aepsbar4(0) = Grads[j+indexFieldfV*nbFF][0];
          Aepsbar4(1) = Grads[j+indexFieldfV*nbFF][1];
          Aepsbar4(2) = Grads[j+indexFieldfV*nbFF][2];

          Aepsbar3(0) = Grads[j+indexFieldfT*nbFF][0];
          Aepsbar3(1) = Grads[j+indexFieldfT*nbFF][1];
          Aepsbar3(2) = Grads[j+indexFieldfT*nbFF][2];

          for(int k=0;k<nbFF;k++){
            double ratio = detJ*weight;
              // x y z
            newB(0) = Grads[k][0];
            newB(1) = Grads[k][1];
            newB(2) = Grads[k][2];

            Bepsbar4(0) = Grads[k+indexFieldfV*nbFF][0];
            Bepsbar4(1) = Grads[k+indexFieldfV*nbFF][1];
            Bepsbar4(2) = Grads[k+indexFieldfV*nbFF][2];

            Bepsbar3(0) = Grads[k+indexFieldfT*nbFF][0];
            Bepsbar3(1) = Grads[k+indexFieldfT*nbFF][1];
            Bepsbar3(2) = Grads[k+indexFieldfT*nbFF][2];

            // pure thermal dfint dt
            for(int m=0; m<3; m++)
            {
              mStiff(j+indexFieldfT*nbFF,k+indexFieldfT*nbFF)+=eqRatio*ratio*(Aepsbar3(m)*djydT.operator()(m)*Vals[k+indexFieldfT*nbFF]);
              for(int n=0; n<3; n++)
              {
              mStiff(j+indexFieldfT*nbFF,k+indexFieldfT*nbFF)+=eqRatio*ratio*(Aepsbar3(m)*djydGradT.operator()(m,n)*Bepsbar3(n));
              }
            }
            mStiff(j+indexFieldfT*nbFF,k+indexFieldfT*nbFF)+=eqRatio*ratio*dwdt*Vals[j+indexFieldfT*nbFF]*Vals[k+indexFieldfT*nbFF];// related to cp


           /*   for(int n=0; n<3; n++)
              {
                 mStiff(j+indexFieldfT*nbFF,k+indexFieldfT*nbFF)+=eqRatio*ratio*(dwdGradT.operator()(n)*Bepsbar3(n))*Vals[j+indexFieldfT*nbFF];  // new related to w
              }//*


                //dfintdv
            for(int m=0; m<3; m++)
            {
              mStiff(j+indexFieldfT*nbFF,k+indexFieldfV*nbFF)+=eqRatio*ratio*(Aepsbar3(m)*djydV.operator()(m)*Vals[k+indexFieldfV*nbFF]);
              for(int n=0; n<3; n++)
              {
                mStiff(j+indexFieldfT*nbFF,k+indexFieldfV*nbFF)+=eqRatio*ratio*(Aepsbar3(m)*djydGradV.operator()(m,n)*Bepsbar4(n));
              }
            }

        /*      for(int m=0; m<3; m++)
            {
              mStiff(j+indexFieldfT*nbFF,k+indexFieldfV*nbFF)+=eqRatio*ratio*(dwdV*Vals[k+indexFieldfV*nbFF]*Vals[j+indexFieldfT*nbFF]);//new related to w
              for(int n=0; n<3; n++)
              {
              mStiff(j+indexFieldfT*nbFF,k+indexFieldfV*nbFF)+=eqRatio*ratio*(dwdGradV.operator()(n)*Bepsbar4(n)*Vals[j+indexFieldfT*nbFF]);//new related to w
              }
            }*//*


                // pure electrical dfintvdv
            for(int m=0; m<3; m++)
            {
              mStiff(j+indexFieldfV*nbFF,k+indexFieldfV*nbFF)+=eqRatio*ratio*(Aepsbar4(m)*dFluxVdV.operator()(m)*Vals[k+indexFieldfV*nbFF]);
              for(int n=0; n<3; n++)
              {
              mStiff(j+indexFieldfV*nbFF,k+indexFieldfV*nbFF)+=eqRatio*ratio*(Aepsbar4(m)*dFluxVdGradV.operator()(m,n)*Bepsbar4(n));
              }
            }
            //dfintvdt
            for(int m=0; m<3; m++)
            {
              mStiff(j+indexFieldfV*nbFF,k+indexFieldfT*nbFF)+=eqRatio*ratio*(Aepsbar4(m)*dFluxVdT.operator()(m)*Vals[k+indexFieldfT*nbFF]);
              for(int n=0; n<3; n++)
              {
                mStiff(j+indexFieldfV*nbFF,k+indexFieldfT*nbFF)+=eqRatio*ratio*(Aepsbar4(m)*dFluxVdGradT.operator()(m,n)*Bepsbar3(n));
              }
            }
                // cross terms
            for(int kk=0;kk<3;kk++)
            {
              for(int m=0; m<3; m++)
              {
                mStiff(j+kk*nbFF,k+indexFieldfV*nbFF)  += ratio*(newA(m)*dPdV.operator()(kk,m)*Vals[k+indexFieldfV*nbFF]);
                mStiff(j+kk*nbFF,k+indexFieldfT*nbFF) += ratio*(newA(m)*dPdT.operator()(kk,m)*Vals[k+indexFieldfT*nbFF]);

                for(int n=0; n<3; n++)
                {
                mStiff(j+indexFieldfV*nbFF,k+kk*nbFF)  += eqRatio*ratio*(Aepsbar4(m)*dFluxVdF.operator()(m,kk,n)*newB(n));
                mStiff(j+indexFieldfT*nbFF,k+kk*nbFF) += eqRatio*ratio*(Aepsbar3(m)*djydF.operator()(m,kk,n)*newB(n));
                //mStiff(j+indexFieldfT*nbFF,k+kk*nbFF) -= eqRatio*ratio*(gradV(m)*dFluxVdF.operator()(m,kk,n)*newB(n)*Vals[j+indexFieldfT*nbFF]); //before combining with w
                }
                mStiff(j+indexFieldfT*nbFF,k+kk*nbFF)+=eqRatio*ratio*(newB(m)*dwdF.operator()(kk,m)*Vals[j+indexFieldfT*nbFF]); // related to cp
              }
            }
          }
        }
      }
    }
 //mStiff.print("bulk");
  }
  else
   printf("not implemented\n");

}*/


/*void ElecTherMechDG3DForceInter::get(MElement *ele, int npts, IntPt *GP, fullVector<double> &m)const
{

  dG3DForceInter::get(ele, npts, GP,m);

  const fullVector<double> &datafield = *_data;
  MInterfaceElement *ie = dynamic_cast<MInterfaceElement*>(ele);
  const int nbFFm = ie->getElem(0)->getNumVertices();
  const int nbFFp = ie->getElem(1)->getNumVertices();
  const int nbdofm = _minusSpace->getNumKeys(ie->getElem(0));
  const int nbdofp = _plusSpace->getNumKeys(ie->getElem(1));
  double jac[3][3];
  if(_fullDg)
  {
    double gamma=0.;
    // characteristic size
    const double hs = ie->getCharacteristicSize();
    const double nbetahs = _ElecTherMechBeta/hs;
    if( _continuity)
    {
      // get value at gauss's point
      const AllIPState::ipstateElementContainer *vips = _ipf->getAips()->getIPstate(ie->getNum());
      // value on minus and plus elements
      IntPt *GPm; IntPt *GPp;
      _interQuad->getIntPoints(ie,GP,&GPm,&GPp);
      // loop on Gauss Point
      for(int i=0;i<npts; i++)
      {
        // IP
        const IPStateBase *ipsm        = (*vips)[i];
        const IPStateBase *ipsp        = (*vips)[i+npts];
        const ElecTherMechDG3DIPVariableBase *ipvm;
        const ElecTherMechDG3DIPVariableBase *ipvp;
        if(_mlawMinus->getType() == materialLaw::fracture )
        {
          const FractureCohesive3DIPVariable* ipvtmp = static_cast<const FractureCohesive3DIPVariable*>(ipsm->getState(IPStateBase::current));
          ipvm = static_cast<const ElecTherMechDG3DIPVariableBase*>(ipvtmp->getIPvBulk());
          ipvtmp = static_cast<const FractureCohesive3DIPVariable*>(ipsp->getState(IPStateBase::current));
          ipvp = static_cast<const ElecTherMechDG3DIPVariableBase*>(ipvtmp->getIPvBulk());
        }
        else
        {
          ipvm = static_cast<const ElecTherMechDG3DIPVariableBase*>(ipsm->getState(IPStateBase::current));
          ipvp = static_cast<const ElecTherMechDG3DIPVariableBase*>(ipsp->getState(IPStateBase::current));
        }


        const double weight = GP[i].weight;
        double detJ = ipvm->getJacobianDeterminant(ie,GP[i]);



        if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())
        {

          const SVector3 &referenceOutwardNormalm = ipvm->getConstRefToReferenceOutwardNormal();
          const SVector3 &referenceOutwardNormalp = ipvp->getConstRefToReferenceOutwardNormal();

          const double &ftJump =ipvp->getConstRefToOneOverTemperatureJump();
          const double &fvJump =ipvp->getConstRefTofvJump();

          const STensor3  l10m = *(ipvm->getConstRefToElecConductivity());
          const STensor3  l10p = *(ipvp->getConstRefToElecConductivity());

          const STensor3  l20m = *(ipvm->getConstRefToCrossElecConductivity());
          const STensor3  l20p = *(ipvp->getConstRefToCrossElecConductivity());

          const SVector3  &fluxVm = (ipvm->getConstRefToFluxV());
          const SVector3  &fluxVp = (ipvp->getConstRefToFluxV());

         //const STensor3 jy1m	= *(ipvm->getConstRefToEnergyConductivity());
         //const STensor3 jy1p	= *(ipvp->getConstRefToEnergyConductivity());

         const STensor3 jy1m	= *(ipvm->getConstRefToTherConductivity());
         const STensor3 jy1p	= *(ipvp->getConstRefToTherConductivity());

         // static STensor3 jy1m, jy1m_1, jy1p, jy1p_1;
         // STensorOperation::zero(jy1m);STensorOperation::zero(jy1m_1);STensorOperation::zero(jy1p_1);STensorOperation::zero(jy1p);
         // STensorOperation::multSTensor3(ipvm->getConstRefTodFluxdGradField()[0][0],ipvm->getConstRefTodGradFielddGradEnergyConjugatedField()[0][0],jy1m_1);
         // STensorOperation::multSTensor3(ipvm->getConstRefTodFluxdGradField()[0][1],ipvm->getConstRefTodGradFielddGradEnergyConjugatedField()[1][0],jy1m);
         // jy1m += jy1m_1;
         // STensorOperation::multSTensor3(ipvp->getConstRefTodFluxdGradField()[0][0],ipvp->getConstRefTodGradFielddGradEnergyConjugatedField()[0][0],jy1p_1);
         // STensorOperation::multSTensor3(ipvp->getConstRefTodFluxdGradField()[0][1],ipvp->getConstRefTodGradFielddGradEnergyConjugatedField()[1][0],jy1p);
         // jy1p += jy1p_1;

          const SVector3  &fluxjym= (ipvm->getConstRefToFluxEnergy());
          const SVector3  &fluxjyp= (ipvp->getConstRefToFluxEnergy());

          static STensor3 Stiff_alphadialitationm;
          Stiff_alphadialitationm = *ipvm->getConstRefToStiff_alphaDilatation();
          static STensor3  Stiff_alphadialitationp;
          Stiff_alphadialitationp = *ipvp->getConstRefToStiff_alphaDilatation();

          const double &fTp = ipvp->getConstRefTofT();
          const double &fTm = ipvm->getConstRefTofT();

          Stiff_alphadialitationm*= (-1.)/(fTm*fTm);
          Stiff_alphadialitationp*= (-1.)/(fTp*fTp);
          const SVector3 &jump = ipvp->getConstRefToJump();


          double eqRatio = _ElecTherMechEqRatio;

          const std::vector<TensorialTraits<double>::GradType> &Gradsm = ipvm->gradf(_minusSpace,ie->getElem(0),GPm[i]);
          const std::vector<TensorialTraits<double>::GradType> &Gradsp = ipvp->gradf(_plusSpace,ie->getElem(1),GPp[i]);

          // possibility of fracture
          bool broken = false;
          if(_mlawMinus->getType() == materialLaw::fracture )
          {
            const FractureCohesive3DIPVariable *ipvwf = static_cast<const FractureCohesive3DIPVariable*>(ipsm->getState(IPStateBase::current)); // broken via minus (OK broken on both sides)
            broken = ipvwf->isbroken();
          }
          double wJ = detJ* weight;
          wJ*=eqRatio;

          if(!broken)
          {
            static SVector3 nu;
            nu = referenceOutwardNormalm;
            nu*=(1./referenceOutwardNormalm.norm());


            const std::vector<TensorialTraits<double>::ValType> &Valsm = ipvm->f(_minusSpace,ie->getElem(0),GPm[i]);
            const std::vector<TensorialTraits<double>::ValType> &Valsp = ipvp->f(_plusSpace,ie->getElem(1),GPp[i]);

            // consistency \mean{cg*grad eps}
            double meantfluxV =0.;
            for(int k =0; k<3; k++)
            {
              meantfluxV += (fluxVm(k)+fluxVp(k))*(nu(k)/2.);
            }
            double meantfluxjy =0.;
            for(int k =0; k<3; k++)
            {
               meantfluxjy += (fluxjym(k)+fluxjyp(k))*(nu(k)/2.);
            }

          double NMeanjy1JumpNBetasc = 0.;
          double NMeanl10JumpNBetasc = 0.;
          double NMeanl20fvJumpNBetasc = 0.;
          double NMeanl20ftJumpNBetasc = 0.;

          static STensor3 meanjy1;
          meanjy1 = (jy1m);
          meanjy1+= (jy1p);
          meanjy1*= (1./2.);
          static STensor3 meanl10;
          meanl10 = (l10m);
          meanl10+= (l10p);
          meanl10*= (1./2.);
          static STensor3 meanl20;
          meanl20 = (l20m);
          meanl20+= (l20p);
          meanl20*= (1./2.);

          for(int l = 0; l <3; l++)
          {
            for(int n = 0; n <3; n++)
            {
              NMeanjy1JumpNBetasc += meanjy1(l,n)*ftJump*nu(n)*nbetahs*nu(l);
              NMeanl10JumpNBetasc += meanl10(l,n)*fvJump*nu(n)*nbetahs*nu(l);
              NMeanl20fvJumpNBetasc += meanl20(l,n)*fvJump*nu(n)*nbetahs*nu(l);
              NMeanl20ftJumpNBetasc += meanl20(l,n)*ftJump*nu(n)*nbetahs*nu(l);
            }
          }

          double Stiff_alphadialitationp_JumpN=0.;double Stiff_alphadialitationm_JumpN=0.;
          for(int k = 0; k <3; k++)
          {
            for(int m = 0; m <3; m++)
            {
              Stiff_alphadialitationp_JumpN += Stiff_alphadialitationp(k,m)*jump(k)*nu(m);
              Stiff_alphadialitationm_JumpN += Stiff_alphadialitationm(k,m)*jump(k)*nu(m);
            }
          }
          int indexFieldfT=3+getNumNonLocalVariable();
          int indexFieldfV=3+getNumNonLocalVariable()+1;
          //fIt (consistency + stability)
          for(int j=0;j<nbFFm;j++)
          {
             m(j+indexFieldfT*nbFFm) -= (meantfluxjy+NMeanjy1JumpNBetasc)*Valsm[j+indexFieldfT*nbFFm]*wJ;
          }
          for(int j=0;j<nbFFm;j++)
          {
            m(j+indexFieldfT*nbFFm) -= (NMeanl20fvJumpNBetasc*Valsm[j+indexFieldfT*nbFFm])*wJ;
          }

          for(int j=0;j<nbFFp;j++)
          {
            m(j+indexFieldfT*nbFFp+nbdofm) += (meantfluxjy+NMeanjy1JumpNBetasc)*Valsp[j+indexFieldfT*nbFFp]*wJ;
          }
          for(int j=0;j<nbFFp;j++)
          {
            m(j+indexFieldfT*nbFFp+nbdofm) += (NMeanl20fvJumpNBetasc*Valsp[j+indexFieldfT*nbFFp])*wJ;
          }

          //fIv
          for(int j=0;j<nbFFm;j++)
          {
            m(j+indexFieldfV*nbFFm) -= (meantfluxV+NMeanl10JumpNBetasc)*Valsm[j+indexFieldfV*nbFFm]*wJ;
          }
          for(int j=0;j<nbFFm;j++)
          {
            m(j+indexFieldfV*nbFFm) -= NMeanl20ftJumpNBetasc*Valsm[j+indexFieldfV*nbFFm]*wJ;
          }

          for(int j=0;j<nbFFp;j++)
          {
            m(j+indexFieldfV*nbFFp+nbdofm) += (meantfluxV+NMeanl10JumpNBetasc)*Valsp[j+indexFieldfV*nbFFp]*wJ;
          }
          for(int j=0;j<nbFFp;j++)
          {
            m(j+indexFieldfV*nbFFp+nbdofm) += (NMeanl20ftJumpNBetasc*Valsp[j+indexFieldfV*nbFFp])*wJ;
          }

          static SVector3 jy1pJumpN, jy1mJumpN;
          static SVector3 l10pJumpN, l10mJumpN;
          static SVector3 l20pfvJumpN, l20mfvJumpN;
          static SVector3 l20pftJumpN, l20mftJumpN;
          for(int k = 0; k <3; k++)
          {
            jy1mJumpN(k) = 0.;
            jy1pJumpN(k) = 0.;
            l10pJumpN(k) = 0.;
            l10mJumpN(k) = 0.;
            l20pfvJumpN(k) = 0.;
            l20mfvJumpN(k) = 0.;
            l20pftJumpN(k) = 0.;
            l20mftJumpN(k) = 0.;
            for(int m = 0; m <3; m++)
            {
              jy1mJumpN(k) += jy1m(k,m)*ftJump*nu(m);
              jy1pJumpN(k) += jy1p(k,m)*ftJump*nu(m);

              l10mJumpN(k) += l10m(k,m)*fvJump*nu(m);
              l10pJumpN(k) += l10p(k,m)*fvJump*nu(m);
              l20mfvJumpN(k) += l20m(k,m)*fvJump*nu(m);
              l20pfvJumpN(k) += l20p(k,m)*fvJump*nu(m);
              l20mftJumpN(k) += l20m(k,m)*ftJump*nu(m);
              l20pftJumpN(k) += l20p(k,m)*ftJump*nu(m);
            }
          }

          // Assembly (loop on shape function)
          //fIt2  (compatibility)
          for(int j=0;j<nbFFm;j++)
          {
            for(int l=0;l<3;l++)
            {
              m(j+indexFieldfT*nbFFm) += ( jy1mJumpN(l)*Gradsm[j+indexFieldfT*nbFFm][l])*(wJ/2.);
              m(j+indexFieldfT*nbFFm) += ( l20mfvJumpN(l)*Gradsm[j+indexFieldfT*nbFFm][l])*(wJ/2.);
            }
          }

          for(int j=0;j<nbFFp;j++)
          {
            for(int l=0;l<3;l++)
            {
              m(j+indexFieldfT*nbFFp+nbdofm) += ( jy1pJumpN(l)*Gradsp[j+indexFieldfT*nbFFp][l])*(wJ/2.);
              m(j+indexFieldfT*nbFFp+nbdofm) += ( l20pfvJumpN(l)*Gradsp[j+indexFieldfT*nbFFp][l])*(wJ/2.);
            }
          }

          //new alphadialitation
          for(int j=0;j<nbFFm;j++)
          {
            m(j+indexFieldfT*nbFFm) +=  gamma*( Stiff_alphadialitationm_JumpN*Valsm[j+indexFieldfT*nbFFm])*(wJ/2.)/eqRatio*(-1.);
          }
          for(int j=0;j<nbFFp;j++)
          {
            m(j+indexFieldfT*nbFFp+nbdofm) += gamma*( Stiff_alphadialitationp_JumpN*Valsp[j+indexFieldfT*nbFFp])*(wJ/2.)/eqRatio*(-1.);
          }

            //fIv2
          for(int j=0;j<nbFFm;j++)
          {
            for(int l=0;l<3;l++)
            {
              m(j+indexFieldfV*nbFFm) += ( l10mJumpN(l)*Gradsm[j+indexFieldfV*nbFFm][l])*(wJ/2.);
              m(j+indexFieldfV*nbFFm) += ( l20mftJumpN(l)*Gradsm[j+indexFieldfV*nbFFm][l])*(wJ/2.);
            }
          }
          for(int j=0;j<nbFFp;j++)
          {
            for(int l=0;l<3;l++)
            {
              m(j+indexFieldfV*nbFFp+nbdofm) += ( l10pJumpN(l)*Gradsp[j+indexFieldfV*nbFFp][l])*(wJ/2.);
              m(j+indexFieldfV*nbFFp+nbdofm) += ( l20pftJumpN(l)*Gradsp[j+indexFieldfV*nbFFp][l])*(wJ/2.);
            }
          }
        }
        else //broken case
        {

        }
      }
        else{
          const SVector3 &referenceOutwardNormalm = ipvm->getConstRefToReferenceOutwardNormal();
          //const SVector3 &CurrentOutwardNormalm = ipvm->getConstRefToCurrentOutwardNormal();

          const SVector3 &referenceOutwardNormalp = ipvp->getConstRefToReferenceOutwardNormal();

          const double &ftJump =ipvp->getConstRefToOneOverTemperatureJump();
          const double &fvJump =ipvp->getConstRefTofvJump();


          const STensor3  l10m = *(ipvm->getConstRefToElecConductivity());
          const STensor3  l10p = *(ipvp->getConstRefToElecConductivity());

          const STensor3  l20m = *(ipvm->getConstRefToCrossElecConductivity());
          const STensor3  l20p = *(ipvp->getConstRefToCrossElecConductivity());

          const SVector3  &fluxVm = (ipvm->getConstRefToFluxV());
          const SVector3  &fluxVp = (ipvp->getConstRefToFluxV());

          //const STensor3  jy1m	= *(ipvm->getConstRefToEnergyConductivity());
          //const STensor3  jy1p	= *(ipvp->getConstRefToEnergyConductivity());

          const STensor3 jy1m	= *(ipvm->getConstRefToTherConductivity());
          const STensor3 jy1p	= *(ipvp->getConstRefToTherConductivity());

         // static STensor3 jy1m, jy1m_1, jy1p, jy1p_1;
         // STensorOperation::zero(jy1m);STensorOperation::zero(jy1m_1);STensorOperation::zero(jy1p_1);STensorOperation::zero(jy1p);
         // STensorOperation::multSTensor3(ipvm->getConstRefTodFluxdGradField()[0][0],ipvm->getConstRefTodGradFielddGradEnergyConjugatedField()[0][0],jy1m);
         // STensorOperation::multSTensor3(ipvm->getConstRefTodFluxdGradField()[0][1],ipvm->getConstRefTodGradFielddGradEnergyConjugatedField()[1][0],jy1m);
         // jy1m += jy1m_1;
         // STensorOperation::multSTensor3(ipvp->getConstRefTodFluxdGradField()[0][0],ipvp->getConstRefTodGradFielddGradEnergyConjugatedField()[0][0],jy1p);
         // STensorOperation::multSTensor3(ipvp->getConstRefTodFluxdGradField()[0][1],ipvp->getConstRefTodGradFielddGradEnergyConjugatedField()[1][0],jy1p);
         // jy1p += jy1p_1;

          const SVector3  &fluxjym= (ipvm->getConstRefToFluxEnergy());
          const SVector3  &fluxjyp= (ipvp->getConstRefToFluxEnergy());

          static STensor3  Stiff_alphadialitationm;
          Stiff_alphadialitationm = *ipvm->getConstRefToStiff_alphaDilatation();
          static STensor3  Stiff_alphadialitationp;
          Stiff_alphadialitationp = *ipvp->getConstRefToStiff_alphaDilatation();

          const double &fTp = ipvp->getConstRefTofT();
          const double &fTm = ipvm->getConstRefTofT();

          Stiff_alphadialitationm*= (-1.)/(fTm*fTm);
          Stiff_alphadialitationp*= (-1.)/(fTp*fTp);
          const SVector3 &jump = ipvp->getConstRefToJump();

          //double length = _mlaw->getLength();
          double eqRatio = _ElecTherMechEqRatio;

          const std::vector<TensorialTraits<double>::GradType> &Gradsm = ipvm->gradf(_minusSpace,ie->getElem(0),GPm[i]);
          const std::vector<TensorialTraits<double>::GradType> &Gradsp = ipvp->gradf(_plusSpace,ie->getElem(1),GPp[i]);

          // possibility of fracture
          bool broken = false;
          if(_mlawMinus->getType() == materialLaw::fracture )
          {
            const FractureCohesive3DIPVariable *ipvwf = static_cast<const FractureCohesive3DIPVariable*>(ipsm->getState(IPStateBase::current)); // broken via minus (OK broken on both sides)
            broken = ipvwf->isbroken();
          }
          double wJ = detJ* weight;
          wJ*=eqRatio;
          if(!broken)
          {
            static SVector3 nu;
            nu = referenceOutwardNormalm;
            nu*=(1./referenceOutwardNormalm.norm());

            const std::vector<TensorialTraits<double>::ValType> &Valsm = ipvm->f(_minusSpace,ie->getElem(0),GPm[i]);
            const std::vector<TensorialTraits<double>::ValType> &Valsp = ipvp->f(_plusSpace,ie->getElem(1),GPp[i]);

            // consistency \mean{cg*grad eps}
            double meantfluxV =0.;
            for(int k =0; k<3; k++)
            {
               meantfluxV += (fluxVm(k)+fluxVp(k))*(nu(k)/2.);
            }
            double meantfluxjy =0.;
            for(int k =0; k<3; k++)
            {
               meantfluxjy += (fluxjym(k)+fluxjyp(k))*(nu(k)/2.);
            }

            double NMeanjy1JumpNBetasc = 0.;
            double NMeanl10JumpNBetasc = 0.;
            double NMeanl20fvJumpNBetasc = 0.;
            double NMeanl20ftJumpNBetasc = 0.;

            static STensor3 meanjy1;
            meanjy1 = (jy1m);
            meanjy1+= (jy1p);
            meanjy1*= (1./2.);
            static STensor3 meanl10;
            meanl10 = (l10m);
            meanl10+= (l10p);
            meanl10*= (1./2.);
            static STensor3 meanl20;
            meanl20 = (l20m);
            meanl20+= (l20p);
            meanl20*= (1./2.);

            for(int l = 0; l <3; l++)
            {
              for(int n = 0; n <3; n++)
              {
                NMeanjy1JumpNBetasc += meanjy1(l,n)*ftJump*nu(n)*nbetahs*nu(l);
                NMeanl10JumpNBetasc += meanl10(l,n)*fvJump*nu(n)*nbetahs*nu(l);
                NMeanl20fvJumpNBetasc += meanl20(l,n)*fvJump*nu(n)*nbetahs*nu(l);
                NMeanl20ftJumpNBetasc += meanl20(l,n)*ftJump*nu(n)*nbetahs*nu(l);
              }
            }

            double Stiff_alphadialitationp_JumpN=0.;double Stiff_alphadialitationm_JumpN=0.;
            for(int k = 0; k <3; k++)
            {
              for(int m = 0; m <3; m++)
              {
                Stiff_alphadialitationp_JumpN += Stiff_alphadialitationp(k,m)*jump(k)*nu(m);
                Stiff_alphadialitationm_JumpN += Stiff_alphadialitationm(k,m)*jump(k)*nu(m);
              }
            }
            int indexFieldfT=3+getNumNonLocalVariable();
            int indexFieldfV=3+getNumNonLocalVariable()+1;

            //fIt1+3 consistency + stability
            for(int j=0;j<nbFFm;j++)
            {
              m(j+indexFieldfT*nbFFm) -= (meantfluxjy+NMeanjy1JumpNBetasc)*Valsm[j+indexFieldfT*nbFFm]*wJ;
            }
            for(int j=0;j<nbFFm;j++)
            {
              m(j+indexFieldfT*nbFFm) -= NMeanl20fvJumpNBetasc*Valsm[j+indexFieldfT*nbFFm]*wJ;
            }

            for(int j=0;j<nbFFp;j++)
            {
              m(j+indexFieldfT*nbFFp+nbdofm) += (meantfluxjy+NMeanjy1JumpNBetasc)*Valsp[j+indexFieldfT*nbFFp]*wJ;
            }
            for(int j=0;j<nbFFp;j++)
            {
              m(j+indexFieldfT*nbFFp+nbdofm) += (NMeanl20fvJumpNBetasc*Valsp[j+indexFieldfT*nbFFp])*wJ;
            }


            //fIv
            for(int j=0;j<nbFFm;j++)
            {
              m(j+indexFieldfV*nbFFm) -= (meantfluxV+NMeanl10JumpNBetasc)*Valsm[j+indexFieldfV*nbFFm]*wJ;
            }
            for(int j=0;j<nbFFm;j++)
            {
              m(j+indexFieldfV*nbFFm) -= (NMeanl20ftJumpNBetasc*Valsm[j+indexFieldfV*nbFFm])*wJ;
            }

            for(int j=0;j<nbFFp;j++)
            {
              m(j+indexFieldfV*nbFFp+nbdofm) += (meantfluxV+NMeanl10JumpNBetasc)*Valsp[j+indexFieldfV*nbFFp]*wJ;
            }
            for(int j=0;j<nbFFp;j++)
            {
              m(j+indexFieldfV*nbFFp+nbdofm) += (NMeanl20ftJumpNBetasc*Valsp[j+indexFieldfV*nbFFp])*wJ;
            }

              static SVector3 jy1pJumpN, jy1mJumpN;
              static SVector3 l10pJumpN, l10mJumpN;
              static SVector3 l20pfvJumpN, l20mfvJumpN;
              static SVector3 l20pftJumpN, l20mftJumpN;
              for(int k = 0; k <3; k++)
              {
                jy1mJumpN(k) = 0.;
                jy1pJumpN(k) = 0.;
                l10mJumpN(k) = 0.;
                l10pJumpN(k) = 0.;
                l20mfvJumpN(k) = 0.;
                l20pfvJumpN(k) = 0.;
                l20mftJumpN(k) = 0.;
                l20pftJumpN(k) = 0.;
                for(int m = 0; m <3; m++)
                {
                  jy1mJumpN(k) += jy1m(k,m)*ftJump*nu(m);
                  jy1pJumpN(k) += jy1p(k,m)*ftJump*nu(m);
                  l10mJumpN(k) += l10m(k,m)*fvJump*nu(m);
                  l10pJumpN(k) += l10p(k,m)*fvJump*nu(m);
                  l20mfvJumpN(k) += l20m(k,m)*fvJump*nu(m);
                  l20pfvJumpN(k) += l20p(k,m)*fvJump*nu(m);
                  l20mftJumpN(k) += l20m(k,m)*ftJump*nu(m);
                  l20pftJumpN(k) += l20p(k,m)*ftJump*nu(m);
                }
              }


            // Assembly (loop on shape function)
            //fIt2  (compatibility)
            for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<3;l++)
              {
                m(j+indexFieldfT*nbFFm) += ( jy1mJumpN(l)*Gradsm[j+indexFieldfT*nbFFm][l])*(wJ/2.);
                m(j+indexFieldfT*nbFFm) += ( l20mfvJumpN(l)*Gradsm[j+indexFieldfT*nbFFm][l])*(wJ/2.);
              }
            }

            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<3;l++)
              {
                m(j+indexFieldfT*nbFFp+nbdofm) += ( jy1pJumpN(l)*Gradsp[j+indexFieldfT*nbFFp][l])*(wJ/2.);
                m(j+indexFieldfT*nbFFp+nbdofm) += ( l20pfvJumpN(l)*Gradsp[j+indexFieldfT*nbFFp][l])*(wJ/2.);
              }
            }

            //new alphadialitation
            for(int j=0;j<nbFFm;j++)
            {
              m(j+indexFieldfT*nbFFm) +=  gamma*( Stiff_alphadialitationm_JumpN*Valsm[j+indexFieldfT*nbFFm])*(wJ/2.)/eqRatio*(-1.);
            }
            for(int j=0;j<nbFFp;j++)
            {
              m(j+indexFieldfT*nbFFp+nbdofm) +=  gamma*( Stiff_alphadialitationp_JumpN*Valsp[j+indexFieldfT*nbFFp])*(wJ/2.)/eqRatio*(-1.);
            }


      //v
            for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<3;l++)
              {
                m(j+indexFieldfV*nbFFm) += ( l10mJumpN(l)*Gradsm[j+indexFieldfV*nbFFm][l])*(wJ/2.);
                m(j+indexFieldfV*nbFFm) += ( l20mftJumpN(l)*Gradsm[j+indexFieldfV*nbFFm][l])*(wJ/2.);
              }
            }
            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<3;l++)
              {
                m(j+indexFieldfV*nbFFp+nbdofm) += ( l10pJumpN(l)*Gradsp[j+indexFieldfV*nbFFp][l])*(wJ/2.);
                m(j+indexFieldfV*nbFFp+nbdofm) += ( l20pftJumpN(l)*Gradsp[j+indexFieldfV*nbFFp][l])*(wJ/2.);
              }
            }
          }

          else //broken case
          {

          }
        }
      }
    }
  }
}*/


/*void ElecTherMechDG3DStiffnessInter::get(MElement *ele,int npts,IntPt *GP,fullMatrix<double> &stiff) const
{
  dG3DStiffnessInter::get(ele,npts,GP,stiff);

  MInterfaceElement *ie = dynamic_cast<MInterfaceElement*>(ele);
  const int nbFFm = ie->getElem(0)->getNumVertices();
  const int nbFFp = ie->getElem(1)->getNumVertices();
  const int nbdofm = _minusSpace->getNumKeys(ie->getElem(0));
  const int nbdofp = _plusSpace->getNumKeys(ie->getElem(1));
  if(_fullDg)
  {
    #define gamma 0
    // characteristic size
    const double hs = ie->getCharacteristicSize();
    const double nbetahs = _ElecTherMechBeta/hs;
    if( _continuity)
    {
      // get value at gauss's point
      const AllIPState::ipstateElementContainer *vips = _ipf->getAips()->getIPstate(ie->getNum());
      // values of Gauss points on minus and plus elements
      IntPt *GPm; IntPt *GPp;
      _interQuad->getIntPoints(ie,GP,&GPm,&GPp);

      // loop on Gauss Point
      for(int i=0;i<npts; i++)
      {
        // IP
        const IPStateBase *ipsm        = (*vips)[i];
        const IPStateBase *ipsp        = (*vips)[i+npts];
        const ElecTherMechDG3DIPVariableBase *ipvm;     //= static_cast<const nonLocalDamageDG3DIPVariable*>(ipsm->getState(IPStateBase::current));
        const ElecTherMechDG3DIPVariableBase *ipvp;     // = static_cast<const nonLocalDamageDG3DIPVariable*>(ipsp->getState(IPStateBase::current));
        if(_mlawMinus->getType() == materialLaw::fracture )
        {
          const FractureCohesive3DIPVariable* ipvtmp = static_cast<const FractureCohesive3DIPVariable*>(ipsm->getState(IPStateBase::current));
          ipvm = static_cast<const ElecTherMechDG3DIPVariableBase*>(ipvtmp->getIPvBulk());
          ipvtmp = static_cast<const FractureCohesive3DIPVariable*>(ipsp->getState(IPStateBase::current));
          ipvp = static_cast<const ElecTherMechDG3DIPVariableBase*>(ipvtmp->getIPvBulk());
        }
        else
        {
          ipvm = static_cast<const ElecTherMechDG3DIPVariableBase*>(ipsm->getState(IPStateBase::current));
          ipvp = static_cast<const ElecTherMechDG3DIPVariableBase*>(ipsp->getState(IPStateBase::current));
        }
        const double weight = GP[i].weight;
        double detJ = ipvm->getJacobianDeterminant(ie,GP[i]);
        double wJ = detJ* weight;

        const SVector3 &referenceOutwardNormalm = ipvm->getConstRefToReferenceOutwardNormal();
        const SVector3 &referenceOutwardNormalp = ipvp->getConstRefToReferenceOutwardNormal();


        if(getConstitutiveExtraDofDiffusionUseEnergyConjugatedField())
        {
          const double &ftJump = ipvp->getConstRefToOneOverTemperatureJump();
          const double &fvJump = ipvp->getConstRefTofvJump();

          const std::vector<TensorialTraits<double>::GradType> &Gradsm = ipvm->gradf(_minusSpace,ie->getElem(0),GPm[i]);
          const std::vector<TensorialTraits<double>::GradType> &Gradsp = ipvp->gradf(_plusSpace,ie->getElem(1),GPp[i]);

          // possibility of fracture
          bool broken = false;
          if(_mlawMinus->getType() == materialLaw::fracture )
          {
            const FractureCohesive3DIPVariable *ipvwf = static_cast<const FractureCohesive3DIPVariable*>(ipsm->getState(IPStateBase::current)); // broken via minus (OK broken on both sides)
            broken = ipvwf->isbroken();
          }

          static SVector3 nu;
          nu = referenceOutwardNormalm;
          nu*=(1./referenceOutwardNormalm.norm());


          const std::vector<TensorialTraits<double>::ValType> &Valsm = ipvm->f(_minusSpace,ie->getElem(0),GPm[i]);
          const std::vector<TensorialTraits<double>::ValType> &Valsp = ipvp->f(_plusSpace,ie->getElem(1),GPp[i]);


          static STensor3  Stiff_alphadialitationm;
          Stiff_alphadialitationm = *(ipvm->getConstRefToStiff_alphaDilatation());
          static STensor3  Stiff_alphadialitationp;
          Stiff_alphadialitationp = *(ipvp->getConstRefToStiff_alphaDilatation());

          const double &fTp = ipvp->getConstRefTofT();
          const double &fTm = ipvm->getConstRefTofT();

          Stiff_alphadialitationm*= (-1.)/(fTm*fTm);
          Stiff_alphadialitationp*= (-1.)/(fTp*fTp);
          const SVector3 &jump = ipvp->getConstRefToJump();

          if(broken)
          {

          }
          else
          {
            const STensor3  *dPdTm     = &(ipvm->getConstRefTodPdT());
            const STensor3  *dPdTp     = &(ipvp->getConstRefTodPdT());
            const STensor33 *djedFm    = &(ipvm->getConstRefTodFluxVdF());
            const STensor33 *djedFp    = &(ipvp->getConstRefTodFluxVdF());
            const STensor3  *dPdVm     = &(ipvm->getConstRefTodPdV());
            const STensor3  *dPdVp     = &(ipvp->getConstRefTodPdV());

            const STensor3  *l10m      = (ipvm->getConstRefToElecConductivity());
            const STensor3  *l10p      = (ipvp->getConstRefToElecConductivity());
            const STensor3  *l20m      = (ipvm->getConstRefToCrossElecConductivity());
            const STensor3  *l20p      = (ipvp->getConstRefToCrossElecConductivity());

            const STensor3  *dl10dTm   = &(ipvm->getConstRefTodElecConductivitydT());
            const STensor3  *dl10dTp   = &(ipvp->getConstRefTodElecConductivitydT());
            const STensor3  *dl20dTm   = &(ipvm->getConstRefTodCrossElecConductivitydT());
            const STensor3  *dl20dTp   = &(ipvp->getConstRefTodCrossElecConductivitydT());

            const STensor3  *dl20dvm   = &(ipvm->getConstRefTodCrossElecConductivitydv());
            const STensor3  *dl20dvp   = &(ipvp->getConstRefTodCrossElecConductivitydv());

            const STensor43 *dl10dFm   = &(ipvm->getConstRefTodElecConductivitydF());
            const STensor43 *dl10dFp   = &(ipvp->getConstRefTodElecConductivitydF());
            const STensor43 *dl20dFm   = &(ipvm->getConstRefTodCrossElecConductivitydF());
            const STensor43 *dl20dFp   = &(ipvp->getConstRefTodCrossElecConductivitydF());

            const STensor3  *djedGradVm = &(ipvm->getConstRefTodFluxVdGradV());
            const STensor3  *djedGradVp = &(ipvp->getConstRefTodFluxVdGradV());
            const SVector3  *djedVm     = &(ipvm->getConstRefTodFluxVdV());
            const SVector3  *djedVp     = &(ipvp->getConstRefTodFluxVdV());
            const STensor3  *djedGradTm = &(ipvm->getConstRefTodFluxVdGradT());
            const STensor3  *djedGradTp = &(ipvp->getConstRefTodFluxVdGradT());
            const SVector3  *djedTm     = &(ipvm->getConstRefTodFluxVdT());
            const SVector3  *djedTp     = &(ipvp->getConstRefTodFluxVdT());

            const STensor3 *dgardVdgradfVm = &(ipvm->getConstRefTodgradVdgradfV());
            const STensor3 *dgardVdgradfTm = &(ipvm->getConstRefTodgradVdgradfT());
            const STensor3 *dgardTdgradfTm = &(ipvm->getConstRefTodgradTdgradfT());
            const STensor3 *dgardTdgradfVm = &(ipvm->getConstRefTodgradTdgradfV());

            const double   &dVdfVm         = (ipvm->getConstRefTodVdfV());
            const double   &dVdfTm         = (ipvm->getConstRefTodVdfT());
            const double   &dTdfTm         = (ipvm->getConstRefTodTdfT());
            const double   &dTdfVm         = (ipvm->getConstRefTodTdfV());

            const SVector3 *dgradVdfVm     = &(ipvm->getConstRefTodgradVdfV());
            const SVector3 *dgradVdfTm     = &(ipvm->getConstRefTodgradVdfT());
            const SVector3 *dgradTdfTm     = &(ipvm->getConstRefTodgradTdfT());
            const SVector3 *dgradTdfVm     = &(ipvm->getConstRefTodgradTdfV());

            const STensor3 *dgardVdgradfVp = &(ipvp->getConstRefTodgradVdgradfV());
            const STensor3 *dgardVdgradfTp = &(ipvp->getConstRefTodgradVdgradfT());
            const STensor3 *dgardTdgradfTp = &(ipvp->getConstRefTodgradTdgradfT());
            const STensor3 *dgardTdgradfVp = &(ipvp->getConstRefTodgradTdgradfV());

            const double   &dVdfVp         = (ipvp->getConstRefTodVdfV());
            const double   &dVdfTp         = (ipvp->getConstRefTodVdfT());
            const double   &dTdfVp         = (ipvp->getConstRefTodTdfV());
            const double   &dTdfTp         = (ipvp->getConstRefTodTdfT());

            const SVector3 *dgradVdfVp     = &(ipvp->getConstRefTodgradVdfV());
            const SVector3 *dgradVdfTp     = &(ipvp->getConstRefTodgradVdfT());
            const SVector3 *dgradTdfTp     = &(ipvp->getConstRefTodgradTdfT());
            const SVector3 *dgradTdfVp     = &(ipvp->getConstRefTodgradTdfV());

          //  const STensor3  *jy1m	= (ipvm->getConstRefToEnergyConductivity());
          //  const STensor43 *djy1dFm    = &(ipvm->getConstRefTodEnergyConductivitydF());
          //  const STensor3  *djy1dVm	= &(ipvm->getConstRefTodEnergyConductivitydV());
          //  const STensor3  *djy1dTm	= &(ipvm->getConstRefTodEnergyConductivitydT());

          //  const STensor3  *jy1p	= (ipvp->getConstRefToEnergyConductivity());
          //  const STensor43 *djy1dFp    = &(ipvp->getConstRefTodEnergyConductivitydF());
          //  const STensor3  *djy1dVp	= &(ipvp->getConstRefTodEnergyConductivitydV());
          //  const STensor3  *djy1dTp	= &(ipvp->getConstRefTodEnergyConductivitydT());  

            const STensor3  *jy1m	= (ipvm->getConstRefToTherConductivity());
            const STensor43 *djy1dFm    = &(ipvm->getConstRefTodTherConductivitydF());
            const STensor3  *djy1dVm	= &(ipvm->getConstRefTodTherConductivitydv());
            const STensor3  *djy1dTm	= &(ipvm->getConstRefTodTherConductivitydT());

            const STensor3  *jy1p	= (ipvp->getConstRefToTherConductivity());
            const STensor43 *djy1dFp    = &(ipvp->getConstRefTodTherConductivitydF());
            const STensor3  *djy1dVp	= &(ipvp->getConstRefTodTherConductivitydv());
            const STensor3  *djy1dTp	= &(ipvp->getConstRefTodTherConductivitydT()); 

        //    static STensor3 jy1m, jy1m_1;
            STensorOperation::zero(jy1m); STensorOperation::zero(jy1m_1);
            STensorOperation::multSTensor3(ipvm->getConstRefTodFluxdGradField()[0][0],ipvm->getConstRefTodGradFielddGradEnergyConjugatedField()[0][0],jy1m_1);
            STensorOperation::multSTensor3(ipvm->getConstRefTodFluxdGradField()[0][1],ipvm->getConstRefTodGradFielddGradEnergyConjugatedField()[1][0],jy1m);
            jy1m += jy1m_1;
            static STensor43 djy1dFm, djy1dFp, djydgradVdFm, djydgradTdFm, djydgradVdFp, djydgradTdFp;
            static STensor3 dgradVdgradfTm, dgradTdgradfTm, dgradVdgradfTp, dgradTdgradfTp; 
            djydgradVdFm=ipvm->getConstRefTodFluxdGradFielddF()[0][1];
            djydgradTdFm=ipvm->getConstRefTodFluxdGradFielddF()[0][0];
            dgradVdgradfTm=ipvm->getConstRefTodGradFielddGradEnergyConjugatedField()[1][0];
            dgradTdgradfTm=ipvm->getConstRefTodGradFielddGradEnergyConjugatedField()[0][0];
            djydgradVdFp=ipvp->getConstRefTodFluxdGradFielddF()[0][1];
            djydgradTdFp=ipvp->getConstRefTodFluxdGradFielddF()[0][0];
            dgradVdgradfTp=ipvp->getConstRefTodGradFielddGradEnergyConjugatedField()[1][0];
            dgradTdgradfTp=ipvp->getConstRefTodGradFielddGradEnergyConjugatedField()[0][0];
            STensorOperation::zero(djy1dFm);
            STensorOperation::zero(djy1dFp);

             for (int i=0; i<3; i++){
              for (int j=0; j<3; j++){
                for (int k=0; k<3; k++){
                 for (int l=0; l<3; l++){
                  for (int q=0; q<3; q++){
                 			  djy1dFm(i,j,k,l) += djydgradVdFm(i,q,k,l)*dgradVdgradfTm(q,j);
                 			  djy1dFm(i,j,k,l) += djydgradTdFm(i,q,k,l)*dgradTdgradfTm(q,j);
                 			  djy1dFp(i,j,k,l) += djydgradVdFp(i,q,k,l)*dgradVdgradfTp(q,j);
                 			  djy1dFp(i,j,k,l) += djydgradTdFp(i,q,k,l)*dgradTdgradfTp(q,j);
                                         }
                                        }
                                       }
                                      }
                                     }

            static STensor3 djy1dVm, djy1dVm_1;
            STensorOperation::zero(djy1dVm); STensorOperation::zero(djy1dVm_1);
            STensorOperation::multSTensor3(ipvm->getConstRefTodFluxdGradFielddField()[0][0][1],ipvm->getConstRefTodGradFielddGradEnergyConjugatedField()[0][0],djy1dVm_1);
            STensorOperation::multSTensor3(ipvm->getConstRefTodFluxdGradFielddField()[0][1][1],ipvm->getConstRefTodGradFielddGradEnergyConjugatedField()[1][0],djy1dVm);
            djy1dVm += djy1dVm_1;
            static STensor3 djy1dTm, djy1dTm_1;
            STensorOperation::zero(djy1dTm); STensorOperation::zero(djy1dTm_1);
            STensorOperation::multSTensor3(ipvm->getConstRefTodFluxdGradFielddField()[0][0][0],ipvm->getConstRefTodGradFielddGradEnergyConjugatedField()[0][0],djy1dTm_1);
            STensorOperation::multSTensor3(ipvm->getConstRefTodFluxdGradFielddField()[0][1][0],ipvm->getConstRefTodGradFielddGradEnergyConjugatedField()[1][0],djy1dTm);
            djy1dTm += djy1dTm_1;

            static STensor3 jy1p, jy1p_1;
            STensorOperation::multSTensor3(ipvp->getConstRefTodFluxdGradField()[0][0],ipvp->getConstRefTodGradFielddGradEnergyConjugatedField()[0][0],jy1p_1);
            STensorOperation::multSTensor3(ipvp->getConstRefTodFluxdGradField()[0][1],ipvp->getConstRefTodGradFielddGradEnergyConjugatedField()[1][0],jy1p);
            jy1p += jy1p_1;

            static STensor3 djy1dVp, djy1dVp_1;
            STensorOperation::zero(djy1dVp); STensorOperation::zero(djy1dVp_1);
            STensorOperation::multSTensor3(ipvp->getConstRefTodFluxdGradFielddField()[0][0][1],ipvp->getConstRefTodGradFielddGradEnergyConjugatedField()[0][0],djy1dVp_1);
            STensorOperation::multSTensor3(ipvp->getConstRefTodFluxdGradFielddField()[0][1][1],ipvp->getConstRefTodGradFielddGradEnergyConjugatedField()[1][0],djy1dVp);
            djy1dVp += djy1dVp_1;
            static STensor3 djy1dTp, djy1dTp_1;
            STensorOperation::zero(djy1dTp); STensorOperation::zero(djy1dTp_1);
            STensorOperation::multSTensor3(ipvp->getConstRefTodFluxdGradFielddField()[0][0][0],ipvp->getConstRefTodGradFielddGradEnergyConjugatedField()[0][0],djy1dTp_1);
            STensorOperation::multSTensor3(ipvp->getConstRefTodFluxdGradFielddField()[0][1][0],ipvp->getConstRefTodGradFielddGradEnergyConjugatedField()[1][0],djy1dTp);
            djy1dTp += djy1dTp_1; //

            const STensor33 *djydFp	= &(ipvp->getConstRefTodFluxEnergydF());
            const STensor3  *djydGradTp	= &(ipvp->getConstRefTodFluxEnergydGradT());
            const SVector3  *djydTp	= &(ipvp->getConstRefTodFluxEnergydT());
            const STensor3  *djydGradVp	= &(ipvp->getConstRefTodFluxEnergydGradV());
            const SVector3  *djydVp	= &(ipvp->getConstRefTodFluxEnergydV());
            const SVector3  *fluxjyp	= &(ipvp->getConstRefToFluxEnergy());

            const STensor33 *djydFm	= &(ipvm->getConstRefTodFluxEnergydF());
            const STensor3  *djydGradTm	= &(ipvm->getConstRefTodFluxEnergydGradT());
            const SVector3  *djydTm	= &(ipvm->getConstRefTodFluxEnergydT());
            const STensor3  *djydGradVm	= &(ipvm->getConstRefTodFluxEnergydGradV());
            const SVector3  *djydVm	= &(ipvm->getConstRefTodFluxEnergydV());
            const SVector3  *fluxjym	= &(ipvm->getConstRefToFluxEnergy());




            double eqRatio = _ElecTherMechEqRatio;
            wJ*=eqRatio;
            // Assembly consistency (from derivative of sigma with extra field)
            int indexFieldfT=3+getNumNonLocalVariable();
            int indexFieldfV=3+getNumNonLocalVariable()+1;
          for(int j=0;j<nbFFm;j++)
            {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+k*nbFFm,l+indexFieldfT*nbFFm) -= (dPdTm->operator()(k,q)*dTdfTm+dPdVm->operator()(k,q)*dVdfTm)*(Valsm[l+indexFieldfT*nbFFm]*Valsm[j+0*nbFFm]*wJ/2.*nu(q)/eqRatio);

                    stiff(j+k*nbFFm,l+indexFieldfV*nbFFm) -= (dPdVm->operator()(k,q)*dVdfVm+dPdTm->operator()(k,q)*dTdfVm)*(Valsm[l+indexFieldfV*nbFFm]*Valsm[j+0*nbFFm]*wJ/2.*nu(q)/eqRatio);
                  }
                }
                for(int l=0;l<nbFFp;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+k*nbFFm,l+indexFieldfT*nbFFp+nbdofm) -= (dPdTp->operator()(k,q)*dTdfTp+dPdVp->operator()(k,q)*dVdfTp)*(Valsp[l+indexFieldfT*nbFFp]*Valsm[j+0*nbFFm]*wJ/2.*nu(q)/eqRatio);

                    stiff(j+k*nbFFm,l+indexFieldfV*nbFFp+nbdofm) -= (dPdVp->operator()(k,q)*dVdfVp+dPdTp->operator()(k,q)*dTdfVp)*(Valsp[l+indexFieldfV*nbFFp]*Valsm[j+0*nbFFm]*wJ/2.*nu(q)/eqRatio);
                  }
                }
              }
            }
            for(int j=0;j<nbFFp;j++)
            {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+k*nbFFp+nbdofm,l+indexFieldfT*nbFFm) += (dPdTm->operator()(k,q)*dTdfTm+dPdVm->operator()(k,q)*dVdfTm)*(Valsm[l+indexFieldfT*nbFFm]*Valsp[j+0*nbFFp]*wJ/2.*nu(q)/eqRatio);

                    stiff(j+k*nbFFp+nbdofm,l+indexFieldfV*nbFFm) += (dPdVm->operator()(k,q)*dVdfVm+dPdTm->operator()(k,q)*dTdfVm)*(Valsm[l+indexFieldfV*nbFFm]*Valsp[j+0*nbFFp]*wJ/2.*nu(q)/eqRatio);
                  }
                }
                for(int l=0;l<nbFFp;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+k*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += (dPdTp->operator()(k,q)*dTdfTp+dPdVp->operator()(k,q)*dVdfTp)*(Valsp[l+indexFieldfT*nbFFp]*Valsp[j+0*nbFFp]*wJ/2.*nu(q)/eqRatio);

                    stiff(j+k*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += (dPdVp->operator()(k,q)*dVdfVp+dPdTp->operator()(k,q)*dTdfVp)*(Valsp[l+indexFieldfV*nbFFp]*Valsp[j+0*nbFFp]*wJ/2.*nu(q)/eqRatio);
                  }
                }
              }
            }


            // Assembly consistency (from derivative of flux with deformation gradient)
            for(int j=0;j<nbFFm;j++)
            {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    for(int N = 0; N< 3; N++)
                    {
                      stiff(j+indexFieldfT*nbFFm,l+k*nbFFm) -= djydFm->operator()(N,k,q)*(Gradsm[l+0*nbFFm][q]*Valsm[j+indexFieldfT*nbFFm]*wJ/2.*nu(N));
                    }
                  }
                }
                for(int l=0;l<nbFFp;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    for(int N = 0; N< 3; N++)
                    {
                      stiff(j+indexFieldfT*nbFFm,l+k*nbFFp+nbdofm) -= djydFp->operator()(N,k,q)*(Gradsp[l+0*nbFFp][q]*Valsm[j+indexFieldfT*nbFFm]*wJ/2.*nu(N));
                    }
                  }
                }
              }
            }
            for(int j=0;j<nbFFp;j++)
            {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    for(int N = 0; N< 3; N++)
                    {
                      stiff(j+indexFieldfT*nbFFp+nbdofm,l+k*nbFFm) += djydFm->operator()(N,k,q)*(Gradsm[l+0*nbFFm][q]*Valsp[j+indexFieldfT*nbFFp]*wJ/2.*nu(N));
                    }
                  }
                }
                for(int l=0;l<nbFFp;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    for(int N = 0; N< 3; N++)
                    {
                      stiff(j+indexFieldfT*nbFFp+nbdofm,l+k*nbFFp+nbdofm) += djydFp->operator()(N,k,q)*(Gradsp[l+0*nbFFp][q]*Valsp[j+indexFieldfT*nbFFp]*wJ/2.*nu(N));
                    }
                  }
                }
              }
            }
              //djedf
            for(int j=0;j<nbFFm;j++)
            {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    for(int N = 0; N< 3; N++)
                    {
                      stiff(j+indexFieldfV*nbFFm,l+k*nbFFm) -= djedFm->operator()(N,k,q)*(Gradsm[l+0*nbFFm][q]*Valsm[j+indexFieldfV*nbFFm]*wJ/2.*nu(N));
                    }
                  }
                }
                for(int l=0;l<nbFFp;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    for(int N = 0; N< 3; N++)
                    {
                      stiff(j+indexFieldfV*nbFFm,l+k*nbFFp+nbdofm) -= djedFp->operator()(N,k,q)*(Gradsp[l+0*nbFFp][q]*Valsm[j+indexFieldfV*nbFFm]*wJ/2.*nu(N));
                    }
                  }
                }
              }
            }
            for(int j=0;j<nbFFp;j++)
            {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    for(int N = 0; N< 3; N++)
                    {
                      stiff(j+indexFieldfV*nbFFp+nbdofm,l+k*nbFFm) += djedFm->operator()(N,k,q)*(Gradsm[l+0*nbFFm][q]*Valsp[j+indexFieldfV*nbFFp]*wJ/2.*nu(N));
                    }
                  }
                }
                for(int l=0;l<nbFFp;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    for(int N = 0; N< 3; N++)
                    {
                      stiff(j+indexFieldfV*nbFFp+nbdofm,l+k*nbFFp+nbdofm) += djedFp->operator()(N,k,q)*(Gradsp[l+0*nbFFp][q]*Valsp[j+indexFieldfV*nbFFp]*wJ/2.*nu(N));
                    }
                  }
                }
              }
            }

             // consistency D\mean{q} DT
            //dfI1tdft+dfI1tdfv
           for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p = 0; p< 3; p++)
                {
                  stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFm) -= (djydTm->operator()(p)*dTdfTm+djydVm->operator()(p)*dVdfTm)*(Valsm[l+indexFieldfT*nbFFm]*Valsm[j+indexFieldfT*nbFFm]*wJ/2.*nu(p));
                  stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFm) -= (djydVm->operator()(p)*dVdfVm+djydTm->operator()(p)*dTdfVm)*(Valsm[l+indexFieldfV*nbFFm]*Valsm[j+indexFieldfT*nbFFm]*wJ/2.*nu(p));
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFm) -=(djydGradTm->operator()(p,q)*dgradTdfTm->operator()(q)+djydGradVm->operator()(p,q)*dgradVdfTm->operator()(q))
                    *(Valsm[l+indexFieldfT*nbFFm]*Valsm[j+indexFieldfT*nbFFm]*wJ/2.*nu(p));
                    stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFm) -=(djydGradTm->operator()(p,q)*dgradTdfVm->operator()(q)+djydGradVm->operator()(p,q)*dgradVdfVm->operator()(q))
                    *(Valsm[j+indexFieldfT*nbFFm]*Valsm[l+indexFieldfV*nbFFm]*wJ/2.*nu(p));

                    for(int o = 0; o< 3; o++)
                    {
                      stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFm) -= (djydGradTm->operator()(p,o)*dgardTdgradfTm->operator()(o,q)+djydGradVm->operator()(p,o)*dgardVdgradfTm->operator()(o,q))
                      *Valsm[j+indexFieldfT*nbFFm]*(Gradsm[l+indexFieldfT*nbFFm][q]*wJ/2.*nu(p));
                      stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFm) -= (djydGradVm->operator()(p,o)*dgardVdgradfVm->operator()(o,q)+djydGradTm->operator()(p,o)*dgardTdgradfVm->operator()(o,q))
                      *Valsm[j+indexFieldfT*nbFFm]*(Gradsm[l+indexFieldfV*nbFFm][q]*wJ/2.*nu(p));
                    }
                  }
                }
              }

              for(int l=0;l<nbFFp;l++)
              {
                for(int p = 0; p< 3; p++)
                {
                  stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFp+nbdofm) -= (djydTp->operator()(p)*dTdfTp+djydVp->operator()(p)*dVdfTp)*(Valsp[l+indexFieldfT*nbFFp]*Valsm[j+indexFieldfT*nbFFm]*wJ/2.*nu(p));
                  stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFp+nbdofm) -= (djydVp->operator()(p)*dVdfVp+djydTp->operator()(p)*dTdfVp)*(Valsp[l+indexFieldfV*nbFFp]*Valsm[j+indexFieldfT*nbFFm]*wJ/2.*nu(p));
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFp+nbdofm) -= (djydGradTp->operator()(p,q)*dgradTdfTp->operator()(q)+djydGradVp->operator()(p,q)*dgradVdfTp->operator()(q))
                    *Valsm[j+indexFieldfT*nbFFm]*(Valsp[l+indexFieldfT*nbFFp]*wJ/2.*nu(p));
                    stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFp+nbdofm) -= (djydGradTp->operator()(p,q)*dgradTdfVp->operator()(q)+djydGradVp->operator()(p,q)*dgradVdfVp->operator()(q))
                    *Valsm[j+indexFieldfT*nbFFm]*(Valsp[l+indexFieldfV*nbFFp]*wJ/2.*nu(p));

                    for(int o = 0; o< 3; o++)
                    {
                      stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFp+nbdofm) -=(djydGradTp->operator()(p,o)*dgardTdgradfTp->operator()(o,q)+djydGradVp->operator()(p,o)*dgardVdgradfTp->operator()(o,q))
                      *Valsm[j+indexFieldfT*nbFFm]*(Gradsp[l+indexFieldfT*nbFFp][q]*wJ/2.*nu(p));
                      stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFp+nbdofm) -=(djydGradVp->operator()(p,o)*dgardVdgradfVp->operator()(o,q)+djydGradTp->operator()(p,o)*dgardTdgradfVp->operator()(o,q))
                      *Valsm[j+indexFieldfT*nbFFm]*(Gradsp[l+indexFieldfV*nbFFp][q]*wJ/2.*nu(p));
                    }
                  }
                }
              }
            }

            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p = 0; p< 3; p++)
                {
                  stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFm) += (djydTm->operator()(p)*dTdfTm+djydVm->operator()(p)*dVdfTm)*(Valsm[l+indexFieldfT*nbFFm]*Valsp[j+indexFieldfT*nbFFp]*wJ/2.*nu(p));
                  stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFm) += (djydVm->operator()(p)*dVdfVm+djydTm->operator()(p)*dTdfVm)*(Valsm[l+indexFieldfV*nbFFm]*Valsp[j+indexFieldfT*nbFFp]*wJ/2.*nu(p));
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFm) += (djydGradTm->operator()(p,q)*dgradTdfTm->operator()(q)+djydGradVm->operator()(p,q)*dgradVdfTm->operator()(q))
                            *Valsp[j+indexFieldfT*nbFFp]*(Valsm[l+indexFieldfT*nbFFm]*wJ/2.*nu(p));
                    stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFm) += (djydGradTm->operator()(p,q)*dgradTdfVm->operator()(q)+djydGradVm->operator()(p,q)*dgradVdfVm->operator()(q))
                            *Valsp[j+indexFieldfT*nbFFp]*(Valsm[l+indexFieldfV*nbFFm]*wJ/2.*nu(p));
                    for(int o = 0; o< 3; o++)
                    {
                      stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFm) += (djydGradTm->operator()(p,o)*dgardTdgradfTm->operator()(o,q)+djydGradVm->operator()(p,o)*dgardVdgradfTm->operator()(o,q))
                              *Valsp[j+indexFieldfT*nbFFp]*(Gradsm[l+indexFieldfT*nbFFm][q]*wJ/2.*nu(p));
                      stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFm) += (djydGradVm->operator()(p,o)*dgardVdgradfVm->operator()(o,q)+djydGradTm->operator()(p,o)*dgardTdgradfVm->operator()(o,q))
                               *Valsp[j+indexFieldfT*nbFFp]*(Gradsm[l+indexFieldfV*nbFFm][q]*wJ/2.*nu(p));
                    }
                  }
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p = 0; p< 3; p++)
                {
                  stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += (djydTp->operator()(p)*dTdfTp+djydVp->operator()(p)*dVdfTp)*(Valsp[l+indexFieldfT*nbFFp]*Valsp[j+indexFieldfT*nbFFp]*wJ/2.*nu(p));
                  stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += (djydVp->operator()(p)*dVdfVp+djydTp->operator()(p)*dTdfVp)*(Valsp[l+indexFieldfV*nbFFp]*Valsp[j+indexFieldfT*nbFFp]*wJ/2.*nu(p));
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += (djydGradTp->operator()(p,q)*dgradTdfTp->operator()(q)+djydGradVp->operator()(p,q)*dgradVdfTp->operator()(q))
                              *Valsp[j+indexFieldfT*nbFFp]*(Valsp[l+indexFieldfT*nbFFp]*wJ/2.*nu(p));
                    stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += (djydGradTp->operator()(p,q)*dgradTdfVp->operator()(q)+djydGradVp->operator()(p,q)*dgradVdfVp->operator()(q))
                              *Valsp[j+indexFieldfT*nbFFp]*(Valsp[l+indexFieldfV*nbFFp]*wJ/2.*nu(p));
                    for(int o = 0; o< 3; o++)
                    {
                      stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm)+=(djydGradTp->operator()(p,o)*dgardTdgradfTp->operator()(o,q)+djydGradVp->operator()(p,o)*dgardVdgradfTp->operator()(o,q))
                              *Valsp[j+indexFieldfT*nbFFp]*(Gradsp[l+indexFieldfT*nbFFp][q]*wJ/2.*nu(p));
                      stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm)+=(djydGradVp->operator()(p,o)*dgardVdgradfVp->operator()(o,q)+djydGradTp->operator()(p,o)*dgardTdgradfVp->operator()(o,q))
                              *Valsp[j+indexFieldfT*nbFFp]*(Gradsp[l+indexFieldfV*nbFFp][q]*wJ/2.*nu(p));
                    }
                  }
                }
              }
            }


            //dfI1vdv+dfI1vdt
            for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p = 0; p< 3; p++)
                {
                  stiff(j+indexFieldfV*nbFFm,l+indexFieldfV*nbFFm) -= (djedVm->operator()(p)*dVdfVm+djedTm->operator()(p)*dTdfVm)*(Valsm[l+indexFieldfV*nbFFm]*Valsm[j+indexFieldfV*nbFFm]*wJ/2.*nu(p));
                  stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFm) -= (djedTm->operator()(p)*dTdfTm+djedVm->operator()(p)*dVdfTm)*(Valsm[l+indexFieldfT*nbFFm]*Valsm[j+indexFieldfV*nbFFm]*wJ/2.*nu(p));
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+indexFieldfV*nbFFm,l+indexFieldfV*nbFFm) -=(djedGradVm->operator()(p,q)*dgradVdfVm->operator()(q)+djedGradTm->operator()(p,q)*dgradTdfVm->operator()(q))
                          *(Valsm[l+indexFieldfV*nbFFm]*Valsm[j+indexFieldfV*nbFFm]*wJ/2.*nu(p));
                    stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFm) -=(djedGradVm->operator()(p,q)*dgradVdfTm->operator()(q)+djedGradTm->operator()(p,q)*dgradTdfTm->operator()(q))
                           *Valsm[j+indexFieldfV*nbFFm]*(Valsm[l+indexFieldfT*nbFFm]*wJ/2.*nu(p));
                    for(int o = 0; o< 3; o++)
                    {
                      stiff(j+indexFieldfV*nbFFm,l+indexFieldfV*nbFFm) -=(djedGradVm->operator()(p,o)*dgardVdgradfVm->operator()(o,q)+djedGradTm->operator()(p,o)*dgardTdgradfVm->operator()(o,q))
                            *Valsm[j+indexFieldfV*nbFFm]*(Gradsm[l+indexFieldfV*nbFFm][q]*wJ/2.*nu(p));
                      stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFm) -=(djedGradTm->operator()(p,o)*dgardTdgradfTm->operator()(o,q)+djedGradVm->operator()(p,o)*dgardVdgradfTm->operator()(o,q))
                            *Valsm[j+indexFieldfV*nbFFm]*(Gradsm[l+indexFieldfT*nbFFm][q]*wJ/2.*nu(p));
                    }
                  }
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p = 0; p< 3; p++)
                {
                  stiff(j+indexFieldfV*nbFFm,l+indexFieldfV*nbFFp+nbdofm) -= (djedVp->operator()(p)*dVdfVp+djedTp->operator()(p)*dTdfVp)*(Valsp[l+indexFieldfV*nbFFp]*Valsm[j+indexFieldfV*nbFFm]*wJ/2.*nu(p));
                  stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFp+nbdofm) -= (djedTp->operator()(p)*dTdfTp+djedVp->operator()(p)*dVdfTp)*(Valsp[l+indexFieldfT*nbFFp]*Valsm[j+indexFieldfV*nbFFm]*wJ/2.*nu(p));
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+indexFieldfV*nbFFm,l+indexFieldfV*nbFFp+nbdofm) -=(djedGradVp->operator()(p,q)*dgradVdfVp->operator()(q)+djedGradTp->operator()(p,q)*dgradTdfVp->operator()(q))
                            *Valsm[j+indexFieldfV*nbFFm]*(Valsp[l+indexFieldfV*nbFFp]*wJ/2.*nu(p));
                    stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFp+nbdofm) -=(djedGradTp->operator()(p,q)*dgradTdfTp->operator()(q)+djedGradVp->operator()(p,q)*dgradVdfTp->operator()(q))
                            *Valsm[j+indexFieldfV*nbFFm]*(Valsp[l+indexFieldfT*nbFFp]*wJ/2.*nu(p));
                    for(int o = 0; o< 3; o++)
                    {
                      stiff(j+indexFieldfV*nbFFm,l+indexFieldfV*nbFFp+nbdofm) -=(djedGradVp->operator()(p,o)*dgardVdgradfVp->operator()(o,q)+djedGradTp->operator()(p,o)*dgardTdgradfVp->operator()(o,q))
                              *Valsm[j+indexFieldfV*nbFFm]*(Gradsp[l+indexFieldfV*nbFFp][q]*wJ/2.*nu(p));
                      stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFp+nbdofm) -=(djedGradTp->operator()(p,o)*dgardTdgradfTp->operator()(o,q)+djedGradVp->operator()(p,o)*dgardVdgradfTp->operator()(o,q))
                              *Valsm[j+indexFieldfV*nbFFm]*(Gradsp[l+indexFieldfT*nbFFp][q]*wJ/2.*nu(p));
                    }
                  }
                }
              }
            }

            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p = 0; p< 3; p++)
                {
                  stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfV*nbFFm) +=( djedVm->operator()(p)*dVdfVm+ djedTm->operator()(p)*dTdfVm)*(Valsm[l+indexFieldfV*nbFFm]*Valsp[j+indexFieldfV*nbFFp]*wJ/2.*nu(p));
                  stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFm) += (djedTm->operator()(p)*dTdfTm+djedVm->operator()(p)*dVdfTm)*(Valsm[l+indexFieldfT*nbFFm]*Valsp[j+indexFieldfV*nbFFp]*wJ/2.*nu(p));
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfV*nbFFm) +=(djedGradVm->operator()(p,q)*dgradVdfVm->operator()(q)+djedGradTm->operator()(p,q)*dgradTdfVm->operator()(q))
                            *Valsp[j+indexFieldfV*nbFFp] *(Valsm[l+indexFieldfV*nbFFm]*wJ/2.*nu(p));
                    stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFm) +=(djedGradVm->operator()(p,q)*dgradVdfTm->operator()(q)+djedGradTm->operator()(p,q)*dgradTdfTm->operator()(q))
                            *Valsp[j+indexFieldfV*nbFFp] *(Valsm[l+indexFieldfT*nbFFm]*wJ/2.*nu(p));
                    for(int o = 0; o< 3; o++)
                    {
                      stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfV*nbFFm) +=(djedGradVm->operator()(p,o)*dgardVdgradfVm->operator()(o,q)+djedGradTm->operator()(p,o)*dgardTdgradfVm->operator()(o,q))
                              *Valsp[j+indexFieldfV*nbFFp]*(Gradsm[l+indexFieldfV*nbFFm][q]*wJ/2.*nu(p));
                      stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFm) +=(djedGradTm->operator()(p,o)*dgardTdgradfTm->operator()(o,q)+djedGradVm->operator()(p,o)*dgardVdgradfTm->operator()(o,q))
                              *Valsp[j+indexFieldfV*nbFFp] *(Gradsm[l+indexFieldfT*nbFFm][q]*wJ/2.*nu(p));
                    }
                  }
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p = 0; p< 3; p++)
                {
                  stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += (djedVp->operator()(p)*dVdfVp+djedTp->operator()(p)*dTdfVp)*(Valsp[l+indexFieldfV*nbFFp]*Valsp[j+indexFieldfV*nbFFp]*wJ/2.*nu(p));
                  stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += (djedTp->operator()(p)*dTdfTp+djedVp->operator()(p)*dVdfTp)*(Valsp[l+indexFieldfT*nbFFp]*Valsp[j+indexFieldfV*nbFFp]*wJ/2.*nu(p));
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += (djedGradVp->operator()(p,q)*dgradVdfVp->operator()(q)+djedGradTp->operator()(p,q)*dgradTdfVp->operator()(q))
                             *(Valsp[l+indexFieldfV*nbFFp]*Valsp[j+indexFieldfV*nbFFp]*wJ/2.*nu(p));
                    stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += (djedGradTp->operator()(p,q)*dgradTdfTp->operator()(q)+djedGradVp->operator()(p,q)*dgradVdfTp->operator()(q))
                             *(Valsp[l+indexFieldfT*nbFFp]*Valsp[j+indexFieldfV*nbFFp]*wJ/2.*nu(p));
                    for(int o = 0; o< 3; o++)
                    {
                      stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += (djedGradVp->operator()(p,o)*dgardVdgradfVp->operator()(o,q)+djedGradTp->operator()(p,o)*dgardTdgradfVp->operator()(o,q))
                                *Valsp[j+indexFieldfV*nbFFp]*(Gradsp[l+indexFieldfV*nbFFp][q]*wJ/2.*nu(p));
                      stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += (djedGradTp->operator()(p,o)*dgardTdgradfTp->operator()(o,q)+djedGradVp->operator()(p,o)*dgardVdgradfTp->operator()(o,q))
                                *Valsp[j+indexFieldfV*nbFFp]*(Gradsp[l+indexFieldfT*nbFFp][q]*wJ/2.*nu(p));
                    }
                  }
                }
              }
            }

            // Stability  N \mean{k0} (N)
            double NBhsNdjy1dfTm=0.;double NBhsNdjy1dfTp=0.;  double NBhsNdL1dfTm=0.; double NBhsNdL1dfTp=0.;
            double NBhsNdL2dfTm=0.;double NBhsNdL2dfTp=0.; double NBhsNdjy1dfvm=0.; double NBhsNdjy1dfvp=0.; double NBhsNdl2dfvm=0.; double NBhsNdl2dfvp=0.;

            double NBhsNdjy1dTm=0.;double NBhsNdjy1dTp=0.; double NBhsNdL1dTm=0.; double NBhsNdL1dTp=0.;
            double NBhsNdL2dTm=0.;double NBhsNdL2dTp=0.; double NBhsNdjy1dvm=0.; double NBhsNdjy1dvp=0.; double NBhsNdL2dVm=0.; double NBhsNdL2dVp=0.;

            for(int l = 0; l <3; l++)
            {
              for(int n = 0; n <3; n++)
              {
                NBhsNdjy1dTm+=djy1dTm->operator()(l,n)*nu(n)*nbetahs*nu(l)/2.;
                NBhsNdjy1dTp+=djy1dTp->operator()(l,n)*nu(n)*nbetahs*nu(l)/2.;

                NBhsNdL1dTm+=dl10dTm->operator()(l,n)*nu(n)*nbetahs*nu(l)/2.;
                NBhsNdL1dTp+=dl10dTp->operator()(l,n)*nu(n)*nbetahs*nu(l)/2.;
                NBhsNdL2dTm+=dl20dTm->operator()(l,n)*nu(n)*nbetahs*nu(l)/2.;
                NBhsNdL2dTp+=dl20dTp->operator()(l,n)*nu(n)*nbetahs*nu(l)/2.;

                NBhsNdjy1dvm+=djy1dVm->operator()(l,n)*nu(n)*nbetahs*nu(l)/2.;
                NBhsNdjy1dvp+=djy1dVp->operator()(l,n)*nu(n)*nbetahs*nu(l)/2.;

                NBhsNdL2dVm+=dl20dvm->operator()(l,n)*nu(n)*nbetahs*nu(l)/2.;
                NBhsNdL2dVp+=dl20dvp->operator()(l,n)*nu(n)*nbetahs*nu(l)/2.;
              }
            }

            NBhsNdjy1dfTm = NBhsNdjy1dTm*dTdfTm+NBhsNdjy1dvm*dVdfTm;
            NBhsNdjy1dfTp = NBhsNdjy1dTp*dTdfTp+NBhsNdjy1dvp*dVdfTp;

            NBhsNdL1dfTm = NBhsNdL1dTm*dTdfTm;
            NBhsNdL1dfTp = NBhsNdL1dTp*dTdfTp;
            NBhsNdL2dfTm = NBhsNdL2dTm*dTdfTm+NBhsNdL2dVm*dVdfTm;
            NBhsNdL2dfTp = NBhsNdL2dTp*dTdfTp+NBhsNdL2dVp*dVdfTp;

            NBhsNdjy1dfvm =NBhsNdjy1dvm*dVdfVm+NBhsNdjy1dTm*dTdfVm;
            NBhsNdjy1dfvp =NBhsNdjy1dvp*dVdfVp+NBhsNdjy1dTp*dTdfVp;

            NBhsNdl2dfvm =NBhsNdL2dVm*dVdfVm+NBhsNdL2dTm*dTdfVm;
            NBhsNdl2dfvp =NBhsNdL2dVp*dVdfVp+NBhsNdL2dTp*dTdfVp;

            static STensor3 NBhsNdjy1dFm, NBhsNdjy1dFp, NBhsNdL1dFm,  NBhsNdL1dFp,  NBhsNdL2dFm,  NBhsNdL2dFp;
            for(int j = 0; j<3; j++)
            {
              for(int p = 0; p<3; p++)
              {
                NBhsNdjy1dFm(j,p)= 0.;
                NBhsNdjy1dFp(j,p) = 0.;
                NBhsNdL1dFm(j,p) = 0.;
                NBhsNdL1dFp(j,p) = 0.;
                NBhsNdL2dFm(j,p) = 0.;
                NBhsNdL2dFp(j,p) = 0.;

                for(int l = 0; l <3; l++)
                {
                  for(int n = 0; n <3; n++)
                  {
                    NBhsNdjy1dFm(j,p)+=djy1dFm->operator()(l,n,j,p)*nu(n)*nbetahs*nu(l)/2.;
                    NBhsNdjy1dFp(j,p)+=djy1dFp->operator()(l,n,j,p)*nu(n)*nbetahs*nu(l)/2.;

                    NBhsNdL1dFm(j,p)+=dl10dFm->operator()(l,n,j,p)*nu(n)*nbetahs*nu(l)/2.;
                    NBhsNdL1dFp(j,p)+=dl10dFp->operator()(l,n,j,p)*nu(n)*nbetahs*nu(l)/2.;
                    NBhsNdL2dFm(j,p)+=dl20dFm->operator()(l,n,j,p)*nu(n)*nbetahs*nu(l)/2.;
                    NBhsNdL2dFp(j,p)+=dl20dFp->operator()(l,n,j,p)*nu(n)*nbetahs*nu(l)/2.;
                  }
                }
              }
            }

            double NMeanjy1NBetasc=0.; double NMeanl1NBetasc=0.; double NMeanl2NBetasc=0.;
            for(int l = 0; l <3; l++)
            {
              for(int n = 0; n <3; n++)
              {
                NMeanjy1NBetasc += (jy1p->operator()(l,n)+jy1m->operator()(l,n))*nu(n)*nbetahs*nu(l)/2.;
                NMeanl1NBetasc  += (l10p->operator()(l,n)+l10m->operator()(l,n))*nu(n)*nbetahs*nu(l)/2.;
                NMeanl2NBetasc  += (l20p->operator()(l,n)+l20m->operator()(l,n))*nu(n)*nbetahs*nu(l)/2.;
              }
            }
            //new terms for dfTI3dfT
            for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFm) -= ftJump*NBhsNdjy1dfTm*Valsm[j+indexFieldfT*nbFFm]*Valsm[l+indexFieldfT*nbFFm]*wJ;
                stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFm) -= fvJump*NBhsNdL2dfTm*Valsm[j+indexFieldfT*nbFFm]*Valsm[l+indexFieldfT*nbFFm]*wJ;

                stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFm) += NMeanjy1NBetasc*Valsm[j+indexFieldfT*nbFFm]*Valsm[l+indexFieldfT*nbFFm]*wJ;
              }
              for(int l=0;l<nbFFp;l++)
              {
                stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFp+nbdofm) -= ftJump*NBhsNdjy1dfTp*Valsm[j+indexFieldfT*nbFFm]*Valsp[l+indexFieldfT*nbFFp]*wJ;
                stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFp+nbdofm) -= fvJump*NBhsNdL2dfTp*Valsm[j+indexFieldfT*nbFFm]*Valsp[l+indexFieldfT*nbFFp]*wJ;

                stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFp+nbdofm) -= NMeanjy1NBetasc*Valsm[j+indexFieldfT*nbFFm]*Valsp[l+indexFieldfT*nbFFp]*wJ;
              }
            }

            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFm) += ftJump*NBhsNdjy1dfTm*Valsp[j+indexFieldfT*nbFFp]*Valsm[l+indexFieldfT*nbFFm]*wJ;
                stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFm) += fvJump*NBhsNdL2dfTm *Valsp[j+indexFieldfT*nbFFp]*Valsm[l+indexFieldfT*nbFFm]*wJ;

                stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFm) -= NMeanjy1NBetasc*Valsp[j+indexFieldfT*nbFFp]*Valsm[l+indexFieldfT*nbFFm]*wJ;
              }
              for(int l=0;l<nbFFp;l++)
              {
                stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += ftJump*NBhsNdjy1dfTp*Valsp[j+indexFieldfT*nbFFp]*Valsp[l+indexFieldfT*nbFFp]*wJ;
                stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += fvJump*NBhsNdL2dfTp *Valsp[j+indexFieldfT*nbFFp]*Valsp[l+indexFieldfT*nbFFp]*wJ;

                stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += NMeanjy1NBetasc*Valsp[j+indexFieldfT*nbFFp]*Valsp[l+indexFieldfT*nbFFp]*wJ;
              }
            }

            // terms for dfTI3dfv
            for(int j=0;j<nbFFm;j++)
            {
            for(int l=0;l<nbFFm;l++)
            {
              stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFm) -= ftJump*NBhsNdjy1dfvm*Valsm[j+indexFieldfT*nbFFm]*Valsm[l+indexFieldfV*nbFFm]*wJ;
              stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFm) -= fvJump*NBhsNdl2dfvm *Valsm[j+indexFieldfT*nbFFm]*Valsm[l+indexFieldfV*nbFFm]*wJ;

              stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFm) += NMeanl2NBetasc*Valsm[j+indexFieldfT*nbFFm]*Valsm[l+indexFieldfV*nbFFm]*wJ;
            }
            for(int l=0;l<nbFFp;l++)
            {
              stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFp+nbdofm) -= ftJump*NBhsNdjy1dfvp*Valsm[j+indexFieldfT*nbFFm]*Valsp[l+indexFieldfV*nbFFp]*wJ;
              stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFp+nbdofm) -= fvJump*NBhsNdl2dfvp *Valsm[j+indexFieldfT*nbFFm]*Valsp[l+indexFieldfV*nbFFp]*wJ;

              stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFp+nbdofm) -= NMeanl2NBetasc*Valsm[j+indexFieldfT*nbFFm]*Valsp[l+indexFieldfV*nbFFp]*wJ;
            }
          }

          for(int j=0;j<nbFFp;j++)
          {
            for(int l=0;l<nbFFm;l++)
            {
              stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFm) += ftJump*NBhsNdjy1dfvm*Valsp[j+indexFieldfT*nbFFp]*Valsm[l+indexFieldfV*nbFFm]*wJ;
              stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFm) += fvJump*NBhsNdl2dfvm *Valsp[j+indexFieldfT*nbFFp]*Valsm[l+indexFieldfV*nbFFm]*wJ;

              stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFm) -= NMeanl2NBetasc*Valsp[j+indexFieldfT*nbFFp]*Valsm[l+indexFieldfV*nbFFm]*wJ;
            }
            for(int l=0;l<nbFFp;l++)
            {
              stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += ftJump*NBhsNdjy1dfvp*Valsp[j+indexFieldfT*nbFFp]*Valsp[l+indexFieldfV*nbFFp]*wJ;
              stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += fvJump*NBhsNdl2dfvp *Valsp[j+indexFieldfT*nbFFp]*Valsp[l+indexFieldfV*nbFFp]*wJ;

              stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) +=NMeanl2NBetasc*Valsp[j+indexFieldfT*nbFFp]*Valsp[l+indexFieldfV*nbFFp]*wJ;
            }
          }

          //Terms for dfvI3dfv
          for(int j=0;j<nbFFm;j++)
          {
            for(int l=0;l<nbFFm;l++)
            {
              stiff(j+indexFieldfV*nbFFm,l+indexFieldfV*nbFFm) -= ftJump*NBhsNdl2dfvm*Valsm[j+indexFieldfV*nbFFm]*Valsm[l+indexFieldfV*nbFFm]*wJ;

              stiff(j+indexFieldfV*nbFFm,l+indexFieldfV*nbFFm) += NMeanl1NBetasc*Valsm[j+indexFieldfV*nbFFm]*Valsm[l+indexFieldfV*nbFFm]*wJ;
            }
            for(int l=0;l<nbFFp;l++)
            {
              stiff(j+indexFieldfV*nbFFm,l+indexFieldfV*nbFFp+nbdofm) -= ftJump*NBhsNdl2dfvp*Valsm[j+indexFieldfV*nbFFm]*Valsp[l+indexFieldfV*nbFFp]*wJ;

              stiff(j+indexFieldfV*nbFFm,l+indexFieldfV*nbFFp+nbdofm) -= NMeanl1NBetasc*Valsm[j+indexFieldfV*nbFFm]*Valsp[l+indexFieldfV*nbFFp]*wJ;
            }
          }

          for(int j=0;j<nbFFp;j++)
          {
            for(int l=0;l<nbFFm;l++)
            {
              stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfV*nbFFm) += ftJump*NBhsNdl2dfvm*Valsp[j+indexFieldfV*nbFFp]*Valsm[l+indexFieldfV*nbFFm]*wJ;

              stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfV*nbFFm) -= NMeanl1NBetasc*Valsp[j+indexFieldfV*nbFFp]*Valsm[l+indexFieldfV*nbFFm]*wJ;
            }
            for(int l=0;l<nbFFp;l++)
            {
              stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += ftJump*NBhsNdl2dfvp*Valsp[j+indexFieldfV*nbFFp]*Valsp[l+indexFieldfV*nbFFp]*wJ;

              stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) +=NMeanl1NBetasc*Valsp[j+indexFieldfV*nbFFp]*Valsp[l+indexFieldfV*nbFFp]*wJ;
            }
          }


          for(int j=0;j<nbFFm;j++)
          {
            for(int l=0;l<nbFFm;l++)
            {
              stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFm) -= fvJump*NBhsNdL1dfTm*Valsm[j+indexFieldfV*nbFFm]*Valsm[l+indexFieldfT*nbFFm]*wJ;
              stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFm) -= ftJump*NBhsNdL2dfTm*Valsm[j+indexFieldfV*nbFFm]*Valsm[l+indexFieldfT*nbFFm]*wJ;

              stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFm) += NMeanl2NBetasc*Valsm[j+indexFieldfV*nbFFm]*Valsm[l+indexFieldfT*nbFFm]*wJ;
            }
            for(int l=0;l<nbFFp;l++)
            {
              stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFp+nbdofm) -= fvJump*NBhsNdL1dfTp*Valsm[j+indexFieldfV*nbFFm]*Valsp[l+indexFieldfT*nbFFp]*wJ;
              stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFp+nbdofm) -= ftJump*NBhsNdL2dfTp*Valsm[j+indexFieldfV*nbFFm]*Valsp[l+indexFieldfT*nbFFp]*wJ;

              stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFp+nbdofm) -= NMeanl2NBetasc*Valsm[j+indexFieldfV*nbFFm]*Valsp[l+indexFieldfT*nbFFp]*wJ;
            }
          }

          for(int j=0;j<nbFFp;j++)
          {
            for(int l=0;l<nbFFm;l++)
            {
            stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFm) += fvJump*NBhsNdL1dfTm*Valsp[j+indexFieldfV*nbFFp]*Valsm[l+indexFieldfT*nbFFm]*wJ;
            stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFm) += ftJump*NBhsNdL2dfTm*Valsp[j+indexFieldfV*nbFFp]*Valsm[l+indexFieldfT*nbFFm]*wJ;

              stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFm) -= NMeanl2NBetasc*Valsp[j+indexFieldfV*nbFFp]*Valsm[l+indexFieldfT*nbFFm]*wJ;

            }
            for(int l=0;l<nbFFp;l++)
            {
              stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += fvJump*NBhsNdL1dfTp*Valsp[j+indexFieldfV*nbFFp]*Valsp[l+indexFieldfT*nbFFp]*wJ;
              stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += ftJump*NBhsNdL2dfTp*Valsp[j+indexFieldfV*nbFFp]*Valsp[l+indexFieldfT*nbFFp]*wJ;

              stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += NMeanl2NBetasc*Valsp[j+indexFieldfV*nbFFp]*Valsp[l+indexFieldfT*nbFFp]*wJ;
            }
          }

          //F
           //new terms for dfvI3dF
          for(int j=0;j<nbFFm;j++)
          {
            for(int k=0;k<3;k++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p=0;p<3;p++)
                {
                  stiff(j+indexFieldfV*nbFFm,l+k*nbFFm) -= fvJump*NBhsNdL1dFm(k,p)*Gradsm[l+0*nbFFm][p]*Valsm[j+indexFieldfV*nbFFm]*wJ;
                  stiff(j+indexFieldfV*nbFFm,l+k*nbFFm) -= ftJump*NBhsNdL2dFm(k,p)*Gradsm[l+0*nbFFm][p]*Valsm[j+indexFieldfV*nbFFm]*wJ;

                  stiff(j+indexFieldfT*nbFFm,l+k*nbFFm) -= ftJump*NBhsNdjy1dFm(k,p)*Gradsm[l+0*nbFFm][p]*Valsm[j+indexFieldfT*nbFFm]*wJ;
                  stiff(j+indexFieldfT*nbFFm,l+k*nbFFm) -= fvJump*NBhsNdL2dFm(k,p)*Gradsm[l+0*nbFFm][p]*Valsm[j+indexFieldfT*nbFFm]*wJ;
                }
              }

              for(int l=0;l<nbFFp;l++)
              {
                for(int p=0;p<3;p++)
                {
                  stiff(j+indexFieldfV*nbFFm,l+k*nbFFp+nbdofm) -= fvJump*NBhsNdL1dFp(k,p)*Gradsp[l+0*nbFFp][p]*Valsm[j+indexFieldfV*nbFFm]*wJ;
                  stiff(j+indexFieldfV*nbFFm,l+k*nbFFp+nbdofm) -= ftJump*NBhsNdL2dFp(k,p)*Gradsp[l+0*nbFFp][p]*Valsm[j+indexFieldfV*nbFFm]*wJ;

                  stiff(j+indexFieldfT*nbFFm,l+k*nbFFp+nbdofm) -= ftJump*NBhsNdjy1dFp(k,p)*Gradsp[l+0*nbFFp][p]*Valsm[j+indexFieldfT*nbFFm]*wJ;
                  stiff(j+indexFieldfT*nbFFm,l+k*nbFFp+nbdofm) -= fvJump*NBhsNdL2dFp(k,p)*Gradsp[l+0*nbFFp][p]*Valsm[j+indexFieldfT*nbFFm]*wJ;
                }
              }
            }
          }

          for(int j=0;j<nbFFp;j++)
          {
            for(int k=0;k<3;k++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p=0;p<3;p++)
                {
                  stiff(j+indexFieldfV*nbFFp+nbdofm,l+k*nbFFm) += fvJump*NBhsNdL1dFm(k,p)*Gradsm[l+0*nbFFm][p]*Valsp[j+indexFieldfV*nbFFp]*wJ;
                  stiff(j+indexFieldfV*nbFFp+nbdofm,l+k*nbFFm) += ftJump*NBhsNdL2dFm(k,p)*Gradsm[l+0*nbFFm][p]*Valsp[j+indexFieldfV*nbFFp]*wJ;

                  stiff(j+indexFieldfT*nbFFp+nbdofm,l+k*nbFFm) += ftJump*NBhsNdjy1dFm(k,p)*Gradsm[l+0*nbFFm][p]*Valsp[j+indexFieldfT*nbFFp]*wJ;
                  stiff(j+indexFieldfT*nbFFp+nbdofm,l+k*nbFFm) += fvJump*NBhsNdL2dFm(k,p)*Gradsm[l+0*nbFFm][p]*Valsp[j+indexFieldfT*nbFFp]*wJ;
                }
              }

              for(int l=0;l<nbFFp;l++)
              {
                for(int p=0;p<3;p++)
                {
                  stiff(j+indexFieldfV*nbFFp+nbdofm,l+k*nbFFp+nbdofm) +=fvJump*NBhsNdL1dFp(k,p)*Gradsp[l+0*nbFFp][p]*Valsp[j+indexFieldfV*nbFFp]*wJ;
                  stiff(j+indexFieldfV*nbFFp+nbdofm,l+k*nbFFp+nbdofm) +=ftJump*NBhsNdL2dFp(k,p)*Gradsp[l+0*nbFFp][p]*Valsp[j+indexFieldfV*nbFFp]*wJ;

                  stiff(j+indexFieldfT*nbFFp+nbdofm,l+k*nbFFp+nbdofm) +=ftJump*NBhsNdjy1dFp(k,p)*Gradsp[l+0*nbFFp][p]*Valsp[j+indexFieldfT*nbFFp]*wJ;
                  stiff(j+indexFieldfT*nbFFp+nbdofm,l+k*nbFFp+nbdofm) +=fvJump*NBhsNdL2dFp(k,p)*Gradsp[l+0*nbFFp][p]*Valsp[j+indexFieldfT*nbFFp]*wJ;
                }
              }
            }
          }

          //dfI2TdfT
          // compatibility
          // Assembly (loop on shape function)
          for(int j=0;j<nbFFm;j++)
          {
            for(int l=0;l<nbFFm;l++)
            {
              for(int p =0; p < 3; p++)
              {
                for(int q =0; q < 3; q++)
                {
                  stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFm) -= (jy1m->operator()(p,q)*Gradsm[j+indexFieldfT*nbFFm][q])*nu(p)*Valsm[l+indexFieldfT*nbFFm]*wJ/2.;
                  stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFm) += ftJump*(djy1dTm->operator()(p,q)*dTdfTm+djy1dVm->operator()(p,q)*dVdfTm)*Gradsm[j+indexFieldfT*nbFFm][q]*nu(p)*Valsm[l+indexFieldfT*nbFFm]*wJ/2.;
                  stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFm) += fvJump*(dl20dTm->operator()(p,q)*dTdfTm+dl20dvm->operator()(p,q)*dVdfTm)*Gradsm[j+indexFieldfT*nbFFm][q]*(nu(p)*Valsm[l+indexFieldfT*nbFFm])*wJ/2.;
                }
              }
            }
            for(int l=0;l<nbFFp;l++)
            {
              for(int p =0; p < 3; p++)
              {
                for(int q =0; q < 3; q++)
                {
                  stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFp+nbdofm) += (jy1m->operator()(p,q)*Gradsm[j+indexFieldfT*nbFFm][q])*nu(p)*Valsp[l+indexFieldfT*nbFFp]*wJ/2.;
                }
              }
            }
          }

          for(int j=0;j<nbFFp;j++)
          {
            for(int l=0;l<nbFFm;l++)
            {
              for(int p =0; p < 3; p++)
              {
                for(int q =0; q < 3; q++)
                {
                  stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFm) -= (jy1p->operator()(p,q)*Gradsp[j+indexFieldfT*nbFFp][q])*nu(p)*Valsm[l+indexFieldfT*nbFFm]*wJ/2.;
                }
              }
            }
            for(int l=0;l<nbFFp;l++)
            {
              for(int p =0; p < 3; p++)
              {
                for(int q =0; q < 3; q++)
                {
                  stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += (jy1p->operator()(p,q)*Gradsp[j+indexFieldfT*nbFFp][q])*nu(p)*Valsp[l+indexFieldfT*nbFFp]*wJ/2.;

                  stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += ftJump*(djy1dTp->operator()(p,q)*dTdfTp+djy1dVp->operator()(p,q)*dVdfTp)*Gradsp[j+indexFieldfT*nbFFp][q]*nu(p)*Valsp[l+indexFieldfT*nbFFp]*wJ/2.;
                  stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += fvJump*(dl20dTp->operator()(p,q)*dTdfTp+dl20dvp->operator()(p,q)*dVdfTp)*Gradsp[j+indexFieldfT*nbFFp][q]*(nu(p)*Valsp[l+indexFieldfT*nbFFp])*wJ/2.;
                }
              }
            }
          }


         //new related to dialetation

          for(int j=0;j<nbFFm;j++)
          {
            for(int l=0;l<nbFFm;l++)
            {
              for(int k =0; k < 3; k++)
              {
                for(int p =0; p < 3; p++)
                {
                  stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFm) +=  gamma*jump(k)*(Stiff_alphadialitationm(k,p)*(-2./fTm)*Valsm[j+indexFieldfT*nbFFm])*nu(p)*Valsm[l+indexFieldfT*nbFFm]*wJ/2.*(-1.)/eqRatio;
                }
              }
            }
          }

          for(int j=0;j<nbFFp;j++)
          {
            for(int l=0;l<nbFFp;l++)
            {
              for(int k =0; k < 3; k++)
              {
                for(int p =0; p< 3; p++)
                {
                  stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) +=  gamma*jump(k)*(Stiff_alphadialitationp(k,p)*(-2./fTp)*Valsp[j+indexFieldfT*nbFFp])*nu(p)*Valsp[l+indexFieldfT*nbFFp]*wJ/2.*(-1.)/eqRatio;
                }
              }
            }
          }


          for(int j=0;j<nbFFm;j++)
          {
            for(int k=0;k<3;k++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p =0; p < 3; p++)
                {
                    stiff(j+indexFieldfT*nbFFm,l+k*nbFFm) -=  gamma*(Stiff_alphadialitationm(k,p)*Valsm[j+indexFieldfT*nbFFm])*(nu(p)*wJ/2.*Valsm[l+0*nbFFm])/eqRatio*(-1.);
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p =0; p < 3; p++)
                {
                    stiff(j+indexFieldfT*nbFFm,l+k*nbFFp+nbdofm) +=  gamma*(Stiff_alphadialitationm(k,p)*Valsm[j+indexFieldfT*nbFFm])*(nu(p)*wJ/2.*Valsp[l+0*nbFFp])/eqRatio*(-1.);
                }
              }
            }
          }

          for(int j=0;j<nbFFp;j++)
          {
            for(int k=0;k<3;k++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  stiff(j+indexFieldfT*nbFFp+nbdofm,l+k*nbFFm) -=  gamma*(Stiff_alphadialitationp(k,p)*Valsp[j+indexFieldfT*nbFFp])*(nu(p)*wJ/2.*Valsm[l+0*nbFFm])/eqRatio*(-1.);
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  stiff(j+indexFieldfT*nbFFp+nbdofm,l+k*nbFFp+nbdofm) +=  gamma*(Stiff_alphadialitationp(k,p)*Valsp[j+indexFieldfT*nbFFp])*(nu(p)*wJ/2.*Valsp[l+0*nbFFp])/eqRatio*(-1.);
                }
              }
            }
          }

          //dfI2Tdfv
          for(int j=0;j<nbFFm;j++)
          {
            for(int l=0;l<nbFFm;l++)
            {
              for(int p =0; p < 3; p++)
              {
                for(int q =0; q < 3; q++)
                {
                  stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFm) -= (l20m->operator()(p,q)*Gradsm[j+indexFieldfT*nbFFm][q])*nu(p)*Valsm[l+indexFieldfV*nbFFm]*wJ/2.;

                  stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFm) += ftJump*(djy1dVm->operator()(p,q)*dVdfVm+djy1dTm->operator()(p,q)*dTdfVm)*Gradsm[j+indexFieldfT*nbFFm][q]*nu(p)*Valsm[l+indexFieldfV*nbFFm]*wJ/2.;
                  stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFm) += fvJump*(dl20dvm->operator()(p,q)*dVdfVm+dl20dTm->operator()(p,q)*dTdfVm)*Gradsm[j+indexFieldfT*nbFFm][q]*nu(p)*Valsm[l+indexFieldfV*nbFFm]*wJ/2.;
                }
              }
            }
            for(int l=0;l<nbFFp;l++)
            {
              for(int p =0; p < 3; p++)
              {
                for(int q =0; q < 3; q++)
                {
                  stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFp+nbdofm) += (l20m->operator()(p,q)*Gradsm[j+indexFieldfT*nbFFm][q])*nu(p)*Valsp[l+indexFieldfV*nbFFp]*wJ/2.;
                }
              }
            }
          }
          for(int j=0;j<nbFFp;j++)
          {
            for(int l=0;l<nbFFm;l++)
            {
              for(int p =0; p < 3; p++)
              {
                for(int q =0; q < 3; q++)
                {
                  stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFm) -= (l20p->operator()(p,q)*Gradsp[j+indexFieldfT*nbFFp][q])*(nu(p)*Valsm[l+indexFieldfV*nbFFm])*wJ/2.;
                }
              }
            }
            for(int l=0;l<nbFFp;l++)
            {
              for(int p =0; p < 3; p++)
              {
                for(int q =0; q < 3; q++)
                {
                  stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += (l20p->operator()(p,q)*Gradsp[j+indexFieldfT*nbFFp][q])*nu(p)*Valsp[l+indexFieldfV*nbFFp]*wJ/2.;

                  stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += ftJump*(djy1dVp->operator()(p,q)*dVdfVp+djy1dTp->operator()(p,q)*dTdfVp)*Gradsp[j+indexFieldfT*nbFFp][q]*nu(p)*Valsp[l+indexFieldfV*nbFFp]*wJ/2.;
                  stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += fvJump*(dl20dvp->operator()(p,q)*dVdfVp+dl20dTp->operator()(p,q)*dTdfVp)*Gradsp[j+indexFieldfT*nbFFp][q]*nu(p)*Valsp[l+indexFieldfV*nbFFp]*wJ/2.;
                }
              }
            }
          }

          //dfI2vdfv
           for(int j=0;j<nbFFm;j++)
          {
            for(int l=0;l<nbFFm;l++)
            {
              for(int p =0; p < 3; p++)
              {
                for(int q =0; q < 3; q++)
                {
                  stiff(j+indexFieldfV*nbFFm,l+indexFieldfV*nbFFm) -= (l10m->operator()(p,q)*Gradsm[j+indexFieldfV*nbFFm][q])*nu(p)*Valsm[l+indexFieldfV*nbFFm]*wJ/2.;

                 stiff(j+indexFieldfV*nbFFm,l+indexFieldfV*nbFFm) += fvJump*(dl10dTm->operator()(p,q)*dTdfVm*Gradsm[j+indexFieldfV*nbFFm][q])*(nu(p)*Valsm[l+indexFieldfV*nbFFm])*wJ/2.;
                 stiff(j+indexFieldfV*nbFFm,l+indexFieldfV*nbFFm) += ftJump*(dl20dvm->operator()(p,q)*dVdfVm+dl20dTm->operator()(p,q)*dTdfVm)*Gradsm[j+indexFieldfV*nbFFm][q]*nu(p)*Valsm[l+indexFieldfV*nbFFm]*wJ/2.;
                }
              }
            }
            for(int l=0;l<nbFFp;l++)
            {
              for(int p =0; p < 3; p++)
              {
                for(int q =0; q < 3; q++)
                {
                  stiff(j+indexFieldfV*nbFFm,l+indexFieldfV*nbFFp+nbdofm) += (l10m->operator()(p,q)*Gradsm[j+indexFieldfV*nbFFm][q])*nu(p)*Valsp[l+indexFieldfV*nbFFp]*wJ/2.;
                }
              }
            }
          }
          for(int j=0;j<nbFFp;j++)
          {
            for(int l=0;l<nbFFm;l++)
            {
              for(int p =0; p < 3; p++)
              {
                for(int q =0; q < 3; q++)
                {
                  stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfV*nbFFm) -= (l10p->operator()(p,q)*Gradsp[j+indexFieldfV*nbFFp][q])*(nu(p)*Valsm[l+indexFieldfV*nbFFm])*wJ/2.;
                }
              }
            }
            for(int l=0;l<nbFFp;l++)
            {
              for(int p =0; p < 3; p++)
              {
                for(int q =0; q < 3; q++)
                {
                  stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += (l10p->operator()(p,q)*Gradsp[j+indexFieldfV*nbFFp][q])*nu(p)*Valsp[l+indexFieldfV*nbFFp]*wJ/2.;

                  stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += fvJump*(dl10dTp->operator()(p,q)*dTdfVp*Gradsp[j+indexFieldfV*nbFFp][q])*(nu(p)*Valsp[l+indexFieldfV*nbFFp])*wJ/2.;
                  stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += ftJump*(dl20dvp->operator()(p,q)*dVdfVp+dl20dTp->operator()(p,q)*dTdfVp)*Gradsp[j+indexFieldfV*nbFFp][q]*nu(p)*Valsp[l+indexFieldfV*nbFFp]*wJ/2.;
                }
              }
            }
          }

          //dfI2vdfT
          for(int j=0;j<nbFFm;j++)
          {
            for(int l=0;l<nbFFm;l++)
            {
              for(int p =0; p < 3; p++)
              {
                for(int q =0; q < 3; q++)
                {
                  stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFm) -= (l20m->operator()(p,q)*Gradsm[j+indexFieldfV*nbFFm][q])*nu(p)*Valsm[l+indexFieldfT*nbFFm]*wJ/2.;

                  stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFm) += fvJump*(dl10dTm->operator()(p,q)*dTdfTm*Gradsm[j+indexFieldfV*nbFFm][q])*(nu(p)*Valsm[l+indexFieldfT*nbFFm])*wJ/2.;
                  stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFm) += ftJump*(dl20dTm->operator()(p,q)*dTdfTm+dl20dvm->operator()(p,q)*dVdfTm)*Gradsm[j+indexFieldfV*nbFFm][q]*nu(p)*Valsm[l+indexFieldfT*nbFFm]*wJ/2.;
                }
              }
            }
            for(int l=0;l<nbFFp;l++)
            {
              for(int p =0; p < 3; p++)
              {
                for(int q =0; q < 3; q++)
                {
                  stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFp+nbdofm) += (l20m->operator()(p,q)*Gradsm[j+indexFieldfV*nbFFm][q])*nu(p)*Valsp[l+indexFieldfT*nbFFp]*wJ/2.;
                }
              }
            }
          }

          for(int j=0;j<nbFFp;j++)
          {
            for(int l=0;l<nbFFm;l++)
            {
              for(int p =0; p < 3; p++)
              {
                for(int q =0; q < 3; q++)
                {
                  stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFm) -= (l20p->operator()(p,q)*Gradsp[j+indexFieldfV*nbFFp][q])*nu(p)*Valsm[l+indexFieldfT*nbFFm]*wJ/2.;
                }
              }
            }
            for(int l=0;l<nbFFp;l++)
            {
              for(int p =0; p < 3; p++)
              {
                for(int q =0; q < 3; q++)
                {
                  stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += (l20p->operator()(p,q)*Gradsp[j+indexFieldfV*nbFFp][q])*nu(p)*Valsp[l+indexFieldfT*nbFFp]*wJ/2.;

                  stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += fvJump*(dl10dTp->operator()(p,q)*dTdfTp*Gradsp[j+indexFieldfV*nbFFp][q])*(nu(p)*Valsp[l+indexFieldfT*nbFFp])*wJ/2.;
                  stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += ftJump*(dl20dTp->operator()(p,q)*dTdfTp+dl20dvp->operator()(p,q)*dVdfTp)*Gradsp[j+indexFieldfV*nbFFp][q]*nu(p)*Valsp[l+indexFieldfT*nbFFp]*wJ/2.;
                }
              }
            }
          }

          //dfTI2dF +dfvI2dF
          for(int j=0;j<nbFFm;j++)
          {
            for(int l=0;l<nbFFm;l++)
            {
              for(int m =0; m < 3; m++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    for(int n =0; n< 3; n++)
                                {
                      stiff(j+indexFieldfT*nbFFm,l+m*nbFFm) += ftJump*(djy1dFm->operator()(p,q,m,n)*Gradsm[j+indexFieldfT*nbFFm][q]*Gradsm[l+0*nbFFm][n])*(nu(p)*wJ/2.);
                      stiff(j+indexFieldfT*nbFFm,l+m*nbFFm) += fvJump*(dl20dFm->operator()(p,q,m,n)*Gradsm[j+indexFieldfT*nbFFm][q]*Gradsm[l+0*nbFFm][n])*(nu(p)*wJ/2.);

                      stiff(j+indexFieldfV*nbFFm,l+m*nbFFm) += fvJump*(dl10dFm->operator()(p,q,m,n)*Gradsm[j+indexFieldfV*nbFFm][q]*Gradsm[l+0*nbFFm][n])*(nu(p)*wJ/2.);
                      stiff(j+indexFieldfV*nbFFm,l+m*nbFFm) += ftJump*(dl20dFm->operator()(p,q,m,n)*Gradsm[j+indexFieldfV*nbFFm][q]*Gradsm[l+0.*nbFFm][n])*(nu(p)*wJ/2.);
                    }
                  }
                }
              }
            }
          }

          for(int j=0;j<nbFFp;j++)
          {
            for(int l=0;l<nbFFp;l++)
            {
              for(int m =0; m < 3; m++)
              {
                for(int n =0; n< 3; n++)
                {
                  for(int p =0; p < 3; p++)
                  {
                    for(int q =0; q < 3; q++)
                    {
                      stiff(j+indexFieldfT*nbFFp+nbdofm,l+m*nbFFp+nbdofm) += ftJump*(djy1dFp->operator()(p,q,m,n)*Gradsp[j+indexFieldfT*nbFFp][q]*Gradsp[l+0*nbFFp][n])*(nu(p)*wJ/2.);
                      stiff(j+indexFieldfT*nbFFp+nbdofm,l+m*nbFFp+nbdofm) += fvJump*(dl20dFp->operator()(p,q,m,n)*Gradsp[j+indexFieldfT*nbFFp][q]*Gradsp[l+0*nbFFp][n])*(nu(p)*wJ/2.);

                      stiff(j+indexFieldfV*nbFFp+nbdofm,l+m*nbFFp+nbdofm) += fvJump*(dl10dFp->operator()(p,q,m,n)*Gradsp[j+indexFieldfV*nbFFp][q]*Gradsp[l+0*nbFFp][n])*(nu(p)*wJ/2.);
                      stiff(j+indexFieldfV*nbFFp+nbdofm,l+m*nbFFp+nbdofm) += ftJump*(dl20dFp->operator()(p,q,m,n)*Gradsp[j+indexFieldfV*nbFFp][q]*Gradsp[l+0*nbFFp][n])*(nu(p)*wJ/2.);
                    }
                  }
                }
              }
            }
          }
        }
      }
        else
        {
          const double &ftJump = ipvp->getConstRefToOneOverTemperatureJump();
          const double &fvJump = ipvp->getConstRefTofvJump();

          const double &dfvJumpdvm = ipvm->getConstRefTodfvjumpdvm();
          const double &dfvJumpdvp = ipvp->getConstRefTodfvjumpdvp();

          const double &dfvJumpdTm = ipvm->getConstRefTodfvjumpdTm();
          const double &dfvJumpdTp = ipvp->getConstRefTodfvjumpdTp();

          const double &dftJumpdTm = ipvm->getConstRefTodOneOverTemperatureJumpdTm();
          const double &dftJumpdTp = ipvp->getConstRefTodOneOverTemperatureJumpdTp();

          //std::vector<TensorialTraits<double>::GradType> &Gradsm = vallm[i]->_vgrads;
          //std::vector<TensorialTraits<double>::GradType> &Gradsp = vallp[i]->_vgrads;

          const std::vector<TensorialTraits<double>::GradType> &Gradsm = ipvm->gradf(_minusSpace,ie->getElem(0),GPm[i]);
          const std::vector<TensorialTraits<double>::GradType> &Gradsp = ipvp->gradf(_plusSpace,ie->getElem(1),GPp[i]);

          // possibility of fracture
          bool broken = false;
          if(_mlawMinus->getType() == materialLaw::fracture )
          {
            const FractureCohesive3DIPVariable *ipvwf = static_cast<const FractureCohesive3DIPVariable*>(ipsm->getState(IPStateBase::current)); // broken via minus (OK broken on both sides)
            broken = ipvwf->isbroken();
          }
          static SVector3 nu;
          nu = referenceOutwardNormalm;
          nu*= (1./referenceOutwardNormalm.norm());

          const std::vector<TensorialTraits<double>::ValType> &Valsm = ipvm->f(_minusSpace,ie->getElem(0),GPm[i]);
          const std::vector<TensorialTraits<double>::ValType> &Valsp = ipvp->f(_plusSpace,ie->getElem(1),GPp[i]);

          //new
          static STensor3  Stiff_alphadialitationm;
          Stiff_alphadialitationm = *(ipvm->getConstRefToStiff_alphaDilatation());
          static STensor3  Stiff_alphadialitationp;
          Stiff_alphadialitationp = *(ipvp->getConstRefToStiff_alphaDilatation());

          const double &Tp = ipvp->getConstRefToTemperature();
          const double &Tm = ipvm->getConstRefToTemperature();

          Stiff_alphadialitationm*= (-1.)*(Tm*Tm);
          Stiff_alphadialitationp*= (-1.)*(Tp*Tp);
          const SVector3 &jump = ipvp->getConstRefToJump();

          if(broken)
          {

          }
          else
          {
            const STensor3  *dPdTm     = &(ipvm->getConstRefTodPdT());
            const STensor3  *dPdTp     = &(ipvp->getConstRefTodPdT());
            const STensor33 *djedFm    = &(ipvm->getConstRefTodFluxVdF());
            const STensor33 *djedFp    = &(ipvp->getConstRefTodFluxVdF());
            const STensor3  *dPdVm     = &(ipvm->getConstRefTodPdV());
            const STensor3  *dPdVp     = &(ipvp->getConstRefTodPdV());

            const STensor3  *l10m = (ipvm->getConstRefToElecConductivity());
            const STensor3  *l10p = (ipvp->getConstRefToElecConductivity());
            const STensor3  *l20m = (ipvm->getConstRefToCrossElecConductivity());
            const STensor3  *l20p = (ipvp->getConstRefToCrossElecConductivity());

            const STensor3  *dl10dTm = &(ipvm->getConstRefTodElecConductivitydT());
            const STensor3  *dl10dTp = &(ipvp->getConstRefTodElecConductivitydT());
            const STensor3  *dl20dTm = &(ipvm->getConstRefTodCrossElecConductivitydT());
            const STensor3  *dl20dTp = &(ipvp->getConstRefTodCrossElecConductivitydT());

            const STensor3  *dl20dvm = &(ipvm->getConstRefTodCrossElecConductivitydv());
            const STensor3  *dl20dvp = &(ipvp->getConstRefTodCrossElecConductivitydv());

            const STensor43 *dl10dFm = &(ipvm->getConstRefTodElecConductivitydF());
            const STensor43 *dl10dFp = &(ipvp->getConstRefTodElecConductivitydF());
            const STensor43 *dl20dFm = &(ipvm->getConstRefTodCrossElecConductivitydF());
            const STensor43 *dl20dFp = &(ipvp->getConstRefTodCrossElecConductivitydF());

            const STensor3  *djedGradVm = &(ipvm->getConstRefTodFluxVdGradV());
            const STensor3  *djedGradVp = &(ipvp->getConstRefTodFluxVdGradV());
            const SVector3  *djedVm     = &(ipvm->getConstRefTodFluxVdV());
            const SVector3  *djedVp     = &(ipvp->getConstRefTodFluxVdV());
            const STensor3  *djedGradTm = &(ipvm->getConstRefTodFluxVdGradT());
            const STensor3  *djedGradTp = &(ipvp->getConstRefTodFluxVdGradT());
            const SVector3  *djedTm     = &(ipvm->getConstRefTodFluxVdT());
            const SVector3  *djedTp     = &(ipvp->getConstRefTodFluxVdT());

         //   const STensor3  *jy1m	= (ipvm->getConstRefToEnergyConductivity());
         //   const STensor43 *djy1dFm    = &(ipvm->getConstRefTodEnergyConductivitydF());
         //   const STensor3  *djy1dVm	= &(ipvm->getConstRefTodEnergyConductivitydV());
         //   const STensor3  *djy1dTm	= &(ipvm->getConstRefTodEnergyConductivitydT());

         //    const STensor3  *jy1p	= (ipvp->getConstRefToEnergyConductivity());
         //   const STensor43 *djy1dFp    = &(ipvp->getConstRefTodEnergyConductivitydF());
         //   const STensor3  *djy1dVp	= &(ipvp->getConstRefTodEnergyConductivitydV());
         //  const STensor3  *djy1dTp	= &(ipvp->getConstRefTodEnergyConductivitydT()); 

            const STensor3  *jy1m	= (ipvm->getConstRefToTherConductivity());
            const STensor43 *djy1dFm    = &(ipvm->getConstRefTodTherConductivitydF());
            const STensor3  *djy1dVm	= &(ipvm->getConstRefTodTherConductivitydv());
            const STensor3  *djy1dTm	= &(ipvm->getConstRefTodTherConductivitydT());

            const STensor3  *jy1p	= (ipvp->getConstRefToTherConductivity());
            const STensor43 *djy1dFp    = &(ipvp->getConstRefTodTherConductivitydF());
            const STensor3  *djy1dVp	= &(ipvp->getConstRefTodTherConductivitydv());
            const STensor3  *djy1dTp	= &(ipvp->getConstRefTodTherConductivitydT()); 

           // static STensor3 jy1m, jy1m_1;
            STensorOperation::zero(jy1m); STensorOperation::zero(jy1m_1);
            STensorOperation::multSTensor3(ipvm->getConstRefTodFluxdGradField()[0][0],ipvm->getConstRefTodGradFielddGradEnergyConjugatedField()[0][0],jy1m_1);
            STensorOperation::multSTensor3(ipvm->getConstRefTodFluxdGradField()[0][1],ipvm->getConstRefTodGradFielddGradEnergyConjugatedField()[1][0],jy1m);
            jy1m += jy1m_1;
            static STensor43 djy1dFm, djy1dFp, djydgradVdFm, djydgradTdFm, djydgradVdFp, djydgradTdFp;
            static STensor3 dgradVdgradfTm, dgradTdgradfTm, dgradVdgradfTp, dgradTdgradfTp; 
            djydgradVdFm=ipvm->getConstRefTodFluxdGradFielddF()[0][1];
            djydgradTdFm=ipvm->getConstRefTodFluxdGradFielddF()[0][0];
            dgradVdgradfTm=ipvm->getConstRefTodGradFielddGradEnergyConjugatedField()[1][0];
            dgradTdgradfTm=ipvm->getConstRefTodGradFielddGradEnergyConjugatedField()[0][0];
            djydgradVdFp=ipvp->getConstRefTodFluxdGradFielddF()[0][1];
            djydgradTdFp=ipvp->getConstRefTodFluxdGradFielddF()[0][0];
            dgradVdgradfTp=ipvp->getConstRefTodGradFielddGradEnergyConjugatedField()[1][0];
            dgradTdgradfTp=ipvp->getConstRefTodGradFielddGradEnergyConjugatedField()[0][0];
            STensorOperation::zero(djy1dFm);
            STensorOperation::zero(djy1dFp);

             for (int i=0; i<3; i++){
              for (int j=0; j<3; j++){
                for (int k=0; k<3; k++){
                 for (int l=0; l<3; l++){
                  for (int q=0; q<3; q++){
                 			  djy1dFm(i,j,k,l) += djydgradVdFm(i,q,k,l)*dgradVdgradfTm(q,j);
                 			  djy1dFm(i,j,k,l) += djydgradTdFm(i,q,k,l)*dgradTdgradfTm(q,j);
                 			  djy1dFp(i,j,k,l) += djydgradVdFp(i,q,k,l)*dgradVdgradfTp(q,j);
                 			  djy1dFp(i,j,k,l) += djydgradTdFp(i,q,k,l)*dgradTdgradfTp(q,j);
                                         }
                                        }
                                       }
                                      }
                                     }

            static STensor3 djy1dVm, djy1dVm_1;
            STensorOperation::zero(djy1dVm); STensorOperation::zero(djy1dVm_1);
            STensorOperation::multSTensor3(ipvm->getConstRefTodFluxdGradFielddField()[0][0][1],ipvm->getConstRefTodGradFielddGradEnergyConjugatedField()[0][0],djy1dVm_1);
            STensorOperation::multSTensor3(ipvm->getConstRefTodFluxdGradFielddField()[0][1][1],ipvm->getConstRefTodGradFielddGradEnergyConjugatedField()[1][0],djy1dVm);
            djy1dVm += djy1dVm_1;
            static STensor3 djy1dTm, djy1dTm_1;
            STensorOperation::zero(djy1dTm); STensorOperation::zero(djy1dTm_1);
            STensorOperation::multSTensor3(ipvm->getConstRefTodFluxdGradFielddField()[0][0][0],ipvm->getConstRefTodGradFielddGradEnergyConjugatedField()[0][0],djy1dTm_1);
            STensorOperation::multSTensor3(ipvm->getConstRefTodFluxdGradFielddField()[0][1][0],ipvm->getConstRefTodGradFielddGradEnergyConjugatedField()[1][0],djy1dTm);
            djy1dTm += djy1dTm_1;

            static STensor3 jy1p, jy1p_1;
            STensorOperation::multSTensor3(ipvp->getConstRefTodFluxdGradField()[0][0],ipvp->getConstRefTodGradFielddGradEnergyConjugatedField()[0][0],jy1p_1);
            STensorOperation::multSTensor3(ipvp->getConstRefTodFluxdGradField()[0][1],ipvp->getConstRefTodGradFielddGradEnergyConjugatedField()[1][0],jy1p);
            jy1p += jy1p_1;

            static STensor3 djy1dVp, djy1dVp_1;
            STensorOperation::zero(djy1dVp); STensorOperation::zero(djy1dVp_1);
            STensorOperation::multSTensor3(ipvp->getConstRefTodFluxdGradFielddField()[0][0][1],ipvp->getConstRefTodGradFielddGradEnergyConjugatedField()[0][0],djy1dVp_1);
            STensorOperation::multSTensor3(ipvp->getConstRefTodFluxdGradFielddField()[0][1][1],ipvp->getConstRefTodGradFielddGradEnergyConjugatedField()[1][0],djy1dVp);
            djy1dVp += djy1dVp_1;
            static STensor3 djy1dTp, djy1dTp_1;
            STensorOperation::zero(djy1dTp); STensorOperation::zero(djy1dTp_1);
            STensorOperation::multSTensor3(ipvp->getConstRefTodFluxdGradFielddField()[0][0][0],ipvp->getConstRefTodGradFielddGradEnergyConjugatedField()[0][0],djy1dTp_1);
            STensorOperation::multSTensor3(ipvp->getConstRefTodFluxdGradFielddField()[0][1][0],ipvp->getConstRefTodGradFielddGradEnergyConjugatedField()[1][0],djy1dTp);
            djy1dTp += djy1dTp_1; //

            const STensor33 *djydFp	= &(ipvp->getConstRefTodFluxEnergydF());
            const STensor3  *djydGradTp	= &(ipvp->getConstRefTodFluxEnergydGradT());
            const SVector3  *djydTp	= &(ipvp->getConstRefTodFluxEnergydT());
            const STensor3  *djydGradVp	= &(ipvp->getConstRefTodFluxEnergydGradV());
            const SVector3  *djydVp	= &(ipvp->getConstRefTodFluxEnergydV());
            const SVector3  *fluxjyp	= &(ipvp->getConstRefToFluxEnergy());

            const STensor33 *djydFm	= &(ipvm->getConstRefTodFluxEnergydF());
            const STensor3  *djydGradTm	= &(ipvm->getConstRefTodFluxEnergydGradT());
            const SVector3  *djydTm	= &(ipvm->getConstRefTodFluxEnergydT());
            const STensor3  *djydGradVm	= &(ipvm->getConstRefTodFluxEnergydGradV());
            const SVector3  *djydVm	= &(ipvm->getConstRefTodFluxEnergydV());
            const SVector3  *fluxjym	= &(ipvm->getConstRefToFluxEnergy());

            double eqRatio = _ElecTherMechEqRatio;
            wJ*=eqRatio;
            int indexFieldfT=3+getNumNonLocalVariable();
            int indexFieldfV=3+getNumNonLocalVariable()+1;

            // Assembly consistency (from derivative of sigma with extra field)
            //dpdV +dpdT
           for(int j=0;j<nbFFm;j++)
            {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+k*nbFFm,l+indexFieldfT*nbFFm) -= dPdTm->operator()(k,q)*(Valsm[l+indexFieldfT*nbFFm]*Valsm[j+0*nbFFm]*wJ/2.*nu(q)/eqRatio);

                    stiff(j+k*nbFFm,l+indexFieldfV*nbFFm) -= dPdVm->operator()(k,q)*(Valsm[l+indexFieldfV*nbFFm]*Valsm[j+0*nbFFm]*wJ/2.*nu(q)/eqRatio);
                  }
                }
                for(int l=0;l<nbFFp;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+k*nbFFm,l+indexFieldfT*nbFFp+nbdofm) -= dPdTp->operator()(k,q)*(Valsp[l+indexFieldfT*nbFFp]*Valsm[j+0*nbFFm]*wJ/2.*nu(q)/eqRatio);

                    stiff(j+k*nbFFm,l+indexFieldfV*nbFFp+nbdofm) -= dPdVp->operator()(k,q)*(Valsp[l+indexFieldfV*nbFFp]*Valsm[j+0*nbFFm]*wJ/2.*nu(q)/eqRatio);
                  }
                }
              }
            }
            for(int j=0;j<nbFFp;j++)
            {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+k*nbFFp+nbdofm,l+indexFieldfT*nbFFm) += dPdTm->operator()(k,q)*(Valsm[l+indexFieldfT*nbFFm]*Valsp[j+0*nbFFp]*wJ/2.*nu(q)/eqRatio);

                    stiff(j+k*nbFFp+nbdofm,l+indexFieldfV*nbFFm) += dPdVm->operator()(k,q)*(Valsm[l+indexFieldfV*nbFFm]*Valsp[j+0*nbFFp]*wJ/2.*nu(q)/eqRatio);
                  }
                }
                for(int l=0;l<nbFFp;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+k*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += dPdTp->operator()(k,q)*(Valsp[l+indexFieldfT*nbFFp]*Valsp[j+0*nbFFp]*wJ/2.*nu(q)/eqRatio);

                    stiff(j+k*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += dPdVp->operator()(k,q)*(Valsp[l+indexFieldfV*nbFFp]*Valsp[j+0*nbFFp]*wJ/2.*nu(q)/eqRatio);
                  }
                }
              }
            }

            // Assembly consistency (from derivative of flux with deformation gradient)
            for(int j=0;j<nbFFm;j++)
            {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    for(int N = 0; N< 3; N++)
                    {
                      stiff(j+indexFieldfT*nbFFm,l+k*nbFFm) -= djydFm->operator()(N,k,q)*(Gradsm[l+0*nbFFm][q]*Valsm[j+indexFieldfT*nbFFm]*wJ/2.*nu(N));
                    }
                  }
                }
                for(int l=0;l<nbFFp;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    for(int N = 0; N< 3; N++)
                    {
                      stiff(j+indexFieldfT*nbFFm,l+k*nbFFp+nbdofm) -= djydFp->operator()(N,k,q)*(Gradsp[l+0*nbFFp][q]*Valsm[j+indexFieldfT*nbFFm]*wJ/2.*nu(N));
                    }
                  }
                }
              }
            }
            for(int j=0;j<nbFFp;j++)
            {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    for(int N = 0; N< 3; N++)
                    {
                      stiff(j+indexFieldfT*nbFFp+nbdofm,l+k*nbFFm) += djydFm->operator()(N,k,q)*(Gradsm[l+0*nbFFm][q]*Valsp[j+indexFieldfT*nbFFp]*wJ/2.*nu(N));
                    }
                  }
                }
                for(int l=0;l<nbFFp;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    for(int N = 0; N< 3; N++)
                    {
                      stiff(j+indexFieldfT*nbFFp+nbdofm,l+k*nbFFp+nbdofm) += djydFp->operator()(N,k,q)*(Gradsp[l+0*nbFFp][q]*Valsp[j+indexFieldfT*nbFFp]*wJ/2.*nu(N));
                    }
                  }
                }
              }
            }

            //djedf  (derivative if cross term)
            for(int j=0;j<nbFFm;j++)
            {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    for(int N = 0; N< 3; N++)
                    {
                      stiff(j+indexFieldfV*nbFFm,l+k*nbFFm) -= djedFm->operator()(N,k,q)*(Gradsm[l+0*nbFFm][q]*Valsm[j+indexFieldfV*nbFFm]*wJ/2.*nu(N));
                    }
                  }
                }
                for(int l=0;l<nbFFp;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    for(int N = 0; N< 3; N++)
                    {
                      stiff(j+indexFieldfV*nbFFm,l+k*nbFFp+nbdofm) -= djedFp->operator()(N,k,q)*(Gradsp[l+0*nbFFp][q]*Valsm[j+indexFieldfV*nbFFm]*wJ/2.*nu(N));
                    }
                  }
                }
              }
            }
            for(int j=0;j<nbFFp;j++)
            {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    for(int N = 0; N< 3; N++)
                    {
                      stiff(j+indexFieldfV*nbFFp+nbdofm,l+k*nbFFm) += djedFm->operator()(N,k,q)*(Gradsm[l+0*nbFFm][q]*Valsp[j+indexFieldfV*nbFFp]*wJ/2.*nu(N));
                    }
                  }
                }
                for(int l=0;l<nbFFp;l++)
                {
                  for(int q = 0; q< 3; q++)
                  {
                    for(int N = 0; N< 3; N++)
                    {
                      stiff(j+indexFieldfV*nbFFp+nbdofm,l+k*nbFFp+nbdofm) += djedFp->operator()(N,k,q)*(Gradsp[l+0*nbFFp][q]*Valsp[j+indexFieldfV*nbFFp]*wJ/2.*nu(N));
                    }
                  }
                }
              }
            }

            // consistency D\mean{q} DT
            // Assembly consistency
            //dfI1tdt+dfI1tdv
            for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p = 0; p< 3; p++)
                {
                  stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFm) -= djydTm->operator()(p)*(Valsm[l+indexFieldfT*nbFFm]*Valsm[j+indexFieldfT*nbFFm]*wJ/2.*nu(p));
                  stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFm) -= djydVm->operator()(p)*(Valsm[l+indexFieldfV*nbFFm]*Valsm[j+indexFieldfT*nbFFm]*wJ/2.*nu(p));
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFm) -= djydGradTm->operator()(q,p)*(Gradsm[l+indexFieldfT*nbFFm][p]*Valsm[j+indexFieldfT*nbFFm]*wJ/2.*nu(q));
                    stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFm) -= djydGradVm->operator()(q,p)*(Gradsm[l+indexFieldfV*nbFFm][p]*Valsm[j+indexFieldfT*nbFFm]*wJ/2.*nu(q));
                  }
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p = 0; p< 3; p++)
                {
                  stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFp+nbdofm) -= djydTp->operator()(p)*(Valsp[l+indexFieldfT*nbFFp]*Valsm[j+indexFieldfT*nbFFm]*wJ/2.*nu(p));
                  stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFp+nbdofm) -= djydVp->operator()(p)*(Valsp[l+indexFieldfV*nbFFp]*Valsm[j+indexFieldfT*nbFFm]*wJ/2.*nu(p));
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFp+nbdofm) -= djydGradTp->operator()(q,p)*(Gradsp[l+indexFieldfT*nbFFp][p]*Valsm[j+indexFieldfT*nbFFm]*wJ/2.*nu(q));
                    stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFp+nbdofm) -= djydGradVp->operator()(q,p)*(Gradsp[l+indexFieldfV*nbFFp][p]*Valsm[j+indexFieldfT*nbFFm]*wJ/2.*nu(q));
                  }
                }
              }
            }
            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p = 0; p< 3; p++)
                {
                  stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFm) += djydTm->operator()(p)*(Valsm[l+indexFieldfT*nbFFm]*Valsp[j+indexFieldfT*nbFFp]*wJ/2.*nu(p));
                  stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFm) += djydVm->operator()(p)*(Valsm[l+indexFieldfV*nbFFm]*Valsp[j+indexFieldfT*nbFFp]*wJ/2.*nu(p));
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFm) += djydGradTm->operator()(q,p)*(Gradsm[l+indexFieldfT*nbFFm][p]*Valsp[j+indexFieldfT*nbFFp]*wJ/2.*nu(q));
                    stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFm) += djydGradVm->operator()(q,p)*(Gradsm[l+indexFieldfV*nbFFm][p]*Valsp[j+indexFieldfT*nbFFp]*wJ/2.*nu(q));
                  }
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p = 0; p< 3; p++)
                {
                  stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += djydTp->operator()(p)*(Valsp[l+indexFieldfT*nbFFp]*Valsp[j+indexFieldfT*nbFFp]*wJ/2.*nu(p));
                  stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += djydVp->operator()(p)*(Valsp[l+indexFieldfV*nbFFp]*Valsp[j+indexFieldfT*nbFFp]*wJ/2.*nu(p));
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += djydGradTp->operator()(q,p)*(Gradsp[l+indexFieldfT*nbFFp][p]*Valsp[j+indexFieldfT*nbFFp]*wJ/2.*nu(q));
                    stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += djydGradVp->operator()(q,p)*(Gradsp[l+indexFieldfV*nbFFp][p]*Valsp[j+indexFieldfT*nbFFp]*wJ/2.*nu(q));
                  }
                }
              }
            }

            //dfI1vdv+dfI1vdt
            for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p = 0; p< 3; p++)
                {
                  stiff(j+indexFieldfV*nbFFm,l+indexFieldfV*nbFFm) -= djedVm->operator()(p)*(Valsm[l+indexFieldfV*nbFFm]*Valsm[j+indexFieldfV*nbFFm]*wJ/2.*nu(p));
                  stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFm) -= djedTm->operator()(p)*(Valsm[l+indexFieldfT*nbFFm]*Valsm[j+indexFieldfV*nbFFm]*wJ/2.*nu(p));
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+indexFieldfV*nbFFm,l+indexFieldfV*nbFFm) -= djedGradVm->operator()(q,p)*(Gradsm[l+indexFieldfV*nbFFm][p]*Valsm[j+indexFieldfV*nbFFm]*wJ/2.*nu(q));
                    stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFm) -= djedGradTm->operator()(q,p)*(Gradsm[l+indexFieldfT*nbFFm][p]*Valsm[j+indexFieldfV*nbFFm]*wJ/2.*nu(q));
                  }
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p = 0; p< 3; p++)
                {
                  stiff(j+indexFieldfV*nbFFm,l+indexFieldfV*nbFFp+nbdofm) -= djedVp->operator()(p)*(Valsp[l+indexFieldfV*nbFFp]*Valsm[j+indexFieldfV*nbFFm]*wJ/2.*nu(p));
                  stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFp+nbdofm) -= djedTp->operator()(p)*(Valsp[l+indexFieldfT*nbFFp]*Valsm[j+indexFieldfV*nbFFm]*wJ/2.*nu(p));

                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+indexFieldfV*nbFFm,l+indexFieldfV*nbFFp+nbdofm) -= djedGradVp->operator()(q,p)*(Gradsp[l+indexFieldfV*nbFFp][p]*Valsm[j+indexFieldfV*nbFFm]*wJ/2.*nu(q));
                    stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFp+nbdofm) -= djedGradTp->operator()(q,p)*(Gradsp[l+indexFieldfT*nbFFp][p]*Valsm[j+indexFieldfV*nbFFm]*wJ/2.*nu(q));
                  }
                }
              }
            }
            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p = 0; p< 3; p++)
                {
                  stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfV*nbFFm) += djedVm->operator()(p)*(Valsm[l+indexFieldfV*nbFFm]*Valsp[j+indexFieldfV*nbFFp]*wJ/2.*nu(p));
                  stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFm) += djedTm->operator()(p)*(Valsm[l+indexFieldfT*nbFFm]*Valsp[j+indexFieldfV*nbFFp]*wJ/2.*nu(p));

                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfV*nbFFm) += djedGradVm->operator()(q,p)*(Gradsm[l+indexFieldfV*nbFFm][p]*Valsp[j+indexFieldfV*nbFFp]*wJ/2.*nu(q));
                    stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFm) += djedGradTm->operator()(q,p)*(Gradsm[l+indexFieldfT*nbFFm][p]*Valsp[j+indexFieldfV*nbFFp]*wJ/2.*nu(q));
                  }
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p = 0; p< 3; p++)
                {
                  stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += djedVp->operator()(p)*(Valsp[l+indexFieldfV*nbFFp]*Valsp[j+indexFieldfV*nbFFp]*wJ/2.*nu(p));
                  stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += djedTp->operator()(p)*(Valsp[l+indexFieldfT*nbFFp]*Valsp[j+indexFieldfV*nbFFp]*wJ/2.*nu(p));
                  for(int q = 0; q< 3; q++)
                  {
                    stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += djedGradVp->operator()(q,p)*(Gradsp[l+indexFieldfV*nbFFp][p]*Valsp[j+indexFieldfV*nbFFp]*wJ/2.*nu(q));
                    stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += djedGradTp->operator()(q,p)*(Gradsp[l+indexFieldfT*nbFFp][p]*Valsp[j+indexFieldfV*nbFFp]*wJ/2.*nu(q));
                  }
                }
              }
            }

                // Stability  N \mean{k0} (N)
            double NBhsNdjy1dTm=0.;double NBhsNdjy1dTp=0.; double NBhsNdjy1dvm=0.; double NBhsNdjy1dvp=0.;
            double NBhsNdL1dTm=0.;double NBhsNdL1dTp=0.; double NBhsNdL2dTm=0.;double NBhsNdL2dTp=0.;double NBhsNdL2dVm=0.; double NBhsNdL2dVp=0.;

            for(int l = 0; l <3; l++)
            {
              for(int n = 0; n <3; n++)
              {
                NBhsNdjy1dTm+=djy1dTm->operator()(l,n)*nu(n)*nbetahs*nu(l)/2.;
                NBhsNdjy1dTp+=djy1dTp->operator()(l,n)*nu(n)*nbetahs*nu(l)/2.;

                NBhsNdL1dTm+=dl10dTm->operator()(l,n)*nu(n)*nbetahs*nu(l)/2.;
                NBhsNdL1dTp+=dl10dTp->operator()(l,n)*nu(n)*nbetahs*nu(l)/2.;
                NBhsNdL2dTm+=dl20dTm->operator()(l,n)*nu(n)*nbetahs*nu(l)/2.;
                NBhsNdL2dTp+=dl20dTp->operator()(l,n)*nu(n)*nbetahs*nu(l)/2.;

                NBhsNdjy1dvm+=djy1dVm->operator()(l,n)*nu(n)*nbetahs*nu(l)/2.;
                NBhsNdjy1dvp+=djy1dVp->operator()(l,n)*nu(n)*nbetahs*nu(l)/2.;

                NBhsNdL2dVm+=dl20dvm->operator()(l,n)*nu(n)*nbetahs*nu(l)/2.;
                NBhsNdL2dVp+=dl20dvp->operator()(l,n)*nu(n)*nbetahs*nu(l)/2.;
              }
            }

            static STensor3 NBhsNdjy1dFm;
            static STensor3 NBhsNdjy1dFp;
            static STensor3 NBhsNdL1dFm;
            static STensor3 NBhsNdL1dFp;
            static STensor3 NBhsNdL2dFm;
            static STensor3 NBhsNdL2dFp;

            for(int j = 0; j<3; j++){
              for(int p = 0; p<3; p++)
              {
                NBhsNdjy1dFm(j,p) = 0.;
                NBhsNdjy1dFp(j,p)= 0.;
                NBhsNdL1dFm(j,p)= 0.;
                NBhsNdL1dFp(j,p)= 0.;
                NBhsNdL2dFm(j,p)= 0.;
                NBhsNdL2dFp(j,p)= 0.;

                for(int l = 0; l <3; l++)
                {
                  for(int n = 0; n <3; n++)
                  {
                    NBhsNdjy1dFm(j,p)+=djy1dFm->operator()(l,n,j,p)*nu(n)*nbetahs*nu(l)/2.;
                    NBhsNdjy1dFp(j,p)+=djy1dFp->operator()(l,n,j,p)*nu(n)*nbetahs*nu(l)/2.;
                    NBhsNdL1dFm(j,p)+= dl10dFm->operator()(l,n,j,p)*nu(n)*nbetahs*nu(l)/2.;
                    NBhsNdL1dFp(j,p)+= dl10dFp->operator()(l,n,j,p)*nu(n)*nbetahs*nu(l)/2.;
                    NBhsNdL2dFm(j,p)+= dl20dFm->operator()(l,n,j,p)*nu(n)*nbetahs*nu(l)/2.;
                    NBhsNdL2dFp(j,p)+= dl20dFp->operator()(l,n,j,p)*nu(n)*nbetahs*nu(l)/2.;
                  }
                }
              }
            }

            // dfIt3tdt
            double NMeanjy1NBetasc=0.;
            double NMeanl1NBetasc=0.; double NMeanl2NBetasc=0.;
            for(int l = 0; l <3; l++)
            {
              for(int n = 0; n <3; n++)
              {
                NMeanjy1NBetasc += (jy1p->operator()(l,n)+jy1m->operator()(l,n))*nu(n)*nbetahs*nu(l)/2.;
                NMeanl1NBetasc +=  (l10p->operator()(l,n)+l10m->operator()(l,n))*nu(n)*nbetahs*nu(l)/2.;
                NMeanl2NBetasc +=  (l20p->operator()(l,n)+l20m->operator()(l,n))*nu(n)*nbetahs*nu(l)/2.;
              }
            }

           for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFm) -= dftJumpdTm*NMeanjy1NBetasc*Valsm[j+indexFieldfT*nbFFm]*Valsm[l+indexFieldfT*nbFFm]*wJ;
                stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFm) -= dfvJumpdTm*NMeanl2NBetasc*Valsm[j+indexFieldfT*nbFFm]*Valsm[l+indexFieldfT*nbFFm]*wJ;

                stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFm) -= ftJump*NBhsNdjy1dTm*Valsm[l+indexFieldfT*nbFFm]*Valsm[j+indexFieldfT*nbFFm]*wJ;
                stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFm) -= fvJump*NBhsNdL2dTm*Valsm[l+indexFieldfT*nbFFm]*Valsm[j+indexFieldfT*nbFFm]*wJ;
              }
              for(int l=0;l<nbFFp;l++)
              {
                stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFp+nbdofm) -= dftJumpdTp*NMeanjy1NBetasc*Valsm[j+indexFieldfT*nbFFm]*Valsp[l+indexFieldfT*nbFFp]*wJ;
                stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFp+nbdofm) -= dfvJumpdTp*NMeanl2NBetasc*Valsm[j+indexFieldfT*nbFFm]*Valsp[l+indexFieldfT*nbFFp]*wJ;

                stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFp+nbdofm) -= ftJump*NBhsNdjy1dTp*Valsm[l+indexFieldfT*nbFFm]*Valsp[j+indexFieldfT*nbFFp]*wJ;
                stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFp+nbdofm) -= fvJump*NBhsNdL2dTp*Valsm[l+indexFieldfT*nbFFm]*Valsp[j+indexFieldfT*nbFFp]*wJ;
              }
            }


            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFm) += dftJumpdTm*NMeanjy1NBetasc*Valsp[j+indexFieldfT*nbFFp]*Valsm[l+indexFieldfT*nbFFm]*wJ;
                stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFm) += dfvJumpdTm*NMeanl2NBetasc*Valsp[j+indexFieldfT*nbFFp]*Valsm[l+indexFieldfT*nbFFm]*wJ;

                stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFm) += ftJump*NBhsNdjy1dTm*Valsp[l+indexFieldfT*nbFFp]*Valsm[j+indexFieldfT*nbFFm]*wJ;
                stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFm) += fvJump*NBhsNdL2dTm*Valsp[l+indexFieldfT*nbFFp]*Valsm[j+indexFieldfT*nbFFm]*wJ;
              }
              for(int l=0;l<nbFFp;l++)
              {
                stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += dftJumpdTp*NMeanjy1NBetasc*Valsp[j+indexFieldfT*nbFFp]*Valsp[l+indexFieldfT*nbFFp]*wJ;
                stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += dfvJumpdTp*NMeanl2NBetasc*Valsp[j+indexFieldfT*nbFFp]*Valsp[l+indexFieldfT*nbFFp]*wJ;

                stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += ftJump*NBhsNdjy1dTp*Valsp[l+indexFieldfT*nbFFp]*Valsp[j+indexFieldfT*nbFFp]*wJ;
                stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += fvJump*NBhsNdL2dTp*Valsp[l+indexFieldfT*nbFFp]*Valsp[j+indexFieldfT*nbFFp]*wJ;
              }
            }

             //Terms for dfTI3dv
            for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFm) -= ftJump*NBhsNdjy1dvm*Valsm[j+indexFieldfT*nbFFm]*Valsm[l+indexFieldfV*nbFFm]*wJ;
                stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFm) -= fvJump*NBhsNdL2dVm*Valsm[j+indexFieldfT*nbFFm]*Valsm[l+indexFieldfV*nbFFm]*wJ;

                stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFm) -= dfvJumpdvm*NMeanl2NBetasc*Valsm[j+indexFieldfT*nbFFm]*Valsm[l+indexFieldfV*nbFFm]*wJ;
              }
              for(int l=0;l<nbFFp;l++)
              {
                stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFp+nbdofm) -= ftJump*NBhsNdjy1dvp*Valsm[j+indexFieldfT*nbFFm]*Valsp[l+indexFieldfV*nbFFp]*wJ;
                stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFp+nbdofm) -= fvJump*NBhsNdL2dVp*Valsm[j+indexFieldfT*nbFFm]*Valsp[l+indexFieldfV*nbFFp]*wJ;

                stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFp+nbdofm) -= dfvJumpdvp*NMeanl2NBetasc*Valsm[j+indexFieldfT*nbFFm]*Valsp[l+indexFieldfV*nbFFp]*wJ;
              }
            }

            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFm) += ftJump*NBhsNdjy1dvm*Valsp[j+indexFieldfT*nbFFp]*Valsm[l+indexFieldfV*nbFFm]*wJ;
                stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFm) += fvJump*NBhsNdL2dVm*Valsp[j+indexFieldfT*nbFFp]*Valsm[l+indexFieldfV*nbFFm]*wJ;

                stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFm) += dfvJumpdvm*NMeanl2NBetasc*Valsp[j+indexFieldfT*nbFFp]*Valsm[l+indexFieldfV*nbFFm]*wJ;
              }
              for(int l=0;l<nbFFp;l++)
              {
                stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += ftJump*NBhsNdjy1dvp*Valsp[j+indexFieldfT*nbFFp]*Valsp[l+indexFieldfV*nbFFp]*wJ;
                stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += fvJump*NBhsNdL2dVp*Valsp[j+indexFieldfT*nbFFp]*Valsp[l+indexFieldfV*nbFFp]*wJ;

                stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) +=dfvJumpdvp*NMeanl2NBetasc*Valsp[j+indexFieldfT*nbFFp]*Valsp[l+indexFieldfV*nbFFp]*wJ;
              }
            }

            //dfI3vdv
            for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {

                stiff(j+indexFieldfV*nbFFm,l+indexFieldfV*nbFFm) -= ftJump*NBhsNdL2dVm*Valsm[j+indexFieldfV*nbFFm]*Valsm[l+indexFieldfV*nbFFm]*wJ;

                stiff(j+indexFieldfV*nbFFm,l+indexFieldfV*nbFFm) -= dfvJumpdvm*NMeanl1NBetasc*Valsm[j+indexFieldfV*nbFFm]*Valsm[l+indexFieldfV*nbFFm]*wJ;
              }
              for(int l=0;l<nbFFp;l++)
              {

                stiff(j+indexFieldfV*nbFFm,l+indexFieldfV*nbFFp+nbdofm) -= ftJump*NBhsNdL2dVp*Valsm[j+indexFieldfV*nbFFm]*Valsp[l+indexFieldfV*nbFFp]*wJ;

                stiff(j+indexFieldfV*nbFFm,l+indexFieldfV*nbFFp+nbdofm) -= dfvJumpdvp*NMeanl1NBetasc*Valsm[j+indexFieldfV*nbFFm]*Valsp[l+indexFieldfV*nbFFp]*wJ;
              }
            }

            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                 stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfV*nbFFm) += ftJump*NBhsNdL2dVm*Valsp[j+indexFieldfV*nbFFp]*Valsm[l+indexFieldfV*nbFFm]*wJ;

                 stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfV*nbFFm) += dfvJumpdvm*NMeanl1NBetasc*Valsp[j+indexFieldfV*nbFFp]*Valsm[l+indexFieldfV*nbFFm]*wJ;

              }
              for(int l=0;l<nbFFp;l++)
              {
                stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += ftJump*NBhsNdL2dVp*Valsp[j+indexFieldfV*nbFFp]*Valsp[l+indexFieldfV*nbFFp]*wJ;

                stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) +=dfvJumpdvp*NMeanl1NBetasc*Valsp[j+indexFieldfV*nbFFp]*Valsp[l+indexFieldfV*nbFFp]*wJ;

              }
            }

            //Terms for dfvI3dT
              for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFm) -= fvJump*NBhsNdL1dTm*Valsm[j+indexFieldfV*nbFFm]*Valsm[l+indexFieldfT*nbFFm]*wJ;
                stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFm) -= ftJump*NBhsNdL2dTm*Valsm[j+indexFieldfV*nbFFm]*Valsm[l+indexFieldfT*nbFFm]*wJ;

                stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFm) -= dfvJumpdTm*NMeanl1NBetasc*Valsm[j+indexFieldfV*nbFFm]*Valsm[l+indexFieldfT*nbFFm]*wJ;
                stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFm) -= dftJumpdTm*NMeanl2NBetasc*Valsm[j+indexFieldfV*nbFFm]*Valsm[l+indexFieldfT*nbFFm]*wJ;
              }
              for(int l=0;l<nbFFp;l++)
              {
                stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFp+nbdofm) -= fvJump*NBhsNdL1dTp*Valsm[j+indexFieldfV*nbFFm]*Valsp[l+indexFieldfT*nbFFp]*wJ;
                stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFp+nbdofm) -= ftJump*NBhsNdL2dTp*Valsm[j+indexFieldfV*nbFFm]*Valsp[l+indexFieldfT*nbFFp]*wJ;

                stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFp+nbdofm) -= dfvJumpdTp*NMeanl1NBetasc*Valsm[j+indexFieldfV*nbFFm]*Valsp[l+indexFieldfT*nbFFp]*wJ;
                stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFp+nbdofm) -= dftJumpdTp*NMeanl2NBetasc*Valsm[j+indexFieldfV*nbFFm]*Valsp[l+indexFieldfT*nbFFp]*wJ;
              }
            }

            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFm) += fvJump*NBhsNdL1dTm*Valsp[j+indexFieldfV*nbFFp]*Valsm[l+indexFieldfT*nbFFm]*wJ;
                stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFm) += ftJump*NBhsNdL2dTm*Valsp[j+indexFieldfV*nbFFp]*Valsm[l+indexFieldfT*nbFFm]*wJ;

                stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFm) += dfvJumpdTm*NMeanl1NBetasc*Valsp[j+indexFieldfV*nbFFp]*Valsm[l+indexFieldfT*nbFFm]*wJ;
                stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFm) += dftJumpdTm*NMeanl2NBetasc*Valsp[j+indexFieldfV*nbFFp]*Valsm[l+indexFieldfT*nbFFm]*wJ;
              }
              for(int l=0;l<nbFFp;l++)
              {
                stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += fvJump*NBhsNdL1dTp*Valsp[j+indexFieldfV*nbFFp]*Valsp[l+indexFieldfT*nbFFp]*wJ;
                stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += ftJump*NBhsNdL2dTp*Valsp[j+indexFieldfV*nbFFp]*Valsp[l+indexFieldfT*nbFFp]*wJ;

                stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) +=dfvJumpdTp*NMeanl1NBetasc*Valsp[j+indexFieldfV*nbFFp]*Valsp[l+indexFieldfT*nbFFp]*wJ;
                stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += dftJumpdTp*NMeanl2NBetasc*Valsp[j+indexFieldfV*nbFFp]*Valsp[l+indexFieldfT*nbFFp]*wJ;
              }
            }

            //F
             //new terms for dfvI3dF +dfTI3dF
            for(int j=0;j<nbFFm;j++)
            {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int p=0;p<3;p++)
                  {
                    stiff(j+indexFieldfV*nbFFm,l+k*nbFFm) -= fvJump*NBhsNdL1dFm(k,p)*Gradsm[l+0*nbFFm][p]*Valsm[j+indexFieldfV*nbFFm]*wJ;
                    stiff(j+indexFieldfV*nbFFm,l+k*nbFFm) -= ftJump*NBhsNdL2dFm(k,p)*Gradsm[l+0*nbFFm][p]*Valsm[j+indexFieldfV*nbFFm]*wJ;

                    stiff(j+indexFieldfT*nbFFm,l+k*nbFFm) -= ftJump*NBhsNdjy1dFm(k,p)*Gradsm[l+0*nbFFm][p]*Valsm[j+indexFieldfT*nbFFm]*wJ;
                    stiff(j+indexFieldfT*nbFFm,l+k*nbFFm) -= fvJump*NBhsNdL2dFm(k,p)*Gradsm[l+0*nbFFm][p]*Valsm[j+indexFieldfT*nbFFm]*wJ;
                  }
                }

                for(int l=0;l<nbFFp;l++)
                {
                  for(int p= 0;p<3;p++)
                  {
                    stiff(j+indexFieldfV*nbFFm,l+k*nbFFp+nbdofm) -= fvJump*NBhsNdL1dFp(k,p)*Gradsp[l+0*nbFFp][p]*Valsm[j+indexFieldfV*nbFFm]*wJ;
                    stiff(j+indexFieldfV*nbFFm,l+k*nbFFp+nbdofm) -= ftJump*NBhsNdL2dFp(k,p)*Gradsp[l+0*nbFFp][p]*Valsm[j+indexFieldfV*nbFFm]*wJ;

                    stiff(j+indexFieldfT*nbFFm,l+k*nbFFp+nbdofm) -= ftJump*NBhsNdjy1dFp(k,p)*Gradsp[l+0*nbFFp][p]*Valsm[j+indexFieldfT*nbFFm]*wJ;
                    stiff(j+indexFieldfT*nbFFm,l+k*nbFFp+nbdofm) -= fvJump*NBhsNdL2dFp(k,p)*Gradsp[l+0*nbFFp][p]*Valsm[j+indexFieldfT*nbFFm]*wJ;
                  }
                }
              }
            }


            for(int j=0;j<nbFFp;j++)
            {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int p=0;p<3;p++)
                  {
                    stiff(j+indexFieldfV*nbFFp+nbdofm,l+k*nbFFm) += fvJump*NBhsNdL1dFm(k,p)*Gradsm[l+0*nbFFm][p]*Valsp[j+indexFieldfV*nbFFp]*wJ;
                    stiff(j+indexFieldfV*nbFFp+nbdofm,l+k*nbFFm) += ftJump*NBhsNdL2dFm(k,p)*Gradsm[l+0*nbFFm][p]*Valsp[j+indexFieldfV*nbFFp]*wJ;

                    stiff(j+indexFieldfT*nbFFp+nbdofm,l+k*nbFFm) += ftJump*NBhsNdjy1dFm(k,p)*Gradsm[l+0*nbFFm][p]*Valsp[j+indexFieldfT*nbFFp]*wJ;
                    stiff(j+indexFieldfT*nbFFp+nbdofm,l+k*nbFFm) += fvJump*NBhsNdL2dFm(k,p)*Gradsm[l+0*nbFFm][p]*Valsp[j+indexFieldfT*nbFFp]*wJ;
                  }
                }

                for(int l=0;l<nbFFp;l++)
                {
                  for(int p=0;p<3;p++)
                  {
                    stiff(j+indexFieldfV*nbFFp+nbdofm,l+k*nbFFp+nbdofm) +=fvJump*NBhsNdL1dFp(k,p)*Gradsp[l+0*nbFFp][p]*Valsp[j+indexFieldfV*nbFFp]*wJ;
                    stiff(j+indexFieldfV*nbFFp+nbdofm,l+k*nbFFp+nbdofm) +=ftJump*NBhsNdL2dFp(k,p)*Gradsp[l+0*nbFFp][p]*Valsp[j+indexFieldfV*nbFFp]*wJ;

                    stiff(j+indexFieldfT*nbFFp+nbdofm,l+k*nbFFp+nbdofm) +=ftJump*NBhsNdjy1dFp(k,p)*Gradsp[l+0*nbFFp][p]*Valsp[j+indexFieldfT*nbFFp]*wJ;
                    stiff(j+indexFieldfT*nbFFp+nbdofm,l+k*nbFFp+nbdofm) +=fvJump*NBhsNdL2dFp(k,p)*Gradsp[l+0*nbFFp][p]*Valsp[j+indexFieldfT*nbFFp]*wJ;
                  }
                }
              }
            }

            //dfI2tdt ( compatibility)
            // Assembly (loop on shape function)
            for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFm) += (jy1m->operator()(p,q)*Gradsm[j+indexFieldfT*nbFFm][q])*nu(p)*Valsm[l+indexFieldfT*nbFFm]*dftJumpdTm*wJ/2.;
                    stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFm) += (l20m->operator()(p,q)*Gradsm[j+indexFieldfT*nbFFm][q])*nu(p)*Valsm[l+indexFieldfT*nbFFm]*dfvJumpdTm*wJ/2.;

                    stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFm) += ftJump*(djy1dTm->operator()(p,q)*Gradsm[j+indexFieldfT*nbFFm][q])*nu(p)*Valsm[l+indexFieldfT*nbFFm]*wJ/2.;
                    stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFm) += fvJump*(dl20dTm->operator()(p,q)*Gradsm[j+indexFieldfT*nbFFm][q])*(nu(p)*Valsm[l+indexFieldfT*nbFFm])*wJ/2.;
                  }
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFp+nbdofm) += (jy1m->operator()(p,q)*Gradsm[j+indexFieldfT*nbFFm][q])*nu(p)*Valsp[l+indexFieldfT*nbFFp]*dftJumpdTp*wJ/2.;
                    stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFp+nbdofm) += (l20m->operator()(p,q)*Gradsm[j+indexFieldfT*nbFFm][q])*nu(p)*Valsp[l+indexFieldfT*nbFFp]*dfvJumpdTp*wJ/2.;
                  }
                }
              }
            }

            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFm) += (jy1p->operator()(p,q)*Gradsp[j+indexFieldfT*nbFFp][q])*nu(p)*Valsm[l+indexFieldfT*nbFFm]*dftJumpdTm*wJ/2.;
                    stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFm) += (l20p->operator()(p,q)*Gradsp[j+indexFieldfT*nbFFp][q])*nu(p)*Valsm[l+indexFieldfT*nbFFm]*dfvJumpdTm*wJ/2.;
                  }
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += (jy1p->operator()(p,q)*Gradsp[j+indexFieldfT*nbFFp][q])*nu(p)*Valsp[l+indexFieldfT*nbFFp]*dftJumpdTp*wJ/2.;
                    stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += (l20p->operator()(p,q)*Gradsp[j+indexFieldfT*nbFFp][q])*nu(p)*Valsp[l+indexFieldfT*nbFFp]*dfvJumpdTp*wJ/2.;

                    stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += ftJump*(djy1dTp->operator()(p,q)*Gradsp[j+indexFieldfT*nbFFp][q])*nu(p)*Valsp[l+indexFieldfT*nbFFp]*wJ/2.;
                    stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += fvJump*(dl20dTp->operator()(p,q)*Gradsp[j+indexFieldfT*nbFFp][q])*(nu(p)*Valsp[l+indexFieldfT*nbFFp])*wJ/2.;
                  }
                }
              }
            }
                //new related to dialetation


            for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int k =0; k < 3; k++)
                {
                  for(int p =0; p < 3; p++)
                  {
                    stiff(j+indexFieldfT*nbFFm,l+indexFieldfT*nbFFm) += gamma*(jump(k)*(Stiff_alphadialitationm(k,p)*(2./Tm)*Valsm[j+indexFieldfT*nbFFm])*nu(p)*Valsm[l+indexFieldfT*nbFFm]*wJ/2.*(-1.)/eqRatio);
                  }
                }
              }
            }

            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<nbFFp;l++)
              {
                for(int k =0; k < 3; k++)
                {
                  for(int p =0; p< 3; p++)
                  {

                    stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += gamma*(jump(k)*(Stiff_alphadialitationp(k,p)*(2./Tp)*Valsp[j+indexFieldfT*nbFFp])*nu(p)*Valsp[l+indexFieldfT*nbFFp]*wJ/2.*(-1.)/eqRatio);
                  }
                }
              }
            }
            for(int j=0;j<nbFFm;j++)
            {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int p =0; p < 3; p++)
                  {
                      stiff(j+indexFieldfT*nbFFm,l+k*nbFFm) -= gamma*((Stiff_alphadialitationm(k,p)*Valsm[j+indexFieldfT*nbFFm])*(nu(p)*wJ/2.*Valsm[l+0*nbFFm])/eqRatio*(-1.));
                  }
                }
                for(int l=0;l<nbFFp;l++)
                {
                  for(int p =0; p < 3; p++)
                  {
                      stiff(j+indexFieldfT*nbFFm,l+k*nbFFp+nbdofm) +=  gamma*((Stiff_alphadialitationm(k,p)*Valsm[j+indexFieldfT*nbFFm])*(nu(p)*wJ/2.*Valsp[l+0*nbFFp])/eqRatio*(-1.));
                  }
                }
              }
            }

            for(int j=0;j<nbFFp;j++)
            {
              for(int k=0;k<3;k++)
              {
                for(int l=0;l<nbFFm;l++)
                {
                  for(int p =0; p < 3; p++)
                  {
                    stiff(j+indexFieldfT*nbFFp+nbdofm,l+k*nbFFm) -=  gamma*((Stiff_alphadialitationp(k,p)*Valsp[j+indexFieldfT*nbFFp])*(nu(p)*wJ/2.*Valsm[l+0*nbFFm])/eqRatio*(-1.));
                  }
                }
                for(int l=0;l<nbFFp;l++)
                {
                  for(int p =0; p < 3; p++)
                  {
                      stiff(j+indexFieldfT*nbFFp+nbdofm,l+k*nbFFp+nbdofm) +=  (gamma*(Stiff_alphadialitationp(k,p)*Valsp[j+indexFieldfT*nbFFp])*(nu(p)*wJ/2.*Valsp[l+0*nbFFp])/eqRatio*(-1.));
                  }
                }
              }
            }
            //dfI2tdv
           for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFm) += (l20m->operator()(p,q)*Gradsm[j+indexFieldfT*nbFFm][q])*nu(p)*Valsm[l+indexFieldfV*nbFFm]*dfvJumpdvm*wJ/2.;

                    stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFm) += ftJump*(djy1dVm->operator()(p,q)*Gradsm[j+indexFieldfT*nbFFm][q])*nu(p)*Valsm[l+indexFieldfV*nbFFm]*wJ/2.;
                    stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFm) += fvJump*(dl20dvm->operator()(p,q)*Gradsm[j+indexFieldfT*nbFFm][q])*nu(p)*Valsm[l+indexFieldfV*nbFFm]*wJ/2.;
                  }
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+indexFieldfT*nbFFm,l+indexFieldfV*nbFFp+nbdofm) += (l20m->operator()(p,q)*Gradsm[j+indexFieldfT*nbFFm][q])*nu(p)*Valsp[l+indexFieldfV*nbFFp]*dfvJumpdvp*wJ/2.;
                  }
                }
              }
            }
            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFm) += (l20p->operator()(p,q)*Gradsp[j+indexFieldfT*nbFFp][q])*(nu(p)*Valsm[l+indexFieldfV*nbFFm])*dfvJumpdvm*wJ/2.;
                  }
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += (l20p->operator()(p,q)*Gradsp[j+indexFieldfT*nbFFp][q])*nu(p)*Valsp[l+indexFieldfV*nbFFp]*dfvJumpdvp*wJ/2.;

                    stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += ftJump*(djy1dVp->operator()(p,q)*Gradsp[j+indexFieldfT*nbFFp][q])*nu(p)*Valsp[l+indexFieldfV*nbFFp]*wJ/2.;
                    stiff(j+indexFieldfT*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += fvJump*(dl20dvp->operator()(p,q)*Gradsp[j+indexFieldfT*nbFFp][q])*nu(p)*Valsp[l+indexFieldfV*nbFFp]*wJ/2.;
                  }
                }
              }
            }

            //dfI2vdv (compatibility)
            for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+indexFieldfV*nbFFm,l+indexFieldfV*nbFFm) += (l10m->operator()(p,q)*Gradsm[j+indexFieldfV*nbFFm][q])*nu(p)*Valsm[l+indexFieldfV*nbFFm]*dfvJumpdvm*wJ/2.;
                    stiff(j+indexFieldfV*nbFFm,l+indexFieldfV*nbFFm) += ftJump*(dl20dvm->operator()(p,q)*Gradsm[j+indexFieldfV*nbFFm][q])*nu(p)*Valsm[l+indexFieldfV*nbFFm]*wJ/2.;
                  }
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+indexFieldfV*nbFFm,l+indexFieldfV*nbFFp+nbdofm) += (l10m->operator()(p,q)*Gradsm[j+indexFieldfV*nbFFm][q])*nu(p)*Valsp[l+indexFieldfV*nbFFp]*dfvJumpdvp*wJ/2.;
                  }
                }
              }
            }
            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfV*nbFFm) += (l10p->operator()(p,q)*Gradsp[j+indexFieldfV*nbFFp][q])*(nu(p)*Valsm[l+indexFieldfV*nbFFm])*dfvJumpdvm*wJ/2.;
                  }
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += (l10p->operator()(p,q)*Gradsp[j+indexFieldfV*nbFFp][q])*nu(p)*Valsp[l+indexFieldfV*nbFFp]*dfvJumpdvp*wJ/2.;

                    stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfV*nbFFp+nbdofm) += ftJump*(dl20dvp->operator()(p,q)*Gradsp[j+indexFieldfV*nbFFp][q])*nu(p)*Valsp[l+indexFieldfV*nbFFp]*wJ/2.;
                  }
                }
              }
            }

           // terms dfvI2dt
           for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFm) += (l10m->operator()(p,q)*Gradsm[j+indexFieldfV*nbFFm][q])*nu(p)*Valsm[l+indexFieldfT*nbFFm]*dfvJumpdTm*wJ/2.;
                    stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFm) += (l20m->operator()(p,q)*Gradsm[j+indexFieldfV*nbFFm][q])*nu(p)*Valsm[l+indexFieldfT*nbFFm]*dftJumpdTm*wJ/2.;

                    stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFm) += fvJump*(dl10dTm->operator()(p,q)*Gradsm[j+indexFieldfV*nbFFm][q])*nu(p)*Valsm[l+indexFieldfT*nbFFm]*wJ/2.;
                    stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFm) += ftJump*(dl20dTm->operator()(p,q)*Gradsm[j+indexFieldfV*nbFFm][q])*nu(p)*Valsm[l+indexFieldfT*nbFFm]*wJ/2.;

                  }
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFp+nbdofm) += (l10m->operator()(p,q)*Gradsm[j+indexFieldfV*nbFFm][q])*nu(p)*Valsp[l+indexFieldfT*nbFFp]*dfvJumpdTp*wJ/2.;
                    stiff(j+indexFieldfV*nbFFm,l+indexFieldfT*nbFFp+nbdofm) += (l20m->operator()(p,q)*Gradsm[j+indexFieldfV*nbFFm][q])*nu(p)*Valsp[l+indexFieldfT*nbFFp]*dftJumpdTp*wJ/2.;
                  }
                }
              }
            }

            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFm) += (l10p->operator()(p,q)*Gradsp[j+indexFieldfV*nbFFp][q])*nu(p)*Valsm[l+indexFieldfT*nbFFm]*dfvJumpdTm*wJ/2.;
                    stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFm) += (l20p->operator()(p,q)*Gradsp[j+indexFieldfV*nbFFp][q])*nu(p)*Valsm[l+indexFieldfT*nbFFm]*dftJumpdTm*wJ/2.;
                  }
                }
              }
              for(int l=0;l<nbFFp;l++)
              {
                for(int p =0; p < 3; p++)
                {
                  for(int q =0; q < 3; q++)
                  {
                    stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += (l10p->operator()(p,q)*Gradsp[j+indexFieldfV*nbFFp][q])*nu(p)*Valsp[l+indexFieldfT*nbFFp]*dfvJumpdTp*wJ/2.;
                    stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += (l20p->operator()(p,q)*Gradsp[j+indexFieldfV*nbFFp][q])*nu(p)*Valsp[l+indexFieldfT*nbFFp]*dftJumpdTp*wJ/2.;

                    stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += fvJump*(dl10dTp->operator()(p,q)*Gradsp[j+indexFieldfV*nbFFp][q])*(nu(p)*Valsp[l+indexFieldfT*nbFFp])*wJ/2.;
                    stiff(j+indexFieldfV*nbFFp+nbdofm,l+indexFieldfT*nbFFp+nbdofm) += ftJump*(dl20dTp->operator()(p,q)*Gradsp[j+indexFieldfV*nbFFp][q])*nu(p)*Valsp[l+indexFieldfT*nbFFp]*wJ/2.;
                  }
                }
              }
            }

          //dfTI2dF+dfvI2dF
         for(int j=0;j<nbFFm;j++)
            {
              for(int l=0;l<nbFFm;l++)
              {
                for(int m =0; m < 3; m++)
                {
                  for(int p =0; p < 3; p++)
                  {
                    for(int q =0; q < 3; q++)
                    {
                      for(int n =0; n< 3; n++)
                                  {
                        stiff(j+indexFieldfT*nbFFm,l+m*nbFFm) += ftJump*(djy1dFm->operator()(p,q,m,n)*Gradsm[j+indexFieldfT*nbFFm][q]*Gradsm[l+0*nbFFm][n])*(nu(p)*wJ/2.);
                        stiff(j+indexFieldfT*nbFFm,l+m*nbFFm) += fvJump*(dl20dFm->operator()(p,q,m,n)*Gradsm[j+indexFieldfT*nbFFm][q]*Gradsm[l+0*nbFFm][n])*(nu(p)*wJ/2.);

                        stiff(j+indexFieldfV*nbFFm,l+m*nbFFm) += fvJump*(dl10dFm->operator()(p,q,m,n)*Gradsm[j+indexFieldfV*nbFFm][q]*Gradsm[l+0*nbFFm][n])*(nu(p)*wJ/2.);
                        stiff(j+indexFieldfV*nbFFm,l+m*nbFFm) += ftJump*(dl20dFm->operator()(p,q,m,n)*Gradsm[j+indexFieldfV*nbFFm][q]*Gradsm[l+0.*nbFFm][n])*(nu(p)*wJ/2.);
                      }
                    }
                  }
                }
              }
            }

            for(int j=0;j<nbFFp;j++)
            {
              for(int l=0;l<nbFFp;l++)
              {
                for(int m =0; m < 3; m++)
                {
                  for(int n =0; n< 3; n++)
                  {
                    for(int p =0; p < 3; p++)
                    {
                      for(int q =0; q < 3; q++)
                      {
                        stiff(j+indexFieldfT*nbFFp+nbdofm,l+m*nbFFp+nbdofm) += ftJump*(djy1dFp->operator()(p,q,m,n)*Gradsp[j+indexFieldfT*nbFFp][q]*Gradsp[l+0*nbFFp][n])*(nu(p)*wJ/2.);
                        stiff(j+indexFieldfT*nbFFp+nbdofm,l+m*nbFFp+nbdofm) += fvJump*(dl20dFp->operator()(p,q,m,n)*Gradsp[j+indexFieldfT*nbFFp][q]*Gradsp[l+0*nbFFp][n])*(nu(p)*wJ/2.);

                        stiff(j+indexFieldfV*nbFFp+nbdofm,l+m*nbFFp+nbdofm) += fvJump*(dl10dFp->operator()(p,q,m,n)*Gradsp[j+indexFieldfV*nbFFp][q]*Gradsp[l+0*nbFFp][n])*(nu(p)*wJ/2.);
                        stiff(j+indexFieldfV*nbFFp+nbdofm,l+m*nbFFp+nbdofm) += ftJump*(dl20dFp->operator()(p,q,m,n)*Gradsp[j+indexFieldfV*nbFFp][q]*Gradsp[l+0*nbFFp][n])*(nu(p)*wJ/2.);
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }

   //stiff.print("dginter last");

};*/



